/**
 * A Concise Dictionary of the Words in the Hebrew Bible
 *    with their Renderings in the King James Version
 * by James Strong, LL.D., S.T.D.
 * 1894
 *
 * JSON version
 * ============
 * Copyright 2010, Open Scriptures. CC-BY-SA. Derived from XML.
 * $Id$
 *
 * XML e-text version
 * ==================
 * From a source by David Instone-Brewer (www.2LetterLookup.com):
 * "The data came from several PD sources which were full of errors. 
 * I cleaned them by comparing them to each other and (where 
 * necessary) looking at the original."
 *
 * Editing, corrections and Unicode transliterations
 * by David Troidl.
 *
 * Comments or corrections:
 * open-scriptures@googlegroups.com
 */

var strongsHebrewDictionary = {"H1":{"lemma":"אָב","xlit":"ʼâb","pron":"awb","derivation":"a primitive word;","strongs_def":"father, in a literal and immediate, or figurative and remote application","kjv_def":"chief, (fore-) father(-less), [idiom] patrimony, principal. Compare names in 'Abi-'."},
"H2":{"lemma":"אַב","xlit":"ʼab","pron":"ab","derivation":"(Aramaic) corresponding to H1 (אָב)","strongs_def":"{father}","kjv_def":"father."},
"H3":{"lemma":"אֵב","xlit":"ʼêb","pron":"abe","derivation":"from the same as H24 (אָבִיב);","strongs_def":"a green plant","kjv_def":"greenness, fruit."},
"H4":{"lemma":"אֵב","xlit":"ʼêb","pron":"abe","derivation":"(Aramaic) corresponding to H3 (אֵב)","strongs_def":"{a green plant}","kjv_def":"fruit."},
"H5":{"lemma":"אֲבַגְתָא","xlit":"ʼĂbagthâʼ","pron":"ab-ag-thaw'","derivation":"of foreign origin;","strongs_def":"Abagtha, a eunuch of Xerxes","kjv_def":"Abagtha."},
"H6":{"lemma":"אָבַד","xlit":"ʼâbad","pron":"aw-bad'","derivation":"a primitive root;","strongs_def":"properly, to wander away, i.e. lose oneself; by implication to perish (causative, destroy)","kjv_def":"break, destroy(-uction), [phrase] not escape, fail, lose, (cause to, make) perish, spend, [idiom] and surely, take, be undone, [idiom] utterly, be void of, have no way to flee."},
"H7":{"lemma":"אֲבַד","xlit":"ʼăbad","pron":"ab-ad'","derivation":"(Aramaic) corresponding to H6 (אָבַד)","strongs_def":"{properly, to wander away, i.e. lose oneself; by implication to perish (causative, destroy)}","kjv_def":"destroy, perish."},
"H8":{"lemma":"אֹבֵד","xlit":"ʼôbêd","pron":"o-bade'","derivation":"active of participle of H6 (אָבַד);","strongs_def":"(concrete) wretched or (abstract) destruction","kjv_def":"perish."},
"H9":{"lemma":"אֲבֵדָה","xlit":"ʼăbêdâh","pron":"ab-ay-daw'","derivation":"from H6 (אָבַד); Compare H10 (אֲבַדֹּה).","strongs_def":"concrete, something lost; abstract, destruction, i.e. Hades","kjv_def":"lost."},
"H10":{"lemma":"אֲבַדֹּה","xlit":"ʼăbaddôh","pron":"ab-ad-do'","derivation":"the same as H9 (אֲבֵדָה), miswritten for H11 (אֲבַדּוֹן);","strongs_def":"a perishing","kjv_def":"destruction."},
"H11":{"lemma":"אֲבַדּוֹן","xlit":"ʼăbaddôwn","pron":"ab-ad-done'","derivation":"intensive from H6 (אָבַד);","strongs_def":"abstract, a perishing; concrete, Hades","kjv_def":"destruction."},
"H12":{"lemma":"אַבְדָן","xlit":"ʼabdân","pron":"ab-dawn'","derivation":"from H6 (אָבַד);","strongs_def":"a perishing","kjv_def":"destruction."},
"H13":{"lemma":"אׇבְדַן","xlit":"ʼobdan","pron":"ob-dawn'","derivation":"from H6 (אָבַד);","strongs_def":"a perishing","kjv_def":"destruction."},
"H14":{"lemma":"אָבָה","xlit":"ʼâbâh","pron":"aw-baw'","derivation":"a primitive root;","strongs_def":"to breathe after, i.e. (figuratively) to be acquiescent","kjv_def":"consent, rest content will, be willing."},
"H15":{"lemma":"אָבֶה","xlit":"ʼâbeh","pron":"aw-beh'","derivation":"from H14 (אָבָה);","strongs_def":"longing","kjv_def":"desire."},
"H16":{"lemma":"אֵבֶה","xlit":"ʼêbeh","pron":"ay-beh'","derivation":"from H14 (אָבָה) (in the sense of bending toward);","strongs_def":"the papyrus","kjv_def":"swift."},
"H17":{"lemma":"אֲבוֹי","xlit":"ʼăbôwy","pron":"ab-o'ee","derivation":"from H14 (אָבָה) (in the sense of desiring);","strongs_def":"want","kjv_def":"sorrow."},
"H18":{"lemma":"אֵבוּס","xlit":"ʼêbûwç","pron":"ay-booce'","derivation":"from H75 (אָבַס);","strongs_def":"a manger or stall","kjv_def":"crib."},
"H19":{"lemma":"אִבְחָה","xlit":"ʼibchâh","pron":"ib-khaw'","derivation":"from an unused root (apparently meaning to turn);","strongs_def":"brandishing of a sword","kjv_def":"point."},
"H20":{"lemma":"אֲבַטִּיחַ","xlit":"ʼăbaṭṭîyach","pron":"ab-at-tee'-akh","derivation":"of uncertain derivation;","strongs_def":"a melon (only plural)","kjv_def":"melon."},
"H21":{"lemma":"אֲבִי","xlit":"ʼĂbîy","pron":"ab-ee'","derivation":"from H1 (אָב); fatherly;","strongs_def":"Abi, Hezekiah's mother","kjv_def":"Abi."},
"H22":{"lemma":"אֲבִיאֵל","xlit":"ʼĂbîyʼêl","pron":"ab-ee-ale'","derivation":"from H1 (אָב) and H410 (אֵל); father (i.e. possessor) of God;","strongs_def":"Abiel, the name of two Israelites","kjv_def":"Abiel."},
"H23":{"lemma":"אֲבִיאָסָף","xlit":"ʼĂbîyʼâçâph","pron":"ab-ee-aw-sawf'","derivation":"from H1 (אָב) and H622 (אָסַף); father of gathering (i.e. gatherer);","strongs_def":"Abiasaph, an Israelite","kjv_def":"Abiasaph."},
"H24":{"lemma":"אָבִיב","xlit":"ʼâbîyb","pron":"aw-beeb'","derivation":"from an unused root (meaning to be tender);","strongs_def":"green, i.e. a young ear of grain; hence, the name of the month Abib or Nisan","kjv_def":"Abib, ear, green ears of corn (not maize)."},
"H25":{"lemma":"אֲבִי גִבְעוֹן","xlit":"ʼĂbîy Gibʻôwn","pron":"ab-ee' ghib-one'","derivation":"from H1 (אָב) and H1391 (גִּבְעוֹן); father (i.e. founder) of Gibon;","strongs_def":"Abi-Gibon, perhaps an Israelite","kjv_def":"father of Gibeon."},
"H26":{"lemma":"אֲבִיגַיִל","xlit":"ʼĂbîygayil","pron":"ab-ee-gah'-yil","derivation":"or shorter אֲבִיגַל; from H1 (אָב) and H1524 (גִּיל); father (i.e. source) of joy;","strongs_def":"Abigail or Abigal, the name of two Israelitesses","kjv_def":"Abigal."},
"H27":{"lemma":"אֲבִידָן","xlit":"ʼĂbîydân","pron":"ab-ee-dawn'","derivation":"from H1 (אָב) and H1777 (דִּין); father of judgment (i.e. judge);","strongs_def":"Abidan, an Israelite","kjv_def":"Abidan."},
"H28":{"lemma":"אֲבִידָע","xlit":"ʼĂbîydâʻ","pron":"ab-ee-daw'","derivation":"from H1 (אָב) and H3045 (יָדַע); father of knowledge (i.e. knowing);","strongs_def":"Abida, a son of Abraham by Keturah","kjv_def":"Abida, Abidah."},
"H29":{"lemma":"אֲבִיָּה","xlit":"ʼĂbîyâh","pron":"ab-ee-yaw'","derivation":"or prolonged אֲבִיָּהוּ ; from H1 (אָב) and H3050 (יָהּ); father (i.e. worshipper) of Jah;","strongs_def":"Abijah, the name of several Israelite men and two Israelitesses","kjv_def":"Abiah, Abijah."},
"H30":{"lemma":"אֲבִיהוּא","xlit":"ʼĂbîyhûwʼ","pron":"ab-ee-hoo'","derivation":"from H1 (אָב) and H1931 (הוּא); father (i.e. worshipper) of Him (i.e. God);","strongs_def":"Abihu, a son of Aaron","kjv_def":"Abihu."},
"H31":{"lemma":"אֲבִיהוּד","xlit":"ʼĂbîyhûwd","pron":"ab-ee-hood'","derivation":"from H1 (אָב) and H1935 (הוֹד); father (i.e. possessor) of renown;","strongs_def":"Abihud, the name of two Israelites","kjv_def":"Abihud."},
"H32":{"lemma":"אֲבִיהַיִל","xlit":"ʼĂbîyhayil","pron":"ab-ee-hah'-yil","derivation":"or (more correctly) אֲבִיחַיִל ; from H1 (אָב) and H2428 (חַיִל); father (i.e. possessor) of might;","strongs_def":"Abihail or Abichail, the name of three Israelites and two Israelitesses","kjv_def":"Abihail."},
"H33":{"lemma":"אֲבִי הָעֶזְרִי","xlit":"ʼĂbîy hâ-ʻEzrîy","pron":"ab-ee'-haw-ez-ree'","derivation":"from H44 (אֲבִיעֶזֶר) with the article inserted; father of the Ezrite;","strongs_def":"an Abiezrite or descendant of Abiezer","kjv_def":"Abiezrite."},
"H34":{"lemma":"אֶבְיוֹן","xlit":"ʼebyôwn","pron":"eb-yone'","derivation":"from H14 (אָבָה), in the sense of want (especially in feeling);","strongs_def":"destitute","kjv_def":"beggar, needy, poor (man)."},
"H35":{"lemma":"אֲבִיּוֹנָה","xlit":"ʼăbîyôwnâh","pron":"ab-ee-yo-naw'","derivation":"from H14 (אָבָה);","strongs_def":"provocative of desire; the caper berry (from its stimulative taste)","kjv_def":"desire."},
"H36":{"lemma":"אֲבִיטוּב","xlit":"ʼĂbîyṭûwb","pron":"ab-ee-toob'","derivation":"from H1 (אָב) and H2898 (טוּב); father of goodness (i.e. good);","strongs_def":"Abitub, an Israelite","kjv_def":"Abitub."},
"H37":{"lemma":"אֲבִיטָל","xlit":"ʼĂbîyṭâl","pron":"ab-ee-tal'","derivation":"from H1 (אָב) and H2919 (טַל); father of dew (i.e. fresh);","strongs_def":"Abital, a wife of King David","kjv_def":"Abital."},
"H38":{"lemma":"אֲבִיָּם","xlit":"ʼĂbîyâm","pron":"ab-ee-yawm'","derivation":"from H1 (אָב) and H3220 (יָם); father of (the) sea (i.e. seaman);","strongs_def":"Abijam (or Abijah), a king of Judah","kjv_def":"Abijam."},
"H39":{"lemma":"אֲבִימָאֵל","xlit":"ʼĂbîymâʼêl","pron":"ab-ee-maw-ale'","derivation":"from H1 (אָב) and an elsewhere unused (probably foreign) word; father of Mael (apparently some Arab tribe);","strongs_def":"Abimael, a son of Joktan","kjv_def":"Abimael."},
"H40":{"lemma":"אֲבִימֶלֶךְ","xlit":"ʼĂbîymelek","pron":"ab-ee-mel'-ek","derivation":"from H1 (אָב) and H4428 (מֶלֶךְ); father of (the) king;","strongs_def":"Abimelek, the name of two Philistine kings and of two Israelites","kjv_def":"Abimelech."},
"H41":{"lemma":"אֲבִינָדָב","xlit":"ʼĂbîynâdâb","pron":"ab-ee-naw-dawb'","derivation":"from H1 (אָב) and H5068 (נָדַב); father of generosity (i.e. liberal);","strongs_def":"Abinadab, the name of four Israelites","kjv_def":"Abinadab."},
"H42":{"lemma":"אֲבִינֹעַם","xlit":"ʼĂbîynôʻam","pron":"ab-ee-no'-am","derivation":"from H1 (אָב) and H5278 (נֹעַם); father of pleasantness (i.e. gracious);","strongs_def":"Abinoam, an Israelite","kjv_def":"Abinoam."},
"H43":{"lemma":"אֶבְיָסָף","xlit":"ʼEbyâçâph","pron":"eb-yaw-sawf'","derivation":"contracted from H23 (אֲבִיאָסָף);","strongs_def":"Ebjasaph, an Israelite","kjv_def":"Ebiasaph."},
"H44":{"lemma":"אֲבִיעֶזֶר","xlit":"ʼĂbîyʻezer","pron":"ab-ee-ay'-zer","derivation":"from H1 (אָב) and H5829 (עֵזֶר); father of help (i.e. helpful);","strongs_def":"Abiezer, the name of two Israelites","kjv_def":"Abiezer."},
"H45":{"lemma":"אֲבִי־עַלְבוֹן","xlit":"ʼĂbîy-ʻalbôwn","pron":"ab-ee-al-bone'","derivation":"from H1 (אָב) and an unused root of uncertain derivation; probably, father of strength (i.e. valiant);","strongs_def":"Abialbon, an Israelite","kjv_def":"Abialbon."},
"H46":{"lemma":"אָבִיר","xlit":"ʼâbîyr","pron":"aw-beer'","derivation":"from H82 (אָבַר);","strongs_def":"mighty (spoken of God)","kjv_def":"mighty (one)."},
"H47":{"lemma":"אַבִּיר","xlit":"ʼabbîyr","pron":"ab-beer'","derivation":"for H46 (אָבִיר)","strongs_def":"{mighty (spoken of God)}","kjv_def":"angel, bull, chiefest, mighty (one), stout(-hearted), strong (one), valiant."},
"H48":{"lemma":"אֲבִירָם","xlit":"ʼĂbîyrâm","pron":"ab-ee-rawm'","derivation":"from H1 (אָב) and H7311 (רוּם); father of height (i.e. lofty);","strongs_def":"Abiram, the name of two Israelites","kjv_def":"Abiram."},
"H49":{"lemma":"אֲבִישַׁג","xlit":"ʼĂbîyshag","pron":"ab-ee-shag'","derivation":"from H1 (אָב) and H7686 (שָׁגָה); father of error (i.e. blundering);","strongs_def":"Abishag, a concubine of David","kjv_def":"Abishag."},
"H50":{"lemma":"אֲבִישׁוּעַ","xlit":"ʼĂbîyshûwaʻ","pron":"ab-ee-shoo'-ah","derivation":"from H1 (אָב) and H7771 (שׁוֹעַ); father of plenty (i.e. prosperous);","strongs_def":"Abishua, the name of two Israelites","kjv_def":"Abishua."},
"H51":{"lemma":"אֲבִישׁוּר","xlit":"ʼĂbîyshûwr","pron":"ab-ee-shoor'","derivation":"from H1 (אָב) and H7791 (שׁוּר); father of (the) wall (i.e. perhaps mason);","strongs_def":"Abishur, an Israelite","kjv_def":"Abishur."},
"H52":{"lemma":"אֲבִישַׁי","xlit":"ʼĂbîyshay","pron":"ab-ee-shah'ee","derivation":"or (shorter) אַבְשַׁי ; from H1 (אָב) and H7862 (שַׁי); father of a gift (i.e. probably generous);","strongs_def":"Abishai, an Israelite","kjv_def":"Abishai."},
"H53":{"lemma":"אֲבִישָׁלוֹם","xlit":"ʼĂbîyshâlôwm","pron":"ab-ee-shaw-lome'","derivation":"or (shortened) אַבְשָׁלוֹם ; from H1 (אָב) and H7965 (שָׁלוֹם); father of peace (i.e. friendly);","strongs_def":"Abshalom, a son of David; also (the fuller form) a later Israelite","kjv_def":"Abishalom, Absalom."},
"H54":{"lemma":"אֶבְיָתָר","xlit":"ʼEbyâthâr","pron":"ab-yaw-thawr'","derivation":"contracted from H1 (אָב) and H3498 (יָתַר); father of abundance (i.e. liberal);","strongs_def":"Ebjathar, an Israelite","kjv_def":"Abiathar."},
"H55":{"lemma":"אָבַךְ","xlit":"ʼâbak","pron":"aw-bak'","derivation":"a primitive root;","strongs_def":"probably to coil upward","kjv_def":"mount up."},
"H56":{"lemma":"אָבַל","xlit":"ʼâbal","pron":"aw-bal'","derivation":"a primitive root;","strongs_def":"to bewail","kjv_def":"lament, mourn."},
"H57":{"lemma":"אָבֵל","xlit":"ʼâbêl","pron":"aw-bale'","derivation":"from H56 (אָבַל);","strongs_def":"lamenting","kjv_def":"mourn(-er, -ing)."},
"H58":{"lemma":"אָבֵל","xlit":"ʼâbêl","pron":"aw-bale'","derivation":"from an unused root (meaning to be grassy);","strongs_def":"a meadow","kjv_def":"plain. Compare also the proper names beginning with Abel-."},
"H59":{"lemma":"אָבֵל","xlit":"ʼÂbêl","pron":"aw-bale'","derivation":"from H58 (אָבֵל); a meadow;","strongs_def":"Abel, the name of two places in Palestine.","kjv_def":"Abel."},
"H60":{"lemma":"אֵבֶל","xlit":"ʼêbel","pron":"ay'-bel","derivation":"from H56 (אָבַל);","strongs_def":"lamentation","kjv_def":"mourning."},
"H61":{"lemma":"אֲבָל","xlit":"ʼăbâl","pron":"ab-awl'","derivation":"apparently from H56 (אָבַל) through the idea of negation;","strongs_def":"nay, i.e. truly or yet","kjv_def":"but, indeed, nevertheless, verily."},
"H62":{"lemma":"אָבֵל בֵּית־מֲעַכָה","xlit":"ʼÂbêl Bêyth-Măʻakâh","pron":"aw-bale' bayth ma-akaw'","derivation":"from H58 (אָבֵל) and H1004 (בַּיִת) and H4601 (מַעֲכָה); meadow of Beth-Maakah;","strongs_def":"Abel of Beth-maakah, a place in Palestine","kjv_def":"Abel-beth-maachah, Abel of Beth-maachah,"},
"H63":{"lemma":"אָבֵל הַשִּׁטִּים","xlit":"ʼÂbêl hash-Shiṭṭîym","pron":"aw-bale' hash-shit-teem'","derivation":"from H58 (אָבֵל) and the plural of H7848 (שִׁטָּה), with the article inserted; meadow of the acacias;","strongs_def":"Abel hash-Shittim, a place in Palestine","kjv_def":"Abelshittim."},
"H64":{"lemma":"אָבֵל כְּרָמִים","xlit":"ʼÂbêl Kᵉrâmîym","pron":"aw-bale' ker-aw-meem'","derivation":"from H58 (אָבֵל) and the plural of H3754 (כֶּרֶם); meadow of vineyards;","strongs_def":"Abel-Keramim, a place in Palestine","kjv_def":"plain of the vineyards."},
"H65":{"lemma":"אָבֵל מְחוֹלָה","xlit":"ʼÂbêl Mᵉchôwlâh","pron":"aw-bale' mekh-o-law'","derivation":"from H58 (אָבֵל) and H4246 (מְחֹלָה)xlit mᵉchôwlâh corrected to mᵉchôlâh; meadow of dancing;","strongs_def":"Abel-Mecholah, a place in Palestine","kjv_def":"Abel-meholah."},
"H66":{"lemma":"אַבֵל מַיִם","xlit":"ʼAbêl Mayim","pron":"aw-bale' mah'-yim","derivation":"from H58 (אָבֵל) and H4325 (מַיִם); meadow of water;","strongs_def":"Abel-Majim, a place in Palestine","kjv_def":"Abel-maim."},
"H67":{"lemma":"אָבֵל מִצְרַיִם","xlit":"ʼÂbêl Mitsrayim","pron":"aw-bale' mits-rah'-yim","derivation":"from H58 (אָבֵל) and H4714 (מִצְרַיִם); meadow of Egypt;","strongs_def":"Abel-Mitsrajim, a place in Palestine","kjv_def":"Abel-mizraim."},
"H68":{"lemma":"אֶבֶן","xlit":"ʼeben","pron":"eh'-ben","derivation":"from the root of H1129 (בָּנָה) through the meaning to build;","strongs_def":"a stone","kjv_def":"[phrase] carbuncle, [phrase] mason, [phrase] plummet, (chalk-, hail-, head-, sling-) stone(-ny), (divers) weight(-s)."},
"H69":{"lemma":"אֶבֶן","xlit":"ʼeben","pron":"eh'-ben","derivation":"(Aramaic) corresponding to H68 (אֶבֶן)","strongs_def":"{a stone}","kjv_def":"stone."},
"H70":{"lemma":"אֹבֶן","xlit":"ʼôben","pron":"o' ben","derivation":"from the same as H68 (אֶבֶן);","strongs_def":"a pair of stones (only dual); a potter's wheel or a midwife's stool (consisting alike of two horizontal disks with a support between)","kjv_def":"wheel, stool."},
"H71":{"lemma":"אֲבָנָה","xlit":"ʼĂbânâh","pron":"ab-aw-naw'","derivation":"perhaps feminine of H68 (אֶבֶן); stony;  Compare H549 (אֲמָנָה).","strongs_def":"Abanah, a river near Damascus","kjv_def":"Abana."},
"H72":{"lemma":"אֶבֶן הָעֵזֶר","xlit":"ʼEben hâ-ʻÊzer","pron":"eh'-ben haw-e'-zer","derivation":"from H68 (אֶבֶן) and H5828 (עֵזֶר) with the article inserted; stone of the help;","strongs_def":"Eben-ha-Ezer, a place in Palestine","kjv_def":"Ebenezer."},
"H73":{"lemma":"אַבְנֵט","xlit":"ʼabnêṭ","pron":"ab-nate'","derivation":"of uncertain derivation;","strongs_def":"a belt","kjv_def":"girdle."},
"H74":{"lemma":"אַבְנֵר","xlit":"ʼAbnêr","pron":"ab-nare'","derivation":"or (fully) אֲבִינֵר ; from H1 (אָב) and H5216 (נִיר); father of light (i.e. enlightening);","strongs_def":"Abner, an Israelite","kjv_def":"Abner."},
"H75":{"lemma":"אָבַס","xlit":"ʼâbaç","pron":"aw-bas'","derivation":"a primitive root;","strongs_def":"to fodder","kjv_def":"fatted, stalled."},
"H76":{"lemma":"אֲבַעְבֻּעָה","xlit":"ʼăbaʻbuʻâh","pron":"ab-ah-boo-aw'","derivation":"(by reduplication) from an unused root (meaning to belch forth);","strongs_def":"an inflammatory pustule (as eruption)","kjv_def":"blains."},
"H77":{"lemma":"אֶבֶץ","xlit":"ʼEbets","pron":"eh'-bets","derivation":"from an unused root probably meaning to gleam; conspicuous;","strongs_def":"Ebets, a place in Palestine","kjv_def":"Abez."},
"H78":{"lemma":"אִבְצָן","xlit":"ʼIbtsân","pron":"ib-tsawn'","derivation":"from the same as H76 (אֲבַעְבֻּעָה); splendid;","strongs_def":"Ibtsan, an Israelite","kjv_def":"Ibzan."},
"H79":{"lemma":"אָבַק","xlit":"ʼâbaq","pron":"aw-bak'","derivation":"a primitive root, probably to float away (as vapor), but used only as denominative from H80 (אָבָק);","strongs_def":"to bedust, i.e. grapple","kjv_def":"wrestle."},
"H80":{"lemma":"אָבָק","xlit":"ʼâbâq","pron":"aw-bawk'","derivation":"from root of H79 (אָבַק);","strongs_def":"light particles (as volatile)","kjv_def":"(small) dust, powder."},
"H81":{"lemma":"אֲבָקָה","xlit":"ʼăbâqâh","pron":"ab-aw-kaw'","derivation":"feminine of H80 (אָבָק)","strongs_def":"{light particles (as volatile)}","kjv_def":"powder."},
"H82":{"lemma":"אָבַר","xlit":"ʼâbar","pron":"aw-bar'","derivation":"a primitive root;","strongs_def":"to soar","kjv_def":"fly."},
"H83":{"lemma":"אֵבֶר","xlit":"ʼêber","pron":"ay-ber'","derivation":"from H82 (אָבַר);","strongs_def":"a pinion","kjv_def":"(long-) wing(-ed)."},
"H84":{"lemma":"אֶבְרָה","xlit":"ʼebrâh","pron":"eb-raw'","derivation":"feminine of H83 (אֵבֶר)","strongs_def":"{a pinion}","kjv_def":"feather, wing."},
"H85":{"lemma":"אַבְרָהָם","xlit":"ʼAbrâhâm","pron":"ab-raw-hawm'","derivation":"contracted from H1 (אָב) and an unused root (probably meaning to be populous); father of a multitude;","strongs_def":"Abraham, the later name of Abram","kjv_def":"Abraham."},
"H86":{"lemma":"אַבְרֵךְ","xlit":"ʼabrêk","pron":"ab-rake'","derivation":"probably an Egyptian word meaning","strongs_def":"kneel","kjv_def":"bow the knee."},
"H87":{"lemma":"אַבְרָם","xlit":"ʼAbrâm","pron":"ab-rawm'","derivation":"contracted from H48 (אֲבִירָם); high father;","strongs_def":"Abram, the original name of Abraham","kjv_def":"Abram."},
"H88":{"lemma":"אֹבֹת","xlit":"ʼôbôth","pron":"o-both'","derivation":"plural of H178 (אוֹב); water-skins;","strongs_def":"Oboth, a place in the Desert","kjv_def":"Oboth."},
"H89":{"lemma":"אָגֵא","xlit":"ʼÂgêʼ","pron":"aw-gay'","derivation":"of uncertain derivation (compare H90 (אֲגַג));","strongs_def":"Age, an Israelite","kjv_def":"Agee."},
"H90":{"lemma":"אֲגַג","xlit":"ʼĂgag","pron":"ag-ag'","derivation":"or אֲגָג; of uncertain derivation (compare H89 (אָגֵא)); flame;","strongs_def":"Agag, a title of Amalekitish kings","kjv_def":"Agag."},
"H91":{"lemma":"אֲגָגִי","xlit":"ʼĂgâgîy","pron":"ag-aw-ghee'","derivation":"patrial or patronymic from H90 (אֲגַג);","strongs_def":"an Agagite or descendent (subject) of Agag","kjv_def":"Agagite."},
"H92":{"lemma":"אֲגֻדָּה","xlit":"ʼăguddâh","pron":"ag-ood-daw'","derivation":"feminine passive participle of an unused root (meaning to bind);","strongs_def":"a band, bundle, knot, or arch","kjv_def":"bunch, burden, troop."},
"H93":{"lemma":"אֱגוֹז","xlit":"ʼĕgôwz","pron":"eg-oze'","derivation":"prob of Persian origin;","strongs_def":"a nut","kjv_def":"nut."},
"H94":{"lemma":"אָגוּר","xlit":"ʼÂgûwr","pron":"aw-goor'","derivation":"passive participle of H103 (אָגַר); gathered (i.e. received among the sages);","strongs_def":"Agur, a fanciful name for Solomon","kjv_def":"Agur."},
"H95":{"lemma":"אֲגוֹרָה","xlit":"ʼăgôwrâh","pron":"ag-o-raw'","derivation":"from the same as H94 (אָגוּר);","strongs_def":"properly, something gathered, i.e. perhaps a grain or berry; used only of a small (silver) coin","kjv_def":"piece (of) silver."},
"H96":{"lemma":"אֶגֶל","xlit":"ʼegel","pron":"eh'-ghel","derivation":"from an unused root (meaning to flow down or together as drops);","strongs_def":"a reservoir","kjv_def":"drop."},
"H97":{"lemma":"אֶגְלַיִם","xlit":"ʼEglayim","pron":"eg-lah'-yim","derivation":"dual of H96 (אֶגֶל).; a double pond;","strongs_def":"Eglajim, a place in Moab","kjv_def":"Eglaim."},
"H98":{"lemma":"אֲגַם","xlit":"ʼăgam","pron":"ag-am'","derivation":"from an unused root (meaning to collect as water);","strongs_def":"a marsh; hence a rush (as growing in swamps); hence a stockade of reeds","kjv_def":"pond, pool, standing (water)."},
"H99":{"lemma":"אָגֵם","xlit":"ʼâgêm","pron":"aw-game'","derivation":"probably from the same as H98 (אֲגַם) (in the sense of stagnant water);","strongs_def":"figuratively, sad","kjv_def":"pond."},
"H100":{"lemma":"אַגְמוֹן","xlit":"ʼagmôwn","pron":"ag-mone'","derivation":"from the same as H98 (אֲגַם); a marshy pool (others from a different root, a kettle); by implication","strongs_def":"a bulrush (as growing there); collectively a rope of bulrushes","kjv_def":"bulrush, caldron, hook, rush."},
"H101":{"lemma":"אַגָּן","xlit":"ʼaggân","pron":"ag-gawn'","derivation":"probably from H5059 (נָגַן);","strongs_def":"a bowl (as pounded out hollow)","kjv_def":"basin, cup, goblet."},
"H102":{"lemma":"אַגָּף","xlit":"ʼaggâph","pron":"ag-gawf'","derivation":"probably from H5062 (נָגַף) (through the idea of impending); a cover or heap; i.e.","strongs_def":"(only plural) wings of an army, or crowds of troops","kjv_def":"bands."},
"H103":{"lemma":"אָגַר","xlit":"ʼâgar","pron":"aw-gar'","derivation":"a primitive root;","strongs_def":"to harvest","kjv_def":"gather."},
"H104":{"lemma":"אִגְּרָא","xlit":"ʼiggᵉrâʼ","pron":"ig-er-aw'","derivation":"(Aramaic) of Persian origin;","strongs_def":"an epistle (as carried by a state courier or postman)","kjv_def":"letter."},
"H105":{"lemma":"אֲגַרְטָל","xlit":"ʼăgarṭâl","pron":"ag-ar-tawl'","derivation":"of uncertain derivation;","strongs_def":"a basin","kjv_def":"charger."},
"H106":{"lemma":"אֶגְרֹף","xlit":"ʼegrôph","pron":"eg-rofe'","derivation":"from H1640 (גָּרַף) (in the sense of grasping);","strongs_def":"the clenched hand","kjv_def":"fist."},
"H107":{"lemma":"אִגֶּרֶת","xlit":"ʼiggereth","pron":"ig-eh'-reth","derivation":"feminine of H104 (אִגְּרָא);","strongs_def":"an epistle","kjv_def":"letter."},
"H108":{"lemma":"אֵד","xlit":"ʼêd","pron":"ade","derivation":"from the same as H181 (אוּד) (in the sense of enveloping);","strongs_def":"a fog","kjv_def":"mist, vapor."},
"H109":{"lemma":"אָדַב","xlit":"ʼâdab","pron":"aw-dab'","derivation":"a primitive root;","strongs_def":"to languish","kjv_def":"grieve."},
"H110":{"lemma":"אַדְבְּאֵל","xlit":"ʼAdbᵉʼêl","pron":"ad-beh-ale'","derivation":"probably from H109 (אָדַב) (in the sense of chastisement) and H410 (אֵל); disciplined of God;","strongs_def":"Adbeel, a son of Ishmael","kjv_def":"Adbeel."},
"H111":{"lemma":"אֲדַד","xlit":"ʼĂdad","pron":"ad-ad'","derivation":"probably an orthographical variation for H2301 (חֲדַד);","strongs_def":"Adad (or Hadad), an Edomite","kjv_def":"Hadad."},
"H112":{"lemma":"אִדּוֹ","xlit":"ʼIddôw","pron":"id-do","derivation":"of uncertain derivation;","strongs_def":"Iddo, an Israelite","kjv_def":"Iddo."},
"H113":{"lemma":"אָדוֹן","xlit":"ʼâdôwn","pron":"aw-done'","derivation":"or (shortened) אָדֹן; from an unused root (meaning to rule);","strongs_def":"sovereign, i.e. controller (human or divine)","kjv_def":"lord, master, owner. Compare also names beginning with 'Adoni-'."},
"H114":{"lemma":"אַדּוֹן","xlit":"ʼAddôwn","pron":"ad-done'","derivation":"probably intensive for H113 (אָדוֹן); powerful;","strongs_def":"Addon, apparently an Israelite","kjv_def":"Addon."},
"H115":{"lemma":"אֲדוֹרַיִם","xlit":"ʼĂdôwrayim","pron":"ad-o-rah'-yim","derivation":"dual from H142 (אָדַר) (in the sense of eminence); double mound;","strongs_def":"Adorajim, a place in Palestine","kjv_def":"Adoraim."},
"H116":{"lemma":"אֱדַיִן","xlit":"ʼĕdayin","pron":"ed-ah'-yin","derivation":"(Aramaic) of uncertain derivation;","strongs_def":"then (of time)","kjv_def":"now, that time, then."},
"H117":{"lemma":"אַדִּיר","xlit":"ʼaddîyr","pron":"ad-deer'","derivation":"from H142 (אָדַר);","strongs_def":"wide or (generally) large; figuratively, powerful","kjv_def":"excellent, famous, gallant, glorious, goodly, lordly, mighty(-ier one), noble, principal, worthy."},
"H118":{"lemma":"אֲדַלְיָא","xlit":"ʼĂdalyâʼ","pron":"ad-al-yaw'","derivation":"of Persian derivation;","strongs_def":"Adalja, a son of Haman","kjv_def":"Adalia."},
"H119":{"lemma":"אָדַם","xlit":"ʼâdam","pron":"aw-dam'","derivation":"to show blood (in the face), i.e.","strongs_def":"flush or turn rosy","kjv_def":"be (dyed, made) red (ruddy)."},
"H120":{"lemma":"אָדָם","xlit":"ʼâdâm","pron":"aw-dawm'","derivation":"from H119 (אָדַם);","strongs_def":"ruddy i.e. a human being (an individual or the species, mankind, etc.)","kjv_def":"[idiom] another, [phrase] hypocrite, [phrase] common sort, [idiom] low, man (mean, of low degree), person."},
"H121":{"lemma":"אָדָם","xlit":"ʼÂdâm","pron":"aw-dawm'","derivation":"the same as H120 (אָדָם);","strongs_def":"Adam the name of the first man, also of a place in Palestine","kjv_def":"Adam."},
"H122":{"lemma":"אָדֹם","xlit":"ʼâdôm","pron":"aw-dome'","derivation":"from H119 (אָדַם);","strongs_def":"rosy","kjv_def":"red, ruddy."},
"H123":{"lemma":"אֱדֹם","xlit":"ʼĔdôm","pron":"ed-ome'","derivation":"or (fully) אֱדוֹם ; from H122 (אָדֹם); red (see Genesis 25:25);","strongs_def":"Edom, the elder twin-brother of Jacob; hence the region (Idumaea) occupied by him","kjv_def":"Edom, Edomites, Idumea."},
"H124":{"lemma":"אֹדֶם","xlit":"ʼôdem","pron":"o'-dem","derivation":"from H119 (אָדַם);","strongs_def":"redness, i.e. the ruby, garnet, or some other red gem","kjv_def":"sardius."},
"H125":{"lemma":"אֲדַמְדָּם","xlit":"ʼădamdâm","pron":"ad-am-dawm'","derivation":"reduplicated from H119 (אָדַם);","strongs_def":"reddish","kjv_def":"(somewhat) reddish."},
"H126":{"lemma":"אַדְמָה","xlit":"ʼAdmâh","pron":"ad-maw'","derivation":"contracted for H127 (אֲדָמָה); earthy;","strongs_def":"Admah, a place near the Dead Sea","kjv_def":"Admah."},
"H127":{"lemma":"אֲדָמָה","xlit":"ʼădâmâh","pron":"ad-aw-maw'","derivation":"from H119 (אָדַם);","strongs_def":"soil (from its general redness)","kjv_def":"country, earth, ground, husband(-man) (-ry), land."},
"H128":{"lemma":"אֲדָמָה","xlit":"ʼĂdâmâh","pron":"ad-aw-maw'","derivation":"the same as H127 (אֲדָמָה);","strongs_def":"Adamah, a place in Palestine","kjv_def":"Adamah."},
"H129":{"lemma":"אֲדָמִי","xlit":"ʼĂdâmîy","pron":"ad-aw-mee'","derivation":"from H127 (אֲדָמָה); earthy;","strongs_def":"Adami, a place in Palestine","kjv_def":"Adami."},
"H130":{"lemma":"אֱדֹמִי","xlit":"ʼĔdômîy","pron":"ed-o-mee'","derivation":"or (fully) אֱדוֹמִי ; patronymic from H123 (אֱדֹם); See H726 (אֲרוֹמִי).","strongs_def":"an Edomite, or descendants from (or inhabitants of) Edom","kjv_def":"Edomite. "},
"H131":{"lemma":"אֲדֻמִּים","xlit":"ʼĂdummîym","pron":"ad-oom-meem'","derivation":"plural of H121 (אָדָם); red spots;","strongs_def":"Adummim, a pass in Palestine","kjv_def":"Adummim."},
"H132":{"lemma":"אַדְמֹנִי","xlit":"ʼadmônîy","pron":"ad-mo-nee'","derivation":"or (fully) אַדְמוֹנִי ; from H119 (אָדַם);","strongs_def":"reddish (of the hair or the complexion)","kjv_def":"red, ruddy."},
"H133":{"lemma":"אַדְמָתָא","xlit":"ʼAdmâthâʼ","pron":"ad-maw-thaw'","derivation":"probably of Persian derivation;","strongs_def":"Admatha, a Persian nobleman","kjv_def":"Admatha."},
"H134":{"lemma":"אֶדֶן","xlit":"ʼeden","pron":"eh'-den","derivation":"from the same as H113 (אָדוֹן) (in the sense of strength);","strongs_def":"a basis (of a building, a column, etc.)","kjv_def":"foundation, socket."},
"H135":{"lemma":"אַדָּן","xlit":"ʼAddân","pron":"ad-dawn'","derivation":"intensive from the same as H134 (אֶדֶן); firm;","strongs_def":"Addan, an Israelite","kjv_def":"Addan."},
"H136":{"lemma":"אֲדֹנָי","xlit":"ʼĂdônây","pron":"ad-o-noy'","derivation":"an emphatic form of H113 (אָדוֹן);","strongs_def":"the Lord (used as a proper name of God only)","kjv_def":"(my) Lord."},
"H137":{"lemma":"אֲדֹנִי־בֶזֶק","xlit":"ʼĂdônîy-Bezeq","pron":"ad-o''-nee-beh'-zek","derivation":"from H113 (אָדוֹן) and H966 (בֶּזֶק); lord of Bezek;","strongs_def":"Adoni-Bezek; a Canaanitish king","kjv_def":"Adonibezek."},
"H138":{"lemma":"אֲדֹנִיָּה","xlit":"ʼĂdônîyâh","pron":"ad-o-nee-yaw'","derivation":"original (prolonged) אֲדֹנִיָּהוּ ; from H113 (אָדוֹן) and H3050 (יָהּ); lord (i.e. worshipper) of Jah;","strongs_def":"Adonijah, the name of three Israelites","kjv_def":"Adonijah."},
"H139":{"lemma":"אֲדֹנִי־צֶדֶק","xlit":"ʼĂdônîy-Tsedeq","pron":"ad-o''-nee-tseh'-dek","derivation":"from H113 (אָדוֹן) and H6664 (צֶדֶק); lord of justice;","strongs_def":"Adoni-Tsedek, a Canaanitish king","kjv_def":"Adonizedec."},
"H140":{"lemma":"אֲדֹנִיקָם","xlit":"ʼĂdônîyqâm","pron":"ad-o-nee-kawm'","derivation":"from H113 (אָדוֹן) and H6965 (קוּם); lord of rising (i.e. high);","strongs_def":"Adonikam, the name of one or two Israelites","kjv_def":"Adonikam."},
"H141":{"lemma":"אֲדֹנִירָם","xlit":"ʼĂdônîyrâm","pron":"ad-o-nee-rawm'","derivation":"from H113 (אָדוֹן) and H7311 (רוּם); lord of height;","strongs_def":"Adoniram, an Israelite","kjv_def":"Adoniram."},
"H142":{"lemma":"אָדַר","xlit":"ʼâdar","pron":"aw-dar'","derivation":"a primitive root;","strongs_def":"to expand, i.e. be great or (figuratively) magnificent","kjv_def":"(become) glorious, honourable."},
"H143":{"lemma":"אֲדָר","xlit":"ʼĂdâr","pron":"ad-awr'","derivation":"probably of foreign derivation; perhaps meaning fire;","strongs_def":"Adar, the 12th Hebrew month","kjv_def":"Adar."},
"H144":{"lemma":"אֲדָר","xlit":"ʼĂdâr","pron":"ad-awr'","derivation":"(Aramaic) corresponding to H143 (אֲדָר)","strongs_def":"{Adar, the 12th Hebrew month}","kjv_def":"Adar."},
"H145":{"lemma":"אֶדֶר","xlit":"ʼeder","pron":"eh'-der","derivation":"from H142 (אָדַר);","strongs_def":"amplitude, i.e. (concrete) a mantle; also (figuratively) splendor","kjv_def":"goodly, robe."},
"H146":{"lemma":"אַדָּר","xlit":"ʼAddâr","pron":"ad-dawr'","derivation":"intensive from H142 (אָדַר); ample;","strongs_def":"Addar, a place in Palestine; also an Israelite","kjv_def":"Addar."},
"H147":{"lemma":"אִדַּר","xlit":"ʼiddar","pron":"id-dar'","derivation":"(Aramaic) intensive, from a root corresponding to H142 (אָדַר);","strongs_def":"ample, i.e. a threshing-floor","kjv_def":"threshingfloor."},
"H148":{"lemma":"אֲדַרְגָּזֵר","xlit":"ʼădargâzêr","pron":"ad-ar''-gaw-zare'","derivation":"(Aramaic) from the same as H147 (אִדַּר), and H1505 (גְּזַר);","strongs_def":"a chief diviner, or astrologer","kjv_def":"judge."},
"H149":{"lemma":"אַדְרַזְדָּא","xlit":"ʼadrazdâʼ","pron":"ad-raz-daw'","derivation":"(Aramaic) probably of Persian origin;","strongs_def":"quickly or carefully","kjv_def":"diligently."},
"H150":{"lemma":"אֲדַרְכֹּן","xlit":"ʼădarkôn","pron":"ad-ar-kone'","derivation":"of Persian origin;","strongs_def":"a daric or Persian coin","kjv_def":"dram."},
"H151":{"lemma":"אֲדֹרָם","xlit":"ʼĂdôrâm","pron":"ad-o-rawm'","derivation":"contracted for H141 (אֲדֹנִירָם);","strongs_def":"Adoram (or Adoniram), an Israelite","kjv_def":"Adoram."},
"H152":{"lemma":"אֲדְרַמֶּלֶךְ","xlit":"ʼĂdrammelek","pron":"ad-ram-meh'-lek","derivation":"from H142 (אָדַר) and H4428 (מֶלֶךְ); splendor of (the) king;","strongs_def":"Adrammelek, the name of an Assyrian idol, also of a son of Sennacherib","kjv_def":"Adrammelech."},
"H153":{"lemma":"אֶדְרָע","xlit":"ʼedrâʻ","pron":"ed-raw'","derivation":"(Aramaic) an orthographical variation for H1872 (דְּרַע);","strongs_def":"an arm, i.e. (figuratively) power","kjv_def":"force."},
"H154":{"lemma":"אֶדְרֶעִי","xlit":"ʼedreʻîy","pron":"ed-reh'-ee","derivation":"from the equivalent of H153 (אֶדְרָע); mighty;","strongs_def":"Edrei, the name of two places in Palestine","kjv_def":"Edrei."},
"H155":{"lemma":"אַדֶּרֶת","xlit":"ʼaddereth","pron":"ad-deh'-reth","derivation":"feminine of H117 (אַדִּיר); ; also the same as H145 (אֶדֶר)","strongs_def":"something ample (as a large vine, a wide dress)","kjv_def":"garment, glory, goodly, mantle, robe."},
"H156":{"lemma":"אָדַשׁ","xlit":"ʼâdash","pron":"aw-dash'","derivation":"a primitive root;","strongs_def":"to tread out (grain)","kjv_def":"thresh."},
"H157":{"lemma":"אָהַב","xlit":"ʼâhab","pron":"aw-hab'","derivation":"or אָהֵב ; a primitive root;","strongs_def":"to have affection for (sexually or otherwise)","kjv_def":"(be-) love(-d, -ly, -r), like, friend."},
"H158":{"lemma":"אַהַב","xlit":"ʼahab","pron":"ah'-hab","derivation":"from H157 (אָהַב);","strongs_def":"affection (in a good or a bad sense)","kjv_def":"love(-r)."},
"H159":{"lemma":"אֹהַב","xlit":"ʼôhab","pron":"o'-hab","derivation":"from H156 (אָדַשׁ); meaning the same as H158 (אַהַב)","strongs_def":"{affection (in a good or a bad sense)}","kjv_def":"love."},
"H160":{"lemma":"אַהֲבָה","xlit":"ʼahăbâh","pron":"a-hab-aw","derivation":"feminine of H158 (אַהַב) and meaning the same","strongs_def":"{affection (in a good or a bad sense)}","kjv_def":"love."},
"H161":{"lemma":"אֹהַד","xlit":"ʼÔhad","pron":"o'-had","derivation":"from an unused root meaning to be united; unity;","strongs_def":"Ohad, an Israelite","kjv_def":"Ohad."},
"H162":{"lemma":"אֲהָהּ","xlit":"ʼăhâhh","pron":"a-haw'","derivation":"apparently a primitive word expressing pain exclamatorily;","strongs_def":"Oh!","kjv_def":"ah, alas."},
"H163":{"lemma":"אַהֲוָא","xlit":"ʼAhăvâʼ","pron":"a-hav-aw'","derivation":"probably of foreign origin;","strongs_def":"Ahava, a river of Babylonia","kjv_def":"Ahava."},
"H164":{"lemma":"אֵהוּד","xlit":"ʼÊhûwd","pron":"ay-hood'","derivation":"from the same as H161 (אֹהַד); united;","strongs_def":"Ehud, the name of two or three Israelites","kjv_def":"Ehud."},
"H165":{"lemma":"אֱהִי","xlit":"ʼĕhîy","pron":"e-hee'","derivation":"apparently an orthographical variation for H346 (אַיֵּה);","strongs_def":"where","kjv_def":"I will be (Hosea 13:10,14) (which is often the rendering of the same Hebrew form from H1961 (הָיָה))."},
"H166":{"lemma":"אָהַל","xlit":"ʼâhal","pron":"aw-hal'","derivation":"a primitive root;","strongs_def":"to be clear","kjv_def":"shine."},
"H167":{"lemma":"אָהַל","xlit":"ʼâhal","pron":"aw-hal'","derivation":"a denominative from H168 (אֹהֶל);","strongs_def":"to tent","kjv_def":"pitch (remove) a tent."},
"H168":{"lemma":"אֹהֶל","xlit":"ʼôhel","pron":"o'-hel","derivation":"from H166 (אָהַל);","strongs_def":"a tent (as clearly conspicuous from a distance)","kjv_def":"covering, (dwelling) (place), home, tabernacle, tent."},
"H169":{"lemma":"אֹהֶל","xlit":"ʼÔhel","pron":"o'-hel","derivation":"the same as H168 (אֹהֶל);","strongs_def":"Ohel, an Israelite","kjv_def":"Ohel."},
"H170":{"lemma":"אׇהֳלָה","xlit":"ʼOhŏlâh","pron":"o-hol-aw'","derivation":"in form a feminine of H168 (אֹהֶל), but in fact for אׇהֳלָהּ; from H168 (אֹהֶל); her tent (i.e. idolatrous sanctuary);","strongs_def":"Oholah, a symbolic name for Samaria","kjv_def":"Aholah."},
"H171":{"lemma":"אׇהֳלִיאָב","xlit":"ʼOhŏlîyʼâb","pron":"o''-hol-e-awb'","derivation":"from H168 (אֹהֶל) and H1 (אָב); tent of (his) father;","strongs_def":"Oholiab, an Israelite","kjv_def":"Aholiab."},
"H172":{"lemma":"אׇהֳלִיבָה","xlit":"ʼOhŏlîybâh","pron":"o''-hol-ee-baw'","derivation":"(similarly with H170 (אׇהֳלָה)) for אׇהֳלִיבָהּ ; from H168 (אֹהֶל); my tent (is) in her;","strongs_def":"Oholibah, a symbolic name for Judah","kjv_def":"Aholibah."},
"H173":{"lemma":"אׇהֳלִיבָמָה","xlit":"ʼOhŏlîybâmâh","pron":"o''-hol-ee-baw-maw'","derivation":"from H168 (אֹהֶל) and H1116 (בָּמָה); tent of (the) height;","strongs_def":"Oholibamah, a wife of Esau","kjv_def":"Aholibamah."},
"H174":{"lemma":"אֲהָלִים","xlit":"ʼăhâlîym","pron":"a-haw-leem'","derivation":"or (feminine) אֲהָלוֹת (only used thus in the plural); of foreign origin;","strongs_def":"aloe wood (i.e. sticks)","kjv_def":"(tree of lign-) aloes."},
"H175":{"lemma":"אַהֲרוֹן","xlit":"ʼAhărôwn","pron":"a-har-one'","derivation":"of uncertain derivation;","strongs_def":"Aharon, the brother of Moses","kjv_def":"Aaron."},
"H176":{"lemma":"אוֹ","xlit":"ʼôw","pron":"o","derivation":"presumed to be the 'constructive' or genitival form of אַו ; short for H185 (אַוָּה);","strongs_def":"desire (and so probably in Proverbs 31:4); hence (by way of alternative) or, also if","kjv_def":"also, and, either, if, at the least, [idiom] nor, or, otherwise, then, whether."},
"H177":{"lemma":"אוּאֵל","xlit":"ʼÛwʼêl","pron":"oo-ale'","derivation":"from H176 (אוֹ) and H410 (אֵל); wish of God;","strongs_def":"Uel, and Israelite","kjv_def":"Uel."},
"H178":{"lemma":"אוֹב","xlit":"ʼôwb","pron":"obe","derivation":"from the same as H1 (אָב) (apparently through the idea of prattling a father's name);","strongs_def":"properly, a mumble, i.e. a water skin (from its hollow sound); hence a necromancer (ventriloquist, as from a jar)","kjv_def":"bottle, familiar spirit."},
"H179":{"lemma":"אוֹבִיל","xlit":"ʼôwbîyl","pron":"o-beel'","derivation":"probably from H56 (אָבַל); mournful;","strongs_def":"Obil, an Ishmaelite","kjv_def":"Obil."},
"H180":{"lemma":"אוּבָל","xlit":"ʼûwbâl","pron":"oo-bawl'","derivation":"or (shortened) אֻבָל ; from H2986 (יָבַל) (in the sense of H2988 (יָבָל));","strongs_def":"a stream","kjv_def":"river."},
"H181":{"lemma":"אוּד","xlit":"ʼûwd","pron":"ood","derivation":"from an unused root meaning to rake together;","strongs_def":"a poker (for turning or gathering embers)","kjv_def":"(fire-) brand."},
"H182":{"lemma":"אוֹדוֹת","xlit":"ʼôwdôwth","pron":"o-doth'","derivation":"or (shortened) אֹדוֹת (only thus in the plural); from the same as H181 (אוּד);","strongs_def":"turnings (i.e. occasions); (adverb) on account of","kjv_def":"(be-) cause, concerning, sake."},
"H183":{"lemma":"אָוָה","xlit":"ʼâvâh","pron":"aw-vaw'","derivation":"a primitive root;","strongs_def":"to wish for","kjv_def":"covet, (greatly) desire, be desirous, long, lust (after)."},
"H184":{"lemma":"אָוָה","xlit":"ʼâvâh","pron":"aw-vaw'","derivation":"a primitive root;","strongs_def":"to extend or mark out","kjv_def":"point out."},
"H185":{"lemma":"אַוָּה","xlit":"ʼavvâh","pron":"av-vaw'","derivation":"from H183 (אָוָה);","strongs_def":"longing","kjv_def":"desire, lust after, pleasure."},
"H186":{"lemma":"אוּזַי","xlit":"ʼÛwzay","pron":"oo-zah'-ee","derivation":"perhaps by permutation for H5813 (עֻזִּי), strong;","strongs_def":"Uzai, an Israelite","kjv_def":"Uzai."},
"H187":{"lemma":"אוּזָל","xlit":"ʼÛwzâl","pron":"oo-zawl'","derivation":"of uncertain derivation;","strongs_def":"Uzal, a son of Joktan","kjv_def":"Uzal."},
"H188":{"lemma":"אוֹי","xlit":"ʼôwy","pron":"o'-ee","derivation":"probably from H183 (אָוָה) (in the sense of crying out after);","strongs_def":"lamentation; also interjectionally Oh!","kjv_def":"alas, woe."},
"H189":{"lemma":"אֱוִי","xlit":"ʼĔvîy","pron":"ev-ee'","derivation":"probably from H183 (אָוָה); desirous;","strongs_def":"Evi, a Midianitish chief","kjv_def":"Evi."},
"H190":{"lemma":"אוֹיָה","xlit":"ʼôwyâh","pron":"o-yaw'","derivation":"feminine of H188 (אוֹי)","strongs_def":"{lamentation; also interjectionally Oh!}","kjv_def":"woe."},
"H191":{"lemma":"אֱוִיל","xlit":"ʼĕvîyl","pron":"ev-eel'","derivation":"from an unused root (meaning to be perverse);","strongs_def":"(figuratively) silly","kjv_def":"fool(-ish) (man)."},
"H192":{"lemma":"אֱוִיל מְרֹדַךְ","xlit":"ʼĔvîyl Mᵉrôdak","pron":"ev-eel' mer-o-dak'","derivation":"of Aramaic derivation and probably meaning soldier of Merodak;","strongs_def":"Evil-Merodak, a Babylonian king","kjv_def":"Evil-merodach."},
"H193":{"lemma":"אוּל","xlit":"ʼûwl","pron":"ool","derivation":"from an unused root meaning to twist, i.e. (by implication) be strong;","strongs_def":"the body (as being rolled together); also powerful","kjv_def":"mighty, strength."},
"H194":{"lemma":"אוּלַי","xlit":"ʼûwlay","pron":"oo-lah'ee","derivation":"or (shortened) אֻלַי; from H176 (אוֹ);","strongs_def":"if not; hence perhaps","kjv_def":"if so be, may be, peradventure, unless."},
"H195":{"lemma":"אוּלַי","xlit":"ʼÛwlay","pron":"oo-lah'ee","derivation":"of Persian derivation;","strongs_def":"the Ulai (or Eulaeus), a river of Persia","kjv_def":"Ulai."},
"H196":{"lemma":"אֱוִלִי","xlit":"ʼĕvilîy","pron":"ev-ee-lee'","derivation":"from H191 (אֱוִיל);","strongs_def":"silly, foolish; hence (morally) impious","kjv_def":"foolish."},
"H197":{"lemma":"אוּלָם","xlit":"ʼûwlâm","pron":"oo-lawm'","derivation":"or (shortened), אֻלָם ; from H481 (אָלַם) (in the sense of tying);","strongs_def":"a vestibule (as bound to the building)","kjv_def":"porch."},
"H198":{"lemma":"אוּלָם","xlit":"ʼÛwlâm","pron":"oo-lawm'","derivation":"appar, from H481 (אָלַם) (in the sense of dumbness); solitary;","strongs_def":"Ulam, the name of two Israelites","kjv_def":"Ulam."},
"H199":{"lemma":"אוּלָם","xlit":"ʼûwlâm","pron":"oo-lawm'","derivation":"apparently a variation of H194 (אוּלַי);","strongs_def":"however or on the contrary","kjv_def":"as for, but, howbeit, in very deed, surely, truly, wherefore."},
"H200":{"lemma":"אִוֶּלֶת","xlit":"ʼivveleth","pron":"iv-veh'-leth","derivation":"from the same as H191 (אֱוִיל);","strongs_def":"silliness","kjv_def":"folly, foolishly(-ness)."},
"H201":{"lemma":"אוֹמָר","xlit":"ʼÔwmâr","pron":"o-mawr'","derivation":"from H559 (אָמַר); talkative;","strongs_def":"Omar, a grandson of Esau","kjv_def":"Omar."},
"H202":{"lemma":"אוֹן","xlit":"ʼôwn","pron":"one","derivation":"probably from the same as H205 (אָוֶן) (in the sense of effort, but successful);","strongs_def":"ability, power, (figuratively) wealth","kjv_def":"force, goods, might, strength, substance."},
"H203":{"lemma":"אוֹן","xlit":"ʼÔwn","pron":"one","derivation":"the same as H202 (אוֹן);","strongs_def":"On, an Israelite","kjv_def":"On."},
"H204":{"lemma":"אוֹן","xlit":"ʼÔwn","pron":"one","derivation":"or (shortened); אֹן; of Egyptian derivation;","strongs_def":"On, a city of Egypt","kjv_def":"On."},
"H205":{"lemma":"אָוֶן","xlit":"ʼâven","pron":"aw-ven'","derivation":"from an unused root perhaps meaning properly, to pant (hence, to exert oneself, usually in vain; to come to naught);","strongs_def":"strictly nothingness; also trouble. vanity, wickedness; specifically an idol","kjv_def":"affliction, evil, false, idol, iniquity, mischief, mourners(-ing), naught, sorrow, unjust, unrighteous, vain, vanity, wicked(-ness). Compare H369 (אַיִן)."},
"H206":{"lemma":"אָוֶן","xlit":"ʼÂven","pron":"aw'-ven","derivation":"the same as H205 (אָוֶן); idolatry;","strongs_def":"Aven, the contemptuous synonym of three places, one in Coele-Syria, one in Egypt (On), and one in Palestine (Bethel)","kjv_def":"Aven. See also H204 (אוֹן), H1007 (בֵּית אָוֶן)."},
"H207":{"lemma":"אוֹנוֹ","xlit":"ʼÔwnôw","pron":"o-no'","derivation":"or (shortened) אֹנוֹ; prolonged from H202 (אוֹן); strong;","strongs_def":"Ono, a place in Palestine","kjv_def":"Ono."},
"H208":{"lemma":"אוֹנָם","xlit":"ʼÔwnâm","pron":"o-nawm'","derivation":"a variation of H209 (אוֹנָן); strong;","strongs_def":"Onam, the name of an Edomite and of an Israelite","kjv_def":"Onam."},
"H209":{"lemma":"אוֹנָן","xlit":"ʼÔwnân","pron":"o-nawn'","derivation":"a variation of H207 (אוֹנוֹ); strong;","strongs_def":"Onan, a son of Judah","kjv_def":"Onan."},
"H210":{"lemma":"אוּפָז","xlit":"ʼÛwphâz","pron":"oo-fawz'","derivation":"perhaps a corruption of H211 (אוֹפִיר);","strongs_def":"Uphaz, a famous gold region","kjv_def":"Uphaz."},
"H211":{"lemma":"אוֹפִיר","xlit":"ʼÔwphîyr","pron":"o-feer'","derivation":"or (shortened) אֹפִיר; and אוֹפִר; of uncertain derivation;","strongs_def":"Ophir, the name of a son of Joktan, and of a gold region in the East","kjv_def":"Ophir."},
"H212":{"lemma":"אוֹפָן","xlit":"ʼôwphân","pron":"o-fawn'","derivation":"or (shortened) אֹפָן; from an unused root meaning to revolve;","strongs_def":"a wheel","kjv_def":"wheel."},
"H213":{"lemma":"אוּץ","xlit":"ʼûwts","pron":"oots","derivation":"a primitive root;","strongs_def":"to press; (by implication) to be close, hurry, withdraw","kjv_def":"(make) haste(-n, -y), labor, be narrow."},
"H214":{"lemma":"אוֹצָר","xlit":"ʼôwtsâr","pron":"o-tsaw'","derivation":"from H686 (אָצַר);","strongs_def":"a depository","kjv_def":"armory, cellar, garner, store(-house), treasure(-house) (-y)."},
"H215":{"lemma":"אוֹר","xlit":"ʼôwr","pron":"ore","derivation":"a primitive root;","strongs_def":"to be (causative, make) luminous (literally and metaphorically)","kjv_def":"[idiom] break of day, glorious, kindle, (be, en-, give, show) light (-en, -ened), set on fire, shine."},
"H216":{"lemma":"אוֹר","xlit":"ʼôwr","pron":"ore","derivation":"from H215 (אוֹר);","strongs_def":"illumination or (concrete) luminary (in every sense, including lightning, happiness, etc.)","kjv_def":"bright, clear, [phrase] day, light (-ning), morning, sun."},
"H217":{"lemma":"אוּר","xlit":"ʼûwr","pron":"ore","derivation":"from H215 (אוֹר);","strongs_def":"flame; hence (in the plural) the East (as being the region of light)","kjv_def":"fire, light. See also H224 (אוּרִים)."},
"H218":{"lemma":"אוּר","xlit":"ʼÛwr","pron":"oor","derivation":"the same as H217 (אוּר);","strongs_def":"Ur, a place in Chaldaea; also an Israelite","kjv_def":"Ur."},
"H219":{"lemma":"אוֹרָה","xlit":"ʼôwrâh","pron":"o-raw'","derivation":"feminine of H216 (אוֹר);","strongs_def":"luminousness, i.e. (figuratively) prosperity; also a plant (as being bright)","kjv_def":"herb, light."},
"H220":{"lemma":"אֲוֵרָה","xlit":"ʼăvêrâh","pron":"av-ay-raw'","derivation":"by transposition for H723 (אֻרְוָה);","strongs_def":"a stall","kjv_def":"cote."},
"H221":{"lemma":"אוּרִי","xlit":"ʼÛwrîy","pron":"oo-ree'","derivation":"from H217 (אוּר); fiery;","strongs_def":"Uri, the name of three Israelites","kjv_def":"Uri."},
"H222":{"lemma":"אוּרִיאֵל","xlit":"ʼÛwrîyʼêl","pron":"oo-ree-ale'","derivation":"from H217 (אוּר) and H410 (אֵל); flame of God;","strongs_def":"Uriel, the name of two Israelites","kjv_def":"Uriel."},
"H223":{"lemma":"אוּרִיָּה","xlit":"ʼÛwrîyâh","pron":"oo-ree-yaw'","derivation":"or (prolonged) אוּרִיָּהוּ ; from H217 (אוּר) and H3050 (יָהּ); flame of Jah;","strongs_def":"Urijah, the name of one Hittite and five Israelites","kjv_def":"Uriah, Urijah."},
"H224":{"lemma":"אוּרִים","xlit":"ʼÛwrîym","pron":"oo-reem'","derivation":"plural of H217 (אוּר); lights;","strongs_def":"Urim, the oracular brilliancy of the figures in the high-priest's breastplate","kjv_def":"Urim."},
"H225":{"lemma":"אוּת","xlit":"ʼûwth","pron":"ooth","derivation":"a primitive root;","strongs_def":"properly, to come, i.e. (implied) to assent","kjv_def":"consent."},
"H226":{"lemma":"אוֹת","xlit":"ʼôwth","pron":"oth","derivation":"probably from H225 (אוּת) (in the sense of appearing);","strongs_def":"a signal (literally or figuratively), as aflag, beacon, monument, omen, prodigy, evidence, etc.","kjv_def":"mark, miracle, (en-) sign, token."},
"H227":{"lemma":"אָז","xlit":"ʼâz","pron":"awz","derivation":"a demonstrative adverb;","strongs_def":"at that time or place; also as a conjunction, therefore","kjv_def":"beginning, for, from, hitherto, now, of old, once, since, then, at which time, yet."},
"H228":{"lemma":"אֲזָא","xlit":"ʼăzâʼ","pron":"az-zaw'","derivation":"(Aramaic) or אֲזָה (Aramaic);","strongs_def":"to kindle; (by implication) to heat","kjv_def":"heat, hot."},
"H229":{"lemma":"אֶזְבַּי","xlit":"ʼEzbay","pron":"ez-bah'ee","derivation":"probably from H231 (אֵזוֹב); hyssop-like;","strongs_def":"Ezbai, an Israelite","kjv_def":"Ezbai."},
"H230":{"lemma":"אֲזָד","xlit":"ʼăzâd","pron":"az-zawd'","derivation":"(Aramaic) of uncertain derivation;","strongs_def":"firm","kjv_def":"be gone."},
"H231":{"lemma":"אֵזוֹב","xlit":"ʼêzôwb","pron":"ay-zobe'","derivation":"probably of foreign derivation;","strongs_def":"hyssop","kjv_def":"hyssop."},
"H232":{"lemma":"אֵזוֹר","xlit":"ʼêzôwr","pron":"ay-zore'","derivation":"from H246 (אֲזִקִּים);","strongs_def":"something girt; a belt, also a band","kjv_def":"girdle."},
"H233":{"lemma":"אֲזַי","xlit":"ʼăzay","pron":"az-ah'ee","derivation":"probably from H227 (אָז);","strongs_def":"at that time","kjv_def":"then."},
"H234":{"lemma":"אַזְכָּרָה","xlit":"ʼazkârâh","pron":"az-kaw-raw'","derivation":"from H2142 (זָכַר);","strongs_def":"a reminder; specifically remembrance-offering","kjv_def":"memorial."},
"H235":{"lemma":"אָזַל","xlit":"ʼâzal","pron":"aw-zal'","derivation":"a primitive root;","strongs_def":"to go away, hence, to disappear","kjv_def":"fail, gad about, go to and fro (but in Ezekiel 27:19 the word is rendered by many 'from Uzal,' by others 'yarn'), be gone (spent)."},
"H236":{"lemma":"אֲזַל","xlit":"ʼăzal","pron":"az-al'","derivation":"(Aramaic) the same as H235 (אָזַל);","strongs_def":"to depart","kjv_def":"go (up)."},
"H237":{"lemma":"אֶזֶל","xlit":"ʼezel","pron":"eh'-zel","derivation":"from H235 (אָזַל); departure;","strongs_def":"Ezel, a memorial stone in Palestine","kjv_def":"Ezel."},
"H238":{"lemma":"אָזַן","xlit":"ʼâzan","pron":"aw-zan'","derivation":"a primitive root; probably to expand; but used only as a denominative from H241 (אֹזֶן);","strongs_def":"to broaden out the ear (with the hand), i.e. (by implication) to listen","kjv_def":"give (perceive by the) ear, hear(-ken). See H239 (אָזַן)."},
"H239":{"lemma":"אָזַן","xlit":"ʼâzan","pron":"aw-zan'","derivation":"a primitive root (rather identical with H238 (אָזַן) through the idea of scales as if two ears);","strongs_def":"to weigh, i.e. (figuratively) ponder","kjv_def":"give good heed."},
"H240":{"lemma":"אָזֵן","xlit":"ʼâzên","pron":"aw-zane'","derivation":"from H238 (אָזַן);","strongs_def":"a spade or paddle (as having a broad end)","kjv_def":"weapon."},
"H241":{"lemma":"אֹזֶן","xlit":"ʼôzen","pron":"o'-zen","derivation":"from H238 (אָזַן);","strongs_def":"broadness. i.e. (concrete) the ear (from its form in man)","kjv_def":"[phrase] advertise, audience, [phrase] displease, ear, hearing, [phrase] show."},
"H242":{"lemma":"אֻזֵּן שֶׁאֱרָה","xlit":"ʼUzzên Sheʼĕrâh","pron":"ooz-zane' sheh-er-aw'","derivation":"from H238 (אָזַן) and H7609 (שֶׁאֱרָה); plot of Sheerah (i.e. settled by him);","strongs_def":"Uzzen-Sheerah, a place in Palestine","kjv_def":"Uzzen-sherah."},
"H243":{"lemma":"אַזְנוֹת תָּבוֹר","xlit":"ʼAznôwth Tâbôwr","pron":"az-noth' taw-bore'","derivation":"from H238 (אָזַן) and H8396 (תָּבוֹר); flats (i.e. tops) of Tabor (i.e. situated on it);","strongs_def":"Aznoth-Tabor, a place in Palestine","kjv_def":"Aznoth-tabor."},
"H244":{"lemma":"אׇזְנִי","xlit":"ʼOznîy","pron":"oz-nee'","derivation":"from H241 (אֹזֶן); having (quick) ears;","strongs_def":"Ozni, an Israelite; also an Oznite (collectively), his descendant","kjv_def":"Ozni, Oznites."},
"H245":{"lemma":"אֲזַנְיָה","xlit":"ʼĂzanyâh","pron":"az-an-yaw'","derivation":"from H238 (אָזַן) and H3050 (יָהּ); heard by Jah;","strongs_def":"Azanjah, an Israelite","kjv_def":"Azaniah."},
"H246":{"lemma":"אֲזִקִּים","xlit":"ʼăziqqîym","pron":"az-ik-keem'","derivation":"a variation for H2131 (זִיקָה);","strongs_def":"manacles","kjv_def":"chains."},
"H247":{"lemma":"אָזַר","xlit":"ʼâzar","pron":"aw-zar'","derivation":"a primitive root;","strongs_def":"to belt","kjv_def":"bind (compass) about, gird (up, with)."},
"H248":{"lemma":"אֶזְרוֹעַ","xlit":"ʼezrôwaʻ","pron":"ez-ro'-a","derivation":"a variation for H2220 (זְרוֹעַ);","strongs_def":"the arm","kjv_def":"arm."},
"H249":{"lemma":"אֶזְרָח","xlit":"ʼezrâch","pron":"ez-rawkh'","derivation":"from H2224 (זָרַח) (in the sense of springing up);","strongs_def":"a spontaneous growth, i.e. native (tree or persons)","kjv_def":"bay tree, (home-) born (in the land), of the (one's own) country (nation)."},
"H250":{"lemma":"אֶזְרָחִי","xlit":"ʼEzrâchîy","pron":"ez-raw-khee'","derivation":"patronymic from H2246 (חֹבָב);","strongs_def":"an Ezrachite or descendant of Zerach","kjv_def":"Ezrahite."},
"H251":{"lemma":"אָח","xlit":"ʼâch","pron":"awkh","derivation":"a primitive word;","strongs_def":"a brother (used in the widest sense of literal relationship and metaphorical affinity or resemblance (like father))","kjv_def":"another, brother(-ly); kindred, like, other. Compare also the proper names beginning with 'Ah-' or 'Ahi-'."},
"H252":{"lemma":"אַח","xlit":"ʼach","pron":"akh","derivation":"(Aramaic) corresponding to H251 (אָח)","strongs_def":"{a brother (used in the widest sense of literal relationship and metaphorical affinity or resemblance)}","kjv_def":"brother."},
"H253":{"lemma":"אָח","xlit":"ʼâch","pron":"awkh","derivation":"a variation for H162 (אֲהָהּ);","strongs_def":"Oh! (expressive of grief or surprise)","kjv_def":"ah, alas."},
"H254":{"lemma":"אָח","xlit":"ʼâch","pron":"awkh","derivation":"of uncertain derivation;","strongs_def":"a fire-pot or chafing dish","kjv_def":"hearth."},
"H255":{"lemma":"אֹחַ","xlit":"ʼôach","pron":"o'-akh","derivation":"probably from H253 (אָח);","strongs_def":"a howler or lonesome wild animal","kjv_def":"doleful creature."},
"H256":{"lemma":"אַחְאָב","xlit":"ʼAchʼâb","pron":"akh-awb'","derivation":"once (by contraction) אֶחָב (Jeremiah 29:22); from H251 (אָח) and H1 (אָב); brother (i.e. friend) of (his) father;","strongs_def":"Achab, the name of a king of Israel and of a prophet at Babylon","kjv_def":"Ahab."},
"H257":{"lemma":"אַחְבָן","xlit":"ʼAchbân","pron":"akh-bawn'","derivation":"from H251 (אָח) and H995 (בִּין); brother (i.e. possessor) of understanding;","strongs_def":"Achban, an Israelite","kjv_def":"Ahban."},
"H258":{"lemma":"אָחַד","xlit":"ʼâchad","pron":"aw-khad'","derivation":"perhaps a primitive root;","strongs_def":"to unify, i.e. (figuratively) collect (one's thoughts)","kjv_def":"go one way or other."},
"H259":{"lemma":"אֶחָד","xlit":"ʼechâd","pron":"ekh-awd'","derivation":"a numeral from H258 (אָחַד);","strongs_def":"properly, united, i.e. one; or (as an ordinal) first","kjv_def":"a, alike, alone, altogether, and, any(-thing), apiece, a certain, (dai-) ly, each (one), [phrase] eleven, every, few, first, [phrase] highway, a man, once, one, only, other, some, together,"},
"H260":{"lemma":"אָחוּ","xlit":"ʼâchûw","pron":"aw'-khoo","derivation":"of uncertain (perhaps Egyptian) derivation;","strongs_def":"a bulrush or any marshy grass (particularly that along the Nile)","kjv_def":"flag, meadow."},
"H261":{"lemma":"אֵחוּד","xlit":"ʼÊchûwd","pron":"ay-khood'","derivation":"from H258 (אָחַד); united;","strongs_def":"Echud, the name of three Israelites","kjv_def":"Ehud."},
"H262":{"lemma":"אַחְוָה","xlit":"ʼachvâh","pron":"akh-vaw'","derivation":"from H2331 (חָוָה) (in the sense of H2324 (חֲוָא));","strongs_def":"an utterance","kjv_def":"declaration."},
"H263":{"lemma":"אַחֲוָה","xlit":"ʼachăvâh","pron":"akh-av-aw'","derivation":"(Aramaic) corresponding to H262 (אַחְוָה);","strongs_def":"solution (of riddles)","kjv_def":"showing."},
"H264":{"lemma":"אַחֲוָה","xlit":"ʼachăvâh","pron":"akh-av-aw'","derivation":"from H251 (אָח);","strongs_def":"fraternity","kjv_def":"brotherhood."},
"H265":{"lemma":"אֲחוֹחַ","xlit":"ʼĂchôwach","pron":"akh-o'-akh","derivation":"by reduplication from H251 (אָח); brotherly;","strongs_def":"Achoach, an Israelite","kjv_def":"Ahoah."},
"H266":{"lemma":"אֲחוֹחִי","xlit":"ʼĂchôwchîy","pron":"akh-o-khee'","derivation":"patronymic from H264 (אַחֲוָה);","strongs_def":"an Achochite or descendant of Achoach","kjv_def":"Ahohite."},
"H267":{"lemma":"אֲחוּמַי","xlit":"ʼĂchûwmay","pron":"akh-oo-mah'-ee","derivation":"perhaps from H251 (אָח) and H4325 (מַיִם); brother (i.e. neighbour) of water;","strongs_def":"Achumai, an Israelite","kjv_def":"Ahumai."},
"H268":{"lemma":"אָחוֹר","xlit":"ʼâchôwr","pron":"aw-khore'","derivation":"or (shortened) אָחֹר ; from H299 (אֲחִירַע);","strongs_def":"the hinder part; hence (adverb) behind, backward; also (as facing north) the West","kjv_def":"after(-ward), back (part, -side, -ward), hereafter, (be-) hind(-er part), time to come, without."},
"H269":{"lemma":"אָחוֹת","xlit":"ʼâchôwth","pron":"aw-khoth'","derivation":"irregular feminine of H251 (אָח);","strongs_def":"a sister (used very widely (like brother), literally and figuratively)","kjv_def":"(an-) other, sister, together."},
"H270":{"lemma":"אָחַז","xlit":"ʼâchaz","pron":"aw-khaz'","derivation":"a primitive root;","strongs_def":"to seize (often with the accessory idea of holding in possession)","kjv_def":"[phrase] be affrighted, bar, (catch, lay, take) hold (back), come upon, fasten, handle, portion, (get, have or take) possess(-ion)."},
"H271":{"lemma":"אָחָז","xlit":"ʼÂchâz","pron":"aw-khawz'","derivation":"from H270 (אָחַז); possessor;","strongs_def":"Achaz, the name of a Jewish king and of an Israelite","kjv_def":"Ahaz."},
"H272":{"lemma":"אֲחֻזָּה","xlit":"ʼăchuzzâh","pron":"akh-ooz-zaw'","derivation":"feminine passive participle from H270 (אָחַז);","strongs_def":"something seized, i.e. a possession (especially of land)","kjv_def":"possession."},
"H273":{"lemma":"אַחְזַי","xlit":"ʼAchzay","pron":"akh-zah'ee","derivation":"from H270 (אָחַז); seizer;","strongs_def":"Achzai, an Israelite","kjv_def":"Ahasai."},
"H274":{"lemma":"אֲחַזְיָה","xlit":"ʼĂchazyâh","pron":"akh-az-yaw'","derivation":"or (prolonged) אֲחַזְיָהוּ ; from H270 (אָחַז) and H3050 (יָהּ); Jah has seized;","strongs_def":"Achazjah, the name of a Jewish and an Israelite king","kjv_def":"Ahaziah."},
"H275":{"lemma":"אֲחֻזָּם","xlit":"ʼĂchuzzâm","pron":"akh-ooz-zawm'","derivation":"from H270 (אָחַז); seizure;","strongs_def":"Achuzzam, an Israelite","kjv_def":"Ahuzam."},
"H276":{"lemma":"אֲחֻזַּת","xlit":"ʼĂchuzzath","pron":"akh-ooz-zath'","derivation":"a variation of H272 (אֲחֻזָּה); possession;","strongs_def":"Achuzzath, a Philistine","kjv_def":"Ahuzzath."},
"H277":{"lemma":"אֲחִי","xlit":"ʼĂchîy","pron":"akh-ee'","derivation":"from H251 (אָח); brotherly;","strongs_def":"Achi, the name of two Israelites","kjv_def":"Ahi."},
"H278":{"lemma":"אֵחִי","xlit":"ʼÊchîy","pron":"ay-khee'","derivation":"probably the same as H277 (אֲחִי);","strongs_def":"Echi, an Israelite","kjv_def":"Ehi."},
"H279":{"lemma":"אֲחִיאָם","xlit":"ʼĂchîyʼâm","pron":"akh-ee-awm'","derivation":"from H251 (אָח) and H517 (אֵם); brother of the mother (i.e. uncle);","strongs_def":"Achiam, an Israelite","kjv_def":"Ahiam."},
"H280":{"lemma":"אֲחִידָה","xlit":"ʼăchîydâh","pron":"akh-ee-daw'","derivation":"(Aramaic) corresponding to H2420 (חִידָה),","strongs_def":"an enigma","kjv_def":"hard sentence."},
"H281":{"lemma":"אֲחִיָּה","xlit":"ʼĂchîyâh","pron":"akh-ee-yaw","derivation":"or (prolonged) אֲחִיָּהוּ ; from H251 (אָח) and H3050 (יָהּ); brother (i.e. worshipper) of Jah;","strongs_def":"Achijah, the name of nine Israelites","kjv_def":"Ahiah, Ahijah."},
"H282":{"lemma":"אֲחִיהוּד","xlit":"ʼĂchîyhûwd","pron":"akh-ee-hood'","derivation":"from H251 (אָח) and H1935 (הוֹד); brother (i.e. possessor) of renown;","strongs_def":"Achihud, an Israelite","kjv_def":"Ahihud."},
"H283":{"lemma":"אַחְיוֹ","xlit":"ʼAchyôw","pron":"akh-yo'","derivation":"prolonged from H251 (אָח); brotherly;","strongs_def":"Achio, the name of three Israelites","kjv_def":"Ahio."},
"H284":{"lemma":"אֲחִיחֻד","xlit":"ʼĂchîychud","pron":"akh-ee-khood'","derivation":"from H251 (אָח) and H2330 (חוּד); brother of a riddle (i.e. mysterious);","strongs_def":"Achichud, an Israelite","kjv_def":"Ahihud."},
"H285":{"lemma":"אֲחִיטוּב","xlit":"ʼĂchîyṭûwb","pron":"akh-ee-toob'","derivation":"from H251 (אָח) and H2898 (טוּב); brother of goodness;","strongs_def":"Achitub, the name of several priests","kjv_def":"Ahitub."},
"H286":{"lemma":"אֲחִילוּד","xlit":"ʼĂchîylûwd","pron":"akh-ee-lood'","derivation":"from H251 (אָח) and H3205 (יָלַד); brother of one born;","strongs_def":"Achilud, an Israelite","kjv_def":"Ahilud."},
"H287":{"lemma":"אֲחִימוֹת","xlit":"ʼĂchîymôwth","pron":"akh-ee-moth'","derivation":"from H251 (אָח) and H4191 (מוּת); brother of death;","strongs_def":"Achimoth, an Israelite","kjv_def":"Ahimoth."},
"H288":{"lemma":"אֲחִימֶלֶךְ","xlit":"ʼĂchîymelek","pron":"akh-ee-meh'-lek","derivation":"from H251 (אָח) and H4428 (מֶלֶךְ); brother of (the) king;","strongs_def":"Achimelek, the name of an Israelite and of a Hittite","kjv_def":"Ahimelech."},
"H289":{"lemma":"אֲחִימַן","xlit":"ʼĂchîyman","pron":"akh-ee-man'","derivation":"or אֲחִימָן ; from H251 (אָח) and H4480 (מִן); brother of a portion (i.e. gift);","strongs_def":"Achiman, the name of an Anakite and of an Israelite","kjv_def":"Ahiman."},
"H290":{"lemma":"אֲחִימַעַץ","xlit":"ʼĂchîymaʻats","pron":"akh-ee-mah'-ats","derivation":"from H251 (אָח) and the equivalent of H4619 (מַעַץ); brother of anger;","strongs_def":"Achimaats, the name of three Israelites","kjv_def":"Ahimaaz."},
"H291":{"lemma":"אַחְיָן","xlit":"ʼAchyân","pron":"akh-yawn'","derivation":"from H251 (אָח); brotherly;","strongs_def":"Achjan, an Israelite","kjv_def":"Ahian."},
"H292":{"lemma":"אֲחִינָדָב","xlit":"ʼĂchîynâdâb","pron":"akh-ee-naw-dawb'","derivation":"from H251 (אָח) and H5068 (נָדַב); brother of liberality;","strongs_def":"Achinadab, an Israelite","kjv_def":"Ahinadab."},
"H293":{"lemma":"אֲחִינֹעַם","xlit":"ʼĂchîynôʻam","pron":"akh-ee-no'-am","derivation":"from H251 (אָח) and H5278 (נֹעַם); brother of pleasantness;","strongs_def":"Achinoam, the name of two Israelitesses","kjv_def":"Ahinoam."},
"H294":{"lemma":"אֲחִיסָמָךְ","xlit":"ʼĂchîyçâmâk","pron":"akh-ee-saw-mawk'","derivation":"from H251 (אָח) and H5564 (סָמַךְ); brother of support;","strongs_def":"Achisamak, an Israelite","kjv_def":"Ahisamach."},
"H295":{"lemma":"אֲחִיעֶזֶר","xlit":"ʼĂchîyʻezer","pron":"akh-ee-eh'-zer","derivation":"from H251 (אָח) and H5828 (עֵזֶר); brother of help;","strongs_def":"Achiezer, the name of two Israelites","kjv_def":"Ahiezer."},
"H296":{"lemma":"אֲחִיקָם","xlit":"ʼĂchîyqâm","pron":"akh-ee-kawm'","derivation":"from H251 (אָח) and H6965 (קוּם); brother of rising (i.e. high);","strongs_def":"Achikam, an Israelite","kjv_def":"Ahikam."},
"H297":{"lemma":"אֲחִירָם","xlit":"ʼĂchîyrâm","pron":"akh-ee-rawm'","derivation":"from H251 (אָח) and H7311 (רוּם); brother of height (i.e. high);","strongs_def":"Achiram, an Israelite","kjv_def":"Ahiram."},
"H298":{"lemma":"אֲחִירָמִי","xlit":"ʼĂchîyrâmîy","pron":"akh-ee-raw-mee'","derivation":"patronymic from H297 (אֲחִירָם);","strongs_def":"an Achiramite or descendant (collectively) of Achiram","kjv_def":"Ahiramites."},
"H299":{"lemma":"אֲחִירַע","xlit":"ʼĂchîyraʻ","pron":"akh-ee-rah'","derivation":"from H251 (אָח) and H7451 (רַע); brother of wrong;","strongs_def":"Achira, an Israelite","kjv_def":"Ahira."},
"H300":{"lemma":"אֲחִישַׁחַר","xlit":"ʼĂchîyshachar","pron":"akh-ee-shakh'-ar","derivation":"from H251 (אָח) and H7837 (שַׁחַר); brother of (the) dawn;","strongs_def":"Achishachar, an Israelite","kjv_def":"Ahishar."},
"H301":{"lemma":"אֲחִישָׁר","xlit":"ʼĂchîyshâr","pron":"akh-ee-shawr'","derivation":"from H251 (אָח) and H7891 (שִׁיר); brother of (the) singer;","strongs_def":"Achishar, an Israelite","kjv_def":"Ahishar."},
"H302":{"lemma":"אֲחִיתֹפֶל","xlit":"ʼĂchîythôphel","pron":"akh-ee-tho'-fel","derivation":"from H251 (אָח) and H8602 (תָּפֵל); brother of folly;","strongs_def":"Achithophel, an Israelite","kjv_def":"Ahithophel."},
"H303":{"lemma":"אַחְלָב","xlit":"ʼAchlâb","pron":"akh-lawb'","derivation":"from the same root as H2459 (חֶלֶב); fatness (i.e. fertile);","strongs_def":"Achlab, a place in Palestine","kjv_def":"Ahlab."},
"H304":{"lemma":"אַחְלַי","xlit":"ʼAchlay","pron":"akh-lah'ee","derivation":"the same as H305 (אַחֲלַי); wishful;","strongs_def":"Achlai, the name of an Israelitess and of an Israelite","kjv_def":"Ahlai."},
"H305":{"lemma":"אַחֲלַי","xlit":"ʼachălay","pron":"akh-al-ah'ee","derivation":"or אַחֲלֵי; prob from H253 (אָח) and a variation of H3863 (לוּא);","strongs_def":"would that!","kjv_def":"O that, would God."},
"H306":{"lemma":"אַחְלָמָה","xlit":"ʼachlâmâh","pron":"akh-law'-maw","derivation":"perhaps from H2492 (חָלַם) (and thus dream-stone);","strongs_def":"a gem, probably the amethyst","kjv_def":"amethyst."},
"H307":{"lemma":"אַחְמְתָא","xlit":"ʼAchmᵉthâʼ","pron":"akh-me-thaw'","derivation":"of Persian derivation;","strongs_def":"Achmetha (i.e. Ecbatana), the summer capital of Persia","kjv_def":"Achmetha."},
"H308":{"lemma":"אֲחַסְבַּי","xlit":"ʼĂchaçbay","pron":"akh-as-bah'ee","derivation":"of uncertain derivation;","strongs_def":"Achasbai, an Israelite","kjv_def":"Ahasbai."},
"H309":{"lemma":"אָחַר","xlit":"ʼâchar","pron":"aw-khar'","derivation":"a primitive root;","strongs_def":"to loiter (i.e. be behind); by implication to procrastinate","kjv_def":"continue, defer, delay, hinder, be late (slack), stay (there), tarry (longer)."},
"H310":{"lemma":"אַחַר","xlit":"ʼachar","pron":"akh-ar'","derivation":"from H309 (אָחַר);","strongs_def":"properly, the hind part; generally used as an adverb or conjunction, after (in various senses)","kjv_def":"after (that, -ward), again, at, away from, back (from, -side), behind, beside, by, follow (after, -ing), forasmuch, from, hereafter, hinder end, [phrase] out (over) live, [phrase] persecute, posterity, pursuing, remnant, seeing, since, thence(-forth), when, with."},
"H311":{"lemma":"אַחַר","xlit":"ʼachar","pron":"akh-ar'","derivation":"(Aramaic) corresponding to H310 (אַחַר);","strongs_def":"after","kjv_def":"(here-) after."},
"H312":{"lemma":"אַחֵר","xlit":"ʼachêr","pron":"akh-air'","derivation":"from H309 (אָחַר);","strongs_def":"properly, hinder; generally, next, other, etc.","kjv_def":"(an-) other man, following, next, strange."},
"H313":{"lemma":"אַחֵר","xlit":"ʼAchêr","pron":"akh-air'","derivation":"the same as H312 (אַחֵר);","strongs_def":"Acher, an Israelite","kjv_def":"Aher."},
"H314":{"lemma":"אַחֲרוֹן","xlit":"ʼachărôwn","pron":"akh-ar-one'","derivation":"or (shortened) אַחֲרֹן; from H309 (אָחַר);","strongs_def":"hinder; generally, late or last; specifically (as facing the east) western","kjv_def":"after (-ward), to come, following, hind(-er, -ermost, -most), last, latter, rereward, ut(ter) most."},
"H315":{"lemma":"אַחְרַח","xlit":"ʼAchrach","pron":"akh-rakh'","derivation":"from H310 (אַחַר) and H251 (אָח); after (his) brother;","strongs_def":"Achrach, an Israelite","kjv_def":"Aharah."},
"H316":{"lemma":"אֲחַרְחֵל","xlit":"ʼĂcharchêl","pron":"akh-ar-kale'","derivation":"from H310 (אַחַר) and H2426 (חֵיל); behind (the) intrenchment (i.e. safe);","strongs_def":"Acharchel, an Israelite","kjv_def":"Aharhel,"},
"H317":{"lemma":"אׇחֳרִי","xlit":"ʼochŏrîy","pron":"okh-or-ee'","derivation":"(Aramaic) from H311 (אַחַר);","strongs_def":"other","kjv_def":"(an-) other."},
"H318":{"lemma":"אׇחֳרֵין","xlit":"ʼochŏrêyn","pron":"okh-or-ane'","derivation":"(Aramaic) or (shortened) אׇחֳרֵן; (Aramaic) from H317 (אׇחֳרִי);","strongs_def":"last","kjv_def":"at last."},
"H319":{"lemma":"אַחֲרִית","xlit":"ʼachărîyth","pron":"akh-ar-eeth'","derivation":"from H310 (אַחַר);","strongs_def":"the last or end, hence, the future; also posterity","kjv_def":"(last, latter) end (time), hinder (utter) -most, length, posterity, remnant, residue, reward."},
"H320":{"lemma":"אַחֲרִית","xlit":"ʼachărîyth","pron":"akh-ar-eeth'","derivation":"(Aramaic) from H311 (אַחַר); the same as H319 (אַחֲרִית);","strongs_def":"later","kjv_def":"latter."},
"H321":{"lemma":"אׇחֳרָן","xlit":"ʼochŏrân","pron":"okh-or-awn'","derivation":"(Aramaic) from H311 (אַחַר); the same as H317 (אׇחֳרִי);","strongs_def":"other","kjv_def":"(an-) other."},
"H322":{"lemma":"אֲחֹרַנִּית","xlit":"ʼăchôrannîyth","pron":"akh-o-ran-neeth'","derivation":"prolonged from H268 (אָחוֹר);","strongs_def":"backwards","kjv_def":"back (-ward, again)."},
"H323":{"lemma":"אֲחַשְׁדַּרְפַּן","xlit":"ʼăchashdarpan","pron":"akh-ash-dar-pan'","derivation":"of Persian derivation;","strongs_def":"a satrap or governorof amain province (of Persia)","kjv_def":"lieutenant."},
"H324":{"lemma":"אֲחַשְׁדַּרְפַּן","xlit":"ʼăchashdarpan","pron":"akh-ash-dar-pan'","derivation":"(Aramaic) corresponding to H323 (אֲחַשְׁדַּרְפַּן)","strongs_def":"{a satrap or governorof amain province (of Persia)}","kjv_def":"prince."},
"H325":{"lemma":"אֲחַשְׁוֵרוֹשׁ","xlit":"ʼĂchashvêrôwsh","pron":"akh-ash-vay-rosh'","derivation":"or (shortened) אַחַשְׁרֹשׁ (Esther 10:1); of Persian origin;","strongs_def":"Achashverosh (i.e. Ahasuerus or Artaxerxes, but in this case Xerxes), the title (rather than name) of a Persian king","kjv_def":"Ahasuerus."},
"H326":{"lemma":"אֲחַשְׁתָּרִי","xlit":"ʼăchashtârîy","pron":"akh-ash-taw-ree'","derivation":"probably of Persian derivation;","strongs_def":"an achastarite (i.e. courier); the designation (rather than name) of an Israelite","kjv_def":"Haakashtari (includ. the article)."},
"H327":{"lemma":"אֲחַשְׁתָּרָן","xlit":"ʼăchashtârân","pron":"akh-ash-taw-rawn'","derivation":"of Persian origin;","strongs_def":"a mule","kjv_def":"camel."},
"H328":{"lemma":"אַט","xlit":"ʼaṭ","pron":"at","derivation":"from an unused root perhaps meaning to move softly;","strongs_def":"(as a noun) a necromancer (from their soft incantations), (as an adverb) gently","kjv_def":"charmer, gently, secret, softly."},
"H329":{"lemma":"אָטָד","xlit":"ʼâṭâd","pron":"aw-tawd'","derivation":"from an unused root probably meaning to pierce or make fast;","strongs_def":"a thorn-tree (especially the buckthorn)","kjv_def":"Atad, bramble, thorn."},
"H330":{"lemma":"אֵטוּן","xlit":"ʼêṭûwn","pron":"ay-toon'","derivation":"from an unused root (probably meaning to bind);","strongs_def":"properly, twisted (yarn), i.e. tapestry","kjv_def":"fine linen."},
"H331":{"lemma":"אָטַם","xlit":"ʼâṭam","pron":"aw-tam'","derivation":"a primitive root;","strongs_def":"to close (the lips or ears); by analology to contract (a window by bevelled jambs)","kjv_def":"narrow, shut, stop."},
"H332":{"lemma":"אָטַר","xlit":"ʼâṭar","pron":"aw-tar'","derivation":"a primitive root;","strongs_def":"to close up","kjv_def":"shut."},
"H333":{"lemma":"אָטֵר","xlit":"ʼÂṭêr","pron":"aw-tare'","derivation":"from H332 (אָטַר); maimed;","strongs_def":"Ater, the name of three Israelites","kjv_def":"Ater."},
"H334":{"lemma":"אִטֵּר","xlit":"ʼiṭṭêr","pron":"it-tare'","derivation":"from H332 (אָטַר);","strongs_def":"shut up, i.e. impeded (as to the use of the right hand)","kjv_def":"[phrase] left-handed."},
"H335":{"lemma":"אַי","xlit":"ʼay","pron":"ah'ee","derivation":"perhaps from H370 (אַיִן);","strongs_def":"where? hence how?","kjv_def":"how, what, whence, where, whether, which (way)."},
"H336":{"lemma":"אִי","xlit":"ʼîy","pron":"ee","derivation":"probably identical with H335 (אַי) (through the idea of a query);","strongs_def":"not","kjv_def":"island (Job 22:30)."},
"H337":{"lemma":"אִי","xlit":"ʼîy","pron":"ee","derivation":"short from H188 (אוֹי);","strongs_def":"alas!","kjv_def":"woe."},
"H338":{"lemma":"אִי","xlit":"ʼîy","pron":"ee","derivation":"probably identical with H337 (אִי) (through the idea of a doleful sound);","strongs_def":"a howler (used only in the plural), i.e. any solitary wild creature","kjv_def":"wild beast of the islands."},
"H339":{"lemma":"אִי","xlit":"ʼîy","pron":"ee","derivation":"from H183 (אָוָה);","strongs_def":"properly, a habitable spot (as desirable); dry land, a coast, an island","kjv_def":"country, isle, island."},
"H340":{"lemma":"אָיַב","xlit":"ʼâyab","pron":"aw-yab'","derivation":"a primitive root;","strongs_def":"to hate (as one of an opposite tribe or party); hence to be hostile","kjv_def":"be an enemy."},
"H341":{"lemma":"אֹיֵב","xlit":"ʼôyêb","pron":"o-yabe'","derivation":"or (fully) אוֹיֵב; active participle of H340 (אָיַב);","strongs_def":"hating; an adversary","kjv_def":"enemy, foe."},
"H342":{"lemma":"אֵיבָה","xlit":"ʼêybâh","pron":"ay-baw'","derivation":"from H340 (אָיַב);","strongs_def":"hostility","kjv_def":"emnity, hatred."},
"H343":{"lemma":"אֵיד","xlit":"ʼêyd","pron":"ade","derivation":"from the same as H181 (אוּד) (in the sense of bending down);","strongs_def":"oppression; by implication misfortune, ruin","kjv_def":"calamity, destruction."},
"H344":{"lemma":"אַיָּה","xlit":"ʼayâh","pron":"ah-yaw'","derivation":"perhaps from H337 (אִי);","strongs_def":"the screamer, i.e. a hawk","kjv_def":"kite, vulture."},
"H345":{"lemma":"אַיָּה","xlit":"ʼAyâh","pron":"ah-yaw'","derivation":"the same as H344 (אַיָּה);","strongs_def":"Ajah, the name of two Israelites","kjv_def":"Aiah, Ajah."},
"H346":{"lemma":"אַיֵּה","xlit":"ʼayêh","pron":"ah-yay'","derivation":"prolonged from H335 (אַי);","strongs_def":"where?","kjv_def":"where."},
"H347":{"lemma":"אִיּוֹב","xlit":"ʼÎyôwb","pron":"ee-yobe'","derivation":"from H340 (אָיַב); hated (i.e. persecuted);","strongs_def":"Ijob, the patriarch famous for his patience","kjv_def":"Job."},
"H348":{"lemma":"אִיזֶבֶל","xlit":"ʼÎyzebel","pron":"ee-zeh'-bel","derivation":"from H336 (אִי)lemma אי missing vowel, corrected to אִי and H2083 (זְבֻל);","strongs_def":"Izebel, the wife of king Ahab","kjv_def":"Jezebel."},
"H349":{"lemma":"אֵיךְ","xlit":"ʼêyk","pron":"ake","derivation":"also אֵיכָה ; and אֵיכָכָה ; prolonged from H335 (אַי);","strongs_def":"how? or how!; also where","kjv_def":"how, what."},
"H350":{"lemma":"אִי־כָבוֹד","xlit":"ʼÎy-kâbôwd","pron":"ee-kaw-bode'","derivation":"from H336 (אִי)lemma אי missing vowel, corrected to אִי and H3519 (כָּבוֹד); (there is) no glory, i.e. inglorious;","strongs_def":"Ikabod, a son of Phineas","kjv_def":"I-chabod."},
"H351":{"lemma":"אֵיכֹה","xlit":"ʼêykôh","pron":"ay-ko","derivation":"probably a variation for H349 (אֵיךְ), but not as an interogative;","strongs_def":"where","kjv_def":"where."},
"H352":{"lemma":"אַיִל","xlit":"ʼayil","pron":"ah'-yil","derivation":"from the same as H193 (אוּל);","strongs_def":"properly, strength; hence, anything strong; specifically a chief (politically); also a ram (from his strength); a pilaster (as a strong support); an oak or other strong tree","kjv_def":"mighty (man), lintel, oak, post, ram, tree."},
"H353":{"lemma":"אֱיָל","xlit":"ʼĕyâl","pron":"eh-yawl'","derivation":"a variation of H352 (אַיִל);","strongs_def":"strength","kjv_def":"strength."},
"H354":{"lemma":"אַיָּל","xlit":"ʼayâl","pron":"ah-yawl'","derivation":"an intensive form of H352 (אַיִל) (in the sense of ram);","strongs_def":"a stag or male deer","kjv_def":"hart."},
"H355":{"lemma":"אַיָּלָה","xlit":"ʼayâlâh","pron":"ah-yaw-law'","derivation":"feminine of H354 (אַיָּל);","strongs_def":"a doe or female deer","kjv_def":"hind."},
"H356":{"lemma":"אֵילוֹן","xlit":"ʼÊylôwn","pron":"ay-lone'","derivation":"or (shortened) אֵלוֹן ; or אֵילֹן ; from H352 (אַיִל); oak-grove;","strongs_def":"Elon, the name of a place in Palestine, and also of one Hittite, two Israelites","kjv_def":"Elon."},
"H357":{"lemma":"אַיָּלוֹן","xlit":"ʼAyâlôwn","pron":"ah-yaw-lone'","derivation":"from H354 (אַיָּל); deer-field;","strongs_def":"Ajalon, the name of five places in Palestine","kjv_def":"Aijalon, Ajalon."},
"H358":{"lemma":"אֵילוֹן בֵּית חָנָן","xlit":"ʼÊylôwn Bêyth Chânân","pron":"ay-lone' bayth-chawnawn'","derivation":"from H356 (אֵילוֹן), H1004 (בַּיִת), and H2603 (חָנַן); oak-grove of (the) house of favor;","strongs_def":"Elon of Beth-chanan, a place in Palestine","kjv_def":"Elon-bethhanan."},
"H359":{"lemma":"אֵילוֹת","xlit":"ʼÊylôwth","pron":"ay-loth'","derivation":"or אֵילַת ; from H352 (אַיִל); trees or a grove (i.e. palms);","strongs_def":"Eloth or Elath, a place on the Red Sea","kjv_def":"Elath, Eloth."},
"H360":{"lemma":"אֱיָלוּת","xlit":"ʼĕyâlûwth","pron":"eh-yaw-looth'","derivation":"feminine of H353 (אֱיָל);","strongs_def":"power; by implication, protection","kjv_def":"strength."},
"H361":{"lemma":"אֵילָם","xlit":"ʼêylâm","pron":"ay-lawm'","derivation":"or (shortened) אֵלָם ; or (feminine) אֵלַמָּה ; probably from H352 (אַיִל);","strongs_def":"a pillar-space (or colonnade), i.e. a pale (or portico)","kjv_def":"arch."},
"H362":{"lemma":"אֵילִם","xlit":"ʼÊylim","pron":"ay-leem'","derivation":"plural of H352 (אַיִל); palm-trees;","strongs_def":"Elim, a place in the Desert","kjv_def":"Elim."},
"H363":{"lemma":"אִילָן","xlit":"ʼîylân","pron":"ee-lawn'","derivation":"(Aramaic) corresponding to H356 (אֵילוֹן);","strongs_def":"a tree","kjv_def":"tree."},
"H364":{"lemma":"אֵיל פָּארָן","xlit":"ʼÊyl Pâʼrân","pron":"ale paw-rawn'","derivation":"from H352 (אַיִל) and H6290 (פָּארָן); oak of Paran;","strongs_def":"El-Paran, a portion of the district of Paran","kjv_def":"El-paran."},
"H365":{"lemma":"אַיֶּלֶת","xlit":"ʼayeleth","pron":"ah-yeh'-leth","derivation":"the same as H355 (אַיָּלָה);","strongs_def":"a doe","kjv_def":"hind, Aijeleth."},
"H366":{"lemma":"אָיֹם","xlit":"ʼâyôm","pron":"aw-yome'","derivation":"from an unused root (meaning to frighten);","strongs_def":"frightful","kjv_def":"terrible."},
"H367":{"lemma":"אֵימָה","xlit":"ʼêymâh","pron":"ay-maw'","derivation":"or (shortened) אֵמָה; from the same as H366 (אָיֹם);","strongs_def":"fright; concrete, an idol (as a bugbear)","kjv_def":"dread, fear, horror, idol, terrible, terror."},
"H368":{"lemma":"אֵימִים","xlit":"ʼÊymîym","pron":"ay-meem'","derivation":"plural of H367 (אֵימָה); terrors;","strongs_def":"Emim, an early Canaanitish (or Maobitish) tribe","kjv_def":"Emims."},
"H369":{"lemma":"אַיִן","xlit":"ʼayin","pron":"ah'-yin","derivation":"as if from a primitive root meaning to be nothing or not exist;","strongs_def":"a non-entity; generally used as a negative particle","kjv_def":"else, except, fail, (father-) less, be gone, in(-curable), neither, never, no (where), none, nor, (any, thing), not, nothing, to nought, past, un(-searchable), well-nigh, without. Compare H370 (אַיִן)."},
"H370":{"lemma":"אַיִן","xlit":"ʼayin","pron":"ah-yin'","derivation":"probably identical with H369 (אַיִן) in the sense of query (compare H336 (אִי)lemma אי missing vowel, corrected to אִי)","strongs_def":"where? (only in connection with prepositional prefix, whence)","kjv_def":"whence, where."},
"H371":{"lemma":"אִין","xlit":"ʼîyn","pron":"een","derivation":"apparently a shortened form of H369 (אַיִן); but (like H370 (אַיִן)) an interrogative;","strongs_def":"is it not?","kjv_def":"not."},
"H372":{"lemma":"אִיעֶזֵר","xlit":"ʼÎyʻezêr","pron":"ee-eh'-zer","derivation":"from H336 (אִי)lemma אי missing vowel, corrected to אִי and H5828 (עֵזֶר); helpless;","strongs_def":"Iezer, an Israelite","kjv_def":"Jeezer."},
"H373":{"lemma":"אִיעֶזְרִי","xlit":"ʼÎyʻezrîy","pron":"ee-ez-ree'","derivation":"patronymic from H372 (אִיעֶזֵר);","strongs_def":"an Iezrite or descendant of Iezer","kjv_def":"Jezerite."},
"H374":{"lemma":"אֵיפָה","xlit":"ʼêyphâh","pron":"ay-faw'","derivation":"or (shortened) אֵפָה ; of Egyptian derivation;","strongs_def":"an ephah or measure for grain; hence, a measure in general","kjv_def":"ephah, (divers) measure(-s)."},
"H375":{"lemma":"אֵיפֹה","xlit":"ʼêyphôh","pron":"ay-fo'","derivation":"from H335 (אַי) and H6311 (פֹּה);","strongs_def":"what place?; also (of time) when?; or (of means) how?","kjv_def":"what manner, where."},
"H376":{"lemma":"אִישׁ","xlit":"ʼîysh","pron":"eesh","derivation":"contracted for H582 (אֱנוֹשׁ) (or perhaps rather from an unused root meaning to be extant);","strongs_def":"a man as an individual or a male person; often used as an adjunct to a more definite term (and in such cases frequently not expressed in translation)","kjv_def":"also, another, any (man), a certain, [phrase] champion, consent, each, every (one), fellow, (foot-, husband-) man, (good-, great, mighty) man, he, high (degree), him (that is), husband, man(-kind), [phrase] none, one, people, person, [phrase] steward, what (man) soever, whoso(-ever), worthy. Compare H802 (אִשָּׁה)."},
"H377":{"lemma":"אִישׁ","xlit":"ʼîysh","pron":"eesh","derivation":"denominative from H376 (אִישׁ);","strongs_def":"to be a man, i.e. act in amanly way","kjv_def":"show (one) self a man."},
"H378":{"lemma":"אִישׁ־בֹּשֶׁת","xlit":"ʼÎysh-Bôsheth","pron":"eesh-bo'-sheth","derivation":"from H376 (אִישׁ) and H1322 (בֹּשֶׁת); man of shame;","strongs_def":"Ish-Bosheth, a son of King Saul","kjv_def":"Ish-bosheth."},
"H379":{"lemma":"אִישְׁהוֹד","xlit":"ʼÎyshᵉhôwd","pron":"eesh-hode'","derivation":"from H376 (אִישׁ) and H1935 (הוֹד); man of renown;","strongs_def":"Ishod, an Israelite","kjv_def":"Ishod."},
"H380":{"lemma":"אִישׁוֹן","xlit":"ʼîyshôwn","pron":"ee-shone'","derivation":"diminutive from H376 (אִישׁ);","strongs_def":"the little man of the eye; the pupil or ball; hence, the middle (of night)","kjv_def":"apple (of the eye), black, obscure."},
"H381":{"lemma":"אִישׁ־חַיִל","xlit":"ʼÎysh-Chayil","pron":"eesh-khah'-yil","derivation":"from H376 (אִישׁ) and H2428 (חַיִל); man of might; by defective transcription (2 Samuel 23:20) אִישׁ־חַיxlit ʼÎsh-Chay corrected to ʼÎysh-Chay; as if from H376 (אִישׁ) and H2416 (חַי); living man;","strongs_def":"Ishchail (or Ish-chai), an Israelite","kjv_def":"a valiant man."},
"H382":{"lemma":"אִישׁ־טוֹב","xlit":"ʼÎysh-Ṭôwb","pron":"eesh-tobe'","derivation":"from H376 (אִישׁ) and H2897 (טוֹב); man of Tob;","strongs_def":"Ish-Tob, a place in Palestine","kjv_def":"Ish-tob."},
"H383":{"lemma":"אִיתַי","xlit":"ʼîythay","pron":"ee-thah'ee","derivation":"(Aramaic) corresponding to H3426 (יֵשׁ);","strongs_def":"properly, entity; used only as aparticle of affirmation, there is","kjv_def":"art thou, can, do ye, have, it be, there is (are), [idiom] we will not."},
"H384":{"lemma":"אִיתִיאֵל","xlit":"ʼÎythîyʼêl","pron":"eeth-ee-ale'","derivation":"perhaps from H837 (אֹשֶׁר) and H410 (אֵל); God has arrived;","strongs_def":"Ithiel, the name of an Israelite, also of a symbolical person","kjv_def":"Ithiel."},
"H385":{"lemma":"אִיתָמָר","xlit":"ʼÎythâmâr","pron":"eeth-aw-mawr'","derivation":"from H339 (אִי) and H8558 (תָּמָר); coast of the palm-tree;","strongs_def":"Ithamar, a son of Aaron","kjv_def":"Ithamar."},
"H386":{"lemma":"אֵיתָן","xlit":"ʼêythân","pron":"ay-thawn'","derivation":"or (shortened) אֵתָן ; from an unused root (meaning to continue);","strongs_def":"permanence; hence (concrete) permanent; specifically a chieftain","kjv_def":"hard, mighty, rough, strength, strong."},
"H387":{"lemma":"אֵיתָן","xlit":"ʼÊythân","pron":"ay-thawn'","derivation":"the same as H386 (אֵיתָן); permanent;","strongs_def":"Ethan, the name of four Israelites","kjv_def":"Ethan."},
"H388":{"lemma":"אֵיתָנִים","xlit":"ʼÊythânîym","pron":"ay-thaw-neem'","derivation":"plural of H386 (אֵיתָן); always with the article; the permanent brooks;","strongs_def":"Ethanim, the name of a month","kjv_def":"Ethanim."},
"H389":{"lemma":"אַךְ","xlit":"ʼak","pron":"ak","derivation":"akin to H403 (אָכֵן);","strongs_def":"a particle of affirmation, surely; hence (by limitation) only","kjv_def":"also, in any wise, at least, but, certainly, even, howbeit, nevertheless, notwithstanding, only, save, surely, of a surety, truly, verily, [phrase] wherefore, yet (but)."},
"H390":{"lemma":"אַכַּד","xlit":"ʼAkkad","pron":"ak-kad'","derivation":"from an unused root probably meaning to strengthen; a fortress;","strongs_def":"Accad, a place in Babylon","kjv_def":"Accad."},
"H391":{"lemma":"אַכְזָב","xlit":"ʼakzâb","pron":"ak-zawb'","derivation":"from H3576 (כָּזַב);","strongs_def":"falsehood; by implication treachery","kjv_def":"liar, lie."},
"H392":{"lemma":"אַכְזִיב","xlit":"ʼAkzîyb","pron":"ak-zeeb'","derivation":"from H391 (אַכְזָב); deceitful (in the sense of a winter-torrent which fails in summer);","strongs_def":"Akzib, the name of two places in Palestine","kjv_def":"Achzib."},
"H393":{"lemma":"אַכְזָר","xlit":"ʼakzâr","pron":"ak-zawr'","derivation":"from an unused root (apparently meaning to act harshly);","strongs_def":"violent; by implication deadly; also (in a good sense) brave","kjv_def":"cruel, fierce."},
"H394":{"lemma":"אַכְזָרִי","xlit":"ʼakzârîy","pron":"ak-zawr-ree'","derivation":"from H393 (אַכְזָר);","strongs_def":"terrible","kjv_def":"cruel (one)."},
"H395":{"lemma":"אַכְזְרִיּוּת","xlit":"ʼakzᵉrîyûwth","pron":"ak-ze-ree-ooth'","derivation":"from H394 (אַכְזָרִי);","strongs_def":"fierceness","kjv_def":"cruel."},
"H396":{"lemma":"אֲכִילָה","xlit":"ʼăkîylâh","pron":"ak-ee-law'","derivation":"feminine from H398 (אָכַל);","strongs_def":"something eatable, i.e. food","kjv_def":"meat."},
"H397":{"lemma":"אֲכִישׁ","xlit":"ʼĂkîysh","pron":"aw-keesh'","derivation":"of uncertain derivation;","strongs_def":"Akish, a Philistine king","kjv_def":"Achish."},
"H398":{"lemma":"אָכַל","xlit":"ʼâkal","pron":"aw-kal'","derivation":"a primitive root;","strongs_def":"to eat (literally or figuratively)","kjv_def":"[idiom] at all, burn up, consume, devour(-er, up), dine, eat(-er, up), feed (with), food, [idiom] freely, [idiom] in...wise(-deed, plenty), (lay) meat, [idiom] quite."},
"H399":{"lemma":"אֲכַל","xlit":"ʼăkal","pron":"ak-al'","derivation":"(Aramaic) corresponding to H398 (אָכַל)","strongs_def":"{to eat (literally or figuratively)}","kjv_def":"[phrase] accuse, devour, eat."},
"H400":{"lemma":"אֹכֶל","xlit":"ʼôkel","pron":"o'-kel","derivation":"from H398 (אָכַל);","strongs_def":"food","kjv_def":"eating, food, meal(-time), meat, prey, victuals."},
"H401":{"lemma":"אֻכָל","xlit":"ʼUkâl","pron":"oo-kawl'","derivation":"or אֻכָּל; apparently from H398 (אָכַל); devoured;","strongs_def":"Ucal, a fancy name","kjv_def":"Ucal."},
"H402":{"lemma":"אׇכְלָה","xlit":"ʼoklâh","pron":"ok-law'","derivation":"feminine of H401 (אֻכָל);","strongs_def":"food","kjv_def":"consume, devour, eat, food, meat."},
"H403":{"lemma":"אָכֵן","xlit":"ʼâkên","pron":"aw-kane'","derivation":"from H3559 (כּוּן) (compare H3651 (כֵּן));","strongs_def":"firmly; figuratively, surely; also (advers.) but","kjv_def":"but, certainly, nevertheless, surely, truly, verily."},
"H404":{"lemma":"אָכַף","xlit":"ʼâkaph","pron":"aw-kaf'","derivation":"a primitive root; apparently meaning to curve (as with a burden);","strongs_def":"to urge","kjv_def":"crave."},
"H405":{"lemma":"אֶכֶף","xlit":"ʼekeph","pron":"eh'-kef","derivation":"from H404 (אָכַף);","strongs_def":"a load; by implication, a stroke (others dignity)","kjv_def":"hand."},
"H406":{"lemma":"אִכָּר","xlit":"ʼikkâr","pron":"ik-kawr'","derivation":"from an unused root meaning to dig;","strongs_def":"a farmer","kjv_def":"husbandman, ploughman."},
"H407":{"lemma":"אַכְשָׁף","xlit":"ʼAkshâph","pron":"ak-shawf'","derivation":"from H3784 (כָּשַׁף); fascination;","strongs_def":"Acshaph, a place in Palestine","kjv_def":"Achshaph."},
"H408":{"lemma":"אַל","xlit":"ʼal","pron":"al","derivation":"a negative particle (akin to H3808 (לֹא));","strongs_def":"not (the qualified negation, used as a deprecative); once (Job 24:25) as a noun, nothing","kjv_def":"nay, neither, [phrase] never, no, nor, not, nothing (worth), rather than."},
"H409":{"lemma":"אַל","xlit":"ʼal","pron":"al","derivation":"(Aramaic) corresponding to H408 (אַל)","strongs_def":"{not (the qualified negation, used as a deprecative); once (Job 24:25) as a noun, nothing}","kjv_def":"not."},
"H410":{"lemma":"אֵל","xlit":"ʼêl","pron":"ale","derivation":"shortened from H352 (אַיִל);","strongs_def":"strength; as adjective, mighty; especially the Almighty (but used also of any deity)","kjv_def":"God (god), [idiom] goodly, [idiom] great, idol, might(-y one), power, strong. Compare names in '-el.'"},
"H411":{"lemma":"אֵל","xlit":"ʼêl","pron":"ale","derivation":"a demonstrative particle (but only in a plural sense)","strongs_def":"these or those","kjv_def":"these, those. Compare H428 (אֵלֶּה)."},
"H412":{"lemma":"אֵל","xlit":"ʼêl","pron":"ale","derivation":"(Aramaic) corresponding to H411 (אֵל)","strongs_def":"{these or those}","kjv_def":"these."},
"H413":{"lemma":"אֵל","xlit":"ʼêl","pron":"ale","derivation":"(but only used in the shortened constructive form אֶל ); a primitive particle; properly, denoting motion towards, but occasionally used of a quiescent position, i.e.","strongs_def":"near, with or among; often in general, to","kjv_def":"about, according to, after, against, among, as for, at, because(-fore, -side), both...and, by, concerning, for, from, [idiom] hath, in(-to), near, (out) of, over, through, to(-ward), under, unto, upon, whether, with(-in)."},
"H414":{"lemma":"אֵלָא","xlit":"ʼÊlâʼ","pron":"ay-law'","derivation":"a variation of H424 (אֵלָה)lemma אִלָה first vowel, corrected to אֵלָה; oak;","strongs_def":"Ela, an Israelite","kjv_def":"Elah."},
"H415":{"lemma":"אֵל אֱלֹהֵי יִשְׂרָאֵל","xlit":"ʼÊl ʼĕlôhêy Yisrâʼêl","pron":"ale el-o-hay' yis-rawale'","derivation":"from H410 (אֵל) and H430 (אֱלֹהִים) and H3478 (יִשְׂרָאֵל); the mighty god if Jisrael;","strongs_def":"El-Elohi-Jisrael, the title given to a consecrated spot by Jacob","kjv_def":"Elelohe-israel."},
"H416":{"lemma":"אֵל בֵּית־אֵל","xlit":"ʼÊl Bêyth-ʼÊl","pron":"ale bayth-ale'","derivation":"from H410 (אֵל) and H1008 (בֵּית־אֵל); the God of Bethel;","strongs_def":"El-Bethel, the title given to a consecrated spot by Jacob","kjv_def":"El-beth-el."},
"H417":{"lemma":"אֶלְגָּבִישׁ","xlit":"ʼelgâbîysh","pron":"el-gaw-beesh'","derivation":"from H410 (אֵל) and H1378 (גָּבִישׁ);","strongs_def":"hail (as if a great pearl)","kjv_def":"great hail(-stones)."},
"H418":{"lemma":"אַלְגּוּמִּים","xlit":"ʼalgûwmmîym","pron":"al-goom-meem'","derivation":"by transposition for H484 (אַלְמֻגִּים);","strongs_def":"sticks of algum wood","kjv_def":"algum (trees)."},
"H419":{"lemma":"אֶלְדָּד","xlit":"ʼEldâd","pron":"el-dad'","derivation":"from H410 (אֵל) and H1730 (דּוֹד); God has loved;","strongs_def":"Eldad, an Israelite","kjv_def":"Eldad."},
"H420":{"lemma":"אֶלְדָּעָה","xlit":"ʼEldâʻâh","pron":"el-daw-aw'","derivation":"from H410 (אֵל) and H3045 (יָדַע); God of knowledge;","strongs_def":"Eldaah, a son of Midian","kjv_def":"Eldaah."},
"H421":{"lemma":"אָלָה","xlit":"ʼâlâh","pron":"aw-law'","derivation":"a primitive root (rather identical with H422 (אָלָה) through the idea of invocation);","strongs_def":"to bewail","kjv_def":"lament."},
"H422":{"lemma":"אָלָה","xlit":"ʼâlâh","pron":"aw-law'","derivation":"a primitive root;","strongs_def":"properly, to adjure, i.e. (usually in a bad sense) imprecate","kjv_def":"adjure, curse, swear."},
"H423":{"lemma":"אָלָה","xlit":"ʼâlâh","pron":"aw-law'","derivation":"from H422 (אָלָה);","strongs_def":"an imprecation","kjv_def":"curse, cursing, execration, oath, swearing."},
"H424":{"lemma":"אֵלָה","xlit":"ʼêlâh","pron":"ay-law'","derivation":"feminine of H352 (אַיִל);","strongs_def":"an oak or other strong tree","kjv_def":"elm, oak, teil-tree."},
"H425":{"lemma":"אֵלָה","xlit":"ʼÊlâh","pron":"ay-law'","derivation":"the same as H424 (אֵלָה)lemma אִלָה first vowel, corrected to אֵלָה;","strongs_def":"Elah, the name of an Edomite, of four Israelites, and also of a place in Palestine","kjv_def":"Elah."},
"H426":{"lemma":"אֱלָהּ","xlit":"ʼĕlâhh","pron":"el-aw'","derivation":"(Aramaic) corresponding to H433 (אֱלוֹהַּ);","strongs_def":"God","kjv_def":"God, god."},
"H427":{"lemma":"אַלָּה","xlit":"ʼallâh","pron":"al-law'","derivation":"A variation of H424 (אֵלָה)lemma אִלָה first vowel, corrected to אֵלָה","strongs_def":"{an oak or other strong tree}","kjv_def":"oak."},
"H428":{"lemma":"אֵלֶּה","xlit":"ʼêl-leh","pron":"ale'-leh","derivation":"prolonged from H411 (אֵל);","strongs_def":"these or those","kjv_def":"an-(the) other; one sort, so, some, such, them, these (same), they, this, those, thus, which, who(-m)."},
"H429":{"lemma":"אֵלֶּה","xlit":"ʼêlleh","pron":"ale'-leh","derivation":"(Aramaic) corresponding to H428 (אֵלֶּה)","strongs_def":"{these or those}","kjv_def":"these."},
"H430":{"lemma":"אֱלֹהִים","xlit":"ʼĕlôhîym","pron":"el-o-heem'","derivation":"plural of H433 (אֱלוֹהַּ);","strongs_def":"gods in the ordinary sense; but specifically used (in the plural thus, especially with the article) of the supreme God; occasionally applied by way of deference to magistrates; and sometimes as a superlative","kjv_def":"angels, [idiom] exceeding, God (gods) (-dess, -ly), [idiom] (very) great, judges, [idiom] mighty."},
"H431":{"lemma":"אֲלוּ","xlit":"ʼălûw","pron":"al-oo'","derivation":"(Aramaic) probably prolonged from H412 (אֵל);","strongs_def":"lo!","kjv_def":"behold."},
"H432":{"lemma":"אִלּוּ","xlit":"ʼillûw","pron":"il-loo'","derivation":"probably from H408 (אַל);","strongs_def":"nay, i.e. (softened) if","kjv_def":"but if, yea though."},
"H433":{"lemma":"אֱלוֹהַּ","xlit":"ʼĕlôwahh","pron":"el-o'-ah","derivation":"rarely (shortened) אֱלֹהַּ; probably prolonged (emphatic) from H410 (אֵל);","strongs_def":"a deity or the Deity","kjv_def":"God, god. See H430 (אֱלֹהִים)."},
"H434":{"lemma":"אֱלוּל","xlit":"ʼĕlûwl","pron":"el-ool'","derivation":"for H457 (אֱלִיל);","strongs_def":"good for nothing","kjv_def":"thing of nought."},
"H435":{"lemma":"אֱלוּל","xlit":"ʼĔlûwl","pron":"el-ool'","derivation":"probably of foreign derivation;","strongs_def":"Elul, the sixth Jewish month","kjv_def":"Elul."},
"H436":{"lemma":"אֵלוֹן","xlit":"ʼêlôwn","pron":"ay-lone'","derivation":"prolonged from H352 (אַיִל);","strongs_def":"an oak or other strong tree","kjv_def":"plain. See also H356 (אֵילוֹן)."},
"H437":{"lemma":"אַלּוֹן","xlit":"ʼallôwn","pron":"al-lone'","derivation":"a variation of H436 (אֵלוֹן)","strongs_def":"{an oak or other strong tree}","kjv_def":"oak."},
"H438":{"lemma":"אַלּוֹן","xlit":"ʼAllôwn","pron":"al-lone'","derivation":"the same as H437 (אַלּוֹן);","strongs_def":"Allon, an Israelite, also a place in Palestine","kjv_def":"Allon."},
"H439":{"lemma":"אַלּוֹן בָּכוּת","xlit":"ʼAllôwn Bâkûwth","pron":"al-lone' baw-kooth'","derivation":"from H437 (אַלּוֹן) and a variation of H1068 (בְּכִית); oak of weeping;","strongs_def":"Allon-Bakuth, a monumental tree","kjv_def":"Allon-bachuth."},
"H440":{"lemma":"אֵלוֹנִי","xlit":"ʼÊlôwnîy","pron":"ay-lo-nee'","derivation":"or rather (shortened) אֵלֹנִי; patron from H438 (אַלּוֹן);","strongs_def":"an Elonite or descendant (collectively) of Elon","kjv_def":"Elonites."},
"H441":{"lemma":"אַלּוּף","xlit":"ʼallûwph","pron":"al-loof'","derivation":"or (shortened) אַלֻּף; from H502 (אָלַף);","strongs_def":"familiar; a friend, also gentle; hence, a bullock (as being tame; applied, although masculine, to a cow); and so, a chieftain (as notable, like neat cattle)","kjv_def":"captain, duke, (chief) friend, governor, guide, ox."},
"H442":{"lemma":"אָלוּשׁ","xlit":"ʼÂlûwsh","pron":"aw-loosh'","derivation":"of uncertain derivation;","strongs_def":"Alush, a place in the Desert","kjv_def":"Alush."},
"H443":{"lemma":"אֶלְזָבָד","xlit":"ʼElzâbâd","pron":"el-zaw-bawd'","derivation":"from H410 (אֵל) and H2064 (זָבַד); God has bestowed;","strongs_def":"Elzabad, the name of two Israelites","kjv_def":"Elzabad."},
"H444":{"lemma":"אָלַח","xlit":"ʼâlach","pron":"aw-lakh'","derivation":"a primitive root;","strongs_def":"to muddle, i.e. (figuratively and intransitive) to turn (morally) corrupt","kjv_def":"become filthy."},
"H445":{"lemma":"אֶלְחָנָן","xlit":"ʼElchânân","pron":"el-khaw-nawn'","derivation":"from H410 (אֵל) and H2603 (חָנַן); God (is) gracious;","strongs_def":"Elchanan, an Israelite","kjv_def":"Elkanan."},
"H446":{"lemma":"אֱלִיאָב","xlit":"ʼĔlîyʼâb","pron":"el-ee-awb'","derivation":"from H410 (אֵל) and H1 (אָב); God of (his) father;","strongs_def":"Eliab, the name of six Israelites","kjv_def":"Eliab."},
"H447":{"lemma":"אֱלִיאֵל","xlit":"ʼĔlîyʼêl","pron":"el-ee-ale'","derivation":"from H410 (אֵל) repeated; God of (his) God;","strongs_def":"Eliel, the name of nine Israelites","kjv_def":"Eliel."},
"H448":{"lemma":"אֱלִיאָתָה","xlit":"ʼĔlîyʼâthâh","pron":"el-ee-aw-thaw'","derivation":"or (contraction) אֱלִיָּתָה; from H410 (אֵל) and H225 (אוּת); God of (his) consent;","strongs_def":"Eliathah, an Israelite","kjv_def":"Eliathah."},
"H449":{"lemma":"אֱלִידָד","xlit":"ʼĔlîydâd","pron":"el-ee-dawd'","derivation":"from the same as H419 (אֶלְדָּד); God of (his) love;","strongs_def":"Elidad, an Israelite","kjv_def":"Elidad."},
"H450":{"lemma":"אֶלְיָדָע","xlit":"ʼElyâdâʻ","pron":"el-yaw-daw'","derivation":"from H410 (אֵל) and H3045 (יָדַע); God (is) knowing;","strongs_def":"Eljada, the name of two Israelites and of an Aramaean leader","kjv_def":"Eliada."},
"H451":{"lemma":"אַלְיָה","xlit":"ʼalyâh","pron":"al-yaw'","derivation":"from H422 (אָלָה) (in the original sense of strength);","strongs_def":"the stout part, i.e. the fat tail of the Oriental sheep","kjv_def":"rump."},
"H452":{"lemma":"אֵלִיָּה","xlit":"ʼÊlîyâh","pron":"ay-lee-yaw'","derivation":"or prolonged אֵלִיָּהוּ; from H410 (אֵל) and H3050 (יָהּ); God of Jehovah;","strongs_def":"Elijah, the name of the famous prophet and of two other Israelites","kjv_def":"Elijah, Eliah."},
"H453":{"lemma":"אֱלִיהוּ","xlit":"ʼĔlîyhûw","pron":"el-ee-hoo'","derivation":"or (fully) אֱלִיהוּא; from H410 (אֵל) and H1931 (הוּא); God of him;","strongs_def":"Elihu, the name of one of Job's friends, and of three Israelites","kjv_def":"Elihu."},
"H454":{"lemma":"אֶלְיְהוֹעֵינַי","xlit":"ʼElyᵉhôwʻêynay","pron":"el-ye-ho-ay-nah'ee","derivation":"or (shortened) אֶלְיוֹעֵינַי; from H413 (אֵל) and H3068 (יְהֹוָה) and H5869 (עַיִן); towards Jehovah (are) my eyes;","strongs_def":"Eljehoenai or Eljoenai, the name of seven Israelites","kjv_def":"Elihoenai, Elionai."},
"H455":{"lemma":"אֶלְיַחְבָּא","xlit":"ʼElyachbâʼ","pron":"el-yakh-baw'","derivation":"from H410 (אֵל) and H2244 (חָבָא); God will hide;","strongs_def":"Eljachba, an Israelite","kjv_def":"Eliahbah."},
"H456":{"lemma":"אֱלִיחֹרֶף","xlit":"ʼĔlîychôreph","pron":"el-ee-kho'-ref","derivation":"from H410 (אֵל) and H2779 (חֹרֶף); God of autumn;","strongs_def":"Elichoreph, an Israelite","kjv_def":"Elihoreph."},
"H457":{"lemma":"אֱלִיל","xlit":"ʼĕlîyl","pron":"el-eel'","derivation":"apparently from H408 (אַל);","strongs_def":"good for nothing, by anal. vain or vanity; specifically an idol","kjv_def":"idol, no value, thing of nought."},
"H458":{"lemma":"אֱלִימֶלֶךְ","xlit":"ʼĔlîymelek","pron":"el-ee-meh'-lek","derivation":"from H410 (אֵל) and H4428 (מֶלֶךְ); God of (the) king;","strongs_def":"Elimelek, an Israelite","kjv_def":"Elimelech."},
"H459":{"lemma":"אִלֵּין","xlit":"ʼillêyn","pron":"il-lane'","derivation":"(Aramaic) or shorter אִלֵּן; prolonged from H412 (אֵל);","strongs_def":"these","kjv_def":"the, these."},
"H460":{"lemma":"אֶלְיָסָף","xlit":"ʼElyâçâph","pron":"el-yaw-sawf'","derivation":"from H410 (אֵל) and H3254 (יָסַף); God (is) gatherer;","strongs_def":"Eljasaph, the name of two Israelites","kjv_def":"Eliasaph."},
"H461":{"lemma":"אֱלִיעֶזֶר","xlit":"ʼĔlîyʻezer","pron":"el-ee-eh'-zer","derivation":"from H410 (אֵל) and H5828 (עֵזֶר); God of help;","strongs_def":"Eliezer, the name of a Damascene and of ten Israelites","kjv_def":"Eliezer."},
"H462":{"lemma":"אֱלִיעֵינַי","xlit":"ʼĔlîyʻêynay","pron":"el-ee-ay-nah'ee","derivation":"probably contracted for H454 (אֶלְיְהוֹעֵינַי);","strongs_def":"Elienai, an Israelite","kjv_def":"Elienai."},
"H463":{"lemma":"אֱלִיעָם","xlit":"ʼĔlîyʻâm","pron":"el-ee-awm'","derivation":"from H410 (אֵל) and H5971 (עַם); God of (the) people;","strongs_def":"Eliam, an Israelite","kjv_def":"Eliam."},
"H464":{"lemma":"אֱלִיפַז","xlit":"ʼĔlîyphaz","pron":"el-ee-faz'","derivation":"from H410 (אֵל) and H6337 (פָּז); God of gold;","strongs_def":"Eliphaz, the name of one of Job's friends, and of a son of Esau","kjv_def":"Eliphaz."},
"H465":{"lemma":"אֱלִיפָל","xlit":"ʼĔlîyphâl","pron":"el-ee-fawl'","derivation":"from H410 (אֵל) and H6419 (פָּלַל); God of judgment;","strongs_def":"Eliphal, an Israelite","kjv_def":"Eliphal."},
"H466":{"lemma":"אֱלִיפְלֵהוּ","xlit":"ʼĔlîyphᵉlêhûw","pron":"el-ee-fe-lay'-hoo","derivation":"from H410 (אֵל) and H6395 (פָּלָה); God of his distinction;","strongs_def":"Eliphelehu, an Israelite","kjv_def":"Elipheleh."},
"H467":{"lemma":"אֱלִיפֶלֶט","xlit":"ʼĔlîypheleṭ","pron":"el-ee-feh'-let","derivation":"or (shortened) אֱלְפֶּלֶט; from H410 (אֵל) and H6405 (פַּלֵּט); God of deliverance;","strongs_def":"Eliphelet or Elpelet, the name of six Israelites","kjv_def":"Eliphalet, Eliphelet, Elpalet."},
"H468":{"lemma":"אֱלִיצוּר","xlit":"ʼĔlîytsûwr","pron":"el-ee-tsoor'","derivation":"from H410 (אֵל) and H6697 (צוּר); God of (the) rock;","strongs_def":"Elitsur, an Israelite","kjv_def":"Elizur."},
"H469":{"lemma":"אֱלִיצָפָן","xlit":"ʼĔlîytsâphân","pron":"el-ee-tsaw-fawn'","derivation":"or (shortened) אֶלְצָפָן; from H410 (אֵל) and H6845 (צָפַן); God of treasure;","strongs_def":"Elitsaphan or Eltsaphan, an Israelite","kjv_def":"Elizaphan, Elzaphan."},
"H470":{"lemma":"אֱלִיקָא","xlit":"ʼĔlîyqâʼ","pron":"el-ee-kaw'","derivation":"from H410 (אֵל) and H6958 (קוֹא); God of rejection;","strongs_def":"Elika, an Israelite","kjv_def":"Elika."},
"H471":{"lemma":"אֶלְיָקִים","xlit":"ʼElyâqîym","pron":"el-yaw-keem'","derivation":"from H410 (אֵל) and H6965 (קוּם); God of raising;","strongs_def":"Eljakim, the name of four Israelites","kjv_def":"Eliakim."},
"H472":{"lemma":"אֱלִישֶׁבַע","xlit":"ʼĔlîyshebaʻ","pron":"el-ee-sheh'-bah","derivation":"from H410 (אֵל) and H7651 (שֶׁבַע) (in the sense of H7650 (שָׁבַע)); God of (the) oath;","strongs_def":"Elisheba, the wife of Aaron","kjv_def":"Elisheba."},
"H473":{"lemma":"אֱלִישָׁה","xlit":"ʼĔlîyshâh","pron":"el-ee-shaw'","derivation":"probably of foreign derivation;","strongs_def":"Elishah, a son of Javan","kjv_def":"Elishah."},
"H474":{"lemma":"אֱלִישׁוּעַ","xlit":"ʼĔlîyshûwaʻ","pron":"el-ee-shoo'-ah","derivation":"from H410 (אֵל) and H7769 (שׁוּעַ); God of supplication (or of riches);","strongs_def":"Elishua, the son of King David","kjv_def":"Elishua."},
"H475":{"lemma":"אֶלְיָשִׁיב","xlit":"ʼElyâshîyb","pron":"el-yaw-sheeb'","derivation":"from H410 (אֵל) and H7725 (שׁוּב); God will restore;","strongs_def":"Eljashib, the name of six Israelites","kjv_def":"Eliashib."},
"H476":{"lemma":"אֱלִישָׁמָע","xlit":"ʼĔlîyshâmâʻ","pron":"el-ee-shaw-maw'","derivation":"from H410 (אֵל) and H8085 (שָׁמַע); God of hearing;","strongs_def":"Elishama, the name of seven Israelites","kjv_def":"Elishama."},
"H477":{"lemma":"אֱלִישָׁע","xlit":"ʼĔlîyshâʻ","pron":"el-ee-shaw'","derivation":"contracted for H474 (אֱלִישׁוּעַ).;","strongs_def":"Elisha, the famous prophet","kjv_def":"Elisha."},
"H478":{"lemma":"אֱלִישָׁפָט","xlit":"ʼĔlîyshâphâṭ","pron":"el-ee-shaw-fawt'","derivation":"from H410 (אֵל) and H8199 (שָׁפַט); God of judgment;","strongs_def":"Elishaphat, an Israelite","kjv_def":"Elishaphat."},
"H479":{"lemma":"אִלֵּךְ","xlit":"ʼillêk","pron":"il-lake'","derivation":"(Aramaic) prolonged from H412 (אֵל);","strongs_def":"these","kjv_def":"these, those."},
"H480":{"lemma":"אַלְלַי","xlit":"ʼalᵉlay","pron":"al-le-lah'ee","derivation":"by reduplication from H421 (אָלָה);","strongs_def":"alas!","kjv_def":"woe."},
"H481":{"lemma":"אָלַם","xlit":"ʼâlam","pron":"aw-lam'","derivation":"a primitive root;","strongs_def":"to tie fast; hence (of the mouth) to be tongue-tied","kjv_def":"bind, be dumb, put to silence."},
"H482":{"lemma":"אֵלֶם","xlit":"ʼêlem","pron":"ay'-lem","derivation":"from H481 (אָלַם);","strongs_def":"silence (i.e. mute justice)","kjv_def":"congregation. Compare H3128 (יוֹנַת אֵלֶם רְחֹקִים)."},
"H483":{"lemma":"אִלֵּם","xlit":"ʼillêm","pron":"il-lame'","derivation":"from H481 (אָלַם);","strongs_def":"speechless","kjv_def":"dumb (man)."},
"H484":{"lemma":"אַלְמֻגִּים","xlit":"ʼalmuggîym","pron":"al-moog-gheem'","derivation":"probably of foreign derivation (used thus only in the plural);","strongs_def":"almug (i.e. probably sandle-wood) sticks","kjv_def":"almug trees. Compare H418 (אַלְגּוּמִּים)."},
"H485":{"lemma":"אֲלֻמָּה","xlit":"ʼălummâh","pron":"al-oom-maw'","derivation":"or (masculine) אָלֻם; passive participle of H481 (אָלַם);","strongs_def":"something bound; a sheaf","kjv_def":"sheaf."},
"H486":{"lemma":"אַלְמוֹדָד","xlit":"ʼAlmôwdâd","pron":"al-mo-dawd'","derivation":"probably of foreign derivation","strongs_def":"Almodad, a son of Joktan","kjv_def":"Almodad."},
"H487":{"lemma":"אַלַּמֶּלֶךְ","xlit":"ʼAllammelek","pron":"al-lam-meh'-lek","derivation":"from H427 (אַלָּה) and H4428 (מֶלֶךְ); oak of (the) king;","strongs_def":"Allammelek, a place in Palestine","kjv_def":"Alammelech."},
"H488":{"lemma":"אַלְמָן","xlit":"ʼalmân","pron":"al-mawn'","derivation":"prolonged from H481 (אָלַם) in the sense of bereavement;","strongs_def":"discarded (as a divorced person)","kjv_def":"forsaken."},
"H489":{"lemma":"אַלְמֹן","xlit":"ʼalmôn","pron":"al-mone'","derivation":"from H481 (אָלַם) as in H488 (אַלְמָן);","strongs_def":"bereavement","kjv_def":"widowhood."},
"H490":{"lemma":"אַלְמָנָה","xlit":"ʼalmânâh","pron":"al-maw-naw'","derivation":"feminine of H488 (אַלְמָן);","strongs_def":"a widow; also a desolate place","kjv_def":"desolate house (palace), widow."},
"H491":{"lemma":"אַלְמָנוּת","xlit":"ʼalmânûwth","pron":"al-maw-nooth'","derivation":"feminine of H488 (אַלְמָן);","strongs_def":"concrete, a widow; abstract, widowhood","kjv_def":"widow, widowhood."},
"H492":{"lemma":"אַלְמֹנִי","xlit":"ʼalmônîy","pron":"al-mo-nee'","derivation":"from H489 (אַלְמֹן) in the sense of concealment;","strongs_def":"some one (i.e. so and so, without giving the name of the person or place)","kjv_def":"one, and such."},
"H493":{"lemma":"אֶלְנַעַם","xlit":"ʼElnaʻam","pron":"el-nah'-am","derivation":"from H410 (אֵל) and H5276 (נָעֵם); God (is his) delight;","strongs_def":"Elnaam, an Israelite","kjv_def":"Elnaam."},
"H494":{"lemma":"אֶלְנָתָן","xlit":"ʼElnâthân","pron":"el-naw-thawn'","derivation":"from H410 (אֵל) and H5414 (נָתַן); God (is the) giver;","strongs_def":"Elnathan, the name of four Israelites","kjv_def":"Elnathan."},
"H495":{"lemma":"אֶלָּסָר","xlit":"ʼEllâçâr","pron":"el-law-sawr'","derivation":"probably of foreign derivation;","strongs_def":"Ellasar, an early country of Asia","kjv_def":"Ellasar."},
"H496":{"lemma":"אֶלְעָד","xlit":"ʼElʻâd","pron":"el-awd'","derivation":"from H410 (אֵל) and H5749 (עוּד); God has testified;","strongs_def":"Elad, an Israelite","kjv_def":"Elead."},
"H497":{"lemma":"אֶלְעָדָה","xlit":"ʼElʻâdâh","pron":"el-aw-daw'","derivation":"from H410 (אֵל) and H5710 (עָדָה); God has decked;","strongs_def":"Eladah, an Israelite","kjv_def":"Eladah."},
"H498":{"lemma":"אֶלְעוּזַי","xlit":"ʼElʻûwzay","pron":"el-oo-zah'ee","derivation":"from H410 (אֵל) and H5756 (עוּז) (in the sense of H5797 (עֹז)); God (is) defensive;","strongs_def":"Eluzai, an Israelite","kjv_def":"Eluzai."},
"H499":{"lemma":"אֶלְעָזָר","xlit":"ʼElʻâzâr","pron":"el-aw-zawr'","derivation":"from H410 (אֵל) and H5826 (עָזַר); God (is) helper;","strongs_def":"Elazar, the name of seven Israelites","kjv_def":"Eleazar."},
"H500":{"lemma":"אֶלְעָלֵא","xlit":"ʼElʻâlêʼ","pron":"el-aw-lay'","derivation":"or (more properly) אֶלְעָלֵה; from H410 (אֵל) and H5927 (עָלָה); God (is) going up;","strongs_def":"Elale or Elaleh, a place east of the Jordan","kjv_def":"Elealeh."},
"H501":{"lemma":"אֶלְעָשָׂה","xlit":"ʼElʻâsâh","pron":"el-aw-saw'","derivation":"from H410 (אֵל) and H6213 (עָשָׂה); God has made;","strongs_def":"Elasah, the name of four Israelites","kjv_def":"Elasah, Eleasah."},
"H502":{"lemma":"אָלַף","xlit":"ʼâlaph","pron":"aw-lof'","derivation":"a primitive root, to associate with;","strongs_def":"hence, to learn (and causatively to teach)","kjv_def":"learn, teach, utter."},
"H503":{"lemma":"אָלַף","xlit":"ʼâlaph","pron":"aw-laf'","derivation":"denominative from H505 (אֶלֶף);","strongs_def":"causative, to make athousandfold","kjv_def":"bring forth thousands."},
"H504":{"lemma":"אֶלֶף","xlit":"ʼeleph","pron":"eh'-lef","derivation":"from H502 (אָלַף);","strongs_def":"a family; also (from the sense of yoking or taming) an ox or cow","kjv_def":"family, kine, oxen."},
"H505":{"lemma":"אֶלֶף","xlit":"ʼeleph","pron":"eh'-lef","derivation":"prop, the same as H504 (אֶלֶף);","strongs_def":"hence (the ox's head being the first letter of the alphabet, and this eventually used as a numeral) a thousand","kjv_def":"thousand."},
"H506":{"lemma":"אֲלַף","xlit":"ʼălaph","pron":"al-af'","derivation":"(Aramaic) or אֶלֶף; (Aramaic), corresponding to H505 (אֶלֶף)","strongs_def":"{hence (the ox's head being the first letter of the alphabet, and this eventually used as a numeral) a thousand}","kjv_def":"thousand."},
"H507":{"lemma":"אֶלֶף","xlit":"ʼEleph","pron":"eh'-lef","derivation":"the same as H505 (אֶלֶף);","strongs_def":"Eleph, a place in Palestine","kjv_def":"Eleph."},
"H508":{"lemma":"אֶלְפַּעַל","xlit":"ʼElpaʻal","pron":"el-pah'-al","derivation":"from H410 (אֵל) and H6466 (פָּעַל); God (is) act;","strongs_def":"Elpaal, an Israelite","kjv_def":"Elpaal."},
"H509":{"lemma":"אָלַץ","xlit":"ʼâlats","pron":"aw-lats'","derivation":"a primitive root;","strongs_def":"to press","kjv_def":"urge."},
"H510":{"lemma":"אַלְקוּם","xlit":"ʼalqûwm","pron":"al-koom'","derivation":"probably from H408 (אַל) and H6965 (קוּם);","strongs_def":"a non-rising (i.e. resistlessness)","kjv_def":"no rising up."},
"H511":{"lemma":"אֶלְקָנָה","xlit":"ʼElqânâh","pron":"el-kaw-naw'","derivation":"from H410 (אֵל) and H7069 (קָנָה); God has obtained;","strongs_def":"Elkanah, the name of several Israelites","kjv_def":"Elkanah."},
"H512":{"lemma":"אֶלְקֹשִׁי","xlit":"ʼElqôshîy","pron":"el-ko-shee'","derivation":"patrial from a name of uncertain derivation;","strongs_def":"an Elkoshite or native of Elkosh","kjv_def":"Elkoshite."},
"H513":{"lemma":"אֶלְתּוֹלַד","xlit":"ʼEltôwlad","pron":"el-to-lad'","derivation":"probably from H410 (אֵל) and a masculine form of H8435 (תּוֹלְדָה) (compare H8434 (תּוֹלָד)); God (is) generator;","strongs_def":"Eltolad, a place in Palestine","kjv_def":"Eltolad."},
"H514":{"lemma":"אֶלְתְּקֵא","xlit":"ʼEltᵉqêʼ","pron":"el-te-kay'","derivation":"or (more properly) אֶלְתְּקֵה; of uncertain derivation;","strongs_def":"Eltekeh or Elteke, a place in Palestine","kjv_def":"Eltekeh."},
"H515":{"lemma":"אֶלְתְּקֹן","xlit":"ʼEltᵉqôn","pron":"el-te-kone'","derivation":"from H410 (אֵל) and H8626 (תָּקַן); God (is) straight;","strongs_def":"Eltekon, a place in Palestine","kjv_def":"Eltekon."},
"H516":{"lemma":"אַל תַּשְׁחֵת","xlit":"ʼAl tashchêth","pron":"al tash-kayth'","derivation":"from H408 (אַל) and H7843 (שָׁחַת);","strongs_def":"'Thou must not destroy'; probably the opening words to a popular song","kjv_def":"Al-taschith."},
"H517":{"lemma":"אֵם","xlit":"ʼêm","pron":"ame","derivation":"a primitive word;","strongs_def":"a mother (as the bond of the family); in a wide sense (both literally and figuratively (like father))","kjv_def":"dam, mother, [idiom] parting."},
"H518":{"lemma":"אִם","xlit":"ʼim","pron":"eem","derivation":"a primitive particle;","strongs_def":"used very widely as demonstrative, lo!; interrogative, whether?; or conditional, if, although; also Oh that!, when; hence, as a negative, not","kjv_def":"(and, can-, doubtless, if, that) (not), [phrase] but, either, [phrase] except, [phrase] more(-over if, than), neither, nevertheless, nor, oh that, or, [phrase] save (only, -ing), seeing, since, sith, [phrase] surely (no more, none, not), though, [phrase] of a truth, [phrase] unless, [phrase] verily, when, whereas, whether, while, [phrase] yet."},
"H519":{"lemma":"אָמָה","xlit":"ʼâmâh","pron":"aw-maw'","derivation":"apparently a primitive word;","strongs_def":"a maidservant or female slave","kjv_def":"(hand-) bondmaid(-woman), maid(-servant)."},
"H520":{"lemma":"אַמָּה","xlit":"ʼammâh","pron":"am-maw'","derivation":"prolonged from H517 (אֵם);","strongs_def":"properly, a mother (i.e. unit of measure, or the fore-arm (below the elbow), i.e. a cubit; also a door-base (as a bond of the entrance)","kjv_def":"cubit, [phrase] hundred (by exchange for H3967 (מֵאָה)), measure, post."},
"H521":{"lemma":"אַמָּה","xlit":"ʼammâh","pron":"am-maw'","derivation":"(Aramaic) corresponding to H520 (אַמָּה)","strongs_def":"{properly, a mother (i.e. unit of measure, or the fore-arm (below the elbow), i.e. a cubit; also a door-base (as a bond of the entrance)}","kjv_def":"cubit."},
"H522":{"lemma":"אַמָּה","xlit":"ʼAmmâh","pron":"am-maw'","derivation":"the same as H520 (אַמָּה);","strongs_def":"Ammah, a hill in Palestine","kjv_def":"Ammah."},
"H523":{"lemma":"אֻמַּה","xlit":"ʼummah","pron":"oom-maw'","derivation":"from the same as H517 (אֵם);","strongs_def":"a collection, i.e. community of persons","kjv_def":"nation, people."},
"H524":{"lemma":"אֻמָּה","xlit":"ʼummâh","pron":"oom-maw'","derivation":"(Aramaic) corresponding to H523 (אֻמַּה)","strongs_def":"{a collection, i.e. community of persons}","kjv_def":"nation."},
"H525":{"lemma":"אָמוֹן","xlit":"ʼâmôwn","pron":"aw-mone'","derivation":"from H539 (אָמַן), probably in the sense of training; like H542 (אָמָן)","strongs_def":"skilled, i.e. an architect","kjv_def":"one brought up."},
"H526":{"lemma":"אָמוֹן","xlit":"ʼÂmôwn","pron":"aw-mone'","derivation":"the same as H525 (אָמוֹן);","strongs_def":"Amon, the name of three Israelites","kjv_def":"Amon."},
"H527":{"lemma":"אָמוֹן","xlit":"ʼâmôwn","pron":"aw-mone'","derivation":"a variation for H1995 (הָמוֹן);","strongs_def":"a throng of people","kjv_def":"multitude."},
"H528":{"lemma":"אָמוֹן","xlit":"ʼÂmôwn","pron":"aw-mone'","derivation":"of Egyptian derivation; used only as an adjunct of H4996 (נֹא)","strongs_def":"Amon (i.e. Ammon or Amn), a deity of Egypt","kjv_def":"multitude, populous."},
"H529":{"lemma":"אֵמוּן","xlit":"ʼêmûwn","pron":"ay-moon'","derivation":"from H539 (אָמַן);","strongs_def":"established, i.e. (figuratively) trusty; also (abstractly) trustworthiness","kjv_def":"faith(-ful), truth."},
"H530":{"lemma":"אֱמוּנָה","xlit":"ʼĕmûwnâh","pron":"em-oo-naw'","derivation":"or (shortened) אֱמֻנָה; feminine of H529 (אֵמוּן);","strongs_def":"literally firmness; figuratively security; morally fidelity","kjv_def":"faith(-ful, -ly, -ness, (man)), set office, stability, steady, truly, truth, verily."},
"H531":{"lemma":"אָמוֹץ","xlit":"ʼÂmôwts","pron":"aw-mohts'","derivation":"from H553 (אָמַץ); strong;","strongs_def":"Amots, an Israelite","kjv_def":"Amoz."},
"H532":{"lemma":"אָמִי","xlit":"ʼÂmîy","pron":"aw-mee'","derivation":"an abbreviated form for H526 (אָמוֹן);","strongs_def":"Ami, an Israelite","kjv_def":"Ami."},
"H533":{"lemma":"אַמִּיץ","xlit":"ʼammîyts","pron":"am-meets'","derivation":"or (shortened) אַמִּץ; from H553 (אָמַץ);","strongs_def":"strong or (abstractly) strength","kjv_def":"courageous, mighty, strong (one)."},
"H534":{"lemma":"אָמִיר","xlit":"ʼâmîyr","pron":"aw-meer'","derivation":"apparently from H559 (אָמַר) (in the sense of self-exaltation);","strongs_def":"a summit (of a tree or mountain","kjv_def":"bough, branch."},
"H535":{"lemma":"אָמַל","xlit":"ʼâmal","pron":"aw-mal'","derivation":"a primitive root;","strongs_def":"to droop; by implication to be sick, to mourn","kjv_def":"languish, be weak, wax feeble."},
"H536":{"lemma":"אֻמְלַל","xlit":"ʼumlal","pron":"oom-lal'","derivation":"from H535 (אָמַל);","strongs_def":"sick","kjv_def":"weak."},
"H537":{"lemma":"אֲמֵלָל","xlit":"ʼămêlâl","pron":"am-ay-lawl'","derivation":"from H535 (אָמַל);","strongs_def":"languid","kjv_def":"feeble."},
"H538":{"lemma":"אֲמָם","xlit":"ʼĂmâm","pron":"am-awm'","derivation":"from H517 (אֵם); gathering-spot;","strongs_def":"Amam, a place in Palestine","kjv_def":"Amam."},
"H539":{"lemma":"אָמַן","xlit":"ʼâman","pron":"aw-man'","derivation":"a primitive root; (Isaiah 30:21; interchangeable with H541 (אָמַן), to go to the right hand)","strongs_def":"properly, to build up or support; to foster as a parent or nurse; figuratively to render (or be) firm or faithful, to trust or believe, to be permanent or quiet; morally to be true or certain;","kjv_def":"hence, assurance, believe, bring up, establish, [phrase] fail, be faithful (of long continuance, stedfast, sure, surely, trusty, verified), nurse, (-ing father), (put), trust, turn to the right."},
"H540":{"lemma":"אֲמַן","xlit":"ʼăman","pron":"am-an'","derivation":"(Aramaic) corresponding to H539 (אָמַן)","strongs_def":"{properly, to build up or support; to foster as a parent or nurse; figuratively to render (or be) firm or faithful, to trust or believe, to be permanent or quiet; morally to be true or certain};","kjv_def":"believe, faithful, sure."},
"H541":{"lemma":"אָמַן","xlit":"ʼâman","pron":"aw-man'","derivation":"denominative from H3225 (יָמִין);","strongs_def":"to take the right hand road","kjv_def":"turn to the right. See H539 (אָמַן)."},
"H542":{"lemma":"אָמָן","xlit":"ʼâmân","pron":"aw-mawn'","derivation":"from H539 (אָמַן) (in the sense of training);","strongs_def":"an expert","kjv_def":"cunning workman."},
"H543":{"lemma":"אָמֵן","xlit":"ʼâmên","pron":"aw-mane'","derivation":"from H539 (אָמַן);","strongs_def":"sure; abstract, faithfulness; adverb, truly","kjv_def":"Amen, so be it, truth."},
"H544":{"lemma":"אֹמֶן","xlit":"ʼômen","pron":"oh-men'","derivation":"from H539 (אָמַן);","strongs_def":"verity","kjv_def":"truth."},
"H545":{"lemma":"אׇמְנָה","xlit":"ʼomnâh","pron":"om-naw'","derivation":"feminine of H544 (אֹמֶן) (in the specific sense of training);","strongs_def":"tutelage","kjv_def":"brought up."},
"H546":{"lemma":"אׇמְנָה","xlit":"ʼomnâh","pron":"om-naw'","derivation":"feminine form of H544 (אֹמֶן) (in its usual sense);","strongs_def":"adverb, surely","kjv_def":"indeed."},
"H547":{"lemma":"אֹמְנָה","xlit":"ʼômᵉnâh","pron":"om-me-naw'","derivation":"feminine active participle of H544 (אֹמֶן) (in the original sense of supporting);","strongs_def":"a column","kjv_def":"pillar."},
"H548":{"lemma":"אֲמָנָה","xlit":"ʼămânâh","pron":"am-aw-naw'","derivation":"feminine of H543 (אָמֵן);","strongs_def":"something fixed, i.e. a covenant. an allowance","kjv_def":"certain portion, sure."},
"H549":{"lemma":"אֲמָנָה","xlit":"ʼĂmânâh","pron":"am-aw-naw'","derivation":"the same as H548 (אֲמָנָה);","strongs_def":"Amanah, a mountain near Damascus","kjv_def":"Amana."},
"H550":{"lemma":"אַמְנוֹן","xlit":"ʼAmnôwn","pron":"am-nohn'","derivation":"or אֲמִינוֹן; from H539 (אָמַן); faithful;","strongs_def":"Amnon (or Aminon), a son of David","kjv_def":"Amnon."},
"H551":{"lemma":"אׇמְנָם","xlit":"ʼomnâm","pron":"om-nawm'","derivation":"adverb from H544 (אֹמֶן);","strongs_def":"verily","kjv_def":"indeed, no doubt, surely, (it is, of a) true(-ly, -th)."},
"H552":{"lemma":"אֻמְנָם","xlit":"ʼumnâm","pron":"oom-nawm'","derivation":"an orthographical variation of H551 (אׇמְנָם)","strongs_def":"{verily}","kjv_def":"in (very) deed; of a surety."},
"H553":{"lemma":"אָמַץ","xlit":"ʼâmats","pron":"aw-mats'","derivation":"a primitive root;","strongs_def":"to be alert, physically (on foot) or mentally (in courage)","kjv_def":"confirm, be courageous (of good courage, stedfastly minded, strong, stronger), establish, fortify, harden, increase, prevail, strengthen (self), make strong (obstinate, speed)."},
"H554":{"lemma":"אָמֹץ","xlit":"ʼâmôts","pron":"aw-mohts'","derivation":"probably from H553 (אָמַץ);","strongs_def":"of a strong color, i.e. red (others fleet)","kjv_def":"bay."},
"H555":{"lemma":"אֹמֶץ","xlit":"ʼômets","pron":"o'-mets","derivation":"from H553 (אָמַץ);","strongs_def":"strength","kjv_def":"stronger."},
"H556":{"lemma":"אַמְצָה","xlit":"ʼamtsâh","pron":"am-tsaw'","derivation":"from H553 (אָמַץ);","strongs_def":"force","kjv_def":"strength."},
"H557":{"lemma":"אַמְצִי","xlit":"ʼAmtsîy","pron":"am-tsee'","derivation":"from H553 (אָמַץ); strong;","strongs_def":"Amtsi, an Israelite","kjv_def":"Amzi."},
"H558":{"lemma":"אֲמַצְיָה","xlit":"ʼĂmatsyâh","pron":"am-ats-yaw'","derivation":"or אֲמַצְיָהוּ; from H553 (אָמַץ) and H3050 (יָהּ); strength of Jah;","strongs_def":"Amatsjah, the name of four Israelites","kjv_def":"Amaziah."},
"H559":{"lemma":"אָמַר","xlit":"ʼâmar","pron":"aw-mar'","derivation":"a primitive root;","strongs_def":"to say (used with great latitude)","kjv_def":"answer, appoint, avouch, bid, boast self, call, certify, challenge, charge, [phrase] (at the, give) command(-ment), commune, consider, declare, demand, [idiom] desire, determine, [idiom] expressly, [idiom] indeed, [idiom] intend, name, [idiom] plainly, promise, publish, report, require, say, speak (against, of), [idiom] still, [idiom] suppose, talk, tell, term, [idiom] that is, [idiom] think, use (speech), utter, [idiom] verily, [idiom] yet."},
"H560":{"lemma":"אֲמַר","xlit":"ʼămar","pron":"am-ar'","derivation":"(Aramaic) corresponding to H559 (אָמַר)","strongs_def":"{to say (used with great latitude)}","kjv_def":"command, declare, say, speak, tell."},
"H561":{"lemma":"אֵמֶר","xlit":"ʼêmer","pron":"ay'-mer","derivation":"from H559 (אָמַר);","strongs_def":"something said","kjv_def":"answer, [idiom] appointed unto him, saying, speech, word."},
"H562":{"lemma":"אֹמֶר","xlit":"ʼômer","pron":"o'-mer","derivation":"the same as H561 (אֵמֶר)","strongs_def":"{something said}","kjv_def":"promise, speech, thing, word."},
"H563":{"lemma":"אִמַּר","xlit":"ʼimmar","pron":"im-mar'","derivation":"(Aramaic) perhaps from H560 (אֲמַר) (in the sense of bringing forth);","strongs_def":"a lamb","kjv_def":"lamb."},
"H564":{"lemma":"אִמֵּר","xlit":"ʼImmêr","pron":"im-mare'","derivation":"from H559 (אָמַר); talkative;","strongs_def":"Immer, the name of five Israelites","kjv_def":"Immer."},
"H565":{"lemma":"אִמְרָה","xlit":"ʼimrâh","pron":"im-raw'","derivation":"or אֶמְרָה; feminine of H561 (אֵמֶר), and meaning the same","strongs_def":"{something said}","kjv_def":"commandment, speech, word."},
"H566":{"lemma":"אִמְרִי","xlit":"ʼImrîy","pron":"im-ree'","derivation":"from H564 (אִמֵּר);","strongs_def":"Imri, the name of two Israelites","kjv_def":"Imri."},
"H567":{"lemma":"אֱמֹרִי","xlit":"ʼĔmôrîy","pron":"em-o-ree'","derivation":"probably a patronymic from an unused name derived from H559 (אָמַר) in the sense of publicity, i.e. prominence; thus, a mountaineer;","strongs_def":"an Emorite, one of the Canaanitish tribes","kjv_def":"Amorite."},
"H568":{"lemma":"אֲמַרְיָה","xlit":"ʼĂmaryâh","pron":"am-ar-yaw'","derivation":"or prolonged אֲמַרְיָהוּ; from H559 (אָמַר) and H3050 (יָהּ); Jah has said (i.e. promised);","strongs_def":"Amarjah, the name of nine Israelites","kjv_def":"Amariah."},
"H569":{"lemma":"אַמְרָפֶל","xlit":"ʼAmrâphel","pron":"am-raw-fel'","derivation":"of uncertain (perhaps foreign) derivation;","strongs_def":"Amraphel, a king of Shinar","kjv_def":"Amraphel."},
"H570":{"lemma":"אֶמֶשׁ","xlit":"ʼemesh","pron":"eh'-mesh","derivation":"time past, i.e.","strongs_def":"yesterday or last night","kjv_def":"former time, yesterday(-night)"},
"H571":{"lemma":"אֶמֶת","xlit":"ʼemeth","pron":"eh'-meth","derivation":"contracted from H539 (אָמַן);","strongs_def":"stability; (figuratively) certainty, truth, trustworthiness","kjv_def":"assured(-ly), establishment, faithful, right, sure, true (-ly, -th), verity."},
"H572":{"lemma":"אַמְתַּחַת","xlit":"ʼamtachath","pron":"am-takh'-ath","derivation":"from H4969 (מָתַח);","strongs_def":"properly, something expansive, i.e. a bag","kjv_def":"sack."},
"H573":{"lemma":"אֲמִתַּי","xlit":"ʼĂmittay","pron":"am-it-tah'ee","derivation":"from H571 (אֶמֶת); veracious;","strongs_def":"Amittai, an Israelite","kjv_def":"Amittai."},
"H574":{"lemma":"אֵמְתָּנִי","xlit":"ʼêmᵉtânîy","pron":"em-taw-nee'","derivation":"(Aramaic) from a root corresponding to that of H4975 (מֹתֶן);","strongs_def":"well-loined (i.e. burly) or mighty","kjv_def":"terrible."},
"H575":{"lemma":"אָן","xlit":"ʼân","pron":"awn","derivation":"or אָנָה; contracted from H370 (אַיִן);","strongs_def":"where?; hence, whither?, when?; also hither and thither","kjv_def":"[phrase] any (no) whither, now, where, whither(-soever)."},
"H576":{"lemma":"אֲנָא","xlit":"ʼănâʼ","pron":"an-aw'","derivation":"(Aramaic) or אֲנָה; (Aramaic), corresponding to H589 (אֲנִי)lemma אֲנַי second vowel, corrected to אֲנִי;","strongs_def":"I","kjv_def":"I, as for me."},
"H577":{"lemma":"אָנָּא","xlit":"ʼânnâʼ","pron":"awn-naw'","derivation":"or אָנָּה; apparent contracted from H160 (אַהֲבָה) and H4994 (נָא);","strongs_def":"oh now!","kjv_def":"I (me) beseech (pray) thee, O."},
"H578":{"lemma":"אָנָה","xlit":"ʼânâh","pron":"aw-naw'","derivation":"a primitive root;","strongs_def":"to groan","kjv_def":"lament, mourn."},
"H579":{"lemma":"אָנָה","xlit":"ʼânâh","pron":"aw-naw'","derivation":"a primitive root (perhaps rather identical with H578 (אָנָה) through the idea of contraction in anguish);","strongs_def":"to approach; hence, to meet in various senses","kjv_def":"befall, deliver, happen, seek a quarrel."},
"H580":{"lemma":"אֲנוּ","xlit":"ʼănûw","pron":"an-oo'","derivation":"contracted for H587 (אֲנַחְנוּ);","strongs_def":"we","kjv_def":"we."},
"H581":{"lemma":"אִנּוּן","xlit":"ʼinnûwn","pron":"in-noon'","derivation":"(Aramaic) or (feminine) אִנִּיןlemma אֵנִּין first vowel, corrected to אִנִּין; (Aramaic), corresponding to H1992 (הֵם);","strongs_def":"they","kjv_def":"[idiom] are, them, these."},
"H582":{"lemma":"אֱנוֹשׁ","xlit":"ʼĕnôwsh","pron":"en-oshe'","derivation":"from H605 (אָנַשׁ); properly, a mortal (and thus differing from the more dignified H120 (אָדָם)); hence,","strongs_def":"a man in general (singly or collectively)","kjv_def":"another, [idiom] (blood-) thirsty, certain, chap(-man); divers, fellow, [idiom] in the flower of their age, husband, (certain, mortal) man, people, person, servant, some ([idiom] of them), [phrase] stranger, those, [phrase] their trade. It is often unexpressed in the English versions, especially when used in apposition with another word. Compare H376 (אִישׁ)."},
"H583":{"lemma":"אֱנוֹשׁ","xlit":"ʼĔnôwsh","pron":"en-ohsh'","derivation":"the same as H582 (אֱנוֹשׁ);","strongs_def":"Enosh, a son of Seth","kjv_def":"Enos."},
"H584":{"lemma":"אָנַח","xlit":"ʼânach","pron":"aw-nakh'","derivation":"a primitive root;","strongs_def":"to sigh","kjv_def":"groan, mourn, sigh."},
"H585":{"lemma":"אֲנָחָה","xlit":"ʼănâchâh","pron":"an-aw-khaw'","derivation":"from H585 (אֲנָחָה);","strongs_def":"sighing","kjv_def":"groaning, mourn, sigh."},
"H586":{"lemma":"אֲנַחְנָא","xlit":"ʼănachnâʼ","pron":"an-akh'-naw","derivation":"(Aramaic) or אֲנַחְנָה; (Aramaic), corresponding to H587 (אֲנַחְנוּ);","strongs_def":"we","kjv_def":"we."},
"H587":{"lemma":"אֲנַחְנוּ","xlit":"ʼănachnûw","pron":"an-akh'-noo","derivation":"apparently from H595 (אָנֹכִי);","strongs_def":"we","kjv_def":"ourselves, us, we."},
"H588":{"lemma":"אֲנָחֲרָת","xlit":"ʼĂnâchărâth","pron":"an-aw-kha-rawth'","derivation":"probably from the same root as H5170 (נַחַר); a gorge or narrow pass;","strongs_def":"Anacharath, a place in Palestine","kjv_def":"Anaharath."},
"H589":{"lemma":"אֲנִי","xlit":"ʼănîy","pron":"an-ee'","derivation":"contracted from H595 (אָנֹכִי);","strongs_def":"I","kjv_def":"I, (as for) me, mine, myself, we, [idiom] which, [idiom] who."},
"H590":{"lemma":"אֳנִי","xlit":"ʼŏnîy","pron":"on-ee'","derivation":"probably from H579 (אָנָה) (in the sense of conveyance);","strongs_def":"-a ship or (collectively) a fleet","kjv_def":"galley, navy (of ships)."},
"H591":{"lemma":"אֳנִיָּה","xlit":"ʼŏnîyâh","pron":"on-ee-yaw'","derivation":"feminine of H590 (אֳנִי);","strongs_def":"a ship","kjv_def":"ship(-men)."},
"H592":{"lemma":"אֲנִיָּה","xlit":"ʼănîyâh","pron":"an-ee-yaw'","derivation":"from H578 (אָנָה);","strongs_def":"groaning","kjv_def":"lamentation, sorrow."},
"H593":{"lemma":"אֲנִיעָם","xlit":"ʼĂnîyʻâm","pron":"an-ee-awm'","derivation":"from H578 (אָנָה) and H5971 (עַם); groaning of (the) people;","strongs_def":"Aniam, an Israelite","kjv_def":"Aniam."},
"H594":{"lemma":"אֲנָךְ","xlit":"ʼănâk","pron":"an-awk'","derivation":"probably from an unused root meaning to be narrow;","strongs_def":"according to most a plumb-line, and to others a hook","kjv_def":"plumb-line"},
"H595":{"lemma":"אָנֹכִי","xlit":"ʼânôkîy","pron":"aw-no-kee'","derivation":"sometimes, ; a primitive pronoun;","strongs_def":"I","kjv_def":"I, me, [idiom] which."},
"H596":{"lemma":"אָנַן","xlit":"ʼânan","pron":"aw-nan'","derivation":"a primitive root;","strongs_def":"to mourn, i.e. complain","kjv_def":"complain."},
"H597":{"lemma":"אָנַס","xlit":"ʼânaç","pron":"aw-nas'","strongs_def":"to insist","kjv_def":"compel."},
"H598":{"lemma":"אֲנַס","xlit":"ʼănaç","pron":"an-as'","derivation":"(Aramaic) corresponding to H597 (אָנַס);","strongs_def":"figuratively, to distress","kjv_def":"trouble."},
"H599":{"lemma":"אָנַף","xlit":"ʼânaph","pron":"aw-naf'","derivation":"a primitive root;","strongs_def":"to breathe hard, i.e. be enraged","kjv_def":"be angry (displeased)."},
"H600":{"lemma":"אֲנַף","xlit":"ʼănaph","pron":"an-af'","derivation":"(Aramaic) corresponding to H639 (אַף) (only in the plural as a singular);","strongs_def":"the face","kjv_def":"face, visage."},
"H601":{"lemma":"אֲנָפָה","xlit":"ʼănâphâh","pron":"an-aw-faw'","derivation":"from H599 (אָנַף);","strongs_def":"an unclean bird, perhaps the parrot (from its irascibility)","kjv_def":"heron."},
"H602":{"lemma":"אָנַק","xlit":"ʼânaq","pron":"aw-nak'","derivation":"a primitive root;","strongs_def":"to shriek","kjv_def":"cry, groan."},
"H603":{"lemma":"אֲנָקָה","xlit":"ʼănâqâh","pron":"an-aw-kaw'","derivation":"from H602 (אָנַק);","strongs_def":"shrieking","kjv_def":"crying out, groaning, sighing."},
"H604":{"lemma":"אֲנָקָה","xlit":"ʼănâqâh","pron":"an-aw-kaw'","derivation":"the same as H603 (אֲנָקָה);","strongs_def":"some kind of lizard, probably the gecko (from its wail)","kjv_def":"ferret."},
"H605":{"lemma":"אָנַשׁ","xlit":"ʼânash","pron":"aw-nash'","derivation":"a primitive root;","strongs_def":"to be frail, feeble, or (figuratively) melancholy","kjv_def":"desperate(-ly wicked), incurable, sick, woeful."},
"H606":{"lemma":"אֱנָשׁ","xlit":"ʼĕnâsh","pron":"en-awsh'","derivation":"(Aramaic) or אֱנַשׁ; (Aramaic), corresponding to H582 (אֱנוֹשׁ);","strongs_def":"a man","kjv_def":"man, [phrase] whosoever."},
"H607":{"lemma":"אַנְתָּה","xlit":"ʼantâh","pron":"an-taw'","derivation":"(Aramaic) corresponding to H859 (אַתָּה);","strongs_def":"thou","kjv_def":"as for thee, thou."},
"H608":{"lemma":"אַנְתּוּן","xlit":"ʼantûwn","pron":"an-toon'","derivation":"(Aramaic) plural of H607 (אַנְתָּה);","strongs_def":"ye","kjv_def":"ye."},
"H609":{"lemma":"אָסָא","xlit":"ʼÂçâʼ","pron":"aw-saw'","derivation":"of uncertain derivation;","strongs_def":"Asa, the name of a king and of a Levite","kjv_def":"Asa."},
"H610":{"lemma":"אָסוּךְ","xlit":"ʼâçûwk","pron":"aw-sook'","derivation":"from H5480 (סוּךְ);","strongs_def":"anointed, i.e. an oil-flask","kjv_def":"pot."},
"H611":{"lemma":"אָסוֹן","xlit":"ʼâçôwn","pron":"aws-sone'","derivation":"of uncertain derivation;","strongs_def":"hurt","kjv_def":"mischief."},
"H612":{"lemma":"אֵסוּר","xlit":"ʼêçûwr","pron":"ay-soor'","derivation":"from H631 (אָסַר);","strongs_def":"a bond (especially manacles of a prisoner)","kjv_def":"band, [phrase] prison."},
"H613":{"lemma":"אֱסוּר","xlit":"ʼĕçûwr","pron":"es-oor'","derivation":"(Aramaic) corresponding to H612 (אֵסוּר)","strongs_def":"{a bond (especially manacles of a prisoner)}","kjv_def":"band, imprisonment."},
"H614":{"lemma":"אָסִיף","xlit":"ʼâçîyph","pron":"aw-seef'","derivation":"or אָסִף; from H622 (אָסַף);","strongs_def":"gathered, i.e. (abstractly) a gathering in of crops","kjv_def":"ingathering."},
"H615":{"lemma":"אָסִיר","xlit":"ʼâçîyr","pron":"aw-sere'","derivation":"from H631 (אָסַר);","strongs_def":"bound, i.e. a captive","kjv_def":"(those which are) bound, prisoner."},
"H616":{"lemma":"אַסִּיר","xlit":"ʼaççîyr","pron":"as-sere'","derivation":"for H615 (אָסִיר)","strongs_def":"{bound, i.e. a captive}","kjv_def":"prisoner."},
"H617":{"lemma":"אַסִּיר","xlit":"ʼAççîyr","pron":"as-sere'","derivation":"the same as H616 (אַסִּיר); prisoner;","strongs_def":"Assir, the name of two Israelites","kjv_def":"Assir."},
"H618":{"lemma":"אָסָם","xlit":"ʼâçâm","pron":"aw-sawm'","derivation":"from an unused root meaning to heap together;","strongs_def":"a storehouse (only in the plural)","kjv_def":"barn, storehouse."},
"H619":{"lemma":"אַסְנָה","xlit":"ʼAçnâh","pron":"as-naw'","derivation":"of uncertain derivation;","strongs_def":"Asnah, one of the Nethinim","kjv_def":"Asnah."},
"H620":{"lemma":"אׇסְנַפַּר","xlit":"ʼOçnappar","pron":"os-nap-par'","derivation":"of foreign derivation;","strongs_def":"Osnappar, an Assyrian king","kjv_def":"Asnapper."},
"H621":{"lemma":"אָסְנַת","xlit":"ʼÂçᵉnath","pron":"aw-se-nath'","derivation":"of Egyptian derivation;","strongs_def":"Asenath, the wife of Joseph","kjv_def":"Asenath."},
"H622":{"lemma":"אָסַף","xlit":"ʼâçaph","pron":"aw-saf'","derivation":"a primitive root;","strongs_def":"to gather for any purpose; hence, to receive, take away, i.e. remove (destroy, leave behind, put up, restore, etc.)","kjv_def":"assemble, bring, consume, destroy, felch, gather (in, together, up again), [idiom] generally, get (him), lose, put all together, receive, recover (another from leprosy), (be) rereward, [idiom] surely, take (away, into, up), [idiom] utterly, withdraw."},
"H623":{"lemma":"אָסָף","xlit":"ʼÂçâph","pron":"aw-sawf'","derivation":"from H622 (אָסַף); collector;","strongs_def":"Asaph, the name of three Israelites, and of the family of the first","kjv_def":"Asaph."},
"H624":{"lemma":"אָסֻף","xlit":"ʼâçuph","pron":"aw-soof'","derivation":"passive participle of H622 (אָסַף);","strongs_def":"collected (only in the plural), i.e. a collection of offerings)","kjv_def":"threshold, Asuppim."},
"H625":{"lemma":"אֹסֶף","xlit":"ʼôçeph","pron":"o'-sef","derivation":"from H622 (אָסַף);","strongs_def":"a collection (of fruits)","kjv_def":"gathering."},
"H626":{"lemma":"אֲסֵפָה","xlit":"ʼăçêphâh","pron":"as-ay-faw'","derivation":"from H622 (אָסַף);","strongs_def":"a collection of people (only adverbial)","kjv_def":"[idiom] together."},
"H627":{"lemma":"אֲסֻפָּה","xlit":"ʼăçuppâh","pron":"as-up-paw'","derivation":"feminine of H624 (אָסֻף);","strongs_def":"a collection of (learned) men (only in the plural)","kjv_def":"assembly."},
"H628":{"lemma":"אֲסְפְּסֻף","xlit":"ʼăçpᵉçuph","pron":"as-pes-oof'","derivation":"by reduplication from H624 (אָסֻף);","strongs_def":"gathered up together, i.e. a promiscuous assemblage (of people)","kjv_def":"mixt multitude."},
"H629":{"lemma":"אׇסְפַּרְנָא","xlit":"ʼoçparnâʼ","pron":"os-par-naw'","derivation":"(Aramaic) of Persian derivation;","strongs_def":"diligently","kjv_def":"fast, forthwith, speed(-ily)."},
"H630":{"lemma":"אַסְפָּתָא","xlit":"ʼAçpâthâʼ","pron":"as-paw-thaw'","derivation":"of Persian derivation;","strongs_def":"Aspatha, a son of Haman","kjv_def":"Aspatha."},
"H631":{"lemma":"אָסַר","xlit":"ʼâçar","pron":"aw-sar'","derivation":"a primitive root;","strongs_def":"to yoke or hitch; by analogy, to fasten in any sense, to join battle","kjv_def":"bind, fast, gird, harness, hold, keep, make ready, order, prepare, prison(-er), put in bonds, set in array, tie."},
"H632":{"lemma":"אֱסָר","xlit":"ʼĕçâr","pron":"es-sawr'","derivation":"or אִסָּר; from H631 (אָסַר);","strongs_def":"an obligation or vow (of abstinence)","kjv_def":"binding, bond."},
"H633":{"lemma":"אֱסָר","xlit":"ʼĕçâr","pron":"es-sawr'","derivation":"(Aramaic) corresponding to H632 (אֱסָר) in a legal sense;","strongs_def":"an interdict","kjv_def":"decree."},
"H634":{"lemma":"אֵסַר־חַדּוֹן","xlit":"ʼÊçar-Chaddôwn","pron":"ay-sar' Chad-dohn'","derivation":"of foreign derivation;","strongs_def":"Esar-chaddon, an Assyrian king","kjv_def":"Esar-haddon."},
"H635":{"lemma":"אֶסְתֵּר","xlit":"ʼEçtêr","pron":"es-tare'","derivation":"of Persian derivation;","strongs_def":"Ester, the Jewish heroine","kjv_def":"Esther."},
"H636":{"lemma":"אָע","xlit":"ʼâʻ","pron":"aw","derivation":"(Aramaic) corresponding to H6086 (עֵץ);","strongs_def":"a tree or wood","kjv_def":"timber, wood."},
"H637":{"lemma":"אַף","xlit":"ʼaph","pron":"af","derivation":"a primitive particle;","strongs_def":"meaning accession (used as an adverb or conjunction); also or yea; adversatively though","kjv_def":"also, [phrase] although, and (furthermore, yet), but, even, [phrase] how much less (more, rather than), moreover, with, yea."},
"H638":{"lemma":"אַף","xlit":"ʼaph","pron":"af","derivation":"(Aramaic) corresponding to H637 (אַף)","strongs_def":"{meaning accession (used as an adverb or conjunction); also or yea; adversatively though}","kjv_def":"also."},
"H639":{"lemma":"אַף","xlit":"ʼaph","pron":"af","derivation":"from H599 (אָנַף);","strongs_def":"properly, the nose or nostril; hence, the face, and occasionally a person; also (from the rapid breathing in passion) ire","kjv_def":"anger(-gry), [phrase] before, countenance, face, [phrase] forebearing, forehead, [phrase] (long-) suffering, nose, nostril, snout, [idiom] worthy, wrath."},
"H640":{"lemma":"אָפַד","xlit":"ʼâphad","pron":"aw-fad'","derivation":"a primitive root (rather a denominative from H646 (אֵפוֹד));","strongs_def":"to gird on (the ephod)","kjv_def":"bind, gird."},
"H641":{"lemma":"אֵפֹד","xlit":"ʼÊphôd","pron":"ay-fode'","derivation":"the same as H646 (אֵפוֹד) shortened;","strongs_def":"Ephod, an Israelite","kjv_def":"Ephod."},
"H642":{"lemma":"אֵפֻדָּה","xlit":"ʼêphuddâh","pron":"ay-food-daw'","derivation":"feminine of H646 (אֵפוֹד);","strongs_def":"a girding on (of the ephod); hence, generally, a plating (of metal)","kjv_def":"ephod, ornament."},
"H643":{"lemma":"אַפֶּדֶן","xlit":"ʼappeden","pron":"ap-peh'-den","derivation":"apparently of foreign derivation;","strongs_def":"a pavilion or palace-tent","kjv_def":"palace."},
"H644":{"lemma":"אָפָה","xlit":"ʼâphâh","pron":"aw-faw'","derivation":"a primitive root;","strongs_def":"to cook, especially to bake","kjv_def":"bake(-r, (-meats))."},
"H645":{"lemma":"אֵפוֹ","xlit":"ʼêphôw","pron":"ay-fo'","derivation":"or אֵפוֹא; from H6311 (פֹּה);","strongs_def":"strictly a demonstrative particle, here; but used of time, now or then","kjv_def":"here, now, where?"},
"H646":{"lemma":"אֵפוֹד","xlit":"ʼêphôwd","pron":"ay-fode'","derivation":"rarely אֵפֹד; probably of foreign derivation","strongs_def":"a girdle; specifically the ephod or highpriest's shoulder-piece; also generally, an image","kjv_def":"ephod."},
"H647":{"lemma":"אֲפִיחַ","xlit":"ʼĂphîyach","pron":"af-ee'-akh","derivation":"perhaps from H6315 (פּוּחַ); breeze;","strongs_def":"Aphiach, an Israelite","kjv_def":"Aphiah."},
"H648":{"lemma":"אָפִיל","xlit":"ʼâphîyl","pron":"aw-feel'","derivation":"from the same as H651 (אָפֵל) (in the sense of weakness);","strongs_def":"unripe","kjv_def":"not grown up."},
"H649":{"lemma":"אַפַּיִם","xlit":"ʼAppayim","pron":"ap-pah'-yim","derivation":"dual of H639 (אַף); two nostrils;","strongs_def":"Appajim, an Israelite","kjv_def":"Appaim."},
"H650":{"lemma":"אָפִיק","xlit":"ʼâphîyq","pron":"aw-feek'","derivation":"from H622 (אָסַף);","strongs_def":"properly, containing, i.e. a tube; also a bed or valley of astream; also a strong thing or a hero","kjv_def":"brook, channel, mighty, river, [phrase] scale, stream, strong piece."},
"H651":{"lemma":"אָפֵל","xlit":"ʼâphêl","pron":"aw-fale'","derivation":"from an unused root meaning to set as the sun;","strongs_def":"dusky","kjv_def":"very dark."},
"H652":{"lemma":"אֹפֶל","xlit":"ʼôphel","pron":"o'fel","derivation":"from the same as H651 (אָפֵל);","strongs_def":"dusk","kjv_def":"darkness, obscurity, privily."},
"H653":{"lemma":"אֲפֵלָה","xlit":"ʼăphêlâh","pron":"af-ay-law'","derivation":"feminine of H651 (אָפֵל);","strongs_def":"duskiness, figuratively, misfortune; concrete, concealment","kjv_def":"dark, darkness, gloominess, [idiom] thick."},
"H654":{"lemma":"אֶפְלָל","xlit":"ʼEphlâl","pron":"ef-lawl'","derivation":"from H6419 (פָּלַל); judge;","strongs_def":"Ephlal, an Israelite","kjv_def":"Ephlal."},
"H655":{"lemma":"אֹפֶן","xlit":"ʼôphen","pron":"o'-fen","derivation":"from an unused root meaning to revolve;","strongs_def":"a turn, i.e. a season","kjv_def":"[phrase] fitly."},
"H656":{"lemma":"אָפֵס","xlit":"ʼâphêç","pron":"aw-face'","derivation":"a primitive root;","strongs_def":"to disappear, i.e. cease","kjv_def":"be clean gone (at an end, brought to nought), fail."},
"H657":{"lemma":"אֶפֶס","xlit":"ʼepheç","pron":"eh'-fes","derivation":"from H656 (אָפֵס); (sometimes like H6466 (פָּעַל))","strongs_def":"cessation, i.e. an end (especially of the earth); often used adverb, no further; also the ankle (in the dual), as being the extremity of the leg or foot","kjv_def":"ankle, but (only), end, howbeit, less than nothing, nevertheless (where), no, none (beside), not (any, -withstanding), thing of nought, save(-ing), there, uttermost part, want, without (cause)."},
"H658":{"lemma":"אֶפֶס דַּמִּים","xlit":"ʼEpheç Dammîym","pron":"eh'-fes dam-meem'","derivation":"from H657 (אֶפֶס) and the plural of H1818 (דָּם); boundary of blood-drops;","strongs_def":"Ephes-Dammim, a place in Palestine","kjv_def":"Ephes-dammim."},
"H659":{"lemma":"אֵפַע","xlit":"ʼêphaʻ","pron":"eh'-fah","derivation":"from an unused root probably meaning to breathe;","strongs_def":"properly, a breath, i.e. nothing","kjv_def":"of nought."},
"H660":{"lemma":"אֶפְעֶה","xlit":"ʼephʻeh","pron":"ef-eh'","derivation":"from H659 (אֵפַע) (in the sense of hissing);","strongs_def":"an asp or other venomous serpent","kjv_def":"viper."},
"H661":{"lemma":"אָפַף","xlit":"ʼâphaph","pron":"aw-faf'","derivation":"a primitive root;","strongs_def":"to surround","kjv_def":"compass."},
"H662":{"lemma":"אָפַק","xlit":"ʼâphaq","pron":"aw-fak'","derivation":"a primitive root;","strongs_def":"to contain, i.e. (reflex.) abstain","kjv_def":"force (oneself), restrain."},
"H663":{"lemma":"אֲפֵק","xlit":"ʼĂphêq","pron":"af-ake'","derivation":"or אֲפִיק; from H662 (אָפַק) (in the sense of strength); fortress;","strongs_def":"Aphek (or Aphik), the name of three places in Palestine","kjv_def":"Aphek, Aphik."},
"H664":{"lemma":"אֲפֵקָה","xlit":"ʼĂphêqâh","pron":"af-ay-kaw'","derivation":"feminine of H663 (אֲפֵק); fortress;","strongs_def":"Aphekah, a place in Palestine","kjv_def":"Aphekah."},
"H665":{"lemma":"אֵפֶר","xlit":"ʼêpher","pron":"ay'-fer","derivation":"from an unused root meaning to bestrew;","strongs_def":"ashes","kjv_def":"ashes."},
"H666":{"lemma":"אֲפֵר","xlit":"ʼăphêr","pron":"af-ayr'","derivation":"from the same as H665 (אֵפֶר) (in the sense of covering);","strongs_def":"a turban","kjv_def":"ashes."},
"H667":{"lemma":"אֶפְרֹחַ","xlit":"ʼephrôach","pron":"ef-ro'-akh","derivation":"from H6524 (פָּרַח) (in the sense of bursting the shell);","strongs_def":"the brood of a bird","kjv_def":"young (one)."},
"H668":{"lemma":"אַפִּרְיוֹן","xlit":"ʼappiryôwn","pron":"ap-pir-yone'","derivation":"probably of Egyptian derivation;","strongs_def":"a palanquin","kjv_def":"chariot."},
"H669":{"lemma":"אֶפְרַיִם","xlit":"ʼEphrayim","pron":"ef-rah'-yim","derivation":"dual of masculine form of H672 (אֶפְרָת); double fruit;","strongs_def":"Ephrajim, a son of Joseph; also the tribe descended from him, and its territory","kjv_def":"Ephraim, Ephraimites."},
"H670":{"lemma":"אֲפָרְסַי","xlit":"ʼĂphârᵉçay","pron":"af-aw-re-sah'ee","derivation":"(Aramaic) of foreign origin (only in the plural);","strongs_def":"an Apherasite or inhabitant of an unknown region of Assyria","kjv_def":"Apharsite."},
"H671":{"lemma":"אֲפַרְסְכַי","xlit":"ʼĂpharçᵉkay","pron":"af-ar-sek-ah'ee","derivation":"(Aramaic) or אֲפַרְסַתְכַי; (Aramaic), of foreign origin (only in the plural);","strongs_def":"an Apharsekite or Apharsathkite, an unknown Assyrian tribe","kjv_def":"Apharsachites, Apharasthchites."},
"H672":{"lemma":"אֶפְרָת","xlit":"ʼEphrâth","pron":"ef-rawth'","derivation":"or אֶפְרָתָה; from H6509 (פָּרָה); fruitfulness;","strongs_def":"Ephrath, another name for Bethlehem;","kjv_def":"once (Psalm 132:6) perhaps for Ephraim; also of an Israelitish woman; Ephrath, Ephratah."},
"H673":{"lemma":"אֶפְרָתִי","xlit":"ʼEphrâthîy","pron":"ef-rawth-ee'","derivation":"patrial from H672 (אֶפְרָת);","strongs_def":"an Ephrathite or an Ephraimite","kjv_def":"Ephraimite, Ephrathite."},
"H674":{"lemma":"אַפְּתֹם","xlit":"ʼappᵉthôm","pron":"ap-pe-thome'","derivation":"(Aramaic) of Persian origin;","strongs_def":"revenue; others at the last","kjv_def":"revenue."},
"H675":{"lemma":"אֶצְבּוֹן","xlit":"ʼEtsbôwn","pron":"ets-bone'","derivation":"or אֶצְבֹּן; of uncertain derivation;","strongs_def":"Etsbon, the name of two Israelites","kjv_def":"Ezbon."},
"H676":{"lemma":"אֶצְבַּע","xlit":"ʼetsbaʻ","pron":"ets-bah'","derivation":"from the same as H6648 (צֶבַע) (in the sense of grasping);","strongs_def":"something to sieze with, i.e. a finger; by analogy, a toe","kjv_def":"finger, toe."},
"H677":{"lemma":"אֶצְבַּע","xlit":"ʼetsbaʻ","pron":"ets-bah'","derivation":"(Aramaic) corresponding to H676 (אֶצְבַּע)","strongs_def":"{something to sieze with, i.e. a finger; by analogy, a toe}","kjv_def":"finger, toe."},
"H678":{"lemma":"אָצִיל","xlit":"ʼâtsîyl","pron":"aw-tseel'","derivation":"from H680 (אָצַל) (in its secondary sense of separation);","strongs_def":"an extremity (Isaiah 41:9), also a noble","kjv_def":"chief man, noble."},
"H679":{"lemma":"אַצִּיל","xlit":"ʼatstsîyl","pron":"ats-tseel'","derivation":"from H680 (אָצַל) (in its primary sense of uniting);","strongs_def":"a joint of the hand (i.e. knuckle); also (according to some) a party-wall (Ezekiel 41:8)","kjv_def":"(arm) hole, great."},
"H680":{"lemma":"אָצַל","xlit":"ʼâtsal","pron":"aw-tsal'","derivation":"a primitive root; properly, to join; used only as a denominative from H681 (אֵצֶל);","strongs_def":"to separate; hence, to select, refuse, contract","kjv_def":"keep, reserve, straiten, take."},
"H681":{"lemma":"אֵצֶל","xlit":"ʼêtsel","pron":"ay'-tsel","derivation":"from H680 (אָצַל) (in the sense of joining);","strongs_def":"a side; (as a preposition) near","kjv_def":"at, (hard) by, (from) (beside), near (unto), toward, with. See also H1018 (בֵּית הָאֵצֶל)."},
"H682":{"lemma":"אָצֵל","xlit":"ʼÂtsêl","pron":"aw-tsale'","derivation":"from H680 (אָצַל); noble;","strongs_def":"Atsel, the name of an Israelite, and of a place in Palestine","kjv_def":"Azal, Azel."},
"H683":{"lemma":"אֲצַלְיָהוּ","xlit":"ʼĂtsalyâhûw","pron":"ats-al-yaw'-hoo","derivation":"from H680 (אָצַל) and H3050 (יָהּ) prolonged; Jah has reserved;","strongs_def":"Atsaljah, an Israelite","kjv_def":"Azaliah."},
"H684":{"lemma":"אֹצֶם","xlit":"ʼÔtsem","pron":"o'-tsem","derivation":"from an unused root probably meaning to be strong; strength (i.e. strong);","strongs_def":"Otsem, the name of two Israelites","kjv_def":"Ozem."},
"H685":{"lemma":"אֶצְעָדָה","xlit":"ʼetsʻâdâh","pron":"ets-aw-daw'","derivation":"a variation from H6807 (צְעָדָה);","strongs_def":"properly, a step-chain; by analogy, a bracelet","kjv_def":"bracelet, chain."},
"H686":{"lemma":"אָצַר","xlit":"ʼâtsar","pron":"aw-tsar'","derivation":"a primitive root;","strongs_def":"to store up","kjv_def":"(lay up in) store, [phrase] (make) treasure(-r)."},
"H687":{"lemma":"אֶצֶר","xlit":"ʼEtser","pron":"ay'-tser","derivation":"from H686 (אָצַר); treasure;","strongs_def":"Etser, an Idumaean","kjv_def":"Ezer."},
"H688":{"lemma":"אֶקְדָּח","xlit":"ʼeqdâch","pron":"ek-dawkh'","derivation":"from H6916 (קִדָּה);","strongs_def":"burning, i.e. a carbuncle or other fiery gem","kjv_def":"carbuncle."},
"H689":{"lemma":"אַקּוֹ","xlit":"ʼaqqôw","pron":"ak-ko'","derivation":"probably from H602 (אָנַק);","strongs_def":"slender, i.e. the ibex","kjv_def":"wild goat."},
"H690":{"lemma":"אֲרָא","xlit":"ʼărâʼ","pron":"ar-aw'","derivation":"probably for H738 (אֲרִי); lion;","strongs_def":"Ara, an Israelite","kjv_def":"Ara."},
"H691":{"lemma":"אֶרְאֵל","xlit":"ʼerʼêl","pron":"er-ale'","derivation":"probably for H739 (אֲרִיאֵל);","strongs_def":"a hero (collectively)","kjv_def":"valiant one."},
"H692":{"lemma":"אַרְאֵלִי","xlit":"ʼArʼêlîy","pron":"ar-ay-lee'","derivation":"from H691 (אֶרְאֵל); heroic;","strongs_def":"Areli (or an Arelite, collectively), an Israelite and his descendants","kjv_def":"Areli, Arelites."},
"H693":{"lemma":"אָרַב","xlit":"ʼârab","pron":"aw-rab'","derivation":"a primitive root;","strongs_def":"to lurk","kjv_def":"(lie in) ambush(-ment), lay (lie in) wait."},
"H694":{"lemma":"אֲרָב","xlit":"ʼĂrâb","pron":"ar-awb'","derivation":"from H693 (אָרַב); ambush;","strongs_def":"Arab, a place in Palestine","kjv_def":"Arab."},
"H695":{"lemma":"אֶרֶב","xlit":"ʼereb","pron":"eh'-reb","derivation":"from H693 (אָרַב);","strongs_def":"ambuscade","kjv_def":"den, lie in wait."},
"H696":{"lemma":"אֹרֶב","xlit":"ʼôreb","pron":"o'-reb","derivation":"the same as H695 (אֶרֶב)","strongs_def":"{ambuscade}","kjv_def":"wait."},
"H697":{"lemma":"אַרְבֶּה","xlit":"ʼarbeh","pron":"ar-beh'","derivation":"from H7235 (רָבָה);","strongs_def":"a locust (from its rapid increase)","kjv_def":"grasshopper, locust."},
"H698":{"lemma":"אׇרֳבָה","xlit":"ʼorŏbâh","pron":"or-ob-aw'","derivation":"feminine of H696 (אֹרֶב) (only in the plural);","strongs_def":"ambuscades","kjv_def":"spoils."},
"H699":{"lemma":"אֲרֻבָּה","xlit":"ʼărubbâh","pron":"ar-oob-baw'","derivation":"feminine participle passive of H693 (אָרַב) (as if for lurking);","strongs_def":"a lattice; (by implication) a window, dovecot (because of the pigeon-holes), chimney (with its apertures for smoke), sluice (with openings for water)","kjv_def":"chimney, window."},
"H700":{"lemma":"אֲרֻבּוֹת","xlit":"ʼĂrubbôwth","pron":"ar-oob-both","derivation":"plural of H699 (אֲרֻבָּה);","strongs_def":"Arubboth, a place in Palestine","kjv_def":"Aruboth."},
"H701":{"lemma":"אַרְבִּי","xlit":"ʼArbîy","pron":"ar-bee'","derivation":"patrial from H694 (אֲרָב);","strongs_def":"an Arbite or native of Arab","kjv_def":"Arbite."},
"H702":{"lemma":"אַרְבַּע","xlit":"ʼarbaʻ","pron":"ar-bah'","derivation":"masculine אַרְבָּעָה; from H7251 (רָבַע);","strongs_def":"four","kjv_def":"four."},
"H703":{"lemma":"אַרְבַּע","xlit":"ʼarbaʻ","pron":"ar-bah'","derivation":"(Aramaic) corresponding to H702 (אַרְבַּע)","strongs_def":"{four}","kjv_def":"four."},
"H704":{"lemma":"אַרְבַּע","xlit":"ʼArbaʻ","pron":"ar-bah'","derivation":"the same as H702 (אַרְבַּע);","strongs_def":"Arba, one of the Anakim","kjv_def":"Arba."},
"H705":{"lemma":"אַרְבָּעִים","xlit":"ʼarbâʻîym","pron":"ar-baw-eem'","derivation":"multiple of H702 (אַרְבַּע);","strongs_def":"forty","kjv_def":"-forty."},
"H706":{"lemma":"אַרְבַּעְתַּיִם","xlit":"ʼarbaʻtayim","pron":"ar-bah-tah'-yim","derivation":"dual of H702 (אַרְבַּע);","strongs_def":"fourfold","kjv_def":"fourfold."},
"H707":{"lemma":"אָרַג","xlit":"ʼârag","pron":"aw-rag'","derivation":"a primitive root;","strongs_def":"to plait or weave","kjv_def":"weaver(-r)."},
"H708":{"lemma":"אֶרֶג","xlit":"ʼereg","pron":"eh'-reg","derivation":"from H707 (אָרַג);","strongs_def":"a weaving; a braid; also a shuttle","kjv_def":"beam, weaver's shuttle."},
"H709":{"lemma":"אַרְגֹּב","xlit":"ʼArgôb","pron":"ar-gobe'","derivation":"from the same as H7263 (רֶגֶב); stony;","strongs_def":"Argob, a district of Palestine","kjv_def":"Argob."},
"H710":{"lemma":"אַרְגְּוָן","xlit":"ʼargᵉvân","pron":"arg-ev-awn'","derivation":"a variation for H713 (אַרְגָּמָן);","strongs_def":"purple","kjv_def":"purple."},
"H711":{"lemma":"אַרְגְּוָן","xlit":"ʼargᵉvân","pron":"arg-ev-awn'","derivation":"(Aramaic) corresponding to H710 (אַרְגְּוָן)","strongs_def":"{purple}","kjv_def":"purple."},
"H712":{"lemma":"אַרְגָּז","xlit":"ʼargâz","pron":"ar-gawz'","derivation":"perhaps from H7264 (רָגַז) (in the sense of being suspended),","strongs_def":"a box (as a pannier)","kjv_def":"coffer."},
"H713":{"lemma":"אַרְגָּמָן","xlit":"ʼargâmân","pron":"ar-gaw-mawn'","derivation":"of foreign origin;","strongs_def":"purple (the color or the dyed stuff)","kjv_def":"purple."},
"H714":{"lemma":"אַרְדְּ","xlit":"ʼArd","pron":"ard","derivation":"from an unused root probably meaning to wander; fugitive;","strongs_def":"Ard, the name of two Israelites","kjv_def":"Ard."},
"H715":{"lemma":"אַרְדּוֹן","xlit":"ʼArdôwn","pron":"ar-dohn'","derivation":"from the same as H714 (אַרְדְּ); roaming;","strongs_def":"Ardon, an Israelite","kjv_def":"Ardon."},
"H716":{"lemma":"אַרְדִּי","xlit":"ʼArdîy","pron":"ar-dee","derivation":"patronymic from H714 (אַרְדְּ);","strongs_def":"an Ardite (collectively) or descendant of Ard","kjv_def":"Ardites."},
"H717":{"lemma":"אָרָה","xlit":"ʼârâh","pron":"aw-raw'","derivation":"a primitive root;","strongs_def":"to pluck","kjv_def":"gather, pluck."},
"H718":{"lemma":"אֲרוּ","xlit":"ʼărûw","pron":"ar-oo'","derivation":"(Aramaic) probably akin to H431 (אֲלוּ);","strongs_def":"lo!","kjv_def":"behold, lo."},
"H719":{"lemma":"אַרְוַד","xlit":"ʼArvad","pron":"ar-vad'","derivation":"probably from H7300 (רוּד); a refuge for the roving;","strongs_def":"Arvad, an island-city of Palestine","kjv_def":"Arvad."},
"H720":{"lemma":"אֲרוֹד","xlit":"ʼĂrôwd","pron":"ar-ode'","derivation":"an orthographical variation of H719 (אַרְוַד); fugitive;","strongs_def":"Arod, an Israelite","kjv_def":"Arod."},
"H721":{"lemma":"אַרְוָדִי","xlit":"ʼArvâdîy","pron":"ar-vaw-dee'","derivation":"patrial from H719 (אַרְוַד);","strongs_def":"an Arvadite or citizen of Arvad","kjv_def":"Arvadite."},
"H722":{"lemma":"אֲרוֹדִי","xlit":"ʼĂrôwdîy","pron":"ar-o-dee'","derivation":"patronymic from H721 (אַרְוָדִי);","strongs_def":"an Arodite or descendant of Arod","kjv_def":"Arodi, Arodites."},
"H723":{"lemma":"אֻרְוָה","xlit":"ʼurvâh","pron":"oor-vaw'","derivation":"or אֲרָיָה; from H717 (אָרָה) (in the sense of feeding);","strongs_def":"a herding-place for an animal","kjv_def":"stall."},
"H724":{"lemma":"אֲרוּכָה","xlit":"ʼărûwkâh","pron":"ar-oo-kaw'","derivation":"or אֲרֻכָה; feminine passive participle of H748 (אָרַךְ) (in the sense of restoring to soundness);","strongs_def":"wholeness (literally or figuratively)","kjv_def":"health, made up, perfected."},
"H725":{"lemma":"אֲרוּמָה","xlit":"ʼĂrûwmâh","pron":"ar-oo-maw'","derivation":"a variation of H7316 (רוּמָה); height;","strongs_def":"Arumah, a place in Palestine","kjv_def":"Arumah."},
"H726":{"lemma":"אֲרוֹמִי","xlit":"ʼĂrôwmîy","pron":"ar-o-mee'","derivation":"a clerical error for H130 (אֱדֹמִי);","strongs_def":"an Edomite (as in the margin)","kjv_def":"Syrian."},
"H727":{"lemma":"אָרוֹן","xlit":"ʼârôwn","pron":"aw-rone'","derivation":"or אָרֹן; from H717 (אָרָה) (in the sense of gathering);","strongs_def":"a box","kjv_def":"ark, chest, coffin."},
"H728":{"lemma":"אֲרַוְנָה","xlit":"ʼĂravnâh","pron":"ar-av-naw'","derivation":"or (by transposition) אוֹרְנָה; or אַרְנִיָּהlemma אַנִיָּה missing resh, corrected to אַרְנִיָּה; all by orthographical variation for H771 (אׇרְנָן);","strongs_def":"Aravnah (or Arnijah or Ornah), a Jebusite","kjv_def":"Araunah."},
"H729":{"lemma":"אָרַז","xlit":"ʼâraz","pron":"aw-raz'","derivation":"a primitive root; to be firm; used only in the passive participle as a denominative from H730 (אֶרֶז);","strongs_def":"of cedar","kjv_def":"made of cedar."},
"H730":{"lemma":"אֶרֶז","xlit":"ʼerez","pron":"eh-rez'","derivation":"from H729 (אָרַז);","strongs_def":"a cedar tree (from the tenacity of its roots)","kjv_def":"cedar (tree)."},
"H731":{"lemma":"אַרְזָה","xlit":"ʼarzâh","pron":"ar-zaw'","derivation":"feminine of H730 (אֶרֶז);","strongs_def":"cedar wainscoating","kjv_def":"cedar work."},
"H732":{"lemma":"אָרַח","xlit":"ʼârach","pron":"aw-rakh'","derivation":"a primitive root;","strongs_def":"to travel","kjv_def":"go, wayfaring (man)."},
"H733":{"lemma":"אָרַח","xlit":"ʼÂrach","pron":"aw-rakh'","derivation":"from H732 (אָרַח); way faring;","strongs_def":"Arach, the name of three Israelites","kjv_def":"Arah."},
"H734":{"lemma":"אֹרַח","xlit":"ʼôrach","pron":"o'-rakh","derivation":"from H732 (אָרַח);","strongs_def":"a well-trodden road (literally or figuratively); also a caravan","kjv_def":"manner, path, race, rank, traveller, troop, (by-, high-) way."},
"H735":{"lemma":"אֹרַח","xlit":"ʼôrach","pron":"o'-rakh","derivation":"(Aramaic) corresponding to H734 (אֹרַח);","strongs_def":"a road","kjv_def":"way."},
"H736":{"lemma":"אֹרְחָה","xlit":"ʼôrᵉchâh","pron":"o-rekh-aw'","derivation":"feminine active participle of H732 (אָרַח);","strongs_def":"a caravan","kjv_def":"(travelling) company."},
"H737":{"lemma":"אֲרֻחָה","xlit":"ʼăruchâh","pron":"ar-oo-khaw'","derivation":"feminine passive participle of H732 (אָרַח) (in the sense of appointing);","strongs_def":"a ration of food","kjv_def":"allowance, diet, dinner, victuals."},
"H738":{"lemma":"אֲרִי","xlit":"ʼărîy","pron":"ar-ee'","derivation":"or (prolonged) []; from H717 (אָרָה) (in the sense of violence);","strongs_def":"a lion","kjv_def":"(young) lion, [phrase] pierce (from the margin)."},
"H739":{"lemma":"אֲרִיאֵל","xlit":"ʼărîyʼêl","pron":"ar-ee-ale'","derivation":"or אֲרִאֵל; from H738 (אֲרִי) and H410 (אֵל);","strongs_def":"lion of God; i.e. heroic","kjv_def":"lionlike men."},
"H740":{"lemma":"אֲרִיאֵל","xlit":"ʼĂrîyʼêl","pron":"ar-ee-ale'","derivation":"the same as H739 (אֲרִיאֵל);","strongs_def":"Ariel, a symbolical name for Jerusalem, also the name of an Israelite","kjv_def":"Ariel."},
"H741":{"lemma":"אֲרִאֵיל","xlit":"ʼăriʼêyl","pron":"ar-ee-ale'","derivation":"either by transposition for H739 (אֲרִיאֵל) or, more probably, an orthographical variation for H2025 (הַרְאֵל);","strongs_def":"the altar of the temple","kjv_def":"altar."},
"H742":{"lemma":"אֲרִידַי","xlit":"ʼĂrîyday","pron":"ar-ee-dah'-ee","derivation":"of Persian origin;","strongs_def":"Aridai, a son of Haman","kjv_def":"Aridai."},
"H743":{"lemma":"אֲרִידָתָא","xlit":"ʼĂrîydâthâʼ","pron":"ar-ee-daw-thaw'","derivation":"of Persian origin;","strongs_def":"Aridatha, a son of Haman","kjv_def":"Aridatha."},
"H744":{"lemma":"אַרְיֵה","xlit":"ʼaryêh","pron":"ar-yay'","derivation":"(Aramaic) corresponding to H738 (אֲרִי)","strongs_def":"{a lion}","kjv_def":"lion."},
"H745":{"lemma":"אַרְיֵה","xlit":"ʼAryêh","pron":"ar-yay'","derivation":"the same as H738 (אֲרִי); lion;","strongs_def":"Arjeh, an Israelite","kjv_def":"Arieh."},
"H746":{"lemma":"אֲרְיוֹךְ","xlit":"ʼĂryôwk","pron":"ar-yoke'","derivation":"of foreign origin;","strongs_def":"Arjok, the name of two Babylonians","kjv_def":"Arioch."},
"H747":{"lemma":"אֲרִיסַי","xlit":"ʼĂrîyçay","pron":"ar-ee-sah'-ee","derivation":"of Persian origin;","strongs_def":"Arisai, a son of Haman","kjv_def":"Arisai."},
"H748":{"lemma":"אָרַךְ","xlit":"ʼârak","pron":"aw-rak'","derivation":"a primitive root;","strongs_def":"to be (causative, make) long (literally or figuratively)","kjv_def":"defer, draw out, lengthen, (be, become, make, pro-) long, [phrase] (out-, over-) live, tarry (long)."},
"H749":{"lemma":"אֲרַךְ","xlit":"ʼărak","pron":"ar-ak'","derivation":"(Aramaic) properly, corresponding to H748 (אָרַךְ), but used only in the sense of reaching to a given point;","strongs_def":"to suit","kjv_def":"be meet."},
"H750":{"lemma":"אָרֵךְ","xlit":"ʼârêk","pron":"aw-rake'","derivation":"from H748 (אָרַךְ);","strongs_def":"long","kjv_def":"long(-suffering, -winged), patient, slow (to anger)."},
"H751":{"lemma":"אֶרֶךְ","xlit":"ʼErek","pron":"eh'-rek","derivation":"from H748 (אָרַךְ); length;","strongs_def":"Erek, a place in Babylon","kjv_def":"Erech."},
"H752":{"lemma":"אָרֹךְ","xlit":"ʼârôk","pron":"aw-roke'","derivation":"from H748 (אָרַךְ);","strongs_def":"long","kjv_def":"long."},
"H753":{"lemma":"אֹרֶךְ","xlit":"ʼôrek","pron":"o'rek'","derivation":"from H748 (אָרַךְ);","strongs_def":"length","kjv_def":"[phrase] forever, length, long."},
"H754":{"lemma":"אַרְכָּא","xlit":"ʼarkâʼ","pron":"ar-kaw'","derivation":"(Aramaic) or אַרְכָּה; (Aramaic), from H749 (אֲרַךְ);","strongs_def":"length","kjv_def":"lengthening, prolonged."},
"H755":{"lemma":"אַרְכֻבָה","xlit":"ʼarkubâh","pron":"ar-koo-baw'","derivation":"(Aramaic) from an unused root corresponding to H7392 (רָכַב) (in the sense of bending the knee);","strongs_def":"the knee","kjv_def":"knee."},
"H756":{"lemma":"אַרְכְּוַי","xlit":"ʼArkᵉvay","pron":"ar-kev-ah'ee","derivation":"(Aramaic) patrial from H751 (אֶרֶךְ);","strongs_def":"an Arkevite (collectively) or native of Erek","kjv_def":"Archevite."},
"H757":{"lemma":"אַרְכִּי","xlit":"ʼArkîy","pron":"ar-kee'","derivation":"patrial from another place (in Palestine) of similar name with H751 (אֶרֶךְ);","strongs_def":"an Arkite or native of Erek","kjv_def":"Archi, Archite."},
"H758":{"lemma":"אֲרָם","xlit":"ʼĂrâm","pron":"arawm'","derivation":"from the same as H759 (אַרְמוֹן); the highland;","strongs_def":"Aram or Syria, and its inhabitants; also the name of the son of Shem, a grandson of Nahor, and of an Israelite","kjv_def":"Aram, Mesopotamia, Syria, Syrians."},
"H759":{"lemma":"אַרְמוֹן","xlit":"ʼarmôwn","pron":"ar-mone'","derivation":"from an unused root (meaning to be elevated);","strongs_def":"a citadel (from its height)","kjv_def":"castle, palace. Compare H2038 (הַרְמוֹן)."},
"H760":{"lemma":"אֲרַם צוֹבָה","xlit":"ʼĂram Tsôwbâh","pron":"ar-am' tso-baw'","derivation":"from H758 (אֲרָם) and H6678 (צוֹבָא);","strongs_def":"Aram of Tsoba (or Coele-Syria)","kjv_def":"Aram-zobah."},
"H761":{"lemma":"אֲרַמִּי","xlit":"ʼĂrammîy","pron":"ar-am-mee'","derivation":"patrial from H758 (אֲרָם);","strongs_def":"an Aramite or Aramaean","kjv_def":"Syrian, Aramitess."},
"H762":{"lemma":"אֲרָמִית","xlit":"ʼĂrâmîyth","pron":"ar-aw-meeth'","derivation":"feminine of H761 (אֲרַמִּי);","strongs_def":"(only adverbial) in Aramean","kjv_def":"in the Syrian language (tongue), in Syriac."},
"H763":{"lemma":"אֲרַם נַהֲרַיִם","xlit":"ʼĂram Nahărayim","pron":"ar-am' nah-har-ah'-yim","derivation":"from H758 (אֲרָם) and the dual of H5104 (נָהָר);","strongs_def":"Aram of (the) two rivers (Euphrates and Tigris) or Mesopotamia","kjv_def":"Aham-naharaim, Mesopotamia."},
"H764":{"lemma":"אַרְמֹנִי","xlit":"ʼArmônîy","pron":"ar-mo-nee'","derivation":"from H759 (אַרְמוֹן); palatial;","strongs_def":"Armoni, an Israelite","kjv_def":"Armoni."},
"H765":{"lemma":"אֲרָן","xlit":"ʼĂrân","pron":"ar-awn'","derivation":"from H7442 (רָנַן); stridulous;","strongs_def":"Aran, an Edomite","kjv_def":"Aran."},
"H766":{"lemma":"אֹרֶן","xlit":"ʼôren","pron":"o'-ren","derivation":"from the same as H765 (אֲרָן) (in the sense of strength);","strongs_def":"the ash tree (from its toughness)","kjv_def":"ash."},
"H767":{"lemma":"אֹרֶן","xlit":"ʼÔren","pron":"o'-ren","derivation":"the same as H766 (אֹרֶן);","strongs_def":"Oren, an Israelite","kjv_def":"Oren."},
"H768":{"lemma":"אַרְנֶבֶת","xlit":"ʼarnebeth","pron":"ar-neh'-beth","derivation":"of uncertain derivation;","strongs_def":"the hare","kjv_def":"hare."},
"H769":{"lemma":"אַרְנוֹן","xlit":"ʼArnôwn","pron":"ar-nohn'","derivation":"or אַרְנֹןlemma אַרְנן missing vowel, corrected to אַרְנֹן; from H7442 (רָנַן); a brawling stream;","strongs_def":"the Arnon, a river east of the Jordan, also its territory","kjv_def":"Arnon."},
"H770":{"lemma":"אַרְנָן","xlit":"ʼArnân","pron":"ar-nawn'","derivation":"probably from the same as H769 (אַרְנוֹן); noisy;","strongs_def":"Arnan, an Israelite","kjv_def":"Arnan."},
"H771":{"lemma":"אׇרְנָן","xlit":"ʼOrnân","pron":"or-nawn'","derivation":"probably from H766 (אֹרֶן); strong;","strongs_def":"Ornan, a Jebusite","kjv_def":"Ornan. See H728 (אֲרַוְנָה)."},
"H772":{"lemma":"אֲרַע","xlit":"ʼăraʻ","pron":"ar-ah'","derivation":"(Aramaic) corresponding to H776 (אֶרֶץ);","strongs_def":"the earth; by implication (figuratively) low","kjv_def":"earth, interior."},
"H773":{"lemma":"אַרְעִית","xlit":"ʼarʻîyth","pron":"arh-eeth'","derivation":"(Aramaic) feminine of H772 (אֲרַע);","strongs_def":"the bottom","kjv_def":"bottom."},
"H774":{"lemma":"אַרְפָּד","xlit":"ʼArpâd","pron":"ar-pawd'","derivation":"from H7502 (רָפַד); spread out;","strongs_def":"Arpad, a place in Syria","kjv_def":"Arpad, Arphad."},
"H775":{"lemma":"אַרְפַּכְשַׁד","xlit":"ʼArpakshad","pron":"ar-pak-shad'","derivation":"probably of foreign origin;","strongs_def":"Arpakshad, a son of Noah; also the region settled by him","kjv_def":"Arphaxad."},
"H776":{"lemma":"אֶרֶץ","xlit":"ʼerets","pron":"eh'-rets","derivation":"from an unused root probably meaning to be firm;","strongs_def":"the earth (at large, or partitively a land)","kjv_def":"[idiom] common, country, earth, field, ground, land, [idiom] natins, way, [phrase] wilderness, world."},
"H777":{"lemma":"אַרְצָא","xlit":"ʼartsâʼ","pron":"ar-tsaw'","derivation":"from H776 (אֶרֶץ); earthiness;","strongs_def":"Artsa, an Israelite","kjv_def":"Arza."},
"H778":{"lemma":"אֲרַק","xlit":"ʼăraq","pron":"ar-ak'","derivation":"(Aramaic) by transmutation for H772 (אֲרַע);","strongs_def":"the earth","kjv_def":"earth."},
"H779":{"lemma":"אָרַר","xlit":"ʼârar","pron":"aw-rar'","derivation":"a primitive root;","strongs_def":"to execrate","kjv_def":"[idiom] bitterly curse."},
"H780":{"lemma":"אֲרָרַט","xlit":"ʼĂrâraṭ","pron":"ar-aw-rat'","derivation":"of foreign origin;","strongs_def":"Ararat (or rather Armenia)","kjv_def":"Ararat, Armenia."},
"H781":{"lemma":"אָרַשׂ","xlit":"ʼâras","pron":"aw-ras'","derivation":"a primitive root;","strongs_def":"to engage for matrimony","kjv_def":"betroth, espouse."},
"H782":{"lemma":"אֲרֶשֶׁת","xlit":"ʼăresheth","pron":"ar-eh'-sheth","derivation":"from H781 (אָרַשׂ) (in the sense of desiring to possess);","strongs_def":"a longing for","kjv_def":"request."},
"H783":{"lemma":"אַרְתַּחְשַׁשְׁתָּא","xlit":"ʼArtachshashtâʼ","pron":"ar-takh-shash-taw'","derivation":"or אַרְתַּחְשַׁשְׁתְּא; or by permutation אַרְתַּחְשַׁסְתְּא; of foreign origin;","strongs_def":"Artachshasta (or Artaxerxes), a title (rather than name) of several Persian kings","kjv_def":"Artaxerxes."},
"H784":{"lemma":"אֵשׁ","xlit":"ʼêsh","pron":"aysh","derivation":"a primitive word;","strongs_def":"fire (literally or figuratively)","kjv_def":"burning, fiery, fire, flaming, hot."},
"H785":{"lemma":"אֵשׁ","xlit":"ʼêsh","pron":"aysh","derivation":"(Aramaic) corresponding to H784 (אֵשׁ)","strongs_def":"{fire (literally or figuratively)}","kjv_def":"flame."},
"H786":{"lemma":"אִשׁ","xlit":"ʼish","pron":"eesh","derivation":"identical (in origin and formation) with H784 (אֵשׁ);","strongs_def":"entity, used only adverbially, there is or are","kjv_def":"are there, none can. Compare H3426 (יֵשׁ)."},
"H787":{"lemma":"אֹשׁ","xlit":"ʼôsh","pron":"ohsh","derivation":"(Aramaic) corresponding (by transposition and abbreviation) to H803 (אֲשׁוּיָה);","strongs_def":"a foundation","kjv_def":"foundation."},
"H788":{"lemma":"אַשְׁבֵּל","xlit":"ʼAshbêl","pron":"ash-bale'","derivation":"probably from the same as H7640 (שֹׁבֶל); flowing;","strongs_def":"Ashbel, an Israelite","kjv_def":"Ashbel."},
"H789":{"lemma":"אַשְׁבֵּלִי","xlit":"ʼAshbêlîy","pron":"ash-bay-lee'","derivation":"patronymic from H788 (אַשְׁבֵּל);","strongs_def":"an Ashbelite (collectively) or descendant of Ashbel","kjv_def":"Ashbelites."},
"H790":{"lemma":"אֶשְׁבָּן","xlit":"ʼEshbân","pron":"esh-bawn'","derivation":"probably from the same as H7644 (שֶׁבְנָא); vigorous;","strongs_def":"Eshban, an Idumaean","kjv_def":"Eshban."},
"H791":{"lemma":"אַשְׁבֵּעַ","xlit":"ʼAshbêaʻ","pron":"ash-bay'-ah","derivation":"from H7650 (שָׁבַע); adjurer;","strongs_def":"Asbea, an Israelite","kjv_def":"Ashbea."},
"H792":{"lemma":"אֶשְׁבַּעַל","xlit":"ʼEshbaʻal","pron":"esh-bah'-al","derivation":"from H376 (אִישׁ) and H1168 (בַּעַל); man of Baal;","strongs_def":"Eshbaal (or Ishbosheth), a son of Saul","kjv_def":"Eshbaal."},
"H793":{"lemma":"אֶשֶׁד","xlit":"ʼeshed","pron":"eh'-shed","derivation":"from an unused root meaning to pour;","strongs_def":"an outpouring","kjv_def":"stream."},
"H794":{"lemma":"אֲשֵׁדָה","xlit":"ʼăshêdâh","pron":"ash-ay-daw'","derivation":"feminine of H793 (אֶשֶׁד);","strongs_def":"a ravine","kjv_def":"springs."},
"H795":{"lemma":"אַשְׁדּוֹד","xlit":"ʼAshdôwd","pron":"ash-dode'","derivation":"from H7703 (שָׁדַד); ravager;","strongs_def":"Ashdod, a place in Palestine","kjv_def":"Ahdod."},
"H796":{"lemma":"אַשְׁדּוֹדִי","xlit":"ʼAshdôwdîy","pron":"ash-do-dee'","derivation":"patrial from H795 (אַשְׁדּוֹד);","strongs_def":"an Ashdodite (often collectively) or inhabitant of Asdod","kjv_def":"Ashdodites, of Ashdod."},
"H797":{"lemma":"אַשְׁדּוֹדִית","xlit":"ʼAshdôwdîyth","pron":"ash-do-deeth'","derivation":"feminine of H796 (אַשְׁדּוֹדִי);","strongs_def":"(only adverb) in the language of Ashdod","kjv_def":"in the speech of Ashdod."},
"H798":{"lemma":"אַשְׁדּוֹת הַפִּסְגָּה","xlit":"ʼAshdôwth hap-Piçgâh","pron":"ash-doth' hap-pisgaw'","derivation":"from the plural of H794 (אֲשֵׁדָה) and H6449 (פִּסְגָּה) with the article interposed; ravines of the Pisgah;","strongs_def":"Ashdoth-Pisgah, a place east of the Jordan","kjv_def":"Ashdoth-pisgah."},
"H799":{"lemma":"אֶשְׁדָּת","xlit":"ʼeshdâth","pron":"esh-dawth'","derivation":"from H784 (אֵשׁ) and H1881 (דָּת);","strongs_def":"a fire-law","kjv_def":"fiery law."},
"H800":{"lemma":"אֶשָּׁה","xlit":"ʼeshshâh","pron":"esh-shaw'","derivation":"feminine of H784 (אֵשׁ);","strongs_def":"fire","kjv_def":"fire."},
"H801":{"lemma":"אִשָּׁה","xlit":"ʼishshâh","pron":"ish-shaw'","derivation":"the same as H800 (אֶשָּׁה), but used in a liturgical sense;","strongs_def":"properly, a burnt-offering; but occasionally of any sacrifice","kjv_def":"(offering, sacrifice), (made) by fire."},
"H802":{"lemma":"אִשָּׁה","xlit":"ʼishshâh","pron":"ish-shaw'","derivation":"feminine of H376 (אִישׁ) or H582 (אֱנוֹשׁ); irregular plural, נָשִׁים;(used in the same wide sense as H582 (אֱנוֹשׁ))","strongs_def":"a woman","kjv_def":"(adulter) ess, each, every, female, [idiom] many, [phrase] none, one, [phrase] together, wife, woman. Often unexpressed in English."},
"H803":{"lemma":"אֲשׁוּיָה","xlit":"ʼăshûwyâh","pron":"ash-oo-yah'","derivation":"feminine passive participle from an unused root meaning to found;","strongs_def":"foundation","kjv_def":"foundation."},
"H804":{"lemma":"אַשּׁוּר","xlit":"ʼAshshûwr","pron":"ash-shoor'","derivation":"or אַשֻּׁר; apparently from H833 (אָשַׁר) (in the sense of successful);","strongs_def":"Ashshur, the second son of Shem; also his descendants and the country occupied by them (i.e. Assyria), its region and its empire","kjv_def":"Asshur, Assur, Assyria, Assyrians. See H838 (אָשֻׁר)."},
"H805":{"lemma":"אֲשׁוּרִי","xlit":"ʼĂshûwrîy","pron":"ash-oo-ree'","derivation":"or אַשּׁוּרִי; from a patrial word of the same form as H804 (אַשּׁוּר);","strongs_def":"an Ashurite (collectively) or inhabitant of Ashur, a district in Palestine","kjv_def":"Asshurim, Ashurites."},
"H806":{"lemma":"אַשְׁחוּר","xlit":"ʼAshchûwr","pron":"ash-khoor'","derivation":"probably from H7835 (שָׁחַר); black;","strongs_def":"Ashchur, an Israelite","kjv_def":"Ashur."},
"H807":{"lemma":"אַשִׁימָא","xlit":"ʼAshîymâʼ","pron":"ash-ee-maw'","derivation":"of foreign origin;","strongs_def":"Ashima, a deity of Hamath","kjv_def":"Ashima."},
"H808":{"lemma":"אָשִׁישׁ","xlit":"ʼâshîysh","pron":"aw-sheesh'","derivation":"from the same as H784 (אֵשׁ) (in the sense of pressing down firmly; compare H803 (אֲשׁוּיָה));","strongs_def":"a (ruined) foundation","kjv_def":"foundation."},
"H809":{"lemma":"אֲשִׁישָׁה","xlit":"ʼăshîyshâh","pron":"ash-ee-shaw'","derivation":"feminine of H808 (אָשִׁישׁ);","strongs_def":"something closely pressed together, i.e. a cake of raisins or other comfits","kjv_def":"flagon."},
"H810":{"lemma":"אֶשֶׁךְ","xlit":"ʼeshek","pron":"eh'-shek","derivation":"from an unused root (probably meaning to bunch together);","strongs_def":"a testicle (as a lump)","kjv_def":"stone."},
"H811":{"lemma":"אֶשְׁכּוֹל","xlit":"ʼeshkôwl","pron":"esh-kole'","derivation":"or אֶשְׁכֹּל; probably prolonged from H810 (אֶשֶׁךְ);","strongs_def":"a bunch of grapes or other fruit","kjv_def":"cluster (of grapes)."},
"H812":{"lemma":"אֶשְׁכֹּל","xlit":"ʼEshkôl","pron":"esh-kole'","derivation":"the same as H811 (אֶשְׁכּוֹל);","strongs_def":"Eshcol, the name of an Amorite, also of a valley in Palestine","kjv_def":"Eshcol."},
"H813":{"lemma":"אַשְׁכְּנַז","xlit":"ʼAshkᵉnaz","pron":"ash-ken-az'","derivation":"of foreign origin;","strongs_def":"Ashkenaz, a Japhethite, also his descendants","kjv_def":"Ashkenaz."},
"H814":{"lemma":"אֶשְׁכָּר","xlit":"ʼeshkâr","pron":"esh-cawr'","derivation":"for H7939 (שָׂכָר);","strongs_def":"a gratuity","kjv_def":"gift, present."},
"H815":{"lemma":"אֵשֶׁל","xlit":"ʼêshel","pron":"ay'-shel","derivation":"from a root of uncertain signification;","strongs_def":"a tamarisk tree; by extension, a grove of any kind","kjv_def":"grove, tree."},
"H816":{"lemma":"אָשַׁם","xlit":"ʼâsham","pron":"aw-sham'","derivation":"or אָשֵׁם; a primitive root;","strongs_def":"to be guilty; by implication to be punished or perish","kjv_def":"[idiom] certainly, be(-come, made) desolate, destroy, [idiom] greatly, be(-come, found, hold) guilty, offend (acknowledge offence), trespass."},
"H817":{"lemma":"אָשָׁם","xlit":"ʼâshâm","pron":"aw-shawm'","derivation":"from H816 (אָשַׁם);","strongs_def":"guilt; by implication, a fault; also a sin-offering","kjv_def":"guiltiness, (offering for) sin, trespass (offering)."},
"H818":{"lemma":"אָשֵׁם","xlit":"ʼâshêm","pron":"aw-shame'","derivation":"from H816 (אָשַׁם);","strongs_def":"guilty; hence, presenting asin-offering","kjv_def":"one which is faulty, guilty."},
"H819":{"lemma":"אַשְׁמָה","xlit":"ʼashmâh","pron":"ash-maw'","derivation":"feminine of H817 (אָשָׁם);","strongs_def":"guiltiness, a fault, the presentation of asin-offering","kjv_def":"offend, sin, (cause of) trespass(-ing, offering)."},
"H820":{"lemma":"אַשְׁמָן","xlit":"ʼashmân","pron":"ash-mawn'","derivation":"probably from H8081 (שֶׁמֶן);","strongs_def":"a fat-field","kjv_def":"desolate place."},
"H821":{"lemma":"אַשְׁמֻרָה","xlit":"ʼashmurâh","pron":"ash-moo-raw'","derivation":"or אַשׁמוּרָה; or אַשׁמֹרֶת; (feminine) from H8104 (שָׁמַר);","strongs_def":"a night watch","kjv_def":"watch."},
"H822":{"lemma":"אֶשְׁנָב","xlit":"ʼeshnâb","pron":"esh-nawb'","derivation":"apparently from an unused root (probably meaning to leave interstices);","strongs_def":"a latticed window","kjv_def":"casement, lattice."},
"H823":{"lemma":"אַשְׁנָה","xlit":"ʼAshnâh","pron":"ash-naw'","derivation":"probably a variation for H3466 (יְשָׁנָה);","strongs_def":"Ashnah, the name of two places in Palestine","kjv_def":"Ashnah."},
"H824":{"lemma":"אֶשְׁעָן","xlit":"ʼEshʻân","pron":"esh-awn'","derivation":"from H8172 (שָׁעַן); support;","strongs_def":"Eshan, a place in Palestine","kjv_def":"Eshean."},
"H825":{"lemma":"אַשָּׁף","xlit":"ʼashshâph","pron":"ash-shawf'","derivation":"from an unused root (probably meaning to lisp, i.e. practice enchantment);","strongs_def":"a conjurer","kjv_def":"astrologer."},
"H826":{"lemma":"אַשָּׁף","xlit":"ʼashshâph","pron":"ash-shawf'","derivation":"(Aramaic) corresponding to H825 (אַשָּׁף)","strongs_def":"{a conjurer}","kjv_def":"astrologer."},
"H827":{"lemma":"אַשְׁפָּה","xlit":"ʼashpâh","pron":"ash-paw'","derivation":"perhaps (feminine) from the same as H825 (אַשָּׁף) (in the sense of covering);","strongs_def":"a quiver or arrow-case","kjv_def":"quiver."},
"H828":{"lemma":"אַשְׁפְּנַז","xlit":"ʼAshpᵉnaz","pron":"ash-pen-az'","derivation":"of foreign origin;","strongs_def":"Ashpenaz, a Babylonian eunuch","kjv_def":"Ashpenaz."},
"H829":{"lemma":"אֶשְׁפָּר","xlit":"ʼeshpâr","pron":"esh-pawr'","derivation":"of uncertain derivation;","strongs_def":"a measured portion","kjv_def":"good piece (of flesh)."},
"H830":{"lemma":"אַשְׁפֹּת","xlit":"ʼashpôth","pron":"ash-pohth'","derivation":"or אַשְׁפּוֹת; or שְׁפֹת; (contraction) plural of a noun of the same form as H827 (אַשְׁפָּה), from H8192 (שָׁפָה) (in the sense of scraping);","strongs_def":"a heap of rubbish or filth","kjv_def":"dung (hill)."},
"H831":{"lemma":"אַשְׁקְלוֹן","xlit":"ʼAshqᵉlôwn","pron":"ash-kel-one'","derivation":"probably from H8254 (שָׁקַל) in the sense of weighing-place (i.e. mart);","strongs_def":"Ashkelon, a place in Palestine","kjv_def":"Ashkelon, Askalon."},
"H832":{"lemma":"אֶשְׁקְלוֹנִי","xlit":"ʼEshqᵉlôwnîy","pron":"esh-kel-o-nee'","derivation":"patrial from H831 (אַשְׁקְלוֹן);","strongs_def":"Ashkelonite (collectively) or inhabitant of Ashkelon","kjv_def":"Eshkalonites."},
"H833":{"lemma":"אָשַׁר","xlit":"ʼâshar","pron":"aw-shar'","derivation":"or אָשֵׁר; a primitive root;","strongs_def":"to be straight (used in the widest sense, especially to be level, right, happy); figuratively, to go forward, be honest, proper","kjv_def":"(call, be) bless(-ed, happy), go, guide, lead, relieve."},
"H834":{"lemma":"אֲשֶׁר","xlit":"ʼăsher","pron":"ash-er'","derivation":"a primitive relative pronoun (of every gender and number);","strongs_def":"who, which, what, that; also (as an adverb and a conjunction) when, where, how, because, in order that, etc.","kjv_def":"[idiom] after, [idiom] alike, as (soon as), because, [idiom] every, for, [phrase] forasmuch, [phrase] from whence, [phrase] how(-soever), [idiom] if, (so) that ((thing) which, wherein), [idiom] though, [phrase] until, [phrase] whatsoever, when, where ([phrase] -as, -in, -of, -on, -soever, -with), which, whilst, [phrase] whither(-soever), who(-m, -soever, -se). As it is indeclinable, it is often accompanied by the personal pronoun expletively, used to show the connection."},
"H835":{"lemma":"אֶשֶׁר","xlit":"ʼesher","pron":"eh'-sher","derivation":"from H833 (אָשַׁר);","strongs_def":"happiness; only in masculine plural construction as interjection, how happy!","kjv_def":"blessed, happy."},
"H836":{"lemma":"אָשֵׁר","xlit":"ʼÂshêr","pron":"aw-share'","derivation":"from H833 (אָשַׁר),","strongs_def":"happy; Asher, a son of Jacob, and the tribe descended from him, with its territory; also a place in Palestine","kjv_def":"Asher"},
"H837":{"lemma":"אֹשֶׁר","xlit":"ʼôsher","pron":"o'-sher","derivation":"from H833 (אָשַׁר);","strongs_def":"happiness","kjv_def":"happy."},
"H838":{"lemma":"אָשֻׁר","xlit":"ʼâshur","pron":"aw-shoor'","derivation":"or אַשֻּׁר; from H833 (אָשַׁר) in the sense of going;","strongs_def":"a step","kjv_def":"going, step."},
"H839":{"lemma":"אֲשֻׁר","xlit":"ʼăshur","pron":"ash-oor'","derivation":"contracted for H8391 (תְּאַשּׁוּר);","strongs_def":"the cedar tree or some other light elastic wood","kjv_def":"Ashurite."},
"H840":{"lemma":"אֲשַׂרְאֵל","xlit":"ʼĂsarʼêl","pron":"as-ar-ale'","derivation":"by orthographical variation from H833 (אָשַׁר) and H410 (אֵל); right of God;","strongs_def":"Asarel, an Israelite","kjv_def":"Asareel."},
"H841":{"lemma":"אֲשַׂרְאֵלָה","xlit":"ʼĂsarʼêlâh","pron":"as-ar-ale'-aw","derivation":"from the same as H840 (אֲשַׂרְאֵל); right toward God;","strongs_def":"Asarelah, an Israelite","kjv_def":"Asarelah. Compare H3480 (יְשַׂרְאֵלָה)."},
"H842":{"lemma":"אֲשֵׁרָה","xlit":"ʼăshêrâh","pron":"ash-ay-raw'","derivation":"or אֲשֵׁירָה; from H833 (אָשַׁר); happy;","strongs_def":"Asherah (or Astarte) a Phoenician goddess; also an image of the same","kjv_def":"grove. Compare H6253 (עַשְׁתֹּרֶת)."},
"H843":{"lemma":"אָשֵׁרִי","xlit":"ʼÂshêrîy","pron":"aw-shay-ree'","derivation":"patronymic from H836 (אָשֵׁר);","strongs_def":"an Asherite (collectively) or descendant of Asher","kjv_def":"Asherites."},
"H844":{"lemma":"אַשְׂרִיאֵל","xlit":"ʼAsrîyʼêl","pron":"as-ree-ale'","derivation":"an orthographical variation for H840 (אֲשַׂרְאֵל);","strongs_def":"Asriel, the name of two Israelites","kjv_def":"Ashriel, Asriel."},
"H845":{"lemma":"אַשְׂרִאֵלִי","xlit":"ʼAsriʼêlîy","pron":"as-ree-ale-ee'","derivation":"patronymic from H844 (אַשְׂרִיאֵל);","strongs_def":"an Asrielite (collectively) or descendant of Asriel","kjv_def":"Asrielites."},
"H846":{"lemma":"אֻשַּׁרְנָא","xlit":"ʼushsharnâʼ","pron":"oosh-ar-naw'","derivation":"(Aramaic) from a root corresponding to H833 (אָשַׁר);","strongs_def":"a wall (from its uprightness)","kjv_def":"wall."},
"H847":{"lemma":"אֶשְׁתָּאֹל","xlit":"ʼEshtâʼôl","pron":"esh-taw-ole'","derivation":"or אֶשְׁתָּאוֹל; probably from H7592 (שָׁאַל); intreaty;","strongs_def":"Eshtaol, a place in Palestine","kjv_def":"Eshtaol."},
"H848":{"lemma":"אֶשְׁתָּאֻלִי","xlit":"ʼEshtâʼulîy","pron":"esh-taw-oo-lee'","derivation":"patrial from H847 (אֶשְׁתָּאֹל);","strongs_def":"an Eshtaolite (collectively) or inhabitant of Eshtaol","kjv_def":"Eshtaulites."},
"H849":{"lemma":"אֶשְׁתַּדּוּר","xlit":"ʼeshtaddûwr","pron":"esh-tad-dure'","derivation":"(Aramaic) from H7712 (שְׁדַר) (in a bad sense);","strongs_def":"rebellion","kjv_def":"sedition."},
"H850":{"lemma":"אֶשְׁתּוֹן","xlit":"ʼEshtôwn","pron":"esh-tone'","derivation":"probably from the same as H7764 (שׁוּנִי); restful;","strongs_def":"Eshton, an Israelite","kjv_def":"Eshton."},
"H851":{"lemma":"אֶשְׁתְּמֹעַ","xlit":"ʼEshtᵉmôaʻ","pron":"esh-tem-o'-ah","derivation":"or אֶשְׁתְּמוֹעַ; or אֶשְׁתְּמֹה; from H8085 (שָׁמַע) (in the sense of obedience);","strongs_def":"Eshtemoa or Eshtemoh, a place in Palestine","kjv_def":"Eshtemoa, Eshtemoh."},
"H852":{"lemma":"אָת","xlit":"ʼâth","pron":"awth","derivation":"(Aramaic) corresponding to H226 (אוֹת);","strongs_def":"a portent","kjv_def":"sign."},
"H853":{"lemma":"אֵת","xlit":"ʼêth","pron":"ayth","derivation":"apparent contracted from H226 (אוֹת) in the demonstrative sense of entity;","strongs_def":"properly, self (but generally used to point out more definitely the object of a verb or preposition, even or namely)","kjv_def":"(as such unrepresented in English)."},
"H854":{"lemma":"אֵת","xlit":"ʼêth","pron":"ayth","derivation":"probably from H579 (אָנָה);","strongs_def":"properly, nearness (used only as a preposition or an adverb), near; hence, generally, with, by, at, among, etc.","kjv_def":"against, among, before, by, for, from, in(-to), (out) of, with. Often with another prepositional prefix."},
"H855":{"lemma":"אֵת","xlit":"ʼêth","pron":"ayth","derivation":"of uncertain derivation;","strongs_def":"a hoe or other digging implement","kjv_def":"coulter, plowshare."},
"H856":{"lemma":"אֶתְבַּעַל","xlit":"ʼEthbaʻal","pron":"eth-bah'-al","derivation":"from H854 (אֵת) and H1168 (בַּעַל); with Baal;","strongs_def":"Ethbaal, a Phoenician king","kjv_def":"Ethbaal."},
"H857":{"lemma":"אָתָה","xlit":"ʼâthâh","pron":"aw-thaw'","derivation":"or אָתָא; a primitive root (collateral to H225 (אוּת) contraction);","strongs_def":"to arrive","kjv_def":"(be-, things to) come (upon), bring."},
"H858":{"lemma":"אָתָה","xlit":"ʼâthâh","pron":"aw-thaw'","derivation":"(Aramaic) or אָתָא; (Aramaic), corresponding to H857 (אָתָה)","strongs_def":"{to arrive}","kjv_def":"(be-) come, bring."},
"H859":{"lemma":"אַתָּה","xlit":"ʼattâh","pron":"at-taw'","derivation":"or (shortened); אַתָּ; or אַת; feminine (irregular) sometimes אַתִּי; plural masculine אַתֶּם; feminine אַתֶּן; or אַתֵּנָהlemma אתֵּנָה missing vowel, corrected to אַתֵּנָה; or אַתֵּנָּה; a primitive pronoun of the second person;","strongs_def":"thou and thee, or (plural) ye and you","kjv_def":"thee, thou, ye, you."},
"H860":{"lemma":"אָתוֹן","xlit":"ʼâthôwn","pron":"aw-thone'","derivation":"probably from the same as H386 (אֵיתָן) (in the sense of patience);","strongs_def":"a female donkey (from its docility)","kjv_def":"(she) ass."},
"H861":{"lemma":"אַתּוּן","xlit":"ʼattûwn","pron":"at-toon'","derivation":"(Aramaic) probably from the corresponding to H784 (אֵשׁ);","strongs_def":"probably a fire-place, i.e. furnace","kjv_def":"furnace."},
"H862":{"lemma":"אַתּוּק","xlit":"ʼattûwq","pron":"at-tooke'","derivation":"or אַתִּיק; from H5423 (נָתַק) in the sense of decreasing;","strongs_def":"a ledge or offset in abuilding","kjv_def":"gallery."},
"H863":{"lemma":"אִתַּי","xlit":"ʼIttay","pron":"it-tah'ee","derivation":"or אִיתַי; from H854 (אֵת); near;","strongs_def":"Ittai or Ithai, the name of a Gittite and of an Israelite","kjv_def":"Ithai, Ittai."},
"H864":{"lemma":"אֵתָם","xlit":"ʼÊthâm","pron":"ay-thawm'","derivation":"of Egyptian derivation;","strongs_def":"Etham, a place in the Desert","kjv_def":"Etham."},
"H865":{"lemma":"אֶתְמוֹל","xlit":"ʼethmôwl","pron":"eth-mole'","derivation":"or אִתְמוֹל; or אֶתְמוּל; probably from H853 (אֵת) or H854 (אֵת) and H4136 (מוּל);","strongs_def":"heretofore; definitely yesterday","kjv_def":"[phrase] before (that) time, [phrase] heretofore, of late (old), [phrase] times past, yester(day)."},
"H866":{"lemma":"אֶתְנָה","xlit":"ʼethnâh","pron":"eth-naw'","derivation":"from H8566 (תָּנָה);","strongs_def":"a present (as the price of harlotry)","kjv_def":"reward."},
"H867":{"lemma":"אֶתְנִי","xlit":"ʼEthnîy","pron":"eth-nee'","derivation":"perhaps from H866 (אֶתְנָה); munificence;","strongs_def":"Ethni, an Israelite","kjv_def":"Ethni."},
"H868":{"lemma":"אֶתְנַן","xlit":"ʼethnan","pron":"eth-nan'","derivation":"the same as H866 (אֶתְנָה);","strongs_def":"a gift (as the price of harlotry or idolatry)","kjv_def":"hire, reward."},
"H869":{"lemma":"אֶתְנַן","xlit":"ʼEthnan","pron":"eth-nan'","derivation":"the same as H868 (אֶתְנַן) in the sense of H867 (אֶתְנִי);","strongs_def":"Ethnan, an Israelite","kjv_def":"Ethnan."},
"H870":{"lemma":"אֲתַר","xlit":"ʼăthar","pron":"ath-ar'","derivation":"(Aramaic) from a root corresponding to that of H871 (אֲתָרִים);","strongs_def":"a place; (adverb) after","kjv_def":"after, place."},
"H871":{"lemma":"אֲתָרִים","xlit":"ʼĂthârîym","pron":"ath-aw-reem'","derivation":"plural from an unused root (probably meaning to step); places;","strongs_def":"Atharim, a place near Palestine","kjv_def":"spies. b"},
"H872":{"lemma":"בְּאָה","xlit":"bᵉʼâh","pron":"be-aw'","derivation":"from H935 (בּוֹא);","strongs_def":"an entrance to a building","kjv_def":"entry."},
"H873":{"lemma":"בִּאוּשׁ","xlit":"biʼûwsh","pron":"be-oosh'","derivation":"(Aramaic) from H888 (בְּאֵשׁ);","strongs_def":"wicked","kjv_def":"bad."},
"H874":{"lemma":"בָּאַר","xlit":"bâʼar","pron":"baw-ar'","derivation":"a primitive root;","strongs_def":"to dig; by analogy, to engrave; figuratively, to explain","kjv_def":"declare, (make) plain(-ly)."},
"H875":{"lemma":"בְּאֵר","xlit":"bᵉʼêr","pron":"be-ayr'","derivation":"from H874 (בָּאַר);","strongs_def":"a pit; especially a well","kjv_def":"pit, well."},
"H876":{"lemma":"בְּאֵר","xlit":"Bᵉʼêr","pron":"be-ayr'","derivation":"the same as H875 (בְּאֵר);","strongs_def":"Beer, a place in the Desert, also one in Palestine","kjv_def":"Beer."},
"H877":{"lemma":"בֹּאר","xlit":"bôʼr","pron":"bore","derivation":"from H874 (בָּאַר);","strongs_def":"a cistern","kjv_def":"cistern."},
"H878":{"lemma":"בְּאֵרָא","xlit":"Bᵉʼêrâʼ","pron":"be-ay-raw'","derivation":"from H875 (בְּאֵר); a well;","strongs_def":"Beera, an Israelite","kjv_def":"Beera."},
"H879":{"lemma":"בְּאֵר אֵלִים","xlit":"Bᵉʼêr ʼÊlîym","pron":"be-ayr' ay-leem'","derivation":"from H875 (בְּאֵר) and the plural of H410 (אֵל); well of heroes;","strongs_def":"Beer-Elim, a place in the Desert","kjv_def":"Beer-elim."},
"H880":{"lemma":"בְּאֵרָה","xlit":"Bᵉʼêrâh","pron":"be-ay-raw'","derivation":"the same as H878 (בְּאֵרָא);","strongs_def":"Beerah, an Israelite","kjv_def":"Beerah."},
"H881":{"lemma":"בְּאֵרוֹת","xlit":"Bᵉʼêrôwth","pron":"be-ay-rohth'","derivation":"feminine plural of H875 (בְּאֵר); wells;","strongs_def":"Beeroth, a place in Palestine","kjv_def":"Beeroth."},
"H882":{"lemma":"בְּאֵרִי","xlit":"Bᵉʼêrîy","pron":"be-ay-ree'","derivation":"from H875 (בְּאֵר); fountained;","strongs_def":"Beeri, the name of a Hittite and of an Israelite","kjv_def":"Beeri."},
"H883":{"lemma":"בְּאֵר לַחַי רֹאִי","xlit":"Bᵉʼêr la-Chay Rôʼîy","pron":"be-ayr' lakh-ah'ee ro-ee'","derivation":"from H875 (בְּאֵר) and H2416 (חַי) (with prefix) and H7203 (רֹאֶה); well of a living (One) my Seer;","strongs_def":"Beer-Lachai-Roi, a place in the Desert","kjv_def":"Beer-lahai-roi."},
"H884":{"lemma":"בְּאֵר שֶׁבַע","xlit":"Bᵉʼêr Shebaʻ","pron":"be-ayr' sheh'-bah","derivation":"from H875 (בְּאֵר) and H7651 (שֶׁבַע) (in the sense of H7650 (שָׁבַע)); well of an oath;","strongs_def":"Beer-Sheba, a place in Palestine","kjv_def":"Beer-shebah."},
"H885":{"lemma":"בְּאֵרֹת בְּנֵי־יַעֲקַן","xlit":"Bᵉʼêrôth Bᵉnêy-Yaʻăqan","pron":"be-ay-roth' be-nay' yah-a-can'","derivation":"from the feminine plural of H875 (בְּאֵר), and the plural contraction of H1121 (בֵּן), and H3292 (יַעֲקָן); wells of (the) sons of Jaakan;","strongs_def":"Beeroth-Bene-Jaakan, a place in the Desert","kjv_def":"Beeroth of the children of Jaakan."},
"H886":{"lemma":"בְּאֵרֹתִי","xlit":"Bᵉʼêrôthîy","pron":"be-ay-ro-thee'","derivation":"patrial from H881 (בְּאֵרוֹת);","strongs_def":"a Beerothite or inhabitant of Beeroth","kjv_def":"Beerothite."},
"H887":{"lemma":"בָּאַשׁ","xlit":"bâʼash","pron":"baw-ash'","derivation":"a primitive root;","strongs_def":"to smell bad; figuratively, to be offensive morally","kjv_def":"(make to) be abhorred (had in abomination, loathsome, odious), (cause a, make to) stink(-ing savour), [idiom] utterly."},
"H888":{"lemma":"בְּאֵשׁ","xlit":"bᵉʼêsh","pron":"be-aysh'","derivation":"(Aramaic) corresponding to H887 (בָּאַשׁ)","strongs_def":"{to smell bad; figuratively, to be offensive morally}","kjv_def":"displease."},
"H889":{"lemma":"בְּאֹשׁ","xlit":"bᵉʼôsh","pron":"be-oshe'","derivation":"from H877 (בֹּאר);","strongs_def":"a stench","kjv_def":"stink."},
"H890":{"lemma":"בׇּאְשָׁה","xlit":"boʼshâh","pron":"bosh-aw'","derivation":"feminine of H889 (בְּאֹשׁ);","strongs_def":"stink-weed or any other noxious or useless plant","kjv_def":"cockle."},
"H891":{"lemma":"בְּאֻשִׁים","xlit":"bᵉʼushîym","pron":"be-oo-sheem'","derivation":"plural of H889 (בְּאֹשׁ);","strongs_def":"poison-berries","kjv_def":"wild grapes."},
"H892":{"lemma":"בָּבָה","xlit":"bâbâh","pron":"baw-baw'","derivation":"feminine active participle of an unused root meaning to hollow out;","strongs_def":"something hollowed (as a gate), i.e. pupil of the eye","kjv_def":"apple (of the eye)."},
"H893":{"lemma":"בֵּבַי","xlit":"Bêbay","pron":"bay-bah'ee","derivation":"probably of foreign origin;","strongs_def":"Bebai, an Israelite","kjv_def":"Bebai."},
"H894":{"lemma":"בָּבֶל","xlit":"Bâbel","pron":"baw-bel'","derivation":"from H1101 (בָּלַל); confusion;","strongs_def":"Babel (i.e. Babylon), including Babylonia and the Babylonian empire","kjv_def":"Babel, Babylon."},
"H895":{"lemma":"בַּבֶל","xlit":"Babel","pron":"baw-bel'","derivation":"(Aramaic) corresponding to H894 (בָּבֶל)","strongs_def":"{Babel (i.e. Babylon), including Babylonia and the Babylonian empire}","kjv_def":"Babylon."},
"H896":{"lemma":"בַּבְלִי","xlit":"Bablîy","pron":"bab-lee'","derivation":"(Aramaic) patrial from H895 (בַּבֶל);","strongs_def":"a Babylonian","kjv_def":"Babylonia."},
"H897":{"lemma":"בַּג","xlit":"bag","pron":"bag","derivation":"a Persian word;","strongs_def":"food","kjv_def":"spoil (from the margin for H957 (בַּז).)"},
"H898":{"lemma":"בָּגַד","xlit":"bâgad","pron":"baw-gad'","derivation":"a primitive root;","strongs_def":"to cover (with a garment); figuratively, to act covertly; by implication, to pillage","kjv_def":"deal deceitfully (treacherously, unfaithfully), offend, transgress(-or), (depart), treacherous (dealer, -ly, man), unfaithful(-ly, man), [idiom] very."},
"H899":{"lemma":"בֶּגֶד","xlit":"beged","pron":"behg'-ed","derivation":"from H898 (בָּגַד);","strongs_def":"a covering, i.e. clothing; also treachery or pillage","kjv_def":"apparel, cloth(-es, ing), garment, lap, rag, raiment, robe, [idiom] very (treacherously), vesture, wardrobe."},
"H900":{"lemma":"בֹּגְדוֹת","xlit":"bôgᵉdôwth","pron":"bohg-ed-ohth","derivation":"feminine plural active participle of H898 (בָּגַד);","strongs_def":"treacheries","kjv_def":"treacherous."},
"H901":{"lemma":"בָּגוֹד","xlit":"bâgôwd","pron":"baw-gode'","derivation":"from H898 (בָּגַד);","strongs_def":"treacherous","kjv_def":"treacherous."},
"H902":{"lemma":"בִּגְוַי","xlit":"Bigvay","pron":"big-vah'ee","derivation":"probably of foreign origin;","strongs_def":"Bigvai, an Israelite","kjv_def":"Bigvai."},
"H903":{"lemma":"בִּגְתָא","xlit":"Bigthâʼ","pron":"big-thaw'","derivation":"of Persian derivation;","strongs_def":"Bigtha, a eunuch of Xerxes","kjv_def":"Bigtha."},
"H904":{"lemma":"בִּגְתָן","xlit":"Bigthân","pron":"big-thawn'","derivation":"or בִּגְתָנָא; of similar derivation to H903 (בִּגְתָא);","strongs_def":"Bigthan or Bigthana, a eunuch of Xerxes","kjv_def":"Bigthan, Bigthana."},
"H905":{"lemma":"בַּד","xlit":"bad","pron":"bad","derivation":"from H909 (בָּדַד);","strongs_def":"properly, separation; by implication, a part of the body, branch of atree, bar forcarrying; figuratively, chief of a city; especially (with prepositional prefix) as an adverb, apart, only, besides","kjv_def":"alone, apart, bar, besides, branch, by self, of each alike, except, only, part, staff, strength."},
"H906":{"lemma":"בַּד","xlit":"bad","pron":"bad","derivation":"perhaps from H909 (בָּדַד) (in the sense of divided fibres);","strongs_def":"flaxen thread or yarn; hence, a linen garment","kjv_def":"linen."},
"H907":{"lemma":"בַּד","xlit":"bad","pron":"bad","derivation":"from H908 (בָּדָא);","strongs_def":"a brag or lie; also a liar","kjv_def":"liar, lie."},
"H908":{"lemma":"בָּדָא","xlit":"bâdâʼ","pron":"baw-daw'","derivation":"a primitive root;","strongs_def":"(figuratively) to invent","kjv_def":"devise, feign."},
"H909":{"lemma":"בָּדַד","xlit":"bâdad","pron":"baw-dad'","derivation":"a primitive root;","strongs_def":"to divide, i.e. (reflex.) be solitary","kjv_def":"alone."},
"H910":{"lemma":"בָּדָד","xlit":"bâdâd","pron":"baw-dawd'","derivation":"from H909 (בָּדַד);","strongs_def":"separate; adverb, separately","kjv_def":"alone, desolate, only, solitary."},
"H911":{"lemma":"בְּדַד","xlit":"Bᵉdad","pron":"bed-ad'","derivation":"from H909 (בָּדַד); separation;","strongs_def":"Bedad, an Edomite","kjv_def":"Bedad."},
"H912":{"lemma":"בֵּדְיָה","xlit":"Bêdᵉyâh","pron":"bay-de-yaw'","derivation":"probably a shortened form H5662 (עֹבַדְיָה); servant of Jehovah;","strongs_def":"Bedejah, an Israelite","kjv_def":"Bedeiah."},
"H913":{"lemma":"בְּדִיל","xlit":"bᵉdîyl","pron":"bed-eel'","derivation":"from H914 (בָּדַל);","strongs_def":"alloy (because removed by smelting); by analogy, tin","kjv_def":"[phrase] plummet, tin."},
"H914":{"lemma":"בָּדַל","xlit":"bâdal","pron":"baw-dal'","derivation":"a primitive root;","strongs_def":"to divide (in variation senses literally or figuratively, separate, distinguish, differ, select, etc.)","kjv_def":"(make, put) difference, divide (asunder), (make) separate (self, -ation), sever (out), [idiom] utterly."},
"H915":{"lemma":"בָּדָל","xlit":"bâdâl","pron":"baw-dawl'","derivation":"from H914 (בָּדַל);","strongs_def":"a part","kjv_def":"piece."},
"H916":{"lemma":"בְּדֹלַח","xlit":"bᵉdôlach","pron":"bed-o'-lakh","derivation":"probably from H914 (בָּדַל);","strongs_def":"something in pieces, i.e. bdellium, a (fragrant) gum (perhaps amber); others a pearl","kjv_def":"bdellium."},
"H917":{"lemma":"בְּדָן","xlit":"Bᵉdân","pron":"bed-awn'","derivation":"probably shortened for H5658 (עַבְדוֹן); servile;","strongs_def":"Bedan, the name of two Israelites","kjv_def":"Bedan."},
"H918":{"lemma":"בָּדַק","xlit":"bâdaq","pron":"baw-dak'","derivation":"a primitive root; to gap open; used only as a denominative from H919 (בֶּדֶק);","strongs_def":"to mend a breach","kjv_def":"repair."},
"H919":{"lemma":"בֶּדֶק","xlit":"bedeq","pron":"beh'-dek","derivation":"from H918 (בָּדַק);","strongs_def":"a gap or leak (in a building or a ship)","kjv_def":"breach, [phrase] calker."},
"H920":{"lemma":"בִּדְקַר","xlit":"Bidqar","pron":"bid-car'","derivation":"probably from H1856 (דָּקַר) with a prepositional prefix; by stabbing, i.e. assassin;","strongs_def":"Bidkar, an Israelite","kjv_def":"Bidkar."},
"H921":{"lemma":"בְּדַר","xlit":"bᵉdar","pron":"bed-ar'","derivation":"(Aramaic) corresponding (by transposition) to H6504 (פָּרַד);","strongs_def":"to scatter","kjv_def":"scatter."},
"H922":{"lemma":"בֹּהוּ","xlit":"bôhûw","pron":"bo'-hoo","derivation":"from an unused root (meaning to be empty);","strongs_def":"a vacuity, i.e. (superficially) an undistinguishable ruin","kjv_def":"emptiness, void."},
"H923":{"lemma":"בַּהַט","xlit":"bahaṭ","pron":"bah'-hat","derivation":"from an unused root (probably meaning to glisten);","strongs_def":"white marble or perhaps alabaster","kjv_def":"red (marble)."},
"H924":{"lemma":"בְּהִילוּ","xlit":"bᵉhîylûw","pron":"be-hee-loo'","derivation":"(Aramaic) from H927 (בְּהַל);","strongs_def":"a hurry; only adverb, hastily","kjv_def":"in haste."},
"H925":{"lemma":"בָּהִיר","xlit":"bâhîyr","pron":"baw-here'","derivation":"from an unused root (meaning to be bright);","strongs_def":"shining","kjv_def":"bright."},
"H926":{"lemma":"בָּהַל","xlit":"bâhal","pron":"baw-hal'","derivation":"a primitive root;","strongs_def":"to tremble inwardly (or palpitate), i.e. (figuratively) be (causative, make) (suddenly) alarmed or agitated; by implication to hasten anxiously","kjv_def":"be (make) affrighted (afraid, amazed, dismayed, rash), (be, get, make) haste(-n, -y, -ily), (give) speedy(-ily), thrust out, trouble, vex."},
"H927":{"lemma":"בְּהַל","xlit":"bᵉhal","pron":"be-hal'","derivation":"(Aramaic) corresponding to H926 (בָּהַל);","strongs_def":"to terrify, hasten","kjv_def":"in haste, trouble."},
"H928":{"lemma":"בֶּהָלָה","xlit":"behâlâh","pron":"beh-haw-law'","derivation":"from H926 (בָּהַל);","strongs_def":"panic, destruction","kjv_def":"terror, trouble."},
"H929":{"lemma":"בְּהֵמָה","xlit":"bᵉhêmâh","pron":"be-hay-maw'","derivation":"from an unused root (probably meaning to be mute);","strongs_def":"properly, a dumb beast; especially any large quadruped or animal (often collective)","kjv_def":"beast, cattle."},
"H930":{"lemma":"בְּהֵמוֹת","xlit":"bᵉhêmôwth","pron":"be-hay-mohth'","derivation":"in form a plural or H929 (בְּהֵמָה), but really a singular of Egyptian derivation;","strongs_def":"a water-ox, i.e. the hippopotamus or Nile-horse","kjv_def":"Behemoth."},
"H931":{"lemma":"בֹּהֶן","xlit":"bôhen","pron":"bo'-hen","derivation":"from an unused root apparently meaning to be thick;","strongs_def":"the thumb of the hand or great toe of the foot","kjv_def":"thumb, great toe."},
"H932":{"lemma":"בֹּהַן","xlit":"Bôhan","pron":"bo'han","derivation":"an orthographical variation of H931 (בֹּהֶן);","strongs_def":"thumb, Bohan, an Israelite","kjv_def":"Bohan."},
"H933":{"lemma":"בֹּהַק","xlit":"bôhaq","pron":"bo'-hak","derivation":"from an unused root meaning to be pale;","strongs_def":"white scurf","kjv_def":"freckled spot."},
"H934":{"lemma":"בֹּהֶרֶת","xlit":"bôhereth","pron":"bo-heh'-reth","derivation":"feminine active participle of the same as H925 (בָּהִיר);","strongs_def":"a whitish spot on the skin","kjv_def":"bright spot."},
"H935":{"lemma":"בּוֹא","xlit":"bôwʼ","pron":"bo","derivation":"a primitive root;","strongs_def":"to go or come (in a wide variety of applications)","kjv_def":"abide, apply, attain, [idiom] be, befall, [phrase] besiege, bring (forth, in, into, to pass), call, carry, [idiom] certainly, (cause, let, thing for) to come (against, in, out, upon, to pass), depart, [idiom] doubtless again, [phrase] eat, [phrase] employ, (cause to) enter (in, into, -tering, -trance, -try), be fallen, fetch, [phrase] follow, get, give, go (down, in, to war), grant, [phrase] have, [idiom] indeed, (in-) vade, lead, lift (up), mention, pull in, put, resort, run (down), send, set, [idiom] (well) stricken (in age), [idiom] surely, take (in), way."},
"H936":{"lemma":"בּוּז","xlit":"bûwz","pron":"booz","derivation":"a primitive root;","strongs_def":"to disrespect","kjv_def":"contemn, despise, [idiom] utterly."},
"H937":{"lemma":"בּוּז","xlit":"bûwz","pron":"booz","derivation":"from H936 (בּוּז);","strongs_def":"disrespect","kjv_def":"contempt(-uously), despised, shamed."},
"H938":{"lemma":"בּוּז","xlit":"Bûwz","pron":"booz","derivation":"the same as H937 (בּוּז);","strongs_def":"Buz, the name of a son of Nahor, and of an Israelite","kjv_def":"Buz."},
"H939":{"lemma":"בּוּזָה","xlit":"bûwzâh","pron":"boo-zaw'","derivation":"feminine passive participle of H936 (בּוּז);","strongs_def":"something scorned; an object of contempt","kjv_def":"despised."},
"H940":{"lemma":"בּוּזִי","xlit":"Bûwzîy","pron":"boo-zee'","derivation":"patronymic from H938 (בּוּז);","strongs_def":"a Buzite or descendant of Buz","kjv_def":"Buzite."},
"H941":{"lemma":"בּוּזִי","xlit":"Bûwzîy","pron":"boo-zee'","derivation":"the same as H940 (בּוּזִי);","strongs_def":"Buzi, an Israelite","kjv_def":"Buzi."},
"H942":{"lemma":"בַּוַּי","xlit":"Bavvay","pron":"bav-vah'ee","derivation":"probably of Persian origin;","strongs_def":"Bavvai, an Israelite","kjv_def":"Bavai."},
"H943":{"lemma":"בּוּךְ","xlit":"bûwk","pron":"book","derivation":"a primitive root;","strongs_def":"to involve (literally or figuratively)","kjv_def":"be entangled, (perplexed)."},
"H944":{"lemma":"בּוּל","xlit":"bûwl","pron":"bool","derivation":"for H2981 (יְבוּל);","strongs_def":"produce (of the earth, etc.)","kjv_def":"food, stock."},
"H945":{"lemma":"בּוּל","xlit":"Bûwl","pron":"bool","derivation":"the same as H944 (בּוּל) (in the sense of rain);","strongs_def":"Bul, the eighth Hebrew month","kjv_def":"Bul."},
"H946":{"lemma":"בּוּנָה","xlit":"Bûwnâh","pron":"boo-naw'","derivation":"from H995 (בִּין); discretion;","strongs_def":"Bunah, an Israelite","kjv_def":"Bunah."},
"H947":{"lemma":"בּוּס","xlit":"bûwç","pron":"boos","derivation":"a primitive root;","strongs_def":"to trample (literally or figuratively)","kjv_def":"loath, tread (down, under (foot)), be polluted."},
"H948":{"lemma":"בּוּץ","xlit":"bûwts","pron":"boots","derivation":"from an unused root (of the same form) meaning to bleach, i.e. (intransitive) be white;","strongs_def":"probably cotton (of some sort)","kjv_def":"fine (white) linen."},
"H949":{"lemma":"בּוֹצֵץ","xlit":"Bôwtsêts","pron":"bo-tsates'","derivation":"from the same as H948 (בּוּץ); shining;","strongs_def":"Botsets, a rock near Michmash","kjv_def":"Bozez."},
"H950":{"lemma":"בּוּקָה","xlit":"bûwqâh","pron":"boo-kaw'","derivation":"feminine passive participle of an unused root (meaning to be hollow);","strongs_def":"emptiness (as adjective)","kjv_def":"empty."},
"H951":{"lemma":"בּוֹקֵר","xlit":"bôwqêr","pron":"bo-kare'","derivation":"properly, active participle from H1239 (בָּקַר) as denominative from H1241 (בָּקָר);","strongs_def":"a cattle-tender","kjv_def":"herdman."},
"H952":{"lemma":"בּוּר","xlit":"bûwr","pron":"boor","derivation":"a primitive root;","strongs_def":"to bore, i.e. (figuratively) examine","kjv_def":"declare."},
"H953":{"lemma":"בּוֹר","xlit":"bôwr","pron":"bore","derivation":"from H952 (בּוּר) (in the sense of H877 (בֹּאר));","strongs_def":"a pit hole (especially one used as a cistern or a prison)","kjv_def":"cistern, dungeon, fountain, pit, well."},
"H954":{"lemma":"בּוּשׁ","xlit":"bûwsh","pron":"boosh","derivation":"a primitive root;","strongs_def":"properly, to pale, i.e. by implication to be ashamed; also (by implication) to be disappointed or delayed","kjv_def":"(be, make, bring to, cause, put to, with, a-) shamed(-d), be (put to) confounded(-fusion), become dry, delay, be long."},
"H955":{"lemma":"בּוּשָׁה","xlit":"bûwshâh","pron":"boo-shaw'","derivation":"feminine participle passive of H954 (בּוּשׁ);","strongs_def":"shame","kjv_def":"shame."},
"H956":{"lemma":"בּוּת","xlit":"bûwth","pron":"booth","derivation":"(Aramaic) apparent denominative from H1005 (בַּיִת);","strongs_def":"to lodge over night","kjv_def":"pass the night."},
"H957":{"lemma":"בַּז","xlit":"baz","pron":"baz","derivation":"from H962 (בָּזַז);","strongs_def":"plunder","kjv_def":"booty, prey, spoil(-ed)."},
"H958":{"lemma":"בָּזָא","xlit":"bâzâʼ","pron":"baw-zaw'","derivation":"a primitive root;","strongs_def":"probably to cleave","kjv_def":"spoil."},
"H959":{"lemma":"בָּזָה","xlit":"bâzâh","pron":"baw-zaw'","derivation":"a primitive root;","strongs_def":"to disesteem","kjv_def":"despise, disdain, contemn(-ptible), [phrase] think to scorn, vile person."},
"H960":{"lemma":"בָּזֹה","xlit":"bâzôh","pron":"baw-zo'","derivation":"from H959 (בָּזָה);","strongs_def":"scorned","kjv_def":"despise."},
"H961":{"lemma":"בִּזָּה","xlit":"bizzâh","pron":"biz-zaw'","derivation":"feminine of H957 (בַּז);","strongs_def":"booty","kjv_def":"prey, spoil."},
"H962":{"lemma":"בָּזַז","xlit":"bâzaz","pron":"baw-zaz'","derivation":"a primitive root;","strongs_def":"to plunder","kjv_def":"catch, gather, (take) for a prey, rob(-ber), spoil, take (away, spoil), [idiom] utterly."},
"H963":{"lemma":"בִּזָּיוֹן","xlit":"bizzâyôwn","pron":"biz-zaw-yone'","derivation":"from H959 (בָּזָה)","strongs_def":"disesteem","kjv_def":"contempt."},
"H964":{"lemma":"בִּזְיוֹתְיָה","xlit":"bizyôwthᵉyâh","pron":"biz-yo-the-yaw'","derivation":"from H959 (בָּזָה) and H3050 (יָהּ); contempts of Jah;","strongs_def":"Bizjothjah, a place in Palestine","kjv_def":"Bizjothjah."},
"H965":{"lemma":"בָּזָק","xlit":"bâzâq","pron":"baw-zawk'","derivation":"from an unused root meaning to lighten;","strongs_def":"a flash of lightning","kjv_def":"flash of lightning."},
"H966":{"lemma":"בֶּזֶק","xlit":"Bezeq","pron":"beh'-zak","derivation":"from H965 (בָּזָק); lightning;","strongs_def":"Bezek, a place in Palestine","kjv_def":"Bezek."},
"H967":{"lemma":"בָּזַר","xlit":"bâzar","pron":"baw-zar'","derivation":"a primitive root;","strongs_def":"to disperse","kjv_def":"scatter."},
"H968":{"lemma":"בִּזְתָא","xlit":"Bizthâʼ","pron":"biz-thaw'","derivation":"of Persian origin;","strongs_def":"Biztha, a eunuch of Xerxes","kjv_def":"Biztha."},
"H969":{"lemma":"בָּחוֹן","xlit":"bâchôwn","pron":"baw-khone'","derivation":"from H974 (בָּחַן);","strongs_def":"an assayer or metals","kjv_def":"tower."},
"H970":{"lemma":"בָּחוּר","xlit":"bâchûwr","pron":"baw-khoor'","derivation":"or בָּחֻר; participle passive of H977 (בָּחַר);","strongs_def":"properly, selected, i.e. a youth (often collective)","kjv_def":"(choice) young (man), chosen, [idiom] hole."},
"H971":{"lemma":"בַּחִין","xlit":"bachîyn","pron":"bakh-een'","derivation":"another form of H975 (בַּחַן);","strongs_def":"a watch-tower of besiegers","kjv_def":"tower."},
"H972":{"lemma":"בָּחִיר","xlit":"bâchîyr","pron":"baw-kheer'","derivation":"from H977 (בָּחַר);","strongs_def":"select","kjv_def":"choose, chosen one, elect."},
"H973":{"lemma":"בָּחַל","xlit":"bâchal","pron":"baw-khal'","derivation":"a primitive root;","strongs_def":"to loath","kjv_def":"abhor, get hastily (from the margin for H926 (בָּהַל))."},
"H974":{"lemma":"בָּחַן","xlit":"bâchan","pron":"baw-khan'","derivation":"a primitive root;","strongs_def":"to test (especially metals); generally and figuratively, to investigate","kjv_def":"examine, prove, tempt, try (trial)."},
"H975":{"lemma":"בַּחַן","xlit":"bachan","pron":"bakh'-an","derivation":"from H974 (בָּחַן) (in the sense of keeping a look-out);","strongs_def":"a watch-tower","kjv_def":"tower."},
"H976":{"lemma":"בֹּחַן","xlit":"bôchan","pron":"bo'-khan","derivation":"from H974 (בָּחַן);","strongs_def":"trial","kjv_def":"tried."},
"H977":{"lemma":"בָּחַר","xlit":"bâchar","pron":"baw-khar'","derivation":"a primitive root;","strongs_def":"properly, to try, i.e. (by implication) select","kjv_def":"acceptable, appoint, choose (choice), excellent, join, be rather, require."},
"H978":{"lemma":"בַּחֲרוּמִי","xlit":"Bachărûwmîy","pron":"bakh-ar-oo-mee'","derivation":"patrial from H980 (בַּחֻרִים) (by transposition);","strongs_def":"a Bacharumite or inhabitant of Bachurim","kjv_def":"Baharumite."},
"H979":{"lemma":"בְּחֻרוֹת","xlit":"bᵉchurôwth","pron":"bekh-oo-rothe'","derivation":"or בְּחוּרוֹת; feminine plural of H970 (בָּחוּר); also (masculine plural) בְּחֻרִים;","strongs_def":"youth (collectively and abstractly)","kjv_def":"young men, youth."},
"H980":{"lemma":"בַּחֻרִים","xlit":"Bachurîym","pron":"bakh-oo-reem'","derivation":"or בַּחוּרִים; masculine plural of H970 (בָּחוּר); young men;","strongs_def":"Bachurim, a place in Palestine","kjv_def":"Bahurim."},
"H981":{"lemma":"בָּטָא","xlit":"bâṭâʼ","pron":"baw-taw'","derivation":"or בָּטָה; a primitive root;","strongs_def":"to babble; hence, to vociferate angrily","kjv_def":"pronounce, speak (unadvisedly)."},
"H982":{"lemma":"בָּטַח","xlit":"bâṭach","pron":"baw-takh'","derivation":"a primitive root; properly, to hide for refuge (but not so precipitately as H2620 (חָסָה));","strongs_def":"figuratively, to trust, be confident or sure","kjv_def":"be bold (confident, secure, sure), careless (one, woman), put confidence, (make to) hope, (put, make to) trust."},
"H983":{"lemma":"בֶּטַח","xlit":"beṭach","pron":"beh'takh","derivation":"from H982 (בָּטַח);","strongs_def":"properly, a place of refuge; abstract, safety, both the fact (security) and the feeling (trust); often (adverb with or without preposition) safely","kjv_def":"assurance, boldly, (without) care(-less), confidence, hope, safe(-ly, -ty), secure, surely."},
"H984":{"lemma":"בֶּטַח","xlit":"Beṭach","pron":"beh'takh","derivation":"the same as H983 (בֶּטַח);","strongs_def":"Betach, a place in Syria","kjv_def":"Betah."},
"H985":{"lemma":"בִּטְחָה","xlit":"biṭchâh","pron":"bit-khaw'","derivation":"feminine of H984 (בֶּטַח);","strongs_def":"trust","kjv_def":"confidence."},
"H986":{"lemma":"בִּטָּחוֹן","xlit":"biṭṭâchôwn","pron":"bit-taw-khone'","derivation":"from H982 (בָּטַח);","strongs_def":"trust","kjv_def":"confidence, hope."},
"H987":{"lemma":"בַּטֻּחוֹת","xlit":"baṭṭuchôwth","pron":"bat-too-khoth'","derivation":"feminine plural from H982 (בָּטַח);","strongs_def":"security","kjv_def":"secure."},
"H988":{"lemma":"בָּטֵל","xlit":"bâṭêl","pron":"baw-tale'","derivation":"a primitive root;","strongs_def":"to desist from labor","kjv_def":"cease."},
"H989":{"lemma":"בְּטֵל","xlit":"bᵉṭêl","pron":"bet-ale'","derivation":"(Aramaic) corresponding to H988 (בָּטֵל);","strongs_def":"to stop","kjv_def":"(cause, make to), cease, hinder."},
"H990":{"lemma":"בֶּטֶן","xlit":"beṭen","pron":"beh'-ten","derivation":"from an unused root probably meaning to be hollow;","strongs_def":"the belly, especially the womb; also the bosom or body of anything","kjv_def":"belly, body, [phrase] as they be born, [phrase] within, womb."},
"H991":{"lemma":"בֶּטֶן","xlit":"Beṭen","pron":"beh'-ten","derivation":"the same as H990 (בֶּטֶן);","strongs_def":"Beten. a place in Palestine","kjv_def":"Beten."},
"H992":{"lemma":"בֹּטֶן","xlit":"bôṭen","pron":"bo'-ten","derivation":"from H990 (בֶּטֶן);","strongs_def":"(only in plural) a pistachio-nut (from its form)","kjv_def":"nut."},
"H993":{"lemma":"בְּטֹנִים","xlit":"Bᵉṭônîym","pron":"bet-o-neem'","derivation":"probably plural from H992 (בֹּטֶן); hollows;","strongs_def":"Betonim, a place in Palestine","kjv_def":"Betonim."},
"H994":{"lemma":"בִּי","xlit":"bîy","pron":"bee","derivation":"perhaps from H1158 (בָּעָה) (in the sense of asking); properly, a request; used only adverbially (always with 'my Lord');","strongs_def":"oh that!; with leave, or if it please","kjv_def":"alas, O, oh."},
"H995":{"lemma":"בִּין","xlit":"bîyn","pron":"bene","derivation":"a primitive root;","strongs_def":"to separate mentally (or distinguish), i.e.(generally) understand","kjv_def":"attend, consider, be cunning, diligently, direct, discern, eloquent, feel, inform, instruct, have intelligence, know, look well to, mark, perceive, be prudent, regard, (can) skill(-full), teach, think, (cause, make to, get, give, have) understand(-ing), view, (deal) wise(-ly, man)."},
"H996":{"lemma":"בֵּין","xlit":"bêyn","pron":"bane","derivation":"(sometimes in the plural masculine or feminine); properly, the constructive form of an otherwise unused noun from H995 (בִּין); a distinction; but used only as a preposition,","strongs_def":"between (repeated before each noun, often with other particles); also as a conjunction, either...or","kjv_def":"among, asunder, at, between (-twixt...and), [phrase] from (the widest), [idiom] in, out of, whether (it be...or), within."},
"H997":{"lemma":"בֵּין","xlit":"bêyn","pron":"bane","derivation":"(Aramaic) corresponding to H996 (בֵּין)","strongs_def":"{between (repeated before each noun, often with other particles); also as a conjunction, either...or}","kjv_def":"among, between."},
"H998":{"lemma":"בִּינָה","xlit":"bîynâh","pron":"bee-naw'","derivation":"from H995 (בִּין);","strongs_def":"understanding","kjv_def":"knowledge, meaning, [idiom] perfectly, understanding, wisdom."},
"H999":{"lemma":"בִּינָה","xlit":"bîynâh","pron":"bee-naw'","derivation":"(Aramaic) corresponding to H998 (בִּינָה)","strongs_def":"{understanding}","kjv_def":"knowledge."},
"H1000":{"lemma":"בֵּיצָה","xlit":"bêytsâh","pron":"bay-tsaw'","derivation":"from the same as H948 (בּוּץ);","strongs_def":"an egg (from its whiteness)","kjv_def":"egg."},
"H1001":{"lemma":"בִּירָא","xlit":"bîyrâʼ","pron":"bee-raw'","derivation":"(Aramaic) corresponding to H1002 (בִּירָה);","strongs_def":"a palace","kjv_def":"palace."},
"H1002":{"lemma":"בִּירָה","xlit":"bîyrâh","pron":"bee-raw'","derivation":"of foreign origin;","strongs_def":"a castle or palace","kjv_def":"palace."},
"H1003":{"lemma":"בִּירָנִית","xlit":"bîyrânîyth","pron":"bee-raw-neeth'","derivation":"from H1002 (בִּירָה);","strongs_def":"a fortress","kjv_def":"castle."},
"H1004":{"lemma":"בַּיִת","xlit":"bayith","pron":"bah'-yith","derivation":"probably from H1129 (בָּנָה) abbreviated;","strongs_def":"a house (in the greatest variation of applications, especially family, etc.)","kjv_def":"court, daughter, door, [phrase] dungeon, family, [phrase] forth of, [idiom] great as would contain, hangings, home(born), (winter) house(-hold), inside(-ward), palace, place, [phrase] prison, [phrase] steward, [phrase] tablet, temple, web, [phrase] within(-out)."},
"H1005":{"lemma":"בַּיִת","xlit":"bayith","pron":"bah-yith","derivation":"(Aramaic) corresponding to H1004 (בַּיִת)","strongs_def":"{a house (in the greatest variation of applications, especially family, etc.)}","kjv_def":"house."},
"H1006":{"lemma":"בַּיִת","xlit":"Bayith","pron":"bah'-yith","derivation":"the same as H1004 (בַּיִת);","strongs_def":"Bajith, a place in Palestine","kjv_def":"Bajith."},
"H1007":{"lemma":"בֵּית אָוֶן","xlit":"Bêyth ʼÂven","pron":"bayth aw'-ven","derivation":"from H1004 (בַּיִת) and H205 (אָוֶן); house of vanity;","strongs_def":"Beth-Aven, a place in Palestine","kjv_def":"Beth-aven."},
"H1008":{"lemma":"בֵּית־אֵל","xlit":"Bêyth-ʼÊl","pron":"bayth-ale'","derivation":"from H1004 (בַּיִת) and H410 (אֵל); house of God;","strongs_def":"Beth-El, a place in Palestine","kjv_def":"Beth-el."},
"H1009":{"lemma":"בֵּית אַרְבֵּאל","xlit":"Bêyth ʼArbêʼl","pron":"bayth ar-bale'","derivation":"from H1004 (בַּיִת) and H695 (אֶרֶב) and H410 (אֵל); house of God's ambush;","strongs_def":"Beth-Arbel, a place in Palestine","kjv_def":"Beth-Arbel."},
"H1010":{"lemma":"בֵּית בַּעַל מְעוֹן","xlit":"Bêyth Baʻal Mᵉʻôwn","pron":"bayth bah'-al me-own'","derivation":"from H1004 (בַּיִת) and H1168 (בַּעַל) and H4583 (מָעוֹן); house of Baal of (the) habitation of (apparently by transposition); or (shorter) בֵּית מְעוֹןlemma בֵּית מעוֹן missing vowel, corrected to בֵּית מְעוֹן; house of habitation of (Baal);","strongs_def":"Beth-Baal-Meon, a place in Palestine","kjv_def":"Beth-baal-meon. Compare H1186 (בַּעַל מְעוֹן) and H1194 (בְּעֹן)."},
"H1011":{"lemma":"בֵּית בִּרְאִי","xlit":"Bêyth Birʼîy","pron":"bayth bir-ee'","derivation":"from H1004 (בַּיִת) and H1254 (בָּרָא); house of a creative one;","strongs_def":"Beth-Biri, a place in Palestine","kjv_def":"Bethbirei."},
"H1012":{"lemma":"בֵּית בָּרָה","xlit":"Bêyth Bârâh","pron":"bayth baw-raw'","derivation":"probably from H1004 (בַּיִת) and H5679 (עֲבָרָה); house of (the) ford;","strongs_def":"Beth-Barah, a place in Palestine","kjv_def":"Beth-barah."},
"H1013":{"lemma":"בֵּית־גָּדֵר","xlit":"Bêyth-Gâdêr","pron":"bayth-gaw-dare'","derivation":"from H1004 (בַּיִת) and H1447 (גָּדֵר); house of (the) wall;","strongs_def":"Beth-Gader, a place in Palestine","kjv_def":"Beth-gader."},
"H1014":{"lemma":"בֵּית גָּמוּל","xlit":"Bêyth Gâmûwl","pron":"bayth gaw-mool'","derivation":"from H1004 (בַּיִת) and the passive participle of H1576 (גְּמוּל); house of (the) weaned;","strongs_def":"Beth-Gamul, a place East of the Jordan","kjv_def":"Beth-gamul."},
"H1015":{"lemma":"בֵּית דִּבְלָתַיִם","xlit":"Bêyth Diblâthayim","pron":"bayth dib-law-thah'-yim","derivation":"from H1004 (בַּיִת) and the dual of H1690 (דְּבֵלָה); house of (the) two figcakes;","strongs_def":"Beth-Diblathajim, a place East of the Jordan","kjv_def":"Beth-diblathaim."},
"H1016":{"lemma":"בֵּית־דָּגוֹן","xlit":"Bêyth-Dâgôwn","pron":"bayth-daw-gohn'","derivation":"from H1004 (בַּיִת) and H1712 (דָּגוֹן); house of Dagon;","strongs_def":"Beth-Dagon, the name of two places in Palestine","kjv_def":"Beth-dagon."},
"H1017":{"lemma":"בֵּית הָאֱלִי","xlit":"Bêyth hâ-ʼĔlîy","pron":"bayth haw-el-ee'","derivation":"patrial from H1008 (בֵּית־אֵל) with the article interposed;","strongs_def":"a Beth-elite, or inhabitant of Bethel","kjv_def":"Bethelite."},
"H1018":{"lemma":"בֵּית הָאֵצֶל","xlit":"Bêyth hâʼêtsel","pron":"bayth haw-ay'-tsel","derivation":"from H1004 (בַּיִת) and H681 (אֵצֶל) with the article interposed; house of the side;","strongs_def":"Beth-ha-Etsel, a place in Palestine","kjv_def":"Beth-ezel."},
"H1019":{"lemma":"בֵּית הַגִּלְגָּל","xlit":"Bêyth hag-Gilgâl","pron":"bayth hag-gil gawl'","derivation":"from H1004 (בַּיִת) and H1537 (גִּלְגָּל) with the article interposed; house of Gilgal (or rolling);","strongs_def":"Beth-hag-Gilgal, a place in Palestine","kjv_def":"Beth-gilgal."},
"H1020":{"lemma":"בֵּית הַיְשִׁימוֹת","xlit":"Bêyth ha-Yshîy-môwth","pron":"bayth hah-yesh-eemoth'","derivation":"from H1004 (בַּיִת) and the plural of H3451 (יְשִׁימַה) with the article interposed; house of the deserts;","strongs_def":"Beth-ha-Jeshimoth, a town East of the Jordan","kjv_def":"Beth-jeshimoth."},
"H1021":{"lemma":"בֵּית הַכֶּרֶם","xlit":"Bêyth hak-Kerem","pron":"bayth hak-keh'-rem","derivation":"from H1004 (בַּיִת) and H3754 (כֶּרֶם) with the article interposed; house of the vineyard;","strongs_def":"Beth-hak-Kerem, a place in Palestine","kjv_def":"Beth-haccerem."},
"H1022":{"lemma":"בֵּית הַלַּחְמִי","xlit":"Bêyth hal-Lachmîy","pron":"bayth hal-lakh-mee'","derivation":"patrial from H1035 (בֵּית לֶחֶם) with the article inserted;","strongs_def":"a Beth-lechemite, or native of Bethlechem","kjv_def":"Bethlehemite."},
"H1023":{"lemma":"בֵּית הַמֶּרְחָק","xlit":"Bêyth ham-Merchâq","pron":"bayth ham-mer-khawk'","derivation":"from H1004 (בַּיִת) and H4801 (מֶרְחָק) with the article interposed; house of the breadth;","strongs_def":"Beth-ham-Merchak, a place in Palestine","kjv_def":"place that was far off."},
"H1024":{"lemma":"בֵּית הַמַּרְכָּבוֹת","xlit":"Bêyth ham-Markâbôwth","pron":"bayth ham-markaw-both'","derivation":"or (shortened) בֵּית מַרְכָּבוֹת; from H1004 (בַּיִת) and the plural of H4818 (מֶרְכָּבָה) (with or without the article interposed); place of (the) chariots;","strongs_def":"Beth-ham-Markaboth or Beth-Markaboth, a place in Palestine","kjv_def":"Bethmarcaboth."},
"H1025":{"lemma":"בֵּית הָעֵמֶק","xlit":"Bêyth hâ-ʻÊmeq","pron":"bayth haw-Ay'-mek","derivation":"from H1004 (בַּיִת) and H6010 (עֵמֶק) with the article interposed; house of the valley;","strongs_def":"Beth-ha-Emek, a place in Palestine","kjv_def":"Beth-emek."},
"H1026":{"lemma":"בֵּית הָעֲרָבָה","xlit":"Bêyth hâ-ʻĂrâbâh","pron":"bayth haw-ar-aw-baw","derivation":"from H1004 (בַּיִת) and H6160 (עֲרָבָה) with the article interposed; house of the Desert;","strongs_def":"Beth-ha-Arabah, a place in Palestine","kjv_def":"Beth-arabah."},
"H1027":{"lemma":"בֵּית הָרָם","xlit":"Bêyth hâ-Râm","pron":"bayth haw-rawm'","derivation":"from H1004 (בַּיִת) and H7311 (רוּם) with the article interposed; house of the height;","strongs_def":"Beth-ha-Ram, a place East of the Jordan","kjv_def":"Beth-aram."},
"H1028":{"lemma":"בֵּית הָרָן","xlit":"Bêyth hâ-Rân","pron":"bayth haw-rawn'","derivation":"probably for H1027 (בֵּית הָרָם);","strongs_def":"Beth-ha-Ram, a place East of the Jordan","kjv_def":"Beth-haran."},
"H1029":{"lemma":"בֵּית הַשִּׁטָּה","xlit":"Bêyth hash-Shiṭṭâh","pron":"bayth hash-shit-taw'","derivation":"from H1004 (בַּיִת) and H7848 (שִׁטָּה) with the article interposed; house of the acacia;","strongs_def":"Beth-hash-Shittah, a place in Palestine","kjv_def":"Beth-shittah."},
"H1030":{"lemma":"בֵּית הַשִּׁמְשִׁי","xlit":"Bêyth hash-Shimshîy","pron":"bayth hash-shim-shee'","derivation":"patrial from H1053 (בֵּית שֶׁמֶשׁ) with the article inserted;","strongs_def":"a Beth-shimshite, or inhabitant of Bethshemesh","kjv_def":"Bethshemite."},
"H1031":{"lemma":"בֵּית חׇגְלָה","xlit":"Bêyth Choglâh","pron":"bayth chog-law'","derivation":"from H1004 (בַּיִת) and the same as H2295 (חׇגְלָה); house of a partridge;","strongs_def":"Beth-Choglah, a place in Palestine","kjv_def":"Beth-hoglah."},
"H1032":{"lemma":"בֵּית חוֹרוֹן","xlit":"Bêyth Chôwrôwn","pron":"bayth kho-rone'","derivation":"from H1004 (בַּיִת) and H2356 (חוֹר); house of hollowness;","strongs_def":"Beth-Choron, the name of two adjoining places in Palestine","kjv_def":"Beth-horon."},
"H1033":{"lemma":"בֵּית כַּר","xlit":"Bêyth Kar","pron":"bayth kar","derivation":"from H1004 (בַּיִת) and H3733 (כַּר); house of pasture;","strongs_def":"Beth-Car, a place in Palestine","kjv_def":"Beth-car."},
"H1034":{"lemma":"בֵּית לְבָאוֹת","xlit":"Bêyth Lᵉbâʼôwth","pron":"bayth leb-aw-oth'","derivation":"from H1004 (בַּיִת) and the plural of H3833 (לָבִיא); house of lionesses;","strongs_def":"Beth-Lebaoth, a place in Palestine","kjv_def":"Beth-lebaoth. Compare H3822 (לְבָאוֹת)."},
"H1035":{"lemma":"בֵּית לֶחֶם","xlit":"Bêyth Lechem","pron":"bayth leh'-khem","derivation":"from H1004 (בַּיִת) and H3899 (לֶחֶם); house of bread;","strongs_def":"Beth-Lechem, a place in Palestine","kjv_def":"Bethlehem."},
"H1036":{"lemma":"בֵּית לְעַפְרָה","xlit":"Bêyth lᵉ-ʻAphrâh","pron":"bayth le-af-raw'","derivation":"from H1004 (בַּיִת) and the feminine of H6083 (עָפָר) (with preposition interposed); house to (i.e. of) dust;","strongs_def":"Beth-le-Aphrah, a place in Palestine","kjv_def":"house of Aphrah."},
"H1037":{"lemma":"בֵּית מִלּוֹא","xlit":"Bêyth Millôwʼ","pron":"bayth mil-lo'","derivation":"or בֵּית מִלֹּא; from H1004 (בַּיִת) and H4407 (מִלּוֹא); house of (the) rampart;","strongs_def":"Beth-Millo, the name of two citadels","kjv_def":"house of Millo."},
"H1038":{"lemma":"בֵּית מַעֲכָה","xlit":"Bêyth Maʻăkâh","pron":"bayth mah-ak-aw'","derivation":"from H1004 (בַּיִת) and H4601 (מַעֲכָה); house of Maakah;","strongs_def":"Beth-Maakah, a place in Palestine","kjv_def":"Beth-maachah."},
"H1039":{"lemma":"בֵּית נִמְרָה","xlit":"Bêyth Nimrâh","pron":"bayth nim-raw'","derivation":"from H1004 (בַּיִת) and the feminine of H5246 (נָמֵר); house of (the) leopard;","strongs_def":"Beth-Nimrah, a place east of the Jordan","kjv_def":"Beth-Nimrah. Compare H5247 (נִמְרָה)."},
"H1040":{"lemma":"בֵּית עֵדֶן","xlit":"Bêyth ʻÊden","pron":"bayth ay'-den","derivation":"from H1004 (בַּיִת) and H5730 (עֵדֶן); house of pleasure;","strongs_def":"Beth-Eden, a place in Syria","kjv_def":"Beth-eden."},
"H1041":{"lemma":"בֵּית עַזְמָוֶת","xlit":"Bêyth ʻAzmâveth","pron":"bayth az-maw'-veth","derivation":"from H1004 (בַּיִת) and H5820 (עַזְמָוֶת);","strongs_def":"house of Azmaveth, a place in Palestine","kjv_def":"Bethaz-maveth. Compare H5820 (עַזְמָוֶת)."},
"H1042":{"lemma":"בֵּית עֲנוֹת","xlit":"Bêyth ʻĂnôwth","pron":"bayth an-oth'","derivation":"from H1004 (בַּיִת) and a plural from H6030 (עָנָה); house of replies;","strongs_def":"Beth-Anoth, a place in Palestine","kjv_def":"Beth-anoth."},
"H1043":{"lemma":"בֵּית עֲנָת","xlit":"Bêyth ʻĂnâth","pron":"bayth an-awth'","derivation":"an orthographical variation for H1042 (בֵּית עֲנוֹת);","strongs_def":"Beth-Anath, a place in Palestine","kjv_def":"Beth-anath."},
"H1044":{"lemma":"בֵּית עֵקֶד","xlit":"Bêyth ʻÊqed","pron":"bayth ay'-ked","derivation":"from H1004 (בַּיִת) and a derivative of H6123 (עָקַד); house of (the) binding (for sheep-shearing);","strongs_def":"Beth-Eked, a place in Palestine","kjv_def":"shearing house."},
"H1045":{"lemma":"בֵּית עַשְׁתָּרוֹת","xlit":"Bêyth ʻAshtârôwth","pron":"bayth ash-taw-roth'","derivation":"; from H1004 (בַּיִת) and H6252 (עַשְׁתָּרוֹת); house of Ashtoreths;","strongs_def":"Beth-Ashtaroth, a place in Palestine","kjv_def":"house of Ashtaroth. Compare H1203 (בְּעֶשְׁתְּרָה), H6252 (עַשְׁתָּרוֹת)."},
"H1046":{"lemma":"בֵּית פֶּלֶט","xlit":"Bêyth Peleṭ","pron":"bayth peh'-let","derivation":"from H1004 (בַּיִת) and H6412 (פָּלִיט); house of escape;","strongs_def":"Beth-Palet, a place in Palestine","kjv_def":"Beth-palet."},
"H1047":{"lemma":"בֵּית פְּעוֹר","xlit":"Bêyth Pᵉʻôwr","pron":"bayth pe-ore'","derivation":"from H1004 (בַּיִת) and H6465 (פְּעוֹר); house of Peor;","strongs_def":"Beth-Peor, a place East of the Jordan","kjv_def":"Bethpeor."},
"H1048":{"lemma":"בֵּית פַּצֵּץ","xlit":"Bêyth Patstsêts","pron":"bayth pats-tsates'","derivation":"from H1004 (בַּיִת) and a derivative from H6327 (פּוּץ); house of dispersion;","strongs_def":"Beth-Patstsets, a place in Palestine","kjv_def":"Beth-pazzez."},
"H1049":{"lemma":"בֵּית צוּר","xlit":"Bêyth Tsûwr","pron":"bayth tsoor'","derivation":"from H1004 (בַּיִת) and H6697 (צוּר); house of (the) rock;","strongs_def":"Beth-Tsur, a place in Palestine","kjv_def":"Beth-zur."},
"H1050":{"lemma":"בֵּית רְחוֹב","xlit":"Bêyth Rᵉchôwb","pron":"bayth re-khobe'","derivation":"from H1004 (בַּיִת) and H7339 (רְחֹב); house of (the) street;","strongs_def":"Beth-Rechob, a place in Palestine","kjv_def":"Beth-rehob."},
"H1051":{"lemma":"בֵּית רָפָא","xlit":"Bêyth Râphâʼ","pron":"bayth raw-faw'","derivation":"from H1004 (בַּיִת) and H7497 (רָפָא); house of (the) giant;","strongs_def":"Beth-Rapha, an Israelite","kjv_def":"Beth-rapha."},
"H1052":{"lemma":"בֵּית שְׁאָן","xlit":"Bêyth Shᵉʼân","pron":"bayth she-awn'","derivation":"or בֵּית שָׁן; from H1004 (בַּיִת) and H7599 (שָׁאַן); house of ease;","strongs_def":"Beth-Shean or Beth-Shan, a place in Palestine","kjv_def":"Beth-shean, Beth-Shan."},
"H1053":{"lemma":"בֵּית שֶׁמֶשׁ","xlit":"Bêyth Shemesh","pron":"bayth sheh'-mesh","derivation":"from H1004 (בַּיִת) and H8121 (שֶׁמֶשׁ); house of (the) sun;","strongs_def":"Beth-Shemesh, a place in Palestine","kjv_def":"Beth-shemesh."},
"H1054":{"lemma":"בֵּית תַּפּוּחַ","xlit":"Bêyth Tappûwach","pron":"bayth tap-poo'-akh","derivation":"from H1004 (בַּיִת) and H8598 (תַּפּוּחַ); house of (the) apple;","strongs_def":"Beth-Tappuach, a place in Palestine","kjv_def":"Beth-tappuah."},
"H1055":{"lemma":"בִּיתָן","xlit":"bîythân","pron":"bee-thawn'","derivation":"probably from H1004 (בַּיִת);","strongs_def":"a palace (i.e. large house)","kjv_def":"palace."},
"H1056":{"lemma":"בָּכָא","xlit":"Bâkâʼ","pron":"baw-kaw'","derivation":"from H1058 (בָּכָה), weeping;","strongs_def":"Baca, a valley in Palestine","kjv_def":"Baca."},
"H1057":{"lemma":"בָּכָא","xlit":"bâkâʼ","pron":"baw-kaw'","derivation":"the same as H1056 (בָּכָא);","strongs_def":"the weeping tree (some gum-distilling tree, perhaps the balsam)","kjv_def":"mulberry tree."},
"H1058":{"lemma":"בָּכָה","xlit":"bâkâh","pron":"baw-kaw'","derivation":"a primitive root;","strongs_def":"to weep; generally to bemoan","kjv_def":"[idiom] at all, bewail, complain, make lamentation, [idiom] more, mourn, [idiom] sore, [idiom] with tears, weep."},
"H1059":{"lemma":"בֶּכֶה","xlit":"bekeh","pron":"beh'-keh","derivation":"from H1058 (בָּכָה);","strongs_def":"a weeping","kjv_def":"[idiom] sore."},
"H1060":{"lemma":"בְּכוֹר","xlit":"bᵉkôwr","pron":"bek-ore'","derivation":"from H1069 (בָּכַר);","strongs_def":"firstborn; hence, chief","kjv_def":"eldest (son), firstborn(-ling)."},
"H1061":{"lemma":"בִּכּוּר","xlit":"bikkûwr","pron":"bik-koor'","derivation":"from H1069 (בָּכַר);","strongs_def":"the first-fruits of the crop","kjv_def":"first fruit (-ripe (figuratively)), hasty fruit."},
"H1062":{"lemma":"בְּכוֹרָה","xlit":"bᵉkôwrâh","pron":"bek-o-raw'","derivation":"or (short) בְּכֹרָה; feminine of H1060 (בְּכוֹר);","strongs_def":"the firstling of man or beast; abstractly primogeniture","kjv_def":"birthright, firstborn(-ling)."},
"H1063":{"lemma":"בִּכּוּרָה","xlit":"bikkûwrâh","pron":"bik-koo-raw'","derivation":"feminine of H1061 (בִּכּוּר);","strongs_def":"the early fig","kjv_def":"firstripe (fruit)."},
"H1064":{"lemma":"בְּכוֹרַת","xlit":"Bᵉkôwrath","pron":"bek-o-rath'","derivation":"feminine of H1062 (בְּכוֹרָה); primogeniture;","strongs_def":"Bekorath, an Israelite","kjv_def":"Bechorath."},
"H1065":{"lemma":"בְּכִי","xlit":"Bᵉkîy","pron":"bek-ee'","derivation":"from H1058 (בָּכָה);","strongs_def":"a weeping; by analogy, a dripping","kjv_def":"overflowing, [idiom] sore, (continual) weeping, wept."},
"H1066":{"lemma":"בֹּכִים","xlit":"Bôkîym","pron":"bo-keem'","derivation":"plural active participle of H1058 (בָּכָה); (with the article) the weepers;","strongs_def":"Bo-kim, a place in Palestine","kjv_def":"Bochim."},
"H1067":{"lemma":"בְּכִירָה","xlit":"bᵉkîyrâh","pron":"bek-ee-raw'","derivation":"feminine from H1069 (בָּכַר);","strongs_def":"the eldest daughter","kjv_def":"firstborn."},
"H1068":{"lemma":"בְּכִית","xlit":"bᵉkîyth","pron":"bek-eeth'","derivation":"from H1058 (בָּכָה);","strongs_def":"a weeping","kjv_def":"mourning."},
"H1069":{"lemma":"בָּכַר","xlit":"bâkar","pron":"baw-kar'","derivation":"a primitive root; properly, to burst the womb, i.e. (causatively) bear or make early fruit (of woman or tree); also (as denominative from H1061 (בִּכּוּר))","strongs_def":"to give the birthright","kjv_def":"make firstborn, be firstling, bring forth first child (new fruit)."},
"H1070":{"lemma":"בֶּכֶר","xlit":"beker","pron":"beh'-ker","derivation":"from H1069 (בָּכַר) (in the sense of youth);","strongs_def":"a young camel","kjv_def":"dromedary."},
"H1071":{"lemma":"בֶכֶר","xlit":"Beker","pron":"beh'-ker","derivation":"the same as H1070 (בֶּכֶר);","strongs_def":"Beker, the name of two Israelites","kjv_def":"Becher."},
"H1072":{"lemma":"בִּכְרָה","xlit":"bikrâh","pron":"bik-raw'","derivation":"feminine of H1070 (בֶּכֶר);","strongs_def":"a young she-camel","kjv_def":"dromedary."},
"H1073":{"lemma":"בַּכֻּרָה","xlit":"bakkurâh","pron":"bak-koo-raw'","derivation":"by orthographical variation for H1063 (בִּכּוּרָה);","strongs_def":"a first-ripe fig","kjv_def":"firstripe."},
"H1074":{"lemma":"בֹּכְרוּ","xlit":"Bôkᵉrûw","pron":"bo-ker-oo'","derivation":"from H1069 (בָּכַר); first-born;","strongs_def":"Bokeru, an Israelite","kjv_def":"Bocheru."},
"H1075":{"lemma":"בִּכְרִי","xlit":"Bikrîy","pron":"bik-ree'","derivation":"from H1069 (בָּכַר); youth-ful;","strongs_def":"Bikri, an Israelite","kjv_def":"Bichri."},
"H1076":{"lemma":"בַּכְרִי","xlit":"Bakrîy","pron":"bak-ree'","derivation":"patronymically from H1071 (בֶכֶר);","strongs_def":"a Bakrite (collectively) or descendants of Beker","kjv_def":"Bachrites."},
"H1077":{"lemma":"בַּל","xlit":"bal","pron":"bal","derivation":"from H1086 (בָּלָה);","strongs_def":"properly, a failure; by implication nothing; usually (adverb) not at all; also lest","kjv_def":"lest, neither, no, none (that...), not (any), nothing."},
"H1078":{"lemma":"בֵּל","xlit":"Bêl","pron":"bale","derivation":"by contraction for H1168 (בַּעַל);","strongs_def":"Bel, the Baal of the Babylonians","kjv_def":"Bel."},
"H1079":{"lemma":"בָּל","xlit":"bâl","pron":"bawl","derivation":"(Aramaic) from H1080 (בְּלָא);","strongs_def":"properly, anxiety, i.e. (by implication) the heart (as its seat)","kjv_def":"heart."},
"H1080":{"lemma":"בְּלָא","xlit":"bᵉlâʼ","pron":"bel-aw'","derivation":"(Aramaic) corresponding to H1086 (בָּלָה) (but used only in a mental sense);","strongs_def":"to afflict","kjv_def":"wear out."},
"H1081":{"lemma":"בַּלְאֲדָן","xlit":"Balʼădân","pron":"bal-ad-awn'","derivation":"from H1078 (בֵּל) and H113 (אָדוֹן) (contracted); Bel (is his) lord;","strongs_def":"Baladan, the name of a Babylonian prince","kjv_def":"Baladan."},
"H1082":{"lemma":"בָּלַג","xlit":"bâlag","pron":"baw-lag'","derivation":"a primitive root;","strongs_def":"to break off or loose (in a favorable or unfavorable sense), i.e. desist (from grief) or invade (with destruction)","kjv_def":"comfort, (recover) strength(-en)."},
"H1083":{"lemma":"בִּלְגַה","xlit":"Bilgah","pron":"bil-gaw'","derivation":"from H1082 (בָּלַג); desistance;","strongs_def":"Bilgah, the name of two Israelites","kjv_def":"Bilgah."},
"H1084":{"lemma":"בִּלְגַּי","xlit":"Bilgay","pron":"bil-gah'ee","derivation":"from H1082 (בָּלַג); desistant;","strongs_def":"Bilgai, an Israelite","kjv_def":"Bilgai."},
"H1085":{"lemma":"בִּלְדַּד","xlit":"Bildad","pron":"bil-dad'","derivation":"of uncertain derivation;","strongs_def":"Bildad, one of Job's friends","kjv_def":"Bildad."},
"H1086":{"lemma":"בָּלָה","xlit":"bâlâh","pron":"baw-law'","derivation":"a primitive root;","strongs_def":"to fail; by implication to wear out, decay (causatively, consume, spend)","kjv_def":"consume, enjoy long, become (make, wax) old, spend, waste."},
"H1087":{"lemma":"בָּלֶה","xlit":"bâleh","pron":"baw-leh'","derivation":"from H1086 (בָּלָה);","strongs_def":"worn out","kjv_def":"old."},
"H1088":{"lemma":"בָּלָה","xlit":"Bâlâh","pron":"baw-law'","derivation":"feminine of H1087 (בָּלֶה); failure;","strongs_def":"Balah, a place in Palestine","kjv_def":"Balah."},
"H1089":{"lemma":"בָּלַהּ","xlit":"bâlahh","pron":"baw-lah'","derivation":"a primitive root (rather by transposition for H926 (בָּהַל));","strongs_def":"to palpitate; hence, (causatively) to terrify","kjv_def":"trouble."},
"H1090":{"lemma":"בִּלְהָה","xlit":"Bilhâh","pron":"bil-haw'","derivation":"from H1089 (בָּלַהּ); timid;","strongs_def":"Bilhah, the name of one of Jacob's concubines; also of a place in Palestine","kjv_def":"Bilhah."},
"H1091":{"lemma":"בַּלָּהָה","xlit":"ballâhâh","pron":"bal-law-haw'","derivation":"from H1089 (בָּלַהּ);","strongs_def":"alarm; hence, destruction","kjv_def":"terror, trouble."},
"H1092":{"lemma":"בִּלְהָן","xlit":"Bilhân","pron":"bil-hawn'","derivation":"from H1089 (בָּלַהּ); timid;","strongs_def":"Bilhan, the name of an Edomite and of an Israelite","kjv_def":"Bilhan."},
"H1093":{"lemma":"בְּלוֹ","xlit":"bᵉlôw","pron":"bel-o'","derivation":"(Aramaic) from a root corresponding to H1086 (בָּלָה);","strongs_def":"excise (on articles consumed)","kjv_def":"tribute."},
"H1094":{"lemma":"בְּלוֹא","xlit":"bᵉlôwʼ","pron":"bel-o'","derivation":"or (fully) בְּלוֹי; from H1086 (בָּלָה);","strongs_def":"(only in plural construction) rags","kjv_def":"old."},
"H1095":{"lemma":"בֵּלְטְשַׁאצַּר","xlit":"Bêlṭᵉshaʼtstsar","pron":"bale-tesh-ats-tsar'","derivation":"of foreign derivation;","strongs_def":"Belteshatstsar, the Babylonian name of Daniel","kjv_def":"Belteshazzar."},
"H1096":{"lemma":"בֵּלְטְשַׁאצַּר","xlit":"Bêlṭᵉshaʼtstsar","pron":"bale-tesh-ats-tsar'","derivation":"(Aramaic) corresponding to H1095 (בֵּלְטְשַׁאצַּר)","strongs_def":"{Belteshatstsar, the Babylonian name of Daniel}","kjv_def":"Belteshazzar."},
"H1097":{"lemma":"בְּלִי","xlit":"bᵉlîy","pron":"bel-ee'","derivation":"from H1086 (בָּלָה);","strongs_def":"properly, failure, i.e. nothing or destruction; usually (with preposition) without, not yet, because not, as long as, etc.","kjv_def":"corruption, ig(norantly), for lack of, where no...is, so that no, none, not, un(awares), without."},
"H1098":{"lemma":"בְּלִיל","xlit":"bᵉlîyl","pron":"bel-eel'","derivation":"from H1101 (בָּלַל);","strongs_def":"mixed, i.e. (specifically) feed (for cattle)","kjv_def":"corn, fodder, provender."},
"H1099":{"lemma":"בְּלִימָה","xlit":"bᵉlîymâh","pron":"bel-ee-mah'","derivation":"from H1097 (בְּלִי) and H4100 (מָה);","strongs_def":"(as indefinitely) nothing whatever","kjv_def":"nothing."},
"H1100":{"lemma":"בְּלִיַּעַל","xlit":"bᵉlîyaʻal","pron":"bel-e-yah'-al","derivation":"from H1097 (בְּלִי) and H3276 (יַעַל);(often in connection with H376 (אִישׁ), H802 (אִשָּׁה), H1121 (בֵּן), etc.)","strongs_def":"without profit, worthlessness; by extension, destruction, wickedness","kjv_def":"Belial, evil, naughty, ungodly (men), wicked."},
"H1101":{"lemma":"בָּלַל","xlit":"bâlal","pron":"baw-lal'","derivation":"a primitive root; (also  denominatively from H1098 (בְּלִיל)) to fodder","strongs_def":"to overflow (specifically with oil.); by implication, to mix; to fodder","kjv_def":"anoint, confound, [idiom] fade, mingle, mix (self), give provender, temper."},
"H1102":{"lemma":"בָּלַם","xlit":"bâlam","pron":"baw-lam'","derivation":"a primitive root;","strongs_def":"to muzzle","kjv_def":"be held in."},
"H1103":{"lemma":"בָּלַס","xlit":"bâlaç","pron":"baw-las'","derivation":"a primitive root;","strongs_def":"to pinch sycamore figs (a process necessary to ripen them)","kjv_def":"gatherer."},
"H1104":{"lemma":"בָּלַע","xlit":"bâlaʻ","pron":"baw-lah'","derivation":"a primitive root;","strongs_def":"to make away with (specifically by swallowing); generally, to destroy","kjv_def":"cover, destroy, devour, eat up, be at end, spend up, swallow down (up)."},
"H1105":{"lemma":"בֶּלַע","xlit":"belaʻ","pron":"beh'-lah","derivation":"from H1104 (בָּלַע);","strongs_def":"a gulp; figuratively, destruction","kjv_def":"devouring, that which he hath swallowed up."},
"H1106":{"lemma":"בֶּלַע","xlit":"Belaʻ","pron":"beh'-lah","derivation":"the same as H1105 (בֶּלַע);","strongs_def":"Bela, the name of a place, also of an Edomite and of two Israelites","kjv_def":"Bela."},
"H1107":{"lemma":"בִּלְעֲדֵי","xlit":"bilʻădêy","pron":"bil-ad-ay'","derivation":"or בַּלְעֲדֵי; constructive plural from H1077 (בַּל) and H5703 (עַד), not till, i.e. (as preposition or adverb)","strongs_def":"except, without, besides","kjv_def":"beside, not (in), save, without."},
"H1108":{"lemma":"בַּלְעִי","xlit":"Balʻîy","pron":"bel-ee'","derivation":"patronymically from H1106 (בֶּלַע);","strongs_def":"a Belaite (collectively) or descendants of Bela","kjv_def":"Belaites."},
"H1109":{"lemma":"בִּלְעָם","xlit":"Bilʻâm","pron":"bil-awm'","derivation":"probably from H1077 (בַּל) and H5971 (עַם); not (of the) people, i.e. foreigner;","strongs_def":"Bilam, a Mesopotamian prophet; also a place in Palestine","kjv_def":"Balaam, Bileam."},
"H1110":{"lemma":"בָּלַק","xlit":"bâlaq","pron":"baw-lak'","derivation":"a primitive root;","strongs_def":"to annihilate","kjv_def":"(make) waste."},
"H1111":{"lemma":"בָּלָק","xlit":"Bâlâq","pron":"baw-lawk'","derivation":"from H1110 (בָּלַק); waster;","strongs_def":"Balak, a Moabitish king","kjv_def":"Balak."},
"H1112":{"lemma":"בֵּלְשַׁאצַּר","xlit":"Bêlshaʼtstsar","pron":"bale-shats-tsar'","derivation":"or בֵּלְאשַׁצַּרlemma בֵּלְאשַׁצּר missing vowel, corrected to בֵּלְאשַׁצַּר; of foreign origin (compare H1095 (בֵּלְטְשַׁאצַּר));","strongs_def":"Belshatstsar, a Babylonian king","kjv_def":"Belshazzar."},
"H1113":{"lemma":"בֵּלְשַׁאצַּר","xlit":"Bêlshaʼtstsar","pron":"bale-shats-tsar'","derivation":"(Aramaic) corresponding to H1112 (בֵּלְשַׁאצַּר)","strongs_def":"{Belshatstsar, a Babylonian king}","kjv_def":"Belshazzar."},
"H1114":{"lemma":"בִּלְשָׁן","xlit":"Bilshân","pron":"bil-shawn'","derivation":"of uncertain derivation;","strongs_def":"Bilshan, an Israelite","kjv_def":"Bilshan."},
"H1115":{"lemma":"בִּלְתִּי","xlit":"biltîy","pron":"bil-tee'","derivation":"constructive feminine of H1086 (בָּלָה) (equivalent to H1097 (בְּלִי));","strongs_def":"properly, a failure of, i.e. (used only as a negative particle, usually with a prepositional prefix) not, except, without, unless, besides, because not, until, etc.","kjv_def":"because un(satiable), beside, but, [phrase] continual, except, from, lest, neither, no more, none, not, nothing, save, that no, without."},
"H1116":{"lemma":"בָּמָה","xlit":"bâmâh","pron":"bam-maw'","derivation":"from an unused root (meaning to be high);","strongs_def":"an elevation","kjv_def":"height, high place, wave."},
"H1117":{"lemma":"בָּמָה","xlit":"Bâmâh","pron":"baw-maw'","derivation":"the same as H1116 (בָּמָה);","strongs_def":"Bamah, a place in Palestine","kjv_def":"Bamah. See also H1120 (בָּמוֹת)."},
"H1118":{"lemma":"בִּמְהָל","xlit":"Bimhâl","pron":"bim-hawl'","derivation":"probably from H4107 (מָהַל) with prepositional prefix; with pruning;","strongs_def":"Bimhal, an Israelite","kjv_def":"Bimhal."},
"H1119":{"lemma":"בְּמוֹ","xlit":"bᵉmôw","pron":"bem-o'","derivation":"prolongation for prepositional prefix;","strongs_def":"in, with, by, etc.","kjv_def":"for, in into, through."},
"H1120":{"lemma":"בָּמוֹת","xlit":"Bâmôwth","pron":"baw-moth'","derivation":"plural of H1116 (בָּמָה); heights; or (fully) בָּמוֹת בַּעַל; from the same and H1168 (בַּעַל); heights of Baal;","strongs_def":"Bamoth or Bamoth-Baal, a place East of the Jordan","kjv_def":"Bamoth, Bamoth-baal."},
"H1121":{"lemma":"בֵּן","xlit":"bên","pron":"bane","derivation":"from H1129 (בָּנָה);","strongs_def":"a son (as a builder of the family name), in the widest sense (of literal and figurative relationship, including grandson, subject, nation, quality or condition, etc., (like father or brother), etc.)","kjv_def":"[phrase] afflicted, age, (Ahoh-) (Ammon-) (Hachmon-) (Lev-) ite, (anoint-) ed one, appointed to, ([phrase]) arrow, (Assyr-) (Babylon-) (Egypt-) (Grec-) ian, one born, bough, branch, breed, [phrase] (young) bullock, [phrase] (young) calf, [idiom] came up in, child, colt, [idiom] common, [idiom] corn, daughter, [idiom] of first, [phrase] firstborn, foal, [phrase] very fruitful, [phrase] postage, [idiom] in, [phrase] kid, [phrase] lamb, ([phrase]) man, meet, [phrase] mighty, [phrase] nephew, old, ([phrase]) people, [phrase] rebel, [phrase] robber, [idiom] servant born, [idiom] soldier, son, [phrase] spark, [phrase] steward, [phrase] stranger, [idiom] surely, them of, [phrase] tumultuous one, [phrase] valiant(-est), whelp, worthy, young (one), youth."},
"H1122":{"lemma":"בֵּן","xlit":"Bên","pron":"bane","derivation":"the same as H1121 (בֵּן);","strongs_def":"Ben, an Israelite","kjv_def":"Ben."},
"H1123":{"lemma":"בֵּן","xlit":"bên","pron":"bane","derivation":"(Aramaic) corresponding to H1121 (בֵּן)","strongs_def":"{a son (as a builder of the family name), in the widest sense}","kjv_def":"child, son, young."},
"H1124":{"lemma":"בְּנָא","xlit":"bᵉnâʼ","pron":"ben-aw'","derivation":"(Aramaic) or בְּנָה; (Aramaic), corresponding to H1129 (בָּנָה);","strongs_def":"to build","kjv_def":"build, make."},
"H1125":{"lemma":"בֶּן־אֲבִינָדָב","xlit":"Ben-ʼĂbîynâdâb","pron":"ben-ab-ee''-naw-dawb'","derivation":"from H1121 (בֵּן) and H40 (אֲבִימֶלֶךְ); (the) son of Abinadab;","strongs_def":"Ben-Abinadab, an Israelite","kjv_def":"the son of Abinadab."},
"H1126":{"lemma":"בֶּן־אוֹנִי","xlit":"Ben-ʼÔwnîy","pron":"ben-o-nee'","derivation":"from H1121 (בֵּן) and H205 (אָוֶן); son of my sorrow;","strongs_def":"Ben-Oni, the original name of Benjamin","kjv_def":"Ben-oni."},
"H1127":{"lemma":"בֶּן־גֶּבֶר","xlit":"Ben-Geber","pron":"ben-gheh'-ber","derivation":"from H1121 (בֵּן) and H1397 (גֶּבֶר); son of (the) hero;","strongs_def":"Ben-Geber, an Israelite","kjv_def":"the son of Geber."},
"H1128":{"lemma":"בֶּן־דֶּקֶר","xlit":"Ben-Deqer","pron":"ben-deh'-ker","derivation":"from H1121 (בֵּן) and a derivative of H1856 (דָּקַר); son of piercing (or of a lance);","strongs_def":"Ben-Deker, an Israelite","kjv_def":"the son of Dekar."},
"H1129":{"lemma":"בָּנָה","xlit":"bânâh","pron":"baw-naw'","derivation":"a primitive root;","strongs_def":"to build (literally and figuratively)","kjv_def":"(begin to) build(-er), obtain children, make, repair, set (up), [idiom] surely."},
"H1130":{"lemma":"בֶּן־הֲדַד","xlit":"Ben-Hădad","pron":"ben-had-ad'","derivation":"from H1121 (בֵּן) and H1908 (הֲדַד); son of Hadad;","strongs_def":"Ben-Hadad, the name of several Syrian kings","kjv_def":"Benhadad."},
"H1131":{"lemma":"בִּנּוּי","xlit":"Binnûwy","pron":"bin-noo'-ee","derivation":"from H1129 (בָּנָה); built up;","strongs_def":"Binnui, an Israelite","kjv_def":"Binnui."},
"H1132":{"lemma":"בֶּן־זוֹחֵת","xlit":"Ben-Zôwchêth","pron":"ben-zo-khayth'","derivation":"from H1121 (בֵּן) and H2105 (זוֹחֵת); son of Zocheth;","strongs_def":"Ben-Zocheth, an Israelite","kjv_def":"Ben-zoketh."},
"H1133":{"lemma":"בֶּן־חוּר","xlit":"Ben-Chûwr","pron":"ben-khoor'","derivation":"from H1121 (בֵּן) and H2354 (חוּר); son of Chur;","strongs_def":"Ben-Chur, an Israelite","kjv_def":"the son of Hur."},
"H1134":{"lemma":"בֶּן־חַיִל","xlit":"Ben-Chayil","pron":"ben-khah'-yil","derivation":"from H1121 (בֵּן) and H2428 (חַיִל); son of might;","strongs_def":"Ben-Chail, an Israelite","kjv_def":"Ben-hail."},
"H1135":{"lemma":"בֶּן־חָנָן","xlit":"Ben-Chânân","pron":"ben-khaw-nawn'","derivation":"from H1121 (בֵּן) and H2605 (חָנָן); son of Chanan;","strongs_def":"Ben-Chanan, an Israelite","kjv_def":"Ben-hanan."},
"H1136":{"lemma":"בֶּן־חֶסֶד","xlit":"Ben-Cheçed","pron":"ben-kheh'-sed","derivation":"from H1121 (בֵּן) and H2617 (חֵסֵד); son of kindness;","strongs_def":"Ben-Chesed, an Israelite","kjv_def":"the son of Hesed."},
"H1137":{"lemma":"בָּנִי","xlit":"Bânîy","pron":"baw-nee'","derivation":"from H1129 (בָּנָה); built;","strongs_def":"Bani, the name of five Israelites","kjv_def":"Bani."},
"H1138":{"lemma":"בֻּנִּי","xlit":"Bunnîy","pron":"boon-nee'","derivation":"or (fuller) בּוּנִי; from H1129 (בָּנָה); built;","strongs_def":"Bunni or Buni, an Israelite","kjv_def":"Bunni."},
"H1139":{"lemma":"בְּנֵי־בְּרַק","xlit":"Bᵉnêy-Bᵉraq","pron":"ben-ay'-ber-ak'","derivation":"from the plural construction of H1121 (בֵּן) and H1300 (בָּרָק);","strongs_def":"sons of lightning, Bene-berak, a place in Palestine","kjv_def":"Bene-barak."},
"H1140":{"lemma":"בִּנְיָה","xlit":"binyâh","pron":"bin-yaw'","derivation":"feminine from H1129 (בָּנָה);","strongs_def":"a structure","kjv_def":"building."},
"H1141":{"lemma":"בְּנָיָה","xlit":"Bᵉnâyâh","pron":"ben-aw-yaw'","derivation":"or בּנָיָהוּ; (prolonged) from H1129 (בָּנָה) and H3050 (יָהּ); Jah has built;","strongs_def":"Benajah, the name of twelve Israelites","kjv_def":"Benaiah."},
"H1142":{"lemma":"בְּנֵי יַעֲקָן","xlit":"Bᵉnêy Yaʻăqân","pron":"ben-ay' yah-ak-awn'","derivation":"from the plural of H1121 (בֵּן) and H3292 (יַעֲקָן); sons of Yaakan;","strongs_def":"Bene-Jaakan, a place in the Desert","kjv_def":"Bene-jaakan."},
"H1143":{"lemma":"בֵּנַיִם","xlit":"bênayim","pron":"bay-nah'-yim","derivation":"dual of H996 (בֵּין);","strongs_def":"a double interval, i.e. the space between two armies","kjv_def":"[phrase] champion."},
"H1144":{"lemma":"בִּנְיָמִין","xlit":"Binyâmîyn","pron":"bin-yaw-mene'","derivation":"from H1121 (בֵּן) and H3225 (יָמִין); son of (the) right hand;","strongs_def":"Binjamin, youngest son of Jacob; also the tribe descended from him, and its territory","kjv_def":"Benjamin."},
"H1145":{"lemma":"בֶּן־יְמִינִי","xlit":"Ben-yᵉmîynîy","pron":"ben-yem-ee-nee'","derivation":"sometimes (with the article inserted) בֵּן־הַיּמִינִי; with H376 (אִישׁ) inserted (1 Samuel 9:1) בֶּן־אִישׁ ימִינִי; son of a man of Jemini; or shortened אִישׁ יְמִינִי; (1 Samuel 9:4; Esther 2:5) a man of Jemini, or (1 Samuel 20:1) simply יְמִינִיxlit Yᵉmînîy corrected to Yᵉmîynîy; a Jeminite; (plural) בְּנֵי יְמִינִיxlit Bᵉnîy corrected to Bᵉnêy; patron from H1144 (בִּנְיָמִין);","strongs_def":"a Benjaminite, or descendent of Benjamin","kjv_def":"Benjamite, of Benjamin."},
"H1146":{"lemma":"בִּנְיָן","xlit":"binyân","pron":"bin-yawn'","derivation":"from H1129 (בָּנָה);","strongs_def":"an edifice","kjv_def":"building."},
"H1147":{"lemma":"בִּנְיָן","xlit":"binyân","pron":"bin-yawn'","derivation":"(Aramaic) corresponding to H1146 (בִּנְיָן)","strongs_def":"{an edifice}","kjv_def":"building."},
"H1148":{"lemma":"בְּנִינוּ","xlit":"Bᵉnîynûw","pron":"ben-ee-noo'","derivation":"probably from H1121 (בֵּן) with pron. suff.; our son;","strongs_def":"Beninu, an Israelite","kjv_def":"Beninu."},
"H1149":{"lemma":"בְּנַס","xlit":"bᵉnaç","pron":"ben-as'","derivation":"(Aramaic) of uncertain affinity;","strongs_def":"to be enraged","kjv_def":"be angry."},
"H1150":{"lemma":"בִּנְעָא","xlit":"Binʻâʼ","pron":"bin-aw'","derivation":"or בִּנְעָה; of uncertain derivation;","strongs_def":"Bina or Binah, an Israelite","kjv_def":"Binea, Bineah."},
"H1151":{"lemma":"בֶּן־עַמִּי","xlit":"Ben-ʻAmmîy","pron":"ben-am-mee'","derivation":"from H1121 (בֵּן) and H5971 (עַם) with pronomial suffix; son of my people;","strongs_def":"Ben-Ammi, a son of Lot","kjv_def":"Ben-ammi."},
"H1152":{"lemma":"בְּסוֹדְיָה","xlit":"Bᵉçôwdᵉyâh","pron":"bes-o-deh-yaw'","derivation":"from H5475 (סוֹד) and H3050 (יָהּ) with prepositional prefix; in (the) counsel of Jehovah;","strongs_def":"Besodejah, an Israelite","kjv_def":"Besodeiah."},
"H1153":{"lemma":"בְּסַי","xlit":"Bᵉçay","pron":"bes-ah'-ee","derivation":"from H947 (בּוּס); domineering;","strongs_def":"Besai, one of the Nethinim","kjv_def":"Besai."},
"H1154":{"lemma":"בֶּסֶר","xlit":"beçer","pron":"beh'-ser","derivation":"from an unused root meaning to be sour;","strongs_def":"an immature grape","kjv_def":"unripe grape."},
"H1155":{"lemma":"בֹּסֶר","xlit":"bôçer","pron":"bo'ser","derivation":"from the same as H1154 (בֶּסֶר)","strongs_def":"{an immature grape}","kjv_def":"sour grape."},
"H1156":{"lemma":"בְּעָא","xlit":"bᵉʻâʼ","pron":"beh-aw'","derivation":"(Aramaic) or בְּעָה; (Aramaic), corresponding to H1158 (בָּעָה);","strongs_def":"to seek or ask","kjv_def":"ask, desire, make (petition), pray, request, seek."},
"H1157":{"lemma":"בְּעַד","xlit":"bᵉʻad","pron":"beh-ad'","derivation":"from H5704 (עַד) with prepositional prefix;","strongs_def":"in up to or over against; generally at, beside, among, behind, for, etc.","kjv_def":"about, at by (means of), for, over, through, up (-on), within."},
"H1158":{"lemma":"בָּעָה","xlit":"bâʻâh","pron":"baw-aw'","derivation":"a primitive root;","strongs_def":"to gush over, i.e. to swell; (figuratively) to desire earnestly; by implication to ask","kjv_def":"cause, inquire, seek up, swell out."},
"H1159":{"lemma":"בָּעוּ","xlit":"bâʻûw","pron":"baw-oo'","derivation":"(Aramaic) from H1156 (בְּעָא);","strongs_def":"a request","kjv_def":"petition."},
"H1160":{"lemma":"בְּעוֹר","xlit":"Bᵉʻôwr","pron":"beh-ore'","derivation":"from H1197 (בָּעַר) (in the sense of burning); a lamp;","strongs_def":"Beor, the name of the father of an Edomitish king; also of that of Balaam","kjv_def":"Beor."},
"H1161":{"lemma":"בִּעוּתִים","xlit":"biʻûwthîym","pron":"be-oo-theme'","derivation":"masculine plural from H1204 (בָּעַת);","strongs_def":"alarms","kjv_def":"terrors."},
"H1162":{"lemma":"בֹּעַז","xlit":"Bôʻaz","pron":"bo'-az","derivation":"from an unused root of uncertain meaning;","strongs_def":"Boaz, the ancestor of David; also the name of a pillar in front of the temple","kjv_def":"Boaz."},
"H1163":{"lemma":"בָּעַט","xlit":"bâʻaṭ","pron":"baw-at'","derivation":"a primitive root;","strongs_def":"to trample down, i.e. (figuratively) despise","kjv_def":"kick."},
"H1164":{"lemma":"בְּעִי","xlit":"bᵉʻîy","pron":"beh-ee'","derivation":"from H1158 (בָּעָה);","strongs_def":"a prayer","kjv_def":"grave."},
"H1165":{"lemma":"בְּעִיר","xlit":"bᵉʻîyr","pron":"beh-ere'","derivation":"from H1197 (בָּעַר) (in the sense of eating);","strongs_def":"cattle","kjv_def":"beast, cattle."},
"H1166":{"lemma":"בָּעַל","xlit":"bâʻal","pron":"baw-al'","derivation":"a primitive root; also as denominative from H1167 (בַּעַל)","strongs_def":"to be master; hence, to marry","kjv_def":"have dominion (over), be husband, marry(-ried, [idiom] wife)."},
"H1167":{"lemma":"בַּעַל","xlit":"baʻal","pron":"bah'-al","derivation":"from H1166 (בָּעַל);","strongs_def":"a master; hence, a husband, or (figuratively) owner (often used with another noun in modifications of this latter sense)","kjv_def":"[phrase] archer, [phrase] babbler, [phrase] bird, captain, chief man, [phrase] confederate, [phrase] have to do, [phrase] dreamer, those to whom it is due, [phrase] furious, those that are given to it, great, [phrase] hairy, he that hath it, have, [phrase] horseman, husband, lord, man, [phrase] married, master, person, [phrase] sworn, they of."},
"H1168":{"lemma":"בַּעַל","xlit":"Baʻal","pron":"bah'-al","derivation":"the same as H1167 (בַּעַל);","strongs_def":"Baal, a Phoenician deity","kjv_def":"Baal, (plural) Baalim."},
"H1169":{"lemma":"בְּעֵל","xlit":"bᵉʻêl","pron":"beh-ale'","derivation":"(Aramaic) corresponding to H1167 (בַּעַל)","strongs_def":"{a master; hence, a husband,}","kjv_def":"[phrase] chancellor."},
"H1170":{"lemma":"בַּעַל בְּרִית","xlit":"Baʻal Bᵉrîyth","pron":"bah'-al ber-eeth'","derivation":"from H1168 (בַּעַל) and H1285 (בְּרִית); Baal of (the) covenant;","strongs_def":"Baal-Berith, a special deity of the Shechemites","kjv_def":"Baal-berith."},
"H1171":{"lemma":"בַּעַל גָּד","xlit":"Baʻal Gâd","pron":"bah'-al gawd","derivation":"from H1168 (בַּעַל) and H1409 (גָּד); Baal of Fortune;","strongs_def":"Baal-Gad, a place in Syria","kjv_def":"Baal-gad."},
"H1172":{"lemma":"בַּעֲלָה","xlit":"baʻălâh","pron":"bah-al-aw'","derivation":"feminine of H1167 (בַּעַל);","strongs_def":"a mistress","kjv_def":"that hath, mistress."},
"H1173":{"lemma":"בַּעֲלָה","xlit":"Baʻălâh","pron":"bah-al-aw'","derivation":"the same as H1172 (בַּעֲלָה);","strongs_def":"Baalah, the name of three places in Palestine","kjv_def":"Baalah."},
"H1174":{"lemma":"בַּעַל הָמוֹן","xlit":"Baʻal Hâmôwn","pron":"bah'-al haw-mone'","derivation":"from H1167 (בַּעַל) and H1995 (הָמוֹן); possessor of a multitude;","strongs_def":"Baal-Hamon, a place in Palestine","kjv_def":"Baal-hamon."},
"H1175":{"lemma":"בְּעָלוֹת","xlit":"Bᵉʻâlôwth","pron":"beh-aw-loth'","derivation":"plural of H1172 (בַּעֲלָה); mistresses;","strongs_def":"Bealoth, a place in Palestine","kjv_def":"Bealoth, in Aloth (by mistake for a plural from H5927 (עָלָה) with prepositional prefix)."},
"H1176":{"lemma":"בַּעַל זְבוּב","xlit":"Baʻal Zᵉbûwb","pron":"bah'-al zeb-oob'","derivation":"from H1168 (בַּעַל) and H2070 (זְבוּב); Baal of (the) Fly;","strongs_def":"Baal-Zebub, a special deity of the Ekronites","kjv_def":"Baal-zebub."},
"H1177":{"lemma":"בַּעַל חָנָן","xlit":"Baʻal Chânân","pron":"bah'-al khaw-nawn'","derivation":"from H1167 (בַּעַל) and H2603 (חָנַן); possessor of grace;","strongs_def":"Baal-Chanan, the name of an Edomite, also of an Israelite","kjv_def":"Baal-hanan."},
"H1178":{"lemma":"בַּעַל חָצוֹר","xlit":"Baʻal Châtsôwr","pron":"bah'-al khaw-tsore'","derivation":"from H1167 (בַּעַל) and a modification of H2691 (חָצֵר); possessor of a village;","strongs_def":"Baal-Chatsor, a place in Palestine","kjv_def":"Baal-hazor."},
"H1179":{"lemma":"בַּעַל חֶרְמוֹן","xlit":"Baʻal Chermôwn","pron":"bah'-al kher-mone'","derivation":"from H1167 (בַּעַל) and H2768 (חֶרְמוֹן); possessor of Hermon;","strongs_def":"Baal-Chermon, a place in Palestine","kjv_def":"Baal-hermon."},
"H1180":{"lemma":"בַּעֲלִי","xlit":"Baʻălîy","pron":"bah-al-ee'","derivation":"from H1167 (בַּעַל) with pron. suff.; my master;","strongs_def":"Baali, a symbolical name for Jehovah","kjv_def":"Baali."},
"H1181":{"lemma":"בַּעֲלֵי בָּמוֹת","xlit":"Baʻălêy Bâmôwth","pron":"bah-al-ay' baw-moth'","derivation":"from the plural of H1168 (בַּעַל) and the plural of H1116 (בָּמָה); Baals of (the) heights;","strongs_def":"Baale-Bamoth, a place East of the Jordan","kjv_def":"lords of the high places."},
"H1182":{"lemma":"בְּעֶלְיָדָע","xlit":"Bᵉʻelyâdâʻ","pron":"beh-el-yaw-daw'","derivation":"from H1168 (בַּעַל) and H3045 (יָדַע); Baal has known;","strongs_def":"Beeljada, an Israelite","kjv_def":"Beeliada."},
"H1183":{"lemma":"בְּעַלְיָה","xlit":"Bᵉʻalyâh","pron":"beh-al-yaw'","derivation":"from H1167 (בַּעַל) and H3050 (יָהּ); Jah (is) master;","strongs_def":"Bealjah, an Israelite","kjv_def":"Bealiah."},
"H1184":{"lemma":"בַּעֲלֵי יְהוּדָה","xlit":"Baʻălêy Yᵉhûwdâh","pron":"bah-al-ay' yeh-hoo-daw'","derivation":"from the plural of H1167 (בַּעַל) and H3063 (יְהוּדָה); masters of Judah;","strongs_def":"Baale-Jehudah, a place in Palestine","kjv_def":"Baale of Judah."},
"H1185":{"lemma":"בַּעֲלִיס","xlit":"Baʻălîyç","pron":"bah-al-ece'","derivation":"probably from a derivative of H5965 (עָלַס) with prepositional prefix; in exultation;","strongs_def":"Baalis, an Ammonitish king","kjv_def":"Baalis."},
"H1186":{"lemma":"בַּעַל מְעוֹן","xlit":"Baʻal Mᵉʻôwn","pron":"bah-al meh-one'","derivation":"from H1168 (בַּעַל) and H4583 (מָעוֹן); Baal of (the) habitation (of) (compare H1010 (בֵּית בַּעַל מְעוֹן));","strongs_def":"Baal-Meon, a place East of the Jordan","kjv_def":"Baal-meon."},
"H1187":{"lemma":"בַּעַל פְּעוֹר","xlit":"Baʻal Pᵉʻôwr","pron":"bah'-al peh-ore'","derivation":"from H1168 (בַּעַל) and H6465 (פְּעוֹר); Baal of Peor;","strongs_def":"Baal-Peor, a Moabitish deity","kjv_def":"Baal-peor."},
"H1188":{"lemma":"בַּעַל פְּרָצִים","xlit":"Baʻal Pᵉrâtsîym","pron":"bah'-al per-aw-tseem'","derivation":"from H1167 (בַּעַל) and the plural of H6556 (פֶּרֶץ); possessor of breaches;","strongs_def":"Baal-Peratsim, a place in Palestine","kjv_def":"Baal-perazim."},
"H1189":{"lemma":"בַּעַל צְפוֹן","xlit":"Baʻal Tsᵉphôwn","pron":"bah'-al tsef-one'","derivation":"from H1168 (בַּעַל) and H6828 (צָפוֹן) (in the sense of cold) (according to others an Egyptian form of Typhon, the destroyer); Baal of winter;","strongs_def":"Baal-Tsephon, a place in Eqypt","kjv_def":"Baal-zephon."},
"H1190":{"lemma":"בַּעַל שָׁלִשָׁה","xlit":"Baʻal Shâlishâh","pron":"bah'-al shaw-lee-shaw'","derivation":"from H1168 (בַּעַל) and H8031 (שָׁלִשָׁה);","strongs_def":"Baal of Shalishah, Baal-Shalishah, a place in Palestine","kjv_def":"Baal-shalisha."},
"H1191":{"lemma":"בַּעֲלָת","xlit":"Baʻălâth","pron":"bah-al-awth'","derivation":"a modification of H1172 (בַּעֲלָה); mistressship;","strongs_def":"Baalath, a place in Palestine","kjv_def":"Baalath."},
"H1192":{"lemma":"בַּעֲלַת בְּאֵר","xlit":"Baʻălath Bᵉʼêr","pron":"bah-al-ath' beh-ayr'","derivation":"from H1172 (בַּעֲלָה) and H875 (בְּאֵר); mistress of a well;","strongs_def":"Baalath-Beer, a place in Palestine","kjv_def":"Baalath-beer."},
"H1193":{"lemma":"בַּעַל תָּמָר","xlit":"Baʻal Tâmâr","pron":"bah'-al taw-mawr'","derivation":"from H1167 (בַּעַל) and H8558 (תָּמָר); possessor of (the) palm-tree;","strongs_def":"Baal-Tamar, a place in Palestine","kjv_def":"Baal-tamar."},
"H1194":{"lemma":"בְּעֹן","xlit":"Bᵉʻôn","pron":"beh-ohn'","derivation":"probably a contraction of H1010 (בֵּית בַּעַל מְעוֹן);","strongs_def":"Beon, a place East of the Jordan","kjv_def":"Beon."},
"H1195":{"lemma":"בַּעֲנָא","xlit":"Baʻănâʼ","pron":"bah-an-aw'","derivation":"the same as H1196 (בַּעֲנָה);","strongs_def":"Banana, the name of four Israelite","kjv_def":"Baana, Baanah."},
"H1196":{"lemma":"בַּעֲנָה","xlit":"Baʻănâh","pron":"bah-an-aw'","derivation":"from a derivative of H6031 (עָנָה) with prepositional prefix; in affliction","strongs_def":"Baanah, the name of four Israelites","kjv_def":"Baanah."},
"H1197":{"lemma":"בָּעַר","xlit":"bâʻar","pron":"baw-ar'","derivation":"a primitive root; also as denominative from H1198 (בַּעַר)","strongs_def":"to kindle, i.e. consume (by fire or by eating);  to be(-come) brutish","kjv_def":"be brutish, bring (put, take) away, burn, (cause to) eat (up), feed, heat, kindle, set (on fire), waste."},
"H1198":{"lemma":"בַּעַר","xlit":"baʻar","pron":"bah'-ar","derivation":"from H1197 (בָּעַר);","strongs_def":"properly, foot (as consumed); i.e. (by exten.) of cattle brutishness; (concretely) stupid","kjv_def":"brutish (person), foolish."},
"H1199":{"lemma":"בָּעֲרָא","xlit":"Bâʻărâʼ","pron":"bah-ar-aw'","derivation":"from H1198 (בַּעַר); brutish;","strongs_def":"Baara, an Israelitish woman","kjv_def":"Baara."},
"H1200":{"lemma":"בְּעֵרָה","xlit":"bᵉʻêrâh","pron":"be-ay-raw'","derivation":"from H1197 (בָּעַר);","strongs_def":"a burning","kjv_def":"fire."},
"H1201":{"lemma":"בַּעְשָׁא","xlit":"Baʻshâʼ","pron":"bah-shaw'","derivation":"from an unused root meaning to stink; offensiveness;","strongs_def":"Basha, a king of Israel","kjv_def":"Baasha."},
"H1202":{"lemma":"בַּעֲשֵׂיָה","xlit":"Baʻăsêyâh","pron":"bah-as-ay-yaw'","derivation":"from H6213 (עָשָׂה) and H3050 (יָהּ) with a prepositional prefix; in (the) work of Jah;","strongs_def":"Baasejah, an Israelite","kjv_def":"Baaseiah."},
"H1203":{"lemma":"בְּעֶשְׁתְּרָה","xlit":"Bᵉʻeshtᵉrâh","pron":"beh-esh-ter-aw'","derivation":"from H6251 (עַשְׁתְּרָה) (as singular of H6252 (עַשְׁתָּרוֹת)) with a prepositional prefix; with Ashtoreth;","strongs_def":"Beeshterah, a place East of the Jordan","kjv_def":"Beeshterah."},
"H1204":{"lemma":"בָּעַת","xlit":"bâʻath","pron":"baw-ath'","derivation":"a primitive root;","strongs_def":"to fear","kjv_def":"affright, be (make) afraid, terrify, trouble."},
"H1205":{"lemma":"בְּעָתָה","xlit":"bᵉʻâthâh","pron":"beh-aw-thaw'","derivation":"from H1204 (בָּעַת);","strongs_def":"fear","kjv_def":"trouble."},
"H1206":{"lemma":"בֹץ","xlit":"bôts","pron":"botse","derivation":"probably the same as H948 (בּוּץ);","strongs_def":"mud (as whitish clay)","kjv_def":"mire."},
"H1207":{"lemma":"בִּצָּה","xlit":"bitstsâh","pron":"bits-tsaw'","derivation":"intensive from H1206 (בֹץ);","strongs_def":"a swamp","kjv_def":"fen, mire(-ry place)."},
"H1208":{"lemma":"בָּצוֹר","xlit":"bâtsôwr","pron":"baw-tsore'","derivation":"from H1219 (בָּצַר);","strongs_def":"inaccessible, i.e. lofty","kjv_def":"vintage (by confusion with H1210 (בָּצִיר))."},
"H1209":{"lemma":"בֵּצַי","xlit":"Bêtsay","pron":"bay-tsah'-ee","derivation":"perhaps the same as H1153 (בְּסַי);","strongs_def":"Betsai, the name of two Israelites","kjv_def":"Bezai."},
"H1210":{"lemma":"בָּצִיר","xlit":"bâtsîyr","pron":"baw-tseer'","derivation":"from H1219 (בָּצַר);","strongs_def":"clipped, i.e. the grape crop","kjv_def":"vintage."},
"H1211":{"lemma":"בֶּצֶל","xlit":"betsel","pron":"beh'-tsel","derivation":"from an unused root apparently meaning to peel;","strongs_def":"an onion","kjv_def":"onion."},
"H1212":{"lemma":"בְּצַלְאֵל","xlit":"Bᵉtsalʼêl","pron":"bets-al-ale'","derivation":"probably from H6738 (צֵל) and H410 (אֵל) with a prepositional prefix; in (the) shadow (i.e. protection) of God;","strongs_def":"Betsalel, the name of two Israelites","kjv_def":"Bezaleel."},
"H1213":{"lemma":"בַּצְלוּת","xlit":"Batslûwth","pron":"bats-looth'","derivation":"or בַּצְלִית; from the same as H1211 (בֶּצֶל)xlit bᵉtsel corrected to betsel; a peeling;","strongs_def":"Batsluth or Batslith, an Israelite","kjv_def":"Bazlith, Bazluth."},
"H1214":{"lemma":"בָּצַע","xlit":"bâtsaʻ","pron":"baw-tsah'","derivation":"a primitive root to break off, i.e. (usually) plunder;","strongs_def":"figuratively, to finish, or (intransitively) stop","kjv_def":"(be) covet(-ous), cut (off), finish, fulfill, gain (greedily), get, be given to (covetousness), greedy, perform, be wounded."},
"H1215":{"lemma":"בֶּצַע","xlit":"betsaʻ","pron":"beh'-tsah","derivation":"from H1214 (בָּצַע);","strongs_def":"plunder; by extension, gain (usually unjust)","kjv_def":"covetousness, (dishonest) gain, lucre, profit."},
"H1216":{"lemma":"בָּצֵק","xlit":"bâtsêq","pron":"baw-tsake'","derivation":"a primitive root;","strongs_def":"perhaps to swell up, i.e. blister","kjv_def":"swell."},
"H1217":{"lemma":"בָּצֵק","xlit":"bâtsêq","pron":"baw-tsake'","derivation":"from H1216 (בָּצֵק);","strongs_def":"dough (as swelling by fermentation)","kjv_def":"dough, flour."},
"H1218":{"lemma":"בׇּצְקַת","xlit":"Botsqath","pron":"bots-cath'","derivation":"from H1216 (בָּצֵק); a swell of ground;","strongs_def":"Botscath, a place in Palestine","kjv_def":"Bozcath, Boskath."},
"H1219":{"lemma":"בָּצַר","xlit":"bâtsar","pron":"baw-tsar'","derivation":"a primitive root; to clip off; specifically (as denominative from H1210 (בָּצִיר))","strongs_def":"to gather grapes; also to be isolated (i.e. inaccessible by height or fortification)","kjv_def":"cut off, (de-) fenced, fortify, (grape) gather(-er), mighty things, restrain, strong, wall (up), withhold."},
"H1220":{"lemma":"בֶּצֶר","xlit":"betser","pron":"beh'-tser","derivation":"from H1219 (בָּצַר);","strongs_def":"strictly a clipping, i.e. gold (as dug out)","kjv_def":"gold defence."},
"H1221":{"lemma":"בֶּצֶר","xlit":"Betser","pron":"beh'-tser","derivation":"the same as H1220 (בֶּצֶר), an inaccessible spot;","strongs_def":"Betser, a place in Palestine; also an Israelite","kjv_def":"Bezer."},
"H1222":{"lemma":"בְּצַר","xlit":"bᵉtsar","pron":"bets-ar'","derivation":"another form for H1220 (בֶּצֶר);","strongs_def":"gold","kjv_def":"gold."},
"H1223":{"lemma":"בׇּצְרָה","xlit":"botsrâh","pron":"bots-raw'","derivation":"feminine from H1219 (בָּצַר);","strongs_def":"an enclosure, i.e. sheep fold","kjv_def":"Bozrah."},
"H1224":{"lemma":"בׇּצְרָה","xlit":"Botsrâh","pron":"bots-raw'","derivation":"the same as H1223 (בׇּצְרָה);","strongs_def":"Botsrah, a place in Edom","kjv_def":"Bozrah."},
"H1225":{"lemma":"בִּצָּרוֹן","xlit":"bitstsârôwn","pron":"bits-tsaw-rone'","derivation":"masculine intensive from H1219 (בָּצַר);","strongs_def":"a fortress","kjv_def":"stronghold."},
"H1226":{"lemma":"בַּצֹּרֶת","xlit":"batstsôreth","pron":"bats-tso'-reth","derivation":"feminine intensive from H1219 (בָּצַר);","strongs_def":"restraint (of rain), i.e. drought","kjv_def":"dearth, drought."},
"H1227":{"lemma":"בַּקְבּוּק","xlit":"Baqbûwq","pron":"bak-book'","derivation":"the same as H1228 (בַּקְבֻּק)xlit baqbûk corrected to baqbuq;","strongs_def":"Bakbuk, one of the Nethinim","kjv_def":"Bakbuk."},
"H1228":{"lemma":"בַּקְבֻּק","xlit":"baqbuq","pron":"bak-book'","derivation":"from H1238 (בָּקַק);","strongs_def":"a bottle (from the gurgling in emptying)","kjv_def":"bottle, cruse."},
"H1229":{"lemma":"בַּקְבֻּקְיָה","xlit":"Baqbuqyâh","pron":"bak-book-yaw'","derivation":"from H1228 (בַּקְבֻּק)xlit baqbûk corrected to baqbuq and H3050 (יָהּ); emptying (i.e. wasting) of Jah;","strongs_def":"Bakbukjah, an Israelite","kjv_def":"Bakbukiah."},
"H1230":{"lemma":"בַּקְבַּקַּר","xlit":"Baqbaqqar","pron":"bak-bak-kar'","derivation":"reduplicated from H1239 (בָּקַר); searcher;","strongs_def":"Bakbakkar, an Israelite","kjv_def":"Bakbakkar."},
"H1231":{"lemma":"בֻּקִּי","xlit":"Buqqîy","pron":"book-kee'","derivation":"from H1238 (בָּקַק); wasteful;","strongs_def":"Bukki, the name of two Israelites","kjv_def":"Bukki."},
"H1232":{"lemma":"בֻּקִּיָּה","xlit":"Buqqîyâh","pron":"book-kee-yaw'","derivation":"from H1238 (בָּקַק) and H3050 (יָהּ); wasting of Jah;","strongs_def":"Bukkijah, an Israelite","kjv_def":"Bukkiah."},
"H1233":{"lemma":"בְּקִיעַ","xlit":"bᵉqîyaʻ","pron":"bek-ee'-ah","derivation":"from H1234 (בָּקַע);","strongs_def":"a fissure","kjv_def":"breach, cleft."},
"H1234":{"lemma":"בָּקַע","xlit":"bâqaʻ","pron":"baw-kah'","derivation":"a primitive root;","strongs_def":"to cleave; generally, to rend, break, rip or open","kjv_def":"make a breach, break forth (into, out, in pieces, through, up), be ready to burst, cleave (asunder), cut out, divide, hatch, rend (asunder), rip up, tear, win."},
"H1235":{"lemma":"בֶּקַע","xlit":"beqaʻ","pron":"beh'-kah","derivation":"from H1234 (בָּקַע);","strongs_def":"a section (half) of ashekel, i.e. a beka (a weight and a coin)","kjv_def":"bekah, half a shekel."},
"H1236":{"lemma":"בִּקְעָא","xlit":"biqʻâʼ","pron":"bik-aw'","derivation":"(Aramaic) corresponding to H1237 (בִּקְעָה)","strongs_def":"{properly, a split, i.e. a wide level valley between mountains}","kjv_def":"plain."},
"H1237":{"lemma":"בִּקְעָה","xlit":"biqʻâh","pron":"bik-aw'","derivation":"from H1234 (בָּקַע);","strongs_def":"properly, a split, i.e. a wide level valley between mountains","kjv_def":"plain, valley."},
"H1238":{"lemma":"בָּקַק","xlit":"bâqaq","pron":"baw-kah'","derivation":"a primitive root;","strongs_def":"to pour out, i.e. to empty, figuratively, to depopulate; by analogy, to spread out (as a fruitful vine)","kjv_def":"(make) empty (out), fail, [idiom] utterly, make void."},
"H1239":{"lemma":"בָּקַר","xlit":"bâqar","pron":"baw-kar","derivation":"a primitive root;","strongs_def":"properly, to plough, or (generally) break forth, i.e. (figuratively) to inspect, admire, care for, consider","kjv_def":"(make) inquire (-ry), (make) search, seek out."},
"H1240":{"lemma":"בְּקַר","xlit":"bᵉqar","pron":"bek-ar'","derivation":"(Aramaic) corresponding to H1239 (בָּקַר)","strongs_def":"{properly, to plough, or (generally) break forth, i.e. (figuratively) to inspect, admire, care for, consider}","kjv_def":"inquire, make search."},
"H1241":{"lemma":"בָּקָר","xlit":"bâqâr","pron":"baw-kawr'","derivation":"from H1239 (בָּקַר);","strongs_def":"beef cattle or an animal of the ox family of either gender (as used for plowing); collectively, a herd","kjv_def":"beeve, bull ([phrase] -ock), [phrase] calf, [phrase] cow, great (cattle), [phrase] heifer, herd, kine, ox."},
"H1242":{"lemma":"בֹּקֶר","xlit":"bôqer","pron":"bo'-ker","derivation":"from H1239 (בָּקַר);","strongs_def":"properly, dawn (as the break of day); generally, morning","kjv_def":"([phrase]) day, early, morning, morrow."},
"H1243":{"lemma":"בַּקָּרָה","xlit":"baqqârâh","pron":"bak-kaw-raw'","derivation":"intensive from H1239 (בָּקַר);","strongs_def":"a looking after","kjv_def":"seek out."},
"H1244":{"lemma":"בִּקֹּרֶת","xlit":"biqqôreth","pron":"bik-ko-reth","derivation":"from H1239 (בָּקַר);","strongs_def":"properly, examination, i.e. (by implication) punishment","kjv_def":"scourged."},
"H1245":{"lemma":"בָּקַשׁ","xlit":"bâqash","pron":"baw-kash'","derivation":"a primitive root;","strongs_def":"to search out (by any method, specifically in worship or prayer); by implication, to strive after","kjv_def":"ask, beg, beseech, desire, enquire, get, make inquisition, procure, (make) request, require, seek (for)."},
"H1246":{"lemma":"בַּקָּשָׁה","xlit":"baqqâshâh","pron":"bak-kaw-shaw'","derivation":"from H1245 (בָּקַשׁ);","strongs_def":"a petition","kjv_def":"request."},
"H1247":{"lemma":"בַּר","xlit":"bar","pron":"bar","derivation":"(Aramaic) corresponding to H1121 (בֵּן);","strongs_def":"a son, grandson, etc.","kjv_def":"[idiom] old, son."},
"H1248":{"lemma":"בַּר","xlit":"bar","pron":"bar","derivation":"borrowed (as a title) from H1247 (בַּר);","strongs_def":"the heir (apparent to the throne)","kjv_def":"son."},
"H1249":{"lemma":"בַּר","xlit":"bar","pron":"bar","derivation":"from H1305 (בָּרַר) (in its various senses);","strongs_def":"beloved; also pure, empty","kjv_def":"choice, clean, clear, pure."},
"H1250":{"lemma":"בָּר","xlit":"bâr","pron":"bawr","derivation":"or בַּר; from H1305 (בָּרַר) (in the sense of winnowing);","strongs_def":"grain of any kind (even while standing in the field); by extension the open country","kjv_def":"corn, wheat."},
"H1251":{"lemma":"בַּר","xlit":"bar","pron":"bar","derivation":"(Aramaic) corresponding to H1250 (בָּר);","strongs_def":"a field","kjv_def":"field."},
"H1252":{"lemma":"בֹּר","xlit":"bôr","pron":"bore","derivation":"from H1305 (בָּרַר);","strongs_def":"purify","kjv_def":"cleanness, pureness."},
"H1253":{"lemma":"בֹּר","xlit":"bôr","pron":"bore","derivation":"the same as H1252 (בֹּר);","strongs_def":"vegetable lye (from its cleansing); used as a soap forwashing, or a flux formetals","kjv_def":"[idiom] never so, purely."},
"H1254":{"lemma":"בָּרָא","xlit":"bârâʼ","pron":"baw-raw'","derivation":"a primitive root;","strongs_def":"(absolutely) to create; (qualified) to cut down (a wood), select, feed (as formative processes)","kjv_def":"choose, create (creator), cut down, dispatch, do, make (fat)."},
"H1255":{"lemma":"בְּרֹאדַךְ בַּלְאֲדָן","xlit":"Bᵉrôʼdak Balʼădân","pron":"ber-o-dak' bal-ad-awn'","derivation":"a variation of H4757 (מְרֹאדַךְ בַּלְאָדָן);","strongs_def":"Berodak-Baladan, a Babylonian king","kjv_def":"Berodach-baladan.Birliy."},
"H1256":{"lemma":"בְּרָאיָה","xlit":"Bᵉrâʼyâh","pron":"ber-aw-yaw'","derivation":"from H1254 (בָּרָא) and H8050 (שְׁמוּאֵל); Jah has created;","strongs_def":"Berajah, an Israelite","kjv_def":"Beraiah."},
"H1257":{"lemma":"בַּרְבֻּר","xlit":"barbur","pron":"bar-boor'","derivation":"by reduplication from H1250 (בָּר);","strongs_def":"a fowl (as fattened on grain)","kjv_def":"fowl."},
"H1258":{"lemma":"בָּרַד","xlit":"bârad","pron":"baw-rad'","derivation":"a primitive root,","strongs_def":"to hail","kjv_def":"hail."},
"H1259":{"lemma":"בָּרָד","xlit":"bârâd","pron":"baw-rawd'","derivation":"from H1258 (בָּרַד);","strongs_def":"hail","kjv_def":"hail(stones)."},
"H1260":{"lemma":"בֶּרֶד","xlit":"Bered","pron":"beh'red","derivation":"from H1258 (בָּרַד); hail;","strongs_def":"Bered, the name of a place south of Palestine, also of an Israelite","kjv_def":"Bered."},
"H1261":{"lemma":"בָּרֹד","xlit":"bârôd","pron":"baw-rode'","derivation":"from H1258 (בָּרַד);","strongs_def":"spotted (as if with hail)","kjv_def":"grisled."},
"H1262":{"lemma":"בָּרָה","xlit":"bârâh","pron":"baw-raw'","derivation":"a primitive root; to select; also (as denominative from H1250 (בָּר)) to feed; also (as equivalent to H1305 (בָּרַר)) to render clear (Ecclesiastes 3:18)","strongs_def":"to select; to feed; to render clear","kjv_def":"choose, (cause to) eat, manifest, (give) meat."},
"H1263":{"lemma":"בָּרוּךְ","xlit":"Bârûwk","pron":"baw-rook'","derivation":"passive participle from H1288 (בָרַךְ); blessed;","strongs_def":"Baruk, the name of three Israelites","kjv_def":"Baruch."},
"H1264":{"lemma":"בְּרוֹם","xlit":"bᵉrôwm","pron":"ber-ome'","derivation":"probably of foreign origin;","strongs_def":"damask (stuff of variegated thread)","kjv_def":"rich apparel."},
"H1265":{"lemma":"בְּרוֹשׁ","xlit":"bᵉrôwsh","pron":"ber-osh'","derivation":"of uncertain derivation;","strongs_def":"a cypress (?) tree; hence, a lance or a musical instrument (as made of that wood)","kjv_def":"fir (tree)."},
"H1266":{"lemma":"בְּרוֹת","xlit":"bᵉrôwth","pron":"ber-oth'","derivation":"a variation of H1265 (בְּרוֹשׁ);","strongs_def":"the cypress (or some elastic tree)","kjv_def":"fir."},
"H1267":{"lemma":"בָּרוּת","xlit":"bârûwth","pron":"baw-rooth","derivation":"from H1262 (בָּרָה);","strongs_def":"food","kjv_def":"meat."},
"H1268":{"lemma":"בֵּרוֹתָה","xlit":"Bêrôwthâh","pron":"bay-ro-thaw'","derivation":"or בֵּרֹתַי; probably from H1266 (בְּרוֹת)lemma בֻּרוֹת first vowel, corrected to בְּרוֹת; cypress or cypresslike;","strongs_def":"Berothah or Berothai, a place north of Palestine","kjv_def":"Berothah, Berothai."},
"H1269":{"lemma":"בִּרְזוֹת","xlit":"Birzôwth","pron":"beer-zoth'","derivation":"probably feminine plural from an unused root (apparently mean to pierce); holes;","strongs_def":"Birzoth, an Israelite","kjv_def":"Birzavith (from the marg.)."},
"H1270":{"lemma":"בַּרְזֶל","xlit":"barzel","pron":"bar-zel'","derivation":"perhaps from the root of H1269 (בִּרְזוֹת);","strongs_def":"iron (as cutting); by extension, an iron implement","kjv_def":"(ax) head, iron."},
"H1271":{"lemma":"בַּרְזִלַּי","xlit":"Barzillay","pron":"bar-zil-lah'-ee","derivation":"from H1270 (בַּרְזֶל); iron-hearted;","strongs_def":"Barzillai, the name of three Israelites","kjv_def":"Barzillai."},
"H1272":{"lemma":"בָּרַח","xlit":"bârach","pron":"baw-rakh'","derivation":"a primitive root;","strongs_def":"to bolt, i.e. figuratively, to flee suddenly","kjv_def":"chase (away); drive away, fain, flee (away), put to flight, make haste, reach, run away, shoot."},
"H1273":{"lemma":"בַּרְחֻמִי","xlit":"Barchumîy","pron":"bar-khoo-mee'","derivation":"by transposition for H978 (בַּחֲרוּמִי);","strongs_def":"a Barchumite, or native of Bachurim","kjv_def":"Barhumite."},
"H1274":{"lemma":"בְּרִי","xlit":"bᵉrîy","pron":"ber-ee'","derivation":"from H1262 (בָּרָה);","strongs_def":"fat","kjv_def":"fat."},
"H1275":{"lemma":"בֵּרִי","xlit":"Bêrîy","pron":"bay-ree'","derivation":"probably by contraction from H882 (בְּאֵרִי);","strongs_def":"Beri, an Israelite","kjv_def":"Beri."},
"H1276":{"lemma":"בֵּרִי","xlit":"Bêrîy","pron":"bay-ree'","derivation":"of uncertain derivation;","strongs_def":"(only in the plural and with the article) the Berites, a place in Palestine","kjv_def":"Berites."},
"H1277":{"lemma":"בָּרִיא","xlit":"bârîyʼ","pron":"baw-ree'","derivation":"from H1254 (בָּרָא) (in the sense of H1262 (בָּרָה));","strongs_def":"fatted or plump","kjv_def":"fat ((fleshed), -ter), fed, firm, plenteous, rank."},
"H1278":{"lemma":"בְּרִיאָה","xlit":"bᵉrîyʼâh","pron":"ber-ee-aw'","derivation":"feminine from H1254 (בָּרָא);","strongs_def":"a creation, i.e. a novelty","kjv_def":"new thing."},
"H1279":{"lemma":"בִּרְיָה","xlit":"biryâh","pron":"beer-yaw'","derivation":"feminine from H1262 (בָּרָה);","strongs_def":"food","kjv_def":"meat."},
"H1280":{"lemma":"בְּרִיחַ","xlit":"bᵉrîyach","pron":"ber-ee'-akh","derivation":"from H1272 (בָּרַח);","strongs_def":"a bolt","kjv_def":"bar, fugitive."},
"H1281":{"lemma":"בָּרִיחַ","xlit":"bârîyach","pron":"baw-ree'-akh","derivation":"or בָּרִחַ; (shortened) from H1272 (בָּרַח);","strongs_def":"a fugitive, i.e. the serpent (as fleeing), and the constellation by that name","kjv_def":"crooked, noble, piercing."},
"H1282":{"lemma":"בָּרִיחַ","xlit":"Bârîyach","pron":"baw-ree'-akh","derivation":"the same as H1281 (בָּרִיחַ);","strongs_def":"Bariach, an Israelite","kjv_def":"Bariah."},
"H1283":{"lemma":"בְּרִיעָה","xlit":"Bᵉrîyʻâh","pron":"ber-ee'-aw","derivation":"apparently from the feminine of H7451 (רַע) with a prepositional prefix; in trouble;","strongs_def":"Beriah, the name of four Israelites","kjv_def":"Beriah."},
"H1284":{"lemma":"בְּרִיעִי","xlit":"Bᵉrîyʻîy","pron":"ber-ee-ee'","derivation":"patronymically from H1283 (בְּרִיעָה);","strongs_def":"a Beriite (collectively) or descendants of Beriah","kjv_def":"Beerites."},
"H1285":{"lemma":"בְּרִית","xlit":"bᵉrîyth","pron":"ber-eeth'","derivation":"from H1262 (בָּרָה) (in the sense of cutting (like H1254 (בָּרָא)));","strongs_def":"a compact (because made by passing between pieces of flesh)","kjv_def":"confederacy, (con-) feder(-ate), covenant, league."},
"H1286":{"lemma":"בְּרִית","xlit":"Bᵉrîyth","pron":"ber-eeth'","derivation":"the same as H1285 (בְּרִית);","strongs_def":"Berith, a Shechemitish deity","kjv_def":"Berith."},
"H1287":{"lemma":"בֹּרִית","xlit":"bôrîyth","pron":"bo-reeth'","derivation":"feminine of H1253 (בֹּר);","strongs_def":"vegetable alkali","kjv_def":"sope."},
"H1288":{"lemma":"בָרַךְ","xlit":"bârak","pron":"baw-rak'","derivation":"a primitive root;","strongs_def":"to kneel; by implication to bless God (as an act of adoration), and (vice-versa) man (as a benefit); also (by euphemism) to curse (God or the king, as treason)","kjv_def":"[idiom] abundantly, [idiom] altogether, [idiom] at all, blaspheme, bless, congratulate, curse, [idiom] greatly, [idiom] indeed, kneel (down), praise, salute, [idiom] still, thank."},
"H1289":{"lemma":"בְּרַךְ","xlit":"bᵉrak","pron":"ber-ak'","derivation":"(Aramaic) corresponding to H1288 (בָרַךְ)","strongs_def":"{to kneel; by implication to bless God (as an act of adoration), and (vice-versa) man (as a benefit); also (by euphemism) to curse (God or the king, as treason)}","kjv_def":"bless, kneel."},
"H1290":{"lemma":"בֶּרֶךְ","xlit":"berek","pron":"beh'-rek","derivation":"from H1288 (בָרַךְ);","strongs_def":"a knee","kjv_def":"knee."},
"H1291":{"lemma":"בֶּרֶךְ","xlit":"berek","pron":"beh'-rek","derivation":"(Aramaic) corresponding to H1290 (בֶּרֶךְ)","strongs_def":"{a knee}","kjv_def":"knee."},
"H1292":{"lemma":"בָּרַכְאֵל","xlit":"Bârakʼêl","pron":"baw-rak-ale'","derivation":"from H1288 (בָרַךְ) and H410 (אֵל), God has blessed;","strongs_def":"Barakel, the father of one of Job's friends","kjv_def":"Barachel."},
"H1293":{"lemma":"בְּרָכָה","xlit":"Bᵉrâkâh","pron":"ber-aw-kaw'","derivation":"from H1288 (בָרַךְ);","strongs_def":"benediction; by implication prosperity","kjv_def":"blessing, liberal, pool, present."},
"H1294":{"lemma":"בְּרָכָה","xlit":"Bᵉrâkâh","pron":"ber-aw-kaw'","derivation":"the same as H1293 (בְּרָכָה);","strongs_def":"Berakah, the name of an Israelite, and also of a valley in Palestine","kjv_def":"Berachah."},
"H1295":{"lemma":"בְּרֵכָה","xlit":"bᵉrêkâh","pron":"ber-ay-kaw'","derivation":"from H1288 (בָרַךְ);","strongs_def":"a reservoir (at which camels kneel as a resting-place)","kjv_def":"(fish-) pool."},
"H1296":{"lemma":"בֶּרֶכְיָה","xlit":"Berekyâh","pron":"beh-rek-yaw'","derivation":"or בֶּרֶכְיָהוּ; from H1290 (בֶּרֶךְ) and H3050 (יָהּ); knee (i.e. blessing) of Jah;","strongs_def":"Berekjah, the name of six Israelites","kjv_def":"Berachiah, Berechiah."},
"H1297":{"lemma":"בְּרַם","xlit":"bᵉram","pron":"ber-am'","derivation":"(Aramaic) perhaps from H7313 (רוּם) with a prepositional prefix;","strongs_def":"properly, highly, i.e. surely; but used adversatively, however","kjv_def":"but, nevertheless, yet."},
"H1298":{"lemma":"בֶּרַע","xlit":"Beraʻ","pron":"beh'-rah","derivation":"of uncertain derivation;","strongs_def":"Bera, a Sodomitish king","kjv_def":"Bera."},
"H1299":{"lemma":"בָּרַק","xlit":"bâraq","pron":"baw-rak'","derivation":"a primitive root;","strongs_def":"to lighten (lightning)","kjv_def":"cast forth."},
"H1300":{"lemma":"בָּרָק","xlit":"bârâq","pron":"baw-rawk'","derivation":"from H1299 (בָּרַק);","strongs_def":"lightning; by analogy, a gleam; concretely, a flashing sword","kjv_def":"bright, glitter(-ing sword), lightning."},
"H1301":{"lemma":"בָּרָק","xlit":"Bârâq","pron":"baw-rawk'","derivation":"the same as H1300 (בָּרָק);","strongs_def":"Barak, an Israelite","kjv_def":"Barak."},
"H1302":{"lemma":"בַּרְקוֹס","xlit":"Barqôwç","pron":"bar-kose'","derivation":"of uncertain derivation;","strongs_def":"Barkos, one of the Nethimim","kjv_def":"Barkos."},
"H1303":{"lemma":"בַּרְקָן","xlit":"barqân","pron":"bar-kwan'","derivation":"from H1300 (בָּרָק);","strongs_def":"a thorn (perhaps as burning brightly)","kjv_def":"brier."},
"H1304":{"lemma":"בָּרֶקֶת","xlit":"bâreqeth","pron":"baw-reh'-keth","derivation":"or בָּרְקַתxlit bârᵉkath corrected to bârᵉqath; from H1300 (בָּרָק);","strongs_def":"a gem (as flashing), perhaps the emerald","kjv_def":"carbuncle."},
"H1305":{"lemma":"בָּרַר","xlit":"bârar","pron":"baw-rar'","derivation":"a primitive root;","strongs_def":"to clarify (i.e. brighten), examine, select","kjv_def":"make bright, choice, chosen, cleanse (be clean), clearly, polished, (shew self) pure(-ify), purge (out)."},
"H1306":{"lemma":"בִּרְשַׁע","xlit":"Birshaʻ","pron":"beer-shah'","derivation":"probably from H7562 (רֶשַׁע) with a prepositional prefix; with wickedness;","strongs_def":"Birsha, a king of Gomorrah","kjv_def":"Birsha."},
"H1307":{"lemma":"בֵּרֹתִי","xlit":"Bêrôthîy","pron":"bay-ro-thee'","derivation":"patrial from H1268 (בֵּרוֹתָה);","strongs_def":"a Berothite, or inhabitant of Berothai","kjv_def":"Berothite."},
"H1308":{"lemma":"בְּשׂוֹר","xlit":"Bᵉsôwr","pron":"bes-ore'","derivation":"from H1319 (בָּשַׂר); cheerful;","strongs_def":"Besor, a stream of Palestine","kjv_def":"Besor."},
"H1309":{"lemma":"בְּשׂוֹרָה","xlit":"bᵉsôwrâh","pron":"bes-o-raw'","derivation":"or (shortened) בְּשֹׂרָה; feminine from H1319 (בָּשַׂר);","strongs_def":"glad tidings; by implication, reward forgood news","kjv_def":"reward for tidings."},
"H1310":{"lemma":"בָּשַׁל","xlit":"bâshal","pron":"baw-shal'","derivation":"a primitive root;","strongs_def":"properly, to boil up; hence, to be done in cooking; figuratively to ripen","kjv_def":"bake, boil, bring forth, roast, seethe, sod (be sodden)."},
"H1311":{"lemma":"בָּשֵׁל","xlit":"bâshêl","pron":"baw-shale'","derivation":"from H1310 (בָּשַׁל);","strongs_def":"boiled","kjv_def":"[idiom] at all, sodden."},
"H1312":{"lemma":"בִּשְׁלָם","xlit":"Bishlâm","pron":"bish-lawm'","derivation":"of foreign derivation;","strongs_def":"Bishlam, a Persian","kjv_def":"Bishlam."},
"H1313":{"lemma":"בָּשָׂם","xlit":"bâsâm","pron":"baw-sawm'","derivation":"from an unused root meaning to be fragrant; (compare H5561 (סַם))","strongs_def":"the balsam plant","kjv_def":"spice."},
"H1314":{"lemma":"בֶּשֶׂם","xlit":"besem","pron":"beh'-sem","derivation":"or בֹּשֶׂם; from the same as H1313 (בָּשָׂם);","strongs_def":"fragrance; by implication, spicery; also the balsam plant","kjv_def":"smell, spice, sweet (odour)."},
"H1315":{"lemma":"בׇּשְׂמַת","xlit":"Bosmath","pron":"bos-math'","derivation":"feminine of H1314 (בֶּשֶׂם) (the second form); fragrance;","strongs_def":"Bosmath, the name of a wife of Esau, and of a daughter of Solomon","kjv_def":"Bashemath, Basmath."},
"H1316":{"lemma":"בָּשָׁן","xlit":"Bâshân","pron":"baw-shawn'","derivation":"of uncertain derivation;","strongs_def":"Bashan (often with the article), a region East of the Jordan","kjv_def":"Bashan."},
"H1317":{"lemma":"בׇּשְׁנָה","xlit":"boshnâh","pron":"bosh-naw'","derivation":"feminine from H954 (בּוּשׁ);","strongs_def":"shamefulness","kjv_def":"shame."},
"H1318":{"lemma":"בָּשַׁס","xlit":"bâshaç","pron":"baw-shas'","derivation":"a primitive root;","strongs_def":"to trample down","kjv_def":"tread."},
"H1319":{"lemma":"בָּשַׂר","xlit":"bâsar","pron":"baw-sar'","derivation":"a primitive root;","strongs_def":"properly, to be fresh, i.e. full (rosy, (figuratively) cheerful); to announce (glad news)","kjv_def":"messenger, preach, publish, shew forth, (bear, bring, carry, preach, good, tell good) tidings."},
"H1320":{"lemma":"בָּשָׂר","xlit":"bâsâr","pron":"baw-sawr'","derivation":"from H1319 (בָּשַׂר);","strongs_def":"flesh (from its freshness); by extension, body, person; also (by euphemistically) the pudenda of aman","kjv_def":"body, (fat, lean) flesh(-ed), kin, (man-) kind, [phrase] nakedness, self, skin."},
"H1321":{"lemma":"בְּשַׁר","xlit":"bᵉshar","pron":"bes-ar'","derivation":"(Aramaic) corresponding to H1320 (בָּשָׂר)","strongs_def":"{flesh (from its freshness); by extension, body, person; also (by euphemistically) the pudenda of aman}","kjv_def":"flesh."},
"H1322":{"lemma":"בֹּשֶׁת","xlit":"bôsheth","pron":"bo'-sheth","derivation":"from H954 (בּוּשׁ);","strongs_def":"shame (the feeling and the condition, as well as its cause); by implication (specifically) an idol","kjv_def":"ashamed, confusion, [phrase] greatly, (put to) shame(-ful thing)."},
"H1323":{"lemma":"בַּת","xlit":"bath","pron":"bath","derivation":"from H1129 (בָּנָה) (as feminine of H1121 (בֵּן));","strongs_def":"a daughter (used in the same wide sense as other terms of relationship, literally and figuratively)","kjv_def":"apple (of the eye), branch, company, daughter, [idiom] first, [idiom] old, [phrase] owl, town, village."},
"H1324":{"lemma":"בַּת","xlit":"bath","pron":"bath","derivation":"probably from the same as H1327 (בַּתָּה);","strongs_def":"a bath or Hebrew measure (as a means of division) of liquids","kjv_def":"bath."},
"H1325":{"lemma":"בַּת","xlit":"bath","pron":"bath","derivation":"(Aramaic) corresponding to H1324 (בַּת)","strongs_def":"{a bath or Hebrew measure (as a means of division) of liquids}","kjv_def":"bath."},
"H1326":{"lemma":"בָּתָה","xlit":"bâthâh","pron":"baw-thaw'","derivation":"probably an orthographical variation for H1327 (בַּתָּה);","strongs_def":"desolation","kjv_def":"waste."},
"H1327":{"lemma":"בַּתָּה","xlit":"battâh","pron":"bat-taw'","derivation":"feminine from an unused root (meaning to break in pieces);","strongs_def":"desolation","kjv_def":"desolate."},
"H1328":{"lemma":"בְּתוּאֵל","xlit":"Bᵉthûwʼêl","pron":"beth-oo-ale'","derivation":"apparently from the same as H1326 (בָּתָה) and H410 (אֵל); destroyed of God;","strongs_def":"Bethuel, the name of a nephew of Abraham, and of a place in Palestine","kjv_def":"Bethuel. Compare H1329 (בְּתוּל)."},
"H1329":{"lemma":"בְּתוּל","xlit":"Bᵉthûwl","pron":"beth-ool'","derivation":"for H1328 (בְּתוּאֵל);","strongs_def":"Bethul (i.e. Bethuel), a place in Palestine","kjv_def":"Bethuel."},
"H1330":{"lemma":"בְּתוּלָה","xlit":"bᵉthûwlâh","pron":"beth-oo-law'","derivation":"feminine passive participle of an unused root meaning to separate;","strongs_def":"a virgin (from her privacy); sometimes (by continuation) a bride; also (figuratively) a city or state","kjv_def":"maid, virgin."},
"H1331":{"lemma":"בְּתוּלִים","xlit":"bᵉthûwlîym","pron":"beth-oo-leem'","derivation":"masculine plural of the same as H1330 (בְּתוּלָה);","strongs_def":"(collectively and abstractly) virginity; by implication and concretely, the tokens of it","kjv_def":"[idiom] maid, virginity."},
"H1332":{"lemma":"בִּתְיָה","xlit":"Bithyâh","pron":"bith-yaw'","derivation":"from H1323 (בַּת) and H3050 (יָהּ); daughter (i.e. worshipper) of Jah;","strongs_def":"Bithjah, an Egyptian woman","kjv_def":"Bithiah."},
"H1333":{"lemma":"בָּתַק","xlit":"bâthaq","pron":"baw-thak'","derivation":"a primitive root;","strongs_def":"to cut in pieces","kjv_def":"thrust through."},
"H1334":{"lemma":"בָּתַר","xlit":"bâthar","pron":"baw-thar'","derivation":"a primitive root,","strongs_def":"to chop up","kjv_def":"divide."},
"H1335":{"lemma":"בֶּתֶר","xlit":"bether","pron":"beh'-ther","derivation":"from H1334 (בָּתַר);","strongs_def":"a section","kjv_def":"part, piece."},
"H1336":{"lemma":"בֶּתֶר","xlit":"Bether","pron":"beh'-ther","derivation":"the same as H1335 (בֶּתֶר);","strongs_def":"Bether, a (craggy) place in Palestine","kjv_def":"Bether."},
"H1337":{"lemma":"בַּת רַבִּים","xlit":"Bath Rabbîym","pron":"bath rab-beem'","derivation":"from H1323 (בַּת) and a masculine plural from H7227 (רַב);","strongs_def":"the daughter (i.e. city) of Rabbah","kjv_def":"Bath-rabbim."},
"H1338":{"lemma":"בִּתְרוֹן","xlit":"Bithrôwn","pron":"bith-rone'","derivation":"from H1334 (בָּתַר); (with the article) the craggy spot;","strongs_def":"Bithron, a place East of the Jordan","kjv_def":"Bithron."},
"H1339":{"lemma":"בַּת־שֶׁבַע","xlit":"Bath-Shebaʻ","pron":"bath-sheh'-bah","derivation":"from H1323 (בַּת) and H7651 (שֶׁבַע) (in the sense of H7650 (שָׁבַע)); daughter of an oath;","strongs_def":"Bath-Sheba, the mother of Solomon","kjv_def":"Bath-sheba."},
"H1340":{"lemma":"בַּת־שׁוּעַ","xlit":"Bath-Shûwaʻ","pron":"bath-shoo'-ah","derivation":"from H1323 (בַּת) and H7771 (שׁוֹעַ); daughter of wealth;  the same as H1339 (בַּת־שֶׁבַע)","strongs_def":"Bath-shua,","kjv_def":"Bath-shua."},
"H1341":{"lemma":"גֵּא","xlit":"gêʼ","pron":"gay'","derivation":"for H1343 (גֵּאֶה);","strongs_def":"haughty","kjv_def":"proud."},
"H1342":{"lemma":"גָּאָה","xlit":"gâʼâh","pron":"gaw-aw'","derivation":"a primitive root;","strongs_def":"to mount up; hence, in general, to rise, (figuratively) be majestic","kjv_def":"gloriously, grow up, increase, be risen, triumph."},
"H1343":{"lemma":"גֵּאֶה","xlit":"gêʼeh","pron":"gay-eh'","derivation":"from H1342 (גָּאָה);","strongs_def":"lofty; figuratively, arrogant","kjv_def":"proud."},
"H1344":{"lemma":"גֵּאָה","xlit":"gêʼâh","pron":"gay-aw'","derivation":"feminine from H1342 (גָּאָה);","strongs_def":"arrogance","kjv_def":"pride."},
"H1345":{"lemma":"גְּאוּאֵל","xlit":"Gᵉʼûwʼêl","pron":"gheh-oo-ale'","derivation":"from H1342 (גָּאָה) and H410 (אֵל); majesty of God;","strongs_def":"Geuel, an Israelite","kjv_def":"Geuel."},
"H1346":{"lemma":"גַּאֲוָה","xlit":"gaʼăvâh","pron":"gah-av-aw'","derivation":"from H1342 (גָּאָה);","strongs_def":"arrogance or majesty; by implication, (concretely) ornament","kjv_def":"excellency, haughtiness, highness, pride, proudly, swelling."},
"H1347":{"lemma":"גָּאוֹן","xlit":"gâʼôwn","pron":"gaw-ohn'","derivation":"from H1342 (גָּאָה); the same as H1346 (גַּאֲוָה)xlit gaʻăvâh corrected to gaʼăvâh","strongs_def":"{arrogance or majesty; by implication, (concretely) ornament}","kjv_def":"arrogancy, excellency(-lent), majesty, pomp, pride, proud, swelling."},
"H1348":{"lemma":"גֵּאוּת","xlit":"gêʼûwth","pron":"gay-ooth'","derivation":"from H1342 (גָּאָה); the same as H1346 (גַּאֲוָה)xlit gaʻăvâh corrected to gaʼăvâh","strongs_def":"{arrogance or majesty; by implication, (concretely) ornament}","kjv_def":"excellent things, lifting up, majesty, pride, proudly, raging."},
"H1349":{"lemma":"גַּאֲיוֹן","xlit":"gaʼăyôwn","pron":"gah-ah-yone'","derivation":"from H1342 (גָּאָה);","strongs_def":"haughty","kjv_def":"proud."},
"H1350":{"lemma":"גָּאַל","xlit":"gâʼal","pron":"gaw-al'","derivation":"a primitive root, to redeem (according to the Oriental law of kinship), i.e.","strongs_def":"to be the next of kin (and as such to buy back a relative's property, marry his widow, etc.)","kjv_def":"[idiom] in any wise, [idiom] at all, avenger, deliver, (do, perform the part of near, next) kinsfolk(-man), purchase, ransom, redeem(-er), revenger."},
"H1351":{"lemma":"גָּאַל","xlit":"gâʼal","pron":"gaw-al'","derivation":"a primitive root, (rather identified with H1350 (גָּאַל), through the idea of freeing, i.e. repudiating);","strongs_def":"to soil or (figuratively) desecrate","kjv_def":"defile, pollute, stain."},
"H1352":{"lemma":"גֹּאֶל","xlit":"gôʼel","pron":"go'-el","derivation":"from H1351 (גָּאַל);","strongs_def":"profanation","kjv_def":"defile."},
"H1353":{"lemma":"גְּאֻלָּה","xlit":"gᵉʼullâh","pron":"gheh-ool-law'","derivation":"feminine passive participle of H1350 (גָּאַל);","strongs_def":"redemption (including the right and the object); by implication, relationship","kjv_def":"kindred, redeem, redemption, right."},
"H1354":{"lemma":"גַב","xlit":"gab","pron":"gab","derivation":"from an unused root meaning to hollow or curve; (compare H1460 (גֵּו) and H1479 (גּוּף))","strongs_def":"the back (as rounded); by analogy, the top or rim, a boss, a vault, arch of eye, bulwarks, etc.","kjv_def":"back, body, boss, eminent (higher) place, (eye) brows, nave, ring."},
"H1355":{"lemma":"גַּב","xlit":"gab","pron":"gab","derivation":"(Aramaic) corresponding to H1354 (גַב)","strongs_def":"{the back (as rounded); by analogy, the top or rim, a boss, a vault, arch of eye, bulwarks, etc.}","kjv_def":"back."},
"H1356":{"lemma":"גֵּב","xlit":"gêb","pron":"gabe","derivation":"from H1461 (גּוּב);","strongs_def":"a log (as cut out); also well or cistern (as dug)","kjv_def":"beam, ditch, pit."},
"H1357":{"lemma":"גֵּב","xlit":"gêb","pron":"gabe","derivation":"probably from H1461 (גּוּב) (compare H1462 (גּוֹב));","strongs_def":"a locust (from its cutting)","kjv_def":"locust."},
"H1358":{"lemma":"גֹּב","xlit":"gôb","pron":"gobe","derivation":"(Aramaic) from a root corresponding to H1461 (גּוּב);","strongs_def":"a pit (for wild animals) (as cut out)","kjv_def":"den."},
"H1359":{"lemma":"גֹּב","xlit":"Gôb","pron":"gobe","derivation":"or (fully) גּוֹב; from H1461 (גּוּב); pit;","strongs_def":"Gob, a place in Palestine","kjv_def":"Gob."},
"H1360":{"lemma":"גֶּבֶא","xlit":"gebeʼ","pron":"geh'-beh","derivation":"from an unused root meaning probably to collect;","strongs_def":"a reservoir; by analogy, a marsh","kjv_def":"marish, pit."},
"H1361":{"lemma":"גָּבַהּ","xlit":"gâbahh","pron":"gaw-bah'","derivation":"a primitive root;","strongs_def":"to soar, i.e. be lofty; figuratively, to be haughty","kjv_def":"exalt, be haughty, be (make) high(-er), lift up, mount up, be proud, raise up great height, upward."},
"H1362":{"lemma":"גָּבָהּ","xlit":"gâbâhh","pron":"gaw-bawh'","derivation":"from H1361 (גָּבַהּ);","strongs_def":"lofty (literally or figuratively)","kjv_def":"high, proud."},
"H1363":{"lemma":"גֹּבַהּ","xlit":"gôbahh","pron":"go'-bah","derivation":"from H1361 (גָּבַהּ);","strongs_def":"elation, grandeur, arrogance","kjv_def":"excellency, haughty, height, high, loftiness, pride."},
"H1364":{"lemma":"גָּבֹהַּ","xlit":"gâbôahh","pron":"gaw-bo'-ah","derivation":"or גָּבוֹהַּ; (fully) from H1361 (גָּבַהּ);","strongs_def":"elevated (or elated), powerful, arrogant","kjv_def":"haughty, height, high(-er), lofty, proud, [idiom] exceeding proudly."},
"H1365":{"lemma":"גַּבְהוּת","xlit":"gabhûwth","pron":"gab-hooth'","derivation":"from H1361 (גָּבַהּ);","strongs_def":"pride","kjv_def":"loftiness, lofty."},
"H1366":{"lemma":"גְּבוּל","xlit":"gᵉbûwl","pron":"gheb-ool'","derivation":"or גְּבֻל; (shortened) from H1379 (גָּבַל);","strongs_def":"properly, a cord (as twisted), i.e. (by implication) a boundary; by extension the territory inclosed","kjv_def":"border, bound, coast, [idiom] great, landmark, limit, quarter, space."},
"H1367":{"lemma":"גְּבוּלָה","xlit":"gᵉbûwlâh","pron":"gheb-oo-law'","derivation":"or גְּבֻלָה; (shortened) feminine of H1366 (גְּבוּל);","strongs_def":"a boundary, region","kjv_def":"border, bound, coast, landmark. place."},
"H1368":{"lemma":"גִּבּוֹר","xlit":"gibbôwr","pron":"ghib-bore'","derivation":"or גִּבֹּר; (shortened) intensive from the same as H1397 (גֶּבֶר);","strongs_def":"powerful; by implication, warrior, tyrant","kjv_def":"champion, chief, [idiom] excel, giant, man, mighty (man, one), strong (man), valiant man."},
"H1369":{"lemma":"גְּבוּרָה","xlit":"gᵉbûwrâh","pron":"gheb-oo-raw'","derivation":"feminine passive participle from the same as H1368 (גִּבּוֹר);","strongs_def":"force (literally or figuratively); by implication, valor, victory","kjv_def":"force, mastery, might, mighty (act, power), power, strength."},
"H1370":{"lemma":"גְּבוּרָה","xlit":"gᵉbûwrâh","pron":"gheb-oo-raw'","derivation":"(Aramaic) corresponding to H1369 (גְּבוּרָה);","strongs_def":"power","kjv_def":"might."},
"H1371":{"lemma":"גִּבֵּחַ","xlit":"gibbêach","pron":"ghib-bay'-akh","derivation":"from an unused root meaning to be high (in the forehead);","strongs_def":"bald in the forehead","kjv_def":"forehead bald."},
"H1372":{"lemma":"גַּבַּחַת","xlit":"gabbachath","pron":"gab-bakh'-ath","derivation":"from the same as H1371 (גִּבֵּחַ);","strongs_def":"baldness in the forehead; by analogy, a bare spot on the right side of cloth","kjv_def":"bald forehead, [idiom] without."},
"H1373":{"lemma":"גַּבַּי","xlit":"Gabbay","pron":"gab-bah'ee","derivation":"from the same as H1354 (גַב); collective","strongs_def":"Gabbai, an Israelite","kjv_def":"Gabbai."},
"H1374":{"lemma":"גֵּבִים","xlit":"Gêbîym","pron":"gay-beem'","derivation":"plural of H1356 (גֵּב); cisterns;","strongs_def":"Gebim, a place in Palestine","kjv_def":"Gebim."},
"H1375":{"lemma":"גְּבִיעַ","xlit":"gᵉbîyaʻ","pron":"gheb-ee'-ah","derivation":"from an unused root (meaning to be convex);","strongs_def":"a goblet; by analogy, the calyx of aflower","kjv_def":"house, cup, pot."},
"H1376":{"lemma":"גְּבִיר","xlit":"gᵉbîyr","pron":"gheb-eer'","derivation":"from H1396 (גָּבַר);","strongs_def":"a master","kjv_def":"lord."},
"H1377":{"lemma":"גְּבִירָה","xlit":"gᵉbîyrâh","pron":"gheb-ee-raw'","derivation":"feminine of H1376 (גְּבִיר);","strongs_def":"a mistress","kjv_def":"queen."},
"H1378":{"lemma":"גָּבִישׁ","xlit":"gâbîysh","pron":"gaw-beesh'","derivation":"from an unused root (probably meaning to freeze);","strongs_def":"crystal (from its resemblance to ice)","kjv_def":"pearl."},
"H1379":{"lemma":"גָּבַל","xlit":"gâbal","pron":"gaw-bal'","derivation":"a primitive root; also as a denominative from H1366 (גְּבוּל)","strongs_def":"properly, to twist as arope;  to bound (as by a line)","kjv_def":"be border, set (bounds about)."},
"H1380":{"lemma":"גְּבַל","xlit":"Gᵉbal","pron":"gheb-al'","derivation":"from H1379 (גָּבַל) (in the sense of a chain of hills); a mountain;","strongs_def":"Gebal, a place in Phoenicia","kjv_def":"Gebal."},
"H1381":{"lemma":"גְּבָל","xlit":"Gᵉbâl","pron":"gheb-awl'","derivation":"the same as H1380 (גְּבַל);","strongs_def":"Gebal, a region in Idumaea","kjv_def":"Gebal."},
"H1382":{"lemma":"גִּבְלִי","xlit":"Giblîy","pron":"ghib-lee'","derivation":"; patrial from H1380 (גְּבַל);","strongs_def":"a Gebalite, or inhabitant of Gebal","kjv_def":"Giblites, stone-squarer."},
"H1383":{"lemma":"גַּבְלֻת","xlit":"gabluth","pron":"gab-looth'","derivation":"from H1379 (גָּבַל);","strongs_def":"a twisted chain or lace","kjv_def":"end."},
"H1384":{"lemma":"גִּבֵּן","xlit":"gibbên","pron":"gib-bane'","derivation":"from an unused root meaning to be arched or contracted;","strongs_def":"hunch-backed","kjv_def":"crookbackt."},
"H1385":{"lemma":"גְּבִנָה","xlit":"gᵉbinâh","pron":"gheb-ee-naw'","derivation":"feminine from the same as H1384 (גִּבֵּן);","strongs_def":"curdled milk","kjv_def":"cheese."},
"H1386":{"lemma":"גַּבְנֹן","xlit":"gabnôn","pron":"gab-nohn'","derivation":"from the same as H1384 (גִּבֵּן);","strongs_def":"a hump or peak of hills","kjv_def":"high."},
"H1387":{"lemma":"גֶּבַע","xlit":"Gebaʻ","pron":"gheh'-bah","derivation":"from the same as H1375 (גְּבִיעַ), a hillock;","strongs_def":"Geba, a place in Palestine","kjv_def":"Gaba, Geba, Gibeah."},
"H1388":{"lemma":"גִּבְעָא","xlit":"Gibʻâʼ","pron":"ghib-aw'","derivation":"by permutation for H1389 (גִּבְעָה); a hill;","strongs_def":"Giba, a place in Palestine","kjv_def":"Gibeah."},
"H1389":{"lemma":"גִּבְעָה","xlit":"gibʻâh","pron":"ghib-aw'","derivation":"feminine from the same as H1387 (גֶּבַע);","strongs_def":"a hillock","kjv_def":"hill, little hill."},
"H1390":{"lemma":"גִּבְעָה","xlit":"Gibʻâh","pron":"ghib-aw'","derivation":"the same as H1389 (גִּבְעָה);","strongs_def":"Gibah; the name of three places in Palestine","kjv_def":"Gibeah, the hill."},
"H1391":{"lemma":"גִּבְעוֹן","xlit":"Gibʻôwn","pron":"ghib-ohn'","derivation":"from the same as H1387 (גֶּבַע); hilly;","strongs_def":"Gibon, a place in Palestine","kjv_def":"Gibeon."},
"H1392":{"lemma":"גִּבְעֹל","xlit":"gibʻôl","pron":"ghib-ole'","derivation":"prolonged from H1375 (גְּבִיעַ);","strongs_def":"the calyx of a flower","kjv_def":"bolled."},
"H1393":{"lemma":"גִּבְעֹנִי","xlit":"Gibʻônîy","pron":"ghib-o-nee'","derivation":"patrial from H1391 (גִּבְעוֹן);","strongs_def":"a Gibonite, or inhabitant of Gibon","kjv_def":"Gibeonite."},
"H1394":{"lemma":"גִּבְעַת","xlit":"Gibʻath","pron":"ghib-ath'","derivation":"from the same as H1375 (גְּבִיעַ); hilliness;","strongs_def":"Gibath","kjv_def":"Gibeath."},
"H1395":{"lemma":"גִּבְעָתִי","xlit":"Gibʻâthîy","pron":"ghib-aw-thee'","derivation":"patrial from H1390 (גִּבְעָה);","strongs_def":"a Gibathite, or inhabitant of Gibath","kjv_def":"Gibeathite."},
"H1396":{"lemma":"גָּבַר","xlit":"gâbar","pron":"gaw-bar'","derivation":"a primitive root;","strongs_def":"to be strong; by implication, to prevail, act insolently","kjv_def":"exceed, confirm, be great, be mighty, prevail, put to more (strength), strengthen, be stronger, be valiant."},
"H1397":{"lemma":"גֶּבֶר","xlit":"geber","pron":"gheh'-ber","derivation":"from H1396 (גָּבַר);","strongs_def":"properly, a valiant man or warrior; generally, a person simply","kjv_def":"every one, man, [idiom] mighty."},
"H1398":{"lemma":"גֶּבֶר","xlit":"Geber","pron":"gheh'-ber","derivation":"the same as H1397 (גֶּבֶר);","strongs_def":"Geber, the name of two Israelites","kjv_def":"Geber."},
"H1399":{"lemma":"גְּבַר","xlit":"gᵉbar","pron":"gheb-ar'","derivation":"from H1396 (גָּבַר); the same as H1397 (גֶּבֶר);","strongs_def":"a person","kjv_def":"man."},
"H1400":{"lemma":"גְּבַר","xlit":"gᵉbar","pron":"gheb-ar'","derivation":"(Aramaic) corresponding to H1399 (גְּבַר)","strongs_def":"{a person}","kjv_def":"certain, man."},
"H1401":{"lemma":"גִּבָּר","xlit":"gibbâr","pron":"ghib-bawr'","derivation":"(Aramaic) intensive of H1400 (גְּבַר);","strongs_def":"valiant, or warrior","kjv_def":"mighty."},
"H1402":{"lemma":"גִּבָּר","xlit":"Gibbâr","pron":"ghib-bawr'","derivation":"intensive of H1399 (גְּבַר);","strongs_def":"Gibbar, an Israelite","kjv_def":"Gibbar."},
"H1403":{"lemma":"גַּבְרִיאֵל","xlit":"Gabrîyʼêl","pron":"gab-ree-ale'","derivation":"from H1397 (גֶּבֶר) and H410 (אֵל); man of God;","strongs_def":"Gabriel, an archangel","kjv_def":"Garbriel."},
"H1404":{"lemma":"גְּבֶרֶת","xlit":"gᵉbereth","pron":"gheb-eh'-reth","derivation":"feminine of H1376 (גְּבִיר);","strongs_def":"mistress","kjv_def":"lady, mistress."},
"H1405":{"lemma":"גִּבְּתוֹן","xlit":"Gibbᵉthôwn","pron":"ghib-beth-one'","derivation":"intensive from H1389 (גִּבְעָה); a hilly spot;","strongs_def":"Gibbethon, a place in Palestine","kjv_def":"Gibbethon."},
"H1406":{"lemma":"גָּג","xlit":"gâg","pron":"gawg","derivation":"probably by reduplication from H1342 (גָּאָה);","strongs_def":"a roof; by analogy, the top of an altar","kjv_def":"roof (of the house), (house) top (of the house)."},
"H1407":{"lemma":"גַּד","xlit":"gad","pron":"gad","derivation":"from H1413 (גָּדַד) (in the sense of cutting);","strongs_def":"coriander seed (from its furrows)","kjv_def":"coriander."},
"H1408":{"lemma":"גַּד","xlit":"Gad","pron":"gad","derivation":"a variation of H1409 (גָּד);","strongs_def":"Fortune, a Babylonian deity","kjv_def":"that troop."},
"H1409":{"lemma":"גָּד","xlit":"gâd","pron":"gawd","derivation":"from H1464 (גּוּד) (in the sense of distributing);","strongs_def":"fortune","kjv_def":"troop."},
"H1410":{"lemma":"גָּד","xlit":"Gâd","pron":"gawd","derivation":"from H1464 (גּוּד);","strongs_def":"Gad, a son of Jacob, including his tribe and its territory; also a prophet","kjv_def":"Gad."},
"H1411":{"lemma":"גְּדָבָר","xlit":"gᵉdâbâr","pron":"ghed-aw-bawr'","derivation":"(Aramaic) corresponding to H1489 (גִּזְבָּר);","strongs_def":"a treasurer","kjv_def":"treasurer."},
"H1412":{"lemma":"גֻּדְגֹּדָה","xlit":"Gudgôdâh","pron":"gud-go'-daw","derivation":"by reduplication from H1413 (גָּדַד) (in the sense of cutting) cleft;","strongs_def":"Gudgodah, a place in the Desert","kjv_def":"Gudgodah."},
"H1413":{"lemma":"גָּדַד","xlit":"gâdad","pron":"gaw-dad'","derivation":"a primitive root (compare H1464 (גּוּד));","strongs_def":"to crowd; also to gash (as if by pressing into)","kjv_def":"assemble (selves by troops), gather (selves together, self in troops), cut selves."},
"H1414":{"lemma":"גְּדַד","xlit":"gᵉdad","pron":"ghed-ad'","derivation":"(Aramaic) corresponding to H1413 (גָּדַד);","strongs_def":"to cut down","kjv_def":"hew down."},
"H1415":{"lemma":"גָּדָה","xlit":"gâdâh","pron":"gaw-daw'","derivation":"from an unused root (meaning to cut off);","strongs_def":"a border of a river (as cut into by the stream)","kjv_def":"bank."},
"H1416":{"lemma":"גְּדוּד","xlit":"gᵉdûwd","pron":"ghed-ood'","derivation":"from H1413 (גָּדַד);","strongs_def":"a crowd (especially of soldiers)","kjv_def":"army, band (of men), company, troop (of robbers)."},
"H1417":{"lemma":"גְּדוּד","xlit":"gᵉdûwd","pron":"ghed-ood'","derivation":"or גְּדֻדָה; (feminine) from H1413 (גָּדַד);","strongs_def":"a furrow (as cut)","kjv_def":"furrow."},
"H1418":{"lemma":"גְּדוּדָה","xlit":"gᵉdûwdâh","pron":"ghed-oo-daw'","derivation":"feminine participle passive of H1413 (גָּדַד);","strongs_def":"an incision","kjv_def":"cutting."},
"H1419":{"lemma":"גָּדוֹל","xlit":"gâdôwl","pron":"gaw-dole'","derivation":"or גָּדֹל; (shortened) from H1431 (גָּדַל);","strongs_def":"great (in any sense); hence, older; also insolent","kjv_def":"[phrase] aloud, elder(-est), [phrase] exceeding(-ly), [phrase] far, (man of) great (man, matter, thing,-er,-ness), high, long, loud, mighty, more, much, noble, proud thing, [idiom] sore, ([idiom]) very."},
"H1420":{"lemma":"גְּדוּלָה","xlit":"gᵉdûwlâh","pron":"ghed-oo-law'","derivation":"or גְּדֻלָּה; (shortened) or (less accurately) גְּדוּלָּה; feminine of H1419 (גָּדוֹל);","strongs_def":"greatness; (concretely) mighty acts","kjv_def":"dignity, great things(-ness), majesty."},
"H1421":{"lemma":"גִּדּוּף","xlit":"giddûwph","pron":"ghid-doof'","derivation":"or (shortened) גִּדֻּף; and (feminine) גִּדּוּפָהxlit giddûphâh corrected to giddûwphâh; or גִּדֻּפָה; from H1422 (גְּדוּפָה);","strongs_def":"vilification","kjv_def":"reproach, reviling."},
"H1422":{"lemma":"גְּדוּפָה","xlit":"gᵉdûwphâh","pron":"ghed-oo-faw'","derivation":"feminine passive participle of H1442 (גָּדַף);","strongs_def":"a revilement","kjv_def":"taunt."},
"H1423":{"lemma":"גְּדִי","xlit":"gᵉdîy","pron":"ghed-ee'","derivation":"from the same as H1415 (גָּדָה);","strongs_def":"a young goat (from browsing)","kjv_def":"kid."},
"H1424":{"lemma":"גָּדִי","xlit":"Gâdîy","pron":"gaw-dee'","derivation":"from H1409 (גָּד); fortunate;","strongs_def":"Gadi, an Israelite","kjv_def":"Gadi."},
"H1425":{"lemma":"גָּדִי","xlit":"Gâdîy","pron":"gaw-dee'","derivation":"patronymically from H1410 (גָּד);","strongs_def":"a Gadite (collectively) or descendants of Gad","kjv_def":"Gadites, children of Gad."},
"H1426":{"lemma":"גַּדִּי","xlit":"Gaddîy","pron":"gad-dee'","derivation":"intensive for H1424 (גָּדִי);","strongs_def":"Gaddi, an Israelite","kjv_def":"Gaddi."},
"H1427":{"lemma":"גַּדִּיאֵל","xlit":"Gaddîyʼêl","pron":"gad-dee-ale'","derivation":"from H1409 (גָּד) and H410 (אֵל); fortune of God;","strongs_def":"Gaddiel, an Israelite","kjv_def":"Gaddiel."},
"H1428":{"lemma":"גִּדְיָה","xlit":"gidyâh","pron":"ghid-yaw'","derivation":"or גַּדיָה; the same as H1415 (גָּדָה);","strongs_def":"a river brink","kjv_def":"bank."},
"H1429":{"lemma":"גְּדִיָּה","xlit":"gᵉdîyâh","pron":"ghed-ee-yaw'","derivation":"feminine of H1423 (גְּדִי);","strongs_def":"a young female goat","kjv_def":"kid."},
"H1430":{"lemma":"גָּדִישׁ","xlit":"gâdîysh","pron":"gaw-deesh'","derivation":"from an unused root (meaning to heap up);","strongs_def":"a stack of sheaves; by analogy, a tomb","kjv_def":"shock (stack) (of corn), tomb."},
"H1431":{"lemma":"גָּדַל","xlit":"gâdal","pron":"gaw-dal'","derivation":"a primitive root; properly, to twist (compare H1434 (גְּדִל)), i.e.","strongs_def":"to be (causatively make) large (in various senses, as in body, mind, estate or honor, also in pride)","kjv_def":"advance, boast, bring up, exceed, excellent, be(-come, do, give, make, wax), great(-er, come to... estate, [phrase] things), grow(up), increase, lift up, magnify(-ifical), be much set by, nourish (up), pass, promote, proudly (spoken), tower."},
"H1432":{"lemma":"גָּדֵל","xlit":"gâdêl","pron":"gaw-dale'","derivation":"from H1431 (גָּדַל);","strongs_def":"large (literally or figuratively)","kjv_def":"great, grew."},
"H1433":{"lemma":"גֹּדֶל","xlit":"gôdel","pron":"go'-del","derivation":"from H1431 (גָּדַל);","strongs_def":"magnitude (literally or figuratively)","kjv_def":"greatness, stout(-ness)."},
"H1434":{"lemma":"גְּדִל","xlit":"gᵉdil","pron":"ghed-eel'","derivation":"from H1431 (גָּדַל) (in the sense of twisting);","strongs_def":"thread, i.e. a tassel or festoon","kjv_def":"fringe, wreath."},
"H1435":{"lemma":"גִּדֵּל","xlit":"Giddêl","pron":"ghid-dale'","derivation":"from H1431 (גָּדַל); stout;","strongs_def":"Giddel, the name of one of the Nethinim, also of one of 'Solomon's servants'","kjv_def":"Giddel."},
"H1436":{"lemma":"גְּדַּלְיָה","xlit":"Gᵉdalyâh","pron":"ghed-al-yaw'","derivation":"or (prolonged) גְּדַלְיָהוּ; from H1431 (גָּדַל) and H3050 (יָהּ); Jah has become great;","strongs_def":"Gedaljah, the name of five Israelites","kjv_def":"Gedaliah."},
"H1437":{"lemma":"גִּדַּלְתִּי","xlit":"Giddaltîy","pron":"ghid-dal'-tee","derivation":"from H1431 (גָּדַל); I have made great;","strongs_def":"Giddalti, an Israelite","kjv_def":"Giddalti."},
"H1438":{"lemma":"גָּדַע","xlit":"gâdaʻ","pron":"gaw-dah'","derivation":"a primitive root;","strongs_def":"to fell a tree; generally, to destroy anything","kjv_def":"cut (asunder, in sunder, down, off), hew down."},
"H1439":{"lemma":"גִּדְעוֹן","xlit":"Gidʻôwn","pron":"ghid-ohn'","derivation":"from H1438 (גָּדַע); feller (i.e. warrior);","strongs_def":"Gidon, an Israelite","kjv_def":"Gideon."},
"H1440":{"lemma":"גִּדְעֹם","xlit":"Gidʻôm","pron":"ghid-ohm'","derivation":"from H1438 (גָּדַע); a cutting (i.e. desolation);","strongs_def":"Gidom, a place in Palestine","kjv_def":"Gidom."},
"H1441":{"lemma":"גִּדְעֹנִי","xlit":"Gidʻônîy","pron":"ghid-o-nee'","derivation":"from H1438 (גָּדַע); warlike (compare H1439 (גִּדְעוֹן));","strongs_def":"Gidoni, an Israelite","kjv_def":"Gideoni."},
"H1442":{"lemma":"גָּדַף","xlit":"gâdaph","pron":"gaw-daf'","derivation":"a primitive root;","strongs_def":"to hack (with words), i.e. revile","kjv_def":"blaspheme, reproach."},
"H1443":{"lemma":"גָּדַר","xlit":"gâdar","pron":"gaw-dar'","derivation":"a primitive root;","strongs_def":"to wall in or around","kjv_def":"close up, fence up, hedge, inclose, make up (a wall), mason, repairer."},
"H1444":{"lemma":"גֶּדֶר","xlit":"geder","pron":"gheh'-der","derivation":"from H1443 (גָּדַר);","strongs_def":"a circumvallation","kjv_def":"wall."},
"H1445":{"lemma":"גֶּדֶר","xlit":"Geder","pron":"gheh'-der","derivation":"; the same as H1444 (גֶּדֶר);","strongs_def":"Geder, a place in Palestine","kjv_def":"Geder."},
"H1446":{"lemma":"גְדֹר","xlit":"Gᵉdôr","pron":"ghed-ore'","derivation":"or (fully) גְּדוֹר; from H1443 (גָּדַר); inclosure;","strongs_def":"Gedor, a place in Palestine; also the name of three Israelites","kjv_def":"Gedor."},
"H1447":{"lemma":"גָּדֵר","xlit":"gâdêr","pron":"gaw-dare'","derivation":"from H1443 (גָּדַר);","strongs_def":"a circumvallation; by implication, an inclosure","kjv_def":"fence, hedge, wall."},
"H1448":{"lemma":"גְּדֵרָה","xlit":"gᵉdêrâh","pron":"ghed-ay-raw'","derivation":"feminine of H1447 (גָּדֵר);","strongs_def":"enclosure (especially for flocks)","kjv_def":"(sheep-) cote (fold) hedge, wall."},
"H1449":{"lemma":"גְּדֵרָה","xlit":"Gᵉdêrâh","pron":"ghed-ay-raw'","derivation":"the same as H1448 (גְּדֵרָה);","strongs_def":"(with the article) Gederah, a place in Palestine","kjv_def":"Gederah, hedges."},
"H1450":{"lemma":"גְּדֵרוֹת","xlit":"Gᵉdêrôwth","pron":"ghed-ay-rohth'","derivation":"plural of H1448 (גְּדֵרָה); walls;","strongs_def":"Gederoth, a place in Palestine","kjv_def":"Gederoth."},
"H1451":{"lemma":"גְּדֵרִי","xlit":"Gᵉdêrîy","pron":"ghed-ay-ree'","derivation":"patrial from H1445 (גֶּדֶר);","strongs_def":"a Gederite, or inhabitant of Geder","kjv_def":"Gederite."},
"H1452":{"lemma":"גְּדֵרָתִי","xlit":"Gᵉdêrâthîy","pron":"ghed-ay-raw-thee'","derivation":"patrial from H1449 (גְּדֵרָה);","strongs_def":"a Gederathite, or inhabitant of Gederah","kjv_def":"Gederathite."},
"H1453":{"lemma":"גְּדֵרֹתַיִם","xlit":"Gᵉdêrôthayim","pron":"ghed-ay-ro-thah'-yim","derivation":"dual of H1448 (גְּדֵרָה); double wall;","strongs_def":"Gederothajim, a place in Palestine","kjv_def":"Gederothaim."},
"H1454":{"lemma":"גֵּה","xlit":"gêh","pron":"gay","derivation":"probably a clerical error for H2088 (זֶה);","strongs_def":"this","kjv_def":"this."},
"H1455":{"lemma":"גָּהָה","xlit":"gâhâh","pron":"gaw-haw'","derivation":"a primitive root;","strongs_def":"to remove (a bandage from a wound, i.e. heal it)","kjv_def":"cure."},
"H1456":{"lemma":"גֵּהָה","xlit":"gêhâh","pron":"gay-haw'","derivation":"from H1455 (גָּהָה);","strongs_def":"a cure","kjv_def":"medicine."},
"H1457":{"lemma":"גָּהַר","xlit":"gâhar","pron":"gaw-har'","derivation":"a primitive root;","strongs_def":"to prostrate oneself","kjv_def":"cast self down, stretch self."},
"H1458":{"lemma":"גַּו","xlit":"gav","pron":"gav","derivation":"another form for H1460 (גֵּו);","strongs_def":"the back","kjv_def":"back."},
"H1459":{"lemma":"גַּו","xlit":"gav","pron":"gav","derivation":"(Aramaic) corresponding to H1460 (גֵּו);","strongs_def":"the middle","kjv_def":"midst, same, there-(where-) in."},
"H1460":{"lemma":"גֵּו","xlit":"gêv","pron":"gave","derivation":"from H1342 (גָּאָה) (corresponding to H1354 (גַב));","strongs_def":"the back; by analogy, the middle","kjv_def":"[phrase] among, back, body."},
"H1461":{"lemma":"גּוּב","xlit":"gûwb","pron":"goob","derivation":"a primitive root;","strongs_def":"to dig","kjv_def":"husbandman."},
"H1462":{"lemma":"גּוֹב","xlit":"gôwb","pron":"gobe","derivation":"from H1461 (גּוּב);","strongs_def":"the locust (from its grubbing as a larvae)","kjv_def":"grasshopper, [idiom] great."},
"H1463":{"lemma":"גּוֹג","xlit":"Gôwg","pron":"gohg","derivation":"of uncertain derivation;","strongs_def":"Gog, the name of an Israelite, also of some nothern nation","kjv_def":"Gog."},
"H1464":{"lemma":"גּוּד","xlit":"gûwd","pron":"goode","derivation":"a primitive root (akin to H1413 (גָּדַד));","strongs_def":"to crowd upon, i.e. attack","kjv_def":"invade, overcome."},
"H1465":{"lemma":"גֵּוָה","xlit":"gêvâh","pron":"gay-vaw'","derivation":"feminine of H1460 (גֵּו);","strongs_def":"the back, i.e. (by extensive) the person","kjv_def":"body."},
"H1466":{"lemma":"גֵּוָה","xlit":"gêvâh","pron":"gay-vaw'","derivation":"the same as H1465 (גֵּוָה);","strongs_def":"exaltation; (figuratively) arrogance","kjv_def":"lifting up, pride."},
"H1467":{"lemma":"גֵּוָה","xlit":"gêvâh","pron":"gay-vaw'","derivation":"(Aramaic) corresponding to H1466 (גֵּוָה)","strongs_def":"{exaltation; (figuratively) arrogance}","kjv_def":"pride."},
"H1468":{"lemma":"גּוּז","xlit":"gûwz","pron":"gooz","derivation":"a primitive root (compare H1494 (גַּזָז));","strongs_def":"properly, to shear off; but used only in the (figuratively) sense of passing rapidly","kjv_def":"bring, cut off."},
"H1469":{"lemma":"גּוֹזָל","xlit":"gôwzâl","pron":"go-zawl'","derivation":"or (shortened) גֹזָלlemma גזָל missing vowel, corrected to גֹזָל; from H1497 (גָּזַל);","strongs_def":"a nestling (as being comparatively nude of feathers)","kjv_def":"young (pigeon)."},
"H1470":{"lemma":"גּוֹזָן","xlit":"Gôwzân","pron":"go-zawn'","derivation":"probably from H1468 (גּוּז); a quarry (as a place of cutting stones);","strongs_def":"Gozan, a province of Assyria","kjv_def":"Gozan."},
"H1471":{"lemma":"גּוֹי","xlit":"gôwy","pron":"go'-ee","derivation":"rarely (shortened) גֹּי; apparently from the same root as H1465 (גֵּוָה) (in the sense of massing);","strongs_def":"a foreign nation; hence, a Gentile; also (figuratively) a troop of animals, or a flight of locusts","kjv_def":"Gentile, heathen, nation, people."},
"H1472":{"lemma":"גְּוִיָּה","xlit":"gᵉvîyâh","pron":"ghev-ee-yaw'","derivation":"prolonged for H1465 (גֵּוָה);","strongs_def":"a body, whether alive or dead","kjv_def":"(dead) body, carcase, corpse."},
"H1473":{"lemma":"גּוֹלָה","xlit":"gôwlâh","pron":"go-law'","derivation":"or (shortened) גֹּלָה; active participle feminine of H1540 (גָּלָה);","strongs_def":"exile; concretely and collectively exiles","kjv_def":"(carried away), captive(-ity), removing."},
"H1474":{"lemma":"גּוֹלָן","xlit":"Gôwlân","pron":"go-lawn'","derivation":"from H1473 (גּוֹלָה); captive;","strongs_def":"Golan, a place east of the Jordan","kjv_def":"Golan."},
"H1475":{"lemma":"גּוּמָּץ","xlit":"gûwmmâts","pron":"goom-mawts'","derivation":"of uncertain derivation;","strongs_def":"a pit","kjv_def":"pit."},
"H1476":{"lemma":"גּוּנִי","xlit":"Gûwnîy","pron":"goo-nee'","derivation":"probably from H1598 (גָּנַן); protected;","strongs_def":"Guni, the name of two Israelites","kjv_def":"Guni."},
"H1477":{"lemma":"גּוּנִי","xlit":"Gûwnîy","pron":"goo-nee'","derivation":"patronymically from H1476 (גּוּנִי);","strongs_def":"a Gunite (collectively with article prefix) or descendants of Guni","kjv_def":"Gunites."},
"H1478":{"lemma":"גָּוַע","xlit":"gâvaʻ","pron":"gaw-vah'","derivation":"a primitive root;","strongs_def":"to breathe out, i.e. (by implication) expire","kjv_def":"die, be dead, give up the ghost, perish."},
"H1479":{"lemma":"גּוּף","xlit":"gûwph","pron":"goof","derivation":"a primitive root;","strongs_def":"properly, to hollow or arch, i.e. (figuratively) close; to shut","kjv_def":"shut."},
"H1480":{"lemma":"גּוּפָה","xlit":"gûwphâh","pron":"goo-faw'","derivation":"from H1479 (גּוּף);","strongs_def":"a corpse (as closed to sense)","kjv_def":"body."},
"H1481":{"lemma":"גּוּר","xlit":"gûwr","pron":"goor","derivation":"a primitive root;","strongs_def":"properly, to turn aside from the road (for a lodging or any other purpose), i.e. sojourn (as a guest); also to shrink, fear (as in a strange place); also to gather forhostility (as afraid)","kjv_def":"abide, assemble, be afraid, dwell, fear, gather (together), inhabitant, remain, sojourn, stand in awe, (be) stranger, [idiom] surely."},
"H1482":{"lemma":"גּוּר","xlit":"gûwr","pron":"goor","derivation":"or (shortened) גֻּר; perhaps from H1481 (גּוּר);","strongs_def":"a cub (as still abiding in the lair), especially of the lion","kjv_def":"whelp, young one."},
"H1483":{"lemma":"גּוּר","xlit":"Gûwr","pron":"goor","derivation":"the same as H1482 (גּוּר);","strongs_def":"Gur, a place in Palestine","kjv_def":"Gur."},
"H1484":{"lemma":"גּוֹר","xlit":"gôwr","pron":"gore","derivation":"or (feminine) גֹּרָה; a variation of H1482 (גּוּר)","strongs_def":"{a cub (as still abiding in the lair), especially of the lion}","kjv_def":"whelp."},
"H1485":{"lemma":"גּוּר־בַּעַל","xlit":"Gûwr-Baʻal","pron":"goor-bah'-al","derivation":"from H1481 (גּוּר) and H1168 (בַּעַל); dwelling of Baal;","strongs_def":"Gur-Baal, a place in Arabia","kjv_def":"Gur-baal."},
"H1486":{"lemma":"גּוֹרָל","xlit":"gôwrâl","pron":"go-rawl'","derivation":"or (shortened) גֹּרָל; from an unused root meaning to be rough (as stone);","strongs_def":"properly, a pebble, i.e. a lot (small stones being used for that purpose); figuratively, a portion or destiny (as if determined by lot)","kjv_def":"lot."},
"H1487":{"lemma":"גּוּשׁ","xlit":"gûwsh","pron":"goosh","derivation":"or rather (by permutation) גִּישׁ; of uncertain derivation;","strongs_def":"a mass of earth","kjv_def":"clod."},
"H1488":{"lemma":"גֵּז","xlit":"gêz","pron":"gaze","derivation":"from H1494 (גַּזָז);","strongs_def":"a fleece (as shorn); also mown grass","kjv_def":"fleece, mowing, mown grass."},
"H1489":{"lemma":"גִּזְבָּר","xlit":"gizbâr","pron":"ghiz-bawr'","derivation":"of foreign derivation;","strongs_def":"treasurer","kjv_def":"treasurer."},
"H1490":{"lemma":"גִּזְבָּר","xlit":"gizbâr","pron":"ghiz-bawr'","derivation":"(Aramaic) corresponding to H1489 (גִּזְבָּר)","strongs_def":"{treasurer}","kjv_def":"treasurer."},
"H1491":{"lemma":"גָּזָה","xlit":"gâzâh","pron":"gaw-zaw'","derivation":"a primitive root (akin to H1468 (גּוּז));","strongs_def":"to cut off, i.e. portion out","kjv_def":"take."},
"H1492":{"lemma":"גָּזַּה","xlit":"gâzzah","pron":"gaz-zaw'","derivation":"feminine from H1494 (גַּזָז);","strongs_def":"a fleece","kjv_def":"fleece."},
"H1493":{"lemma":"גִּזוֹנִי","xlit":"Gizôwnîy","pron":"ghee-zo-nee'","derivation":"patrial from the unused name of a place apparently in Palestine;","strongs_def":"a Gizonite or inhabitant of Gizoh","kjv_def":"Gizonite."},
"H1494":{"lemma":"גַּזָז","xlit":"gazâz","pron":"gaw-zaz'","derivation":"a primitive root (akin to H1468 (גּוּז));","strongs_def":"to cut off; specifically to shear aflock or shave the hair; figuratively to destroy an enemy","kjv_def":"cut off (down), poll, shave, (sheep-) shear(-er)."},
"H1495":{"lemma":"גָּזֵז","xlit":"Gâzêz","pron":"gaw-zaze'","derivation":"from H1494 (גַּזָז); shearer;","strongs_def":"Gazez, the name of two Israelites","kjv_def":"Gazez."},
"H1496":{"lemma":"גָּזִית","xlit":"gâzîyth","pron":"gaw-zeeth'","derivation":"from H1491 (גָּזָה);","strongs_def":"something cut, i.e. dressed stone","kjv_def":"hewed, hewn stone, wrought."},
"H1497":{"lemma":"גָּזַל","xlit":"gâzal","pron":"gaw-zal'","derivation":"a primitive root;","strongs_def":"to pluck off; specifically to flay, strip or rob","kjv_def":"catch, consume, exercise (robbery), pluck (off), rob, spoil, take away (by force, violence), tear."},
"H1498":{"lemma":"גָּזֵל","xlit":"gâzêl","pron":"gaw-zale'","derivation":"from H1497 (גָּזַל);","strongs_def":"robbery, or (concretely) plunder","kjv_def":"robbery, thing taken away by violence."},
"H1499":{"lemma":"גֵּזֶל","xlit":"gêzel","pron":"ghe'-zel","derivation":"from H1497 (גָּזַל);","strongs_def":"plunder, i.e. violence","kjv_def":"violence, violent perverting.. gozal. See H1469 (גּוֹזָל)."},
"H1500":{"lemma":"גְּזֵלָה","xlit":"gᵉzêlâh","pron":"ghez-ay-law'","derivation":"feminine of H1498 (גָּזֵל) and mean the same","strongs_def":"{robbery, or (concretely) plunder}","kjv_def":"that (he had robbed) (which he took violently away), spoil, violence."},
"H1501":{"lemma":"גָּזָם","xlit":"gâzâm","pron":"gaw-zawm'","derivation":"from an unused root meaning to devour;","strongs_def":"a kind of locust","kjv_def":"palmer-worm."},
"H1502":{"lemma":"גַּזָּם","xlit":"Gazzâm","pron":"gaz-zawm'","derivation":"from the same as H1501 (גָּזָם); devourer","strongs_def":"Gazzam, one of the Nethinim","kjv_def":"Gazzam."},
"H1503":{"lemma":"גֶּזַע","xlit":"gezaʻ","pron":"geh'-zah","derivation":"from an unused root meaning to cut down (trees);","strongs_def":"the trunk or stump of atree (as felled or as planted)","kjv_def":"stem, stock."},
"H1504":{"lemma":"גָּזַר","xlit":"gâzar","pron":"gaw-zar'","derivation":"a primitive root;","strongs_def":"to cut down or off; (figuratively) to destroy, divide, exclude, or decide","kjv_def":"cut down (off), decree, divide, snatch."},
"H1505":{"lemma":"גְּזַר","xlit":"gᵉzar","pron":"ghez-ar'","derivation":"(Aramaic) corresponding to H1504 (גָּזַר);","strongs_def":"to quarry; determine","kjv_def":"cut out, soothsayer."},
"H1506":{"lemma":"גֶּזֶר","xlit":"gezer","pron":"gheh'-zer","derivation":"from H1504 (גָּזַר);","strongs_def":"something cut off; a portion","kjv_def":"part, piece."},
"H1507":{"lemma":"גֶּזֶר","xlit":"Gezer","pron":"gheh'-zer","derivation":"the same as H1506 (גֶּזֶר);","strongs_def":"Gezer, a place in Palestine","kjv_def":"Gazer, Gezer."},
"H1508":{"lemma":"גִּזְרָה","xlit":"gizrâh","pron":"ghiz-raw'","derivation":"feminine of H1506 (גֶּזֶר);","strongs_def":"the figure or person (as if cut out); also an inclosure (as separated)","kjv_def":"polishing, separate place."},
"H1509":{"lemma":"גְּזֵרָה","xlit":"gᵉzêrâh","pron":"ghez-ay-raw'","derivation":"from H1504 (גָּזַר);","strongs_def":"a desert (as separated)","kjv_def":"not inhabited."},
"H1510":{"lemma":"גְּזֵרָה","xlit":"gᵉzêrâh","pron":"ghez-ay-raw'","derivation":"(Aramaic) from H1505 (גְּזַר) (as H1504 (גָּזַר));","strongs_def":"a decree","kjv_def":"decree."},
"H1511":{"lemma":"גִּזְרִי","xlit":"Gizrîy","pron":"ghiz-ree'","derivation":"(in the m patrial from H1507 (גֶּזֶר); a Gezerite (collectively) or inhabitants of Gezer; but better (as in the text) bytransposition גִּזְרִיxlit Girzîy corrected to Gizrîy; patrial of H1630 (גְּרִזִים);","strongs_def":"a Grizite (collectively) or member of a native tribe in Palestine","kjv_def":"Gezrites."},
"H1512":{"lemma":"גָּחוֹן","xlit":"gâchôwn","pron":"gaw-khone'","derivation":"probably from H1518 (גִּיחַ)lemma גּיחַ missing vowel, corrected to גִּיחַ; (compare H1521 (גִּיחוֹן))","strongs_def":"the external abdomen, belly (as the source of the faetus )","kjv_def":"belly."},
"H1513":{"lemma":"גֶּחֶל","xlit":"gechel","pron":"geh'-khel","derivation":"or (feminine) גַּחֶלֶת; from an unused root meaning to glow or kindle;","strongs_def":"an ember","kjv_def":"(burning) coal."},
"H1514":{"lemma":"גַּחַם","xlit":"Gacham","pron":"gah'-kham","derivation":"from an unused root meaning to burn; flame;","strongs_def":"Gacham, a son of Nahor","kjv_def":"Gaham."},
"H1515":{"lemma":"גַּחַר","xlit":"Gachar","pron":"gah'-khar","derivation":"from an unused root meaning to hide; lurker;","strongs_def":"Gachar, one of the Nethinim","kjv_def":"Gahar."},
"H1516":{"lemma":"גַּיְא","xlit":"gayʼ","pron":"gah'-ee","derivation":"or (shortened) גַּי; probably (by transmutation) from the same root as H1466 (גֵּוָה) (abbreviated);","strongs_def":"a gorge (from its lofty sides; hence, narrow, but not a gully or winter-torrent)","kjv_def":"valley."},
"H1517":{"lemma":"גִּיד","xlit":"gîyd","pron":"gheed","derivation":"probably from H1464 (גּוּד);","strongs_def":"a thong (as compressing); by analogy, a tendon","kjv_def":"sinew."},
"H1518":{"lemma":"גִּיחַ","xlit":"gîyach","pron":"ghee'-akh","derivation":"or (shortened) גֹּחַ; a primitive root;","strongs_def":"to gush forth (as water), generally to issue","kjv_def":"break forth, labor to bring forth, come forth, draw up, take out."},
"H1519":{"lemma":"גִּיחַ","xlit":"gîyach","pron":"ghee'-akh","derivation":"(Aramaic) or (shortened) גּוּחַ; (Aramaic), corresponding to H1518 (גִּיחַ)lemma גּיחַ missing vowel, corrected to גִּיחַ;","strongs_def":"to rush forth","kjv_def":"strive."},
"H1520":{"lemma":"גִּיחַ","xlit":"Gîyach","pron":"ghee'-akh","derivation":"from H1518 (גִּיחַ)lemma גּיחַ missing vowel, corrected to גִּיחַ; a fountain;","strongs_def":"Giach, a place in Palestine","kjv_def":"Giah."},
"H1521":{"lemma":"גִּיחוֹן","xlit":"Gîychôwn","pron":"ghee-khone'","derivation":"or (shortened) גִּחוֹן; from H1518 (גִּיחַ)lemma גּיחַ missing vowel, corrected to גִּיחַ; stream;","strongs_def":"Gichon, a river of Paradise; also a valley (or pool) near Jerusalem","kjv_def":"Gihon."},
"H1522":{"lemma":"גֵּיחֲזִי","xlit":"Gêychăzîy","pron":"gay-khah-zee'","derivation":"or גֵּחֲזִי; apparently from H1516 (גַּיְא) and H2372 (חָזָה); valley of a visionary;","strongs_def":"Gechazi, the servant of Elisha","kjv_def":"Gehazi."},
"H1523":{"lemma":"גִּיל","xlit":"gîyl","pron":"gheel","derivation":"or (by permutation) גּוּל; a primitive root;","strongs_def":"properly, to spin round (under the influence of any violent emotion), i.e. usually rejoice, or (as cringing) fear","kjv_def":"be glad, joy, be joyful, rejoice."},
"H1524":{"lemma":"גִּיל","xlit":"gîyl","pron":"gheel","derivation":"from H1523 (גִּיל);","strongs_def":"a revolution (of time, i.e. an age); also joy","kjv_def":"[idiom] exceedingly, gladness, [idiom] greatly, joy, rejoice(-ing), sort."},
"H1525":{"lemma":"גִּילָה","xlit":"gîylâh","pron":"ghee-law'","derivation":"or גִּילַת; feminine of H1524 (גִּיל);","strongs_def":"joy","kjv_def":"joy, rejoicing."},
"H1526":{"lemma":"גִּילֹנִי","xlit":"Gîylônîy","pron":"ghee-lo-nee'","derivation":"patrial from H1542 (גִּלֹה);","strongs_def":"a Gilonite or inhabitant of Giloh","kjv_def":"Gilonite."},
"H1527":{"lemma":"גִּינַת","xlit":"Gîynath","pron":"ghee-nath'","derivation":"of uncertain derivation;","strongs_def":"Ginath, an Israelite","kjv_def":"Ginath."},
"H1528":{"lemma":"גִּיר","xlit":"gîyr","pron":"gheer","derivation":"(Aramaic) corresponding to H1615 (גִּר);","strongs_def":"lime","kjv_def":"plaster."},
"H1529":{"lemma":"גֵּישָׁן","xlit":"Gêyshân","pron":"gay-shawn'","derivation":"from the same as H1487 (גּוּשׁ); lumpish;","strongs_def":"Geshan, an Israelite","kjv_def":"Geshan."},
"H1530":{"lemma":"גַּל","xlit":"gal","pron":"gal","derivation":"from H1556 (גָּלַל);","strongs_def":"something rolled, i.e. a heap of stone or dung (plural ruins), by analogy, a spring of water (plural waves)","kjv_def":"billow, heap, spring, wave."},
"H1531":{"lemma":"גֹּל","xlit":"gôl","pron":"gole","derivation":"from H1556 (גָּלַל);","strongs_def":"a cup for oil (as round)","kjv_def":"bowl."},
"H1532":{"lemma":"גַּלָּב","xlit":"gallâb","pron":"gal-lawb'","derivation":"from an unused root meaning to shave;","strongs_def":"a barber","kjv_def":"barber."},
"H1533":{"lemma":"גִּלְבֹּעַ","xlit":"Gilbôaʻ","pron":"ghil-bo'-ah","derivation":"from H1530 (גַּל) and H1158 (בָּעָה); fountain of ebullition;","strongs_def":"Gilboa, a mountain of Palestine","kjv_def":"Gilboa."},
"H1534":{"lemma":"גַּלְגַּל","xlit":"galgal","pron":"gal-gal'","derivation":"by reduplication from H1556 (גָּלַל);","strongs_def":"a wheel; by analogy, a whirlwind; also dust (as whirled)","kjv_def":"heaven, rolling thing, wheel."},
"H1535":{"lemma":"גַּלְגַּל","xlit":"galgal","pron":"gal-gal'","derivation":"(Aramaic) corresponding to H1534 (גַּלְגַּל);","strongs_def":"a wheel","kjv_def":"wheel."},
"H1536":{"lemma":"גִּלְגָּל","xlit":"gilgâl","pron":"ghil-gawl'","derivation":"a variation of H1534 (גַּלְגַּל)","strongs_def":"{a wheel; by analogy, a whirlwind; also dust (as whirled)}","kjv_def":"wheel."},
"H1537":{"lemma":"גִּלְגָּל","xlit":"Gilgâl","pron":"ghil-gawl'","derivation":"the same as H1536 (גִּלְגָּל) (with the article as a properly, noun);","strongs_def":"Gilgal, the name of three places in Palestine","kjv_def":"Gilgal. See also H1019 (בֵּית הַגִּלְגָּל)."},
"H1538":{"lemma":"גֻּלְגֹּלֶת","xlit":"gulgôleth","pron":"gul-go'-leth","derivation":"by reduplication from H1556 (גָּלַל);","strongs_def":"a skull (as round); by implication, a head (in enumeration of persons)","kjv_def":"head, every man, poll, skull."},
"H1539":{"lemma":"גֶּלֶד","xlit":"geled","pron":"ghe'-led","derivation":"from an unused root probably meaning to polish;","strongs_def":"the (human) skin (as smooth)","kjv_def":"skin."},
"H1540":{"lemma":"גָּלָה","xlit":"gâlâh","pron":"gaw-law'","derivation":"a primitive root;","strongs_def":"to denude (especially in a disgraceful sense); by implication, to exile (captives being usually stripped); figuratively, to reveal","kjv_def":"[phrase] advertise, appear, bewray, bring, (carry, lead, go) captive (into captivity), depart, disclose, discover, exile, be gone, open, [idiom] plainly, publish, remove, reveal, [idiom] shamelessly, shew, [idiom] surely, tell, uncover."},
"H1541":{"lemma":"גְּלָה","xlit":"gᵉlâh","pron":"ghel-aw'","derivation":"(Aramaic) or גְּלָא; (Aramaic), corresponding to H1540 (גָּלָה)","strongs_def":"{to denude (especially in a disgraceful sense); by implication, to exile (captives being usually stripped); figuratively, to reveal}","kjv_def":"bring over, carry away, reveal."},
"H1542":{"lemma":"גִּלֹה","xlit":"Gilôh","pron":"ghee-lo'","derivation":"or (fully) גִּילֹה; from H1540 (גָּלָה); open;","strongs_def":"Giloh, a place in Palestine","kjv_def":"Giloh."},
"H1543":{"lemma":"גֻּלָּה","xlit":"gullâh","pron":"gool-law'","derivation":"feminine from H1556 (גָּלַל);","strongs_def":"a fountain, bowl or globe (all as round)","kjv_def":"bowl, pommel, spring."},
"H1544":{"lemma":"גִּלּוּל","xlit":"gillûwl","pron":"ghil-lool'","derivation":"or (shortened) גִּלֻּל; from H1556 (גָּלַל);","strongs_def":"properly, a log (as round); by implication, an idol","kjv_def":"idol."},
"H1545":{"lemma":"גְּלוֹם","xlit":"gᵉlôwm","pron":"ghel-ome'","derivation":"from H1563 (גָּלַם);","strongs_def":"clothing (as wrapped)","kjv_def":"clothes."},
"H1546":{"lemma":"גָּלוּת","xlit":"gâlûwth","pron":"gaw-looth'","derivation":"feminine from H1540 (גָּלָה);","strongs_def":"captivity; concretely, exiles (collectively)","kjv_def":"(they that are carried away) captives(-ity)."},
"H1547":{"lemma":"גָּלוּת","xlit":"gâlûwth","pron":"gaw-looth'","derivation":"(Aramaic) corresponding to H1546 (גָּלוּת)","strongs_def":"{captivity; concretely, exiles (collectively)}","kjv_def":"captivity."},
"H1548":{"lemma":"גָּלַח","xlit":"gâlach","pron":"gaw-lakh'","derivation":"a primitive root;","strongs_def":"properly, to be bald, i.e. (causatively) to shave; figuratively to lay waste","kjv_def":"poll, shave (off)."},
"H1549":{"lemma":"גִּלָּיוֹן","xlit":"gillâyôwn","pron":"ghil-law-yone'","derivation":"or גִּלְיוֹן; from H1540 (גָּלָה);","strongs_def":"a tablet for writing (as bare); by analogy, a mirror (as a plate)","kjv_def":"glass, roll."},
"H1550":{"lemma":"גָּלִיל","xlit":"gâlîyl","pron":"gaw-leel'","derivation":"from H1556 (גָּלַל);","strongs_def":"a valve of a folding door (as turning); also a ring (as round)","kjv_def":"folding, ring."},
"H1551":{"lemma":"גָּלִיל","xlit":"Gâlîyl","pron":"gaw-leel'","derivation":"or (prolonged) גָּלִילָה; the same as H1550 (גָּלִיל); a circle (with the article);","strongs_def":"Galil (as a special circuit) in the North of Palestine","kjv_def":"Galilee."},
"H1552":{"lemma":"גְלִילָה","xlit":"gᵉlîylâh","pron":"ghel-ee-law'","derivation":"feminine of H1550 (גָּלִיל);","strongs_def":"a circuit or region","kjv_def":"border, coast, country."},
"H1553":{"lemma":"גְּלִילוֹת","xlit":"Gᵉlîylôwth","pron":"ghel-ee-lowth'","derivation":"plural of H1552 (גְלִילָה); circles;","strongs_def":"Geliloth, a place in Palestine","kjv_def":"Geliloth."},
"H1554":{"lemma":"גַּלִּים","xlit":"Gallîym","pron":"gal-leem'","derivation":"plural of H1530 (גַּל); springs;","strongs_def":"Gallim, a place in Palestine","kjv_def":"Gallim."},
"H1555":{"lemma":"גׇּלְיַת","xlit":"Golyath","pron":"gol-yath'","derivation":"perhaps from H1540 (גָּלָה); exile;","strongs_def":"Goljath, a Philistine","kjv_def":"Goliath."},
"H1556":{"lemma":"גָּלַל","xlit":"gâlal","pron":"gaw-lal'","derivation":"a primitive root;","strongs_def":"to roll (literally or figuratively)","kjv_def":"commit, remove, roll (away, down, together), run down, seek occasion, trust, wallow."},
"H1557":{"lemma":"גָּלָל","xlit":"gâlâl","pron":"gaw-lawl'","derivation":"from H1556 (גָּלַל);","strongs_def":"dung (as in balls)","kjv_def":"dung."},
"H1558":{"lemma":"גָּלָל","xlit":"gâlâl","pron":"gaw-lawl'","derivation":"from H1556 (גָּלַל);","strongs_def":"a circumstance (as rolled around); only used adverbially, on account of","kjv_def":"because of, for (sake)."},
"H1559":{"lemma":"גָּלָל","xlit":"Gâlâl","pron":"gaw-lawl'","derivation":"from H1556 (גָּלַל), in the sense of H1560 (גְּלָל); great;","strongs_def":"Galal, the name of two Israelites","kjv_def":"Galal."},
"H1560":{"lemma":"גְּלָל","xlit":"gᵉlâl","pron":"ghel-awl'","derivation":"(Aramaic) from a root corresponding to H1556 (גָּלַל);","strongs_def":"weight or size (as if rolled)","kjv_def":"great."},
"H1561":{"lemma":"גֵּלֶל","xlit":"gêlel","pron":"gay'-lel","derivation":"a variation of H1557 (גָּלָל);","strongs_def":"dung (plural balls of dung)","kjv_def":"dung."},
"H1562":{"lemma":"גִּלֲלַי","xlit":"Gilălay","pron":"ghe-lal-ah'-ee","derivation":"from H1561 (גֵּלֶל); dungy;","strongs_def":"Gilalai, an Israelite","kjv_def":"Gilalai."},
"H1563":{"lemma":"גָּלַם","xlit":"gâlam","pron":"gaw-lam'","derivation":"a primitive root;","strongs_def":"to fold","kjv_def":"wrap together."},
"H1564":{"lemma":"גֹּלֶם","xlit":"gôlem","pron":"go'-lem","derivation":"from H1563 (גָּלַם);","strongs_def":"a wrapped (and unformed mass, i.e. as the embryo)","kjv_def":"substance yet being unperfect."},
"H1565":{"lemma":"גַּלְמוּד","xlit":"galmûwd","pron":"gal-mood'","derivation":"probably by prolonged from H1563 (גָּלַם);","strongs_def":"sterile (as wrapped up too hard); figuratively, desolate","kjv_def":"desolate, solitary."},
"H1566":{"lemma":"גָּלַע","xlit":"gâlaʻ","pron":"gaw-lah'","derivation":"a primitive root;","strongs_def":"to be obstinate","kjv_def":"(inter-) meddle (with)."},
"H1567":{"lemma":"גַּלְעֵד","xlit":"Galʻêd","pron":"gal-ade'","derivation":"from H1530 (גַּל) and H5707 (עֵד); heap of testimony;","strongs_def":"Galed, a memorial cairn East of the Jordan","kjv_def":"Galeed."},
"H1568":{"lemma":"גִּלְעָד","xlit":"Gilʻâd","pron":"ghil-awd'","derivation":"probably from H1567 (גַּלְעֵד);","strongs_def":"Gilad, a region East of the Jordan; also the name of three Israelites","kjv_def":"Gilead, Gileadite."},
"H1569":{"lemma":"גִּלְעָדִי","xlit":"Gilʻâdîy","pron":"ghil-aw-dee'","derivation":"patronymically from H1568 (גִּלְעָד);","strongs_def":"a Giladite or descendant of Gilad","kjv_def":"Gileadite."},
"H1570":{"lemma":"גָּלַשׁ","xlit":"gâlash","pron":"gaw-lash'","derivation":"a primitive root;","strongs_def":"probably to caper (as a goat)","kjv_def":"appear."},
"H1571":{"lemma":"גַּם","xlit":"gam","pron":"gam","derivation":"by contraction from an unused root meaning to gather;","strongs_def":"properly, assemblage; used only adverbially also, even, yea, though; often repeated as correl. both...and","kjv_def":"again, alike, also, (so much) as (soon), both (so)...and, but, either...or, even, for all, (in) likewise (manner), moreover, nay...neither, one, then(-refore), though, what, with, yea."},
"H1572":{"lemma":"גָּמָא","xlit":"gâmâʼ","pron":"gaw-maw'","derivation":"a primitive root (literally or figuratively) to absorb","strongs_def":"to absorb","kjv_def":"swallow, drink."},
"H1573":{"lemma":"גֹּמֶא","xlit":"gômeʼ","pron":"go'-meh","derivation":"from H1572 (גָּמָא);","strongs_def":"properly, an absorbent, i.e. the bulrush (from its porosity); specifically the papyrus","kjv_def":"(bul-) rush."},
"H1574":{"lemma":"גֹּמֶד","xlit":"gômed","pron":"go'-med","derivation":"from an unused root apparently meaning to grasp;","strongs_def":"properly, a span","kjv_def":"cubit."},
"H1575":{"lemma":"גַּמָּד","xlit":"gammâd","pron":"gam-mawd'","derivation":"from the same as H1574 (גֹּמֶד);","strongs_def":"a warrior (as grasping weapons)","kjv_def":"Gammadims."},
"H1576":{"lemma":"גְּמוּל","xlit":"gᵉmûwl","pron":"ghem-ool'","derivation":"from H1580 (גָּמַל);","strongs_def":"treatment, i.e. an act (of good or ill); by implication, service or requital","kjv_def":"[phrase] as hast served, benefit, desert, deserving, that which he hath given, recompense, reward."},
"H1577":{"lemma":"גָּמוּל","xlit":"gâmûwl","pron":"gaw-mool'","derivation":"passive participle of H1580 (גָּמַל); rewarded;","strongs_def":"Gamul, an Israelite","kjv_def":"Gamul. See also H1014 (בֵּית גָּמוּל)."},
"H1578":{"lemma":"גְּמוּלָה","xlit":"gᵉmûwlâh","pron":"ghem-oo-law'","derivation":"feminine of H1576 (גְּמוּל);","strongs_def":"meaning the same","kjv_def":"deed, recompense, such a reward."},
"H1579":{"lemma":"גִּמְזוֹ","xlit":"Gimzôw","pron":"ghim-zo'","derivation":"of uncertain derivation;","strongs_def":"Gimzo, a place in Palestine","kjv_def":"Gimzo."},
"H1580":{"lemma":"גָּמַל","xlit":"gâmal","pron":"gaw-mal'","derivation":"a primitive root;","strongs_def":"to treat a person (well or ill), i.e. benefit or requite; by implication (of toil), to ripen, i.e. (specifically) to wean","kjv_def":"bestow on, deal bountifully, do (good), recompense, requite, reward, ripen, [phrase] serve, mean, yield."},
"H1581":{"lemma":"גָּמָל","xlit":"gâmâl","pron":"gaw-mawl'","derivation":"apparently from H1580 (גָּמַל) (in the sense of labor or burden-bearing);","strongs_def":"a camel","kjv_def":"camel."},
"H1582":{"lemma":"גְּמַלִּי","xlit":"Gᵉmallîy","pron":"ghem-al-lee'","derivation":"probably from H1581 (גָּמָל); cameldriver;","strongs_def":"Gemalli, an Israelite","kjv_def":"Gemalli."},
"H1583":{"lemma":"גַּמְלִיאֵל","xlit":"Gamlîyʼêl","pron":"gam-lee-ale'","derivation":"from H1580 (גָּמַל) and H410 (אֵל); reward of God;","strongs_def":"Gamliel, an Israelite","kjv_def":"Gamaliel."},
"H1584":{"lemma":"גָּמַר","xlit":"gâmar","pron":"gaw-mar'","derivation":"a primitive root;","strongs_def":"to end (in the sense of completion or failure)","kjv_def":"cease, come to an end, fail, perfect, perform."},
"H1585":{"lemma":"גְּמַר","xlit":"gᵉmar","pron":"ghem-ar'","derivation":"(Aramaic) corresponding to H1584 (גָּמַר)","strongs_def":"{to end (in the sense of completion or failure)}","kjv_def":"perfect."},
"H1586":{"lemma":"גֹּמֶר","xlit":"Gômer","pron":"go'-mer","derivation":"from H1584 (גָּמַר); completion;","strongs_def":"Gomer, the name of a son of Japheth and of his descendants; also of a Hebrewess","kjv_def":"Gomer."},
"H1587":{"lemma":"גְּמַרְיָה","xlit":"Gᵉmaryâh","pron":"ghem-ar-yaw'","derivation":"or גְּמַרְיָהוּ; from H1584 (גָּמַר) and H3050 (יָהּ); Jah has perfected;","strongs_def":"Gemarjah, the name of two Israelites","kjv_def":"Gemariah."},
"H1588":{"lemma":"גַּן","xlit":"gan","pron":"gan","derivation":"from H1598 (גָּנַן);","strongs_def":"a garden (as fenced)","kjv_def":"garden."},
"H1589":{"lemma":"גָּנַב","xlit":"gânab","pron":"gaw-nab'","derivation":"a primitive root;","strongs_def":"to thieve (literally or figuratively); by implication, to deceive","kjv_def":"carry away, [idiom] indeed, secretly bring, steal (away), get by stealth."},
"H1590":{"lemma":"גַּנָּב","xlit":"gannâb","pron":"gaw-nab'","derivation":"from H1589 (גָּנַב);","strongs_def":"a stealer","kjv_def":"thief."},
"H1591":{"lemma":"גְּנֵבָה","xlit":"gᵉnêbâh","pron":"ghen-ay-baw'","derivation":"from H1589 (גָּנַב);","strongs_def":"stealing, i.e. (concretely) something stolen","kjv_def":"theft."},
"H1592":{"lemma":"גְּנֻבַת","xlit":"Gᵉnubath","pron":"ghen-oo-bath'","derivation":"from H1589 (גָּנַב); theft;","strongs_def":"Genubath, an Edomitish prince","kjv_def":"Genubath."},
"H1593":{"lemma":"גַּנָּה","xlit":"gannâh","pron":"gan-naw'","derivation":"feminine of H1588 (גַּן);","strongs_def":"a garden","kjv_def":"garden."},
"H1594":{"lemma":"גִּנָּה","xlit":"ginnâh","pron":"ghin-naw'","derivation":"another form for H1593 (גַּנָּה)","strongs_def":"{a garden}","kjv_def":"garden."},
"H1595":{"lemma":"גֶּנֶז","xlit":"genez","pron":"gheh'-nez","derivation":"from an unused root meaning to store;","strongs_def":"treasure; by implication, a coffer","kjv_def":"chest, treasury."},
"H1596":{"lemma":"גְּנַז","xlit":"gᵉnaz","pron":"ghen-az'","derivation":"(Aramaic) corresponding to H1595 (גֶּנֶז);","strongs_def":"treasure","kjv_def":"treasure."},
"H1597":{"lemma":"גִּנְזַךְ","xlit":"ginzak","pron":"ghin-zak'","derivation":"prolonged from H1595 (גֶּנֶז);","strongs_def":"a treasury","kjv_def":"treasury."},
"H1598":{"lemma":"גָּנַן","xlit":"gânan","pron":"gaw-nan'","derivation":"a primitive root;","strongs_def":"to hedge about, i.e. (generally) protect","kjv_def":"defend."},
"H1599":{"lemma":"גִּנְּתוֹן","xlit":"Ginnᵉthôwn","pron":"ghin-neth-one","derivation":"or גִּנְּתוֹ; from H1598 (גָּנַן); gardener;","strongs_def":"Ginnethon or Ginnetho, an Israelite","kjv_def":"Ginnetho, Ginnethon."},
"H1600":{"lemma":"גָּעָה","xlit":"gâʻâh","pron":"gaw-aw'","derivation":"a primitive root;","strongs_def":"to bellow (as cattle)","kjv_def":"low."},
"H1601":{"lemma":"גֹּעָה","xlit":"Gôʻâh","pron":"go-aw'","derivation":"feminine active participle of H1600 (גָּעָה); lowing;","strongs_def":"Goah, a place near Jerusalem","kjv_def":"Goath."},
"H1602":{"lemma":"גָּעַל","xlit":"gâʻal","pron":"gaw-al'","derivation":"a primitive root;","strongs_def":"to detest; by implication, to reject","kjv_def":"abhor, fail, lothe, vilely cast away."},
"H1603":{"lemma":"גַּעַל","xlit":"Gaʻal","pron":"gah'-al","derivation":"from H1602 (גָּעַל); loathing;","strongs_def":"Gaal, an Israelite","kjv_def":"Gaal."},
"H1604":{"lemma":"גֹּעַל","xlit":"gôʻal","pron":"go'-al","derivation":"from H1602 (גָּעַל);","strongs_def":"abhorrence","kjv_def":"loathing."},
"H1605":{"lemma":"גָּעַר","xlit":"gâʻar","pron":"gaw-ar'","derivation":"a primitive root;","strongs_def":"to chide","kjv_def":"corrupt, rebuke, reprove."},
"H1606":{"lemma":"גְּעָרָה","xlit":"gᵉʻârâh","pron":"gheh-aw-raw'","derivation":"from H1605 (גָּעַר);","strongs_def":"a chiding","kjv_def":"rebuke(-ing), reproof."},
"H1607":{"lemma":"גָּעַשׁ","xlit":"gâʻash","pron":"gaw-ash'","derivation":"a primitive root","strongs_def":"to agitate violently","kjv_def":"move, shake, toss, trouble."},
"H1608":{"lemma":"גַּעַשׁ","xlit":"Gaʻash","pron":"ga'-ash","derivation":"from H1607 (גָּעַשׁ); a quaking;","strongs_def":"Gaash, a hill in Palestine","kjv_def":"Gaash."},
"H1609":{"lemma":"גַּעְתָּם","xlit":"Gaʻtâm","pron":"gah-tawm'","derivation":"of uncertain derivation;","strongs_def":"Gatam, an Edomite","kjv_def":"Gatam."},
"H1610":{"lemma":"גַּף","xlit":"gaph","pron":"gaf","derivation":"from an unused root meaning to arch;","strongs_def":"the back; by extension the body or self","kjv_def":"[phrase] highest places, himself."},
"H1611":{"lemma":"גַּף","xlit":"gaph","pron":"gaf","derivation":"(Aramaic) corresponding to H1610 (גַּף);","strongs_def":"a wing","kjv_def":"wing."},
"H1612":{"lemma":"גֶּפֶן","xlit":"gephen","pron":"gheh'-fen","derivation":"from an unused root meaning to bend;","strongs_def":"a vine (as twining), especially the grape","kjv_def":"vine, tree."},
"H1613":{"lemma":"גֹּפֶר","xlit":"gôpher","pron":"go'-fer","derivation":"from an unused root, probably meaning to house in;","strongs_def":"a kind of tree or wood (as used for building), apparently the cypress","kjv_def":"gopher."},
"H1614":{"lemma":"גׇּפְרִית","xlit":"gophrîyth","pron":"gof-reeth'","derivation":"probably feminine of H1613 (גֹּפֶר);","strongs_def":"properly, cypress-resin; by analogy, sulphur (as equally inflammable)","kjv_def":"brimstone."},
"H1615":{"lemma":"גִּר","xlit":"gir","pron":"gheer","derivation":"perhaps from H3564 (כּוּר);","strongs_def":"lime (from being burned in a kiln)","kjv_def":"chalk(-stone)."},
"H1616":{"lemma":"גֵּר","xlit":"gêr","pron":"gare","derivation":"or (fully) geyr (gare); from H1481 (גּוּר);","strongs_def":"properly, a guest; by implication, a foreigner","kjv_def":"alien, sojourner, stranger."},
"H1617":{"lemma":"גֵּרָא","xlit":"Gêrâʼ","pron":"gay-raw'","derivation":"perhaps from H1626 (גֵּרָה); a grain;","strongs_def":"Gera, the name of six Israelites","kjv_def":"Gera."},
"H1618":{"lemma":"גָּרָב","xlit":"gârâb","pron":"gaw-rawb'","derivation":"from an unused root meaning to scratch;","strongs_def":"scurf (from itching)","kjv_def":"scab, scurvy."},
"H1619":{"lemma":"גָּרֵב","xlit":"Gârêb","pron":"gaw-rabe'","derivation":"from the same as H1618 (גָּרָב); scabby;","strongs_def":"Gareb, the name of an Israelite, also of a hill near Jerusalem","kjv_def":"Gareb."},
"H1620":{"lemma":"גַּרְגַּר","xlit":"gargar","pron":"gar-gar'","derivation":"by reduplication from H1641 (גָּרַר);","strongs_def":"a berry (as if a pellet of rumination)","kjv_def":"berry."},
"H1621":{"lemma":"גַּרְגְּרוֹת","xlit":"gargᵉrôwth","pron":"gar-gher-owth'","derivation":"feminine plural from H1641 (גָּרַר);","strongs_def":"the throat (as used in rumination)","kjv_def":"neck."},
"H1622":{"lemma":"גִּרְגָּשִׁי","xlit":"Girgâshîy","pron":"ghir-gaw-shee'","derivation":"patrial from an unused name (of uncertain derivation);","strongs_def":"a Girgashite, one of the native tribes of Canaan","kjv_def":"Girgashite, Girgasite."},
"H1623":{"lemma":"גָּרַד","xlit":"gârad","pron":"gaw-rad'","derivation":"a primitive root;","strongs_def":"to abrade","kjv_def":"scrape."},
"H1624":{"lemma":"גָּרָה","xlit":"gârâh","pron":"gaw-raw'","derivation":"a primitive root;","strongs_def":"properly, to grate, i.e. (figuratively) to anger","kjv_def":"contend, meddle, stir up, strive."},
"H1625":{"lemma":"גֵּרָה","xlit":"gêrâh","pron":"gay-raw'","derivation":"from H1641 (גָּרַר);","strongs_def":"the cud (as scraping the throat)","kjv_def":"cud."},
"H1626":{"lemma":"גֵּרָה","xlit":"gêrâh","pron":"gay-raw'","derivation":"from H1641 (גָּרַר) (as in H1625 (גֵּרָה)); properly, (like H1620 (גַּרְגַּר)) a kernel (round as if scraped), i.e.","strongs_def":"a gerah or small weight (and coin)","kjv_def":"gerah."},
"H1627":{"lemma":"גָּרוֹן","xlit":"gârôwn","pron":"gaw-rone'","derivation":"or (shortened) גָּרֹן; from H1641 (גָּרַר); (compare H1621 (גַּרְגְּרוֹת))","strongs_def":"the throat (as roughened by swallowing)","kjv_def":"[idiom] aloud, mouth, neck, throat."},
"H1628":{"lemma":"גֵּרוּת","xlit":"gêrûwth","pron":"gay-rooth'","derivation":"from H1481 (גּוּר);","strongs_def":"a (temporary) residence","kjv_def":"habitation."},
"H1629":{"lemma":"גָּרַז","xlit":"gâraz","pron":"gaw-raz'","derivation":"a primitive root;","strongs_def":"to cut off","kjv_def":"cut off."},
"H1630":{"lemma":"גְּרִזִים","xlit":"Gᵉrizîym","pron":"gher-ee-zeem'","derivation":"plural of an unused noun from H1629 (גָּרַז) (compare H1511 (גִּזְרִי)), cut up (i.e. rocky);","strongs_def":"Gerizim, a mountain of Palestine","kjv_def":"Gerizim."},
"H1631":{"lemma":"גַּרְזֶן","xlit":"garzen","pron":"gar-zen'","derivation":"from H1629 (גָּרַז);","strongs_def":"an axe","kjv_def":"ax."},
"H1632":{"lemma":"גָּרֹל","xlit":"gârôl","pron":"gaw-role'","derivation":"from the same as H1486 (גּוֹרָל);","strongs_def":"harsh","kjv_def":"man of great (as in the margin which reads H1419 (גָּדוֹל))."},
"H1633":{"lemma":"גָּרַם","xlit":"gâram","pron":"gaw-ram'","derivation":"a primitive root; to be spare or skeleton-like; used only as a denominative from H1634 (גֶּרֶם);","strongs_def":"(causative) to bone, i.e. denude (by extensive, craunch) the bones","kjv_def":"gnaw the bones, break."},
"H1634":{"lemma":"גֶּרֶם","xlit":"gerem","pron":"gheh'-rem","derivation":"from H1633 (גָּרַם);","strongs_def":"a bone (as the skeleton of the body); hence, self, i.e. (figuratively) very","kjv_def":"bone, strong, top."},
"H1635":{"lemma":"גֶּרֶם","xlit":"gerem","pron":"gheh'-rem","derivation":"(Aramaic) corresponding to H1634 (גֶּרֶם);","strongs_def":"a bone","kjv_def":"bone."},
"H1636":{"lemma":"גַּרְמִי","xlit":"Garmîy","pron":"gar-mee'","derivation":"from H1634 (גֶּרֶם);","strongs_def":"bony, i.e. strong","kjv_def":"Garmite."},
"H1637":{"lemma":"גֹּרֶן","xlit":"gôren","pron":"go'-ren","derivation":"from an unused root meaning to smooth;","strongs_def":"a threshing-floor (as made even); by analogy, any open area","kjv_def":"(barn, corn, threshing-) floor, (threshing-, void) place."},
"H1638":{"lemma":"גָּרַס","xlit":"gâraç","pron":"gaw-ras'","derivation":"a primitive root;","strongs_def":"to crush; also (intransitively and figuratively) to dissolve","kjv_def":"break."},
"H1639":{"lemma":"גָּרַע","xlit":"gâraʻ","pron":"gaw-rah'","derivation":"a primitive root;","strongs_def":"to scrape off; by implication, to shave, remove, lessen, withhold","kjv_def":"abate, clip, (di-) minish, do (take) away, keep back, restrain, make small, withdraw."},
"H1640":{"lemma":"גָּרַף","xlit":"gâraph","pron":"gaw-raf'","derivation":"a primitive root;","strongs_def":"to bear off violently","kjv_def":"sweep away."},
"H1641":{"lemma":"גָּרַר","xlit":"gârar","pron":"gaw-rar'","derivation":"a primitive root;","strongs_def":"to drag off roughly; by implication, to bring up the cud (i.e. ruminate); by analogy, to saw","kjv_def":"catch, chew, [idiom] continuing, destroy, saw."},
"H1642":{"lemma":"גְּרָר","xlit":"Gᵉrâr","pron":"gher-awr'","derivation":"probably from H1641 (גָּרַר); a rolling country;","strongs_def":"Gerar, a Philistine city","kjv_def":"Gerar."},
"H1643":{"lemma":"גֶּרֶשׂ","xlit":"geres","pron":"gheh'-res","derivation":"from an unused root meaning to husk;","strongs_def":"a kernel (collectively), i.e. grain","kjv_def":"beaten corn."},
"H1644":{"lemma":"גָּרַשׁ","xlit":"gârash","pron":"gaw-rash'","derivation":"a primitive root;","strongs_def":"to drive out from a possession; especially to expatriate or divorce","kjv_def":"cast up (out), divorced (woman), drive away (forth, out), expel, [idiom] surely put away, trouble, thrust out."},
"H1645":{"lemma":"גֶּרֶשׁ","xlit":"geresh","pron":"gheh'-resh","derivation":"from H1644 (גָּרַשׁ);","strongs_def":"produce (as if expelled)","kjv_def":"put forth."},
"H1646":{"lemma":"גְרֻשָׁה","xlit":"gᵉrushâh","pron":"gher-oo-shaw'","derivation":"feminine passive participle of H1644 (גָּרַשׁ);","strongs_def":"(abstractly) dispossession","kjv_def":"exaction."},
"H1647":{"lemma":"גֵּרְשֹׁם","xlit":"Gêrᵉshôm","pron":"gay-resh-ome'","derivation":"for H1648 (גֵּרְשׁוֹן);","strongs_def":"Gereshom, the name of four Israelites","kjv_def":"Gershom."},
"H1648":{"lemma":"גֵּרְשׁוֹן","xlit":"Gêrᵉshôwn","pron":"gay-resh-one'","derivation":"or גֵּרְשׁוֹם; from H1644 (גָּרַשׁ); a refugee;","strongs_def":"Gereshon or Gereshom, an Israelite","kjv_def":"Gershon, Gershom."},
"H1649":{"lemma":"גֵּרְשֻׁנִּי","xlit":"Gêrᵉshunnîy","pron":"gay-resh-oon-nee'","derivation":"patronymically from H1648 (גֵּרְשׁוֹן);","strongs_def":"a Gereshonite or descendant of Gereshon","kjv_def":"Gershonite, sons of Gershon."},
"H1650":{"lemma":"גְּשׁוּר","xlit":"Gᵉshûwr","pron":"ghesh-oor'","derivation":"from an unused root (meaning to join); bridge;","strongs_def":"Geshur, a district of Syria","kjv_def":"Geshur, Geshurite."},
"H1651":{"lemma":"גְּשׁוּרִי","xlit":"Gᵉshûwrîy","pron":"ghe-shoo-ree'","derivation":"patrial from H1650 (גְּשׁוּר);","strongs_def":"a Geshurite (also collectively) or inhabitants of Geshur","kjv_def":"Geshuri, Geshurites."},
"H1652":{"lemma":"גָּשַׁם","xlit":"gâsham","pron":"gaw-sham'","derivation":"a primitive root;","strongs_def":"to shower violently","kjv_def":"(cause to) rain."},
"H1653":{"lemma":"גֶּשֶׁם","xlit":"geshem","pron":"gheh'-shem","derivation":"from H1652 (גָּשַׁם);","strongs_def":"a shower","kjv_def":"rain, shower."},
"H1654":{"lemma":"גֶּשֶׁם","xlit":"Geshem","pron":"gheh'-shem","derivation":"or (prolonged) גַּשְׁמוּ; the same as H1653 (גֶּשֶׁם);","strongs_def":"Geshem or Gashmu, an Arabian","kjv_def":"Geshem, Gashmu."},
"H1655":{"lemma":"גֶּשֶׁם","xlit":"geshem","pron":"gheh'-shem","derivation":"(Aramaic) apparently the same as H1653 (גֶּשֶׁם);","strongs_def":"used in a peculiar sense, the body (probably for the (figuratively) idea of a hard rain)","kjv_def":"body."},
"H1656":{"lemma":"גֹּשֶׁם","xlit":"gôshem","pron":"go'-shem","derivation":"from H1652 (גָּשַׁם); equivalent to H1653 (גֶּשֶׁם)","strongs_def":"{a shower}","kjv_def":"rained upon."},
"H1657":{"lemma":"גֹּשֶׁן","xlit":"Gôshen","pron":"go'-shen","derivation":"probably of Egyptian origin;","strongs_def":"Goshen, the residence of the Israelites in Egypt; also a place in Palestine","kjv_def":"Goshen."},
"H1658":{"lemma":"גִּשְׁפָּא","xlit":"Gishpâʼ","pron":"ghish-paw'","derivation":"of uncertain derivation;","strongs_def":"Gishpa, an Israelite","kjv_def":"Gispa."},
"H1659":{"lemma":"גָּשַׁשׁ","xlit":"gâshash","pron":"gaw-shash'","derivation":"a primitive root; apparently","strongs_def":"to feel about","kjv_def":"grope."},
"H1660":{"lemma":"גַּת","xlit":"gath","pron":"gath","derivation":"probably from H5059 (נָגַן) (in the sense of treading out grapes);","strongs_def":"a wine-press (or vat for holding the grapes in pressing them)","kjv_def":"(wine-) press (fat)."},
"H1661":{"lemma":"גַּת","xlit":"Gath","pron":"gath","derivation":"the same as H1660 (גַּת);","strongs_def":"Gath, a Philistine city","kjv_def":"Gath."},
"H1662":{"lemma":"גַּת־הַחֵפֶר","xlit":"Gath-ha-Chêpher","pron":"gath-hah-khay'-fer","derivation":"or (abridged) גִּתָּה־חֵפֶר; from H1660 (גַּת) and H2658 (חָפַר) with the article inserted; wine-press of (the) well;","strongs_def":"Gath-Chepher, a place in Palestine","kjv_def":"Gath-kephr, Gittah-kephr."},
"H1663":{"lemma":"גִּתִּי","xlit":"Gittîy","pron":"ghit-tee'","derivation":"patrial from H1661 (גַּת);","strongs_def":"a Gittite or inhabitant of Gath","kjv_def":"Gittite."},
"H1664":{"lemma":"גִּתַּיִם","xlit":"Gittayim","pron":"ghit-tah'-yim","derivation":"dual of H1660 (גַּת); double winepress;","strongs_def":"Gittajim, a place in Palestine","kjv_def":"Gittaim."},
"H1665":{"lemma":"גִּתִּית","xlit":"Gittîyth","pron":"ghit-teeth'","derivation":"feminine of H1663 (גִּתִּי);","strongs_def":"a Gittite harp","kjv_def":"Gittith."},
"H1666":{"lemma":"גֶּתֶר","xlit":"Gether","pron":"gheh'-ther","derivation":"of uncertain derivation;","strongs_def":"Gether, a son of Aram, and the region settled by him","kjv_def":"Gether."},
"H1667":{"lemma":"גַּת־רִמּוֹן","xlit":"Gath-Rimmôwn","pron":"gath-rim-mone'","derivation":"from H1660 (גַּת) and H7416 (רִמּוֹן); wine-press of (the) pomegranate;","strongs_def":"Gath-Rimmon, a place in Palestine","kjv_def":"Gath-rimmon. d"},
"H1668":{"lemma":"דָּא","xlit":"dâʼ","pron":"daw","derivation":"(Aramaic) corresponding to H2088 (זֶה);","strongs_def":"this","kjv_def":"one..another, this."},
"H1669":{"lemma":"דָּאַב","xlit":"dâʼab","pron":"daw-ab'","derivation":"a primitive root;","strongs_def":"to pine","kjv_def":"mourn, sorrow(-ful)."},
"H1670":{"lemma":"דְּאָבָה","xlit":"dᵉʼâbâh","pron":"deh-aw-baw'","derivation":"from H1669 (דָּאַב);","strongs_def":"properly, pining; by analogy, fear","kjv_def":"sorrow."},
"H1671":{"lemma":"דְּאָבוֹן","xlit":"dᵉʼâbôwn","pron":"deh-aw-bone'","derivation":"from H1669 (דָּאַב);","strongs_def":"pining","kjv_def":"sorrow."},
"H1672":{"lemma":"דָּאַג","xlit":"dâʼag","pron":"daw-ag'","derivation":"a primitive root;","strongs_def":"be anxious","kjv_def":"be afraid (careful, sorry), sorrow, take thought."},
"H1673":{"lemma":"דֹּאֵג","xlit":"Dôʼêg","pron":"do-ayg'","derivation":"or (fully) דּוֹאֵג; active participle of H1672 (דָּאַג); anxious;","strongs_def":"Doeg, an Edomite","kjv_def":"Doeg."},
"H1674":{"lemma":"דְּאָגָה","xlit":"dᵉʼâgâh","pron":"deh-aw-gaw'","derivation":"from H1672 (דָּאַג);","strongs_def":"anxiety","kjv_def":"care(-fulness), fear, heaviness, sorrow."},
"H1675":{"lemma":"דָּאָה","xlit":"dâʼâh","pron":"daw-aw'","derivation":"a primitive root;","strongs_def":"to dart, i.e. fly rapidly","kjv_def":"fly."},
"H1676":{"lemma":"דָּאָה","xlit":"dâʼâh","pron":"daw-aw'","derivation":"from H1675 (דָּאָה);","strongs_def":"the kite (from its rapid flight)","kjv_def":"vulture. See H7201 (רָאָה)."},
"H1677":{"lemma":"דֹּב","xlit":"dôb","pron":"dobe","derivation":"or (fully) דּוֹב; from H1680 (דָּבַב);","strongs_def":"the bear (as slow)","kjv_def":"bear."},
"H1678":{"lemma":"דֹּב","xlit":"dôb","pron":"dobe","derivation":"(Aramaic) corresponding to H1677 (דֹּב)","strongs_def":"{the bear (as slow)}","kjv_def":"bear."},
"H1679":{"lemma":"דֹּבֶא","xlit":"dôbeʼ","pron":"do'-beh","derivation":"from an unused root (compare H1680 (דָּבַב)) (probably meaning to be sluggish, i.e. restful);","strongs_def":"quiet","kjv_def":"strength."},
"H1680":{"lemma":"דָּבַב","xlit":"dâbab","pron":"daw-bab'","derivation":"a primitive root (compare H1679 (דֹּבֶא));","strongs_def":"to move slowly, i.e. glide","kjv_def":"cause to speak."},
"H1681":{"lemma":"דִּבָּה","xlit":"dibbâh","pron":"dib-baw'","derivation":"from H1680 (דָּבַב) (in the sense of furtive motion);","strongs_def":"slander","kjv_def":"defaming, evil report, infamy, slander."},
"H1682":{"lemma":"דְּבוֹרָה","xlit":"dᵉbôwrâh","pron":"deb-o-raw'","derivation":"or (shortened) דְּבֹרָה; from H1696 (דָבַר) (in the sense of orderly motion);","strongs_def":"the bee (from its systematic instincts)","kjv_def":"bee."},
"H1683":{"lemma":"דְּבּוֹרָה","xlit":"Dᵉbôwrâh","pron":"deb-o-raw'","derivation":"or (shortened) דְּבֹרָה; the same as H1682 (דְּבוֹרָה);","strongs_def":"Deborah, the name of two Hebrewesses","kjv_def":"Deborah."},
"H1684":{"lemma":"דְּבַח","xlit":"dᵉbach","pron":"deb-akh'","derivation":"(Aramaic) corresponding to H2076 (זָבַח);","strongs_def":"to sacrifice (an animal)","kjv_def":"offer (sacrifice)."},
"H1685":{"lemma":"דְּבַח","xlit":"dᵉbach","pron":"deb-akh'","derivation":"(Aramaic) from H1684 (דְּבַח);","strongs_def":"a sacrifice","kjv_def":"sacrifice."},
"H1686":{"lemma":"דִּבְיוֹן","xlit":"dibyôwn","pron":"dib-yone'","derivation":"in the margin for the textual reading, חֶרְיוֹן; both, (in the plural only and) of uncertain derivation,","strongs_def":"probably some cheap vegetable, perhaps a bulbous root","kjv_def":"dove's dung."},
"H1687":{"lemma":"דְּבִיר","xlit":"dᵉbîyr","pron":"deb-eer'","derivation":"or (shortened) דְּבִר; from H1696 (דָבַר) (apparently in the sense of oracle);","strongs_def":"the shrine or innermost part of the sanctuary","kjv_def":"oracle."},
"H1688":{"lemma":"דְּבִיר","xlit":"Dᵉbîyr","pron":"deb-eer'","derivation":"or (shortened) דְּבִר; (Joshua 13:26 (but see H3810 (לֹא דְבַר))), the same as H1687 (דְּבִיר);","strongs_def":"Debir, the name of an Amoritish king and of two places in Palestine","kjv_def":"Debir."},
"H1689":{"lemma":"דִּבְלָה","xlit":"Diblâh","pron":"dib-law'","derivation":"probably an orthographical error for H7247 (רִבְלָה);","strongs_def":"Diblah, a place in Syria","kjv_def":"Diblath."},
"H1690":{"lemma":"דְּבֵלָה","xlit":"dᵉbêlâh","pron":"deb-ay-law'","derivation":"from an unused root (akin to H2082 (זָבַל)) probably meaning to press together;","strongs_def":"a cake of pressed figs","kjv_def":"cake (lump) of figs."},
"H1691":{"lemma":"דִּבְלַיִם","xlit":"Diblayim","pron":"dib-lah'-yim","derivation":"dual from the masculine of H1690 (דְּבֵלָה); two cakes;","strongs_def":"Diblajim, a symbolic name","kjv_def":"Diblaim."},
"H1692":{"lemma":"דָּבַק","xlit":"dâbaq","pron":"daw-bak'","derivation":"a primitive root;","strongs_def":"properly, to impinge, i.e. cling or adhere; figuratively, to catch by pursuit","kjv_def":"abide fast, cleave (fast together), follow close (hard after), be joined (together), keep (fast), overtake, pursue hard, stick, take."},
"H1693":{"lemma":"דְּבַק","xlit":"dᵉbaq","pron":"deb-ak'","derivation":"(Aramaic) corresponding to H1692 (דָּבַק);","strongs_def":"to stick to","kjv_def":"cleave."},
"H1694":{"lemma":"דֶּבֶק","xlit":"debeq","pron":"deh'-bek","derivation":"from H1692 (דָּבַק);","strongs_def":"a joint; by implication, solder","kjv_def":"joint, solder."},
"H1695":{"lemma":"דָּבֵק","xlit":"dâbêq","pron":"daw-bake'","derivation":"from H1692 (דָּבַק);","strongs_def":"adhering","kjv_def":"cleave, joining, stick closer."},
"H1696":{"lemma":"דָבַר","xlit":"dâbar","pron":"daw-bar'","derivation":"a primitive root;","strongs_def":"perhaps properly, to arrange; but used figuratively (of words), to speak; rarely (in a destructive sense) to subdue","kjv_def":"answer, appoint, bid, command, commune, declare, destroy, give, name, promise, pronounce, rehearse, say, speak, be spokesman, subdue, talk, teach, tell, think, use (entreaties), utter, [idiom] well, [idiom] work."},
"H1697":{"lemma":"דָּבָר","xlit":"dâbâr","pron":"daw-baw'","derivation":"from H1696 (דָבַר);","strongs_def":"a word; by implication, a matter (as spoken of) or thing; adverbially, a cause","kjv_def":"act, advice, affair, answer, [idiom] any such (thing), because of, book, business, care, case, cause, certain rate, [phrase] chronicles, commandment, [idiom] commune(-ication), [phrase] concern(-ing), [phrase] confer, counsel, [phrase] dearth, decree, deed, [idiom] disease, due, duty, effect, [phrase] eloquent, errand, (evil favoured-) ness, [phrase] glory, [phrase] harm, hurt, [phrase] iniquity, [phrase] judgment, language, [phrase] lying, manner, matter, message, (no) thing, oracle, [idiom] ought, [idiom] parts, [phrase] pertaining, [phrase] please, portion, [phrase] power, promise, provision, purpose, question, rate, reason, report, request, [idiom] (as hast) said, sake, saying, sentence, [phrase] sign, [phrase] so, some (uncleanness), somewhat to say, [phrase] song, speech, [idiom] spoken, talk, task, [phrase] that, [idiom] there done, thing (concerning), thought, [phrase] thus, tidings, what(-soever), [phrase] wherewith, which, word, work."},
"H1698":{"lemma":"דֶּבֶר","xlit":"deber","pron":"deh'-ber","derivation":"from H1696 (דָבַר) (in the sense of destroying);","strongs_def":"a pestilence","kjv_def":"murrain, pestilence, plague."},
"H1699":{"lemma":"דֹּבֶר","xlit":"dôber","pron":"do'-ber","derivation":"from H1696 (דָבַר) (in its original sense);","strongs_def":"a pasture (from its arrangement of the flock)","kjv_def":"fold, manner."},
"H1700":{"lemma":"דִּבְרָה","xlit":"dibrâh","pron":"dib-raw'","derivation":"feminine of H1697 (דָּבָר);","strongs_def":"a reason, suit or style","kjv_def":"cause, end, estate, order, regard."},
"H1701":{"lemma":"דִּבְרָה","xlit":"dibrâh","pron":"dib-raw'","derivation":"(Aramaic) corresponding to H1700 (דִּבְרָה)","strongs_def":"{a reason, suit or style}","kjv_def":"intent, sake."},
"H1702":{"lemma":"דֹּבְרָה","xlit":"dôbᵉrâh","pron":"do-ber-aw'","derivation":"feminine active participle of H1696 (דָבַר) in the sense of driving (compare H1699 (דֹּבֶר));","strongs_def":"a raft","kjv_def":"float."},
"H1703":{"lemma":"דַּבָּרָה","xlit":"dabbârâh","pron":"dab-baw-raw'","derivation":"intensive from H1696 (דָבַר);","strongs_def":"a word","kjv_def":"word."},
"H1704":{"lemma":"דִּבְרִי","xlit":"Dibrîy","pron":"dib-ree'","derivation":"from H1697 (דָּבָר); wordy;","strongs_def":"Dibri, an Israelite","kjv_def":"Dibri."},
"H1705":{"lemma":"דֲּבְרַת","xlit":"Dăbrath","pron":"daw-ber-ath'","derivation":"from H1697 (דָּבָר) (perhaps in the sense of H1699 (דֹּבֶר));","strongs_def":"Daberath, a place in Palestine","kjv_def":"Dabareh, Daberath."},
"H1706":{"lemma":"דְּבַשׁ","xlit":"dᵉbash","pron":"deb-ash'","derivation":"from an unused root meaning to be gummy;","strongs_def":"honey (from its stickiness); by analogy, syrup","kjv_def":"honey(-comb)."},
"H1707":{"lemma":"דַּבֶּשֶׁת","xlit":"dabbesheth","pron":"dab-beh'-sheth","derivation":"intensive from the same as H1706 (דְּבַשׁ);","strongs_def":"a sticky mass, i.e. the hump of acamel","kjv_def":"hunch (of a camel)."},
"H1708":{"lemma":"דַּבֶּשֶׁת","xlit":"Dabbesheth","pron":"dab-beh'-sheth","derivation":"the same as H1707 (דַּבֶּשֶׁת);","strongs_def":"Dabbesheth, a place in Palestine","kjv_def":"Dabbesheth."},
"H1709":{"lemma":"דָּג","xlit":"dâg","pron":"dawg","derivation":"or (fully) דָּאג; (Nehemiah 13:16), from H1711 (דָּגָה); a fish (as prolific); or perhaps rather from H1672 (דָּאַג) (as timid); but still better from H1672 (דָּאַג) (in the sense of squirming, i.e. moving by the vibratory action of the tail);","strongs_def":"a fish (often used collectively)","kjv_def":"fish."},
"H1710":{"lemma":"דָּגָה","xlit":"dâgâh","pron":"daw-gaw'","derivation":"feminine of H1709 (דָּג), and meaning the same","strongs_def":"{a fish (often used collectively)}","kjv_def":"fish."},
"H1711":{"lemma":"דָּגָה","xlit":"dâgâh","pron":"daw-gaw'","derivation":"a primitive root; to move rapidly; used only as a denominative from H1709 (דָּג);","strongs_def":"to spawn, i.e. become numerous","kjv_def":"grow."},
"H1712":{"lemma":"דָּגוֹן","xlit":"Dâgôwn","pron":"daw-gohn'","derivation":"from H1709 (דָּג); the fish-god;","strongs_def":"Dagon, a Philistine deity","kjv_def":"Dagon."},
"H1713":{"lemma":"דָּגַל","xlit":"dâgal","pron":"daw-gal'","derivation":"a primitive root;","strongs_def":"to flaunt, i.e. raise a flag; figuratively, to be conspicuous","kjv_def":"(set up, with) banners, chiefest."},
"H1714":{"lemma":"דֶּגֶל","xlit":"degel","pron":"deh'-gel","derivation":"from H1713 (דָּגַל);","strongs_def":"a flag","kjv_def":"banner, standard."},
"H1715":{"lemma":"דָּגָן","xlit":"dâgân","pron":"daw-gawn'","derivation":"from H1711 (דָּגָה);","strongs_def":"properly, increase, i.e. grain","kjv_def":"corn (floor), wheat."},
"H1716":{"lemma":"דָּגַר","xlit":"dâgar","pron":"daw-gar'","derivation":"a primitive root,","strongs_def":"to brood over eggs or young","kjv_def":"gather, sit."},
"H1717":{"lemma":"דַּד","xlit":"dad","pron":"dad","derivation":"apparently from the same as H1730 (דּוֹד);","strongs_def":"the breast (as the seat of love, or from its shape)","kjv_def":"breast, teat."},
"H1718":{"lemma":"דָּדָה","xlit":"dâdâh","pron":"daw-daw'","derivation":"a doubtful root;","strongs_def":"to walk gently","kjv_def":"go (softly, with)."},
"H1719":{"lemma":"דְּדָן","xlit":"Dᵉdân","pron":"ded-awn'","derivation":"or (prolonged) דְּדָנֶה; (Ezekiel 25:13), of uncertain derivation;","strongs_def":"Dedan, the name of two Cushites and of their territory","kjv_def":"Dedan."},
"H1720":{"lemma":"דְּדָנִים","xlit":"Dᵉdânîym","pron":"ded-aw-neem'","derivation":"plural of H1719 (דְּדָן) (as patrial);","strongs_def":"Dedanites, the descendants or inhabitants of Dedan","kjv_def":"Dedanim."},
"H1721":{"lemma":"דֹּדָנִים","xlit":"Dôdânîym","pron":"do-daw-neem'","derivation":"or (by orthographical error) רֹדָנִים (1 Chronicles 1:7); a plural of uncertain derivation;","strongs_def":"Dodanites, or descendants of a son of Javan","kjv_def":"Dodanim."},
"H1722":{"lemma":"דְּהַב","xlit":"dᵉhab","pron":"deh-hab'","derivation":"(Aramaic) corresponding to H2091 (זָהָב);","strongs_def":"gold","kjv_def":"gold(-en)."},
"H1723":{"lemma":"דַּהֲוָא","xlit":"Dahăvâʼ","pron":"dah-hav-aw'","derivation":"(Aramaic) of uncertain derivation;","strongs_def":"Dahava, a people colonized in Samaria","kjv_def":"Dehavites."},
"H1724":{"lemma":"דָּהַם","xlit":"dâham","pron":"daw-ham'","derivation":"a primitive root (compare H1740 (דּוּחַ));","strongs_def":"to be dumb, i.e. (figuratively) dumbfounded","kjv_def":"astonished."},
"H1725":{"lemma":"דָּהַר","xlit":"dâhar","pron":"daw-har'","derivation":"a primitive root;","strongs_def":"to curvet or move irregularly","kjv_def":"pause."},
"H1726":{"lemma":"דַּהֲהַר","xlit":"dahăhar","pron":"dah-hah-har'","derivation":"by reduplication from H1725 (דָּהַר);","strongs_def":"a gallop","kjv_def":"pransing."},
"H1727":{"lemma":"דּוּב","xlit":"dûwb","pron":"doob","derivation":"a primitive root;","strongs_def":"to mope, i.e. (figuratively) pine","kjv_def":"sorrow."},
"H1728":{"lemma":"דַּוָּג","xlit":"davvâg","pron":"dav-vawg'","derivation":"an orthographical variation of H1709 (דָּג) as a denominative (H1771 (דַּיָּג));","strongs_def":"a fisherman","kjv_def":"fisher."},
"H1729":{"lemma":"דּוּגָה","xlit":"dûwgâh","pron":"doo-gaw'","derivation":"feminine from the same as H1728 (דַּוָּג);","strongs_def":"properly, fishery, i.e. a hook forfishing","kjv_def":"fish (hook)."},
"H1730":{"lemma":"דּוֹד","xlit":"dôwd","pron":"dode","derivation":"or (shortened) דֹּד; from an unused root meaning properly, to boil, i.e.","strongs_def":"(figuratively) to love; by implication, a love-token, lover, friend; specifically an uncle","kjv_def":"(well-) beloved, father's brother, love, uncle."},
"H1731":{"lemma":"דּוּד","xlit":"dûwd","pron":"dood","derivation":"from the same as H1730 (דּוֹד);","strongs_def":"a pot (for boiling); also (by resemblance of shape) a basket","kjv_def":"basket, caldron, kettle, (seething) pot."},
"H1732":{"lemma":"דָּוִד","xlit":"Dâvid","pron":"daw-veed'","derivation":"rarely (fully); דָּוִיד; from the same as H1730 (דּוֹד); loving;","strongs_def":"David, the youngest son of Jesse","kjv_def":"David."},
"H1733":{"lemma":"דּוֹדָה","xlit":"dôwdâh","pron":"do-daw'","derivation":"feminine of H1730 (דּוֹד);","strongs_def":"an aunt","kjv_def":"aunt, father's sister, uncle's wife."},
"H1734":{"lemma":"דּוֹדוֹ","xlit":"Dôwdôw","pron":"do-do'","derivation":"from H1730 (דּוֹד); loving;","strongs_def":"Dodo, the name of three Israelites","kjv_def":"Dodo."},
"H1735":{"lemma":"דּוֹדָוָהוּ","xlit":"Dôwdâvâhûw","pron":"do-daw-vaw'-hoo","derivation":"from H1730 (דּוֹד) and H3050 (יָהּ); love of Jah;","strongs_def":"Dodavah, an Israelite","kjv_def":"Dodavah."},
"H1736":{"lemma":"דּוּדַי","xlit":"dûwday","pron":"doo-dah'-ee","derivation":"from H1731 (דּוּד);","strongs_def":"a boiler or basket; also the mandrake (as an aphrodisiac)","kjv_def":"basket, mandrake."},
"H1737":{"lemma":"דּוֹדַי","xlit":"Dôwday","pron":"do-dah'ee","derivation":"formed like H1736 (דּוּדַי); amatory;","strongs_def":"Dodai, an Israelite","kjv_def":"Dodai."},
"H1738":{"lemma":"דָּוָה","xlit":"dâvâh","pron":"daw-vaw'","derivation":"a primitive root;","strongs_def":"to be sick (as if in menstruation)","kjv_def":"infirmity."},
"H1739":{"lemma":"דָּוֶה","xlit":"dâveh","pron":"daw-veh'","derivation":"from H1738 (דָּוָה);","strongs_def":"sick (especially in menstruation)","kjv_def":"faint, menstruous cloth, she that is sick, having sickness."},
"H1740":{"lemma":"דּוּחַ","xlit":"dûwach","pron":"doo'-akh","derivation":"a primitive root;","strongs_def":"to thrust away; figuratively, to cleanse","kjv_def":"cast out, purge, wash."},
"H1741":{"lemma":"דְּוַי","xlit":"dᵉvay","pron":"dev-ah'ee","derivation":"from H1739 (דָּוֶה);","strongs_def":"sickness; figuratively, loathing","kjv_def":"languishing, sorrowful."},
"H1742":{"lemma":"דַּוָּי","xlit":"davvây","pron":"dav-voy'","derivation":"from H1739 (דָּוֶה);","strongs_def":"sick; figuratively, troubled","kjv_def":"faint."},
"H1743":{"lemma":"דּוּךְ","xlit":"dûwk","pron":"dook","derivation":"a primitive root;","strongs_def":"to bruise in a mortar","kjv_def":"beat."},
"H1744":{"lemma":"דּוּכִיפַת","xlit":"dûwkîyphath","pron":"doo-kee-fath'","derivation":"of uncertain derivation;","strongs_def":"the hoopoe or else the grouse","kjv_def":"lapwing."},
"H1745":{"lemma":"דּוּמָה","xlit":"dûwmâh","pron":"doo-maw'","derivation":"from an unused root meaning to be dumb (compare H1820 (דָּמָה));","strongs_def":"silence; figuratively, death","kjv_def":"silence."},
"H1746":{"lemma":"דּוּמָה","xlit":"Dûwmâh","pron":"doo-maw'","derivation":"the same as H1745 (דּוּמָה);","strongs_def":"Dumah, a tribe and region of Arabia","kjv_def":"Dumah."},
"H1747":{"lemma":"דּוּמִיָּה","xlit":"dûwmîyâh","pron":"doo-me-yaw'","derivation":"from H1820 (דָּמָה);","strongs_def":"stillness; adverbially, silently; abstractly quiet, trust","kjv_def":"silence, silent, waiteth."},
"H1748":{"lemma":"דּוּמָם","xlit":"dûwmâm","pron":"doo-mawm'","derivation":"from H1826 (דָּמַם);","strongs_def":"still; adverbially, silently","kjv_def":"dumb, silent, quietly wait."},
"H1749":{"lemma":"דּוֹנַג","xlit":"dôwnag","pron":"do-nag'","derivation":"of uncertain derivation;","strongs_def":"wax;","kjv_def":"wax."},
"H1750":{"lemma":"דּוּץ","xlit":"dûwts","pron":"doots","derivation":"a primitive root;","strongs_def":"to leap","kjv_def":"be turned."},
"H1751":{"lemma":"דּוּק","xlit":"dûwq","pron":"dook","derivation":"(Aramaic) corresponding to H1854 (דָּקַק);","strongs_def":"to crumble","kjv_def":"be broken to pieces."},
"H1752":{"lemma":"דּוּר","xlit":"dûwr","pron":"dure","derivation":"a primitive root;","strongs_def":"properly, to gyrate (or move in a circle), i.e. to remain","kjv_def":"dwell."},
"H1753":{"lemma":"דּוּר","xlit":"dûwr","pron":"dure","derivation":"(Aramaic) corresponding to H1752 (דּוּר);","strongs_def":"to reside","kjv_def":"dwell."},
"H1754":{"lemma":"דּוּר","xlit":"dûwr","pron":"dure","derivation":"from H1752 (דּוּר);","strongs_def":"a circle, ball or pile","kjv_def":"ball, turn, round about."},
"H1755":{"lemma":"דּוֹר","xlit":"dôwr","pron":"dore","derivation":"or (shortened) דֹּר; from H1752 (דּוּר);","strongs_def":"properly, a revolution of time, i.e. an age or generation; also a dwelling","kjv_def":"age, [idiom] evermore, generation, (n-) ever, posterity."},
"H1756":{"lemma":"דּוֹר","xlit":"Dôwr","pron":"dore","derivation":"or (by permutation) דֹּאר; (Joshua 17:11; 1 Kings 4:11), from H1755 (דּוֹר); dwelling;","strongs_def":"Dor, a place in Palestine","kjv_def":"Dor."},
"H1757":{"lemma":"דּוּרָא","xlit":"Dûwrâʼ","pron":"doo-raw'","derivation":"(Aramaic) probably from H1753 (דּוּר); circle or dwelling;","strongs_def":"Dura, a place in Babylonia","kjv_def":"Dura."},
"H1758":{"lemma":"דּוּשׁ","xlit":"dûwsh","pron":"doosh","derivation":"or דּוֹשׁ; or דִּישׁ; a primitive root;","strongs_def":"to trample or thresh","kjv_def":"break, tear, thresh, tread out (down), at grass (Jeremiah 50:11, by mistake for H1877 (דֶּשֶׁא))."},
"H1759":{"lemma":"דּוּשׁ","xlit":"dûwsh","pron":"doosh","derivation":"(Aramaic) corresponding to H1758 (דּוּשׁ);","strongs_def":"to trample","kjv_def":"tread down."},
"H1760":{"lemma":"דָּחָה","xlit":"dâchâh","pron":"daw-khaw'","derivation":"or דָּחַח; (Jeremiah 23:12), a primitive root;","strongs_def":"to push down","kjv_def":"chase, drive away (on), overthrow, outcast, [idiom] sore, thrust, totter."},
"H1761":{"lemma":"דַּחֲוָה","xlit":"dachăvâh","pron":"dakh-av-aw'","derivation":"(Aramaic) from the equivalent of H1760 (דָּחָה);","strongs_def":"probably a musical instrument (as being struck)","kjv_def":"instrument of music."},
"H1762":{"lemma":"דְּחִי","xlit":"dᵉchîy","pron":"deh-khee'","derivation":"from H1760 (דָּחָה);","strongs_def":"a push, i.e. (by implication) a fall","kjv_def":"falling."},
"H1763":{"lemma":"דְּחַל","xlit":"dᵉchal","pron":"deh-khal'","derivation":"(Aramaic) corresponding to H2119 (זָחַל);","strongs_def":"to slink, i.e. (by implication) to fear, or (causatively) be formidable","kjv_def":"make afraid, dreadful, fear, terrible."},
"H1764":{"lemma":"דֹּחַן","xlit":"dôchan","pron":"do'-khan","derivation":"of uncertain derivation;","strongs_def":"millet","kjv_def":"millet."},
"H1765":{"lemma":"דָּחַף","xlit":"dâchaph","pron":"daw-khaf'","derivation":"a primitive root;","strongs_def":"to urge, i.e. hasten","kjv_def":"(be) haste(-ned), pressed on."},
"H1766":{"lemma":"דָּחַק","xlit":"dâchaq","pron":"daw-khak'","derivation":"a primitive root;","strongs_def":"to press, i.e. oppress","kjv_def":"thrust, vex."},
"H1767":{"lemma":"דַּי","xlit":"day","pron":"dahee","derivation":"of uncertain derivation;","strongs_def":"enough (as noun or adverb), used chiefly with preposition in phrases","kjv_def":"able, according to, after (ability), among, as (oft as), (more than) enough, from, in, since, (much as is) sufficient(-ly), too much, very, when."},
"H1768":{"lemma":"דִּי","xlit":"dîy","pron":"dee","derivation":"(Aramaic) apparently for H1668 (דָּא);","strongs_def":"that, used as relative conjunction, and especially (with a preposition) in adverbial phrases; also as preposition of","kjv_def":"[idiom] as, but, for(-asmuch [phrase]), [phrase] now, of, seeing, than, that, therefore, until, [phrase] what (-soever), when, which, whom, whose."},
"H1769":{"lemma":"דִּיבוֹן","xlit":"Dîybôwn","pron":"dee-bome'","derivation":"or (shortened) דִּיבֹן; from H1727 (דּוּב); pining","strongs_def":"Dibon, the name of three places in Palestine","kjv_def":"Dibon. (Also, with H1410 (גָּד) added, Dibon-gad.)"},
"H1770":{"lemma":"דִּיג","xlit":"dîyg","pron":"deeg","derivation":"denominative from H1709 (דָּג);","strongs_def":"to fish","kjv_def":"fish."},
"H1771":{"lemma":"דַּיָּג","xlit":"dayâg","pron":"dah-yawg'","derivation":"from H1770 (דִּיג);","strongs_def":"a fisherman","kjv_def":"fisher."},
"H1772":{"lemma":"דַּיָּה","xlit":"dayâh","pron":"dah-yaw'","derivation":"intensive from H1675 (דָּאָה);","strongs_def":"a falcon (from its rapid flight)","kjv_def":"vulture."},
"H1773":{"lemma":"דְּיוֹ","xlit":"dᵉyôw","pron":"deh-yo'","derivation":"of uncertain derivation;","strongs_def":"ink","kjv_def":"ink."},
"H1774":{"lemma":"דִּי זָהָב","xlit":"Dîy zâhâb","pron":"dee zaw-hawb'","derivation":"as if from H1768 (דִּי) and H2091 (זָהָב); of gold;","strongs_def":"Dizahab, a place in the Desert","kjv_def":"Dizahab."},
"H1775":{"lemma":"דִּימוֹן","xlit":"Dîymôwn","pron":"dee-mone'","derivation":"perhaps for H1769 (דִּיבוֹן);","strongs_def":"Dimon, a place in Palestine","kjv_def":"Dimon."},
"H1776":{"lemma":"דִּימוֹנָה","xlit":"Dîymôwnâh","pron":"dee-mo-naw'","derivation":"feminine of H1775 (דִּימוֹן);","strongs_def":"Dimonah, a place in Palestine","kjv_def":"Dimonah."},
"H1777":{"lemma":"דִּין","xlit":"dîyn","pron":"deen","derivation":"or (Genesis 6:3) דּוּן; a primitive root;","strongs_def":"a straight course, i.e. sail direct","kjv_def":"(come) with a straight course."},
"H1778":{"lemma":"דִּין","xlit":"dîyn","pron":"deen","derivation":"(Aramaic) corresponding to H1777 (דִּין);","strongs_def":"to judge","kjv_def":"judge."},
"H1779":{"lemma":"דִּין","xlit":"dîyn","pron":"deen","derivation":"or (Job 19:29) דּוּן; from H1777 (דִּין);","strongs_def":"judgement (the suit, justice, sentence or tribunal); by implication also strife","kjv_def":"cause, judgement, plea, strife."},
"H1780":{"lemma":"דִּין","xlit":"dîyn","pron":"deen","derivation":"(Aramaic) corresponding to H1779 (דִּין)","strongs_def":"{judgement (the suit, justice, sentence or tribunal); by implication also strife}","kjv_def":"judgement."},
"H1781":{"lemma":"דַּיָּן","xlit":"dayân","pron":"dah-yawn'","derivation":"from H1777 (דִּין);","strongs_def":"a judge or advocate","kjv_def":"judge."},
"H1782":{"lemma":"דַּיָּן","xlit":"dayân","pron":"dah-yawn'","derivation":"(Aramaic) corresponding to H1781 (דַּיָּן)","strongs_def":"{a judge or advocate}","kjv_def":"judge."},
"H1783":{"lemma":"דִּינָה","xlit":"Dîynâh","pron":"dee-naw'","derivation":"feminine of H1779 (דִּין); justice;","strongs_def":"Dinah, the daughter of Jacob","kjv_def":"Dinah."},
"H1784":{"lemma":"דִּינַי","xlit":"Dîynay","pron":"dee-nah'-ee","derivation":"(Aramaic) partial from uncertain primitive;","strongs_def":"a Dinaite or inhabitant of some unknown Assyria province","kjv_def":"Dinaite."},
"H1785":{"lemma":"דָּיֵק","xlit":"dâyêq","pron":"daw-yake'","derivation":"from a root corresp. to H1751 (דּוּק);","strongs_def":"a battering-tower","kjv_def":"fort."},
"H1786":{"lemma":"דַּיִשׁ","xlit":"dayish","pron":"dah-yish'","derivation":"from H1758 (דּוּשׁ);","strongs_def":"threshing-time","kjv_def":"threshing."},
"H1787":{"lemma":"דִּישׁוֹן","xlit":"Dîyshôwn","pron":"dee-shone'","derivation":"(Diyshon, the same as H1788 (דִּישֹׁן);","strongs_def":"Dishon, the name of two Edomites","kjv_def":"Dishon."},
"H1788":{"lemma":"דִּישֹׁן","xlit":"dîyshôn","pron":"dee-shone'","derivation":"from H1758 (דּוּשׁ);","strongs_def":"the leaper, i.e. an antelope","kjv_def":"pygarg."},
"H1789":{"lemma":"דִּישָׁן","xlit":"Dîyshân","pron":"dee-shawn'","derivation":"another form of H1787 (דִּישׁוֹן)","strongs_def":"Dishan, an Edomite","kjv_def":"Dishan, Dishon."},
"H1790":{"lemma":"דַּךְ","xlit":"dak","pron":"dak","derivation":"from an unused root (compare H1794 (דָּכָה));","strongs_def":"crushed, i.e. (figuratively) injured","kjv_def":"afflicted, oppressed."},
"H1791":{"lemma":"דֵּךְ","xlit":"dêk","pron":"dake","derivation":"(Aramaic) or דָּךְ; (Aramaic), prolonged from H1668 (דָּא);","strongs_def":"this","kjv_def":"the same, this."},
"H1792":{"lemma":"דָּכָא","xlit":"dâkâʼ","pron":"daw-kaw'","derivation":"a primitive root (compare H1794 (דָּכָה));","strongs_def":"to crumble; transitively, to bruise (literally or figuratively)","kjv_def":"beat to pieces, break (in pieces), bruise, contrite, crush, destroy, humble, oppress, smite."},
"H1793":{"lemma":"דַּכָּא","xlit":"dakkâʼ","pron":"dak-kaw'","derivation":"from H1792 (דָּכָא);","strongs_def":"crushed (literally powder, or figuratively, contrite)","kjv_def":"contrite, destruction."},
"H1794":{"lemma":"דָּכָה","xlit":"dâkâh","pron":"daw-kaw'","derivation":"a primitive root (compare H1790 (דַּךְ), H1792 (דָּכָא));","strongs_def":"to collapse (phys. or mentally)","kjv_def":"break (sore), contrite, crouch."},
"H1795":{"lemma":"דַּכָּה","xlit":"dakkâh","pron":"dak-kaw'","derivation":"from H1794 (דָּכָה) like H1793 (דַּכָּא);","strongs_def":"mutilated","kjv_def":"[phrase] wounded."},
"H1796":{"lemma":"דֳּכִי","xlit":"dŏkîy","pron":"dok-ee'","derivation":"from H1794 (דָּכָה);","strongs_def":"a dashing of surf","kjv_def":"wave."},
"H1797":{"lemma":"דִּכֵּן","xlit":"dikkên","pron":"dik-kane'","derivation":"(Aramaic) prolonged from H1791 (דֵּךְ);","strongs_def":"this","kjv_def":"same, that, this."},
"H1798":{"lemma":"דְּכַר","xlit":"dᵉkar","pron":"dek-ar'","derivation":"(Aramaic) corresponding to H2145 (זָכָר);","strongs_def":"properly, a male, i.e. of sheep","kjv_def":"ram."},
"H1799":{"lemma":"דִּכְרוֹן","xlit":"dikrôwn","pron":"dik-rone'","derivation":"(Aramaic) or דׇּכְרָן (Aramaic); corresponding to H2146 (זִכְרוֹן);","strongs_def":"a register","kjv_def":"record."},
"H1800":{"lemma":"דַּל","xlit":"dal","pron":"dal","derivation":"from H1809 (דָּלַל);","strongs_def":"properly, dangling, i.e. (by implication) weak or thin","kjv_def":"lean, needy, poor (man), weaker."},
"H1801":{"lemma":"דָּלַג","xlit":"dâlag","pron":"daw-lag'","derivation":"a primitive root;","strongs_def":"to spring","kjv_def":"leap."},
"H1802":{"lemma":"דָּלָה","xlit":"dâlâh","pron":"daw-law'","derivation":"a primitive root (compare H1809 (דָּלַל));","strongs_def":"properly, to dangle, i.e. to let down abucket (for drawing out water); figuratively, to deliver","kjv_def":"draw (out), [idiom] enough, lift up."},
"H1803":{"lemma":"דַּלָּה","xlit":"dallâh","pron":"dal-law'","derivation":"from H1802 (דָּלָה);","strongs_def":"properly, something dangling, i.e. a loose thread or hair; figuratively, indigent","kjv_def":"hair, pining sickness, poor(-est sort)."},
"H1804":{"lemma":"דָּלַח","xlit":"dâlach","pron":"daw-lakh'","derivation":"a primitive root;","strongs_def":"to roil water","kjv_def":"trouble."},
"H1805":{"lemma":"דְּלִי","xlit":"dᵉlîy","pron":"del-ee'","derivation":"or דֳּלִי; from H1802 (דָּלָה);","strongs_def":"a pail or jar (for drawing water)","kjv_def":"bucket."},
"H1806":{"lemma":"דְּלָיָה","xlit":"Dᵉlâyâh","pron":"del-aw-yaw'","derivation":"or (prolonged) דְּלָיָהוּxlit Dᵉlâyâhhûw corrected to Dᵉlâyâhûw; from H1802 (דָּלָה) and H3050 (יָהּ); Jah has delivered;","strongs_def":"Delajah, the name of five Israelites","kjv_def":"Dalaiah, Delaiah."},
"H1807":{"lemma":"דְּלִילָה","xlit":"Dᵉlîylâh","pron":"del-ee-law'","derivation":"from H1809 (דָּלַל); languishing","strongs_def":"Delilah, a Philistine woman","kjv_def":"Delilah."},
"H1808":{"lemma":"דָּלִיָּה","xlit":"dâlîyâh","pron":"daw-lee-yaw'","derivation":"from H1802 (דָּלָה);","strongs_def":"something dangling, i.e. a bough","kjv_def":"branch."},
"H1809":{"lemma":"דָּלַל","xlit":"dâlal","pron":"daw-lal'","derivation":"a primitive root (compare H1802 (דָּלָה));","strongs_def":"to slacken or be feeble; figuratively, to be oppressed","kjv_def":"bring low, dry up, be emptied, be not equal, fail, be impoverished, be made thin."},
"H1810":{"lemma":"דִּלְעָן","xlit":"Dilʻân","pron":"dil-awn'","derivation":"of uncertain derivation;","strongs_def":"Dilan, a place in Palestine","kjv_def":"Dilean."},
"H1811":{"lemma":"דָּלַף","xlit":"dâlaph","pron":"daw-laf'","derivation":"a primitive root;","strongs_def":"to drip; by implication, to weep","kjv_def":"drop through, melt, pour out."},
"H1812":{"lemma":"דֶּלֶף","xlit":"deleph","pron":"deh'-lef","derivation":"from H1811 (דָּלַף);","strongs_def":"a dripping","kjv_def":"dropping."},
"H1813":{"lemma":"דַּלְפוֹן","xlit":"Dalphôwn","pron":"dal-fone'","derivation":"from H1811 (דָּלַף); dripping;","strongs_def":"Dalphon, a son of Haman","kjv_def":"Dalphon."},
"H1814":{"lemma":"דָּלַק","xlit":"dâlaq","pron":"daw-lak'","derivation":"a primitive root;","strongs_def":"to flame (literally or figuratively)","kjv_def":"burning, chase, inflame, kindle, persecute(-or), pursue hotly."},
"H1815":{"lemma":"דְּלַק","xlit":"dᵉlaq","pron":"del-ak'","derivation":"(Aramaic) corresponding to H1814 (דָּלַק)","strongs_def":"{to flame (literally or figuratively)}","kjv_def":"burn."},
"H1816":{"lemma":"דַּלֶּקֶת","xlit":"dalleqeth","pron":"dal-lek'-keth","derivation":"; from H1814 (דָּלַק);","strongs_def":"a burning fever","kjv_def":"inflammation."},
"H1817":{"lemma":"דֶּלֶת","xlit":"deleth","pron":"deh'-leth","derivation":"from H1802 (דָּלָה);","strongs_def":"something swinging, i.e. the valve of adoor","kjv_def":"door (two-leaved), gate, leaf, lid. (Psalm 141:3)."},
"H1818":{"lemma":"דָּם","xlit":"dâm","pron":"dawm","derivation":"from H1826 (דָּמַם) (compare H119 (אָדַם));","strongs_def":"blood (as that which when shed causes death) of man or an animal; by analogy, the juice of the grape; figuratively (especially in the plural) bloodshed (i.e. drops of blood)","kjv_def":"blood(-y, -guiltiness, (-thirsty), [phrase] innocent."},
"H1819":{"lemma":"דָּמָה","xlit":"dâmâh","pron":"daw-maw'","derivation":"a primitive root;","strongs_def":"to compare; by implication, to resemble, liken, consider","kjv_def":"compare, devise, (be) like(-n), mean, think, use similitudes."},
"H1820":{"lemma":"דָּמָה","xlit":"dâmâh","pron":"daw-mam'","derivation":"a primitive root;","strongs_def":"to be dumb or silent; hence, to fail or perish; trans. to destroy","kjv_def":"cease, be cut down (off), destroy, be brought to silence, be undone, [idiom] utterly."},
"H1821":{"lemma":"דְּמָה","xlit":"dᵉmâh","pron":"dem-aw'","derivation":"(Aramaic) corresponding to H1819 (דָּמָה);","strongs_def":"to resemble","kjv_def":"be like."},
"H1822":{"lemma":"דֻּמָּה","xlit":"dummâh","pron":"doom-maw'","derivation":"from H1820 (דָּמָה);","strongs_def":"desolation; concretely, desolate","kjv_def":"destroy."},
"H1823":{"lemma":"דְּמוּת","xlit":"dᵉmûwth","pron":"dem-ooth'","derivation":"from H1819 (דָּמָה);","strongs_def":"resemblance; concretely, model, shape; adverbially, like","kjv_def":"fashion, like (-ness, as), manner, similitude."},
"H1824":{"lemma":"דְּמִי","xlit":"dᵉmîy","pron":"dem-ee'","derivation":"or דֳּמִי; from H1820 (דָּמָה);","strongs_def":"quiet","kjv_def":"cutting off, rest, silence."},
"H1825":{"lemma":"דִּמְיוֹן","xlit":"dimyôwn","pron":"dim-yone'","derivation":"from H1819 (דָּמָה);","strongs_def":"resemblance","kjv_def":"[idiom] like."},
"H1826":{"lemma":"דָּמַם","xlit":"dâmam","pron":"daw-man'","derivation":"a primitive root (compare H1724 (דָּהַם), H1820 (דָּמָה));","strongs_def":"to be dumb; by implication, to be astonished, to stop; also to perish","kjv_def":"cease, be cut down (off), forbear, hold peace, quiet self, rest, be silent, keep (put to) silence, be (stand) still, tarry, wait."},
"H1827":{"lemma":"דְּמָמָה","xlit":"dᵉmâmâh","pron":"dem-aw-maw'","derivation":"feminine from H1826 (דָּמַם);","strongs_def":"quiet","kjv_def":"calm, silence, still."},
"H1828":{"lemma":"דֹּמֶן","xlit":"dômen","pron":"do'-men","derivation":"of uncertain derivation;","strongs_def":"manure","kjv_def":"dung."},
"H1829":{"lemma":"דִּמְנָה","xlit":"Dimnâh","pron":"dim-naw'","derivation":"feminine from the same as H1828 (דֹּמֶן); a dung-heap;","strongs_def":"Dimnah, a place in Palestine","kjv_def":"Dimnah."},
"H1830":{"lemma":"דָּמַע","xlit":"dâmaʻ","pron":"daw-mah'","derivation":"a primitive root;","strongs_def":"to weep","kjv_def":"[idiom] sore, weep."},
"H1831":{"lemma":"דֶּמַע","xlit":"demaʻ","pron":"dah'-mah","derivation":"from H1830 (דָּמַע);","strongs_def":"a tear; figuratively, juice","kjv_def":"liquor."},
"H1832":{"lemma":"דִּמְעָה","xlit":"dimʻâh","pron":"dim-aw'","derivation":"feminine of H1831 (דֶּמַע);","strongs_def":"weeping","kjv_def":"tears."},
"H1833":{"lemma":"דְּמֶשֶׁק","xlit":"dᵉmesheq","pron":"dem-eh'-shek","derivation":"by orthographical variation from H1834 (דַּמֶּשֶׂק);","strongs_def":"damask (as a fabric of Damascus)","kjv_def":"in Damascus."},
"H1834":{"lemma":"דַּמֶּשֶׂק","xlit":"Dammeseq","pron":"dam-meh'-sek","derivation":"or דּוּמֶשֶׂק; or דַּרְמֶשֶׂק; of foreign origin;","strongs_def":"Damascus, a city of Syria","kjv_def":"Damascus."},
"H1835":{"lemma":"דָּן","xlit":"Dân","pron":"dawn","derivation":"from H1777 (דִּין); judge;","strongs_def":"Dan, one of the sons of Jacob; also the tribe descended from him, and its territory; likewise a place in Palestine colonized by them","kjv_def":"Daniel"},
"H1836":{"lemma":"דֵּן","xlit":"dên","pron":"dane","derivation":"(Aramaic) an orthographical variation of H1791 (דֵּךְ);","strongs_def":"this","kjv_def":"(afore-) time, [phrase] after this manner, here (-after), one...another, such, there(-fore), these, this (matter), [phrase] thus, where(-fore), which."},
"H1837":{"lemma":"דַּנָּה","xlit":"Dannâh","pron":"dan-naw'","derivation":"of uncertain derivation;","strongs_def":"Dannah, a place in Palestine","kjv_def":"Dannah."},
"H1838":{"lemma":"דִּנְהָבָה","xlit":"Dinhâbâh","pron":"din-haw-baw'","derivation":"of uncertain derivation;","strongs_def":"Dinhabah, an Edomitish town","kjv_def":"Dinhaban."},
"H1839":{"lemma":"דָּנִי","xlit":"Dânîy","pron":"daw-nee'","derivation":"patronymically from H1835 (דָּן);","strongs_def":"a Danite (often collectively) or descendants (or inhabitants) of Dan","kjv_def":"Danites, of Daniel"},
"H1840":{"lemma":"דָנִיֵּאל","xlit":"Dânîyêʼl","pron":"daw-nee-yale'","derivation":"in Ezekiel it is; דָּנִאֵל; from H1835 (דָּן) and H410 (אֵל); judge of God;","strongs_def":"Daniel or Danijel, the name of two Israelites","kjv_def":"Daniel."},
"H1841":{"lemma":"דָּנִיֵּאל","xlit":"Dânîyêʼl","pron":"daw-nee-yale'","derivation":"(Aramaic) corresponding to H1840 (דָנִיֵּאל);","strongs_def":"Danijel, the Hebrew prophet","kjv_def":"Daniel."},
"H1842":{"lemma":"דָּן יַעַן","xlit":"Dân Yaʻan","pron":"dawn yah'-an","derivation":"from H1835 (דָּן) and (apparently) H3282 (יַעַן); judge of purpose;","strongs_def":"Dan-Jaan, a place in Palestine","kjv_def":"Dan-jaan."},
"H1843":{"lemma":"דֵּעַ","xlit":"dêaʻ","pron":"day'-ah","derivation":"from H3045 (יָדַע);","strongs_def":"knowledge","kjv_def":"knowledge, opinion."},
"H1844":{"lemma":"דֵּעָה","xlit":"dêʻâh","pron":"day-aw'","derivation":"feminine of H1843 (דֵּעַ);","strongs_def":"knowledge","kjv_def":"knowledge."},
"H1845":{"lemma":"דְּעוּאֵל","xlit":"Dᵉʻûwʼêl","pron":"deh-oo-ale'","derivation":"from H3045 (יָדַע) and H410 (אֵל); known of God;","strongs_def":"Deuel, an Israelite","kjv_def":"Deuel."},
"H1846":{"lemma":"דָּעַךְ","xlit":"dâʻak","pron":"daw-ak'","derivation":"a primitive root;","strongs_def":"to be extinguished; figuratively, to expire or be dried up","kjv_def":"be extinct, consumed, put out, quenched."},
"H1847":{"lemma":"דַּעַת","xlit":"daʻath","pron":"dah'-ath","derivation":"from H3045 (יָדַע);","strongs_def":"knowledge","kjv_def":"cunning, (ig-) norantly, know(-ledge), (un-) awares (wittingly)."},
"H1848":{"lemma":"דׇּפִי","xlit":"dophîy","pron":"dof'-ee","derivation":"from an unused root (meaning to push over);","strongs_def":"a stumbling-block","kjv_def":"slanderest."},
"H1849":{"lemma":"דָּפַק","xlit":"dâphaq","pron":"daw-fak'","derivation":"a primitive root;","strongs_def":"to knock; by analogy, to press severely","kjv_def":"beat, knock, overdrive."},
"H1850":{"lemma":"דׇּפְקָה","xlit":"Dophqâh","pron":"dof-kaw'","derivation":"from H1849 (דָּפַק); a knock;","strongs_def":"Dophkah, a place in the Desert","kjv_def":"Dophkah."},
"H1851":{"lemma":"דַּק","xlit":"daq","pron":"dak","derivation":"from H1854 (דָּקַק);","strongs_def":"crushed, i.e. (by implication) small or thin","kjv_def":"dwarf, lean(-fleshed), very little thing, small, thin."},
"H1852":{"lemma":"דֹּק","xlit":"dôq","pron":"doke","derivation":"from H1854 (דָּקַק);","strongs_def":"something crumbling, i.e. fine (as a thin cloth)","kjv_def":"curtain."},
"H1853":{"lemma":"דִּקְלָה","xlit":"Diqlâh","pron":"dik-law'","derivation":"of foreign origin;","strongs_def":"Diklah, a region of Arabia","kjv_def":"Diklah."},
"H1854":{"lemma":"דָּקַק","xlit":"dâqaq","pron":"daw-kak'","derivation":"a primitive root (compare H1915 (הָדַךְ)lemma הָדךְ missing vowel, corrected to הָדַךְ);","strongs_def":"to crush (or intransitively) crumble","kjv_def":"beat in pieces (small), bruise, make dust, (into) [idiom] powder, (be, very) small, stamp (small)."},
"H1855":{"lemma":"דְּקַק","xlit":"dᵉqaq","pron":"dek-ak'","derivation":"(Aramaic) corresponding to H1854 (דָּקַק);","strongs_def":"to crumble or (trans.) crush","kjv_def":"break to pieces."},
"H1856":{"lemma":"דָּקַר","xlit":"dâqar","pron":"daw-kar'","derivation":"a primitive root;","strongs_def":"to stab; by analogy, to starve; figuratively, to revile","kjv_def":"pierce, strike (thrust) through, wound."},
"H1857":{"lemma":"דֶּקֶר","xlit":"Deqer","pron":"deh'-ker","derivation":"from H1856 (דָּקַר); a stab;","strongs_def":"Deker, an Israelite","kjv_def":"Dekar."},
"H1858":{"lemma":"דַּר","xlit":"dar","pron":"dar","derivation":"apparently from the same as H1865 (דְּרוֹר);","strongs_def":"properly, a pearl (from its sheen as rapidly turned); by analogy, pearl-stone, i.e. mother-of-pearl or alabaster","kjv_def":"[idiom] white."},
"H1859":{"lemma":"דָּר","xlit":"dâr","pron":"dawr","derivation":"(Aramaic) corresponding to H1755 (דּוֹר);","strongs_def":"an age","kjv_def":"generation."},
"H1860":{"lemma":"דְּרָאוֹן","xlit":"dᵉrâʼôwn","pron":"der-aw-one'","derivation":"or דֵּרָאוֹן; from an unused root (meaning to repulse);","strongs_def":"an object of aversion","kjv_def":"abhorring, contempt."},
"H1861":{"lemma":"דׇּרְבוֹן","xlit":"dorbôwn","pron":"dor-bone'","derivation":"(also dor-bawn'); of uncertain derivation;","strongs_def":"a goad","kjv_def":"goad."},
"H1862":{"lemma":"דַּרְדַּע","xlit":"Dardaʻ","pron":"dar-dah'","derivation":"apparently from H1858 (דַּר) and H1843 (דֵּעַ); pearl of knowledge;","strongs_def":"Darda, an Israelite","kjv_def":"Darda."},
"H1863":{"lemma":"דַּרְדַּר","xlit":"dardar","pron":"dar-dar'","derivation":"of uncertain derivation;","strongs_def":"a thorn","kjv_def":"thistle."},
"H1864":{"lemma":"דָּרוֹם","xlit":"dârôwm","pron":"daw-rome'","derivation":"of uncertain derivation;","strongs_def":"the south; poet. the south wind","kjv_def":"south."},
"H1865":{"lemma":"דְּרוֹר","xlit":"dᵉrôwr","pron":"der-ore'","derivation":"from an unused root (meaning to move rapidly);","strongs_def":"freedom; hence, spontaneity of outflow, and so clear","kjv_def":"liberty, pure."},
"H1866":{"lemma":"דְּרוֹר","xlit":"dᵉrôwr","pron":"der-ore'","derivation":"the same as H1865 (דְּרוֹר), applied to a bird;","strongs_def":"the swift, a kind of swallow","kjv_def":"swallow."},
"H1867":{"lemma":"דָּֽרְיָוֵשׁ","xlit":"Dârᵉyâvêsh","pron":"daw-reh-yaw-vaysh'","derivation":"of Persian origin;","strongs_def":"Darejavesh, a title (rather than name) of several Persian kings","kjv_def":"Darius."},
"H1868":{"lemma":"דָּֽרְיָוֵשׁ","xlit":"Dârᵉyâvêsh","pron":"daw-reh-yaw-vaysh'","derivation":"(Aramaic) corresponding to H1867 (דָּֽרְיָוֵשׁ)","strongs_def":"{Darejavesh, a title (rather than name) of several Persian kings}","kjv_def":"Darius."},
"H1869":{"lemma":"דָּרַךְ","xlit":"dârak","pron":"daw-rak'","derivation":"a primitive root;","strongs_def":"to tread; by implication, to walk; also to string abow (by treading on it in bending)","kjv_def":"archer, bend, come, draw, go (over), guide, lead (forth), thresh, tread (down), walk."},
"H1870":{"lemma":"דֶּרֶךְ","xlit":"derek","pron":"deh'-rek","derivation":"from H1869 (דָּרַךְ);","strongs_def":"a road (as trodden); figuratively, a course of life or mode of action, often adverb","kjv_def":"along, away, because of, [phrase] by, conversation, custom, (east-) ward, journey, manner, passenger, through, toward, (high-) (path-) way(-side), whither(-soever)."},
"H1871":{"lemma":"דַּרְכְּמוֹן","xlit":"darkᵉmôwn","pron":"dar-kem-one'","derivation":"of Persian origin;","strongs_def":"a 'drachma', or coin","kjv_def":"dram."},
"H1872":{"lemma":"דְּרַע","xlit":"dᵉraʻ","pron":"der-aw'","derivation":"(Aramaic) corresponding to H2220 (זְרוֹעַ);","strongs_def":"an arm","kjv_def":"arm."},
"H1873":{"lemma":"דָּרַע","xlit":"Dâraʻ","pron":"daw-rah'","derivation":"probably a contraction from H1862 (דַּרְדַּע);","strongs_def":"Dara, an Israelite","kjv_def":"Dara."},
"H1874":{"lemma":"דַּרְקוֹן","xlit":"Darqôwn","pron":"dar-kone'","derivation":"of uncertain derivation;","strongs_def":"Darkon, one of 'Solomon's servants'","kjv_def":"Darkon."},
"H1875":{"lemma":"דָּרַשׁ","xlit":"dârash","pron":"daw-rash'","derivation":"a primitive root;","strongs_def":"properly, to tread or frequent; usually to follow (for pursuit or search); by implication, to seek or ask; specifically to worship","kjv_def":"ask, [idiom] at all, care for, [idiom] diligently, inquire, make inquisition, (necro-) mancer, question, require, search, seek (for, out), [idiom] surely."},
"H1876":{"lemma":"דָּשָׁא","xlit":"dâshâʼ","pron":"daw-shaw'","derivation":"a primitive root;","strongs_def":"to sprout","kjv_def":"bring forth, spring."},
"H1877":{"lemma":"דֶּשֶׁא","xlit":"desheʼ","pron":"deh'-sheh","derivation":"from H1876 (דָּשָׁא);","strongs_def":"a sprout; by analogy, grass","kjv_def":"(tender) grass, green, (tender) herb."},
"H1878":{"lemma":"דָּשֵׁן","xlit":"dâshên","pron":"daw-shane'","derivation":"a primitive root; also denominatively (from H1880 (דֶּשֶׁן))","strongs_def":"to be fat; transitively, to fatten (or regard as fat); specifically to anoint; figuratively, to satisfy;  to remove (fat) ashes (of sacrifices)","kjv_def":"accept, anoint, take away the (receive) ashes (from), make (wax) fat."},
"H1879":{"lemma":"דָּשֵׁן","xlit":"dâshên","pron":"daw-shane'","derivation":"from H1878 (דָּשֵׁן);","strongs_def":"fat; figuratively, rich, fertile","kjv_def":"fat."},
"H1880":{"lemma":"דֶּשֶׁן","xlit":"deshen","pron":"deh'-shen","derivation":"from H1878 (דָּשֵׁן);","strongs_def":"the fat; abstractly fatness, i.e. (figuratively) abundance; specifically the (fatty) ashes of sacrifices","kjv_def":"ashes, fatness."},
"H1881":{"lemma":"דָּת","xlit":"dâth","pron":"dawth","derivation":"of uncertain (perhaps foreign) derivation;","strongs_def":"a royal edict or statute","kjv_def":"commandment, commission, decree, law, manner."},
"H1882":{"lemma":"דָּת","xlit":"dâth","pron":"dawth","derivation":"(Aramaic) corresponding to H1881 (דָּת);","strongs_def":"{a royal edict or statute}","kjv_def":"decree, law."},
"H1883":{"lemma":"דֶּתֶא","xlit":"detheʼ","pron":"deh'-thay","derivation":"(Aramaic) corresponding to H1877 (דֶּשֶׁא)","strongs_def":"{a sprout; by analogy, grass}","kjv_def":"tender grass."},
"H1884":{"lemma":"דְּתָבָר","xlit":"dᵉthâbâr","pron":"deth-aw-bawr'","derivation":"(Aramaic) of Persian origin;","strongs_def":"meaning one skilled in law; a judge","kjv_def":"counsellor."},
"H1885":{"lemma":"דָּתָן","xlit":"Dâthân","pron":"daw-thawn'","derivation":"of uncertain derivation;","strongs_def":"Dathan, an Israelite","kjv_def":"Dathan."},
"H1886":{"lemma":"דֹּתָן","xlit":"Dôthân","pron":"do'-thawn","derivation":"or (Aramaic dual) דֹּתַיִן; (Genesis 37:17), of uncertain derivation;","strongs_def":"Dothan, a place in Palestine","kjv_def":"Dothan. h"},
"H1887":{"lemma":"הֵא","xlit":"hêʼ","pron":"hay","derivation":"a primitive particle;","strongs_def":"lo!","kjv_def":"behold, lo."},
"H1888":{"lemma":"הֵא","xlit":"hêʼ","pron":"hay","derivation":"(Aramaic) or הָא; (Aramaic), corresponding to H1887 (הֵא)","strongs_def":"{lo!}","kjv_def":"even, lo."},
"H1889":{"lemma":"הֶאָח","xlit":"heʼâch","pron":"heh-awkh'","derivation":"from H1887 (הֵא) and H253 (אָח);","strongs_def":"aha!","kjv_def":"ah, aha, ha."},
"H1890":{"lemma":"הַבְהָב","xlit":"habhâb","pron":"hab-hawb'","derivation":"by reduplication from H3051 (יָהַב);","strongs_def":"gift (in sacrifice), i.e. holocaust","kjv_def":"offering."},
"H1891":{"lemma":"הָבַל","xlit":"hâbal","pron":"haw-bal'","derivation":"a primitive root;","strongs_def":"to be vain in act, word, or expectation; specifically to lead astray","kjv_def":"be (become, make) vain."},
"H1892":{"lemma":"הֶבֶל","xlit":"hebel","pron":"heh'bel","derivation":"or (rarely in the abs.) הֲבֵל; from H1891 (הָבַל);","strongs_def":"emptiness or vanity; figuratively, something transitory and unsatisfactory; often used as an adverb","kjv_def":"[idiom] altogether, vain, vanity."},
"H1893":{"lemma":"הֶבֶל","xlit":"Hebel","pron":"heh'-bel","derivation":"the same as H1892 (הֶבֶל);","strongs_def":"Hebel, the son of Adam","kjv_def":"Abel."},
"H1894":{"lemma":"הֹבֶן","xlit":"hôben","pron":"ho'-ben","derivation":"only in plural, from an unused root meaning to be hard;","strongs_def":"ebony","kjv_def":"ebony."},
"H1895":{"lemma":"הָבַר","xlit":"hâbar","pron":"haw-bar'","derivation":"a primitive root of uncertain (perhaps foreign) derivation;","strongs_def":"to be a horoscopist","kjv_def":"[phrase] (astro-) loger."},
"H1896":{"lemma":"הֵגֵא","xlit":"Hêgêʼ","pron":"hay-gay'","derivation":"or (by permutation) הֵגַי; probably of Persian origin;","strongs_def":"Hege or Hegai, a eunuch of Xerxes","kjv_def":"Hegai, Hege."},
"H1897":{"lemma":"הָגָה","xlit":"hâgâh","pron":"daw-gaw'","derivation":"a primitive root (compare H1901 (הָגִיג));","strongs_def":"to murmur (in pleasure or anger); by implication, to ponder","kjv_def":"imagine, meditate, mourn, mutter, roar, [idiom] sore, speak, study, talk, utter."},
"H1898":{"lemma":"הָגָה","xlit":"hâgâh","pron":"haw-gaw'","derivation":"a primitive root;","strongs_def":"to remove","kjv_def":"stay, stay away."},
"H1899":{"lemma":"הֶגֶה","xlit":"hegeh","pron":"heh'-geh","derivation":"from H1897 (הָגָה);","strongs_def":"a muttering (in sighing, thought, or as thunder)","kjv_def":"mourning, sound, tale."},
"H1900":{"lemma":"הָגוּת","xlit":"hâgûwth","pron":"haw-gooth'","derivation":"from H1897 (הָגָה);","strongs_def":"musing","kjv_def":"meditation."},
"H1901":{"lemma":"הָגִיג","xlit":"hâgîyg","pron":"haw-gheeg'","derivation":"from an unused root akin to H1897 (הָגָה);","strongs_def":"properly, a murmur, i.e. complaint","kjv_def":"meditation, musing."},
"H1902":{"lemma":"הִגָּיוֹן","xlit":"higgâyôwn","pron":"hig-gaw-yone'","derivation":"intensive from H1897 (הָגָה);","strongs_def":"a murmuring sound, i.e. a musical notation (probably similar to the modern affettuoso to indicate solemnity of movement); by implication, a machination","kjv_def":"device, Higgaion, meditation, solemn sound."},
"H1903":{"lemma":"הָגִין","xlit":"hâgîyn","pron":"haw-gheen'","derivation":"of uncertain derivation;","strongs_def":"perhaps suitable or turning","kjv_def":"directly."},
"H1904":{"lemma":"הָגָר","xlit":"Hâgâr","pron":"haw-gawr'","derivation":"of uncertain (perhaps foreign) derivation;","strongs_def":"Hagar, the mother of Ishmael","kjv_def":"Hagar."},
"H1905":{"lemma":"הַגְרִי","xlit":"Hagrîy","pron":"hag-ree'","derivation":"or (prolonged) הַגְרִיאxlit Hagrîʼ corrected to Hagrîyʼ; perhaps patronymically from H1904 (הָגָר);","strongs_def":"a Hagrite or member of a certain Arabian clan","kjv_def":"Hagarene, Hagarite, Haggeri."},
"H1906":{"lemma":"הֵד","xlit":"hêd","pron":"hade","derivation":"for H1959 (הֵידָד);","strongs_def":"a shout","kjv_def":"sounding again."},
"H1907":{"lemma":"הַדָּבָר","xlit":"haddâbâr","pron":"had-daw-bawr'","derivation":"(Aramaic) probably of foreign origin;","strongs_def":"a vizier","kjv_def":"counsellor."},
"H1908":{"lemma":"הֲדַד","xlit":"Hădad","pron":"had-ad'","derivation":"probably of foreign origin (compare H111 (אֲדַד));","strongs_def":"Hadad, the name of an idol, and of several kings of Edom","kjv_def":"Hadad."},
"H1909":{"lemma":"הֲדַדְעֶזֶר","xlit":"Hădadʻezer","pron":"had-ad-eh'-zer","derivation":"from H1908 (הֲדַד) and H5828 (עֵזֶר); Hadad (is his) help;","strongs_def":"Hadadezer, a Syrian king","kjv_def":"Hadadezer. Compare H1928 (הֲדַרְעֶזֶר)."},
"H1910":{"lemma":"הֲדַדְרִמּוֹן","xlit":"Hădadrimmôwn","pron":"had-ad-rim-mone'","derivation":"from H1908 (הֲדַד) and H7417 (רִמּוֹן);","strongs_def":"Hadad-Rimmon, a place in Palestine","kjv_def":"Hadad-rimmon."},
"H1911":{"lemma":"הָדָה","xlit":"hâdâh","pron":"haw-daw'","derivation":"a primitive root (compare H3034 (יָדָה));","strongs_def":"to stretch forth the hand","kjv_def":"put."},
"H1912":{"lemma":"הֹדוּ","xlit":"Hôdûw","pron":"ho'-doo","derivation":"of foreign origin;","strongs_def":"Hodu (i.e. Hindustan)","kjv_def":"India."},
"H1913":{"lemma":"הֲדוֹרָם","xlit":"Hădôwrâm","pron":"had-o-rawm'","derivation":"or הֲדֹרָם; probably of foreign derivation;","strongs_def":"Hadoram, a son of Joktan, and the tribe descended from him","kjv_def":"Hadoram."},
"H1914":{"lemma":"הִדַּי","xlit":"Hidday","pron":"hid-dah'ee","derivation":"of uncertain derivation;","strongs_def":"Hiddai, an Israelite","kjv_def":"Hiddai."},
"H1915":{"lemma":"הָדַךְ","xlit":"hâdak","pron":"haw-dak'","derivation":"a primitive root (compare H1854 (דָּקַק));","strongs_def":"to crush with the foot","kjv_def":"tread down."},
"H1916":{"lemma":"הֲדֹם","xlit":"hădôm","pron":"had-ome'","derivation":"from an unused root meaning to stamp upon;","strongs_def":"a foot stool","kjv_def":"(foot-) stool."},
"H1917":{"lemma":"הַדָּם","xlit":"haddâm","pron":"had-dawm'","derivation":"(Aramaic) from a root corresponding to that of H1916 (הֲדֹם);","strongs_def":"something stamped to pieces, i.e. a bit","kjv_def":"piece."},
"H1918":{"lemma":"הֲדַס","xlit":"hădaç","pron":"had-as'","derivation":"of uncertain derivation;","strongs_def":"the myrtle","kjv_def":"myrtle (tree)."},
"H1919":{"lemma":"הֲדַסָּה","xlit":"Hădaççâh","pron":"had-as-saw'","derivation":"feminine of H1918 (הֲדַס);","strongs_def":"Hadassah (or Esther)","kjv_def":"Hadassah."},
"H1920":{"lemma":"הָדַף","xlit":"Hâdaph","pron":"haw-daf'","derivation":"a primitive root;","strongs_def":"to push away or down","kjv_def":"cast away (out), drive, expel, thrust (away)."},
"H1921":{"lemma":"הָדַר","xlit":"hâdar","pron":"haw-dar'","derivation":"a primitive root;","strongs_def":"to swell up (literally or figuratively, active or passive); by implication, to favor or honour, be high or proud","kjv_def":"countenance, crooked place, glorious, honour, put forth."},
"H1922":{"lemma":"הֲדַר","xlit":"hădar","pron":"had-ar'","derivation":"(Aramaic) corresponding to H1921 (הָדַר);","strongs_def":"to magnify (figuratively)","kjv_def":"glorify, honour."},
"H1923":{"lemma":"הֲדַר","xlit":"hădar","pron":"had-ar'","derivation":"(Aramaic) from H1922 (הֲדַר)lemma הֲדַּר extra dagesh, corrected to הֲדַר;","strongs_def":"magnificence","kjv_def":"honour, majesty."},
"H1924":{"lemma":"הֲדַר","xlit":"Hădar","pron":"had-ar'","derivation":"the same as H1926 (הָדָר);","strongs_def":"Hadar, an Edomite","kjv_def":"Hadar."},
"H1925":{"lemma":"הֶדֶר","xlit":"heder","pron":"heh'-der","derivation":"from H1921 (הָדַר);","strongs_def":"honour; used (figuratively) for the capital city (Jerusalem)","kjv_def":"glory."},
"H1926":{"lemma":"הָדָר","xlit":"hâdâr","pron":"haw-dawr'","derivation":"from H1921 (הָדַר);","strongs_def":"magnificence, i.e. ornament or splendor","kjv_def":"beauty, comeliness, excellency, glorious, glory, goodly, honour, majesty."},
"H1927":{"lemma":"הֲדָרָה","xlit":"hădârâh","pron":"had-aw-raw'","derivation":"feminine of H1926 (הָדָר);","strongs_def":"decoration","kjv_def":"beauty, honour."},
"H1928":{"lemma":"הֲדַרְעֶזֶר","xlit":"Hădarʻezer","pron":"had-ar-eh'-zer","derivation":"from H1924 (הֲדַר) and H5828 (עֵזֶר); Hadar (i.e. Hadad, H1908 (הֲדַד)) is his help; (i.e. Hadadezer, H1909 (הֲדַדְעֶזֶר))","strongs_def":"Hadarezer,  a Syrian king","kjv_def":"Hadarezer."},
"H1929":{"lemma":"הָהּ","xlit":"hâhh","pron":"haw","derivation":"a shortened form of H162 (אֲהָהּ);","strongs_def":"ah! expressing grief","kjv_def":"woe worth."},
"H1930":{"lemma":"הוֹ","xlit":"hôw","pron":"ho","derivation":"by permutation from H1929 (הָהּ);","strongs_def":"oh!","kjv_def":"alas."},
"H1931":{"lemma":"הוּא","xlit":"hûwʼ","pron":"hoo","derivation":"of which the feminine (beyond the Pentateuch) is הִיא; a primitive word, the third person pronoun singular;","strongs_def":"he (she or it); only expressed when emphatic or without a verb; also (intensively) self, or (especially with the article) the same; sometimes (as demonstrative) this or that; occasionally (instead of copula) as or are","kjv_def":"he, as for her, him(-self), it, the same, she (herself), such, that (...it), these, they, this, those, which (is), who."},
"H1932":{"lemma":"הוּא","xlit":"hûwʼ","pron":"hoo","derivation":"(Aramaic) or (feminine) הִיא; (Aramaic), corresponding to H1931 (הוּא)","strongs_def":"{he (she or it); self, or (especially with the article) the same; sometimes (as demonstrative) this or that; occasionally (instead of copula) as or are}","kjv_def":"[idiom] are, it, this."},
"H1933":{"lemma":"הָוָא","xlit":"hâvâʼ","pron":"haw-vaw'","derivation":"or הָוָה; a primitive root (compare H183 (אָוָה), H1961 (הָיָה)) supposed to mean properly, to breathe;","strongs_def":"to be (in the sense of existence)","kjv_def":"be, [idiom] have."},
"H1934":{"lemma":"הָוָא","xlit":"hâvâʼ","pron":"hav-aw'","derivation":"(Aramaic) or הָוָה; (Aramaic), corresponding to H1933 (הָוָא);","strongs_def":"to exist; used in a great variety of applications (especially in connection with other words)","kjv_def":"be, become, [phrase] behold, [phrase] came (to pass), [phrase] cease, [phrase] cleave, [phrase] consider, [phrase] do, [phrase] give, [phrase] have, [phrase] judge, [phrase] keep, [phrase] labour, [phrase] mingle (self), [phrase] put, [phrase] see, [phrase] seek, [phrase] set, [phrase] slay, [phrase] take heed, tremble, [phrase] walk, [phrase] would."},
"H1935":{"lemma":"הוֹד","xlit":"hôwd","pron":"hode","derivation":"from an unused root;","strongs_def":"grandeur (i.e. an imposing form and appearance)","kjv_def":"beauty, comeliness, excellency, glorious, glory, goodly, honour, majesty."},
"H1936":{"lemma":"הוֹד","xlit":"Hôwd","pron":"hode","derivation":"the same as H1935 (הוֹד);","strongs_def":"Hod, an Israelite","kjv_def":"Hod."},
"H1937":{"lemma":"הוֹדְוָה","xlit":"Hôwdᵉvâh","pron":"ho-dev-aw'","derivation":"a form of H1938 (הוֹדַוְיָה);","strongs_def":"Hodevah (or Hodevjah), an Israelite","kjv_def":"Hodevah."},
"H1938":{"lemma":"הוֹדַוְיָה","xlit":"Hôwdavyâh","pron":"ho-dav-yaw'","derivation":"from H1935 (הוֹד) and H3050 (יָהּ); majesty of Jah;","strongs_def":"Hodavjah, the name of three Israelites","kjv_def":"Hodaviah."},
"H1939":{"lemma":"הוֹדַיְוָהוּ","xlit":"Hôwdayvâhûw","pron":"ho-dah-yeh-vaw'-hoo","derivation":"a form of H1938 (הוֹדַוְיָה);","strongs_def":"Hodajvah, an Israelite","kjv_def":"Hodaiah."},
"H1940":{"lemma":"הוֹדִיָּה","xlit":"Hôwdîyâh","pron":"ho-dee-yaw'","derivation":"a form for the feminine of H3064 (יְהוּדִי);","strongs_def":"a Jewess","kjv_def":"Hodiah."},
"H1941":{"lemma":"הוֹדִיָּה","xlit":"Hôwdîyâh","pron":"ho-dee-yaw'","derivation":"a form of H1938 (הוֹדַוְיָה);","strongs_def":"Hodijah, the name of three Israelites","kjv_def":"Hodijah."},
"H1942":{"lemma":"הַוָּה","xlit":"havvâh","pron":"hav-vaw'","derivation":"from H1933 (הָוָא) (in the sense of eagerly coveting and rushing upon;","strongs_def":"by implication, of falling); desire; also ruin","kjv_def":"calamity, iniquity, mischief, mischievous (thing), naughtiness, naughty, noisome, perverse thing, substance, very wickedness."},
"H1943":{"lemma":"הֹוָה","xlit":"hôvâh","pron":"ho-vaw'","derivation":"another form for H1942 (הַוָּה);","strongs_def":"ruin","kjv_def":"mischief."},
"H1944":{"lemma":"הוֹהָם","xlit":"Hôwhâm","pron":"ho-hawm'","derivation":"of uncertain derivation;","strongs_def":"Hoham, a Canaanitish king","kjv_def":"Hoham."},
"H1945":{"lemma":"הוֹי","xlit":"hôwy","pron":"hoh'ee","derivation":"a prolonged form of H1930 (הוֹ) (akin to H188 (אוֹי));","strongs_def":"oh!","kjv_def":"ah, alas, ho, O, woe."},
"H1946":{"lemma":"הוּךְ","xlit":"hûwk","pron":"hook","derivation":"(Aramaic) corresponding to H1981 (הֲלַךְ);","strongs_def":"to go; causatively, to bring","kjv_def":"bring again, come, go (up)."},
"H1947":{"lemma":"הוֹלֵלָה","xlit":"hôwlêlâh","pron":"ho-lay-law'","derivation":"feminine active participle of H1984 (הָלַל);","strongs_def":"folly","kjv_def":"madness."},
"H1948":{"lemma":"הוֹלֵלוּת","xlit":"hôwlêlûwth","pron":"ho-lay-looth'","derivation":"from active participle of H1984 (הָלַל);","strongs_def":"folly","kjv_def":"madness."},
"H1949":{"lemma":"הוּם","xlit":"hûwm","pron":"hoom","derivation":"a primitive root (compare H2000 (הָמַם));","strongs_def":"to make an uproar, or agitate greatly","kjv_def":"destroy, move, make a noise, put, ring again."},
"H1950":{"lemma":"הוֹמָם","xlit":"Hôwmâm","pron":"ho-mawm'","derivation":"from H2000 (הָמַם); raging;","strongs_def":"Homam, an Edomitish chieftain","kjv_def":"Homam. Compare H1967 (הֵימָם)."},
"H1951":{"lemma":"הוּן","xlit":"hûwn","pron":"hoon","derivation":"a primitive root;","strongs_def":"properly, to be naught, i.e. (figuratively) to be (causatively, act) light","kjv_def":"be ready."},
"H1952":{"lemma":"הוֹן","xlit":"hôwn","pron":"hone","derivation":"from the same as H1951 (הוּן) in the sense of H202 (אוֹן);","strongs_def":"wealth; by implication, enough","kjv_def":"enough, [phrase] for nought, riches, substance, wealth."},
"H1953":{"lemma":"הוֹשָׁמָע","xlit":"Hôwshâmâʻ","pron":"ho-shaw-maw'","derivation":"from H3068 (יְהֹוָה) and H8085 (שָׁמַע); Jehovah has heard;","strongs_def":"Hoshama, an Israelite","kjv_def":"Hoshama."},
"H1954":{"lemma":"הוֹשֵׁעַ","xlit":"Hôwshêaʻ","pron":"ho-shay'-ah","derivation":"from H3467 (יָשַׁע); deliverer;","strongs_def":"Hoshea, the name of five Israelites","kjv_def":"Hosea, Hoshea, Oshea."},
"H1955":{"lemma":"הוֹשַׁעְיָה","xlit":"Hôwshaʻyâh","pron":"ho-shah-yaw'","derivation":"from H3467 (יָשַׁע) and H3050 (יָהּ); Jah has saved;","strongs_def":"Hoshajah, the name of two Israelites","kjv_def":"Hoshaiah."},
"H1956":{"lemma":"הוֹתִיר","xlit":"Hôwthîyr","pron":"ho-theer'","derivation":"from H3498 (יָתַר); he has caused to remain;","strongs_def":"Hothir, an Israelite","kjv_def":"Hothir."},
"H1957":{"lemma":"הָזָה","xlit":"hâzâh","pron":"haw-zaw'","derivation":"a primitive root (compare H2372 (חָזָה));","strongs_def":"to dream","kjv_def":"sleep."},
"H1958":{"lemma":"הִי","xlit":"hîy","pron":"he","derivation":"for H5092 (נְהִי);","strongs_def":"lamentation","kjv_def":"woe. (For hiyr. See H1931 (הוּא), H1932 (הוּא).)"},
"H1959":{"lemma":"הֵידָד","xlit":"hêydâd","pron":"hay-dawd'","derivation":"from an unused root (meaning to shout);","strongs_def":"acclamation","kjv_def":"shout(-ing)."},
"H1960":{"lemma":"הֻיְּדָה","xlit":"huyᵉdâh","pron":"hoo-yed-aw'","derivation":"from the same as H1959 (הֵידָד);","strongs_def":"properly, an acclaim, i.e. a choir of singers","kjv_def":"thanksgiving."},
"H1961":{"lemma":"הָיָה","xlit":"hâyâh","pron":"haw-yaw","derivation":"a primitive root (compare H1933 (הָוָא));","strongs_def":"to exist, i.e. be or become, come to pass (always emphatic, and not a mere copula or auxiliary)","kjv_def":"beacon, [idiom] altogether, be(-come), accomplished, committed, like), break, cause, come (to pass), do, faint, fall, [phrase] follow, happen, [idiom] have, last, pertain, quit (one-) self, require, [idiom] use."},
"H1962":{"lemma":"הַיָּה","xlit":"hayâh","pron":"hah-yaw'","derivation":"another form for H1943 (הֹוָה);","strongs_def":"ruin","kjv_def":"calamity."},
"H1963":{"lemma":"הֵיךְ","xlit":"hêyk","pron":"hake","derivation":"another form for H349 (אֵיךְ);","strongs_def":"how?","kjv_def":"how."},
"H1964":{"lemma":"הֵיכָל","xlit":"hêykâl","pron":"hay-kawl'","derivation":"probably from H3201 (יָכֹל) (in the sense of capacity);","strongs_def":"a large public building, such as a palace or temple","kjv_def":"palace, temple."},
"H1965":{"lemma":"הֵיכַל","xlit":"hêykal","pron":"hay-kal'","derivation":"(Aramaic) corresponding to H1964 (הֵיכָל)","strongs_def":"{a large public building, such as a palace or temple}","kjv_def":"palace, temple."},
"H1966":{"lemma":"הֵילֵל","xlit":"hêylêl","pron":"hay-lale'","derivation":"from H1984 (הָלַל) (in the sense of brightness);","strongs_def":"the morning-star","kjv_def":"lucifer."},
"H1967":{"lemma":"הֵימָם","xlit":"Hêymâm","pron":"hay-mawm'","derivation":"another form for H1950 (הוֹמָם);","strongs_def":"Hemam, an Idumaean","kjv_def":"Hemam."},
"H1968":{"lemma":"הֵימָן","xlit":"Hêymân","pron":"hay-mawn'","derivation":"probably from H539 (אָמַן); faithful;","strongs_def":"Heman, the name of at least two Israelites","kjv_def":"Heman."},
"H1969":{"lemma":"הִין","xlit":"hîyn","pron":"heen","derivation":"probably of Egyptian origin;","strongs_def":"a hin or liquid measure","kjv_def":"hin."},
"H1970":{"lemma":"הָכַר","xlit":"hâkar","pron":"haw-kar'","derivation":"a primitive root; apparently","strongs_def":"to injure","kjv_def":"make self strange."},
"H1971":{"lemma":"הַכָּרָה","xlit":"hakkârâh","pron":"hak-kaw-raw'","derivation":"from H5234 (נָכַר);","strongs_def":"respect, i.e. partiality","kjv_def":"shew."},
"H1972":{"lemma":"הָלָא","xlit":"hâlâʼ","pron":"haw-law'","derivation":"probably denominatively from H1973 (הָלְאָה)xlit hâlᵉâh corrected to hâlᵉʼâh;","strongs_def":"to remove or be remote","kjv_def":"cast far off."},
"H1973":{"lemma":"הָלְאָה","xlit":"hâlᵉʼâh","pron":"haw-leh-aw'","derivation":"from the primitive form of the article (hal);","strongs_def":"to the distance, i.e. far away; also (of time) thus far","kjv_def":"back, beyond, (hence,-) forward, hitherto, thence, forth, yonder."},
"H1974":{"lemma":"הִלּוּל","xlit":"hillûwl","pron":"hil-lool'","derivation":"from H1984 (הָלַל) (in the sense of rejoicing);","strongs_def":"a celebration of thanksgiving for harvest","kjv_def":"merry, praise."},
"H1975":{"lemma":"הַלָּז","xlit":"hallâz","pron":"hal-lawz'","derivation":"from H1976 (הַלָּזֶה);","strongs_def":"this or that","kjv_def":"side, that, this."},
"H1976":{"lemma":"הַלָּזֶה","xlit":"hallâzeh","pron":"hal-law-zeh'","derivation":"from the article (see H1973 (הָלְאָה)xlit hâlᵉâh corrected to hâlᵉʼâh) and H2088 (זֶה);","strongs_def":"this very","kjv_def":"this."},
"H1977":{"lemma":"הַלֵּזוּ","xlit":"hallêzûw","pron":"hal-lay-zoo'","derivation":"another form of H1976 (הַלָּזֶה);","strongs_def":"that","kjv_def":"this."},
"H1978":{"lemma":"הָלִיךְ","xlit":"hâlîyk","pron":"haw-leek'","derivation":"from H1980 (הָלַךְ);","strongs_def":"a walk, i.e. (by implication) a step","kjv_def":"step."},
"H1979":{"lemma":"הֲלִיכָה","xlit":"hălîykâh","pron":"hal-ee-kaw'","derivation":"feminine of H1978 (הָלִיךְ);","strongs_def":"a walking; by implication, a procession or march, a caravan","kjv_def":"company, going, walk, way."},
"H1980":{"lemma":"הָלַךְ","xlit":"hâlak","pron":"haw-lak'","derivation":"akin to H3212 (יָלַךְ); a primitive root;","strongs_def":"to walk (in a great variety of applications, literally and figuratively)","kjv_def":"(all) along, apace, behave (self), come, (on) continually, be conversant, depart, [phrase] be eased, enter, exercise (self), [phrase] follow, forth, forward, get, go (about, abroad, along, away, forward, on, out, up and down), [phrase] greater, grow, be wont to haunt, lead, march, [idiom] more and more, move (self), needs, on, pass (away), be at the point, quite, run (along), [phrase] send, speedily, spread, still, surely, [phrase] tale-bearer, [phrase] travel(-ler), walk (abroad, on, to and fro, up and down, to places), wander, wax, (way-) faring man, [idiom] be weak, whirl."},
"H1981":{"lemma":"הֲלַךְ","xlit":"hălak","pron":"hal-ak'","derivation":"(Aramaic) corresponding to H1980 (הָלַךְ) (compare H1946 (הוּךְ));","strongs_def":"to walk","kjv_def":"walk."},
"H1982":{"lemma":"הֵלֶךְ","xlit":"hêlek","pron":"hay'-lek","derivation":"from H1980 (הָלַךְ);","strongs_def":"properly, a journey, i.e. (by implication) a wayfarer; also a flowing","kjv_def":"[idiom] dropped, traveller."},
"H1983":{"lemma":"הֲלָךְ","xlit":"hălâk","pron":"hal-awk'","derivation":"(Aramaic) from H1981 (הֲלַךְ);","strongs_def":"properly, a journey, i.e. (by implication) toll on goods at aroad","kjv_def":"custom."},
"H1984":{"lemma":"הָלַל","xlit":"hâlal","pron":"haw-lal'","derivation":"a primitive root;","strongs_def":"to be clear (orig. of sound, but usually of color); to shine; hence, to make ashow, to boast; and thus to be (clamorously) foolish; to rave; causatively, to celebrate; also to stultify","kjv_def":"(make) boast (self), celebrate, commend, (deal, make), fool(-ish, -ly), glory, give (light), be (make, feign self) mad (against), give in marriage, (sing, be worthy of) praise, rage, renowned, shine."},
"H1985":{"lemma":"הִלֵּל","xlit":"Hillêl","pron":"hil-layl'","derivation":"from H1984 (הָלַל); praising (namely God);","strongs_def":"Hillel, an Israelite","kjv_def":"Hillel."},
"H1986":{"lemma":"הָלַם","xlit":"hâlam","pron":"haw-lam'","derivation":"a primitive root;","strongs_def":"to strike down; by implication, to hammer, stamp, conquer, disband","kjv_def":"beat (down), break (down), overcome, smite (with the hammer)."},
"H1987":{"lemma":"הֶלֶם","xlit":"Helem","pron":"hay'-lem","derivation":"from H1986 (הָלַם); smiter;","strongs_def":"Helem, the name of two Israelites","kjv_def":"Helem."},
"H1988":{"lemma":"הֲלֹם","xlit":"hălôm","pron":"hal-ome'","derivation":"from the article (see H1973 (הָלְאָה)xlit hâlᵉâh corrected to hâlᵉʼâh);","strongs_def":"hither","kjv_def":"here, hither(-(to)), thither."},
"H1989":{"lemma":"הַלְמוּת","xlit":"halmûwth","pron":"hal-mooth'","derivation":"from H1986 (הָלַם);","strongs_def":"a hammer (or mallet)","kjv_def":"hammer."},
"H1990":{"lemma":"הָם","xlit":"Hâm","pron":"hawm","derivation":"of uncertain derivation;","strongs_def":"Ham, a region of Palestine","kjv_def":"Ham."},
"H1991":{"lemma":"הֵם","xlit":"hêm","pron":"haym","derivation":"from H1993 (הָמָה);","strongs_def":"abundance, i.e. wealth","kjv_def":"any of theirs."},
"H1992":{"lemma":"הֵם","xlit":"hêm","pron":"haym","derivation":"or (prolonged) הֵמָּה; masculine plural from H1981 (הֲלַךְ);","strongs_def":"they (only used when emphatic)","kjv_def":"it, like, [idiom] (how, so) many (soever, more as) they (be), (the) same, [idiom] so, [idiom] such, their, them, these, they, those, which, who, whom, withal, ye."},
"H1993":{"lemma":"הָמָה","xlit":"hâmâh","pron":"haw-maw'","derivation":"a primitive root (compare H1949 (הוּם));","strongs_def":"to make a loud sound (like English 'hum'); by implication, to be in great commotion or tumult, to rage, war, moan, clamor","kjv_def":"clamorous, concourse, cry aloud, be disquieted, loud, mourn, be moved, make a noise, rage, roar, sound, be troubled, make in tumult, tumultuous, be in an uproar."},
"H1994":{"lemma":"הִמּוֹ","xlit":"himmôw","pron":"him-mo'","derivation":"(Aramaic) or (prolonged) הִמּוֹן; (Aramaic), corresponding to H1992 (הֵם);","strongs_def":"they","kjv_def":"[idiom] are, them, those."},
"H1995":{"lemma":"הָמוֹן","xlit":"hâmôwn","pron":"haw-mone'","derivation":"or הָמֹן; (Ezekiel 5:7), from H1993 (הָמָה);","strongs_def":"a noise, tumult, crowd; also disquietude, wealth","kjv_def":"abundance, company, many, multitude, multiply, noise, riches, rumbling, sounding, store, tumult."},
"H1996":{"lemma":"הֲמוֹן גּוֹג","xlit":"Hămôwn Gôwg","pron":"ham-one' gohg","derivation":"from H1995 (הָמוֹן) and H1463 (גּוֹג);","strongs_def":"the multitude of Gog; the fanciful name of an emblematic place in Palestine","kjv_def":"Hamogog."},
"H1997":{"lemma":"הֲמוֹנָה","xlit":"Hămôwnâh","pron":"ham-o-naw'","derivation":"feminine of H1995 (הָמוֹן); multitude; , the same as H1996 (הֲמוֹן גּוֹג)","strongs_def":"Hamonah","kjv_def":"Hamonah."},
"H1998":{"lemma":"הֶמְיָה","xlit":"hemyâh","pron":"hem-yaw'","derivation":"from H1993 (הָמָה);","strongs_def":"sound","kjv_def":"noise."},
"H1999":{"lemma":"הֲמֻלָּה","xlit":"hămullâh","pron":"ham-ool-law'","derivation":"or (too fully) הֲמוּלָּה; (Jeremiah 11:16), feminine passive participle of an unused root meaning to rush (as rain with a windy roar);","strongs_def":"a sound","kjv_def":"speech, tumult."},
"H2000":{"lemma":"הָמַם","xlit":"hâmam","pron":"haw-mam'","derivation":"a primitive root (compare H1949 (הוּם), H1993 (הָמָה));","strongs_def":"properly, to put in commotion; by implication, to disturb, drive, destroy","kjv_def":"break, consume, crush, destroy, discomfit, trouble, vex."},
"H2001":{"lemma":"הָמָן","xlit":"Hâmân","pron":"haw-mawn'","derivation":"of foreign derivation;","strongs_def":"Haman, a Persian vizier","kjv_def":"Haman."},
"H2002":{"lemma":"הַמְנִיךְ","xlit":"hamnîyk","pron":"ham-neek'","derivation":"(Aramaic) but the text is הֲמוּנֵךְ; of foreign origin;","strongs_def":"a necklace","kjv_def":"chain."},
"H2003":{"lemma":"הָמָס","xlit":"hâmâç","pron":"haw-mawce'","derivation":"from an unused root apparently meaning to crackle;","strongs_def":"a dry twig or brushwood","kjv_def":"melting."},
"H2004":{"lemma":"הֵן","xlit":"hên","pron":"hane","derivation":"feminine plural from H1931 (הוּא);","strongs_def":"they (only used when emphatic)","kjv_def":"[idiom] in, such like, (with) them, thereby, therein, (more than) they, wherein, in which, whom, withal."},
"H2005":{"lemma":"הֵן","xlit":"hên","pron":"hane","derivation":"a primitive particle;","strongs_def":"lo!; also (as expressing surprise) if","kjv_def":"behold, if, lo, though."},
"H2006":{"lemma":"הֵן","xlit":"hên","pron":"hane","derivation":"(Aramaic) corresponding to H2005 (הֵן);","strongs_def":"lo! also there(-fore), (un-) less, whether, but, if","kjv_def":"(that) if, or, whether."},
"H2007":{"lemma":"הֵנָּה","xlit":"hênnâh","pron":"hane'-naw","derivation":"prolongation for H2004 (הֵן);","strongs_def":"themselves (often used emphatic for the copula, also in indirect relation)","kjv_def":"[idiom] in, [idiom] such (and such things), their, (into) them, thence, therein, these, they (had), on this side, whose, wherein."},
"H2008":{"lemma":"הֵנָּה","xlit":"hênnâh","pron":"hane'-naw","derivation":"from H2004 (הֵן);","strongs_def":"hither or thither (but used both of place and time)","kjv_def":"here, hither(-to), now, on this (that) side, [phrase] since, this (that) way, thitherward, [phrase] thus far, to...fro, [phrase] yet."},
"H2009":{"lemma":"הִנֵּה","xlit":"hinnêh","pron":"hin-nay'","derivation":"prolongation for H2005 (הֵן);","strongs_def":"lo!","kjv_def":"behold, lo, see."},
"H2010":{"lemma":"הֲנָחָה","xlit":"hănâchâh","pron":"han-aw-khaw'","derivation":"from H5117 (נוּחַ);","strongs_def":"permission of rest, i.e. quiet","kjv_def":"release."},
"H2011":{"lemma":"הִנֹּם","xlit":"Hinnôm","pron":"hin-nome'","derivation":"probably of foreign origin;","strongs_def":"Hinnom, apparently a Jebusite","kjv_def":"Hinnom."},
"H2012":{"lemma":"הֵנַע","xlit":"Hênaʻ","pron":"hay-nah'","derivation":"probably of foreign derivation;","strongs_def":"Hena, a place apparently in Mesopotamia","kjv_def":"Hena."},
"H2013":{"lemma":"הָסָה","xlit":"hâçâh","pron":"haw-saw'","derivation":"a primitive root;","strongs_def":"to hush","kjv_def":"hold peace (tongue), (keep) silence, be silent, still."},
"H2014":{"lemma":"הֲפֻגָה","xlit":"hăphugâh","pron":"haf-oo-gaw'","derivation":"from H6313 (פּוּג);","strongs_def":"relaxation","kjv_def":"intermission."},
"H2015":{"lemma":"הָפַךְ","xlit":"hâphak","pron":"haw-fak'","derivation":"a primitive root;","strongs_def":"to turn about or over; by implication, to change, overturn, return, pervert","kjv_def":"[idiom] become, change, come, be converted, give, make (a bed), overthrow (-turn), perverse, retire, tumble, turn (again, aside, back, to the contrary, every way)."},
"H2016":{"lemma":"הֶפֶךְ","xlit":"hephek","pron":"heh'-fek","derivation":"; or הֵפֶךְ; from H2015 (הָפַךְ);","strongs_def":"a turn, i.e. the reverse","kjv_def":"contrary."},
"H2017":{"lemma":"הֹפֶךְ","xlit":"hôphek","pron":"ho'-fek","derivation":"from H2015 (הָפַךְ);","strongs_def":"an upset, i.e. (abstract) perversity","kjv_def":"turning of things upside down."},
"H2018":{"lemma":"הֲפֵכָה","xlit":"hăphêkâh","pron":"haf-ay-kaw'","derivation":"feminine of H2016 (הֶפֶךְ);","strongs_def":"destruction","kjv_def":"overthrow."},
"H2019":{"lemma":"הֲפַכְפַּךְ","xlit":"hăphakpak","pron":"haf-ak-pak'","derivation":"by reduplication from H2015 (הָפַךְ);","strongs_def":"very perverse","kjv_def":"froward."},
"H2020":{"lemma":"הַצָּלָה","xlit":"hatstsâlâh","pron":"hats-tsaw-law'","derivation":"from H5337 (נָצַל);","strongs_def":"rescue","kjv_def":"deliverance."},
"H2021":{"lemma":"הֹצֶן","xlit":"hôtsen","pron":"ho'-tsen","derivation":"from an unused root meaning apparently to be sharp or strong;","strongs_def":"a weapon of war","kjv_def":"chariot."},
"H2022":{"lemma":"הַר","xlit":"har","pron":"har","derivation":"a shortened form of H2042 (הָרָר);","strongs_def":"a mountain or range of hills (sometimes used figuratively)","kjv_def":"hill (country), mount(-ain), [idiom] promotion."},
"H2023":{"lemma":"הֹר","xlit":"Hôr","pron":"hore","derivation":"another form of H2022 (הַר); mountain;","strongs_def":"Hor, the name of a peak in Idumaea and of one in Syria","kjv_def":"Hor."},
"H2024":{"lemma":"הָרָא","xlit":"Hârâʼ","pron":"haw-raw'","derivation":"perhaps from H2022 (הַר); mountainousness;","strongs_def":"Hara, a region of Media","kjv_def":"Hara."},
"H2025":{"lemma":"הַרְאֵל","xlit":"harʼêl","pron":"har-ale'","derivation":"from H2022 (הַר) and H410 (אֵל);","strongs_def":"mount of God; figuratively, the altar of burnt-offering","kjv_def":"altar. Compare H739 (אֲרִיאֵל)."},
"H2026":{"lemma":"הָרַג","xlit":"hârag","pron":"haw-rag'","derivation":"a primitive root;","strongs_def":"to smite with deadly intent","kjv_def":"destroy, out of hand, kill, murder(-er), put to (death), make (slaughter), slay(-er), [idiom] surely."},
"H2027":{"lemma":"הֶרֶג","xlit":"hereg","pron":"heh'-reg","derivation":"from H2026 (הָרַג);","strongs_def":"slaughter","kjv_def":"be slain, slaughter."},
"H2028":{"lemma":"הֲרֵגָה","xlit":"hărêgâh","pron":"har-ay-gaw'","derivation":"feminine of H2027 (הֶרֶג);","strongs_def":"slaughter","kjv_def":"slaughter."},
"H2029":{"lemma":"הָרָה","xlit":"hârâh","pron":"haw-raw'","derivation":"a primitive root;","strongs_def":"to be (or become) pregnant, conceive (literally or figuratively)","kjv_def":"been, be with child, conceive, progenitor."},
"H2030":{"lemma":"הָרֶה","xlit":"hâreh","pron":"haw-reh'","derivation":"or הָרִי; (Hosea 14:1), from H2029 (הָרָה);","strongs_def":"pregnant","kjv_def":"(be, woman) with child, conceive, [idiom] great."},
"H2031":{"lemma":"הַרְהֹר","xlit":"harhôr","pron":"har-hor'","derivation":"(Aramaic) from a root corresponding to H2029 (הָרָה);","strongs_def":"a mental conception","kjv_def":"thought."},
"H2032":{"lemma":"הֵרוֹן","xlit":"hêrôwn","pron":"hay-rone'","derivation":"or הֵרָיוֹן; from H2029 (הָרָה);","strongs_def":"pregnancy","kjv_def":"conception."},
"H2033":{"lemma":"הֲרוֹרִי","xlit":"Hărôwrîy","pron":"har-o-ree'","derivation":"another form for H2043 (הֲרָרִי);","strongs_def":"a Harorite or mountaineer","kjv_def":"Harorite."},
"H2034":{"lemma":"הֲרִיסָה","xlit":"hărîyçâh","pron":"har-ee-saw'","derivation":"from H2040 (הָרַס);","strongs_def":"something demolished","kjv_def":"ruin."},
"H2035":{"lemma":"הֲרִיסוּת","xlit":"hărîyçûwth","pron":"har-ee-sooth'","derivation":"from H2040 (הָרַס);","strongs_def":"demolition","kjv_def":"destruction."},
"H2036":{"lemma":"הֹרָם","xlit":"Hôrâm","pron":"ho-rawm'","derivation":"from an unused root (meaning to tower up); high;","strongs_def":"Horam, a Canaanitish king","kjv_def":"Horam."},
"H2037":{"lemma":"הָרֻם","xlit":"Hârum","pron":"haw-room'","derivation":"passive participle of the same as H2036 (הֹרָם); high;","strongs_def":"Harum, an Israelite","kjv_def":"Harum."},
"H2038":{"lemma":"הַרְמוֹן","xlit":"harmôwn","pron":"har-mone'","derivation":"from the same as H2036 (הֹרָם);","strongs_def":"a castle (from its height)","kjv_def":"palace."},
"H2039":{"lemma":"הָרָן","xlit":"Hârân","pron":"haw-rawn'","derivation":"perhaps from H2022 (הַר); mountaineer;","strongs_def":"Haran, the name of two men","kjv_def":"Haran."},
"H2040":{"lemma":"הָרַס","xlit":"hâraç","pron":"haw-ras'","derivation":"a primitive root;","strongs_def":"to pull down or in pieces, break, destroy","kjv_def":"beat down, break (down, through), destroy, overthrow, pluck down, pull down, ruin, throw down, [idiom] utterly."},
"H2041":{"lemma":"הֶרֶס","xlit":"hereç","pron":"heh'-res","derivation":"from H2040 (הָרַס);","strongs_def":"demolition","kjv_def":"destruction."},
"H2042":{"lemma":"הָרָר","xlit":"hârâr","pron":"haw-rawr'","derivation":"from an unused root meaning to loom up;","strongs_def":"a mountain","kjv_def":"hill, mount(-ain)."},
"H2043":{"lemma":"הֲרָרִי","xlit":"Hărârîy","pron":"hah-raw-ree'","derivation":"or הָרָרִי; (2 Samuel 23:11), or הָארָרִי; (2 Samuel 23:34, last clause [), apparently from H2042 (הָרָר); a mountaineer","strongs_def":"{Hararite}","kjv_def":"Hararite."},
"H2044":{"lemma":"הָשֵׁם","xlit":"Hâshêm","pron":"haw-shame'","derivation":"perhaps from the same as H2828 (חָשֻׁם); wealthy;","strongs_def":"Hashem, an Israelite","kjv_def":"Hashem."},
"H2045":{"lemma":"הַשְׁמָעוּת","xlit":"hashmâʻûwth","pron":"hashmaw-ooth'","derivation":"from H8085 (שָׁמַע);","strongs_def":"announcement","kjv_def":"to cause to hear."},
"H2046":{"lemma":"הִתּוּךְ","xlit":"hittûwk","pron":"hit-took'","derivation":"from H5413 (נָתַךְ);","strongs_def":"a melting","kjv_def":"is melted."},
"H2047":{"lemma":"הֲתָךְ","xlit":"Hăthâk","pron":"hath-awk'","derivation":"probably of foreign origin;","strongs_def":"Hathak, a Persian eunuch","kjv_def":"Hatach."},
"H2048":{"lemma":"הָתַל","xlit":"hâthal","pron":"haw-thal'","derivation":"a primitive root;","strongs_def":"to deride; by implication, to cheat","kjv_def":"deal deceitfully, deceive, mock."},
"H2049":{"lemma":"הָתֹל","xlit":"hâthôl","pron":"haw-thole'","derivation":"from H2048 (הָתַל) (only in plural collectively);","strongs_def":"a derision","kjv_def":"mocker."},
"H2050":{"lemma":"הָתַת","xlit":"hâthath","pron":"haw-thath'","derivation":"a primitive root;","strongs_def":"properly, to break in upon, i.e. to assail","kjv_def":"imagine mischief. w"},
"H2051":{"lemma":"וְדָן","xlit":"Vᵉdân","pron":"ved-awn'","derivation":"perhaps for H5730 (עֵדֶן);","strongs_def":"Vedan (or Aden), a place in Arabia","kjv_def":"Dan also."},
"H2052":{"lemma":"וָהֵב","xlit":"Vâhêb","pron":"vaw-habe'","derivation":"of uncertain derivation;","strongs_def":"Vaheb, a place in Moab","kjv_def":"what he did."},
"H2053":{"lemma":"וָו","xlit":"vâv","pron":"vaw","derivation":"probably","strongs_def":"a hook (the name of the sixth Hebrew letter)","kjv_def":"hook."},
"H2054":{"lemma":"וָזָר","xlit":"vâzâr","pron":"vaw-zawr'","derivation":"presumed to be from an unused root meaning to bear guilt;","strongs_def":"crime","kjv_def":"[idiom] strange."},
"H2055":{"lemma":"וַיְזָתָא","xlit":"Vayzâthâʼ","pron":"vah-yez-aw'-thaw","derivation":"of foreign origin;","strongs_def":"Vajezatha, a son of Haman","kjv_def":"Vajezatha."},
"H2056":{"lemma":"וָלָד","xlit":"vâlâd","pron":"vaw-lawd'","derivation":"for H3206 (יֶלֶד);","strongs_def":"a boy","kjv_def":"child."},
"H2057":{"lemma":"וַנְיָה","xlit":"Vanyâh","pron":"van-yaw'","derivation":"perhaps for H6043 (עֲנָיָה);","strongs_def":"Vanjah, an Israelite","kjv_def":"Vaniah."},
"H2058":{"lemma":"וׇפְסִי","xlit":"Vophçîy","pron":"vof-see'","derivation":"probably from H3254 (יָסַף); additional;","strongs_def":"Vophsi, an Israelite","kjv_def":"Vophsi."},
"H2059":{"lemma":"וַשְׁנִי","xlit":"Vashnîy","pron":"vash-nee'","derivation":"probably from H3461 (יִשְׁמְרַי); weak;","strongs_def":"Vashni, an Israelite","kjv_def":"Vashni."},
"H2060":{"lemma":"וַשְׁתִּי","xlit":"Vashtîy","pron":"vash-tee'","derivation":"of Persian origin;","strongs_def":"Vashti, the queen of Xerxes","kjv_def":"Vashti. z"},
"H2061":{"lemma":"זְאֵב","xlit":"zᵉʼêb","pron":"zeh-abe'","derivation":"from an unused root meaning to be yellow;","strongs_def":"a wolf","kjv_def":"wolf."},
"H2062":{"lemma":"זְאֵב","xlit":"Zᵉʼêb","pron":"zeh-abe'","derivation":"the same as H2061 (זְאֵב);","strongs_def":"Zeeb, a Midianitish prince","kjv_def":"Zeeb."},
"H2063":{"lemma":"זֹאת","xlit":"zôʼth","pron":"zothe'","derivation":"irregular feminine of H2089 (זֶה);","strongs_def":"this (often used adverb)","kjv_def":"hereby (-in, -with), it, likewise, the one (other, same), she, so (much), such (deed), that, therefore, these, this (thing), thus."},
"H2064":{"lemma":"זָבַד","xlit":"zâbad","pron":"zaw-bad'","derivation":"a primitive root;","strongs_def":"to confer","kjv_def":"endure."},
"H2065":{"lemma":"זֶבֶד","xlit":"zebed","pron":"zeh'-bed","derivation":"from H2064 (זָבַד);","strongs_def":"a gift","kjv_def":"dowry."},
"H2066":{"lemma":"זָבָד","xlit":"Zâbâd","pron":"zaw-bawd'","derivation":"from H2064 (זָבַד); giver;","strongs_def":"Zabad, the name of seven Israelites","kjv_def":"Zabad."},
"H2067":{"lemma":"זַבְדִּי","xlit":"Zabdîy","pron":"zab-dee'","derivation":"from H2065 (זֶבֶד); giving;","strongs_def":"Zabdi, the name of four Israelites","kjv_def":"Zabdi."},
"H2068":{"lemma":"זַבְדִּיאֵל","xlit":"Zabdîyʼêl","pron":"zab-dee-ale'","derivation":"from H2065 (זֶבֶד) and H410 (אֵל);","strongs_def":"gift of God; zabdiel, the name of two Israelites","kjv_def":"Zabdiel."},
"H2069":{"lemma":"זְבַדְיָה","xlit":"Zᵉbadyâh","pron":"zeb-ad-yaw'","derivation":"or זְבַדְיָהוּ; from H2064 (זָבַד) and H3050 (יָהּ); Jah has given;","strongs_def":"Zebadjah, the name of nine Israelites","kjv_def":"Zebadiah."},
"H2070":{"lemma":"זְבוּב","xlit":"zᵉbûwb","pron":"zeb-oob'","derivation":"from an unused root (meaning to flit);","strongs_def":"a fly (especially one of a stinging nature)","kjv_def":"fly."},
"H2071":{"lemma":"זָבוּד","xlit":"Zâbûwd","pron":"zaw-bood'","derivation":"from H2064 (זָבַד);","strongs_def":"given, Zabud, an Israelite","kjv_def":"Zabud."},
"H2072":{"lemma":"זַבּוּד","xlit":"Zabbûwd","pron":"zab-bood'","derivation":"a form of H2071 (זָבוּד); given;","strongs_def":"Zabbud, an Israelite","kjv_def":"Zabbud."},
"H2073":{"lemma":"זְבוּל","xlit":"zᵉbûwl","pron":"ze-bool'","derivation":"or זְבֻל; from H2082 (זָבַל);","strongs_def":"a residence","kjv_def":"dwell in, dwelling, habitation."},
"H2074":{"lemma":"זְבוּלוּן","xlit":"Zᵉbûwlûwn","pron":"zeb-oo-loon'","derivation":"or זְבֻלוּן; or זְבוּלֻן; from H2082 (זָבַל); habitation;","strongs_def":"Zebulon, a son of Jacob; also his territory and tribe","kjv_def":"Zebulun."},
"H2075":{"lemma":"זְבוּלֹנִי","xlit":"Zᵉbûwlônîy","pron":"zeb-oo-lo-nee'","derivation":"patronymically from H2074 (זְבוּלוּן);","strongs_def":"a Zebulonite or descendant of Zebulun","kjv_def":"Zebulonite."},
"H2076":{"lemma":"זָבַח","xlit":"zâbach","pron":"zaw-bakh'","derivation":"a primitive root;","strongs_def":"to slaughter an animal (usually in sacrifice)","kjv_def":"kill, offer, (do) sacrifice, slay."},
"H2077":{"lemma":"זֶבַח","xlit":"zebach","pron":"zeh'-bakh","derivation":"from H2076 (זָבַח);","strongs_def":"properly, a slaughter, i.e. the flesh of an animal; by implication, a sacrifice (the victim or the act)","kjv_def":"offer(-ing), sacrifice."},
"H2078":{"lemma":"זֶבַח","xlit":"Zebach","pron":"zeh'-bakh","derivation":"the same as H2077 (זֶבַח); sacrifice;","strongs_def":"Zebach, a Midianitish prince","kjv_def":"Zebah."},
"H2079":{"lemma":"זַבַּי","xlit":"Zabbay","pron":"zab-bah'-ee","derivation":"probably by orthographical error for H2140 (זַכַּי);","strongs_def":"Zabbai (or Zaccai), an Israelite","kjv_def":"Zabbai."},
"H2080":{"lemma":"זְבִידָה","xlit":"Zᵉbîydâh","pron":"zeb-ee-daw'","derivation":"feminine from H2064 (זָבַד); giving;","strongs_def":"Zebidah, an Israelitess","kjv_def":"Zebudah."},
"H2081":{"lemma":"זְבִינָא","xlit":"Zᵉbîynâʼ","pron":"zeb-ee-naw'","derivation":"from an unused root (meaning to purchase); gainfulness;","strongs_def":"Zebina, an Israelite","kjv_def":"Zebina."},
"H2082":{"lemma":"זָבַל","xlit":"zâbal","pron":"zaw-bal'","derivation":"a primitive root; apparently properly, to inclose, i.e.","strongs_def":"to reside","kjv_def":"dwell with."},
"H2083":{"lemma":"זְבֻל","xlit":"Zᵉbul","pron":"zeb-ool'","derivation":"the same as H2073 (זְבוּל); dwelling;","strongs_def":"Zebul, an Israelite","kjv_def":"Zebul. Compare H2073 (זְבוּל)."},
"H2084":{"lemma":"זְבַן","xlit":"zᵉban","pron":"zeb-an'","derivation":"(Aramaic) corresponding to the root of H2081 (זְבִינָא);","strongs_def":"to acquire by purchase","kjv_def":"gain."},
"H2085":{"lemma":"זָג","xlit":"zâg","pron":"zawg","derivation":"from an unused root probably meaning to inclose;","strongs_def":"the skin of a grape","kjv_def":"husk."},
"H2086":{"lemma":"זֵד","xlit":"zêd","pron":"zade'","derivation":"from H2102 (זוּד);","strongs_def":"arrogant","kjv_def":"presumptuous, proud."},
"H2087":{"lemma":"זָדוֹן","xlit":"zâdôwn","pron":"zaw-done'","derivation":"from H2102 (זוּד);","strongs_def":"arrogance","kjv_def":"presumptuously, pride, proud (man)."},
"H2088":{"lemma":"זֶה","xlit":"zeh","pron":"zeh","derivation":"a primitive word;","strongs_def":"the masculine demonstrative pronoun, this or that","kjv_def":"he, [idiom] hence, [idiom] here, it(-self), [idiom] now, [idiom] of him, the one...the other, [idiom] than the other, ([idiom] out of) the (self) same, such (a one) that, these, this (hath, man), on this side...on that side, [idiom] thus, very, which. Compare H2063 (זֹאת), H2090 (זֹה), H2097 (זוֹ), H2098 (זוּ)."},
"H2089":{"lemma":"זֶה","xlit":"zeh","pron":"zeh","derivation":"(1 Samuel by permutation for H7716 (שֶׂה);","strongs_def":"a sheep","kjv_def":"lamb."},
"H2090":{"lemma":"זֹה","xlit":"zôh","pron":"zo","derivation":"for H2088 (זֶה);","strongs_def":"this or that","kjv_def":"as well as another, it, this, that, thus and thus."},
"H2091":{"lemma":"זָהָב","xlit":"zâhâb","pron":"zaw-hawb'","derivation":"from an unused root meaning to shimmer;","strongs_def":"gold, figuratively, something gold-colored (i.e. yellow), as oil, a clear sky","kjv_def":"gold(-en), fair weather."},
"H2092":{"lemma":"זָהַם","xlit":"zâham","pron":"zaw-ham'","derivation":"a primitive root;","strongs_def":"to be rancid, i.e. (transitively) to loathe","kjv_def":"abhor."},
"H2093":{"lemma":"זַהַם","xlit":"Zaham","pron":"zah'-ham","derivation":"from H2092 (זָהַם); loathing;","strongs_def":"Zaham, an Israelite","kjv_def":"Zaham."},
"H2094":{"lemma":"זָהַר","xlit":"zâhar","pron":"zaw-har'","derivation":"a primitive root;","strongs_def":"to gleam; figuratively, to enlighten (by caution)","kjv_def":"admonish, shine, teach, (give) warn(-ing)."},
"H2095":{"lemma":"זְהַר","xlit":"zᵉhar","pron":"zeh-har'","derivation":"(Aramaic) corresponding to H2094 (זָהַר);","strongs_def":"(passively) be admonished","kjv_def":"take heed."},
"H2096":{"lemma":"זֹהַר","xlit":"zôhar","pron":"zo'-har","derivation":"from H2094 (זָהַר);","strongs_def":"brilliancy","kjv_def":"brightness."},
"H2097":{"lemma":"זוֹ","xlit":"zôw","pron":"zo","derivation":"for H2088 (זֶה);","strongs_def":"this or that","kjv_def":"that, this."},
"H2098":{"lemma":"זוּ","xlit":"zûw","pron":"zoo","derivation":"for H2088 (זֶה);","strongs_def":"this or that","kjv_def":"that, this, [idiom] wherein, which, whom."},
"H2099":{"lemma":"זִו","xlit":"Ziv","pron":"zeev'","derivation":"probably from an unused root meaning to be prominent; properly, brightness (compare H2122 (זִיו)), i.e. (figuratively) the month of flowers;","strongs_def":"Ziv (corresponding to Ijar or May)","kjv_def":"Zif."},
"H2100":{"lemma":"זוּב","xlit":"zûwb","pron":"zoob","derivation":"a primitive root;","strongs_def":"to flow freely (as water), i.e. (specifically) to have a (sexual) flux; figuratively, to waste away; also to overflow","kjv_def":"flow, gush out, have a (running) issue, pine away, run."},
"H2101":{"lemma":"זוֹב","xlit":"zôwb","pron":"zobe","derivation":"from H2100 (זוּב);","strongs_def":"a seminal or menstrual flux","kjv_def":"issue."},
"H2102":{"lemma":"זוּד","xlit":"zûwd","pron":"zood","derivation":"or (by permutation) זִיד; a primitive root;","strongs_def":"to seethe; figuratively, to be insolent","kjv_def":"be proud, deal proudly, presume, (come) presumptuously, sod."},
"H2103":{"lemma":"זוּד","xlit":"zûwd","pron":"zood","derivation":"(Aramaic) corresponding to H2102 (זוּד);","strongs_def":"to be proud","kjv_def":"in pride."},
"H2104":{"lemma":"זוּזִים","xlit":"Zûwzîym","pron":"zoo-zeem'","derivation":"plural probably from the same as H2123 (זִיז); prominent;","strongs_def":"Zuzites, an aboriginal tribe of Palestine","kjv_def":"Zuzims."},
"H2105":{"lemma":"זוֹחֵת","xlit":"Zôwchêth","pron":"zo-khayth'","derivation":"of uncertain origin;","strongs_def":"Zocheth, an Israelite","kjv_def":"Zoheth."},
"H2106":{"lemma":"זָוִית","xlit":"zâvîyth","pron":"zaw-veeth'","derivation":"apparently from the same root as H2099 (זִו) (in the sense of prominence);","strongs_def":"an angle (as projecting), i.e. (by implication) a corner-column (or anta)","kjv_def":"corner(stone)."},
"H2107":{"lemma":"זוּל","xlit":"zûwl","pron":"zool","derivation":"a primitive root (compare H2151 (זָלַל));","strongs_def":"probably to shake out, i.e. (by implication) to scatter profusely; figuratively, to treat lightly","kjv_def":"lavish, despise."},
"H2108":{"lemma":"זוּלָה","xlit":"zûwlâh","pron":"zoo-law'","derivation":"from H2107 (זוּל);","strongs_def":"probably scattering, i.e. removal; used adverbially, except","kjv_def":"beside, but, only, save."},
"H2109":{"lemma":"זוּן","xlit":"zûwn","pron":"zoon","derivation":"a primitive root;","strongs_def":"perhaps properly, to be plump, i.e. (transitively) to nourish","kjv_def":"feed."},
"H2110":{"lemma":"זוּן","xlit":"zûwn","pron":"zoon","derivation":"(Aramaic) corresponding to H2109 (זוּן)","strongs_def":"{perhaps properly, to be plump, i.e. (transitively) to nourish}","kjv_def":"feed."},
"H2111":{"lemma":"זוּעַ","xlit":"zûwaʻ","pron":"zoo'-ah","derivation":"a primitive root;","strongs_def":"properly, to shake off, i.e. (figuratively) to agitate (as with fear)","kjv_def":"move, tremble, vex."},
"H2112":{"lemma":"זוּעַ","xlit":"zûwaʻ","pron":"zoo'-ah","derivation":"(Aramaic) corresponding to H2111 (זוּעַ);","strongs_def":"to shake (with fear)","kjv_def":"tremble."},
"H2113":{"lemma":"זְוָעָה","xlit":"zᵉvâʻâh","pron":"zev-aw-aw'","derivation":"from H2111 (זוּעַ);","strongs_def":"agitation, fear","kjv_def":"be removed, trouble, vexation. Compare H2189 (זַעֲוָה)."},
"H2114":{"lemma":"זוּר","xlit":"zûwr","pron":"zoor","derivation":"a primitive root;","strongs_def":"to turn aside (especially for lodging); hence to be aforeigner, strange, profane; specifically (active participle) to commit adultery","kjv_def":"(come from) another (man, place), fanner, go away, (e-) strange(-r, thing, woman)."},
"H2115":{"lemma":"זוּר","xlit":"zûwr","pron":"zoor","derivation":"a primitive root (compare H6695 (צוֹק));","strongs_def":"to press together, tighten","kjv_def":"close, rush, thrust together."},
"H2116":{"lemma":"זוּרֶה","xlit":"zûwreh","pron":"zoo-reh'","derivation":"from H2115 (זוּר);","strongs_def":"trodden on","kjv_def":"that which is crushed."},
"H2117":{"lemma":"זָזָא","xlit":"zâzâʼ","pron":"zaw-zaw'","derivation":"probably from the root of H2123 (זִיז); prominent;","strongs_def":"Zaza, an Israelite","kjv_def":"Zaza."},
"H2118":{"lemma":"זָחַח","xlit":"zâchach","pron":"zaw-khakh'","derivation":"a primitive root;","strongs_def":"to shove or displace","kjv_def":"loose."},
"H2119":{"lemma":"זָחַל","xlit":"zâchal","pron":"zaw-khal'","derivation":"a primitive root;","strongs_def":"to crawl; by implication, to fear","kjv_def":"be afraid, serpent, worm."},
"H2120":{"lemma":"זֹחֶלֶת","xlit":"Zôcheleth","pron":"zo-kheh'-leth","derivation":"feminine active participle of H2119 (זָחַל); crawling (i.e. serpent);","strongs_def":"Zocheleth, a boundary stone in. Palestine","kjv_def":"Zoheleth."},
"H2121":{"lemma":"זֵידוֹן","xlit":"zêydôwn","pron":"zay-dohn'","derivation":"from H2102 (זוּד);","strongs_def":"boiling of water, i.e. wave","kjv_def":"proud."},
"H2122":{"lemma":"זִיו","xlit":"zîyv","pron":"zeev","derivation":"(Aramaic) corresponding to H2099 (זִו);","strongs_def":"(figuratively) cheerfulness","kjv_def":"brightness, countenance."},
"H2123":{"lemma":"זִיז","xlit":"zîyz","pron":"zeez","derivation":"from an unused root apparently meaning to be conspicuous;","strongs_def":"fulness of the breast; also a moving creature","kjv_def":"abundance, wild beast."},
"H2124":{"lemma":"זִיזָא","xlit":"Zîyzâʼ","pron":"zee-zaw'","derivation":"apparently from the same as H2123 (זִיז); prominence;","strongs_def":"Ziza, the name of two Israelites","kjv_def":"Ziza."},
"H2125":{"lemma":"זִיזָה","xlit":"Zîyzâh","pron":"zee-zaw'","derivation":"another form for H2124 (זִיזָא);","strongs_def":"Zizah, an Israelite","kjv_def":"Zizah."},
"H2126":{"lemma":"זִינָא","xlit":"Zîynâʼ","pron":"zee-naw'","derivation":"from H2109 (זוּן); well-fed; or perhaps an orthographical error for H2124 (זִיזָא);","strongs_def":"Zina, an Israelite","kjv_def":"Zina."},
"H2127":{"lemma":"זִיעַ","xlit":"Zîyaʻ","pron":"zee'-ah","derivation":"from H2111 (זוּעַ); agitation;","strongs_def":"Zia, an Israelite","kjv_def":"Zia."},
"H2128":{"lemma":"זִיף","xlit":"Zîyph","pron":"zeef","derivation":"from the same as H2203 (זֶפֶת); flowing;","strongs_def":"Ziph, the name of a place in Palestine; also of an Israelite","kjv_def":"Ziph."},
"H2129":{"lemma":"זִיפָה","xlit":"Zîyphâh","pron":"zee-faw'","derivation":"feminine of H2128 (זִיף); a flowing;","strongs_def":"Ziphah, an Israelite","kjv_def":"Ziphah."},
"H2130":{"lemma":"זִיפִי","xlit":"Zîyphîy","pron":"zee-fee'","derivation":"patrial from H2128 (זִיף);","strongs_def":"a Ziphite or inhabitant of Ziph","kjv_def":"Ziphim, Ziphite."},
"H2131":{"lemma":"זִיקָה","xlit":"zîyqâh","pron":"zee-kaw'","derivation":"(Isaiah 50: (feminine); and זִק; or זֵק; from H2187 (זָנַק);","strongs_def":"properly, what leaps forth, i.e. flash of fire, or a burning arrow; also (from the original sense of the root) a bond","kjv_def":"chain, fetter, firebrand, spark."},
"H2132":{"lemma":"זַיִת","xlit":"zayith","pron":"zay'-yith","derivation":"probably from an unused root (akin to H2099 (זִו));","strongs_def":"an olive (as yielding illuminating oil), the tree, the branch or the berry","kjv_def":"olive (tree, -yard), Olivet."},
"H2133":{"lemma":"זֵיתָן","xlit":"Zêythân","pron":"zay-thawn'","derivation":"from H2132 (זַיִת); olive grove;","strongs_def":"Zethan, an Israelite","kjv_def":"Zethan."},
"H2134":{"lemma":"זַךְ","xlit":"zak","pron":"zak","derivation":"from H2141 (זָכַךְ);","strongs_def":"clear","kjv_def":"clean, pure."},
"H2135":{"lemma":"זָכָה","xlit":"zâkâh","pron":"zaw-kaw'","derivation":"a primitive root (compare H2141 (זָכַךְ));","strongs_def":"to be translucent; figuratively, to be innocent","kjv_def":"be (make) clean, cleanse, be clear, count pure."},
"H2136":{"lemma":"זָכוּ","xlit":"zâkûw","pron":"zaw-koo'","derivation":"(Aramaic) from a root corresponding to H2135 (זָכָה);","strongs_def":"purity","kjv_def":"innocency."},
"H2137":{"lemma":"זְכוּכִית","xlit":"zᵉkûwkîyth","pron":"zek-oo-keeth","derivation":"from H2135 (זָכָה);","strongs_def":"properly, transparency, i.e. glass","kjv_def":"crystal."},
"H2138":{"lemma":"זָכוּר","xlit":"zâkûwr","pron":"zaw-koor'","derivation":"properly, passive participle of H2142 (זָכַר), but used for H2145 (זָכָר);","strongs_def":"a male (of man or animals)","kjv_def":"males, men-children."},
"H2139":{"lemma":"זַכּוּר","xlit":"Zakkûwr","pron":"zaw-koor'","derivation":"from H2142 (זָכַר); mindful;","strongs_def":"Zakkur, the name of seven Israelites","kjv_def":"Zaccur, Zacchur."},
"H2140":{"lemma":"זַכַּי","xlit":"Zakkay","pron":"zak-kah'-ee","derivation":"from H2141 (זָכַךְ); pure;","strongs_def":"Zakkai, an Israelite","kjv_def":"Zaccai."},
"H2141":{"lemma":"זָכַךְ","xlit":"zâkak","pron":"zaw-kak'","derivation":"a primitive root (compare H2135 (זָכָה));","strongs_def":"to be transparent or clean (phys. or morally)","kjv_def":"be (make) clean, be pure(-r)."},
"H2142":{"lemma":"זָכַר","xlit":"zâkar","pron":"zaw-kar'","derivation":"a primitive root; also as denominative from H2145 (זָכָר)","strongs_def":"properly, to mark (so as to be recognized), i.e. to remember; by implication, to mention;  to be male","kjv_def":"[idiom] burn (incense), [idiom] earnestly, be male, (make) mention (of), be mindful, recount, record(-er), remember, make to be remembered, bring (call, come, keep, put) to (in) remembrance, [idiom] still, think on, [idiom] well."},
"H2143":{"lemma":"זֵכֶר","xlit":"zêker","pron":"zay'-ker","derivation":"or זֶכֶר; from H2142 (זָכַר);","strongs_def":"a memento, abstractly recollection (rarely if ever); by implication, commemoration","kjv_def":"memorial, memory, remembrance, scent."},
"H2144":{"lemma":"זֶכֶר","xlit":"Zeker","pron":"zeh'-ker","derivation":"the same as H2143 (זֵכֶר);","strongs_def":"Zeker, an Israelite","kjv_def":"Zeker."},
"H2145":{"lemma":"זָכָר","xlit":"zâkâr","pron":"zaw-kawr'","derivation":"from H2142 (זָכַר);","strongs_def":"properly, remembered, i.e. a male (of man or animals, as being the most noteworthy sex)","kjv_def":"[idiom] him, male, man(child, -kind)."},
"H2146":{"lemma":"זִכְרוֹן","xlit":"zikrôwn","pron":"zik-rone'","derivation":"from H2142 (זָכַר);","strongs_def":"a memento (or memorable thing, day or writing)","kjv_def":"memorial, record."},
"H2147":{"lemma":"זִכְרִי","xlit":"Zikrîy","pron":"zik-ree'","derivation":"from H2142 (זָכַר); memorable;","strongs_def":"Zicri, the name of twelve Israelites","kjv_def":"Zichri."},
"H2148":{"lemma":"זְכַרְיָה","xlit":"Zᵉkaryâh","pron":"zek-ar-yaw'","derivation":"or זְכַרְיָהוּ; from H2142 (זָכַר) and H3050 (יָהּ); Jah has remembered;","strongs_def":"Zecarjah, the name of twenty-nine Israelites","kjv_def":"Zachariah, Zechariah."},
"H2149":{"lemma":"זֻלּוּת","xlit":"zullûwth","pron":"zool-looth'","derivation":"from H2151 (זָלַל);","strongs_def":"properly, a shaking, i.e. perhaps a tempest","kjv_def":"vilest."},
"H2150":{"lemma":"זַלְזַל","xlit":"zalzal","pron":"zal-zal'","derivation":"by reduplication from H2151 (זָלַל);","strongs_def":"tremulous, i.e. a twig","kjv_def":"sprig."},
"H2151":{"lemma":"זָלַל","xlit":"zâlal","pron":"zaw-lal'","derivation":"a primitive root (compare H2107 (זוּל));","strongs_def":"to shake (as in the wind), i.e. to quake; figuratively, to be loose morally, worthless or prodigal","kjv_def":"blow down, glutton, riotous (eater), vile."},
"H2152":{"lemma":"זַלְעָפָה","xlit":"zalʻâphâh","pron":"zal-aw-faw'","derivation":"or זִלְעָפָהxlit zilʻâphâph corrected to zilʻâphâh; from H2196 (זָעַף);","strongs_def":"a glow (of wind or anger); also a famine (as consuming)","kjv_def":"horrible, horror, terrible."},
"H2153":{"lemma":"זִלְפָּה","xlit":"Zilpâh","pron":"zil-paw","derivation":"; from an unused root apparently meaning to trickle, as myrrh; fragrant dropping;","strongs_def":"Zilpah, Leah's maid","kjv_def":"Zilpah."},
"H2154":{"lemma":"זִמָּה","xlit":"zimmâh","pron":"zim-maw'","derivation":"or זַמָּה; from H2161 (זָמַם);","strongs_def":"a plan, especially a bad one","kjv_def":"heinous crime, lewd(-ly, -ness), mischief, purpose, thought, wicked (device, mind, -ness)."},
"H2155":{"lemma":"זִמָּה","xlit":"Zimmâh","pron":"zim-maw'","derivation":"the same as H2154 (זִמָּה);","strongs_def":"Zimmah, the name of two Israelites","kjv_def":"Zimmah."},
"H2156":{"lemma":"זְמוֹרָה","xlit":"zᵉmôwrâh","pron":"zem-o-raw'","derivation":"or זְמֹרָה (feminine); and זְמֹר (masculine); from H2168 (זָמַר);","strongs_def":"a twig (as pruned)","kjv_def":"vine, branch, slip."},
"H2157":{"lemma":"זַמְזֹם","xlit":"Zamzôm","pron":"zam-zome'","derivation":"from H2161 (זָמַם); intriguing;","strongs_def":"a Zamzumite, or native tribe of Palestine","kjv_def":"Zamzummim."},
"H2158":{"lemma":"זָמִיר","xlit":"zâmîyr","pron":"zaw-meer'","derivation":"or זָמִר; and (feminine) זְמִרָה; from H2167 (זָמַר);","strongs_def":"a song to be accompanied with instrumental music","kjv_def":"psalm(-ist), singing, song."},
"H2159":{"lemma":"זָמִיר","xlit":"zâmîyr","pron":"zaw-meer'","derivation":"from H2168 (זָמַר);","strongs_def":"a twig (as pruned)","kjv_def":"branch."},
"H2160":{"lemma":"זְמִירָה","xlit":"Zᵉmîyrâh","pron":"zem-ee-raw'","derivation":"feminine of H2158 (זָמִיר); song;","strongs_def":"Zemirah, an Israelite","kjv_def":"Zemira."},
"H2161":{"lemma":"זָמַם","xlit":"zâmam","pron":"zaw-mam'","derivation":"a primitive root;","strongs_def":"to plan, usually in a bad sense","kjv_def":"consider, devise, imagine, plot, purpose, think (evil)."},
"H2162":{"lemma":"זָמָם","xlit":"zâmâm","pron":"zaw-mawm'","derivation":"from H2161 (זָמַם);","strongs_def":"a plot","kjv_def":"wicked device."},
"H2163":{"lemma":"זָמַן","xlit":"zâman","pron":"zaw-man'","derivation":"a primitive root;","strongs_def":"to fix (a time)","kjv_def":"appoint."},
"H2164":{"lemma":"זְמַן","xlit":"zᵉman","pron":"zem-an'","derivation":"(Aramaic) corresponding to H2163 (זָמַן);","strongs_def":"to agree (on a time and place)","kjv_def":"prepare."},
"H2165":{"lemma":"זְמָן","xlit":"zᵉmân","pron":"zem-awn'","derivation":"from H2163 (זָמַן);","strongs_def":"an appointed occasion","kjv_def":"season, time."},
"H2166":{"lemma":"זְמָן","xlit":"zᵉmân","pron":"zem-awn'","derivation":"(Aramaic) from H2165 (זְמָן); the same as H2165 (זְמָן)","strongs_def":"{an appointed occasion}","kjv_def":"season, time."},
"H2167":{"lemma":"זָמַר","xlit":"zâmar","pron":"zaw-mar'","derivation":"a primitive root (perhaps identical with H2168 (זָמַר) through the idea of striking with the fingers); properly, to touch the strings or parts of a musical instrument, i.e.","strongs_def":"play upon it; to make music, accompanied by the voice; hence to celebrate in song and music","kjv_def":"give praise, sing forth praises, psalms."},
"H2168":{"lemma":"זָמַר","xlit":"zâmar","pron":"zaw-mar'","derivation":"a primitive root (compare H2167 (זָמַר), H5568 (סָמַר), H6785 (צֶמֶר));","strongs_def":"to trim (a vine)","kjv_def":"prune."},
"H2169":{"lemma":"זֶמֶר","xlit":"zemer","pron":"zeh'-mer","derivation":"apparently from H2167 (זָמַר) or H2168 (זָמַר);","strongs_def":"a gazelle (from its lightly touching the ground)","kjv_def":"chamois."},
"H2170":{"lemma":"זְמָר","xlit":"zᵉmâr","pron":"zem-awr'","derivation":"(Aramaic) from a root corresponding to H2167 (זָמַר);","strongs_def":"instrumental music","kjv_def":"musick."},
"H2171":{"lemma":"זַמָּר","xlit":"zammâr","pron":"zam-mawr'","derivation":"(Aramaic) from the same as H2170 (זְמָר);","strongs_def":"an instrumental musician","kjv_def":"singer."},
"H2172":{"lemma":"זִמְרָה","xlit":"zimrâh","pron":"zim-raw'","derivation":"from H2167 (זָמַר);","strongs_def":"a musical piece or song to be accompanied by an instrument","kjv_def":"melody, psalm."},
"H2173":{"lemma":"זִמְרָה","xlit":"zimrâh","pron":"zim-raw'","derivation":"from H2168 (זָמַר);","strongs_def":"pruned (i.e. choice) fruit","kjv_def":"best fruit."},
"H2174":{"lemma":"זִמְרִי","xlit":"Zimrîy","pron":"zim-ree'","derivation":"from H2167 (זָמַר); musical;","strongs_def":"Zimri, the name of five Israelites, and of an Arabian tribe","kjv_def":"Zimri."},
"H2175":{"lemma":"זִמְרָן","xlit":"Zimrân","pron":"zim-rawn'","derivation":"from H2167 (זָמַר); musical;","strongs_def":"Zimran, a son of Abraham by Keturah","kjv_def":"Zimran."},
"H2176":{"lemma":"זִמְרָת","xlit":"zimrâth","pron":"zim-rawth'","derivation":"from H2167 (זָמַר);","strongs_def":"instrumental music; by implication, praise","kjv_def":"song."},
"H2177":{"lemma":"זַן","xlit":"zan","pron":"zan","derivation":"from H2109 (זוּן);","strongs_def":"properly, nourished (or fully developed), i.e. a form or sort","kjv_def":"divers kinds, [idiom] all manner of store."},
"H2178":{"lemma":"זַן","xlit":"zan","pron":"zan","derivation":"(Aramaic) corresponding to H2177 (זַן);","strongs_def":"sort","kjv_def":"kind."},
"H2179":{"lemma":"זָנַב","xlit":"zânab","pron":"zaw-nab'","derivation":"a primitive root meaning to wag; used only as a denominative from H2180 (זָנָב);","strongs_def":"to curtail, i.e. cut off the rear","kjv_def":"smite the hindmost."},
"H2180":{"lemma":"זָנָב","xlit":"zânâb","pron":"zaw-nawb'","derivation":"from H2179 (זָנַב) (in the original sense of flapping);","strongs_def":"the tail (literally or figuratively)","kjv_def":"tail."},
"H2181":{"lemma":"זָנָה","xlit":"zânâh","pron":"zaw-naw'","derivation":"a primitive root (highly-fed and therefore wanton);","strongs_def":"to commit adultery (usually of the female, and less often of simple fornication, rarely of involuntary ravishment); figuratively, to commit idolatry (the Jewish people being regarded as the spouse of Jehovah)","kjv_def":"(cause to) commit fornication, [idiom] continually, [idiom] great, (be an, play the) harlot, (cause to be, play the) whore, (commit, fall to) whoredom, (cause to) go a-whoring, whorish."},
"H2182":{"lemma":"זָנוֹחַ","xlit":"Zânôwach","pron":"zaw-no'-akh","derivation":"from H2186 (זָנַח); rejected;","strongs_def":"Zanoach, the name of two places in Palestine","kjv_def":"Zanoah."},
"H2183":{"lemma":"זָנוּן","xlit":"zânûwn","pron":"zaw-noon'","derivation":"from H2181 (זָנָה);","strongs_def":"adultery; figuratively, idolatry","kjv_def":"whoredom."},
"H2184":{"lemma":"זְנוּת","xlit":"zᵉnûwth","pron":"zen-ooth'","derivation":"from H2181 (זָנָה);","strongs_def":"adultery, i.e. (figuratively) infidelity, idolatry","kjv_def":"whoredom."},
"H2185":{"lemma":"זֹנוֹת","xlit":"zônôwth","pron":"zo-noth'","derivation":"regarded by some as if from H2109 (זוּן) or an unused root, and applied to military equipments; but evidently the feminine plural active participle of H2181 (זָנָה);","strongs_def":"harlots","kjv_def":"armour."},
"H2186":{"lemma":"זָנַח","xlit":"zânach","pron":"zaw-nakh'","derivation":"a primitive root meaning to push aside, i.e.","strongs_def":"reject, forsake, fail","kjv_def":"cast away (off), remove far away (off)."},
"H2187":{"lemma":"זָנַק","xlit":"zânaq","pron":"zaw-nak'","derivation":"a primitive root;","strongs_def":"properly, to draw together the feet (as an animal about to dart upon its prey), i.e. to spring forward","kjv_def":"leap."},
"H2188":{"lemma":"זֵעָה","xlit":"zêʻâh","pron":"zay-aw'","derivation":"from H2111 (זוּעַ) (in the sense of H3154 (יֶזַע));","strongs_def":"perspiration","kjv_def":"sweat."},
"H2189":{"lemma":"זַעֲוָה","xlit":"zaʻăvâh","pron":"zah-av-aw'","derivation":"by transposition for H2113 (זְוָעָה);","strongs_def":"agitation, maltreatment","kjv_def":"[idiom] removed, trouble."},
"H2190":{"lemma":"זַעֲוָן","xlit":"Zaʻăvân","pron":"zah-av-awn'","derivation":"from H2111 (זוּעַ); disquiet;","strongs_def":"Zaavan, an Idumaean","kjv_def":"Zaavan."},
"H2191":{"lemma":"זְעֵיר","xlit":"zᵉʻêyr","pron":"zeh-ayr'","derivation":"from an unused root (akin (by permutation) to H6819 (צָעַר)), meaning to dwindle;","strongs_def":"small","kjv_def":"little."},
"H2192":{"lemma":"זְעֵיר","xlit":"zᵉʻêyr","pron":"zeh-ayr'","derivation":"(Aramaic) corresponding to H2191 (זְעֵיר)xlit zeʻêyr corrected to zᵉʻêyr","strongs_def":"{small}","kjv_def":"little."},
"H2193":{"lemma":"זָעַךְ","xlit":"zâʻak","pron":"zaw-ak'","derivation":"a primitive root;","strongs_def":"to extinguish","kjv_def":"be extinct."},
"H2194":{"lemma":"זָעַם","xlit":"zâʻam","pron":"zaw-am'","derivation":"a primitive root;","strongs_def":"properly, to foam at the mouth, i.e. to be enraged","kjv_def":"abhor, abominable, (be) angry, defy, (have) indignation."},
"H2195":{"lemma":"זַעַם","xlit":"zaʻam","pron":"zah'-am","derivation":"from H2194 (זָעַם);","strongs_def":"strictly froth at the mouth, i.e. (figuratively) fury (especially of God's displeasure with sin)","kjv_def":"angry, indignation, rage."},
"H2196":{"lemma":"זָעַף","xlit":"zâʻaph","pron":"zaw-af'","derivation":"a primitive root;","strongs_def":"properly, to boil up, i.e. (figuratively) to be peevish or angry","kjv_def":"fret, sad, worse liking, be wroth."},
"H2197":{"lemma":"זַעַף","xlit":"zaʻaph","pron":"zah'-af","derivation":"from H2196 (זָעַף);","strongs_def":"anger","kjv_def":"indignation, rage(-ing), wrath."},
"H2198":{"lemma":"זָעֵף","xlit":"zâʻêph","pron":"zaw-afe'","derivation":"from H2196 (זָעַף);","strongs_def":"angry","kjv_def":"displeased."},
"H2199":{"lemma":"זָעַק","xlit":"zâʻaq","pron":"zaw-ak'","derivation":"a primitive root;","strongs_def":"to shriek (from anguish or danger); by analogy, (as a herald) to announce or convene publicly","kjv_def":"assemble, call (together), (make a) cry (out), come with such a company, gather (together), cause to be proclaimed."},
"H2200":{"lemma":"זְעִק","xlit":"zᵉʻiq","pron":"zek'-eek","derivation":"(Aramaic) corresponding to H2199 (זָעַק);","strongs_def":"to make an outcry","kjv_def":"cry."},
"H2201":{"lemma":"זַעַק","xlit":"zaʻaq","pron":"zah'-ak","derivation":"and (feminine) זְעָקָהxlit zᵉâqâh corrected to zᵉʻâqâh; from H2199 (זָעַק);","strongs_def":"a shriek or outcry","kjv_def":"cry(-ing)."},
"H2202":{"lemma":"זִפְרֹן","xlit":"Ziphrôn","pron":"zi-frone'","derivation":"from an unused root (meaning to be fragrant);","strongs_def":"Ziphron, a place in Palestine","kjv_def":"Ziphron."},
"H2203":{"lemma":"זֶפֶת","xlit":"zepheth","pron":"zeh'-feth","derivation":"from an unused root (meaning to liquify);","strongs_def":"asphalt (from its tendency to soften in the sun)","kjv_def":"pitch."},
"H2204":{"lemma":"זָקֵן","xlit":"zâqên","pron":"zaw-kane'","derivation":"a primitive root;","strongs_def":"to be old","kjv_def":"aged man, be (wax) old (man)."},
"H2205":{"lemma":"זָקֵן","xlit":"zâqên","pron":"zaw-kane'","derivation":"from H2204 (זָקֵן);","strongs_def":"old","kjv_def":"aged, ancient (man), elder(-est), old (man, men and...women), senator."},
"H2206":{"lemma":"זָקָן","xlit":"zâqân","pron":"zaw-kawn'","derivation":"from H2204 (זָקֵן);","strongs_def":"the beard (as indicating age)","kjv_def":"beard."},
"H2207":{"lemma":"זֹקֶן","xlit":"zôqen","pron":"zo'-ken","derivation":"from H2204 (זָקֵן);","strongs_def":"old age","kjv_def":"age."},
"H2208":{"lemma":"זָקֻן","xlit":"zâqun","pron":"zaw-koon'","derivation":"properly, passive participle of H2204 (זָקֵן) (used only in the plural as a noun);","strongs_def":"old age","kjv_def":"old age."},
"H2209":{"lemma":"זִקְנָה","xlit":"ziqnâh","pron":"zik-naw'","derivation":"feminine of H2205 (זָקֵן);","strongs_def":"old age","kjv_def":"old (age)."},
"H2210":{"lemma":"זָקַף","xlit":"zâqaph","pron":"zaw-kaf'","derivation":"a primitive root;","strongs_def":"to life, i.e. (figuratively) comfort","kjv_def":"raise (up)."},
"H2211":{"lemma":"זְקַף","xlit":"zᵉqaph","pron":"zek-af'","derivation":"(Aramaic) corresponding to H2210 (זָקַף);","strongs_def":"to hang, i.e. impale","kjv_def":"set up."},
"H2212":{"lemma":"זָקַק","xlit":"zâqaq","pron":"zaw-kak'","derivation":"a primitive root;","strongs_def":"to strain, (figuratively) extract, clarify","kjv_def":"fine, pour down, purge, purify, refine."},
"H2213":{"lemma":"זֵר","xlit":"zêr","pron":"zare","derivation":"from H2237 (זָרַר) (in the sense of scattering);","strongs_def":"a chaplet (as spread around the top), i.e. (specifically) a border moulding","kjv_def":"crown."},
"H2214":{"lemma":"זָרָא","xlit":"zârâʼ","pron":"zaw-raw'","derivation":"from H2114 (זוּר) (in the sense of estrangement) (compare H2219 (זָרָה));","strongs_def":"disgust","kjv_def":"loathsome."},
"H2215":{"lemma":"זָרַב","xlit":"zârab","pron":"zaw-rab'","derivation":"a primitive root;","strongs_def":"to flow away","kjv_def":"wax warm."},
"H2216":{"lemma":"זְרֻבָּבֶל","xlit":"Zᵉrubbâbel","pron":"zer-oob-baw-bel'","derivation":"from H2215 (זָרַב) and H894 (בָּבֶל); descended of (i.e. from) Babylon, i.e. born there;","strongs_def":"Zerubbabel, an Israelite","kjv_def":"Zerubbabel."},
"H2217":{"lemma":"זְרֻבָּבֶל","xlit":"Zᵉrubbâbel","pron":"zer-oob-baw-bel'","derivation":"(Aramaic) corresponding to H2216 (זְרֻבָּבֶל)","strongs_def":"{Zerubbabel, an Israelite}","kjv_def":"Zerubbabel."},
"H2218":{"lemma":"זֶרֶד","xlit":"Zered","pron":"zeh'-red","derivation":"from an unused root meaning to be exuberant in growth; lined with shrubbery;","strongs_def":"Zered, a brook East of the Dead Sea","kjv_def":"Zared, Zered."},
"H2219":{"lemma":"זָרָה","xlit":"zârâh","pron":"zaw-raw'","derivation":"a primitive root (compare H2114 (זוּר));","strongs_def":"to toss about; by implication, to diffuse, winnow","kjv_def":"cast away, compass, disperse, fan, scatter (away), spread, strew, winnow."},
"H2220":{"lemma":"זְרוֹעַ","xlit":"zᵉrôwaʻ","pron":"zer-o'-ah","derivation":"or (shortened) זְרֹעַ; and (feminine) זְרוֹעָה; or זְרֹעָה; from H2232 (זָרַע);","strongs_def":"the arm (as stretched out), or (of animals) the foreleg; figuratively, force","kjv_def":"arm, [phrase] help, mighty, power, shoulder, strength."},
"H2221":{"lemma":"זֵרוּעַ","xlit":"zêrûwaʻ","pron":"zay-roo'-ah","derivation":"from H2232 (זָרַע);","strongs_def":"something sown, i.e. a plant","kjv_def":"sowing, thing that is sown."},
"H2222":{"lemma":"זַרְזִיף","xlit":"zarzîyph","pron":"zar-zeef'","derivation":"by reduplication from an unused root meaning to flow;","strongs_def":"a pouring rain","kjv_def":"water."},
"H2223":{"lemma":"זַרְזִיר","xlit":"zarzîyr","pron":"zar-zeer'","derivation":"by reduplication from H2115 (זוּר);","strongs_def":"properly, tightly girt, i.e. probably a racer, or some fleet animal (as being slender in the waist)","kjv_def":"[phrase] greyhound."},
"H2224":{"lemma":"זָרַח","xlit":"zârach","pron":"zaw-rakh'","derivation":"a primitive root;","strongs_def":"properly, to irradiate (or shoot forth beams), i.e. to rise (as the sun); specifically, to appear (as a symptom of leprosy)","kjv_def":"arise, rise (up), as soon as it is up."},
"H2225":{"lemma":"זֶרַח","xlit":"zerach","pron":"zeh'-rakh","derivation":"from H2224 (זָרַח);","strongs_def":"a rising of light","kjv_def":"rising."},
"H2226":{"lemma":"זֶרַח","xlit":"Zerach","pron":"zeh'-rakh","derivation":"the same as H2225 (זֶרַח);","strongs_def":"Zerach, the name of three Israelites, also of an Idumaean and an Ethiopian prince","kjv_def":"Zarah, Zerah."},
"H2227":{"lemma":"זַרְחִי","xlit":"Zarchîy","pron":"zar-khee'","derivation":"patronymically from H2226 (זֶרַח);","strongs_def":"a Zarchite or descendant of Zerach","kjv_def":"Zarchite."},
"H2228":{"lemma":"זְרַחְיָה","xlit":"Zᵉrachyâh","pron":"zer-akh-yaw'","derivation":"from H2225 (זֶרַח) and H3050 (יָהּ); Jah has risen;","strongs_def":"Zerachjah, the name of two Israelites","kjv_def":"Zerahiah."},
"H2229":{"lemma":"זָרַם","xlit":"zâram","pron":"zaw-ram'","derivation":"a primitive root;","strongs_def":"to gush (as water)","kjv_def":"carry away as with a flood, pour out."},
"H2230":{"lemma":"זֶרֶם","xlit":"zerem","pron":"zeh'-rem","derivation":"from H2229 (זָרַם);","strongs_def":"a gush of water","kjv_def":"flood, overflowing, shower, storm, tempest."},
"H2231":{"lemma":"זִרְמָה","xlit":"zirmâh","pron":"zir-maw'","derivation":"feminine of H2230 (זֶרֶם);","strongs_def":"a gushing of fluid (semen)","kjv_def":"issue."},
"H2232":{"lemma":"זָרַע","xlit":"zâraʻ","pron":"zaw-rah'","derivation":"a primitive root;","strongs_def":"to sow; figuratively, to disseminate, plant, fructify","kjv_def":"bear, conceive seed, set with sow(-er), yield."},
"H2233":{"lemma":"זֶרַע","xlit":"zeraʻ","pron":"zeh'-rah","derivation":"from H2232 (זָרַע);","strongs_def":"seed; figuratively, fruit, plant, sowing-time, posterity","kjv_def":"[idiom] carnally, child, fruitful, seed(-time), sowing-time."},
"H2234":{"lemma":"זְרַע","xlit":"zᵉraʻ","pron":"zer-ah'","derivation":"(Aramaic) corresponding to H2233 (זֶרַע);","strongs_def":"posterity","kjv_def":"seed."},
"H2235":{"lemma":"זֵרֹעַ","xlit":"zêrôaʻ","pron":"zay-ro'-ah","derivation":"or זֵרָעֹן; from H2232 (זָרַע);","strongs_def":"something sown (only in the plural), i.e. a vegetable (as food)","kjv_def":"pulse."},
"H2236":{"lemma":"זָרַק","xlit":"zâraq","pron":"zaw-rak'","derivation":"a primitive root;","strongs_def":"to sprinkle (fluid or solid particles)","kjv_def":"be here and there, scatter, sprinkle, strew."},
"H2237":{"lemma":"זָרַר","xlit":"zârar","pron":"zaw-rar'","derivation":"a primitive root (compare H2114 (זוּר));","strongs_def":"perhaps to diffuse, i.e. (specifically) to sneeze","kjv_def":"sneeze."},
"H2238":{"lemma":"זֶרֶשׁ","xlit":"Zeresh","pron":"zeh'-resh","derivation":"of Persian origin;","strongs_def":"Zeresh, Haman's wife","kjv_def":"Zeresh."},
"H2239":{"lemma":"זֶרֶת","xlit":"zereth","pron":"zeh'-reth","derivation":"from H2219 (זָרָה);","strongs_def":"the spread of the fingers, i.e. a span","kjv_def":"span."},
"H2240":{"lemma":"זַתּוּא","xlit":"Zattûwʼ","pron":"zat-too'","derivation":"of uncertain derivation;","strongs_def":"Zattu, an Israelite","kjv_def":"Zattu."},
"H2241":{"lemma":"זֵתָם","xlit":"Zêthâm","pron":"zay-thawm'","derivation":"apparently a variation for H2133 (זֵיתָן);","strongs_def":"Zetham, an Israelite","kjv_def":"Zetham."},
"H2242":{"lemma":"זֵתַר","xlit":"Zêthar","pron":"zay-thar'","derivation":"of Persian origin;","strongs_def":"Zethar, a eunuch of Xerxes","kjv_def":"Zethar. j"},
"H2243":{"lemma":"חֹב","xlit":"chôb","pron":"khobe","derivation":"by contraction from H2245 (חָבַב);","strongs_def":"properly, a cherisher, i.e. the bosom","kjv_def":"bosom."},
"H2244":{"lemma":"חָבָא","xlit":"châbâʼ","pron":"khaw-baw'","derivation":"a primitive root (compare H2245 (חָבַב));","strongs_def":"to secrete","kjv_def":"[idiom] held, hide (self), do secretly."},
"H2245":{"lemma":"חָבַב","xlit":"châbab","pron":"khaw-bab'","derivation":"a primitive root (compare H2244 (חָבָא), H2247 (חָבָה));","strongs_def":"properly, to hide (as in the bosom), i.e. to cherish (with affection)","kjv_def":"love."},
"H2246":{"lemma":"חֹבָב","xlit":"Chôbâb","pron":"kho-bawb'","derivation":"from H2245 (חָבַב); cherished;","strongs_def":"Chobab, father-in-law of Moses","kjv_def":"Hobab."},
"H2247":{"lemma":"חָבָה","xlit":"châbâh","pron":"khaw-bah'","derivation":"a primitive root (compare H2245 (חָבַב));","strongs_def":"to secrete","kjv_def":"hide (self)."},
"H2248":{"lemma":"חֲבוּלָה","xlit":"chăbûwlâh","pron":"khab-oo-law'","derivation":"(Aramaic) from H2255 (חֲבַל);","strongs_def":"properly, overthrown, i.e. (morally) crime","kjv_def":"hurt."},
"H2249":{"lemma":"חָבוֹר","xlit":"Châbôwr","pron":"khaw-bore'","derivation":"from H2266 (חָבַר); united;","strongs_def":"Chabor, a river of Assyria","kjv_def":"Habor."},
"H2250":{"lemma":"חַבּוּרָה","xlit":"chabbûwrâh","pron":"khab-boo-raw'","derivation":"or חַבֻּרָה; or חֲבֻרָה; from H2266 (חָבַר);","strongs_def":"properly, bound (with stripes), i.e. a weal (or black-and-blue mark itself)","kjv_def":"blueness, bruise, hurt, stripe, wound."},
"H2251":{"lemma":"חָבַט","xlit":"châbaṭ","pron":"khaw-bat'","derivation":"a primitive root;","strongs_def":"to knock out or off","kjv_def":"beat (off, out), thresh."},
"H2252":{"lemma":"חֲבַיָּה","xlit":"Chăbayâh","pron":"khab-ah-yaw'","derivation":"or חֲבָיָהlemma חֲבָיה missing vowel, corrected to חֲבָיָה; from H2247 (חָבָה) and H3050 (יָהּ); Jah has hidden;","strongs_def":"Chabajah, an Israelite","kjv_def":"Habaiah."},
"H2253":{"lemma":"חֶבְיוֹן","xlit":"chebyôwn","pron":"kheb-yone'","derivation":"from H2247 (חָבָה);","strongs_def":"a concealment","kjv_def":"hiding."},
"H2254":{"lemma":"חָבַל","xlit":"châbal","pron":"khaw-bal'","derivation":"a primitive root;","strongs_def":"to wind tightly (as a rope), i.e. to bind; specifically, by a pledge; figuratively, to pervert, destroy; also to writhe in pain (especially of parturition)","kjv_def":"[idiom] at all, band, bring forth, (deal) corrupt(-ly), destroy, offend, lay to (take a) pledge, spoil, travail, [idiom] very, withhold."},
"H2255":{"lemma":"חֲבַל","xlit":"chăbal","pron":"khab-al'","derivation":"(Aramaic) corresponding to H2254 (חָבַל);","strongs_def":"to ruin","kjv_def":"destroy, hurt."},
"H2256":{"lemma":"חֶבֶל","xlit":"chebel","pron":"kheh'-bel","derivation":"or חֵבֶל; from H2254 (חָבַל);","strongs_def":"a rope (as twisted), especially a measuring line; by implication, a district or inheritance (as measured); or a noose (as of cords); figuratively, a company (as if tied together); also a throe (especially of parturition); also ruin","kjv_def":"band, coast, company, cord, country, destruction, line, lot, pain, pang, portion, region, rope, snare, sorrow, tackling."},
"H2257":{"lemma":"חֲבַל","xlit":"chăbal","pron":"khab-al'","derivation":"(Aramaic) from H2255 (חֲבַל);","strongs_def":"harm (personal or pecuniary)","kjv_def":"damage, hurt."},
"H2258":{"lemma":"חֲבֹל","xlit":"chăbôl","pron":"khab-ole'","derivation":"or (feminine) חֲבֹלָה; from H2254 (חָבַל);","strongs_def":"a pawn (as security for debt)","kjv_def":"pledge."},
"H2259":{"lemma":"חֹבֵל","xlit":"chôbêl","pron":"kho-bale'","derivation":"active participle from H2254 (חָבַל) (in the sense of handling ropes);","strongs_def":"a sailor","kjv_def":"pilot, shipmaster."},
"H2260":{"lemma":"חִבֵּל","xlit":"chibbêl","pron":"khib-bale'","derivation":"from H2254 (חָבַל) (in the sense of furnished with ropes);","strongs_def":"a mast","kjv_def":"mast."},
"H2261":{"lemma":"חֲבַצֶּלֶת","xlit":"chăbatstseleth","pron":"khab-ats-tseh'-leth","derivation":"of uncertain derivation;","strongs_def":"probably meadow-saffron","kjv_def":"rose."},
"H2262":{"lemma":"חֲבַצַּנְיָה","xlit":"Chăbatstsanyâh","pron":"khab-ats-tsan-yaw'","derivation":"of uncertain derivation;","strongs_def":"Chabatstsanjah, a Rechabite","kjv_def":"Habazaniah."},
"H2263":{"lemma":"חָבַק","xlit":"châbaq","pron":"khaw-bak'","derivation":"a primitive root;","strongs_def":"to clasp (the hands or in embrace)","kjv_def":"embrace, fold."},
"H2264":{"lemma":"חִבֻּק","xlit":"chibbuq","pron":"khib-book'","derivation":"from H2263 (חָבַק);","strongs_def":"a clasping of the hands (in idleness)","kjv_def":"fold."},
"H2265":{"lemma":"חֲבַקּוּק","xlit":"Chăbaqqûwq","pron":"khab-ak-kook'","derivation":"by reduplication from H2263 (חָבַק); embrace;","strongs_def":"Chabakkuk, the prophet","kjv_def":"Habakkuk."},
"H2266":{"lemma":"חָבַר","xlit":"châbar","pron":"khaw-bar'","derivation":"a primitive root;","strongs_def":"to join (literally or figuratively); specifically (by means of spells) to fascinate","kjv_def":"charm(-er), be compact, couple (together), have fellowship with, heap up, join (self, together), league."},
"H2267":{"lemma":"חֶבֶר","xlit":"cheber","pron":"kheh'-ber","derivation":"from H2266 (חָבַר);","strongs_def":"a society; also a spell","kjv_def":"[phrase] charmer(-ing), company, enchantment, [idiom] wide."},
"H2268":{"lemma":"חֶבֶר","xlit":"Cheber","pron":"kheh'-ber","derivation":"the same as H2267 (חֶבֶר); community;","strongs_def":"Cheber, the name of a Kenite and of three Israelites","kjv_def":"Heber."},
"H2269":{"lemma":"חֲבַר","xlit":"chăbar","pron":"khab-ar'","derivation":"(Aramaic) from a root corresponding to H2266 (חָבַר);","strongs_def":"an associate","kjv_def":"companion, fellow."},
"H2270":{"lemma":"חָבֵר","xlit":"châbêr","pron":"khaw-bare'","derivation":"from H2266 (חָבַר);","strongs_def":"an associate","kjv_def":"companion, fellow, knit together."},
"H2271":{"lemma":"חַבָּר","xlit":"chabbâr","pron":"khab-bawr'","derivation":"from H2266 (חָבַר);","strongs_def":"a partner","kjv_def":"companion."},
"H2272":{"lemma":"חֲבַרְבֻּרָה","xlit":"chăbarburâh","pron":"khab-ar-boo-raw'","derivation":"by reduplication from H2266 (חָבַר);","strongs_def":"a streak (like a line), as on the tiger","kjv_def":"spot."},
"H2273":{"lemma":"חַבְרָה","xlit":"chabrâh","pron":"khab-raw'","derivation":"(Aramaic) feminine of H2269 (חֲבַר);","strongs_def":"an associate","kjv_def":"other."},
"H2274":{"lemma":"חֶבְרָה","xlit":"chebrâh","pron":"kheb-raw'","derivation":"feminine of H2267 (חֶבֶר);","strongs_def":"association","kjv_def":"company."},
"H2275":{"lemma":"חֶבְרוֹן","xlit":"Chebrôwn","pron":"kheb-rone'","derivation":"from H2267 (חֶבֶר); seat of association;","strongs_def":"Chebron, a place in Palestine, also the name of two Israelites","kjv_def":"Hebron."},
"H2276":{"lemma":"חֶבְרוֹנִי","xlit":"Chebrôwnîy","pron":"kheb-ro-nee'","derivation":"or חֶבְרֹנִי; patronymically from H2275 (חֶבְרוֹן);","strongs_def":"Chebronite (collectively), an inhabitant of Chebron","kjv_def":"Hebronites."},
"H2277":{"lemma":"חֶבְרִי","xlit":"Chebrîy","pron":"kheb-ree'","derivation":"patronymically from H2268 (חֶבֶר);","strongs_def":"a Chebrite (collectively) or descendants of Cheber","kjv_def":"Heberites."},
"H2278":{"lemma":"חֲבֶרֶת","xlit":"chăbereth","pron":"khab-eh'-reth","derivation":"feminine of H2270 (חָבֵר);","strongs_def":"a consort","kjv_def":"companion."},
"H2279":{"lemma":"חֹבֶרֶת","xlit":"chôbereth","pron":"kho-beh'-reth","derivation":"feminine active participle of H2266 (חָבַר);","strongs_def":"a joint","kjv_def":"which coupleth, coupling."},
"H2280":{"lemma":"חָבַשׁ","xlit":"châbash","pron":"khaw-bash'","derivation":"a primitive root;","strongs_def":"to wrap firmly (especially a turban, compress, or saddle); figuratively, to stop, to rule","kjv_def":"bind (up), gird about, govern, healer, put, saddle, wrap about."},
"H2281":{"lemma":"חָבֵת","xlit":"châbêth","pron":"khaw-bayth'","derivation":"from an unused root probably meaning to cook (compare H4227 (מַחֲבַת));","strongs_def":"something fried, probably a griddle-cake","kjv_def":"pan."},
"H2282":{"lemma":"חַג","xlit":"chag","pron":"khag","derivation":"or חָג; from H2287 (חָגַג);","strongs_def":"a festival, or a victim therefor","kjv_def":"(solemn) feast (day), sacrifice, solemnity."},
"H2283":{"lemma":"חָגָא","xlit":"châgâʼ","pron":"khaw-gaw'","derivation":"from an unused root meaning to revolve (compare H2287 (חָגַג));","strongs_def":"properly, vertigo, i.e. (figuratively) fear","kjv_def":"terror."},
"H2284":{"lemma":"חָגָב","xlit":"châgâb","pron":"khaw-gawb'","derivation":"of uncertain derivation;","strongs_def":"a locust","kjv_def":"locust."},
"H2285":{"lemma":"חָגָב","xlit":"Châgâb","pron":"khaw-gawb'","derivation":"the same as H2284 (חָגָב); locust;","strongs_def":"Chagab, one of the Nethinim","kjv_def":"Hagab."},
"H2286":{"lemma":"חֲגָבָא","xlit":"Chăgâbâʼ","pron":"khag-aw-baw'","derivation":"or חֲגָבָה; feminine of H2285 (חָגָב); locust;","strongs_def":"Chagaba or Chagabah, one of the Nethinim","kjv_def":"Hagaba, Hagabah."},
"H2287":{"lemma":"חָגַג","xlit":"châgag","pron":"khaw-gag'","derivation":"a primitive root (compare H2283 (חָגָא), H2328 (חוּג));","strongs_def":"properly, to move in acircle, i.e. (specifically) to march in asacred procession, to observe afestival; by implication, to be giddy","kjv_def":"celebrate, dance, (keep, hold) a (solemn) feast (holiday), reel to and fro."},
"H2288":{"lemma":"חֲגָו","xlit":"chăgâv","pron":"khag-awv'","derivation":"from an unused root meaning to take refuge;","strongs_def":"a rift in rocks","kjv_def":"cleft."},
"H2289":{"lemma":"חָגוֹר","xlit":"châgôwr","pron":"khaw-gore'","derivation":"from H2296 (חָגַר);","strongs_def":"belted","kjv_def":"girded with."},
"H2290":{"lemma":"חֲגוֹר","xlit":"chăgôwr","pron":"khag-ore'","derivation":"or חֲגֹר; and (feminine) חֲגוֹרָה; or חֲגֹרָה; from H2296 (חָגַר);","strongs_def":"a belt (for the waist)","kjv_def":"apron, armour, gird(-le)."},
"H2291":{"lemma":"חַגִּי","xlit":"Chaggîy","pron":"khag-ghee'","derivation":"from H2287 (חָגַג); festive","strongs_def":"Chaggi, an Israelite; also (patronymically) a Chaggite, or descendant of the same","kjv_def":"Haggi, Haggites."},
"H2292":{"lemma":"חַגַּי","xlit":"Chaggay","pron":"khag-gah'-ee","derivation":"from H2282 (חַג); festive;","strongs_def":"Chaggai, a Hebrew prophet","kjv_def":"Haggai."},
"H2293":{"lemma":"חַגִּיָּה","xlit":"Chaggîyâh","pron":"khag-ghee-yaw'","derivation":"from H2282 (חַג) and H3050 (יָהּ); festival of Jah;","strongs_def":"Chaggijah, an Israelite","kjv_def":"Haggiah."},
"H2294":{"lemma":"חַגִּית","xlit":"Chaggîyth","pron":"khag-gheeth'","derivation":"feminine of H2291 (חַגִּי); festive;","strongs_def":"Chaggith, a wife of David","kjv_def":"Haggith."},
"H2295":{"lemma":"חׇגְלָה","xlit":"Choglâh","pron":"khog-law'","derivation":"of uncertain derivation; probably a partridge;","strongs_def":"Choglah, an Israelitess","kjv_def":"Hoglah. See also H1031 (בֵּית חׇגְלָה)."},
"H2296":{"lemma":"חָגַר","xlit":"châgar","pron":"khaw-gar'","derivation":"a primitive root;","strongs_def":"to gird on (as a belt, armor, etc.)","kjv_def":"be able to put on, be afraid, appointed, gird, restrain, [idiom] on every side."},
"H2297":{"lemma":"חַד","xlit":"chad","pron":"khad","derivation":"abridged from H259 (אֶחָד);","strongs_def":"one","kjv_def":"one."},
"H2298":{"lemma":"חַד","xlit":"chad","pron":"khad","derivation":"(Aramaic) corresponding to H2297 (חַד);","strongs_def":"as card. one; as article single; as an ordinal, first; adverbially, at once","kjv_def":"a, first, one, together."},
"H2299":{"lemma":"חַד","xlit":"chad","pron":"khad","derivation":"from H2300 (חָדַד);","strongs_def":"sharp","kjv_def":"sharp."},
"H2300":{"lemma":"חָדַד","xlit":"châdad","pron":"khaw-dad'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) sharp or (figuratively) severe","kjv_def":"be fierce, sharpen."},
"H2301":{"lemma":"חֲדַד","xlit":"Chădad","pron":"khad-ad'","derivation":"from H2300 (חָדַד); fierce;","strongs_def":"Chadad, an Ishmaelite","kjv_def":"Hadad."},
"H2302":{"lemma":"חָדָה","xlit":"châdâh","pron":"khaw-daw'","derivation":"a primitive root;","strongs_def":"to rejoice","kjv_def":"make glad, be joined, rejoice."},
"H2303":{"lemma":"חַדּוּד","xlit":"chaddûwd","pron":"khad-dood'","derivation":"from H2300 (חָדַד);","strongs_def":"a point","kjv_def":"sharp."},
"H2304":{"lemma":"חֶדְוָה","xlit":"chedvâh","pron":"khed-vaw'","derivation":"from H2302 (חָדָה);","strongs_def":"rejoicing","kjv_def":"gladness, joy."},
"H2305":{"lemma":"חֶדְוָה","xlit":"chedvâh","pron":"khed-vaw'","derivation":"(Aramaic) corresponding to H2304 (חֶדְוָה)","strongs_def":"{rejoicing}","kjv_def":"joy."},
"H2306":{"lemma":"חֲדִי","xlit":"chădîy","pron":"khad-ee'","derivation":"(Aramaic) corresponding to H2373 (חָזֶה);","strongs_def":"a breast","kjv_def":"breast."},
"H2307":{"lemma":"חָדִיד","xlit":"Châdîyd","pron":"khaw-deed'","derivation":"from H2300 (חָדַד); a peak;","strongs_def":"Chadid, a place in Palestine","kjv_def":"Hadid."},
"H2308":{"lemma":"חָדַל","xlit":"châdal","pron":"khaw-dal'","derivation":"a primitive root;","strongs_def":"properly, to be flabby, i.e. (by implication) desist; (figuratively) be lacking or idle","kjv_def":"cease, end, fall, forbear, forsake, leave (off), let alone, rest, be unoccupied, want."},
"H2309":{"lemma":"חֶדֶל","xlit":"chedel","pron":"kheh'-del","derivation":"from H2308 (חָדַל);","strongs_def":"rest, i.e. the state of the dead","kjv_def":"world."},
"H2310":{"lemma":"חָדֵל","xlit":"châdêl","pron":"khaw-dale'","derivation":"from H2308 (חָדַל);","strongs_def":"vacant, i.e. ceasing or destitute","kjv_def":"he that forbeareth, frail, rejected."},
"H2311":{"lemma":"חַדְלַי","xlit":"Chadlay","pron":"khad-lah'-ee","derivation":"from H2309 (חֶדֶל); idle;","strongs_def":"Chadlai, an Israelite","kjv_def":"Hadlai."},
"H2312":{"lemma":"חֵדֶק","xlit":"chêdeq","pron":"khay'-dek","derivation":"from an unused root meaning to sting;","strongs_def":"a prickly plant","kjv_def":"brier, thorn."},
"H2313":{"lemma":"חִדֶּקֶל","xlit":"Chiddeqel","pron":"khid-deh'-kel","derivation":"probably of foreign origin;","strongs_def":"the Chiddekel (or Tigris) river","kjv_def":"Hiddekel."},
"H2314":{"lemma":"חָדַר","xlit":"châdar","pron":"khaw-dar'","derivation":"a primitive root;","strongs_def":"properly, to inclose (as a room), i.e. (by analogy,) to beset (as in a siege)","kjv_def":"enter a privy chamber."},
"H2315":{"lemma":"חֶדֶר","xlit":"cheder","pron":"kheh'-der","derivation":"from H2314 (חָדַר);","strongs_def":"an apartment (usually literal)","kjv_def":"((bed) inner) chamber, innermost(-ward) part, parlour, [phrase] south, [idiom] within."},
"H2316":{"lemma":"חֲדַר","xlit":"Chădar","pron":"khad-ar'","derivation":"another form for H2315 (חֶדֶר); chamber;","strongs_def":"Chadar, an Ishmaelite","kjv_def":"Hadar."},
"H2317":{"lemma":"חַדְרָךְ","xlit":"Chadrâk","pron":"khad-rawk'","derivation":"of uncertain derivation;","strongs_def":"Chadrak, a Syrian deity","kjv_def":"Hadrach."},
"H2318":{"lemma":"חָדַשׁ","xlit":"châdash","pron":"khaw-dash'","derivation":"a primitive root;","strongs_def":"to be new; causatively, to rebuild","kjv_def":"renew, repair."},
"H2319":{"lemma":"חָדָשׁ","xlit":"châdâsh","pron":"khaw-dawsh'","derivation":"from H2318 (חָדַשׁ);","strongs_def":"new","kjv_def":"fresh, new thing."},
"H2320":{"lemma":"חֹדֶשׁ","xlit":"chôdesh","pron":"kho'-desh","derivation":"from H2318 (חָדַשׁ);","strongs_def":"the new moon; by implication, a month","kjv_def":"month(-ly), new moon."},
"H2321":{"lemma":"חֹדֶשׁ","xlit":"Chôdesh","pron":"kho'-desh","derivation":"the same as H2320 (חֹדֶשׁ);","strongs_def":"Chodesh, an Israelitess","kjv_def":"Hodesh."},
"H2322":{"lemma":"חֲדָשָׁה","xlit":"Chădâshâh","pron":"khad-aw-shaw'","derivation":"feminine of H2319 (חָדָשׁ); new;","strongs_def":"Chadashah, a place in Palestine","kjv_def":"Hadashah."},
"H2323":{"lemma":"חֲדָת","xlit":"chădâth","pron":"khad-ath'","derivation":"(Aramaic) corresponding to H2319 (חָדָשׁ);","strongs_def":"new","kjv_def":"new."},
"H2324":{"lemma":"חֲוָא","xlit":"chăvâʼ","pron":"khav-aw'","derivation":"(Aramaic) corresponding to H2331 (חָוָה);","strongs_def":"to show","kjv_def":"shew."},
"H2325":{"lemma":"חוּב","xlit":"chûwb","pron":"khoob","derivation":"also חָיַב; a primitive root;","strongs_def":"properly, perhaps to tie, i.e. (figuratively and reflexively) to owe, or (by implication) to forfeit","kjv_def":"make endanger."},
"H2326":{"lemma":"חוֹב","xlit":"chôwb","pron":"khobe","derivation":"from H2325 (חוּב);","strongs_def":"debt","kjv_def":"debtor."},
"H2327":{"lemma":"חוֹבָה","xlit":"chôwbâh","pron":"kho-baw'","derivation":"feminine active participle of H2247 (חָבָה); hiding place;","strongs_def":"Chobah, a place in Syria","kjv_def":"Hobah."},
"H2328":{"lemma":"חוּג","xlit":"chûwg","pron":"khoog","derivation":"a primitive root (compare H2287 (חָגַג));","strongs_def":"to describe a circle","kjv_def":"compass."},
"H2329":{"lemma":"חוּג","xlit":"chûwg","pron":"khoog","derivation":"from H2328 (חוּג);","strongs_def":"a circle","kjv_def":"circle, circuit, compass."},
"H2330":{"lemma":"חוּד","xlit":"chûwd","pron":"khood","derivation":"a primitive root;","strongs_def":"properly, to tie a knot, i.e. (figuratively) to propound a riddle","kjv_def":"put forth."},
"H2331":{"lemma":"חָוָה","xlit":"châvâh","pron":"khaw-vah'","derivation":"a primitive root; (compare H2324 (חֲוָא), H2421 (חָיָה));","strongs_def":"properly, to live; by implication (intensively) to declare or show","kjv_def":"show."},
"H2332":{"lemma":"חַוָּה","xlit":"Chavvâh","pron":"khav-vaw'","derivation":"causatively from H2331 (חָוָה); life-giver;","strongs_def":"Chavvah (or Eve), the first woman","kjv_def":"Eve."},
"H2333":{"lemma":"חַוָּה","xlit":"chavvâh","pron":"khav-vaw'","derivation":"properly, the same as H2332 (חַוָּה) (lifegiving, i.e. living-place);","strongs_def":"by implication, an encampment or village","kjv_def":"(small) town."},
"H2334":{"lemma":"חַוּוֹת יָעִיר","xlit":"Chavvôwth Yâʻîyr","pron":"khav-vothe' yaw-eer'","derivation":"from the plural of H2333 (חַוָּה) and a modification of H3265 (יָעוּר);","strongs_def":"hamlets of Jair, a region of Palestine","kjv_def":"(Bashan-) Havoth-jair."},
"H2335":{"lemma":"חוֹזַי","xlit":"Chôwzay","pron":"kho-zah'-ee","derivation":"from H2374 (חֹזֶה); visionary;","strongs_def":"Chozai, an Israelite","kjv_def":"the seers."},
"H2336":{"lemma":"חוֹחַ","xlit":"chôwach","pron":"kho'-akh","derivation":"from an unused root apparently meaning to pierce;","strongs_def":"a thorn; by analogy, a ring forthe nose","kjv_def":"bramble, thistle, thorn."},
"H2337":{"lemma":"חָוָח","xlit":"châvâch","pron":"khaw-vawkh'","derivation":"perhaps the same as H2336 (חוֹחַ);","strongs_def":"a dell or crevice (as if pierced in the earth)","kjv_def":"thicket."},
"H2338":{"lemma":"חוּט","xlit":"chûwṭ","pron":"khoot","derivation":"(Aramaic) corresponding to the root of H2339 (חוּט), perhaps as a denominative;","strongs_def":"to string together, i.e. (figuratively) to repair","kjv_def":"join."},
"H2339":{"lemma":"חוּט","xlit":"chûwṭ","pron":"khoot","derivation":"from an unused root probably meaning to sew;","strongs_def":"a string; by implication, a measuring tape","kjv_def":"cord, fillet, line, thread."},
"H2340":{"lemma":"חִוִּי","xlit":"Chivvîy","pron":"khiv-vee'","derivation":"perhaps from H2333 (חַוָּה); a villager;","strongs_def":"a Chivvite, one of the aboriginal tribes of Palestine","kjv_def":"Hivite."},
"H2341":{"lemma":"חֲוִילָה","xlit":"Chăvîylâh","pron":"khav-ee-law'","derivation":"probably from H2342 (חוּל); circular;","strongs_def":"Chavilah, the name of two or three eastern regions; also perhaps of two men","kjv_def":"Havilah."},
"H2342":{"lemma":"חוּל","xlit":"chûwl","pron":"khool","derivation":"or חִיל; a primitive root;","strongs_def":"properly, to twist or whirl (in a circular or spiral manner), i.e. (specifically) to dance, to writhe in pain (especially of parturition) or fear; figuratively, to wait, to pervert","kjv_def":"bear, (make to) bring forth, (make to) calve, dance, drive away, fall grievously (with pain), fear, form, great, grieve, (be) grievous, hope, look, make, be in pain, be much (sore) pained, rest, shake, shapen, (be) sorrow(-ful), stay, tarry, travail (with pain), tremble, trust, wait carefully (patiently), be wounded."},
"H2343":{"lemma":"חוּל","xlit":"Chûwl","pron":"khool","derivation":"from H2342 (חוּל); a circle;","strongs_def":"Chul, a son of Aram; also the region settled by him","kjv_def":"Hul."},
"H2344":{"lemma":"חוֹל","xlit":"chôwl","pron":"khole","derivation":"from H2342 (חוּל);","strongs_def":"sand (as round or whirling particles)","kjv_def":"sand."},
"H2345":{"lemma":"חוּם","xlit":"chûwm","pron":"khoom","derivation":"from an unused root meaning to be warm, i.e. (by implication)","strongs_def":"sunburnt or swarthy (blackish)","kjv_def":"brown."},
"H2346":{"lemma":"חוֹמָה","xlit":"chôwmâh","pron":"kho-maw'","derivation":"feminine active participle of an unused root apparently meaning to join;","strongs_def":"a wall of protection","kjv_def":"wall, walled."},
"H2347":{"lemma":"חוּס","xlit":"chûwç","pron":"khoos","derivation":"a primitive root;","strongs_def":"properly, to cover, i.e. (figuratively) to compassionate","kjv_def":"pity, regard, spare."},
"H2348":{"lemma":"חוֹף","xlit":"chôwph","pron":"khofe","derivation":"from an unused root meaning to cover;","strongs_def":"a cove (as a sheltered bay)","kjv_def":"coast (of the sea), haven, shore, (sea-) side."},
"H2349":{"lemma":"חוּפָם","xlit":"Chûwphâm","pron":"khoo-fawm'","derivation":"from the same as H2348 (חוֹף); protection;","strongs_def":"Chupham, an Israelite","kjv_def":"Hupham."},
"H2350":{"lemma":"חוּפָמִי","xlit":"Chûwphâmîy","pron":"khoo-faw-mee'","derivation":"patronymically from H2349 (חוּפָם);","strongs_def":"a Chuphamite or descendant of Chupham","kjv_def":"Huphamites."},
"H2351":{"lemma":"חוּץ","xlit":"chûwts","pron":"khoots","derivation":"or (shortened) חֻץ; (both forms feminine in the plural) from an unused root meaning to sever;","strongs_def":"properly, separate by awall, i.e. outside, outdoors","kjv_def":"abroad, field, forth, highway, more, out(-side, -ward), street, without."},
"H2352":{"lemma":"חוּר","xlit":"chûwr","pron":"khoor","derivation":"or (shortened) חֻר; from an unused root probably meaning to bore;","strongs_def":"the crevice of a serpent; the cell of aprison","kjv_def":"hole."},
"H2353":{"lemma":"חוּר","xlit":"chûwr","pron":"khoor","derivation":"from H2357 (חָוַר);","strongs_def":"white linen","kjv_def":"white."},
"H2354":{"lemma":"חוּר","xlit":"Chûwr","pron":"khoor","derivation":"the same as H2353 (חוּר) or H2352 (חוּר);","strongs_def":"Chur, the name of four Israelites and one Midianite","kjv_def":"Hur."},
"H2355":{"lemma":"חוֹר","xlit":"chôwr","pron":"khore","derivation":"the same as H2353 (חוּר);","strongs_def":"white linen","kjv_def":"network. Compare H2715 (חֹר)."},
"H2356":{"lemma":"חוֹר","xlit":"chôwr","pron":"khore","derivation":"or (shortened) חֹר; the same as H2352 (חוּר);","strongs_def":"a cavity, socket, den","kjv_def":"cave, hole."},
"H2357":{"lemma":"חָוַר","xlit":"châvar","pron":"khaw-var'","derivation":"a primitive root;","strongs_def":"to blanch (as with shame)","kjv_def":"wax pale."},
"H2358":{"lemma":"חִוָּר","xlit":"chivvâr","pron":"khiv-vawr'","derivation":"(Aramaic) from a root corresponding to H2357 (חָוַר);","strongs_def":"white","kjv_def":"white."},
"H2359":{"lemma":"חוּרִי","xlit":"Chûwrîy","pron":"khoo-ree'","derivation":"probably from H2353 (חוּר); linen-worker;","strongs_def":"Churi, an Israelite","kjv_def":"Huri."},
"H2360":{"lemma":"חוּרַי","xlit":"Chûwray","pron":"khoo-rah'ee","derivation":"probably an orthographical variation for H2359 (חוּרִי);","strongs_def":"Churai, an Israelite","kjv_def":"Hurai."},
"H2361":{"lemma":"חוּרָם","xlit":"Chûwrâm","pron":"khoo-rawm'","derivation":"probably from H2353 (חוּר); whiteness, i.e. noble;","strongs_def":"Churam, the name of an Israelite and two Syrians","kjv_def":"Huram. Compare H2438 (חִירָם)."},
"H2362":{"lemma":"חַוְרָן","xlit":"Chavrân","pron":"khav-rawn'","derivation":"apparently from H2357 (חָוַר) (in the sense of H2352 (חוּר)); cavernous;","strongs_def":"Chavran, a region East of the Jordan","kjv_def":"Hauran."},
"H2363":{"lemma":"חוּשׁ","xlit":"chûwsh","pron":"koosh","derivation":"a primitive root;","strongs_def":"to hurry; figuratively, to be eager with excitement or enjoyment","kjv_def":"(make) haste(-n), ready."},
"H2364":{"lemma":"חוּשָׁה","xlit":"Chûwshâh","pron":"khoo-shaw'","derivation":"from H2363 (חוּשׁ); haste;","strongs_def":"Chushah, an Israelite","kjv_def":"Hushah."},
"H2365":{"lemma":"חוּשַׁי","xlit":"Chûwshay","pron":"khoo-shah'-ee","derivation":"from H2363 (חוּשׁ); hasty;","strongs_def":"Chushai, an Israelite","kjv_def":"Hushai."},
"H2366":{"lemma":"חוּשִׁים","xlit":"Chûwshîym","pron":"khoo-sheem'","derivation":"or חֻשִׁים; or חֻשִׁם; plural from H2363 (חוּשׁ); hasters;","strongs_def":"Chushim, the name of three Israelites","kjv_def":"Hushim."},
"H2367":{"lemma":"חוּשָׁם","xlit":"Chûwshâm","pron":"khoo-shawm'","derivation":"or חֻשָׁם; from H2363 (חוּשׁ); hastily;","strongs_def":"Chusham, an Idumaean","kjv_def":"Husham."},
"H2368":{"lemma":"חוֹתָם","xlit":"chôwthâm","pron":"kho-thawm'","derivation":"or חֹתָם; from H2856 (חָתַם);","strongs_def":"a signature-ring","kjv_def":"seal, signet."},
"H2369":{"lemma":"חוֹתָם","xlit":"Chôwthâm","pron":"kho-thawm'","derivation":"the same as H2368 (חוֹתָם); seal;","strongs_def":"Chotham, the name of two Israelites","kjv_def":"Hotham, Hothan."},
"H2370":{"lemma":"חֲזָא","xlit":"chăzâʼ","pron":"khaz-aw'","derivation":"(Aramaic) or חֲזָה; (Aramaic), corresponding to H2372 (חָזָה);","strongs_def":"to gaze upon; mentally to dream, be usual (i.e. seem)","kjv_def":"behold, have (a dream), see, be wont."},
"H2371":{"lemma":"חֲזָאֵל","xlit":"Chăzâʼêl","pron":"khaz-aw-ale'","derivation":"or חֲזָהאֵל; from H2372 (חָזָה) and H410 (אֵל); God has seen;","strongs_def":"Chazael, a king of Syria","kjv_def":"Hazael."},
"H2372":{"lemma":"חָזָה","xlit":"châzâh","pron":"khaw-zaw'","derivation":"a primitive root;","strongs_def":"to gaze at; mentally to perceive, contemplate (with pleasure); specifically, to have avision of","kjv_def":"behold, look, prophesy, provide, see."},
"H2373":{"lemma":"חָזֶה","xlit":"châzeh","pron":"khaw-zeh'","derivation":"from H2372 (חָזָה);","strongs_def":"the breast (as most seen in front)","kjv_def":"breast."},
"H2374":{"lemma":"חֹזֶה","xlit":"chôzeh","pron":"kho-zeh'","derivation":"active participle of H2372 (חָזָה);","strongs_def":"a beholder in vision; also a compact (as looked upon with approval)","kjv_def":"agreement, prophet, see that, seer, (star-) gazer."},
"H2375":{"lemma":"חֲזוֹ","xlit":"Chăzôw","pron":"khaz-o'","derivation":"from H2372 (חָזָה); seer;","strongs_def":"Chazo, a nephew of Abraham","kjv_def":"Hazo."},
"H2376":{"lemma":"חֵזֵו","xlit":"chêzêv","pron":"khay'-zev","derivation":"(Aramaic) from H2370 (חֲזָא);","strongs_def":"a sight","kjv_def":"look, vision."},
"H2377":{"lemma":"חָזוֹן","xlit":"châzôwn","pron":"khaw-zone'","derivation":"from H2372 (חָזָה);","strongs_def":"a sight (mentally), i.e. a dream, revelation, or oracle","kjv_def":"vision."},
"H2378":{"lemma":"חָזוֹת","xlit":"châzôwth","pron":"khaw-zooth'","derivation":"from H2372 (חָזָה);","strongs_def":"a revelation","kjv_def":"vision."},
"H2379":{"lemma":"חֲזוֹת","xlit":"chăzôwth","pron":"khaz-oth'","derivation":"(Aramaic) from H2370 (חֲזָא);","strongs_def":"a view","kjv_def":"sight."},
"H2380":{"lemma":"חָזוּת","xlit":"châzûwth","pron":"khaw-zooth'","derivation":"from H2372 (חָזָה);","strongs_def":"a look; hence (figuratively) striking appearance, revelation, or (by implication) compact","kjv_def":"agreement, notable (one), vision."},
"H2381":{"lemma":"חֲזִיאֵל","xlit":"Chăzîyʼêl","pron":"khaz-ee-ale'","derivation":"from H2372 (חָזָה) and H410 (אֵל); seen of God;","strongs_def":"Chaziel, a Levite","kjv_def":"Haziel."},
"H2382":{"lemma":"חֲזָיָה","xlit":"Chăzâyâh","pron":"khaz-aw-yaw'","derivation":"from H2372 (חָזָה) and H3050 (יָהּ); Jah has seen;","strongs_def":"Chazajah, an Israelite","kjv_def":"Hazaiah."},
"H2383":{"lemma":"חֶזְיוֹן","xlit":"Chezyôwn","pron":"khez-yone'","derivation":"from H2372 (חָזָה); vision;","strongs_def":"Chezjon, a Syrian","kjv_def":"Hezion."},
"H2384":{"lemma":"חִזָּיוֹן","xlit":"chizzâyôwn","pron":"khiz-zaw-yone'","derivation":"from H2372 (חָזָה);","strongs_def":"a revelation, expectation by dream","kjv_def":"vision."},
"H2385":{"lemma":"חֲזִיז","xlit":"chăzîyz","pron":"khaw-zeez'","derivation":"from an unused root meaning to glare;","strongs_def":"a flash of lightning","kjv_def":"bright cloud, lightning."},
"H2386":{"lemma":"חֲזִיר","xlit":"chăzîyr","pron":"khaz-eer'","derivation":"from an unused root probably meaning to enclose;","strongs_def":"a hog (perhaps as penned)","kjv_def":"boar, swine."},
"H2387":{"lemma":"חֵזִיר","xlit":"Chêzîyr","pron":"khay-zeer'","derivation":"from the same as H2386 (חֲזִיר); perhaps protected;","strongs_def":"Chezir, the name of two Israelites","kjv_def":"Hezir."},
"H2388":{"lemma":"חָזַק","xlit":"châzaq","pron":"khaw-zak'","derivation":"a primitive root;","strongs_def":"to fasten upon; hence, to seize, be strong (figuratively, courageous, causatively strengthen, cure, help, repair, fortify), obstinate; to bind, restrain, conquer","kjv_def":"aid, amend, [idiom] calker, catch, cleave, confirm, be constant, constrain, continue, be of good (take) courage(-ous, -ly), encourage (self), be established, fasten, force, fortify, make hard, harden, help, (lay) hold (fast), lean, maintain, play the man, mend, become (wax) mighty, prevail, be recovered, repair, retain, seize, be (wax) sore, strengthen (self), be stout, be (make, shew, wax) strong(-er), be sure, take (hold), be urgent, behave self valiantly, withstand."},
"H2389":{"lemma":"חָזָק","xlit":"châzâq","pron":"khaw-zawk'","derivation":"from H2388 (חָזַק);","strongs_def":"strong (usu. in a bad sense, hard, bold, violent)","kjv_def":"harder, hottest, [phrase] impudent, loud, mighty, sore, stiff(-hearted), strong(-er)."},
"H2390":{"lemma":"חָזֵק","xlit":"châzêq","pron":"khaw-zake'","derivation":"from H2388 (חָזַק);","strongs_def":"powerful","kjv_def":"[idiom] wax louder, stronger."},
"H2391":{"lemma":"חֵזֶק","xlit":"chêzeq","pron":"khay'-zek","derivation":"from H2388 (חָזַק);","strongs_def":"help","kjv_def":"strength."},
"H2392":{"lemma":"חֹזֶק","xlit":"chôzeq","pron":"kho'-zek","derivation":"from H2388 (חָזַק);","strongs_def":"power","kjv_def":"strength."},
"H2393":{"lemma":"חֶזְקָה","xlit":"chezqâh","pron":"khez-kaw'","derivation":"feminine of H2391 (חֵזֶק);","strongs_def":"prevailing power","kjv_def":"strength(-en self), (was) strong."},
"H2394":{"lemma":"חׇזְקָה","xlit":"chozqâh","pron":"khoz-kaw'","derivation":"feminine of H2392 (חֹזֶק);","strongs_def":"vehemence (usually in a bad sense)","kjv_def":"force, mightily, repair, sharply."},
"H2395":{"lemma":"חִזְקִי","xlit":"Chizqîy","pron":"khiz-kee'","derivation":"from H2388 (חָזַק); strong;","strongs_def":"Chizki, an Israelite","kjv_def":"Hezeki."},
"H2396":{"lemma":"חִזְקִיָּה","xlit":"Chizqîyâh","pron":"khiz-kee-yaw'","derivation":"or חִזְקִיָּהוּ; also יְחִזְקִיָּה; or יְחִזְקִיָּהוּ; from H2388 (חָזַק) and H3050 (יָהּ); strengthened of Jah;","strongs_def":"Chizkijah, a king of Judah, also the name of two other Israelites","kjv_def":"Hezekiah, Hizkiah, Hizkijah. Compare H3169 (יְחִזְקִיָּה)."},
"H2397":{"lemma":"חָח","xlit":"châch","pron":"khawkh","derivation":"once (Ezekiel 29:4) חָחִי; from the same as H2336 (חוֹחַ);","strongs_def":"a ring for the nose (or lips)","kjv_def":"bracelet, chain, hook."},
"H2398":{"lemma":"חָטָא","xlit":"châṭâʼ","pron":"khaw-taw'","derivation":"a primitive root;","strongs_def":"properly, to miss; hence (figuratively and generally) to sin; by inference, to forfeit, lack, expiate, repent, (causatively) lead astray, condemn","kjv_def":"bear the blame, cleanse, commit (sin), by fault, harm he hath done, loss, miss, (make) offend(-er), offer for sin, purge, purify (self), make reconciliation, (cause, make) sin(-ful, -ness), trespass."},
"H2399":{"lemma":"חֵטְא","xlit":"chêṭᵉʼ","pron":"khate","derivation":"from H2398 (חָטָא);","strongs_def":"a crime or its penalty","kjv_def":"fault, [idiom] grievously, offence, (punishment of) sin."},
"H2400":{"lemma":"חַטָּא","xlit":"chaṭṭâʼ","pron":"khat-taw'","derivation":"intensively from H2398 (חָטָא);","strongs_def":"a criminal, or one accounted guilty","kjv_def":"offender, sinful, sinner."},
"H2401":{"lemma":"חֲטָאָה","xlit":"chăṭâʼâh","pron":"khat-aw-aw'","derivation":"feminine of H2399 (חֵטְא);","strongs_def":"an offence, or a sacrifice forit","kjv_def":"sin (offering)."},
"H2402":{"lemma":"חַטָּאָה","xlit":"chaṭṭâʼâh","pron":"khat-taw-aw'","derivation":"(Aramaic) corresponding to H2401 (חֲטָאָה);","strongs_def":"an offence, and the penalty or sacrifice forit","kjv_def":"sin (offering)."},
"H2403":{"lemma":"חַטָּאָה","xlit":"chaṭṭâʼâh","pron":"khat-taw-aw'","derivation":"or חַטָּאת; from H2398 (חָטָא);","strongs_def":"an offence (sometimes habitual sinfulness), and its penalty, occasion, sacrifice, or expiation; also (concretely) an offender","kjv_def":"punishment (of sin), purifying(-fication for sin), sin(-ner, offering)."},
"H2404":{"lemma":"חָטַב","xlit":"châṭab","pron":"khaw-tab'","derivation":"a primitive root;","strongs_def":"to chop or carve wood","kjv_def":"cut down, hew(-er), polish."},
"H2405":{"lemma":"חֲטֻבָה","xlit":"chăṭubâh","pron":"khat-oo-baw'","derivation":"feminine passive participle of H2404 (חָטַב);","strongs_def":"properly, a carving; hence, a tapestry (as figured)","kjv_def":"carved."},
"H2406":{"lemma":"חִטָּה","xlit":"chiṭṭâh","pron":"khit-taw'","derivation":"of uncertain derivation;","strongs_def":"wheat, whether the grain or the plant","kjv_def":"wheat(-en)."},
"H2407":{"lemma":"חַטּוּשׁ","xlit":"Chaṭṭûwsh","pron":"khat-toosh'","derivation":"from an unused root of uncertain signification;","strongs_def":"Chattush, the name of four or five Israelites","kjv_def":"Hattush."},
"H2408":{"lemma":"חֲטִי","xlit":"chăṭîy","pron":"khat-ee'","derivation":"(Aramaic) from a root corresponding to H2398 (חָטָא);","strongs_def":"an offence","kjv_def":"sin."},
"H2409":{"lemma":"חֲטָּיָא","xlit":"chăṭṭâyâʼ","pron":"khat-taw-yaw'","derivation":"(Aramaic) from the same as H2408 (חֲטִי);","strongs_def":"an expiation","kjv_def":"sin offering."},
"H2410":{"lemma":"חֲטִיטָא","xlit":"Chăṭîyṭâʼ","pron":"khat-ee-taw'","derivation":"from an unused root apparently meaning to dig out; explorer;","strongs_def":"Chatita, a temple porter","kjv_def":"Hatita."},
"H2411":{"lemma":"חַטִּיל","xlit":"Chaṭṭîyl","pron":"khat-teel'","derivation":"from an unused root apparently meaning to wave; fluctuating;","strongs_def":"Chattil, one of 'Solomon's servants'","kjv_def":"Hattil."},
"H2412":{"lemma":"חֲטִיפָא","xlit":"Chăṭîyphâʼ","pron":"khat-ee-faw'","derivation":"from H2414 (חָטַף); robber;","strongs_def":"Chatipha, one of the Nethinim","kjv_def":"Hatipha."},
"H2413":{"lemma":"חָטַם","xlit":"châṭam","pron":"khaw-tam'","derivation":"a primitive root;","strongs_def":"to stop","kjv_def":"refrain."},
"H2414":{"lemma":"חָטַף","xlit":"châṭaph","pron":"khaw-taf'","derivation":"a primitive root;","strongs_def":"to clutch; hence, to seize as aprisoner","kjv_def":"catch."},
"H2415":{"lemma":"חֹטֵר","xlit":"chôṭêr","pron":"kho'-ter","derivation":"from an unused root of uncertain signification;","strongs_def":"a twig","kjv_def":"rod."},
"H2416":{"lemma":"חַי","xlit":"chay","pron":"khah'-ee","derivation":"from H2421 (חָיָה);","strongs_def":"alive; hence, raw (flesh); fresh (plant, water, year), strong; also (as noun, especially in the feminine singular and masculine plural) life (or living thing), whether literally or figuratively","kjv_def":"[phrase] age, alive, appetite, (wild) beast, company, congregation, life(-time), live(-ly), living (creature, thing), maintenance, [phrase] merry, multitude, [phrase] (be) old, quick, raw, running, springing, troop."},
"H2417":{"lemma":"חַי","xlit":"chay","pron":"khah'-ee","derivation":"(Aramaic) from H2418 (חֲיָא);","strongs_def":"alive; also (as noun in plural) life","kjv_def":"life, that liveth, living."},
"H2418":{"lemma":"חֲיָא","xlit":"chăyâʼ","pron":"khah-yaw'","derivation":"(Aramaic) or חֲיָה; (Aramaic), corresponding to H2421 (חָיָה);","strongs_def":"to live","kjv_def":"live, keep alive."},
"H2419":{"lemma":"חִיאֵל","xlit":"Chîyʼêl","pron":"khee-ale'","derivation":"from H2416 (חַי) and H410 (אֵל); living of God;","strongs_def":"Chiel, an Israelite","kjv_def":"Hiel."},
"H2420":{"lemma":"חִידָה","xlit":"chîydâh","pron":"khee-daw'","derivation":"from H2330 (חוּד);","strongs_def":"a puzzle, hence, a trick, conundrum, sententious maxim","kjv_def":"dark saying (sentence, speech), hard question, proverb, riddle."},
"H2421":{"lemma":"חָיָה","xlit":"châyâh","pron":"khaw-yaw'","derivation":"a primitive root (compare H2331 (חָוָה), H2421 (חָיָה));","strongs_def":"to live, whether literally or figuratively; causatively, to revive","kjv_def":"keep (leave, make) alive, [idiom] certainly, give (promise) life, (let, suffer to) live, nourish up, preserve (alive), quicken, recover, repair, restore (to life), revive, ([idiom] God) save (alive, life, lives), [idiom] surely, be whole."},
"H2422":{"lemma":"חָיֶה","xlit":"châyeh","pron":"khaw-yeh'","derivation":"from H2421 (חָיָה);","strongs_def":"vigorous","kjv_def":"lively."},
"H2423":{"lemma":"חֵיוָא","xlit":"chêyvâʼ","pron":"khay-vaw'","derivation":"(Aramaic) from H2418 (חֲיָא);","strongs_def":"an animal","kjv_def":"beast."},
"H2424":{"lemma":"חַיּוּת","xlit":"chayûwth","pron":"khah-yooth'","derivation":"from H2421 (חָיָה);","strongs_def":"life","kjv_def":"[idiom] living."},
"H2425":{"lemma":"חָיַי","xlit":"châyay","pron":"khaw-yah'-ee","derivation":"a primitive root (compare H2421 (חָיָה));","strongs_def":"to live; causatively to revive","kjv_def":"live, save life."},
"H2426":{"lemma":"חֵיל","xlit":"chêyl","pron":"khale","derivation":"or (shortened) חֵל; a collateral form of H2428 (חַיִל);","strongs_def":"an army; also (by analogy,) an intrenchment","kjv_def":"army, bulwark, host, [phrase] poor, rampart, trench, wall."},
"H2427":{"lemma":"חִיל","xlit":"chîyl","pron":"kheel","derivation":"and (feminine) חִילָה; from H2342 (חוּל);","strongs_def":"a throe (expectant of childbirth)","kjv_def":"pain, pang, sorrow."},
"H2428":{"lemma":"חַיִל","xlit":"chayil","pron":"khah'-yil","derivation":"from H2342 (חוּל);","strongs_def":"probably a force, whether of men, means or other resources; an army, wealth, virtue, valor, strength","kjv_def":"able, activity, ([phrase]) army, band of men (soldiers), company, (great) forces, goods, host, might, power, riches, strength, strong, substance, train, ([phrase]) valiant(-ly), valour, virtuous(-ly), war, worthy(-ily)."},
"H2429":{"lemma":"חַיִל","xlit":"chayil","pron":"khah'-yil","derivation":"(Aramaic) corresponding to H2428 (חַיִל);","strongs_def":"an army, or strength","kjv_def":"aloud, army, [idiom] most (mighty), power."},
"H2430":{"lemma":"חֵילָה","xlit":"chêylâh","pron":"khay-law'","derivation":"feminine of H2428 (חַיִל);","strongs_def":"an intrenchment","kjv_def":"bulwark."},
"H2431":{"lemma":"חֵילָם","xlit":"Chêylâm","pron":"khay-lawm'","derivation":"or חֵלָאםxlit Chêlʼâm corrected to Chêlâʼm; from H2428 (חַיִל); fortress;","strongs_def":"Chelam, a place East of Palestine","kjv_def":"Helam."},
"H2432":{"lemma":"חִילֵן","xlit":"Chîylên","pron":"khee-lane'","derivation":"from H2428 (חַיִל); fortress;","strongs_def":"Chilen, a place in Palestine","kjv_def":"Hilen."},
"H2433":{"lemma":"חִין","xlit":"chîyn","pron":"kheen","derivation":"another form for H2580 (חֵן);","strongs_def":"beauty","kjv_def":"comely."},
"H2434":{"lemma":"חַיִץ","xlit":"chayits","pron":"khah'-yits","derivation":"another form for H2351 (חוּץ);","strongs_def":"a wall","kjv_def":"wall."},
"H2435":{"lemma":"חִיצוֹן","xlit":"chîytsôwn","pron":"khee-tsone'","derivation":"from H2434 (חַיִץ);","strongs_def":"properly, the (outer) wall side; hence, exterior; figuratively, secular (as opposed to sacred)","kjv_def":"outer, outward, utter, without."},
"H2436":{"lemma":"חֵיק","xlit":"chêyq","pron":"khake","derivation":"or חֵק; and חוֹק; from an unused root, apparently meaning to inclose;","strongs_def":"the bosom (literally or figuratively)","kjv_def":"bosom, bottom, lap, midst, within."},
"H2437":{"lemma":"חִירָה","xlit":"Chîyrâh","pron":"khee-raw'","derivation":"from H2357 (חָוַר) in the sense of splendor;","strongs_def":"Chirah, an Adullamite","kjv_def":"Hirah."},
"H2438":{"lemma":"חִירָם","xlit":"Chîyrâm","pron":"khee-rawm'","derivation":"or חִירוֹםlemma חִירֹם missing vav, corrected to חִירוֹם; another form of H2361 (חוּרָם);","strongs_def":"Chiram or Chirom, the name of two Tyrians","kjv_def":"Hiram, Huram."},
"H2439":{"lemma":"חִישׁ","xlit":"chîysh","pron":"kheesh","derivation":"another form of H2363 (חוּשׁ);","strongs_def":"to hurry","kjv_def":"make haste."},
"H2440":{"lemma":"חִישׁ","xlit":"chîysh","pron":"kheesh","derivation":"from H2439 (חִישׁ);","strongs_def":"properly, a hurry; hence (adverb) quickly","kjv_def":"soon."},
"H2441":{"lemma":"חֵךְ","xlit":"chêk","pron":"khake","derivation":"probably from H2596 (חָנַךְ) in the sense of tasting;","strongs_def":"properly, the palate or inside of the mouth; hence, the mouth itself (as the organ of speech, taste and kissing)","kjv_def":"(roof of the) mouth, taste."},
"H2442":{"lemma":"חָכָה","xlit":"châkâh","pron":"khaw-kaw'","derivation":"a primitive root (apparently akin to H2707 (חָקָה) through the idea of piercing);","strongs_def":"properly, to adhere to; hence, to await","kjv_def":"long, tarry, wait."},
"H2443":{"lemma":"חַכָּה","xlit":"chakkâh","pron":"khak-kaw'","derivation":"probably from H2442 (חָכָה);","strongs_def":"a hook (as adhering)","kjv_def":"angle, hook."},
"H2444":{"lemma":"חֲכִילָה","xlit":"Chăkîylâh","pron":"khak-ee-law'","derivation":"from the same as H2447 (חַכְלִיל); dark;","strongs_def":"Chakilah, a hill in Palestine","kjv_def":"Hachilah."},
"H2445":{"lemma":"חַכִּים","xlit":"chakkîym","pron":"khak-keem'","derivation":"(Aramaic) from a root corresponding to H2449 (חָכַם);","strongs_def":"wise, i.e. a Magian","kjv_def":"wise."},
"H2446":{"lemma":"חֲכַלְיָה","xlit":"Chăkalyâh","pron":"khak-al-yaw'","derivation":"from the base of H2447 (חַכְלִיל) and H3050 (יָהּ); darkness of Jah;","strongs_def":"Chakaljah, an Israelite","kjv_def":"Hachaliah."},
"H2447":{"lemma":"חַכְלִיל","xlit":"chaklîyl","pron":"khak-leel'","derivation":"by reduplication from an unused root apparently meaning to be dark;","strongs_def":"darkly flashing (only of the eyes); in a good sense, brilliant (as stimulated by wine)","kjv_def":"red."},
"H2448":{"lemma":"חַכְלִלוּת","xlit":"chaklilûwth","pron":"khak-lee-looth'","derivation":"from H2447 (חַכְלִיל);","strongs_def":"flash (of the eyes); in a bad sense, blearedness","kjv_def":"redness."},
"H2449":{"lemma":"חָכַם","xlit":"châkam","pron":"khaw-kam'","derivation":"a primitive root,","strongs_def":"to be wise (in mind, word or act)","kjv_def":"[idiom] exceeding, teach wisdom, be (make self, shew self) wise, deal (never so) wisely, make wiser."},
"H2450":{"lemma":"חָכָם","xlit":"châkâm","pron":"khaw-kawm'","derivation":"from H2449 (חָכַם);","strongs_def":"wise, (i.e. intelligent, skilful or artful)","kjv_def":"cunning (man), subtil, (un-), wise((hearted), man)."},
"H2451":{"lemma":"חׇכְמָה","xlit":"chokmâh","pron":"khok-maw'","derivation":"from H2449 (חָכַם);","strongs_def":"wisdom (in a good sense)","kjv_def":"skilful, wisdom, wisely, wit."},
"H2452":{"lemma":"חׇכְמָה","xlit":"chokmâh","pron":"khok-maw'","derivation":"(Aramaic) corresponding to H2451 (חׇכְמָה);","strongs_def":"wisdom","kjv_def":"wisdom."},
"H2453":{"lemma":"חַכְמוֹנִי","xlit":"Chakmôwnîy","pron":"khak-mo-nee'","derivation":"from H2449 (חָכַם); skilful;","strongs_def":"Chakmoni, an Israelite","kjv_def":"Hachmoni, Hachmonite."},
"H2454":{"lemma":"חׇכְמוֹת","xlit":"chokmôwth","pron":"khok-moth'","derivation":"or חַכְמוֹת; collateral forms of H2451 (חׇכְמָה);","strongs_def":"wisdom","kjv_def":"wisdom, every wise (woman)."},
"H2455":{"lemma":"חֹל","xlit":"chôl","pron":"khole","derivation":"from H2490 (חָלַל);","strongs_def":"properly, exposed; hence, profane","kjv_def":"common, profane (place), unholy."},
"H2456":{"lemma":"חָלָא","xlit":"châlâʼ","pron":"khaw-law'","derivation":"a primitive root (compare H2470 (חָלָה));","strongs_def":"to be sick","kjv_def":"be diseased."},
"H2457":{"lemma":"חֶלְאָה","xlit":"chelʼâh","pron":"khel-aw'","derivation":"from H2456 (חָלָא);","strongs_def":"properly, disease; hence, rust","kjv_def":"scum."},
"H2458":{"lemma":"חֶלְאָה","xlit":"Chelʼâh","pron":"khel-aw'","derivation":"the same as H2457 (חֶלְאָה);","strongs_def":"Chelah, an Israelitess","kjv_def":"Helah."},
"H2459":{"lemma":"חֶלֶב","xlit":"cheleb","pron":"kheh'-leb","derivation":"or חֵלֶב; from an unused root meaning to be fat;","strongs_def":"fat, whether literally or figuratively; hence, the richest or choice part","kjv_def":"[idiom] best, fat(-ness), [idiom] finest, grease, marrow."},
"H2460":{"lemma":"חֵלֶב","xlit":"Chêleb","pron":"khay'-leb","derivation":"the same as H2459 (חֶלֶב); fatness;","strongs_def":"Cheleb, an Israelite","kjv_def":"Heleb."},
"H2461":{"lemma":"חָלָב","xlit":"châlâb","pron":"khaw-lawb'","derivation":"from the same as H2459 (חֶלֶב);","strongs_def":"milk (as the richness of kine)","kjv_def":"[phrase] cheese, milk, sucking."},
"H2462":{"lemma":"חֶלְבָּה","xlit":"Chelbâh","pron":"khel-baw'","derivation":"feminine of H2459 (חֶלֶב); fertility;","strongs_def":"Chelbah, a place in Palestine","kjv_def":"Helbah."},
"H2463":{"lemma":"חֶלְבּוֹן","xlit":"Chelbôwn","pron":"khel-bone'","derivation":"from H2459 (חֶלֶב); fruitful;","strongs_def":"Chelbon, a place in Syria","kjv_def":"Helbah."},
"H2464":{"lemma":"חֶלְבְּנָה","xlit":"chelbᵉnâh","pron":"khel-ben-aw'","derivation":"from H2459 (חֶלֶב);","strongs_def":"galbanam, an odorous gum (as if fatty)","kjv_def":"galbanum."},
"H2465":{"lemma":"חֶלֶד","xlit":"cheled","pron":"kheh'-led","derivation":"from an unused root apparently meaning to glide swiftly;","strongs_def":"life (as a fleeting portion of time); hence, the world (as transient)","kjv_def":"age, short time, world."},
"H2466":{"lemma":"חֵלֶד","xlit":"chêled","pron":"khay'-led","derivation":"the same as H2465 (חֶלֶד);","strongs_def":"Cheled, an Israelite","kjv_def":"Heled."},
"H2467":{"lemma":"חֹלֶד","xlit":"chôled","pron":"kho'-led","derivation":"from the same as H2465 (חֶלֶד);","strongs_def":"a weasel (from its gliding motion)","kjv_def":"weasel."},
"H2468":{"lemma":"חֻלְדָּה","xlit":"Chuldâh","pron":"khool-daw'","derivation":"feminine of H2467 (חֹלֶד);","strongs_def":"Chuldah, an Israelitess","kjv_def":"Huldah."},
"H2469":{"lemma":"חֶלְדַּי","xlit":"Chelday","pron":"khel-dah'-ee","derivation":"from H2466 (חֵלֶד); worldliness;","strongs_def":"Cheldai, the name of two Israelites","kjv_def":"Heldai."},
"H2470":{"lemma":"חָלָה","xlit":"châlâh","pron":"khaw-law'","derivation":"a primitive root (compare H2342 (חוּל), H2470 (חָלָה), H2490 (חָלַל));","strongs_def":"properly, to be rubbed or worn; hence (figuratively) to be weak, sick, afflicted; or (causatively) to grieve, make sick; also to stroke (in flattering), entreat","kjv_def":"beseech, (be) diseased, (put to) grief, be grieved, (be) grievous, infirmity, intreat, lay to, put to pain, [idiom] pray, make prayer, be (fall, make) sick, sore, be sorry, make suit ([idiom] supplication), woman in travail, be (become) weak, be wounded."},
"H2471":{"lemma":"חַלָּה","xlit":"challâh","pron":"khal-law'","derivation":"from H2490 (חָלַל);","strongs_def":"a cake (as usually punctured)","kjv_def":"cake."},
"H2472":{"lemma":"חֲלוֹם","xlit":"chălôwm","pron":"khal-ome'","derivation":"or (shortened) חֲלֹם; from H2492 (חָלַם);","strongs_def":"a dream","kjv_def":"dream(-er)."},
"H2473":{"lemma":"חֹלוֹן","xlit":"Chôlôwn","pron":"kho-lone'","derivation":"or (shortened) חֹלֹן; probably from H2344 (חוֹל); sandy;","strongs_def":"Cholon, the name of two places in Palestine","kjv_def":"Holon."},
"H2474":{"lemma":"חַלּוֹן","xlit":"challôwn","pron":"khal-lone'","strongs_def":"a window (as perforated)","kjv_def":"window."},
"H2475":{"lemma":"חֲלוֹף","xlit":"chălôwph","pron":"khal-ofe'","derivation":"from H2498 (חָלַף);","strongs_def":"properly, surviving; by implication (collectively) orphans","kjv_def":"[idiom] destruction."},
"H2476":{"lemma":"חֲלוּשָׁה","xlit":"chălûwshâh","pron":"khal-oo-shaw'","derivation":"feminine passive participle of H2522 (חָלַשׁ);","strongs_def":"defeat","kjv_def":"being overcome."},
"H2477":{"lemma":"חֲלַח","xlit":"Chălach","pron":"khal-akh'","derivation":"probably of foreign origin;","strongs_def":"Chalach, a region of Assyria","kjv_def":"Halah."},
"H2478":{"lemma":"חַלְחוּל","xlit":"Chalchûwl","pron":"khal-khool'","derivation":"by reduplication from H2342 (חוּל); contorted;","strongs_def":"Chalchul, a place in Palestine","kjv_def":"Halhul."},
"H2479":{"lemma":"חַלְחָלָה","xlit":"chalchâlâh","pron":"khal-khaw-law'","derivation":"feminine from the same as H2478 (חַלְחוּל);","strongs_def":"writhing (in childbirth); by implication, terror","kjv_def":"(great, much) pain."},
"H2480":{"lemma":"חָלַט","xlit":"châlaṭ","pron":"khaw-lat'","derivation":"a primitive root;","strongs_def":"to snatch at","kjv_def":"catch."},
"H2481":{"lemma":"חֲלִי","xlit":"chălîy","pron":"khal-ee'","derivation":"from H2470 (חָלָה);","strongs_def":"a trinket (as polished)","kjv_def":"jewel, ornament."},
"H2482":{"lemma":"חֲלִי","xlit":"Chălîy","pron":"khal-ee'","derivation":"the same as H2481 (חֲלִי);","strongs_def":"Chali, a place in Palestine","kjv_def":"Hali."},
"H2483":{"lemma":"חֳלִי","xlit":"chŏlîy","pron":"khol-ee'","derivation":"from H2470 (חָלָה);","strongs_def":"malady, anxiety, calamity","kjv_def":"disease, grief, (is) sick(-ness)."},
"H2484":{"lemma":"חֶלְיָה","xlit":"chelyâh","pron":"khel-yaw'","derivation":"feminine of H2481 (חֲלִי);","strongs_def":"a trinket","kjv_def":"jewel."},
"H2485":{"lemma":"חָלִיל","xlit":"châlîyl","pron":"khaw-leel'","derivation":"from H2490 (חָלַל);","strongs_def":"a flute (as perforated)","kjv_def":"pipe."},
"H2486":{"lemma":"חָלִילָה","xlit":"châlîylâh","pron":"khaw-lee'-law","derivation":"or חָלִלָה; a directive from H2490 (חָלַל);","strongs_def":"literal fora profaned thing; used (interj.) far be it!","kjv_def":"be far, ([idiom] God) forbid."},
"H2487":{"lemma":"חֲלִיפָה","xlit":"chălîyphâh","pron":"khal-ee-faw'","derivation":"from H2498 (חָלַף);","strongs_def":"alternation","kjv_def":"change, course."},
"H2488":{"lemma":"חֲלִיצָה","xlit":"chălîytsâh","pron":"khal-ee-tsaw'","derivation":"from H2503 (חֶלֶץ);","strongs_def":"spoil","kjv_def":"armour."},
"H2489":{"lemma":"חֵלְכָא","xlit":"chêlᵉkâʼ","pron":"khay-lek-aw'","derivation":"or חֵלְכָה; apparently from an unused root probably meaning to be dark or (figuratively) unhappy;","strongs_def":"a wretch, i.e. unfortunate","kjv_def":"poor."},
"H2490":{"lemma":"חָלַל","xlit":"châlal","pron":"khaw-lal'","derivation":"a primitive root (compare H2470 (חָלָה)); also  denominative (from H2485 (חָלִיל))","strongs_def":"properly, to bore, i.e. (by implication) to wound, to dissolve; figuratively, to profane (a person, place or thing), to break (one's word), to begin (as if by an 'opening wedge'); to play (the flute)","kjv_def":"begin ([idiom] men began), defile, [idiom] break, defile, [idiom] eat (as common things), [idiom] first, [idiom] gather the grape thereof, [idiom] take inheritance, pipe, player on instruments, pollute, (cast as) profane (self), prostitute, slay (slain), sorrow, stain, wound."},
"H2491":{"lemma":"חָלָל","xlit":"châlâl","pron":"khaw-lawl'","derivation":"from H2490 (חָלַל);","strongs_def":"pierced (especially to death); figuratively, polluted","kjv_def":"kill, profane, slain (man), [idiom] slew, (deadly) wounded."},
"H2492":{"lemma":"חָלַם","xlit":"châlam","pron":"khaw-lam'","derivation":"a primitive root;","strongs_def":"properly, to bind firmly, i.e. (by implication) to be (causatively to make) plump; also (through the figurative sense of dumbness) to dream","kjv_def":"(cause to) dream(-er), be in good liking, recover."},
"H2493":{"lemma":"חֵלֶם","xlit":"chêlem","pron":"khay'-lem","derivation":"(Aramaic) from a root corresponding to H2492 (חָלַם);","strongs_def":"a dream","kjv_def":"dream."},
"H2494":{"lemma":"חֵלֶם","xlit":"Chêlem","pron":"khay'lem","derivation":"from H2492 (חָלַם); a dream;","strongs_def":"Chelem, an Israelite","kjv_def":"Helem. Compare H2469 (חֶלְדַּי)."},
"H2495":{"lemma":"חַלָּמוּת","xlit":"challâmûwth","pron":"khal-law-mooth'","derivation":"from H2492 (חָלַם) (in the sense of insipidity);","strongs_def":"probably purslain","kjv_def":"egg."},
"H2496":{"lemma":"חַלָּמִישׁ","xlit":"challâmîysh","pron":"klal-law-meesh'","derivation":"probably from H2492 (חָלַם) (in the sense of hardness);","strongs_def":"flint","kjv_def":"flint(-y), rock."},
"H2497":{"lemma":"חֵלֹן","xlit":"Chêlôn","pron":"khay-lone'","derivation":"from H2428 (חַיִל); strong;","strongs_def":"Chelon, an Israelite","kjv_def":"Helon."},
"H2498":{"lemma":"חָלַף","xlit":"châlaph","pron":"khaw-laf'","derivation":"a primitive root;","strongs_def":"properly, to slide by, i.e. (by implication) to hasten away, pass on, spring up, pierce or change","kjv_def":"abolish, alter, change, cut off, go on forward, grow up, be over, pass (away, on, through), renew, sprout, strike through."},
"H2499":{"lemma":"חֲלַף","xlit":"chălaph","pron":"khal-af'","derivation":"(Aramaic) corresponding to H2498 (חָלַף);","strongs_def":"to pass on (of time)","kjv_def":"pass."},
"H2500":{"lemma":"חֵלֶף","xlit":"chêleph","pron":"klay'-lef","derivation":"from H2498 (חָלַף);","strongs_def":"properly, exchange; hence (as preposition) instead of","kjv_def":"[idiom] for."},
"H2501":{"lemma":"חֶלֶף","xlit":"Cheleph","pron":"kheh'lef","derivation":"the same as H2500 (חֵלֶף); change;","strongs_def":"Cheleph, a place in Palestine","kjv_def":"Heleph."},
"H2502":{"lemma":"חָלַץ","xlit":"châlats","pron":"khaw-lats'","derivation":"a primitive root;","strongs_def":"to pull off; hence (intensively) to strip, (reflex.) to depart; by implication, to deliver, equip (for fight); present, strengthen","kjv_def":"arm (self), (go, ready) armed ([idiom] man, soldier), deliver, draw out, make fat, loose, (ready) prepared, put off, take away, withdraw self."},
"H2503":{"lemma":"חֶלֶץ","xlit":"Chelets","pron":"kheh'-lets","derivation":"or חֵלֶץ; from H2502 (חָלַץ); perhaps, strength;","strongs_def":"Chelets, the name of two Israelites","kjv_def":"Helez."},
"H2504":{"lemma":"חָלָץ","xlit":"châlâts","pron":"khaw-lawts'","derivation":"from H2502 (חָלַץ) (in the sense of strength); only in the dual;","strongs_def":"the loins (as the seat of vigor)","kjv_def":"loins, reins."},
"H2505":{"lemma":"חָלַק","xlit":"châlaq","pron":"khaw-lak'","derivation":"a primitive root;","strongs_def":"to be smooth (figuratively); by implication (as smooth stones were used for lots) to apportion or separate","kjv_def":"deal, distribute, divide, flatter, give, (have, im-) part(-ner), take away a portion, receive, separate self, (be) smooth(-er)."},
"H2506":{"lemma":"חֵלֶק","xlit":"chêleq","pron":"khay'lek","derivation":"from H2505 (חָלַק);","strongs_def":"properly, smoothness (of the tongue); also an allotment","kjv_def":"flattery, inheritance, part, [idiom] partake, portion."},
"H2507":{"lemma":"חֵלֶק","xlit":"Chêleq","pron":"khay'-lek","derivation":"the same as H2506 (חֵלֶק); portion;","strongs_def":"Chelek, an Israelite","kjv_def":"Helek."},
"H2508":{"lemma":"חֲלָק","xlit":"chălâq","pron":"khal-awk'","derivation":"(Aramaic) from a root corresponding to H2505 (חָלַק);","strongs_def":"a part","kjv_def":"portion."},
"H2509":{"lemma":"חָלָק","xlit":"châlâq","pron":"khaw-lawk'","derivation":"from H2505 (חָלַק);","strongs_def":"smooth (especially of tongue)","kjv_def":"flattering, smooth."},
"H2510":{"lemma":"חָלָק","xlit":"Châlâq","pron":"khaw-lawk'","derivation":"the same as H2509 (חָלָק); bare;","strongs_def":"Chalak, a mountain of Idumaea","kjv_def":"Halak."},
"H2511":{"lemma":"חַלָּק","xlit":"challâq","pron":"khal-lawk'","derivation":"from H2505 (חָלַק);","strongs_def":"smooth","kjv_def":"smooth."},
"H2512":{"lemma":"חַלֻּק","xlit":"challuq","pron":"khal-look'","derivation":"from H2505 (חָלַק);","strongs_def":"smooth","kjv_def":"smooth."},
"H2513":{"lemma":"חֶלְקָה","xlit":"chelqâh","pron":"khel-kaw'","derivation":"feminine of H2506 (חֵלֶק);","strongs_def":"properly, smoothness; figuratively, flattery; also an allotment","kjv_def":"field, flattering(-ry), ground, parcel, part, piece of land (ground), plat, portion, slippery place, smooth (thing)."},
"H2514":{"lemma":"חֲלַקָּה","xlit":"chălaqqâh","pron":"khal-ak-kaw'","derivation":"feminine from H2505 (חָלַק);","strongs_def":"flattery","kjv_def":"flattery."},
"H2515":{"lemma":"חֲלֻקָּה","xlit":"chăluqqâh","pron":"khal-ook-kaw'","derivation":"feminine of H2512 (חַלֻּק);","strongs_def":"a distribution","kjv_def":"division."},
"H2516":{"lemma":"חֶלְקִי","xlit":"Chelqîy","pron":"khel-kee'","derivation":"patronymically from H2507 (חֵלֶק);","strongs_def":"a Chelkite or descendant of Chelek","kjv_def":"Helkites."},
"H2517":{"lemma":"חֶלְקַי","xlit":"Chelqay","pron":"khel-kah'ee","derivation":"from H2505 (חָלַק); apportioned;","strongs_def":"Chelkai, an Israelite","kjv_def":"Helkai."},
"H2518":{"lemma":"חִלְקִיָּה","xlit":"Chilqîyâh","pron":"khil-kee-yaw'","derivation":"or חִלְקִיָּהוּ; from H2506 (חֵלֶק) and H3050 (יָהּ); portion of Jah;","strongs_def":"Chilhijah, the name of eight Israelites","kjv_def":"Hillkiah."},
"H2519":{"lemma":"חֲלַקְלַקָּה","xlit":"chălaqlaqqâh","pron":"khal-ak-lak-kaw'","derivation":"by reduplication from H2505 (חָלַק);","strongs_def":"properly, something very smooth; i.e. a treacherous spot; figuratively, blandishment","kjv_def":"flattery, slippery."},
"H2520":{"lemma":"חֶלְקַת","xlit":"Chelqath","pron":"khel-kath'","derivation":"a form of H2513 (חֶלְקָה); smoothness;","strongs_def":"Chelkath, a place in Palestine","kjv_def":"Helkath."},
"H2521":{"lemma":"חֶלְקַת הַצֻּרִים","xlit":"Chelqath hats-Tsurîym","pron":"khel-kath' hatstsoo-reem'","derivation":"from H2520 (חֶלְקַת) and the plural of H6697 (צוּר), with the article inserted; smoothness of the rocks;","strongs_def":"Chelkath Hats-tsurim, a place in Palestine","kjv_def":"Helkath-hazzurim."},
"H2522":{"lemma":"חָלַשׁ","xlit":"châlash","pron":"khaw-lash'","derivation":"a primitive root;","strongs_def":"to prostrate; by implication, to overthrow, decay","kjv_def":"discomfit, waste away, weaken."},
"H2523":{"lemma":"חַלָּשׁ","xlit":"challâsh","pron":"khal-lawsh'","derivation":"from H2522 (חָלַשׁ);","strongs_def":"frail","kjv_def":"weak."},
"H2524":{"lemma":"חָם","xlit":"châm","pron":"khawm","derivation":"from the same as H2346 (חוֹמָה);","strongs_def":"a father-in-law (as in affinity)","kjv_def":"father in law."},
"H2525":{"lemma":"חָם","xlit":"châm","pron":"khawm","derivation":"from H2552 (חָמַם);","strongs_def":"hot","kjv_def":"hot, warm."},
"H2526":{"lemma":"חָם","xlit":"Châm","pron":"khawm","derivation":"the same as H2525 (חָם); hot (from the tropical habitat);","strongs_def":"Cham, a son of Noah; also (as a patronymic) his descendants or their country","kjv_def":"Ham."},
"H2527":{"lemma":"חֹם","xlit":"chôm","pron":"khome","derivation":"from H2552 (חָמַם);","strongs_def":"heat","kjv_def":"heat, to be hot (warm)."},
"H2528":{"lemma":"חֱמָא","xlit":"chĕmâʼ","pron":"khem-aw'","derivation":"(Aramaic) or חֲמָה; (Aramaic), corresponding to H2534 (חֵמָה);","strongs_def":"anger","kjv_def":"fury."},
"H2529":{"lemma":"חֶמְאָה","xlit":"chemʼâh","pron":"khem-aw'","derivation":"or (shortened) חֵמָה; from the same root as H2346 (חוֹמָה);","strongs_def":"curdled milk or cheese","kjv_def":"butter."},
"H2530":{"lemma":"חָמַד","xlit":"châmad","pron":"khaw-mad'","derivation":"a primitive root;","strongs_def":"to delight in","kjv_def":"beauty, greatly beloved, covet, delectable thing, ([idiom] great) delight, desire, goodly, lust, (be) pleasant (thing), precious (thing)."},
"H2531":{"lemma":"חֶמֶד","xlit":"chemed","pron":"kheh'-med","derivation":"from H2530 (חָמַד);","strongs_def":"delight","kjv_def":"desirable, pleasant."},
"H2532":{"lemma":"חֶמְדָּה","xlit":"chemdâh","pron":"khem-daw'","derivation":"feminine of H2531 (חֶמֶד);","strongs_def":"delight","kjv_def":"desire, goodly, pleasant, precious."},
"H2533":{"lemma":"חֶמְדָּן","xlit":"Chemdân","pron":"khem-dawn'","derivation":"from H2531 (חֶמֶד); pleasant;","strongs_def":"Chemdan, an Idumaean","kjv_def":"Hemdan."},
"H2534":{"lemma":"חֵמָה","xlit":"chêmâh","pron":"khay-maw'","derivation":"or (Daniel 11:44) חֵמָא; from H3179 (יָחַם);","strongs_def":"heat; figuratively, anger, poison (from its fever)","kjv_def":"anger, bottles, hot displeasure, furious(-ly, -ry), heat, indignation, poison, rage, wrath(-ful). See H2529 (חֶמְאָה)."},
"H2535":{"lemma":"חַמָּה","xlit":"chammâh","pron":"kham-maw'","derivation":"from H2525 (חָם);","strongs_def":"heat; by implication, the sun","kjv_def":"heat, sun."},
"H2536":{"lemma":"חַמּוּאֵל","xlit":"Chammûwʼêl","pron":"kham-moo-ale'","derivation":"from H2535 (חַמָּה) and H410 (אֵל); anger of God;","strongs_def":"Chammuel, an Israelite","kjv_def":"Hamuel."},
"H2537":{"lemma":"חֲמוּטַל","xlit":"Chămûwṭal","pron":"kham-oo-tal'","derivation":"or חֲמִיטַל; from H2524 (חָם) and H2919 (טַל); father-in-law of dew;","strongs_def":"Chamutal or Chamital, an Israelitess","kjv_def":"Hamutal."},
"H2538":{"lemma":"חָמוּל","xlit":"Châmûwl","pron":"khaw-mool'","derivation":"from H2550 (חָמַל); pitied;","strongs_def":"Chamul, an Israelite","kjv_def":"Hamul."},
"H2539":{"lemma":"חָמוּלִי","xlit":"Châmûwlîy","pron":"khaw-moo-lee'","derivation":"patronymically from H2538 (חָמוּל);","strongs_def":"a Chamulite (collectively) or descendants of Chamul","kjv_def":"Hamulites."},
"H2540":{"lemma":"חַמּוֹן","xlit":"Chammôwn","pron":"kham-mone'","derivation":"from H2552 (חָמַם); warm spring;","strongs_def":"Chammon, the name of two places in Palestine","kjv_def":"Hammon."},
"H2541":{"lemma":"חָמוֹץ","xlit":"châmôwts","pron":"khaw-motse'","derivation":"from H2556 (חָמֵץ);","strongs_def":"properly, violent; by implication, a robber","kjv_def":"oppressed."},
"H2542":{"lemma":"חַמּוּק","xlit":"chammûwq","pron":"kham-mook'","derivation":"from H2559 (חָמַק);","strongs_def":"a wrapping, i.e. drawers","kjv_def":"joints."},
"H2543":{"lemma":"חֲמוֹר","xlit":"chămôwr","pron":"kham-ore'","derivation":"or (shortened) חֲמֹר; from H2560 (חָמַר);","strongs_def":"a male ass (from its dun red)","kjv_def":"(he) ass."},
"H2544":{"lemma":"חֲמוֹר","xlit":"Chămôwr","pron":"kham-ore'","derivation":"the same as H2543 (חֲמוֹר); donkey;","strongs_def":"Chamor, a Canaanite","kjv_def":"Hamor."},
"H2545":{"lemma":"חֲמוֹת","xlit":"chămôwth","pron":"kham-oth'","derivation":"or (shortened) חֲמֹת; feminine of H2524 (חָם);","strongs_def":"a mother-in-law","kjv_def":"mother in law."},
"H2546":{"lemma":"חֹמֶט","xlit":"chômeṭ","pron":"kho'met","derivation":"from an unused root probably meaning, to lie low;","strongs_def":"a lizard (as creeping)","kjv_def":"snail."},
"H2547":{"lemma":"חֻמְטָה","xlit":"Chumṭâh","pron":"khoom-taw'","derivation":"feminine of H2546 (חֹמֶט); low;","strongs_def":"Chumtah, a place in Palestine","kjv_def":"Humtah."},
"H2548":{"lemma":"חָמִיץ","xlit":"châmîyts","pron":"khaw-meets'","derivation":"from H2556 (חָמֵץ);","strongs_def":"seasoned, i.e. salt provender","kjv_def":"clean."},
"H2549":{"lemma":"חֲמִישִׁי","xlit":"chămîyshîy","pron":"kham-ee-shee'","derivation":"or חֲמִשִּׁי; ordinal from H2568 (חָמֵשׁ);","strongs_def":"fifth; also a fifth","kjv_def":"fifth (part)."},
"H2550":{"lemma":"חָמַל","xlit":"châmal","pron":"khaw-mal'","derivation":"a primitive root;","strongs_def":"to commiserate; by implication, to spare","kjv_def":"have compassion, (have) pity, spare."},
"H2551":{"lemma":"חֶמְלָה","xlit":"chemlâh","pron":"khem-law'","derivation":"from H2550 (חָמַל);","strongs_def":"commiseration","kjv_def":"merciful, pity."},
"H2552":{"lemma":"חָמַם","xlit":"châmam","pron":"khaw-mam'","derivation":"a primitive root;","strongs_def":"to be hot (literally or figuratively)","kjv_def":"enflame self, get (have) heat, be (wax) hot, (be, wax) warm (self, at)."},
"H2553":{"lemma":"חַמָּן","xlit":"chammân","pron":"kham-mawn'","derivation":"from H2535 (חַמָּה);","strongs_def":"a sun-pillar","kjv_def":"idol, image."},
"H2554":{"lemma":"חָמַס","xlit":"châmaç","pron":"khaw-mas'","derivation":"a primitive root;","strongs_def":"to be violent; by implication, to maltreat","kjv_def":"make bare, shake off, violate, do violence, take away violently, wrong, imagine wrongfully."},
"H2555":{"lemma":"חָמָס","xlit":"châmâç","pron":"khaw-mawce'","derivation":"from H2554 (חָמַס);","strongs_def":"violence; by implication, wrong; by meton. unjust gain","kjv_def":"cruel(-ty), damage, false, injustice, [idiom] oppressor, unrighteous, violence (against, done), violent (dealing), wrong."},
"H2556":{"lemma":"חָמֵץ","xlit":"châmêts","pron":"khaw-mates'","derivation":"a primitive root;","strongs_def":"to be pungent; i.e. in taste (sour, i.e. literally fermented, or figuratively, harsh), in color (dazzling)","kjv_def":"cruel (man), dyed, be grieved, leavened."},
"H2557":{"lemma":"חָמֵץ","xlit":"châmêts","pron":"khaw-mates'","derivation":"from H2556 (חָמֵץ);","strongs_def":"ferment, (figuratively) extortion","kjv_def":"leaven, leavened (bread)."},
"H2558":{"lemma":"חֹמֶץ","xlit":"chômets","pron":"kho'-mets","derivation":"from H2556 (חָמֵץ);","strongs_def":"vinegar","kjv_def":"vinegar."},
"H2559":{"lemma":"חָמַק","xlit":"châmaq","pron":"khaw-mak'","derivation":"a primitive root;","strongs_def":"properly, to wrap; hence, to depart (i.e. turn about)","kjv_def":"go about, withdraw self."},
"H2560":{"lemma":"חָמַר","xlit":"châmar","pron":"khaw-mar'","derivation":"a primitive root; also  as denominative (from H2564 (חֵמָר))","strongs_def":"properly, to boil up; hence, to ferment (with scum); to glow (with redness); to smear with pitch","kjv_def":"daub, befoul, be red, trouble."},
"H2561":{"lemma":"חֶמֶר","xlit":"chemer","pron":"kheh'-mer","derivation":"from H2560 (חָמַר);","strongs_def":"wine (as fermenting)","kjv_def":"[idiom] pure, red wine."},
"H2562":{"lemma":"חֲמַר","xlit":"chămar","pron":"kham-ar'","derivation":"(Aramaic) corresponding to H2561 (חֶמֶר);","strongs_def":"wine","kjv_def":"wine."},
"H2563":{"lemma":"חֹמֶר","xlit":"chômer","pron":"kho'mer","derivation":"from H2560 (חָמַר);","strongs_def":"properly, a bubbling up, i.e. of water, a wave; of earth, mire or clay (cement); also a heap; hence, a chomer or dry measure","kjv_def":"clay, heap, homer, mire, motion."},
"H2564":{"lemma":"חֵמָר","xlit":"chêmâr","pron":"khay-mawr'","derivation":"from H2560 (חָמַר);","strongs_def":"bitumen (as rising to the surface)","kjv_def":"slime(-pit)."},
"H2565":{"lemma":"חֲמֹרָה","xlit":"chămôrâh","pron":"kham-o-raw'","derivation":"from H2560 (חָמַר) (compare H2563 (חֹמֶר));","strongs_def":"a heap","kjv_def":"heap."},
"H2566":{"lemma":"חַמְרָן","xlit":"Chamrân","pron":"kham-rawn'","derivation":"from H2560 (חָמַר); red;","strongs_def":"Chamran, an Idumaean","kjv_def":"Amran."},
"H2567":{"lemma":"חָמַשׁ","xlit":"châmash","pron":"khaw-mash'","derivation":"a denominative from H2568 (חָמֵשׁ);","strongs_def":"to tax a fifth","kjv_def":"take up the fifth participle"},
"H2568":{"lemma":"חָמֵשׁ","xlit":"châmêsh","pron":"khaw-maysh'","derivation":"masculine חֲמִשָּׁה; a primitive numeral;","strongs_def":"five","kjv_def":"fif(-teen), fifth, five ([idiom] apiece)."},
"H2569":{"lemma":"חֹמֶשׁ","xlit":"chômesh","pron":"kho'-mesh","derivation":"from H2567 (חָמַשׁ);","strongs_def":"a fifth tax","kjv_def":"fifth participle"},
"H2570":{"lemma":"חֹמֶשׁ","xlit":"chômesh","pron":"kho'-mesh","derivation":"from an unused root probably meaning, to be stout;","strongs_def":"the abdomen (as obese)","kjv_def":"fifth (rib)."},
"H2571":{"lemma":"חָמֻשׁ","xlit":"châmush","pron":"khaw-moosh'","derivation":"passive participle of the same as H2570 (חֹמֶשׁ);","strongs_def":"staunch, i.e. able-bodied soldiers","kjv_def":"armed (men), harnessed."},
"H2572":{"lemma":"חֲמִשִּׁים","xlit":"chămishshîym","pron":"kham-ish-sheem'","derivation":"multiple of H2568 (חָמֵשׁ);","strongs_def":"fifty","kjv_def":"fifty."},
"H2573":{"lemma":"חֵמֶת","xlit":"chêmeth","pron":"klay'-meth","derivation":"from the same as H2346 (חוֹמָה);","strongs_def":"a skin bottle (as tied up)","kjv_def":"bottle."},
"H2574":{"lemma":"חֲמָת","xlit":"Chămâth","pron":"kham-awth'","derivation":"from the same as H2346 (חוֹמָה); walled;","strongs_def":"Chamath, a place in Syria","kjv_def":"Hamath, Hemath."},
"H2575":{"lemma":"חַמַּת","xlit":"Chammath","pron":"klam-math'","derivation":"a variation for the first part of H2576 (חַמֹּת דֹּאר); hot springs;","strongs_def":"Chammath, a place in Palestine","kjv_def":"Hammath."},
"H2576":{"lemma":"חַמֹּת דֹּאר","xlit":"Chammôth Dôʼr","pron":"kham-moth' dore","derivation":"from the plural of H2535 (חַמָּה) and H1756 (דּוֹר); hot springs of Dor;","strongs_def":"Chammath-Dor, a place in Palestine","kjv_def":"Hamath-Dor."},
"H2577":{"lemma":"חֲמָתִי","xlit":"Chămâthîy","pron":"kham-aw-thee'","derivation":"patrial from H2574 (חֲמָת);","strongs_def":"a Chamathite or native of Chamath","kjv_def":"Hamathite."},
"H2578":{"lemma":"חֲמַת צוֹבָה","xlit":"Chămath Tsôwbâh","pron":"kham-ath' tso-baw'","derivation":"from H2574 (חֲמָת) and H6678 (צוֹבָא); Chamath of Tsobah; probably the same as H2574 (חֲמָת)","strongs_def":"Chamath-Tsobah","kjv_def":"Hamath-Zobah."},
"H2579":{"lemma":"חֲמַת רַבָּה","xlit":"Chămath Rabbâh","pron":"kham-ath' rab-baw'","derivation":"from H2574 (חֲמָת) and H7237 (רַבָּה); Chamath of Rabbah; probably the same as H2574 (חֲמָת).","strongs_def":"Chamath-Rabbah"},
"H2580":{"lemma":"חֵן","xlit":"chên","pron":"khane","derivation":"from H2603 (חָנַן);","strongs_def":"graciousness, i.e. subjective (kindness, favor) or objective (beauty)","kjv_def":"favour, grace(-ious), pleasant, precious, (well-) favoured."},
"H2581":{"lemma":"חֵן","xlit":"Chên","pron":"khane","derivation":"the same as H2580 (חֵן); grace;","strongs_def":"Chen, a figurative name for an Israelite","kjv_def":"Hen."},
"H2582":{"lemma":"חֵנָדָד","xlit":"Chênâdâd","pron":"khay-naw-dawd'","derivation":"probably from H2580 (חֵן) and H1908 (הֲדַד); favor of Hadad;","strongs_def":"Chenadad, an Israelite","kjv_def":"Henadad."},
"H2583":{"lemma":"חָנָה","xlit":"chânâh","pron":"khaw-naw'","derivation":"a primitive root (compare H2603 (חָנַן));","strongs_def":"properly, to incline; by implication, to decline (of the slanting rays of evening); specifically, to pitch atent; gen. to encamp (for abode or siege)","kjv_def":"abide (in tents), camp, dwell, encamp, grow to an end, lie, pitch (tent), rest in tent."},
"H2584":{"lemma":"חַנָּה","xlit":"Channâh","pron":"khan-naw'","derivation":"from H2603 (חָנַן); favored;","strongs_def":"Channah, an Israelitess","kjv_def":"Hannah."},
"H2585":{"lemma":"חֲנוֹךְ","xlit":"Chănôwk","pron":"khan-oke'","derivation":"from H2596 (חָנַךְ); initiated;","strongs_def":"Chanok, an antediluvian patriach","kjv_def":"Enoch."},
"H2586":{"lemma":"חָנוּן","xlit":"Chânûwn","pron":"khaw-noon'","derivation":"from H2603 (חָנַן); favored;","strongs_def":"Chanun, the name of an Ammonite and of two Israelites","kjv_def":"Hanun."},
"H2587":{"lemma":"חַנּוּן","xlit":"channûwn","pron":"khan-noon'","derivation":"from H2603 (חָנַן);","strongs_def":"gracious","kjv_def":"gracious."},
"H2588":{"lemma":"חָנוּת","xlit":"chânûwth","pron":"khaw-nooth'","derivation":"from H2583 (חָנָה);","strongs_def":"properly, a vault or cell (with an arch); by implication, a prison","kjv_def":"cabin."},
"H2589":{"lemma":"חַנּוֹת","xlit":"channôwth","pron":"klan-noth'","derivation":"from H2603 (חָנַן) (in the sense of prayer);","strongs_def":"supplication","kjv_def":"be gracious, intreated."},
"H2590":{"lemma":"חָנַט","xlit":"chânaṭ","pron":"khaw-nat'","derivation":"a primitive root;","strongs_def":"to spice; by implication, to embalm; also to ripen","kjv_def":"embalm, put forth."},
"H2591":{"lemma":"חִנְטָא","xlit":"chinṭâʼ","pron":"khint-taw'","derivation":"(Aramaic) corresponding to H2406 (חִטָּה);","strongs_def":"wheat","kjv_def":"wheat."},
"H2592":{"lemma":"חַנִּיאֵל","xlit":"Channîyʼêl","pron":"khan-nee-ale'","derivation":"from H2603 (חָנַן) and H410 (אֵל); favor of God;","strongs_def":"Channiel, the name of two Israelites","kjv_def":"Hanniel."},
"H2593":{"lemma":"חָנִיךְ","xlit":"chânîyk","pron":"kaw-neek'","derivation":"from H2596 (חָנַךְ);","strongs_def":"initiated; i.e. practiced","kjv_def":"trained."},
"H2594":{"lemma":"חֲנִינָה","xlit":"chănîynâh","pron":"khan-ee-naw'","derivation":"from H2603 (חָנַן);","strongs_def":"graciousness","kjv_def":"favour."},
"H2595":{"lemma":"חֲנִית","xlit":"chănîyth","pron":"khan-eeth'","derivation":"from H2583 (חָנָה);","strongs_def":"a lance (for thrusting, like pitching a tent)","kjv_def":"javelin, spear."},
"H2596":{"lemma":"חָנַךְ","xlit":"chânak","pron":"khaw-nak'","derivation":"a primitive root; (compare H2614 (חָנַק))","strongs_def":"properly, to narrow; figuratively, to initiate or discipline","kjv_def":"dedicate, train up."},
"H2597":{"lemma":"חֲנֻכָּא","xlit":"chănukkâʼ","pron":"chan-ook-kaw'","derivation":"(Aramaic) corresponding to H2598 (חֲנֻכָּה);","strongs_def":"consecration","kjv_def":"dedication."},
"H2598":{"lemma":"חֲנֻכָּה","xlit":"chănukkâh","pron":"khan-ook-kaw'","derivation":"from H2596 (חָנַךְ);","strongs_def":"initiation, i.e. consecration","kjv_def":"dedicating(-tion)."},
"H2599":{"lemma":"חֲנֹכִי","xlit":"Chănôkîy","pron":"khan-o-kee'","derivation":"patronymically from H2585 (חֲנוֹךְ);","strongs_def":"a Chanokite (collectively) or descendants of Chanok","kjv_def":"Hanochites."},
"H2600":{"lemma":"חִנָּם","xlit":"chinnâm","pron":"khin-nawm'","derivation":"from H2580 (חֵן);","strongs_def":"gratis, i.e. devoid of cost, reason or advantage","kjv_def":"without a cause (cost, wages), causeless, to cost nothing, free(-ly), innocent, for nothing (nought, in vain."},
"H2601":{"lemma":"חֲנַמְאֵל","xlit":"Chănamʼêl","pron":"khan-am-ale'","derivation":"probably by orthographical variation for H2606 (חֲנַנְאֵל);","strongs_def":"Chanamel, an Israelite","kjv_def":"Hanameel."},
"H2602":{"lemma":"חֲנָמָל","xlit":"chănâmâl","pron":"khan-aw-mawl'","derivation":"of uncertain derivation;","strongs_def":"perhaps the aphis or plantlouse","kjv_def":"frost."},
"H2603":{"lemma":"חָנַן","xlit":"chânan","pron":"khaw-nan'","derivation":"a primitive root (compare H2583 (חָנָה));","strongs_def":"properly, to bend or stoop in kindness to an inferior; to favor, bestow; causatively to implore (i.e. move to favor by petition)","kjv_def":"beseech, [idiom] fair, (be, find, shew) favour(-able), be (deal, give, grant (gracious(-ly), intreat, (be) merciful, have (shew) mercy (on, upon), have pity upon, pray, make supplication, [idiom] very."},
"H2604":{"lemma":"חֲנַן","xlit":"chănan","pron":"khan-an'","derivation":"(Aramaic) corresponding to H2603 (חָנַן);","strongs_def":"to favor or (causatively) to entreat","kjv_def":"shew mercy, make supplication."},
"H2605":{"lemma":"חָנָן","xlit":"Chânân","pron":"khaw-nawn'","derivation":"from H2603 (חָנַן); favor;","strongs_def":"Chanan, the name of seven Israelites","kjv_def":"Canan."},
"H2606":{"lemma":"חֲנַנְאֵל","xlit":"Chănanʼêl","pron":"khan-an-ale'","derivation":"from H2603 (חָנַן) and H410 (אֵל); God has favored;","strongs_def":"Chananel, probably an Israelite, from whom a tower of Jerusalem was named","kjv_def":"Hananeel."},
"H2607":{"lemma":"חֲנָנִי","xlit":"Chănânîy","pron":"khan-aw-nee'","derivation":"from H2603 (חָנַן); gracious;","strongs_def":"Chanani, the name of six Israelites","kjv_def":"Hanani."},
"H2608":{"lemma":"חֲנַנְיָה","xlit":"Chănanyâh","pron":"khan-an-yaw'","derivation":"or חֲנַנְיָהוּ; from H2603 (חָנַן) and H3050 (יָהּ); Jah has favored;","strongs_def":"Chananjah, the name of thirteen Israelites","kjv_def":"Hananiah."},
"H2609":{"lemma":"חָנֵס","xlit":"Chânêç","pron":"khaw-nace'","derivation":"of Egyptian derivation;","strongs_def":"Chanes, a place in Egypt","kjv_def":"Hanes."},
"H2610":{"lemma":"חָנֵף","xlit":"chânêph","pron":"khaw-nafe'","derivation":"a primitive root;","strongs_def":"to soil, especially in a moral sense","kjv_def":"corrupt, defile, [idiom] greatly, pollute, profane."},
"H2611":{"lemma":"חָנֵף","xlit":"chânêph","pron":"khaw-nafe'","derivation":"from H2610 (חָנֵף);","strongs_def":"soiled (i.e. with sin), impious","kjv_def":"hypocrite(-ical)."},
"H2612":{"lemma":"חֹנֶף","xlit":"chôneph","pron":"kho'-nef","derivation":"from H2610 (חָנֵף);","strongs_def":"moral filth, i.e. wickedness","kjv_def":"hypocrisy."},
"H2613":{"lemma":"חֲנֻפָה","xlit":"chănuphâh","pron":"khan-oo-faw'","derivation":"feminine from H2610 (חָנֵף);","strongs_def":"impiety","kjv_def":"profaneness."},
"H2614":{"lemma":"חָנַק","xlit":"chânaq","pron":"khaw-nak'","derivation":"a primitive root (compare H2596 (חָנַךְ));","strongs_def":"to be narrow; by implication, to throttle, or (reflex.) to choke oneself to death (by a rope)","kjv_def":"hang self, strangle."},
"H2615":{"lemma":"חַנָּתֹן","xlit":"Channâthôn","pron":"khan-naw-thone'","derivation":"probably from H2603 (חָנַן); favored;","strongs_def":"Channathon, a place in Palestine","kjv_def":"Hannathon."},
"H2616":{"lemma":"חָסַד","xlit":"châçad","pron":"khaw-sad'","derivation":"a primitive root; (compare H2603 (חָנַן)); (compare H1288 (בָרַךְ))","strongs_def":"properly, perhaps to bow (the neck only  in courtesy to an equal), i.e. to be kind; also (by euphemistically, but rarely) to reprove","kjv_def":"shew self merciful, put to shame."},
"H2617":{"lemma":"חֵסֵד","xlit":"chêçêd","pron":"kheh'-sed","derivation":"from H2616 (חָסַד);","strongs_def":"kindness; by implication (towards God) piety; rarely (by opposition) reproof, or (subject.) beauty","kjv_def":"favour, good deed(-liness, -ness), kindly, (loving-) kindness, merciful (kindness), mercy, pity, reproach, wicked thing."},
"H2618":{"lemma":"חֶסֶד","xlit":"Cheçed","pron":"kheh'-sed","derivation":"the same as H2617 (חֵסֵד); favor;","strongs_def":"Chesed, an Israelite","kjv_def":"Hesed."},
"H2619":{"lemma":"חֲסַדְיָה","xlit":"Chăçadyâh","pron":"khas-ad-yaw'","derivation":"from H2617 (חֵסֵד) and H3050 (יָהּ); Jah has favored;","strongs_def":"Chasadjah, an Israelite","kjv_def":"Hasadiah."},
"H2620":{"lemma":"חָסָה","xlit":"châçâh","pron":"khaw-saw'","derivation":"a primitive root; (compare H982 (בָּטַח))","strongs_def":"to flee for protection; figuratively, to confide in","kjv_def":"have hope, make refuge, (put) trust."},
"H2621":{"lemma":"חֹסָה","xlit":"Chôçâh","pron":"kho-saw'","derivation":"from H2620 (חָסָה); hopeful;","strongs_def":"Chosah, an Israelite; also a place in Palestine","kjv_def":"Hosah."},
"H2622":{"lemma":"חָסוּת","xlit":"châçûwth","pron":"khaw-sooth'","derivation":"from H2620 (חָסָה);","strongs_def":"confidence","kjv_def":"trust."},
"H2623":{"lemma":"חָסִיד","xlit":"châçîyd","pron":"khaw-seed'","derivation":"from H2616 (חָסַד);","strongs_def":"properly, kind, i.e. (religiously) pious (a saint)","kjv_def":"godly (man), good, holy (one), merciful, saint, (un-) godly."},
"H2624":{"lemma":"חֲסִידָה","xlit":"chăçîydâh","pron":"khas-ee-daw'","derivation":"feminine of H2623 (חָסִיד);","strongs_def":"the kind (maternal) bird, i.e. a stork","kjv_def":"[idiom] feather, stork."},
"H2625":{"lemma":"חָסִיל","xlit":"châçîyl","pron":"khaw-seel'","derivation":"from H2628 (חָסַל);","strongs_def":"the ravager, i.e. a locust","kjv_def":"caterpillar."},
"H2626":{"lemma":"חֲסִין","xlit":"chăçîyn","pron":"khas-een'","derivation":"from H2630 (חָסַן);","strongs_def":"properly, firm, i.e. (by implication) mighty","kjv_def":"strong."},
"H2627":{"lemma":"חַסִּיר","xlit":"chaççîyr","pron":"khas-seer'","derivation":"(Aramaic) from a root corresponding to H2637 (חָסֵר);","strongs_def":"deficient","kjv_def":"wanting."},
"H2628":{"lemma":"חָסַל","xlit":"châçal","pron":"khaw-sal'","derivation":"a primitive root;","strongs_def":"to eat off","kjv_def":"consume."},
"H2629":{"lemma":"חָסַם","xlit":"châçam","pron":"khaw-sam'","derivation":"a primitive root;","strongs_def":"to muzzle; by analogy, to stop the nose","kjv_def":"muzzle, stop."},
"H2630":{"lemma":"חָסַן","xlit":"châçan","pron":"khaw-san'","derivation":"a primitive root;","strongs_def":"properly, to (be) compact; by implication, to hoard","kjv_def":"lay up."},
"H2631":{"lemma":"חֲסַן","xlit":"chăçan","pron":"khas-an'","derivation":"(Aramaic) corresponding to H2630 (חָסַן);","strongs_def":"to hold in occupancy","kjv_def":"possess."},
"H2632":{"lemma":"חֵסֶן","xlit":"chêçen","pron":"khay'-sen","derivation":"(Aramaic) from H2631 (חֲסַן);","strongs_def":"strength","kjv_def":"power."},
"H2633":{"lemma":"חֹסֶן","xlit":"chôçen","pron":"kho'-sen","derivation":"from H2630 (חָסַן);","strongs_def":"wealth","kjv_def":"riches, strength, treasure."},
"H2634":{"lemma":"חָסֹן","xlit":"châçôn","pron":"khaw-sone'","derivation":"from H2630 (חָסַן);","strongs_def":"powerful","kjv_def":"strong."},
"H2635":{"lemma":"חֲסַף","xlit":"chăçaph","pron":"khas-af'","derivation":"(Aramaic) from a root corresponding to that of H2636 (חַסְפַּס);","strongs_def":"a clod","kjv_def":"clay."},
"H2636":{"lemma":"חַסְפַּס","xlit":"chaçpaç","pron":"khas-pas'","derivation":"reduplicated from an unused root meaning apparently to peel;","strongs_def":"a shred or scale","kjv_def":"round thing."},
"H2637":{"lemma":"חָסֵר","xlit":"châçêr","pron":"khaw-sare'","derivation":"a primitive root;","strongs_def":"to lack; by implication, to fail, want, lessen","kjv_def":"be abated, bereave, decrease, (cause to) fail, (have) lack, make lower, want."},
"H2638":{"lemma":"חָסֵר","xlit":"châçêr","pron":"khaw-sare'","derivation":"from H2637 (חָסֵר);","strongs_def":"lacking; hence, without","kjv_def":"destitute, fail, lack, have need, void, want."},
"H2639":{"lemma":"חֶסֶר","xlit":"cheçer","pron":"kheh'-ler","derivation":"from H2637 (חָסֵר);","strongs_def":"lack; hence, destitution","kjv_def":"poverty, want."},
"H2640":{"lemma":"חֹסֶר","xlit":"chôçer","pron":"kho'-ser","derivation":"from H2637 (חָסֵר);","strongs_def":"poverty","kjv_def":"in want of."},
"H2641":{"lemma":"חַסְרָה","xlit":"Chaçrâh","pron":"khas-raw'","derivation":"from H2637 (חָסֵר); want","strongs_def":"Chasrah, an Israelite","kjv_def":"Hasrah."},
"H2642":{"lemma":"חֶסְרוֹן","xlit":"cheçrôwn","pron":"khes-rone'","derivation":"from H2637 (חָסֵר);","strongs_def":"deficiency","kjv_def":"wanting."},
"H2643":{"lemma":"חַף","xlit":"chaph","pron":"khaf","derivation":"from H2653 (חׇפַף) (in the moral sense of covered from soil);","strongs_def":"pure","kjv_def":"innocent."},
"H2644":{"lemma":"חָפָא","xlit":"châphâʼ","pron":"khaw-faw'","derivation":"an orthographical variation of H2645 (חָפָה);","strongs_def":"properly, to cover, i.e. (in a sinister sense) to act covertly","kjv_def":"do secretly."},
"H2645":{"lemma":"חָפָה","xlit":"châphâh","pron":"khaw-faw'","derivation":"a primitive root (compare H2644 (חָפָא), H2653 (חׇפַף));","strongs_def":"to cover; by implication, to veil, to encase, protect","kjv_def":"ceil, cover, overlay."},
"H2646":{"lemma":"חֻפָּה","xlit":"chuppâh","pron":"khoop-paw'","derivation":"from H2645 (חָפָה);","strongs_def":"a canopy","kjv_def":"chamber, closet, defence."},
"H2647":{"lemma":"חֻפָּה","xlit":"Chuppâh","pron":"khoop-paw'","derivation":"the same as H2646 (חֻפָּה);","strongs_def":"Chuppah, an Israelite","kjv_def":"Huppah."},
"H2648":{"lemma":"חָפַז","xlit":"châphaz","pron":"khaw-faz'","derivation":"a primitive root;","strongs_def":"properly, to start up suddenly, i.e. (by implication) to hasten away, to fear","kjv_def":"(make) haste (away), tremble."},
"H2649":{"lemma":"חִפָּזוֹן","xlit":"chippâzôwn","pron":"khip-paw-zone'","derivation":"from H2468 (חֻלְדָּה);","strongs_def":"hasty flight","kjv_def":"haste."},
"H2650":{"lemma":"חֻפִּים","xlit":"Chuppîym","pron":"khoop-peem'","derivation":"plural of H2646 (חֻפָּה) (compare H2349 (חוּפָם));","strongs_def":"Chuppim, an Israelite","kjv_def":"Huppim."},
"H2651":{"lemma":"חֹפֶן","xlit":"chôphen","pron":"kho'-fen","derivation":"from an unused root of uncertain signification;","strongs_def":"a fist (only in the dual)","kjv_def":"fists, (both) hands, hand(-ful)."},
"H2652":{"lemma":"חׇפְנִי","xlit":"Chophnîy","pron":"khof-nee'","derivation":"from H2651 (חֹפֶן); perhaps pugilist;","strongs_def":"Chophni, an Israelite","kjv_def":"Hophni."},
"H2653":{"lemma":"חׇפַף","xlit":"chophaph","pron":"khaw-faf'","derivation":"a primitive root (compare H2645 (חָפָה), H3182 (יָחֵף));","strongs_def":"to cover (in protection)","kjv_def":"cover."},
"H2654":{"lemma":"חָפֵץ","xlit":"châphêts","pron":"khaw-fates'","derivation":"a primitive root;","strongs_def":"properly, to incline to; by implication (literally but rarely) to bend; figuratively, to be pleased with, desire","kjv_def":"[idiom] any at all, (have, take) delight, desire, favour, like, move, be (well) pleased, have pleasure, will, would."},
"H2655":{"lemma":"חָפֵץ","xlit":"châphêts","pron":"khaw-fates'","derivation":"from H2654 (חָפֵץ);","strongs_def":"pleased with","kjv_def":"delight in, desire, favour, please, have pleasure, whosoever would, willing, wish."},
"H2656":{"lemma":"חֵפֶץ","xlit":"chêphets","pron":"khay'-fets","derivation":"from H2654 (חָפֵץ);","strongs_def":"pleasure; hence (abstractly) desire; concretely, a valuable thing; hence (by extension) a matter (as something in mind)","kjv_def":"acceptable, delight(-some), desire, things desired, matter, pleasant(-ure), purpose, willingly."},
"H2657":{"lemma":"חֶפְצִי בָּהּ","xlit":"Chephtsîy bâhh","pron":"khef-tsee'baw","derivation":"from H2656 (חֵפֶץ) with suffixes; my delight (is) in her;","strongs_def":"Cheptsi-bah, a fanciful name for Palestine","kjv_def":"Hephzi-bah."},
"H2658":{"lemma":"חָפַר","xlit":"châphar","pron":"khaw-far'","derivation":"a primitive root;","strongs_def":"properly, to pry into; by implication, to delve, to explore","kjv_def":"dig, paw, search out, seek."},
"H2659":{"lemma":"חָפֵר","xlit":"châphêr","pron":"khaw-fare'","derivation":"a primitive root (perhaps rather the same as H2658 (חָפַר) through the idea of detection);","strongs_def":"to blush; figuratively, to be ashamed, disappointed; causatively, to shame, reproach","kjv_def":"be ashamed, be confounded, be brought to confusion (unto shame), come (be put to) shame, bring reproach."},
"H2660":{"lemma":"חֵפֶר","xlit":"Chêpher","pron":"khay'-fer","derivation":"from H2658 (חָפַר) or H2659 (חָפֵר); a pit of shame;","strongs_def":"Chepher, a place in Palestine; also the name of three Israelites","kjv_def":"Hepher."},
"H2661":{"lemma":"חֲפֹר","xlit":"chăphôr","pron":"khaf-ore'","derivation":"from H2658 (חָפַר); a hole; only in connection with H6512 (פֵּרָה), which ought rather to be joined as one word, thus חַפַרְפֵּרָה; by reduplication from H2658 (חָפַר);","strongs_def":"a burrower, i.e. probably a rat","kjv_def":"[phrase] mole."},
"H2662":{"lemma":"חֶפְרִי","xlit":"Chephrîy","pron":"khef-ree'","derivation":"patronymically from H2660 (חֵפֶר);","strongs_def":"a Chephrite (collectively) or descendants of Chepher","kjv_def":"Hepherites."},
"H2663":{"lemma":"חֲפָרַיִם","xlit":"Chăphârayim","pron":"khaf-aw-rah'-yim","derivation":"dual of H2660 (חֵפֶר); double pit;","strongs_def":"Chapharajim, a place in Palestine","kjv_def":"Haphraim."},
"H2664":{"lemma":"חָפַשׂ","xlit":"châphas","pron":"khaw-fas'","derivation":"a primitive root;","strongs_def":"to seek; causatively, to conceal oneself (i.e. let be sought), or mask","kjv_def":"change, (make) diligent (search), disquise self, hide, search (for, out)."},
"H2665":{"lemma":"חֵפֶשׂ","xlit":"chêphes","pron":"khay'-fes","derivation":"from H2664 (חָפַשׂ);","strongs_def":"something covert, i.e. a trick","kjv_def":"search."},
"H2666":{"lemma":"חָפַשׁ","xlit":"châphash","pron":"khaw-fash'","derivation":"a primitive root;","strongs_def":"to spread loose; figuratively, to manumit","kjv_def":"be free."},
"H2667":{"lemma":"חֹפֶשׁ","xlit":"Chôphesh","pron":"kho'-fesh","derivation":"from H2666 (חָפַשׁ);","strongs_def":"something spread loosely, i.e. a carpet","kjv_def":"precious."},
"H2668":{"lemma":"חֻפְשָׁה","xlit":"chuphshâh","pron":"khoof-shaw'","derivation":"from H2666 (חָפַשׁ);","strongs_def":"liberty (from slavery)","kjv_def":"freedom."},
"H2669":{"lemma":"חׇפְשׁוּת","xlit":"chophshûwth","pron":"khof-shooth'","derivation":"and חׇפְשִׁית; from H2666 (חָפַשׁ); (with H1004 (בַּיִת), a hospital)","strongs_def":"prostration by sickness","kjv_def":"several."},
"H2670":{"lemma":"חׇפְשִׁי","xlit":"chophshîy","pron":"khof-shee'","derivation":"from H2666 (חָפַשׁ);","strongs_def":"exempt (from bondage, tax or care)","kjv_def":"free, liberty."},
"H2671":{"lemma":"חֵץ","xlit":"chêts","pron":"khayts","derivation":"from H2686 (חָצַץ); also by interchange for H6086 (עֵץ)","strongs_def":"properly, a piercer, i.e. an arrow; by implication, a wound; figuratively, (of God) thunderbolt; the shaft of aspear","kjv_def":"[phrase] archer, arrow, dart, shaft, staff, wound."},
"H2672":{"lemma":"חָצַב","xlit":"châtsab","pron":"khaw-tsab'","derivation":"or חָצֵב; a primitive root;","strongs_def":"to cut or carve (wood, stone or other material); by implication, to hew, split, square, quarry, engrave","kjv_def":"cut, dig, divide, grave, hew (out, -er), made, mason."},
"H2673":{"lemma":"חָצָה","xlit":"châtsâh","pron":"khaw-tsaw'","derivation":"a primitive root (compare H2086 (זֵד)));","strongs_def":"to cut or split in two; to halve","kjv_def":"divide, [idiom] live out half, reach to the midst, participle"},
"H2674":{"lemma":"חָצוֹר","xlit":"Châtsôwr","pron":"khaw-tsore'","derivation":"a collective form of H2691 (חָצֵר); village;","strongs_def":"Chatsor, the name (thus simply) of two places in Palestine and of one in Arabia","kjv_def":"Hazor."},
"H2675":{"lemma":"חָצוֹר חֲדַתָּה","xlit":"Châtsôwr Chădattâh","pron":"khaw-tsore' khad-attaw'","derivation":"from H2674 (חָצוֹר) and a Aramaizing form of the feminine of H2319 (חָדָשׁ) (compare H2323 (חֲדָת));","strongs_def":"new Chatsor, a place in Palestine","kjv_def":"Hazor, Hadattah (as if two places)."},
"H2676":{"lemma":"חָצוֹת","xlit":"châtsôwth","pron":"khaw-tsoth'","derivation":"from H2673 (חָצָה)lemma חָעָה ayin, corrected to חָצָה;","strongs_def":"the middle (of the night)","kjv_def":"mid(-night)."},
"H2677":{"lemma":"חֵצִי","xlit":"chêtsîy","pron":"khay-tsee'","derivation":"from H2673 (חָצָה)lemma חָעָה ayin, corrected to חָצָה;","strongs_def":"the half or middle","kjv_def":"half, middle, mid(-night), midst, part, two parts."},
"H2678":{"lemma":"חִצִּי","xlit":"chitstsîy","pron":"khits-tsee'","derivation":"or חֵצִי; prolongation from H2671 (חֵץ);","strongs_def":"an arrow","kjv_def":"arrow."},
"H2679":{"lemma":"חֲצִי הַמְּנֻחוֹת","xlit":"Chătsîy ham-Mᵉnuchôwth","pron":"chat-tsee' hammen-oo-khoth'","derivation":"from H2677 (חֵצִי) and the plural of H4496 (מְנוּחָה), with the article interposed; midst of the resting-places;","strongs_def":"Chatsi-ham-Menuchoth, an Israelite","kjv_def":"half of the Manahethites."},
"H2680":{"lemma":"חֲצִי הַמְּנַחְתִּי","xlit":"Chătsîy ham-Mᵉnachtîy","pron":"khat-see' ham-menakh-tee'","derivation":"patronymically from H2679 (חֲצִי הַמְּנֻחוֹת);","strongs_def":"a Chatsi-ham-Menachtite or descendant of Chatsi-ham-Menuchoth","kjv_def":"half of the Manahethites."},
"H2681":{"lemma":"חָצִיר","xlit":"châtsîyr","pron":"khaw-tseer'","derivation":"a collateral form of H2691 (חָצֵר);","strongs_def":"a court or abode","kjv_def":"court."},
"H2682":{"lemma":"חָצִיר","xlit":"châtsîyr","pron":"khaw-tseer'","derivation":"perhaps originally the same as H2681 (חָצִיר), from the greenness of a courtyard;","strongs_def":"grass; also a leek (collectively)","kjv_def":"grass, hay, herb, leek."},
"H2683":{"lemma":"חֵצֶן","xlit":"chêtsen","pron":"khay'-tsen","derivation":"from an unused root meaning to hold firmly;","strongs_def":"the bosom (as comprised between the arms)","kjv_def":"bosom."},
"H2684":{"lemma":"חֹצֶן","xlit":"chôtsen","pron":"kho'tsen","derivation":"a collateral form of H2683 (חֵצֶן), and meaning the same","strongs_def":"{the bosom (as comprised between the arms)}","kjv_def":"arm, lap."},
"H2685":{"lemma":"חֲצַף","xlit":"chătsaph","pron":"khats-af'","derivation":"(Aramaic) a primitive root;","strongs_def":"properly, to shear or cut close; figuratively, to be severe","kjv_def":"hasty, be urgent."},
"H2686":{"lemma":"חָצַץ","xlit":"châtsats","pron":"khaw-tsats'","derivation":"a primitive root (compare H2673 (חָצָה)lemma חָעָה ayin, corrected to חָצָה); also as denominative from H2671 (חֵץ)","strongs_def":"properly, to chop into, pierce or sever; hence, to curtail, to distribute (into ranks);  to shoot an arrow","kjv_def":"archer, [idiom] bands, cut off in the midst."},
"H2687":{"lemma":"חָצָץ","xlit":"châtsâts","pron":"khaw-tsawts'","derivation":"from H2687 (חָצָץ);  also like H2671 (חֵץ)","strongs_def":"properly, something cutting; hence, gravel (as grit); an arrow","kjv_def":"arrow, gravel (stone)."},
"H2688":{"lemma":"חַצְצוֹן תָּמָר","xlit":"Chatsᵉtsôwn Tâmâr","pron":"khats-ets-one' taw-mawr'","derivation":"or חַצֲצֹן תָּמָר; from H2686 (חָצַץ) and H8558 (תָּמָר); division (i.e. perhaps row) of (the) palm-tree;","strongs_def":"Chatsetson-tamar, a place in Palestine","kjv_def":"Hazezon-tamar."},
"H2689":{"lemma":"חֲצֹצְרָה","xlit":"chătsôtsᵉrâh","pron":"khats-o-tser-aw'","derivation":"by reduplication from H2690 (חָצַר);","strongs_def":"a trumpet (from its sundered or quavering note)","kjv_def":"trumpet(-er)."},
"H2690":{"lemma":"חָצַר","xlit":"châtsar","pron":"khaw-tsar'","derivation":"a primitive root; properly, to surround with a stockade, and thus separate from the open country; but used only in the reduplicated form חֲצֹצֵר; or (2 Chronicles 5:12) חֲצֹרֵר; as dem. from H2689 (חֲצֹצְרָה);","strongs_def":"to trumpet, i.e. blow on that instrument","kjv_def":"blow, sound, trumpeter."},
"H2691":{"lemma":"חָצֵר","xlit":"châtsêr","pron":"khaw-tsare'","derivation":"(masculine and feminine); from H2690 (חָצַר) in its original sense;","strongs_def":"a yard (as inclosed by a fence); also a hamlet (as similarly surrounded with walls)","kjv_def":"court, tower, village."},
"H2692":{"lemma":"חֲצַר אַדָּר","xlit":"Chătsar ʼAddâr","pron":"khats-ar' addawr'","derivation":"from H2691 (חָצֵר) and H146 (אַדָּר); (the) village of Addar;","strongs_def":"Chatsar-Addar, a place in Palestine","kjv_def":"Hazar-addar."},
"H2693":{"lemma":"חֲצַר גַּדָּה","xlit":"Chătsar Gaddâh","pron":"khats-ar' gad-daw'","derivation":"from H2691 (חָצֵר) and a feminine of H1408 (גַּד); (the) village of (female) Fortune;","strongs_def":"Chatsar-Gaddah, a place in Palestine","kjv_def":"Hazar-gaddah."},
"H2694":{"lemma":"חֲצַר הַתִּיכוֹן","xlit":"Chătsar hat-Tîykôwn","pron":"khats-ar' hat-tee-kone'","derivation":"from H2691 (חָצֵר) and H8484 (תִּיכוֹן) with the article interposed; village of the middle;","strongs_def":"Chatsar-hat-Tikon, a place in Palestine","kjv_def":"Hazar-hatticon."},
"H2695":{"lemma":"חֶצְרוֹ","xlit":"Chetsrôw","pron":"khets-ro'","derivation":"by an orthographical variation for H2696 (חֶצְרוֹן); enclosure;","strongs_def":"Chetsro, an Israelite","kjv_def":"Hezro, Hezrai."},
"H2696":{"lemma":"חֶצְרוֹן","xlit":"Chetsrôwn","pron":"khets-rone'","derivation":"from H2691 (חָצֵר); court-yard;","strongs_def":"Chetsron, the name of a place in Palestine; also of two Israelites","kjv_def":"Hezron."},
"H2697":{"lemma":"חֶצְרוֹנִי","xlit":"Chetsrôwnîy","pron":"khets-ro-nee'","derivation":"patronymically from H2696 (חֶצְרוֹן);","strongs_def":"a Chetsronite or (collectively) descendants of Chetsron","kjv_def":"Hezronites."},
"H2698":{"lemma":"חֲצֵרוֹת","xlit":"Chătsêrôwth","pron":"khats-ay-roth'","derivation":"feminine plural of H2691 (חָצֵר); yards;","strongs_def":"Chatseroth, a place in Palestine","kjv_def":"Hazeroth."},
"H2699":{"lemma":"חֲצֵרִים","xlit":"Chătsêrîym","pron":"khats-ay-reem'","derivation":"plural masculine of H2691 (חָצֵר); yards;","strongs_def":"Chatserim, a place in Palestine","kjv_def":"Hazerim."},
"H2700":{"lemma":"חֲצַרְמָוֶת","xlit":"Chătsarmâveth","pron":"khats-ar-maw'-veth","derivation":"from H2691 (חָצֵר) and H4194 (מָוֶת); village of death;","strongs_def":"Chatsarmaveth, a place in Arabia","kjv_def":"Hazarmaveth."},
"H2701":{"lemma":"חֲצַר סוּסָה","xlit":"Chătsar Çûwçâh","pron":"khats-ar'soo-saw'","derivation":"from H2691 (חָצֵר) and H5484 (סוּסָה); village of cavalry;","strongs_def":"Chatsar-Susah, a place in Palestine","kjv_def":"Hazar-susah."},
"H2702":{"lemma":"חֲצַר סוּסִים","xlit":"Chătsar Çûwçîym","pron":"khats-ar' soo-seem'","derivation":"from H2691 (חָצֵר) and the plural of H5483 (סוּס); village of horses;","strongs_def":"Chatsar-Susim, a place in Palestine","kjv_def":"Hazar-susim."},
"H2703":{"lemma":"חֲצַר עֵינוֹן","xlit":"Chătsar ʻÊynôwn","pron":"khats-ar' ay-none'","derivation":"from H2691 (חָצֵר) and a derivative of H5869 (עַיִן); village of springs;","strongs_def":"Chatsar-Enon, a place in Palestine","kjv_def":"Hazar-enon."},
"H2704":{"lemma":"חֲצַר עֵינָן","xlit":"Chătsar ʻÊynân","pron":"khats-ar' ay-nawn'","derivation":"from H2691 (חָצֵר) and the same as H5881 (עֵינָן); village of springs;","strongs_def":"Chatsar-Enan, a place in Palestine","kjv_def":"Hazar-enan."},
"H2705":{"lemma":"חֲצַר שׁוּעָל","xlit":"Chătsar Shûwʻâl","pron":"khats-ar' shoo-awl'","derivation":"from H2691 (חָצֵר) and H7776 (שׁוּעָל); village of (the) fox;","strongs_def":"Chatsar-Shual, a place in Palestine","kjv_def":"Hazar-shual."},
"H2706":{"lemma":"חֹק","xlit":"chôq","pron":"khoke","derivation":"from H2710 (חָקַק);","strongs_def":"an enactment; hence, an appointment (of time, space, quantity, labor or usage)","kjv_def":"appointed, bound, commandment, convenient, custom, decree(-d), due, law, measure, [idiom] necessary, ordinance(-nary), portion, set time, statute, task."},
"H2707":{"lemma":"חָקָה","xlit":"châqâh","pron":"khaw-kaw'","derivation":"a primitive root;","strongs_def":"to carve; by implication, to delineate; also to entrench","kjv_def":"carved work, portrayed, set a print."},
"H2708":{"lemma":"חֻקָּה","xlit":"chuqqâh","pron":"khook-kaw'","derivation":"feminine of H2706 (חֹק), and meaning substantially the same","strongs_def":"{an enactment; hence, an appointment (of time, space, quantity, labor or usage)}","kjv_def":"appointed, custom, manner, ordinance, site, statute."},
"H2709":{"lemma":"חֲקוּפָא","xlit":"Chăqûwphâʼ","pron":"khah-oo-faw'","derivation":"from an unused root probably meaning to bend; crooked;","strongs_def":"Chakupha, one of the Nethinim","kjv_def":"Hakupha."},
"H2710":{"lemma":"חָקַק","xlit":"châqaq","pron":"khaw-kak'","derivation":"a primitive root;","strongs_def":"properly, to hack, i.e. engrave (Judges 5:14, to be a scribe simply); by implication, to enact (laws being cut in stone or metal tablets in primitive times) or (gen.) prescribe","kjv_def":"appoint, decree, governor, grave, lawgiver, note, pourtray, print, set."},
"H2711":{"lemma":"חֵקֶק","xlit":"chêqeq","pron":"khay'-kek","derivation":"from H2710 (חָקַק);","strongs_def":"an enactment, a resolution","kjv_def":"decree, thought."},
"H2712":{"lemma":"חֻקֹּק","xlit":"Chuqqôq","pron":"Khook-koke'","derivation":"or (fully) חוּקֹק; from H2710 (חָקַק); appointed;","strongs_def":"Chukkok or Chukok, a place in Palestine","kjv_def":"Hukkok, Hukok."},
"H2713":{"lemma":"חָקַר","xlit":"châqar","pron":"khaw-kar'","derivation":"a primitive root;","strongs_def":"properly, to penetrate; hence, to examine intimately","kjv_def":"find out, (make) search (out), seek (out), sound, try."},
"H2714":{"lemma":"חֵקֶר","xlit":"chêqer","pron":"khay'-ker","derivation":"from H2713 (חָקַר);","strongs_def":"examination, enumeration, deliberation","kjv_def":"finding out, number, (un-) search(-able, -ed, out, -ing)."},
"H2715":{"lemma":"חֹר","xlit":"chôr","pron":"khore","derivation":"or (fully) חוֹר; from H2787 (חָרַר); (compare H2751 (חֹרִי))","strongs_def":"properly, white or pure (from the cleansing or shining power of fire; hence (figuratively) noble (in rank)","kjv_def":"noble."},
"H2716":{"lemma":"חֶרֶא","xlit":"chereʼ","pron":"kheh'-reh","derivation":"from an unused (and vulgar) root probably meaning to evacuate the bowels;","strongs_def":"excrement","kjv_def":"dung. Also חֲרִי."},
"H2717":{"lemma":"חָרַב","xlit":"chârab","pron":"khaw-rab'","derivation":"or חֲרֵב; a primitive root;","strongs_def":"to parch (through drought) i.e. (by analogy,) to desolate, destroy, kill","kjv_def":"decay, (be) desolate, destroy(-er), (be) dry (up), slay, [idiom] surely, (lay, lie, make) waste."},
"H2718":{"lemma":"חֲרַב","xlit":"chărab","pron":"khar-ab'","derivation":"(Aramaic) a root corresponding to H2717 (חָרַב);","strongs_def":"to demolish","kjv_def":"destroy."},
"H2719":{"lemma":"חֶרֶב","xlit":"chereb","pron":"kheh'-reb","derivation":"from H2717 (חָרַב);","strongs_def":"drought; also a cutting instrument (from its destructive effect), as a knife, sword, or other sharp implement","kjv_def":"axe, dagger, knife, mattock, sword, tool."},
"H2720":{"lemma":"חָרֵב","xlit":"chârêb","pron":"khaw-rabe'","derivation":"from H2717 (חָרַב);","strongs_def":"parched or ruined","kjv_def":"desolate, dry, waste."},
"H2721":{"lemma":"חֹרֶב","xlit":"chôreb","pron":"kho'-reb","derivation":"a collaterally form of H2719 (חֶרֶב);","strongs_def":"drought or desolation","kjv_def":"desolation, drought, dry, heat, [idiom] utterly, waste."},
"H2722":{"lemma":"חֹרֵב","xlit":"Chôrêb","pron":"kho-rabe'","derivation":"from H2717 (חָרַב); desolate;","strongs_def":"Choreb, a (generic) name for the Sinaitic mountains","kjv_def":"Horeb."},
"H2723":{"lemma":"חׇרְבָּה","xlit":"chorbâh","pron":"khor-baw'","derivation":"feminine of H2721 (חֹרֶב);","strongs_def":"properly, drought, i.e. (by implication) a desolation","kjv_def":"decayed place, desolate (place, -tion), destruction, (laid) waste (place)."},
"H2724":{"lemma":"חָרָבָה","xlit":"chârâbâh","pron":"khaw-raw-baw'","derivation":"feminine of H2720 (חָרֵב);","strongs_def":"a desert","kjv_def":"dry (ground, land)."},
"H2725":{"lemma":"חֲרָבוֹן","xlit":"chărâbôwn","pron":"khar-aw-bone'","derivation":"from H2717 (חָרַב);","strongs_def":"parching heat","kjv_def":"drought."},
"H2726":{"lemma":"חַרְבוֹנָא","xlit":"Charbôwnâʼ","pron":"khar-bo-naw'","derivation":"or חַרְבוֹנָה; of Persian origin;","strongs_def":"Charbona or Charbonah, a eunuch of Xerxes","kjv_def":"Harbona, Harbonah."},
"H2727":{"lemma":"חָרַג","xlit":"chârag","pron":"khaw-rag'","derivation":"a primitive root;","strongs_def":"properly, to leap suddenly, i.e. (by implication) to be dismayed","kjv_def":"be afraid."},
"H2728":{"lemma":"חָרְגֹּל","xlit":"chârᵉgôl","pron":"khar-gole'","derivation":"from H2727 (חָרַג);","strongs_def":"the leaping insect, i.e. a locust","kjv_def":"beetle."},
"H2729":{"lemma":"חָרַד","xlit":"chârad","pron":"khaw-rad'","derivation":"a primitive root;","strongs_def":"to shudder with terror; hence, to fear; also to hasten (with anxiety)","kjv_def":"be (make) afraid, be careful, discomfit, fray (away), quake, tremble."},
"H2730":{"lemma":"חָרֵד","xlit":"chârêd","pron":"khaw-rade'","derivation":"from H2729 (חָרַד);","strongs_def":"fearful; also reverential","kjv_def":"afraid, trembling."},
"H2731":{"lemma":"חֲרָדָה","xlit":"chărâdâh","pron":"khar-aw-daw'","derivation":"feminine of H2730 (חָרֵד);","strongs_def":"fear, anxiety","kjv_def":"care, [idiom] exceedingly, fear, quaking, trembling."},
"H2732":{"lemma":"חֲרָדָה","xlit":"Chărâdâh","pron":"khar-aw-daw'","derivation":"the same as H2731 (חֲרָדָה);","strongs_def":"Charadah, a place in the Desert","kjv_def":"Haradah."},
"H2733":{"lemma":"חֲרֹדִי","xlit":"Chărôdîy","pron":"khar-o-dee'","derivation":"patrial from a derivative of H2729 (חָרַד) (compare H5878 (עֵין חֲרֹד));","strongs_def":"a Charodite, or inhabitant of Charod","kjv_def":"Harodite."},
"H2734":{"lemma":"חָרָה","xlit":"chârâh","pron":"khaw-raw'","derivation":"a primitive root (compare H2787 (חָרַר));","strongs_def":"to glow or grow warm; figuratively (usually) to blaze up, of anger, zeal, jealousy","kjv_def":"be angry, burn, be displeased, [idiom] earnestly, fret self, grieve, be (wax) hot, be incensed, kindle, [idiom] very, be wroth. See H8474 (תַּחָרָה)."},
"H2735":{"lemma":"חֹר הַגִּדְגָּד","xlit":"Chôr hag-Gidgâd","pron":"khore hag-ghid-gawd'","derivation":"from H2356 (חוֹר) and a collateral (masculine) form of H1412 (גֻּדְגֹּדָה), with the article interposed; hole of the cleft;","strongs_def":"Chor-hag-Gidgad, a place in the Desert","kjv_def":"Hor-hagidgad."},
"H2736":{"lemma":"חַרְהֲיָה","xlit":"Charhăyâh","pron":"khar-hah-yaw'","derivation":"; from H2734 (חָרָה) and H3050 (יָהּ); fearing Jah;","strongs_def":"Charhajah, an Israelite","kjv_def":"Harhaiah."},
"H2737":{"lemma":"חָרוּז","xlit":"chârûwz","pron":"khaw-rooz'","derivation":"from an unused root meaning to perforate;","strongs_def":"properly, pierced, i.e. a bead of pearl, gems or jewels (as strung)","kjv_def":"chain."},
"H2738":{"lemma":"חָרוּל","xlit":"chârûwl","pron":"khaw-rool'","derivation":"or (shortened) חָרֻל; apparently, a passive participle of an unused root probably meaning to be prickly;","strongs_def":"properly, pointed, i.e. a bramble or other thorny weed","kjv_def":"nettle."},
"H2739":{"lemma":"חֲרוּמַף","xlit":"chărûwmaph","pron":"khar-oo-maf'","derivation":"from passive participle of H2763 (חָרַם) and H639 (אַף); snub-nosed;","strongs_def":"Charumaph, an Israelite","kjv_def":"Harumaph."},
"H2740":{"lemma":"חָרוֹן","xlit":"chârôwn","pron":"khaw-rone'","derivation":"or (shortened) חָרֹן; from H2734 (חָרָה);","strongs_def":"a burning of anger","kjv_def":"sore displeasure, fierce(-ness), fury, (fierce) wrath(-ful)."},
"H2741":{"lemma":"חֲרוּפִי","xlit":"Chărûwphîy","pron":"khar-oo-fee'","derivation":"a patrial from (probably) a collateral form of H2756 (חָרִיף);","strongs_def":"a Charuphite or inhabitant of Charuph (or Chariph)","kjv_def":"Haruphite."},
"H2742":{"lemma":"חֲרוּץ","xlit":"chărûwts","pron":"khaw-roots'","derivation":"or חָרֻץ; passive participle of H2782 (חָרַץ);","strongs_def":"properly, incised or (active) incisive; hence (as noun masculine or feminine) a trench (as dug), gold (as mined), a threshing-sledge (having sharp teeth); (figuratively) determination; also eager","kjv_def":"decision, diligent, (fine) gold, pointed things, sharp, threshing instrument, wall."},
"H2743":{"lemma":"חָרוּץ","xlit":"Chârûwts","pron":"khaw-roots'","derivation":"the same as H2742 (חֲרוּץ); earnest;","strongs_def":"Charuts, an Israelite","kjv_def":"Haruz."},
"H2744":{"lemma":"חַרְחוּר","xlit":"Charchûwr","pron":"khar-khoor'","derivation":"a fuller form of H2746 (חַרְחֻר); inflammation;","strongs_def":"Charchur, one of the Nethinim","kjv_def":"Harhur."},
"H2745":{"lemma":"חַרְחַס","xlit":"Charchaç","pron":"khar-khas'","derivation":"from the same as H2775 (חֶרֶס); perhaps shining;","strongs_def":"Charchas, an Israelite","kjv_def":"Harhas."},
"H2746":{"lemma":"חַרְחֻר","xlit":"charchur","pron":"khar-khoor'","derivation":"from H2787 (חָרַר);","strongs_def":"fever (as hot)","kjv_def":"extreme burning."},
"H2747":{"lemma":"חֶרֶט","xlit":"chereṭ","pron":"kheh'-ret","derivation":"from a primitive root meaning to engrave;","strongs_def":"a chisel or graver; also a style forwriting","kjv_def":"graving tool, pen."},
"H2748":{"lemma":"חַרְטֹם","xlit":"charṭôm","pron":"khar-tome'","derivation":"from the same as H2747 (חֶרֶט);","strongs_def":"a horoscopist (as drawing magical lines or circles)","kjv_def":"magician."},
"H2749":{"lemma":"חַרְטֹם","xlit":"charṭôm","pron":"khar-tome'","derivation":"(Aramaic) the same as H2748 (חַרְטֹם)","strongs_def":"{a horoscopist (as drawing magical lines or circles)}","kjv_def":"magician."},
"H2750":{"lemma":"חֳרִי","xlit":"chŏrîy","pron":"khor-ee'","derivation":"from H2734 (חָרָה);","strongs_def":"a burning (i.e. intense) anger","kjv_def":"fierce, [idiom] great, heat."},
"H2751":{"lemma":"חֹרִי","xlit":"chôrîy","pron":"kho-ree'","derivation":"from the same as H2353 (חוּר);","strongs_def":"white bread","kjv_def":"white."},
"H2752":{"lemma":"חֹרִי","xlit":"Chôrîy","pron":"kho-ree'","derivation":"from H2356 (חוֹר); cave-dweller or troglodyte;","strongs_def":"a Chorite or aboriginal Idumaean","kjv_def":"Horims, Horites."},
"H2753":{"lemma":"חֹרִי","xlit":"Chôrîy","pron":"kho-ree'","derivation":"or חוֹרִי; the same as H2752 (חֹרִי);","strongs_def":"Chori, the name of two men","kjv_def":"Hori."},
"H2754":{"lemma":"חָרִיט","xlit":"chârîyṭ","pron":"khaw-reet'","derivation":"or חָרִט; from the same as H2747 (חֶרֶט);","strongs_def":"properly, cut out (or hollow), i.e. (by implication) a pocket","kjv_def":"bag, crisping pin."},
"H2755":{"lemma":"חֲרֵי־יוֹנִים","xlit":"chărêy-yôwnîym","pron":"khar-ay'-yo-neem'","derivation":"from the plural of H2716 (חֶרֶא) and the plural of H3123 (יוֹנָה); or perhaps rather the plural of a single word חֲרָאיוֹן, probably a kind of vegetable","strongs_def":"excrements of doves {or a vegetable}","kjv_def":"doves' dung."},
"H2756":{"lemma":"חָרִיף","xlit":"Chârîyph","pron":"khaw-reef'","derivation":"from H2778 (חָרַף); autumnal;","strongs_def":"Chariph, the name of two Israelites","kjv_def":"Hariph."},
"H2757":{"lemma":"חָרִיץ","xlit":"chârîyts","pron":"khaw-reets'","derivation":"or חֲרִץ; from H2782 (חָרַץ); (compare H2742 (חֲרוּץ))","strongs_def":"properly, incisure or (passively) incised; hence, a threshing-sledge (with sharp teeth); also a slice (as cut)","kjv_def":"[phrase] cheese, harrow."},
"H2758":{"lemma":"חָרִישׁ","xlit":"chârîysh","pron":"khaw-reesh'","derivation":"from H2790 (חָרַשׁ);","strongs_def":"ploughing or its season","kjv_def":"earing (time), ground."},
"H2759":{"lemma":"חֲרִישִׁי","xlit":"chărîyshîy","pron":"khar-ee-shee'","derivation":"from H2790 (חָרַשׁ) in the sense of silence;","strongs_def":"quiet, i.e. sultry (as feminine noun, the sirocco or hot east wind)","kjv_def":"vehement."},
"H2760":{"lemma":"חָרַךְ","xlit":"chârak","pron":"khaw-rak'","derivation":"a primitive root;","strongs_def":"to braid (i.e. to entangle or snare) or catch (game) in anet","kjv_def":"roast."},
"H2761":{"lemma":"חֲרַךְ","xlit":"chărak","pron":"khar-ak'","derivation":"(Aramaic) a root probably allied to the equivalent of H2787 (חָרַר);","strongs_def":"to scorch","kjv_def":"singe."},
"H2762":{"lemma":"חֶרֶךְ","xlit":"cherek","pron":"kheh'-rek","derivation":"from H2760 (חָרַךְ);","strongs_def":"properly, a net, i.e. (by analogy) lattice","kjv_def":"lattice."},
"H2763":{"lemma":"חָרַם","xlit":"châram","pron":"khaw-ram'","derivation":"a primitive root;","strongs_def":"to seclude; specifically (by a ban) to devote to religious uses (especially destruction); physical and reflexive, to be blunt as to the nose","kjv_def":"make accursed, consecrate, (utterly) destroy, devote, forfeit, have a flat nose, utterly (slay, make away)."},
"H2764":{"lemma":"חֵרֶם","xlit":"chêrem","pron":"khay'-rem","derivation":"or (Zechariah 14:11) חֶרֶם; from H2763 (חָרַם);","strongs_def":"physical (as shutting in) a net (either literally or figuratively); usually a doomed object; abstractly extermination","kjv_def":"(ac-) curse(-d, -d thing), dedicated thing, things which should have been utterly destroyed, (appointed to) utter destruction, devoted (thing), net."},
"H2765":{"lemma":"חֳרֵם","xlit":"Chŏrêm","pron":"khor-ame'","derivation":"from H2763 (חָרַם); devoted;","strongs_def":"Chorem, a place in Palestine","kjv_def":"Horem."},
"H2766":{"lemma":"חָרִם","xlit":"Chârim","pron":"khaw-reem'","derivation":"from H2763 (חָרַם); snub-nosed;","strongs_def":"Charim, an Israelite","kjv_def":"Harim."},
"H2767":{"lemma":"חׇרְמָה","xlit":"Chormâh","pron":"khor-maw'","derivation":"from H2763 (חָרַם); devoted;","strongs_def":"Chormah, a place in Palestine","kjv_def":"Hormah."},
"H2768":{"lemma":"חֶרְמוֹן","xlit":"Chermôwn","pron":"kher-mone'","derivation":"from H2763 (חָרַם); abrupt;","strongs_def":"Chermon, a mount of Palestine","kjv_def":"Hermon."},
"H2769":{"lemma":"חֶרְמוֹנִים","xlit":"Chermôwnîym","pron":"kher-mo-neem'","derivation":"plural of H2768 (חֶרְמוֹן);","strongs_def":"Hermons, i.e. its peaks","kjv_def":"the Hermonites."},
"H2770":{"lemma":"חֶרְמֵשׁ","xlit":"chermêsh","pron":"kher-mashe'","derivation":"from H2763 (חָרַם);","strongs_def":"a sickle (as cutting)","kjv_def":"sickle."},
"H2771":{"lemma":"חָרָן","xlit":"Chârân","pron":"kaw-rawn'","derivation":"from H2787 (חָרַר); parched;","strongs_def":"Charan, the name of a man and also of a place","kjv_def":"Haran."},
"H2772":{"lemma":"חֹרֹנִי","xlit":"Chôrônîy","pron":"kho-ro-nee'","derivation":"patrial from H2773 (חֹרֹנַיִם);","strongs_def":"a Choronite or inhabitant of Choronaim","kjv_def":"Horonite."},
"H2773":{"lemma":"חֹרֹנַיִם","xlit":"Chôrônayim","pron":"kho-ro-nah'-yim","derivation":"dual of a derivative from H2356 (חוֹר); double cave-town;","strongs_def":"Choronajim, a place in Moab","kjv_def":"Horonaim."},
"H2774":{"lemma":"חַרְנֶפֶר","xlit":"Charnepher","pron":"khar-neh'fer","derivation":"of uncertain derivation;","strongs_def":"Charnepher, an Israelite","kjv_def":"Harnepher."},
"H2775":{"lemma":"חֶרֶס","xlit":"chereç","pron":"kheh'-res","derivation":"or (with a directive enclitic) חַרְסָה; from an unused root meaning to scrape; also perhaps from the mediating idea of H2777 (חַרְסוּת)","strongs_def":"the itch; the sun","kjv_def":"itch, sun."},
"H2776":{"lemma":"חֶרֶס","xlit":"Chereç","pron":"kheh'-res","derivation":"the same as H2775 (חֶרֶס); shining;","strongs_def":"Cheres, a mountain in Palestine","kjv_def":"Heres."},
"H2777":{"lemma":"חַרְסוּת","xlit":"charçûwth","pron":"khar-sooth'","derivation":"from H2775 (חֶרֶס) (apparently in the sense of a red tile used for scraping);","strongs_def":"a potsherd, i.e. (by implication) a pottery; the name of a gate at Jerusalem","kjv_def":"east."},
"H2778":{"lemma":"חָרַף","xlit":"châraph","pron":"khaw-raf'","derivation":"a primitive root; also  denominative (from H2779 (חֹרֶף))","strongs_def":"to pull off, i.e. (by implication) to expose (as by stripping); specifically, to betroth (as if a surrender); figuratively, to carp at, i.e. defame; to spend the winter","kjv_def":"betroth, blaspheme, defy, jeopard, rail, reproach, upbraid."},
"H2779":{"lemma":"חֹרֶף","xlit":"chôreph","pron":"kho'-ref","derivation":"from H2778 (חָרַף);","strongs_def":"properly, the crop gathered, i.e. (by implication) the autumn (and winter) season; figuratively, ripeness of age","kjv_def":"cold, winter (-house), youth."},
"H2780":{"lemma":"חָרֵף","xlit":"Chârêph","pron":"khaw-rafe'","derivation":"from H2778 (חָרַף); reproachful;","strongs_def":"Chareph, an Israelite","kjv_def":"Hareph."},
"H2781":{"lemma":"חֶרְפָּה","xlit":"cherpâh","pron":"kher-paw'","derivation":"from H2778 (חָרַף);","strongs_def":"contumely, disgrace, the pudenda","kjv_def":"rebuke, reproach(-fully), shame."},
"H2782":{"lemma":"חָרַץ","xlit":"chârats","pron":"khaw-rats'","derivation":"a primitive root;","strongs_def":"properly, to point sharply, i.e. (literally) to wound; figuratively, to be alert, to decide","kjv_def":"bestir self, decide, decree, determine, maim, move."},
"H2783":{"lemma":"חֲרַץ","xlit":"chărats","pron":"khar-ats'","derivation":"(Aramaic) from a root corresponding to H2782 (חָרַץ) in the sense of vigor;","strongs_def":"the loin (as the seat of strength)","kjv_def":"loin."},
"H2784":{"lemma":"חַרְצֻבָּה","xlit":"chartsubbâh","pron":"khar-tsoob-baw'","derivation":"of uncertain derivation;","strongs_def":"a fetter; figuratively, a pain","kjv_def":"band."},
"H2785":{"lemma":"חַרְצַן","xlit":"chartsan","pron":"kchar-tsan'","derivation":"from H2782 (חָרַץ);","strongs_def":"a sour grape (as sharp in taste)","kjv_def":"kernel."},
"H2786":{"lemma":"חָרַק","xlit":"châraq","pron":"khaw-rak'","derivation":"a primitive root;","strongs_def":"to grate the teeth","kjv_def":"gnash."},
"H2787":{"lemma":"חָרַר","xlit":"chârar","pron":"khaw-rar'","derivation":"a primitive root;","strongs_def":"to glow, i.e. literally (to melt, burn, dry up) or figuratively (to show or incite passion)","kjv_def":"be angry, burn, dry, kindle."},
"H2788":{"lemma":"חָרֵר","xlit":"chârêr","pron":"khaw-rare'","derivation":"from H2787 (חָרַר);","strongs_def":"arid","kjv_def":"parched place."},
"H2789":{"lemma":"חֶרֶשׂ","xlit":"cheres","pron":"kheh'-res","derivation":"a collateral form mediating between H2775 (חֶרֶס) and H2791 (חֶרֶשׁ);","strongs_def":"a piece of pottery","kjv_def":"earth(-en), (pot-) sherd, [phrase] stone."},
"H2790":{"lemma":"חָרַשׁ","xlit":"chârash","pron":"khaw-rash'","derivation":"a primitive root;","strongs_def":"to scratch, i.e. (by implication) to engrave, plough; hence (from the use of tools) to fabricate (of any material); figuratively, to devise (in a bad sense); hence (from the idea of secrecy) to be silent, to let alone; hence (by implication) to be deaf (as an accompaniment of dumbness)","kjv_def":"[idiom] altogether, cease, conceal, be deaf, devise, ear, graven, imagine, leave off speaking, hold peace, plow(-er, man), be quiet, rest, practise secretly, keep silence, be silent, speak not a word, be still, hold tongue, worker."},
"H2791":{"lemma":"חֶרֶשׁ","xlit":"cheresh","pron":"kheh'-resh","derivation":"from H2790 (חָרַשׁ);","strongs_def":"magical craft; also silence","kjv_def":"cunning, secretly."},
"H2792":{"lemma":"חֶרֶשׁ","xlit":"Cheresh","pron":"kheh'-resh","derivation":"the same as H2791 (חֶרֶשׁ)","strongs_def":"Cheresh, a Levite","kjv_def":"Heresh."},
"H2793":{"lemma":"חֹרֶשׁ","xlit":"chôresh","pron":"kho'-resh","derivation":"from H2790 (חָרַשׁ);","strongs_def":"a forest (perhaps as furnishing the material for fabric)","kjv_def":"bough, forest, shroud, wood."},
"H2794":{"lemma":"חֹרֵשׁ","xlit":"chôrêsh","pron":"kho-rashe'","derivation":"active participle of H2790 (חָרַשׁ);","strongs_def":"a fabricator or mechanic","kjv_def":"artificer."},
"H2795":{"lemma":"חֵרֵשׁ","xlit":"chêrêsh","pron":"khay-rashe'","derivation":"from H2790 (חָרַשׁ);","strongs_def":"deaf (whether literally or spiritual)","kjv_def":"deaf."},
"H2796":{"lemma":"חָרָשׁ","xlit":"chârâsh","pron":"khaw-rawsh'","derivation":"from H2790 (חָרַשׁ);","strongs_def":"a fabricator or any material","kjv_def":"artificer, ([phrase]) carpenter, craftsman, engraver, maker, [phrase] mason, skilful, ([phrase]) smith, worker, workman, such as wrought."},
"H2797":{"lemma":"חַרְשָׁא","xlit":"Charshâʼ","pron":"khar-shaw'","derivation":"from H2792 (חֶרֶשׁ); magician;","strongs_def":"Charsha, one of the Nethinim","kjv_def":"Harsha."},
"H2798":{"lemma":"חֲרָשִׁים","xlit":"Chărâshîym","pron":"khar-aw-sheem'","derivation":"plural of H2796 (חָרָשׁ);","strongs_def":"mechanics, the name of a valley in Jerusalem","kjv_def":"Charashim, craftsmen."},
"H2799":{"lemma":"חֲרֹשֶׁת","xlit":"chărôsheth","pron":"khar-o'-sheth","derivation":"from H2790 (חָרַשׁ);","strongs_def":"mechanical work","kjv_def":"carving, cutting."},
"H2800":{"lemma":"חֲרֹשֶׁת","xlit":"Chărôsheth","pron":"khar-o'-sheth","derivation":"the same as H2799 (חֲרֹשֶׁת);","strongs_def":"Charosheth, a place in Palestine","kjv_def":"Harosheth."},
"H2801":{"lemma":"חָרַת","xlit":"chârath","pron":"khaw-rath'","derivation":"a primitive root;","strongs_def":"to engrave","kjv_def":"graven."},
"H2802":{"lemma":"חֶרֶת","xlit":"Chereth","pron":"kheh'-reth","derivation":"from H2801 (חָרַת) (but equivalent to H2793 (חֹרֶשׁ)); forest;","strongs_def":"Chereth, a thicket in Palestine","kjv_def":"Hereth."},
"H2803":{"lemma":"חָשַׁב","xlit":"châshab","pron":"khaw-shab'","derivation":"a primitive root;","strongs_def":"properly, to plait or interpenetrate, i.e. (literally) to weave or (generally) to fabricate; figuratively, to plot or contrive (usually in a malicious sense); hence (from the mental effort) to think, regard, value, compute","kjv_def":"(make) account (of), conceive, consider, count, cunning (man, work, workman), devise, esteem, find out, forecast, hold, imagine, impute, invent, be like, mean, purpose, reckon(-ing be made), regard, think."},
"H2804":{"lemma":"חֲשַׁב","xlit":"chăshab","pron":"khash-ab'","derivation":"(Aramaic) corresponding to H2803 (חָשַׁב);","strongs_def":"to regard","kjv_def":"repute."},
"H2805":{"lemma":"חֵשֶׁב","xlit":"chêsheb","pron":"khay'-sheb","derivation":"from H2803 (חָשַׁב);","strongs_def":"a belt or strap (as being interlaced)","kjv_def":"curious girdle."},
"H2806":{"lemma":"חַשְׁבַּדָּנָה","xlit":"Chashbaddânâh","pron":"khash-bad-daw'-naw","derivation":"from H2803 (חָשַׁב) and H1777 (דִּין); considerate judge;","strongs_def":"Chasbaddanah, an Israelite","kjv_def":"Hasbadana."},
"H2807":{"lemma":"חֲשֻׁבָה","xlit":"Chăshubâh","pron":"khash-oo-baw'","derivation":"from H2803 (חָשַׁב); estimation;","strongs_def":"Cashubah, an Israelite","kjv_def":"Hashubah."},
"H2808":{"lemma":"חֶשְׁבּוֹן","xlit":"cheshbôwn","pron":"khesh-bone'","derivation":"from H2803 (חָשַׁב);","strongs_def":"properly, contrivance; by implication, intelligence","kjv_def":"account, device, reason."},
"H2809":{"lemma":"חֶשְׁבּוֹן","xlit":"Cheshbôwn","pron":"khesh-bone'","derivation":"the same as H2808 (חֶשְׁבּוֹן);","strongs_def":"Cheshbon, a place East of the Jordan","kjv_def":"Heshbon."},
"H2810":{"lemma":"חִשָּׁבוֹן","xlit":"chishshâbôwn","pron":"khish-shaw-bone'","derivation":"from H2803 (חָשַׁב);","strongs_def":"a contrivance, i.e. actual (a warlike machine) or mental (a machination)","kjv_def":"engine, invention."},
"H2811":{"lemma":"חֲשַׁבְיָה","xlit":"Chăshabyâh","pron":"khash-ab-yaw'","derivation":"or חֲשַׁבְיָהוּ; from H2803 (חָשַׁב) and H3050 (יָהּ); Jah has regarded;","strongs_def":"Chashabjah, the name of nine Israelites","kjv_def":"Hashabiah."},
"H2812":{"lemma":"חֲשַׁבְנָה","xlit":"Chăshabnâh","pron":"khash-ab-naw'","derivation":"feminine of H2808 (חֶשְׁבּוֹן); inventiveness;","strongs_def":"Chashnah, an Israelite","kjv_def":"Hashabnah."},
"H2813":{"lemma":"חֲשַׁבְנְיָה","xlit":"Chăshabnᵉyâh","pron":"khash-ab-neh-yaw'","derivation":"from H2808 (חֶשְׁבּוֹן) and H3050 (יָהּ); thought of Jah;","strongs_def":"Chashabnejah, the name of two Israelites","kjv_def":"Hashabniah."},
"H2814":{"lemma":"חָשָׁה","xlit":"châshâh","pron":"khaw-shaw'","derivation":"a primitive root;","strongs_def":"to hush or keep quiet","kjv_def":"hold peace, keep silence, be silent, (be) still."},
"H2815":{"lemma":"חַשּׁוּב","xlit":"Chashshûwb","pron":"khash-shoob'","derivation":"from H2803 (חָשַׁב); intelligent;","strongs_def":"Chashshub, the name of two or three Israelites","kjv_def":"Hashub, Hasshub."},
"H2816":{"lemma":"חֲשׁוֹךְ","xlit":"chăshôwk","pron":"khash-oke'","derivation":"(Aramaic) from a root corresponding to H2821 (חָשַׁךְ);","strongs_def":"the dark","kjv_def":"darkness."},
"H2817":{"lemma":"חֲשׂוּפָא","xlit":"Chăsûwphâʼ","pron":"khas-oo-faw'","derivation":"or חֲשֻׂפָא; from H2834 (חָשַׂף); nakedness;","strongs_def":"Chasupha, one of the Nethinim","kjv_def":"Hashupha, Hasupha."},
"H2818":{"lemma":"חֲשַׁח","xlit":"chăshach","pron":"khash-akh'","derivation":"(Aramaic) a collateral root to one corresponding to H2363 (חוּשׁ) in the sense of readiness;","strongs_def":"to be necessary (from the idea of convenience) or (transitively) to need","kjv_def":"careful, have need of."},
"H2819":{"lemma":"חַשְׁחוּת","xlit":"chashchûwth","pron":"khash-khooth'","derivation":"from a root corresponding to H2818 (חֲשַׁח);","strongs_def":"necessity","kjv_def":"be needful."},
"H2820":{"lemma":"חָשַׂךְ","xlit":"châsak","pron":"khaw-sak'","derivation":"a primitive root; also interchangeable  with H2821 (חָשַׁךְ)","strongs_def":"to restrain or (reflex.) refrain; by implication, to refuse, spare, preserve; to observe","kjv_def":"assuage, [idiom] darken, forbear, hinder, hold back, keep (back), punish, refrain, reserve, spare, withhold."},
"H2821":{"lemma":"חָשַׁךְ","xlit":"châshak","pron":"khaw-shak'","derivation":"a primitive root;","strongs_def":"to be dark (as withholding light); transitively, to darken","kjv_def":"be black, be (make) dark, darken, cause darkness, be dim, hide."},
"H2822":{"lemma":"חֹשֶׁךְ","xlit":"chôshek","pron":"kho-shek'","derivation":"from H2821 (חָשַׁךְ);","strongs_def":"the dark; hence (literally) darkness; figuratively, misery, destruction, death, ignorance, sorrow, wickedness","kjv_def":"dark(-ness), night, obscurity."},
"H2823":{"lemma":"חָשֹׁךְ","xlit":"châshôk","pron":"khaw-shoke'","derivation":"from H2821 (חָשַׁךְ);","strongs_def":"dark (figuratively, i.e. obscure)","kjv_def":"meaning"},
"H2824":{"lemma":"חֶשְׁכָה","xlit":"cheshkâh","pron":"khesh-kaw'","derivation":"from H2821 (חָשַׁךְ);","strongs_def":"darkness","kjv_def":"dark."},
"H2825":{"lemma":"חֲשֵׁכָה","xlit":"chăshêkâh","pron":"khash-ay-kaw'","derivation":"or חֲשֵׁיכָהxlit chăshêkâh corrected to chăshêykâh; from H2821 (חָשַׁךְ);","strongs_def":"darkness; figuratively, misery","kjv_def":"darkness."},
"H2826":{"lemma":"חָשַׁל","xlit":"châshal","pron":"khaw-shal'","derivation":"a primitive root;","strongs_def":"to make (intrans. be) unsteady, i.e. weak","kjv_def":"feeble."},
"H2827":{"lemma":"חֲשַׁל","xlit":"chăshal","pron":"khash-al'","derivation":"(Aramaic) a root corresponding to H2826 (חָשַׁל);","strongs_def":"to weaken, i.e. crush","kjv_def":"subdue."},
"H2828":{"lemma":"חָשֻׁם","xlit":"Châshum","pron":"khaw-shoom'","derivation":"from the same as H2831 (חַשְׁמַן); enriched;","strongs_def":"Chashum, the name of two or three Israelites","kjv_def":"Hashum."},
"H2829":{"lemma":"חֶשְׁמוֹן","xlit":"Cheshmôwn","pron":"klesh-mone'","derivation":"the same as H2831 (חַשְׁמַן); opulent;","strongs_def":"Cheshmon, a place in Palestine","kjv_def":"Heshmon."},
"H2830":{"lemma":"חַשְׁמַל","xlit":"chashmal","pron":"khash-mal'","derivation":"of uncertain derivation;","strongs_def":"probably bronze or polished spectrum metal","kjv_def":"amber."},
"H2831":{"lemma":"חַשְׁמַן","xlit":"chashman","pron":"khash-man'","derivation":"from an unused root (probably meaning firm or capacious in resources); apparently","strongs_def":"wealthy","kjv_def":"princes."},
"H2832":{"lemma":"חַשְׁמֹנָה","xlit":"Chashmônâh","pron":"khash-mo-naw'","derivation":"feminine of H2831 (חַשְׁמַן); fertile;","strongs_def":"Chasmonah, a place in the Desert","kjv_def":"Hashmonah."},
"H2833":{"lemma":"חֹשֶׁן","xlit":"chôshen","pron":"kho'-shen","derivation":"from an unused root probably meaning to contain or sparkle;","strongs_def":"perhaps a pocket (as holding the Urim and Thummim), or rich (as containing gems), used only of the gorget of the highpriest","kjv_def":"breastplate."},
"H2834":{"lemma":"חָשַׂף","xlit":"châsaph","pron":"khaw-saf'","derivation":"a primitive root;","strongs_def":"to strip off, i.e. generally to make naked (for exertion or in disgrace), to drain away or bail up (a liquid)","kjv_def":"make bare, clean, discover, draw out, take, uncover."},
"H2835":{"lemma":"חָשִׂף","xlit":"châsiph","pron":"khaw-seef'","derivation":"from H2834 (חָשַׂף);","strongs_def":"properly, drawn off, i.e. separated; hence, a small company (as divided from the rest)","kjv_def":"little flock."},
"H2836":{"lemma":"חָשַׁק","xlit":"châshaq","pron":"khaw-shak'","derivation":"a primitive root; also interchangeable for H2820 (חָשַׂךְ)","strongs_def":"to cling, i.e. join, (figuratively) to love, delight in; elliptically; to deliver","kjv_def":"have a delight, (have a) desire, fillet, long, set (in) love."},
"H2837":{"lemma":"חֵשֶׁק","xlit":"chêsheq","pron":"khay'-shek","derivation":"from H2836 (חָשַׁק);","strongs_def":"delight","kjv_def":"desire, pleasure."},
"H2838":{"lemma":"חָשֻׁק","xlit":"châshuq","pron":"khaw-shook'","derivation":"or חָשׁוּק; past participle of H2836 (חָשַׁק);","strongs_def":"attached, i.e. a fence-rail or rod connecting the posts or pillars","kjv_def":"fillet."},
"H2839":{"lemma":"חִשֻּׁק","xlit":"chishshuq","pron":"khish-shook'","derivation":"from H2836 (חָשַׁק);","strongs_def":"conjoined, i.e. a wheel-spoke or rod connecting the hub with the rim","kjv_def":"felloe."},
"H2840":{"lemma":"חִשֻּׁר","xlit":"chishshur","pron":"khish-shoor'","derivation":"from an unused root meaning to bind together;","strongs_def":"combined, i.e. the nave or hub of awheel (as holding the spokes together)","kjv_def":"spoke."},
"H2841":{"lemma":"חַשְׁרָה","xlit":"chashrâh","pron":"khash-raw'","derivation":"from the same as H2840 (חִשֻּׁר);","strongs_def":"properly, a combination or gathering, i.e. of watery clouds","kjv_def":"dark."},
"H2842":{"lemma":"חָשַׁשׁ","xlit":"châshash","pron":"khaw-shash'","derivation":"by variation for H7179 (קַשׁ);","strongs_def":"dry grass","kjv_def":"chaff."},
"H2843":{"lemma":"חֻשָׁתִי","xlit":"Chushâthîy","pron":"khoo-shaw-thee'","derivation":"patronymically from H2364 (חוּשָׁה);","strongs_def":"a Chushathite or descendant of Chushah","kjv_def":"Hushathite."},
"H2844":{"lemma":"חַת","xlit":"chath","pron":"khath","derivation":"from H2865 (חָתַת);","strongs_def":"concretely, crushed; also afraid; abstractly, terror","kjv_def":"broken, dismayed, dread, fear."},
"H2845":{"lemma":"חֵת","xlit":"Chêth","pron":"khayth","derivation":"from H2865 (חָתַת); terror;","strongs_def":"Cheth, an aboriginal Canaanite","kjv_def":"Heth."},
"H2846":{"lemma":"חָתָה","xlit":"châthâh","pron":"khaw-thaw'","derivation":"a primitive root;","strongs_def":"to lay hold of; especially to pick up fire","kjv_def":"heap, take (away)."},
"H2847":{"lemma":"חִתָּה","xlit":"chittâh","pron":"khit-taw'","derivation":"from H2865 (חָתַת);","strongs_def":"fear","kjv_def":"terror."},
"H2848":{"lemma":"חִתּוּל","xlit":"chittûwl","pron":"khit-tool'","derivation":"from H2853 (חָתַל);","strongs_def":"swathed, i.e. a bandage","kjv_def":"roller."},
"H2849":{"lemma":"חַתְחַת","xlit":"chathchath","pron":"khath-khath'","derivation":"from H2844 (חַת);","strongs_def":"terror","kjv_def":"fear."},
"H2850":{"lemma":"חִתִּי","xlit":"Chittîy","pron":"khit-tee'","derivation":"patronymically from H2845 (חֵת);","strongs_def":"a Chittite, or descendant of Cheth","kjv_def":"Hittite, Hittities."},
"H2851":{"lemma":"חִתִּית","xlit":"chittîyth","pron":"khit-teeth'","derivation":"from H2865 (חָתַת);","strongs_def":"fear","kjv_def":"terror."},
"H2852":{"lemma":"חָתַךְ","xlit":"châthak","pron":"khaw-thak'","derivation":"a primitive root;","strongs_def":"properly, to cut off, i.e. (figuratively) to decree","kjv_def":"determine."},
"H2853":{"lemma":"חָתַל","xlit":"châthal","pron":"khaw-thal'","derivation":"a primitive root;","strongs_def":"to swathe","kjv_def":"[idiom] at all, swaddle."},
"H2854":{"lemma":"חֲתֻלָּה","xlit":"chăthullâh","pron":"khath-ool-law'","derivation":"from H2853 (חָתַל);","strongs_def":"a swathing cloth (figuratively)","kjv_def":"swaddling band."},
"H2855":{"lemma":"חֶתְלֹן","xlit":"Chethlôn","pron":"kheth-lone'","derivation":"from H2853 (חָתַל); enswathed;","strongs_def":"Chethlon, a place in Palestine","kjv_def":"Hethlon."},
"H2856":{"lemma":"חָתַם","xlit":"châtham","pron":"khaw-tham'","derivation":"a primitive root;","strongs_def":"to close up; especially to seal","kjv_def":"make an end, mark, seal (up), stop."},
"H2857":{"lemma":"חֲתַם","xlit":"chătham","pron":"khath-am'","derivation":"(Aramaic) a root corresponding to H2856 (חָתַם);","strongs_def":"to seal","kjv_def":"seal."},
"H2858":{"lemma":"חֹתֶמֶת","xlit":"chôthemeth","pron":"kho-the-meth","derivation":"feminine active participle of H2856 (חָתַם);","strongs_def":"a seal","kjv_def":"signet."},
"H2859":{"lemma":"חָתַן","xlit":"châthan","pron":"khaw-than'","derivation":"a primitive root;","strongs_def":"to give (a daughter) away in marriage; hence (generally) to contract affinity by marriage","kjv_def":"join in affinity, father in law, make marriages, mother in law, son in law."},
"H2860":{"lemma":"חָתָן","xlit":"châthân","pron":"khaw-thawn'","derivation":"from H2859 (חָתַן);","strongs_def":"a relative by marriage (especially through the bride); figuratively, a circumcised child (as a species of religious espousal)","kjv_def":"bridegroom, husband, son in law."},
"H2861":{"lemma":"חֲתֻנָּה","xlit":"chăthunnâh","pron":"khath-oon-naw'","derivation":"from H2859 (חָתַן);","strongs_def":"a wedding","kjv_def":"espousal."},
"H2862":{"lemma":"חָתַף","xlit":"châthaph","pron":"khaw-thaf'","derivation":"a primitive root;","strongs_def":"to clutch","kjv_def":"take away."},
"H2863":{"lemma":"חֶתֶף","xlit":"chetheph","pron":"kheh'-thef","derivation":"from H2862 (חָתַף);","strongs_def":"properly, rapine; figuratively, robbery","kjv_def":"prey."},
"H2864":{"lemma":"חָתַר","xlit":"châthar","pron":"khaw-thar'","derivation":"a primitive root;","strongs_def":"to force a passage, as by burglary; figuratively, with oars","kjv_def":"dig (through), row."},
"H2865":{"lemma":"חָתַת","xlit":"châthath","pron":"khaw-thath'","derivation":"a primitive root;","strongs_def":"properly, to prostrate; hence, to break down, either (literally) by violence, or (figuratively) by confusion and fear","kjv_def":"abolish, affright, be (make) afraid, amaze, beat down, discourage, (cause to) dismay, go down, scare, terrify."},
"H2866":{"lemma":"חֲתַת","xlit":"chăthath","pron":"khath-ath'","derivation":"from H2865 (חָתַת);","strongs_def":"dismay","kjv_def":"casting down."},
"H2867":{"lemma":"חֲתַת","xlit":"Chăthath","pron":"khath-ath'","derivation":"the same as H2866 (חֲתַת);","strongs_def":"Chathath, an Israelite","kjv_def":"cathath. f"},
"H2868":{"lemma":"טְאֵב","xlit":"ṭᵉʼêb","pron":"teh-abe'","derivation":"a primitive root;","strongs_def":"to rejoice","kjv_def":"be glad"},
"H2869":{"lemma":"טָב","xlit":"ṭâb","pron":"teh-abe'","derivation":"the same as H2896 (טוֹב);","strongs_def":"good","kjv_def":"fine, good."},
"H2870":{"lemma":"טָבְאֵל","xlit":"ṭâbᵉʼêl","pron":"taw-beh-ale'","derivation":"from H2895 (טוֹב) and H410 (אֵל); pleasing (to) God;","strongs_def":"Tabeel, the name of a Syrian and of a Persian","kjv_def":"Tabeal, Tabeel."},
"H2871":{"lemma":"טָבוּל","xlit":"ṭâbûwl","pron":"taw-bool'","derivation":"passive participle of H2881 (טָבַל);","strongs_def":"properly, dyed, i.e. a turban (probably as of colored stuff)","kjv_def":"dyed attire."},
"H2872":{"lemma":"טַבּוּר","xlit":"ṭabbûwr","pron":"tab-boor'","derivation":"from an unused root meaning to pile up;","strongs_def":"properly, accumulated; i.e. (by implication) a summit","kjv_def":"middle, midst."},
"H2873":{"lemma":"טָבַח","xlit":"ṭâbach","pron":"taw-bakh'","derivation":"a primitive root;","strongs_def":"to slaughter (animals or men)","kjv_def":"kill, (make) slaughter, slay."},
"H2874":{"lemma":"טֶבַח","xlit":"ṭebach","pron":"teh'-bakh","derivation":"from H2873 (טָבַח);","strongs_def":"properly, something slaughtered; hence, a beast (or meat, as butchered); abstractly butchery (or concretely, a place of slaughter)","kjv_def":"[idiom] beast, slaughter, [idiom] slay, [idiom] sore."},
"H2875":{"lemma":"טֶבַח","xlit":"Ṭebach","pron":"teh'-bakh","derivation":"the same as H2874 (טֶבַח); massacre;","strongs_def":"Tebach, the name of a Mesopotamian and of an Israelite","kjv_def":"Tebah."},
"H2876":{"lemma":"טַבָּח","xlit":"ṭabbâch","pron":"tab-bawkh'","derivation":"from H2873 (טָבַח);","strongs_def":"properly, a butcher; hence, a lifeguardsman (because he was acting as an executioner); also a cook (usually slaughtering the animal for food)","kjv_def":"cook, guard."},
"H2877":{"lemma":"טַבָּח","xlit":"ṭabbâch","pron":"tab-bawkh'","derivation":"(Aramaic) the same as H2876 (טַבָּח);","strongs_def":"a lifeguardsman","kjv_def":"guard."},
"H2878":{"lemma":"טִבְחָה","xlit":"ṭibchâh","pron":"tib-khaw'","derivation":"feminine of H2874 (טֶבַח) and meaning the same","strongs_def":"{properly, something slaughtered; hence, a beast (or meat, as butchered); abstractly butchery (or concretely, a place of slaughter)}","kjv_def":"flesh, slaughter."},
"H2879":{"lemma":"טַבָּחָה","xlit":"ṭabbâchâh","pron":"tab-baw-khaw'","derivation":"feminine of H2876 (טַבָּח);","strongs_def":"a female cook","kjv_def":"cook."},
"H2880":{"lemma":"טִבְחַת","xlit":"Ṭibchath","pron":"tib-khath'","derivation":"from H2878 (טִבְחָה); slaughter;","strongs_def":"Tibchath, a place in Syria","kjv_def":"Tibhath."},
"H2881":{"lemma":"טָבַל","xlit":"ṭâbal","pron":"taw-bal'","derivation":"a primitive root;","strongs_def":"to dip, to immerse","kjv_def":"dip, plunge."},
"H2882":{"lemma":"טְבַלְיָהוּ","xlit":"Ṭᵉbalyâhûw","pron":"teb-al-yaw'-hoo","derivation":"from H2881 (טָבַל) and H3050 (יָהּ); Jah has dipped;","strongs_def":"Tebaljah, an Israelite","kjv_def":"Tebaliah."},
"H2883":{"lemma":"טָבַע","xlit":"ṭâbaʻ","pron":"taw-bah'","derivation":"a primitive root;","strongs_def":"to sink","kjv_def":"drown, fasten, settle, sink."},
"H2884":{"lemma":"טַבָּעוֹת","xlit":"Ṭabbâʻôwth","pron":"tab-baw-othe'","derivation":"plural of H2885 (טַבַּעַת); rings;","strongs_def":"Tabbaoth, one of the Nethinim","kjv_def":"Tabaoth."},
"H2885":{"lemma":"טַבַּעַת","xlit":"ṭabbaʻath","pron":"tab-bah'-ath","derivation":"from H2883 (טָבַע);","strongs_def":"properly, a seal (as sunk into the wax), i.e. signet (for sealing); hence (generally) a ring of any kind","kjv_def":"ring."},
"H2886":{"lemma":"טַבְרִמּוֹן","xlit":"Ṭabrimmôwn","pron":"tab-rim-mone'","derivation":"from H2895 (טוֹב) and H7417 (רִמּוֹן); pleasing (to) Rimmon;","strongs_def":"Tabrimmon, a Syrian","kjv_def":"Tabrimmon."},
"H2887":{"lemma":"טֵבֶת","xlit":"Ṭêbeth","pron":"tay'-beth","derivation":"probably of foreign derivation;","strongs_def":"Tebeth, the tenth Hebrew month","kjv_def":"Tebeth."},
"H2888":{"lemma":"טַבַּת","xlit":"Ṭabbath","pron":"tab-bath'","derivation":"of uncertain derivation;","strongs_def":"Tabbath, a place East of the Jordan","kjv_def":"Tabbath."},
"H2889":{"lemma":"טָהוֹר","xlit":"ṭâhôwr","pron":"taw-hore'","derivation":"or טָהֹר; from H2891 (טָהֵר);","strongs_def":"pure (in a physical, chemical, ceremonial or moral sense)","kjv_def":"clean, fair, pure(-ness)."},
"H2890":{"lemma":"טְהוֹר","xlit":"ṭᵉhôwr","pron":"teh-hore'","derivation":"from H2891 (טָהֵר);","strongs_def":"purity","kjv_def":"pureness."},
"H2891":{"lemma":"טָהֵר","xlit":"ṭâhêr","pron":"taw-hare'","derivation":"a primitive root; properly, to be bright; i.e. (by implication)","strongs_def":"to be pure (physical sound, clear, unadulterated; Levitically, uncontaminated; morally, innocent or holy)","kjv_def":"be (make, make self, pronounce) clean, cleanse (self), purge, purify(-ier, self)."},
"H2892":{"lemma":"טֹהַר","xlit":"ṭôhar","pron":"to'-har","derivation":"from H2891 (טָהֵר);","strongs_def":"literally brightness; ceremonially purification","kjv_def":"clearness, glory, purifying."},
"H2893":{"lemma":"טׇהֳרָה","xlit":"ṭohŏrâh","pron":"toh-or-aw'","derivation":"feminine of H2892 (טֹהַר);","strongs_def":"ceremonial purification; moral purity","kjv_def":"[idiom] is cleansed, cleansing, purification(-fying)."},
"H2894":{"lemma":"טוּא","xlit":"ṭûwʼ","pron":"too","derivation":"a primitive root;","strongs_def":"to sweep away","kjv_def":"sweep."},
"H2895":{"lemma":"טוֹב","xlit":"ṭôwb","pron":"tobe","derivation":"a primitive root,","strongs_def":"to be (transitively, do or make) good (or well) in the widest sense","kjv_def":"be (do) better, cheer, be (do, seem) good, (make) goodly, [idiom] please, (be, do, go, play) well."},
"H2896":{"lemma":"טוֹב","xlit":"ṭôwb","pron":"tobe","derivation":"from H2895 (טוֹב);","strongs_def":"good (as an adjective) in the widest sense; used likewise as a noun, both in the masculine and the feminine, the singular and the plural (good, a good or good thing, a good man or woman; the good, goods or good things, good men or women), also as an adverb (well)","kjv_def":"beautiful, best, better, bountiful, cheerful, at ease, [idiom] fair (word), (be in) favour, fine, glad, good (deed, -lier, -liest, -ly, -ness, -s), graciously, joyful, kindly, kindness, liketh (best), loving, merry, [idiom] most, pleasant, [phrase] pleaseth, pleasure, precious, prosperity, ready, sweet, wealth, welfare, (be) well(-favoured)."},
"H2897":{"lemma":"טוֹב","xlit":"Ṭôwb","pron":"tobe","derivation":"the same as H2896 (טוֹב); good;","strongs_def":"Tob, a region apparently East of the Jordan","kjv_def":"Tob."},
"H2898":{"lemma":"טוּב","xlit":"ṭûwb","pron":"toob","derivation":"from H2895 (טוֹב);","strongs_def":"good (as a noun), in the widest sense, especially goodness (superlative concretely, the best), beauty, gladness, welfare","kjv_def":"fair, gladness, good(-ness, thing, -s), joy, go well with."},
"H2899":{"lemma":"טוֹב אֲדֹנִיָּהוּ","xlit":"Ṭôwb ʼĂdônîyâhûw","pron":"tobe ado-nee-yah'-hoo","derivation":"from H2896 (טוֹב) and H138 (אֲדֹנִיָּה); pleasing (to) Adonijah;","strongs_def":"Tob-Adonijah, an Israelite","kjv_def":"Tob-adonijah."},
"H2900":{"lemma":"טוֹבִיָּה","xlit":"Ṭôwbîyâh","pron":"to-bee-yaw'","derivation":"or טוֹבִיָּהוּ; from H2896 (טוֹב) and H3050 (יָהּ); goodness of Jehovah;","strongs_def":"Tobijah, the name of three Israelites and of one Samaritan","kjv_def":"Tobiah, Tobijah."},
"H2901":{"lemma":"טָוָה","xlit":"ṭâvâh","pron":"taw-vaw'","derivation":"a primitive root;","strongs_def":"to spin","kjv_def":"spin."},
"H2902":{"lemma":"טוּחַ","xlit":"ṭûwach","pron":"too'-akh","derivation":"a primitive root;","strongs_def":"to smear, especially with lime","kjv_def":"daub, overlay, plaister, smut."},
"H2903":{"lemma":"טוֹפָפָה","xlit":"ṭôwphâphâh","pron":"to-faw-faw'","derivation":"from an unused root meaning to go around or bind;","strongs_def":"a fillet for the forehead","kjv_def":"frontlet."},
"H2904":{"lemma":"טוּל","xlit":"ṭûwl","pron":"tool","derivation":"a primitive root;","strongs_def":"to pitch over or reel; hence (transitively) to cast down or out","kjv_def":"carry away, (utterly) cast (down, forth, out), send out."},
"H2905":{"lemma":"טוּר","xlit":"ṭûwr","pron":"toor","derivation":"from an unused root meaning to range in a regular manner;","strongs_def":"a row; hence, a wall","kjv_def":"row."},
"H2906":{"lemma":"טוּר","xlit":"ṭûwr","pron":"toor","derivation":"(Aramaic) corresponding to H6697 (צוּר);","strongs_def":"a rock or hill","kjv_def":"mountain."},
"H2907":{"lemma":"טוּשׂ","xlit":"ṭûws","pron":"toos","derivation":"a primitive root;","strongs_def":"to pounce as a bird of prey","kjv_def":"haste."},
"H2908":{"lemma":"טְוָת","xlit":"ṭᵉvâth","pron":"tev-awth'","derivation":"(Aramaic) from a root corresponding to H2901 (טָוָה);","strongs_def":"hunger (as twisting)","kjv_def":"fasting."},
"H2909":{"lemma":"טָחָה","xlit":"ṭâchâh","pron":"taw-khaw'","derivation":"a primitive root;","strongs_def":"to stretch a bow, as an archer","kjv_def":"(bow-) shot."},
"H2910":{"lemma":"טֻחָה","xlit":"ṭuchâh","pron":"too-khaw'","derivation":"from H2909 (טָחָה) (or H2902 (טוּחַ)) in the sense of overlaying;","strongs_def":"(in the plural only) the kidneys (as being covered); hence (figuratively) the inmost thought","kjv_def":"inward parts."},
"H2911":{"lemma":"טְחוֹן","xlit":"ṭᵉchôwn","pron":"tekh-one'","derivation":"from H2912 (טָחַן);","strongs_def":"a hand mill; hence, a millstone","kjv_def":"to grind."},
"H2912":{"lemma":"טָחַן","xlit":"ṭâchan","pron":"taw-khan'","derivation":"a primitive root;","strongs_def":"to grind meal; hence, to be aconcubine (that being their employment)","kjv_def":"grind(-er)."},
"H2913":{"lemma":"טַחֲנָה","xlit":"ṭachănâh","pron":"takh-an-aw'","derivation":"from H2912 (טָחַן);","strongs_def":"a hand mill; hence (figuratively) chewing","kjv_def":"grinding."},
"H2914":{"lemma":"טְחֹר","xlit":"ṭᵉchôr","pron":"tekh-ore'","derivation":"from an unused root meaning to burn;","strongs_def":"a boil or ulcer (from the inflammation), especially a tumorin the anus or pudenda (the piles)","kjv_def":"emerod."},
"H2915":{"lemma":"טִיחַ","xlit":"ṭîyach","pron":"tee'akh","derivation":"from (the equiv. of) H2902 (טוּחַ);","strongs_def":"mortar or plaster","kjv_def":"daubing."},
"H2916":{"lemma":"טִיט","xlit":"ṭîyṭ","pron":"teet","derivation":"from an unused root meaning apparently to be sticky (rather perb. a demonstrative); from H2894 (טוּא), through the idea of dirt to be swept away);","strongs_def":"mud or clay; figuratively, calamity","kjv_def":"clay, dirt, mire."},
"H2917":{"lemma":"טִין","xlit":"ṭîyn","pron":"teen","derivation":"(Aramaic) perhaps, by interchange, for a word corresponding to H2916 (טִיט);","strongs_def":"clay","kjv_def":"miry."},
"H2918":{"lemma":"טִירָה","xlit":"ṭîyrâh","pron":"tee-raw'","derivation":"feminine of (an equivalent to) H2905 (טוּר);","strongs_def":"a wall; hence, a fortress or a hamlet","kjv_def":"(goodly) castle, habitation, palace, row."},
"H2919":{"lemma":"טַל","xlit":"ṭal","pron":"tal","derivation":"from H2926 (טָלַל);","strongs_def":"dew (as covering vegetation)","kjv_def":"dew."},
"H2920":{"lemma":"טַל","xlit":"ṭal","pron":"tal","derivation":"(Aramaic) the same as H2919 (טַל)","strongs_def":"{dew (as covering vegetation)}","kjv_def":"dew."},
"H2921":{"lemma":"טָלָא","xlit":"ṭâlâʼ","pron":"taw-law'","derivation":"a primitive root;","strongs_def":"properly, to cover with pieces; i.e. (by implication) to spot or variegate (as tapestry)","kjv_def":"clouted, with divers colours, spotted."},
"H2922":{"lemma":"טְלָא","xlit":"ṭᵉlâʼ","pron":"tel-aw'","derivation":"apparently from H2921 (טָלָא) in the (orig.) sense of covering (for protection); (compare H2924 (טָלֶה))","strongs_def":"a lamb","kjv_def":"lamb."},
"H2923":{"lemma":"טְלָאִים","xlit":"Ṭᵉlâʼîym","pron":"tel-aw-eem'","derivation":"from the plural of H2922 (טְלָא); lambs;","strongs_def":"Telaim, a place in Palestine","kjv_def":"Telaim."},
"H2924":{"lemma":"טָלֶה","xlit":"ṭâleh","pron":"taw-leh'","derivation":"by variation for H2922 (טְלָא);","strongs_def":"a lamb","kjv_def":"lamb."},
"H2925":{"lemma":"טַלְטֵלָה","xlit":"ṭalṭêlâh","pron":"tal-tay-law'","derivation":"from H2904 (טוּל);","strongs_def":"overthrow or rejection","kjv_def":"captivity."},
"H2926":{"lemma":"טָלַל","xlit":"ṭâlal","pron":"taw-lal'","derivation":"a primitive root;","strongs_def":"properly, to strew over, i.e. (by implication) to cover in or plate (with beams)","kjv_def":"cover."},
"H2927":{"lemma":"טְלַל","xlit":"ṭᵉlal","pron":"tel-al'","derivation":"(Aramaic) corresponding to H2926 (טָלַל);","strongs_def":"to cover with shade","kjv_def":"have a shadow."},
"H2928":{"lemma":"טֶלֶם","xlit":"Ṭelem","pron":"teh'-lem","derivation":"from an unused root meaning to break up or treat violently; oppression;","strongs_def":"Telem, the name of a place in Idumaea, also of a temple doorkeeper","kjv_def":"Telem."},
"H2929":{"lemma":"טַלְמוֹן","xlit":"Ṭalmôwn","pron":"tal-mone'","derivation":"from the same as H2728 (חָרְגֹּל); oppressive;","strongs_def":"Talmon, a temple doorkeeper","kjv_def":"Talmon."},
"H2930":{"lemma":"טָמֵא","xlit":"ṭâmêʼ","pron":"taw-may'","derivation":"a primitive root;","strongs_def":"to be foul, especially in a ceremial or moral sense (contaminated)","kjv_def":"defile (self), pollute (self), be (make, make self, pronounce) unclean, [idiom] utterly."},
"H2931":{"lemma":"טָמֵא","xlit":"ṭâmêʼ","pron":"taw-may'","derivation":"from H2930 (טָמֵא);","strongs_def":"foul in a religious sense","kjv_def":"defiled, [phrase] infamous, polluted(-tion), unclean."},
"H2932":{"lemma":"טֻמְאָה","xlit":"ṭumʼâh","pron":"toom-aw'","derivation":"from H2930 (טָמֵא);","strongs_def":"religious impurity","kjv_def":"filthiness, unclean(-ness)."},
"H2933":{"lemma":"טָמָה","xlit":"ṭâmâh","pron":"taw-maw'","derivation":"a collateral form of H2930 (טָמֵא);","strongs_def":"to be impure in a religious sense","kjv_def":"be defiled, be reputed vile."},
"H2934":{"lemma":"טָמַן","xlit":"ṭâman","pron":"taw-man'","derivation":"a primitive root;","strongs_def":"to hide (by covering over)","kjv_def":"hide, lay privily, in secret."},
"H2935":{"lemma":"טֶנֶא","xlit":"ṭeneʼ","pron":"teh'-neh","derivation":"from an unused root probably meaning to weave;","strongs_def":"a basket (of interlaced osiers)","kjv_def":"basket."},
"H2936":{"lemma":"טָנַף","xlit":"ṭânaph","pron":"taw-naf'","derivation":"a primitive root;","strongs_def":"to soil","kjv_def":"defile."},
"H2937":{"lemma":"טָעָה","xlit":"ṭâʻâh","pron":"taw-aw'","derivation":"a primitive root;","strongs_def":"to wander; causatively to lead astray","kjv_def":"seduce."},
"H2938":{"lemma":"טָעַם","xlit":"ṭâʻam","pron":"taw-am'","derivation":"a primitive root;","strongs_def":"to taste; figuratively, to perceive","kjv_def":"[idiom] but, perceive, taste."},
"H2939":{"lemma":"טְעַם","xlit":"ṭᵉʻam","pron":"teh-am'","derivation":"(Aramaic) corresponding to H2938 (טָעַם);","strongs_def":"to taste; causatively to feed","kjv_def":"make to eat, feed."},
"H2940":{"lemma":"טַעַם","xlit":"ṭaʻam","pron":"tah'-am","derivation":"from H2938 (טָעַם);","strongs_def":"properly, a taste, i.e. (figuratively) perception; by implication, intelligence; transitively, a mandate","kjv_def":"advice, behaviour, decree, discretion, judgment, reason, taste, understanding."},
"H2941":{"lemma":"טַעַם","xlit":"ṭaʻam","pron":"tah'-am","derivation":"(Aramaic) from H2939 (טְעַם); compare H2940 (טַעַם))","strongs_def":"properly, a taste, i.e.  ajudicial sentence","kjv_def":"account, [idiom] to be commanded, commandment, matter."},
"H2942":{"lemma":"טְעֵם","xlit":"ṭᵉʻêm","pron":"teh-ame'","derivation":"(Aramaic) from H2939 (טְעַם), and equivalent to H2941 (טַעַם);","strongs_def":"properly, flavor; figuratively, judgment (both subjective and objective); hence, account (both subjectively and objectively)","kjv_def":"[phrase] chancellor, [phrase] command, commandment, decree, [phrase] regard, taste, wisdom."},
"H2943":{"lemma":"טָעַן","xlit":"ṭâʻan","pron":"taw-an'","derivation":"a primitive root;","strongs_def":"to load a beast","kjv_def":"lade."},
"H2944":{"lemma":"טָעַן","xlit":"ṭâʻan","pron":"taw-an'","derivation":"a primitive root;","strongs_def":"to stab","kjv_def":"thrust through."},
"H2945":{"lemma":"טַף","xlit":"ṭaph","pron":"taf","derivation":"from H2952 (טָפַף) (perhaps referring to the tripping gait of children);","strongs_def":"a family (mostly used collectively in the singular)","kjv_def":"(little) children (ones), families."},
"H2946":{"lemma":"טָפַח","xlit":"ṭâphach","pron":"taw-fakh'","derivation":"a primitive root; or perhaps a denominative from H2947 (טֵפַח), from dandling on the palms","strongs_def":"to flatten out or extend (as a tent); figuratively, to nurse achild (as promotive of growth);","kjv_def":"span, swaddle."},
"H2947":{"lemma":"טֵפַח","xlit":"ṭêphach","pron":"tay'-fakh","derivation":"from H2946 (טָפַח);","strongs_def":"a spread of the hand, i.e. a palm-breadth (not 'span' of the fingers); architecturally, a corbel (as a supporting palm)","kjv_def":"coping, hand-breadth."},
"H2948":{"lemma":"טֹפַח","xlit":"ṭôphach","pron":"to'-fakh","derivation":"from H2946 (טָפַח) (the same as H2947 (טֵפַח))","strongs_def":"{a spread of the hand, i.e. a palm-breadth (not 'span' of the fingers); architecturally, a corbel (as a supporting palm)}","kjv_def":"hand-breadth (broad)."},
"H2949":{"lemma":"טִפֻּח","xlit":"ṭippuch","pron":"tip-pookh'","derivation":"from H2946 (טָפַח);","strongs_def":"nursing","kjv_def":"span long."},
"H2950":{"lemma":"טָפַל","xlit":"ṭâphal","pron":"taw-fal'","derivation":"a primitive root;","strongs_def":"properly, to stick on as apatch; figuratively, to impute falsely","kjv_def":"forge(-r), sew up."},
"H2951":{"lemma":"טִפְסַר","xlit":"ṭiphçar","pron":"tif-sar'","derivation":"of foreign derivation;","strongs_def":"a military governor","kjv_def":"captain."},
"H2952":{"lemma":"טָפַף","xlit":"ṭâphaph","pron":"taw-faf'","derivation":"a primitive root; apparently","strongs_def":"to trip (with short steps) coquettishly","kjv_def":"mince."},
"H2953":{"lemma":"טְפַר","xlit":"ṭᵉphar","pron":"tef-ar'","derivation":"(Aramaic) from a root corresponding to H6852 (צָפַר), and meaning the same as H6856 (צִפֹּרֶן);","strongs_def":"a finger-nail; also a hoof or claw","kjv_def":"nail."},
"H2954":{"lemma":"טָפַשׁ","xlit":"ṭâphash","pron":"taw-fash'","derivation":"a primitive root;","strongs_def":"properly, apparently to be thick; figuratively, to be stupid","kjv_def":"be fat."},
"H2955":{"lemma":"טָפַת","xlit":"Ṭâphath","pron":"taw-fath'","derivation":"probably from H5197 (נָטַף); a dropping (of ointment);","strongs_def":"Taphath, an Israelitess","kjv_def":"Taphath."},
"H2956":{"lemma":"טָרַד","xlit":"ṭârad","pron":"taw-rad'","derivation":"a primitive root;","strongs_def":"to drive on; figuratively, to follow close","kjv_def":"continual."},
"H2957":{"lemma":"טְרַד","xlit":"ṭᵉrad","pron":"ter-ad'","derivation":"(Aramaic) corresponding to H2956 (טָרַד);","strongs_def":"to expel","kjv_def":"drive."},
"H2958":{"lemma":"טְרוֹם","xlit":"ṭᵉrôwm","pron":"ter-ome'","derivation":"a variation of H2962 (טֶרֶם);","strongs_def":"not yet","kjv_def":"before."},
"H2959":{"lemma":"טָרַח","xlit":"ṭârach","pron":"taw-rakh'","derivation":"a primitive root;","strongs_def":"to overburden","kjv_def":"weary."},
"H2960":{"lemma":"טֹרַח","xlit":"ṭôrach","pron":"to'-rakh","derivation":"from H2959 (טָרַח);","strongs_def":"a burden","kjv_def":"cumbrance, trouble."},
"H2961":{"lemma":"טָרִי","xlit":"ṭârîy","pron":"taw-ree'","derivation":"from an unused root apparently meaning to be moist;","strongs_def":"properly, dripping; hence, fresh (i.e. recently made such)","kjv_def":"new, putrefying."},
"H2962":{"lemma":"טֶרֶם","xlit":"ṭerem","pron":"teh'-rem","derivation":"from an unused root apparently meaning to interrupt or suspend;","strongs_def":"properly, non-occurrence; used adverbially, not yet or before","kjv_def":"before, ere, not yet."},
"H2963":{"lemma":"טָרַף","xlit":"ṭâraph","pron":"taw-raf'","derivation":"a primitive root;","strongs_def":"to pluck off or pull to pieces; causatively to supply with food (as in morsels)","kjv_def":"catch, [idiom] without doubt, feed, ravin, rend in pieces, [idiom] surely, tear (in pieces)."},
"H2964":{"lemma":"טֶרֶף","xlit":"ṭereph","pron":"teh'-ref","derivation":"from H2963 (טָרַף);","strongs_def":"something torn, i.e. a fragment, e.g. a fresh leaf, prey, food","kjv_def":"leaf, meat, prey, spoil."},
"H2965":{"lemma":"טָרָף","xlit":"ṭârâph","pron":"taw-rawf'","derivation":"from H2963 (טָרַף);","strongs_def":"recently torn off, i.e. fresh","kjv_def":"pluckt off."},
"H2966":{"lemma":"טְרֵפָה","xlit":"ṭᵉrêphâh","pron":"ter-ay-faw'","derivation":"feminine (collectively) of H2964 (טֶרֶף);","strongs_def":"prey, i.e. flocks devoured by animals","kjv_def":"ravin, (that which was) torn (of beasts, in pieces)."},
"H2967":{"lemma":"טַרְפְּלַי","xlit":"Ṭarpᵉlay","pron":"tar-pel-ah'-ee","derivation":"(Aramaic) from a name of foreign derivation;","strongs_def":"a Tarpelite (collectively) or inhabitants of Tarpel, a place in Assyria","kjv_def":"Tarpelites. y"},
"H2968":{"lemma":"יָאַב","xlit":"yâʼab","pron":"yaw-ab'","derivation":"a primitive root;","strongs_def":"to desire","kjv_def":"long."},
"H2969":{"lemma":"יָאָה","xlit":"yâʼâh","pron":"yaw-aw'","derivation":"a primitive root;","strongs_def":"to be suitable","kjv_def":"appertain."},
"H2970":{"lemma":"יַאֲזַנְיָה","xlit":"Yaʼăzanyâh","pron":"yah-az-an-yaw'","derivation":"or יַאֲזַנְיָהוּ; from H238 (אָזַן) and H3050 (יָהּ); heard of Jah;","strongs_def":"Jaazanjah, the name of four Israelites","kjv_def":"Jaazaniah. Compare H3153 (יְזַנְיָה)."},
"H2971":{"lemma":"יָאִיר","xlit":"Yâʼîyr","pron":"yaw-ere'","derivation":"from H215 (אוֹר); enlightener;","strongs_def":"Jair, the name of four Israelites","kjv_def":"Jair."},
"H2972":{"lemma":"יָאִרִי","xlit":"Yâʼirîy","pron":"yaw-ee-ree'","derivation":"patronymically from H2971 (יָאִיר);","strongs_def":"a Jairite or descendant of Jair","kjv_def":"Jairite."},
"H2973":{"lemma":"יָאַל","xlit":"yâʼal","pron":"yaw-al'","derivation":"a primitive root;","strongs_def":"properly, to be slack, i.e. (figuratively) to be foolish","kjv_def":"dote, be (become, do) foolish(-ly)."},
"H2974":{"lemma":"יָאַל","xlit":"yâʼal","pron":"yaw-al'","derivation":"a primitive root (probably rather the same as H2973 (יָאַל) through the idea of mental weakness);","strongs_def":"properly, to yield, especially assent; hence (pos.) to undertake as an act of volition","kjv_def":"assay, begin, be content, please, take upon, [idiom] willingly, would."},
"H2975":{"lemma":"יְאֹר","xlit":"yᵉʼôr","pron":"yeh-ore'","derivation":"of Egyptian origin;","strongs_def":"a channel, e.g. a fosse, canal, shaft; specifically the Nile, as the one river of Egypt, including its collateral trenches; also the Tigris, as the main river of Assyria","kjv_def":"brook, flood, river, stream."},
"H2976":{"lemma":"יָאַשׁ","xlit":"yâʼash","pron":"yaw-ash'","derivation":"a primitive root;","strongs_def":"to desist, i.e. (figuratively) to despond","kjv_def":"(cause to) despair, one that is desperate, be no hope."},
"H2977":{"lemma":"יֹאשִׁיָּה","xlit":"Yôʼshîyâh","pron":"yo-shee-yaw'","derivation":"or יֹאשִׁיָּהוּ; from the same root as H803 (אֲשׁוּיָה) and H3050 (יָהּ); founded of Jah;","strongs_def":"Joshijah, the name of two Israelites","kjv_def":"Josiah."},
"H2978":{"lemma":"יְאִתוֹן","xlit":"yᵉʼithôwn","pron":"yeh-ee-thone'","derivation":"from H857 (אָתָה);","strongs_def":"an entry","kjv_def":"entrance."},
"H2979":{"lemma":"יְאָתְרַי","xlit":"yᵉʼâthᵉray","pron":"yeh-aw-ther-ah'ee","derivation":"from the same as H871 (אֲתָרִים); stepping;","strongs_def":"Jeatherai, an Israelite","kjv_def":"Jeaterai."},
"H2980":{"lemma":"יָבַב","xlit":"yâbab","pron":"yaw-bab","derivation":"a primitive root;","strongs_def":"to bawl","kjv_def":"cry out."},
"H2981":{"lemma":"יְבוּל","xlit":"yᵉbûwl","pron":"yeb-ool'","derivation":"from H2986 (יָבַל);","strongs_def":"produce, i.e. a crop or (figuratively) wealth","kjv_def":"fruit, increase."},
"H2982":{"lemma":"יְבוּס","xlit":"Yᵉbûwç","pron":"yeb-oos'","derivation":"from H947 (בּוּס); trodden, i.e. threshing-place;","strongs_def":"Jebus, the original name of Jerusalem","kjv_def":"Jebus."},
"H2983":{"lemma":"יְבוּסִי","xlit":"Yᵉbûwçîy","pron":"yeb-oo-see'","derivation":"patrial from H2982 (יְבוּס);","strongs_def":"a Jebusite or inhabitant of Jebus","kjv_def":"Jebusite(-s)."},
"H2984":{"lemma":"יִבְחַר","xlit":"Yibchar","pron":"yib-khar'","derivation":"from H977 (בָּחַר); choice;","strongs_def":"Jibchar, an Israelite","kjv_def":"Ibhar."},
"H2985":{"lemma":"יָבִין","xlit":"Yâbîyn","pron":"yaw-bene'","derivation":"from H995 (בִּין); intelligent;","strongs_def":"Jabin, the name of two Canaanitish kings","kjv_def":"Jabin."},
"H2986":{"lemma":"יָבַל","xlit":"yâbal","pron":"yaw-bal'","derivation":"a primitive root;","strongs_def":"properly, to flow; causatively, to bring (especially with pomp)","kjv_def":"bring (forth), carry, lead (forth)."},
"H2987":{"lemma":"יְבַל","xlit":"yᵉbal","pron":"yeb-al'","derivation":"(Aramaic) corresponding to H2986 (יָבַל);","strongs_def":"to bring","kjv_def":"bring, carry."},
"H2988":{"lemma":"יָבָל","xlit":"yâbâl","pron":"yaw-bawl'","derivation":"from H2986 (יָבַל);","strongs_def":"a stream","kjv_def":"(water-) course, stream."},
"H2989":{"lemma":"יָבָל","xlit":"Yâbâl","pron":"yaw-bawl'","derivation":"the same as H2988 (יָבָל);","strongs_def":"Jabal, an antediluvian","kjv_def":"Jabal."},
"H2990":{"lemma":"יַבֵּל","xlit":"yabbêl","pron":"yab-bale'","derivation":"from H2986 (יָבַל);","strongs_def":"having running sores","kjv_def":"wen."},
"H2991":{"lemma":"יִבְלְעָם","xlit":"Yiblᵉʻâm","pron":"yib-leh-awm'","derivation":"from H1104 (בָּלַע) and H5971 (עַם); devouring people;","strongs_def":"Jibleam, a place in Palestine","kjv_def":"Ibleam."},
"H2992":{"lemma":"יָבַם","xlit":"yâbam","pron":"yaw-bam'","derivation":"a primitive root of doubtful meaning; used only as a denominative from H2993 (יָבָם);","strongs_def":"to marry a (deceased) brother's widow","kjv_def":"perform the duty of a husband's brother, marry."},
"H2993":{"lemma":"יָבָם","xlit":"yâbâm","pron":"yaw-bawm'","derivation":"from (the orig. of) H2992 (יָבַם);","strongs_def":"a brotherin-law","kjv_def":"husband's brother."},
"H2994":{"lemma":"יְבֵמֶת","xlit":"Yᵉbêmeth","pron":"yeb-ay'-meth","derivation":"feminine participle of H2992 (יָבַם);","strongs_def":"a sister-in-law","kjv_def":"brother's wife, sister in law."},
"H2995":{"lemma":"יַבְנְאֵל","xlit":"Yabnᵉʼêl","pron":"yab-neh-ale'","derivation":"from H1129 (בָּנָה) and H410 (אֵל); built of God;","strongs_def":"Jabneel, the name of two places in Palestine","kjv_def":"Jabneel."},
"H2996":{"lemma":"יַבְנֶה","xlit":"Yabneh","pron":"yab-neh'","derivation":"from H1129 (בָּנָה); a building;","strongs_def":"Jabneh, a place in Palestine","kjv_def":"Jabneh."},
"H2997":{"lemma":"יִבְנְיָה","xlit":"Yibnᵉyâh","pron":"yib-neh-yaw'","derivation":"from H1129 (בָּנָה) and H3050 (יָהּ); built of Jah;","strongs_def":"Jibnejah, an Israelite","kjv_def":"Ibneiah."},
"H2998":{"lemma":"יִבְנִיָּה","xlit":"Yibnîyâh","pron":"yib-nee-yaw'","derivation":"from H1129 (בָּנָה) and H3050 (יָהּ); building of Jah;","strongs_def":"Jibnijah, an Israelite","kjv_def":"Ibnijah."},
"H2999":{"lemma":"יַבֹּק","xlit":"Yabbôq","pron":"yab-boke'","derivation":"probably from H1238 (בָּקַק); pouring forth;","strongs_def":"Jabbok, a river east of the Jordan","kjv_def":"Jabbok."},
"H3000":{"lemma":"יְבֶרֶכְיָהוּ","xlit":"Yᵉberekyâhûw","pron":"yeb-eh-rek-yaw'-hoo","derivation":"from H1288 (בָרַךְ) and H3050 (יָהּ); blessed of Jah;","strongs_def":"Jeberekjah, an Israelite","kjv_def":"Jeberechiah."},
"H3001":{"lemma":"יָבֵשׁ","xlit":"yâbêsh","pron":"yaw-bashe'","derivation":"a primitive root;","strongs_def":"to be ashamed, confused or disappointed; also (as failing) to dry up (as water) or wither (as herbage)","kjv_def":"be ashamed, clean, be confounded, (make) dry (up), (do) shame(-fully), [idiom] utterly, wither (away)."},
"H3002":{"lemma":"יָבֵשׁ","xlit":"yâbêsh","pron":"yaw-bashe'","derivation":"from H3001 (יָבֵשׁ);","strongs_def":"dry","kjv_def":"dried (away), dry."},
"H3003":{"lemma":"יָבֵשׁ","xlit":"Yâbêsh","pron":"yaw-bashe'","derivation":"the same as H3002 (יָבֵשׁ) (also יָבֵישׁ; often with the addition of H1568 (גִּלְעָד), i.e. Jabesh of Gilad);","strongs_def":"Jobesh, the name of an Israelite and of a place in Palestine","kjv_def":"Jobesh (-Gilead)."},
"H3004":{"lemma":"יַבָּשָׁה","xlit":"yabbâshâh","pron":"yab-baw-shaw'","derivation":"from H3001 (יָבֵשׁ);","strongs_def":"dry ground","kjv_def":"dry (ground, land)."},
"H3005":{"lemma":"יִבְשָׂם","xlit":"Yibsâm","pron":"yib-sawm'","derivation":"from the same as H1314 (בֶּשֶׂם); fragrant;","strongs_def":"Jibsam, an Israelite","kjv_def":"Jibsam."},
"H3006":{"lemma":"יַבֶּשֶׁת","xlit":"yabbesheth","pron":"yab-beh'-sheth","derivation":"a variation of H3004 (יַבָּשָׁה);","strongs_def":"dry ground","kjv_def":"dry land."},
"H3007":{"lemma":"יַבֶּשֶׁת","xlit":"yabbesheth","pron":"yab-beh'-sheth","derivation":"(Aramaic) corresponding to H3006 (יַבֶּשֶׁת);","strongs_def":"dry land","kjv_def":"earth."},
"H3008":{"lemma":"יִגְאָל","xlit":"Yigʼâl","pron":"yig-awl'","derivation":"from H1350 (גָּאַל); avenger;","strongs_def":"Jigal, the name of three Israelites","kjv_def":"Igal, Igeal."},
"H3009":{"lemma":"יָגַב","xlit":"yâgab","pron":"yaw-gab'","derivation":"a primitive root;","strongs_def":"to dig or plow","kjv_def":"husbandman."},
"H3010":{"lemma":"יָגֵב","xlit":"yâgêb","pron":"yaw-gabe'","derivation":"from H3009 (יָגַב);","strongs_def":"a plowed field","kjv_def":"field."},
"H3011":{"lemma":"יׇגְבְּהָה","xlit":"Yogbᵉhâh","pron":"yog-beh-haw'","derivation":"feminine from H1361 (גָּבַהּ); hillock;","strongs_def":"Jogbehah, a place East of the Jordan","kjv_def":"Jogbehah."},
"H3012":{"lemma":"יִגְדַּלְיָהוּ","xlit":"Yigdalyâhûw","pron":"yig-dal-yaw'-hoo","derivation":"from H1431 (גָּדַל) and H3050 (יָהּ); magnified of Jah;","strongs_def":"Jigdaljah, an Israelite","kjv_def":"Igdaliah."},
"H3013":{"lemma":"יָגָה","xlit":"yâgâh","pron":"yaw-gaw'","derivation":"a primitive root;","strongs_def":"to grieve","kjv_def":"afflict, cause grief, grieve, sorrowful, vex."},
"H3014":{"lemma":"יָגָה","xlit":"yâgâh","pron":"yaw-gaw'","derivation":"a primitive root (probably rather the same as H3013 (יָגָה) through the common idea of dissatisfaction);","strongs_def":"to push away","kjv_def":"be removed."},
"H3015":{"lemma":"יָגוֹן","xlit":"yâgôwn","pron":"yaw-gohn'","derivation":"from H3013 (יָגָה);","strongs_def":"affliction","kjv_def":"grief, sorrow."},
"H3016":{"lemma":"יָגוֹר","xlit":"yâgôwr","pron":"yaw-gore'","derivation":"from H3025 (יָגֹר);","strongs_def":"fearful","kjv_def":"afraid, fearest."},
"H3017":{"lemma":"יָגוּר","xlit":"Yâgûwr","pron":"yaw-goor'","derivation":"probably from H1481 (גּוּר); a lodging;","strongs_def":"Jagur, a place in Palestine","kjv_def":"Jagur."},
"H3018":{"lemma":"יְגִיעַ","xlit":"yᵉgîyaʻ","pron":"yeg-ee'-ah","derivation":"from H3021 (יָגַע);","strongs_def":"toil; hence, a work, produce, property (as the result of labor)","kjv_def":"labour, work."},
"H3019":{"lemma":"יָגִיעַ","xlit":"yâgîyaʻ","pron":"yaw-ghee'-ah","derivation":"from H3021 (יָגַע);","strongs_def":"tired","kjv_def":"weary."},
"H3020":{"lemma":"יׇגְלִי","xlit":"Yoglîy","pron":"yog-lee'","derivation":"from H1540 (גָּלָה); exiled;","strongs_def":"Jogli, an Israelite","kjv_def":"Jogli."},
"H3021":{"lemma":"יָגַע","xlit":"yâgaʻ","pron":"yaw-gah'","derivation":"a primitive root;","strongs_def":"properly, to gasp; hence, to be exhausted, to tire, to toil","kjv_def":"faint, (make to) labour, (be) weary."},
"H3022":{"lemma":"יָגָע","xlit":"yâgâʻ","pron":"yaw-gaw'","derivation":"from H3021 (יָגַע);","strongs_def":"earnings (as the product of toil)","kjv_def":"that which he laboured for."},
"H3023":{"lemma":"יָגֵעַ","xlit":"yâgêaʻ","pron":"yaw-gay'-ah","derivation":"from H3021 (יָגַע);","strongs_def":"tired; hence (transitive) tiresome","kjv_def":"full of labour, weary."},
"H3024":{"lemma":"יְגִעָה","xlit":"yᵉgiʻâh","pron":"yeg-ee-aw'","derivation":"feminine of H3019 (יָגִיעַ);","strongs_def":"fatigue","kjv_def":"weariness."},
"H3025":{"lemma":"יָגֹר","xlit":"yâgôr","pron":"yaw-gore'","derivation":"a primitive root;","strongs_def":"to fear","kjv_def":"be afraid, fear."},
"H3026":{"lemma":"יְגַר שַׂהֲדוּתָא","xlit":"Yᵉgar Sahădûwthâʼ","pron":"yegar' sah-had-oo-thaw'","derivation":"(Aramaic) from a word derived from an unused root (meaning to gather) and a derivation of a root corresponding to H7717 (שָׂהֵד); heap of the testimony;","strongs_def":"Jegar-Sahadutha, a cairn East of the Jordan","kjv_def":"Jegar-Sahadutha."},
"H3027":{"lemma":"יָד","xlit":"yâd","pron":"yawd","derivation":"a primitive word; in distinction from H3709 (כַּף), the closed one); used (as noun, adverb, etc.) in a great variety of applications, both literally and figuratively, both proximate and remote (as follows)","strongs_def":"a hand (the open one (indicating power, means, direction, etc.),","kjv_def":"([phrase] be) able, [idiom] about, [phrase] armholes, at, axletree, because of, beside, border, [idiom] bounty, [phrase] broad, (broken-) handed, [idiom] by, charge, coast, [phrase] consecrate, [phrase] creditor, custody, debt, dominion, [idiom] enough, [phrase] fellowship, force, [idiom] from, hand(-staves, -y work), [idiom] he, himself, [idiom] in, labour, [phrase] large, ledge, (left-) handed, means, [idiom] mine, ministry, near, [idiom] of, [idiom] order, ordinance, [idiom] our, parts, pain, power, [idiom] presumptuously, service, side, sore, state, stay, draw with strength, stroke, [phrase] swear, terror, [idiom] thee, [idiom] by them, [idiom] themselves, [idiom] thine own, [idiom] thou, through, [idiom] throwing, [phrase] thumb, times, [idiom] to, [idiom] under, [idiom] us, [idiom] wait on, (way-) side, where, [phrase] wide, [idiom] with (him, me, you), work, [phrase] yield, [idiom] yourselves."},
"H3028":{"lemma":"יַד","xlit":"yad","pron":"yad","derivation":"(Aramaic) corresponding to H3027 (יָד)","strongs_def":"{a hand (the open one (indicating power, means, direction, etc.)}","kjv_def":"hand, power."},
"H3029":{"lemma":"יְדָא","xlit":"yᵉdâʼ","pron":"yed-aw'","derivation":"(Aramaic) corresponding to H3034 (יָדָה);","strongs_def":"to praise","kjv_def":"(give) thank(-s)."},
"H3030":{"lemma":"יִדְאֲלָה","xlit":"Yidʼălâh","pron":"yid-al-aw'","derivation":"of uncertain derivation;","strongs_def":"Jidalah, a place in Palestine","kjv_def":"Idalah."},
"H3031":{"lemma":"יִדְבָּשׁ","xlit":"Yidbâsh","pron":"yid-bawsh'","derivation":"from the same as H1706 (דְּבַשׁ); perhaps honeyed;","strongs_def":"Jidbash, an Israelite","kjv_def":"Idbash."},
"H3032":{"lemma":"יָדַד","xlit":"yâdad","pron":"yaw-dad'","derivation":"a primitive root; (compare H3034 (יָדָה))","strongs_def":"properly, to handle, i.e. to throw, e.g. lots","kjv_def":"cast."},
"H3033":{"lemma":"יְדִדוּת","xlit":"yᵉdidûwth","pron":"yed-ee-dooth'","derivation":"from H3039 (יְדִיד);","strongs_def":"properly, affection; concretely, a darling object","kjv_def":"dearly beloved."},
"H3034":{"lemma":"יָדָה","xlit":"yâdâh","pron":"yaw-daw'","derivation":"a primitive root; used only as denominative from H3027 (יָד); literally, to use (i.e. hold out) the hand;","strongs_def":"physically, to throw (a stone, an arrow) at or away; especially to revere or worship (with extended hands); intensively, to bemoan (by wringing the hands)","kjv_def":"cast (out), (make) confess(-ion), praise, shoot, (give) thank(-ful, -s, -sgiving)."},
"H3035":{"lemma":"יִדּוֹ","xlit":"Yiddôw","pron":"yid-do'","derivation":"from H3034 (יָדָה); praised;","strongs_def":"Jiddo, an Israelite","kjv_def":"Iddo."},
"H3036":{"lemma":"יָדוֹן","xlit":"Yâdôwn","pron":"yaw-done'","derivation":"from H3034 (יָדָה); thankful;","strongs_def":"Jadon, an Israelite","kjv_def":"Jadon."},
"H3037":{"lemma":"יַדּוּעַ","xlit":"Yaddûwaʻ","pron":"yad-doo'-ah","derivation":"from H3045 (יָדַע); knowing;","strongs_def":"Jaddua, the name of two Israelites","kjv_def":"Jaddua."},
"H3038":{"lemma":"יְדוּתוּן","xlit":"Yᵉdûwthûwn","pron":"yed-oo-thoon'","derivation":"or יְדֻתוּן; or יְדִיתוּן; probably from H3034 (יָדָה); laudatory;","strongs_def":"Jeduthun, an Israelite","kjv_def":"Jeduthun."},
"H3039":{"lemma":"יְדִיד","xlit":"yᵉdîyd","pron":"yed-eed'","derivation":"from the same as H1730 (דּוֹד);","strongs_def":"loved","kjv_def":"amiable, (well-) beloved, loves."},
"H3040":{"lemma":"יְדִידָה","xlit":"Yᵉdîydâh","pron":"yed-ee-daw'","derivation":"feminine of H3039 (יְדִיד); beloved;","strongs_def":"Jedidah, an Israelitess","kjv_def":"Jedidah."},
"H3041":{"lemma":"יְדִידְיָה","xlit":"Yᵉdîydᵉyâh","pron":"yed-ee-deh-yaw'","derivation":"from H3039 (יְדִיד) and H3050 (יָהּ); beloved of Jah;","strongs_def":"Jedidejah, a name of Solomon","kjv_def":"Jedidiah."},
"H3042":{"lemma":"יְדָיָה","xlit":"Yᵉdâyâh","pron":"yed-aw-yaw'","derivation":"from H3034 (יָדָה) and H3050 (יָהּ); praised of Jah;","strongs_def":"Jedajah, the name of two Israelites","kjv_def":"Jedaiah."},
"H3043":{"lemma":"יְדִיעֲאֵל","xlit":"Yᵉdîyʻăʼêl","pron":"yed-ee-ah-ale'","derivation":"from H3045 (יָדַע) and H410 (אֵל); knowing God;","strongs_def":"Jediael, the name of three Israelites","kjv_def":"Jediael."},
"H3044":{"lemma":"יִדְלָף","xlit":"Yidlâph","pron":"yid-lawf'","derivation":"from H1811 (דָּלַף); tearful;","strongs_def":"Jidlaph, a Mesopotamian","kjv_def":"Jidlaph."},
"H3045":{"lemma":"יָדַע","xlit":"yâdaʻ","pron":"yaw-dah'","derivation":"a primitive root;","strongs_def":"to know (properly, to ascertain by seeing); used in a great variety of senses, figuratively, literally, euphemistically and inferentially (including observation, care, recognition; and causatively, instruction, designation, punishment, etc.)","kjv_def":"acknowledge, acquaintance(-ted with), advise, answer, appoint, assuredly, be aware, (un-) awares, can(-not), certainly, comprehend, consider, [idiom] could they, cunning, declare, be diligent, (can, cause to) discern, discover, endued with, familiar friend, famous, feel, can have, be (ig-) norant, instruct, kinsfolk, kinsman, (cause to let, make) know, (come to give, have, take) knowledge, have (knowledge), (be, make, make to be, make self) known, [phrase] be learned, [phrase] lie by man, mark, perceive, privy to, [idiom] prognosticator, regard, have respect, skilful, shew, can (man of) skill, be sure, of a surety, teach, (can) tell, understand, have (understanding), [idiom] will be, wist, wit, wot."},
"H3046":{"lemma":"יְדַע","xlit":"yᵉdaʻ","pron":"yed-ah'","derivation":"(Aramaic) corresponding to H3045 (יָדַע)","strongs_def":"{to know (properly, to ascertain by seeing); used in a great variety of senses, figuratively, literally, euphemistically and inferentially (including observation, care, recognition; and causatively, instruction, designation, punishment, etc.)}","kjv_def":"certify, know, make known, teach."},
"H3047":{"lemma":"יָדָע","xlit":"Yâdâʻ","pron":"yaw-daw'","derivation":"from H3045 (יָדַע); knowing;","strongs_def":"Jada, an Israelite","kjv_def":"Jada."},
"H3048":{"lemma":"יְדַעְיָה","xlit":"Yᵉdaʻyâh","pron":"yed-ah-yaw'","derivation":"from H3045 (יָדַע) and H3050 (יָהּ); Jah has known;","strongs_def":"Jedajah, the name of two Israelites","kjv_def":"Jedaiah."},
"H3049":{"lemma":"יִדְּעֹנִי","xlit":"yiddᵉʻônîy","pron":"yid-deh-o-nee'","derivation":"from H3045 (יָדַע);","strongs_def":"properly, a knowing one; specifically, a conjurer; (by impl) a ghost","kjv_def":"wizard."},
"H3050":{"lemma":"יָהּ","xlit":"Yâhh","pron":"yaw","derivation":"contraction for H3068 (יְהֹוָה), and meaning the same;","strongs_def":"Jah, the sacred name","kjv_def":"Jah, the Lord, most vehement. Compare names in '-iah,' '-jah.'"},
"H3051":{"lemma":"יָהַב","xlit":"yâhab","pron":"yaw-hab'","derivation":"a primitive root;","strongs_def":"to give (whether literal or figurative); generally, to put; imperatively (reflexive) come","kjv_def":"ascribe, bring, come on, give, go, set, take."},
"H3052":{"lemma":"יְהַב","xlit":"yᵉhab","pron":"yeh-hab'","derivation":"(Aramaic) corresponding to H3051 (יָהַב)","strongs_def":"{to give (whether literal or figurative); generally, to put; imperatively (reflexive) come}","kjv_def":"deliver, give, lay, [phrase] prolong, pay, yield."},
"H3053":{"lemma":"יְהָב","xlit":"yᵉhâb","pron":"ye-hawb'","derivation":"from H3051 (יָהַב);","strongs_def":"properly, what is given (by Providence), i.e. a lot","kjv_def":"burden."},
"H3054":{"lemma":"יָהַד","xlit":"yâhad","pron":"yaw-had'","derivation":"denominative from a form corresponding to H3061 (יְהוּד)lemma יְהיּד yod, corrected to יְהוּד;","strongs_def":"to Judaize, i.e. become Jewish","kjv_def":"become Jews."},
"H3055":{"lemma":"יְהֻד","xlit":"Yᵉhud","pron":"yeh-hood'","derivation":"a briefer form of one corresponding to H3061 (יְהוּד)lemma יְהיּד yod, corrected to יְהוּד;","strongs_def":"Jehud, a place in Palestine","kjv_def":"Jehud."},
"H3056":{"lemma":"יֶהְדַי","xlit":"Yehday","pron":"yeh-dah'-ee","derivation":"perhaps from a form corresponding to H3061 (יְהוּד)lemma יְהיּד yod, corrected to יְהוּד; Judaistic;","strongs_def":"Jehdai, an Israelite","kjv_def":"Jehdai."},
"H3057":{"lemma":"יְהֻדִיָּה","xlit":"Yᵉhudîyâh","pron":"yeh-hoo-dee-yaw'","derivation":"feminine of H3064 (יְהוּדִי);","strongs_def":"Jehudijah, a Jewess","kjv_def":"Jehudijah."},
"H3058":{"lemma":"יֵהוּא","xlit":"Yêhûwʼ","pron":"yay-hoo'","derivation":"from H3068 (יְהֹוָה) and H1931 (הוּא); Jehovah (is) He;","strongs_def":"Jehu, the name of five Israelites","kjv_def":"Jehu."},
"H3059":{"lemma":"יְהוֹאָחָז","xlit":"Yᵉhôwʼâchâz","pron":"yeh-ho-aw-khawz'","derivation":"from H3068 (יְהֹוָה) and H270 (אָחַז); Jehovah-seized;","strongs_def":"Jehoachaz, the name of three Israelites","kjv_def":"Jehoahaz. Compare H3099 (יוֹאָחָז)."},
"H3060":{"lemma":"יְהוֹאָשׁ","xlit":"Yᵉhôwʼâsh","pron":"yeh-ho-awsh'","derivation":"from H3068 (יְהֹוָה) and (perhaps) H784 (אֵשׁ); Jehovah-fired;","strongs_def":"Jehoash, the name of two Israelite kings","kjv_def":"Jehoash. Compare H3101 (יוֹאָשׁ)."},
"H3061":{"lemma":"יְהוּד","xlit":"Yᵉhûwd","pron":"yeh-hood'","derivation":"(Aramaic) contracted from a form corresponding to H3063 (יְהוּדָה);","strongs_def":"properly, Judah, hence, Judaea","kjv_def":"Jewry, Judah, Judea."},
"H3062":{"lemma":"יְהוּדָאִי","xlit":"Yᵉhûwdâʼîy","pron":"yeh-hoo-daw-ee'","derivation":"(Aramaic) patrial from H3061 (יְהוּד)lemma יְהיּד yod, corrected to יְהוּד;","strongs_def":"a Jehudaite (or Judaite), i.e. Jew","kjv_def":"Jew."},
"H3063":{"lemma":"יְהוּדָה","xlit":"Yᵉhûwdâh","pron":"yeh-hoo-daw'","derivation":"from H3034 (יָדָה); celebrated;","strongs_def":"Jehudah (or Judah), the name of five Israelites; also of the tribe descended from the first, and of its territory","kjv_def":"Judah."},
"H3064":{"lemma":"יְהוּדִי","xlit":"Yᵉhûwdîy","pron":"yeh-hoo-dee'","derivation":"patronymically from H3063 (יְהוּדָה);","strongs_def":"a Jehudite (i.e. Judaite or Jew), or descendant of Jehudah (i.e. Judah)","kjv_def":"Jew."},
"H3065":{"lemma":"יְהוּדִי","xlit":"Yᵉhûwdîy","pron":"yeh-hoo-dee'","derivation":"the same as H3064 (יְהוּדִי);","strongs_def":"Jehudi, an Israelite","kjv_def":"Jehudi."},
"H3066":{"lemma":"יְהוּדִית","xlit":"Yᵉhûwdîyth","pron":"yeh-hoo-deeth'","derivation":"feminine of H3064 (יְהוּדִי);","strongs_def":"the Jewish (used adverbially) language","kjv_def":"in the Jews' language."},
"H3067":{"lemma":"יְהוּדִית","xlit":"Yᵉhûwdîyth","pron":"yeh-ho-deeth'","derivation":"the same as H3066 (יְהוּדִית); Jewess;","strongs_def":"Jehudith, a Canaanitess","kjv_def":"Judith."},
"H3068":{"lemma":"יְהֹוָה","xlit":"Yᵉhôvâh","pron":"yeh-ho-vaw'","derivation":"from H1961 (הָיָה); (the) self-Existent or Eternal;","strongs_def":"Jehovah, Jewish national name of God","kjv_def":"Jehovah, the Lord. Compare H3050 (יָהּ), H3069 (יְהֹוִה)."},
"H3069":{"lemma":"יְהֹוִה","xlit":"Yᵉhôvih","pron":"yeh-ho-vee'","derivation":"a variation of H3068 (יְהֹוָה) (used after H136 (אֲדֹנָי), and pronounced by Jews as H430 (אֱלֹהִים), in order to prevent the repetition of the same sound, since they elsewhere pronounce H3068 (יְהֹוָה) as H136 (אֲדֹנָי))","strongs_def":"{YHWH}","kjv_def":"God."},
"H3070":{"lemma":"יְהֹוָה יִרְאֶה","xlit":"Yᵉhôvâh yirʼeh","pron":"yeh-ho-vaw' yir-eh'","derivation":"from H3068 (יְהֹוָה) and H7200 (רָאָה); Jehovah will see (to it);","strongs_def":"Jehovah-Jireh, a symbolical name for Mount Moriah","kjv_def":"Jehovah-jireh."},
"H3071":{"lemma":"יְהֹוָה נִסִּי","xlit":"Yᵉhôvâh niççîy","pron":"yeh-ho-vaw' nis-see'","derivation":"from H3068 (יְהֹוָה) and H5251 (נֵס) with the prononimal suffix; Jehovah (is) my banner;","strongs_def":"Jehovah-Nissi, a symbolical name of an altar in the Desert","kjv_def":"Jehovah-nissi."},
"H3072":{"lemma":"יְהֹוָה צִדְקֵנוּ","xlit":"Yᵉhôvâh tsidqênûw","pron":"ye-ho-vaw' tsid-kay'-noo","derivation":"from H3068 (יְהֹוָה) and H6664 (צֶדֶק) with pronominal suffix; Jehovah (is) our right;","strongs_def":"Jehovah-Tsidkenu, a symbolical epithet of the Messiah and of Jerusalem","kjv_def":"the Lord our righteousness."},
"H3073":{"lemma":"יְהֹוָה שָׁלוֹם","xlit":"Yᵉhôvâh shâlôwm","pron":"yeh-ho-vaw' shaw-lome'","derivation":"from H3068 (יְהֹוָה) and H7965 (שָׁלוֹם); Jehovah (is) peace;","strongs_def":"Jehovah-Shalom, a symbolical name of an altar in Palestine","kjv_def":"Jehovah-shalom."},
"H3074":{"lemma":"יְהֹוָה שָׁמָּה","xlit":"Yᵉhôvâh shâmmâh","pron":"yeh-ho-vaw' shawm'-maw","derivation":"from H3068 (יְהֹוָה) and H8033 (שָׁם) with directive enclitic; Jehovah (is) thither;","strongs_def":"Jehovah-Shammah, a symbolic title of Jerusalem","kjv_def":"Jehovahshammah."},
"H3075":{"lemma":"יְהוֹזָבָד","xlit":"Yᵉhôwzâbâd","pron":"yeh-ho-zaw-bawd'","derivation":"from H3068 (יְהֹוָה) and H2064 (זָבַד); Jehovah-endowed;","strongs_def":"Jehozabad, the name of three Israelites","kjv_def":"Jehozabad. Compare H3107 (יוֹזָבָד)."},
"H3076":{"lemma":"יְהוֹחָנָן","xlit":"Yᵉhôwchânân","pron":"yeh-ho-khaw-nawn'","derivation":"from H3068 (יְהֹוָה) and H2603 (חָנַן); Jehovah-favored;","strongs_def":"Jehochanan, the name of eight Israelites","kjv_def":"Jehohanan, Johanan. Compare H3110 (יוֹחָנָן)."},
"H3077":{"lemma":"יְהוֹיָדָע","xlit":"Yᵉhôwyâdâʻ","pron":"yeh-ho-yaw-daw'","derivation":"from H3068 (יְהֹוָה) and H3045 (יָדַע); Jehovah-known;","strongs_def":"Jehojada, the name of three Israelites","kjv_def":"Jehoiada. Compare H3111 (יוֹיָדָע)."},
"H3078":{"lemma":"יְהוֹיָכִין","xlit":"Yᵉhôwyâkîyn","pron":"yeh-ho-yaw-keen'","derivation":"from H3068 (יְהֹוָה) and H3559 (כּוּן); Jehovah will establish;","strongs_def":"Jehojakin, a Jewish king","kjv_def":"Jehoiachin. Compare H3112 (יוֹיָכִין)."},
"H3079":{"lemma":"יְהוֹיָקִים","xlit":"Yᵉhôwyâqîym","pron":"yeh-ho-yaw-keem'","derivation":"from H3068 (יְהֹוָה) abbreviated and H6965 (קוּם); Jehovah will raise;","strongs_def":"Jehojakim, a Jewish king","kjv_def":"Jehoiakim. Compare H3113 (יוֹיָקִים)."},
"H3080":{"lemma":"יְהוֹיָרִיב","xlit":"Yᵉhôwyârîyb","pron":"yeh-ho-yaw-reeb'","derivation":"from H3068 (יְהֹוָה) and H7378 (רִיב); Jehovah will contend;","strongs_def":"Jehojarib, the name of two Israelites","kjv_def":"Jehoiarib. Compare H3114 (יוֹיָרִיב)."},
"H3081":{"lemma":"יְהוּכַל","xlit":"Yᵉhûwkal","pron":"yeh-hoo-kal'","derivation":"from H3201 (יָכֹל); potent;","strongs_def":"Jehukal, an Israelite","kjv_def":"Jehucal. Compare H3116 (יוּכַל)."},
"H3082":{"lemma":"יְהוֹנָדָב","xlit":"Yᵉhôwnâdâb","pron":"yeh-ho-naw-dawb'","derivation":"from H3068 (יְהֹוָה) and H5068 (נָדַב); Jehovah-largessed;","strongs_def":"Jehonadab, the name of an Israelite and of an Arab","kjv_def":"Jehonadab, Jonadab. Compare H3122 (יוֹנָדָב)."},
"H3083":{"lemma":"יְהוֹנָתָן","xlit":"Yᵉhôwnâthân","pron":"yeh-ho-naw-thawn'","derivation":"from H3068 (יְהֹוָה) and H5414 (נָתַן); Jehovah-given;","strongs_def":"Jehonathan, the name of four Israelites","kjv_def":"Jonathan. Compare H3129 (יוֹנָתָן)."},
"H3084":{"lemma":"יְהוֹסֵף","xlit":"Yᵉhôwçêph","pron":"yeh-ho-safe'","derivation":"a fuller form of H3130 (יוֹסֵף);","strongs_def":"Jehoseph (i.e. Joseph), a son of Jacob","kjv_def":"Joseph."},
"H3085":{"lemma":"יְהוֹעַדָּה","xlit":"Yᵉhôwʻaddâh","pron":"yeh-ho-ad-daw'","derivation":"from H3068 (יְהֹוָה) and H5710 (עָדָה); Jehovah-adorned;","strongs_def":"Jehoaddah, an Israelite","kjv_def":"Jehoada."},
"H3086":{"lemma":"יְהוֹעַדִּין","xlit":"Yᵉhôwʻaddîyn","pron":"yeh-ho-ad-deen'","derivation":"or יְהוֹעַדָּן; from H3068 (יְהֹוָה) and H5727 (עָדַן); Jehovahpleased;","strongs_def":"Jehoaddin or Jehoaddan, an Israelitess","kjv_def":"Jehoaddan."},
"H3087":{"lemma":"יְהוֹצָדָק","xlit":"Yᵉhôwtsâdâq","pron":"yeh-ho-tsaw-dawk'","derivation":"from H3068 (יְהֹוָה) and H6663 (צָדַק); Jehovah-righted;","strongs_def":"Jehotsadak, an Israelite","kjv_def":"Jehozadek, Josedech. Compare H3136 (יוֹצָדָק)."},
"H3088":{"lemma":"יְהוֹרָם","xlit":"Yᵉhôwrâm","pron":"yeh-ho-rawm'","derivation":"from H3068 (יְהֹוָה) and H7311 (רוּם); Jehovah-raised;","strongs_def":"Jehoram, the name of a Syrian and of three Israelites","kjv_def":"Jehoram, Joram. Compare H3141 (יוֹרָם)."},
"H3089":{"lemma":"יְהוֹשֶׁבַע","xlit":"Yᵉhôwshebaʻ","pron":"yeh-ho-sheh'-bah","derivation":"from H3068 (יְהֹוָה) and H7650 (שָׁבַע); Jehovah-sworn;","strongs_def":"Jehosheba, an Israelitess","kjv_def":"Jehosheba. Compare H3090 (יְהוֹשַׁבְעַת)."},
"H3090":{"lemma":"יְהוֹשַׁבְעַת","xlit":"Yᵉhôwshabʻath","pron":"yeh-ho-shab-ath'","derivation":"a form of H3089 (יְהוֹשֶׁבַע);","strongs_def":"Jehoshabath, an Israelitess","kjv_def":"Jehoshabeath."},
"H3091":{"lemma":"יְהוֹשׁוּעַ","xlit":"Yᵉhôwshûwaʻ","pron":"yeh-ho-shoo'-ah","derivation":"or יְהוֹשֻׁעַ; from H3068 (יְהֹוָה) and H3467 (יָשַׁע); Jehovah-saved;","strongs_def":"Jehoshua (i.e. Joshua), the Jewish leader","kjv_def":"Jehoshua, Jehoshuah, Joshua. Compare H1954 (הוֹשֵׁעַ), H3442 (יֵשׁוּעַ)."},
"H3092":{"lemma":"יְהוֹשָׁפָט","xlit":"Yᵉhôwshâphâṭ","pron":"yeh-ho-shaw-fawt'","derivation":"from H3068 (יְהֹוָה) and H8199 (שָׁפַט); Jehovah-judged;","strongs_def":"Jehoshaphat, the name of six Israelites; also of a valley near Jerusalem","kjv_def":"Jehoshaphat. Compare H3146 (יוֹשָׁפָט)."},
"H3093":{"lemma":"יָהִיר","xlit":"yâhîyr","pron":"yaw-here'","derivation":"probably from the same as H2022 (הַר);","strongs_def":"elated; hence, arrogant","kjv_def":"haughty, proud."},
"H3094":{"lemma":"יְהַלֶּלְאֵל","xlit":"Yᵉhallelʼêl","pron":"yeh-hal-lel-ale'","derivation":"from H1984 (הָלַל) and H410 (אֵל); praising God;","strongs_def":"Jehallelel, the name of two Israelites","kjv_def":"Jehalellel, Jehalelel."},
"H3095":{"lemma":"יַהֲלֹם","xlit":"yahălôm","pron":"yah-hal-ome'","derivation":"from H1986 (הָלַם) (in the sense of hardness);","strongs_def":"a precious stone, probably onyx","kjv_def":"diamond."},
"H3096":{"lemma":"יַהַץ","xlit":"Yahats","pron":"yah'-hats","derivation":"or יַהְצָה; or (feminine) יַהְצָה; from an unused root meaning to stamp; perhaps threshing-floor;","strongs_def":"Jahats or Jahtsah, a place East of the Jordan","kjv_def":"Jahaz, Jahazah, Jahzah."},
"H3097":{"lemma":"יוֹאָב","xlit":"Yôwʼâb","pron":"yo-awb'","derivation":"from H3068 (יְהֹוָה) and H1 (אָב); Jehovah-fathered;","strongs_def":"Joab, the name of three Israelites","kjv_def":"Joab."},
"H3098":{"lemma":"יוֹאָח","xlit":"Yôwʼâch","pron":"yo-awkh'","derivation":"from H3068 (יְהֹוָה) and H251 (אָח); Jehovahbrothered;","strongs_def":"Joach, the name of four Israelites","kjv_def":"Joah."},
"H3099":{"lemma":"יוֹאָחָז","xlit":"Yôwʼâchâz","pron":"yo-aw-khawz'","derivation":"a form of H3059 (יְהוֹאָחָז);","strongs_def":"Joachaz, the name of two Israelites","kjv_def":"Jehoahaz, Joahaz."},
"H3100":{"lemma":"יוֹאֵל","xlit":"Yôwʼêl","pron":"yo-ale'","derivation":"from H3068 (יְהֹוָה) and H410 (אֵל); Jehovah (is his) God;","strongs_def":"Joel, the name of twelve Israelites","kjv_def":"Joel."},
"H3101":{"lemma":"יוֹאָשׁ","xlit":"Yôwʼâsh","pron":"yo-awsh'","derivation":"or יֹאָשׁ; (2 Chronicles 24:1), a form of H3060 (יְהוֹאָשׁ);","strongs_def":"Joash, the name of six Israelites","kjv_def":"Joash."},
"H3102":{"lemma":"יוֹב","xlit":"Yôwb","pron":"yobe","derivation":"perhaps a form of H3103 (יוֹבָב), but more probably by erroneous transcription for H3437 (יָשׁוּב);","strongs_def":"Job, an Israelite","kjv_def":"Job."},
"H3103":{"lemma":"יוֹבָב","xlit":"Yôwbâb","pron":"yo-bawb'","derivation":"from H2980 (יָבַב); howler;","strongs_def":"Jobab, the name of two Israelites and of three foreigners","kjv_def":"Jobab."},
"H3104":{"lemma":"יוֹבֵל","xlit":"yôwbêl","pron":"yo-bale'","derivation":"or יֹבֵל; apparently from H2986 (יָבַל);","strongs_def":"the blast of a horn (from its continuous sound); specifically, the signal of the silver trumpets; hence, the instrument itself and the festival thus introduced","kjv_def":"jubile, ram's horn, trumpet."},
"H3105":{"lemma":"יוּבַל","xlit":"yûwbal","pron":"yoo-bal'","derivation":"from H2986 (יָבַל);","strongs_def":"a stream","kjv_def":"river."},
"H3106":{"lemma":"יוּבַל","xlit":"Yûwbal","pron":"yoo-bawl'","derivation":"from H2986 (יָבַל); stream;","strongs_def":"Jubal, an antediluvian","kjv_def":"Jubal."},
"H3107":{"lemma":"יוֹזָבָד","xlit":"Yôwzâbâd","pron":"yo-zaw-bawd'","derivation":"a form of H3075 (יְהוֹזָבָד);","strongs_def":"Jozabad, the name of ten Israelites","kjv_def":"Josabad, Jozabad."},
"H3108":{"lemma":"יוֹזָכָר","xlit":"Yôwzâkâr","pron":"yo-zaw-kawr'","derivation":"from H3068 (יְהֹוָה) and H2142 (זָכַר); Jehovah-remembered;","strongs_def":"Jozacar, an Israelite","kjv_def":"Jozachar."},
"H3109":{"lemma":"יוֹחָא","xlit":"Yôwchâʼ","pron":"yo-khaw'","derivation":"probably from H3068 (יְהֹוָה) and a variation of H2421 (חָיָה); Jehovah-revived;","strongs_def":"Jocha, the name of two Israelites","kjv_def":"Joha."},
"H3110":{"lemma":"יוֹחָנָן","xlit":"Yôwchânân","pron":"yo-khaw-nawn'","derivation":"a form of H3076 (יְהוֹחָנָן);","strongs_def":"Jochanan, the name of nine Israelites","kjv_def":"Johanan."},
"H3111":{"lemma":"יוֹיָדָע","xlit":"Yôwyâdâʻ","pron":"yo-yaw-daw'","derivation":"a form of H3077 (יְהוֹיָדָע);","strongs_def":"Jojada, the name of two Israelites","kjv_def":"Jehoiada, Joiada."},
"H3112":{"lemma":"יוֹיָכִין","xlit":"Yôwyâkîyn","pron":"yo-yaw-keen'","derivation":"a form of H3078 (יְהוֹיָכִין);","strongs_def":"Jojakin, an Israelite king","kjv_def":"Jehoiachin."},
"H3113":{"lemma":"יוֹיָקִים","xlit":"Yôwyâqîym","pron":"yo-yaw-keem'","derivation":"a form of H3079 (יְהוֹיָקִים);","strongs_def":"Jojakim, an Israelite","kjv_def":"Joiakim. Compare H3137 (יוֹקִים)."},
"H3114":{"lemma":"יוֹיָרִיב","xlit":"Yôwyârîyb","pron":"yo-yaw-reeb'","derivation":"a form of H3080 (יְהוֹיָרִיב);","strongs_def":"Jojarib, the name of four Israelites","kjv_def":"Joiarib."},
"H3115":{"lemma":"יוֹכֶבֶד","xlit":"Yôwkebed","pron":"yo-keh'-bed","derivation":"from H3068 (יְהֹוָה) contracted and H3513 (כָּבַד); Jehovah-gloried;","strongs_def":"Jokebed, the mother of Moses","kjv_def":"Jochebed."},
"H3116":{"lemma":"יוּכַל","xlit":"Yûwkal","pron":"yoo-kal'","derivation":"a form of H3081 (יְהוּכַל);","strongs_def":"Jukal, an Israelite","kjv_def":"Jucal."},
"H3117":{"lemma":"יוֹם","xlit":"yôwm","pron":"yome","derivation":"from an unused root meaning to be hot;","strongs_def":"a day (as the warm hours), whether literal (from sunrise to sunset, or from one sunset to the next), or figurative (a space of time defined by an associated term), (often used adverb)","kjv_def":"age, [phrase] always, [phrase] chronicals, continually(-ance), daily, ((birth-), each, to) day, (now a, two) days (agone), [phrase] elder, [idiom] end, [phrase] evening, [phrase] (for) ever(-lasting, -more), [idiom] full, life, as (so) long as (... live), (even) now, [phrase] old, [phrase] outlived, [phrase] perpetually, presently, [phrase] remaineth, [idiom] required, season, [idiom] since, space, then, (process of) time, [phrase] as at other times, [phrase] in trouble, weather, (as) when, (a, the, within a) while (that), [idiom] whole ([phrase] age), (full) year(-ly), [phrase] younger."},
"H3118":{"lemma":"יוֹם","xlit":"yôwm","pron":"yome","derivation":"(Aramaic) corresponding to H3117 (יוֹם);","strongs_def":"a day","kjv_def":"day (by day), time."},
"H3119":{"lemma":"יוֹמָם","xlit":"yôwmâm","pron":"yo-mawm'","derivation":"from H3117 (יוֹם);","strongs_def":"daily","kjv_def":"daily, (by, in the) day(-time)."},
"H3120":{"lemma":"יָוָן","xlit":"Yâvân","pron":"yaw-vawn'","derivation":"probably from the same as H3196 (יַיִן); effervescing (i.e. hot and active);","strongs_def":"Javan, the name of a son of Joktan, and of the race (Ionians, i.e. Greeks) descended from him, with their territory; also of a place in Arabia","kjv_def":"Javan."},
"H3121":{"lemma":"יָוֵן","xlit":"yâvên","pron":"yaw-ven'","derivation":"from the same as H3196 (יַיִן);","strongs_def":"properly, dregs (as effervescing); hence, mud","kjv_def":"mire, miry."},
"H3122":{"lemma":"יוֹנָדָב","xlit":"Yôwnâdâb","pron":"yo-naw-dawb'","derivation":"a form of H3082 (יְהוֹנָדָב);","strongs_def":"Jonadab, the name of an Israelite and of a Rechabite","kjv_def":"Jonadab."},
"H3123":{"lemma":"יוֹנָה","xlit":"yôwnâh","pron":"yo-naw'","derivation":"probably from the same as H3196 (יַיִן);","strongs_def":"a dove (apparently from the warmth of their mating)","kjv_def":"dove, pigeon."},
"H3124":{"lemma":"יוֹנָה","xlit":"Yôwnâh","pron":"yo-naw'","derivation":"the same as H3123 (יוֹנָה);","strongs_def":"Jonah, an Israelite","kjv_def":"Jonah."},
"H3125":{"lemma":"יְוָנִי","xlit":"Yᵉvânîy","pron":"yev-aw-nee'","derivation":"patronymically from H3121 (יָוֵן);","strongs_def":"a Jevanite, or descendant of Javan","kjv_def":"Grecian."},
"H3126":{"lemma":"יוֹנֵק","xlit":"yôwnêq","pron":"yo-nake'","derivation":"active participle of H3243 (יָנַק);","strongs_def":"a sucker; hence, a twig (of a tree felled and sprouting)","kjv_def":"tender plant."},
"H3127":{"lemma":"יוֹנֶקֶת","xlit":"yôwneqeth","pron":"yo-neh'-keth","derivation":"feminine of H3126 (יוֹנֵק);","strongs_def":"a sprout","kjv_def":"(tender) branch, young twig."},
"H3128":{"lemma":"יוֹנַת אֵלֶם רְחֹקִים","xlit":"yôwnath ʼêlem rᵉchôqîym","pron":"yo-nath' ay'-lem rekh-o-keem'","derivation":"from H3123 (יוֹנָה) and H482 (אֵלֶם) and the plural of H7350 (רָחוֹק);","strongs_def":"dove of (the) silence (i.e. dumb Israel) of (i.e. among) distances (i.e. strangers); the title of a ditty (used for a name of its melody)","kjv_def":"Jonath-elem-rechokim."},
"H3129":{"lemma":"יוֹנָתָן","xlit":"Yôwnâthân","pron":"yo-naw-thawn'","derivation":"a form of H3083 (יְהוֹנָתָן);","strongs_def":"Jonathan, the name of ten Israelites","kjv_def":"Jonathan."},
"H3130":{"lemma":"יוֹסֵף","xlit":"Yôwçêph","pron":"yo-safe'","derivation":"future of H3254 (יָסַף); let him add (or perhaps simply active participle adding);","strongs_def":"Joseph, the name of seven Israelites","kjv_def":"Joseph. Compare H3084 (יְהוֹסֵף)."},
"H3131":{"lemma":"יוֹסִפְיָה","xlit":"Yôwçiphyâh","pron":"yo-sif-yaw'","derivation":"from active participle of H3254 (יָסַף) and H3050 (יָהּ); Jah (is) adding;","strongs_def":"Josiphjah, an Israelite","kjv_def":"Josiphiah."},
"H3132":{"lemma":"יוֹעֵאלָה","xlit":"Yôwʻêʼlâh","pron":"yo-ay-law'","derivation":"perhaps feminine active participle of H3276 (יַעַל); furthermore;","strongs_def":"Joelah, an Israelite","kjv_def":"Joelah."},
"H3133":{"lemma":"יוֹעֵד","xlit":"Yôwʻêd","pron":"yo-ade'","derivation":"apparently the active participle of H3259 (יָעַד); appointer;","strongs_def":"Joed, an Israelite","kjv_def":"Joed."},
"H3134":{"lemma":"יוֹעֶזֶר","xlit":"Yôwʻezer","pron":"yo-eh'-zer","derivation":"from H3068 (יְהֹוָה) and H5828 (עֵזֶר); Jehovah (is his) help;","strongs_def":"Joezer, an Israelite","kjv_def":"Joezer."},
"H3135":{"lemma":"יוֹעָשׁ","xlit":"Yôwʻâsh","pron":"yo-awsh'","derivation":"from H3068 (יְהֹוָה) and H5789 (עוּשׁ); Jehovahhastened;","strongs_def":"Joash, the name of two Israelites","kjv_def":"Joash."},
"H3136":{"lemma":"יוֹצָדָק","xlit":"Yôwtsâdâq","pron":"yo-tsaw-dawk'","derivation":"a form of H3087 (יְהוֹצָדָק);","strongs_def":"Jotsadak, an Israelite","kjv_def":"Jozadak."},
"H3137":{"lemma":"יוֹקִים","xlit":"Yôwqîym","pron":"yo-keem'","derivation":"a form of H3113 (יוֹיָקִים);","strongs_def":"Jokim, an Israelite","kjv_def":"Jokim."},
"H3138":{"lemma":"יוֹרֶה","xlit":"yôwreh","pron":"yo-reh'","derivation":"active participle of H3384 (יָרָה);","strongs_def":"sprinkling; hence, a sprinkling (or autumnal showers)","kjv_def":"first rain, former (rain)."},
"H3139":{"lemma":"יוֹרָה","xlit":"Yôwrâh","pron":"yo-raw'","derivation":"from H3384 (יָרָה); rainy;","strongs_def":"Jorah, an Israelite","kjv_def":"Jorah."},
"H3140":{"lemma":"יוֹרַי","xlit":"Yôwray","pron":"yo-rah'-ee","derivation":"from H3384 (יָרָה); rainy;","strongs_def":"Jorai, an Israelite","kjv_def":"Jorai."},
"H3141":{"lemma":"יוֹרָם","xlit":"Yôwrâm","pron":"yo-rawm'","derivation":"a form of H3088 (יְהוֹרָם);","strongs_def":"Joram, the name of three Israelites and one Syrian","kjv_def":"Joram."},
"H3142":{"lemma":"יוּשַׁב חֶסֶד","xlit":"Yûwshab Cheçed","pron":"yoo-shab' kheh'-sed","derivation":"from H7725 (שׁוּב) and H2617 (חֵסֵד); kindness will be returned;","strongs_def":"Jushab-Chesed, an Israelite","kjv_def":"Jushab-hesed."},
"H3143":{"lemma":"יוֹשִׁבְיָה","xlit":"Yôwshibyâh","pron":"yo-shi-yaw'","derivation":"from H3427 (יָשַׁב) and H3050 (יָהּ); Jehovah will cause to dwell;","strongs_def":"Josibjah, an Israelite","kjv_def":"Josibiah."},
"H3144":{"lemma":"יוֹשָׁה","xlit":"Yôwshâh","pron":"yo-shaw'","derivation":"probably a form of H3145 (יוֹשַׁוְיָה);","strongs_def":"Joshah, an Israelite","kjv_def":"Joshah."},
"H3145":{"lemma":"יוֹשַׁוְיָה","xlit":"Yôwshavyâh","pron":"yo-shav-yaw'","derivation":"from H3068 (יְהֹוָה) and H7737 (שָׁוָה); Jehovah-set;","strongs_def":"Joshavjah, an Israelite","kjv_def":"Joshaviah. Compare H3144 (יוֹשָׁה)."},
"H3146":{"lemma":"יוֹשָׁפָט","xlit":"Yôwshâphâṭ","pron":"yo-shaw-fawt'","derivation":"a form of H3092 (יְהוֹשָׁפָט);","strongs_def":"Joshaphat, an Israelite","kjv_def":"Joshaphat."},
"H3147":{"lemma":"יוֹתָם","xlit":"Yôwthâm","pron":"yo-thawm'","derivation":"from H3068 (יְהֹוָה) and H8535 (תָּם); Jehovah (is) perfect;","strongs_def":"Jotham, the name of three Israelites","kjv_def":"Jotham."},
"H3148":{"lemma":"יוֹתֵר","xlit":"yôwthêr","pron":"yo-thare'","derivation":"active participle of H3498 (יָתַר);","strongs_def":"properly, redundant; hence, over and above, as adjective, noun, adverb or conjunction","kjv_def":"better, more(-over), over, profit."},
"H3149":{"lemma":"יְזַוְאֵל","xlit":"Yᵉzavʼêl","pron":"yez-av-ale'","derivation":"from an unused root (meaning to sprinkle) and H410 (אֵל); sprinkled of God;","strongs_def":"Jezavel, an Israelite","kjv_def":"Jeziel (from the margin)."},
"H3150":{"lemma":"יִזִּיָּה","xlit":"Yizzîyâh","pron":"yiz-zee-yaw'","derivation":"from the same as the first part of H3149 (יְזַוְאֵל) and H3050 (יָהּ); sprinkled of Jah;","strongs_def":"Jizzijah, an Israelite","kjv_def":"Jeziah."},
"H3151":{"lemma":"יָזִיז","xlit":"Yâzîyz","pron":"yaw-zeez'","derivation":"from the same as H2123 (זִיז); he will make prominent;","strongs_def":"Jaziz, an Israelite","kjv_def":"Jaziz."},
"H3152":{"lemma":"יִזְלִיאָה","xlit":"Yizlîyʼâh","pron":"yiz-lee-aw'","derivation":"perhaps from an unused root (meaning to draw up); he will draw out;","strongs_def":"Jizliah, an Israelite","kjv_def":"Jezliah."},
"H3153":{"lemma":"יְזַנְיָה","xlit":"Yᵉzanyâh","pron":"yez-an-yaw'","derivation":"or יְזַנְיָהוּ; probably for H2970 (יַאֲזַנְיָה);","strongs_def":"Jezanjah, an Israelite","kjv_def":"Jezaniah."},
"H3154":{"lemma":"יֶזַע","xlit":"yezaʻ","pron":"yeh'-zah","derivation":"from an unused root mean to ooze;","strongs_def":"sweat, i.e. (by implication) a sweating dress","kjv_def":"any thing that causeth sweat."},
"H3155":{"lemma":"יִזְרָח","xlit":"Yizrâch","pron":"yiz-rawkh'","derivation":"a variation for H250 (אֶזְרָחִי);","strongs_def":"a Jizrach (i.e. Ezrachite or Zarchite) or descendant of Zerach","kjv_def":"Izrahite."},
"H3156":{"lemma":"יִזְרַחְיָה","xlit":"Yizrachyâh","pron":"yiz-rakh-yaw'","derivation":"from H2224 (זָרַח) and H3050 (יָהּ); Jah will shine;","strongs_def":"Jizrachjah, the name of two Israelites","kjv_def":"Izrahiah, Jezrahiah."},
"H3157":{"lemma":"יִזְרְעֵאל","xlit":"Yizrᵉʻêʼl","pron":"yiz-reh-ale'","derivation":"from H2232 (זָרַע) and H410 (אֵל); God will sow;","strongs_def":"Jizreel, the name of two places in Palestine and of two Israelites","kjv_def":"Jezreel."},
"H3158":{"lemma":"יִזְרְעֵאלִי","xlit":"Yizrᵉʻêʼlîy","pron":"yiz-reh-ay-lee'","derivation":"patronymically from H3157 (יִזְרְעֵאל);","strongs_def":"a Jizreelite or native of Jizreel","kjv_def":"Jezreelite."},
"H3159":{"lemma":"יִזְרְעֵאלִית","xlit":"Yizrᵉʻêʼlîyth","pron":"yiz-reh-ay-leeth'","derivation":"feminine of H3158 (יִזְרְעֵאלִי);","strongs_def":"a Jezreelitess","kjv_def":"Jezreelitess."},
"H3160":{"lemma":"יְחֻבָּה","xlit":"Yᵉchubbâh","pron":"yekh-oob-baw'","derivation":"from H2247 (חָבָה); hidden;","strongs_def":"Jechubbah, an Israelite","kjv_def":"Jehubbah."},
"H3161":{"lemma":"יָחַד","xlit":"yâchad","pron":"yaw-khad'","derivation":"a primitive root;","strongs_def":"to be (or become) one","kjv_def":"join, unite."},
"H3162":{"lemma":"יַחַד","xlit":"yachad","pron":"yakh'-ad","derivation":"from H3161 (יָחַד);","strongs_def":"properly, a unit, i.e. (adverb) unitedly","kjv_def":"alike, at all (once), both, likewise, only, (al-) together, withal."},
"H3163":{"lemma":"יַחְדוֹ","xlit":"Yachdôw","pron":"yakh-doe'","derivation":"from H3162 (יַחַד) with pronominal suffix; his unity, i.e. (adverb) together;","strongs_def":"Jachdo, an Israelite","kjv_def":"Jahdo."},
"H3164":{"lemma":"יַחְדִּיאֵל","xlit":"Yachdîyʼêl","pron":"yakh-dee-ale'","derivation":"from H3162 (יַחַד) and H410 (אֵל); unity of God;","strongs_def":"Jachdiel, an Israelite","kjv_def":"Jahdiel."},
"H3165":{"lemma":"יֶחְדִּיָּהוּ","xlit":"Yechdîyâhûw","pron":"yekh-dee-yaw'-hoo","derivation":"from H3162 (יַחַד) and H3050 (יָהּ); unity of Jah;","strongs_def":"Jechdijah, the name of two Israelites","kjv_def":"Jehdeiah."},
"H3166":{"lemma":"יַחֲזִיאֵל","xlit":"Yachăzîyʼêl","pron":"yakh-az-ee-ale'","derivation":"from H2372 (חָזָה) and H410 (אֵל); beheld of God;","strongs_def":"Jachaziel, the name of five Israelites","kjv_def":"Jahaziel, Jahziel."},
"H3167":{"lemma":"יַחְזְיָה","xlit":"Yachzᵉyâh","pron":"yakh-zeh-yaw'","derivation":"from H2372 (חָזָה) and H3050 (יָהּ); Jah will behold;","strongs_def":"Jachzejah, an Israelite","kjv_def":"Jahaziah."},
"H3168":{"lemma":"יְחֶזְקֵאל","xlit":"Yᵉchezqêʼl","pron":"yekh-ez-kale'","derivation":"from H2388 (חָזַק) and H410 (אֵל); God will strengthen;","strongs_def":"Jechezkel, the name of two Israelites","kjv_def":"Ezekiel, Jehezekel."},
"H3169":{"lemma":"יְחִזְקִיָּה","xlit":"Yᵉchizqîyâh","pron":"yekh-iz-kee-yaw'","derivation":"or יְחִזְקִיָּהוּ; from H3388 (יְרוּשָׁא) and H3050 (יָהּ); strengthened of Jah;","strongs_def":"Jechizkijah, the name of five Israelites","kjv_def":"Hezekiah, Jehizkiah. Compare H2396 (חִזְקִיָּה)."},
"H3170":{"lemma":"יַחְזֵרָה","xlit":"Yachzêrâh","pron":"yakh-zay-raw'","derivation":"from the same as H2386 (חֲזִיר); perhaps protection;","strongs_def":"Jachzerah, an Israelite","kjv_def":"Jahzerah."},
"H3171":{"lemma":"יְחִיאֵל","xlit":"Yᵉchîyʼêl","pron":"yekh-ee-ale'","derivation":"or (2 Chronicles 29:14) יְחַוְאֵל; from H2421 (חָיָה) and H410 (אֵל); God will live;","strongs_def":"Jechiel (or Jechavel), the name of eight Israelites","kjv_def":"Jehiel."},
"H3172":{"lemma":"יְחִיאֵלִי","xlit":"Yᵉchîyʼêlîy","pron":"yekh-ee-ay-lee'","derivation":"patronymically from H3171 (יְחִיאֵל);","strongs_def":"a Jechielite or descendant of Jechiel","kjv_def":"Jehieli."},
"H3173":{"lemma":"יָחִיד","xlit":"yâchîyd","pron":"yaw-kheed'","derivation":"from H3161 (יָחַד);","strongs_def":"properly, united, i.e. sole; by implication, beloved; also lonely; (feminine) the life (as not to be replaced)","kjv_def":"darling, desolate, only (child, son), solitary."},
"H3174":{"lemma":"יְחִיָּה","xlit":"Yᵉchîyâh","pron":"yekh-ee-yaw'","derivation":"from H2421 (חָיָה) and H3050 (יָהּ); Jah will live;","strongs_def":"Jechijah, an Israelite","kjv_def":"Jehiah."},
"H3175":{"lemma":"יָחִיל","xlit":"yâchîyl","pron":"yaw-kheel'","derivation":"from H3176 (יָחַל);","strongs_def":"expectant","kjv_def":"should hope."},
"H3176":{"lemma":"יָחַל","xlit":"yâchal","pron":"yaw-chal'","derivation":"a primitive root;","strongs_def":"to wait; by implication, to be patient, hope","kjv_def":"(cause to, have, make to) hope, be pained, stay, tarry, trust, wait."},
"H3177":{"lemma":"יַחְלְאֵל","xlit":"Yachlᵉʼêl","pron":"yakh-leh-ale'","derivation":"from H3176 (יָחַל) and H410 (אֵל); expectant of God;","strongs_def":"Jachleel, an Israelite","kjv_def":"Jahleel."},
"H3178":{"lemma":"יַחְלְאֵלִי","xlit":"Yachlᵉʼêlîy","pron":"yakh-leh-ay-lee'","derivation":"patronymically from H3177 (יַחְלְאֵל);","strongs_def":"a Jachleelite or descendant of Jachleel","kjv_def":"Jahleelites."},
"H3179":{"lemma":"יָחַם","xlit":"yâcham","pron":"yaw-kham'","derivation":"a primitive root;","strongs_def":"probably to be hot; figuratively, to conceive","kjv_def":"get heat, be hot, conceive, be warm."},
"H3180":{"lemma":"יַחְמוּר","xlit":"yachmûwr","pron":"yakh-moor'","derivation":"from H2560 (חָמַר);","strongs_def":"a kind of deer","kjv_def":"fallow deer."},
"H3181":{"lemma":"יַחְמַי","xlit":"Yachmay","pron":"yakh-mah'-ee","derivation":"probably from H3179 (יָחַם); hot;","strongs_def":"Jachmai, an Israelite","kjv_def":"Jahmai."},
"H3182":{"lemma":"יָחֵף","xlit":"yâchêph","pron":"yaw-khafe'","derivation":"from an unused root meaning to take off the shoes;","strongs_def":"unsandalled","kjv_def":"barefoot, being unshod."},
"H3183":{"lemma":"יַחְצְאֵל","xlit":"Yachtsᵉʼêl","pron":"yakh-tseh-ale'","derivation":"from H2673 (חָצָה)lemma חָעָה ayin, corrected to חָצָה and H410 (אֵל); God will allot;","strongs_def":"Jachtseel, an Israelite","kjv_def":"Jahzeel. Compare H3185 (יַחְצִיאֵל)."},
"H3184":{"lemma":"יַחְצְאֵלִי","xlit":"Yachtsᵉʼêlîy","pron":"yakh-tseh-ay-lee'","derivation":"patronymically from H3183 (יַחְצְאֵל);","strongs_def":"a Jachtseelite (collectively) or descendants of Jachtseel","kjv_def":"Jahzeelites."},
"H3185":{"lemma":"יַחְצִיאֵל","xlit":"Yachtsîyʼêl","pron":"yakh-tsee-ale'","derivation":"from H2673 (חָצָה)lemma חָעָה ayin, corrected to חָצָה and H410 (אֵל); allotted of God;","strongs_def":"Jachtsiel, an Israelite","kjv_def":"Jahziel. Compare H3183 (יַחְצְאֵל)."},
"H3186":{"lemma":"יָחַר","xlit":"yâchar","pron":"yaw-khar'","derivation":"a primitive root;","strongs_def":"to delay","kjv_def":"tarry longer."},
"H3187":{"lemma":"יָחַשׂ","xlit":"yâchas","pron":"yaw-khas'","derivation":"a primitive root; to sprout; used only as denominative from H3188 (יַחַשׂ);","strongs_def":"to enroll by pedigree","kjv_def":"(number after, number throughout the) genealogy (to be reckoned), be reckoned by genealogies."},
"H3188":{"lemma":"יַחַשׂ","xlit":"yachas","pron":"yakh'-as","derivation":"from H3187 (יָחַשׂ);","strongs_def":"a pedigree or family list (as growing spontaneously)","kjv_def":"genealogy."},
"H3189":{"lemma":"יַחַת","xlit":"Yachath","pron":"yakh'-ath","derivation":"from H3161 (יָחַד); unity;","strongs_def":"Jachath, the name of four Israelites","kjv_def":"Jahath."},
"H3190":{"lemma":"יָטַב","xlit":"yâṭab","pron":"yaw-tab'","derivation":"a primitive root;","strongs_def":"to be (causative) make well, literally (sound, beautiful) or figuratively (happy, successful, right)","kjv_def":"be accepted, amend, use aright, benefit, be (make) better, seem best, make cheerful, be comely, [phrase] be content, diligent(-ly), dress, earnestly, find favour, give, be glad, do (be, make) good(-ness), be (make) merry, please ([phrase] well), shew more (kindness), skilfully, [idiom] very small, surely, make sweet, thoroughly, tire, trim, very, be (can, deal, entreat, go, have) well (said, seen)."},
"H3191":{"lemma":"יְטַב","xlit":"yᵉṭab","pron":"yet-ab'","derivation":"(Aramaic) corresponding to H3190 (יָטַב)","strongs_def":"{to be (causative) make well, literally (sound, beautiful) or figuratively (happy, successful, right)}","kjv_def":"seem good."},
"H3192":{"lemma":"יׇטְבָה","xlit":"Yoṭbâh","pron":"yot-baw'","derivation":"from H3190 (יָטַב); pleasantness;","strongs_def":"Jotbah, a place in Palestine","kjv_def":"Jotbah."},
"H3193":{"lemma":"יׇטְבָתָה","xlit":"Yoṭbâthâh","pron":"yot-baw'-thaw","derivation":"from H3192 (יׇטְבָה);","strongs_def":"Jotbathah, a place in the Desert","kjv_def":"Jotbath, Jotbathah."},
"H3194":{"lemma":"יֻטָּה","xlit":"Yuṭṭâh","pron":"yoo-taw'","derivation":"or יוּטָה; from H5186 (נָטָה); extended;","strongs_def":"Juttah (or Jutah), a place in Palestine","kjv_def":"Juttah."},
"H3195":{"lemma":"יְטוּר","xlit":"Yᵉṭûwr","pron":"yet-oor'","derivation":"probably from the same as H2905 (טוּר); encircled (i.e. inclosed);","strongs_def":"Jetur, a son of Ishmael","kjv_def":"Jetur."},
"H3196":{"lemma":"יַיִן","xlit":"yayin","pron":"yah'-yin","derivation":"from an unused root meaning to effervesce;","strongs_def":"wine (as fermented); by implication, intoxication","kjv_def":"banqueting, wine, wine(-bibber)."},
"H3197":{"lemma":"יַךְ","xlit":"yak","pron":"yak","derivation":"by erroneous transcription for H3027 (יָד);","strongs_def":"a hand or side","kjv_def":"(way-) side."},
"H3198":{"lemma":"יָכַח","xlit":"yâkach","pron":"yaw-kahh'","derivation":"a primitive root;","strongs_def":"to be right (i.e. correct); reciprocal, to argue; causatively, to decide, justify or convict","kjv_def":"appoint, argue, chasten, convince, correct(-ion), daysman, dispute, judge, maintain, plead, reason (together), rebuke, reprove(-r), surely, in any wise."},
"H3199":{"lemma":"יָכִין","xlit":"Yâkîyn","pron":"yaw-keen'","derivation":"from H3559 (כּוּן); he (or it) will establish;","strongs_def":"Jakin, the name of three Israelites and of a temple pillar","kjv_def":"Jachin."},
"H3200":{"lemma":"יָכִינִי","xlit":"Yâkîynîy","pron":"yaw-kee-nee'","derivation":"patronymically from H3199 (יָכִין)lemma יִכִין first vowel, corrected to יָכִין;","strongs_def":"a Jakinite (collectively) or descendants of Jakin","kjv_def":"Jachinites."},
"H3201":{"lemma":"יָכֹל","xlit":"yâkôl","pron":"yaw-kole'","derivation":"or (fuller) יָכוֹל; a primitive root;","strongs_def":"to be able, literally (can, could) or morally (may, might)","kjv_def":"be able, any at all (ways), attain, can (away with, (-not)), could, endure, might, overcome, have power, prevail, still, suffer."},
"H3202":{"lemma":"יְכֵל","xlit":"yᵉkêl","pron":"yek-ale'","derivation":"(Aramaic) or יְכִיל; (Aramaic), to H3201 (יָכֹל)","strongs_def":"{to be able, literally (can, could) or morally (may, might)}","kjv_def":"be able, can, couldest, prevail."},
"H3203":{"lemma":"יְכׇלְיָה","xlit":"Yᵉkolyâh","pron":"yek-ol-yaw'","derivation":"and יְכׇלְיָהוּ; or (2 Chronicles 26:3) יְכִילְיָה; from H3201 (יָכֹל) and H3050 (יָהּ); Jah will enable;","strongs_def":"Jekoljah or Jekiljah, an Israelitess","kjv_def":"Jecholiah, Jecoliah."},
"H3204":{"lemma":"יְכׇנְיָה","xlit":"Yᵉkonyâh","pron":"yek-on-yaw'","derivation":"and יְכׇנְיָהוּ; or (Jeremiah 27:20) יְכוֹנְיָה; from H3559 (כּוּן) and H3050 (יָהּ); Jah will establish;","strongs_def":"Jekonjah, a Jewish king","kjv_def":"Jeconiah. Compare H3659 (כׇּנְיָהוּ)."},
"H3205":{"lemma":"יָלַד","xlit":"yâlad","pron":"yaw-lad'","derivation":"a primitive root;","strongs_def":"to bear young; causatively, to beget; medically, to act as midwife; specifically, to show lineage","kjv_def":"bear, beget, birth(-day), born, (make to) bring forth (children, young), bring up, calve, child, come, be delivered (of a child), time of delivery, gender, hatch, labour, (do the office of a) midwife, declare pedigrees, be the son of, (woman in, woman that) travail(-eth, -ing woman)."},
"H3206":{"lemma":"יֶלֶד","xlit":"yeled","pron":"yeh'-led","derivation":"from H3205 (יָלַד);","strongs_def":"something born, i.e. a lad or offspring","kjv_def":"boy, child, fruit, son, young man (one)."},
"H3207":{"lemma":"יַלְדָּה","xlit":"yaldâh","pron":"yal-daw'","derivation":"feminine of H3206 (יֶלֶד);","strongs_def":"a lass","kjv_def":"damsel, girl."},
"H3208":{"lemma":"יַלְדוּת","xlit":"yaldûwth","pron":"yal-dooth'","derivation":"abstractly from H3206 (יֶלֶד);","strongs_def":"boyhood (or girlhood)","kjv_def":"childhood, youth."},
"H3209":{"lemma":"יִלּוֹד","xlit":"yillôwd","pron":"yil-lode'","derivation":"passive from H3205 (יָלַד);","strongs_def":"born","kjv_def":"born."},
"H3210":{"lemma":"יָלוֹן","xlit":"Yâlôwn","pron":"yaw-lone'","derivation":"from H3885 (לוּן); lodging;","strongs_def":"Jalon, an Israelite","kjv_def":"Jalon."},
"H3211":{"lemma":"יָלִיד","xlit":"yâlîyd","pron":"yaw-leed'","derivation":"from H3205 (יָלַד);","strongs_def":"born","kjv_def":"(home-) born, child, son."},
"H3212":{"lemma":"יָלַךְ","xlit":"yâlak","pron":"yaw-lak'","derivation":"a primitive root (compare H1980 (הָלַךְ));","strongs_def":"to walk (literally or figuratively); causatively, to carry (in various senses)","kjv_def":"[idiom] again, away, bear, bring, carry (away), come (away), depart, flow, [phrase] follow(-ing), get (away, hence, him), (cause to, made) go (away, -ing, -ne, one's way, out), grow, lead (forth), let down, march, prosper, [phrase] pursue, cause to run, spread, take away (-journey), vanish, (cause to) walk(-ing), wax, [idiom] be weak."},
"H3213":{"lemma":"יָלַל","xlit":"yâlal","pron":"yaw-lal'","derivation":"a primitive root;","strongs_def":"to howl (with a wailing tone) or yell (with a boisterous one)","kjv_def":"(make to) howl, be howling."},
"H3214":{"lemma":"יְלֵל","xlit":"yᵉlêl","pron":"yel-ale'","derivation":"from H3213 (יָלַל);","strongs_def":"a howl","kjv_def":"howling."},
"H3215":{"lemma":"יְלָלָה","xlit":"yᵉlâlâh","pron":"yel-aw-law'","derivation":"feminine of H3214 (יְלֵל)","strongs_def":"{a howl}","kjv_def":"a howling."},
"H3216":{"lemma":"יָלַע","xlit":"yâlaʻ","pron":"yaw-lah'","derivation":"a primitive root;","strongs_def":"to blurt or utter inconsiderately","kjv_def":"devour."},
"H3217":{"lemma":"יַלֶּפֶת","xlit":"yallepheth","pron":"yal-leh'-feth","derivation":"from an unused root apparently meaning to stick or scrape;","strongs_def":"scurf or tetter","kjv_def":"scabbed."},
"H3218":{"lemma":"יֶלֶק","xlit":"yeleq","pron":"yeh'-lek","derivation":"from an unused root meaning to lick up;","strongs_def":"a devourer; specifically, the young locust","kjv_def":"cankerworm, caterpillar."},
"H3219":{"lemma":"יַלְקוּט","xlit":"yalqûwṭ","pron":"yal-koot'","derivation":"from H3950 (לָקַט);","strongs_def":"a travelling pouch (as if for gleanings)","kjv_def":"scrip."},
"H3220":{"lemma":"יָם","xlit":"yâm","pron":"yawm","derivation":"from an unused root meaning to roar;","strongs_def":"a sea (as breaking in noisy surf) or large body of water; specifically (with the article), the Mediterranean Sea; sometimes a large river, or an artifical basin; locally, the west, or (rarely) the south","kjv_def":"sea ([idiom] -faring man, (-shore)), south, west (-ern, side, -ward)."},
"H3221":{"lemma":"יָם","xlit":"yâm","pron":"yawm","derivation":"(Aramaic) corresponding to H3220 (יָם)","strongs_def":"{a sea (as breaking in noisy surf) or large body of water; specifically (with the article), the Mediterranean Sea; sometimes a large river, or an artifical basin; locally, the west, or (rarely) the south}","kjv_def":"sea."},
"H3222":{"lemma":"יֵם","xlit":"yêm","pron":"yame","derivation":"from the same as H3117 (יוֹם);","strongs_def":"a warm spring","kjv_def":"mule."},
"H3223":{"lemma":"יְמוּאֵל","xlit":"Yᵉmûwʼêl","pron":"yem-oo-ale'","derivation":"from H3117 (יוֹם) and H410 (אֵל); day of God;","strongs_def":"Jemuel, an Israelite","kjv_def":"Jemuel."},
"H3224":{"lemma":"יְמִימָה","xlit":"Yᵉmîymâh","pron":"yem-ee-maw'","derivation":"perhaps from the same as H3117 (יוֹם); properly, warm, i.e. affectionate; hence, dove (compare H3123 (יוֹנָה));","strongs_def":"Jemimah, one of Job's daughters","kjv_def":"Jemimah."},
"H3225":{"lemma":"יָמִין","xlit":"yâmîyn","pron":"yaw-meen'","derivation":"from H3231 (יָמַן);","strongs_def":"the right hand or side (leg, eye) of a person or other object (as the stronger and more dexterous); locally, the south","kjv_def":"[phrase] left-handed, right (hand, side), south."},
"H3226":{"lemma":"יָמִין","xlit":"Yâmîyn","pron":"yaw-meen'","derivation":"the same as H3225 (יָמִין);","strongs_def":"Jamin, the name of three Israelites","kjv_def":"Jamin. See also H1144 (בִּנְיָמִין)."},
"H3227":{"lemma":"יְמִינִי","xlit":"yᵉmîynîy","pron":"yem-ee-nee'","derivation":"for H3225 (יָמִין);","strongs_def":"right","kjv_def":"(on the) right (hand)."},
"H3228":{"lemma":"יְמִינִי","xlit":"Yᵉmîynîy","pron":"yem-ee-nee'","derivation":"patronymically from H3226 (יָמִין);","strongs_def":"a Jeminite (collectively) or descendants of Jamin","kjv_def":"Jaminites. See also H1145 (בֶּן־יְמִינִי)."},
"H3229":{"lemma":"יִמְלָא","xlit":"Yimlâʼ","pron":"yeem-law'","derivation":"or יִמְלָה; from H4390 (מָלֵא); full;","strongs_def":"Jimla or Jimlah, an Israelite","kjv_def":"Imla, Imlah."},
"H3230":{"lemma":"יַמְלֵךְ","xlit":"Yamlêk","pron":"yam-lake'","derivation":"from H4427 (מָלַךְ); he will make king;","strongs_def":"Jamlek, an Israelite","kjv_def":"Jamlech."},
"H3231":{"lemma":"יָמַן","xlit":"yâman","pron":"yaw-man'","derivation":"a primitive root; to be (physically) right (i.e. firm); but used only as denominative from H3225 (יָמִין) and transitive,","strongs_def":"to be right-handed or take the right-hand side","kjv_def":"go (turn) to (on, use) the right hand."},
"H3232":{"lemma":"יִמְנָה","xlit":"Yimnâh","pron":"yim-naw'","derivation":"from H3231 (יָמַן); prosperity (as betokened by the right hand);","strongs_def":"Jimnah, the name of two Israelites; also (with the article) of the posterity of one of them","kjv_def":"Imna, Imnah, Jimnah, Jimnites."},
"H3233":{"lemma":"יְמָנִי","xlit":"yᵉmânîy","pron":"yem-aw-nee'","derivation":"from H3231 (יָמַן);","strongs_def":"right (i.e. at the right hand)","kjv_def":"(on the) right (hand)."},
"H3234":{"lemma":"יִמְנָע","xlit":"Yimnâʻ","pron":"yim-naw'","derivation":"from H4513 (מָנַע); he will restrain;","strongs_def":"Jimna, an Israelite","kjv_def":"Imna."},
"H3235":{"lemma":"יָמַר","xlit":"yâmar","pron":"yaw-mar'","derivation":"a primitive root;","strongs_def":"to exchange; by implication, to change places","kjv_def":"boast selves, change."},
"H3236":{"lemma":"יִמְרָה","xlit":"Yimrâh","pron":"yim-raw'","derivation":"probably from H3235 (יָמַר); interchange;","strongs_def":"Jimrah, an Israelite","kjv_def":"Imrah."},
"H3237":{"lemma":"יָמַשׁ","xlit":"yâmash","pron":"yaw-mash'","derivation":"a primitive root;","strongs_def":"to touch","kjv_def":"feel."},
"H3238":{"lemma":"יָנָה","xlit":"yânâh","pron":"yaw-naw'","derivation":"a primitive root;","strongs_def":"to rage or be violent; by implication, to suppress, to maltreat","kjv_def":"destroy, (thrust out by) oppress(-ing, -ion, -or), proud, vex, do violence."},
"H3239":{"lemma":"יָנוֹחַ","xlit":"Yânôwach","pron":"yaw-no'-akh or (with enclitic) Yanowchah yaw-no'-khaw","derivation":"from H3240 (יָנַח); quiet;","strongs_def":"Janoach or Janochah, a place in Palestine","kjv_def":"Janoah, Janohah."},
"H3240":{"lemma":"יָנַח","xlit":"yânach","pron":"yaw-nakh'","derivation":"a primitive root;","strongs_def":"to deposit; by implication, to allow to stay","kjv_def":"bestow, cast down, lay (down, up), leave (off), let alone (remain), pacify, place, put, set (down), suffer, withdraw, withhold. (The Hiphil forms with the dagesh are here referred to, in accordance with the older grammarians; but if any distinction of the kind is to be made, these should rather be referred to H5117 (נוּחַ), and the others here.)"},
"H3241":{"lemma":"יָנִים","xlit":"Yânîym","pron":"yaw-neem'","derivation":"from H5123 (נוּם); asleep;","strongs_def":"Janim, a place in Palestine","kjv_def":"Janum (from the margin)."},
"H3242":{"lemma":"יְנִיקָה","xlit":"yᵉnîyqâh","pron":"yen-ee-kaw'","derivation":"from H3243 (יָנַק);","strongs_def":"a sucker or sapling","kjv_def":"young twig."},
"H3243":{"lemma":"יָנַק","xlit":"yânaq","pron":"yaw-nak'","derivation":"a primitive root;","strongs_def":"to suck; causatively, to give milk","kjv_def":"milch, nurse(-ing mother), (give, make to) suck(-ing child, -ling)."},
"H3244":{"lemma":"יַנְשׁוּף","xlit":"yanshûwph","pron":"yan-shoof'","derivation":"or יַנְשׁוֹף; apparently from H5398 (נָשַׁף); (compare H5399 (נֶשֶׁף))","strongs_def":"an unclean (acquatic) bird; probably the heron (perhaps from its blowing cry, or because the nightheron is meant ))","kjv_def":"(great) owl.s"},
"H3245":{"lemma":"יָסַד","xlit":"yâçad","pron":"yaw-sad'","derivation":"a primitive root;","strongs_def":"to set (literally or figuratively); intensively, to found; reflexively, to sit down together, i.e. settle, consult","kjv_def":"appoint, take counsel, establish, (lay the, lay for a) found(-ation), instruct, lay, ordain, set, [idiom] sure."},
"H3246":{"lemma":"יְסֻד","xlit":"yᵉçud","pron":"yes-ood'","derivation":"from H3245 (יָסַד);","strongs_def":"a foundation (figuratively, i.e. beginning)","kjv_def":"[idiom] began."},
"H3247":{"lemma":"יְסוֹד","xlit":"yᵉçôwd","pron":"yes-ode'","derivation":"from H3245 (יָסַד);","strongs_def":"a foundation (literally or figuratively)","kjv_def":"bottom, foundation, repairing"},
"H3248":{"lemma":"יְסוּדָה","xlit":"yᵉçûwdâh","pron":"yes-oo-daw'","derivation":"feminine of H3246 (יְסֻד);","strongs_def":"a foundation","kjv_def":"foundation."},
"H3249":{"lemma":"יָסוּר","xlit":"yâçûwr","pron":"yaw-soor'","derivation":"from H5493 (סוּר);","strongs_def":"departing","kjv_def":"they that depart"},
"H3250":{"lemma":"יִסּוֹר","xlit":"yiççôwr","pron":"yis-sore'","derivation":"from H3256 (יָסַר);","strongs_def":"a reprover","kjv_def":"instruct."},
"H3251":{"lemma":"יָסַךְ","xlit":"yâçak","pron":"yaw-sak'","derivation":"a primitive root;","strongs_def":"to pour (intransitive)","kjv_def":"be poured."},
"H3252":{"lemma":"יִסְכָּה","xlit":"Yiçkâh","pron":"yis-kaw'","derivation":"from an unused root meaning to watch; observant;","strongs_def":"Jiskah, sister of Lot","kjv_def":"Iscah."},
"H3253":{"lemma":"יִסְמַכְיָהוּ","xlit":"Yiçmakyâhûw","pron":"yis-mak-yaw-hoo'","derivation":"from H5564 (סָמַךְ) and H3050 (יָהּ); Jah will sustain;","strongs_def":"Jismakjah, an Israelite","kjv_def":"Ismachiah."},
"H3254":{"lemma":"יָסַף","xlit":"yâçaph","pron":"yaw-saf'","derivation":"a primitive root;","strongs_def":"to add or augment (often adverbial, to continue to do a thing)","kjv_def":"add, [idiom] again, [idiom] any more, [idiom] cease, [idiom] come more, [phrase] conceive again, continue, exceed, [idiom] further, [idiom] gather together, get more, give more-over, [idiom] henceforth, increase (more and more), join, [idiom] longer (bring, do, make, much, put), [idiom] (the, much, yet) more (and more), proceed (further), prolong, put, be (strong-) er, [idiom] yet, yield."},
"H3255":{"lemma":"יְסַף","xlit":"yᵉçaph","pron":"yes-af'","derivation":"(Aramaic) corresponding to H3254 (יָסַף)","strongs_def":"{to add or augment (often adverbial, to continue to do a thing)}","kjv_def":"add."},
"H3256":{"lemma":"יָסַר","xlit":"yâçar","pron":"yaw-sar'","derivation":"a primitive root;","strongs_def":"to chastise, literally (with blows) or figuratively (with words); hence, to instruct","kjv_def":"bind, chasten, chastise, correct, instruct, punish, reform, reprove, sore, teach."},
"H3257":{"lemma":"יָע","xlit":"yâʻ","pron":"yaw","derivation":"from H3261 (יָעָה);","strongs_def":"a shovel","kjv_def":"shovel."},
"H3258":{"lemma":"יַעְבֵּץ","xlit":"Yaʻbêts","pron":"yah-bates'","derivation":"from an unused root probably meaning to grieve; sorrowful;","strongs_def":"Jabets, the name of an Israelite, and also of a place in Palestine","kjv_def":"Jabez."},
"H3259":{"lemma":"יָעַד","xlit":"yâʻad","pron":"yaw-ad'","derivation":"a primitive root;","strongs_def":"to fix upon (by agreement or appointment); by implication, to meet (at a stated time), to summon (to trial), to direct (in a certain quarter or position), to engage (for marriage)","kjv_def":"agree,(maxke an) appoint(-ment, a time), assemble (selves), betroth, gather (selves, together), meet (together), set (a time)."},
"H3260":{"lemma":"יֶעְדִּי","xlit":"Yeʻdîy","pron":"yed-ee'","derivation":"from H3259 (יָעַד); appointed;","strongs_def":"Jedi, an Israelite","kjv_def":"Iddo (from the margin) See H3035 (יִדּוֹ)."},
"H3261":{"lemma":"יָעָה","xlit":"yâʻâh","pron":"yaw-aw'","derivation":"a primitive root; apparently","strongs_def":"to brush aside","kjv_def":"sweep away."},
"H3262":{"lemma":"יְעוּאֵל","xlit":"Yᵉʻûwʼêl","pron":"yeh-oo-ale'","derivation":"from H3261 (יָעָה) and H410 (אֵל); carried away of God;","strongs_def":"Jeuel, the name of four Israelites","kjv_def":"Jehiel, Jeiel, Jeuel. Compare H3273 (יְעִיאֵל)."},
"H3263":{"lemma":"יְעוּץ","xlit":"Yᵉʻûwts","pron":"yeh-oots'","derivation":"from H5779 (עוּץ); counsellor;","strongs_def":"Jeuts, an Israelite","kjv_def":"Jeuz"},
"H3264":{"lemma":"יָעוֹר","xlit":"yâʻôwr","pron":"yaw-ore'","derivation":"a variation of H3298 (יַעֲרֶשְׁיָה);","strongs_def":"a forest","kjv_def":"wood."},
"H3265":{"lemma":"יָעוּר","xlit":"Yâʻûwr","pron":"yaw-oor'","derivation":"apparently the passive participle of the same as H3293 (יַעַר); wooded;","strongs_def":"Jaur, an Israelite","kjv_def":"Jair(from the margin)."},
"H3266":{"lemma":"יְעוּשׁ","xlit":"Yᵉʻûwsh","pron":"yeh-oosh'","derivation":"from H5789 (עוּשׁ); hasty;","strongs_def":"Jeush, the name of an Edomite and of four Israelites","kjv_def":"Jehush, Jeush. Compare H3274 (יְעִישׁ)."},
"H3267":{"lemma":"יָעַז","xlit":"yâʻaz","pron":"yaw-az'","derivation":"a primitive root;","strongs_def":"to be bold or obstinate","kjv_def":"fierce."},
"H3268":{"lemma":"יַעֲזִיאֵל","xlit":"Yaʻăzîyʼêl","pron":"yah-az-ee-ale'","derivation":"from H3267 (יָעַז) and H410 (אֵל); emboldened of God;","strongs_def":"Jaaziel, an Israelite","kjv_def":"Jaaziel."},
"H3269":{"lemma":"יַעֲזִיָּהוּ","xlit":"Yaʻăzîyâhûw","pron":"yah-az-ee-yaw'-hoo","derivation":"from H3267 (יָעַז) and H3050 (יָהּ); emboldened of Jah;","strongs_def":"Jaazijah, an Israelite","kjv_def":"Jaaziah."},
"H3270":{"lemma":"יַעֲזֵיר","xlit":"Yaʻăzêyr","pron":"yah-az-ayr'","derivation":"or יַעְזֵר; from H5826 (עָזַר); helpful;","strongs_def":"Jaazer or Jazer, a place East of the Jordan","kjv_def":"Jaazer, Jazer."},
"H3271":{"lemma":"יָעַט","xlit":"yâʻaṭ","pron":"yaw-at'","derivation":"a primitive root;","strongs_def":"to clothe","kjv_def":"cover."},
"H3272":{"lemma":"יְעַט","xlit":"yᵉʻaṭ","pron":"yeh-at'","derivation":"(Aramaic) corresponding to H3289 (יָעַץ);","strongs_def":"to counsel; reflexively, to consult","kjv_def":"counsellor, consult together."},
"H3273":{"lemma":"יְעִיאֵל","xlit":"Yᵉʻîyʼêl","pron":"yeh-ee-ale'","derivation":"from H3261 (יָעָה) and H410 (אֵל); carried away of God;","strongs_def":"Jeiel, the name of six Israelites","kjv_def":"Jeiel, Jehiel. Compare H3262 (יְעוּאֵל)."},
"H3274":{"lemma":"יְעִישׁ","xlit":"Yᵉʻîysh","pron":"yeh-eesh'","derivation":"from H5789 (עוּשׁ); hasty;","strongs_def":"Jeish, the name of an Edomite and of a an Israelite","kjv_def":"Jeush (from the margin). Compare H3266 (יְעוּשׁ)."},
"H3275":{"lemma":"יַעְכָּן","xlit":"Yaʻkân","pron":"yah-kawn'","derivation":"from the same as H5912 (עָכָן); troublesome;","strongs_def":"Jakan, an Israelite","kjv_def":"Jachan."},
"H3276":{"lemma":"יַעַל","xlit":"yaʻal","pron":"yaw-al'","derivation":"a primitive root;","strongs_def":"properly, to ascend; figuratively, to be valuable (objectively; useful, subjectively; benefited)","kjv_def":"[idiom] at all, set forward, can do good, (be, have) profit, (able)."},
"H3277":{"lemma":"יָעֵל","xlit":"yâʻêl","pron":"yaw-ale'","derivation":"from H3276 (יַעַל);","strongs_def":"an ibex (as climbing)","kjv_def":"wild goat."},
"H3278":{"lemma":"יָעֵל","xlit":"Yâʻêl","pron":"yaw-ale'","derivation":"the same as H3277 (יָעֵל);","strongs_def":"Jael, a Canaanite","kjv_def":"Jael."},
"H3279":{"lemma":"יַעֲלָא","xlit":"Yaʻălâʼ","pron":"yah-al-aw'","derivation":"or יַעֲלָה; the same as H3280 (יַעֲלָה) or direct from H3276 (יַעַל);","strongs_def":"Jaala or Jaalah, one of the Nethinim","kjv_def":"Jaala, Jaalah."},
"H3280":{"lemma":"יַעֲלָה","xlit":"yaʻălâh","pron":"yah-al-aw'","derivation":"feminine of H3277 (יָעֵל)","strongs_def":"{an ibex (as climbing)}","kjv_def":"roe."},
"H3281":{"lemma":"יַעְלָם","xlit":"Yaʻlâm","pron":"yah-lawm'","derivation":"from H5956 (עָלַם); occult;","strongs_def":"Jalam, an Edomite","kjv_def":"Jalam."},
"H3282":{"lemma":"יַעַן","xlit":"yaʻan","pron":"yah'-an","derivation":"from an unused root meaning to pay attention;","strongs_def":"properly, heed; by implication, purpose (sake or account); used adverbially to indicate the reason or cause","kjv_def":"because (that), forasmuch ([phrase] as), seeing then, [phrase] that, [phrase] wheras, [phrase] why."},
"H3283":{"lemma":"יָעֵן","xlit":"yâʻên","pron":"yaw-ane'","derivation":"from the same as H3282 (יַעַן);","strongs_def":"the ostrich (probably from its answering cry","kjv_def":"ostrich."},
"H3284":{"lemma":"יַעֲנָה","xlit":"yaʻănâh","pron":"yah-an-aw'","derivation":"feminine of H3283 (יָעֵן), and meaning the same","strongs_def":"{the ostrich (probably from its answering cry}","kjv_def":"[phrase] owl."},
"H3285":{"lemma":"יַעֲנַי","xlit":"Yaʻănay","pron":"yah-an-ah'ee","derivation":"from the same as H3283 (יָעֵן); responsive;","strongs_def":"Jaanai, an Israelite","kjv_def":"Jaanai."},
"H3286":{"lemma":"יָעַף","xlit":"yâʻaph","pron":"yaw-af'","derivation":"a primitive root;","strongs_def":"to tire (as if from wearisome flight)","kjv_def":"faint, cause to fly, (be) weary (self)."},
"H3287":{"lemma":"יָעֵף","xlit":"yâʻêph","pron":"yaw-afe'","derivation":"from H3286 (יָעַף);","strongs_def":"fatigued; figuratively, exhausted","kjv_def":"faint, weary."},
"H3288":{"lemma":"יְעָף","xlit":"yᵉʻâph","pron":"yeh-awf'","derivation":"from H3286 (יָעַף);","strongs_def":"fatigue (adverb, utterly exhausted)","kjv_def":"swiftly.e"},
"H3289":{"lemma":"יָעַץ","xlit":"yâʻats","pron":"yaw-ats'","derivation":"a primitive root;","strongs_def":"to advise; reflexively, to deliberate or resolve","kjv_def":"advertise, take advise, advise (well), consult, (give, take) counsel(-lor), determine, devise, guide, purpose."},
"H3290":{"lemma":"יַעֲקֹב","xlit":"Yaʻăqôb","pron":"yah-ak-obe'","derivation":"from H6117 (עָקַב); heel-catcher (i.e. supplanter);","strongs_def":"Jaakob, the Israelitish patriarch","kjv_def":"Jacob."},
"H3291":{"lemma":"יַעֲקֹבָה","xlit":"Yaʻăqôbâh","pron":"yah-ak-o'-baw","derivation":"from H3290 (יַעֲקֹב);","strongs_def":"Jaakobah, an Israelite","kjv_def":"Jaakobah."},
"H3292":{"lemma":"יַעֲקָן","xlit":"Yaʻăqân","pron":"yah-ak-awn'","derivation":"from the same as H6130 (עָקָן);","strongs_def":"Jaakan, an Idumaean","kjv_def":"Jaakan. Compare H1142 (בְּנֵי יַעֲקָן)."},
"H3293":{"lemma":"יַעַר","xlit":"yaʻar","pron":"yah'-ar","derivation":"from an unused root probably meaning to thicken with verdure;","strongs_def":"a copse of bushes; hence, a forest; hence, honey in the comb (as hived in trees)","kjv_def":"(honey-) comb, forest, wood."},
"H3294":{"lemma":"יַעְרָה","xlit":"Yaʻrâh","pron":"yah-raw'","derivation":"a form of H3295 (יַעֲרָה);","strongs_def":"Jarah, an Israelite","kjv_def":"Jarah."},
"H3295":{"lemma":"יַעֲרָה","xlit":"yaʻărâh","pron":"yah-ar-aw'","derivation":"feminine of H3293 (יַעַר), and meaning the same","strongs_def":"{a copse of bushes; hence, a forest; hence, honey in the comb (as hived in trees)}","kjv_def":"(honey-) comb, forest."},
"H3296":{"lemma":"יַעֲרֵי אֹרְגִים","xlit":"Yaʻărêy ʼÔrᵉgîym","pron":"yah-ar-ay' o-reg-eem'","derivation":"from the plural of H3293 (יַעַר) and the masculine plural active participle of H707 (אָרַג); woods of weavers;","strongs_def":"Jaare-Oregim, an Israelite","kjv_def":"Jaare-oregim."},
"H3297":{"lemma":"יְעָרִים","xlit":"Yᵉʻârîym","pron":"yeh-aw-reem'","derivation":"plural of H3293 (יַעַר); forests;","strongs_def":"Jearim, a place in Palestine","kjv_def":"Jearim. Compare H7157 (קִרְיַת יְעָרִים)."},
"H3298":{"lemma":"יַעֲרֶשְׁיָה","xlit":"Yaʻăreshyâh","pron":"yah-ar-esh-yaw'","derivation":"from an unused root of uncertain signification and H3050 (יָהּ);","strongs_def":"Jaareshjah, an Israelite","kjv_def":"Jaresiah."},
"H3299":{"lemma":"יַעֲשׂוּ","xlit":"Yaʻăsûw","pron":"yah-as-oo'","derivation":"from H6213 (עָשָׂה); they will do;","strongs_def":"Jaasu, an Israelite","kjv_def":"Jaasau."},
"H3300":{"lemma":"יַעֲשִׂיאֵל","xlit":"Yaʻăsîyʼêl","pron":"yah-as-ee-ale'","derivation":"from H6213 (עָשָׂה) and H410 (אֵל); made of God;","strongs_def":"Jaasiel, an Israelite","kjv_def":"Jaasiel, Jasiel."},
"H3301":{"lemma":"יִפְדְּיָה","xlit":"Yiphdᵉyâh","pron":"yif-deh-yaw'","derivation":"from H6299 (פָּדָה) and H3050 (יָהּ); Jah will liberate;","strongs_def":"Jiphdejah, an Israelite","kjv_def":"Iphedeiah."},
"H3302":{"lemma":"יָפָה","xlit":"yâphâh","pron":"yaw-faw'","derivation":"a primitive root;","strongs_def":"properly, to be bright, i.e. (by implication) beautiful","kjv_def":"be beautiful, be (make self) fair(-r), deck."},
"H3303":{"lemma":"יָפֶה","xlit":"yâpheh","pron":"yaw-feh'","derivation":"from H3302 (יָפָה);","strongs_def":"beautiful (literally or figuratively)","kjv_def":"[phrase] beautiful, beauty, comely, fair(-est, one), [phrase] goodly, pleasant, well."},
"H3304":{"lemma":"יְפֵה־פִיָּה","xlit":"yᵉphêh-phîyâh","pron":"yef-eh' fee-yaw'","derivation":"from H3302 (יָפָה) by reduplication;","strongs_def":"very beautiful","kjv_def":"very fair."},
"H3305":{"lemma":"יָפוֹ","xlit":"Yâphôw","pron":"yaw-fo'","derivation":"or יָפוֹא; (Ezra 3:7), from H3302 (יָפָה); beautiful;","strongs_def":"Japho, a place in Palestine","kjv_def":"Japha, Joppa."},
"H3306":{"lemma":"יָפַח","xlit":"yâphach","pron":"yaw-fakh'","derivation":"a primitive root;","strongs_def":"properly, to breathe hard, i.e. (by implication) to sigh","kjv_def":"bewail self."},
"H3307":{"lemma":"יָפֵחַ","xlit":"yâphêach","pron":"yaw-fay'-akh","derivation":"from H3306 (יָפַח);","strongs_def":"properly, puffing, i.e. (figuratively) meditating","kjv_def":"such as breathe out."},
"H3308":{"lemma":"יֳפִי","xlit":"yŏphîy","pron":"yof-ee'","derivation":"from H3302 (יָפָה);","strongs_def":"beauty","kjv_def":"beauty."},
"H3309":{"lemma":"יָפִיעַ","xlit":"Yâphîyaʻ","pron":"yaw-fee'-ah","derivation":"from H3313 (יָפַע); bright;","strongs_def":"Japhia, the name of a Canaanite, an Israelite, and a place in Palestine","kjv_def":"Japhia."},
"H3310":{"lemma":"יַפְלֵט","xlit":"Yaphlêṭ","pron":"yaf-late'","derivation":"from H6403 (פָּלַט); he will deliver;","strongs_def":"Japhlet, an Israelite","kjv_def":"Japhlet."},
"H3311":{"lemma":"יַפְלֵטִי","xlit":"Yaphlêṭîy","pron":"yaf-lay-tee'","derivation":"patronymically from H3310 (יַפְלֵט);","strongs_def":"a Japhletite or descendant of Japhlet","kjv_def":"Japhleti."},
"H3312":{"lemma":"יְפֻנֶּה","xlit":"Yᵉphunneh","pron":"yef-oon-neh'","derivation":"from H6437 (פָּנָה); he will be prepared;","strongs_def":"Jephunneh, the name of two Israelites","kjv_def":"Jephunneh."},
"H3313":{"lemma":"יָפַע","xlit":"yâphaʻ","pron":"yaw-fah'","derivation":"a primitive root;","strongs_def":"to shine","kjv_def":"be light, shew self, (cause to) shine (forth)."},
"H3314":{"lemma":"יִפְעָה","xlit":"yiphʻâh","pron":"yif-aw'","derivation":"from H3313 (יָפַע);","strongs_def":"splendor or (figuratively) beauty","kjv_def":"brightness."},
"H3315":{"lemma":"יֶפֶת","xlit":"Yepheth","pron":"yeh'-feth","derivation":"from H6601 (פָּתָה); expansion;","strongs_def":"Jepheth, a son of Noah; also his posterity","kjv_def":"Japheth."},
"H3316":{"lemma":"יִפְתָּח","xlit":"Yiphtâch","pron":"yif-tawkh'","derivation":"from H6605 (פָּתַח); he will open;","strongs_def":"Jiphtach, an Israelite; also a place in Palestine","kjv_def":"Jephthah, Jiphtah."},
"H3317":{"lemma":"יִפְתַּח־אֵל","xlit":"Yiphtach-ʼêl","pron":"yif-tach-ale'","derivation":"from H6605 (פָּתַח) and H410 (אֵל); God will open;","strongs_def":"Jiphtach-el, a place in Palestine","kjv_def":"Jiphthah-el."},
"H3318":{"lemma":"יָצָא","xlit":"yâtsâʼ","pron":"yaw-tsaw'","derivation":"a primitive root;","strongs_def":"to go (causatively, bring) out, in a great variety of applications, literally and figuratively, direct and proxim.","kjv_def":"[idiom] after, appear, [idiom] assuredly, bear out, [idiom] begotten, break out, bring forth (out, up), carry out, come (abroad, out, thereat, without), [phrase] be condemned, depart(-ing, -ure), draw forth, in the end, escape, exact, fail, fall (out), fetch forth (out), get away (forth, hence, out), (able to, cause to, let) go abroad (forth, on, out), going out, grow, have forth (out), issue out, lay (lie) out, lead out, pluck out, proceed, pull out, put away, be risen, [idiom] scarce, send with commandment, shoot forth, spread, spring out, stand out, [idiom] still, [idiom] surely, take forth (out), at any time, [idiom] to (and fro), utter."},
"H3319":{"lemma":"יְצָא","xlit":"yᵉtsâʼ","pron":"yets-aw'","derivation":"(Aramaic) corresponding to H3318 (יָצָא)","strongs_def":"{to go (causatively, bring) out, in a great variety of applications, literally and figuratively, direct and proxim.}","kjv_def":"finish."},
"H3320":{"lemma":"יָצַב","xlit":"yâtsab","pron":"yaw-tsab'","derivation":"a primitive root;","strongs_def":"to place (any thing so as to stay); reflexively, to station, offer, continue","kjv_def":"present selves, remaining, resort, set (selves), (be able to, can, with-) stand (fast, forth, -ing, still, up)."},
"H3321":{"lemma":"יְצַב","xlit":"yᵉtsab","pron":"yets-abe'","derivation":"(Aramaic) corresponding to H3320 (יָצַב);","strongs_def":"to be firm; hence, to speak surely","kjv_def":"truth."},
"H3322":{"lemma":"יָצַג","xlit":"yâtsag","pron":"yaw-tsag'","derivation":"a primitive root;","strongs_def":"to place permanently","kjv_def":"establish, leave, make, present, put, set, stay."},
"H3323":{"lemma":"יִצְהָר","xlit":"yitshâr","pron":"yits-hawr'","derivation":"from H6671 (צָהַר);","strongs_def":"oil (as producing light); figuratively, anointing","kjv_def":"[phrase] anointed oil."},
"H3324":{"lemma":"יִצְהָר","xlit":"Yitshâr","pron":"yits-hawr'","derivation":"the same as H3323 (יִצְהָר);","strongs_def":"Jitshar, an Israelite","kjv_def":"Izhar."},
"H3325":{"lemma":"יִצְהָרִי","xlit":"Yitshârîy","pron":"yits-haw-ree'","derivation":"patronymically from H3324 (יִצְהָר);","strongs_def":"a Jitsharite or descendant of Jitshar","kjv_def":"Izeharites, Izharites."},
"H3326":{"lemma":"יָצוּעַ","xlit":"yâtsûwaʻ","pron":"yaw-tsoo'-ah","derivation":"passive participle of H3331 (יַצַע);","strongs_def":"spread, i.e. a bed; (architecture) an extension, i.e. wing or lean-to (a single story or collectively)","kjv_def":"bed, chamber, couch."},
"H3327":{"lemma":"יִצְחָק","xlit":"Yitschâq","pron":"yits-khawk'","derivation":"from H6711 (צָחַק); laughter (i.e. mochery);","strongs_def":"Jitschak (or Isaac), son of Abraham","kjv_def":"Isaac. Compare H3446 (יִשְׂחָק)."},
"H3328":{"lemma":"יִצְחַר","xlit":"Yitschar","pron":"yits-khar'","derivation":"from the same as H6713 (צַחַר); he will shine;","strongs_def":"Jitschar, an Israelite","kjv_def":"and Zehoar (from the margin)."},
"H3329":{"lemma":"יָצִיא","xlit":"yâtsîyʼ","pron":"yaw-tsee'","derivation":"from H3318 (יָצָא);","strongs_def":"issue, i.e. offspring","kjv_def":"those that came forth."},
"H3330":{"lemma":"יַצִּיב","xlit":"yatstsîyb","pron":"yats-tseeb'","derivation":"(Aramaic) from H3321 (יְצַב)lemma יְצב missing vowel, corrected to יְצַבxlit yᵉtsêb corrected to yᵉtsab;","strongs_def":"fixed, sure; concretely, certainty","kjv_def":"certain(-ty), true, truth."},
"H3331":{"lemma":"יַצַע","xlit":"yatsaʻ","pron":"yaw-tsah'","derivation":"a primitive root;","strongs_def":"to strew as a surface","kjv_def":"make (one's) bed, [idiom] lie, spread."},
"H3332":{"lemma":"יָצַק","xlit":"yâtsaq","pron":"yaw-tsak'","derivation":"a primitive root;","strongs_def":"properly, to pour out (transitive or intransitive); by implication, to melt or cast as metal; by extension, to place firmly, to stiffen or grow hard","kjv_def":"cast, cleave fast, be (as) firm, grow, be hard, lay out, molten, overflow, pour (out), run out, set down, stedfast."},
"H3333":{"lemma":"יְצֻקָה","xlit":"yᵉtsuqâh","pron":"yets-oo-kaw'","derivation":"passive participle feminine of H3332 (יָצַק);","strongs_def":"poured out, i.e. run into amould","kjv_def":"when it was cast."},
"H3334":{"lemma":"יָצַר","xlit":"yâtsar","pron":"yaw-tsar'","derivation":"a primitive root;","strongs_def":"to press (intransitive), i.e. be narrow; figuratively, be in distress","kjv_def":"be distressed, be narrow, be straitened (in straits), be vexed."},
"H3335":{"lemma":"יָצַר","xlit":"yâtsar","pron":"yaw-tsar'","derivation":"probably identical with H3334 (יָצַר) (through the squeezing into shape); (compare H3331 (יַצַע));","strongs_def":"to mould into a form; especially as apotter; figuratively, to determine (i.e. form a resolution)","kjv_def":"[idiom] earthen, fashion, form, frame, make(-r), potter, purpose."},
"H3336":{"lemma":"יֵצֶר","xlit":"yêtser","pron":"yay'-tser","derivation":"from H3335 (יָצַר);","strongs_def":"a form; figuratively, conception (i.e. purpose)","kjv_def":"frame, thing framed, imagination, mind, work."},
"H3337":{"lemma":"יֵצֶר","xlit":"Yêtser","pron":"yay-tser","derivation":"the same as H3336 (יֵצֶר);","strongs_def":"Jetser, an Israelite","kjv_def":"Jezer."},
"H3338":{"lemma":"יָצֻר","xlit":"yâtsur","pron":"yaw-tsoor'","derivation":"passive participle of H3335 (יָצַר);","strongs_def":"structure, i.e. limb or part","kjv_def":"member.ql"},
"H3339":{"lemma":"יִצְרִי","xlit":"Yitsrîy","pron":"yits-ree'","derivation":"from H3335 (יָצַר); formative;","strongs_def":"Jitsri, an Israelite","kjv_def":"Isri."},
"H3340":{"lemma":"יִצְרִי","xlit":"Yitsrîy","pron":"yits-ree'","derivation":"; patronymically from H3337 (יֵצֶר);","strongs_def":"a Jitsrite (collectively) or descendants of Jetser","kjv_def":"Jezerites."},
"H3341":{"lemma":"יָצַת","xlit":"yâtsath","pron":"yaw-tsath'","derivation":"a primitive root;","strongs_def":"to burn or set on fire; figuratively, to desolate","kjv_def":"burn (up), be desolate, set (on) fire (fire), kindle."},
"H3342":{"lemma":"יֶקֶב","xlit":"yeqeb","pron":"yeh'-keb","derivation":"from an unused root meaning to excavate;","strongs_def":"a trough (as dug out); specifically, a wine-vat (whether the lower one, into which the juice drains; or the upper, in which the grapes are crushed)","kjv_def":"fats, presses, press-fat, wine(-press)."},
"H3343":{"lemma":"יְקַבְצְאֵל","xlit":"Yᵉqabtsᵉʼêl","pron":"yek-ab-tseh-ale'","derivation":"from H6908 (קָבַץ) and H410 (אֵל); God will gather;","strongs_def":"Jekabtseel, a place in Palestine","kjv_def":"Jekabzeel. Compare H6909 (קַבְצְאֵל)."},
"H3344":{"lemma":"יָקַד","xlit":"yâqad","pron":"yaw-kad'","derivation":"a primitive root;","strongs_def":"to burn","kjv_def":"(be) burn(-ing), [idiom] from the hearth, kindle."},
"H3345":{"lemma":"יְקַד","xlit":"yᵉqad","pron":"yek-ad'","derivation":"(Aramaic) corresponding to H3344 (יָקַד)","strongs_def":"{to burn}","kjv_def":"burning."},
"H3346":{"lemma":"יְקֵדָא","xlit":"yᵉqêdâʼ","pron":"yek-ay-daw'","derivation":"(Aramaic) from H3345 (יְקַד);","strongs_def":"a conflagration","kjv_def":"burning."},
"H3347":{"lemma":"יׇקְדְעָם","xlit":"Yoqdᵉʻâm","pron":"yok-deh-awm'","derivation":"from H3344 (יָקַד) and H5971 (עַם); burning of (the) people;","strongs_def":"Jokdeam, a place in Palestine","kjv_def":"Jokdeam."},
"H3348":{"lemma":"יָקֶה","xlit":"Yâqeh","pron":"yaw-keh'","derivation":"from an unused root probably meaning to obey; obedient;","strongs_def":"Jakeh, a symbolical name (for Solomon)","kjv_def":"Jakeh."},
"H3349":{"lemma":"יִקָּהָה","xlit":"yiqqâhâh","pron":"yik-kaw-haw'","derivation":"from the same as H3348 (יָקֶה);","strongs_def":"obedience","kjv_def":"gathering, to obey."},
"H3350":{"lemma":"יְקוֹד","xlit":"yᵉqôwd","pron":"yek-ode'","derivation":"from H3344 (יָקַד);","strongs_def":"a burning","kjv_def":"burning."},
"H3351":{"lemma":"יְקוּם","xlit":"yᵉqûwm","pron":"yek-oom'","derivation":"from H6965 (קוּם);","strongs_def":"properly, standing (extant), i.e. by implication, a living thing","kjv_def":"(living) substance."},
"H3352":{"lemma":"יָקוֹשׁ","xlit":"yâqôwsh","pron":"yaw-koshe'","derivation":"from H3369 (יָקֹשׁ);","strongs_def":"properly, entangling; hence, a snarer","kjv_def":"fowler."},
"H3353":{"lemma":"יָקוּשׁ","xlit":"yâqûwsh","pron":"yaw-koosh'","derivation":"passive participle of H3369 (יָקֹשׁ);","strongs_def":"properly, entangled, i.e. by implication (intransitively) a snare, or (transitive) a snarer","kjv_def":"fowler, snare."},
"H3354":{"lemma":"יְקוּתִיאֵל","xlit":"Yᵉqûwthîyʼêl","pron":"yek-ooth-ee'-ale","derivation":"from the same as H3348 (יָקֶה) and H410 (אֵל); obedience of God;","strongs_def":"Jekuthiel, an Israelite","kjv_def":"Jekuthiel."},
"H3355":{"lemma":"יׇקְטָן","xlit":"Yoqṭân","pron":"yok-tawn'","derivation":"from H6994 (קָטֹן); he will be made little;","strongs_def":"Joktan, an Arabian patriarch","kjv_def":"Joktan."},
"H3356":{"lemma":"יָקִים","xlit":"Yâqîym","pron":"yaw-keem'","derivation":"from H6965 (קוּם); he will raise;","strongs_def":"Jakim, the name of two Israelites","kjv_def":"Jakim. Compare H3079 (יְהוֹיָקִים)."},
"H3357":{"lemma":"יַקִּיר","xlit":"yaqqîyr","pron":"yak-keer'","derivation":"from H3365 (יָקַר);","strongs_def":"precious","kjv_def":"dear."},
"H3358":{"lemma":"יַקִּיר","xlit":"yaqqîyr","pron":"yak-keer'","derivation":"(Aramaic) corresponding to H3357 (יַקִּיר)","strongs_def":"{precious}","kjv_def":"noble, rare."},
"H3359":{"lemma":"יְקַמְיָה","xlit":"Yᵉqamyâh","pron":"yek-am-yaw'","derivation":"from H6965 (קוּם) and H3050 (יָהּ); Jah will rise;","strongs_def":"Jekamjah, the name of two Israelites","kjv_def":"Jekamiah. Compare H3079 (יְהוֹיָקִים)."},
"H3360":{"lemma":"יְקַמְעָם","xlit":"Yᵉqamʻâm","pron":"yek-am'-awm","derivation":"from H6965 (קוּם) and H5971 (עַם); (the) people will rise;","strongs_def":"Jekamam, an Israelite","kjv_def":"Jekameam. Compare H3079 (יְהוֹיָקִים), H3361 (יׇקְמְעָם)."},
"H3361":{"lemma":"יׇקְמְעָם","xlit":"Yoqmᵉʻâm","pron":"yok-meh-awm'","derivation":"from H6965 (קוּם) and H5971 (עַם); (the) people will be raised;","strongs_def":"Jokmeam, a place in Palestine","kjv_def":"Jokmeam. Compare H3360 (יְקַמְעָם), H3362 (יׇקְנְעָם)."},
"H3362":{"lemma":"יׇקְנְעָם","xlit":"Yoqnᵉʻâm","pron":"yok-neh-awm'","derivation":"from H6969 (קוּן) and H5971 (עַם); (the) people will be lamented;","strongs_def":"Jokneam, a place in Palestine","kjv_def":"Jokneam."},
"H3363":{"lemma":"יָקַע","xlit":"yâqaʻ","pron":"yaw-kah'","derivation":"a primitive root;","strongs_def":"properly, to sever oneself, i.e. (by implication) to be dislocated; figuratively, to abandon; causatively, to impale (and thus allow to drop to pieces by rotting)","kjv_def":"be alienated, depart, hang (up), be out of joint."},
"H3364":{"lemma":"יָקַץ","xlit":"yâqats","pron":"yaw-kats'","derivation":"a primitive root;","strongs_def":"to awake (intransitive)","kjv_def":"(be) awake(-d)."},
"H3365":{"lemma":"יָקַר","xlit":"yâqar","pron":"yaw-kar'","derivation":"a primitive root;","strongs_def":"properly, apparently, to be heavy, i.e. (figuratively) valuable; causatively, to make rare (figuratively, to inhibit)","kjv_def":"be (make) precious, be prized, be set by, withdraw."},
"H3366":{"lemma":"יְקָר","xlit":"yᵉqâr","pron":"yek-awr'","derivation":"from H3365 (יָקַר);","strongs_def":"value, i.e. (concretely) wealth; abstractly, costliness, dignity","kjv_def":"honour, precious (things), price."},
"H3367":{"lemma":"יְקָר","xlit":"yᵉqâr","pron":"yek-awr'","derivation":"(Aramaic) corresponding to H3366 (יְקָר)","strongs_def":"{value, i.e. (concretely) wealth; abstractly, costliness, dignity}","kjv_def":"glory, honour."},
"H3368":{"lemma":"יָקָר","xlit":"yâqâr","pron":"yaw-kawr'","derivation":"from H3365 (יָקַר);","strongs_def":"valuable (objectively or subjectively)","kjv_def":"brightness, clear, costly, excellent, fat, honourable women, precious, reputation."},
"H3369":{"lemma":"יָקֹשׁ","xlit":"yâqôsh","pron":"yaw-koshe'","derivation":"a primitive root;","strongs_def":"to ensnare (literally or figuratively)","kjv_def":"fowler (lay a) snare."},
"H3370":{"lemma":"יׇקְשָׁן","xlit":"Yoqshân","pron":"yok-shawn'","derivation":"from H3369 (יָקֹשׁ); insidious;","strongs_def":"Jokshan, an Arabian patriarch","kjv_def":"Jokshan."},
"H3371":{"lemma":"יׇקְתְאֵל","xlit":"Yoqthᵉʼêl","pron":"yok-theh-ale'","derivation":"probably from the same as H3348 (יָקֶה) and H410 (אֵל); veneration of God (compare H3354 (יְקוּתִיאֵל));","strongs_def":"Joktheel, the name of a place in Palestine, and of one in Idumaea","kjv_def":"Joktheel."},
"H3372":{"lemma":"יָרֵא","xlit":"yârêʼ","pron":"yaw-ray'","derivation":"a primitive root;","strongs_def":"to fear; morally, to revere; caus. to frighten","kjv_def":"affright, be (make) afraid, dread(-ful), (put in) fear(-ful, -fully, -ing), (be had in) reverence(-end), [idiom] see, terrible (act, -ness, thing)."},
"H3373":{"lemma":"יָרֵא","xlit":"yârêʼ","pron":"yaw-ray'","derivation":"from H3372 (יָרֵא);","strongs_def":"fearing; morally, reverent","kjv_def":"afraid, fear (-ful)."},
"H3374":{"lemma":"יִרְאָה","xlit":"yirʼâh","pron":"yir-aw'","derivation":"feminine of H3373 (יָרֵא);","strongs_def":"fear (also used as infinitive); morally, reverence","kjv_def":"[idiom] dreadful, [idiom] exceedingly, fear(-fulness)."},
"H3375":{"lemma":"יִרְאוֹן","xlit":"Yirʼôwn","pron":"yir-ohn'","derivation":"from H3372 (יָרֵא); fearfulness;","strongs_def":"Jiron, a place in Palestine","kjv_def":"Iron."},
"H3376":{"lemma":"יִרְאִיָּיה","xlit":"Yirʼîyâyh","pron":"yir-ee-yaw'","derivation":"from H3373 (יָרֵא) and H3050 (יָהּ); fearful of Jah;","strongs_def":"Jirijah, an Israelite","kjv_def":"Irijah."},
"H3377":{"lemma":"יָרֵב","xlit":"Yârêb","pron":"yaw-rabe'","derivation":"from H7378 (רִיב); he will contend;","strongs_def":"Jareb, a symbolical name for Assyria","kjv_def":"Jareb. Compare H3402 (יָרִיב)."},
"H3378":{"lemma":"יְרֻבַּעַל","xlit":"Yᵉrubbaʻal","pron":"yer-oob-bah'-al","derivation":"from H7378 (רִיב) and H1168 (בַּעַל); Baal will contend;","strongs_def":"Jerubbaal, a symbolic name of Gideon","kjv_def":"Jerubbaal."},
"H3379":{"lemma":"יָרׇבְעָם","xlit":"Yârobʻâm","pron":"yaw-rob-awm'","derivation":"from H7378 (רִיב) and H5971 (עַם); (the) people will contend;","strongs_def":"Jarobam, the name of two Israelite kings","kjv_def":"Jeroboam."},
"H3380":{"lemma":"יְרֻבֶּשֶׁת","xlit":"Yᵉrubbesheth","pron":"yer-oob-beh'-sheth","derivation":"from H7378 (רִיב) and H1322 (בֹּשֶׁת); shame (i.e. the idol) will contend;","strongs_def":"Jerubbesheth, a symbolic name for Gideon","kjv_def":"Jerubbesheth."},
"H3381":{"lemma":"יָרַד","xlit":"yârad","pron":"yaw-rad'","derivation":"a primitive root;","strongs_def":"to descend (literally, to go downwards; or conventionally to a lower region, as the shore, a boundary, the enemy, etc.; or figuratively, to fall); causatively, to bring down (in all the above applications)","kjv_def":"[idiom] abundantly, bring down, carry down, cast down, (cause to) come(-ing) down, fall (down), get down, go(-ing) down(-ward), hang down, [idiom] indeed, let down, light (down), put down (off), (cause to, let) run down, sink, subdue, take down."},
"H3382":{"lemma":"יֶרֶד","xlit":"Yered","pron":"yeh'-red","derivation":"from H3381 (יָרַד); a descent;","strongs_def":"Jered, the name of an antediluvian, and of an Israelite","kjv_def":"Jared."},
"H3383":{"lemma":"יַרְדֵּן","xlit":"Yardên","pron":"yar-dane'","derivation":"from H3381 (יָרַד); a descender;","strongs_def":"Jarden, the principal river of Palestine","kjv_def":"Jordan."},
"H3384":{"lemma":"יָרָה","xlit":"yârâh","pron":"yaw-raw'","derivation":"or (2 Chronicles 26:15) יָרָא; a primitive root;","strongs_def":"properly, to flow as water (i.e. to rain); transitively, to lay or throw (especially an arrow, i.e. to shoot); figuratively, to point out (as if by aiming the finger), to teach","kjv_def":"([phrase]) archer, cast, direct, inform, instruct, lay, shew, shoot, teach(-er,-ing), through."},
"H3385":{"lemma":"יְרוּאֵל","xlit":"Yᵉrûwʼêl","pron":"yer-oo-ale'","derivation":"from H3384 (יָרָה) and H410 (אֵל); founded of God;","strongs_def":"Jeruel, a place in Palestine","kjv_def":"Jeruel."},
"H3386":{"lemma":"יָרוֹחַ","xlit":"Yârôwach","pron":"yaw-ro'-akh","derivation":"perhaps denominative from H3394 (יָרֵחַ); (born at the) new moon;","strongs_def":"Jaroach, an Israelite","kjv_def":"Jaroah."},
"H3387":{"lemma":"יָרוֹק","xlit":"yârôwq","pron":"yaw-roke'","derivation":"from H3417 (יָרָק);","strongs_def":"green, i.e. an herb","kjv_def":"green thing."},
"H3388":{"lemma":"יְרוּשָׁא","xlit":"Yᵉrûwshâʼ","pron":"yer-oo-shaw'","derivation":"or יְרוּשָׁה feminine passive participle of H3423 (יָרַשׁ); possessed;","strongs_def":"Jerusha or Jerushah, as Israelitess","kjv_def":"Jerusha, Jerushah."},
"H3389":{"lemma":"יְרוּשָׁלַ͏ִם","xlit":"Yᵉrûwshâlaim","pron":"yer-oo-shaw-lah'-im","derivation":"rarely יְרוּשָׁלַיִם; a dual (in allusion to its two main hills (the true pointing, at least of the former reading, seems to be that of H3390 (יְרוּשָׁלֵם))); probably from (the passive participle of) H3384 (יָרָה) and H7999 (שָׁלַם); founded peaceful;","strongs_def":"Jerushalaim or Jerushalem, the capital city of Palestine","kjv_def":"Jerusalem."},
"H3390":{"lemma":"יְרוּשָׁלֵם","xlit":"Yᵉrûwshâlêm","pron":"yer-oo-shaw-lame'","derivation":"(Aramaic) corresponding to H3389 (יְרוּשָׁלַ͏ִם)","strongs_def":"{Jerusalem}","kjv_def":"{Jerusalem}"},
"H3391":{"lemma":"יֶרַח","xlit":"yerach","pron":"yeh'-rakh","derivation":"from a unused root of uncertain signification;","strongs_def":"a lunation, i.e. month","kjv_def":"month, moon."},
"H3392":{"lemma":"יֶרַח","xlit":"Yerach","pron":"yeh'-rakh","derivation":"the same as H3391 (יֶרַח);","strongs_def":"Jerach, an Arabian patriarch","kjv_def":"Jerah."},
"H3393":{"lemma":"יְרַח","xlit":"yᵉrach","pron":"yeh-rakh'","derivation":"(Aramaic) corresponding to H3391 (יֶרַח);","strongs_def":"a month","kjv_def":"month."},
"H3394":{"lemma":"יָרֵחַ","xlit":"yârêach","pron":"yaw-ray'-akh","derivation":"from the same as H3391 (יֶרַח);","strongs_def":"the moon","kjv_def":"moon. Yrechow. See H3405 (יְרִיחוֹ)."},
"H3395":{"lemma":"יְרֹחָם","xlit":"Yᵉrôchâm","pron":"yer-o-khawm'","derivation":"from H7355 (רָחַם); compassionate;","strongs_def":"Jerocham, the name of seven or eight Israelites","kjv_def":"Jeroham."},
"H3396":{"lemma":"יְרַחְמְאֵל","xlit":"Yᵉrachmᵉʼêl","pron":"yer-akh-meh-ale'","derivation":"from H7355 (רָחַם) and H410 (אֵל); God will compassionate;","strongs_def":"Jerachmeel, the name of three Israelites","kjv_def":"Jerahmeel."},
"H3397":{"lemma":"יְרַחְמְאֵלִי","xlit":"Yᵉrachmᵉʼêlîy","pron":"yer-akh-meh-ay-lee'","derivation":"patronymically from H3396 (יְרַחְמְאֵל);","strongs_def":"a Jerachmeelite or descendant of Jerachmeel","kjv_def":"Jerahmeelites."},
"H3398":{"lemma":"יַרְחָע","xlit":"Yarchâʻ","pron":"yar-khaw'","derivation":"probably of Egyptian origin;","strongs_def":"Jarcha, an Egyptian","kjv_def":"Jarha."},
"H3399":{"lemma":"יָרַט","xlit":"yâraṭ","pron":"yaw-rat'","derivation":"a primitive root;","strongs_def":"to precipitate or hurl (rush) headlong; (intransitively) to be rash","kjv_def":"be perverse, turn over."},
"H3400":{"lemma":"יְרִיאֵל","xlit":"Yᵉrîyʼêl","pron":"yer-ee-ale'","derivation":"from H3384 (יָרָה) and H410 (אֵל); thrown of God;","strongs_def":"Jeriel, an Israelite","kjv_def":"Jeriel. Compare H3385 (יְרוּאֵל)."},
"H3401":{"lemma":"יָרִיב","xlit":"yârîyb","pron":"yaw-rebe'","derivation":"from H7378 (רִיב);","strongs_def":"literally he will contend; properly, adjective, contentious; used as noun, an adversary","kjv_def":"that content(-eth), that strive."},
"H3402":{"lemma":"יָרִיב","xlit":"Yârîyb","pron":"yaw-rebe'","derivation":"the same as H3401 (יָרִיב);","strongs_def":"Jarib, the name of three Israelites","kjv_def":"Jarib."},
"H3403":{"lemma":"יְרִיבַי","xlit":"Yᵉrîybay","pron":"yer-eeb-ah'ee","derivation":"from H3401 (יָרִיב); contentious;","strongs_def":"Jeribai, an Israelite","kjv_def":"Jeribai."},
"H3404":{"lemma":"יְרִיָּה","xlit":"Yᵉrîyâh","pron":"yer-ee-yaw'","derivation":"or יְרִיָּהוּ; from H3384 (יָרָה) and H3050 (יָהּ); Jah will throw;","strongs_def":"Jerijah, an Israelite","kjv_def":"Jeriah, Jerijah."},
"H3405":{"lemma":"יְרִיחוֹ","xlit":"Yᵉrîychôw","pron":"yer-ee-kho'","derivation":"or יְרֵחוֹ; or variation (1 Kings 16:34) יְרִיחֹה; perhaps from H3394 (יָרֵחַ); its month; or else from H7306 (רוּחַ); fragrant;","strongs_def":"Jericho or Jerecho, a place in Palestine","kjv_def":"Jericho."},
"H3406":{"lemma":"יְרִימוֹת","xlit":"Yᵉrîymôwth","pron":"yer-ee-mohth'","derivation":"or יְרֵימוֹת; or יְרֵמוֹת; feminine plural from H7311 (רוּם); elevations;","strongs_def":"Jerimoth or Jeremoth, the name of twelve Israelites","kjv_def":"Jermoth, Jerimoth, and Ramoth (from the margin)."},
"H3407":{"lemma":"יְרִיעָה","xlit":"yᵉrîyʻâh","pron":"yer-ee-aw'","derivation":"from H3415 (יָרַע);","strongs_def":"a hanging (as tremulous)","kjv_def":"curtain."},
"H3408":{"lemma":"יְרִיעוֹת","xlit":"Yᵉrîyʻôwth","pron":"yer-ee-ohth'","derivation":"plural of H3407 (יְרִיעָה); curtains;","strongs_def":"Jerioth, an Israelitess","kjv_def":"Jerioth."},
"H3409":{"lemma":"יָרֵךְ","xlit":"yârêk","pron":"yaw-rake'","derivation":"from an unused root meaning to be soft;","strongs_def":"the thigh (from its fleshy softness); by euphemistically the generative parts; figuratively, a shank, flank, side","kjv_def":"[idiom] body, loins, shaft, side, thigh."},
"H3410":{"lemma":"יַרְכָא","xlit":"yarkâʼ","pron":"yar-kaw'","derivation":"(Aramaic) corresponding to H3411 (יְרֵכָה);","strongs_def":"a thigh","kjv_def":"thigh."},
"H3411":{"lemma":"יְרֵכָה","xlit":"yᵉrêkâh","pron":"yer-ay-kaw'","derivation":"feminine of H3409 (יָרֵךְ);","strongs_def":"properly, the flank; but used only figuratively, the rear or recess","kjv_def":"border, coast, part, quarter, side."},
"H3412":{"lemma":"יַרְמוּת","xlit":"Yarmûwth","pron":"yar-mooth'","derivation":"from H7311 (רוּם); elevation;","strongs_def":"Jarmuth, the name of two places in Palestine","kjv_def":"Jarmuth."},
"H3413":{"lemma":"יְרֵמַי","xlit":"Yᵉrêmay","pron":"yer-ay-mah'-ee","derivation":"from H7311 (רוּם); elevated;","strongs_def":"Jeremai, an Israelite","kjv_def":"Jeremai."},
"H3414":{"lemma":"יִרְמְיָה","xlit":"Yirmᵉyâh","pron":"yir-meh-yaw'","derivation":"or יִרְמְיָהוּ; from H7311 (רוּם) and H3050 (יָהּ); Jah will rise;","strongs_def":"Jirmejah, the name of eight or nine Israelites","kjv_def":"Jeremiah."},
"H3415":{"lemma":"יָרַע","xlit":"yâraʻ","pron":"yaw-rah'","derivation":"a primitive root;","strongs_def":"properly, to be broken up (with any violent action) i.e. (figuratively) to fear","kjv_def":"be grevious (only Isaiah 15:4; the rest belong to H7489 (רָעַע))."},
"H3416":{"lemma":"יִרְפְּאֵל","xlit":"Yirpᵉʼêl","pron":"yir-peh-ale'","derivation":"from H7495 (רָפָא) and H410 (אֵל); God will heal;","strongs_def":"Jirpeel, a place in Palestine","kjv_def":"Irpeel."},
"H3417":{"lemma":"יָרָק","xlit":"yârâq","pron":"yaw-rak'","derivation":"a primitive root;","strongs_def":"to spit","kjv_def":"[idiom] but, spit."},
"H3418":{"lemma":"יֶרֶק","xlit":"yereq","pron":"yeh'-rek","derivation":"from H3417 (יָרָק) (in the sense of vacuity of color);","strongs_def":"properly, pallor, i.e. hence, the yellowish green of young and sickly vegetation; concretely, verdure, i.e. grass or vegetation","kjv_def":"grass, green (thing)."},
"H3419":{"lemma":"יָרָק","xlit":"yârâq","pron":"yaw-rawk'","derivation":"from the same as H3418 (יֶרֶק);","strongs_def":"properly, green; concretely, a vegetable","kjv_def":"green, herbs."},
"H3420":{"lemma":"יֵרָקוֹן","xlit":"yêrâqôwn","pron":"yay-raw-kone'","derivation":"from H3418 (יֶרֶק);","strongs_def":"paleness, whether of persons (from fright), or of plants (from drought)","kjv_def":"greenish, yellow."},
"H3421":{"lemma":"יׇרְקְעָם","xlit":"Yorqᵉʻâm","pron":"yor-keh-awm'","derivation":"from H7324 (רוּק) and H5971 (עַם); people will be poured forth;","strongs_def":"Jorkeam, a place in Palestine","kjv_def":"Jorkeam."},
"H3422":{"lemma":"יְרַקְרַק","xlit":"yᵉraqraq","pron":"yer-ak-rak'","derivation":"from the same as H3418 (יֶרֶק);","strongs_def":"yellowishness","kjv_def":"greenish, yellow."},
"H3423":{"lemma":"יָרַשׁ","xlit":"yârash","pron":"yaw-rash'","derivation":"or יָרֵשׁ; a primitive root;","strongs_def":"to occupy (by driving out previous tenants, and possessing in their place); by implication, to seize, to rob, to inherit; also to expel, to impoverish, to ruin","kjv_def":"cast out, consume, destroy, disinherit, dispossess, drive(-ing) out, enjoy, expel, [idiom] without fail, (give to, leave for) inherit(-ance, -or) [phrase] magistrate, be (make) poor, come to poverty, (give to, make to) possess, get (have) in (take) possession, seize upon, succeed, [idiom] utterly."},
"H3424":{"lemma":"יְרֵשָׁה","xlit":"yᵉrêshâh","pron":"yer-ay-shaw'","derivation":"from H3423 (יָרַשׁ);","strongs_def":"occupancy","kjv_def":"possession."},
"H3425":{"lemma":"יְרֻשָּׁה","xlit":"yᵉrushshâh","pron":"yer-oosh-shaw'","derivation":"from H3423 (יָרַשׁ);","strongs_def":"something occupied; a conquest; also a patrimony","kjv_def":"heritage, inheritance, possession."},
"H3426":{"lemma":"יֵשׁ","xlit":"yêsh","pron":"yaysh","derivation":"perhaps from an unused root meaning to stand out, or exist; entity; used adverbially or as a copula for the substantive verb (H1961 (הָיָה));","strongs_def":"there is or are (or any other form of the verb to be, as may suit the connection)","kjv_def":"(there) are, (he, it, shall, there, there may, there shall, there should) be, thou do, had, hast, (which) hath, (I, shalt, that) have, (he, it, there) is, substance, it (there) was, (there) were, ye will, thou wilt, wouldest."},
"H3427":{"lemma":"יָשַׁב","xlit":"yâshab","pron":"yaw-shab'","derivation":"a primitive root;","strongs_def":"properly, to sit down (specifically as judge. in ambush, in quiet); by implication, to dwell, to remain; causatively, to settle, to marry","kjv_def":"(make to) abide(-ing), continue, (cause to, make to) dwell(-ing), ease self, endure, establish, [idiom] fail, habitation, haunt, (make to) inhabit(-ant), make to keep (house), lurking, [idiom] marry(-ing), (bring again to) place, remain, return, seat, set(-tle), (down-) sit(-down, still, -ting down, -ting (place) -uate), take, tarry."},
"H3428":{"lemma":"יֶשֶׁבְאָב","xlit":"Yeshebʼâb","pron":"yeh-sheb-awb'","derivation":"from H3427 (יָשַׁב) and H1 (אָב); seat of (his) father;","strongs_def":"Jeshebab, an Israelite","kjv_def":"Jeshebeab."},
"H3429":{"lemma":"יֹשֵׁב בַּשֶּׁבֶת","xlit":"Yôshêb bash-Shebeth","pron":"yo-shabe' bash-sheh'-beth","derivation":"from the active participle of H3427 (יָשַׁב) and H7674 (שֶׁבֶת), with a preposition and the article interposed; sitting in the seat;","strongs_def":"Joshebbash-Shebeth, an Israelite","kjv_def":"that sat in the seat."},
"H3430":{"lemma":"יִשְׁבּוֹ בְּנֹב","xlit":"Yishbôw bᵉ-Nôb","pron":"yish-bo'beh-nobe","derivation":"from H3427 (יָשַׁב) and H5011 (נֹב), with a pronominal suffix and a preposition interposed; his dwelling (is) in Nob;","strongs_def":"Jishbo-be-Nob, a Philistine","kjv_def":"Ishbi-benob (from the margin)."},
"H3431":{"lemma":"יִשְׁבַּח","xlit":"Yishbach","pron":"yish-bakh'","derivation":"from H7623 (שָׁבַח); he will praise;","strongs_def":"Jishbach, an Israelite","kjv_def":"Ishbah."},
"H3432":{"lemma":"יָשֻׁבִי","xlit":"Yâshubîy","pron":"yaw-shoo-bee'","derivation":"patronymically from H3437 (יָשׁוּב);","strongs_def":"a Jashubite, or descendant of Jashub","kjv_def":"Jashubites."},
"H3433":{"lemma":"יָשֻׁבִי לֶחֶם","xlit":"Yâshubîy Lechem","pron":"yaw-shoo-bee' leh'-khem","derivation":"from H7725 (שׁוּב) and H3899 (לֶחֶם); returner of bread;","strongs_def":"Jashubi-Lechem, an Israelite","kjv_def":"Jashubi-lehem. (Prob. the text should be pointed יֹשְׁבֵי לֶחֶם and rendered '(they were) inhabitants of Lechem,' i.e. of Bethlehem (by contraction). Compare H3902 (לַחְמִי))."},
"H3434":{"lemma":"יָשׇׁבְעָם","xlit":"Yâshobʻâm","pron":"yaw-shob-awm'","derivation":"from H7725 (שׁוּב) and H5971 (עַם); people will return;","strongs_def":"Jashobam, the name of two or three Israelites","kjv_def":"Jashobeam."},
"H3435":{"lemma":"יִשְׁבָּק","xlit":"Yishbâq","pron":"yish-bawk'","derivation":"from an unused root corresponding to H7662 (שְׁבַק); he will leave;","strongs_def":"Jishbak, a son of Abraham","kjv_def":"Ishbak."},
"H3436":{"lemma":"יׇשְׁבְּקָשָׁה","xlit":"Yoshbᵉqâshâh","pron":"yosh-bek-aw-shaw'","derivation":"from H3427 (יָשַׁב) and H7186 (קָשֶׁה); a hard seat;","strongs_def":"Joshbekashah, an Israelite","kjv_def":"Joshbekashah."},
"H3437":{"lemma":"יָשׁוּב","xlit":"Yâshûwb","pron":"yaw-shoob'","derivation":"or יָשִׁיב; from H7725 (שׁוּב); he will return;","strongs_def":"Jashub, the name of two Israelites","kjv_def":"Jashub."},
"H3438":{"lemma":"יִשְׁוָה","xlit":"Yishvâh","pron":"yish-vaw'","derivation":"from H7737 (שָׁוָה); he will level;","strongs_def":"Jishvah, an Israelite","kjv_def":"Ishvah, Isvah."},
"H3439":{"lemma":"יְשׁוֹחָיָה","xlit":"Yᵉshôwchâyâh","pron":"yesh-o-khaw-yaw'","derivation":"from the same as H3445 (יֶשַׁח) and H3050 (יָהּ); Jah will empty;","strongs_def":"Jeshochajah, an Israelite","kjv_def":"Jeshoaiah."},
"H3440":{"lemma":"יִשְׁוִי","xlit":"Yishvîy","pron":"yish-vee'","derivation":"from H7737 (שָׁוָה); level;","strongs_def":"Jishvi, the name of two Israelites","kjv_def":"Ishuai, Ishvi, Isui, Jesui."},
"H3441":{"lemma":"יִשְׁוִי","xlit":"Yishvîy","pron":"yish-vee'","derivation":"patronymically from H3440 (יִשְׁוִי);","strongs_def":"a Jishvite (collectively) or descendants of Jishvi","kjv_def":"Jesuites."},
"H3442":{"lemma":"יֵשׁוּעַ","xlit":"Yêshûwaʻ","pron":"yay-shoo'-ah","derivation":"for H3091 (יְהוֹשׁוּעַ); he will save;","strongs_def":"Jeshua, the name of ten Israelites, also of a place in Palestine","kjv_def":"Jeshua."},
"H3443":{"lemma":"יֵשׁוּעַ","xlit":"Yêshûwaʻ","pron":"yay-shoo'-ah","derivation":"(Aramaic) corresponding to H3442 (יֵשׁוּעַ)","strongs_def":"{Jeshua, the name of ten Israelites, also of a place in Palestine}","kjv_def":"Jeshua."},
"H3444":{"lemma":"יְשׁוּעָה","xlit":"yᵉshûwʻâh","pron":"yesh-oo'-aw","derivation":"feminine passive participle of H3467 (יָשַׁע);","strongs_def":"something saved, i.e. (abstractly) deliverance; hence, aid, victory, prosperity","kjv_def":"deliverance, health, help(-ing), salvation, save, saving (health), welfare."},
"H3445":{"lemma":"יֶשַׁח","xlit":"yeshach","pron":"yeh'-shakh","derivation":"from an unused root meaning to gape (as the empty stomach);","strongs_def":"hunger","kjv_def":"casting down."},
"H3446":{"lemma":"יִשְׂחָק","xlit":"Yischâq","pron":"yis-khawk'","derivation":"from H7831 (שַׁחֲצוֹם); he will laugh;","strongs_def":"Jischak, the heir of Abraham","kjv_def":"Isaac. Compare H3327 (יִצְחָק)."},
"H3447":{"lemma":"יָשַׁט","xlit":"yâshaṭ","pron":"yaw-shat'","derivation":"a primitive root;","strongs_def":"to extend","kjv_def":"hold out."},
"H3448":{"lemma":"יִשַׁי","xlit":"Yishay","pron":"yee-shah'-ee","derivation":"by Aramaic אִישַׁי; from the same as H3426 (יֵשׁ); extant;","strongs_def":"Jishai, David's father","kjv_def":"Jesse."},
"H3449":{"lemma":"יִשִּׁיָּה","xlit":"Yishshîyâh","pron":"yish-shee-yaw'","derivation":"or יִשִּׁיָּהוּ; from H5383 (נָשָׁה) and H3050 (יָהּ); Jah will lend;","strongs_def":"Jishshijah, the name of five Israelites","kjv_def":"Ishiah, Isshiah, Ishijah, Jesiah."},
"H3450":{"lemma":"יְשִׁימָאֵל","xlit":"Yᵉshîymâʼêl","pron":"yes-eem-aw-ale'","derivation":"from H7760 (שׂוּם) and H410 (אֵל); God will place;","strongs_def":"Jesimael, an Israelite","kjv_def":"Jesimael."},
"H3451":{"lemma":"יְשִׁימַה","xlit":"yᵉshîymah","pron":"yesh-ee-maw'","derivation":"from H3456 (יָשַׁם);","strongs_def":"desolation","kjv_def":"let death seize (from the margin)."},
"H3452":{"lemma":"יְשִׁימוֹן","xlit":"yᵉshîymôwn","pron":"yesh-ee-mone'","derivation":"from H3456 (יָשַׁם);","strongs_def":"a desolation","kjv_def":"desert, Jeshimon, solitary, wilderness."},
"H3453":{"lemma":"יָשִׁישׁ","xlit":"yâshîysh","pron":"yaw-sheesh'","derivation":"from H3486 (יָשֵׁשׁ);","strongs_def":"an old man","kjv_def":"(very) aged (man), ancient, very old."},
"H3454":{"lemma":"יְשִׁישָׁי","xlit":"Yᵉshîyshây","pron":"yesh-ee-shah'-ee","derivation":"from H3453 (יָשִׁישׁ); aged;","strongs_def":"Jeshishai, an Israelite","kjv_def":"Jeshishai."},
"H3455":{"lemma":"יָשַׂם","xlit":"yâsam","pron":"yaw-sam'","derivation":"a primitive root;","strongs_def":"to place; intransitively, to be placed","kjv_def":"be put (set)."},
"H3456":{"lemma":"יָשַׁם","xlit":"yâsham","pron":"yaw-sham'","derivation":"a primitive root;","strongs_def":"to lie waste","kjv_def":"be desolate."},
"H3457":{"lemma":"יִשְׁמָא","xlit":"Yishmâʼ","pron":"yish-maw'","derivation":"from H3456 (יָשַׁם); desolate;","strongs_def":"Jishma, an Israelite","kjv_def":"Ishma."},
"H3458":{"lemma":"יִשְׁמָעֵאל","xlit":"Yishmâʻêʼl","pron":"yish-maw-ale'","derivation":"from H8085 (שָׁמַע) and H410 (אֵל); God will hear;","strongs_def":"Jishmael, the name of Abraham's oldest son, and of five Israelites","kjv_def":"Ishmael."},
"H3459":{"lemma":"יִשְׁמָעֵאלִי","xlit":"Yishmâʻêʼlîy","pron":"yish-maw-ay-lee'","derivation":"patronymically from H3458 (יִשְׁמָעֵאל);","strongs_def":"a Jishmaelite or descendant of Jishmael","kjv_def":"Ishmaelite."},
"H3460":{"lemma":"יִשְׁמַעְיָה","xlit":"Yishmaʻyâh","pron":"yish-mah-yaw'","derivation":"or יִשְׁמַעְיָהוּlemma יְשְׁמַעְיָהוּ first vowel, corrected to יִשְׁמַעְיָהוּ; from H8085 (שָׁמַע) and H3050 (יָהּ); Jah will hear;","strongs_def":"Jishmajah, the name of two Israelites","kjv_def":"Ishmaiah."},
"H3461":{"lemma":"יִשְׁמְרַי","xlit":"Yishmᵉray","pron":"yish-mer-ah'-ee","derivation":"from H8104 (שָׁמַר); preservative;","strongs_def":"Jishmerai, an Israelite","kjv_def":"Ishmerai."},
"H3462":{"lemma":"יָשֵׁן","xlit":"yâshên","pron":"yaw-shane'","derivation":"a primitive root;","strongs_def":"properly, to be slack or languid, i.e. (by implication) sleep (figuratively, to die); also to grow old, stale or inveterate","kjv_def":"old (store), remain long, (make to) sleep."},
"H3463":{"lemma":"יָשֵׁן","xlit":"yâshên","pron":"yaw-shane'","derivation":"from H3462 (יָשֵׁן);","strongs_def":"sleepy","kjv_def":"asleep, (one out of) sleep(-eth, -ing), slept."},
"H3464":{"lemma":"יָשֵׁן","xlit":"Yâshên","pron":"yaw-shane'","derivation":"the same as H3463 (יָשֵׁן);","strongs_def":"Jashen, an Israelite","kjv_def":"Jashen."},
"H3465":{"lemma":"יָשָׁן","xlit":"yâshân","pron":"yaw-shawn'","derivation":"from H3462 (יָשֵׁן);","strongs_def":"old","kjv_def":"old."},
"H3466":{"lemma":"יְשָׁנָה","xlit":"Yᵉshânâh","pron":"yesh-aw-naw'","derivation":"feminine of H3465 (יָשָׁן);","strongs_def":"Jeshanah, a place in Palestine","kjv_def":"Jeshanah."},
"H3467":{"lemma":"יָשַׁע","xlit":"yâshaʻ","pron":"yaw-shah'","derivation":"a primitive root;","strongs_def":"properly, to be open, wide or free, i.e. (by implication) to be safe; causatively, to free or succor","kjv_def":"[idiom] at all, avenging, defend, deliver(-er), help, preserve, rescue, be safe, bring (having) salvation, save(-iour), get victory."},
"H3468":{"lemma":"יֶשַׁע","xlit":"yeshaʻ","pron":"yeh'-shah","derivation":"or יֵשַׁע; from H3467 (יָשַׁע);","strongs_def":"liberty, deliverance, prosperity","kjv_def":"safety, salvation, saving."},
"H3469":{"lemma":"יִשְׁעִי","xlit":"Yishʻîy","pron":"yish-ee'","derivation":"from H3467 (יָשַׁע); saving;","strongs_def":"Jishi, the name of four Israelites","kjv_def":"Ishi."},
"H3470":{"lemma":"יְשַׁעְיָה","xlit":"Yᵉshaʻyâh","pron":"yesh-ah-yaw'","derivation":"or יְשַׁעְיָהוּ; from H3467 (יָשַׁע) and H3050 (יָהּ); Jah has saved;","strongs_def":"Jeshajah, the name of seven Israelites","kjv_def":"Isaiah, Jesaiah, Jeshaiah."},
"H3471":{"lemma":"יָשְׁפֵה","xlit":"yâshᵉphêh","pron":"yaw-shef-ay'","derivation":"from an unused root meaning to polish;","strongs_def":"a gem supposed to be jasper (from the resemblance in name)","kjv_def":"jasper."},
"H3472":{"lemma":"יִשְׁפָּה","xlit":"Yishpâh","pron":"yish-paw'","derivation":"perhaps from H8192 (שָׁפָה); he will scratch;","strongs_def":"Jishpah, an Israelite","kjv_def":"Ispah."},
"H3473":{"lemma":"יִשְׁפָּן","xlit":"Yishpân","pron":"yish-pawn'","derivation":"probably from the same as H8227 (שָׁפָן); he will hide;","strongs_def":"Jishpan, an Israelite","kjv_def":"Ishpan."},
"H3474":{"lemma":"יָשַׁר","xlit":"yâshar","pron":"yaw-shar'","derivation":"a primitive root;","strongs_def":"to be straight or even; figuratively, to be (causatively, to make) right, pleasant, prosperous","kjv_def":"direct, fit, seem good (meet), [phrase] please (will), be (esteem, go) right (on), bring (look, make, take the) straight (way), be upright(-ly)."},
"H3475":{"lemma":"יֵשֶׁר","xlit":"Yêsher","pron":"yay'-sher","derivation":"from H3474 (יָשַׁר); the right;","strongs_def":"Jesher, an Israelite","kjv_def":"Jesher."},
"H3476":{"lemma":"יֹשֶׁר","xlit":"yôsher","pron":"yo'-sher","derivation":"from H3474 (יָשַׁר);","strongs_def":"the right","kjv_def":"equity, meet, right, upright(-ness)."},
"H3477":{"lemma":"יָשָׁר","xlit":"yâshâr","pron":"yaw-shawr'","derivation":"from H3474 (יָשַׁר);","strongs_def":"straight (literally or figuratively)","kjv_def":"convenient, equity, Jasher, just, meet(-est), [phrase] pleased well right(-eous), straight, (most) upright(-ly, -ness)."},
"H3478":{"lemma":"יִשְׂרָאֵל","xlit":"Yisrâʼêl","pron":"yis-raw-ale'","derivation":"from H8280 (שָׂרָה) and H410 (אֵל); he will rule as God;","strongs_def":"Jisrael, a symbolical name of Jacob; also (typically) of his posterity","kjv_def":"Israel."},
"H3479":{"lemma":"יִשְׂרָאֵל","xlit":"Yisrâʼêl","pron":"yis-raw-ale'","derivation":"(Aramaic) corresponding to H3478 (יִשְׂרָאֵל)","strongs_def":"{Jisrael, a symbolical name of Jacob; also (typically) of his posterity}","kjv_def":"Israel."},
"H3480":{"lemma":"יְשַׂרְאֵלָה","xlit":"Yᵉsarʼêlâh","pron":"yes-ar-ale'-aw","derivation":"by variation from H3477 (יָשָׁר) and H410 (אֵל) with directive enclitic; right towards God;","strongs_def":"Jesarelah, an Israelite","kjv_def":"Jesharelah."},
"H3481":{"lemma":"יִשְׂרְאֵלִי","xlit":"Yisrᵉʼêlîy","pron":"yis-reh-ay-lee'","derivation":"patronymically from H3478 (יִשְׂרָאֵל);","strongs_def":"a Jisreelite or descendant of Jisrael","kjv_def":"of Israel, Israelite."},
"H3482":{"lemma":"יִשְׂרְאֵלִית","xlit":"Yisrᵉʼêlîyth","pron":"yis-reh-ay-leeth'","derivation":"feminine of H3481 (יִשְׂרְאֵלִי);","strongs_def":"a Jisreelitess or female descendant of Jisrael","kjv_def":"Israelitish."},
"H3483":{"lemma":"יִשְׁרָה","xlit":"yishrâh","pron":"yish-raw'","derivation":"feminine of H3477 (יָשָׁר);","strongs_def":"rectitude","kjv_def":"uprightness."},
"H3484":{"lemma":"יְשֻׁרוּן","xlit":"Yᵉshurûwn","pron":"yesh-oo-roon'","derivation":"from H3474 (יָשַׁר); upright;","strongs_def":"Jeshurun, a symbolic name for Israel","kjv_def":"Jeshurun."},
"H3485":{"lemma":"יִשָּׂשכָר","xlit":"Yissâˢkâr","pron":"yis-saw-kawr'","derivation":"(strictly yis-saws-kawr'); from H5375 (נָשָׂא) and H7939 (שָׂכָר); he will bring a reward;","strongs_def":"Jissaskar, a son of Jacob","kjv_def":"Issachar."},
"H3486":{"lemma":"יָשֵׁשׁ","xlit":"yâshêsh","pron":"yaw-shaysh'","derivation":"from an unused root meaning to blanch;","strongs_def":"gray-haired, i.e. an aged man","kjv_def":"stoop for age."},
"H3487":{"lemma":"יַת","xlit":"yath","pron":"yath","derivation":"(Aramaic) corresponding to H853 (אֵת);","strongs_def":"a sign of the object of a verb","kjv_def":"[phrase] whom."},
"H3488":{"lemma":"יְתִב","xlit":"yᵉthib","pron":"yeth-eeb'","derivation":"(Aramaic) corresponding to H3427 (יָשַׁב);","strongs_def":"to sit or dwell","kjv_def":"dwell, (be) set, sit."},
"H3489":{"lemma":"יָתֵד","xlit":"yâthêd","pron":"yaw-thade'","derivation":"from an unused root meaning to pin through or fast;","strongs_def":"a peg","kjv_def":"nail, paddle, pin, stake."},
"H3490":{"lemma":"יָתוֹם","xlit":"yâthôwm","pron":"yaw-thome'","derivation":"from an unused root meaning to be lonely;","strongs_def":"a bereaved person","kjv_def":"fatherless (child), orphan."},
"H3491":{"lemma":"יָתוּר","xlit":"yâthûwr","pron":"yaw-thoor'","derivation":"passive participle of H3498 (יָתַר);","strongs_def":"properly, what is left, i.e. (by implication) a gleaning","kjv_def":"range."},
"H3492":{"lemma":"יַתִּיר","xlit":"Yattîyr","pron":"yat-teer'","derivation":"from H3498 (יָתַר); redundant;","strongs_def":"Jattir, a place in Palestine","kjv_def":"Jattir."},
"H3493":{"lemma":"יַתִּיר","xlit":"yattîyr","pron":"yat-teer'","derivation":"(Aramaic) corresponding to H3492 (יַתִּיר);","strongs_def":"preeminent; as an adverb, very","kjv_def":"exceeding(-ly), excellent."},
"H3494":{"lemma":"יִתְלָה","xlit":"Yithlâh","pron":"yith-law'","derivation":"probably from H8518 (תָּלָה); it will hang, i.e. be high;","strongs_def":"Jithlah, a place in Palestine","kjv_def":"Jethlah."},
"H3495":{"lemma":"יִתְמָה","xlit":"Yithmâh","pron":"yith-maw'","derivation":"from the same as H3490 (יָתוֹם); orphanage;","strongs_def":"Jithmah, an Israelite","kjv_def":"Ithmah."},
"H3496":{"lemma":"יַתְנִיאֵל","xlit":"Yathnîyʼêl","pron":"yath-nee-ale'","derivation":"from an unused root meaning to endure, and H410 (אֵל); continued of God;","strongs_def":"Jathniel, an Israelite","kjv_def":"Jathniel."},
"H3497":{"lemma":"יִתְנָן","xlit":"Yithnân","pron":"yith-nawn'","derivation":"from the same as H8577 (תַּנִּין); extensive;","strongs_def":"Jithnan, a place in Palestine","kjv_def":"Ithnan."},
"H3498":{"lemma":"יָתַר","xlit":"yâthar","pron":"yaw-thar'","derivation":"a primitive root;","strongs_def":"to jut over or exceed; by implication, to excel; (intransitively) to remain or be left; causatively, to leave, cause to abound, preserve","kjv_def":"excel, leave (a remnant), left behind, too much, make plenteous, preserve, (be, let) remain(-der, -ing, -nant), reserve, residue, rest."},
"H3499":{"lemma":"יֶתֶר","xlit":"yether","pron":"yeh'-ther","derivation":"from H3498 (יָתַר);","strongs_def":"properly, an overhanging, i.e. (by implication) an excess, superiority, remainder; also a small rope (as hanging free)","kjv_def":"[phrase] abundant, cord, exceeding, excellancy(-ent), what they leave, that hath left, plentifully, remnant, residue, rest, string, with."},
"H3500":{"lemma":"יֶתֶר","xlit":"Yether","pron":"yeh'-ther","derivation":"the same as H3499 (יֶתֶר);","strongs_def":"Jether, the name of five or six Israelites and of one Midianite","kjv_def":"Jether, Jethro. Compare H3503 (יִתְרוֹ)."},
"H3501":{"lemma":"יִתְרָא","xlit":"Yithrâʼ","pron":"yith-raw'","derivation":"by variation for H3502 (יִתְרָה);","strongs_def":"Jithra, an Israelite (or Ishmaelite)","kjv_def":"Ithra."},
"H3502":{"lemma":"יִתְרָה","xlit":"yithrâh","pron":"yith-raw'","derivation":"feminine of H3499 (יֶתֶר);","strongs_def":"properly, excellence, i.e. (by implication) wealth","kjv_def":"abundance, riches."},
"H3503":{"lemma":"יִתְרוֹ","xlit":"Yithrôw","pron":"yith-ro'","derivation":"from H3499 (יֶתֶר) with pron. suffix; his excellence;","strongs_def":"Jethro, Moses' father-in-law","kjv_def":"Jethro. Compare H3500 (יֶתֶר)."},
"H3504":{"lemma":"יִתְרוֹן","xlit":"yithrôwn","pron":"yith-rone'","derivation":"from H3498 (יָתַר);","strongs_def":"preeminence, gain","kjv_def":"better, excellency(-leth), profit(-able)."},
"H3505":{"lemma":"יִתְרִי","xlit":"Yithrîy","pron":"yith-ree'","derivation":"patronymically from H3500 (יֶתֶר);","strongs_def":"a Jithrite or descendant of Jether","kjv_def":"Ithrite."},
"H3506":{"lemma":"יִתְרָן","xlit":"Yithrân","pron":"yith-rawn'","derivation":"from H3498 (יָתַר); excellent;","strongs_def":"Jithran, the name of an Edomite and of an Israelite","kjv_def":"Ithran."},
"H3507":{"lemma":"יִתְרְעָם","xlit":"Yithrᵉʻâm","pron":"yith-reh-awm'","derivation":"from H3499 (יֶתֶר) and H5971 (עַם); excellence of people;","strongs_def":"Jithream, a son of David","kjv_def":"Ithream."},
"H3508":{"lemma":"יֹתֶרֶת","xlit":"yôthereth","pron":"yo-theh'-reth","derivation":"feminine active participle of H3498 (יָתַר);","strongs_def":"the lobe or flap of the liver (as if redundant or outhanging)","kjv_def":"caul."},
"H3509":{"lemma":"יְתֵת","xlit":"Yᵉthêth","pron":"yeh-thayth'","derivation":"of uncertain derivation;","strongs_def":"Jetheth, an Edomite","kjv_def":"Jetheth. k"},
"H3510":{"lemma":"כָּאַב","xlit":"kâʼab","pron":"kaw-ab'","derivation":"a primitive root;","strongs_def":"properly, to feel pain; by implication, to grieve; figuratively, to spoil","kjv_def":"grieving, mar, have pain, make sad (sore), (be) sorrowful."},
"H3511":{"lemma":"כְּאֵב","xlit":"kᵉʼêb","pron":"keh-abe'","derivation":"from H3510 (כָּאַב);","strongs_def":"suffering (physical or mental), adversity","kjv_def":"grief, pain, sorrow."},
"H3512":{"lemma":"כָּאָה","xlit":"kâʼâh","pron":"kaw-aw'","derivation":"a primitive root;","strongs_def":"to despond; causatively, to deject","kjv_def":"broken, be grieved, make sad."},
"H3513":{"lemma":"כָּבַד","xlit":"kâbad","pron":"kaw-bad'","derivation":"or כָּבֵד; a primitive root;","strongs_def":"to be heavy, i.e. in a bad sense (burdensome, severe, dull) or in a good sense (numerous, rich, honorable; causatively, to make weighty (in the same two senses)","kjv_def":"abounding with, more grievously afflict, boast, be chargeable, [idiom] be dim, glorify, be (make) glorious (things), glory, (very) great, be grievous, harden, be (make) heavy, be heavier, lay heavily, (bring to, come to, do, get, be had in) honour (self), (be) honourable (man), lade, [idiom] more be laid, make self many, nobles, prevail, promote (to honour), be rich, be (go) sore, stop."},
"H3514":{"lemma":"כֹּבֶד","xlit":"kôbed","pron":"ko'-bed","derivation":"from H3513 (כָּבַד);","strongs_def":"weight, multitude, vehemence","kjv_def":"grievousness, heavy, great number."},
"H3515":{"lemma":"כָּבֵד","xlit":"kâbêd","pron":"kaw-bade'","derivation":"from H3513 (כָּבַד);","strongs_def":"heavy; figuratively in a good sense (numerous) or in a bad sense (severe, difficult, stupid)","kjv_def":"(so) great, grievous, hard(-ened), (too) heavy(-ier), laden, much, slow, sore, thick."},
"H3516":{"lemma":"כָּבֵד","xlit":"kâbêd","pron":"kaw-bade'","derivation":"the same as H3515 (כָּבֵד);","strongs_def":"the liver (as the heaviest of the viscera)","kjv_def":"liver."},
"H3517":{"lemma":"כְּבֵדֻת","xlit":"kᵉbêduth","pron":"keb-ay-dooth'","derivation":"feminine of H3515 (כָּבֵד);","strongs_def":"difficulty","kjv_def":"[idiom] heavily."},
"H3518":{"lemma":"כָּבָה","xlit":"kâbâh","pron":"kaw-baw'","derivation":"a primitive root;","strongs_def":"to expire or (causatively) to extinguish (fire, light, anger)","kjv_def":"go (put) out, quench."},
"H3519":{"lemma":"כָּבוֹד","xlit":"kâbôwd","pron":"kaw-bode'","derivation":"rarely כָּבֹד; from H3513 (כָּבַד);","strongs_def":"properly, weight, but only figuratively in a good sense, splendor or copiousness","kjv_def":"glorious(-ly), glory, honour(-able)."},
"H3520":{"lemma":"כְּבוּדָּה","xlit":"kᵉbûwddâh","pron":"keb-ood-daw'","derivation":"irregular feminine passive participle of H3513 (כָּבַד);","strongs_def":"weightiness, i.e. magnificence, wealth","kjv_def":"carriage, all glorious, stately."},
"H3521":{"lemma":"כָּבוּל","xlit":"Kâbûwl","pron":"kaw-bool'","derivation":"from the same as H3525 (כֶּבֶל) in the sense of limitation; sterile;","strongs_def":"Cabul, the name of two places in Palestine","kjv_def":"Cabul."},
"H3522":{"lemma":"כַּבּוֹן","xlit":"Kabbôwn","pron":"kab-bone'","derivation":"from an unused root meaning to heap up; hilly;","strongs_def":"Cabon, a place in Palestine","kjv_def":"Cabbon."},
"H3523":{"lemma":"כְּבִיר","xlit":"kᵉbîyr","pron":"keb-eer","derivation":"from H3527 (כָּבַר) in the original sense of plaiting;","strongs_def":"a matrass (of intertwined materials)","kjv_def":"pillow."},
"H3524":{"lemma":"כַּבִּיר","xlit":"kabbîyr","pron":"kab-beer'","derivation":"from H3527 (כָּבַר);","strongs_def":"vast, whether in extent (figuratively, of power, mighty; of time, aged), or in number, many","kjv_def":"[phrase] feeble, mighty, most, much, strong, valiant."},
"H3525":{"lemma":"כֶּבֶל","xlit":"kebel","pron":"keh'-bel","derivation":"from an unused root meaning to twine or braid together;","strongs_def":"a fetter","kjv_def":"fetter."},
"H3526":{"lemma":"כָּבַס","xlit":"kâbaç","pron":"kaw-bas'","derivation":"a primitive root;","strongs_def":"to trample; hence, to wash (properly, by stamping with the feet), whether literal (including the fulling process) or figurative","kjv_def":"fuller, wash(-ing)."},
"H3527":{"lemma":"כָּבַר","xlit":"kâbar","pron":"kaw-bar'","derivation":"a primitive root;","strongs_def":"properly, to plait together, i.e. (figuratively) to augment (especially in number or quantity, to accumulate)","kjv_def":"in abundance, multiply."},
"H3528":{"lemma":"כְּבָר","xlit":"kᵉbâr","pron":"keb-awr'","derivation":"from H3527 (כָּבַר);","strongs_def":"properly, extent of time, i.e. a great while; hence, long ago, formerly, hitherto","kjv_def":"already, (seeing that which), now."},
"H3529":{"lemma":"כְּבָר","xlit":"Kᵉbâr","pron":"keb-awr'","derivation":"the same as H3528 (כְּבָר); length;","strongs_def":"Kebar, a river of Mesopotamia","kjv_def":"Chebar. Compare H2249 (חָבוֹר)."},
"H3530":{"lemma":"כִּבְרָה","xlit":"kibrâh","pron":"kib-raw'","derivation":"feminine of H3528 (כְּבָר);","strongs_def":"properly, length, i.e. a measure (of uncertain dimension)","kjv_def":"[idiom] little."},
"H3531":{"lemma":"כְּבָרָה","xlit":"kᵉbârâh","pron":"keb-aw-raw'","derivation":"from H3527 (כָּבַר) in its original sense;","strongs_def":"a sieve (as netted)","kjv_def":"sieve."},
"H3532":{"lemma":"כֶּבֶשׂ","xlit":"kebes","pron":"keh-bes'","derivation":"from an unused root meaning to dominate;","strongs_def":"a ram (just old enough to butt)","kjv_def":"lamb, sheep."},
"H3533":{"lemma":"כָּבַשׁ","xlit":"kâbash","pron":"kaw-bash'","derivation":"a primitive root;","strongs_def":"to tread down; hence, negatively, to disregard; positively, to conquer, subjugate, violate","kjv_def":"bring into bondage, force, keep under, subdue, bring into subjection."},
"H3534":{"lemma":"כֶּבֶשׁ","xlit":"kebesh","pron":"keh'-besh","derivation":"from H3533 (כָּבַשׁ);","strongs_def":"a footstool (as trodden upon)","kjv_def":"footstool."},
"H3535":{"lemma":"כִּבְשָׂה","xlit":"kibsâh","pron":"kib-saw'","derivation":"or כַּבְשָׂה; feminine of H3532 (כֶּבֶשׂ);","strongs_def":"a ewe","kjv_def":"(ewe) lamb."},
"H3536":{"lemma":"כִּבְשָׁן","xlit":"kibshân","pron":"kib-shawn'","derivation":"from H3533 (כָּבַשׁ);","strongs_def":"a smelting furnace (as reducing metals)","kjv_def":"furnace."},
"H3537":{"lemma":"כַּד","xlit":"kad","pron":"kad","derivation":"from an unused root meaning to deepen;","strongs_def":"properly, a pail; but generally of earthenware; a jar  for domestic purposes","kjv_def":"barrel, pitcher."},
"H3538":{"lemma":"כְּדַב","xlit":"kᵉdab","pron":"ked-ab'","derivation":"(Aramaic) from a root corresponding to H3576 (כָּזַב);","strongs_def":"false","kjv_def":"lying."},
"H3539":{"lemma":"כַּדְכֹּד","xlit":"kadkôd","pron":"kad-kode'","derivation":"from the same as H3537 (כַּד) in the sense of striking fire from a metal forged;","strongs_def":"a sparkling gem, probably the ruby","kjv_def":"agate."},
"H3540":{"lemma":"כְּדׇרְלָעֹמֶר","xlit":"Kᵉdorlâʻômer","pron":"ked-or-law-o'-mer","derivation":"of foreign origin;","strongs_def":"Kedorlaomer, an early Persian king","kjv_def":"Chedorlaomer."},
"H3541":{"lemma":"כֹּה","xlit":"kôh","pron":"ko","derivation":"from the prefix k and H1931 (הוּא);","strongs_def":"properly, like this, i.e. by implication, (of manner) thus (or so); also (of place) here (or hither); or (of time) now","kjv_def":"also, here, + hitherto, like, on the other side, so (and much), such, on that manner, (on) this (manner, side, way, way and that way), + mean while, yonder."},
"H3542":{"lemma":"כָּה","xlit":"kâh","pron":"kaw","derivation":"(Aramaic) corresponding to H3541 (כֹּה)","strongs_def":"{properly, like this, i.e. by implication, (of manner) thus (or so); also (of place) here (or hither); or (of time) now}","kjv_def":"hitherto."},
"H3543":{"lemma":"כָּהָה","xlit":"kâhâh","pron":"kaw-haw'","derivation":"a primitive root;","strongs_def":"to be weak, i.e. (figuratively) to despond (causatively, rebuke), or (of light, the eye) to grow dull","kjv_def":"darken, be dim, fail, faint, restrain, [idiom] utterly."},
"H3544":{"lemma":"כֵּהֶה","xlit":"kêheh","pron":"kay-heh'","derivation":"from H3543 (כָּהָה);","strongs_def":"feeble, obscure","kjv_def":"somewhat dark, darkish, wax dim, heaviness, smoking."},
"H3545":{"lemma":"כֵּהָה","xlit":"kêhâh","pron":"kay-haw'","derivation":"feminine of H3544 (כֵּהֶה);","strongs_def":"properly, a weakening; figuratively, alleviation, i.e. cure","kjv_def":"healing."},
"H3546":{"lemma":"כְּהַל","xlit":"kᵉhal","pron":"keh-hal'","derivation":"(Aramaic) a root corresponding to and H3557 (כּוּל);","strongs_def":"to be able","kjv_def":"be able, could."},
"H3547":{"lemma":"כָּהַן","xlit":"kâhan","pron":"kaw-han'","derivation":"a primitive root, apparently meaning to mediate in religious services; but used only as denominative from H3548 (כֹּהֵן);","strongs_def":"to officiate as a priest; figuratively, to put on regalia","kjv_def":"deck, be (do the office of a, execute the, minister in the) priest('s office)."},
"H3548":{"lemma":"כֹּהֵן","xlit":"kôhên","pron":"ko-hane'","derivation":"active participle of H3547 (כָּהַן);","strongs_def":"literally one officiating, a priest; also (by courtesy) an acting priest (although a layman)","kjv_def":"chief ruler, [idiom] own, priest, prince, principal officer."},
"H3549":{"lemma":"כָּהֵן","xlit":"kâhên","pron":"kaw-hane'","derivation":"(Aramaic) corresponding to H3548 (כֹּהֵן)","strongs_def":"{one officiating, a priest; also (by courtesy) an acting priest (although a layman)}","kjv_def":"priest."},
"H3550":{"lemma":"כְּהֻנָּה","xlit":"kᵉhunnâh","pron":"keh-hoon-naw'","derivation":"from H3547 (כָּהַן);","strongs_def":"priesthood","kjv_def":"priesthood, priest's office."},
"H3551":{"lemma":"כַּו","xlit":"kav","pron":"kav","derivation":"(Aramaic) from a root corresponding to H3854 (לַהַג) in the sense of piercing;","strongs_def":"a window (as a perforation)","kjv_def":"window."},
"H3552":{"lemma":"כּוּב","xlit":"Kûwb","pron":"koob","derivation":"of foreign derivation;","strongs_def":"Kub, a country near Egypt","kjv_def":"Chub."},
"H3553":{"lemma":"כּוֹבַע","xlit":"kôwbaʻ","pron":"ko'-bah","derivation":"from an unused root meaning to be high or rounded;","strongs_def":"a helmet (as arched)","kjv_def":"helmet. Compare H6959 (קוֹבַע)."},
"H3554":{"lemma":"כָּוָה","xlit":"kâvâh","pron":"kaw-vaw'","derivation":"a primitive root;","strongs_def":"properly, to prick or penetrate; hence, to blister (as smarting or eating into)","kjv_def":"burn."},
"H3555":{"lemma":"כְּוִיָּה","xlit":"kᵉvîyâh","pron":"kev-ee-yaw'","derivation":"from H3554 (כָּוָה);","strongs_def":"a branding","kjv_def":"burning."},
"H3556":{"lemma":"כּוֹכָב","xlit":"kôwkâb","pron":"ko-kawb'","derivation":"probably from the same as H3522 (כַּבּוֹן) (in the sense of rolling) or H3554 (כָּוָה) (in the sense of blazing);","strongs_def":"a star (as round or as shining); figuratively, a prince","kjv_def":"star(-gazer)."},
"H3557":{"lemma":"כּוּל","xlit":"kûwl","pron":"kool","derivation":"a primitive root;","strongs_def":"properly, to keep in; hence, to measure; figuratively, to maintain (in various senses)","kjv_def":"(be able to, can) abide, bear, comprehend, contain, feed, forbearing, guide, hold(-ing in), nourish(-er), be present, make provision, receive, sustain, provide sustenance (victuals)."},
"H3558":{"lemma":"כּוּמָז","xlit":"kûwmâz","pron":"koo-mawz'","derivation":"from an unused root meaning to store away;","strongs_def":"a jewel (probably gold beads)","kjv_def":"tablet."},
"H3559":{"lemma":"כּוּן","xlit":"kûwn","pron":"koon","derivation":"a primitive root;","strongs_def":"properly, to be erect (i.e. stand perpendicular); hence (causatively) to set up, in a great variety of applications, whether literal (establish, fix, prepare, apply), or figurative (appoint, render sure, proper or prosperous)","kjv_def":"certain(-ty), confirm, direct, faithfulness, fashion, fasten, firm, be fitted, be fixed, frame, be meet, ordain, order, perfect, (make) preparation, prepare (self), provide, make provision, (be, make) ready, right, set (aright, fast, forth), be stable, (e-) stablish, stand, tarry, [idiom] very deed."},
"H3560":{"lemma":"כּוּן","xlit":"Kûwn","pron":"koon","derivation":"probably from H3559 (כּוּן); established;","strongs_def":"Kun, a place in Syria","kjv_def":"Chun."},
"H3561":{"lemma":"כַּוָּן","xlit":"kavvân","pron":"kav-vawn'","derivation":"from H3559 (כּוּן);","strongs_def":"something prepared, i.e. a sacrificial wafer","kjv_def":"cake."},
"H3562":{"lemma":"כּוֹנַנְיָהוּ","xlit":"Kôwnanyâhûw","pron":"ko-nan-yaw'-hoo","derivation":"from H3559 (כּוּן) and H3050 (יָהּ); Jah has sustained;","strongs_def":"Conanjah, the name of two Israelites","kjv_def":"Conaniah, Cononiah. Compare H3663 (כְּנַנְיָה)."},
"H3563":{"lemma":"כּוֹס","xlit":"kôwç","pron":"koce","derivation":"from an unused root meaning to hold together;","strongs_def":"a cup (as a container), often figuratively, a lot (as if a potion); also some unclean bird, probably an owl (perhaps from the cup-like cavity of its eye)","kjv_def":"cup, (small) owl. Compare H3599 (כִּיס)."},
"H3564":{"lemma":"כּוּר","xlit":"kûwr","pron":"koor","derivation":"from an unused root meaning properly, to dig through;","strongs_def":"a pot or furnace (as if excavated)","kjv_def":"furnace. Compare H3600 (כִּיר)."},
"H3565":{"lemma":"כּוֹר עָשָׁן","xlit":"Kôwr ʻÂshân","pron":"kore aw-shawn'","derivation":"from H3564 (כּוּר) and H6227 (עָשָׁן); furnace of smoke;","strongs_def":"Cor-Ashan, a place in Palestine","kjv_def":"Chorashan."},
"H3566":{"lemma":"כּוֹרֶשׁ","xlit":"Kôwresh","pron":"ko'-resh","derivation":"or (Ezra 1:1 (last time),2) כֹּרֶשׁ; from the Persians;","strongs_def":"Koresh (or Cyrus), the Persian king","kjv_def":"Cyrus."},
"H3567":{"lemma":"כּוֹרֶשׁ","xlit":"Kôwresh","pron":"ko'-resh","derivation":"(Aramaic) corresponding to H3566 (כּוֹרֶשׁ)","strongs_def":"{Koresh (or Cyrus), the Persian king}","kjv_def":"Cyrus."},
"H3568":{"lemma":"כּוּשׁ","xlit":"Kûwsh","pron":"koosh","derivation":"probably of foreign origin;","strongs_def":"Cush (or Ethiopia), the name of a son of Ham, and of his territory; also of an Israelite","kjv_def":"Chush, Cush, Ethiopia."},
"H3569":{"lemma":"כּוּשִׁי","xlit":"Kûwshîy","pron":"koo-shee'","derivation":"patronymically from H3568 (כּוּשׁ);","strongs_def":"a Cushite, or descendant of Cush","kjv_def":"Cushi, Cushite, Ethiopian(-s)."},
"H3570":{"lemma":"כּוּשִׁי","xlit":"Kûwshîy","pron":"koo-shee'","derivation":"the same as H3569 (כּוּשִׁי);","strongs_def":"Cushi, the name of two Israelites","kjv_def":"Cushi."},
"H3571":{"lemma":"כּוּשִׁית","xlit":"Kûwshîyth","pron":"koo-sheeth'","derivation":"feminine of H3569 (כּוּשִׁי);","strongs_def":"a Cushite woman","kjv_def":"Ethiopian."},
"H3572":{"lemma":"כּוּשָׁן","xlit":"Kûwshân","pron":"koo-shawn'","derivation":"perhaps from H3568 (כּוּשׁ);","strongs_def":"Cushan, a region of Arabia","kjv_def":"Cushan."},
"H3573":{"lemma":"כּוּשַׁן רִשְׁעָתַיִם","xlit":"Kûwshan Rishʻâthayim","pron":"koo-shan' rish-awthah'-yim","derivation":"apparently from H3572 (כּוּשָׁן) and the dual of H7564 (רִשְׁעָה); Cushan of double wickedness;","strongs_def":"Cushan-Rishathajim, a Mesopotamian king","kjv_def":"Chushan-rishathayim."},
"H3574":{"lemma":"כּוֹשָׁרָה","xlit":"kôwshârâh","pron":"ko-shaw-raw'","derivation":"from H3787 (כָּשֵׁר);","strongs_def":"prosperity; in plural freedom","kjv_def":"[idiom] chain."},
"H3575":{"lemma":"כּוּת","xlit":"Kûwth","pron":"kooth","derivation":"or (feminine) כּוּתָה; of foreign origin;","strongs_def":"Cuth or Cuthah, a province of Assyria","kjv_def":"Cuth."},
"H3576":{"lemma":"כָּזַב","xlit":"kâzab","pron":"kaw-zab'","derivation":"a primitive root;","strongs_def":"to lie (i.e. deceive), literally or figuratively","kjv_def":"fail, (be found a, make a) liar, lie, lying, be in vain."},
"H3577":{"lemma":"כָּזָב","xlit":"kâzâb","pron":"kaw-zawb'","derivation":"from H3576 (כָּזַב);","strongs_def":"falsehood; literally (untruth) or figuratively (idol)","kjv_def":"deceitful, false, leasing, + liar, lie, lying."},
"H3578":{"lemma":"כֹּזְבָא","xlit":"Kôzᵉbâʼ","pron":"ko-zeb-aw'","derivation":"from H3576 (כָּזַב); fallacious;","strongs_def":"Cozeba, a place in Palestine","kjv_def":"Choseba."},
"H3579":{"lemma":"כֹּזְבִי","xlit":"Kôzᵉbîy","pron":"koz-bee'","derivation":"from H3576 (כָּזַב); false;","strongs_def":"Cozbi, a Midianitess","kjv_def":"Cozbi."},
"H3580":{"lemma":"כְּזִיב","xlit":"Kᵉzîyb","pron":"kez-eeb'","derivation":"from H3576 (כָּזַב); falsified;","strongs_def":"Kezib, a place in Palestine","kjv_def":"Chezib."},
"H3581":{"lemma":"כֹּחַ","xlit":"kôach","pron":"ko'-akh","derivation":"or (Daniel 11:6) כּוֹחַ; from an unused root meaning to be firm;","strongs_def":"vigor, literally (force, in a good or a bad sense) or figuratively (capacity, means, produce); also (from its hardiness) a large lizard","kjv_def":"ability, able, chameleon, force, fruits, might, power(-ful), strength, substance, wealth."},
"H3582":{"lemma":"כָּחַד","xlit":"kâchad","pron":"kaw-khad'","derivation":"a primitive root;","strongs_def":"to secrete, by act or word; hence (intensively) to destroy","kjv_def":"conceal, cut down (off), desolate, hide."},
"H3583":{"lemma":"כָּחַל","xlit":"kâchal","pron":"kaw-khal'","derivation":"a primitive root;","strongs_def":"to paint (with stibium)","kjv_def":"paint."},
"H3584":{"lemma":"כָּחַשׁ","xlit":"kâchash","pron":"kaw-khash'","derivation":"a primitive root;","strongs_def":"to be untrue, in word (to lie, feign, disown) or deed (to disappoint, fail, cringe)","kjv_def":"deceive, deny, dissemble, fail, deal falsely, be found liars, (be-) lie, lying, submit selves."},
"H3585":{"lemma":"כַּחַשׁ","xlit":"kachash","pron":"kakh'-ash","derivation":"from H3584 (כָּחַשׁ);","strongs_def":"literally a failure of flesh, i.e. emaciation; figuratively, hypocrisy","kjv_def":"leanness, lies, lying."},
"H3586":{"lemma":"כֶּחָשׁ","xlit":"kechâsh","pron":"kekh-awsh'","derivation":"from H3584 (כָּחַשׁ);","strongs_def":"faithless","kjv_def":"lying."},
"H3587":{"lemma":"כִּי","xlit":"kîy","pron":"kee","derivation":"from H3554 (כָּוָה);","strongs_def":"a brand or scar","kjv_def":"burning."},
"H3588":{"lemma":"כִּי","xlit":"kîy","pron":"kee","derivation":"a primitive particle (the full form of the prepositional prefix) indicating causal relations of all kinds, antecedent or consequent;","strongs_def":"(by implication) very widely used as a relative conjunction or adverb (as below); often largely modified by other particles annexed","kjv_def":"and, + (forasmuch, inasmuch, where-) as, assured(-ly), + but, certainly, doubtless, + else, even, + except, for, how, (because, in, so, than) that, + nevertheless, now, rightly, seeing, since, surely, then, therefore, + (al-) though, + till, truly, + until, when, whether, while, whom, yea, yet."},
"H3589":{"lemma":"כִּיד","xlit":"kîyd","pron":"keed","derivation":"from a primitive root meaning to strike;","strongs_def":"a crushing; figuratively, calamity","kjv_def":"destruction."},
"H3590":{"lemma":"כִידוֹד","xlit":"kîydôwd","pron":"kee-dode'","derivation":"from the same as H3589 (כִּיד) (compare H3539 (כַּדְכֹּד));","strongs_def":"properly, something struck off, i.e. a spark (as struck)","kjv_def":"spark."},
"H3591":{"lemma":"כִּידוֹן","xlit":"kîydôwn","pron":"kee-dohn'","derivation":"from the same as H3589 (כִּיד); (perhaps smaller than H2595 (חֲנִית)lemma חֲניִת second vowel, corrected to חֲנִית)","strongs_def":"properly, something to strike with, i.e. a dart","kjv_def":"lance, shield, spear, target."},
"H3592":{"lemma":"כִּידוֹן","xlit":"Kîydôwn","pron":"kee-dohn'","derivation":"the same as H3591 (כִּידוֹן);","strongs_def":"Kidon, a place in Palestine","kjv_def":"Chidon."},
"H3593":{"lemma":"כִּידוֹר","xlit":"kîydôwr","pron":"kee-dore'","derivation":"of uncertain derivation;","strongs_def":"perhaps tumult","kjv_def":"battle."},
"H3594":{"lemma":"כִּיּוּן","xlit":"Kîyûwn","pron":"kee-yoon'","derivation":"from H3559 (כּוּן);","strongs_def":"properly, a statue, i.e. idol; but used (by euphemism) forsome heathen deity (perhaps corresponding to Priapus or Baal-peor)","kjv_def":"Chiun."},
"H3595":{"lemma":"כִּיּוֹר","xlit":"kîyôwr","pron":"kee-yore'","derivation":"or כִּיֹּר; from the same as H3564 (כּוּר);","strongs_def":"properly, something round (as excavated or bored), i.e. a chafing-dish forcoals or a caldron forcooking; hence (from similarity of form) a washbowl; also (for the same reason) a pulpit or platform","kjv_def":"hearth, laver, pan, scaffold."},
"H3596":{"lemma":"כִּילַי","xlit":"kîylay","pron":"kee-lah'-ee","derivation":"or כֵּלַי; from H3557 (כּוּל) in the sense of withholding;","strongs_def":"niggardly","kjv_def":"churl."},
"H3597":{"lemma":"כֵּילַף","xlit":"kêylaph","pron":"kay-laf'","derivation":"from an unused root meaning to clap or strike with noise;","strongs_def":"a club or sledge-hammer","kjv_def":"hammer."},
"H3598":{"lemma":"כִּימָה","xlit":"Kîymâh","pron":"kee-maw'","derivation":"from the same as H3558 (כּוּמָז);","strongs_def":"a cluster of stars, i.e. the Pleiades","kjv_def":"Pleiades, seven stars."},
"H3599":{"lemma":"כִּיס","xlit":"kîyç","pron":"keece","derivation":"a form for H3563 (כּוֹס);","strongs_def":"a cup; also a bag formoney or weights","kjv_def":"bag, cup, purse."},
"H3600":{"lemma":"כִּיר","xlit":"kîyr","pron":"keer","derivation":"a form for H3564 (כּוּר) (only in the dual);","strongs_def":"a cooking range (consisting of two parallel stones, across which the boiler is set)","kjv_def":"ranges for pots."},
"H3601":{"lemma":"כִּישׁוֹר","xlit":"kîyshôwr","pron":"kee-shore'","derivation":"from H3787 (כָּשֵׁר); compare H6418 (פֶּלֶךְ)","strongs_def":"literally a director, i.e. the spindle or shank of adistaff, by which it is twirled","kjv_def":"spindle."},
"H3602":{"lemma":"כָּכָה","xlit":"kâkâh","pron":"kaw'-kaw","derivation":"from H3541 (כֹּה);","strongs_def":"just so, referring to the previous or following context","kjv_def":"after that (this) manner, this matter, (even) so, in such a case, thus."},
"H3603":{"lemma":"כִּכָּר","xlit":"kikkâr","pron":"kik-kawr'","derivation":"from H3769 (כָּרַר);","strongs_def":"a circle, i.e. (by implication) a circumjacent tract or region, especially the Ghor or valley of the Jordan; also a (round) loaf; also a talent (or large [round] coin)","kjv_def":"loaf, morsel, piece, plain, talent."},
"H3604":{"lemma":"כִּכֵּר","xlit":"kikkêr","pron":"kik-kare'","derivation":"(Aramaic) corresponding to H3603 (כִּכָּר);","strongs_def":"a talent","kjv_def":"talent."},
"H3605":{"lemma":"כֹּל","xlit":"kôl","pron":"kole","derivation":"or (Jeremiah 33:8) כּוֹל; from H3634 (כָּלַל);","strongs_def":"properly, the whole; hence, all, any or every (in the singular only, but often in a plural sense)","kjv_def":"(in) all (manner, (ye)), altogether, any (manner), enough, every (one, place, thing), howsoever, as many as, (no-) thing, ought, whatsoever, (the) whole, whoso(-ever)."},
"H3606":{"lemma":"כֹּל","xlit":"kôl","pron":"kole","derivation":"(Aramaic) corresponding to H3605 (כֹּל)","strongs_def":"{properly, the whole; hence, all, any or every (in the singular only, but often in a plural sense)}","kjv_def":"all, any, + (forasmuch) as, + be-(for this) cause, every, + no (manner, -ne), + there (where) -fore, + though, what (where, who) -soever, (the) whole."},
"H3607":{"lemma":"כָּלָא","xlit":"kâlâʼ","pron":"kaw-law'","derivation":"a primitive root;","strongs_def":"to restrict, by act (hold back or in) or word (prohibit)","kjv_def":"finish, forbid, keep (back), refrain, restrain, retain, shut up, be stayed, withhold."},
"H3608":{"lemma":"כֶּלֶא","xlit":"keleʼ","pron":"keh'-leh","derivation":"from H3607 (כָּלָא);","strongs_def":"a prison","kjv_def":"prison. Compare H3610 (כִּלְאַיִם), H3628 (כְּלִיא)."},
"H3609":{"lemma":"כִּלְאָב","xlit":"Kilʼâb","pron":"kil-awb'","derivation":"apparently from H3607 (כָּלָא) and H1 (אָב); restraint of (his) father;","strongs_def":"Kilab, an Israelite","kjv_def":"Chileab."},
"H3610":{"lemma":"כִּלְאַיִם","xlit":"kilʼayim","pron":"kil-ah'-yim","derivation":"dual of H3608 (כֶּלֶא) in the original sense of separation;","strongs_def":"two heterogeneities","kjv_def":"divers seeds (-e kinds), mingled (seed)."},
"H3611":{"lemma":"כֶּלֶב","xlit":"keleb","pron":"keh'-leb","derivation":"from an unused root means. to yelp, or else to attack;","strongs_def":"a dog; hence (by euphemism) a male prostitute","kjv_def":"dog."},
"H3612":{"lemma":"כָּלֵב","xlit":"Kâlêb","pron":"kaw-labe'","derivation":"perhaps a form of H3611 (כֶּלֶב), or else from the same root in the sense of forcible;","strongs_def":"Caleb, the name of three Israelites","kjv_def":"Caleb."},
"H3613":{"lemma":"כָּלֵב אֶפְרָתָה","xlit":"Kâlêb ʼEphrâthâh","pron":"kaw-labe' ef-raw'-thaw","derivation":"from H3612 (כָּלֵב) and H672 (אֶפְרָת);","strongs_def":"Caleb-Ephrathah, a place in Egypt (if the text is correct)","kjv_def":"Caleb-ephrathah."},
"H3614":{"lemma":"כָּלִבּוֹ","xlit":"Kâlibbôw","pron":"kaw-lib-bo'","derivation":"probably by erroneous transcription for כָּלֵבִי; patronymically from H3612 (כָּלֵב);","strongs_def":"a Calebite or descendant of Caleb","kjv_def":"of the house of Caleb."},
"H3615":{"lemma":"כָּלָה","xlit":"kâlâh","pron":"kaw-law'","derivation":"a primitive root;","strongs_def":"to end, whether intransitive (to cease, be finished, perish) or transitived (to complete, prepare, consume)","kjv_def":"accomplish, cease, consume (away), determine, destroy (utterly), be (when... were) done, (be an) end (of), expire, (cause to) fail, faint, finish, fulfil, [idiom] fully, [idiom] have, leave (off), long, bring to pass, wholly reap, make clean riddance, spend, quite take away, waste."},
"H3616":{"lemma":"כָּלֶה","xlit":"kâleh","pron":"kaw-leh'","derivation":"from H3615 (כָּלָה);","strongs_def":"pining","kjv_def":"fail."},
"H3617":{"lemma":"כָּלָה","xlit":"kâlâh","pron":"kaw-law'","derivation":"from H3615 (כָּלָה);","strongs_def":"a completion; adverb, completely; also destruction","kjv_def":"altogether, (be, utterly) consume(-d), consummation(-ption), was determined, (full, utter) end, riddance."},
"H3618":{"lemma":"כַּלָּה","xlit":"kallâh","pron":"kal-law'","derivation":"from H3634 (כָּלַל);","strongs_def":"a bride (as if perfect); hence, a son's wife","kjv_def":"bride, daughter-in-law, spouse."},
"H3619":{"lemma":"כְּלוּב","xlit":"kᵉlûwb","pron":"kel-oob'","derivation":"from the same as H3611 (כֶּלֶב);","strongs_def":"a bird-trap (as furnished with a clap-stick or treadle to spring it); hence, a basket (as resembling a wicker cage)","kjv_def":"basket, cage."},
"H3620":{"lemma":"כְּלוּב","xlit":"Kᵉlûwb","pron":"kel-oob'","derivation":"the same as H3619 (כְּלוּב)xlit kᵉlûb corrected to kᵉlûwb;","strongs_def":"Kelub, the name of two Israelites","kjv_def":"Chelub."},
"H3621":{"lemma":"כְּלוּבַי","xlit":"Kᵉlûwbay","pron":"kel-oo-bay'-ee","derivation":"a form of H3612 (כָּלֵב);","strongs_def":"Kelubai, an Israelite","kjv_def":"Chelubai."},
"H3622":{"lemma":"כְּלוּהַי","xlit":"Kᵉlûwhay","pron":"kel-oo-hah'-ee","derivation":"from H3615 (כָּלָה); completed;","strongs_def":"Keluhai, an Israelite","kjv_def":"Chelluh."},
"H3623":{"lemma":"כְּלוּלָה","xlit":"kᵉlûwlâh","pron":"kel-oo-law'","derivation":"denominative passive participle from H3618 (כַּלָּה)lemma כַּלָה missing dagesh, corrected to כַּלָּה;","strongs_def":"bridehood (only in the plural)","kjv_def":"espousal."},
"H3624":{"lemma":"כֶּלַח","xlit":"kelach","pron":"keh'-lakh","derivation":"from an unused root meaning to be complete;","strongs_def":"maturity","kjv_def":"full (old) age."},
"H3625":{"lemma":"כֶּלַח","xlit":"Kelach","pron":"keh'-lakh","derivation":"the same as H3624 (כֶּלַח);","strongs_def":"Kelach, a place in Assyria","kjv_def":"Calah."},
"H3626":{"lemma":"כׇּל־חֹזֶה","xlit":"Kol-Chôzeh","pron":"kol-kho-zeh'","derivation":"from H3605 (כֹּל) and H2374 (חֹזֶה); every seer;","strongs_def":"Col-Chozeh, an Israelite","kjv_def":"Col-hozeh."},
"H3627":{"lemma":"כְּלִי","xlit":"kᵉlîy","pron":"kel-ee'","derivation":"from H3615 (כָּלָה);","strongs_def":"something prepared, i.e. any apparatus (as an implement, utensil, dress, vessel or weapon)","kjv_def":"armour(-bearer), artillery, bag, carriage, [phrase] furnish, furniture, instrument, jewel, that is made of, [idiom] one from another, that which pertaineth, pot, [phrase] psaltery, sack, stuff, thing, tool, vessel, ware, weapon, [phrase] whatsoever."},
"H3628":{"lemma":"כְּלִיא","xlit":"kᵉlîyʼ","pron":"kel-ee'","derivation":"or כְּלוּא; from H3627 (כְּלִי) (compare H3608 (כֶּלֶא));","strongs_def":"a prison","kjv_def":"prison."},
"H3629":{"lemma":"כִּלְיָה","xlit":"kilyâh","pron":"kil-yaw'","derivation":"feminine of H3627 (כְּלִי) (only in the plural);","strongs_def":"a kidney (as an essential organ); figuratively, the mind (as the interior self)","kjv_def":"kidneys, reins."},
"H3630":{"lemma":"כִּלְיוֹן","xlit":"Kilyôwn","pron":"kil-yone'","derivation":"a form of H3631 (כִּלָּיוֹן);","strongs_def":"Kiljon, an Israelite","kjv_def":"Chilion."},
"H3631":{"lemma":"כִּלָּיוֹן","xlit":"killâyôwn","pron":"kil-law-yone'","derivation":"from H3615 (כָּלָה);","strongs_def":"pining, destruction","kjv_def":"consumption, failing."},
"H3632":{"lemma":"כָּלִיל","xlit":"kâlîyl","pron":"kaw-leel'","derivation":"from H3634 (כָּלַל);","strongs_def":"complete; as noun, the whole (specifically, a sacrifice entirely consumed); as adverb, fully","kjv_def":"all, every whit, flame, perfect(-ion), utterly, whole burnt offering (sacrifice), wholly."},
"H3633":{"lemma":"כַּלְכֹּל","xlit":"Kalkôl","pron":"kal-kole'","derivation":"from H3557 (כּוּל); sustenance;","strongs_def":"Calcol, an Israelite","kjv_def":"Calcol, Chalcol."},
"H3634":{"lemma":"כָּלַל","xlit":"kâlal","pron":"kaw-lal'","derivation":"a primitive root;","strongs_def":"to complete","kjv_def":"(make) perfect."},
"H3635":{"lemma":"כְּלַל","xlit":"kᵉlal","pron":"kel-al'","derivation":"(Aramaic) corresponding to H3634 (כָּלַל);","strongs_def":"to complete","kjv_def":"finish, make (set) up."},
"H3636":{"lemma":"כְּלָל","xlit":"Kᵉlâl","pron":"kel-awl'","derivation":"from H3634 (כָּלַל); complete;","strongs_def":"Kelal, an Israelite","kjv_def":"Chelal."},
"H3637":{"lemma":"כָּלַם","xlit":"kâlam","pron":"kaw-lawm'","derivation":"a primitive root;","strongs_def":"properly, to wound; but only figuratively, to taunt or insult","kjv_def":"be (make) ashamed, blush, be confounded, be put to confusion, hurt, reproach, (do, put to) shame."},
"H3638":{"lemma":"כִּלְמָד","xlit":"Kilmâd","pron":"kil-mawd'","derivation":"of foreign derivation;","strongs_def":"Kilmad, a place apparently in the Assyrian empire","kjv_def":"Chilmad."},
"H3639":{"lemma":"כְּלִמָּה","xlit":"kᵉlimmâh","pron":"kel-im-maw'","derivation":"from H3637 (כָּלַם);","strongs_def":"disgrace","kjv_def":"confusion, dishonour, reproach, shame."},
"H3640":{"lemma":"כְּלִמּוּת","xlit":"kᵉlimmûwth","pron":"kel-im-mooth'","derivation":"from H3639 (כְּלִמָּה);","strongs_def":"disgrace","kjv_def":"shame."},
"H3641":{"lemma":"כַּלְנֶה","xlit":"Kalneh","pron":"kal-neh'","derivation":"or כַּלְנֵה; also כַּלְנוֹ; of foreign derivation;","strongs_def":"Calneh or Calno, a place in the Assyrian empire","kjv_def":"Calneh, Calno. Compare H3656 (כַּנֶּה)."},
"H3642":{"lemma":"כָּמַהּ","xlit":"kâmahh","pron":"kaw-mah","derivation":"a primitive root;","strongs_def":"to pine after","kjv_def":"long."},
"H3643":{"lemma":"כִּמְהָם","xlit":"Kimhâm","pron":"kim-hawm'","derivation":"from H3642 (כָּמַהּ); pining;","strongs_def":"Kimham, an Israelite","kjv_def":"Chimham."},
"H3644":{"lemma":"כְּמוֹ","xlit":"kᵉmôw","pron":"kem-o'","derivation":"or כָּמוֹ; (compare H3651 (כֵּן));","strongs_def":"a form of the prefix 'k-', but used separately  as, thus, so","kjv_def":"according to, (such) as (it were, well as), in comparison of, like (as, to, unto), thus, when, worth."},
"H3645":{"lemma":"כְּמוֹשׁ","xlit":"Kᵉmôwsh","pron":"kem-oshe'","derivation":"or (Jeremiah 48:7) כְּמִישׁ; from an unused root meaning to subdue; the powerful;","strongs_def":"Kemosh, the god of the Moabites","kjv_def":"Chemosh."},
"H3646":{"lemma":"כַּמֹּן","xlit":"kammôn","pron":"kam-mone'","derivation":"from an unused root meaning to store up or preserve;","strongs_def":"'cummin' (from its use as a condiment)","kjv_def":"cummin."},
"H3647":{"lemma":"כָּמַס","xlit":"kâmaç","pron":"kaw-mas'","derivation":"a primitive root;","strongs_def":"to store away, i.e. (figuratively) in the memory","kjv_def":"lay up in store."},
"H3648":{"lemma":"כָּמַר","xlit":"kâmar","pron":"kaw-mar'","derivation":"a primitive root;","strongs_def":"properly, to intertwine or contract, i.e. (by implication) to shrivel (as with heat); figuratively, to be deeply affected with passion (love or pity)","kjv_def":"be black, be kindled, yearn."},
"H3649":{"lemma":"כָּמָר","xlit":"kâmâr","pron":"kaw-mawr'","derivation":"from H3648 (כָּמַר);","strongs_def":"properly, an ascetic (as if shrunk with self-maceration), i.e. an idolatrous priest (only in plural)","kjv_def":"Chemarims (idolatrous) priests."},
"H3650":{"lemma":"כִּמְרִיר","xlit":"kimrîyr","pron":"kim-reer'","derivation":"redupl. from H3648 (כָּמַר);","strongs_def":"obscuration (as if from shrinkage of light, i.e. an eclipse (only in plural)","kjv_def":"blackness."},
"H3651":{"lemma":"כֵּן","xlit":"kên","pron":"kane","derivation":"from H3559 (כּוּן);","strongs_def":"properly, set upright; hence (figuratively as adjective) just; but usually (as adverb or conjunction) rightly or so (in various applications to manner, time and relation; often with other particles)","kjv_def":"[phrase] after that (this, -ward, -wards), as... as, [phrase] (for-) asmuch as yet, [phrase] be (for which) cause, [phrase] following, howbeit, in (the) like (manner, -wise), [idiom] the more, right, (even) so, state, straightway, such (thing), surely, [phrase] there (where) -fore, this, thus, true, well, [idiom] you."},
"H3652":{"lemma":"כֵּן","xlit":"kên","pron":"kane","derivation":"(Aramaic) corresponding to H3651 (כֵּן);","strongs_def":"so","kjv_def":"thus."},
"H3653":{"lemma":"כֵּן","xlit":"kên","pron":"kane","derivation":"the same as H3651 (כֵּן), used as a noun;","strongs_def":"a stand, i.e. pedestal or station","kjv_def":"base, estate, foot, office, place, well."},
"H3654":{"lemma":"כֵּן","xlit":"kên","pron":"kane","derivation":"from H3661 (כָּנַן) in the sense of fastening; (from infixing its sting; used only in plural (and irregular in Exodus 8:17,18; Hebrews 13:14))","strongs_def":"a gnat","kjv_def":"lice, [idiom] manner."},
"H3655":{"lemma":"כָּנָה","xlit":"kânâh","pron":"kaw-naw'","derivation":"a primitive root;","strongs_def":"to address by an additional name; hence, to eulogize","kjv_def":"give flattering titles, surname (himself)."},
"H3656":{"lemma":"כַּנֶּה","xlit":"Kanneh","pron":"kan-neh'","derivation":"for H3641 (כַּלְנֶה);","strongs_def":"Canneh, a place in Assyria","kjv_def":"Canneh."},
"H3657":{"lemma":"כַּנָּה","xlit":"kannâh","pron":"kaw-naw'","derivation":"from H3661 (כָּנַן);","strongs_def":"a plant (as set)","kjv_def":"[idiom] vineyard."},
"H3658":{"lemma":"כִּנּוֹר","xlit":"kinnôwr","pron":"kin-nore'","derivation":"from a unused root meaning to twang;","strongs_def":"a harp","kjv_def":"harp."},
"H3659":{"lemma":"כׇּנְיָהוּ","xlit":"Konyâhûw","pron":"kon-yaw'-hoo","derivation":"for H3204 (יְכׇנְיָה);","strongs_def":"Conjah, an Israelite king","kjv_def":"Coniah."},
"H3660":{"lemma":"כְּנֵמָא","xlit":"kᵉnêmâʼ","pron":"ken-ay-maw'","derivation":"(Aramaic) corresponding to H3644 (כְּמוֹ);","strongs_def":"so or thus","kjv_def":"so, (in) this manner (sort), thus."},
"H3661":{"lemma":"כָּנַן","xlit":"kânan","pron":"kaw-nan'","derivation":"a primitive root;","strongs_def":"to set out, i.e. plant","kjv_def":"[idiom] vineyard."},
"H3662":{"lemma":"כְּנָנִי","xlit":"Kᵉnânîy","pron":"ken-aw-nee'","derivation":"from H3661 (כָּנַן); planted;","strongs_def":"Kenani, an Israelite","kjv_def":"Chenani."},
"H3663":{"lemma":"כְּנַנְיָה","xlit":"Kᵉnanyâh","pron":"ken-an-yaw'","derivation":"or כְּנַנְיָהוּ; from H3661 (כָּנַן) and H3050 (יָהּ); Jah has planted;","strongs_def":"Kenanjah, an Israelite","kjv_def":"Chenaniah."},
"H3664":{"lemma":"כָּנַס","xlit":"kânaç","pron":"kaw-nas'","derivation":"a primitive root;","strongs_def":"to collect; hence, to enfold","kjv_def":"gather (together), heap up, wrap self."},
"H3665":{"lemma":"כָּנַע","xlit":"kânaʻ","pron":"kaw-nah'","derivation":"a primitive root;","strongs_def":"properly, to bend the knee; hence, to humiliate, vanquish","kjv_def":"bring down (low), into subjection, under, humble (self), subdue."},
"H3666":{"lemma":"כִּנְעָה","xlit":"kinʻâh","pron":"kin-aw'","derivation":"from H3665 (כָּנַע) in the sense of folding (compare H3664 (כָּנַס));","strongs_def":"a package","kjv_def":"wares."},
"H3667":{"lemma":"כְּנַעַן","xlit":"Kᵉnaʻan","pron":"ken-ah'-an","derivation":"from H3665 (כָּנַע); humiliated;","strongs_def":"Kenaan, a son a Ham; also the country inhabited by him","kjv_def":"Canaan, merchant, traffick."},
"H3668":{"lemma":"כְּנַעֲנָה","xlit":"Kᵉnaʻănâh","pron":"ken-ah-an-aw'","derivation":"feminine of H3667 (כְּנַעַן);","strongs_def":"Kenaanah, the name of two Israelites","kjv_def":"Chenaanah."},
"H3669":{"lemma":"כְּנַעַנִי","xlit":"Kᵉnaʻanîy","pron":"ken-ah-an-ee'","derivation":"patrial from H3667 (כְּנַעַן);","strongs_def":"a Kenaanite or inhabitant of Kenaan; by implication, a pedlar (the Canaanites standing for their neighbors the Ishmaelites, who conducted mercantile caravans)","kjv_def":"Canaanite, merchant, trafficker."},
"H3670":{"lemma":"כָּנַף","xlit":"kânaph","pron":"kaw-naf'","derivation":"a primitive root;","strongs_def":"properly, to project laterally, i.e. probably (reflexive) to withdraw","kjv_def":"be removed."},
"H3671":{"lemma":"כָּנָף","xlit":"kânâph","pron":"kaw-nawf'","derivation":"from H3670 (כָּנַף);","strongs_def":"an edge or extremity; specifically (of a bird or army) a wing, (of a garment or bedclothing) a flap, (of the earth) a quarter, (of a building) a pinnacle","kjv_def":"[phrase] bird, border, corner, end, feather(-ed), [idiom] flying, [phrase] (one an-) other, overspreading, [idiom] quarters, skirt, [idiom] sort, uttermost part, wing(-ed)."},
"H3672":{"lemma":"כִּנְּרוֹת","xlit":"Kinnᵉrôwth","pron":"kin-ner-oth'","derivation":"or כִּנֶּרֶתlemma כִּנֶרֶת missing dagesh, corrected to כִּנֶּרֶת; respectively plural and singular feminine from the same as H3658 (כִּנּוֹר); perhaps harp-shaped;","strongs_def":"Kinneroth or Kinnereth, a place in Palestine","kjv_def":"Chinnereth, Chinneroth, Cinneroth."},
"H3673":{"lemma":"כָּנַשׁ","xlit":"kânash","pron":"kaw-nash'","derivation":"(Aramaic) corresponding to H3664 (כָּנַס);","strongs_def":"to assemble","kjv_def":"gather together."},
"H3674":{"lemma":"כְּנָת","xlit":"kᵉnâth","pron":"ken-awth'","derivation":"from H3655 (כָּנָה);","strongs_def":"a colleague (as having the same title)","kjv_def":"companion."},
"H3675":{"lemma":"כְּנָת","xlit":"kᵉnâth","pron":"ken-awth'","derivation":"(Aramaic) corresponding to H3674 (כְּנָת)","strongs_def":"{a colleague (as having the same title)}","kjv_def":"companion."},
"H3676":{"lemma":"כֵּס","xlit":"kêç","pron":"kace","derivation":"apparently a contraction for H3678 (כִּסֵּא), but probably by erroneous transcription for H5251 (נֵס)","strongs_def":"{a flag; also a sail; by implication, a flagstaff; generally a signal; figuratively, a token}","kjv_def":"sworn."},
"H3677":{"lemma":"כֶּסֶא","xlit":"keçeʼ","pron":"keh'-seh","derivation":"or כֶּסֶה; apparently from H3680 (כָּסָה);","strongs_def":"properly, fulness or the full moon, i.e. its festival","kjv_def":"(time) appointed."},
"H3678":{"lemma":"כִּסֵּא","xlit":"kiççêʼ","pron":"kis-say'","derivation":"or כִּסֵּה; from H3680 (כָּסָה);","strongs_def":"properly, covered, i.e. a throne (as canopied)","kjv_def":"seat, stool, throne."},
"H3679":{"lemma":"כַּסְדַּי","xlit":"Kaçday","pron":"kas-dah'-ee","derivation":"for H3778 (כַּשְׂדִּי)","strongs_def":"{a Kasdite, or descendant of Kesed; by implication, a Chaldaean (as if so descended); also an astrologer (as if proverbial of that people}","kjv_def":"Chaldean."},
"H3680":{"lemma":"כָּסָה","xlit":"kâçâh","pron":"kaw-saw'","derivation":"a primitive root;","strongs_def":"properly, to plump, i.e. fill up hollows; by implication, to cover (for clothing or secrecy)","kjv_def":"clad self, close, clothe, conceal, cover (self), (flee to) hide, overwhelm. Compare H3780 (כָּשָׂה)."},
"H3681":{"lemma":"כָּסוּי","xlit":"kâçûwy","pron":"kaw-soo'-ee","derivation":"passive participle of H3680 (כָּסָה);","strongs_def":"properly, covered, i.e. (as noun) a covering","kjv_def":"covering."},
"H3682":{"lemma":"כְּסוּת","xlit":"kᵉçûwth","pron":"kes-ooth'","derivation":"from H3680 (כָּסָה);","strongs_def":"a cover (garment); figuratively, a veiling","kjv_def":"covering, raiment, vesture."},
"H3683":{"lemma":"כָּסַח","xlit":"kâçach","pron":"kaw-sakh'","derivation":"a primitive root;","strongs_def":"to cut off","kjv_def":"cut down (up)."},
"H3684":{"lemma":"כְּסִיל","xlit":"kᵉçîyl","pron":"kes-eel'","derivation":"from H3688 (כָּסַל);","strongs_def":"properly, fat, i.e. (figuratively) stupid or silly","kjv_def":"fool(-ish)."},
"H3685":{"lemma":"כְּסִיל","xlit":"Kᵉçîyl","pron":"kes-eel'","derivation":"the same as H3684 (כְּסִיל);","strongs_def":"any notable constellation; specifically Orion (as if a burly one)","kjv_def":"constellation, Orion."},
"H3686":{"lemma":"כְּסִיל","xlit":"Kᵉçîyl","pron":"kes-eel'","derivation":"the same as H3684 (כְּסִיל);","strongs_def":"Kesil, a place in Palestine","kjv_def":"Chesil."},
"H3687":{"lemma":"כְּסִילוּת","xlit":"kᵉçîylûwth","pron":"kes-eel-ooth'","derivation":"from H3684 (כְּסִיל);","strongs_def":"silliness","kjv_def":"foolish."},
"H3688":{"lemma":"כָּסַל","xlit":"kâçal","pron":"kaw-sal'","derivation":"a primitive root;","strongs_def":"properly, to be fat, i.e. (figuratively) silly","kjv_def":"be foolish."},
"H3689":{"lemma":"כֶּסֶל","xlit":"keçel","pron":"keh'-sel","derivation":"from H3688 (כָּסַל);","strongs_def":"properly, fatness, i.e. by implication (literally) the loin (as the seat of the leaf fat) or (generally) the viscera; also (figuratively) silliness or (in a good sense) trust","kjv_def":"confidence, flank, folly, hope, loin."},
"H3690":{"lemma":"כִּסְלָה","xlit":"kiçlâh","pron":"kis-law'","derivation":"feminine of H3689 (כֶּסֶל);","strongs_def":"in a good sense, trust; in a bad one, silliness","kjv_def":"confidence, folly."},
"H3691":{"lemma":"כִּסְלֵו","xlit":"Kiçlêv","pron":"kis-lave'","derivation":"probably of foreign origin;","strongs_def":"Kisleu, the 9th Hebrew month","kjv_def":"Chisleu."},
"H3692":{"lemma":"כִּסְלוֹן","xlit":"Kiçlôwn","pron":"kis-lone'","derivation":"from H3688 (כָּסַל); hopeful;","strongs_def":"Kislon, an Israelite","kjv_def":"Chislon."},
"H3693":{"lemma":"כְּסָלוֹן","xlit":"Kᵉçâlôwn","pron":"kes-aw-lone'","derivation":"from H3688 (כָּסַל); fertile;","strongs_def":"Kesalon, a place in Palestine","kjv_def":"Chesalon."},
"H3694":{"lemma":"כְּסֻלּוֹת","xlit":"Kᵉçullôwth","pron":"kes-ool-loth'","derivation":"feminine plural of passive participle of H3688 (כָּסַל); fattened;","strongs_def":"Kesulloth, a place in Palestine","kjv_def":"Chesulloth."},
"H3695":{"lemma":"כַּסְלֻחִים","xlit":"Kaçluchîym","pron":"kas-loo'-kheem","derivation":"a plural probably of foreign derivation;","strongs_def":"Casluchim, a people cognate to the Egyptians","kjv_def":"Casluhim."},
"H3696":{"lemma":"כִּסְלֹת תָּבֹר","xlit":"Kiçlôth Tâbôr","pron":"kis-loth' taw-bore'","derivation":"from the feminine plural of H3689 (כֶּסֶל) and H8396 (תָּבוֹר); flanks of Tabor;","strongs_def":"Kisloth-Tabor, a place in Palestine","kjv_def":"Chisloth-tabor."},
"H3697":{"lemma":"כָּסַם","xlit":"kâçam","pron":"kaw-sam'","derivation":"a primitive root;","strongs_def":"to shear","kjv_def":"[idiom] only, poll. Compare H3765 (כִּרְסֵם)."},
"H3698":{"lemma":"כֻּסֶּמֶת","xlit":"kuççemeth","pron":"koos-seh'-meth","derivation":"from H3697 (כָּסַם);","strongs_def":"spelt (from its bristliness as if just shorn)","kjv_def":"fitches, rie."},
"H3699":{"lemma":"כָּסַס","xlit":"kâçaç","pron":"kaw-sas'","derivation":"a primitive root;","strongs_def":"to estimate","kjv_def":"make count."},
"H3700":{"lemma":"כָּסַף","xlit":"kâçaph","pron":"kaw-saf'","derivation":"a primitive root;","strongs_def":"properly, to become pale, i.e. (by implication) to pine after; also to fear","kjv_def":"(have) desire, be greedy, long, sore."},
"H3701":{"lemma":"כֶּסֶף","xlit":"keçeph","pron":"keh'-sef","derivation":"from H3700 (כָּסַף);","strongs_def":"silver (from its pale color); by implication, money","kjv_def":"money, price, silver(-ling)."},
"H3702":{"lemma":"כְּסַף","xlit":"kᵉçaph","pron":"kes-af'","derivation":"(Aramaic) corresponding to H3701 (כֶּסֶף)","strongs_def":"{silver (from its pale color); by implication, money}","kjv_def":"money, silver."},
"H3703":{"lemma":"כָּסִפְיָא","xlit":"Kâçiphyâʼ","pron":"kaw-sif-yaw'","derivation":"perhaps from H3701 (כֶּסֶף); silvery;","strongs_def":"Casiphja, a place in Babylon","kjv_def":"Casiphia."},
"H3704":{"lemma":"כֶּסֶת","xlit":"keçeth","pron":"keh'-seth","derivation":"from H3680 (כָּסָה);","strongs_def":"a cushion or pillow (as covering a seat or bed)","kjv_def":"pillow."},
"H3705":{"lemma":"כְּעַן","xlit":"kᵉʻan","pron":"keh-an'","derivation":"(Aramaic) probably from H3652 (כֵּן);","strongs_def":"now","kjv_def":"now."},
"H3706":{"lemma":"כְּעֶנֶת","xlit":"kᵉʻeneth","pron":"keh-eh'-neth","derivation":"(Aramaic) or כְּעֶת; (Aramaic), feminine of H3705 (כְּעַן);","strongs_def":"thus (only in the formula 'and so forth')","kjv_def":"at such a time."},
"H3707":{"lemma":"כַּעַס","xlit":"kaʻaç","pron":"kaw-as'","derivation":"a primitive root;","strongs_def":"to trouble; by implication, to grieve, rage, be indignant","kjv_def":"be angry, be grieved, take indignation, provoke (to anger, unto wrath), have sorrow, vex, be wroth."},
"H3708":{"lemma":"כַּעַס","xlit":"kaʻaç","pron":"kah'-as","derivation":"or (in Job) כַּעַשׂ; from H3707 (כַּעַס);","strongs_def":"vexation","kjv_def":"anger, angry, grief, indignation, provocation, provoking, [idiom] sore, sorrow, spite, wrath."},
"H3709":{"lemma":"כַּף","xlit":"kaph","pron":"kaf","derivation":"from H3721 (כָּפַף);","strongs_def":"the hollow hand or palm (so of the paw of an animal, of the sole, and even of the bowl of a dish or sling, the handle of a bolt, the leaves of a palm-tree); figuratively, power","kjv_def":"branch, [phrase] foot, hand((-ful), -dle, (-led)), hollow, middle, palm, paw, power, sole, spoon."},
"H3710":{"lemma":"כֵּף","xlit":"kêph","pron":"kafe","derivation":"from H3721 (כָּפַף);","strongs_def":"a hollow rock","kjv_def":"rock."},
"H3711":{"lemma":"כָּפָה","xlit":"kâphâh","pron":"kaw-faw'","derivation":"a primitive root;","strongs_def":"properly, to bend, i.e. (figuratively) to tame or subdue","kjv_def":"pacify."},
"H3712":{"lemma":"כִּפָּה","xlit":"kippâh","pron":"kip-paw'","derivation":"feminine of H3709 (כַּף);","strongs_def":"a leaf of a palmtree","kjv_def":"branch."},
"H3713":{"lemma":"כְּפוֹר","xlit":"kᵉphôwr","pron":"kef-ore'","derivation":"from H3722 (כָּפַר);","strongs_def":"properly, a cover, i.e. (by implication) a tankard (or covered goblet); also white frost (as covering the ground)","kjv_def":"bason, hoar(-y) frost."},
"H3714":{"lemma":"כָּפִיס","xlit":"kâphîyç","pron":"kaw-fece'","derivation":"from an unused root meaning to connect;","strongs_def":"a girder","kjv_def":"beam."},
"H3715":{"lemma":"כְּפִיר","xlit":"kᵉphîyr","pron":"kef-eer'","derivation":"from H3722 (כָּפַר);","strongs_def":"a village (as covered in by walls); also a young lion (perhaps as covered with a mane)","kjv_def":"(young) lion, village. Compare H3723 (כָּפָר)."},
"H3716":{"lemma":"כְּפִירָה","xlit":"Kᵉphîyrâh","pron":"kef-ee-raw'","derivation":"feminine of H3715 (כְּפִיר); the village (always with the art.);","strongs_def":"Kephirah, a place in Palestine","kjv_def":"Chephirah."},
"H3717":{"lemma":"כָּפַל","xlit":"kâphal","pron":"kaw-fal'","derivation":"a primitive root;","strongs_def":"to fold together; figuratively, to repeat","kjv_def":"double."},
"H3718":{"lemma":"כֶּפֶל","xlit":"kephel","pron":"keh'-fel","derivation":"from H3717 (כָּפַל);","strongs_def":"a duplicate","kjv_def":"double."},
"H3719":{"lemma":"כָּפַן","xlit":"kâphan","pron":"kaw-fan'","derivation":"a primitive root;","strongs_def":"to bend","kjv_def":"bend."},
"H3720":{"lemma":"כָּפָן","xlit":"kâphân","pron":"kaw-fawn'","derivation":"from H3719 (כָּפַן);","strongs_def":"hunger (as making to stoop with emptiness and pain)","kjv_def":"famine."},
"H3721":{"lemma":"כָּפַף","xlit":"kâphaph","pron":"kaw-faf'","derivation":"a primitive root;","strongs_def":"to curve","kjv_def":"bow down (self)."},
"H3722":{"lemma":"כָּפַר","xlit":"kâphar","pron":"kaw-far'","derivation":"a primitive root;","strongs_def":"to cover (specifically with bitumen); figuratively, to expiate or condone, to placate or cancel","kjv_def":"appease, make (an atonement, cleanse, disannul, forgive, be merciful, pacify, pardon, purge (away), put off, (make) reconcile(-liation)."},
"H3723":{"lemma":"כָּפָר","xlit":"kâphâr","pron":"kaw-fawr'","derivation":"from H3722 (כָּפַר);","strongs_def":"a village (as protected by walls)","kjv_def":"village. Compare H3715 (כְּפִיר)."},
"H3724":{"lemma":"כֹּפֶר","xlit":"kôpher","pron":"ko'-fer","derivation":"from H3722 (כָּפַר);","strongs_def":"properly, a cover, i.e. (literally) a village (as covered in); (specifically) bitumen (as used for coating), and the henna plant (as used for dyeing); figuratively, a redemption-price","kjv_def":"bribe, camphire, pitch, ransom, satisfaction, sum of money, village."},
"H3725":{"lemma":"כִּפֻּר","xlit":"kippur","pron":"kip-poor'","derivation":"from H3722 (כָּפַר);","strongs_def":"expiation (only in plural)","kjv_def":"atonement."},
"H3726":{"lemma":"כְּפַר הָעַמּוֹנִי","xlit":"Kᵉphar hâ-ʻAmmôwnîy","pron":"kef-ar' haw-am-monee'","derivation":"from H3723 (כָּפָר) and H5984 (עַמּוֹנִי), with the article interposed; village of the Ammonite;","strongs_def":"Kefar-ha-Ammoni, a place in Palestine","kjv_def":"Chefarhaamonai."},
"H3727":{"lemma":"כַּפֹּרֶת","xlit":"kappôreth","pron":"kap-po'-reth","derivation":"from H3722 (כָּפַר);","strongs_def":"a lid (used only of the cover of the sacred Ark)","kjv_def":"mercy seat."},
"H3728":{"lemma":"כָּפַשׁ","xlit":"kâphash","pron":"kaw-fash'","derivation":"a primitive root;","strongs_def":"to tread down; figuratively, to humiliate","kjv_def":"cover."},
"H3729":{"lemma":"כְּפַת","xlit":"kᵉphath","pron":"kef-ath'","derivation":"(Aramaic) a root of uncertain correspondence;","strongs_def":"to fetter","kjv_def":"bind."},
"H3730":{"lemma":"כַּפְתֹּר","xlit":"kaphtôr","pron":"kaf-tore'","derivation":"or (Amos 9:1) כַּפְתּוֹר; probably from an unused root meaning to encircle;","strongs_def":"a chaplet; but used only in an architectonic sense, i.e. the capital of acolumn, or a wreath-like button or disk on the candelabrum","kjv_def":"knop, (upper) lintel."},
"H3731":{"lemma":"כַּפְתֹּר","xlit":"Kaphtôr","pron":"kaf-tore'","derivation":"or (Amos 9:7) כַּפְתּוֹר; apparently the same as H3730 (כַּפְתֹּר);","strongs_def":"Caphtor (i.e. a wreath-shaped island), the original seat of the Philistines","kjv_def":"Caphtor."},
"H3732":{"lemma":"כַּפְתֹּרִי","xlit":"Kaphtôrîy","pron":"kaf-to-ree'","derivation":"patrial from H3731 (כַּפְתֹּר);","strongs_def":"a Caphtorite (collectively) or native of Caphtor","kjv_def":"Caphthorim, Caphtorim(-s)."},
"H3733":{"lemma":"כַּר","xlit":"kar","pron":"kar","derivation":"from H3769 (כָּרַר) in the sense of plumpness;","strongs_def":"a ram (as full-grown and fat), including a battering-ram (as butting); hence, a meadow (as for sheep); also a pad or camel's saddle (as puffed out)","kjv_def":"captain, furniture, lamb, (large) pasture, ram. See also H1033 (בֵּית כַּר), H3746 (כָּרִי)."},
"H3734":{"lemma":"כֹּר","xlit":"kôr","pron":"kore","derivation":"from the same as H3564 (כּוּר);","strongs_def":"properly, a deep round vessel, i.e. (specifically) a cor or measure for things dry","kjv_def":"cor, measure. Aramaic the same."},
"H3735":{"lemma":"כָּרָא","xlit":"Kârâʼ","pron":"kaw-raw'","derivation":"(Aramaic) probably corresponding to H3738 (כָּרָה) in the sense of piercing(figuratively);","strongs_def":"to grieve","kjv_def":"be grieved."},
"H3736":{"lemma":"כַּרְבֵּל","xlit":"karbêl","pron":"kar-bale'","derivation":"from the same as H3525 (כֶּבֶל);","strongs_def":"to gird or clothe","kjv_def":"clothed."},
"H3737":{"lemma":"כַּרְבְּלָא","xlit":"karbᵉlâʼ","pron":"kar-bel-aw'","derivation":"(Aramaic) from a verb corresponding to that of H3736 (כַּרְבֵּל);","strongs_def":"a mantle","kjv_def":"hat."},
"H3738":{"lemma":"כָּרָה","xlit":"kârâh","pron":"kaw-raw'","derivation":"a primitive root;","strongs_def":"properly, to dig; figuratively, to plot; generally, to bore or open","kjv_def":"dig, [idiom] make (a banquet), open."},
"H3739":{"lemma":"כָּרָה","xlit":"kârâh","pron":"kaw-raw'","derivation":"usually assigned as a primitive root, but probably only a special application of H3738 (כָּרָה) (through the common idea of planning implied in a bargain);","strongs_def":"to purchase","kjv_def":"buy, prepare."},
"H3740":{"lemma":"כֵּרָה","xlit":"kêrâh","pron":"kay-raw'","derivation":"from H3739 (כָּרָה);","strongs_def":"a purchase","kjv_def":"provision."},
"H3741":{"lemma":"כָּרָה","xlit":"kârâh","pron":"kaw-raw'","derivation":"feminine of H3733 (כַּר);","strongs_def":"a meadow","kjv_def":"cottage."},
"H3742":{"lemma":"כְּרוּב","xlit":"kᵉrûwb","pron":"ker-oob'","derivation":"of uncertain derivation;","strongs_def":"a cherub or imaginary figure","kjv_def":"cherub, (plural) cherubims."},
"H3743":{"lemma":"כְּרוּב","xlit":"Kᵉrûwb","pron":"ker-oob'","derivation":"the same as H3742 (כְּרוּב);","strongs_def":"Kerub, a place in Babylon","kjv_def":"Cherub."},
"H3744":{"lemma":"כָּרוֹז","xlit":"kârôwz","pron":"kaw-roze'","derivation":"(Aramaic) from H3745 (כְּרַז);","strongs_def":"a herald","kjv_def":"herald."},
"H3745":{"lemma":"כְּרַז","xlit":"kᵉraz","pron":"ker-az'","derivation":"(Aramaic) probably of Greek origin;","strongs_def":"to proclaim","kjv_def":"make a proclamation."},
"H3746":{"lemma":"כָּרִי","xlit":"kârîy","pron":"kaw-ree'","derivation":"perhaps an abridged plural of H3733 (כַּר) in the sense of leader (of the flock);","strongs_def":"a life-guardsman","kjv_def":"captains, Cherethites (from the margin)."},
"H3747":{"lemma":"כְּרִית","xlit":"Kᵉrîyth","pron":"ker-eeth'","derivation":"from H3772 (כָּרַת); a cut;","strongs_def":"Kerith, a brook of Palestine","kjv_def":"Cherith."},
"H3748":{"lemma":"כְּרִיתוּת","xlit":"kᵉrîythûwth","pron":"ker-ee-thooth'","derivation":"from H3772 (כָּרַת);","strongs_def":"a cutting (of the matrimonial bond), i.e. divorce","kjv_def":"divorce(-ment)."},
"H3749":{"lemma":"כַּרְכֹּב","xlit":"karkôb","pron":"kar-kobe'","derivation":"expanded from the same as H3522 (כַּבּוֹן);","strongs_def":"a rim or top margin","kjv_def":"compass."},
"H3750":{"lemma":"כַּרְכֹּם","xlit":"karkôm","pron":"kar-kome'","derivation":"probably of foreign origin;","strongs_def":"the crocus","kjv_def":"saffron."},
"H3751":{"lemma":"כַּרְכְּמִישׁ","xlit":"Karkᵉmîysh","pron":"kar-kem-eesh'","derivation":"of foreign derivation;","strongs_def":"Karkemish, a place in Syria","kjv_def":"Carchemish."},
"H3752":{"lemma":"כַּרְכַּס","xlit":"Karkaç","pron":"kar-kas'","derivation":"of Persian origin;","strongs_def":"Karkas, a eunuch of Xerxes","kjv_def":"Carcas."},
"H3753":{"lemma":"כַּרְכָּרָה","xlit":"karkârâh","pron":"kar-kaw-raw'","derivation":"from H3769 (כָּרַר);","strongs_def":"a dromedary (from its rapid motion as if dancing)","kjv_def":"swift beast."},
"H3754":{"lemma":"כֶּרֶם","xlit":"kerem","pron":"keh'-rem","derivation":"from an unused root of uncertain meaning;","strongs_def":"a garden or vineyard","kjv_def":"vines, (increase of the) vineyard(-s), vintage. See also H1021 (בֵּית הַכֶּרֶם)."},
"H3755":{"lemma":"כֹּרֵם","xlit":"kôrêm","pron":"ko-rame'","derivation":"active participle of an imaginary denominative from H3754 (כֶּרֶם);","strongs_def":"a vinedresser","kjv_def":"vine dresser (as one or two words)."},
"H3756":{"lemma":"כַּרְמִי","xlit":"Karmîy","pron":"kar-mee'","derivation":"from H3754 (כֶּרֶם); gardener;","strongs_def":"Karmi, the name of three Israelites","kjv_def":"Carmi."},
"H3757":{"lemma":"כַּרְמִי","xlit":"Karmîy","pron":"kar-mee'","derivation":"patronymically from H3756 (כַּרְמִי);","strongs_def":"a Karmite or descendant of Karmi","kjv_def":"Carmites."},
"H3758":{"lemma":"כַּרְמִיל","xlit":"karmîyl","pron":"kar-mele'","derivation":"probably of foreign origin;","strongs_def":"carmine, a deep red","kjv_def":"crimson."},
"H3759":{"lemma":"כַּרְמֶל","xlit":"karmel","pron":"kar-mel'","derivation":"from H3754 (כֶּרֶם);","strongs_def":"a planted field (garden, orchard, vineyard or park); by implication, garden produce","kjv_def":"full (green) ears (of corn), fruitful field (place), plentiful (field)."},
"H3760":{"lemma":"כַּרְמֶל","xlit":"Karmel","pron":"kar-mel'","derivation":"the same as H3759 (כַּרְמֶל);","strongs_def":"Karmel, the name of a hill and of a town in Palestine","kjv_def":"Carmel, fruitful (plentiful) field, (place)."},
"H3761":{"lemma":"כַּרְמְלִי","xlit":"Karmᵉlîy","pron":"kar-mel-ee'","derivation":"patron from H3760 (כַּרְמֶל);","strongs_def":"a Karmelite or inhabitant of Karmel (the town)","kjv_def":"Carmelite."},
"H3762":{"lemma":"כַּרְמְלִית","xlit":"Karmᵉlîyth","pron":"kar-mel-eeth'","derivation":"feminine of H3761 (כַּרְמְלִי);","strongs_def":"a Karmelitess or female inhabitant of Karmel","kjv_def":"Carmelitess."},
"H3763":{"lemma":"כְּרָן","xlit":"Kᵉrân","pron":"ker-awn'","derivation":"of uncertain derivation;","strongs_def":"Keran, an aboriginal Idumaean","kjv_def":"Cheran."},
"H3764":{"lemma":"כׇּרְסֵא","xlit":"korçêʼ","pron":"kor-say'","derivation":"(Aramaic) corresponding to H3678 (כִּסֵּא);","strongs_def":"a throne","kjv_def":"throne."},
"H3765":{"lemma":"כִּרְסֵם","xlit":"kirçêm","pron":"kir-same'","derivation":"from H3697 (כָּסַם);","strongs_def":"to lay waste","kjv_def":"waste."},
"H3766":{"lemma":"כָּרַע","xlit":"kâraʻ","pron":"kaw-rah'","derivation":"a primitive root;","strongs_def":"to bend the knee; by implication, to sink, to prostrate","kjv_def":"bow (down, self), bring down (low), cast down, couch, fall, feeble, kneeling, sink, smite (stoop) down, subdue, [idiom] very."},
"H3767":{"lemma":"כָּרָע","xlit":"kârâʻ","pron":"kaw-raw'","derivation":"from H3766 (כָּרַע);","strongs_def":"the leg (from the knee to the ankle) of men or locusts (only in the dual)","kjv_def":"leg."},
"H3768":{"lemma":"כַּרְפַּס","xlit":"karpaç","pron":"kar-pas'","derivation":"of foreign origin;","strongs_def":"byssus or fine vegetable wool","kjv_def":"green."},
"H3769":{"lemma":"כָּרַר","xlit":"kârar","pron":"kaw-rar'","derivation":"a primitive root;","strongs_def":"to dance (i.e. whirl)","kjv_def":"dance(-ing)."},
"H3770":{"lemma":"כְּרֵשׂ","xlit":"kᵉrês","pron":"ker-ace'","derivation":"by variation from H7164 (קָרַס);","strongs_def":"the paunch or belly (as swelling out)","kjv_def":"belly."},
"H3771":{"lemma":"כַּרְשְׁנָא","xlit":"Karshᵉnâʼ","pron":"kar-shen-aw'","derivation":"of foreign origin;","strongs_def":"Karshena, a courtier of Xerxes","kjv_def":"Carshena."},
"H3772":{"lemma":"כָּרַת","xlit":"kârath","pron":"kaw-rath'","derivation":"a primitive root;","strongs_def":"to cut (off, down or asunder); by implication, to destroy or consume; specifically, to covenant (i.e. make an alliance or bargain, originally by cutting flesh and passing between the pieces)","kjv_def":"be chewed, be con-(feder-) ate, covenant, cut (down, off), destroy, fail, feller, be freed, hew (down), make a league (covenant), [idiom] lose, perish, [idiom] utterly, [idiom] want."},
"H3773":{"lemma":"כָּרֻתָה","xlit":"kâruthâh","pron":"kaw-rooth-aw'","derivation":"passive participle feminine of H3772 (כָּרַת);","strongs_def":"something cut, i.e. a hewn timber","kjv_def":"beam."},
"H3774":{"lemma":"כְּרֵתִי","xlit":"Kᵉrêthîy","pron":"ker-ay-thee'","derivation":"probably from H3772 (כָּרַת) in the sense of executioner; (compare H2876 (טַבָּח)) (only collectively in the singular as plural)","strongs_def":"a Kerethite or life-guardsman","kjv_def":"Cherethims, Cherethites."},
"H3775":{"lemma":"כֶּשֶׂב","xlit":"keseb","pron":"keh'-seb","derivation":"apparently by transposition for H3532 (כֶּבֶשׂ);","strongs_def":"a young sheep","kjv_def":"lamb."},
"H3776":{"lemma":"כִּשְׂבָּה","xlit":"kisbâh","pron":"kis-baw'","derivation":"feminine of H3775 (כֶּשֶׂב);","strongs_def":"a young ewe","kjv_def":"lamb."},
"H3777":{"lemma":"כֶּשֶׂד","xlit":"Kesed","pron":"keh'-sed","derivation":"from an unused root of uncertain meaning;","strongs_def":"Kesed, a relative of Abraham","kjv_def":"Chesed."},
"H3778":{"lemma":"כַּשְׂדִּי","xlit":"Kasdîy","pron":"kas-dee'","derivation":"(occasionally with enclitic) כַּשְׂדִּימָה; towards the Kasdites into Chaldea), patronymically from H3777 (כֶּשֶׂד) (only in the plural);","strongs_def":"a Kasdite, or descendant of Kesed; by implication, a Chaldaean (as if so descended); also an astrologer (as if proverbial of that people","kjv_def":"Chaldeans, Chaldees, inhabitants of Chaldea."},
"H3779":{"lemma":"כַּשְׂדַּי","xlit":"Kasday","pron":"kas-dah'-ee","derivation":"(Aramaic) corresponding to H3778 (כַּשְׂדִּי);","strongs_def":"a Chaldaean or inhabitant of Chaldaea; by implication, a Magian or professional astrologer","kjv_def":"Chaldean."},
"H3780":{"lemma":"כָּשָׂה","xlit":"kâsâh","pron":"kaw-saw'","derivation":"a primitive root;","strongs_def":"to grow fat (i.e. be covered with flesh)","kjv_def":"be covered. Compare H3680 (כָּסָה)."},
"H3781":{"lemma":"כַּשִּׁיל","xlit":"kashshîyl","pron":"kash-sheel'","derivation":"from H3782 (כָּשַׁל);","strongs_def":"properly, a feller, i.e. an axe","kjv_def":"ax."},
"H3782":{"lemma":"כָּשַׁל","xlit":"kâshal","pron":"kaw-shal'","derivation":"a primitive root;","strongs_def":"to totter or waver (through weakness of the legs, especially the ankle); by implication, to falter, stumble, faint or fall","kjv_def":"bereave (from the margin), cast down, be decayed, (cause to) fail, (cause, make to) fall (down, -ing), feeble, be (the) ruin(-ed, of), (be) overthrown, (cause to) stumble, [idiom] utterly, be weak."},
"H3783":{"lemma":"כִּשָּׁלוֹן","xlit":"kishshâlôwn","pron":"kish-shaw-lone'","derivation":"from H3782 (כָּשַׁל);","strongs_def":"properly, a tottering, i.e. ruin","kjv_def":"fall."},
"H3784":{"lemma":"כָּשַׁף","xlit":"kâshaph","pron":"kaw-shaf'","derivation":"a primitive root;","strongs_def":"properly, to whisper aspell, i.e. to inchant or practise magic","kjv_def":"sorcerer, (use) witch(-craft)."},
"H3785":{"lemma":"כֶּשֶׁף","xlit":"kesheph","pron":"keh'-shef","derivation":"from H3784 (כָּשַׁף);","strongs_def":"magic","kjv_def":"sorcery, witchcraft."},
"H3786":{"lemma":"כַּשָּׁף","xlit":"kashshâph","pron":"kash-shawf'","derivation":"from H3784 (כָּשַׁף);","strongs_def":"a magician","kjv_def":"sorcerer."},
"H3787":{"lemma":"כָּשֵׁר","xlit":"kâshêr","pron":"kaw-share'","derivation":"a primitive root properly, to be straight or right;","strongs_def":"by implication, to be acceptable; also to succeed or prosper","kjv_def":"direct, be right, prosper."},
"H3788":{"lemma":"כִּשְׁרוֹן","xlit":"kishrôwn","pron":"kish-rone'","derivation":"from H3787 (כָּשֵׁר);","strongs_def":"success, advantage","kjv_def":"equity, good, right."},
"H3789":{"lemma":"כָּתַב","xlit":"kâthab","pron":"kaw-thab'","derivation":"a primitive root;","strongs_def":"to grave, by implication, to write (describe, inscribe, prescribe, subscribe)","kjv_def":"describe, record, prescribe, subscribe, write(-ing, -ten)."},
"H3790":{"lemma":"כְּתַב","xlit":"kᵉthab","pron":"keth-ab'","derivation":"(Aramaic) corresponding to H3789 (כָּתַב)","strongs_def":"{to grave, by implication, to write (describe, inscribe, prescribe, subscribe)}","kjv_def":"write(-ten)."},
"H3791":{"lemma":"כָּתָב","xlit":"kâthâb","pron":"kaw-thawb'","derivation":"from H3789 (כָּתַב);","strongs_def":"something written, i.e. a writing, record or book","kjv_def":"register, scripture, writing."},
"H3792":{"lemma":"כְּתָב","xlit":"kᵉthâb","pron":"keth-awb'","derivation":"(Aramaic) corresponding to H3791 (כָּתָב)","strongs_def":"{something written, i.e. a writing, record or book}","kjv_def":"prescribing, writing(-ten)."},
"H3793":{"lemma":"כְּתֹבֶת","xlit":"kᵉthôbeth","pron":"keth-o'-beth","derivation":"from H3789 (כָּתַב);","strongs_def":"a letter or other mark branded on the skin","kjv_def":"[idiom] any (mark)."},
"H3794":{"lemma":"כִּתִּי","xlit":"Kittîy","pron":"kit-tee'","derivation":"or כִּתִּיִּי; patrial from an unused name denoting Cyprus (only in the plural);","strongs_def":"a Kittite or Cypriote; hence, an islander in general, i.e. the Greeks or Romans on the shores opposite Palestine","kjv_def":"Chittim, Kittim."},
"H3795":{"lemma":"כָּתִית","xlit":"kâthîyth","pron":"kaw-theeth'","derivation":"from H3807 (כָּתַת);","strongs_def":"beaten, i.e. pure (oil)","kjv_def":"beaten."},
"H3796":{"lemma":"כֹּתֶל","xlit":"kôthel","pron":"ko'-thel","derivation":"from an unused root meaning to compact;","strongs_def":"a wall (as gathering inmates)","kjv_def":"wall."},
"H3797":{"lemma":"כְּתַל","xlit":"kᵉthal","pron":"keth-al'","derivation":"(Aramaic) corresponding to H3796 (כֹּתֶל)","strongs_def":"{a wall (as gathering inmates)}","kjv_def":"wall."},
"H3798":{"lemma":"כִּתְלִישׁ","xlit":"Kithlîysh","pron":"kith-leesh'","derivation":"from H3796 (כֹּתֶל) and H376 (אִישׁ); wall of a man;","strongs_def":"Kithlish, a place in Palestine","kjv_def":"Kithlish."},
"H3799":{"lemma":"כָּתַם","xlit":"kâtham","pron":"kaw-tham'","derivation":"a primitive root;","strongs_def":"properly, to carve or engrave, i.e. (by implication) to inscribe indelibly","kjv_def":"mark."},
"H3800":{"lemma":"כֶּתֶם","xlit":"kethem","pron":"keh'-them","derivation":"from H3799 (כָּתַם);","strongs_def":"properly, something carved out, i.e. ore; hence, gold (pure as originally mined)","kjv_def":"((most) fine, pure) gold(-en wedge)."},
"H3801":{"lemma":"כְּתֹנֶת","xlit":"kᵉthôneth","pron":"keth-o'-neth","derivation":"or כֻּתֹּנֶת; from an unused root meaning to cover (compare H3802 (כָּתֵף));","strongs_def":"a shirt","kjv_def":"coat, garment, robe."},
"H3802":{"lemma":"כָּתֵף","xlit":"kâthêph","pron":"kaw-thafe'","derivation":"from an unused root meaning to clothe;","strongs_def":"the shoulder (proper, i.e. upper end of the arm; as being the spot where the garments hang); figuratively, side-piece or lateral projection of anything","kjv_def":"arm, corner, shoulder(-piece), side, undersetter."},
"H3803":{"lemma":"כָּתַר","xlit":"kâthar","pron":"kaw-thar'","derivation":"a primitive root;","strongs_def":"to enclose; hence (in a friendly sense) to crown, (in a hostile one) to besiege; also to wait (as restraining oneself)","kjv_def":"beset round, compass about, be crowned inclose round, suffer."},
"H3804":{"lemma":"כֶּתֶר","xlit":"kether","pron":"keh'-ther","derivation":"from H3803 (כָּתַר);","strongs_def":"properly, a circlet, i.e. a diadem","kjv_def":"crown."},
"H3805":{"lemma":"כֹתֶרֶת","xlit":"kôthereth","pron":"ko-theh'-reth","derivation":"feminine active participle of H3803 (כָּתַר);","strongs_def":"the capital of a column","kjv_def":"chapiter."},
"H3806":{"lemma":"כָּתַשׁ","xlit":"kâthash","pron":"kaw-thash'","derivation":"a primitive root;","strongs_def":"to butt or pound","kjv_def":"bray."},
"H3807":{"lemma":"כָּתַת","xlit":"kâthath","pron":"kaw-thath'","derivation":"a primitive root;","strongs_def":"to bruise or violently strike","kjv_def":"beat (down, to pieces), break in pieces, crushed, destroy, discomfit, smite, stamp. l"},
"H3808":{"lemma":"לֹא","xlit":"lôʼ","pron":"lo","derivation":"or לוֹא; or לֹה; (Deuteronomy 3:11), a primitive particle;","strongs_def":"not (the simple or abs. negation); by implication, no; often used with other particles","kjv_def":"[idiom] before, [phrase] or else, ere, [phrase] except, ig(-norant), much, less, nay, neither, never, no((-ne), -r, (-thing)), ([idiom] as though...,(can-), for) not (out of), of nought, otherwise, out of, [phrase] surely, [phrase] as truly as, [phrase] of a truth, [phrase] verily, for want, [phrase] whether, without."},
"H3809":{"lemma":"לָא","xlit":"lâʼ","pron":"law","derivation":"(Aramaic) or לָה; (Aramaic) (Daniel 4:32), corresponding to H3808 (לֹא)","strongs_def":"{not (the simple or abs. negation); by implication, no;}","kjv_def":"or even, neither, no(-ne, -r), (can-) not, as nothing, without."},
"H3810":{"lemma":"לֹא דְבַר","xlit":"Lôʼ Dᵉbar","pron":"lo deb-ar'","derivation":"or לוֹ דבַר; (Samuel 9:4,5), or לִדְבִר; (Joshua 13:26), (probably rather לֹדְבַר ); from H3808 (לֹא) and H1699 (דֹּבֶר); pastureless;","strongs_def":"Lo-Debar, a place in Palestine","kjv_def":"Debir, Lodebar."},
"H3811":{"lemma":"לָאָה","xlit":"lâʼâh","pron":"law-aw'","derivation":"a primitive root;","strongs_def":"to tire; (figuratively) to be (or make) disgusted","kjv_def":"faint, grieve, lothe, (be, make) weary (selves)."},
"H3812":{"lemma":"לֵאָה","xlit":"Lêʼâh","pron":"lay-aw'","derivation":"from H3811 (לָאָה); weary;","strongs_def":"Leah, a wife of Jacob","kjv_def":"Leah."},
"H3813":{"lemma":"לָאַט","xlit":"lâʼaṭ","pron":"law-at'","derivation":"a primitive root;","strongs_def":"to muffle","kjv_def":"cover."},
"H3814":{"lemma":"לָאט","xlit":"lâʼṭ","pron":"lawt","derivation":"from H3813 (לָאַט) (or perhaps for active participle of H3874 (לוּט));","strongs_def":"properly, muffled, i.e. silently","kjv_def":"softly."},
"H3815":{"lemma":"לָאֵל","xlit":"Lâʼêl","pron":"law-ale'","derivation":"from the prepositional prefix and H410 (אֵל); (belonging) to God;","strongs_def":"Lael an Israelite","kjv_def":"Lael."},
"H3816":{"lemma":"לְאֹם","xlit":"lᵉʼôm","pron":"leh-ome'","derivation":"or לְאוֹם; from an unused root meaning to gather;","strongs_def":"a community","kjv_def":"nation, people."},
"H3817":{"lemma":"לְאֻמִּים","xlit":"Lᵉʼummîym","pron":"leh-oom-meem'","derivation":"plural of H3816 (לְאֹם); communities;","strongs_def":"Leum mim, an Arabian","kjv_def":"Leummim."},
"H3818":{"lemma":"לֹא עַמִּי","xlit":"Lôʼ ʻAmmîy","pron":"lo am-mee'","derivation":"from H3808 (לֹא) and H5971 (עַם) with pronominal suffix; not my people;","strongs_def":"Lo-Ammi, the symbolic name of a son of Hosea","kjv_def":"Lo-ammi."},
"H3819":{"lemma":"לֹא רֻחָמָה","xlit":"Lôʼ Ruchâmâh","pron":"lo roo-khaw-maw'","derivation":"from H3808 (לֹא) and H7355 (רָחַם); not pitied;","strongs_def":"Lo-Ruchamah, the symbolic name of a son of Hosea","kjv_def":"Lo-ruhamah."},
"H3820":{"lemma":"לֵב","xlit":"lêb","pron":"labe","derivation":"a form of H3824 (לֵבָב);","strongs_def":"the heart; also used (figuratively) very widely for the feelings, the will and even the intellect; likewise for the centre of anything","kjv_def":"[phrase] care for, comfortably, consent, [idiom] considered, courag(-eous), friend(-ly), ((broken-), (hard-), (merry-), (stiff-), (stout-), double) heart(-ed), [idiom] heed, [idiom] I, kindly, midst, mind(-ed), [idiom] regard(-ed), [idiom] themselves, [idiom] unawares, understanding, [idiom] well, willingly, wisdom."},
"H3821":{"lemma":"לֵב","xlit":"lêb","pron":"labe","derivation":"(Aramaic) corresponding to H3820 (לֵב)","strongs_def":"{the heart; also used (figuratively) very widely for the feelings, the will and even the intellect; likewise for the centre of anything}","kjv_def":"heart."},
"H3822":{"lemma":"לְבָאוֹת","xlit":"Lᵉbâʼôwth","pron":"leb-aw-oth'","derivation":"plural of H3833 (לָבִיא); lionesses;","strongs_def":"Lebaoth, a place in Palestine","kjv_def":"Lebaoth. See also H1034 (בֵּית לְבָאוֹת)."},
"H3823":{"lemma":"לָבַב","xlit":"lâbab","pron":"law-bab'","derivation":"a primitive root; properly, to be enclosed (as if with fat); by implication (as denominative from H3824 (לֵבָב));  to unheart;  also as denominative from H3834 (לָבִיבָה) i.e.","strongs_def":"(in a good sense) transport (with love), or (in a bad sense) stultify; to make cakes","kjv_def":"make cakes, ravish, be wise."},
"H3824":{"lemma":"לֵבָב","xlit":"lêbâb","pron":"lay-bawb'","derivation":"from H3823 (לָבַב); used also like H3820 (לֵב)","strongs_def":"the heart (as the most interior organ);","kjv_def":"[phrase] bethink themselves, breast, comfortably, courage, ((faint), (tender-) heart(-ed), midst, mind, [idiom] unawares, understanding."},
"H3825":{"lemma":"לְבַב","xlit":"lᵉbab","pron":"leb-ab'","derivation":"(Aramaic) corresponding to H3824 (לֵבָב)","strongs_def":"{the heart (as the most interior organ);}","kjv_def":"heart."},
"H3826":{"lemma":"לִבָּה","xlit":"libbâh","pron":"lib-baw'","derivation":"feminine of H3820 (לֵב);","strongs_def":"the heart","kjv_def":"heart."},
"H3827":{"lemma":"לַבָּה","xlit":"labbâh","pron":"lab-baw'","derivation":"for H3852 (לֶהָבָה);","strongs_def":"flame","kjv_def":"flame."},
"H3828":{"lemma":"לְבוֹנָה","xlit":"lᵉbôwnâh","pron":"leb-o-naw'","derivation":"or לְבֹנָה; from H3836 (לָבָן);","strongs_def":"frankincense (from its whiteness or perhaps that of its smoke)","kjv_def":"(frank-) incense."},
"H3829":{"lemma":"לְבוֹנָה","xlit":"Lᵉbôwnâh","pron":"leb-o-naw'","derivation":"the same as H3828 (לְבוֹנָה);","strongs_def":"Lebonah, a place in Palestine","kjv_def":"Lebonah."},
"H3830":{"lemma":"לְבוּשׁ","xlit":"lᵉbûwsh","pron":"leb-oosh'","derivation":"or לְבֻשׁ; from H3847 (לָבַשׁ);","strongs_def":"a garment (literally or figuratively); by implication (euphemistically) a wife","kjv_def":"apparel, clothed with, clothing, garment, raiment, vestment, vesture."},
"H3831":{"lemma":"לְבוּשׁ","xlit":"lᵉbûwsh","pron":"leb-oosh'","derivation":"(Aramaic) corresponding to H3830 (לְבוּשׁ)","strongs_def":"{a garment (literally or figuratively); by implication (euphemistically) a wife}","kjv_def":"garment."},
"H3832":{"lemma":"לָבַט","xlit":"lâbaṭ","pron":"law-bat'","derivation":"a primitive root;","strongs_def":"to overthrow; intransposed, to fall","kjv_def":"fall."},
"H3833":{"lemma":"לָבִיא","xlit":"lâbîyʼ","pron":"law-bee'","derivation":"or (Ezekiel 19:2) לְבִיָּא; irregular masculine plural לְבָאִים; irregular feminine plural לְבָאוֹת; from an unused root meaning;  compare H738 (אֲרִי)","strongs_def":"to roar; a lion (properly, a lioness as the fiercer (although not a roarer;))","kjv_def":"(great, old, stout) lion, lioness, young (lion)."},
"H3834":{"lemma":"לָבִיבָה","xlit":"lâbîybâh","pron":"law-bee-baw'","derivation":"or rather לְבִבָה; from H3823 (לָבַב) in its original sense of fatness (or perhaps of folding);","strongs_def":"a cake (either as fried or turned)","kjv_def":"cake."},
"H3835":{"lemma":"לָבַן","xlit":"lâban","pron":"law-ban'","derivation":"a primitive root; also as denominative from H3843 (לְבֵנָה)","strongs_def":"to be (or become) white;  to make bricks","kjv_def":"make brick, be (made, make) white(-r)."},
"H3836":{"lemma":"לָבָן","xlit":"lâbân","pron":"law-bawn'","derivation":"or (Genesis 49:12) לָבֵן; from H3835 (לָבַן);","strongs_def":"white","kjv_def":"white."},
"H3837":{"lemma":"לָבָן","xlit":"Lâbân","pron":"law-bawn'","derivation":"the same as H3836 (לָבָן);","strongs_def":"Laban, a Mesopotamian; also a place in the Desert","kjv_def":"Laban."},
"H3838":{"lemma":"לְבָנָא","xlit":"Lᵉbânâʼ","pron":"leb-aw-naw'","derivation":"or לְבָנָה; the same as H3842 (לְבָנָה);","strongs_def":"Lebana or Lebanah, one of the Nethinim","kjv_def":"Lebana, Lebanah."},
"H3839":{"lemma":"לִבְנֶה","xlit":"libneh","pron":"lib-neh'","derivation":"from H3835 (לָבַן);","strongs_def":"some sort of whitish tree, perhaps the storax","kjv_def":"poplar."},
"H3840":{"lemma":"לִבְנָה","xlit":"libnâh","pron":"lib-naw'","derivation":"from H3835 (לָבַן);","strongs_def":"properly, whiteness, i.e. (by implication) transparency","kjv_def":"paved."},
"H3841":{"lemma":"לִבְנָה","xlit":"Libnâh","pron":"lib-naw'","derivation":"the same as H3839 (לִבְנֶה);","strongs_def":"Libnah, a place in the Desert and one in Palestine","kjv_def":"Libnah."},
"H3842":{"lemma":"לְבָנָה","xlit":"lᵉbânâh","pron":"leb-aw-naw'","derivation":"from H3835 (לָבַן);","strongs_def":"properly, (the) white, i.e. the moon","kjv_def":"moon. See also H3838 (לְבָנָא)."},
"H3843":{"lemma":"לְבֵנָה","xlit":"lᵉbênâh","pron":"leb-ay-naw'","derivation":"from H3835 (לָבַן);","strongs_def":"a brick (from the whiteness of the clay)","kjv_def":"(altar of) brick, tile."},
"H3844":{"lemma":"לְבָנוֹן","xlit":"Lᵉbânôwn","pron":"leb-aw-nohn'","derivation":"from H3825 (לְבַב); (the) white mountain (from its snow);","strongs_def":"Lebanon, a mountain range in Palestine","kjv_def":"Lebanon."},
"H3845":{"lemma":"לִבְנִי","xlit":"Libnîy","pron":"lib-nee'","derivation":"from H3835 (לָבַן); white;","strongs_def":"Libni, an Israelite","kjv_def":"Libni."},
"H3846":{"lemma":"לִבְנִי","xlit":"Libnîy","pron":"lib-nee'","derivation":"patronymically from H3845 (לִבְנִי);","strongs_def":"a Libnite or descendants of Libni (collectively)","kjv_def":"Libnites."},
"H3847":{"lemma":"לָבַשׁ","xlit":"lâbash","pron":"law-bash'","derivation":"or לָבֵשׁ; a primitive root;","strongs_def":"properly, wrap around, i.e. (by implication) to put on agarment or clothe (oneself, or another), literally or figuratively","kjv_def":"(in) apparel, arm, array (self), clothe (self), come upon, put (on, upon), wear."},
"H3848":{"lemma":"לְבַשׁ","xlit":"lᵉbash","pron":"leb-ash'","derivation":"(Aramaic) corresponding to H3847 (לָבַשׁ)","strongs_def":"{properly, wrap around, i.e. (by implication) to put on agarment or clothe (oneself, or another), literally or figuratively}","kjv_def":"clothe."},
"H3849":{"lemma":"לֹג","xlit":"lôg","pron":"lohg","derivation":"from an unused root apparently meaning to deepen or hollow (like H3537 (כַּד));","strongs_def":"a log or measure forliquids","kjv_def":"log (of oil)."},
"H3850":{"lemma":"לֹד","xlit":"Lôd","pron":"lode","derivation":"from an unused root of uncertain signification;","strongs_def":"Lod, a place in Palestine","kjv_def":"Lod."},
"H3851":{"lemma":"לַהַב","xlit":"lahab","pron":"lah'-hab","derivation":"from an usused root meaning to gleam;","strongs_def":"a flash; figuratively, a sharply polished blade or point of aweapon","kjv_def":"blade, bright, flame, glittering."},
"H3852":{"lemma":"לֶהָבָה","xlit":"lehâbâh","pron":"leh-aw-baw'","derivation":"or לַהֶבֶת; feminine of H3851 (לַהַב), and meaning the same","strongs_def":"{a flash; figuratively, a sharply polished blade or point of aweapon}","kjv_def":"flame(-ming), head (of a spear)."},
"H3853":{"lemma":"לְהָבִים","xlit":"Lᵉhâbîym","pron":"leh-haw-beem'","derivation":"plural of H3851 (לַהַב); flames;","strongs_def":"Lehabim, a son of Mizraim, and his descendants","kjv_def":"Lehabim."},
"H3854":{"lemma":"לַהַג","xlit":"lahag","pron":"lah'-hag","derivation":"from an unused root meaning to be eager;","strongs_def":"intense mental application","kjv_def":"study."},
"H3855":{"lemma":"לַהַד","xlit":"Lahad","pron":"lah'-had","derivation":"from an unused root meaning to glow (compare H3851 (לַהַב)) or else to be earnest (compare H3854 (לַהַג));","strongs_def":"Lahad, an Israelite","kjv_def":"Lahad."},
"H3856":{"lemma":"לָהַהּ","xlit":"lâhahh","pron":"law-hah'","derivation":"a primitive root meaning properly, to burn, i.e. (by implication)","strongs_def":"to be rabid (figuratively, insane); also (from the exhaustion of frenzy) to languish","kjv_def":"faint, mad."},
"H3857":{"lemma":"לָהַט","xlit":"lâhaṭ","pron":"law-hat'","derivation":"a primitive root;","strongs_def":"properly, to lick, i.e. (by implication) to blaze","kjv_def":"burn (up), set on fire, flaming, kindle."},
"H3858":{"lemma":"לַהַט","xlit":"lahaṭ","pron":"lah'-hat","derivation":"from H3857 (לָהַט);","strongs_def":"a blaze; also (from the idea of enwrapping) magic (as covert)","kjv_def":"flaming, enchantment."},
"H3859":{"lemma":"לָהַם","xlit":"lâham","pron":"law-ham'","derivation":"a primitive root;","strongs_def":"properly, to burn in, i.e. (figuratively) to rankle","kjv_def":"wound."},
"H3860":{"lemma":"לָהֵן","xlit":"lâhên","pron":"law-hane'","derivation":"from the prepositional prefix meaning to or for and H2005 (הֵן);","strongs_def":"popularly for if; hence, therefore","kjv_def":"for them (by mistake for prepositional suffix)."},
"H3861":{"lemma":"לָהֵן","xlit":"lâhên","pron":"law-hane'","derivation":"(Aramaic) corresponding to H3860 (לָהֵן);","strongs_def":"therefore; also except","kjv_def":"but, except, save, therefore, wherefore."},
"H3862":{"lemma":"לַהֲקָה","xlit":"lahăqâh","pron":"lah-hak-aw'","derivation":"probably from an unused root meaning to gather;","strongs_def":"an assembly","kjv_def":"company."},
"H3863":{"lemma":"לוּא","xlit":"lûwʼ","pron":"loo","derivation":"or לֻא; or לוּ;","strongs_def":"a conditional particle; if; by implication (interj. as a wish) would that!","kjv_def":"if (haply), peradventure, I pray thee, though, I would, would God (that)."},
"H3864":{"lemma":"לוּבִי","xlit":"Lûwbîy","pron":"loo-bee'","derivation":"or לֻבִּי; (Daniel 11:43), partrial from a name probably derived from an unused root meaning to thirst, i.e. a dry region; apparently","strongs_def":"a Libyan or inhabitant of interior Africa (only in plural)","kjv_def":"Lubim(-s), Libyans."},
"H3865":{"lemma":"לוּד","xlit":"Lûwd","pron":"lood","derivation":"probably of foreign derivation;","strongs_def":"Lud, the name of two nations","kjv_def":"Lud, Lydia."},
"H3866":{"lemma":"לוּדִי","xlit":"Lûwdîy","pron":"loo-dee'","derivation":"or לוּדִיִּי; patrial from H3865 (לוּד);","strongs_def":"a Ludite or inhabitants of Lud (only in plural)","kjv_def":"Ludim. Lydians."},
"H3867":{"lemma":"לָוָה","xlit":"lâvâh","pron":"law-vaw'","derivation":"a primitive root;","strongs_def":"properly, to twine, i.e. (by implication) to unite, to remain; also to borrow (as a form of obligation) or (caus.) to lend","kjv_def":"abide with, borrow(-er), cleave, join (self), lend(-er)."},
"H3868":{"lemma":"לוּז","xlit":"lûwz","pron":"looz","derivation":"a primitive root; (compare H3867 (לָוָה), H3874 (לוּט) and H3885 (לוּן))","strongs_def":"to turn aside , i.e. (literally) to depart, (figuratively) be perverse","kjv_def":"depart, froward, perverse(-ness)."},
"H3869":{"lemma":"לוּז","xlit":"lûwz","pron":"looz","derivation":"probably of foreign origin;","strongs_def":"some kind of nuttree, perhaps the almond","kjv_def":"hazel."},
"H3870":{"lemma":"לוּז","xlit":"Lûwz","pron":"looz","derivation":"probably from H3869 (לוּז) (as growing there);","strongs_def":"Luz, the name of two places in Palestine","kjv_def":"Luz."},
"H3871":{"lemma":"לוּחַ","xlit":"lûwach","pron":"loo'-akh","derivation":"or לֻחַ; from a primitive root;","strongs_def":"probably meaning to glisten; a tablet (as polished), of stone, wood or metal","kjv_def":"board, plate, table."},
"H3872":{"lemma":"לוּחִית","xlit":"Lûwchîyth","pron":"loo-kheeth'","derivation":"or לֻחוֹת; (Jeremiah 48:5), from the same as H3871 (לוּחַ); floored;","strongs_def":"Luchith, a place East of the Jordan","kjv_def":"Luhith."},
"H3873":{"lemma":"לוֹחֵשׁ","xlit":"Lôwchêsh","pron":"lo-khashe'","derivation":"active participle of H3907 (לָחַשׁ); (the) enchanter;","strongs_def":"Lochesh, an Israelite","kjv_def":"Hallohesh, Haloshesh (includ. the article)."},
"H3874":{"lemma":"לוּט","xlit":"lûwṭ","pron":"loot","derivation":"a primitive root;","strongs_def":"to wrap up","kjv_def":"cast, wrap."},
"H3875":{"lemma":"לוֹט","xlit":"lôwṭ","pron":"lote","derivation":"from H3874 (לוּט);","strongs_def":"a veil","kjv_def":"covering."},
"H3876":{"lemma":"לוֹט","xlit":"Lôwṭ","pron":"lote","derivation":"the same as H3875 (לוֹט);","strongs_def":"Lot, Abraham's nephew","kjv_def":"Lot."},
"H3877":{"lemma":"לוֹטָן","xlit":"Lôwṭân","pron":"lo-tawn'","derivation":"from H3875 (לוֹט); covering;","strongs_def":"Lotan, an Idumaean","kjv_def":"Lotan."},
"H3878":{"lemma":"לֵוִי","xlit":"Lêvîy","pron":"lay-vee'","derivation":"from H3867 (לָוָה); attached;","strongs_def":"Levi, a son of Jacob","kjv_def":"Levi. See also H3879 (לֵוִי), H3881 (לֵוִיִּי)."},
"H3879":{"lemma":"לֵוִי","xlit":"Lêvîy","pron":"lay-vee'","derivation":"(Aramaic) corresponding to H3880 (לִוְיָה)","strongs_def":"{something attached, i.e. a wreath}","kjv_def":"Levite."},
"H3880":{"lemma":"לִוְיָה","xlit":"livyâh","pron":"liv-yaw'","derivation":"from H3867 (לָוָה);","strongs_def":"something attached, i.e. a wreath","kjv_def":"ornament."},
"H3881":{"lemma":"לֵוִיִּי","xlit":"Lêvîyîy","pron":"lay-vee-ee'","derivation":"or לֵוִי; patronymically from H3878 (לֵוִי);","strongs_def":"a Levite or descendant of Levi","kjv_def":"Leviite."},
"H3882":{"lemma":"לִוְיָתָן","xlit":"livyâthân","pron":"liv-yaw-thawn'","derivation":"from H3867 (לָוָה);","strongs_def":"a wreathed animal, i.e. a serpent (especially the crocodile or some other large sea-monster); figuratively, the constellation of the dragon; also as a symbol of Babylon","kjv_def":"leviathan, mourning."},
"H3883":{"lemma":"לוּל","xlit":"lûwl","pron":"lool","derivation":"from an unused root meaning to fold back;","strongs_def":"a spiral step","kjv_def":"winding stair. Compare H3924 (לֻלָאָה)."},
"H3884":{"lemma":"לוּלֵא","xlit":"lûwlêʼ","pron":"loo-lay'","derivation":"or לוּלֵי; from H3863 (לוּא) and H3808 (לֹא);","strongs_def":"if not","kjv_def":"except, had not, if (...not), unless, were it not that."},
"H3885":{"lemma":"לוּן","xlit":"lûwn","pron":"loon","derivation":"or לִין; a primitive root;","strongs_def":"to stop (usually over night); by implication, to stay permanently; hence (in a bad sense) to be obstinate (especially in words, to complain)","kjv_def":"abide (all night), continue, dwell, endure, grudge, be left, lie all night, (cause to) lodge (all night, in, -ing, this night), (make to) murmur, remain, tarry (all night, that night)."},
"H3886":{"lemma":"לוּעַ","xlit":"lûwaʻ","pron":"loo'-ah","derivation":"a primitive root;","strongs_def":"to gulp; figuratively, to be rash","kjv_def":"swallow down (up)."},
"H3887":{"lemma":"לוּץ","xlit":"lûwts","pron":"loots","derivation":"a primitive root;","strongs_def":"properly, to make mouths at, i.e. to scoff; hence (from the effort to pronounce a foreign language) to interpret, or (generally) intercede","kjv_def":"ambassador, have in derision, interpreter, make a mock, mocker, scorn(-er, -ful), teacher."},
"H3888":{"lemma":"לוּשׁ","xlit":"lûwsh","pron":"loosh","derivation":"a primitive root;","strongs_def":"to knead","kjv_def":"knead."},
"H3889":{"lemma":"לוּשׁ","xlit":"Lûwsh","pron":"loosh","derivation":"from H3888 (לוּשׁ); kneading;","strongs_def":"Lush, a place in Palestine","kjv_def":"Laish (from the margin). Compare H3919 (לַיִשׁ)."},
"H3890":{"lemma":"לְוָת","xlit":"lᵉvâth","pron":"lev-awth'","derivation":"(Aramaic) from a root corresponding to H3867 (לָוָה);","strongs_def":"properly, adhesion, i.e. (as preposition) with","kjv_def":"[idiom] thee."},
"H3891":{"lemma":"לְזוּת","xlit":"lᵉzûwth","pron":"lez-ooth'","derivation":"from H3868 (לוּז);","strongs_def":"perverseness","kjv_def":"perverse."},
"H3892":{"lemma":"לַח","xlit":"lach","pron":"lakh","derivation":"from an unused root meaning to be new;","strongs_def":"fresh, i.e. unused or undried","kjv_def":"green, moist."},
"H3893":{"lemma":"לֵחַ","xlit":"lêach","pron":"lay'-akh","derivation":"from the same as H3892 (לַח);","strongs_def":"freshness, i.e. vigor","kjv_def":"natural force."},
"H3894":{"lemma":"לָחוּם","xlit":"lâchûwm","pron":"law-khoom'","derivation":"or לָחֻם; passive participle of H3898 (לָחַם);","strongs_def":"properly, eaten, i.e. food; also flesh, i.e. body","kjv_def":"while...is eating, flesh."},
"H3895":{"lemma":"לְחִי","xlit":"lᵉchîy","pron":"lekh-ee'","derivation":"from an unused root meaning to be soft;","strongs_def":"the cheek (from its fleshiness); hence, the jaw-bone","kjv_def":"cheek (bone), jaw (bone)."},
"H3896":{"lemma":"לֶחִי","xlit":"Lechîy","pron":"lekh'-ee","derivation":"a form of H3895 (לְחִי);","strongs_def":"Lechi, a place in Palestine","kjv_def":"Lehi. Compare also H7437 (רָמַת לֶחִי)."},
"H3897":{"lemma":"לָחַךְ","xlit":"lâchak","pron":"law-khak'","derivation":"a primitive root;","strongs_def":"to lick","kjv_def":"lick (up)."},
"H3898":{"lemma":"לָחַם","xlit":"lâcham","pron":"law-kham'","derivation":"a primitive root;","strongs_def":"to feed on; figuratively, to consume; by implication, to battle (as destruction)","kjv_def":"devour, eat, [idiom] ever, fight(-ing), overcome, prevail, (make) war(-ring)."},
"H3899":{"lemma":"לֶחֶם","xlit":"lechem","pron":"lekh'-em","derivation":"from H3898 (לָחַם); See also H1036 (בֵּית לְעַפְרָה)","strongs_def":"food (for man or beast), especially bread, or grain (for making it)","kjv_def":"(shew-) bread, [idiom] eat, food, fruit, loaf, meat, victuals."},
"H3900":{"lemma":"לְחֶם","xlit":"lᵉchem","pron":"lekh-em'","derivation":"(Aramaic) corresponding to H3899 (לֶחֶם)","strongs_def":"{(for man or beast), especially bread, or grain (for making it)}","kjv_def":"feast."},
"H3901":{"lemma":"לָחֶם","xlit":"lâchem","pron":"law-khem'","derivation":"from H3898 (לָחַם),","strongs_def":"battle","kjv_def":"war."},
"H3902":{"lemma":"לַחְמִי","xlit":"Lachmîy","pron":"lakh-mee'","derivation":"from H3899 (לֶחֶם); foodful; (or perhaps erroneous transcription) for H1022 (בֵּית הַלַּחְמִי)","strongs_def":"Lachmi, an Israelite; or rather probably a brief form","kjv_def":"Lahmi. See also H3433 (יָשֻׁבִי לֶחֶם)."},
"H3903":{"lemma":"לַחְמָס","xlit":"Lachmâç","pron":"lakh-maws'","derivation":"probably by erroneous transcription for לַחְמָם; from H3899 (לֶחֶם); food-like;","strongs_def":"Lachmam or Lachmas, a place in Palestine","kjv_def":"Lahmam."},
"H3904":{"lemma":"לְחֵנָה","xlit":"lᵉchênâh","pron":"lekh-ay-naw'","derivation":"(Aramaic) from an unused root of uncertain meaning;","strongs_def":"a concubine","kjv_def":"concubine."},
"H3905":{"lemma":"לָחַץ","xlit":"lâchats","pron":"law-khats'","derivation":"a primitive root;","strongs_def":"properly, to press, i.e. (figuratively) to distress","kjv_def":"afflict, crush, force, hold fast, oppress(-or), thrust self."},
"H3906":{"lemma":"לַחַץ","xlit":"lachats","pron":"lakh'-ats","derivation":"from H3905 (לָחַץ);","strongs_def":"distress","kjv_def":"affliction, oppression."},
"H3907":{"lemma":"לָחַשׁ","xlit":"lâchash","pron":"law-khash'","derivation":"a primitive root;","strongs_def":"to whisper; by implication, to mumble aspell (as a magician)","kjv_def":"charmer, whisper (together)."},
"H3908":{"lemma":"לַחַשׁ","xlit":"lachash","pron":"lakh'-ash","derivation":"from H3907 (לָחַשׁ);","strongs_def":"properly, a whisper, i.e. by implication, (in a good sense) a private prayer, (in a bad one) an incantation; concretely, an amulet","kjv_def":"charmed, earring, enchantment, orator, prayer."},
"H3909":{"lemma":"לָט","xlit":"lâṭ","pron":"lawt","derivation":"a form of H3814 (לָאט) or else participle from H3874 (לוּט);","strongs_def":"properly, covered, i.e. secret; by implication, incantation; also secrecy or (adverb) covertly","kjv_def":"enchantment, privily, secretly, softly."},
"H3910":{"lemma":"לֹט","xlit":"lôṭ","pron":"lote","derivation":"probably from H3874 (לוּט);","strongs_def":"a gum (from its sticky nature), probably ladanum","kjv_def":"myrrh."},
"H3911":{"lemma":"לְטָאָה","xlit":"lᵉṭâʼâh","pron":"let-aw-aw'","derivation":"from an unused root meaning to hide;","strongs_def":"a kind of lizard (from its covert habits)","kjv_def":"lizard."},
"H3912":{"lemma":"לְטוּשִׁם","xlit":"Lᵉṭûwshim","pron":"let-oo-sheem'","derivation":"masculine plural of passive participle of H3913 (לָטַשׁ); hammered (i.e. oppressed) ones;","strongs_def":"Letushim, an Arabian tribe","kjv_def":"Letushim."},
"H3913":{"lemma":"לָטַשׁ","xlit":"lâṭash","pron":"law-tash'","derivation":"a primitive root;","strongs_def":"properly, to hammer out (an edge), i.e. to sharpen","kjv_def":"instructer, sharp(-en), whet."},
"H3914":{"lemma":"לֹיָה","xlit":"lôyâh","pron":"lo-yaw'","derivation":"a form of H3880 (לִוְיָה);","strongs_def":"a wreath","kjv_def":"addition."},
"H3915":{"lemma":"לַיִל","xlit":"layil","pron":"lah'-yil","derivation":"or (Isaiah 21:11) לֵיל; also לַיְלָה; from the same as H3883 (לוּל);","strongs_def":"properly, a twist (away of the light), i.e. night; figuratively, adversity","kjv_def":"(mid-)night (season)."},
"H3916":{"lemma":"לֵילְיָא","xlit":"lêylᵉyâʼ","pron":"lay-leh-yaw'","derivation":"(Aramaic) corresponding to H3815 (לָאֵל)","strongs_def":"{Lael an Israelite}","kjv_def":"night."},
"H3917":{"lemma":"לִילִית","xlit":"lîylîyth","pron":"lee-leeth'","derivation":"from H3915 (לַיִל);","strongs_def":"a night spectre","kjv_def":"screech owl."},
"H3918":{"lemma":"לַיִשׁ","xlit":"layish","pron":"lah'-yish","derivation":"from H3888 (לוּשׁ) in the sense of crushing;","strongs_def":"a lion (from his destructive blows)","kjv_def":"(old) lion."},
"H3919":{"lemma":"לַיִשׁ","xlit":"Layish","pron":"lah'-yish","derivation":"the same as H3918 (לַיִשׁ);","strongs_def":"Laish, the name of two places in Palestine","kjv_def":"Laish. Compare H3889 (לוּשׁ)."},
"H3920":{"lemma":"לָכַד","xlit":"lâkad","pron":"law-kad'","derivation":"a primitive root;","strongs_def":"to catch (in a net, trap or pit); generally, to capture or occupy; also to choose (by lot); figuratively, to cohere","kjv_def":"[idiom] at all, catch (self), be frozen, be holden, stick together, take."},
"H3921":{"lemma":"לֶכֶד","xlit":"leked","pron":"leh'ked","derivation":"from H3920 (לָכַד);","strongs_def":"something to capture with, i.e. a noose","kjv_def":"being taken."},
"H3922":{"lemma":"לֵכָה","xlit":"lêkâh","pron":"lay-kaw'","derivation":"from H3212 (יָלַךְ); a journey;","strongs_def":"Lekah, a place in Palestine","kjv_def":"Lecah."},
"H3923":{"lemma":"לָכִישׁ","xlit":"Lâkîysh","pron":"law-keesh'","derivation":"from an unused root of uncertain meaning;","strongs_def":"Lakish, a place in Palestine","kjv_def":"Lachish."},
"H3924":{"lemma":"לֻלָאָה","xlit":"lulâʼâh","pron":"loo-law-aw'","derivation":"from the same as H3883 (לוּל);","strongs_def":"a loop","kjv_def":"loop."},
"H3925":{"lemma":"לָמַד","xlit":"lâmad","pron":"law-mad'","derivation":"a primitive root;","strongs_def":"properly, to goad, i.e. (by implication) to teach (the rod being an Oriental incentive)","kjv_def":"(un-) accustomed, [idiom] diligently, expert, instruct, learn, skilful, teach(-er, -ing)."},
"H3926":{"lemma":"לְמוֹ","xlit":"lᵉmôw","pron":"lem-o'","derivation":"a prolonged and separable form of the prepositional prefix;","strongs_def":"to or for","kjv_def":"at, for, to, upon."},
"H3927":{"lemma":"לְמוּאֵל","xlit":"Lᵉmûwʼêl","pron":"lem-oo-ale'","derivation":"or לְמוֹאֵל; from H3926 (לְמוֹ) and H410 (אֵל); (belonging) to God;","strongs_def":"Lemuel or Lemoel, a symbolic name of Solomon","kjv_def":"Lemuel."},
"H3928":{"lemma":"לִמּוּד","xlit":"limmûwd","pron":"lim-mood'","derivation":"or לִמֻּד; from H3925 (לָמַד);","strongs_def":"instructed","kjv_def":"accustomed, disciple, learned, taught, used."},
"H3929":{"lemma":"לֶמֶךְ","xlit":"Lemek","pron":"leh'-mek","derivation":"from an unused root of uncertain meaning;","strongs_def":"Lemek, the name of two antediluvian patriarchs","kjv_def":"Lamech."},
"H3930":{"lemma":"לֹעַ","xlit":"lôaʻ","pron":"lo'ah","derivation":"from H3886 (לוּעַ);","strongs_def":"the gullet","kjv_def":"throat."},
"H3931":{"lemma":"לָעַב","xlit":"lâʻab","pron":"law-ab'","derivation":"a primitive root;","strongs_def":"to deride","kjv_def":"mock."},
"H3932":{"lemma":"לָעַג","xlit":"lâʻag","pron":"law-ag'","derivation":"a primitive root;","strongs_def":"to deride; by implication (as if imitating a foreigner) to speak unintelligibly","kjv_def":"have in derision, laugh (to scorn), mock (on), stammering."},
"H3933":{"lemma":"לַעַג","xlit":"laʻag","pron":"lah'-ag","derivation":"from H3932 (לָעַג);","strongs_def":"derision, scoffing","kjv_def":"derision, scorn (-ing)."},
"H3934":{"lemma":"לָעֵג","xlit":"lâʻêg","pron":"law-ayg'","derivation":"from H3932 (לָעַג);","strongs_def":"a buffoon; also a foreigner","kjv_def":"mocker, stammering."},
"H3935":{"lemma":"לַעְדָּה","xlit":"Laʻdâh","pron":"lah-daw'","derivation":"from an unused root of uncertain meaning;","strongs_def":"Ladah, an Israelite","kjv_def":"Laadah."},
"H3936":{"lemma":"לַעְדָּן","xlit":"Laʻdân","pron":"lah-dawn'","derivation":"from the same as H3935 (לַעְדָּה);","strongs_def":"Ladan, the name of two Israelites","kjv_def":"Laadan."},
"H3937":{"lemma":"לָעַז","xlit":"lâʻaz","pron":"law-az'","derivation":"a primitive root;","strongs_def":"to speak in a foreign tongue","kjv_def":"strange language."},
"H3938":{"lemma":"לָעַט","xlit":"lâʻaṭ","pron":"law-at'","derivation":"a primitive root;","strongs_def":"to swallow greedily; causatively, to feed","kjv_def":"feed."},
"H3939":{"lemma":"לַעֲנָה","xlit":"laʻănâh","pron":"lah-an-aw'","derivation":"from an unused root supposed to mean to curse;","strongs_def":"wormwood (regarded as poisonous, and therefore accursed)","kjv_def":"hemlock, wormwood."},
"H3940":{"lemma":"לַפִּיד","xlit":"lappîyd","pron":"lap-peed'","derivation":"or לַפִּד; from an unused root probably meaning to shine;","strongs_def":"a flambeau, lamp or flame","kjv_def":"(fire-) brand, (burning) lamp, lightning, torch."},
"H3941":{"lemma":"לַפִּידוֹת","xlit":"Lappîydôwth","pron":"lap-pee-doth'","derivation":"feminine plural of H3940 (לַפִּיד);","strongs_def":"Lappidoth, the husband of Deborah","kjv_def":"Lappidoth."},
"H3942":{"lemma":"לִפְנַי","xlit":"liphnay","pron":"lif-nah'ee","derivation":"from the prepositional prefix (to or for) and H6440 (פָּנִים);","strongs_def":"anterior","kjv_def":"before."},
"H3943":{"lemma":"לָפַת","xlit":"lâphath","pron":"law-fath'","derivation":"a primitive root;","strongs_def":"properly, to bend, i.e. (by implication) to clasp; also (reflexively) to turn around or aside","kjv_def":"take hold, turn aside (self)."},
"H3944":{"lemma":"לָצוֹן","xlit":"lâtsôwn","pron":"law-tsone'","derivation":"from H3887 (לוּץ);","strongs_def":"derision","kjv_def":"scornful(-ning)."},
"H3945":{"lemma":"לָצַץ","xlit":"lâtsats","pron":"law-tsats'","derivation":"a primitive root;","strongs_def":"to deride","kjv_def":"scorn."},
"H3946":{"lemma":"לַקּוּם","xlit":"Laqqûwm","pron":"lak-koom'","derivation":"from an unused root thought to mean to stop up by a barricade; perhaps fortification;","strongs_def":"Lakkum, a place in Palestine","kjv_def":"Lakum."},
"H3947":{"lemma":"לָקַח","xlit":"lâqach","pron":"law-kakh'","derivation":"a primitive root;","strongs_def":"to take (in the widest variety of applications)","kjv_def":"accept, bring, buy, carry away, drawn, fetch, get, infold, [idiom] many, mingle, place, receive(-ing), reserve, seize, send for, take (away, -ing, up), use, win."},
"H3948":{"lemma":"לֶקַח","xlit":"leqach","pron":"leh'-kakh","derivation":"from H3947 (לָקַח);","strongs_def":"properly, something received, i.e. (mentally) instruction (whether on the part of the teacher or hearer); also (in an active and sinister sense) inveiglement","kjv_def":"doctrine, learning, fair speech."},
"H3949":{"lemma":"לִקְחִי","xlit":"Liqchîy","pron":"lik-khee'","derivation":"from H3947 (לָקַח); learned;","strongs_def":"Likchi, an Israelite","kjv_def":"Likhi."},
"H3950":{"lemma":"לָקַט","xlit":"lâqaṭ","pron":"law-kat'","derivation":"a primitive root;","strongs_def":"properly, to pick up, i.e. (generally) to gather; specifically, to glean","kjv_def":"gather (up), glean."},
"H3951":{"lemma":"לֶקֶט","xlit":"leqeṭ","pron":"leh'-ket","derivation":"from H3950 (לָקַט);","strongs_def":"the gleaning","kjv_def":"gleaning."},
"H3952":{"lemma":"לָקַק","xlit":"lâqaq","pron":"law-kak'","derivation":"a primitive root;","strongs_def":"to lick or lap","kjv_def":"lap, lick."},
"H3953":{"lemma":"לָקַשׁ","xlit":"lâqash","pron":"law-kash'","derivation":"a primitive root;","strongs_def":"to gather the after crop","kjv_def":"gather."},
"H3954":{"lemma":"לֶקֶשׁ","xlit":"leqesh","pron":"leh'-kesh","derivation":"from H3953 (לָקַשׁ);","strongs_def":"the after crop","kjv_def":"latter growth."},
"H3955":{"lemma":"לְשַׁד","xlit":"lᵉshad","pron":"lesh-ad'","derivation":"from an unused root of uncertain meaning; apparently juice, i.e.","strongs_def":"(figuratively) vigor; also a sweet or fat cake","kjv_def":"fresh, moisture."},
"H3956":{"lemma":"לָשׁוֹן","xlit":"lâshôwn","pron":"law-shone'","derivation":"or לָשֹׁן; also (in plural) feminine לְשֹׁנָה; from H3960 (לָשַׁן);","strongs_def":"the tongue (of man or animals), used literally (as the instrument of licking, eating, or speech), and figuratively (speech, an ingot, a fork of flame, a cove of water)","kjv_def":"[phrase] babbler, bay, [phrase] evil speaker, language, talker, tongue, wedge."},
"H3957":{"lemma":"לִשְׁכָּה","xlit":"lishkâh","pron":"lish-kaw'","derivation":"from an unused root of uncertain meaning;","strongs_def":"a room in a building (whether for storage, eating, or lodging)","kjv_def":"chamber, parlour. Compare H5393 (נִשְׁכָּה)."},
"H3958":{"lemma":"לֶשֶׁם","xlit":"leshem","pron":"leh'-shem","derivation":"from an unused root of uncertain meaning;","strongs_def":"a gem, perhaps the jacinth","kjv_def":"ligure."},
"H3959":{"lemma":"לֶשֶׁם","xlit":"Leshem","pron":"leh'-shem","derivation":"the same as H3958 (לֶשֶׁם);","strongs_def":"Leshem, a place in Palestine","kjv_def":"Leshem."},
"H3960":{"lemma":"לָשַׁן","xlit":"lâshan","pron":"law-shan'","derivation":"a primitive root; properly, to lick; but used only as a denominative from H3956 (לָשׁוֹן);","strongs_def":"to wag the tongue, i.e. to calumniate","kjv_def":"accuse, slander."},
"H3961":{"lemma":"לִשָּׁן","xlit":"lishshân","pron":"lish-shawn'","derivation":"(Aramaic) corresponding to H3956 (לָשׁוֹן);","strongs_def":"speech, i.e. a nation","kjv_def":"language."},
"H3962":{"lemma":"לֶשַׁע","xlit":"Leshaʻ","pron":"leh'-shah","derivation":"from an unused root thought to mean to break through; a boiling spring;","strongs_def":"Lesha, a place probably East of the Jordan","kjv_def":"Lasha."},
"H3963":{"lemma":"לֶתֶךְ","xlit":"lethek","pron":"leh'-thek","derivation":"from an unused root of uncertain meaning;","strongs_def":"a measure for things dry","kjv_def":"half homer. m"},
"H3964":{"lemma":"מָא","xlit":"mâʼ","pron":"maw","derivation":"(Aramaic) corresponding to H4100 (מָה);","strongs_def":"(as indefinitely) that","kjv_def":"[phrase] what."},
"H3965":{"lemma":"מַאֲבוּס","xlit":"maʼăbûwç","pron":"mah-ab-ooce'","derivation":"from H75 (אָבַס);","strongs_def":"a granary","kjv_def":"storehouse."},
"H3966":{"lemma":"מְאֹד","xlit":"mᵉʼôd","pron":"meh-ode'","derivation":"from the same as H181 (אוּד);","strongs_def":"properly, vehemence, i.e. (with or without preposition) vehemently; by implication, wholly, speedily, etc. (often with other words as an intensive or superlative; especially when repeated)","kjv_def":"diligently, especially, exceeding(-ly), far, fast, good, great(-ly), [idiom] louder and louder, might(-ily, -y), (so) much, quickly, (so) sore, utterly, very ([phrase] much, sore), well."},
"H3967":{"lemma":"מֵאָה","xlit":"mêʼâh","pron":"may-aw'","derivation":"or מֵאיָה; properly, a primitive numeral","strongs_def":"a hundred; also as a multiplicative and a fraction","kjv_def":"hundred((-fold), -th), [phrase] sixscore."},
"H3968":{"lemma":"מֵאָה","xlit":"Mêʼâh","pron":"may-aw'","derivation":"the same as H3967 (מֵאָה);","strongs_def":"Meah, a tower in Jerusalem","kjv_def":"Meah."},
"H3969":{"lemma":"מְאָה","xlit":"mᵉʼâh","pron":"meh-aw'","derivation":"(Aramaic) corresponding to H3967 (מֵאָה); properly, a primitive numeral","strongs_def":"a hundred; also as a multiplicative and a fraction}","kjv_def":"hundred."},
"H3970":{"lemma":"מַאֲוַי","xlit":"maʼăvay","pron":"mah-av-ah'ee","derivation":"from H183 (אָוָה);","strongs_def":"a desire","kjv_def":"desire."},
"H3971":{"lemma":"מאוּם","xlit":"mʼûwm","pron":"moom","derivation":"usually מוּם; as if passive participle from an unused root probably meaning","strongs_def":"to stain; a blemish (physically or morally)","kjv_def":"blemish, blot, spot."},
"H3972":{"lemma":"מְאוּמָה","xlit":"mᵉʼûwmâh","pron":"meh-oo'-maw","derivation":"apparently a form of H3971 (מאוּם);","strongs_def":"properly, a speck or point, i.e. (by implication) something; with negative, nothing","kjv_def":"fault, [phrase] no(-ught), ought, somewhat, any (no-)thing."},
"H3973":{"lemma":"מָאוֹס","xlit":"mâʼôwç","pron":"maw-oce'","derivation":"from H3988 (מָאַס);","strongs_def":"refuse","kjv_def":"refuse."},
"H3974":{"lemma":"מָאוֹר","xlit":"mâʼôwr","pron":"maw-ore'","derivation":"or מָאֹר; also (in plural) feminine מְאוֹרָה; or מְאֹרָה; from H215 (אוֹר);","strongs_def":"properly, a luminous body or luminary, i.e. (abstractly) light (as an element); figuratively, brightness, i.e.cheerfulness; specifically, a chandelier","kjv_def":"bright, light."},
"H3975":{"lemma":"מְאוּרָה","xlit":"mᵉʼûwrâh","pron":"meh-oo-raw'","derivation":"feminine passive participle of H215 (אוֹר);","strongs_def":"something lighted, i.e. an aperture; by implication, a crevice or hole (of a serpent)","kjv_def":"den."},
"H3976":{"lemma":"מֹאזֵן","xlit":"môʼzên","pron":"mo-zane'","derivation":"from H239 (אָזַן);","strongs_def":"(only in the dual) a pair of scales","kjv_def":"balances."},
"H3977":{"lemma":"מֹאזֵן","xlit":"môʼzên","pron":"mo-zane'","derivation":"(Aramaic) corresponding to H3976 (מֹאזֵן)","strongs_def":"{(only in the dual) a pair of scales}","kjv_def":"balances."},
"H3978":{"lemma":"מַאֲכָל","xlit":"maʼăkâl","pron":"mah-ak-awl'","derivation":"from H398 (אָכַל);","strongs_def":"an eatable (includ. provender, flesh and fruit)","kjv_def":"food, fruit, (bake-)meat(-s), victual."},
"H3979":{"lemma":"מַאֲכֶלֶת","xlit":"maʼăkeleth","pron":"mah-ak-eh'-leth","derivation":"from H398 (אָכַל);","strongs_def":"something to eat with,i.e. a knife","kjv_def":"knife."},
"H3980":{"lemma":"מַאֲכֹלֶת","xlit":"maʼăkôleth","pron":"mah-ak-o'-leth","derivation":"from H398 (אָכַל);","strongs_def":"something eaten (by fire), i.e. fuel","kjv_def":"fuel."},
"H3981":{"lemma":"מַאֲמָץ","xlit":"maʼămâts","pron":"mah-am-awts'","derivation":"from H553 (אָמַץ);","strongs_def":"strength, i.e. (plural) resources","kjv_def":"force."},
"H3982":{"lemma":"מַאֲמַר","xlit":"maʼămar","pron":"mah-am-ar'","derivation":"from H559 (אָמַר);","strongs_def":"something (authoritatively) said, i.e. an edict","kjv_def":"commandment, decree."},
"H3983":{"lemma":"מֵאמַר","xlit":"mêʼmar","pron":"may-mar'","derivation":"(Aramaic) corresponding to H3982 (מַאֲמַר)","strongs_def":"{something (authoritatively) said, i.e. an edict}","kjv_def":"appointment, word."},
"H3984":{"lemma":"מָאן","xlit":"mâʼn","pron":"mawn","derivation":"(Aramaic) probably from a root corresponding to H579 (אָנָה) in the sense of an inclosure by sides;","strongs_def":"a utensil","kjv_def":"vessel."},
"H3985":{"lemma":"מָאֵן","xlit":"mâʼên","pron":"maw-ane'","derivation":"a primitive root;","strongs_def":"to refuse","kjv_def":"refuse, [idiom] utterly."},
"H3986":{"lemma":"מָאֵן","xlit":"mâʼên","pron":"maw-ane'","derivation":"from H3985 (מָאֵן);","strongs_def":"unwilling","kjv_def":"refuse."},
"H3987":{"lemma":"מֵאֵן","xlit":"mêʼên","pron":"may-ane'","derivation":"from H3985 (מָאֵן);","strongs_def":"refractory","kjv_def":"refuse."},
"H3988":{"lemma":"מָאַס","xlit":"mâʼaç","pron":"maw-as'","derivation":"a primitive root;","strongs_def":"to spurn; also (intransitively) to disappear","kjv_def":"abhor, cast away (off), contemn, despise, disdain, (become) loathe(some), melt away, refuse, reject, reprobate, [idiom] utterly, vile person."},
"H3989":{"lemma":"מַאֲפֶה","xlit":"maʼăpheh","pron":"mah-af-eh'","derivation":"from H644 (אָפָה);","strongs_def":"something baked, i.e. a batch","kjv_def":"baken."},
"H3990":{"lemma":"מַאֲפֵל","xlit":"maʼăphêl","pron":"mah-af-ale'","derivation":"from the same as H651 (אָפֵל);","strongs_def":"something opaque","kjv_def":"darkness."},
"H3991":{"lemma":"מַאְפֵלְיָה","xlit":"maʼphêlᵉyâh","pron":"mah-af-ay-leh-yaw'","derivation":"prolonged feminine of H3990 (מַאֲפֵל);","strongs_def":"opaqueness","kjv_def":"darkness."},
"H3992":{"lemma":"מָאַר","xlit":"mâʼar","pron":"maw-ar'","derivation":"a primitive root;","strongs_def":"to be bitter or (causatively) to embitter, i.e. be painful","kjv_def":"fretting, picking."},
"H3993":{"lemma":"מַאֲרָב","xlit":"maʼărâb","pron":"mah-ar-awb'","derivation":"from H693 (אָרַב);","strongs_def":"an ambuscade","kjv_def":"lie in ambush, ambushment, lurking place, lying in wait."},
"H3994":{"lemma":"מְאֵרָה","xlit":"mᵉʼêrâh","pron":"meh-ay-raw'","derivation":"from H779 (אָרַר);","strongs_def":"an execration","kjv_def":"curse."},
"H3995":{"lemma":"מִבְדָּלָה","xlit":"mibdâlâh","pron":"mib-daw-law'","derivation":"from H914 (בָּדַל);","strongs_def":"a separation, i.e. (concretely) a separate place","kjv_def":"separate."},
"H3996":{"lemma":"מָבוֹא","xlit":"mâbôwʼ","pron":"maw-bo'","derivation":"from H935 (בּוֹא); (with or without H8121 (שֶׁמֶשׁ))","strongs_def":"an entrance (the place or the act); specifically sunset or the west; also (adverb with preposition) towards","kjv_def":"by which came, as cometh, in coming, as men enter into, entering, entrance into, entry, where goeth, going down, [phrase] westward. Compare H4126 (מוֹבָא)."},
"H3997":{"lemma":"מְבוֹאָה","xlit":"mᵉbôwʼâh","pron":"meb-o-aw'","derivation":"feminine of H3996 (מָבוֹא);","strongs_def":"a haven","kjv_def":"entry."},
"H3998":{"lemma":"מְבוּכָה","xlit":"mᵉbûwkâh","pron":"meb-oo-kaw'","derivation":"from H943 (בּוּךְ);","strongs_def":"perplexity","kjv_def":"perplexity."},
"H3999":{"lemma":"מַבּוּל","xlit":"mabbûwl","pron":"mab-bool'","derivation":"from H2986 (יָבַל) in the sense of flowing;","strongs_def":"a deluge","kjv_def":"flood."},
"H4000":{"lemma":"מָבוֹן","xlit":"mâbôwn","pron":"maw-bone'","derivation":"from H995 (בִּין);","strongs_def":"instructing","kjv_def":"taught."},
"H4001":{"lemma":"מְבוּסָה","xlit":"mᵉbûwçâh","pron":"meb-oo-saw'","derivation":"from H947 (בּוּס);","strongs_def":"a trampling","kjv_def":"treading (trodden) down (under foot)."},
"H4002":{"lemma":"מַבּוּעַ","xlit":"mabbûwaʻ","pron":"mab-boo'-ah","derivation":"from H5042 (נָבַע);","strongs_def":"a fountain","kjv_def":"fountain, spring."},
"H4003":{"lemma":"מְבוּקָה","xlit":"mᵉbûwqâh","pron":"meb-oo-kah'","derivation":"from the same as H950 (בּוּקָה);","strongs_def":"emptiness","kjv_def":"void."},
"H4004":{"lemma":"מִבְחוֹר","xlit":"mibchôwr","pron":"mib-khore'","derivation":"from H977 (בָּחַר);","strongs_def":"select, i.e. well fortified","kjv_def":"choice."},
"H4005":{"lemma":"מִבְחָר","xlit":"mibchâr","pron":"mib-khawr'","derivation":"from H977 (בָּחַר);","strongs_def":"select, i.e. best","kjv_def":"choice(-st), chosen."},
"H4006":{"lemma":"מִבְחָר","xlit":"Mibchâr","pron":"mib-khawr'","derivation":"the same as H4005 (מִבְחָר);","strongs_def":"Mibchar, an Israelite","kjv_def":"Mibhar."},
"H4007":{"lemma":"מַבָּט","xlit":"mabbâṭ","pron":"mab-bawt'","derivation":"or מֶבָּט; from H5027 (נָבַט);","strongs_def":"something expected, i.e. (abstractly) expectation","kjv_def":"expectation."},
"H4008":{"lemma":"מִבְטָא","xlit":"mibṭâʼ","pron":"mib-taw'","derivation":"from H981 (בָּטָא);","strongs_def":"a rash utterance (hasty vow)","kjv_def":"(that which...) uttered (out of)."},
"H4009":{"lemma":"מִבְטָח","xlit":"mibṭâch","pron":"mib-tawkh'","derivation":"from H982 (בָּטַח);","strongs_def":"properly, a refuge, i.e. (objective) security, or (subjective) assurance","kjv_def":"confidence, hope, sure, trust."},
"H4010":{"lemma":"מַבְלִיגִית","xlit":"mablîygîyth","pron":"mab-leeg-eeth'","derivation":"from H1082 (בָּלַג);","strongs_def":"desistance (or rather desolation)","kjv_def":"comfort self."},
"H4011":{"lemma":"מִבְנֶה","xlit":"mibneh","pron":"mib-neh'","derivation":"from H1129 (בָּנָה);","strongs_def":"a building","kjv_def":"frame."},
"H4012":{"lemma":"מְבֻנַּי","xlit":"Mᵉbunnay","pron":"meb-oon-hah'-ee","derivation":"from H1129 (בָּנָה); built up;","strongs_def":"Mebunnai, an Israelite","kjv_def":"Mebunnai."},
"H4013":{"lemma":"מִבְצָר","xlit":"mibtsâr","pron":"mib-tsawr'","derivation":"also (in plural) feminine (Daniel 11:l5) מִבְצָרָה; from H1219 (בָּצַר);","strongs_def":"a fortification, castle, or fortified city; figuratively, a defender","kjv_def":"(de-, most) fenced, fortress, (most) strong (hold)."},
"H4014":{"lemma":"מִבְצָר","xlit":"Mibtsâr","pron":"mib-tsawr'","derivation":"the same as H4013 (מִבְצָר);","strongs_def":"Mibtsar, an Idumaean","kjv_def":"Mibzar."},
"H4015":{"lemma":"מִבְרָח","xlit":"mibrâch","pron":"mib-rawkh'","derivation":"from H1272 (בָּרַח);","strongs_def":"a refugee","kjv_def":"fugitive."},
"H4016":{"lemma":"מָבֻשׁ","xlit":"mâbush","pron":"maw-boosh'","derivation":"from H954 (בּוּשׁ);","strongs_def":"(plural) the (male) pudenda","kjv_def":"secrets."},
"H4017":{"lemma":"מִבְשָׂם","xlit":"Mibsâm","pron":"mib-sawm'","derivation":"from the same as H1314 (בֶּשֶׂם); fragrant;","strongs_def":"Mibsam, the name of an Ishmaelite and of an Israelite","kjv_def":"Mibsam."},
"H4018":{"lemma":"מְבַשְּׁלָה","xlit":"mᵉbashshᵉlâh","pron":"meb-ash-shel-aw'","derivation":"from H1310 (בָּשַׁל);","strongs_def":"a cooking hearth","kjv_def":"boiling-place."},
"H4019":{"lemma":"מַגְבִּישׁ","xlit":"Magbîysh","pron":"mag-beesh'","derivation":"from the same as H1378 (גָּבִישׁ); stiffening;","strongs_def":"Magbish, an Israelite, or a place in Palestine","kjv_def":"Magbish."},
"H4020":{"lemma":"מִגְבָּלָה","xlit":"migbâlâh","pron":"mig-baw-law'","derivation":"from H1379 (גָּבַל);","strongs_def":"a border","kjv_def":"end."},
"H4021":{"lemma":"מִגְבָּעָה","xlit":"migbâʻâh","pron":"mig-baw-aw'","derivation":"from the same as H1389 (גִּבְעָה);","strongs_def":"a cap (as hemispherical)","kjv_def":"bonnet."},
"H4022":{"lemma":"מֶגֶד","xlit":"meged","pron":"meh'-ghed","derivation":"from an unused root probably meaning to be eminent;","strongs_def":"properly, a distinguished thing; hence something valuable, as aproduct or fruit","kjv_def":"pleasant, precious fruit (thing)."},
"H4023":{"lemma":"מְגִדּוֹן","xlit":"Mᵉgiddôwn","pron":"meg-id-done'","derivation":"(Zechariah H12 (אַבְדָן)) or מְגִדּוֹ; from H1413 (גָּדַד); rendezvous;","strongs_def":"Megiddon or Megiddo, a place in Palestine","kjv_def":"Megiddo, Megiddon."},
"H4024":{"lemma":"מִגְדּוֹל","xlit":"Migdôwl","pron":"mig-dole'","derivation":"or מִגְדֹּל; probably of Egyptian origin;","strongs_def":"Migdol, a place in Egypt","kjv_def":"Migdol, tower."},
"H4025":{"lemma":"מַגְדִּיאֵל","xlit":"Magdîyʼêl","pron":"mag-dee-ale'","derivation":"from H4022 (מֶגֶד) and H410 (אֵל); preciousness of God;","strongs_def":"Magdiel, an Idumaean","kjv_def":"Magdiel."},
"H4026":{"lemma":"מִגְדָּל","xlit":"migdâl","pron":"mig-dawl'","derivation":"also (in plural) feminine מִגְדָּלָה; from H1431 (גָּדַל);","strongs_def":"a tower (from its size or height); by analogy, a rostrum; figuratively, a (pyramidal) bed of flowers","kjv_def":"castle, flower, tower. Compare the names following."},
"H4027":{"lemma":"מִגְדַּל־אֵל","xlit":"Migdal-ʼÊl","pron":"mig-dal-ale'","derivation":"from H4026 (מִגְדָּל) and H410 (אֵל); tower of God;","strongs_def":"Migdal-El, a place in Palestine","kjv_def":"Migdal-el."},
"H4028":{"lemma":"מִגְדַּל־גָּד","xlit":"Migdal-Gâd","pron":"migdal-gawd'","derivation":"from H4026 (מִגְדָּל) and H1408 (גַּד); tower of Fortune;","strongs_def":"Migdal-Gad, a place in Palestine","kjv_def":"Migdal-gad."},
"H4029":{"lemma":"מִגְדַּל־עֵדֶר","xlit":"Migdal-ʻÊder","pron":"mig-dal'-ay'-der","derivation":"from H4026 (מִגְדָּל) and H5739 (עֵדֶר); tower of a flock;","strongs_def":"Migdal-Eder, a place in Palestine","kjv_def":"Migdal-eder, tower of the flock."},
"H4030":{"lemma":"מִגְדָּנָה","xlit":"migdânâh","pron":"mig-daw-naw'","derivation":"from the same as H4022 (מֶגֶד);","strongs_def":"preciousness, i.e. a gem","kjv_def":"precious thing, present."},
"H4031":{"lemma":"מָגוֹג","xlit":"Mâgôwg","pron":"maw-gogue'","derivation":"from H1463 (גּוֹג);","strongs_def":"Magog, a son of Japheth; also a barbarous northern region","kjv_def":"Magog."},
"H4032":{"lemma":"מָגוֹר","xlit":"mâgôwr","pron":"maw-gore'","derivation":"or (Lamentations 2:22) מָגוּר; from H1481 (גּוּר) in the sense of fearing;","strongs_def":"a fright (objective or subjective)","kjv_def":"fear, terror. Compare H4036 (מָגוֹר מִסָּבִיב)."},
"H4033":{"lemma":"מָגוּר","xlit":"mâgûwr","pron":"maw-goor'","derivation":"or מָגֻר; from H1481 (גּוּר) in the sense of lodging;","strongs_def":"a temporary abode; by extension, a permanent residence","kjv_def":"dwelling, pilgrimage, where sojourn, be a stranger. Compare H4032 (מָגוֹר)."},
"H4034":{"lemma":"מְגוֹרַה","xlit":"mᵉgôwrah","pron":"meg-o-raw'","derivation":"feminine of H4032 (מָגוֹר);","strongs_def":"affright","kjv_def":"fear."},
"H4035":{"lemma":"מְגוּרָה","xlit":"mᵉgûwrâh","pron":"meg-oo-raw'","derivation":"feminine of H4032 (מָגוֹר) or of H4033 (מָגוּר);","strongs_def":"a fright; also a granary","kjv_def":"barn, fear."},
"H4036":{"lemma":"מָגוֹר מִסָּבִיב","xlit":"Mâgôwr miç-Çâbîyb","pron":"maw-gore' mis-saw-beeb'","derivation":"from H4032 (מָגוֹר) and H5439 (סָבִיב) with the preposition inserted; affright from around;","strongs_def":"Magor-mis-Sabib, a symbolic name of Pashur","kjv_def":"Magormissabib."},
"H4037":{"lemma":"מַגְזֵרָה","xlit":"magzêrâh","pron":"mag-zay-raw'","derivation":"from H1504 (גָּזַר);","strongs_def":"a cutting implement, i.e. a blade","kjv_def":"axe."},
"H4038":{"lemma":"מַגָּל","xlit":"maggâl","pron":"mag-gawl'","derivation":"from an unused root meaning to reap;","strongs_def":"a sickle","kjv_def":"sickle."},
"H4039":{"lemma":"מְגִלָּה","xlit":"mᵉgillâh","pron":"meg-il-law'","derivation":"from H1556 (גָּלַל);","strongs_def":"a roll","kjv_def":"roll, volume."},
"H4040":{"lemma":"מְגִלָּה","xlit":"mᵉgillâh","pron":"meg-il-law'","derivation":"(Aramaic) corresponding to H4039 (מְגִלָּה)","strongs_def":"{a roll}","kjv_def":"roll."},
"H4041":{"lemma":"מְגַמָּה","xlit":"mᵉgammâh","pron":"meg-am-maw'","derivation":"from the same as H1571 (גַּם);","strongs_def":"properly, accumulation, i.e. impulse or direction","kjv_def":"sup up."},
"H4042":{"lemma":"מָגַן","xlit":"mâgan","pron":"maw-gan'","derivation":"a denominative from H4043 (מָגֵן);","strongs_def":"properly, to shield; encompass with; figuratively, to rescue, to hand safely over (i.e. surrender)","kjv_def":"deliver."},
"H4043":{"lemma":"מָגֵן","xlit":"mâgên","pron":"maw-gane'","derivation":"also (in plural) feminine מְגִנָּה; from H1598 (גָּנַן);","strongs_def":"a shield (i.e. the small one or buckler); figuratively, a protector; also the scaly hide of the crocodile","kjv_def":"[idiom] armed, buckler, defence, ruler, [phrase] scale, shield."},
"H4044":{"lemma":"מְגִנָּה","xlit":"mᵉginnâh","pron":"meg-in-naw'","derivation":"from H4042 (מָגַן);","strongs_def":"a covering (in a bad sense), i.e. blindness or obduracy","kjv_def":"sorrow. See also H4043 (מָגֵן)."},
"H4045":{"lemma":"מִגְעֶרֶת","xlit":"migʻereth","pron":"mig-eh'-reth","derivation":"from H1605 (גָּעַר);","strongs_def":"reproof (i.e. curse)","kjv_def":"rebuke."},
"H4046":{"lemma":"מַגֵּפָה","xlit":"maggêphâh","pron":"mag-gay-faw'","derivation":"from H5062 (נָגַף);","strongs_def":"a pestilence; by analogy, defeat","kjv_def":"([idiom] be) plague(-d), slaughter, stroke."},
"H4047":{"lemma":"מַגְפִּיעָשׁ","xlit":"Magpîyʻâsh","pron":"mag-pee-awsh'","derivation":"apparently from H1479 (גּוּף) or H5062 (נָגַף) and H6211 (עָשׁ); exterminator of (the) moth;","strongs_def":"Magpiash, an Israelite","kjv_def":"Magpiash."},
"H4048":{"lemma":"מָגַר","xlit":"mâgar","pron":"maw-gar'","derivation":"a primitive root;","strongs_def":"to yield up; intensively, to precipitate","kjv_def":"cast down, terror."},
"H4049":{"lemma":"מְגַר","xlit":"mᵉgar","pron":"meg-ar'","derivation":"(Aramaic) corresponding to H4048 (מָגַר);","strongs_def":"to overthrow","kjv_def":"destroy."},
"H4050":{"lemma":"מְגֵרָה","xlit":"mᵉgêrâh","pron":"meg-ay-raw'","derivation":"from H1641 (גָּרַר);","strongs_def":"a saw","kjv_def":"axe, saw."},
"H4051":{"lemma":"מִגְרוֹן","xlit":"Migrôwn","pron":"mig-rone'","derivation":"from H4048 (מָגַר); precipice;","strongs_def":"Migron, a place in Palestine","kjv_def":"Migron."},
"H4052":{"lemma":"מִגְרָעָה","xlit":"migrâʻâh","pron":"mig-raw-aw'","derivation":"from H1639 (גָּרַע);","strongs_def":"a ledge or offset","kjv_def":"narrowed rest."},
"H4053":{"lemma":"מִגְרָפָה","xlit":"migrâphâh","pron":"mig-raw-faw'","derivation":"from H1640 (גָּרַף);","strongs_def":"something thrown off (by the spade), i.e. a clod","kjv_def":"clod."},
"H4054":{"lemma":"מִגְרָשׁ","xlit":"migrâsh","pron":"mig-rawsh'","derivation":"also (in plural) feminine (Ezekiel 27:28) מִגְרָשָׁה; from H1644 (גָּרַשׁ);","strongs_def":"a suburb (i.e. open country whither flocks are driven from pasture); hence, the area around abuilding, or the margin of the sea","kjv_def":"cast out, suburb."},
"H4055":{"lemma":"מַד","xlit":"mad","pron":"mad","derivation":"or מֵד; from H4058 (מָדַד);","strongs_def":"properly, extent, i.e. height; also a measure; by implication, a vesture (as measured); also a carpet","kjv_def":"armour, clothes, garment, judgment, measure, raiment, stature."},
"H4056":{"lemma":"מַדְבַּח","xlit":"madbach","pron":"mad-bakh'","derivation":"(Aramaic) from H1684 (דְּבַח);","strongs_def":"a sacrificial altar","kjv_def":"altar."},
"H4057":{"lemma":"מִדְבָּר","xlit":"midbâr","pron":"mid-bawr'","derivation":"from H1696 (דָבַר) in the sense of driving;","strongs_def":"a pasture (i.e. open field, whither cattle are driven); by implication, a desert; also speech (including its organs)","kjv_def":"desert, south, speech, wilderness."},
"H4058":{"lemma":"מָדַד","xlit":"mâdad","pron":"maw-dad'","derivation":"a primitive root;","strongs_def":"properly, to stretch; by implication, to measure (as if by stretching a line); figuratively, to be extended","kjv_def":"measure, mete, stretch self."},
"H4059":{"lemma":"מִדַּד","xlit":"middad","pron":"mid-dad'","derivation":"from H5074 (נָדַד);","strongs_def":"flight","kjv_def":"be gone."},
"H4060":{"lemma":"מִדָּה","xlit":"middâh","pron":"mid-daw'","derivation":"feminine of H4055 (מַד);","strongs_def":"properly, extension, i.e. height or breadth; also a measure (including its standard); hence a portion (as measured) or a vestment; specifically, tribute (as measured)","kjv_def":"garment, measure(-ing, meteyard, piece, size, (great) stature, tribute, wide."},
"H4061":{"lemma":"מִדָּה","xlit":"middâh","pron":"mid-daw'","derivation":"(Aramaic) or מִנְדָּה; (Aramaic), corresponding to H4060 (מִדָּה);","strongs_def":"tribute in money","kjv_def":"toll, tribute."},
"H4062":{"lemma":"מַדְהֵבָה","xlit":"madhêbâh","pron":"mad-hay-baw'","derivation":"perhaps from the equivalent of H1722 (דְּהַב);","strongs_def":"goldmaking, i.e. exactness","kjv_def":"golden city."},
"H4063":{"lemma":"מֶדֶו","xlit":"medev","pron":"meh'-dev","derivation":"from an unused root meaning to stretch;","strongs_def":"properly, extent, i.e. measure; by implication, a dress (as measured)","kjv_def":"garment."},
"H4064":{"lemma":"מַדְוֶה","xlit":"madveh","pron":"mad-veh'","derivation":"from H1738 (דָּוָה);","strongs_def":"sickness","kjv_def":"disease."},
"H4065":{"lemma":"מַדּוּחַ","xlit":"maddûwach","pron":"mad-doo'akh","derivation":"from H5080 (נָדַח);","strongs_def":"seduction","kjv_def":"cause of banishment."},
"H4066":{"lemma":"מָדוֹן","xlit":"mâdôwn","pron":"maw-dohn'","derivation":"from H1777 (דִּין);","strongs_def":"a contest or quarrel","kjv_def":"brawling, contention(-ous), discord, strife. Compare H4079 (מִדְיָן), H4090 (מְדָן)."},
"H4067":{"lemma":"מָדוֹן","xlit":"mâdôwn","pron":"maw-dohn'","derivation":"from the same as H4063 (מֶדֶו);","strongs_def":"extensiveness, i.e. height","kjv_def":"stature."},
"H4068":{"lemma":"מָדוֹן","xlit":"Mâdôwn","pron":"maw-dohn'","derivation":"the same as H4067 (מָדוֹן);","strongs_def":"Madon, a place in Palestine","kjv_def":"Madon."},
"H4069":{"lemma":"מַדּוּעַ","xlit":"maddûwaʻ","pron":"mad-doo'-ah","derivation":"or מַדֻּעַ; from H4100 (מָה) and the passive participle of H3045 (יָדַע);","strongs_def":"what (is) known?; i.e. (by implication) (adverbially) why?","kjv_def":"how, wherefore, why."},
"H4070":{"lemma":"מְדוֹר","xlit":"mᵉdôwr","pron":"med-ore'","derivation":"(Aramaic) or מְדֹר; (Aramaic), or מְדָר; (Aramaic), from H1753 (דּוּר);","strongs_def":"a dwelling","kjv_def":"dwelling."},
"H4071":{"lemma":"מְדוּרָה","xlit":"mᵉdûwrâh","pron":"med-oo-raw'","derivation":"or מְדֻרָה; from H1752 (דּוּר) in the sense of accumulation;","strongs_def":"a pile of fuel","kjv_def":"pile (for fire)."},
"H4072":{"lemma":"מִדְחֶה","xlit":"midcheh","pron":"mid-kheh'","derivation":"from H1760 (דָּחָה);","strongs_def":"overthrow","kjv_def":"ruin."},
"H4073":{"lemma":"מְדַחְפָה","xlit":"mᵉdachphâh","pron":"med-akh-faw'","derivation":"from H1765 (דָּחַף);","strongs_def":"a push, i.e. ruin","kjv_def":"overthrow."},
"H4074":{"lemma":"מָדַי","xlit":"Mâday","pron":"maw-dah'-ee","derivation":"of foreign derivation;","strongs_def":"Madai, a country of central Asia","kjv_def":"Madai, Medes, Media."},
"H4075":{"lemma":"מָדַי","xlit":"Mâday","pron":"maw-dah'-ee","derivation":"patrial from H4074 (מָדַי);","strongs_def":"a Madian or native of Madai","kjv_def":"Mede."},
"H4076":{"lemma":"מָדַי","xlit":"Mâday","pron":"maw-dah'-ee","derivation":"(Aramaic) corresponding to H4074 (מָדַי)","strongs_def":"{Madai, a country of central Asia}","kjv_def":"Mede(-s)."},
"H4077":{"lemma":"מָדַי","xlit":"Mâday","pron":"maw-dah'-ee","derivation":"(Aramaic) corresponding to H4075 (מָדַי)","strongs_def":"{a Madian or native of Madai}","kjv_def":"Median."},
"H4078":{"lemma":"מַדַּי","xlit":"madday","pron":"mad-dah'-ee","derivation":"from H4100 (מָה) and H1767 (דַּי);","strongs_def":"what (is) enough, i.e. sufficiently","kjv_def":"sufficiently."},
"H4079":{"lemma":"מִדְיָן","xlit":"midyân","pron":"mid-yawn'","derivation":"a variation for H4066 (מָדוֹן)","strongs_def":"{a contest or quarrel}","kjv_def":"brawling, contention(-ous)."},
"H4080":{"lemma":"מִדְיָן","xlit":"Midyân","pron":"mid-yawn'","derivation":"the same as H4079 (מִדְיָן);","strongs_def":"Midjan, a son of Abraham; also his country and (collectively) his descendants","kjv_def":"Midian, Midianite."},
"H4081":{"lemma":"מִדִּין","xlit":"Middîyn","pron":"mid-deen'","derivation":"a variation for H4080 (מִדְיָן)","strongs_def":"{Midjan, a son of Abraham; also his country and (collectively) his descendants}","kjv_def":"Middin."},
"H4082":{"lemma":"מְדִינָה","xlit":"mᵉdîynâh","pron":"med-ee-naw'","derivation":"from H1777 (דִּין);","strongs_def":"properly, a judgeship, i.e. jurisdiction; by implication, a district (as ruled by a judge); generally, a region","kjv_def":"([idiom] every) province."},
"H4083":{"lemma":"מְדִינָה","xlit":"mᵉdîynâh","pron":"med-ee-naw'","derivation":"(Aramaic) corresponding to H4082 (מְדִינָה)","strongs_def":"{properly, a judgeship, i.e. jurisdiction; by implication, a district (as ruled by a judge); generally, a region}","kjv_def":"province."},
"H4084":{"lemma":"מִדְיָנִי","xlit":"Midyânîy","pron":"mid-yaw-nee'","derivation":"patronymical or patrial from H4080 (מִדְיָן);","strongs_def":"a Midjanite or descendant (native) of Midjan","kjv_def":"Midianite. Compare H4092 (מְדָנִי)."},
"H4085":{"lemma":"מְדֹכָה","xlit":"mᵉdôkâh","pron":"med-o-kaw'","derivation":"from H1743 (דּוּךְ);","strongs_def":"a mortar","kjv_def":"mortar."},
"H4086":{"lemma":"מַדְמֵן","xlit":"Madmên","pron":"mad-mane'","derivation":"from the same as H1828 (דֹּמֶן); dunghill;","strongs_def":"Madmen, a place in Palestine","kjv_def":"Madmen."},
"H4087":{"lemma":"מַדְמֵנָה","xlit":"madmênâh","pron":"mad-may-naw'","derivation":"feminine from the same as H1828 (דֹּמֶן);","strongs_def":"a dunghill","kjv_def":"dunghill."},
"H4088":{"lemma":"מַדְמֵנָה","xlit":"Madmênâh","pron":"mad-may-naw'","derivation":"the same as H4087 (מַדְמֵנָה);","strongs_def":"Madmenah, a place in Palestine","kjv_def":"Madmenah."},
"H4089":{"lemma":"מַדְמַנָּה","xlit":"Madmannâh","pron":"mad-man-naw'","derivation":"a variation for H4087 (מַדְמֵנָה);","strongs_def":"Madmannah, a place in Palestine","kjv_def":"Madmannah."},
"H4090":{"lemma":"מְדָן","xlit":"mᵉdân","pron":"med-awn'","derivation":"a form of H4066 (מָדוֹן)","strongs_def":"{a contest or quarrel}","kjv_def":"discord, strife."},
"H4091":{"lemma":"מְדָן","xlit":"Mᵉdân","pron":"med-awn'","derivation":"the same as H4090 (מְדָן);","strongs_def":"Medan, a son of Abraham","kjv_def":"Medan."},
"H4092":{"lemma":"מְדָנִי","xlit":"Mᵉdânîy","pron":"med-aw-nee'","derivation":"a variation of H4084 (מִדְיָנִי)","strongs_def":"{a Midjanite or descendant (native) of Midjan}","kjv_def":"Midianite."},
"H4093":{"lemma":"מַדָּע","xlit":"maddâʻ","pron":"mad-daw'","derivation":"or מַדַּע; from H3045 (יָדַע);","strongs_def":"intelligence or consciousness","kjv_def":"knowledge, science, thought."},
"H4094":{"lemma":"מַדְקָרָה","xlit":"madqârâh","pron":"mad-kaw-raw'","derivation":"from H1856 (דָּקַר);","strongs_def":"a wound","kjv_def":"piercing."},
"H4095":{"lemma":"מַדְרֵגָה","xlit":"madrêgâh","pron":"mad-ray-gaw'","derivation":"from an unused root meaning to step;","strongs_def":"properly, a step; by implication, a steep or inaccessible place","kjv_def":"stair, steep place."},
"H4096":{"lemma":"מִדְרָךְ","xlit":"midrâk","pron":"mid-rawk'","derivation":"from H1869 (דָּרַךְ);","strongs_def":"a treading, i.e. a place for stepping on","kjv_def":"(foot-) breadth."},
"H4097":{"lemma":"מִדְרָשׁ","xlit":"midrâsh","pron":"mid-rawsh'","derivation":"from H1875 (דָּרַשׁ);","strongs_def":"properly, an investigation, i.e. (by implication) a treatise or elaborate compilation","kjv_def":"story."},
"H4098":{"lemma":"מְדֻשָּׁה","xlit":"mᵉdushshâh","pron":"med-oosh-shaw'","derivation":"from H1758 (דּוּשׁ);","strongs_def":"a threshing, i.e. (concretely and figuratively) down-trodden people","kjv_def":"threshing."},
"H4099":{"lemma":"מְדָתָא","xlit":"Mᵉdâthâʼ","pron":"med-aw-thaw'","derivation":"of Persian origin;","strongs_def":"Medatha, the father of Haman","kjv_def":"Hammedatha (including the article)."},
"H4100":{"lemma":"מָה","xlit":"mâh","pron":"maw","derivation":"or מַה; or מָ; or מַ; also מֶה; a primitive particle;","strongs_def":"properly, interrogative what? (including how? why? when?); but also exclamation, what! (including how!), or indefinitely what (including whatever, and even relatively, that which); often used with prefixes in various adverbial or conjunctive senses","kjv_def":"how (long, oft, (-soever)), (no-) thing, what (end, good, purpose, thing), whereby(-fore, -in, -to, -with), (for) why."},
"H4101":{"lemma":"מָה","xlit":"mâh","pron":"maw","derivation":"(Aramaic) corresponding to H4100 (מָה)","strongs_def":"{properly, interrogative what? (including how? why? when?); but also exclamation, what! (including how!), or indefinitely what}","kjv_def":"how great (mighty), that which, what(-soever), why."},
"H4102":{"lemma":"מָהַהּ","xlit":"mâhahh","pron":"maw-hah'","derivation":"apparently a denominative from H4100 (מָה);","strongs_def":"properly, to question or hesitate, i.e. (by implication) to be reluctant","kjv_def":"delay, linger, stay selves, tarry."},
"H4103":{"lemma":"מְהוּמָה","xlit":"mᵉhûwmâh","pron":"meh-hoo-maw'","derivation":"from H1949 (הוּם);","strongs_def":"confusion or uproar","kjv_def":"destruction, discomfiture, trouble, tumult, vexation, vexed."},
"H4104":{"lemma":"מְהוּמָן","xlit":"Mᵉhûwmân","pron":"meh-hoo-mawn'","derivation":"of Persian origin;","strongs_def":"Mehuman, a eunuch of Xerxes","kjv_def":"Mehuman."},
"H4105":{"lemma":"מְהֵיטַבְאֵל","xlit":"Mᵉhêyṭabʼêl","pron":"meh-hay-tab-ale'","derivation":"from H3190 (יָטַב) (augmented) and H410 (אֵל); bettered of God;","strongs_def":"Mehetabel, the name of an Edomitish man and woman","kjv_def":"Mehetabeel, Mehetabel."},
"H4106":{"lemma":"מָהִיר","xlit":"mâhîyr","pron":"maw-here'","derivation":"or מָהִר; from H4116 (מָהַר);","strongs_def":"quick; hence, skilful","kjv_def":"diligent, hasty, ready."},
"H4107":{"lemma":"מָהַל","xlit":"mâhal","pron":"maw-hal'","derivation":"a primitive root;","strongs_def":"properly, to cut down or reduce, i.e. by implication, to adulterate","kjv_def":"mixed."},
"H4108":{"lemma":"מַהְלֵךְ","xlit":"mahlêk","pron":"mah-lake'","derivation":"from H1980 (הָלַךְ);","strongs_def":"a walking (plural collectively), i.e. access","kjv_def":"place to walk."},
"H4109":{"lemma":"מַהֲלָךְ","xlit":"mahălâk","pron":"mah-hal-awk'","derivation":"from H1980 (הָלַךְ);","strongs_def":"a walk, i.e. a passage or a distance","kjv_def":"journey, walk."},
"H4110":{"lemma":"מַהֲלָל","xlit":"mahălâl","pron":"mah-hal-awl'","derivation":"from H1984 (הָלַל);","strongs_def":"fame","kjv_def":"praise."},
"H4111":{"lemma":"מַהֲלַלְאֵל","xlit":"Mahălalʼêl","pron":"mah-hal-al-ale'","derivation":"from H4110 (מַהֲלָל) and H410 (אֵל); praise of God;","strongs_def":"Mahalalel, the name of an antediluvian patriarch and of an Israelite","kjv_def":"Mahalaleel."},
"H4112":{"lemma":"מַהֲלֻמָּה","xlit":"mahălummâh","pron":"mah-hal-oom-maw'","derivation":"from H1986 (הָלַם);","strongs_def":"a blow","kjv_def":"stripe, stroke."},
"H4113":{"lemma":"מַהֲמֹרָה","xlit":"mahămôrâh","pron":"mah-ham-o-raw'","derivation":"from an unused root of uncertain meaning;","strongs_def":"perhaps an abyss","kjv_def":"deep pit."},
"H4114":{"lemma":"מַהְפֵּכָה","xlit":"mahpêkâh","pron":"mah-pay-kaw'","derivation":"from H2015 (הָפַךְ);","strongs_def":"a destruction","kjv_def":"when...overthrew, overthrow(-n)."},
"H4115":{"lemma":"מַהְפֶּכֶת","xlit":"mahpeketh","pron":"mah-peh'-keth","derivation":"from H2015 (הָפַךְ);","strongs_def":"a wrench, i.e. the stocks","kjv_def":"prison, stocks."},
"H4116":{"lemma":"מָהַר","xlit":"mâhar","pron":"maw-har'","derivation":"a primitive root;","strongs_def":"properly, to be liquid or flow easily, i.e. (by implication); to hurry (in a good or a bad sense); often used (with another verb) adverbially, promptly","kjv_def":"be carried headlong, fearful, (cause to make, in, make) haste(-n, -ily), (be) hasty, (fetch, make ready) [idiom] quickly, rash, [idiom] shortly, (be so) [idiom] soon, make speed, [idiom] speedily, [idiom] straightway, [idiom] suddenly, swift."},
"H4117":{"lemma":"מָהַר","xlit":"mâhar","pron":"maw-har'","derivation":"a primitive root (perhaps rather the same as H4116 (מָהַר) through the idea of readiness in assent);","strongs_def":"to bargain (for a wife), i.e. to wed","kjv_def":"endow, [idiom] surely."},
"H4118":{"lemma":"מַהֵר","xlit":"mahêr","pron":"mah-hare'","derivation":"from H4116 (מָהַר);","strongs_def":"properly, hurrying; hence (adverbially) in ahurry","kjv_def":"hasteth, hastily, at once, quickly, soon, speedily, suddenly."},
"H4119":{"lemma":"מֹהַר","xlit":"môhar","pron":"mo'-har","derivation":"from H4117 (מָהַר);","strongs_def":"a price (for a wife)","kjv_def":"dowry."},
"H4120":{"lemma":"מְהֵרָה","xlit":"mᵉhêrâh","pron":"meh-hay-raw'","derivation":"feminine of H4118 (מַהֵר);","strongs_def":"properly, a hurry; hence (adverbially) promptly","kjv_def":"hastily, quickly, shortly, soon, make (with) speed(-ily), swiftly."},
"H4121":{"lemma":"מַהֲרַי","xlit":"Mahăray","pron":"mah-har-ah'-ee","derivation":"from H4116 (מָהַר); hasty;","strongs_def":"Maharai, an Israelite","kjv_def":"Maharai."},
"H4122":{"lemma":"מַהֵר שָׁלָל חָשׁ בַּז","xlit":"Mahêr Shâlâl Châsh Baz","pron":"mah-hare' shaw-lawl' khawsh baz","derivation":"from H4118 (מַהֵר) and H7998 (שָׁלָל) and H2363 (חוּשׁ) and H957 (בַּז); hasting (is he (the enemy) to the) booty, swift (to the) prey;","strongs_def":"Maher-Shalal-Chash-Baz; the symbolical name of the son of Isaiah","kjv_def":"Maher-sha-lal-bash-baz."},
"H4123":{"lemma":"מַהֲתַלָּה","xlit":"mahăthallâh","pron":"mah-hath-al-law'","derivation":"from H2048 (הָתַל);","strongs_def":"a delusion","kjv_def":"deceit."},
"H4124":{"lemma":"מוֹאָב","xlit":"Môwʼâb","pron":"mo-awb","derivation":"from a prolonged form of the prepositional prefix m-and H1 (אָב); from (her (the mother's)) father;","strongs_def":"Moab, an incestuous son of Lot; also his territory and descendants","kjv_def":"Moab."},
"H4125":{"lemma":"מוֹאָבִי","xlit":"Môwʼâbîy","pron":"mo-aw-bee'","derivation":"feminine מוֹאָבִיָּה; or מוֹאָבִיתxlit Môwâbîyth corrected to Môwʼâbîyth; patronymical from H4124 (מוֹאָב);","strongs_def":"a Moabite or Moabitess, i.e. a descendant from Moab","kjv_def":"(woman) of Moab, Moabite(-ish, -ss)."},
"H4126":{"lemma":"מוֹבָא","xlit":"môwbâʼ","pron":"mo-baw'","derivation":"by transp. for H3996 (מָבוֹא);","strongs_def":"an entrance","kjv_def":"coming."},
"H4127":{"lemma":"מוּג","xlit":"mûwg","pron":"moog","derivation":"a primitive root;","strongs_def":"to melt, i.e. literally (to soften, flow down, disappear), or figuratively (to fear, faint)","kjv_def":"consume, dissolve, (be) faint(-hearted), melt (away), make soft."},
"H4128":{"lemma":"מוּד","xlit":"mûwd","pron":"mood","derivation":"a primitive root;","strongs_def":"to shake","kjv_def":"measure."},
"H4129":{"lemma":"מוֹדַע","xlit":"môwdaʻ","pron":"mo-dah'","derivation":"or rather מֹדָע; from H3045 (יָדַע);","strongs_def":"an acquaintance","kjv_def":"kinswoman."},
"H4130":{"lemma":"מוֹדַעַת","xlit":"môwdaʻath","pron":"mo-dah'-ath","derivation":"from H3045 (יָדַע);","strongs_def":"acquaintance","kjv_def":"kindred."},
"H4131":{"lemma":"מוֹט","xlit":"môwṭ","pron":"mote","derivation":"a primitive root;","strongs_def":"to waver; by implication, to slip, shake, fall","kjv_def":"be carried, cast, be out of course, be fallen in decay, [idiom] exceedingly, fall(-ing down), be (re-) moved, be ready, shake, slide, slip."},
"H4132":{"lemma":"מוֹט","xlit":"môwṭ","pron":"mote","derivation":"from H4131 (מוֹט);","strongs_def":"a wavering, i.e. fall; by implication, a pole (as shaking); hence, a yoke (as essentially a bent pole)","kjv_def":"bar, be moved, staff, yoke."},
"H4133":{"lemma":"מוֹטָה","xlit":"môwṭâh","pron":"mo-taw'","derivation":"feminine of H4132 (מוֹט);","strongs_def":"a pole; by implication, an ox-bow; hence, a yoke (either literal or figurative)","kjv_def":"bands, heavy, staves, yoke."},
"H4134":{"lemma":"מוּךְ","xlit":"mûwk","pron":"mook","derivation":"a primitive root;","strongs_def":"to become thin, i.e. (figuratively) be impoverished","kjv_def":"be (waxen) poor(-er)."},
"H4135":{"lemma":"מוּל","xlit":"mûwl","pron":"mool","derivation":"a primitive root;","strongs_def":"to cut short, i.e. curtail (specifically the prepuce, i.e. to circumcise); by implication, to blunt; figuratively, to destroy","kjv_def":"circumcise(-ing), selves), cut down (in pieces), destroy, [idiom] must needs."},
"H4136":{"lemma":"מוּל","xlit":"mûwl","pron":"mool","derivation":"or מוֹל; (Deuteronomy 1:1), or מוֹאל; (Nehemiah 12:38), or מֻל; (Numbers 22:5), from H4135 (מוּל);","strongs_def":"properly, abrupt, i.e. a precipice; by implication, the front; used only adverbially (with prepositional prefix) opposite","kjv_def":"(over) against, before, (fore-) front, from, (God-) ward, toward, with."},
"H4137":{"lemma":"מוֹלָדָה","xlit":"Môwlâdâh","pron":"mo-law-daw'","derivation":"from H3205 (יָלַד); birth;","strongs_def":"Moladah, a place in Palestine","kjv_def":"Moladah."},
"H4138":{"lemma":"מוֹלֶדֶת","xlit":"môwledeth","pron":"mo-leh'-deth","derivation":"from H3205 (יָלַד);","strongs_def":"nativity (plural birth-place); by implication, lineage, native country; also offspring, family","kjv_def":"begotten, born, issue, kindred, native(-ity)."},
"H4139":{"lemma":"מוּלָה","xlit":"mûwlâh","pron":"moo-law'","derivation":"from H4135 (מוּל);","strongs_def":"circumcision","kjv_def":"circumcision."},
"H4140":{"lemma":"מוֹלִיד","xlit":"Môwlîyd","pron":"mo-leed'","derivation":"from H3205 (יָלַד); genitor;","strongs_def":"Molid, an Israelite","kjv_def":"Molid."},
"H4141":{"lemma":"מוּסָב","xlit":"mûwçâb","pron":"moo-sawb'","derivation":"from H5437 (סָבַב);","strongs_def":"a turn, i.e. circuit (of a building)","kjv_def":"winding about."},
"H4142":{"lemma":"מוּסַבָּה","xlit":"mûwçabbâh","pron":"moo-sab-baw'","derivation":"or מֻסַבָּה; feminine of H4141 (מוּסָב);","strongs_def":"a reversal, i.e. the backside (of a gem), fold (of a double-leaved door), transmutation (of a name)","kjv_def":"being changed, inclosed, be set, turning."},
"H4143":{"lemma":"מוּסָד","xlit":"mûwçâd","pron":"moo-sawd'","derivation":"from H3245 (יָסַד);","strongs_def":"a foundation","kjv_def":"foundation."},
"H4144":{"lemma":"מוֹסָד","xlit":"môwçâd","pron":"mo-sawd'","derivation":"from H3245 (יָסַד);","strongs_def":"a foundation","kjv_def":"foundation."},
"H4145":{"lemma":"מוּסָדָה","xlit":"mûwçâdâh","pron":"moo-saw-daw'","derivation":"feminine of H4143 (מוּסָד);","strongs_def":"a foundation; figuratively, an appointment","kjv_def":"foundation, grounded. Compare H4328 (מְיֻסָּדָה)."},
"H4146":{"lemma":"מוֹסָדָה","xlit":"môwçâdâh","pron":"mo-saw-daw'","derivation":"or מֹסָדָה; feminine of H4144 (מוֹסָד);","strongs_def":"a foundation","kjv_def":"foundation."},
"H4147":{"lemma":"מוֹסֵר","xlit":"môwçêr","pron":"mo-sare'","derivation":"also (in plural) feminine מוֹסֵרָה; or מֹסְרָה; from H3256 (יָסַר);","strongs_def":"properly, chastisement, i.e. (by implication) a halter; figuratively, restraint","kjv_def":"band, bond."},
"H4148":{"lemma":"מוּסָר","xlit":"mûwçâr","pron":"moo-sawr'","derivation":"from H3256 (יָסַר);","strongs_def":"properly, chastisement; figuratively, reproof, warning or instruction; also restraint","kjv_def":"bond, chastening(-eth), chastisement, check, correction, discipline, doctrine, instruction, rebuke."},
"H4149":{"lemma":"מוֹסֵרָה","xlit":"Môwçêrâh","pron":"mo-say-raw'","derivation":"or (plural) מֹסְרוֹת feminine of H4147 (מוֹסֵר); correction or corrections;","strongs_def":"Moserah or Moseroth, a place in the Desert","kjv_def":"Mosera, Moseroth."},
"H4150":{"lemma":"מוֹעֵד","xlit":"môwʻêd","pron":"mo-ade'","derivation":"or מֹעֵד; or (feminine) מוֹעָדָה; (2 Chronicles 8:13), from H3259 (יָעַד);","strongs_def":"properly, an appointment, i.e. a fixed time or season; specifically, a festival; conventionally ayear; by implication, an assembly (as convened for a definite purpose); technically the congregation; by extension, the place of meeting; also a signal (as appointed beforehand)","kjv_def":"appointed (sign, time), (place of, solemn) assembly, congregation, (set, solemn) feast, (appointed, due) season, solemn(-ity), synogogue, (set) time (appointed)."},
"H4151":{"lemma":"מוֹעָד","xlit":"môwʻâd","pron":"mo-awd'","derivation":"from H3259 (יָעַד); (as in H4150 (מוֹעֵד));","strongs_def":"properly, an assembly  figuratively, a troop","kjv_def":"appointed time."},
"H4152":{"lemma":"מוּעָדָה","xlit":"mûwʻâdâh","pron":"moo-aw-daw'","derivation":"from H3259 (יָעַד);","strongs_def":"an appointed place, i.e. asylum","kjv_def":"appointed."},
"H4153":{"lemma":"מוֹעַדְיָה","xlit":"Môwʻadyâh","pron":"mo-ad-yaw'","derivation":"from H4151 (מוֹעָד) and H3050 (יָהּ); assembly of Jah;","strongs_def":"Moadjah, an Israelite","kjv_def":"Moadiah. Compare H4573 (מַעֲדְיָה)."},
"H4154":{"lemma":"מוּעֶדֶת","xlit":"mûwʻedeth","pron":"moo-ay'-deth","derivation":"feminine passive participle of H4571 (מָעַד);","strongs_def":"properly, made to slip, i.e. dislocated","kjv_def":"out of joint."},
"H4155":{"lemma":"מוּעָף","xlit":"mûwʻâph","pron":"moo-awf'","derivation":"from H5774 (עוּף);","strongs_def":"properly, covered, i.e. dark; abstractly, obscurity, i.e. distress","kjv_def":"dimness."},
"H4156":{"lemma":"מוֹעֵצָה","xlit":"môwʻêtsâh","pron":"mo-ay-tsaw'","derivation":"from H3289 (יָעַץ);","strongs_def":"a purpose","kjv_def":"counsel, device."},
"H4157":{"lemma":"מוּעָקָה","xlit":"mûwʻâqâh","pron":"moo-aw-kaw'","derivation":"from H5781 (עוּק);","strongs_def":"pressure, i.e. (figuratively) distress","kjv_def":"affliction."},
"H4158":{"lemma":"מוֹפַעַת","xlit":"Môwphaʻath","pron":"mo-fah'-ath","derivation":"(Jeremiah 48:21) or מֵיפַעַת; or מֵפַעַת; from H3313 (יָפַע); illuminative;","strongs_def":"Mophaath or Mephaath, a place in Palestine","kjv_def":"Mephaath."},
"H4159":{"lemma":"מוֹפֵת","xlit":"môwphêth","pron":"mo-faith'","derivation":"or מֹפֵת; from H3302 (יָפָה) in the sense of conspicuousness;","strongs_def":"a miracle; by implication, a token or omen","kjv_def":"miracle, sign, wonder(-ed at)."},
"H4160":{"lemma":"מוּץ","xlit":"mûwts","pron":"moots","derivation":"a primitive root;","strongs_def":"to press, i.e. (figuratively) to oppress","kjv_def":"extortioner."},
"H4161":{"lemma":"מוֹצָא","xlit":"môwtsâʼ","pron":"mo-tsaw'","derivation":"or מֹצָאxlit môtsâ corrected to môtsâʼ; from H3318 (יָצָא);","strongs_def":"a going forth, i.e. (the act) an egress, or (the place) an exit; hence, a source or product; specifically, dawn, the rising of the sun (the East), exportation, utterance, a gate, a fountain, a mine, a meadow (as producing grass)","kjv_def":"brought out, bud, that which came out, east, going forth, goings out, that which (thing that) is gone out, outgoing, proceeded out, spring, vein, (water-) course (springs)."},
"H4162":{"lemma":"מוֹצָא","xlit":"môwtsâʼ","pron":"mo-tsaw'","derivation":"the same as H4161 (מוֹצָא);","strongs_def":"Motsa, the name of two Israelites","kjv_def":"Moza."},
"H4163":{"lemma":"מוֹצָאָה","xlit":"môwtsâʼâh","pron":"mo-tsaw-aw'","derivation":"feminine of H4161 (מוֹצָא); (marg.; compare H6675 (צוֹאָה))","strongs_def":"a family descent; also a sewer","kjv_def":"draught house; going forth."},
"H4164":{"lemma":"מוּצַק","xlit":"mûwtsaq","pron":"moo-tsak'","derivation":"or מוּצָק; from H3332 (יָצַק);","strongs_def":"narrowness; figuratively, distress","kjv_def":"anguish, is straitened, straitness."},
"H4165":{"lemma":"מוּצָק","xlit":"mûwtsâq","pron":"moo-tsawk'","derivation":"from H5694 (עָגִיל);","strongs_def":"properly, fusion, i.e. literally, a casting (of metal); figuratively, a mass (of clay)","kjv_def":"casting, hardness."},
"H4166":{"lemma":"מוּצָקָה","xlit":"mûwtsâqâh","pron":"moo-tsaw-kaw'","derivation":"or מֻצָקָה; from H3332 (יָצַק);","strongs_def":"properly, something poured out, i.e. a casting (of metal); by implication, a tube (as cast)","kjv_def":"when it was cast, pipe."},
"H4167":{"lemma":"מוּק","xlit":"mûwq","pron":"mook","derivation":"a primitive root;","strongs_def":"to jeer, i.e. (intens.) blaspheme","kjv_def":"be corrupt."},
"H4168":{"lemma":"מוֹקֵד","xlit":"môwqêd","pron":"mo-kade'","derivation":"from H3344 (יָקַד);","strongs_def":"a fire or fuel; abstractly, a conflagration","kjv_def":"burning, hearth."},
"H4169":{"lemma":"מוֹקְדָה","xlit":"môwqᵉdâh","pron":"mo-ked-aw'","derivation":"feminine of H4168 (מוֹקֵד);","strongs_def":"fuel","kjv_def":"burning."},
"H4170":{"lemma":"מוֹקֵשׁ","xlit":"môwqêsh","pron":"mo-kashe'","derivation":"or מֹקֵשׁ; from H3369 (יָקֹשׁ);","strongs_def":"a noose (for catching animals) (literally or figuratively); by implication, a hook (for the nose)","kjv_def":"be ensnared, gin, (is) snare(-d), trap."},
"H4171":{"lemma":"מוּר","xlit":"mûwr","pron":"moor","derivation":"a primitive root;","strongs_def":"to alter; by implication, to barter, to dispose of","kjv_def":"[idiom] at all, (ex-) change, remove."},
"H4172":{"lemma":"מוֹרָא","xlit":"môwrâʼ","pron":"mo-raw'","derivation":"or מֹרָא; or מוֹרָהxlit môrâh corrected to môwrâh; (Psalm 9:20), from H3372 (יָרֵא);","strongs_def":"fear; by implication, a fearful thing or deed","kjv_def":"dread, (that ought to be) fear(-ed), terribleness, terror."},
"H4173":{"lemma":"מוֹרַג","xlit":"môwrag","pron":"mo-rag'","derivation":"or מֹרַג; from an unused root meaning to triturate;","strongs_def":"a threshing sledge","kjv_def":"threshing instrument."},
"H4174":{"lemma":"מוֹרָד","xlit":"môwrâd","pron":"mo-rawd'","derivation":"from H3381 (יָרַד);","strongs_def":"a descent; as architecture, an ornamental appendage, perhaps a festoon","kjv_def":"going down, steep place, thin work."},
"H4175":{"lemma":"מוֹרֶה","xlit":"môwreh","pron":"mo-reh'","derivation":"from H3384 (יָרָה); (see H3138 (יוֹרֶה))","strongs_def":"an archer; also teacher or teaching; also the early rain","kjv_def":"(early) rain."},
"H4176":{"lemma":"מוֹרֶה","xlit":"Môwreh","pron":"mo-reh'","derivation":"or מֹרֶה; the same as H4175 (מוֹרֶה);","strongs_def":"Moreh, a Canaanite; also a hill (perhaps named from him)","kjv_def":"Moreh."},
"H4177":{"lemma":"מוֹרָה","xlit":"môwrâh","pron":"mo-raw'","derivation":"from H4171 (מוּר) in the sense of shearing;","strongs_def":"a razor","kjv_def":"razor."},
"H4178":{"lemma":"מוֹרָט","xlit":"môwrâṭ","pron":"mo-rawt'","derivation":"from H3399 (יָרַט);","strongs_def":"obstinate, i.e. independent","kjv_def":"peeled."},
"H4179":{"lemma":"מוֹרִיָּה","xlit":"Môwrîyâh","pron":"mo-ree-yaw'","derivation":"or מֹרִיָּה; from H7200 (רָאָה) and H3050 (יָהּ); seen of Jah;","strongs_def":"Morijah, a hill in Palestine","kjv_def":"Moriah."},
"H4180":{"lemma":"מוֹרָשׁ","xlit":"môwrâsh","pron":"mo-rawsh'","derivation":"from H3423 (יָרַשׁ);","strongs_def":"a possession; figuratively, delight","kjv_def":"possession, thought."},
"H4181":{"lemma":"מוֹרָשָׁה","xlit":"môwrâshâh","pron":"mo-raw-shaw'","derivation":"feminine of H4180 (מוֹרָשׁ);","strongs_def":"a possession","kjv_def":"heritage, inheritance, possession."},
"H4182":{"lemma":"מוֹרֶשֶׁת גַּת","xlit":"Môwresheth Gath","pron":"mo-reh'-sheth gath","derivation":"from H3423 (יָרַשׁ) and H1661 (גַּת); possession of Gath;","strongs_def":"Moresheth-Gath, a place in Palestine","kjv_def":"Moresheth-gath."},
"H4183":{"lemma":"מוֹרַשְׁתִּי","xlit":"Môwrashtîy","pron":"mo-rash-tee'","derivation":"patrial from H4182 (מוֹרֶשֶׁת גַּת);","strongs_def":"a Morashtite or inhabitant of Moresheth-Gath","kjv_def":"Morashthite."},
"H4184":{"lemma":"מוּשׁ","xlit":"mûwsh","pron":"moosh","derivation":"a primitive root;","strongs_def":"to touch","kjv_def":"feel, handle."},
"H4185":{"lemma":"מוּשׁ","xlit":"mûwsh","pron":"moosh","derivation":"a primitive root (perhaps rather the same as H4184 (מוּשׁ) through the idea of receding by contact);","strongs_def":"to withdraw (both literally and figuratively, whether intransitive or transitive)","kjv_def":"cease, depart, go back, remove, take away."},
"H4186":{"lemma":"מוֹשָׁב","xlit":"môwshâb","pron":"mo-shawb'","derivation":"or מֹשָׁב; from H3427 (יָשַׁב);","strongs_def":"a seat; figuratively, a site; abstractly, a session; by extension an abode (the place or the time); by implication, population","kjv_def":"assembly, dwell in, dwelling(-place), wherein (that) dwelt (in), inhabited place, seat, sitting, situation, sojourning."},
"H4187":{"lemma":"מוּשִׁי","xlit":"Mûwshîy","pron":"moo-shee'","derivation":"or מֻשִּׁי; from H4184 (מוּשׁ); sensitive;","strongs_def":"Mushi, a Levite","kjv_def":"Mushi."},
"H4188":{"lemma":"מוּשִׁי","xlit":"Mûwshîy","pron":"moo-shee'","derivation":"patronymical from H4187 (מוּשִׁי);","strongs_def":"a Mushite (collectively) or descendants of Mushi","kjv_def":"Mushites."},
"H4189":{"lemma":"מוֹשְׁכָה","xlit":"môwshᵉkâh","pron":"mo-shek-aw'","derivation":"act participle feminine of H4900 (מָשַׁךְ);","strongs_def":"something drawing, i.e. (figuratively) a cord","kjv_def":"band."},
"H4190":{"lemma":"מוֹשָׁעָה","xlit":"môwshâʻâh","pron":"mo-shaw-aw'","derivation":"from H3467 (יָשַׁע);","strongs_def":"deliverance","kjv_def":"salvation."},
"H4191":{"lemma":"מוּת","xlit":"mûwth","pron":"mooth","derivation":"a primitive root;","strongs_def":"to die (literally or figuratively); causatively, to kill","kjv_def":"[idiom] at all, [idiom] crying, (be) dead (body, man, one), (put to, worthy of) death, destroy(-er), (cause to, be like to, must) die, kill, necro(-mancer), [idiom] must needs, slay, [idiom] surely, [idiom] very suddenly, [idiom] in (no) wise."},
"H4192":{"lemma":"מוּת","xlit":"Mûwth","pron":"mooth","derivation":"(Psalm H48 (אֲבִירָם) or מוּת לַבֵּן; from H4191 (מוּת) and H1121 (בֵּן) with the preposition and article interposed;","strongs_def":"'To die for the son', probably the title of a popular song","kjv_def":"death, Muthlabben."},
"H4193":{"lemma":"מוֹת","xlit":"môwth","pron":"mohth","derivation":"(Aramaic) corresponding to H4194 (מָוֶת);","strongs_def":"death","kjv_def":"death."},
"H4194":{"lemma":"מָוֶת","xlit":"mâveth","pron":"maw'-veth","derivation":"from H4191 (מוּת);","strongs_def":"death (natural or violent); concretely, the dead, their place or state (hades); figuratively, pestilence, ruin","kjv_def":"(be) dead(-ly), death, die(-d)."},
"H4195":{"lemma":"מוֹתָר","xlit":"môwthâr","pron":"mo-thar'","derivation":"from H3498 (יָתַר);","strongs_def":"literally gain; figuratively, superiority","kjv_def":"plenteousness, preeminence, profit."},
"H4196":{"lemma":"מִזְבֵּחַ","xlit":"mizbêach","pron":"miz-bay'-akh","derivation":"from H2076 (זָבַח);","strongs_def":"an altar","kjv_def":"altar."},
"H4197":{"lemma":"מֶזֶג","xlit":"mezeg","pron":"meh'-zeg","derivation":"from an unused root meaning to mingle (water with wine);","strongs_def":"tempered wine","kjv_def":"liquor."},
"H4198":{"lemma":"מָזֶה","xlit":"mâzeh","pron":"maw-zeh'","derivation":"from an unused root meaning to suck out;","strongs_def":"exhausted","kjv_def":"burnt."},
"H4199":{"lemma":"מִזָּה","xlit":"Mizzâh","pron":"miz-zaw'","derivation":"probably from an unused root meaning to faint with fear; terror;","strongs_def":"Mizzah, an Edomite","kjv_def":"Mizzah."},
"H4200":{"lemma":"מֶזֶו","xlit":"mezev","pron":"meh'-zev","derivation":"probably from an unused root meaning to gather in;","strongs_def":"a granary","kjv_def":"garner."},
"H4201":{"lemma":"מְזוּזָה","xlit":"mᵉzûwzâh","pron":"mez-oo-zaw'","derivation":"or מְזֻזָה; from the same as H2123 (זִיז);","strongs_def":"a door-post (as prominent)","kjv_def":"(door, side) post."},
"H4202":{"lemma":"מָזוֹן","xlit":"mâzôwn","pron":"maw-zone'","derivation":"from H2109 (זוּן);","strongs_def":"food","kjv_def":"meat, victual."},
"H4203":{"lemma":"מָזוֹן","xlit":"mâzôwn","pron":"maw-zone'","derivation":"(Aramaic) corresponding to H4202 (מָזוֹן)","strongs_def":"{food}","kjv_def":"meat."},
"H4204":{"lemma":"מָזוֹר","xlit":"mâzôwr","pron":"maw-zore'","derivation":"from H2114 (זוּר) in the sense of turning aside from truth;","strongs_def":"treachery, i.e. a plot","kjv_def":"wound."},
"H4205":{"lemma":"מָזוֹר","xlit":"mâzôwr","pron":"maw-zore'","derivation":"or מָזֹר; from H2115 (זוּר) in the sense of binding up;","strongs_def":"a bandage, i.e. remedy; hence, a sore (as needing a compress)","kjv_def":"bound up, wound."},
"H4206":{"lemma":"מָזִיחַ","xlit":"mâzîyach","pron":"maw-zee'-akh","derivation":"or מֵזַח; from H2118 (זָחַח);","strongs_def":"a belt (as movable)","kjv_def":"girdle, strength."},
"H4207":{"lemma":"מַזְלֵג","xlit":"mazlêg","pron":"maz-layg'","derivation":"or (feminine) מִזְלָגָה; from an unused root meaning to draw up;","strongs_def":"a fork","kjv_def":"fleshhook."},
"H4208":{"lemma":"מַזָּלָה","xlit":"mazzâlâh","pron":"maz-zaw-law'","derivation":"apparently from H5140 (נָזַל) in the sense of raining;","strongs_def":"a constellation, i.e. Zodiacal sign (perhaps as affecting the weather)","kjv_def":"planet. Compare H4216 (מַזָּרָה)."},
"H4209":{"lemma":"מְזִמָּה","xlit":"mᵉzimmâh","pron":"mez-im-maw'","derivation":"from H2161 (זָמַם);","strongs_def":"a plan, usually evil (machination), sometimes good (sagacity)","kjv_def":"(wicked) device, discretion, intent, witty invention, lewdness, mischievous (device), thought, wickedly."},
"H4210":{"lemma":"מִזְמוֹר","xlit":"mizmôwr","pron":"miz-more'","derivation":"from H2167 (זָמַר);","strongs_def":"properly, instrumental music; by implication, a poem set to notes","kjv_def":"psalm."},
"H4211":{"lemma":"מַזְמֵרָה","xlit":"mazmêrâh","pron":"maz-may-raw'","derivation":"from H2168 (זָמַר);","strongs_def":"a pruningknife","kjv_def":"pruning-hook."},
"H4212":{"lemma":"מְזַמְּרָה","xlit":"mᵉzammᵉrâh","pron":"mez-am-mer-aw'","derivation":"from H2168 (זָמַר);","strongs_def":"a tweezer (only in the plural)","kjv_def":"snuffers."},
"H4213":{"lemma":"מִזְעָר","xlit":"mizʻâr","pron":"miz-awr'","derivation":"from the same as H2191 (זְעֵיר)xlit zeʻêyr corrected to zᵉʻêyr;","strongs_def":"fewness; by implication, as superl. diminutiveness","kjv_def":"few, [idiom] very."},
"H4214":{"lemma":"מִזְרֶה","xlit":"mizreh","pron":"miz-reh'","derivation":"from H2219 (זָרָה);","strongs_def":"a winnowing shovel (as scattering the chaff)","kjv_def":"fan."},
"H4215":{"lemma":"מְזָרֶה","xlit":"mᵉzâreh","pron":"mez-aw-reh'","derivation":"apparently from H2219 (זָרָה);","strongs_def":"properly, a scatterer, i.e. the north wind (as dispersing clouds; only in plural)","kjv_def":"north."},
"H4216":{"lemma":"מַזָּרָה","xlit":"mazzârâh","pron":"maz-zaw-raw'","derivation":"apparently from H5144 (נָזַר) in the sense of distinction;","strongs_def":"some noted constellation (only in the plural), perhaps collectively, the zodiac","kjv_def":"Mazzoroth. Compare H4208 (מַזָּלָה)."},
"H4217":{"lemma":"מִזְרָח","xlit":"mizrâch","pron":"miz-rawkh'","derivation":"from H2224 (זָרַח);","strongs_def":"sunrise, i.e. the east","kjv_def":"east (side, -ward), (sun-) rising (of the sun)."},
"H4218":{"lemma":"מִזְרָע","xlit":"mizrâʻ","pron":"miz-raw'","derivation":"from H2232 (זָרַע);","strongs_def":"a planted field","kjv_def":"thing sown."},
"H4219":{"lemma":"מִזְרָק","xlit":"mizrâq","pron":"miz-rawk'","derivation":"from H2236 (זָרַק);","strongs_def":"a bowl (as if for sprinkling)","kjv_def":"bason, bowl."},
"H4220":{"lemma":"מֵחַ","xlit":"mêach","pron":"may'-akh","derivation":"from H4229 (מָחָה) in the sense of greasing;","strongs_def":"fat; figuratively, rich","kjv_def":"fatling (one)."},
"H4221":{"lemma":"מֹחַ","xlit":"môach","pron":"mo'-akh","derivation":"from the same as H4220 (מֵחַ);","strongs_def":"fat, i.e. marrow","kjv_def":"marrow."},
"H4222":{"lemma":"מָחָא","xlit":"mâchâʼ","pron":"maw-khaw'","derivation":"a primitive root;","strongs_def":"to rub or strike the hands together (in exultation)","kjv_def":"clap."},
"H4223":{"lemma":"מְחָא","xlit":"mᵉchâʼ","pron":"mekh-aw'","derivation":"(Aramaic) corresponding to H4222 (מָחָא);","strongs_def":"to strike in pieces; also to arrest; specifically to impale","kjv_def":"hang, smite, stay."},
"H4224":{"lemma":"מַחֲבֵא","xlit":"machăbêʼ","pron":"makh-ab-ay'","derivation":"or מַחֲבֹא; from H2244 (חָבָא);","strongs_def":"a refuge","kjv_def":"hiding (lurking) place."},
"H4225":{"lemma":"מַחְבֶּרֶת","xlit":"machbereth","pron":"makh-beh'-reth","derivation":"from H2266 (חָבַר);","strongs_def":"a junction, i.e. seam or sewed piece","kjv_def":"coupling."},
"H4226":{"lemma":"מְחַבְּרָה","xlit":"mᵉchabbᵉrâh","pron":"mekh-ab-ber-aw'","derivation":"from H2266 (חָבַר);","strongs_def":"a joiner, i.e. brace or cramp","kjv_def":"coupling, joining."},
"H4227":{"lemma":"מַחֲבַת","xlit":"machăbath","pron":"makh-ab-ath'","derivation":"from the same as H2281 (חָבֵת);","strongs_def":"a pan for baking in","kjv_def":"pan."},
"H4228":{"lemma":"מַחֲגֹרֶת","xlit":"machăgôreth","pron":"makh-ag-o'-reth","derivation":"from H2296 (חָגַר);","strongs_def":"a girdle","kjv_def":"girding."},
"H4229":{"lemma":"מָחָה","xlit":"mâchâh","pron":"maw-khaw'","derivation":"a primitive root;","strongs_def":"properly, to stroke or rub; by implication, to erase; also to smooth (as if with oil), i.e. grease or make fat; also to touch, i.e. reach to","kjv_def":"abolish, blot out, destroy, full of marrow, put out, reach unto, [idiom] utterly, wipe (away, out)."},
"H4230":{"lemma":"מְחוּגָה","xlit":"mᵉchûwgâh","pron":"mekk-oo-gaw'","derivation":"from H2328 (חוּג);","strongs_def":"an instrument for marking a circle, i.e. compasses","kjv_def":"compass."},
"H4231":{"lemma":"מָחוֹז","xlit":"mâchôwz","pron":"maw-khoze'","derivation":"from an unused root meaning to enclose;","strongs_def":"a harbor (as shut in by the shore)","kjv_def":"haven."},
"H4232":{"lemma":"מְחוּיָאֵל","xlit":"Mᵉchûwyâʼêl","pron":"mekh-oo-yaw-ale'","derivation":"or מְחִיּיָאֵל; from H4229 (מָחָה) and H410 (אֵל); smitten of God;","strongs_def":"Mechujael or Mechijael, an antediluvian patriarch","kjv_def":"Mehujael."},
"H4233":{"lemma":"מַחֲוִים","xlit":"Machăvîym","pron":"makh-av-eem'","derivation":"apparently a patrial, but from an unknown place (in the plural only for a singular);","strongs_def":"a Machavite or inhabitant of some place named Machaveh","kjv_def":"Mahavite."},
"H4234":{"lemma":"מָחוֹל","xlit":"mâchôwl","pron":"maw-khole'","derivation":"from H2342 (חוּל);","strongs_def":"a (round) dance","kjv_def":"dance(-cing)."},
"H4235":{"lemma":"מָחוֹל","xlit":"Mâchôwl","pron":"maw-khole'","derivation":"the same as H4234 (מָחוֹל); dancing;","strongs_def":"Machol, an Israelite","kjv_def":"Mahol."},
"H4236":{"lemma":"מַחֲזֶה","xlit":"machăzeh","pron":"makh-az-eh'","derivation":"from H2372 (חָזָה);","strongs_def":"a vision","kjv_def":"vision."},
"H4237":{"lemma":"מֶחֱזָה","xlit":"mechĕzâh","pron":"mekh-ez-aw'","derivation":"from H2372 (חָזָה);","strongs_def":"a window","kjv_def":"light."},
"H4238":{"lemma":"מַחֲזִיאוֹת","xlit":"Machăzîyʼôwth","pron":"makh-az-ee-oth'","derivation":"feminine plural from H2372 (חָזָה); visions;","strongs_def":"Machazioth, an Israelite","kjv_def":"Mahazioth."},
"H4239":{"lemma":"מְחִי","xlit":"mᵉchîy","pron":"mekh-ee'","derivation":"from H4229 (מָחָה);","strongs_def":"a stroke, i.e. battering-ram","kjv_def":"engines."},
"H4240":{"lemma":"מְחִידָא","xlit":"Mᵉchîydâʼ","pron":"mek-ee-daw'","derivation":"from H2330 (חוּד); junction;","strongs_def":"Mechida, one of the Nethinim","kjv_def":"Mehida."},
"H4241":{"lemma":"מִחְיָה","xlit":"michyâh","pron":"mikh-yaw'","derivation":"from H2421 (חָיָה);","strongs_def":"preservation of life; hence, sustenance; also the live flesh, i.e. the quick","kjv_def":"preserve life, quick, recover selves, reviving, sustenance, victuals."},
"H4242":{"lemma":"מְחִיר","xlit":"mᵉchîyr","pron":"mekk-eer'","derivation":"from an unused root meaning to buy;","strongs_def":"price, payment, wages","kjv_def":"gain, hire, price, sold, worth."},
"H4243":{"lemma":"מְחִיר","xlit":"Mᵉchîyr","pron":"mekh-eer'","derivation":"the same as H4242 (מְחִיר); price;","strongs_def":"Mechir, an Israelite","kjv_def":"Mehir."},
"H4244":{"lemma":"מַחְלָה","xlit":"Machlâh","pron":"makh-law'","derivation":"from H2470 (חָלָה); sickness;","strongs_def":"Machlah, the name apparently of two Israelitesses","kjv_def":"Mahlah."},
"H4245":{"lemma":"מַחֲלֶה","xlit":"machăleh","pron":"makh-al-eh'","derivation":"or (feminine) מַחֲלָה; from H2470 (חָלָה);","strongs_def":"sickness","kjv_def":"disease, infirmity, sickness."},
"H4246":{"lemma":"מְחֹלָה","xlit":"mᵉchôlâh","pron":"mek-o-law'","derivation":"feminine of H4234 (מָחוֹל);","strongs_def":"a dance","kjv_def":"company, dances(-cing)."},
"H4247":{"lemma":"מְחִלָּה","xlit":"mᵉchillâh","pron":"mekh-il-law'","derivation":"from H2490 (חָלַל);","strongs_def":"a cavern (as if excavated)","kjv_def":"cave."},
"H4248":{"lemma":"מַחְלוֹן","xlit":"Machlôwn","pron":"makh-lone'","derivation":"from H2470 (חָלָה); sick;","strongs_def":"Machlon, an Israelite","kjv_def":"Mahlon."},
"H4249":{"lemma":"מַחְלִי","xlit":"Machlîy","pron":"makh-lee'","derivation":"from H2470 (חָלָה); sick;","strongs_def":"Machli, the name of two Israelites","kjv_def":"Mahli."},
"H4250":{"lemma":"מַחְלִי","xlit":"Machlîy","pron":"makh-lee'","derivation":"patronymical from H4249 (מַחְלִי);","strongs_def":"a Machlite or (collectively) descendants of Machli","kjv_def":"Mahlites."},
"H4251":{"lemma":"מַחְלֻי","xlit":"machluy","pron":"makh-loo'-ee","derivation":"from H2470 (חָלָה);","strongs_def":"a disease","kjv_def":"disease."},
"H4252":{"lemma":"מַחֲלָף","xlit":"machălâph","pron":"makh-al-awf'","derivation":"from H2498 (חָלַף);","strongs_def":"a (sacrificial) knife (as gliding through the flesh)","kjv_def":"knife."},
"H4253":{"lemma":"מַחְלָפָה","xlit":"machlâphâh","pron":"makh-law-faw'","derivation":"from H2498 (חָלַף);","strongs_def":"a ringlet of hair (as gliding over each other)","kjv_def":"lock."},
"H4254":{"lemma":"מַחֲלָצָה","xlit":"machălâtsâh","pron":"makh-al-aw-tsaw'","derivation":"from H2502 (חָלַץ);","strongs_def":"a mantle (as easily drawn off)","kjv_def":"changeable suit of apparel, change of raiment."},
"H4255":{"lemma":"מַחְלְקָה","xlit":"machlᵉqâh","pron":"makh-lek-aw'","derivation":"(Aramaic) corresponding to H4256 (מַחֲלֹקֶת);","strongs_def":"a section (of the Levites)","kjv_def":"course."},
"H4256":{"lemma":"מַחֲלֹקֶת","xlit":"machălôqeth","pron":"makh-al-o'-keth","derivation":"from H2505 (חָלַק);","strongs_def":"a section (of Levites, people or soldiers)","kjv_def":"company, course, division, portion. See also H5555 (סֶלַע הַמַּחְלְקוֹת)."},
"H4257":{"lemma":"מַחֲלַת","xlit":"machălath","pron":"makh-al-ath'","derivation":"from H2470 (חָלָה); sickness;","strongs_def":"'Machalath', probably the title (initial word) of a popular song","kjv_def":"Mahalath."},
"H4258":{"lemma":"מַחֲלַת","xlit":"Machălath","pron":"makh-al-ath'","derivation":"the same as H4257 (מַחֲלַת); sickness;","strongs_def":"Machalath, the name of an Ishmaelitess and of an Israelitess","kjv_def":"Mahalath."},
"H4259":{"lemma":"מְחֹלָתִי","xlit":"Mᵉchôlâthîy","pron":"mekh-o-law-thee'","derivation":"patrial from H65 (אָבֵל מְחוֹלָה);","strongs_def":"a Mecholathite or inhabitant of Abel-Mecholah","kjv_def":"Mecholathite."},
"H4260":{"lemma":"מַחֲמָאָה","xlit":"machămâʼâh","pron":"makh-am-aw-aw'","derivation":"a denominative from H2529 (חֶמְאָה);","strongs_def":"something buttery (i.e. unctuous and pleasant), as (figuratively) flattery","kjv_def":"[idiom] than butter."},
"H4261":{"lemma":"מַחְמָד","xlit":"machmâd","pron":"makh-mawd'","derivation":"from H2530 (חָמַד);","strongs_def":"delightful; hence, a delight, i.e. object of affection or desire","kjv_def":"beloved, desire, goodly, lovely, pleasant (thing)."},
"H4262":{"lemma":"מַחְמֻד","xlit":"machmud","pron":"makh-mood'","derivation":"or מַחְמוּד; from H2530 (חָמַד);","strongs_def":"desired; hence, a valuable","kjv_def":"pleasant thing."},
"H4263":{"lemma":"מַחְמָל","xlit":"machmâl","pron":"makh-mawl'","derivation":"from H2550 (חָמַל); (by paronomasia with H4261 (מַחְמָד)) delight","strongs_def":"properly, sympathy;","kjv_def":"pitieth."},
"H4264":{"lemma":"מַחֲנֶה","xlit":"machăneh","pron":"makh-an-eh'","derivation":"from H2583 (חָנָה);","strongs_def":"an encampment (of travellers or troops); hence, an army, whether literal (of soldiers) or figurative (of dancers, angels, cattle, locusts, stars; or even the sacred courts)","kjv_def":"army, band, battle, camp, company, drove, host, tents."},
"H4265":{"lemma":"מַחֲנֵה־דָן","xlit":"Machănêh-Dân","pron":"makh-an-ay'-dawn","derivation":"from H4264 (מַחֲנֶה) and H1835 (דָּן); camp of Dan;","strongs_def":"Machaneh-Dan, a place in Palestine","kjv_def":"Mahaneh-dan."},
"H4266":{"lemma":"מַחֲנַיִם","xlit":"Machănayim","pron":"makh-an-ah'-yim","derivation":"dual of H4264 (מַחֲנֶה); double camp;","strongs_def":"Machanajim, a place in Palestine","kjv_def":"Mahanaim."},
"H4267":{"lemma":"מַחֲנַק","xlit":"machănaq","pron":"makh-an-ak'","derivation":"from H2614 (חָנַק);","strongs_def":"choking","kjv_def":"strangling."},
"H4268":{"lemma":"מַחֲסֶה","xlit":"machăçeh","pron":"makh-as-eh'","derivation":"or מַחְסֶה; from H2620 (חָסָה);","strongs_def":"a shelter (literally or figuratively)","kjv_def":"hope, (place of) refuge, shelter, trust."},
"H4269":{"lemma":"מַחְסוֹם","xlit":"machçôwm","pron":"makh-sohm'","derivation":"from H2629 (חָסַם);","strongs_def":"a muzzle","kjv_def":"bridle."},
"H4270":{"lemma":"מַחְסוֹר","xlit":"machçôwr","pron":"makh-sore'","derivation":"or מַחְסֹר; from H2637 (חָסֵר);","strongs_def":"deficiency; hence, impoverishment","kjv_def":"lack, need, penury, poor, poverty, want."},
"H4271":{"lemma":"מַחְסֵיָה","xlit":"Machçêyâh","pron":"makh-say-yaw'","derivation":"from H4268 (מַחֲסֶה) and H3050 (יָהּ); refuge of (i.e. in) Jah;","strongs_def":"Machsejah, an Israelite","kjv_def":"Maaseiah."},
"H4272":{"lemma":"מָחַץ","xlit":"mâchats","pron":"maw-khats'","derivation":"a primitive root;","strongs_def":"to dash asunder; by implication, to crush, smash or violently plunge; figuratively, to subdue or destroy","kjv_def":"dip, pierce (through), smite (through), strike through, wound."},
"H4273":{"lemma":"מַחַץ","xlit":"machats","pron":"makh'-ats","derivation":"from H4272 (מָחַץ);","strongs_def":"a contusion","kjv_def":"stroke."},
"H4274":{"lemma":"מַחְצֵב","xlit":"machtsêb","pron":"makh-tsabe'","derivation":"from H2672 (חָצַב);","strongs_def":"properly, a hewing; concretely, a quarry","kjv_def":"hewed(-n)."},
"H4275":{"lemma":"מֶחֱצָה","xlit":"mechĕtsâh","pron":"mekh-ets-aw'","derivation":"from H2673 (חָצָה)lemma חָעָה ayin, corrected to חָצָה;","strongs_def":"a halving","kjv_def":"half."},
"H4276":{"lemma":"מַחֲצִית","xlit":"machătsîyth","pron":"makh-ats-eeth'","derivation":"from H2673 (חָצָה)lemma חָעָה ayin, corrected to חָצָה;","strongs_def":"a halving or the middle","kjv_def":"half (so much), mid(-day)."},
"H4277":{"lemma":"מָחַק","xlit":"mâchaq","pron":"maw-khak'","derivation":"a primitive root;","strongs_def":"to crush","kjv_def":"smite off."},
"H4278":{"lemma":"מֶחְקָר","xlit":"mechqâr","pron":"mekh-kawr'","derivation":"from H2713 (חָקַר);","strongs_def":"properly, scrutinized, i.e. (by implication) a recess","kjv_def":"deep place."},
"H4279":{"lemma":"מָחָר","xlit":"mâchâr","pron":"maw-khar'","derivation":"probably from H309 (אָחַר);","strongs_def":"properly, deferred, i.e. the morrow; usually (adverbially) tomorrow; indefinitely, hereafter","kjv_def":"time to come, tomorrow."},
"H4280":{"lemma":"מַחֲרָאָה","xlit":"machărâʼâh","pron":"makh-ar-aw-aw'","derivation":"from the same as H2716 (חֶרֶא);","strongs_def":"a sink","kjv_def":"draught house."},
"H4281":{"lemma":"מַחֲרֵשָׁה","xlit":"machărêshâh","pron":"makh-ar-ay-shaw'","derivation":"from H2790 (חָרַשׁ);","strongs_def":"probably a pick-axe","kjv_def":"mattock."},
"H4282":{"lemma":"מַחֲרֶשֶׁת","xlit":"machăresheth","pron":"makh-ar-eh'-sheth","derivation":"from H2790 (חָרַשׁ);","strongs_def":"probably a hoe","kjv_def":"share."},
"H4283":{"lemma":"מׇחֳרָת","xlit":"mochŏrâth","pron":"mokh-or-awth'","derivation":"or מׇחֳרָתָם; (1 Samuel 30:17), feminine from the same as H4279 (מָחָר);","strongs_def":"the morrow or (adverbially) tomorrow","kjv_def":"morrow, next day."},
"H4284":{"lemma":"מַחֲשָׁבָה","xlit":"machăshâbâh","pron":"makh-ash-aw-baw'","derivation":"or מַחֲשֶׁבֶת; from H2803 (חָשַׁב);","strongs_def":"a contrivance, i.e. (concretely) a texture, machine, or (abstractly) intention, plan (whether bad, a plot; or good, advice)","kjv_def":"cunning (work), curious work, device(-sed), imagination, invented, means, purpose, thought."},
"H4285":{"lemma":"מַחְשָׁךְ","xlit":"machshâk","pron":"makh-shawk'","derivation":"from H2821 (חָשַׁךְ);","strongs_def":"darkness; concretely, a dark place","kjv_def":"dark(-ness, place)."},
"H4286":{"lemma":"מַחְשֹׂף","xlit":"machsôph","pron":"makh-sofe'","derivation":"from H2834 (חָשַׂף);","strongs_def":"a peeling","kjv_def":"made appear."},
"H4287":{"lemma":"מַחַת","xlit":"Machath","pron":"makh'-ath","derivation":"probably from H4229 (מָחָה); erasure;","strongs_def":"Machath, the name of two Israelites","kjv_def":"Mahath."},
"H4288":{"lemma":"מְחִתָּה","xlit":"mᵉchittâh","pron":"mekh-it-taw'","derivation":"from H2846 (חָתָה);","strongs_def":"properly, a dissolution; concretely, a ruin, or (abstractly) consternation","kjv_def":"destruction, dismaying, ruin, terror."},
"H4289":{"lemma":"מַחְתָּה","xlit":"machtâh","pron":"makh-taw'","derivation":"the same as H4288 (מְחִתָּה) in the sense of removal;","strongs_def":"a pan for live coals","kjv_def":"censer, firepan, snuffdish."},
"H4290":{"lemma":"מַחְתֶּרֶת","xlit":"machtereth","pron":"makh-teh'-reth","derivation":"from H2864 (חָתַר);","strongs_def":"a burglary; figuratively, unexpected examination","kjv_def":"breaking up, secret search."},
"H4291":{"lemma":"מְטָא","xlit":"mᵉṭâʼ","pron":"met-aw'","derivation":"(Aramaic) or מְטָה; (Aramaic), apparently corresponding to H4672 (מָצָא) in the intransitive sense of being found present;","strongs_def":"to arrive, extend or happen","kjv_def":"come, reach."},
"H4292":{"lemma":"מַטְאֲטֵא","xlit":"maṭʼăṭêʼ","pron":"mat-at-ay'","derivation":"apparently a denominative from H2916 (טִיט);","strongs_def":"a broom (as removing dirt (compare English 'to dust', i.e. remove dust))","kjv_def":"besom."},
"H4293":{"lemma":"מַטְבֵּחַ","xlit":"maṭbêach","pron":"mat-bay'-akh","derivation":"from H2873 (טָבַח);","strongs_def":"slaughter","kjv_def":"slaughter."},
"H4294":{"lemma":"מַטֶּה","xlit":"maṭṭeh","pron":"mat-teh'","derivation":"or (feminine) מַטָּה; from H5186 (נָטָה);","strongs_def":"a branch (as extending); figuratively, a tribe; also a rod, whether for chastising (figuratively, correction), ruling (a sceptre), throwing (a lance), or walking (a staff; figuratively, a support of life, e.g. bread)","kjv_def":"rod, staff, tribe."},
"H4295":{"lemma":"מַטָּה","xlit":"maṭṭâh","pron":"mat'-taw","derivation":"from H5786 (עָוַר) with directive enclitic appended;","strongs_def":"downward, below or beneath; often adverbially with or without prefixes","kjv_def":"beneath, down(-ward), less, very low, under(-neath)."},
"H4296":{"lemma":"מִטָּה","xlit":"miṭṭâh","pron":"mit-taw'","derivation":"from H5186 (נָטָה);","strongs_def":"a bed (as extended) forsleeping or eating; by analogy, a sofa, litter or bier","kjv_def":"bed(-chamber), bier."},
"H4297":{"lemma":"מֻטֶּה","xlit":"muṭṭeh","pron":"moot-teh'","derivation":"from H5186 (נָטָה);","strongs_def":"a stretching, i.e. distortion (figuratively, iniquity)","kjv_def":"perverseness."},
"H4298":{"lemma":"מֻטָּה","xlit":"muṭṭâh","pron":"moot-taw'","derivation":"from H5186 (נָטָה);","strongs_def":"expansion","kjv_def":"stretching out."},
"H4299":{"lemma":"מַטְוֶה","xlit":"maṭveh","pron":"mat-veh'","derivation":"from H2901 (טָוָה);","strongs_def":"something spun","kjv_def":"spun."},
"H4300":{"lemma":"מְטִיל","xlit":"mᵉṭîyl","pron":"met-eel'","derivation":"from H2904 (טוּל) in the sense of hammering out;","strongs_def":"an iron bar (as forged)","kjv_def":"bar."},
"H4301":{"lemma":"מַטְמוֹן","xlit":"maṭmôwn","pron":"mat-mone'","derivation":"or מַטְמֹן; or מַטְמֻן; from H2934 (טָמַן);","strongs_def":"a secret storehouse; hence, a secreted valuable (buried); generally money","kjv_def":"hidden riches, (hid) treasure(-s)."},
"H4302":{"lemma":"מַטָּע","xlit":"maṭṭâʻ","pron":"mat-taw'","derivation":"from H5193 (נָטַע);","strongs_def":"something planted, i.e. the place (a garden or vineyard), or the thing (a plant, figuratively or men); by implication, the act, planting","kjv_def":"plant(-ation, -ing)."},
"H4303":{"lemma":"מַטְעַם","xlit":"maṭʻam","pron":"mat-am'","derivation":"or (feminine) מַטְעַמָּה; from H2938 (טָעַם);","strongs_def":"a delicacy","kjv_def":"dainty (meat), savoury meat."},
"H4304":{"lemma":"מִטְפַּחַת","xlit":"miṭpachath","pron":"mit-pakh'-ath","derivation":"from H2946 (טָפַח);","strongs_def":"a wide cloak (for a woman)","kjv_def":"vail, wimple."},
"H4305":{"lemma":"מָטַר","xlit":"mâṭar","pron":"maw-tar'","derivation":"a primitive root;","strongs_def":"to rain","kjv_def":"(cause to) rain (upon)."},
"H4306":{"lemma":"מָטַר","xlit":"mâṭar","pron":"maw-tawr'","derivation":"from H4305 (מָטַר);","strongs_def":"rain","kjv_def":"rain."},
"H4307":{"lemma":"מַטָּרָא","xlit":"maṭṭârâʼ","pron":"mat-taw-raw'","derivation":"or מַטָּרָה; from H5201 (נָטַר);","strongs_def":"a jail (as a guard-house); also an aim (as being closely watched)","kjv_def":"mark, prison."},
"H4308":{"lemma":"מַטְרֵד","xlit":"Maṭrêd","pron":"mat-rade'","derivation":"from H2956 (טָרַד); propulsive;","strongs_def":"Matred, an Edomitess","kjv_def":"Matred."},
"H4309":{"lemma":"מַטְרִי","xlit":"Maṭrîy","pron":"mat-ree'","derivation":"from H4305 (מָטַר); rainy;","strongs_def":"Matri, an Israelite","kjv_def":"Matri."},
"H4310":{"lemma":"מִי","xlit":"mîy","pron":"me","derivation":"an interrogative pronoun of persons, as H4100 (מָה) is of things,","strongs_def":"who? (occasionally, by a peculiar idiom, of things); also (indefinitely) whoever; often used in oblique construction with prefix or suffix","kjv_def":"any (man), [idiom] he, [idiom] him, [phrase] O that! what, which, who(-m, -se, -soever), [phrase] would to God."},
"H4311":{"lemma":"מֵידְבָא","xlit":"Mêydᵉbâʼ","pron":"may-deb-aw'","derivation":"from H4325 (מַיִם) and H1679 (דֹּבֶא); water of quiet;","strongs_def":"Medeba, a place in Palestine","kjv_def":"Medeba."},
"H4312":{"lemma":"מֵידָד","xlit":"Mêydâd","pron":"may-dawd'","derivation":"from H3032 (יָדַד) in the sense of loving; affectionate;","strongs_def":"Medad, an Israelite","kjv_def":"Medad."},
"H4313":{"lemma":"מֵי הַיַּרְקוֹן","xlit":"Mêy hay-Yarqôwn","pron":"may hah'-ee-yar-kone'","derivation":"from H4325 (מַיִם) and H3420 (יֵרָקוֹן) with the art. interposed; water of the yellowness;","strongs_def":"Me-haj-Jarkon, a place in Palestine","kjv_def":"Me-jarkon."},
"H4314":{"lemma":"מֵי זָהָב","xlit":"Mêy Zâhâb","pron":"may zaw-hawb'","derivation":"from H4325 (מַיִם) and H2091 (זָהָב), water of gold;","strongs_def":"Me-Zahab, an Edomite","kjv_def":"Mezahab."},
"H4315":{"lemma":"מֵיטָב","xlit":"mêyṭâb","pron":"may-tawb'","derivation":"from H3190 (יָטַב);","strongs_def":"the best part","kjv_def":"best."},
"H4316":{"lemma":"מִיכָא","xlit":"Mîykâʼ","pron":"mee-kaw'","derivation":"a variation for H4318 (מִיכָה);","strongs_def":"Mica, the name of two Israelites","kjv_def":"Micha."},
"H4317":{"lemma":"מִיכָאֵל","xlit":"Mîykâʼêl","pron":"me-kaw-ale'","derivation":"from H4310 (מִי) and (the prefix derivative from) H3588 (כִּי) and H410 (אֵל); who (is) like God?;","strongs_def":"Mikael, the name of an archangel and of nine Israelites","kjv_def":"Michael."},
"H4318":{"lemma":"מִיכָה","xlit":"Mîykâh","pron":"mee-kaw'","derivation":"an abbreviated form of H4320 (מִיכָיָה);","strongs_def":"Micah, the name of seven Israelites","kjv_def":"Micah, Micaiah, Michah."},
"H4319":{"lemma":"מִיכָהוּ","xlit":"Mîykâhûw","pron":"me-kaw'-hoo","derivation":"a contr. for H4321 (מִיכָיְהוּ);","strongs_def":"Mikehu, an Israelite prophet","kjv_def":"Micaiah (2 Chronicles 18:8)."},
"H4320":{"lemma":"מִיכָיָה","xlit":"Mîykâyâh","pron":"me-kaw-yaw'","derivation":"from H4310 (מִי) and (the prefix derivative from) H3588 (כִּי) and H3050 (יָהּ); who (is) like Jah?;","strongs_def":"Micajah, the name of two Israelites","kjv_def":"Micah, Michaiah. Compare H4318 (מִיכָה)."},
"H4321":{"lemma":"מִיכָיְהוּ","xlit":"Mîykâyᵉhûw","pron":"me-kaw-yeh-hoo'","derivation":"or מִכָיְהוּ; (Jeremiah 36:11), abbreviated form for H4322 (מִיכָיָהוּ);","strongs_def":"Mikajah, the name of three Israelites","kjv_def":"Micah, Micaiah, Michaiah."},
"H4322":{"lemma":"מִיכָיָהוּ","xlit":"Mîykâyâhûw","pron":"me-kaw-yaw'-hoo","derivation":"for H4320 (מִיכָיָה);","strongs_def":"Mikajah, the name of an Israelite and an Israelitess","kjv_def":"Michaiah."},
"H4323":{"lemma":"מִיכָל","xlit":"mîykâl","pron":"me-kawl'","derivation":"from H3201 (יָכֹל);","strongs_def":"properly, a container, i.e. a streamlet","kjv_def":"brook."},
"H4324":{"lemma":"מִיכָל","xlit":"Mîykâl","pron":"me-kawl'","derivation":"apparently the same as H4323 (מִיכָל); revulet;","strongs_def":"Mikal, Saul's daughter","kjv_def":"Michal."},
"H4325":{"lemma":"מַיִם","xlit":"mayim","pron":"mah'-yim","derivation":"dual of a primitive noun (but used in a singular sense);","strongs_def":"water; figuratively, juice; by euphemism, urine, semen","kjv_def":"[phrase] piss, wasting, water(-ing, (-course, -flood, -spring))."},
"H4326":{"lemma":"מִיָּמִן","xlit":"Mîyâmin","pron":"me-yaw-meem'","derivation":"a form for H4509 (מִנְיָמִין);","strongs_def":"Mijamin, the name of three Israelites","kjv_def":"Miamin, Mijamin."},
"H4327":{"lemma":"מִין","xlit":"mîyn","pron":"meen","derivation":"from an unused root meaning to portion out;","strongs_def":"a sort, i.e. species","kjv_def":"kind. Compare H4480 (מִן)."},
"H4328":{"lemma":"מְיֻסָּדָה","xlit":"mᵉyuççâdâh","pron":"meh-yoos-saw-daw'","derivation":"properly, feminine passive participle of H3245 (יָסַד);","strongs_def":"something founded, i.e. a foundation","kjv_def":"foundation."},
"H4329":{"lemma":"מֵיסָךְ","xlit":"mêyçâk","pron":"may-sawk'","derivation":"from H5526 (סָכַךְ);","strongs_def":"a portico (as covered)","kjv_def":"covert."},
"H4330":{"lemma":"מִיץ","xlit":"mîyts","pron":"meets","derivation":"from H4160 (מוּץ);","strongs_def":"pressure","kjv_def":"churning, forcing, wringing."},
"H4331":{"lemma":"מֵישָׁא","xlit":"Mêyshâʼ","pron":"may-shaw'","derivation":"from H4185 (מוּשׁ); departure;","strongs_def":"Mesha, a place in Arabia; also an Israelite","kjv_def":"Mesha."},
"H4332":{"lemma":"מִישָׁאֵל","xlit":"Mîyshâʼêl","pron":"mee-shaw-ale'","derivation":"from H4310 (מִי) and H410 (אֵל) with the abbreviated form insep. relatively (see H834 (אֲשֶׁר)) interposed; who (is) what God (is) ?;","strongs_def":"Mishael, the name of three Israelites","kjv_def":"Mishael."},
"H4333":{"lemma":"מִישָׁאֵל","xlit":"Mîyshâʼêl","pron":"mee-shaw-ale'","derivation":"(Aramaic) corresponding to H4332 (מִישָׁאֵל);","strongs_def":"Mishael, an Israelite","kjv_def":"Mishael."},
"H4334":{"lemma":"מִישׁוֹר","xlit":"mîyshôwr","pron":"mee-shore'","derivation":"or מִישֹׁר; from H3474 (יָשַׁר);","strongs_def":"a level, i.e. a plain (often used (with the article prefix) as a proper name of certain districts); figuratively, concord; also straightness, i.e. (figuratively) justice (sometimes adverbially, justly)","kjv_def":"equity, even place, plain, right(-eously), (made) straight, uprightness."},
"H4335":{"lemma":"מֵישַׁךְ","xlit":"Mêyshak","pron":"may-shak'","derivation":"borrowed from H4336 (מֵישַׁךְ);","strongs_def":"Meshak, an Israelite","kjv_def":"Meshak."},
"H4336":{"lemma":"מֵישַׁךְ","xlit":"Mêyshak","pron":"may-shak'","derivation":"(Aramaic) of foreign origin and doubtful significance; name of H4333 (מִישָׁאֵל)","strongs_def":"Meshak, the Babylonian","kjv_def":"Meshak."},
"H4337":{"lemma":"מֵישָׁע","xlit":"Mêyshâʻ","pron":"may-shah'","derivation":"from H3467 (יָשַׁע); safety;","strongs_def":"Mesha, an Israelite","kjv_def":"Mesha."},
"H4338":{"lemma":"מֵישַׁע","xlit":"Mêyshaʻ","pron":"may-shaw'","derivation":"a variation for H4337 (מֵישָׁע); safety;","strongs_def":"Mesha, a Moabite","kjv_def":"Mesha."},
"H4339":{"lemma":"מֵישָׁר","xlit":"mêyshâr","pron":"may-shawr'","derivation":"from H3474 (יָשַׁר);","strongs_def":"evenness, i.e. (figuratively) prosperity or concord; also straightness, i.e. (figuratively) rectitude (only in plural with singular sense; often adverbially)","kjv_def":"agreement, aright, that are equal, equity, (things that are) right(-eously, things), sweetly, upright(-ly, -ness)."},
"H4340":{"lemma":"מֵיתָר","xlit":"mêythâr","pron":"may-thar'","derivation":"from H3498 (יָתַר); (compare H3499 (יֶתֶר))","strongs_def":"a cord (of a tent); or the string (of a bow)","kjv_def":"cord, string."},
"H4341":{"lemma":"מַכְאֹב","xlit":"makʼôb","pron":"mak-obe'","derivation":"sometimes מַכְאוֹב; also (feminine Isaiah 53:3) מַכְאֹבָה; from H3510 (כָּאַב);","strongs_def":"anguish or (figuratively) affliction","kjv_def":"grief, pain, sorrow."},
"H4342":{"lemma":"מַכְבִּיר","xlit":"makbîyr","pron":"mak-beer'","derivation":"transitive participle of H3527 (כָּבַר);","strongs_def":"plenty","kjv_def":"abundance."},
"H4343":{"lemma":"מַכְבֵּנָא","xlit":"Makbênâʼ","pron":"mak-bay-naw'","derivation":"from the same as H3522 (כַּבּוֹן); knoll;","strongs_def":"Macbena, a place in Palestine settled by him","kjv_def":"Machbenah."},
"H4344":{"lemma":"מַכְבַּנַּי","xlit":"Makbannay","pron":"mak-ban-nah'-ee","derivation":"patrial from H4343 (מַכְבֵּנָא);","strongs_def":"a Macbannite or native of Macbena","kjv_def":"Machbanai."},
"H4345":{"lemma":"מַכְבֵּר","xlit":"makbêr","pron":"mak-bare'","derivation":"from H3527 (כָּבַר) in the sense of covering (compare H3531 (כְּבָרָה));","strongs_def":"a grate","kjv_def":"grate."},
"H4346":{"lemma":"מַכְבָּר","xlit":"makbâr","pron":"mak-bawr'","derivation":"from H3527 (כָּבַר) in the sense of covering; (compare H4345 (מַכְבֵּר))","strongs_def":"a cloth (as netted)","kjv_def":"thick cloth."},
"H4347":{"lemma":"מַכָּה","xlit":"makkâh","pron":"mak-kaw'","derivation":"or (masculine) מַכֶּה; (plural only) from H5221 (נָכָה); a blow (in 2 Chronicles 2:10, of the flail); by implication,","strongs_def":"a wound; figuratively, carnage, also pestilence","kjv_def":"beaten, blow, plague, slaughter, smote, [idiom] sore, stripe, stroke, wound(-ed)."},
"H4348":{"lemma":"מִכְוָה","xlit":"mikvâh","pron":"mik-vaw'","derivation":"from H3554 (כָּוָה);","strongs_def":"a burn","kjv_def":"that burneth, burning."},
"H4349":{"lemma":"מָכוֹן","xlit":"mâkôwn","pron":"maw-kone'","derivation":"from H3559 (כּוּן);","strongs_def":"properly, a fixture, i.e. a basis; generally a place, especially as an abode","kjv_def":"foundation, habitation, (dwelling-, settled) place."},
"H4350":{"lemma":"מְכוֹנָה","xlit":"mᵉkôwnâh","pron":"mek-o-naw'","derivation":"or מְכֹנָה; feminine of H4349 (מָכוֹן);","strongs_def":"a pedestal, also a spot","kjv_def":"base."},
"H4351":{"lemma":"מְכוּרָה","xlit":"mᵉkûwrâh","pron":"mek-oo-raw'","derivation":"or מְכֹרָה; from the same as H3564 (כּוּר) in the sense of dipping;","strongs_def":"origin (as if a mine)","kjv_def":"birth, habitation, nativity."},
"H4352":{"lemma":"מָכִי","xlit":"Mâkîy","pron":"maw-kee'","derivation":"probably from H4134 (מוּךְ); pining;","strongs_def":"Maki, an Israelite","kjv_def":"Machi."},
"H4353":{"lemma":"מָכִיר","xlit":"Mâkîyr","pron":"maw-keer'","derivation":"from H4376 (מָכַר); salesman;","strongs_def":"Makir, an Israelite","kjv_def":"Machir."},
"H4354":{"lemma":"מָכִירִי","xlit":"Mâkîyrîy","pron":"maw-kee-ree'","derivation":"patronymical from H4353 (מָכִיר);","strongs_def":"a Makirite or descendant of Makir","kjv_def":"of Machir."},
"H4355":{"lemma":"מָכַךְ","xlit":"mâkak","pron":"maw-kak'","derivation":"a primitive root;","strongs_def":"to tumble (in ruins); figuratively, to perish","kjv_def":"be brought low, decay."},
"H4356":{"lemma":"מִכְלָאָה","xlit":"miklâʼâh","pron":"mik-law-aw'","derivation":"or מִכְלָה; from H3607 (כָּלָא);","strongs_def":"a pen (for flocks)","kjv_def":"(sheep-)fold. Compare H4357 (מִכְלָה)."},
"H4357":{"lemma":"מִכְלָה","xlit":"miklâh","pron":"mik-law'","derivation":"from H3615 (כָּלָה);","strongs_def":"completion (in plural concrete adverbial, wholly)","kjv_def":"perfect. Compare H4356 (מִכְלָאָה)."},
"H4358":{"lemma":"מִכְלוֹל","xlit":"miklôwl","pron":"mik-lole'","derivation":"from H3634 (כָּלַל);","strongs_def":"perfection (i.e. concrete adverbial, splendidly)","kjv_def":"most gorgeously, all sorts."},
"H4359":{"lemma":"מִכְלָל","xlit":"miklâl","pron":"mik-lawl'","derivation":"from H3634 (כָּלַל);","strongs_def":"perfection (of beauty)","kjv_def":"perfection."},
"H4360":{"lemma":"מִכְלֻל","xlit":"miklul","pron":"mik-lool'","derivation":"from H3634 (כָּלַל);","strongs_def":"something perfect, i.e. a splendid garment","kjv_def":"all sorts."},
"H4361":{"lemma":"מַכֹּלֶת","xlit":"makkôleth","pron":"mak-ko'-leth","derivation":"from H398 (אָכַל);","strongs_def":"nourishment","kjv_def":"food."},
"H4362":{"lemma":"מִכְמַן","xlit":"mikman","pron":"mik-man'","derivation":"from the same as H3646 (כַּמֹּן) in the sense of hiding;","strongs_def":"treasure (as hidden)","kjv_def":"treasure."},
"H4363":{"lemma":"מִכְמָס","xlit":"Mikmâç","pron":"mik-maws'","derivation":"(Ezra 2:2 or מִכְמָשׁ; or מִכְמַשׁ; (Nehemiah 11:31), from H3647 (כָּמַס); hidden;","strongs_def":"Mikmas or Mikmash, a place in Palestine","kjv_def":"Mikmas, Mikmash."},
"H4364":{"lemma":"מַכְמָר","xlit":"makmâr","pron":"mak-mawr'","derivation":"or מִכְמֹר; from H3648 (כָּמַר) in the sense of blackening by heat;","strongs_def":"a (hunter's) net (as dark from concealment)","kjv_def":"net."},
"H4365":{"lemma":"מִכְמֶרֶת","xlit":"mikmereth","pron":"mik-meh'-reth","derivation":"or מִכְמֹרֶת; feminine of H4364 (מַכְמָר);","strongs_def":"a (fisher's) net","kjv_def":"drag, net."},
"H4366":{"lemma":"מִכְמְתָת","xlit":"Mikmᵉthâth","pron":"mik-meth-awth'","derivation":"apparently from an unused root meaning to hide; concealment;","strongs_def":"Mikmethath, a place in Palestine","kjv_def":"Michmethath."},
"H4367":{"lemma":"מַכְנַדְבַי","xlit":"Maknadbay","pron":"mak-nad-bah'-ee","derivation":"from H4100 (מָה) and H5068 (נָדַב) with a particle interposed; what (is) like (a) liberal (man) ?;","strongs_def":"Maknadbai, an Israelite","kjv_def":"Machnadebai."},
"H4368":{"lemma":"מְכֹנָה","xlit":"Mᵉkônâh","pron":"mek-o-naw'","derivation":"the same as H4350 (מְכוֹנָה); a base;","strongs_def":"Mekonah, a place in Palestine","kjv_def":"Mekonah."},
"H4369":{"lemma":"מְכֻנָה","xlit":"mᵉkunâh","pron":"mek-oo-naw'","derivation":"the same as H4350 (מְכוֹנָה);","strongs_def":"a spot","kjv_def":"base."},
"H4370":{"lemma":"מִכְנָס","xlit":"miknâç","pron":"mik-nawce'","derivation":"from H3647 (כָּמַס) in the sense of hiding;","strongs_def":"(only in dual) drawers (from concealing the private parts)","kjv_def":"breeches."},
"H4371":{"lemma":"מֶכֶס","xlit":"mekeç","pron":"meh'-kes","derivation":"probably from an unused root meaning to enumerate;","strongs_def":"an assessment (as based upon a census)","kjv_def":"tribute."},
"H4372":{"lemma":"מִכְסֶה","xlit":"mikçeh","pron":"mik-seh'","derivation":"from H3680 (כָּסָה);","strongs_def":"a covering, i.e. weatherboarding","kjv_def":"covering."},
"H4373":{"lemma":"מִכְסָה","xlit":"mikçâh","pron":"mik-saw'","derivation":"feminine of H4371 (מֶכֶס);","strongs_def":"an enumeration; by implication, a valuation","kjv_def":"number, worth."},
"H4374":{"lemma":"מְכַסֶּה","xlit":"mᵉkaççeh","pron":"mek-as-seh'","derivation":"from H3680 (כָּסָה);","strongs_def":"a covering, i.e. garment; specifically, a coverlet (for a bed), an awning (from the sun); also the omentum (as covering the intestines)","kjv_def":"clothing, to cover, that which covereth."},
"H4375":{"lemma":"מַכְפֵּלָה","xlit":"Makpêlâh","pron":"mak-pay-law'","derivation":"from H3717 (כָּפַל); a fold;","strongs_def":"Makpelah, a place in Palestine","kjv_def":"Machpelah."},
"H4376":{"lemma":"מָכַר","xlit":"mâkar","pron":"maw-kar'","derivation":"a primitive root;","strongs_def":"to sell, literally (as merchandise, a daughter in marriage, into slavery), or figuratively (to surrender)","kjv_def":"[idiom] at all, sell (away, -er, self)."},
"H4377":{"lemma":"מֶכֶר","xlit":"meker","pron":"meh'-ker","derivation":"from H4376 (מָכַר);","strongs_def":"merchandise; also value","kjv_def":"pay, price, ware."},
"H4378":{"lemma":"מַכָּר","xlit":"makkâr","pron":"mak-kawr'","derivation":"from H5234 (נָכַר);","strongs_def":"an acquaintance","kjv_def":"acquaintance."},
"H4379":{"lemma":"מִכְרֶה","xlit":"mikreh","pron":"mik-reh'","derivation":"from H3738 (כָּרָה);","strongs_def":"a pit (for salt)","kjv_def":"(salt-) pit."},
"H4380":{"lemma":"מְכֵרָה","xlit":"mᵉkêrâh","pron":"mek-ay-raw'","derivation":"probably from the same as H3564 (כּוּר) in the sense of stabbing;","strongs_def":"a sword","kjv_def":"habitation."},
"H4381":{"lemma":"מִכְרִי","xlit":"Mikrîy","pron":"mik-ree'","derivation":"from H4376 (מָכַר); salesman;","strongs_def":"Mikri, an Israelite","kjv_def":"Michri."},
"H4382":{"lemma":"מְכֵרָתִי","xlit":"Mᵉkêrâthîy","pron":"mek-ay-raw-thee'","derivation":"patrial from an unused name (the same as H4380 (מְכֵרָה)) of a place in Palestine;","strongs_def":"a Mekerathite, or inhabitant of Mekerah","kjv_def":"Mecherathite."},
"H4383":{"lemma":"מִכְשׁוֹל","xlit":"mikshôwl","pron":"mik-shole'","derivation":"or מִכְשֹׁל; masculine from H3782 (כָּשַׁל);","strongs_def":"a stumbling-block, literally or figuratively (obstacle, enticement (specifically an idol), scruple)","kjv_def":"caused to fall, offence, [idiom] (no-) thing offered, ruin, stumbling-block."},
"H4384":{"lemma":"מַכְשֵׁלָה","xlit":"makshêlâh","pron":"mak-shay-law'","derivation":"feminine from H3782 (כָּשַׁל);","strongs_def":"a stumbling-block, but only figuratively (fall, enticement (idol))","kjv_def":"ruin, stumbling-block."},
"H4385":{"lemma":"מִכְתָּב","xlit":"miktâb","pron":"mik-tawb'","derivation":"from H3789 (כָּתַב);","strongs_def":"a thing written, the characters, or a document (letter, copy, edict, poem)","kjv_def":"writing."},
"H4386":{"lemma":"מְכִתָּה","xlit":"mᵉkittâh","pron":"mek-it-taw'","derivation":"from H3807 (כָּתַת);","strongs_def":"a fracture","kjv_def":"bursting."},
"H4387":{"lemma":"מִכְתָּם","xlit":"miktâm","pron":"mik-tawm'","derivation":"from H3799 (כָּתַם);","strongs_def":"an engraving, i.e. (techn.) a poem","kjv_def":"Michtam."},
"H4388":{"lemma":"מַכְתֵּשׁ","xlit":"maktêsh","pron":"mak-taysh'","derivation":"from H3806 (כָּתַשׁ);","strongs_def":"a mortar; by analogy, a socket (of a tooth)","kjv_def":"hollow place, mortar."},
"H4389":{"lemma":"מַכְתֵּשׁ","xlit":"Maktêsh","pron":"mak-taysh'","derivation":"the same as H4388 (מַכְתֵּשׁ);","strongs_def":"dell; the Maktesh, a place in Jerusalem","kjv_def":"Maktesh."},
"H4390":{"lemma":"מָלֵא","xlit":"mâlêʼ","pron":"maw-lay'","derivation":"or מָלָא; (Esther 7:5), a primitive root;","strongs_def":"to fill or (intransitively) be full of, in a wide application (literally and figuratively)","kjv_def":"accomplish, confirm, [phrase] consecrate, be at an end, be expired, be fenced, fill, fulfil, (be, become, [idiom] draw, give in, go) full(-ly, -ly set, tale), (over-) flow, fulness, furnish, gather (selves, together), presume, replenish, satisfy, set, space, take a (hand-) full, [phrase] have wholly."},
"H4391":{"lemma":"מְלָא","xlit":"mᵉlâʼ","pron":"mel-aw'","derivation":"(Aramaic) corresponding to H4390 (מָלֵא);","strongs_def":"to fill","kjv_def":"fill, be full."},
"H4392":{"lemma":"מָלֵא","xlit":"mâlêʼ","pron":"maw-lay'","derivation":"from H4390 (מָלֵא);","strongs_def":"full (literally or figuratively) or filling (literally); also (concretely) fulness; adverbially, fully","kjv_def":"[idiom] she that was with child, fill(-ed, -ed with), full(-ly), multitude, as is worth."},
"H4393":{"lemma":"מְלֹא","xlit":"mᵉlôʼ","pron":"mel-o'","derivation":"rarely מְלוֹא; or מְלוֹ; (Ezekiel [41:8), from H4390 (מָלֵא);","strongs_def":"fulness (literally or figuratively)","kjv_def":"[idiom] all along, [idiom] all that is (there-) in, fill, ([idiom] that whereof...was) full, fulness, (hand-) full, multitude."},
"H4394":{"lemma":"מִלֻּא","xlit":"milluʼ","pron":"mil-loo'","derivation":"from H4390 (מָלֵא);","strongs_def":"a fulfilling (only in plural), i.e. (literally) a setting (of gems), or (technically) consecration (also concretely, a dedicatory sacrifice)","kjv_def":"consecration, be set."},
"H4395":{"lemma":"מְלֵאָה","xlit":"mᵉlêʼâh","pron":"mel-ay-aw'","derivation":"feminine of H4392 (מָלֵא);","strongs_def":"something fulfilled, i.e. abundance (of produce)","kjv_def":"(first of ripe) fruit, fulness."},
"H4396":{"lemma":"מִלֻּאָה","xlit":"milluʼâh","pron":"mil-loo-aw'","derivation":"feminine of H4394 (מִלֻּא);","strongs_def":"a filling, i.e. setting (of gems)","kjv_def":"enclosing, setting."},
"H4397":{"lemma":"מֲלְאָךְ","xlit":"mălʼâk","pron":"mal-awk'","derivation":"from an unused root meaning to despatch as a deputy;","strongs_def":"a messenger; specifically, of God, i.e. an angel (also a prophet, priest or teacher)","kjv_def":"ambassador, angel, king, messenger."},
"H4398":{"lemma":"מַלְאַךְ","xlit":"malʼak","pron":"mal-ak'","derivation":"(Aramaic) corresponding to H4397 (מֲלְאָךְ);","strongs_def":"an angel","kjv_def":"angel."},
"H4399":{"lemma":"מְלָאכָה","xlit":"mᵉlâʼkâh","pron":"mel-aw-kaw'","derivation":"from the same as H4397 (מֲלְאָךְ);","strongs_def":"properly, deputyship, i.e. ministry; generally, employment (never servile) or work (abstractly or concretely); also property (as the result of labor)","kjv_def":"business, [phrase] cattle, [phrase] industrious, occupation, ([phrase] -pied), [phrase] officer, thing (made), use, (manner of) work((-man), -manship)."},
"H4400":{"lemma":"מַלְאֲכוּת","xlit":"malʼăkûwth","pron":"mal-ak-ooth'","derivation":"from the same as H4397 (מֲלְאָךְ);","strongs_def":"a message","kjv_def":"message."},
"H4401":{"lemma":"מַלְאָכִי","xlit":"Malʼâkîy","pron":"mal-aw-kee'","derivation":"from the same as H4397 (מֲלְאָךְ); ministrative;","strongs_def":"Malaki, a prophet","kjv_def":"Malachi."},
"H4402":{"lemma":"מִלֵּאת","xlit":"millêʼth","pron":"mil-layth'","derivation":"from H4390 (מָלֵא);","strongs_def":"fulness, i.e. (concretely) a plump socket (of the eye)","kjv_def":"[idiom] fitly."},
"H4403":{"lemma":"מַלְבּוּשׁ","xlit":"malbûwsh","pron":"mal-boosh'","derivation":"or מַלְבֻּשׁ; from H3847 (לָבַשׁ);","strongs_def":"a garment, or (collectively) clothing","kjv_def":"apparel, raiment, vestment."},
"H4404":{"lemma":"מַלְבֵּן","xlit":"malbên","pron":"mal-bane'","derivation":"from H3835 (לָבַן) (denominative);","strongs_def":"a brickkiln","kjv_def":"brickkiln."},
"H4405":{"lemma":"מִלָּה","xlit":"millâh","pron":"mil-law'","derivation":"from H4448 (מָלַל) (plural masculine as if from מִלֶּה;","strongs_def":"a word; collectively, a discourse; figuratively, a topic","kjv_def":"[phrase] answer, by-word, matter, any thing (what) to say, to speak(-ing), speak, talking, word."},
"H4406":{"lemma":"מִלָּה","xlit":"millâh","pron":"mil-law'","derivation":"(Aramaic) corresponding to H4405 (מִלָּה);","strongs_def":"a word, command, discourse, or subject","kjv_def":"commandment, matter, thing. word."},
"H4407":{"lemma":"מִלּוֹא","xlit":"millôwʼ","pron":"mil-lo'","derivation":"or מִלֹּא; (2 Kings 12:20), from H4390 (מָלֵא);","strongs_def":"a rampart (as filled in), i.e. the citadel","kjv_def":"Millo. See also H1037 (בֵּית מִלּוֹא)."},
"H4408":{"lemma":"מַלּוּחַ","xlit":"mallûwach","pron":"mal-loo'-akh","derivation":"from H4414 (מָלַח);","strongs_def":"sea-purslain (from its saltness)","kjv_def":"mallows."},
"H4409":{"lemma":"מַלּוּךְ","xlit":"Mallûwk","pron":"mal-luke'","derivation":"or מַלּוּכִי; (Nehemiah 12:14), from H4427 (מָלַךְ); remnant;","strongs_def":"Malluk, the name of five Israelites","kjv_def":"Malluch, Melichu (from the margin)."},
"H4410":{"lemma":"מְלוּכָה","xlit":"mᵉlûwkâh","pron":"mel-oo-kaw'","derivation":"feminine passive participle of H4427 (מָלַךְ);","strongs_def":"something ruled, i.e. a realm","kjv_def":"kingsom, king's, [idiom] royal."},
"H4411":{"lemma":"מָלוֹן","xlit":"mâlôwn","pron":"maw-lone'","derivation":"from H3885 (לוּן);","strongs_def":"a lodgment, i.e. caravanserai or encampment","kjv_def":"inn, place where...lodge, lodging (place)."},
"H4412":{"lemma":"מְלוּנָה","xlit":"mᵉlûwnâh","pron":"mel-oo-naw'","derivation":"feminine from H3885 (לוּן);","strongs_def":"a hut, a hammock","kjv_def":"cottage, lodge."},
"H4413":{"lemma":"מַלּוֹתִי","xlit":"Mallôwthîy","pron":"mal-lo'-thee","derivation":"apparently from H4448 (מָלַל); I have talked (i.e. loquacious)","strongs_def":"Mallothi, an Israelite","kjv_def":"Mallothi."},
"H4414":{"lemma":"מָלַח","xlit":"mâlach","pron":"maw-lakh'","derivation":"a primitive root; also as denominative from H4417 (מֶלַח)","strongs_def":"properly, to rub to pieces or pulverize; intransitively, to disappear as dust; to salt whether internally (to season with salt) or externally (to rub with salt)","kjv_def":"[idiom] at all, salt, season, temper together, vanish away."},
"H4415":{"lemma":"מְלַח","xlit":"mᵉlach","pron":"mel-akh'","derivation":"(Aramaic) corresponding to H4414 (מָלַח);","strongs_def":"to eat salt, i.e. (generally) subsist","kjv_def":"[phrase] have maintenance."},
"H4416":{"lemma":"מְלַח","xlit":"mᵉlach","pron":"mel-akh'","derivation":"(Aramaic) from H4415 (מְלַח);","strongs_def":"salt","kjv_def":"[phrase] maintenance, salt."},
"H4417":{"lemma":"מֶלַח","xlit":"melach","pron":"meh'-lakh","derivation":"from H4414 (מָלַח);","strongs_def":"properly, powder, i.e. (specifically) salt (as easily pulverized and dissolved)","kjv_def":"salt(-pit)."},
"H4418":{"lemma":"מָלָח","xlit":"mâlâch","pron":"maw-lawkh'","derivation":"from H4414 (מָלַח) in its original sense;","strongs_def":"a rag or old garment","kjv_def":"rotten rag."},
"H4419":{"lemma":"מַלָּח","xlit":"mallâch","pron":"mal-lawkh'","derivation":"from H4414 (מָלַח) in its second. sense;","strongs_def":"a sailor (as following 'the salt')","kjv_def":"mariner."},
"H4420":{"lemma":"מְלֵחָה","xlit":"mᵉlêchâh","pron":"mel-ay-khaw'","derivation":"from H4414 (מָלַח) (in its denominative sense); understood as land (H776 (אֶרֶץ))","strongs_def":"properly, salted (i.e. land), i.e. a desert","kjv_def":"barren land(-ness), salt (land)."},
"H4421":{"lemma":"מִלְחָמָה","xlit":"milchâmâh","pron":"mil-khaw-maw'","derivation":"from H3898 (לָחַם) (in the sense of fighting);","strongs_def":"a battle (i.e. the engagement); generally, war (i.e. warfare)","kjv_def":"battle, fight(-ing), war(-rior)."},
"H4422":{"lemma":"מָלַט","xlit":"mâlaṭ","pron":"maw-lat'","derivation":"a primitive root;","strongs_def":"properly, to be smooth, i.e. (by implication) to escape (as if by slipperiness); causatively, to release or rescue; specifically, to bring forth young, emit sparks","kjv_def":"deliver (self), escape, lay, leap out, let alone, let go, preserve, save, [idiom] speedily, [idiom] surely."},
"H4423":{"lemma":"מֶלֶט","xlit":"meleṭ","pron":"meh'-let","derivation":"from H4422 (מָלַט),","strongs_def":"cement (from its plastic smoothness)","kjv_def":"clay."},
"H4424":{"lemma":"מְלַטְיָה","xlit":"Mᵉlaṭyâh","pron":"mel-at-yaw'","derivation":"from H4423 (מֶלֶט) and H3050 (יָהּ); (whom) Jah has delivered;","strongs_def":"Melatjah, a Gibeonite","kjv_def":"Melatiah."},
"H4425":{"lemma":"מְלִילָה","xlit":"mᵉlîylâh","pron":"mel-ee-law'","derivation":"from H4449 (מְלַל) (in the sense of cropping (compare H4135 (מוּל)));","strongs_def":"a head of grain (as cut off)","kjv_def":"ear."},
"H4426":{"lemma":"מְלִיצָה","xlit":"mᵉlîytsâh","pron":"mel-ee-tsaw'","derivation":"from H3887 (לוּץ);","strongs_def":"an aphorism; also a satire","kjv_def":"interpretation, taunting."},
"H4427":{"lemma":"מָלַךְ","xlit":"mâlak","pron":"maw-lak'","derivation":"a primitive root;","strongs_def":"to reign; inceptively, to ascend the throne; causatively, to induct into royalty; hence (by implication) to take counsel","kjv_def":"consult, [idiom] indeed, be (make, set a, set up) king, be (make) queen, (begin to, make to) reign(-ing), rule, [idiom] surely."},
"H4428":{"lemma":"מֶלֶךְ","xlit":"melek","pron":"meh'-lek","derivation":"from H4427 (מָלַךְ);","strongs_def":"a king","kjv_def":"king, royal."},
"H4429":{"lemma":"מֶלֶךְ","xlit":"Melek","pron":"meh'-lek","derivation":"the same as H4428 (מֶלֶךְ); king;","strongs_def":"Melek, the name of two Israelites","kjv_def":"Melech, Hammelech (by including the article)."},
"H4430":{"lemma":"מֶלֶךְ","xlit":"melek","pron":"meh'-lek","derivation":"(Aramaic) corresponding to H4428 (מֶלֶךְ);","strongs_def":"a king","kjv_def":"king, royal."},
"H4431":{"lemma":"מְלַךְ","xlit":"mᵉlak","pron":"mel-ak'","derivation":"(Aramaic) from a root corresponding to H4427 (מָלַךְ) in the sense of consultation;","strongs_def":"advice","kjv_def":"counsel."},
"H4432":{"lemma":"מֹלֶךְ","xlit":"Môlek","pron":"mo'-lek","derivation":"from H4427 (מָלַךְ);","strongs_def":"Molek (i.e. king), the chief deity of the Ammonites","kjv_def":"Molech. Compare H4445 (מַלְכָּם)."},
"H4433":{"lemma":"מַלְכָּא","xlit":"malkâʼ","pron":"mal-kaw'","derivation":"(Aramaic) corresponding to H4436 (מַלְכָּה);","strongs_def":"a queen","kjv_def":"queen."},
"H4434":{"lemma":"מַלְכֹּדֶת","xlit":"malkôdeth","pron":"mal-ko'-deth","derivation":"from H3920 (לָכַד);","strongs_def":"a snare","kjv_def":"trap."},
"H4435":{"lemma":"מִלְכָּה","xlit":"Milkâh","pron":"mil-kaw'","derivation":"a form of H4436 (מַלְכָּה); queen;","strongs_def":"Milcah, the name of a Hebrewess and of an Israelite","kjv_def":"Milcah."},
"H4436":{"lemma":"מַלְכָּה","xlit":"malkâh","pron":"mal-kaw'","derivation":"feminine of H4428 (מֶלֶךְ);","strongs_def":"a queen","kjv_def":"queen."},
"H4437":{"lemma":"מַלְכוּ","xlit":"malkûw","pron":"mal-koo'","derivation":"(Aramaic) corresponding to H4438 (מַלְכוּת);","strongs_def":"dominion (abstractly or concretely)","kjv_def":"kingdom, kingly, realm, reign."},
"H4438":{"lemma":"מַלְכוּת","xlit":"malkûwth","pron":"mal-kooth'","derivation":"or מַלְכֻת; or (in plural) מַלְכֻיָּה; from H4427 (מָלַךְ);","strongs_def":"a rule; concretely, a dominion","kjv_def":"empire, kingdom, realm, reign, royal."},
"H4439":{"lemma":"מַלְכִּיאֵל","xlit":"Malkîyʼêl","pron":"mal-kee-ale'","derivation":"from H4428 (מֶלֶךְ) and H410 (אֵל); king of (i.e. appointed by) God;","strongs_def":"Malkiel, an Israelite","kjv_def":"Malchiel."},
"H4440":{"lemma":"מַלְכִּיאֵלִי","xlit":"Malkîyʼêlîy","pron":"mal-kee-ay-lee'","derivation":"patronymical from H4439 (מַלְכִּיאֵל);","strongs_def":"a Malkielite or descendant of Malkiel","kjv_def":"Malchielite."},
"H4441":{"lemma":"מַלְכִּיָּה","xlit":"Malkîyâh","pron":"mal-kee-yaw'","derivation":"or מַלְכִּיָּהוּlemma מַלְכִּיָהוּ missing dagesh, corrected to מַלְכִּיָּהוּ; (Jeremiah [38:6), from H4428 (מֶלֶךְ) and H3050 (יָהּ); king of (i.e. appointed by) Jah;","strongs_def":"Malkijah, the name of ten Israelites","kjv_def":"Malchiah, Malchijah."},
"H4442":{"lemma":"מַלְכִּי־צֶדֶק","xlit":"Malkîy-Tsedeq","pron":"mal-kee-tseh'-dek","derivation":"from H4428 (מֶלֶךְ) and H6664 (צֶדֶק); king of right;","strongs_def":"Malki-Tsedek, an early king in Palestine","kjv_def":"Melchizedek."},
"H4443":{"lemma":"מַלְכִּירָם","xlit":"Malkîyrâm","pron":"mal-kee-rawm'","derivation":"from H4428 (מֶלֶךְ) and H7311 (רוּם); king of a high one (i.e. of exaltation);","strongs_def":"Malkiram, an Israelite","kjv_def":"Malchiram."},
"H4444":{"lemma":"מַלְכִּישׁוּעַ","xlit":"Malkîyshûwaʻ","pron":"mal-kee-shoo'-ah","derivation":"from H4428 (מֶלֶךְ) and H7769 (שׁוּעַ); king of wealth;","strongs_def":"Malkishua, an Israelite","kjv_def":"Malchishua."},
"H4445":{"lemma":"מַלְכָּם","xlit":"Malkâm","pron":"mal-kawm'","derivation":"or מִלְכּוֹם; from H4428 (מֶלֶךְ) for H4432 (מֹלֶךְ);","strongs_def":"Malcam or Milcom, the national idol of the Ammonites","kjv_def":"Malcham, Milcom."},
"H4446":{"lemma":"מְלֶכֶת","xlit":"mᵉleketh","pron":"mel-eh'-keth","derivation":"from H4427 (מָלַךְ);","strongs_def":"a queen","kjv_def":"queen."},
"H4447":{"lemma":"מֹלֶכֶת","xlit":"Môleketh","pron":"mo-leh'-keth","derivation":"feminine active participle of H4427 (מָלַךְ); queen;","strongs_def":"Moleketh, an Israelitess","kjv_def":"Hammoleketh (including the article)."},
"H4448":{"lemma":"מָלַל","xlit":"mâlal","pron":"maw-lal'","derivation":"a primitive root;","strongs_def":"to speak (mostly poetical) or say","kjv_def":"say, speak, utter."},
"H4449":{"lemma":"מְלַל","xlit":"mᵉlal","pron":"mel-al'","derivation":"(Aramaic) corresponding to H4448 (מָלַל);","strongs_def":"to speak","kjv_def":"say, speak(-ing)."},
"H4450":{"lemma":"מִלֲלַי","xlit":"Milălay","pron":"mee-lal-ah'-ee","derivation":"from H4448 (מָלַל); talkative;","strongs_def":"Milalai, an Israelite","kjv_def":"Milalai."},
"H4451":{"lemma":"מַלְמָד","xlit":"malmâd","pron":"mal-mawd'","derivation":"from H3925 (לָמַד);","strongs_def":"a goad for oxen","kjv_def":"goad."},
"H4452":{"lemma":"מָלַץ","xlit":"mâlats","pron":"maw-lats'","derivation":"a primitive root;","strongs_def":"to be smooth, i.e. (figuratively) pleasant","kjv_def":"be sweet."},
"H4453":{"lemma":"מֶלְצָר","xlit":"meltsâr","pron":"mel-tsawr'","derivation":"of Persian derivation;","strongs_def":"the butler or other officer in the Babylonian court","kjv_def":"Melzar."},
"H4454":{"lemma":"מָלַק","xlit":"mâlaq","pron":"maw-lak'","derivation":"a primitive root;","strongs_def":"to crack a joint; by implication, to wring the neck of afowl (without separating it)","kjv_def":"wring off."},
"H4455":{"lemma":"מַלְקוֹחַ","xlit":"malqôwach","pron":"mal-ko'-akh","derivation":"from H3947 (לָקַח);","strongs_def":"transitively (in dual) the jaws (as taking food); intransitively, spoil (and captives) (as taken)","kjv_def":"booty, jaws, prey."},
"H4456":{"lemma":"מַלְקוֹשׁ","xlit":"malqôwsh","pron":"mal-koshe'","derivation":"from H3953 (לָקַשׁ); (compare H3954 (לֶקֶשׁ))","strongs_def":"the spring rain; figuratively, eloquence","kjv_def":"latter rain."},
"H4457":{"lemma":"מֶלְקָח","xlit":"melqâch","pron":"mel-kawkh'","derivation":"or מַלְקָח; from H3947 (לָקַח);","strongs_def":"(only in dual) tweezers","kjv_def":"snuffers, tongs."},
"H4458":{"lemma":"מֶלְתָּחָה","xlit":"meltâchâh","pron":"mel-taw-khaw'","derivation":"from an unused root meaning to spread out;","strongs_def":"a wardrobe (i.e. room where clothing is spread)","kjv_def":"vestry."},
"H4459":{"lemma":"מַלְתָּעָה","xlit":"maltâʻâh","pron":"mal-taw-aw'","derivation":"transp. for H4973 (מְתַלְּעָה);","strongs_def":"a grinder, i.e. back tooth","kjv_def":"great tooth."},
"H4460":{"lemma":"מַמְּגֻרָה","xlit":"mammᵉgurâh","pron":"mam-meg-oo-raw'","derivation":"from H4048 (מָגַר) (in the sense of depositing);","strongs_def":"a granary","kjv_def":"barn."},
"H4461":{"lemma":"מֵמַד","xlit":"mêmad","pron":"may-mad'","derivation":"from H4058 (מָדַד);","strongs_def":"a measure","kjv_def":"measure."},
"H4462":{"lemma":"מְמוּכָן","xlit":"Mᵉmûwkân","pron":"mem-oo-kawn'","derivation":"or (transp.) מוֹמֻכָן; (Esther 1:16), of Persian derivation;","strongs_def":"Memucan or Momucan, a Persian satrap","kjv_def":"Memucan."},
"H4463":{"lemma":"מָמוֹת","xlit":"mâmôwth","pron":"maw-mothe'","derivation":"from H4191 (מוּת);","strongs_def":"a mortal disease; concretely, a corpse","kjv_def":"death."},
"H4464":{"lemma":"מַמְזֵר","xlit":"mamzêr","pron":"mam-zare'","derivation":"from an unused root meaning to alienate;","strongs_def":"a mongrel, i.e. born of aJewish father and aheathen mother","kjv_def":"bastard."},
"H4465":{"lemma":"מִמְכָּר","xlit":"mimkâr","pron":"mim-kawr'","derivation":"from H4376 (מָכַר);","strongs_def":"merchandise; abstractly, a selling","kjv_def":"[idiom] ought, (that which cometh of) sale, that which...sold, ware."},
"H4466":{"lemma":"מִמְכֶּרֶת","xlit":"mimkereth","pron":"mim-keh'-reth","derivation":"feminine of H4465 (מִמְכָּר);","strongs_def":"a sale","kjv_def":"[phrase] sold as."},
"H4467":{"lemma":"מַמְלָכָה","xlit":"mamlâkâh","pron":"mam-law-kaw'","derivation":"from H4427 (מָלַךְ);","strongs_def":"dominion, i.e. (abstractly) the estate (rule) or (concretely) the country (realm)","kjv_def":"kingdom, king's, reign, royal."},
"H4468":{"lemma":"מַמְלָכוּת","xlit":"mamlâkûwth","pron":"mam-law-kooth'","derivation":"a form of H4467 (מַמְלָכָה) and equiv. to it","strongs_def":"{dominion, i.e. (abstractly) the estate (rule) or (concretely) the country (realm)}","kjv_def":"kingdom, reign."},
"H4469":{"lemma":"מַמְסָךְ","xlit":"mamçâk","pron":"mam-sawk'","derivation":"from H4537 (מָסַךְ);","strongs_def":"mixture, i.e. (specifically) wine mixed (with water or spices)","kjv_def":"drink-offering, mixed wine."},
"H4470":{"lemma":"מֶמֶר","xlit":"memer","pron":"meh'-mer","derivation":"from an unused root meaning to grieve;","strongs_def":"sorrow","kjv_def":"bitterness."},
"H4471":{"lemma":"מַמְרֵא","xlit":"Mamrêʼ","pron":"mam-ray'","derivation":"from H4754 (מָרָא)(in the sense of vigor); lusty;","strongs_def":"Mamre, an Amorite","kjv_def":"Mamre."},
"H4472":{"lemma":"מַמְרֹר","xlit":"mamrôr","pron":"mam-rore'","derivation":"from H4843 (מָרַר);","strongs_def":"a bitterness, i.e. (figuratively) calamity","kjv_def":"bitterness."},
"H4473":{"lemma":"מִמְשַׁח","xlit":"mimshach","pron":"mim-shakh'","derivation":"from H4886 (מָשַׁח), in the sense of expansion;","strongs_def":"outspread (i.e. with outstretched wings)","kjv_def":"anointed."},
"H4474":{"lemma":"מִמְשָׁל","xlit":"mimshâl","pron":"mim-shawl'","derivation":"from H4910 (מָשַׁל);","strongs_def":"a ruler or (abstractly) rule","kjv_def":"dominion, that ruled."},
"H4475":{"lemma":"מֶמְשָׁלָה","xlit":"memshâlâh","pron":"mem-shaw-law'","derivation":"feminine of H4474 (מִמְשָׁל);","strongs_def":"rule; also (concretely in plural) a realm or a ruler","kjv_def":"dominion, government, power, to rule."},
"H4476":{"lemma":"מִמְשָׁק","xlit":"mimshâq","pron":"mim-shawk'","derivation":"from the same as H4943 (מֶשֶׁק);","strongs_def":"a possession","kjv_def":"breeding."},
"H4477":{"lemma":"מַמְתַּק","xlit":"mamtaq","pron":"mam-tak'","derivation":"from H4985 (מָתַק);","strongs_def":"something sweet (literally or figuratively)","kjv_def":"(most) sweet."},
"H4478":{"lemma":"מָן","xlit":"mân","pron":"mawn","derivation":"from H4100 (מָה);","strongs_def":"literally a whatness (so to speak), i.e. manna (so called from the question about it)","kjv_def":"manna."},
"H4479":{"lemma":"מָן","xlit":"mân","pron":"mawn","derivation":"(Aramaic) from H4101 (מָה);","strongs_def":"who or what (properly, interrogatively, hence, also indefinitely and relatively)","kjv_def":"what, who(-msoever, [phrase] -so)."},
"H4480":{"lemma":"מִן","xlit":"min","pron":"min","derivation":"or מִנִּי; or מִנֵּי; (constructive plural) (Isaiah 30:11); for H4482 (מֵן);","strongs_def":"properly, a part of; hence (prepositionally), from or out of in many senses","kjv_def":"above, after, among, at, because of, by (reason of), from (among), in, [idiom] neither, [idiom] nor, (out) of, over, since, [idiom] then, through, [idiom] whether, with."},
"H4481":{"lemma":"מִן","xlit":"min","pron":"min","derivation":"(Aramaic) corresponding to H4480 (מִן)","strongs_def":"{properly, a part of; hence (prepositionally), from or out of}","kjv_def":"according, after, [phrase] because, [phrase] before, by, for, from, [idiom] him, [idiom] more than, (out) of, part, since, [idiom] these, to, upon, [phrase] when."},
"H4482":{"lemma":"מֵן","xlit":"mên","pron":"mane","derivation":"from an unused root meaning to apportion; a part; hence,","strongs_def":"a musical chord (as parted into strings)","kjv_def":"in (the same) (Psalm 68:23), stringed instrument (Psalm 150:4), whereby (Psalm 45:8 (defective plural))."},
"H4483":{"lemma":"מְנָא","xlit":"mᵉnâʼ","pron":"men-aw'","derivation":"(Aramaic) or מְנָה; (Aramaic), corresponding to H4487 (מָנָה);","strongs_def":"to count, appoint","kjv_def":"number, ordain, set."},
"H4484":{"lemma":"מְנֵא","xlit":"mᵉnêʼ","pron":"men-ay'","derivation":"(Aramaic) passive participle of H4483 (מְנָא);","strongs_def":"numbered","kjv_def":"Mene."},
"H4485":{"lemma":"מַנְגִּינָה","xlit":"mangîynâh","pron":"man-ghee-naw'","derivation":"from H5059 (נָגַן);","strongs_def":"a satire","kjv_def":"music."},
"H4486":{"lemma":"מַנְדַּע","xlit":"mandaʻ","pron":"man-dah'","derivation":"(Aramaic) corresponding to H4093 (מַדָּע);","strongs_def":"wisdom or intelligence","kjv_def":"knowledge, reason, understanding."},
"H4487":{"lemma":"מָנָה","xlit":"mânâh","pron":"maw-naw'","derivation":"a primitive root;","strongs_def":"properly, to weigh out; by implication, to allot or constitute officially; also to enumerate or enroll","kjv_def":"appoint, count, number, prepare, set, tell."},
"H4488":{"lemma":"מָנֶה","xlit":"mâneh","pron":"maw-neh'","derivation":"from H4487 (מָנָה);","strongs_def":"properly, a fixed weight or measured amount, i.e. (techn.) a maneh or mina","kjv_def":"maneh, pound."},
"H4489":{"lemma":"מֹנֶה","xlit":"môneh","pron":"mo-neh'","derivation":"from H4487 (מָנָה);","strongs_def":"properly, something weighed out, i.e. (figuratively) a portion of time, i.e. an instance","kjv_def":"time."},
"H4490":{"lemma":"מָנָה","xlit":"mânâh","pron":"maw-naw'","derivation":"from H4487 (מָנָה);","strongs_def":"properly, something weighed out, i.e. (generally) a division; specifically (of food) a ration; also a lot","kjv_def":"such things as belonged, part, portion."},
"H4491":{"lemma":"מִנְהָג","xlit":"minhâg","pron":"min-hawg'","derivation":"from H5090 (נָהַג);","strongs_def":"the driving (of a chariot)","kjv_def":"driving."},
"H4492":{"lemma":"מִנְהָרָה","xlit":"minhârâh","pron":"min-haw-raw'","derivation":"from H5102 (נָהַר);","strongs_def":"properly, a channel or fissure, i.e. (by implication) a cavern","kjv_def":"den."},
"H4493":{"lemma":"מָנוֹד","xlit":"mânôwd","pron":"maw-node'","derivation":"from H5110 (נוּד)","strongs_def":"a nodding or toss (of the head in derision)","kjv_def":"shaking."},
"H4494":{"lemma":"מָנוֹחַ","xlit":"mânôwach","pron":"maw-no'-akh","derivation":"from H5117 (נוּחַ);","strongs_def":"quiet, i.e. (concretely) a settled spot, or (figuratively) a home","kjv_def":"(place of) rest."},
"H4495":{"lemma":"מָנוֹחַ","xlit":"Mânôwach","pron":"maw-no'-akh","derivation":"the same as H4494 (מָנוֹחַ); rest;","strongs_def":"Manoach, an Israelite","kjv_def":"Manoah."},
"H4496":{"lemma":"מְנוּחָה","xlit":"mᵉnûwchâh","pron":"men-oo-khaw'","derivation":"or מְנֻחָה; feminine of H4495 (מָנוֹחַ);","strongs_def":"repose or (adverbially) peacefully; figuratively, consolation (specifically, matrimony); hence (concretely) an abode","kjv_def":"comfortable, ease, quiet, rest(-ing place), still."},
"H4497":{"lemma":"מָנוֹן","xlit":"mânôwn","pron":"maw-nohn'","derivation":"from H5125 (נוּן);","strongs_def":"a continuator, i.e. heir","kjv_def":"son."},
"H4498":{"lemma":"מָנוֹס","xlit":"mânôwç","pron":"maw-noce'","derivation":"from H5127 (נוּס);","strongs_def":"a retreat (literally or figuratively); abstractly, a fleeing","kjv_def":"[idiom] apace, escape, way to flee, flight, refuge."},
"H4499":{"lemma":"מְנוּסָה","xlit":"mᵉnûwçâh","pron":"men-oo-saw'","derivation":"or מְנֻסָה; feminine of H4498 (מָנוֹס);","strongs_def":"retreat","kjv_def":"fleeing, flight."},
"H4500":{"lemma":"מָנוֹר","xlit":"mânôwr","pron":"maw-nore'","derivation":"from H5214 (נִיר);","strongs_def":"a yoke (properly, for plowing), i.e. the frame of aloom","kjv_def":"beam."},
"H4501":{"lemma":"מְנוֹרָה","xlit":"mᵉnôwrâh","pron":"men-o-raw'","derivation":"or מְנֹרָה; feminine of H4500 (מָנוֹר) (in the original sense of H5216 (נִיר));","strongs_def":"a chandelier","kjv_def":"candlestick."},
"H4502":{"lemma":"מִנְּזָר","xlit":"minnᵉzâr","pron":"min-ez-awr'","derivation":"from H5144 (נָזַר);","strongs_def":"a prince","kjv_def":"crowned."},
"H4503":{"lemma":"מִנְחָה","xlit":"minchâh","pron":"min-khaw'","derivation":"from an unused root meaning to apportion, i.e. bestow;","strongs_def":"a donation; euphemistically, tribute; specifically a sacrificial offering (usually bloodless and voluntary)","kjv_def":"gift, oblation, (meat) offering, present, sacrifice."},
"H4504":{"lemma":"מִנְחָה","xlit":"minchâh","pron":"min-khaw'","derivation":"(Aramaic) corresponding to H4503 (מִנְחָה);","strongs_def":"a sacrificial offering","kjv_def":"oblation, meat offering."},
"H4505":{"lemma":"מְנַחֵם","xlit":"Mᵉnachêm","pron":"men-akh-ame'","derivation":"from H5162 (נָחַם); comforter;","strongs_def":"Menachem, an Israelite","kjv_def":"Menahem."},
"H4506":{"lemma":"מָנַחַת","xlit":"Mânachath","pron":"maw-nakh'-ath","derivation":"from H5117 (נוּחַ); rest;","strongs_def":"Manachath, the name of an Edomite and of a place in Moab","kjv_def":"Manahath."},
"H4507":{"lemma":"מְנִי","xlit":"Mᵉnîy","pron":"men-ee'","derivation":"from H4487 (מָנָה);","strongs_def":"the Apportioner, i.e. Fate (as an idol)","kjv_def":"number."},
"H4508":{"lemma":"מִנִּי","xlit":"Minnîy","pron":"min-nee'","derivation":"of foreign derivation;","strongs_def":"Minni, an Armenian province","kjv_def":"Minni."},
"H4509":{"lemma":"מִנְיָמִין","xlit":"Minyâmîyn","pron":"min-yaw-meen'","derivation":"from H4480 (מִן) and H3225 (יָמִין); from (the) right hand;","strongs_def":"Minjamin, the name of two Israelites","kjv_def":"Miniamin. Compare H4326 (מִיָּמִן)."},
"H4510":{"lemma":"מִנְיָן","xlit":"minyân","pron":"min-yawn'","derivation":"(Aramaic) from H4483 (מְנָא);","strongs_def":"enumeration","kjv_def":"number."},
"H4511":{"lemma":"מִנִּית","xlit":"Minnîyth","pron":"min-neeth'","derivation":"from the same as H4482 (מֵן); enumeration;","strongs_def":"Minnith, a place East of the Jordan","kjv_def":"Minnith."},
"H4512":{"lemma":"מִנְלֶה","xlit":"minleh","pron":"min-leh'","derivation":"from H5239 (נָלָה);","strongs_def":"completion, i.e. (in produce) wealth","kjv_def":"perfection."},
"H4513":{"lemma":"מָנַע","xlit":"mânaʻ","pron":"maw-nah'","derivation":"a primitive root;","strongs_def":"to debar (negatively or positively) from benefit or injury","kjv_def":"deny, keep (back), refrain, restrain, withhold."},
"H4514":{"lemma":"מַנְעוּל","xlit":"manʻûwl","pron":"man-ool'","derivation":"or מַנְעֻל; from H5274 (נָעַל);","strongs_def":"a bolt","kjv_def":"lock."},
"H4515":{"lemma":"מִנְעָל","xlit":"minʻâl","pron":"man-awl'","derivation":"from H5274 (נָעַל);","strongs_def":"a bolt","kjv_def":"shoe."},
"H4516":{"lemma":"מַנְעַם","xlit":"manʻam","pron":"man-am'","derivation":"from H5276 (נָעֵם);","strongs_def":"a delicacy","kjv_def":"dainty."},
"H4517":{"lemma":"מְנַעְנַע","xlit":"mᵉnaʻnaʻ","pron":"men-ah-ah'","derivation":"from H5128 (נוּעַ);","strongs_def":"a sistrum (so called from its rattling sound)","kjv_def":"cornet."},
"H4518":{"lemma":"מְנַקִּית","xlit":"mᵉnaqqîyth","pron":"men-ak-keeth'","derivation":"from H5352 (נָקָה);","strongs_def":"a sacrificial basin (for holding blood)","kjv_def":"bowl."},
"H4519":{"lemma":"מְנַשֶּׁה","xlit":"Mᵉnashsheh","pron":"men-ash-sheh'","derivation":"from H5382 (נָשָׁה); causing to forget;","strongs_def":"Menashsheh, a grandson of Jacob, also the tribe descended from him, and its territory","kjv_def":"Manasseh."},
"H4520":{"lemma":"מְנַשִּׁי","xlit":"Mᵉnashshîy","pron":"men-ash-shee'","derivation":"from H4519 (מְנַשֶּׁה);","strongs_def":"a Menashshite or descendant of Menashsheh","kjv_def":"of Manasseh, Manassites."},
"H4521":{"lemma":"מְנָת","xlit":"mᵉnâth","pron":"men-awth'","derivation":"from H4487 (מָנָה);","strongs_def":"an allotment (by courtesy, law or providence)","kjv_def":"portion."},
"H4522":{"lemma":"מַס","xlit":"maç","pron":"mas","derivation":"or מִס; from H4549 (מָסַס);","strongs_def":"properly, a burden (as causing to faint), i.e. a tax in the form of forced labor","kjv_def":"discomfited, levy, task(-master), tribute(-tary)."},
"H4523":{"lemma":"מָס","xlit":"mâç","pron":"mawce","derivation":"from H4549 (מָסַס);","strongs_def":"fainting, i.e. (figuratively) disconsolate","kjv_def":"is afflicted."},
"H4524":{"lemma":"מֵסַב","xlit":"mêçab","pron":"may-sab'","derivation":"plural masculine מְסִבִּים; or feminine מְסִבּוֹת; from H5437 (סָבַב);","strongs_def":"a divan (as enclosing the room); abstractly (adverbial) around","kjv_def":"that compass about, (place) round about, at table."},
"H4525":{"lemma":"מַסְגֵּר","xlit":"maçgêr","pron":"mas-gare'","derivation":"from H5462 (סָגַר);","strongs_def":"a fastener, i.e. (of a person) a smith, (of a thing) a prison","kjv_def":"prison, smith."},
"H4526":{"lemma":"מִסְגֶּרֶת","xlit":"miçgereth","pron":"mis-gheh'-reth","derivation":"from H5462 (סָגַר);","strongs_def":"something enclosing, i.e. a margin (of a region, of a panel); concretely, a stronghold","kjv_def":"border, close place, hole."},
"H4527":{"lemma":"מַסַּד","xlit":"maççad","pron":"mas-sad'","derivation":"from H3245 (יָסַד);","strongs_def":"a foundation","kjv_def":"foundation."},
"H4528":{"lemma":"מִסְדְּרוֹן","xlit":"miçdᵉrôwn","pron":"mis-der-ohn'","derivation":"from the same as H5468 (סֶדֶר);","strongs_def":"a colonnade or internal portico (from its rows of pillars)","kjv_def":"porch."},
"H4529":{"lemma":"מָסָה","xlit":"mâçâh","pron":"maw-saw'","derivation":"a primitive root;","strongs_def":"to dissolve","kjv_def":"make to consume away, (make to) melt, water."},
"H4530":{"lemma":"מִסָּה","xlit":"miççâh","pron":"mis-saw'","derivation":"from H4549 (מָסַס) (in the sense of flowing);","strongs_def":"abundance, i.e. (adverbially) liberally","kjv_def":"tribute."},
"H4531":{"lemma":"מַסָּה","xlit":"maççâh","pron":"mas-saw'","derivation":"from H5254 (נָסָה);","strongs_def":"a testing, of men (judicial) or of God (querulous)","kjv_def":"temptation, trial."},
"H4532":{"lemma":"מַסָּה","xlit":"Maççâh","pron":"mas-saw'","derivation":"the same as H4531 (מַסָּה);","strongs_def":"Massah, a place in the Desert","kjv_def":"Massah."},
"H4533":{"lemma":"מַסְוֶה","xlit":"maçveh","pron":"mas-veh'","derivation":"apparently from an unused root meaning to cover;","strongs_def":"a veil","kjv_def":"vail."},
"H4534":{"lemma":"מְסוּכָה","xlit":"mᵉçûwkâh","pron":"mes-oo-kaw'","derivation":"for H4881 (מְשׂוּכָה);","strongs_def":"a hedge","kjv_def":"thorn hedge."},
"H4535":{"lemma":"מַסָּח","xlit":"maççâch","pron":"mas-sawkh'","derivation":"from H5255 (נָסַח) in the sense of staving off;","strongs_def":"a cordon, (adverbially) or (as a) military barrier","kjv_def":"broken down."},
"H4536":{"lemma":"מִסְחָר","xlit":"miçchâr","pron":"mis-khawr'","derivation":"from H5503 (סָחַר);","strongs_def":"trade","kjv_def":"traffic."},
"H4537":{"lemma":"מָסַךְ","xlit":"mâçak","pron":"maw-sak'","derivation":"a primitive root;","strongs_def":"to mix, especially wine (with spices)","kjv_def":"mingle."},
"H4538":{"lemma":"מֶסֶךְ","xlit":"meçek","pron":"meh'-sek","derivation":"from H4537 (מָסַךְ);","strongs_def":"a mixture, i.e. of wine with spices","kjv_def":"mixture."},
"H4539":{"lemma":"מָסָךְ","xlit":"mâçâk","pron":"maw-sawk'","derivation":"from H5526 (סָכַךְ);","strongs_def":"a cover, i.e. veil","kjv_def":"covering, curtain, hanging."},
"H4540":{"lemma":"מְסֻכָּה","xlit":"mᵉçukkâh","pron":"mes-ook-kaw'","derivation":"from H5526 (סָכַךְ);","strongs_def":"a covering, i.e. garniture","kjv_def":"covering."},
"H4541":{"lemma":"מַסֵּכָה","xlit":"maççêkâh","pron":"mas-say-kaw'","derivation":"from H5258 (נָסַךְ);","strongs_def":"properly, a pouring over, i.e. fusion of metal (especially a cast image); by implication, a libation, i.e. league; concretely a coverlet (as if poured out)","kjv_def":"covering, molten (image), vail."},
"H4542":{"lemma":"מִסְכֵּן","xlit":"miçkên","pron":"mis-kane'","derivation":"from H5531 (סִכְלוּת);","strongs_def":"indigent","kjv_def":"poor (man)."},
"H4543":{"lemma":"מִסְכְּנָה","xlit":"miçkᵉnâh","pron":"mis-ken-aw'","derivation":"by transp. from H3664 (כָּנַס);","strongs_def":"a magazine","kjv_def":"store(-house), treasure."},
"H4544":{"lemma":"מִסְכֵּנֻת","xlit":"miçkênuth","pron":"mis-kay-nooth'","derivation":"from H4542 (מִסְכֵּן);","strongs_def":"indigence","kjv_def":"scarceness."},
"H4545":{"lemma":"מַסֶּכֶת","xlit":"maççeketh","pron":"mas-seh'-keth","derivation":"from H5259 (נָסַךְ) in the sense of spreading out;","strongs_def":"something expanded, i.e. the warp in aloom (as stretched out to receive the woof)","kjv_def":"web."},
"H4546":{"lemma":"מְסִלָּה","xlit":"mᵉçillâh","pron":"mes-il-law'","derivation":"from H5549 (סָלַל);","strongs_def":"a thoroughfare (as turnpiked), literally or figuratively; specifically a viaduct, a staircase","kjv_def":"causeway, course, highway, path, terrace."},
"H4547":{"lemma":"מַסְלוּל","xlit":"maçlûwl","pron":"mas-lool'","derivation":"from H5549 (סָלַל);","strongs_def":"a thoroughfare (as turnpiked)","kjv_def":"highway."},
"H4548":{"lemma":"מַסְמֵר","xlit":"maçmêr","pron":"mas-mare'","derivation":"or מִסְמֵר; also (feminine) מַסְמְרָה; or מִסְמְרָה; or even מַשְׂמְרָה; (Ecclesiastes 12:11), from H5568 (סָמַר);","strongs_def":"a peg (as bristling from the surface)","kjv_def":"nail."},
"H4549":{"lemma":"מָסַס","xlit":"mâçaç","pron":"maw-sas'","derivation":"a primitive root;","strongs_def":"to liquefy; figuratively, to waste (with disease), to faint (with fatigue, fear or grief)","kjv_def":"discourage, faint, be loosed, melt (away), refuse, [idiom] utterly."},
"H4550":{"lemma":"מַסַּע","xlit":"maççaʻ","pron":"mas-sah'","derivation":"from H5265 (נָסַע);","strongs_def":"a departure (from striking the tents), i.e. march (not necessarily a single day's travel); by implication, a station (or point of departure)","kjv_def":"journey(-ing)."},
"H4551":{"lemma":"מַסָּע","xlit":"maççâʻ","pron":"mas-saw'","derivation":"from H5265 (נָסַע) in the sense of projecting;","strongs_def":"a missile (spear or arrow); also a quarry (whence stones are, as it were, ejected)","kjv_def":"before it was brought, dart."},
"H4552":{"lemma":"מִסְעָד","xlit":"miçʻâd","pron":"mis-awd'","derivation":"' from H5582 (סָעַד);","strongs_def":"a balustrade (for stairs)","kjv_def":"pillar."},
"H4553":{"lemma":"מִסְפֵּד","xlit":"miçpêd","pron":"mis-pade'","derivation":"from H5594 (סָפַד);","strongs_def":"a lamentation","kjv_def":"lamentation, one mourneth, mourning, wailing."},
"H4554":{"lemma":"מִסְפּוֹא","xlit":"miçpôwʼ","pron":"mis-po'","derivation":"from an unused root meaning to collect;","strongs_def":"fodder","kjv_def":"provender."},
"H4555":{"lemma":"מִסְפָּחָה","xlit":"miçpâchâh","pron":"mis-paw-khaw'","derivation":"from H5596 (סָפַח);","strongs_def":"a veil (as spread out)","kjv_def":"kerchief."},
"H4556":{"lemma":"מִסְפַּחַת","xlit":"miçpachath","pron":"mis-pakh'-ath","derivation":"from H5596 (סָפַח);","strongs_def":"scruf (as spreading over the surface)","kjv_def":"scab."},
"H4557":{"lemma":"מִסְפָּר","xlit":"miçpâr","pron":"mis-pawr'","derivation":"from H5608 (סָפַר);","strongs_def":"a number, definite (arithmetical) or indefinite (large, innumerable; small, a few); also (abstractly) narration","kjv_def":"[phrase] abundance, account, [idiom] all, [idiom] few, (in-) finite, (certain) number(-ed), tale, telling, [phrase] time."},
"H4558":{"lemma":"מִסְפָּר","xlit":"Miçpâr","pron":"mis-pawr'","derivation":"the same as H4457 (מֶלְקָח); number;","strongs_def":"Mispar, an Israelite","kjv_def":"Mizpar. Compare H4559 (מִסְפֶּרֶת)."},
"H4559":{"lemma":"מִסְפֶּרֶת","xlit":"Miçpereth","pron":"mis-peh'-reth","derivation":"feminine of H4437 (מַלְכוּ); enumeration;","strongs_def":"Mispereth, an Israelite","kjv_def":"Mispereth. Compare H4458 (מֶלְתָּחָה)."},
"H4560":{"lemma":"מָסַר","xlit":"mâçar","pron":"maw-sar'","derivation":"a primitive root;","strongs_def":"to sunder, i.e. (transitively) set apart, or (reflex.) apostatize","kjv_def":"commit, deliver."},
"H4561":{"lemma":"מֹסָר","xlit":"môçâr","pron":"mo-sawr'","derivation":"from H3256 (יָסַר);","strongs_def":"admonition","kjv_def":"instruction."},
"H4562":{"lemma":"מָסֹרֶת","xlit":"mâçôreth","pron":"maw-so'-reth","derivation":"from H631 (אָסַר);","strongs_def":"a band","kjv_def":"bond."},
"H4563":{"lemma":"מִסְתּוֹר","xlit":"miçtôwr","pron":"mis-tore'","derivation":"from H5641 (סָתַר);","strongs_def":"a refuge","kjv_def":"covert."},
"H4564":{"lemma":"מַסְתֵּר","xlit":"maçtêr","pron":"mas-tare'","derivation":"from H5641 (סָתַר);","strongs_def":"properly, a hider, i.e. (abstractly) a hiding, i.e. aversion","kjv_def":"hid."},
"H4565":{"lemma":"מִסְתָּר","xlit":"miçtâr","pron":"mis-tawr'","derivation":"from H5641 (סָתַר);","strongs_def":"properly, a concealer, i.e. a covert","kjv_def":"secret(-ly, place)."},
"H4566":{"lemma":"מַעְבָּד","xlit":"maʻbâd","pron":"mah-bawd'","derivation":"from H5647 (עָבַד);","strongs_def":"an act","kjv_def":"work."},
"H4567":{"lemma":"מַעְבָד","xlit":"maʻbâd","pron":"mah-bawd'","derivation":"(Aramaic) corresponding to H4566 (מַעְבָּד);","strongs_def":"an act","kjv_def":"work."},
"H4568":{"lemma":"מַעֲבֶה","xlit":"maʻăbeh","pron":"mah-ab-eh'","derivation":"from H5666 (עָבָה);","strongs_def":"properly, compact (part of soil), i.e. loam","kjv_def":"clay."},
"H4569":{"lemma":"מַעֲבָר","xlit":"maʻăbâr","pron":"mah-ab-awr'","derivation":"or feminine מַעֲבָרָה; from H5674 (עָבַר);","strongs_def":"a crossing-place (of a river, a ford; of a mountain, a pass); abstractly, a transit, i.e. (figuratively) overwhelming","kjv_def":"ford, place where...pass, passage."},
"H4570":{"lemma":"מַעְגָּל","xlit":"maʻgâl","pron":"mah-gawl'","derivation":"or feminine מַעְגָּלָה; from the same as H5696 (עָגֹל);","strongs_def":"a track (literally or figuratively); also a rampart (as circular)","kjv_def":"going, path, trench, way(-side)."},
"H4571":{"lemma":"מָעַד","xlit":"mâʻad","pron":"maw-ad'","derivation":"a primitive root;","strongs_def":"to waver","kjv_def":"make to shake, slide, slip."},
"H4572":{"lemma":"מַעֲדַי","xlit":"Maʻăday","pron":"mah-ad-ah'-ee","derivation":"from H5710 (עָדָה); ornamental;","strongs_def":"Maadai, an Israelite","kjv_def":"Maadai."},
"H4573":{"lemma":"מַעֲדְיָה","xlit":"Maʻădyâh","pron":"mah-ad-yaw'","derivation":"from H5710 (עָדָה) and H3050 (יָהּ); ornament of Jah;","strongs_def":"Maadjah, an Israelite","kjv_def":"Maadiah. Compare H4153 (מוֹעַדְיָה)."},
"H4574":{"lemma":"מַעֲדָן","xlit":"maʻădân","pron":"mah-ad-awn'","derivation":"or (feminine) מַעֲדַנָּה; from H5727 (עָדַן);","strongs_def":"a delicacy or (abstractly) pleasure (adverbially, cheerfully)","kjv_def":"dainty, delicately, delight."},
"H4575":{"lemma":"מַעֲדַנָּה","xlit":"maʻădannâh","pron":"mah-ad-an-naw'","derivation":"by transitive from H6029 (עָנַד);","strongs_def":"a bond, i.e. group","kjv_def":"influence."},
"H4576":{"lemma":"מַעְדֵּר","xlit":"maʻdêr","pron":"mah-dare'","derivation":"from H5737 (עֲדַר);","strongs_def":"a (weeding) hoe","kjv_def":"mattock."},
"H4577":{"lemma":"מְעָה","xlit":"mᵉʻâh","pron":"meh-aw'","derivation":"(Aramaic) or מְעָא; (Aramaic), corresponding to H4578 (מֵעֶה)xlit mêʻâh corrected to mêʻeh;","strongs_def":"only in plural the bowels","kjv_def":"belly."},
"H4578":{"lemma":"מֵעֶה","xlit":"mêʻeh","pron":"may-aw'","derivation":"from an unused root probably meaning to be soft;","strongs_def":"used only in plural the intestines, or (collectively) the abdomen, figuratively, sympathy; by implication, a vest; by extension the stomach, the uterus (or of men, the seat of generation), the heart (figuratively)","kjv_def":"belly, bowels, [idiom] heart, womb."},
"H4579":{"lemma":"מֵעָה","xlit":"mêʻâh","pron":"may-aw'","derivation":"feminine of H4578 (מֵעֶה)xlit mêʻâh corrected to mêʻeh;","strongs_def":"the belly, i.e. (figuratively) interior","kjv_def":"gravel."},
"H4580":{"lemma":"מָעוֹג","xlit":"mâʻôwg","pron":"maw-ogue'","derivation":"from H5746 (עוּג); also with H3934 (לָעֵג) a table-buffoon, i.e.","strongs_def":"a cake of bread; parasite","kjv_def":"cake, feast."},
"H4581":{"lemma":"מָעוֹז","xlit":"mâʻôwz","pron":"maw-oze'","derivation":"(also מָעוּז ); or מָעֹז (also מָעֻז; from H5810 (עָזַז);","strongs_def":"a fortified place; figuratively, a defence","kjv_def":"force, fort(-ress), rock, strength(-en), ([idiom] most) strong (hold)."},
"H4582":{"lemma":"מָעוֹךְ","xlit":"Mâʻôwk","pron":"maw-oke'","derivation":"from H4600 (מָעַךְ); oppressed;","strongs_def":"Maok, a Philistine","kjv_def":"Maoch."},
"H4583":{"lemma":"מָעוֹן","xlit":"mâʻôwn","pron":"maw-ohn'","derivation":"or מָעִין; (1 Chronicles 4:41), from the same as H5772 (עוֹנָה);","strongs_def":"an abode, of God (the Tabernacle or the Temple), men (their home) or animals (their lair); hence, a retreat (asylum)","kjv_def":"den, dwelling((-) place), habitation."},
"H4584":{"lemma":"מָעוֹן","xlit":"Mâʻôwn","pron":"maw-ohn'","derivation":"the same as H4583 (מָעוֹן); a residence;","strongs_def":"Maon, the name of an Israelite and of a place in Palestine","kjv_def":"Maon, Maonites. Compare H1010 (בֵּית בַּעַל מְעוֹן), H4586 (מְעוּנִי)lemma מְעוּנָי third vowel, corrected to מְעוּנִי."},
"H4585":{"lemma":"מְעוֹנָה","xlit":"mᵉʻôwnâh","pron":"meh-o-naw'","derivation":"or מְעֹנָה; feminine of H4583 (מָעוֹן), and meaning the same","strongs_def":"{an abode, of God (the Tabernacle or the Temple), men (their home) or animals (their lair); hence, a retreat (asylum)}","kjv_def":"den, habitation, (dwelling) place, refuge."},
"H4586":{"lemma":"מְעוּנִי","xlit":"Mᵉʻûwnîy","pron":"meh-oo-nee'","derivation":"or מְעִינִי; probably patrial from H4584 (מָעוֹן);","strongs_def":"a Meunite, or inhabitant of Maon (only in plural)","kjv_def":"Mehunim(-s), Meunim."},
"H4587":{"lemma":"מְעוֹנֹתַי","xlit":"Mᵉʻôwnôthay","pron":"meh-o-no-thah'-ee","derivation":"plural of H4585 (מְעוֹנָה); habitative;","strongs_def":"Meonothai, an Israelite","kjv_def":"Meonothai."},
"H4588":{"lemma":"מָעוּף","xlit":"mâʻûwph","pron":"maw-off'","derivation":"from H5774 (עוּף) in the sense of covering with shade (compare H4155 (מוּעָף));","strongs_def":"darkness","kjv_def":"dimness."},
"H4589":{"lemma":"מָעוֹר","xlit":"mâʻôwr","pron":"maw-ore'","derivation":"from H5783 (עוּר);","strongs_def":"nakedness, i.e. (in plural) the pudenda","kjv_def":"nakedness."},
"H4590":{"lemma":"מַעַזְיָה","xlit":"Maʻazyâh","pron":"mah-az-yaw'","derivation":"or מַעַזְיָהוּ; probably from H5756 (עוּז) (in the sense of protection) and H3050 (יָהּ); rescue of Jah;","strongs_def":"Maazjah, the name of two Israelites","kjv_def":"Maaziah."},
"H4591":{"lemma":"מָעַט","xlit":"mâʻaṭ","pron":"maw-at'","derivation":"a primitive root;","strongs_def":"properly, to pare off, i.e. lessen; intransitively, to be (or causatively, to make) small or few (or figuratively, ineffective)","kjv_def":"suffer to decrease, diminish, (be, [idiom] borrow a, give, make) few (in number, -ness), gather least (little), be (seem) little, ([idiom] give the) less, be minished, bring to nothing."},
"H4592":{"lemma":"מְעַט","xlit":"mᵉʻaṭ","pron":"meh-at'","derivation":"or מְעָט; from H4591 (מָעַט);","strongs_def":"a little or few (often adverbial or compar.)","kjv_def":"almost (some, very) few(-er, -est), lightly, little (while), (very) small (matter, thing), some, soon, [idiom] very."},
"H4593":{"lemma":"מָעֹט","xlit":"mâʻôṭ","pron":"maw-ote'","derivation":"passive adjective of H4591 (מָעַט);","strongs_def":"thinned (as to the edge), i.e. sharp","kjv_def":"wrapped up."},
"H4594":{"lemma":"מַעֲטֶה","xlit":"maʻăṭeh","pron":"mah-at-eh'","derivation":"from H5844 (עָטָה);","strongs_def":"a vestment","kjv_def":"garment."},
"H4595":{"lemma":"מַעֲטָפָה","xlit":"maʻăṭâphâh","pron":"mah-at-aw-faw'","derivation":"from H5848 (עָטַף);","strongs_def":"a cloak","kjv_def":"mantle."},
"H4596":{"lemma":"מְעִי","xlit":"mᵉʻîy","pron":"meh-ee'","derivation":"from H5753 (עָוָה); (compare H5856 (עִי))","strongs_def":"a pile of rubbish (as contorted), i.e. a ruin","kjv_def":"heap."},
"H4597":{"lemma":"מָעַי","xlit":"Mâʻay","pron":"maw-ah'-ee","derivation":"probably from H4578 (מֵעֶה)xlit mêʻâh corrected to mêʻeh; sympathetic;","strongs_def":"Maai, an Israelite","kjv_def":"Maai."},
"H4598":{"lemma":"מְעִיל","xlit":"mᵉʻîyl","pron":"meh-eel'","derivation":"from H4603 (מָעַל) in the sense of covering;","strongs_def":"a robe (i.e. upper and outer garment)","kjv_def":"cloke, coat, mantle, robe."},
"H4599":{"lemma":"מַעְיָן","xlit":"maʻyân","pron":"mah-yawn'","derivation":"or מַעְיְנוֹ; (Psalm 114:8), or (feminine) מַעְיָנָה; from H5869 (עַיִן) (as a denominative in the sense of a spring);","strongs_def":"a fountain (also collectively), figuratively, a source (of satisfaction)","kjv_def":"fountain, spring, well."},
"H4600":{"lemma":"מָעַךְ","xlit":"mâʻak","pron":"maw-ak'","derivation":"a primitive root;","strongs_def":"to press, i.e. to pierce, emasculate, handle","kjv_def":"bruised, stuck, be pressed."},
"H4601":{"lemma":"מַעֲכָה","xlit":"Maʻăkâh","pron":"mah-ak-aw'","derivation":"or מַעֲכָת; (Joshua 13:13), from H4600 (מָעַךְ); depression;","strongs_def":"Maakah (or Maakath), the name of a place in Syria, also of a Mesopotamian, of three Israelites, and of four Israelitesses and one Syrian woman","kjv_def":"Maachah, Maachathites. See also H1038 (בֵּית מַעֲכָה)."},
"H4602":{"lemma":"מַעֲכָתִי","xlit":"Maʻăkâthîy","pron":"mah-ak-aw-thee'","derivation":"patrial from H4601 (מַעֲכָה);","strongs_def":"a Maakathite, or inhabitant of Maakah","kjv_def":"Maachathite."},
"H4603":{"lemma":"מָעַל","xlit":"mâʻal","pron":"maw-al'","derivation":"a primitive root;","strongs_def":"properly, to cover up; used only figuratively, to act covertly, i.e. treacherously","kjv_def":"transgress, (commit, do a) trespass(-ing)."},
"H4604":{"lemma":"מַעַל","xlit":"maʻal","pron":"mah'-al","derivation":"from H4603 (מָעַל);","strongs_def":"treachery, i.e. sin","kjv_def":"falsehood, grievously, sore, transgression, trespass, [idiom] very."},
"H4605":{"lemma":"מַעַל","xlit":"maʻal","pron":"mah'al","derivation":"from H5927 (עָלָה);","strongs_def":"properly, the upper part, used only adverbially with prefix upward, above, overhead, from the top, etc.","kjv_def":"above, exceeding(-ly), forward, on ([idiom] very) high, over, up(-on, -ward), very."},
"H4606":{"lemma":"מֵעָל","xlit":"mêʻâl","pron":"may-awl'","derivation":"(Aramaic) from H5954 (עֲלַל);","strongs_def":"(only in plural as singular) the setting (of the sun)","kjv_def":"going down."},
"H4607":{"lemma":"מֹעַל","xlit":"môʻal","pron":"mo'-al","derivation":"from H5927 (עָלָה);","strongs_def":"a raising (of the hands)","kjv_def":"lifting up."},
"H4608":{"lemma":"מַעֲלֶה","xlit":"maʻăleh","pron":"mah-al-eh'","derivation":"from H5927 (עָלָה);","strongs_def":"an elevation, i.e. (concretely) acclivity or platform; abstractly (the relation or state) a rise or (figuratively) priority","kjv_def":"ascent, before, chiefest, cliff, that goeth up, going up, hill, mounting up, stairs."},
"H4609":{"lemma":"מַעֲלָה","xlit":"maʻălâh","pron":"mah-al-aw'","derivation":"feminine of H4608 (מַעֲלֶה);","strongs_def":"elevation, i.e. the act (literally, a journey to a higher place, figuratively, a thought arising), or (concretely) the condition (literally, a step or grademark, figuratively, a superiority of station); specifically a climactic progression (in certain Psalms)","kjv_def":"things that come up, (high) degree, deal, go up, stair, step, story."},
"H4610":{"lemma":"מַעֲלֵה עַקְרַבִּים","xlit":"Maʻălêh ʻAqrabbîym","pron":"mah-al-ay' ak-rabbeem'","derivation":"from H4608 (מַעֲלֶה) and (the plural of) H6137 (עַקְרָב);","strongs_def":"Steep of Scorpions, a place in the Desert","kjv_def":"Maaleh-accrabim, the ascent (going up) of Akrabbim."},
"H4611":{"lemma":"מַעֲלָל","xlit":"maʻălâl","pron":"mah-al-awl'","derivation":"from H5953 (עָלַל);","strongs_def":"an act (good or bad)","kjv_def":"doing, endeavour, invention, work."},
"H4612":{"lemma":"מַעֲמָד","xlit":"maʻămâd","pron":"mah-am-awd'","derivation":"from H5975 (עָמַד);","strongs_def":"(figuratively) a position","kjv_def":"attendance, office, place, state."},
"H4613":{"lemma":"מׇעֳמָד","xlit":"moʻŏmâd","pron":"moh-om-awd'","derivation":"from H5975 (עָמַד);","strongs_def":"literally a foothold","kjv_def":"standing."},
"H4614":{"lemma":"מַעֲמָסָה","xlit":"maʻămâçâh","pron":"mah-am-aw-saw'","derivation":"from H6006 (עָמַס);","strongs_def":"burdensomeness","kjv_def":"burdensome."},
"H4615":{"lemma":"מַעֲמָק","xlit":"maʻămâq","pron":"mah-am-awk'","derivation":"from H6009 (עָמַק);","strongs_def":"a deep","kjv_def":"deep, depth."},
"H4616":{"lemma":"מַעַן","xlit":"maʻan","pron":"mah'-an","derivation":"from H6030 (עָנָה);","strongs_def":"properly, heed, i.e. purpose; used only adverbially, on account of (as a motive or an aim), teleologically, in order that","kjv_def":"because of, to the end (intent) that, for (to,... 's sake), [phrase] lest, that, to."},
"H4617":{"lemma":"מַעֲנֶה","xlit":"maʻăneh","pron":"mah-an-eh'","derivation":"from H6030 (עָנָה);","strongs_def":"a reply (favorable or contradictory)","kjv_def":"answer, [idiom] himself."},
"H4618":{"lemma":"מַעֲנָה","xlit":"maʻănâh","pron":"mah-an-aw'","derivation":"from H6031 (עָנָה), in the sense of depression or tilling;","strongs_def":"a furrow","kjv_def":"[phrase] acre, furrow."},
"H4619":{"lemma":"מַעַץ","xlit":"Maʻats","pron":"mah'-ats","derivation":"from H6095 (עָצָה); closure;","strongs_def":"Maats, an Israelite","kjv_def":"Maaz."},
"H4620":{"lemma":"מַעֲצֵבָה","xlit":"maʻătsêbâh","pron":"mah-ats-ay-baw'","derivation":"from H6087 (עָצַב);","strongs_def":"anguish","kjv_def":"sorrow."},
"H4621":{"lemma":"מַעֲצָד","xlit":"maʻătsâd","pron":"mah-ats-awd'","derivation":"from an unused root meaning to hew;","strongs_def":"an axe","kjv_def":"ax, tongs."},
"H4622":{"lemma":"מַעְצוֹר","xlit":"maʻtsôwr","pron":"mah-tsore'","derivation":"from H6113 (עָצָר);","strongs_def":"objectively, a hindrance","kjv_def":"restraint."},
"H4623":{"lemma":"מַעְצָר","xlit":"maʻtsâr","pron":"mah-tsawr'","derivation":"from H6113 (עָצָר);","strongs_def":"subjectively, control","kjv_def":"rule."},
"H4624":{"lemma":"מַעֲקֶה","xlit":"maʻăqeh","pron":"mah-ak-eh'","derivation":"from an unused root meaning to repress;","strongs_def":"a parapet","kjv_def":"battlement."},
"H4625":{"lemma":"מַעֲקָשׁ","xlit":"maʻăqâsh","pron":"mah-ak-awsh'","derivation":"from H6140 (עָקַשׁ);","strongs_def":"a crook (in a road)","kjv_def":"crooked thing."},
"H4626":{"lemma":"מַעַר","xlit":"maʻar","pron":"mah'-ar","derivation":"from H6168 (עָרָה);","strongs_def":"a nude place, i.e. (literally) the pudenda, or (figuratively) a vacant space","kjv_def":"nakedness, proportion."},
"H4627":{"lemma":"מַעֲרָב","xlit":"maʻărâb","pron":"mah-ar-awb'","derivation":"from H6148 (עָרַב), in the sense of trading;","strongs_def":"traffic; by implication, mercantile goods","kjv_def":"market, merchandise."},
"H4628":{"lemma":"מַעֲרָב","xlit":"maʻărâb","pron":"mah-ar-awb'","derivation":"or (feminine) מַעֲרָבָה; from H6150 (עָרַב), in the sense of shading;","strongs_def":"the west (as a region of the evening sun)","kjv_def":"west."},
"H4629":{"lemma":"מַעֲרֶה","xlit":"maʻăreh","pron":"mah-ar-eh'","derivation":"from H6168 (עָרָה);","strongs_def":"a nude place, i.e. a common","kjv_def":"meadows."},
"H4630":{"lemma":"מַעֲרָה","xlit":"maʻărâh","pron":"mah-ar-aw'","derivation":"feminine of H4629 (מַעֲרֶה);","strongs_def":"an open spot","kjv_def":"army (from the margin)."},
"H4631":{"lemma":"מְעָרָה","xlit":"mᵉʻârâh","pron":"meh-aw-raw'","derivation":"from H5783 (עוּר);","strongs_def":"a cavern (as dark)","kjv_def":"cave, den, hole."},
"H4632":{"lemma":"מְעָרָה","xlit":"Mᵉʻârâh","pron":"meh-aw-raw'","derivation":"the same as H4631 (מְעָרָה); cave;","strongs_def":"Mearah, a place in Palestine","kjv_def":"Mearah."},
"H4633":{"lemma":"מַעֲרָךְ","xlit":"maʻărâk","pron":"mah-ar-awk'","derivation":"from H6186 (עָרַךְ);","strongs_def":"an arrangement, i.e. (figuratively) mental disposition","kjv_def":"preparation."},
"H4634":{"lemma":"מַעֲרָכָה","xlit":"maʻărâkâh","pron":"mah-ar-aw-kaw'","derivation":"feminine of H4633 (מַעֲרָךְ);","strongs_def":"an arrangement; concretely, a pile; specifically a military array","kjv_def":"army, fight, be set in order, ordered place, rank, row."},
"H4635":{"lemma":"מַעֲרֶכֶת","xlit":"maʻăreketh","pron":"mah-ar-eh'-keth","derivation":"from H6186 (עָרַךְ);","strongs_def":"an arrangement, i.e. (concretely) a pile (of loaves)","kjv_def":"row, shewbread."},
"H4636":{"lemma":"מַעֲרֹם","xlit":"maʻărôm","pron":"mah-ar-ome'","derivation":"from H6191 (עָרַם), in the sense of stripping;","strongs_def":"bare","kjv_def":"naked."},
"H4637":{"lemma":"מַעֲרָצָה","xlit":"maʻărâtsâh","pron":"mah-ar-aw-tsaw'","derivation":"from H6206 (עָרַץ);","strongs_def":"violence","kjv_def":"terror."},
"H4638":{"lemma":"מַעֲרָת","xlit":"Maʻărâth","pron":"mah-ar-awth'","derivation":"a form of H4630 (מַעֲרָה); waste;","strongs_def":"Maarath, a place in Palestine","kjv_def":"Maarath."},
"H4639":{"lemma":"מַעֲשֶׂה","xlit":"maʻăseh","pron":"mah-as-eh'","derivation":"from H6213 (עָשָׂה);","strongs_def":"an action (good or bad); generally, a transaction; abstractly, activity; by implication, a product (specifically, a poem) or (generally) property","kjv_def":"act, art, [phrase] bakemeat, business, deed, do(-ing), labor, thing made, ware of making, occupation, thing offered, operation, possession, [idiom] well, (handy-, needle-, net-) work(ing, -manship), wrought."},
"H4640":{"lemma":"מַעֲשַׂי","xlit":"Maʻăsay","pron":"mah-as-ah'ee","derivation":"from H6213 (עָשָׂה); operative;","strongs_def":"Maasai, an Israelite","kjv_def":"Maasiai."},
"H4641":{"lemma":"מַעֲשֵׂיָה","xlit":"Maʻăsêyâh","pron":"mah-as-ay-yaw'","derivation":"or מַעֲשֵׂיָהוּ; from H4639 (מַעֲשֶׂה) and H3050 (יָהּ); work of Jah;","strongs_def":"Maasejah, the name of sixteen Israelites","kjv_def":"Maaseiah."},
"H4642":{"lemma":"מַעֲשַׁקָּה","xlit":"maʻăshaqqâh","pron":"mah-ash-ak-kaw'","derivation":"from H6231 (עָשַׁק);","strongs_def":"oppression","kjv_def":"oppression, [idiom] oppressor."},
"H4643":{"lemma":"מַעֲשֵׂר","xlit":"maʻăsêr","pron":"mah-as-ayr'","derivation":"or מַעֲשַׂר; and (in plural) feminine מַעַשְׂרָה; from H6240 (עָשָׂר);","strongs_def":"a tenth; especially a tithe","kjv_def":"tenth (part), tithe(-ing)."},
"H4644":{"lemma":"מֹף","xlit":"Môph","pron":"mofe","derivation":"of Egyptian origin;","strongs_def":"Moph, the capital of Lower Egypt","kjv_def":"Memphis. Compare H5297 (נֹף)."},
"H4645":{"lemma":"מִפְגָּע","xlit":"miphgâʻ","pron":"mif-gaw'","derivation":"from H6293 (פָּגַע);","strongs_def":"an object of attack","kjv_def":"mark."},
"H4646":{"lemma":"מַפָּח","xlit":"mappâch","pron":"map-pawkh'","derivation":"from H5301 (נָפַח);","strongs_def":"a breathing out (of life), i.e. expiring","kjv_def":"giving up."},
"H4647":{"lemma":"מַפֻּחַ","xlit":"mappuach","pron":"map-poo'-akh","derivation":"from H5301 (נָפַח);","strongs_def":"the bellows (i.e. blower) of aforge","kjv_def":"bellows."},
"H4648":{"lemma":"מְפִיבֹשֶׁת","xlit":"Mᵉphîybôsheth","pron":"mef-ee-bo'-sheth","derivation":"or מְפִבֹשֶׁת; probably from H6284 (פָּאָה) and H1322 (בֹּשֶׁת); dispeller of shame (i.e. of Baal);","strongs_def":"Mephibosheth, the name of two Israelites","kjv_def":"Mephibosheth."},
"H4649":{"lemma":"מֻפִּים","xlit":"Muppîym","pron":"moop-peem'","derivation":"a plural apparently from H5130 (נוּף); wavings;","strongs_def":"Muppim, an Israelite","kjv_def":"Muppim. Compare H8206 (שֻׁפִּים)."},
"H4650":{"lemma":"מֵפִיץ","xlit":"mêphîyts","pron":"may-feets'","derivation":"from H6327 (פּוּץ);","strongs_def":"a breaker, i.e. mallet","kjv_def":"maul."},
"H4651":{"lemma":"מַפָּל","xlit":"mappâl","pron":"map-pawl'","derivation":"from H5307 (נָפַל);","strongs_def":"a falling off, i.e. chaff; also something pendulous, i.e. a flap","kjv_def":"flake, refuse."},
"H4652":{"lemma":"מִפְלָאָה","xlit":"miphlâʼâh","pron":"mif-law-aw'","derivation":"from H6381 (פָּלָא);","strongs_def":"a miracle","kjv_def":"wondrous work."},
"H4653":{"lemma":"מִפְלַגָּה","xlit":"miphlaggâh","pron":"mif-lag-gaw'","derivation":"from H6385 (פָּלַג);","strongs_def":"a classification","kjv_def":"division."},
"H4654":{"lemma":"מַפָּלָה","xlit":"mappâlâh","pron":"map-paw-law'","derivation":"or מַפֵּלָה; from H5307 (נָפַל);","strongs_def":"something fallen, i.e. a ruin","kjv_def":"ruin(-ous)."},
"H4655":{"lemma":"מִפְלָט","xlit":"miphlâṭ","pron":"mif-lawt'","derivation":"from H6403 (פָּלַט);","strongs_def":"an escape","kjv_def":"escape."},
"H4656":{"lemma":"מִפְלֶצֶת","xlit":"miphletseth","pron":"mif-leh'-tseth","derivation":"from H6426 (פָּלַץ);","strongs_def":"a terror, i.e. an idol","kjv_def":"idol."},
"H4657":{"lemma":"מִפְלָשׂ","xlit":"miphlâs","pron":"mif-lawce'","derivation":"from an unused root meaning to balance;","strongs_def":"a poising","kjv_def":"balancing."},
"H4658":{"lemma":"מַפֶּלֶת","xlit":"mappeleth","pron":"map-peh'-leth","derivation":"from H5307 (נָפַל);","strongs_def":"fall, i.e. decadence; concretely, a ruin; specifically a carcase","kjv_def":"carcase, fall, ruin."},
"H4659":{"lemma":"מִפְעָל","xlit":"miphʻâl","pron":"mif-awl'","derivation":"or (feminine) מִפְעָלָה; from H6466 (פָּעַל);","strongs_def":"a performance","kjv_def":"work."},
"H4660":{"lemma":"מַפָּץ","xlit":"mappâts","pron":"map-pawts'","derivation":"from H5310 (נָפַץ);","strongs_def":"a smiting to pieces","kjv_def":"slaughter."},
"H4661":{"lemma":"מַפֵּץ","xlit":"mappêts","pron":"map-pates'","derivation":"from H5310 (נָפַץ);","strongs_def":"a smiter, i.e. a war club","kjv_def":"battle ax."},
"H4662":{"lemma":"מִפְקָד","xlit":"miphqâd","pron":"mif-kawd'","derivation":"from H6485 (פָּקַד);","strongs_def":"an appointment, i.e. mandate; concretely, a designated spot; specifically, a census","kjv_def":"appointed place, commandment, number."},
"H4663":{"lemma":"מִפְקָד","xlit":"Miphqâd","pron":"mif-kawd'","derivation":"the same as H4662 (מִפְקָד); assignment;","strongs_def":"Miphkad, the name of a gate in Jerusalem","kjv_def":"Miphkad."},
"H4664":{"lemma":"מִפְרָץ","xlit":"miphrâts","pron":"mif-rawts'","derivation":"from H6555 (פָּרַץ)lemma פּרַץ missing vowel, corrected to פָּרַץ;","strongs_def":"a break (in the shore), i.e. a haven","kjv_def":"breach."},
"H4665":{"lemma":"מִפְרֶקֶת","xlit":"miphreqeth","pron":"mif-reh'-keth","derivation":"from H6561 (פָּרַק);","strongs_def":"properly, a fracture, i.e. joint (vertebrae) of the neck","kjv_def":"neck."},
"H4666":{"lemma":"מִפְרָשׂ","xlit":"miphrâs","pron":"mif-rawce'","derivation":"from H6566 (פָּרַשׂ);","strongs_def":"an expansion","kjv_def":"that which...spreadest forth, spreading."},
"H4667":{"lemma":"מִפְשָׂעָה","xlit":"miphsâʻâh","pron":"mif-saw-aw'","derivation":"from H6585 (פָּשַׂע);","strongs_def":"a stride, i.e. (by euphemism) the crotch","kjv_def":"buttocks."},
"H4668":{"lemma":"מַפְתֵּחַ","xlit":"maphtêach","pron":"maf-tay'-akh","derivation":"from H6605 (פָּתַח);","strongs_def":"an opener, i.e. a key","kjv_def":"key."},
"H4669":{"lemma":"מִפְתָּח","xlit":"miphtâch","pron":"mif-tawkh'","derivation":"from H6605 (פָּתַח);","strongs_def":"an aperture, i.e. (figuratively) utterance","kjv_def":"opening."},
"H4670":{"lemma":"מִפְתָּן","xlit":"miphtân","pron":"mif-tawn'","derivation":"from the same as H6620 (פֶּתֶן);","strongs_def":"a stretcher, i.e. a sill","kjv_def":"threshold."},
"H4671":{"lemma":"מֹץ","xlit":"môts","pron":"motes","derivation":"or מוֹץ; (Zephaniah 2:2), from H4160 (מוּץ);","strongs_def":"chaff (as pressed out, i.e. winnowed or (rather) threshed loose)","kjv_def":"chaff."},
"H4672":{"lemma":"מָצָא","xlit":"mâtsâʼ","pron":"maw-tsaw'","derivation":"a primitive root;","strongs_def":"properly, to come forth to, i.e. appear or exist; transitively, to attain, i.e. find or acquire; figuratively, to occur, meet or be present","kjv_def":"[phrase] be able, befall, being, catch, [idiom] certainly, (cause to) come (on, to, to hand), deliver, be enough (cause to) find(-ing, occasion, out), get (hold upon), [idiom] have (here), be here, hit, be left, light (up-) on, meet (with), [idiom] occasion serve, (be) present, ready, speed, suffice, take hold on."},
"H4673":{"lemma":"מַצָּב","xlit":"matstsâb","pron":"mats-tsawb'","derivation":"from H5324 (נָצַב);","strongs_def":"a fixed spot; figuratively, an office, a military post","kjv_def":"garrison, station, place where...stood."},
"H4674":{"lemma":"מֻצָּב","xlit":"mutstsâb","pron":"moots-tsawb'","derivation":"from H5324 (נָצַב);","strongs_def":"a station, i.e. military post","kjv_def":"mount."},
"H4675":{"lemma":"מַצָּבָה","xlit":"matstsâbâh","pron":"mats-tsaw-baw'","derivation":"or מִצָּבָה; feminine of H4673 (מַצָּב);","strongs_def":"a military guard","kjv_def":"army, garrison."},
"H4676":{"lemma":"מַצֵּבָה","xlit":"matstsêbâh","pron":"mats-tsay-baw'","derivation":"feminine (causatively) participle of H5324 (נָצַב);","strongs_def":"something stationed, i.e. a column or (memorial stone); by analogy, an idol","kjv_def":"garrison, (standing) image, pillar."},
"H4677":{"lemma":"מְצֹבָיָה","xlit":"Mᵉtsôbâyâh","pron":"mets-o-baw-yaw'","derivation":"apparently from H4672 (מָצָא) and H3050 (יָהּ); found of Jah;","strongs_def":"Metsobajah, a place in Palestine","kjv_def":"Mesobaite."},
"H4678":{"lemma":"מַצֶּבֶת","xlit":"matstsebeth","pron":"mats-tseh'-beth","derivation":"from H5324 (נָצַב);","strongs_def":"something stationary, i.e. a monumental stone; also the stock of atree","kjv_def":"pillar, substance."},
"H4679":{"lemma":"מְצַד","xlit":"mᵉtsad","pron":"mets-ad'","derivation":"or מְצָד; or (feminine) מְצָדָה; from H6679 (צוּד);","strongs_def":"a fastness (as a covert of ambush)","kjv_def":"castle, fort, (strong) hold, munition."},
"H4680":{"lemma":"מָצָה","xlit":"mâtsâh","pron":"maw-tsaw'","derivation":"a primitive root;","strongs_def":"to suck out; by implication, to drain, to squeeze out","kjv_def":"suck, wring (out)."},
"H4681":{"lemma":"מֹצָה","xlit":"Môtsâh","pron":"mo-tsaw'","derivation":"active participle feminine of H4680 (מָצָה); drained;","strongs_def":"Motsah, a place in Palestine","kjv_def":"Mozah."},
"H4682":{"lemma":"מַצָּה","xlit":"matstsâh","pron":"mats-tsaw'","derivation":"from H4711 (מָצַץ) in the sense of greedily devouring for sweetness;","strongs_def":"properly, sweetness; concretely, sweet (i.e. not soured or bittered with yeast); specifically, an unfermented cake or loaf, or (elliptically) the festival of Passover (because no leaven was then used)","kjv_def":"unleaved (bread, cake), without leaven."},
"H4683":{"lemma":"מַצָּה","xlit":"matstsâh","pron":"mats-tsaw'","derivation":"from H5327 (נָצָה);","strongs_def":"a quarrel","kjv_def":"contention, debate, strife."},
"H4684":{"lemma":"מַצְהָלָה","xlit":"matshâlâh","pron":"mats-haw-law'","derivation":"from H6670 (צָהַל);","strongs_def":"a whinnying (through impatience for battle or lust)","kjv_def":"neighing."},
"H4685":{"lemma":"מָצוֹד","xlit":"mâtsôwd","pron":"maw-tsode'","derivation":"or (feminine) מְצוֹדָה; or מצֹדָה; from H6679 (צוּד); also (by interchange for H4679 (מְצַד))","strongs_def":"a net (for capturing animals or fishes);  afastness or (beseiging) tower","kjv_def":"bulwark, hold, munition, net, snare."},
"H4686":{"lemma":"מָצוּד","xlit":"mâtsûwd","pron":"maw-tsood'","derivation":"or (feminine) מְצוּדָה; or מְצֻדָה; for H4685 (מָצוֹד);","strongs_def":"a net, or (abstractly) capture; also a fastness","kjv_def":"castle, defense, fort(-ress), (strong) hold, be hunted, net, snare, strong place."},
"H4687":{"lemma":"מִצְוָה","xlit":"mitsvâh","pron":"mits-vaw'","derivation":"from H6680 (צָוָה);","strongs_def":"a command, whether human or divine (collectively, the Law)","kjv_def":"(which was) commanded(-ment), law, ordinance, precept."},
"H4688":{"lemma":"מְצוֹלָה","xlit":"mᵉtsôwlâh","pron":"mets-o-law'","derivation":"or מְצֹלָה; also מְצוּלָה; or מְצֻלָה; from the same as H6683 (צוּלָה);","strongs_def":"a deep place (of water or mud)","kjv_def":"bottom, deep, depth."},
"H4689":{"lemma":"מָצוֹק","xlit":"mâtsôwq","pron":"maw-tsoke'","derivation":"from H6693 (צוּק);","strongs_def":"a narrow place, i.e. (abstractly and figuratively) confinement or disability","kjv_def":"anguish, distress, straitness."},
"H4690":{"lemma":"מָצוּק","xlit":"mâtsûwq","pron":"maw-tsook'","derivation":"or מָצֻק; from H6693 (צוּק);","strongs_def":"something narrow, i.e. a column or hilltop","kjv_def":"pillar, situate."},
"H4691":{"lemma":"מְצוּקָה","xlit":"mᵉtsûwqâh","pron":"mets-oo-kaw'","derivation":"or מְצֻקָה; feminine of H4690 (מָצוּק);","strongs_def":"narrowness, i.e. (figuratively) trouble","kjv_def":"anguish, distress."},
"H4692":{"lemma":"מָצוֹר","xlit":"mâtsôwr","pron":"maw-tsore'","derivation":"or מָצוּר; from H6696 (צוּר);","strongs_def":"something hemming in, i.e. (objectively) a mound (of besiegers), (abstractly) a siege, (figuratively) distress; or (subjectively) a fastness","kjv_def":"besieged, bulwark, defence, fenced, fortress, siege, strong (hold), tower."},
"H4693":{"lemma":"מָצוֹר","xlit":"mâtsôwr","pron":"maw-tsore'","derivation":"the same as H4692 (מָצוֹר) in the sense of a limit;","strongs_def":"Egypt (as the border of Palestine)","kjv_def":"besieged places, defense, fortified."},
"H4694":{"lemma":"מְצוּרָה","xlit":"mᵉtsûwrâh","pron":"mets-oo-raw'","derivation":"or מְצֻרָה; feminine of H4692 (מָצוֹר);","strongs_def":"a hemming in, i.e. (objectively) a mound (of siege), or (subjectively) a rampart (of protection), (abstractly) fortification","kjv_def":"fenced (city, fort, munition, strong hold."},
"H4695":{"lemma":"מַצּוּת","xlit":"matstsûwth","pron":"mats-tsooth'","derivation":"from H5327 (נָצָה);","strongs_def":"a quarrel","kjv_def":"that contended."},
"H4696":{"lemma":"מֵצַח","xlit":"mêtsach","pron":"may'-tsakh","derivation":"from an unused root meaning to be clear, i.e. conspicuous;","strongs_def":"the forehead (as open and prominent)","kjv_def":"brow, forehead, [phrase] impudent."},
"H4697":{"lemma":"מִצְחָה","xlit":"mitschâh","pron":"mits-khaw'","derivation":"from the same as H4696 (מֵצַח);","strongs_def":"a shinpiece of armor (as prominent), only plural","kjv_def":"greaves."},
"H4698":{"lemma":"מְצִלָּה","xlit":"mᵉtsillâh","pron":"mets-il-law'","derivation":"from H6750 (צָלַל);","strongs_def":"a tinkler, i.e. a bell","kjv_def":"bell."},
"H4699":{"lemma":"מְצֻלָּה","xlit":"mᵉtsullâh","pron":"mets-ool-law'","derivation":"from H6751 (צָלַל);","strongs_def":"shade","kjv_def":"bottom."},
"H4700":{"lemma":"מְצֵלֶת","xlit":"mᵉtsêleth","pron":"mets-ay'-leth","derivation":"from H6750 (צָלַל);","strongs_def":"(only dual) double tinklers, i.e. cymbals","kjv_def":"cymbals."},
"H4701":{"lemma":"מִצְנֶפֶת","xlit":"mitsnepheth","pron":"mits-neh'-feth","derivation":"from H6801 (צָנַף);","strongs_def":"a tiara, i.e. official turban (of a king or high priest)","kjv_def":"diadem, mitre."},
"H4702":{"lemma":"מַצָּע","xlit":"matstsâʻ","pron":"mats-tsaw'","derivation":"from H3331 (יַצַע);","strongs_def":"a couch","kjv_def":"bed."},
"H4703":{"lemma":"מִצְעָד","xlit":"mitsʻâd","pron":"mits-awd'","derivation":"from H6805 (צָעַד);","strongs_def":"a step; figuratively, companionship","kjv_def":"going, step."},
"H4704":{"lemma":"מִצְּעִירָה","xlit":"mitstsᵉʻîyrâh","pron":"mits-tseh-ee-raw'","derivation":"feminine of H4705 (מִצְעָר);","strongs_def":"properly, littleness; concretely, diminutive","kjv_def":"little."},
"H4705":{"lemma":"מִצְעָר","xlit":"mitsʻâr","pron":"mits-awr'","derivation":"from H6819 (צָעַר);","strongs_def":"petty (in size or number); adverbially, a short (time)","kjv_def":"little one (while), small."},
"H4706":{"lemma":"מִצְעָר","xlit":"Mitsʻâr","pron":"mits-awr'","derivation":"the same as H4705 (מִצְעָר);","strongs_def":"Mitsar, a peak of Lebanon","kjv_def":"Mizar."},
"H4707":{"lemma":"מִצְפֶּה","xlit":"mitspeh","pron":"mits-peh'","derivation":"from H6822 (צָפָה);","strongs_def":"an observatory, especially formilitary purposes","kjv_def":"watch tower."},
"H4708":{"lemma":"מִצְפֶּה","xlit":"Mitspeh","pron":"mits-peh'","derivation":"the same as H4707 (מִצְפֶּה);","strongs_def":"Mitspeh, the name of five places in Palestine","kjv_def":"Mizpeh, watch tower. Compare H4709 (מִצְפָּה)."},
"H4709":{"lemma":"מִצְפָּה","xlit":"Mitspâh","pron":"mits-paw'","derivation":"feminine of H4708 (מִצְפֶּה);","strongs_def":"Mitspah, the name of two places in Palestine","kjv_def":"Mitspah. (This seems rather to be only an orthographic variation of H4708 (מִצְפֶּה) when 'in pause'.)"},
"H4710":{"lemma":"מִצְפֻּן","xlit":"mitspun","pron":"mits-poon'","derivation":"from H6845 (צָפַן);","strongs_def":"a secret (place or thing, perhaps, treasure)","kjv_def":"hidden thing."},
"H4711":{"lemma":"מָצַץ","xlit":"mâtsats","pron":"maw-tsats'","derivation":"a primitive root;","strongs_def":"to suck","kjv_def":"milk."},
"H4712":{"lemma":"מֵצַר","xlit":"mêtsar","pron":"may-tsar'","derivation":"from H6896 (קֵבָה);","strongs_def":"something tight, i.e. (figuratively) trouble","kjv_def":"distress, pain, strait."},
"H4713":{"lemma":"מִצְרִי","xlit":"Mitsrîy","pron":"mits-ree'","derivation":"from H4714 (מִצְרַיִם);","strongs_def":"a Mitsrite, or inhabitant of Mitsrajim","kjv_def":"Egyptian, of Egypt."},
"H4714":{"lemma":"מִצְרַיִם","xlit":"Mitsrayim","pron":"mits-rah'-yim","derivation":"dual of H4693 (מָצוֹר);","strongs_def":"Mitsrajim, i.e. Upper and Lower Egypt","kjv_def":"Egypt, Egyptians, Mizraim."},
"H4715":{"lemma":"מִצְרֵף","xlit":"mitsrêph","pron":"mits-rafe'","derivation":"from H6884 (צָרַף);","strongs_def":"a crucible","kjv_def":"fining pot."},
"H4716":{"lemma":"מַק","xlit":"maq","pron":"mak","derivation":"from H4743 (מָקַק);","strongs_def":"properly, a melting, i.e. putridity","kjv_def":"rottenness, stink."},
"H4717":{"lemma":"מַקָּבָה","xlit":"maqqâbâh","pron":"mak-kaw-baw'","derivation":"from H5344 (נָקַב);","strongs_def":"properly, a perforatrix, i.e. a hammer (as piercing)","kjv_def":"hammer."},
"H4718":{"lemma":"מַקֶּבֶת","xlit":"maqqebeth","pron":"mak-keh'-beth","derivation":"from H5344 (נָקַב);","strongs_def":"properly, a perforator, i.e. a hammer (as piercing); also (intransitively) a perforation, i.e. a quarry","kjv_def":"hammer, hole."},
"H4719":{"lemma":"מַקֵּדָה","xlit":"Maqqêdâh","pron":"mak-kay-daw'","derivation":"from the same as H5348 (נָקֹד) in the denominative sense of herding (compare H5349 (נֹקֵד)); fold;","strongs_def":"Makkedah, a place in Palestine","kjv_def":"Makkedah."},
"H4720":{"lemma":"מִקְדָּשׁ","xlit":"miqdâsh","pron":"mik-dawsh'","derivation":"or מִקְּדָשׁ; (Exodus 15:17), from H6942 (קָדַשׁ);","strongs_def":"a consecrated thing or place, especially, a palace, sanctuary (whether of Jehovah or of idols) or asylum","kjv_def":"chapel, hallowed part, holy place, sanctuary."},
"H4721":{"lemma":"מַקְהֵל","xlit":"maqhêl","pron":"mak-hale'","derivation":"or (feminine) מַקְהֵלָה; from H6950 (קָהַל);","strongs_def":"an assembly","kjv_def":"congregation."},
"H4722":{"lemma":"מַקְהֵלֹת","xlit":"Maqhêlôth","pron":"mak-hay-loth'","derivation":"plural of H4721 (מַקְהֵל) (feminine); assemblies;","strongs_def":"Makheloth, a place in the Desert","kjv_def":"Makheloth."},
"H4723":{"lemma":"מִקְוֶה","xlit":"miqveh","pron":"mik-veh'","derivation":"or מִקְוֵה; (1 Kings 10:28), or מִקְוֵאlemma מִּקְוֵא extra dagesh, corrected to מִקְוֵא; (2 Chronicles 1:16), from H6960 (קָוָה);","strongs_def":"something waited for, i.e. confidence (objective or subjective); also a collection, i.e. (of water) a pond, or (of men and horses) a caravan or drove","kjv_def":"abiding, gathering together, hope, linen yarn, plenty (of water), pool."},
"H4724":{"lemma":"מִקְוָה","xlit":"miqvâh","pron":"mik-vaw'","derivation":"feminine of H4723 (מִקְוֶה);","strongs_def":"a collection, i.e. (of water) a reservoir","kjv_def":"ditch."},
"H4725":{"lemma":"מָקוֹם","xlit":"mâqôwm","pron":"maw-kome'","derivation":"or מָקֹם; also (feminine) מְקוֹמָה; or מְקֹמָה; from H6965 (קוּם);","strongs_def":"properly, a standing, i.e. a spot; but used widely of a locality (general or specific); also (figuratively) of a condition (of body or mind)","kjv_def":"country, [idiom] home, [idiom] open, place, room, space, [idiom] whither(-soever)."},
"H4726":{"lemma":"מָקוֹר","xlit":"mâqôwr","pron":"maw-kore'","derivation":"or מָקֹר; from H6979 (קוּר);","strongs_def":"properly, something dug, i.e. a (general) source (of water, even when naturally flowing; also of tears, blood (by euphemism, of the female pudenda); figuratively, of happiness, wisdom, progeny)","kjv_def":"fountain, issue, spring, well(-spring)."},
"H4727":{"lemma":"מִקָּח","xlit":"miqqâch","pron":"mik-kawkh'","derivation":"from H3947 (לָקַח);","strongs_def":"reception","kjv_def":"taking."},
"H4728":{"lemma":"מַקָּחָה","xlit":"maqqâchâh","pron":"mak-kaw-khaw'","derivation":"from H3947 (לָקַח);","strongs_def":"something received, i.e. merchandise (purchased)","kjv_def":"ware."},
"H4729":{"lemma":"מִקְטָר","xlit":"miqṭâr","pron":"mik-tawr'","derivation":"from H6999 (קָטַר);","strongs_def":"something to fume (incense) on i.e. a hearth place","kjv_def":"to burn...upon."},
"H4730":{"lemma":"מִקְטֶרֶת","xlit":"miqṭereth","pron":"mik-teh'-reth","derivation":"feminine of H4729 (מִקְטָר);","strongs_def":"something to fume (incense) in, i.e. a coal-pan","kjv_def":"censer."},
"H4731":{"lemma":"מַקֵּל","xlit":"maqqêl","pron":"mak-kale","derivation":"or (feminine) מַקְּלָה; from an unused root meaning apparently to germinate;","strongs_def":"a shoot, i.e. stick (with leaves on, or for walking, striking, guiding, divining)","kjv_def":"rod, (hand-)staff."},
"H4732":{"lemma":"מִקְלוֹת","xlit":"Miqlôwth","pron":"mik-lohth'","derivation":"(or perhaps mik-kel-ohth') plural of (feminine) H4731 (מַקֵּל); rods;","strongs_def":"Mikloth, a place in the Desert","kjv_def":"Mikloth."},
"H4733":{"lemma":"מִקְלָט","xlit":"miqlâṭ","pron":"mik-lawt'","derivation":"from H7038 (קָלַט) in the sense of taking in;","strongs_def":"an asylum (as a receptacle)","kjv_def":"refuge."},
"H4734":{"lemma":"מִקְלַעַת","xlit":"miqlaʻath","pron":"mik-lah'-ath","derivation":"from H7049 (קָלַע);","strongs_def":"a sculpture (probably in bas-relief)","kjv_def":"carved (figure), carving, graving."},
"H4735":{"lemma":"מִקְנֶה","xlit":"miqneh","pron":"mik-neh'","derivation":"from H7069 (קָנָה);","strongs_def":"something bought, i.e. property, but only livestock; abstractly, acquisition","kjv_def":"cattle, flock, herd, possession, purchase, substance."},
"H4736":{"lemma":"מִקְנָה","xlit":"miqnâh","pron":"mik-naw'","derivation":"feminine of H4735 (מִקְנֶה);","strongs_def":"properly, a buying, i.e. acquisition; concretely, a piece of property (land or living); also the sum paid","kjv_def":"(he that is) bought, possession, piece, purchase."},
"H4737":{"lemma":"מִקְנֵיָהוּ","xlit":"Miqnêyâhûw","pron":"mik-nay-yaw'-hoo","derivation":"from H4735 (מִקְנֶה) and H3050 (יָהּ); possession of Jah;","strongs_def":"Miknejah, an Israelite","kjv_def":"Mikneiah."},
"H4738":{"lemma":"מִקְסָם","xlit":"miqçâm","pron":"mik-sawn'","derivation":"from H7080 (קָסַם);","strongs_def":"an augury","kjv_def":"divination."},
"H4739":{"lemma":"מָקַץ","xlit":"Mâqats","pron":"maw-kats'","derivation":"from H7112 (קָצַץ); end;","strongs_def":"Makats, a place in Palestine","kjv_def":"Makaz."},
"H4740":{"lemma":"מַקְצוֹעַ","xlit":"maqtsôwaʻ","pron":"mak-tso'-ah","derivation":"or מַקְצֹעַ; or (feminine) מַקְצֹעָה; from H7106 (קָצַע) in the denominative sense of bending;","strongs_def":"an angle or recess","kjv_def":"corner, turning."},
"H4741":{"lemma":"מַקְצֻעָה","xlit":"maqtsuʻâh","pron":"mak-tsoo-aw'","derivation":"from H7106 (קָצַע);","strongs_def":"a scraper, i.e. a carving chisel","kjv_def":"plane."},
"H4742":{"lemma":"מְקֻצְעָה","xlit":"mᵉqutsʻâh","pron":"mek-oots-aw'","derivation":"from H7106 (קָצַע) in the denominative sense of bending;","strongs_def":"an angle","kjv_def":"corner."},
"H4743":{"lemma":"מָקַק","xlit":"mâqaq","pron":"maw-kak'","derivation":"' a primitive root;","strongs_def":"to melt; figuratively, to flow, dwindle, vanish","kjv_def":"consume away, be corrupt, dissolve, pine away."},
"H4744":{"lemma":"מִקְרָא","xlit":"miqrâʼ","pron":"mik-raw'","derivation":"from H7121 (קָרָא);","strongs_def":"something called out, i.e. a public meeting (the act, the persons, or the place); also a rehearsal","kjv_def":"assembly, calling, convocation, reading."},
"H4745":{"lemma":"מִקְרֶה","xlit":"miqreh","pron":"mik-reh'","derivation":"from H7136 (קָרָה);","strongs_def":"something met with, i.e. an accident or fortune","kjv_def":"something befallen, befalleth, chance, event, hap(-peneth)."},
"H4746":{"lemma":"מְקָרֶה","xlit":"mᵉqâreh","pron":"mek-aw-reh'","derivation":"from H7136 (קָרָה);","strongs_def":"properly, something meeting, i.e. a frame (of timbers)","kjv_def":"building."},
"H4747":{"lemma":"מְקֵרָה","xlit":"mᵉqêrâh","pron":"mek-ay-raw'","derivation":"from the same as H7119 (קַר);","strongs_def":"a cooling off","kjv_def":"[idiom] summer."},
"H4748":{"lemma":"מִקְשֶׁה","xlit":"miqsheh","pron":"mik-sheh'","derivation":"from H7185 (קָשָׁה) in the sense of knotting up round and hard;","strongs_def":"something turned (rounded), i.e. a curl (of tresses)","kjv_def":"[idiom] well (set) hair."},
"H4749":{"lemma":"מִקְשָׁה","xlit":"miqshâh","pron":"mik-shaw'","derivation":"feminine of H4748 (מִקְשֶׁה);","strongs_def":"rounded work, i.e. moulded by hammering (repousse)","kjv_def":"beaten (out of one piece, work), upright, whole piece."},
"H4750":{"lemma":"מִקְשָׁה","xlit":"miqshâh","pron":"mik-shaw'","derivation":"denominative from H7180 (קִשֻּׁא);","strongs_def":"literally a cucumbered field, i.e. a cucumber patch","kjv_def":"garden of cucumbers."},
"H4751":{"lemma":"מַר","xlit":"mar","pron":"mar","derivation":"or (feminine) מָרָה; from H4843 (מָרַר);","strongs_def":"bitter (literally or figuratively); also (as noun) bitterness, or (adverbially) bitterly","kjv_def":"[phrase] angry, bitter(-ly, -ness), chafed, discontented, [idiom] great, heavy."},
"H4752":{"lemma":"מַר","xlit":"mar","pron":"mar","derivation":"from H4843 (מָרַר) in its original sense of distillation;","strongs_def":"a drop","kjv_def":"drop."},
"H4753":{"lemma":"מֹר","xlit":"môr","pron":"more","derivation":"or מוֹר; from H4843 (מָרַר);","strongs_def":"myrrh (as distilling in drops, and also as bitter)","kjv_def":"myrrh."},
"H4754":{"lemma":"מָרָא","xlit":"mârâʼ","pron":"maw-raw'","derivation":"a primitive root;","strongs_def":"to rebel; hence (through the idea of maltreating) to whip, i.e. lash (self with wings, as the ostrich in running)","kjv_def":"be filthy, lift up self."},
"H4755":{"lemma":"מָרָא","xlit":"Mârâʼ","pron":"maw-raw'","derivation":"for H4751 (מַר) feminine; bitter;","strongs_def":"Mara, a symbolic name of Naomi","kjv_def":"Mara."},
"H4756":{"lemma":"מָרֵא","xlit":"mârêʼ","pron":"maw-ray'","derivation":"(Aramaic) from a root corresponding to H4754 (מָרָא) in the sense of domineering;","strongs_def":"a master","kjv_def":"lord, Lord."},
"H4757":{"lemma":"מְרֹאדַךְ בַּלְאָדָן","xlit":"Mᵉrôʼdak Balʼâdân","pron":"mer-o-dak' bal-awdawn'","derivation":"of foreign derivation;","strongs_def":"Merodak-Baladan, a Babylonian king","kjv_def":"Merodach-baladan. Compare H4781 (מְרֹדָךְ)."},
"H4758":{"lemma":"מַרְאֶה","xlit":"marʼeh","pron":"mar-eh'","derivation":"from H7200 (רָאָה);","strongs_def":"a view (the act of seeing); also an appearance (the thing seen), whether (real) a shape (especially if handsome, comeliness; often plural the looks), or (mental) a vision","kjv_def":"[idiom] apparently, appearance(-reth), [idiom] as soon as beautiful(-ly), countenance, fair, favoured, form, goodly, to look (up) on (to), look(-eth), pattern, to see, seem, sight, visage, vision."},
"H4759":{"lemma":"מַרְאָה","xlit":"marʼâh","pron":"mar-aw'","derivation":"feminine of H4758 (מַרְאֶה);","strongs_def":"a vision; also (causatively) a mirror","kjv_def":"looking glass, vision."},
"H4760":{"lemma":"מֻרְאָה","xlit":"murʼâh","pron":"moor-aw'","derivation":"apparently feminine passive causative participle of H7200 (רָאָה);","strongs_def":"something conspicuous, i.e. the craw of abird (from its prominence)","kjv_def":"crop."},
"H4761":{"lemma":"מַרְאָשָׁה","xlit":"marʼâshâh","pron":"mar-aw-shaw'","derivation":"denominative from H7218 (רֹאשׁ);","strongs_def":"properly, headship, i.e. (plural for collective) dominion","kjv_def":"principality."},
"H4762":{"lemma":"מַרְאֵשָׁה","xlit":"Marʼêshâh","pron":"mar-ay-shaw'","derivation":"or מַרֵשָׁה; formed like H4761 (מַרְאָשָׁה); summit;","strongs_def":"Mareshah, the name of two Israelites and of a place in Palestine","kjv_def":"Mareshah."},
"H4763":{"lemma":"מְרַאֲשָׁה","xlit":"mᵉraʼăshâh","pron":"mer-ah-ash-aw'","derivation":"formed like H4761 (מַרְאָשָׁה);","strongs_def":"properly, a headpiece, i.e. (plural for adverbial) at (or as) the head-rest (or pillow)","kjv_def":"bolster, head, pillow. Compare H4772 (מַרְגְלָה)."},
"H4764":{"lemma":"מֵרָב","xlit":"Mêrâb","pron":"may-rawb'","derivation":"from H7231 (רָבַב); increase;","strongs_def":"Merab, a daughter of Saul","kjv_def":"Merab."},
"H4765":{"lemma":"מַרְבַד","xlit":"marbad","pron":"mar-bad'","derivation":"from H7234 (רָבַד);","strongs_def":"a coverlet","kjv_def":"covering of tapestry."},
"H4766":{"lemma":"מַרְבֶה","xlit":"marbeh","pron":"mar-beh'","derivation":"from H7235 (רָבָה);","strongs_def":"properly, increasing; as noun, greatness, or (adverbially) greatly","kjv_def":"great, increase."},
"H4767":{"lemma":"מִרְבָּה","xlit":"mirbâh","pron":"meer-baw'","derivation":"from H7235 (רָבָה);","strongs_def":"abundance, i.e. a great quantity","kjv_def":"much."},
"H4768":{"lemma":"מַרְבִּית","xlit":"marbîyth","pron":"mar-beeth'","derivation":"from H7235 (רָבָה);","strongs_def":"a multitude; also offspring; specifically interest (on capital)","kjv_def":"greatest part, greatness, increase, multitude."},
"H4769":{"lemma":"מַרְבֵּץ","xlit":"marbêts","pron":"mar-bates'","derivation":"from H7257 (רָבַץ);","strongs_def":"a reclining place, i.e. fold (for flocks)","kjv_def":"couching place, place to lie down."},
"H4770":{"lemma":"מַרְבֵּק","xlit":"marbêq","pron":"mar-bake'","derivation":"from an unused root meaning to tie up;","strongs_def":"a stall (for cattle)","kjv_def":"[idiom] fat(-ted), stall."},
"H4771":{"lemma":"מַרְגּוֹעַ","xlit":"margôwaʻ","pron":"mar-go'-ah","derivation":"from H7280 (רָגַע);","strongs_def":"a resting place","kjv_def":"rest."},
"H4772":{"lemma":"מַרְגְלָה","xlit":"margᵉlâh","pron":"mar-ghel-aw'","derivation":"denominative from H7272 (רֶגֶל);","strongs_def":"(plural for collective) a footpiece, i.e. (adverbially) at the foot, or (direct.) the foot itself","kjv_def":"feet. Compare H4763 (מְרַאֲשָׁה)."},
"H4773":{"lemma":"מַרְגֵּמָה","xlit":"margêmâh","pron":"mar-gay-maw'","derivation":"from H7275 (רָגַם);","strongs_def":"a stone-heap","kjv_def":"sling."},
"H4774":{"lemma":"מַרְגֵּעָה","xlit":"margêʻâh","pron":"mar-gay-aw'","derivation":"from H7280 (רָגַע);","strongs_def":"rest","kjv_def":"refreshing."},
"H4775":{"lemma":"מָרַד","xlit":"mârad","pron":"maw-rad'","derivation":"a primitive root;","strongs_def":"to rebel","kjv_def":"rebel(-lious)."},
"H4776":{"lemma":"מְרַד","xlit":"mᵉrad","pron":"mer-ad'","derivation":"(Aramaic) from a root corresponding to H4775 (מָרַד);","strongs_def":"rebellion","kjv_def":"rebellion."},
"H4777":{"lemma":"מֶרֶד","xlit":"mered","pron":"meh'-red","derivation":"from H4775 (מָרַד);","strongs_def":"rebellion","kjv_def":"rebellion."},
"H4778":{"lemma":"מֶרֶד","xlit":"Mered","pron":"meh'-red","derivation":"the same as H4777 (מֶרֶד);","strongs_def":"Mered, an Israelite","kjv_def":"Mered."},
"H4779":{"lemma":"מָרָד","xlit":"mârâd","pron":"maw-rawd'","derivation":"(Aramaic) from the same as H4776 (מְרַד);","strongs_def":"rebellious","kjv_def":"rebellious."},
"H4780":{"lemma":"מַרְדּוּת","xlit":"mardûwth","pron":"mar-dooth'","derivation":"from H4775 (מָרַד);","strongs_def":"rebelliousness","kjv_def":"[idiom] rebellious."},
"H4781":{"lemma":"מְרֹדָךְ","xlit":"Mᵉrôdâk","pron":"mer-o-dawk'","derivation":"of foreign derivation;","strongs_def":"Merodak, a Babylonian idol","kjv_def":"Merodach. Compare H4757 (מְרֹאדַךְ בַּלְאָדָן)."},
"H4782":{"lemma":"מׇרְדְּכַי","xlit":"Mordᵉkay","pron":"mor-dek-ah'-ee","derivation":"of foreign derivation;","strongs_def":"Mordecai, an Israelite","kjv_def":"Mordecai."},
"H4783":{"lemma":"מֻרְדָּף","xlit":"murdâph","pron":"moor-dawf'","derivation":"from H7291 (רָדַף);","strongs_def":"persecuted","kjv_def":"persecuted."},
"H4784":{"lemma":"מָרָה","xlit":"mârâh","pron":"maw-raw'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) bitter (or unpleasant); (figuratively) to rebel (or resist; causatively, to provoke)","kjv_def":"bitter, change, be disobedient, disobey, grievously, provocation, provoke(-ing), (be) rebel (against, -lious)."},
"H4785":{"lemma":"מָרָה","xlit":"Mârâh","pron":"maw-raw'","derivation":"the same as H4751 (מַר) feminine; bitter;","strongs_def":"Marah, a place in the Desert","kjv_def":"Marah."},
"H4786":{"lemma":"מֹרָה","xlit":"môrâh","pron":"mo-raw'","derivation":"from H4843 (מָרַר);","strongs_def":"bitterness, i.e. (figuratively) trouble","kjv_def":"grief."},
"H4787":{"lemma":"מׇרָּה","xlit":"morrâh","pron":"mor-raw'","derivation":"a form of H4786 (מֹרָה);","strongs_def":"trouble","kjv_def":"bitterness."},
"H4788":{"lemma":"מָרוּד","xlit":"mârûwd","pron":"maw-rood'","derivation":"from H7300 (רוּד) in the sense of maltreatment;","strongs_def":"an outcast; (abstractly) destitution","kjv_def":"cast out, misery."},
"H4789":{"lemma":"מֵרוֹז","xlit":"Mêrôwz","pron":"may-roze'","derivation":"of uncertain derivation;","strongs_def":"Meroz, a place in Palestine","kjv_def":"Meroz."},
"H4790":{"lemma":"מְרוֹחַ","xlit":"mᵉrôwach","pron":"mer-o-akh'","derivation":"from H4799 (מָרַח);","strongs_def":"bruised, i.e. emasculated","kjv_def":"broken."},
"H4791":{"lemma":"מָרוֹם","xlit":"mârôwm","pron":"maw-rome'","derivation":"from H7311 (רוּם);","strongs_def":"altitude, i.e. concretely (an elevated place), abstractly (elevation, figuratively (elation), or adverbially (aloft)","kjv_def":"(far) above, dignity, haughty, height, (most, on) high (one, place), loftily, upward."},
"H4792":{"lemma":"מֵרוֹם","xlit":"Mêrôwm","pron":"may-rome'","derivation":"formed like H4791 (מָרוֹם); height;","strongs_def":"Merom, a lake in Palestine","kjv_def":"Merom."},
"H4793":{"lemma":"מֵרוֹץ","xlit":"mêrôwts","pron":"may-rotes'","derivation":"from H7323 (רוּץ);","strongs_def":"a run (the trial of speed)","kjv_def":"race."},
"H4794":{"lemma":"מְרוּצָה","xlit":"mᵉrûwtsâh","pron":"mer-oo-tsaw'","derivation":"or מְרֻצָה; feminine of H4793 (מֵרוֹץ);","strongs_def":"a race (the act), whether the manner or the progress","kjv_def":"course, running. Compare H4835 (מְרֻצָה)."},
"H4795":{"lemma":"מָרוּק","xlit":"mârûwq","pron":"maw-rook'","derivation":"from H4838 (מָרַק);","strongs_def":"properly, rubbed; but used abstractly, a rubbing (with perfumery)","kjv_def":"purification."},
"H4796":{"lemma":"מָרוֹת","xlit":"Mârôwth","pron":"maw-rohth'","derivation":"plural of H4751 (מַר) feminine; bitter springs;","strongs_def":"Maroth, a place in Palestine","kjv_def":"Maroth."},
"H4797":{"lemma":"מִרְזַח","xlit":"mirzach","pron":"meer-zakh'","derivation":"from an unused root meaning to scream;","strongs_def":"a cry, i.e. (of job), a revel","kjv_def":"banquet."},
"H4798":{"lemma":"מַרְזֵחַ","xlit":"marzêach","pron":"mar-zay'-akh","derivation":"formed like H4797 (מִרְזַח);","strongs_def":"a cry, i.e. (of grief) a lamentation","kjv_def":"mourning."},
"H4799":{"lemma":"מָרַח","xlit":"mârach","pron":"maw-rakh'","derivation":"a primitive root;","strongs_def":"properly, to soften by rubbing or pressure; hence (medicinally) to apply as an emollient","kjv_def":"lay for a plaister."},
"H4800":{"lemma":"מֶרְחָב","xlit":"merchâb","pron":"mer-khawb'","derivation":"from H7337 (רָחַב);","strongs_def":"enlargement, either literally (an open space, usually in a good sense), or figuratively (liberty)","kjv_def":"breadth, large place (room)."},
"H4801":{"lemma":"מֶרְחָק","xlit":"merchâq","pron":"mer-khawk'","derivation":"; from H7368 (רָחַק);","strongs_def":"remoteness, i.e. (concretely) a distant place; often (adverbially) from afar","kjv_def":"(a-, dwell in, very) far (country, off). See also H1023 (בֵּית הַמֶּרְחָק)."},
"H4802":{"lemma":"מַרְחֶשֶׁת","xlit":"marchesheth","pron":"mar-kheh'-sheth","derivation":"from H7370 (רָחַשׁ);","strongs_def":"a stewpan","kjv_def":"fryingpan."},
"H4803":{"lemma":"מָרַט","xlit":"mâraṭ","pron":"maw-rat'","derivation":"a primitive root;","strongs_def":"to polish; by implication, to make bald (the head), to gall (the shoulder); also, to sharpen","kjv_def":"bright, furbish, (have his) hair (be) fallen off, peeled, pluck off (hair)."},
"H4804":{"lemma":"מְרַט","xlit":"mᵉraṭ","pron":"mer-at'","derivation":"(Aramaic) corresponding to H4803 (מָרַט);","strongs_def":"to pull off","kjv_def":"be plucked."},
"H4805":{"lemma":"מְרִי","xlit":"mᵉrîy","pron":"mer-ee'","derivation":"from H4784 (מָרָה);","strongs_def":"bitterness, i.e. (figuratively) rebellion; concretely, bitter, or rebellious","kjv_def":"bitter, (most) rebel(-lion, -lious)."},
"H4806":{"lemma":"מְרִיא","xlit":"mᵉrîyʼ","pron":"mer-ee'","derivation":"from H4754 (מָרָא) in the sense of grossness, through the idea of domineering (compare H4756 (מָרֵא));","strongs_def":"stall-fed; often (as noun) a beeve","kjv_def":"fat (fed) beast (cattle, -ling)."},
"H4807":{"lemma":"מְרִיב בַּעַל","xlit":"Mᵉrîyb Baʻal","pron":"mer-eeb' bah'-al","derivation":"from H7378 (רִיב) and H1168 (בַּעַל);","strongs_def":"quarreller of Baal, Merib-Baal, an epithet of Gideon","kjv_def":"Merib-baal. Compare H4810 (מְרִי בַעַל)."},
"H4808":{"lemma":"מְרִיבָה","xlit":"mᵉrîybâh","pron":"mer-ee-baw'","derivation":"from H7378 (רִיב);","strongs_def":"quarrel","kjv_def":"provocation, strife."},
"H4809":{"lemma":"מְרִיבָה","xlit":"Mᵉrîybâh","pron":"mer-ee-baw'","derivation":"the same as H4808 (מְרִיבָה);","strongs_def":"Meribah, the name of two places in the Desert","kjv_def":"Meribah."},
"H4810":{"lemma":"מְרִי בַעַל","xlit":"Mᵉrîy Baʻal","pron":"mer-ee' bah'-al","derivation":"from H4805 (מְרִי) and H1168 (בַּעַל); rebellion of (i.e. against) Baal;","strongs_def":"Meri-Baal, an epithet of Gideon","kjv_def":"Meri-baal. Compare H4807 (מְרִיב בַּעַל)."},
"H4811":{"lemma":"מְרָיָה","xlit":"Mᵉrâyâh","pron":"mer-aw-yaw'","derivation":"from H4784 (מָרָה); rebellion;","strongs_def":"Merajah, an Israelite","kjv_def":"Meraiah. Compare H3236 (יִמְרָה)."},
"H4812":{"lemma":"מְרָיוֹת","xlit":"Mᵉrâyôwth","pron":"mer-aw-yohth'","derivation":"plural of H4811 (מְרָיָה); rebellious;","strongs_def":"Merajoth, the name of two Israelites","kjv_def":"Meraioth."},
"H4813":{"lemma":"מִרְיָם","xlit":"Miryâm","pron":"meer-yawm'","derivation":"from H4805 (מְרִי); rebelliously;","strongs_def":"Mirjam, the name of two Israelitesses","kjv_def":"Miriam."},
"H4814":{"lemma":"מְרִירוּת","xlit":"mᵉrîyrûwth","pron":"mer-ee-rooth'","derivation":"from H4843 (מָרַר);","strongs_def":"bitterness, i.e. (figuratively) grief","kjv_def":"bitterness."},
"H4815":{"lemma":"מְרִירִי","xlit":"mᵉrîyrîy","pron":"mer-ee-ree'","derivation":"from H4843 (מָרַר);","strongs_def":"bitter, i.e. poisonous","kjv_def":"bitter."},
"H4816":{"lemma":"מֹרֶךְ","xlit":"môrek","pron":"mo'-rek","derivation":"perhaps from H7401 (רָכַךְ);","strongs_def":"softness, i.e. (figuratively) fear","kjv_def":"faintness."},
"H4817":{"lemma":"מֶרְכָּב","xlit":"merkâb","pron":"mer-kawb'","derivation":"from H7392 (רָכַב);","strongs_def":"a chariot; also a seat (in a vehicle)","kjv_def":"chariot, covering, saddle."},
"H4818":{"lemma":"מֶרְכָּבָה","xlit":"merkâbâh","pron":"mer-kaw-baw'","derivation":"feminine of H4817 (מֶרְכָּב);","strongs_def":"a chariot","kjv_def":"chariot. See also H1024 (בֵּית הַמַּרְכָּבוֹת)."},
"H4819":{"lemma":"מַרְכֹּלֶת","xlit":"markôleth","pron":"mar-ko'-leth","derivation":"from H7402 (רָכַל);","strongs_def":"a mart","kjv_def":"merchandise."},
"H4820":{"lemma":"מִרְמָה","xlit":"mirmâh","pron":"meer-maw'","derivation":"from H7411 (רָמָה) in the sense of deceiving;","strongs_def":"fraud","kjv_def":"craft, deceit(-ful, -fully), false, feigned, guile, subtilly, treachery."},
"H4821":{"lemma":"מִרְמָה","xlit":"Mirmâh","pron":"meer-maw'","derivation":"the same as H4820 (מִרְמָה);","strongs_def":"Mirmah, an Israelite","kjv_def":"Mirma."},
"H4822":{"lemma":"מְרֵמוֹת","xlit":"Mᵉrêmôwth","pron":"mer-ay-mohth'","derivation":"plural from H7311 (רוּם); heights;","strongs_def":"Meremoth, the name of two Israelites","kjv_def":"Meremoth."},
"H4823":{"lemma":"מִרְמָס","xlit":"mirmâç","pron":"meer-mawce'","derivation":"from H7429 (רָמַס);","strongs_def":"abasement (the act or the thing)","kjv_def":"tread (down) -ing, (to be) trodden (down) under foot."},
"H4824":{"lemma":"מֵרֹנֹתִי","xlit":"Mêrônôthîy","pron":"may-ro-no-thee'","derivation":"patrial from an unused noun;","strongs_def":"a Meronothite, or inhabitant of some (otherwise unknown) Meronoth.","kjv_def":"Meronothite."},
"H4825":{"lemma":"מֶרֶס","xlit":"Mereç","pron":"meh'-res","derivation":"of foreign derivation;","strongs_def":"Meres, a Persian","kjv_def":"Meres."},
"H4826":{"lemma":"מַרְסְנָא","xlit":"Marçᵉnâʼ","pron":"mar-sen-aw'","derivation":"of foreign derivation;","strongs_def":"Marsena, a Persian","kjv_def":"Marsena."},
"H4827":{"lemma":"מֵרַע","xlit":"mêraʻ","pron":"may-rah'","derivation":"from H7489 (רָעַע);","strongs_def":"used as (abstract) noun, wickedness","kjv_def":"do mischief."},
"H4828":{"lemma":"מֵרֵעַ","xlit":"mêrêaʻ","pron":"may-ray'-ah","derivation":"from H7462 (רָעָה) in the sense of companionship;","strongs_def":"a friend","kjv_def":"companion, friend."},
"H4829":{"lemma":"מִרְעֶה","xlit":"mirʻeh","pron":"meer-eh'","derivation":"from H7462 (רָעָה) in the sense of feeding;","strongs_def":"pasture (the place or the act); also the haunt of wild animals","kjv_def":"feeding place, pasture."},
"H4830":{"lemma":"מִרְעִית","xlit":"mirʻîyth","pron":"meer-eeth'","derivation":"from H7462 (רָעָה) in the sense of feeding;","strongs_def":"pasturage; concretely, a flock","kjv_def":"flock, pasture."},
"H4831":{"lemma":"מַרְעֲלָה","xlit":"Marʻălâh","pron":"mar-al-aw'","derivation":"from H7477 (רָעַל); perhaps, earthquake;","strongs_def":"Maralah, a place in Palestine","kjv_def":"Maralah."},
"H4832":{"lemma":"מַרְפֵּא","xlit":"marpêʼ","pron":"mar-pay'","derivation":"from H7495 (רָפָא);","strongs_def":"properly, curative, i.e. literally (concretely) a medicine, or (abstractly) a cure; figuratively (concretely) deliverance, or (abstractly) placidity","kjv_def":"(in-)cure(-able), healing(-lth), remedy, sound, wholesome, yielding."},
"H4833":{"lemma":"מִרְפָּשׂ","xlit":"mirpâs","pron":"meer-paws'","derivation":"from H7515 (רָפַשׂ);","strongs_def":"muddled water","kjv_def":"that which...have fouled."},
"H4834":{"lemma":"מָרַץ","xlit":"mârats","pron":"maw-rats'","derivation":"a primitive root;","strongs_def":"properly, to press, i.e. (figuratively) to be pungent or vehement; to irritate","kjv_def":"embolden, be forcible, grievous, sore."},
"H4835":{"lemma":"מְרֻצָה","xlit":"mᵉrutsâh","pron":"mer-oo-tsaw'","derivation":"from H7533 (רָצַץ);","strongs_def":"oppression","kjv_def":"violence. See also H4794 (מְרוּצָה)."},
"H4836":{"lemma":"מַרְצֵעַ","xlit":"martsêaʻ","pron":"mar-tsay'-ah","derivation":"from H7527 (רָצַע);","strongs_def":"an awl","kjv_def":"aul."},
"H4837":{"lemma":"מַרְצֶפֶת","xlit":"martsepheth","pron":"mar-tseh'-feth","derivation":"from H7528 (רָצַף);","strongs_def":"a pavement","kjv_def":"pavement."},
"H4838":{"lemma":"מָרַק","xlit":"mâraq","pron":"maw-rak'","derivation":"a primitive root;","strongs_def":"to polish; by implication, to sharpen; also to rinse","kjv_def":"bright, furbish, scour."},
"H4839":{"lemma":"מָרָק","xlit":"mârâq","pron":"maw-rawk'","derivation":"from H4838 (מָרַק);","strongs_def":"soup (as if a rinsing)","kjv_def":"broth. See also H6564 (פָּרָק)."},
"H4840":{"lemma":"מֶרְקָח","xlit":"merqâch","pron":"mer-kawkh'","derivation":"from H7543 (רָקַח);","strongs_def":"a spicy herb","kjv_def":"[idiom] sweet."},
"H4841":{"lemma":"מֶרְקָחָה","xlit":"merqâchâh","pron":"mer-kaw-khaw'","derivation":"feminine of H4840 (מֶרְקָח);","strongs_def":"abstractly, a seasoning (with spicery); concretely, an unguentkettle (for preparing spiced oil)","kjv_def":"pot of ointment, [idiom] well."},
"H4842":{"lemma":"מִרְקַחַת","xlit":"mirqachath","pron":"meer-kakh'-ath","derivation":"from H7543 (רָקַח);","strongs_def":"an aromatic unguent; also an unguent-pot","kjv_def":"prepared by the apothecaries' art, compound, ointment."},
"H4843":{"lemma":"מָרַר","xlit":"mârar","pron":"maw-rar'","derivation":"a primitive root; properly, to trickle (see H4752 (מַר)); but used only as a denominative from H4751 (מַר);","strongs_def":"to be (causatively, make) bitter (literally or figuratively)","kjv_def":"(be, be in, deal, have, make) bitter(-ly, -ness), be moved with choler, (be, have sorely, it) grieved(-eth), provoke, vex."},
"H4844":{"lemma":"מְרֹר","xlit":"mᵉrôr","pron":"mer-ore'","derivation":"or מְרוֹר; from H4843 (מָרַר);","strongs_def":"a bitter herb","kjv_def":"bitter(-ness)."},
"H4845":{"lemma":"מְרֵרָה","xlit":"mᵉrêrâh","pron":"mer-ay-raw'","derivation":"from H4843 (מָרַר);","strongs_def":"bile (from its bitterness)","kjv_def":"gall."},
"H4846":{"lemma":"מְרֹרָה","xlit":"mᵉrôrâh","pron":"mer-o-raw'","derivation":"or מְרוֹרָה; from H4843 (מָרַר);","strongs_def":"properly, bitterness; concretely, a bitter thing; specifically bile; also venom (of a serpent)","kjv_def":"bitter (thing), gall."},
"H4847":{"lemma":"מְרָרִי","xlit":"Mᵉrârîy","pron":"mer-aw-ree'","derivation":"from H4843 (מָרַר); bitter;","strongs_def":"Merari, an Israelite","kjv_def":"Merari. See also H4848 (מְרָרִי)."},
"H4848":{"lemma":"מְרָרִי","xlit":"Mᵉrârîy","pron":"mer-aw-ree'","derivation":"from H4847 (מְרָרִי);","strongs_def":"a Merarite (collectively), or decendants of Merari","kjv_def":"Merarites."},
"H4849":{"lemma":"מִרְשַׁעַת","xlit":"mirshaʻath","pron":"meer-shah'-ath","derivation":"from H7561 (רָשַׁע);","strongs_def":"a female wicked doer","kjv_def":"wicked woman."},
"H4850":{"lemma":"מְרָתַיִם","xlit":"Mᵉrâthayim","pron":"mer-aw-thah'-yim","derivation":"dual of H4751 (מַר) feminine; double bitterness;","strongs_def":"Merathajim, an epithet of Babylon","kjv_def":"Merathaim."},
"H4851":{"lemma":"מַשׁ","xlit":"Mash","pron":"mash","derivation":"of foreign derivation;","strongs_def":"Mash, a son of Aram, and the people descended from him","kjv_def":"Mash."},
"H4852":{"lemma":"מֵשָׁא","xlit":"Mêshâʼ","pron":"may-shaw'","derivation":"of foreign derivation;","strongs_def":"Mesha, a place in Arabia","kjv_def":"Mesha."},
"H4853":{"lemma":"מַשָּׂא","xlit":"massâʼ","pron":"mas-saw'","derivation":"from H5375 (נָשָׂא);","strongs_def":"a burden; specifically, tribute, or (abstractly) porterage; figuratively, an utterance, chiefly adoom, especially singing; mental, desire","kjv_def":"burden, carry away, prophecy, [idiom] they set, song, tribute."},
"H4854":{"lemma":"מַשָּׂא","xlit":"Massâʼ","pron":"mas-saw'","derivation":"the same as H4853 (מַשָּׂא); burden;","strongs_def":"Massa, a son of Ishmael","kjv_def":"Massa."},
"H4855":{"lemma":"מַשָּׁא","xlit":"mashshâʼ","pron":"mash-shaw'","derivation":"from H5383 (נָשָׁה);","strongs_def":"a loan; by implication, interest on adebt","kjv_def":"exaction, usury."},
"H4856":{"lemma":"מַשֹּׂא","xlit":"massôʼ","pron":"mas-so'","derivation":"from H5375 (נָשָׂא);","strongs_def":"partiality (as a lifting up)","kjv_def":"respect."},
"H4857":{"lemma":"מַשְׁאָב","xlit":"mashʼâb","pron":"mash-awb'","derivation":"from H7579 (שָׁאַב);","strongs_def":"a trough for cattle to drink from","kjv_def":"place of drawing water."},
"H4858":{"lemma":"מַשָּׂאָה","xlit":"massâʼâh","pron":"mas-saw-aw'","derivation":"from H5375 (נָשָׂא);","strongs_def":"a conflagration (from the rising of smoke)","kjv_def":"burden."},
"H4859":{"lemma":"מַשָּׁאָה","xlit":"mashshâʼâh","pron":"mash-shaw-aw'","derivation":"feminine of H4855 (מַשָּׁא);","strongs_def":"a loan","kjv_def":"[idiom] any(-thing), debt."},
"H4860":{"lemma":"מַשָּׁאוֹן","xlit":"mashshâʼôwn","pron":"mash-shaw-ohn'","derivation":"from H5377 (נָשָׁא);","strongs_def":"dissimulation","kjv_def":"deceit."},
"H4861":{"lemma":"מִשְׁאָל","xlit":"Mishʼâl","pron":"mish-awl'","derivation":"from H7592 (שָׁאַל); request;","strongs_def":"Mishal, a place in Palestine","kjv_def":"Mishal, Misheal. Compare H4913 (מָשָׁל)."},
"H4862":{"lemma":"מִשְׁאָלָה","xlit":"mishʼâlâh","pron":"mish-aw-law'","derivation":"from H7592 (שָׁאַל);","strongs_def":"a request","kjv_def":"desire, petition."},
"H4863":{"lemma":"מִשְׁאֶרֶת","xlit":"mishʼereth","pron":"mish-eh'-reth","derivation":"from H7604 (שָׁאַר) in the original sense of swelling;","strongs_def":"a kneading-trough (in which the dough rises)","kjv_def":"kneading trough, store."},
"H4864":{"lemma":"מַשְׂאֵת","xlit":"masʼêth","pron":"mas-ayth'","derivation":"from H5375 (נָשָׂא);","strongs_def":"properly, (abstractly) a raising (as of the hands in prayer), or rising (of flame); figuratively, an utterance; concretely, a beacon (as raised); a present (as taken), mess, or tribute; figuratively, a reproach (as a burden)","kjv_def":"burden, collection, sign of fire, (great) flame, gift, lifting up, mess, oblation, reward."},
"H4865":{"lemma":"מִשְׁבְּצָה","xlit":"mishbᵉtsâh","pron":"mish-bets-aw'","derivation":"from H7660 (שָׁבַץ);","strongs_def":"a brocade; by analogy, a (reticulated) setting of agem","kjv_def":"ouch, wrought."},
"H4866":{"lemma":"מִשְׁבֵּר","xlit":"mishbêr","pron":"mish-bare'","derivation":"from H7665 (שָׁבַר);","strongs_def":"the orifice of the womb (from which the fetus breaks forth)","kjv_def":"birth, breaking forth."},
"H4867":{"lemma":"מִשְׁבָּר","xlit":"mishbâr","pron":"mish-bawr'","derivation":"from H7665 (שָׁבַר);","strongs_def":"a breaker (of the sea)","kjv_def":"billow, wave."},
"H4868":{"lemma":"מִשְׁבָּת","xlit":"mishbâth","pron":"mish-bawth'","derivation":"from H7673 (שָׁבַת);","strongs_def":"cessation, i.e. destruction","kjv_def":"sabbath."},
"H4869":{"lemma":"מִשְׂגָּב","xlit":"misgâb","pron":"mis-gawb'","derivation":"from H7682 (שָׂגַב); properly, a cliff (or other lofty or inaccessible place); abstractly, altitude; figuratively, a refuge","strongs_def":"defence, high fort (tower), refuge, Misgab, a place in Moab","kjv_def":"Misgab."},
"H4870":{"lemma":"מִשְׁגֶּה","xlit":"mishgeh","pron":"mish-gay'","derivation":"from H7686 (שָׁגָה);","strongs_def":"an error","kjv_def":"oversight."},
"H4871":{"lemma":"מָשָׁה","xlit":"mâshâh","pron":"maw-shaw'","derivation":"a primitive root;","strongs_def":"to pull out (literally or figuratively)","kjv_def":"draw(out)."},
"H4872":{"lemma":"מֹשֶׁה","xlit":"Môsheh","pron":"mo-sheh'","derivation":"from H4871 (מָשָׁה); drawing out (of the water), i.e. rescued;","strongs_def":"Mosheh, the Israelite lawgiver","kjv_def":"Moses."},
"H4873":{"lemma":"מֹשֶׁה","xlit":"Môsheh","pron":"mo-sheh'","derivation":"(Aramaic) corresponding to H4872 (מֹשֶׁה)","strongs_def":"{Mosheh, the Israelite lawgiver}","kjv_def":"Moses."},
"H4874":{"lemma":"מַשֶּׁה","xlit":"mashsheh","pron":"mash-sheh'","derivation":"from H5383 (נָשָׁה);","strongs_def":"a debt","kjv_def":"[phrase] creditor."},
"H4875":{"lemma":"מְשׁוֹאָה","xlit":"mᵉshôwʼâh","pron":"meh-o-aw'","derivation":"or מְשֹׁאָה; from the same as H7722 (שׁוֹא);","strongs_def":"(a) ruin, abstractly (the act) or concretely (the wreck)","kjv_def":"desolation, waste."},
"H4876":{"lemma":"מַשּׁוּאָה","xlit":"mashshûwʼâh","pron":"mash-shoo-aw'","derivation":"or מַשֻּׁאָה; for H4875 (מְשׁוֹאָה);","strongs_def":"ruin","kjv_def":"desolation, destruction."},
"H4877":{"lemma":"מְשׁוֹבָב","xlit":"Mᵉshôwbâb","pron":"mesh-o-bawb'","derivation":"from H7725 (שׁוּב); returned;","strongs_def":"Meshobab, an Israelite","kjv_def":"Meshobab."},
"H4878":{"lemma":"מְשׁוּבָה","xlit":"mᵉshûwbâh","pron":"mesh-oo-baw'","derivation":"or מְשֻׁבָה; from H7725 (שׁוּב);","strongs_def":"apostasy","kjv_def":"backsliding, turning away."},
"H4879":{"lemma":"מְשׁוּגָה","xlit":"mᵉshûwgâh","pron":"mesh-oo-gaw'","derivation":"from an unused root meaning to stray;","strongs_def":"mistake","kjv_def":"error."},
"H4880":{"lemma":"מָשׁוֹט","xlit":"mâshôwṭ","pron":"maw-shote'","derivation":"or מִשּׁוֹט; from H7751 (שׁוּט);","strongs_def":"an oar","kjv_def":"oar."},
"H4881":{"lemma":"מְשׂוּכָה","xlit":"mᵉsûwkâh","pron":"mes-oo-kaw'","derivation":"or מְשֻׂכָה; from H7753 (שׂוּךְ);","strongs_def":"a hedge","kjv_def":"hedge."},
"H4882":{"lemma":"מְשׁוּסָה","xlit":"mᵉshûwçâh","pron":"mesh-oo-saw'","derivation":"from an unused root meaning to plunder;","strongs_def":"spoilation","kjv_def":"spoil."},
"H4883":{"lemma":"מַשּׂוֹר","xlit":"massôwr","pron":"mas-sore'","derivation":"from an unused root meaning to rasp;","strongs_def":"a saw","kjv_def":"saw."},
"H4884":{"lemma":"מְשׂוּרָה","xlit":"mᵉsûwrâh","pron":"mes-oo-raw'","derivation":"from an unused root meaning apparently to divide;","strongs_def":"a measure (for liquids)","kjv_def":"measure."},
"H4885":{"lemma":"מָשׂוֹשׂ","xlit":"mâsôws","pron":"maw-soce'","derivation":"from H7797 (שׂוּשׂ);","strongs_def":"delight, concretely (the cause or object) or abstractly (the feeling)","kjv_def":"joy, mirth, rejoice."},
"H4886":{"lemma":"מָשַׁח","xlit":"mâshach","pron":"maw-shakh'","derivation":"a primitive root;","strongs_def":"to rub with oil, i.e. to anoint; by implication, to consecrate; also to paint","kjv_def":"anoint, paint."},
"H4887":{"lemma":"מְשַׁח","xlit":"mᵉshach","pron":"mesh-akh'","derivation":"(Aramaic) from a root corresponding to H4886 (מָשַׁח);","strongs_def":"oil","kjv_def":"oil."},
"H4888":{"lemma":"מִשְׁחָה","xlit":"mishchâh","pron":"meesh-khaw'","derivation":"or מׇשְׁחָה; from H4886 (מָשַׁח);","strongs_def":"unction (the act); by implication, a consecratory gift","kjv_def":"(to be) anointed(-ing), ointment."},
"H4889":{"lemma":"מַשְׁחִית","xlit":"mashchîyth","pron":"mash-kheeth'","derivation":"from H7843 (שָׁחַת);","strongs_def":"destructive, i.e. (as noun) destruction, literally (specifically a snare) or figuratively (corruption)","kjv_def":"corruption, (to) destroy(-ing), destruction, trap, [idiom] utterly."},
"H4890":{"lemma":"מִשְׂחַק","xlit":"mischaq","pron":"mis-khawk'","derivation":"from H7831 (שַׁחֲצוֹם);","strongs_def":"a laughing-stock","kjv_def":"scorn."},
"H4891":{"lemma":"מִשְׁחָר","xlit":"mishchâr","pron":"mish-khawr'","derivation":"from H7836 (שָׁחַר) in the sense of day breaking;","strongs_def":"dawn","kjv_def":"morning."},
"H4892":{"lemma":"מַשְׁחֵת","xlit":"mashchêth","pron":"mash-khayth'","derivation":"for H4889 (מַשְׁחִית);","strongs_def":"destruction","kjv_def":"destroying."},
"H4893":{"lemma":"מִשְׁחָת","xlit":"mishchâth","pron":"mish-khawth'","derivation":"or מׇשְׁחָת; from H7843 (שָׁחַת);","strongs_def":"disfigurement","kjv_def":"corruption, marred."},
"H4894":{"lemma":"מִשְׁטוֹחַ","xlit":"mishṭôwach","pron":"mish-to'-akh","derivation":"or מִשְׁטַח; from H7849 (שָׁטַח);","strongs_def":"a spreading-place","kjv_def":"(to) spread (forth, -ing, upon)."},
"H4895":{"lemma":"מַשְׂטֵמָה","xlit":"masṭêmâh","pron":"mas-tay-maw'","derivation":"from the same as H7850 (שֹׁטֵט);","strongs_def":"enmity","kjv_def":"hatred."},
"H4896":{"lemma":"מִשְׁטָר","xlit":"mishṭâr","pron":"mish-tawr'","derivation":"from H7860 (שֹׁטֵר);","strongs_def":"jurisdiction","kjv_def":"dominion."},
"H4897":{"lemma":"מֶשִׁי","xlit":"meshîy","pron":"meh'-shee","derivation":"from H4871 (מָשָׁה);","strongs_def":"silk (as drawn from the cocoon)","kjv_def":"silk."},
"H4898":{"lemma":"מְשֵׁיזַבְאֵל","xlit":"Mᵉshêyzabʼêl","pron":"mesh-ay-zab-ale'","derivation":"from an equiv. to H7804 (שְׁזַב) and H410 (אֵל); delivered of God;","strongs_def":"Meshezabel, an Israelite","kjv_def":"Meshezabeel."},
"H4899":{"lemma":"מָשִׁיחַ","xlit":"mâshîyach","pron":"maw-shee'-akh","derivation":"from H4886 (מָשַׁח);","strongs_def":"anointed; usually a consecrated person (as a king, priest, or saint); specifically, the Messiah","kjv_def":"anointed, Messiah."},
"H4900":{"lemma":"מָשַׁךְ","xlit":"mâshak","pron":"maw-shak'","derivation":"a primitive root;","strongs_def":"to draw, used in a great variety of applications (including to sow, to sound, to prolong, to develop, to march, to remove, to delay, to be tall, etc.)","kjv_def":"draw (along, out), continue, defer, extend, forbear, [idiom] give, handle, make (pro-, sound) long, [idiom] sow, scatter, stretch out."},
"H4901":{"lemma":"מֶשֶׁךְ","xlit":"meshek","pron":"meh'shek","derivation":"from H4900 (מָשַׁךְ);","strongs_def":"a sowing; also a possession","kjv_def":"precious, price."},
"H4902":{"lemma":"מֶשֶׁךְ","xlit":"Meshek","pron":"meh'-shek","derivation":"the same in form as H4901 (מֶשֶׁךְ), but probably of foreign derivation;","strongs_def":"Meshek, a son of Japheth, and the people descended from him","kjv_def":"Mesech, Meshech."},
"H4903":{"lemma":"מִשְׁכַּב","xlit":"mishkab","pron":"mish-kab'","derivation":"(Aramaic) corresponding to H4904 (מִשְׁכָּב);","strongs_def":"a bed","kjv_def":"bed."},
"H4904":{"lemma":"מִשְׁכָּב","xlit":"mishkâb","pron":"mish-kawb'","derivation":"from H7901 (שָׁכַב);","strongs_def":"a bed (figuratively, a bier); abstractly, sleep; by euphemism, carnal intercourse","kjv_def":"bed(-chamber), couch, lieth (lying) with."},
"H4905":{"lemma":"מַשְׂכִּיל","xlit":"maskîyl","pron":"mas-keel'","derivation":"from H7919 (שָׂכַל);","strongs_def":"instructive, i.e. a didactic poem","kjv_def":"Maschil."},
"H4906":{"lemma":"מַשְׂכִּית","xlit":"maskîyth","pron":"mas-keeth'","derivation":"from the same as H7906 (שֵׂכוּ);","strongs_def":"a figure (carved on stone, the wall, or any object); figuratively, imagination","kjv_def":"conceit, image(-ry), picture, [idiom] wish."},
"H4907":{"lemma":"מִשְׁכַּן","xlit":"mishkan","pron":"mish-kan'","derivation":"(Aramaic) corresponding to H4908 (מִשְׁכָּן);","strongs_def":"residence","kjv_def":"habitation."},
"H4908":{"lemma":"מִשְׁכָּן","xlit":"mishkân","pron":"mish-kawn'","derivation":"from H7931 (שָׁכַן);","strongs_def":"a residence (including a shepherd's hut, the lair of animals, figuratively, the grave; also the Temple); specifically, the Tabernacle (properly, its wooden walls)","kjv_def":"dwelleth, dwelling (place), habitation, tabernacle, tent."},
"H4909":{"lemma":"מַשְׂכֹּרֶת","xlit":"maskôreth","pron":"mas-koh'-reth","derivation":"from H7936 (שָׂכַר);","strongs_def":"wages or a reward","kjv_def":"reward, wages."},
"H4910":{"lemma":"מָשַׁל","xlit":"mâshal","pron":"maw-shal'","derivation":"a primitive root;","strongs_def":"to rule","kjv_def":"(have, make to have) dominion, governor, [idiom] indeed, reign, (bear, cause to, have) rule(-ing, -r), have power."},
"H4911":{"lemma":"מָשַׁל","xlit":"mâshal","pron":"maw-shal'","derivation":"denominative from H4912 (מָשָׁל);","strongs_def":"to liken, i.e. (transitively) to use figurative language (an allegory, adage, song or the like); intransitively, to resemble","kjv_def":"be(-come) like, compare, use (as a) proverb, speak (in proverbs), utter."},
"H4912":{"lemma":"מָשָׁל","xlit":"mâshâl","pron":"maw-shawl'","derivation":"apparently from H4910 (מָשַׁל) in some original sense of superiority in mental action;","strongs_def":"properly, a pithy maxim, usually of metaphorical nature; hence, a simile (as an adage, poem, discourse)","kjv_def":"byword, like, parable, proverb."},
"H4913":{"lemma":"מָשָׁל","xlit":"Mâshâl","pron":"maw-shawl'","derivation":"for H4861 (מִשְׁאָל);","strongs_def":"Mashal, a place in Palestine","kjv_def":"Mashal."},
"H4914":{"lemma":"מְשֹׁל","xlit":"mᵉshôl","pron":"mesh-ol'","derivation":"from H4911 (מָשַׁל);","strongs_def":"a satire","kjv_def":"byword."},
"H4915":{"lemma":"מֹשֶׁל","xlit":"môshel","pron":"mo'-shel","derivation":"(1) from H4910 (מָשַׁל); (2) from H4911 (מָשַׁל);","strongs_def":"empire; a parallel","kjv_def":"dominion, like."},
"H4916":{"lemma":"מִשְׁלוֹחַ","xlit":"mishlôwach","pron":"mish-lo'-akh","derivation":"or מִשְׁלֹחַ; also מִשְׁלָח; from H7971 (שָׁלַח);","strongs_def":"a sending out, i.e. (abstractly) presentation (favorable), or seizure (unfavorable); also (concretely) a place of dismissal, or a business to be discharged","kjv_def":"to lay, to put, sending (forth), to set."},
"H4917":{"lemma":"מִשְׁלַחַת","xlit":"mishlachath","pron":"mish-lakh'-ath","derivation":"feminine of H4916 (מִשְׁלוֹחַ);","strongs_def":"a mission, i.e. (abstractly and favorable) release, or (concretely and unfavorable) an army","kjv_def":"discharge, sending."},
"H4918":{"lemma":"מְשֻׁלָּם","xlit":"Mᵉshullâm","pron":"mesh-ool-lawm'","derivation":"from H7999 (שָׁלַם); allied;","strongs_def":"Meshullam, the name of seventeen Israelites","kjv_def":"Meshullam."},
"H4919":{"lemma":"מְשִׁלֵּמוֹת","xlit":"Mᵉshillêmôwth","pron":"mesh-il-lay-mohth'","derivation":"plural from H7999 (שָׁלַם); reconciliations","strongs_def":"Meshillemoth, an Israelite","kjv_def":"Meshillemoth. Compare H4921 (מְשִׁלֵּמִית)."},
"H4920":{"lemma":"מְשֶׁלֶמְיָה","xlit":"Mᵉshelemyâh","pron":"mesh-eh-lem-yaw'","derivation":"or מְשֶׁלֶמְיָהוּ; from H7999 (שָׁלַם) and H3050 (יָהּ); ally of Jah;","strongs_def":"Meshelemjah, an Israelite","kjv_def":"Meshelemiah."},
"H4921":{"lemma":"מְשִׁלֵּמִית","xlit":"Mᵉshillêmîyth","pron":"mesh-il-lay-meeth'","derivation":"from H7999 (שָׁלַם); reconciliation;","strongs_def":"Meshillemith, an Israelite","kjv_def":"Meshillemith. Compare H4919 (מְשִׁלֵּמוֹת)."},
"H4922":{"lemma":"מְשֻׁלֶּמֶת","xlit":"Mᵉshullemeth","pron":"mesh-ool-leh'-meth'","derivation":"feminine of H4918 (מְשֻׁלָּם);","strongs_def":"Meshullemeth, an Israelitess","kjv_def":"Meshullemeth."},
"H4923":{"lemma":"מְשַׁמָּה","xlit":"mᵉshammâh","pron":"mesh-am-maw'","derivation":"from H8074 (שָׁמֵם);","strongs_def":"a waste or amazement","kjv_def":"astonishment, desolate."},
"H4924":{"lemma":"מַשְׁמָן","xlit":"mashmân","pron":"mash-mawn'","derivation":"from H8080 (שָׁמַן);","strongs_def":"fat, i.e. (literally and abstractly) fatness; but usually (figuratively and concretely) a rich dish, a fertile field, a robust man","kjv_def":"fat (one, -ness, -test, -test place)."},
"H4925":{"lemma":"מִשְׁמַנָּה","xlit":"Mishmannâh","pron":"mish-man-naw'","derivation":"from H8080 (שָׁמַן); fatness;","strongs_def":"Mashmannah, an Israelite","kjv_def":"Mishmannah."},
"H4926":{"lemma":"מִשְׁמָע","xlit":"mishmâʻ","pron":"mish-maw'","derivation":"from H8085 (שָׁמַע);","strongs_def":"a report","kjv_def":"hearing."},
"H4927":{"lemma":"מִשְׁמָע","xlit":"Mishmâʻ","pron":"mish-maw'","derivation":"the same as H4926 (מִשְׁמָע);","strongs_def":"Mishma, the name of a son of Ishmael, and of an Israelite","kjv_def":"Mishma."},
"H4928":{"lemma":"מִשְׁמַעַת","xlit":"mishmaʻath","pron":"mish-mah'-ath","derivation":"feminine of H4926 (מִשְׁמָע);","strongs_def":"audience, i.e. the royal court; also obedience, i.e. (concretely) a subject","kjv_def":"bidding, guard, obey."},
"H4929":{"lemma":"מִשְׁמָר","xlit":"mishmâr","pron":"mish-mawr'","derivation":"from H8104 (שָׁמַר);","strongs_def":"a guard (the man, the post or the prison); a deposit (figuratively); also (as observed) a usage (abstractly), or an example (concretely)","kjv_def":"diligence, guard, office, prison, ward, watch."},
"H4930":{"lemma":"מַשְׂמְרָה","xlit":"masmᵉrâh","pron":"mas-mer-aw'","derivation":"for H4548 (מַסְמֵר) fem.;","strongs_def":"a peg","kjv_def":"nail."},
"H4931":{"lemma":"מִשְׁמֶרֶת","xlit":"mishmereth","pron":"mish-mer'-reth","derivation":"feminine of H4929 (מִשְׁמָר);","strongs_def":"watch, i.e. the act (custody), or (concretely) the sentry, the post; objectively preservation, or (concretely) safe; figuratively observance, i.e. (abstractly) duty or (objectively) a usage or party","kjv_def":"charge, keep, or to be kept, office, ordinace, safeguard, ward, watch."},
"H4932":{"lemma":"מִשְׁנֶה","xlit":"mishneh","pron":"mish-neh'","derivation":"from H8138 (שָׁנָה);","strongs_def":"properly, a repetition, i.e. a duplicate (copy of a document), or a double (in amount); by implication, a second (in order, rank, age, quality or location)","kjv_def":"college, copy, double, fatlings, next, second (order), twice as much."},
"H4933":{"lemma":"מְשִׁסָּה","xlit":"mᵉshiççâh","pron":"mesh-is-saw'","derivation":"from H8155 (שָׁסַס);","strongs_def":"plunder","kjv_def":"booty, spoil."},
"H4934":{"lemma":"מִשְׁעוֹל","xlit":"mishʻôwl","pron":"mish-ole'","derivation":"from the same as H8168 (שֹׁעַל);","strongs_def":"a hollow, i.e. a narrow passage","kjv_def":"path."},
"H4935":{"lemma":"מִשְׁעִי","xlit":"mishʻîy","pron":"mish-ee'","derivation":"probably from H8159 (שָׁעָה);","strongs_def":"inspection","kjv_def":"to supple."},
"H4936":{"lemma":"מִשְׁעָם","xlit":"Mishʻâm","pron":"mish-awm'","derivation":"apparently from H8159 (שָׁעָה); inspection;","strongs_def":"Misham, an Israelite","kjv_def":"Misham."},
"H4937":{"lemma":"מִשְׁעֵן","xlit":"mishʻên","pron":"mish-ane'","derivation":"or מִשְׁעָן; from H8172 (שָׁעַן);","strongs_def":"a support (concretely), i.e. (figuratively) a protector or sustenance","kjv_def":"stay."},
"H4938":{"lemma":"מִשְׁעֵנָה","xlit":"mishʻênâh","pron":"mish-ay-naw'","derivation":"or מִשְׁעֶנֶת; feminine of H4937 (מִשְׁעֵן);","strongs_def":"support (abstractly), i.e. (figuratively) sustenance or (concretely) a walking-stick","kjv_def":"staff."},
"H4939":{"lemma":"מִשְׂפָּח","xlit":"mispâch","pron":"mis-pawkh'","derivation":"from H5596 (סָפַח);","strongs_def":"slaughter","kjv_def":"oppression."},
"H4940":{"lemma":"מִשְׁפָּחָה","xlit":"mishpâchâh","pron":"mish-paw-khaw'","derivation":"from H8192 (שָׁפָה) (compare H8198 (שִׁפְחָה));","strongs_def":"a family, i.e. circle of relatives; figuratively, a class (of persons), a species (of animals) or sort (of things); by extension a tribe or people","kjv_def":"family, kind(-red)."},
"H4941":{"lemma":"מִשְׁפָּט","xlit":"mishpâṭ","pron":"mish-pawt'","derivation":"from H8199 (שָׁפַט);","strongs_def":"properly, a verdict (favorable or unfavorable) pronounced judicially, especially a sentence or formal decree (human or (participant's) divine law, individual or collective), including the act, the place, the suit, the crime, and the penalty; abstractly, justice, including a participant's right or privilege (statutory or customary), or even a style","kjv_def":"[phrase] adversary, ceremony, charge, [idiom] crime, custom, desert, determination, discretion, disposing, due, fashion, form, to be judged, judgment, just(-ice, -ly), (manner of) law(-ful), manner, measure, (due) order, ordinance, right, sentence, usest, [idiom] worthy, [phrase] wrong."},
"H4942":{"lemma":"מִשְׁפָּת","xlit":"mishpâth","pron":"mish-pawth'","derivation":"from H8192 (שָׁפָה);","strongs_def":"a stall for cattle (only dual)","kjv_def":"burden, sheepfold."},
"H4943":{"lemma":"מֶשֶׁק","xlit":"mesheq","pron":"meh'-shek","derivation":"from an unused root meaning to hold;","strongs_def":"possession","kjv_def":"[phrase] steward."},
"H4944":{"lemma":"מַשָּׁק","xlit":"mashshâq","pron":"mash-shawk'","derivation":"from H8264 (שָׁקַק);","strongs_def":"a traversing, i.e. rapid motion","kjv_def":"running to and fro."},
"H4945":{"lemma":"מַשְׁקֶה","xlit":"mashqeh","pron":"mash-keh'","derivation":"from H8248 (שָׁקָה);","strongs_def":"properly, causing to drink, i.e. a butler; by implication (intransitively), drink (itself); figuratively, a well-watered region","kjv_def":"butler(-ship), cupbearer, drink(-ing), fat pasture, watered."},
"H4946":{"lemma":"מִשְׁקוֹל","xlit":"mishqôwl","pron":"mish-kole'","derivation":"from H8254 (שָׁקַל);","strongs_def":"weight","kjv_def":"weight."},
"H4947":{"lemma":"מַשְׁקוֹף","xlit":"mashqôwph","pron":"mash-kofe'","derivation":"from H8259 (שָׁקַף) in its original sense of overhanging;","strongs_def":"a lintel","kjv_def":"lintel."},
"H4948":{"lemma":"מִשְׁקָל","xlit":"mishqâl","pron":"mish-kawl'","derivation":"from H8254 (שָׁקַל);","strongs_def":"weight (numerically estimated); hence, weighing (the act)","kjv_def":"(full) weight."},
"H4949":{"lemma":"מִשְׁקֶלֶת","xlit":"mishqeleth","pron":"mish-keh'-leth","derivation":"or מִשְׁקֹלֶת; feminine of H4948 (מִשְׁקָל) or H4947 (מַשְׁקוֹף);","strongs_def":"a weight, i.e. a plummet (with line attached)","kjv_def":"plummet."},
"H4950":{"lemma":"מִשְׁקָע","xlit":"mishqâʻ","pron":"mish-kaw'","derivation":"from H8257 (שָׁקַע);","strongs_def":"a settling place (of water), i.e. a pond","kjv_def":"deep."},
"H4951":{"lemma":"מִשְׂרָה","xlit":"misrâh","pron":"mis-raw'","derivation":"from H8280 (שָׂרָה);","strongs_def":"empire","kjv_def":"government."},
"H4952":{"lemma":"מִשְׁרָה","xlit":"mishrâh","pron":"mish-raw'","derivation":"from H8281 (שָׁרָה) in the sense of loosening;","strongs_def":"maceration, i.e. steeped juice","kjv_def":"liquor."},
"H4953":{"lemma":"מַשְׁרוֹקִי","xlit":"mashrôwqîy","pron":"mash-ro-kee'","derivation":"(Aramaic) from a root corresponding to H8319 (שָׁרַק);","strongs_def":"a (musical) pipe (from its whistling sound)","kjv_def":"flute."},
"H4954":{"lemma":"מִשְׁרָעִי","xlit":"Mishrâʻîy","pron":"mish-raw-ee'","derivation":"patrial from an unused noun from an unused root; probably meaning to stretch out; extension;","strongs_def":"a Mishraite, or inhabitant (collectively) of Mishra","kjv_def":"Mishraites."},
"H4955":{"lemma":"מִשְׂרָפָה","xlit":"misrâphâh","pron":"mis-raw-faw'","derivation":"from H8313 (שָׂרַף);","strongs_def":"combustion, i.e. cremation (of a corpse), or calcination (of lime)","kjv_def":"burning."},
"H4956":{"lemma":"מִשְׂרְפוֹת מַיִם","xlit":"Misrᵉphôwth mayim","pron":"mis-ref-ohth' mah'-yim","derivation":"from the plural of H4955 (מִשְׂרָפָה) and H4325 (מַיִם); burnings of water;","strongs_def":"Misrephoth-Majim, a place in Palestine","kjv_def":"Misrephoth-mayim."},
"H4957":{"lemma":"מַשְׂרֵקָה","xlit":"Masrêqâh","pron":"mas-ray-kaw'","derivation":"a form for H7796 (שׂוֹרֵק) used denominatively; vineyard;","strongs_def":"Masrekah, a place in Idumaea","kjv_def":"Masrekah."},
"H4958":{"lemma":"מַשְׂרֵת","xlit":"masrêth","pron":"mas-rayth'","derivation":"apparently from an unused root meaning to perforate, i.e. hollow out;","strongs_def":"a pan","kjv_def":"pan."},
"H4959":{"lemma":"מָשַׁשׁ","xlit":"mâshash","pron":"maw-shash'","derivation":"a primitive root;","strongs_def":"to feel of; by implication, to grope","kjv_def":"feel, grope, search."},
"H4960":{"lemma":"מִשְׁתֶּה","xlit":"mishteh","pron":"mish-teh'","derivation":"from H8354 (שָׁתָה);","strongs_def":"drink, by implication, drinking (the act); also (by implication) a banquet or (generally) feast","kjv_def":"banquet, drank, drink, feast((-ed), -ing)."},
"H4961":{"lemma":"מִשְׁתֶּה","xlit":"mishteh","pron":"mish-teh'","derivation":"(Aramaic) corresponding to H4960 (מִשְׁתֶּה);","strongs_def":"a banquet","kjv_def":"banquet."},
"H4962":{"lemma":"מַת","xlit":"math","pron":"math","derivation":"from the same as H4970 (מָתַי);","strongs_def":"properly, an adult (as of full length); by implication, a man (only in the plural)","kjv_def":"[phrase] few, [idiom] friends, men, persons, [idiom] small."},
"H4963":{"lemma":"מַתְבֵּן","xlit":"mathbên","pron":"math-bane'","derivation":"denominative from H8401 (תֶּבֶן);","strongs_def":"straw in the heap","kjv_def":"straw."},
"H4964":{"lemma":"מֶתֶג","xlit":"metheg","pron":"meh-theg","derivation":"from an unused root meaning to curb;","strongs_def":"a bit","kjv_def":"bit, bridle."},
"H4965":{"lemma":"מֶתֶג הָאַמָּה","xlit":"Metheg hâ-ʼAmmâh","pron":"meh'-theg haw-am-maw'","derivation":"from H4964 (מֶתֶג) and H520 (אַמָּה) with the art. interposed; bit of the metropolis;","strongs_def":"Metheg-ha-Ammah, an epithet of Gath","kjv_def":"Metheg-ammah."},
"H4966":{"lemma":"מָתוֹק","xlit":"mâthôwq","pron":"maw-thoke'","derivation":"or מָתוּק; from H4985 (מָתַק);","strongs_def":"sweet","kjv_def":"sweet(-er, -ness)."},
"H4967":{"lemma":"מְתוּשָׁאֵל","xlit":"Mᵉthûwshâʼêl","pron":"meth-oo-shaw-ale'","derivation":"from H4962 (מַת) and H410 (אֵל), with the relative interposed; man who (is) of God;","strongs_def":"Methusael, an antediluvian patriarch","kjv_def":"Methusael."},
"H4968":{"lemma":"מְתוּשֶׁלַח","xlit":"Mᵉthûwshelach","pron":"meth-oo-sheh'-lakh","derivation":"from H4962 (מַת) and H7973 (שֶׁלַח); man of a dart;","strongs_def":"Methushelach, an antediluvian patriarch","kjv_def":"Methuselah."},
"H4969":{"lemma":"מָתַח","xlit":"mâthach","pron":"maw-thakh'","derivation":"a primitive root;","strongs_def":"to stretch out","kjv_def":"spread out."},
"H4970":{"lemma":"מָתַי","xlit":"mâthay","pron":"maw-thah'ee","derivation":"from an unused root meaning to extend;","strongs_def":"properly, extent (of time); but used only adverbially (especially with other particle prefixes), when (either relative or interrogative)","kjv_def":"long, when."},
"H4971":{"lemma":"מַתְכֹנֶת","xlit":"mathkôneth","pron":"math-ko'-neth","derivation":"or מַתְכֻנֶת; from H8505 (תָּכַן) in the transferred sense of measuring;","strongs_def":"proportion (in size, number or ingredients)","kjv_def":"composition, measure, state, tale."},
"H4972":{"lemma":"מַתְּלָאָה","xlit":"mattᵉlâʼâh","pron":"mat-tel-aw-aw'","derivation":"from H4100 (מָה) and H8513 (תְּלָאָה);","strongs_def":"what a trouble!","kjv_def":"what a weariness."},
"H4973":{"lemma":"מְתַלְּעָה","xlit":"mᵉthallᵉʻâh","pron":"meth-al-leh-aw'","derivation":"contr. from H3216 (יָלַע);","strongs_def":"properly, a biter, i.e. a tooth","kjv_def":"cheek (jaw) tooth, jaw."},
"H4974":{"lemma":"מְתֹם","xlit":"mᵉthôm","pron":"meth-ohm'","derivation":"from H8552 (תָּמַם);","strongs_def":"wholesomeness; also (adverb) completely","kjv_def":"men (by reading H4962 (מַת)), soundness."},
"H4975":{"lemma":"מֹתֶן","xlit":"môthen","pron":"mo'-then","derivation":"from an unused root meaning to be slender;","strongs_def":"properly, the waist or small of the back; only in plural the loins","kjv_def":"[phrase] greyhound, loins, side."},
"H4976":{"lemma":"מַתָּן","xlit":"mattân","pron":"mat-tawn'","derivation":"from H5414 (נָתַן);","strongs_def":"a present","kjv_def":"gift, to give, reward."},
"H4977":{"lemma":"מַתָּן","xlit":"Mattân","pron":"mat-tawn'","derivation":"the same as H4976 (מַתָּן);","strongs_def":"Mattan, the name of a priest of Baal, and of an Israelite","kjv_def":"Mattan."},
"H4978":{"lemma":"מַתְּנָא","xlit":"mattᵉnâʼ","pron":"mat-ten-aw'","derivation":"(Aramaic) corresponding to H4979 (מַתָּנָה)","strongs_def":"{a present; specifically (in a good sense), a sacrificial offering, (in a bad sense) a bribe}","kjv_def":"gift."},
"H4979":{"lemma":"מַתָּנָה","xlit":"mattânâh","pron":"mat-taw-naw'","derivation":"feminine of H4976 (מַתָּן);","strongs_def":"a present; specifically (in a good sense), a sacrificial offering, (in a bad sense) a bribe","kjv_def":"gift."},
"H4980":{"lemma":"מַתָּנָה","xlit":"Mattânâh","pron":"mat-taw-naw'","derivation":"the same as H4979 (מַתָּנָה);","strongs_def":"Mattanah, a place in the Desert","kjv_def":"Mattanah."},
"H4981":{"lemma":"מִתְנִי","xlit":"Mithnîy","pron":"mith-nee'","derivation":"probably patrial from an unused noun meaning slenderness;","strongs_def":"a Mithnite, or inhabitant of Methen","kjv_def":"Mithnite."},
"H4982":{"lemma":"מַתְּנַי","xlit":"Mattᵉnay","pron":"mat-ten-ah'ee","derivation":"from H4976 (מַתָּן); liberal;","strongs_def":"Mattenai, the name of three Israelites","kjv_def":"Mattenai."},
"H4983":{"lemma":"מַתַּנְיָה","xlit":"Mattanyâh","pron":"mat-tan-yaw'","derivation":"or מַתַּנְיָהוּ; from H4976 (מַתָּן) and H3050 (יָהּ); gift of Jah;","strongs_def":"Mattanjah, the name of ten Israelites","kjv_def":"Mattaniah."},
"H4984":{"lemma":"מִתְנַשֵּׂא","xlit":"mithnassêʼ","pron":"mith-nas-say'","derivation":"from H5375 (נָשָׂא);","strongs_def":"(used as abstractly) supreme exaltation","kjv_def":"exalted."},
"H4985":{"lemma":"מָתַק","xlit":"mâthaq","pron":"maw-thak'","derivation":"a primitive root;","strongs_def":"to suck, by implication, to relish, or (intransitively) be sweet","kjv_def":"be (made, [idiom] take) sweet."},
"H4986":{"lemma":"מֶתֶק","xlit":"metheq","pron":"meh'-thek","derivation":"from H4985 (מָתַק);","strongs_def":"figuratively, pleasantness (of discourse)","kjv_def":"sweetness."},
"H4987":{"lemma":"מֹתֶק","xlit":"môtheq","pron":"mo'-thek","derivation":"from H4985 (מָתַק);","strongs_def":"sweetness","kjv_def":"sweetness."},
"H4988":{"lemma":"מָתָק","xlit":"mâthâq","pron":"maw-thawk'","derivation":"from H4985 (מָתַק);","strongs_def":"a dainty, i.e. (generally) food","kjv_def":"feed sweetly."},
"H4989":{"lemma":"מִתְקָה","xlit":"Mithqâh","pron":"mith-kaw'","derivation":"feminine of H4987 (מֹתֶק); sweetness;","strongs_def":"Mithkah, a place in the Desert","kjv_def":"Mithcah."},
"H4990":{"lemma":"מִתְרְדָת","xlit":"Mithrᵉdâth","pron":"mith-red-awth'","derivation":"of Persian origin;","strongs_def":"Mithredath, the name of two Persians","kjv_def":"Mithredath."},
"H4991":{"lemma":"מַתָּת","xlit":"mattâth","pron":"mat-tawth'","derivation":"feminine of H4976 (מַתָּן) abbreviated form;","strongs_def":"a present","kjv_def":"gift."},
"H4992":{"lemma":"מַתַּתָּה","xlit":"Mattattâh","pron":"mat-tat-taw'","derivation":"for H4993 (מַתִּתְיָה); gift of Jah;","strongs_def":"Mattattah, an Israelite","kjv_def":"Mattathah."},
"H4993":{"lemma":"מַתִּתְיָה","xlit":"Mattithyâh","pron":"mat-tith-yaw'","derivation":"or מַתִּתְיָהוּ; from H4991 (מַתָּת) and H3050 (יָהּ); gift of Jah;","strongs_def":"Mattithjah, the name of four Israelites","kjv_def":"Mattithiah. n"},
"H4994":{"lemma":"נָא","xlit":"nâʼ","pron":"naw","derivation":"a primitive particle of incitement and entreaty, which may usually be rendered;","strongs_def":"'I pray', 'now', or 'then'; added mostly to verbs (in the Imperative or Future), or to interjections, occasionally to an adverb or conjunction","kjv_def":"I beseech (pray) thee (you), go to, now, oh."},
"H4995":{"lemma":"נָא","xlit":"nâʼ","pron":"naw","derivation":"apparently from H5106 (נוּא) in the sense of harshness from refusal;","strongs_def":"properly, tough, i.e. uncooked (flesh)","kjv_def":"raw."},
"H4996":{"lemma":"נֹא","xlit":"Nôʼ","pron":"no","derivation":"of Egyptian origin;","strongs_def":"No (i.e. Thebes), the capital of Upper Egypt","kjv_def":"No. Compare H528 (אָמוֹן)."},
"H4997":{"lemma":"נֹאד","xlit":"nôʼd","pron":"node","derivation":"or נאוֹדlemma נֹאוד vowel, corrected to נאוֹדxlit nôʼwd corrected to nʼôwd; also (feminine) נֹאדָה; from an unused root of uncertain signification;","strongs_def":"a (skin or leather) bag (for fluids)","kjv_def":"bottle."},
"H4998":{"lemma":"נָאָה","xlit":"nâʼâh","pron":"naw-aw'","derivation":"a primitive root;","strongs_def":"properly, to be at home, i.e. (by implication) to be pleasant (or suitable), i.e. beautiful","kjv_def":"be beautiful, become, be comely."},
"H4999":{"lemma":"נָאָה","xlit":"nâʼâh","pron":"naw-aw'","derivation":"from H4998 (נָאָה);","strongs_def":"a home; figuratively, a pasture","kjv_def":"habitation, house, pasture, pleasant place."},
"H5000":{"lemma":"נָאוֶה","xlit":"nâʼveh","pron":"naw-veh'","derivation":"from H4998 (נָאָה) or H5116 (נָוֶה);","strongs_def":"suitable, or beautiful","kjv_def":"becometh, comely, seemly."},
"H5001":{"lemma":"נָאַם","xlit":"nâʼam","pron":"naw-am'","derivation":"a primitive root;","strongs_def":"properly, to whisper, i.e. (by implication) to utter as aoracle","kjv_def":"say."},
"H5002":{"lemma":"נְאֻם","xlit":"nᵉʼum","pron":"neh-oom'","derivation":"from H5001 (נָאַם);","strongs_def":"an oracle","kjv_def":"(hath) said, saith."},
"H5003":{"lemma":"נָאַף","xlit":"nâʼaph","pron":"naw-af'","derivation":"a primitive root;","strongs_def":"to commit adultery; figuratively, to apostatize","kjv_def":"adulterer(-ess), commit(-ing) adultery, woman that breaketh wedlock."},
"H5004":{"lemma":"נִאֻף","xlit":"niʼuph","pron":"nee-oof'","derivation":"from H5003 (נָאַף);","strongs_def":"adultery","kjv_def":"adultery."},
"H5005":{"lemma":"נַאֲפוּף","xlit":"naʼăphûwph","pron":"nah-af-oof'","derivation":"from H5003 (נָאַף);","strongs_def":"adultery","kjv_def":"adultery."},
"H5006":{"lemma":"נָאַץ","xlit":"nâʼats","pron":"naw-ats'","derivation":"a primitive root; in Ecclesiastes 12:5, by interchange for H5132 (נוּץ), to bloom","strongs_def":"to scorn;","kjv_def":"abhor, (give occasion to) blaspheme, contemn, despise, flourish, [idiom] great, provoke."},
"H5007":{"lemma":"נְאָצָה","xlit":"nᵉʼâtsâh","pron":"neh-aw-tsaw'","derivation":"or נֶאָצָה; from H5006 (נָאַץ);","strongs_def":"scorn","kjv_def":"blasphemy."},
"H5008":{"lemma":"נָאַק","xlit":"nâʼaq","pron":"naw-ak'","derivation":"a primitive root;","strongs_def":"to groan","kjv_def":"groan."},
"H5009":{"lemma":"נְאָקָה","xlit":"nᵉʼâqâh","pron":"neh-aw-kaw'","derivation":"from H5008 (נָאַק);","strongs_def":"a groan","kjv_def":"groaning."},
"H5010":{"lemma":"נָאַר","xlit":"nâʼar","pron":"naw-ar'","derivation":"a primitive root;","strongs_def":"to reject","kjv_def":"abhor, make void."},
"H5011":{"lemma":"נֹב","xlit":"Nôb","pron":"nobe","derivation":"the same as H5108 (נוֹב); fruit;","strongs_def":"Nob, a place in Palestine","kjv_def":"Nob."},
"H5012":{"lemma":"נָבָא","xlit":"nâbâʼ","pron":"naw-baw'","derivation":"a primitive root;","strongs_def":"to prophesy, i.e. speak (or sing) by inspiration (in prediction or simple discourse)","kjv_def":"prophesy(-ing), make self a prophet."},
"H5013":{"lemma":"נְבָא","xlit":"nᵉbâʼ","pron":"neb-aw'","derivation":"(Aramaic) corresponding to H5012 (נָבָא)","strongs_def":"{to prophesy, i.e. speak (or sing) by inspiration (in prediction or simple discourse) to prophesy, i.e. speak (or sing) by inspiration (in prediction or simple discourse)}","kjv_def":"prophesy."},
"H5014":{"lemma":"נָבַב","xlit":"nâbab","pron":"naw-bab'","derivation":"a primitive root;","strongs_def":"to pierce; to be hollow, or (figuratively) foolish","kjv_def":"hollow, vain."},
"H5015":{"lemma":"נְבוֹ","xlit":"Nᵉbôw","pron":"neb-o'","derivation":"probably of foreign derivation;","strongs_def":"Nebo, the name of a Babylonian deity, also of a mountain in Moab, and of a place in Palestine","kjv_def":"Nebo."},
"H5016":{"lemma":"נְבוּאָה","xlit":"nᵉbûwʼâh","pron":"neb-oo-aw'","derivation":"from H5012 (נָבָא);","strongs_def":"a prediction (spoken or written)","kjv_def":"prophecy."},
"H5017":{"lemma":"נְבוּאָה","xlit":"nᵉbûwʼâh","pron":"neb-oo-aw","derivation":"(Aramaic) corresponding to H5016 (נְבוּאָה);","strongs_def":"inspired teaching","kjv_def":"prophesying."},
"H5018":{"lemma":"נְבוּזַרְאֲדָן","xlit":"Nᵉbûwzarʼădân","pron":"neb-oo-zar-ad-awn'","derivation":"of foreign origin;","strongs_def":"Nebuzaradan, a Babylonian general","kjv_def":"Nebuzaradan."},
"H5019":{"lemma":"נְבוּכַדְנֶאצַּר","xlit":"Nᵉbûwkadneʼtstsar","pron":"neb-oo-kad-nets-tsar'","derivation":"or נְבֻּכַדְנֶאצַּר; (2 Kings 24:1,10), or נְבוּכַדְנֶצַּר; (Esther 2:6; Daniel 1:18), or נְבוּכַדְרֶאצַּר; or נְבוּכַדְרֶאצּוֹר; (Ezra 2:1; Jeremiah 49:28), or foreign derivation;","strongs_def":"Nebukadnetstsar (or -retstsar, or -retstsor), king of Babylon","kjv_def":"Nebuchadnezzar, Nebuchadrezzar."},
"H5020":{"lemma":"נְבוּכַדְנֶצַּר","xlit":"Nᵉbûwkadnetstsar","pron":"neb-oo-kad-nets-tsar'","derivation":"(Aramaic) corresponding to H5019 (נְבוּכַדְנֶאצַּר)","strongs_def":"{Nebukadnetstsar (or -retstsar, or -retstsor), king of Babylon}","kjv_def":"Nebuchadnezzar."},
"H5021":{"lemma":"נְבוּשַׁזְבָּן","xlit":"Nᵉbûwshazbân","pron":"neb-oo-shaz-bawn'","derivation":"of foreign derivation;","strongs_def":"Nebushazban, Nebuchadnezzar's chief eunuch","kjv_def":"Nebushazban."},
"H5022":{"lemma":"נָבוֹת","xlit":"Nâbôwth","pron":"naw-both'","derivation":"feminine plural from the same as H5011 (נֹב); fruits;","strongs_def":"Naboth, an Israelite","kjv_def":"Naboth."},
"H5023":{"lemma":"נְבִזְבָּה","xlit":"nᵉbizbâh","pron":"neb-iz-baw'","derivation":"(Aramaic) of uncertain derivation;","strongs_def":"a largess","kjv_def":"reward."},
"H5024":{"lemma":"נָבַח","xlit":"nâbach","pron":"naw-bakh'","derivation":"a primitive root;","strongs_def":"to bark (as a dog)","kjv_def":"bark."},
"H5025":{"lemma":"נֹבַח","xlit":"Nôbach","pron":"no'-bach","derivation":"from H5024 (נָבַח); a bark;","strongs_def":"Nobach, the name of an Israelite, and of a place East of the Jordan","kjv_def":"Nobah."},
"H5026":{"lemma":"נִבְחַז","xlit":"Nibchaz","pron":"nib-khaz'","derivation":"of foreign origin;","strongs_def":"Nibchaz, a deity of the Avites","kjv_def":"Nibhaz."},
"H5027":{"lemma":"נָבַט","xlit":"nâbaṭ","pron":"naw-bat'","derivation":"a primitive root;","strongs_def":"to scan, i.e. look intently at; by implication, to regard with pleasure, favor or care","kjv_def":"(cause to) behold, consider, look (down), regard, have respect, see."},
"H5028":{"lemma":"נְבָט","xlit":"Nᵉbâṭ","pron":"neb-awt'","derivation":"from H5027 (נָבַט); regard;","strongs_def":"Nebat, the father of Jeroboam I","kjv_def":"Nebat."},
"H5029":{"lemma":"נְבִיא","xlit":"nᵉbîyʼ","pron":"neb-ee'","derivation":"(Aramaic) corresponding to H5030 (נָבִיא);","strongs_def":"a prophet","kjv_def":"prophet."},
"H5030":{"lemma":"נָבִיא","xlit":"nâbîyʼ","pron":"naw-bee'","derivation":"from H5012 (נָבָא);","strongs_def":"a prophet or (generally) inspired man","kjv_def":"prophecy, that prophesy, prophet."},
"H5031":{"lemma":"נְבִיאָה","xlit":"nᵉbîyʼâh","pron":"neb-ee-yaw'","derivation":"feminine of H5030 (נָבִיא);","strongs_def":"a prophetess or (generally) inspired woman; by implication, a poetess; by association a prophet's wife","kjv_def":"prophetess."},
"H5032":{"lemma":"נְבָיוֹת","xlit":"Nᵉbâyôwth","pron":"neb-aw-yoth'","derivation":"or נְבָיֹת; feminine plural from H5107 (נוּב); fruitfulnesses;","strongs_def":"Nebajoth, a son of Ismael, and the country settled by him","kjv_def":"Nebaioth, Nebajoth."},
"H5033":{"lemma":"נֵבֶךְ","xlit":"nêbek","pron":"nay'-bek","derivation":"from an unused root meaning to burst forth;","strongs_def":"a fountain","kjv_def":"spring."},
"H5034":{"lemma":"נָבֵל","xlit":"nâbêl","pron":"naw-bale'","derivation":"a primitive root;","strongs_def":"to wilt; generally, to fall away, fail, faint; figuratively, to be foolish or (morally) wicked; causatively, to despise, disgrace","kjv_def":"disgrace, dishounour, lightly esteem, fade (away, -ing), fall (down, -ling, off), do foolishly, come to nought, [idiom] surely, make vile, wither."},
"H5035":{"lemma":"נֶבֶל","xlit":"nebel","pron":"neh'-bel","derivation":"or נֵבֶל; from H5034 (נָבֵל);","strongs_def":"a skin-bag for liquids (from collapsing when empty); hence, a vase (as similar in shape when full); also a lyre (as having a body of like form)","kjv_def":"bottle, pitcher, psaltery, vessel, viol."},
"H5036":{"lemma":"נָבָל","xlit":"nâbâl","pron":"naw-bawl'","derivation":"from H5034 (נָבֵל);","strongs_def":"stupid; wicked (especially impious)","kjv_def":"fool(-ish, -ish man, -ish woman), vile person."},
"H5037":{"lemma":"נָבָל","xlit":"Nâbâl","pron":"naw-bawl'","derivation":"the same as H5036 (נָבָל); dolt;","strongs_def":"Nabal, an Israelite","kjv_def":"Nabal."},
"H5038":{"lemma":"נְבֵלָה","xlit":"nᵉbêlâh","pron":"neb-ay-law'","derivation":"from H5034 (נָבֵל);","strongs_def":"a flabby thing, i.e. a carcase or carrion (human or bestial, often collectively); figuratively, an idol","kjv_def":"(dead) body, (dead) carcase, dead of itself, which died, (beast) that (which) dieth of itself."},
"H5039":{"lemma":"נְבָלָה","xlit":"nᵉbâlâh","pron":"neb-aw-law'","derivation":"feminine of H5036 (נָבָל);","strongs_def":"foolishness, i.e. (morally) wickedness; concretely, a crime; by extension, punishment","kjv_def":"folly, vile, villany."},
"H5040":{"lemma":"נַבְלוּת","xlit":"nablûwth","pron":"nab-looth'","derivation":"from H5036 (נָבָל);","strongs_def":"properly, disgrace, i.e. the (female) pudenda","kjv_def":"lewdness."},
"H5041":{"lemma":"נְבַלָּט","xlit":"Nᵉballâṭ","pron":"neb-al-lawt'","derivation":"apparently from H5036 (נָבָל) and H3909 (לָט); foolish secrecy;","strongs_def":"Neballat, a place in Palestine","kjv_def":"Neballat."},
"H5042":{"lemma":"נָבַע","xlit":"nâbaʻ","pron":"naw-bah'","derivation":"a primitive root;","strongs_def":"to gush forth; figuratively, to utter (good or bad words); specifically, to emit (a foul odor)","kjv_def":"belch out, flowing, pour out, send forth, utter (abundantly)."},
"H5043":{"lemma":"נֶבְרְשָׁא","xlit":"nebrᵉshâʼ","pron":"neb-reh-shaw'","derivation":"(Aramaic) from an unused root meaning to shine;","strongs_def":"a light; plural (collectively) a chandelier","kjv_def":"candlestick."},
"H5044":{"lemma":"נִבְשָׁן","xlit":"Nibshân","pron":"nib-shawn'","derivation":"of uncertain derivation;","strongs_def":"Nibshan, a place in Palestine","kjv_def":"Nibshan."},
"H5045":{"lemma":"נֶגֶב","xlit":"negeb","pron":"neh'-gheb","derivation":"from an unused root meaning to be parched;","strongs_def":"the south (from its drought); specifically, the Negeb or southern district of Judah, occasionally, Egypt (as south to Palestine)","kjv_def":"south (country, side, -ward)."},
"H5046":{"lemma":"נָגַד","xlit":"nâgad","pron":"naw-gad'","derivation":"a primitive root;","strongs_def":"properly, to front, i.e. stand boldly out opposite; by implication (causatively), to manifest; figuratively, to announce (always by word of mouth to one present); specifically, to expose, predict, explain, praise","kjv_def":"bewray, [idiom] certainly, certify, declare(-ing), denounce, expound, [idiom] fully, messenger, plainly, profess, rehearse, report, shew (forth), speak, [idiom] surely, tell, utter."},
"H5047":{"lemma":"נְגַד","xlit":"nᵉgad","pron":"neg-ad'","derivation":"(Aramaic) corresponding to H5046 (נָגַד);","strongs_def":"to flow (through the idea of clearing the way)","kjv_def":"issue."},
"H5048":{"lemma":"נֶגֶד","xlit":"neged","pron":"neh'-ghed","derivation":"from H5046 (נָגַד);","strongs_def":"a front, i.e. part opposite; specifically a counterpart, or mate; usually (adverbial, especially with preposition) over against or before","kjv_def":"about, (over) against, [idiom] aloof, [idiom] far (off), [idiom] from, over, presence, [idiom] other side, sight, [idiom] to view."},
"H5049":{"lemma":"נֶגֶד","xlit":"neged","pron":"neh'-ghed","derivation":"(Aramaic) corresponding to H5048 (נֶגֶד);","strongs_def":"opposite","kjv_def":"toward."},
"H5050":{"lemma":"נָגַהּ","xlit":"nâgahh","pron":"naw-gah'","derivation":"a primitive root;","strongs_def":"to glitter; causatively, to illuminate","kjv_def":"(en-) lighten, (cause to) shine."},
"H5051":{"lemma":"נֹגַהּ","xlit":"nôgahh","pron":"no'-gah","derivation":"from H5050 (נָגַהּ);","strongs_def":"brilliancy (literally or figuratively)","kjv_def":"bright(-ness), light, (clear) shining."},
"H5052":{"lemma":"נֹגַהּ","xlit":"Nôgahh","pron":"no'-gah","derivation":"the same as H5051 (נֹגַהּ);","strongs_def":"Nogah, a son of David","kjv_def":"Nogah."},
"H5053":{"lemma":"נֹגַהּ","xlit":"nôgahh","pron":"no'-gah","derivation":"(Aramaic) corresponding to H5051 (נֹגַהּ);","strongs_def":"dawn","kjv_def":"morning."},
"H5054":{"lemma":"נְגֹהָה","xlit":"nᵉgôhâh","pron":"neg-o-haw'","derivation":"feminine of H5051 (נֹגַהּ);","strongs_def":"splendor","kjv_def":"brightness."},
"H5055":{"lemma":"נָגַח","xlit":"nâgach","pron":"naw-gakh'","derivation":"a primitive root;","strongs_def":"to butt with the horns; figuratively, to war against","kjv_def":"gore, push (down, -ing)."},
"H5056":{"lemma":"נַגָּח","xlit":"naggâch","pron":"nag-gawkh'","derivation":"from H5055 (נָגַח);","strongs_def":"butting, i.e. vicious","kjv_def":"used (wont) to push."},
"H5057":{"lemma":"נָגִיד","xlit":"nâgîyd","pron":"naw-gheed'","derivation":"or נָגִד; from H5046 (נָגַד);","strongs_def":"a commander (as occupying the front), civil, military or religious; generally (abstractly, plural), honorable themes","kjv_def":"captain, chief, excellent thing, (chief) governor, leader, noble, prince, (chief) ruler."},
"H5058":{"lemma":"נְגִינָה","xlit":"nᵉgîynâh","pron":"neg-ee-naw'","derivation":"or נְגִינַת; (Psalm 61:title), from H5059 (נָגַן);","strongs_def":"properly, instrumental music; by implication, a stringed instrument; by extension, a poem set to music; specifically, an epigram","kjv_def":"stringed instrument, musick, Neginoth (plural), song."},
"H5059":{"lemma":"נָגַן","xlit":"nâgan","pron":"naw-gan'","derivation":"a primitive root;","strongs_def":"properly, to thrum, i.e. beat atune with the fingers; expectation. to play on astringed instrument; hence (generally), to make music","kjv_def":"player on instruments, sing to the stringed instruments, melody, ministrel, play(-er, -ing)."},
"H5060":{"lemma":"נָגַע","xlit":"nâgaʻ","pron":"naw-gah'","derivation":"a primitive root;","strongs_def":"properly, to touch, i.e. lay the hand upon (for any purpose; euphemistically, to lie with a woman); by implication, to reach (figuratively, to arrive, acquire); violently, to strike (punish, defeat, destroy, etc.)","kjv_def":"beat, ([idiom] be able to) bring (down), cast, come (nigh), draw near (nigh), get up, happen, join, near, plague, reach (up), smite, strike, touch."},
"H5061":{"lemma":"נֶגַע","xlit":"negaʻ","pron":"neh'-gah","derivation":"from H5060 (נָגַע);","strongs_def":"a blow (figuratively, infliction); also (by implication) a spot (concretely, a leprous person or dress)","kjv_def":"plague, sore, stricken, stripe, stroke, wound."},
"H5062":{"lemma":"נָגַף","xlit":"nâgaph","pron":"naw-gaf'","derivation":"a primitive root;","strongs_def":"to push, gore, defeat, stub (the toe), inflict (a disease)","kjv_def":"beat, dash, hurt, plague, slay, smite (down), strike, stumble, [idiom] surely, put to the worse."},
"H5063":{"lemma":"נֶגֶף","xlit":"negeph","pron":"neh'-ghef","derivation":"from H5062 (נָגַף);","strongs_def":"a trip (of the foot); figuratively, an infliction (of disease)","kjv_def":"plague, stumbling."},
"H5064":{"lemma":"נָגַר","xlit":"nâgar","pron":"naw-gar'","derivation":"a primitive root;","strongs_def":"to flow; figuratively, to stretch out; causatively, to pour out or down; figuratively, to deliver over","kjv_def":"fall, flow away, pour down (out), run, shed, spilt, trickle down."},
"H5065":{"lemma":"נָגַשׂ","xlit":"nâgas","pron":"naw-gas'","derivation":"a primitive root;","strongs_def":"to drive (an animal, a workman, a debtor, an army); by implication, to tax, harass, tyrannize","kjv_def":"distress, driver, exact(-or), oppress(-or), [idiom] raiser of taxes, taskmaster."},
"H5066":{"lemma":"נָגַשׁ","xlit":"nâgash","pron":"naw-gash'","derivation":"a primitive root;","strongs_def":"to be or come (causatively, bring) near (for any purpose); euphemistically, to lie with a woman; as an enemy, to attack; religious to worship; causatively, to present; figuratively, to adduce an argument; by reversal, to stand back","kjv_def":"(make to) approach (nigh), bring (forth, hither, near), (cause to) come (hither, near, nigh), give place, go hard (up), (be, draw, go) near (nigh), offer, overtake, present, put, stand."},
"H5067":{"lemma":"נֵד","xlit":"nêd","pron":"nade","derivation":"from H5110 (נוּד) in the sense of piling up;","strongs_def":"a mound, i.e. wave","kjv_def":"heap."},
"H5068":{"lemma":"נָדַב","xlit":"nâdab","pron":"naw-dab'","derivation":"a primitive root;","strongs_def":"to impel; hence, to volunteer (as a soldier), to present spontaneously","kjv_def":"offer freely, be (give, make, offer self) willing(-ly)."},
"H5069":{"lemma":"נְדַב","xlit":"nᵉdab","pron":"ned-ab'","derivation":"(Aramaic) corresponding to H5068 (נָדַב);","strongs_def":"be (or give) liberal(-ly)","kjv_def":"(be minded of...own) freewill (offering), offer freely (willingly)."},
"H5070":{"lemma":"נָדָב","xlit":"Nâdâb","pron":"naw-dawb'","derivation":"from H5068 (נָדַב); liberal;","strongs_def":"Nadab, the name of four Israelites","kjv_def":"Nadab."},
"H5071":{"lemma":"נְדָבָה","xlit":"nᵉdâbâh","pron":"ned-aw-baw'","derivation":"from H5068 (נָדַב);","strongs_def":"properly (abstractly) spontaneity, or (adjectively) spontaneous; also (concretely) a spontaneous or (by inference, in plural) abundant gift","kjv_def":"free(-will) offering, freely, plentiful, voluntary(-ily, offering), willing(-ly), offering)."},
"H5072":{"lemma":"נְדַבְיָה","xlit":"Nᵉdabyâh","pron":"ned-ab-yaw'","derivation":"from H5068 (נָדַב) and H3050 (יָהּ); largess of Jah;","strongs_def":"Nedabjah, an Israelite","kjv_def":"Nedabiah."},
"H5073":{"lemma":"נִדְבָּךְ","xlit":"nidbâk","pron":"nid-bawk'","derivation":"(Aramaic) from a root meaning to stick;","strongs_def":"a layer (of building materials)","kjv_def":"row."},
"H5074":{"lemma":"נָדַד","xlit":"nâdad","pron":"naw-dad'","derivation":"a primitive root;","strongs_def":"properly, to wave to and fro (rarely to flap up and down); figuratively, to rove, flee, or (causatively) to drive away","kjv_def":"chase (away), [idiom] could not, depart, flee ([idiom] apace, away), (re-) move, thrust away, wander (abroad, -er, -ing)."},
"H5075":{"lemma":"נְדַד","xlit":"nᵉdad","pron":"ned-ad'","derivation":"(Aramaic) corresponding to H5074 (נָדַד);","strongs_def":"to depart","kjv_def":"go from."},
"H5076":{"lemma":"נָדֻד","xlit":"nâdud","pron":"naw-dood'","derivation":"passive participle of H5074 (נָדַד);","strongs_def":"properly, tossed; abstractly, a rolling (on the bed)","kjv_def":"tossing to and fro."},
"H5077":{"lemma":"נָדָה","xlit":"nâdâh","pron":"naw-daw'","derivation":"or נָדָא; (2 Kings 17:21), a primitive root;","strongs_def":"properly, to toss; figuratively, to exclude, i.e. banish, postpone, prohibit","kjv_def":"cast out, drive, put far away."},
"H5078":{"lemma":"נֵדֶה","xlit":"nêdeh","pron":"nay'-deh","derivation":"from H5077 (נָדָה) in the sense of freely flinging money;","strongs_def":"a bounty (for prostitution)","kjv_def":"gifts."},
"H5079":{"lemma":"נִדָּה","xlit":"niddâh","pron":"nid-daw'","derivation":"from H5074 (נָדַד);","strongs_def":"properly, rejection; by implication, impurity, especially personal (menstruation) or moral (idolatry, incest)","kjv_def":"[idiom] far, filthiness, [idiom] flowers, menstruous (woman), put apart, [idiom] removed (woman), separation, set apart, unclean(-ness, thing, with filthiness)."},
"H5080":{"lemma":"נָדַח","xlit":"nâdach","pron":"naw-dakh'","derivation":"a primitive root;","strongs_def":"to push off; used in a great variety of applications, literally and figuratively (to expel, mislead, strike, inflict, etc.)","kjv_def":"banish, bring, cast down (out), chase, compel, draw away, drive (away, out, quite), fetch a stroke, force, go away, outcast, thrust away (out), withdraw."},
"H5081":{"lemma":"נָדִיב","xlit":"nâdîyb","pron":"naw-deeb'","derivation":"from H5068 (נָדַב);","strongs_def":"properly, voluntary, i.e. generous; hence, magnanimous; as noun, a grandee (sometimes a tyrant)","kjv_def":"free, liberal (things), noble, prince, willing (hearted)."},
"H5082":{"lemma":"נְדִיבָה","xlit":"nᵉdîybâh","pron":"ned-ee-baw'","derivation":"feminine of H5081 (נָדִיב);","strongs_def":"properly, nobility, i.e. reputation","kjv_def":"soul."},
"H5083":{"lemma":"נָדָן","xlit":"nâdân","pron":"naw-dawn'","derivation":"probably from an unused root meaning to give;","strongs_def":"a present (for prostitution)","kjv_def":"gift."},
"H5084":{"lemma":"נָדָן","xlit":"nâdân","pron":"naw-dawn'","derivation":"of uncertain derivation;","strongs_def":"a sheath (of a sword)","kjv_def":"sheath."},
"H5085":{"lemma":"נִדְנֶה","xlit":"nidneh","pron":"nid-neh'","derivation":"(Aramaic) from the same as H5084 (נָדָן);","strongs_def":"a sheath; figuratively, the body (as the receptacle of the soul)","kjv_def":"body."},
"H5086":{"lemma":"נָדַף","xlit":"nâdaph","pron":"naw-daf'","derivation":"a primitive root;","strongs_def":"to shove asunder, i.e. disperse","kjv_def":"drive (away, to and fro), thrust down, shaken, tossed to and fro."},
"H5087":{"lemma":"נָדַר","xlit":"nâdar","pron":"naw-dar'","derivation":"a primitive root;","strongs_def":"to promise (pos., to do or give something to God)","kjv_def":"(make a) vow."},
"H5088":{"lemma":"נֶדֶר","xlit":"neder","pron":"neh'-der","derivation":"or נֵדֶר; from H5087 (נָדַר);","strongs_def":"a promise (to God); also (concretely) a thing promised","kjv_def":"vow(-ed)."},
"H5089":{"lemma":"נֹהַּ","xlit":"nôahh","pron":"no'-ah","derivation":"from an unused root meaning to lament;","strongs_def":"lamentation","kjv_def":"wailing."},
"H5090":{"lemma":"נָהַג","xlit":"nâhag","pron":"naw-hag'","derivation":"a primitive root;","strongs_def":"to drive forth (a person, an animal or chariot), i.e. lead, carry away; reflexively, to proceed (i.e. impel or guide oneself); also (from the panting induced by effort), to sigh","kjv_def":"acquaint, bring (away), carry away, drive (away), lead (away, forth), (be) guide, lead (away, forth)."},
"H5091":{"lemma":"נָהָה","xlit":"nâhâh","pron":"naw-haw'","derivation":"a primitive root;","strongs_def":"to groan, i.e. bewail; hence (through the idea of crying aloud), to assemble (as if on proclamation)","kjv_def":"lament, wail."},
"H5092":{"lemma":"נְהִי","xlit":"nᵉhîy","pron":"neh-hee'","derivation":"from H5091 (נָהָה);","strongs_def":"an elegy","kjv_def":"lamentation, wailing."},
"H5093":{"lemma":"נִהְיָה","xlit":"nihyâh","pron":"nih-yaw'","derivation":"feminine of H5092 (נְהִי);","strongs_def":"lamentation","kjv_def":"doleful."},
"H5094":{"lemma":"נְהִיר","xlit":"nᵉhîyr","pron":"neh-heere'","derivation":"(Aramaic) or נְהִירוּ; (Aramaic), from the same as H5105 (נְהָרָה);","strongs_def":"illumination, i.e. (figuratively) wisdom","kjv_def":"light."},
"H5095":{"lemma":"נָהַל","xlit":"nâhal","pron":"naw-hal'","derivation":"a primitive root;","strongs_def":"properly, to run with asparkle, i.e. flow; hence (transitively), to conduct, and (by inference) to protect, sustain","kjv_def":"carry, feed, guide, lead (gently, on)."},
"H5096":{"lemma":"נַהֲלָל","xlit":"Nahălâl","pron":"nah-hal-awl'","derivation":"or נַהֲלֹל; the same as H5097 (נַהֲלֹל);","strongs_def":"Nahalal or Nahalol, a place in Palestine","kjv_def":"Nahalal, Nahallal, Nahalol."},
"H5097":{"lemma":"נַהֲלֹל","xlit":"nahălôl","pron":"nah-hal-ole'","derivation":"from H5095 (נָהַל);","strongs_def":"pasture","kjv_def":"bush."},
"H5098":{"lemma":"נָהַם","xlit":"nâham","pron":"naw-ham'","derivation":"a primitive root;","strongs_def":"to growl","kjv_def":"mourn, roar(-ing)."},
"H5099":{"lemma":"נַהַם","xlit":"naham","pron":"nah'-ham","derivation":"from H5098 (נָהַם);","strongs_def":"a snarl","kjv_def":"roaring."},
"H5100":{"lemma":"נְהָמָה","xlit":"nᵉhâmâh","pron":"neh-haw-maw'","derivation":"feminine of H5099 (נַהַם);","strongs_def":"snarling","kjv_def":"disquietness, roaring."},
"H5101":{"lemma":"נָהַק","xlit":"nâhaq","pron":"naw-hak'","derivation":"a primitive root;","strongs_def":"to bray (as an ass), scream (from hunger)","kjv_def":"bray."},
"H5102":{"lemma":"נָהַר","xlit":"nâhar","pron":"naw-har'","derivation":"a primitive root;","strongs_def":"to sparkle, i.e. (figuratively) be cheerful; hence (from the sheen of a running stream) to flow, i.e. (figuratively) assemble","kjv_def":"flow (together), be lightened."},
"H5103":{"lemma":"נְהַר","xlit":"nᵉhar","pron":"neh-har'","derivation":"(Aramaic) from a root corresponding to H5102 (נָהַר);","strongs_def":"a river, especially the Euphrates","kjv_def":"river, stream."},
"H5104":{"lemma":"נָהָר","xlit":"nâhâr","pron":"naw-hawr'","derivation":"from H5102 (נָהַר);","strongs_def":"a stream (including the sea; expectation the Nile, Euphrates, etc.); figuratively, prosperity","kjv_def":"flood, river."},
"H5105":{"lemma":"נְהָרָה","xlit":"nᵉhârâh","pron":"neh-haw-raw'","derivation":"from H5102 (נָהַר) in its original sense;","strongs_def":"daylight","kjv_def":"light."},
"H5106":{"lemma":"נוּא","xlit":"nûwʼ","pron":"noo","derivation":"a primitive root;","strongs_def":"to refuse, forbid, dissuade, or neutralize","kjv_def":"break, disallow, discourage, make of none effect."},
"H5107":{"lemma":"נוּב","xlit":"nûwb","pron":"noob","derivation":"a primitive root;","strongs_def":"to germinate, i.e. (figuratively) to (causatively, make) flourish; also (of words), to utter","kjv_def":"bring forth (fruit), make cheerful, increase."},
"H5108":{"lemma":"נוֹב","xlit":"nôwb","pron":"nobe","derivation":"or נֵיב; from H5107 (נוּב);","strongs_def":"produce, literally or figuratively","kjv_def":"fruit."},
"H5109":{"lemma":"נוֹבַי","xlit":"Nôwbay","pron":"no-bah'ee","derivation":"from H5108 (נוֹב); fruitful;","strongs_def":"Nobai, an Israelite","kjv_def":"Nebai (from the margin)."},
"H5110":{"lemma":"נוּד","xlit":"nûwd","pron":"nood","derivation":"a primitive root;","strongs_def":"to nod, i.e. waver; figuratively, to wander, flee, disappear; also (from shaking the head in sympathy), to console, deplore, or (from tossing the head in scorn) taunt","kjv_def":"bemoan, flee, get, mourn, make to move, take pity, remove, shake, skip for joy, be sorry, vagabond, way, wandering."},
"H5111":{"lemma":"נוּד","xlit":"nûwd","pron":"nood","derivation":"(Aramaic) corresponding to H5116 (נָוֶה);","strongs_def":"to flee","kjv_def":"get away."},
"H5112":{"lemma":"נוֹד","xlit":"nôwd","pron":"node","derivation":"(only defective נֹד ); from H5110 (נוּד);","strongs_def":"exile","kjv_def":"wandering."},
"H5113":{"lemma":"נוֹד","xlit":"Nôwd","pron":"node","derivation":"the same as H5112 (נוֹד)lemma נוֹדּ extra dagesh, corrected to נוֹד; vagrancy;","strongs_def":"Nod, the land of Cain","kjv_def":"Nod."},
"H5114":{"lemma":"נוֹדָב","xlit":"Nôwdâb","pron":"no-dawb'","derivation":"from H5068 (נָדַב); noble;","strongs_def":"Nodab, an Arab tribe","kjv_def":"Nodab."},
"H5115":{"lemma":"נָוָה","xlit":"nâvâh","pron":"naw-vaw'","derivation":"a primitive root; (compare H5116 (נָוֶה))","strongs_def":"to rest (as at home); causatively (through the implied idea of beauty ), to celebrate (with praises)","kjv_def":"keept at home, prepare an habitation."},
"H5116":{"lemma":"נָוֶה","xlit":"nâveh","pron":"naw-veh'","derivation":"or (feminine) נָוָה; from H5115 (נָוָה);","strongs_def":"(adjectively) at home; hence (by implication of satisfaction) lovely; also (noun) a home, of God (temple), men (residence), flocks (pasture), or wild animals (den)","kjv_def":"comely, dwelling (place), fold, habitation, pleasant place, sheepcote, stable, tarried."},
"H5117":{"lemma":"נוּחַ","xlit":"nûwach","pron":"noo'-akh","derivation":"a primitive root;","strongs_def":"to rest, i.e. settle down; used in a great variety of applications, literal and figurative, intransitive, transitive and causative (to dwell, stay, let fall, place, let alone, withdraw, give comfort, etc.)","kjv_def":"cease, be confederate, lay, let down, (be) quiet, remain, (cause to, be at, give, have, make to) rest, set down. Compare H3241 (יָנִים)."},
"H5118":{"lemma":"נוּחַ","xlit":"nûwach","pron":"noo'-akh","derivation":"or נוֹחַ; from H5117 (נוּחַ);","strongs_def":"quiet","kjv_def":"rest(-ed, -ing place)."},
"H5119":{"lemma":"נוֹחָה","xlit":"Nôwchâh","pron":"no-chaw'","derivation":"feminine of H5118 (נוּחַ); quietude;","strongs_def":"Nochah, an Israelite","kjv_def":"Nohah."},
"H5120":{"lemma":"נוּט","xlit":"nûwṭ","pron":"noot","strongs_def":"to quake","kjv_def":"be moved."},
"H5121":{"lemma":"נָוִית","xlit":"Nâvîyth","pron":"naw-veeth'","derivation":"from H5115 (נָוָה); residence;","strongs_def":"Navith, a place in Palestine","kjv_def":"Naioth (from the margin)."},
"H5122":{"lemma":"נְוָלוּ","xlit":"nᵉvâlûw","pron":"nev-aw-loo'","derivation":"(Aramaic) or נְוָלִי; (Aramaic), from an unused root probably meaning to be foul;","strongs_def":"a sink","kjv_def":"dunghill."},
"H5123":{"lemma":"נוּם","xlit":"nûwm","pron":"noom","derivation":"a primitive root;","strongs_def":"to slumber (from drowsiness)","kjv_def":"sleep, slumber."},
"H5124":{"lemma":"נוּמָה","xlit":"nûwmâh","pron":"noo-maw'","derivation":"from H5123 (נוּם);","strongs_def":"sleepiness","kjv_def":"drowsiness."},
"H5125":{"lemma":"נוּן","xlit":"nûwn","pron":"noon","derivation":"a primitive root;","strongs_def":"to resprout, i.e. propagate by shoots; figuratively, to be perpetual","kjv_def":"be continued."},
"H5126":{"lemma":"נוּן","xlit":"Nûwn","pron":"noon","derivation":"or נוֹן; (1 Chronicles 7:27), from H5125 (נוּן); perpetuity,","strongs_def":"Nun or Non, the father of Joshua","kjv_def":"Non, Nun."},
"H5127":{"lemma":"נוּס","xlit":"nûwç","pron":"noos","derivation":"a primitive root;","strongs_def":"to flit, i.e. vanish away (subside, escape; causatively, chase, impel, deliver)","kjv_def":"[idiom] abate, away, be displayed, (make to) flee (away, -ing), put to flight, [idiom] hide, lift up a standard."},
"H5128":{"lemma":"נוּעַ","xlit":"nûwaʻ","pron":"noo'-ah","derivation":"a primitive root;","strongs_def":"to waver, in a great variety of applications, literally and figuratively (as subjoined)","kjv_def":"continually, fugitive, [idiom] make, to (go) up and down, be gone away, (be) move(-able, -d), be promoted, reel, remove, scatter, set, shake, sift, stagger, to and fro, be vagabond, wag, (make) wander (up and down)."},
"H5129":{"lemma":"נוֹעַדְיָה","xlit":"Nôwʻadyâh","pron":"no-ad-yaw'","derivation":"from H3259 (יָעַד) and H3050 (יָהּ); convened of Jah;","strongs_def":"Noadjah, the name of an Israelite, and a false prophetess","kjv_def":"Noadiah."},
"H5130":{"lemma":"נוּף","xlit":"nûwph","pron":"noof","derivation":"a primitive root;","strongs_def":"to quiver (i.e. vibrate up and down, or rock to and fro); used in a great variety of applications (including sprinkling, beckoning, rubbing, bastinadoing, sawing, waving, etc.)","kjv_def":"lift up, move, offer, perfume, send, shake, sift, strike, wave."},
"H5131":{"lemma":"נוֹף","xlit":"nôwph","pron":"nofe","derivation":"from H5130 (נוּף);","strongs_def":"elevation","kjv_def":"situation. Compare H5297 (נֹף)."},
"H5132":{"lemma":"נוּץ","xlit":"nûwts","pron":"noots","derivation":"a primitive root;","strongs_def":"properly, to flash; hence, to blossom (from the brilliancy of color); also, to fly away (from the quickness of motion)","kjv_def":"flee away, bud (forth)."},
"H5133":{"lemma":"נוֹצָה","xlit":"nôwtsâh","pron":"no-tsaw'","derivation":"or נֹצָה; feminine active participle of H5327 (נָצָה) in the sense of flying;","strongs_def":"a pinion (or wing feather); often (collectively) plumage","kjv_def":"feather(-s), ostrich."},
"H5134":{"lemma":"נוּק","xlit":"nûwq","pron":"nook","derivation":"a primitive root;","strongs_def":"to suckle","kjv_def":"nurse."},
"H5135":{"lemma":"נוּר","xlit":"nûwr","pron":"noor","derivation":"(Aramaic) from an unused root (corresponding to that of H5216 (נִיר)) meaning to shine;","strongs_def":"fire","kjv_def":"fiery, fire."},
"H5136":{"lemma":"נוּשׁ","xlit":"nûwsh","pron":"noosh","derivation":"a primitive root;","strongs_def":"to be sick, i.e. (figuratively) distressed","kjv_def":"be full of heaviness."},
"H5137":{"lemma":"נָזָה","xlit":"nâzâh","pron":"naw-zaw'","derivation":"a primitive root;","strongs_def":"to spirt, i.e. besprinkle (especially in expiation)","kjv_def":"sprinkle."},
"H5138":{"lemma":"נָזִיד","xlit":"nâzîyd","pron":"naw-zeed'","derivation":"from H2102 (זוּד);","strongs_def":"something boiled, i.e. soup","kjv_def":"pottage."},
"H5139":{"lemma":"נָזִיר","xlit":"nâzîyr","pron":"naw-zeer'","derivation":"or נָזִר; from H5144 (נָזַר);","strongs_def":"separate, i.e. consecrated (as prince, a Nazirite); hence (figuratively from the latter) an unpruned vine (like an unshorn Nazirite)","kjv_def":"Nazarite (by a false alliteration with Nazareth), separate(-d), vine undressed."},
"H5140":{"lemma":"נָזַל","xlit":"nâzal","pron":"naw-zal'","derivation":"a primitive root;","strongs_def":"to drip, or shed by trickling","kjv_def":"distil, drop, flood, (cause to) flow(-ing), gush out, melt, pour (down), running water, stream."},
"H5141":{"lemma":"נֶזֶם","xlit":"nezem","pron":"neh'-zem","derivation":"from an unused root of uncertain meaning;","strongs_def":"a nose-ring","kjv_def":"earring, jewel."},
"H5142":{"lemma":"נְזַק","xlit":"nᵉzaq","pron":"nez-ak'","derivation":"(Aramaic) corresponding to the root of H5143 (נֵזֶק);","strongs_def":"to suffer (causatively, inflict) loss","kjv_def":"have (en-) damage, hurt(-ful)."},
"H5143":{"lemma":"נֵזֶק","xlit":"nêzeq","pron":"nay'zek","derivation":"from an unused root meaning to injure;","strongs_def":"loss","kjv_def":"damage."},
"H5144":{"lemma":"נָזַר","xlit":"nâzar","pron":"naw-zar'","derivation":"a primitive root;","strongs_def":"to hold aloof, i.e. (intransitivey) abstain (from food and drink, from impurity, and even from divine worship (i.e. apostatize)); specifically, to set apart (to sacred purposes), i.e. devote","kjv_def":"consecrate, separate(-ing, self)."},
"H5145":{"lemma":"נֶזֶר","xlit":"nezer","pron":"neh'-zer","derivation":"or נֵזֶר; from H5144 (נָזַר);","strongs_def":"properly, something set apart, i.e. (abstractly) dedication (of a priet or Nazirite); hence (concretely) unshorn locks; also (by implication) a chaplet (especially of royalty)","kjv_def":"consecration, crown, hair, separation."},
"H5146":{"lemma":"נֹחַ","xlit":"Nôach","pron":"no'-akh","derivation":"the same as H5118 (נוּחַ); rest;","strongs_def":"Noach, the patriarch of the flood","kjv_def":"Noah."},
"H5147":{"lemma":"נַחְבִּי","xlit":"Nachbîy","pron":"nakh-bee'","derivation":"from H2247 (חָבָה); occult;","strongs_def":"Nachbi, an Israelite","kjv_def":"Nakbi."},
"H5148":{"lemma":"נָחָה","xlit":"nâchâh","pron":"naw-khaw'","derivation":"a primitive root;","strongs_def":"to guide; by implication, to transport (into exile, or as colonists)","kjv_def":"bestow, bring, govern, guide, lead (forth), put, straiten."},
"H5149":{"lemma":"נְחוּם","xlit":"Nᵉchûwm","pron":"neh-khoom'","derivation":"from H5162 (נָחַם); comforted;","strongs_def":"Nechum, an Israelite","kjv_def":"Nehum."},
"H5150":{"lemma":"נִחוּם","xlit":"nichûwm","pron":"nee-khoom'","derivation":"or נִחֻם; from H5162 (נָחַם);","strongs_def":"properly, consoled; abstractly, solace","kjv_def":"comfort(-able), repenting."},
"H5151":{"lemma":"נַחוּם","xlit":"Nachûwm","pron":"nakh-oom'","derivation":"from H5162 (נָחַם); comfortable;","strongs_def":"Nachum, an Israelite prophet","kjv_def":"Nahum."},
"H5152":{"lemma":"נָחוֹר","xlit":"Nâchôwr","pron":"naw-khore'","derivation":"from the same as H5170 (נַחַר); snorer;","strongs_def":"Nochor, the name of the grandfather and a brother of Abraham","kjv_def":"Nahor."},
"H5153":{"lemma":"נָחוּשׁ","xlit":"nâchûwsh","pron":"naw-khoosh'","derivation":"apparently passive participle of H5172 (נָחַשׁ) (perhaps in the sense of ringing), i.e. bell-metal; or from the red color of the throat of a serpent (H5175 (נָחָשׁ), as denominative) when hissing);","strongs_def":"coppery, i.e. (figuratively) hard","kjv_def":"of brass."},
"H5154":{"lemma":"נְחוּשָׁה","xlit":"nᵉchûwshâh","pron":"nekh-oo-shaw'","derivation":"or נְחֻשָׁה; feminine of H5153 (נָחוּשׁ);","strongs_def":"copper","kjv_def":"brass, steel. Compare H5176 (נָחָשׁ)."},
"H5155":{"lemma":"נְחִילָה","xlit":"nᵉchîylâh","pron":"nekh-ee-law'","derivation":"probably denominative from H2485 (חָלִיל);","strongs_def":"a flute","kjv_def":"(plural) Nehiloth."},
"H5156":{"lemma":"נְחִיר","xlit":"nᵉchîyr","pron":"nekh-eer'","derivation":"from the same as H5170 (נַחַר);","strongs_def":"a nostril","kjv_def":"(dual) nostrils."},
"H5157":{"lemma":"נָחַל","xlit":"nâchal","pron":"naw-khal'","derivation":"a primitive root;","strongs_def":"to inherit (as a (figurative) mode of descent), or (generally) to occupy; causatively, to bequeath, or (generally) distribute, instate","kjv_def":"divide, have (inheritance), take as a heritage, (cause to, give to, make to) inherit, (distribute for, divide (for, for an, by), give for, have, leave for, take (for)) inheritance, (have in, cause to, be made to) possess(-ion)."},
"H5158":{"lemma":"נַחַל","xlit":"nachal","pron":"nakh'-al","derivation":"or (feminine) נַחְלָה; (Psalm 124:4), or נַחֲלָה; (Ezekiel 47:19; Ezekiel 48:28), from H5157 (נָחַל) in its original sense;","strongs_def":"a stream, especially a winter torrent; (by implication) a (narrow) valley (in which a brook runs); also a shaft (of a mine)","kjv_def":"brook, flood, river, stream, valley."},
"H5159":{"lemma":"נַחֲלָה","xlit":"nachălâh","pron":"nakh-al-aw'","derivation":"from H5157 (נָחַל) (in its usual sense);","strongs_def":"properly, something inherited, i.e. (abstractly) occupancy, or (concretely) an heirloom; generally an estate, patrimony or portion","kjv_def":"heritage, to inherit, inheritance, possession. Compare H5158 (נַחַל)."},
"H5160":{"lemma":"נַחֲלִיאֵל","xlit":"Nachălîyʼêl","pron":"nakh-al-ee-ale'","derivation":"from H5158 (נַחַל) and H410 (אֵל); valley of God;","strongs_def":"Nachaliel, a place in the Desert","kjv_def":"Nahaliel."},
"H5161":{"lemma":"נֶחֱלָמִי","xlit":"Nechĕlâmîy","pron":"nekh-el-aw-mee'","derivation":"apparently a patronymic from an unused name (apparently passive participle of H2492 (חָלַם)); dreamed;","strongs_def":"a Nechelamite, or descendant of Nechlam","kjv_def":"Nehelamite."},
"H5162":{"lemma":"נָחַם","xlit":"nâcham","pron":"naw-kham'","derivation":"a primitive root;","strongs_def":"properly, to sigh, i.e. breathe strongly; by implication, to be sorry, i.e. (in a favorable sense) to pity, console or (reflexively) rue; or (unfavorably) to avenge (oneself)","kjv_def":"comfort (self), ease (one's self), repent(-er,-ing, self)."},
"H5163":{"lemma":"נַחַם","xlit":"Nacham","pron":"nakh'-am","derivation":"from H5162 (נָחַם); consolation;","strongs_def":"Nacham, an Israelite","kjv_def":"Naham."},
"H5164":{"lemma":"נֹחַם","xlit":"nôcham","pron":"no'-kham","derivation":"from H5162 (נָחַם);","strongs_def":"ruefulness, i.e. desistance","kjv_def":"repentance."},
"H5165":{"lemma":"נֶחָמָה","xlit":"nechâmâh","pron":"nekh-aw-maw'","derivation":"from H5162 (נָחַם);","strongs_def":"consolation","kjv_def":"comfort."},
"H5166":{"lemma":"נְחֶמְיָה","xlit":"Nᵉchemyâh","pron":"nekh-em-yaw'","derivation":"from H5162 (נָחַם) and H3050 (יָהּ); consolation of Jah;","strongs_def":"Nechemjah, the name of three Israelites","kjv_def":"Nehemiah."},
"H5167":{"lemma":"נַחֲמָנִי","xlit":"Nachămânîy","pron":"nakh-am-aw-nee'","derivation":"from H5162 (נָחַם); consolatory;","strongs_def":"Nachamani, an Israelite","kjv_def":"Nahamani."},
"H5168":{"lemma":"נַחְנוּ","xlit":"nachnûw","pron":"nakh-noo'","derivation":"for H587 (אֲנַחְנוּ);","strongs_def":"we","kjv_def":"we."},
"H5169":{"lemma":"נָחַץ","xlit":"nâchats","pron":"naw-khats'","derivation":"a primitive root;","strongs_def":"to be urgent","kjv_def":"require haste."},
"H5170":{"lemma":"נַחַר","xlit":"nachar","pron":"nakh'-ar","derivation":"and (feminine) נַחֲרָה; from an unused root meaning to snort or snore;","strongs_def":"a snorting","kjv_def":"nostrils, snorting."},
"H5171":{"lemma":"נַחֲרַי","xlit":"Nachăray","pron":"nakh-ar-ah'-ee","derivation":"or נַחְרַי; from the same as H5170 (נַחַר); snorer;","strongs_def":"Nacharai or Nachrai, an Israelite","kjv_def":"Naharai, Nahari."},
"H5172":{"lemma":"נָחַשׁ","xlit":"nâchash","pron":"naw-khash'","derivation":"a primitive root;","strongs_def":"properly, to hiss, i.e. whisper a (magic) spell; generally, to prognosticate","kjv_def":"[idiom] certainly, divine, enchanter, (use) [idiom] enchantment, learn by experience, [idiom] indeed, diligently observe."},
"H5173":{"lemma":"נַחַשׁ","xlit":"nachash","pron":"nakh'-ash","derivation":"from H5172 (נָחַשׁ);","strongs_def":"an incantation or augury","kjv_def":"enchantment."},
"H5174":{"lemma":"נְחָשׁ","xlit":"nᵉchâsh","pron":"nekh-awsh'","derivation":"(Aramaic) corresponding to H5154 (נְחוּשָׁה);","strongs_def":"copper","kjv_def":"brass."},
"H5175":{"lemma":"נָחָשׁ","xlit":"nâchâsh","pron":"naw-khawsh'","derivation":"from H5172 (נָחַשׁ);","strongs_def":"a snake (from its hiss)","kjv_def":"serpent."},
"H5176":{"lemma":"נָחָשׁ","xlit":"Nâchâsh","pron":"naw-khawsh'","derivation":"the same as H5175 (נָחָשׁ);","strongs_def":"Nachash, the name of two persons apparently non-Israelite","kjv_def":"Nahash."},
"H5177":{"lemma":"נַחְשׁוֹן","xlit":"Nachshôwn","pron":"nakh-shone'","derivation":"from H5172 (נָחַשׁ); enchanter;","strongs_def":"Nachshon, an Israelite","kjv_def":"Naashon, Nahshon."},
"H5178":{"lemma":"נְחֹשֶׁת","xlit":"nᵉchôsheth","pron":"nekh-o'-sheth","derivation":"for H5154 (נְחוּשָׁה);","strongs_def":"copper, hence, something made of that metal, i.e. coin, a fetter; figuratively, base (as compared with gold or silver)","kjv_def":"brasen, brass, chain, copper, fetter (of brass), filthiness, steel."},
"H5179":{"lemma":"נְחֻשְׁתָּא","xlit":"Nᵉchushtâʼ","pron":"nekh-oosh-taw'","derivation":"from H5178 (נְחֹשֶׁת); copper;","strongs_def":"Nechushta, an Israelitess","kjv_def":"Nehushta."},
"H5180":{"lemma":"נְחֻשְׁתָּן","xlit":"Nᵉchushtân","pron":"nekh-oosh-tawn'","derivation":"from H5178 (נְחֹשֶׁת);","strongs_def":"something made of copper, i.e. the copper serpent of the Desert","kjv_def":"Nehushtan."},
"H5181":{"lemma":"נָחַת","xlit":"nâchath","pron":"naw-khath'","derivation":"a primitive root;","strongs_def":"to sink, i.e. descend; causatively, to press or lead down","kjv_def":"be broken, (cause to) come down, enter, go down, press sore, settle, stick fast."},
"H5182":{"lemma":"נְחַת","xlit":"nᵉchath","pron":"nekh-ath'","derivation":"(Aramaic) corresponding to H5181 (נָחַת);","strongs_def":"to descend; causatively, to bring away, deposit, depose","kjv_def":"carry, come down, depose, lay up, place."},
"H5183":{"lemma":"נַחַת","xlit":"Nachath","pron":"nakh'-ath","derivation":"from H5182 (נְחַת); also (intransitively; perhaps from H5117 (נוּחַ))","strongs_def":"a descent, i.e. imposition, unfavorable (punishment) or favorable (food); restfulness","kjv_def":"lighting down, quiet(-ness), to rest, be set on."},
"H5184":{"lemma":"נַחַת","xlit":"Nachath","pron":"nakh'-ath","derivation":"the same as H5183 (נַחַת); quiet;","strongs_def":"Nachath, the name of an Edomite and of two Israelites","kjv_def":"Nahath."},
"H5185":{"lemma":"נָחֵת","xlit":"nâchêth","pron":"naw-khayth'","derivation":"from H5181 (נָחַת);","strongs_def":"descending","kjv_def":"come down."},
"H5186":{"lemma":"נָטָה","xlit":"nâṭâh","pron":"naw-taw'","derivation":"a primitive root;","strongs_def":"to stretch or spread out; by implication, to bend away (including moral deflection); used in a great variety of application","kjv_def":"[phrase] afternoon, apply, bow (down, -ing), carry aside, decline, deliver, extend, go down, be gone, incline, intend, lay, let down, offer, outstretched, overthrown, pervert, pitch, prolong, put away, shew, spread (out), stretch (forth, out), take (aside), turn (aside, away), wrest, cause to yield."},
"H5187":{"lemma":"נְטִיל","xlit":"nᵉṭîyl","pron":"net-eel'","derivation":"from H5190 (נָטַל);","strongs_def":"laden","kjv_def":"that bear."},
"H5188":{"lemma":"נְטִיפָה","xlit":"nᵉṭîyphâh","pron":"net-ee-faw'","derivation":"from H5197 (נָטַף);","strongs_def":"a pendant for the ears (especially of pearls)","kjv_def":"chain, collar."},
"H5189":{"lemma":"נְטִישָׁה","xlit":"nᵉṭîyshâh","pron":"net-ee-shaw'","derivation":"from H5203 (נָטַשׁ);","strongs_def":"a tendril (as an offshoot)","kjv_def":"battlement, branch, plant."},
"H5190":{"lemma":"נָטַל","xlit":"nâṭal","pron":"naw-tal'","derivation":"a primitive root;","strongs_def":"to lift; by implication, to impose","kjv_def":"bear, offer, take up."},
"H5191":{"lemma":"נְטַל","xlit":"nᵉṭal","pron":"net-al'","derivation":"(Aramaic) corresponding to H5190 (נָטַל);","strongs_def":"to raise","kjv_def":"take up."},
"H5192":{"lemma":"נֵטֶל","xlit":"nêṭel","pron":"nay'-tel","derivation":"; from H5190 (נָטַל);","strongs_def":"a burden","kjv_def":"weighty."},
"H5193":{"lemma":"נָטַע","xlit":"nâṭaʻ","pron":"naw-tah'","derivation":"a primitive root;","strongs_def":"properly, to strike in, i.e. fix; specifically, to plant (literally or figuratively)","kjv_def":"fastened, plant(-er)."},
"H5194":{"lemma":"נֶטַע","xlit":"neṭaʻ","pron":"neh'-tah","derivation":"from H5193 (נָטַע);","strongs_def":"a plant; collectively, a plantation; abstractly, a planting","kjv_def":"plant."},
"H5195":{"lemma":"נָטִיעַ","xlit":"nâṭîyaʻ","pron":"naw-tee'-ah","derivation":"from H5193 (נָטַע);","strongs_def":"a plant","kjv_def":"plant."},
"H5196":{"lemma":"נְטָעִים","xlit":"Nᵉṭâʻîym","pron":"net-aw-eem'","derivation":"plural of H5194 (נֶטַע);","strongs_def":"Netaim, a place in Palestine","kjv_def":"plants."},
"H5197":{"lemma":"נָטַף","xlit":"nâṭaph","pron":"naw-taf'","derivation":"a primitive root;","strongs_def":"to ooze, i.e. distil gradually; by implication, to fall in drops; figuratively, to speak by inspiration","kjv_def":"drop(-ping), prophesy(-et)."},
"H5198":{"lemma":"נָטָף","xlit":"nâṭâph","pron":"naw-tawf'","derivation":"from H5197 (נָטַף);","strongs_def":"a drop; specifically, an aromatic gum (probably stacte)","kjv_def":"drop, stacte."},
"H5199":{"lemma":"נְטֹפָה","xlit":"Nᵉṭôphâh","pron":"net-o-faw'","derivation":"from H5197 (נָטַף); distillation;","strongs_def":"Netophah, a place in Palestine","kjv_def":"Netophah."},
"H5200":{"lemma":"נְטֹפָתִי","xlit":"Nᵉṭôphâthîy","pron":"net-o-faw-thee'","derivation":"patronymic from H5199 (נְטֹפָה);","strongs_def":"a Netophathite, or inhabitant of Netophah","kjv_def":"Netophathite."},
"H5201":{"lemma":"נָטַר","xlit":"nâṭar","pron":"naw-tar'","derivation":"a primitive root;","strongs_def":"to guard; figuratively, to cherish (anger)","kjv_def":"bear grudge, keep(-er), reserve."},
"H5202":{"lemma":"נְטַר","xlit":"nᵉṭar","pron":"net-ar'","derivation":"(Aramaic) corresponding to H5201 (נָטַר);","strongs_def":"to retain","kjv_def":"keep."},
"H5203":{"lemma":"נָטַשׁ","xlit":"nâṭash","pron":"naw-tash'","derivation":"a primitive root;","strongs_def":"properly, to pound, i.e. smite; by implication (as if beating out, and thus expanding) to disperse; also, to thrust off, down, out or upon (inclusively, reject, let alone, permit, remit, etc.)","kjv_def":"cast off, drawn, let fall, forsake, join (battle), leave (off), lie still, loose, spread (self) abroad, stretch out, suffer."},
"H5204":{"lemma":"נִי","xlit":"nîy","pron":"nee","derivation":"a doubtful word; apparently from H5091 (נָהָה);","strongs_def":"lamentation","kjv_def":"wailing."},
"H5205":{"lemma":"נִיד","xlit":"nîyd","pron":"need","derivation":"from H5110 (נוּד);","strongs_def":"motion (of the lips in speech)","kjv_def":"moving."},
"H5206":{"lemma":"נִידָה","xlit":"nîydâh","pron":"nee-daw'","derivation":"feminine of H5205 (נִיד);","strongs_def":"removal, i.e. exile","kjv_def":"removed."},
"H5207":{"lemma":"נִיחוֹחַ","xlit":"nîychôwach","pron":"nee-kho'-akh","derivation":"or נִיחֹחַ; from H5117 (נוּחַ);","strongs_def":"properly, restful, i.e. pleasant; abstractly, delight","kjv_def":"sweet (odour)."},
"H5208":{"lemma":"נִיחוֹחַ","xlit":"nîychôwach","pron":"nee-kho'-akh","derivation":"(Aramaic) or נִיחֹחַ; (shorter) (Aramaic), corresponding to H5207 (נִיחוֹחַ);","strongs_def":"pleasure","kjv_def":"sweet odour (savour)."},
"H5209":{"lemma":"נִין","xlit":"nîyn","pron":"neen","derivation":"from H5125 (נוּן);","strongs_def":"progeny","kjv_def":"son."},
"H5210":{"lemma":"נִינְוֵה","xlit":"Nîynᵉvêh","pron":"nee-nev-ay'","derivation":"of foreign origin;","strongs_def":"Nineveh, the capital of Assyria","kjv_def":"Nineveh."},
"H5211":{"lemma":"נִיס","xlit":"nîyç","pron":"neece","derivation":"from H5127 (נוּס);","strongs_def":"fugitive","kjv_def":"that fleeth."},
"H5212":{"lemma":"נִיסָן","xlit":"Nîyçân","pron":"nee-sawn'","derivation":"probably of foreign origin;","strongs_def":"Nisan, the first month of the Jewish sacred year","kjv_def":"Nisan."},
"H5213":{"lemma":"נִיצוֹץ","xlit":"nîytsôwts","pron":"nee-tsotes'","derivation":"from H5340 (נָצַץ);","strongs_def":"a spark","kjv_def":"spark."},
"H5214":{"lemma":"נִיר","xlit":"nîyr","pron":"neer","derivation":"a root probably identical with that of H5216 (נִיר), through the idea of the gleam of a fresh furrow;","strongs_def":"to till the soil","kjv_def":"break up."},
"H5215":{"lemma":"נִיר","xlit":"nîyr","pron":"neer","derivation":"or נִר; from H5214 (נִיר);","strongs_def":"properly, plowing, i.e. (concretely) freshly plowed land","kjv_def":"fallow ground, plowing, tillage."},
"H5216":{"lemma":"נִיר","xlit":"nîyr","pron":"neer","derivation":"or נִר; also נֵיר; or נֵר; or (feminine) נֵרָה; from a primitive root (see H5214 (נִיר); H5135 (נוּר)) properly, meaning to glisten;","strongs_def":"a lamp (i.e. the burner) or light (literally or figuratively)","kjv_def":"candle, lamp, light."},
"H5217":{"lemma":"נָכָא","xlit":"nâkâʼ","pron":"naw-kaw'","derivation":"a primitive root;","strongs_def":"to smite, i.e. drive away","kjv_def":"be viler."},
"H5218":{"lemma":"נָכֵא","xlit":"nâkêʼ","pron":"naw-kay'","derivation":"or נָכָא; from H5217 (נָכָא);","strongs_def":"smitten, i.e. (figuratively) afflicted","kjv_def":"broken, stricken, wounded."},
"H5219":{"lemma":"נְכֹאת","xlit":"nᵉkôʼth","pron":"nek-ohth'","derivation":"from H5218 (נָכֵא);","strongs_def":"properly, a smiting, i.e. (concretely) an aromatic gum (perhaps styrax) (as powdered)","kjv_def":"spicery(-ces)."},
"H5220":{"lemma":"נֶכֶד","xlit":"neked","pron":"neh'-ked","derivation":"from an unused root meaning to propagate;","strongs_def":"offspring","kjv_def":"nephew, son's son."},
"H5221":{"lemma":"נָכָה","xlit":"nâkâh","pron":"naw-kaw'","derivation":"a primitive root;","strongs_def":"to strike (lightly or severely, literally or figuratively)","kjv_def":"beat, cast forth, clap, give (wounds), [idiom] go forward, [idiom] indeed, kill, make (slaughter), murderer, punish, slaughter, slay(-er, -ing), smite(-r, -ing), strike, be stricken, (give) stripes, [idiom] surely, wound."},
"H5222":{"lemma":"נֵכֶה","xlit":"nêkeh","pron":"nay-keh'","derivation":"from H5221 (נָכָה);","strongs_def":"a smiter, i.e. (figuratively) traducer","kjv_def":"abject."},
"H5223":{"lemma":"נָכֶה","xlit":"nâkeh","pron":"naw-keh'","derivation":"","strongs_def":"smitten, i.e. (literally) maimed, or (figuratively) dejected","kjv_def":"contrite, lame."},
"H5224":{"lemma":"נְכוֹ","xlit":"Nᵉkôw","pron":"nek-o'","derivation":"probably of Egyptian origin;","strongs_def":"Neko, an Egyptian king","kjv_def":"Necho. Compare H6549 (פַּרְעֹה נְכֹה)."},
"H5225":{"lemma":"נָכוֹן","xlit":"Nâkôwn","pron":"naw-kone'","derivation":"from H3559 (כּוּן); prepared;","strongs_def":"Nakon, probably an Israelite","kjv_def":"Nachon."},
"H5226":{"lemma":"נֵכַח","xlit":"nêkach","pron":"nay'-kakh","derivation":"from an unused root meaning to be straightforward;","strongs_def":"properly, the fore part; used adverbially, opposite","kjv_def":"before, over against."},
"H5227":{"lemma":"נֹכַח","xlit":"nôkach","pron":"no'-kakh","derivation":"from the same as H5226 (נֵכַח);","strongs_def":"properly, the front part; used adverbially (especially with preposition), opposite, in front of, forward, in behalf of","kjv_def":"(over) against, before, direct(-ly), for, right (on)."},
"H5228":{"lemma":"נָכֹחַ","xlit":"nâkôach","pron":"naw-ko'-akh","derivation":"from the same as H5226 (נֵכַח);","strongs_def":"straightforward, i.e. (figuratively), equitable, correct, or (abstractly), integrity","kjv_def":"plain, right, uprightness."},
"H5229":{"lemma":"נְכֹחָה","xlit":"nᵉkôchâh","pron":"nek-o-khaw'","derivation":"feminine of H5228 (נָכֹחַ);","strongs_def":"properly, straightforwardness, i.e. (figuratively) integrity, or (concretely) a truth","kjv_def":"equity, right (thing), uprightness."},
"H5230":{"lemma":"נָכַל","xlit":"nâkal","pron":"naw-kal'","derivation":"a primitive root;","strongs_def":"to defraud, i.e. act treacherously","kjv_def":"beguile, conspire, deceiver, deal subtilly."},
"H5231":{"lemma":"נֵכֶל","xlit":"nêkel","pron":"nay'-kel","derivation":"from H5230 (נָכַל);","strongs_def":"deceit","kjv_def":"wile."},
"H5232":{"lemma":"נְכַס","xlit":"nᵉkaç","pron":"nek-as'","derivation":"(Aramaic) corresponding to H5233 (נֶכֶס)","strongs_def":"{treasure}","kjv_def":"goods."},
"H5233":{"lemma":"נֶכֶס","xlit":"nekeç","pron":"neh'-kes","derivation":"from an unused root meaning to accumulate;","strongs_def":"treasure","kjv_def":"riches, wealth."},
"H5234":{"lemma":"נָכַר","xlit":"nâkar","pron":"naw-kar'","derivation":"a primitive root;","strongs_def":"properly, to scrutinize, i.e. look intently at; hence (with recognition implied), to acknowledge, be acquainted with, care for, respect, revere, or (with suspicion implied), to disregard, ignore, be strange toward, reject, resign, dissimulate (as if ignorant or disowning)","kjv_def":"acknowledge, [idiom] could, deliver, discern, dissemble, estrange, feign self to be another, know, take knowledge (notice), perceive, regard, (have) respect, behave (make) self strange(-ly)."},
"H5235":{"lemma":"נֶכֶר","xlit":"neker","pron":"neh'-ker","derivation":"or נֹכֶר; from H5234 (נָכַר);","strongs_def":"something strange, i.e. unexpected calamity","kjv_def":"strange."},
"H5236":{"lemma":"נֵכָר","xlit":"nêkâr","pron":"nay-kawr'","derivation":"' from H5234 (נָכַר);","strongs_def":"foreign, or (concretely) a foreigner, or (abstractly) heathendom","kjv_def":"alien, strange ([phrase] -er)."},
"H5237":{"lemma":"נׇכְרִי","xlit":"nokrîy","pron":"nok-ree'","derivation":"from H5235 (נֶכֶר) (second form);","strongs_def":"strange, in a variety of degrees and applications (foreign, non-relative, adulterous, different, wonderful)","kjv_def":"alien, foreigner, outlandish, strange(-r, woman)."},
"H5238":{"lemma":"נְכֹת","xlit":"nᵉkôth","pron":"nek-oth'","derivation":"probably for H5219 (נְכֹאת);","strongs_def":"spicery, i.e. (generally) valuables","kjv_def":"precious things."},
"H5239":{"lemma":"נָלָה","xlit":"nâlâh","pron":"naw-law'","derivation":"apparently a primitive root;","strongs_def":"to complete","kjv_def":"make an end."},
"H5240":{"lemma":"נְמִבְזֶה","xlit":"nᵉmibzeh","pron":"nem-ib-zeh'","derivation":"from H959 (בָּזָה),","strongs_def":"despised","kjv_def":"vile."},
"H5241":{"lemma":"נְמוּאֵל","xlit":"Nᵉmûwʼêl","pron":"nem-oo-ale'","derivation":"apparently for H3223 (יְמוּאֵל);","strongs_def":"Nemuel, the name of two Israelites","kjv_def":"Nemuel."},
"H5242":{"lemma":"נְמוּאֵלִי","xlit":"Nᵉmûwʼêlîy","pron":"nem-oo-ay-lee'","derivation":"from H5241 (נְמוּאֵל);","strongs_def":"a Nemuelite, or descendant of Nemuel","kjv_def":"Nemuelite."},
"H5243":{"lemma":"נָמַל","xlit":"nâmal","pron":"naw-mal'","derivation":"a primitive root;","strongs_def":"to become clipped or (specifically) circumcised","kjv_def":"(branch to) be cut down (off), circumcise."},
"H5244":{"lemma":"נְמָלָה","xlit":"nᵉmâlâh","pron":"nem-aw-law'","derivation":"feminine from H5243 (נָמַל);","strongs_def":"an ant (probably from its almost bisected form)","kjv_def":"ant."},
"H5245":{"lemma":"נְמַר","xlit":"nᵉmar","pron":"nem-ar'","derivation":"(Aramaic) corresponding to H5246 (נָמֵר)","strongs_def":"{a leopard (from its stripes)}","kjv_def":"leopard."},
"H5246":{"lemma":"נָמֵר","xlit":"nâmêr","pron":"naw-mare'","derivation":"from an unused root meaning properly, to filtrate, i.e. be limpid (comp H5247 (נִמְרָה) and H5249 (נִמְרִים)); and thus to spot or stain as if by dripping;","strongs_def":"a leopard (from its stripes)","kjv_def":"leopard."},
"H5247":{"lemma":"נִמְרָה","xlit":"Nimrâh","pron":"nim-raw'","derivation":"from the same as H5246 (נָמֵר); clear water;","strongs_def":"Nimrah, a place East of the Jordan","kjv_def":"Nimrah. See also H1039 (בֵּית נִמְרָה), H5249 (נִמְרִים)."},
"H5248":{"lemma":"נִמְרוֹד","xlit":"Nimrôwd","pron":"nim-rode'","derivation":"or נִמְרֹד; probably of foreign origin;","strongs_def":"Nimrod, a son of Cush","kjv_def":"Nimrod."},
"H5249":{"lemma":"נִמְרִים","xlit":"Nimrîym","pron":"nim-reem'","derivation":"plural of a masculine corresponding to H5247 (נִמְרָה); clear waters;","strongs_def":"Nimrim, a place East of the Jordan","kjv_def":"Nimrim. Compare H1039 (בֵּית נִמְרָה)."},
"H5250":{"lemma":"נִמְשִׁי","xlit":"Nimshîy","pron":"nim-shee'","derivation":"probably from H4871 (מָשָׁה); extricated;","strongs_def":"Nimshi, the (grand-) father of Jehu","kjv_def":"Nimshi."},
"H5251":{"lemma":"נֵס","xlit":"nêç","pron":"nace","derivation":"from H5264 (נָסַס);","strongs_def":"a flag; also a sail; by implication, a flagstaff; generally a signal; figuratively, a token","kjv_def":"banner, pole, sail, (en-) sign, standard."},
"H5252":{"lemma":"נְסִבָּה","xlit":"nᵉçibbâh","pron":"nes-ib-baw'","derivation":"feminine participle passive of H5437 (סָבַב);","strongs_def":"properly, an environment, i.e. circumstance or turn of affairs","kjv_def":"cause."},
"H5253":{"lemma":"נָסַג","xlit":"nâçag","pron":"naw-sag'","derivation":"a primitive root;","strongs_def":"to retreat","kjv_def":"departing away, remove, take (hold), turn away."},
"H5254":{"lemma":"נָסָה","xlit":"nâçâh","pron":"naw-saw'","derivation":"a primitive root;","strongs_def":"to test; by implication, to attempt","kjv_def":"adventure, assay, prove, tempt, try."},
"H5255":{"lemma":"נָסַח","xlit":"nâçach","pron":"naw-sakh'","derivation":"a primitive root;","strongs_def":"to tear away","kjv_def":"destroy, pluck, root."},
"H5256":{"lemma":"נְסַח","xlit":"nᵉçach","pron":"nes-akh'","derivation":"(Aramaic) corresponding to H5255 (נָסַח)","strongs_def":"{to tear away}","kjv_def":"pull down."},
"H5257":{"lemma":"נְסִיךְ","xlit":"nᵉçîyk","pron":"nes-eek'","derivation":"from H5258 (נָסַךְ);","strongs_def":"properly, something poured out, i.e. a libation; also a molten image; by implication, a prince (as anointed)","kjv_def":"drink offering, duke, prince(-ipal)."},
"H5258":{"lemma":"נָסַךְ","xlit":"nâçak","pron":"naw-sak'","derivation":"a primitive root;","strongs_def":"to pour out, especially a libation, or to cast (metal); by analogy, to anoint aking","kjv_def":"cover, melt, offer, (cause to) pour (out), set (up)."},
"H5259":{"lemma":"נָסַךְ","xlit":"nâçak","pron":"naw-sak'","derivation":"a primitive root (probably identical with H5258 (נָסַךְ) through the idea of fusion);","strongs_def":"to interweave, i.e. (figuratively) to overspread","kjv_def":"that is spread."},
"H5260":{"lemma":"נְסַךְ","xlit":"nᵉçak","pron":"nes-ak'","derivation":"(Aramaic) corresponding to H5258 (נָסַךְ);","strongs_def":"to pour out a libation","kjv_def":"offer."},
"H5261":{"lemma":"נְסַךְ","xlit":"nᵉçak","pron":"nes-ak'","derivation":"(Aramaic) corresponding to H5262 (נֶסֶךְ);","strongs_def":"a libation","kjv_def":"drink offering."},
"H5262":{"lemma":"נֶסֶךְ","xlit":"neçek","pron":"neh'-sek","derivation":"or נֵסֶךְ; from H5258 (נָסַךְ);","strongs_def":"a libation; also a cast idol","kjv_def":"cover, drink offering, molten image."},
"H5263":{"lemma":"נָסַס","xlit":"nâçaç","pron":"naw-sas'","derivation":"a primitive root;","strongs_def":"to wane, i.e. be sick","kjv_def":"faint."},
"H5264":{"lemma":"נָסַס","xlit":"nâçaç","pron":"naw-sas'","derivation":"a primitive root; to gleam from afar, i.e. to be conspicuous as a signal; or rather perhaps a denominative from H5251 (נֵס) (and identical with H5263 (נָסַס), through the idea of a flag as fluttering in the wind);","strongs_def":"to raise a beacon","kjv_def":"lift up as an ensign."},
"H5265":{"lemma":"נָסַע","xlit":"nâçaʻ","pron":"naw-sah'","derivation":"a primitive root;","strongs_def":"properly, to pull up, especially the tent-pins, i.e. start on ajourney","kjv_def":"cause to blow, bring, get, (make to) go (away, forth, forward, onward, out), (take) journey, march, remove, set aside (forward), [idiom] still, be on his (go their) way."},
"H5266":{"lemma":"נָסַק","xlit":"nâçaq","pron":"naw-sak'","derivation":"a primitive root;","strongs_def":"to go up","kjv_def":"ascend."},
"H5267":{"lemma":"נְסַק","xlit":"nᵉçaq","pron":"nes-ak'","derivation":"(Aramaic) corresponding to H5266 (נָסַק)","strongs_def":"{to go up}","kjv_def":"take up."},
"H5268":{"lemma":"נִסְרֹךְ","xlit":"Niçrôk","pron":"nis-roke'","derivation":"of foreign origin;","strongs_def":"Nisrok, a Babylonian idol","kjv_def":"Nisroch."},
"H5269":{"lemma":"נֵעָה","xlit":"Nêʻâh","pron":"nay-aw'","derivation":"from H5128 (נוּעַ); motion;","strongs_def":"Neah, a place in Palestine","kjv_def":"Neah."},
"H5270":{"lemma":"נֹעָה","xlit":"Nôʻâh","pron":"no-aw'","derivation":"from H5128 (נוּעַ); movement;","strongs_def":"Noah, an Israelitess","kjv_def":"Noah."},
"H5271":{"lemma":"נָעוּר","xlit":"nâʻûwr","pron":"naw-oor'","derivation":"or נָעֻר; and (feminine) נְעֻרָה; properly, passive participle from H5288 (נַעַר) as denominative;","strongs_def":"(only in plural collectively or emphatic form) youth, the state (juvenility) or the persons (young people)","kjv_def":"childhood, youth."},
"H5272":{"lemma":"נְעִיאֵל","xlit":"Nᵉʻîyʼêl","pron":"neh-ee-ale'","derivation":"from H5128 (נוּעַ) and H410 (אֵל); moved of God;","strongs_def":"Neiel, a place in Palestine","kjv_def":"Neiel."},
"H5273":{"lemma":"נָעִים","xlit":"nâʻîym","pron":"naw-eem'","derivation":"from H5276 (נָעֵם);","strongs_def":"delightful (objective or subjective, literal or figurative)","kjv_def":"pleasant(-ure), sweet."},
"H5274":{"lemma":"נָעַל","xlit":"nâʻal","pron":"naw-al'","derivation":"a primitive root; also denominative from H5275 (נַעַל)","strongs_def":"properly, to fasten up, i.e. with abar or cord; hence, to sandal, i.e. furnish with slippers","kjv_def":"bolt, inclose, lock, shoe, shut up."},
"H5275":{"lemma":"נַעַל","xlit":"naʻal","pron":"nah'-al","derivation":"or (feminine) נַעֲלָה; from H5274 (נָעַל);","strongs_def":"properly, a sandal tongue; by extension a sandal or slipper (sometimes as a symbol of occupancy, a refusal to marry, or of something valueless)","kjv_def":"dryshod, (pair of) shoe((-latchet), -s)."},
"H5276":{"lemma":"נָעֵם","xlit":"nâʻêm","pron":"naw-ame'","derivation":"a primitive root;","strongs_def":"to be agreeable (literally or figuratively)","kjv_def":"pass in beauty, be delight, be pleasant, be sweet."},
"H5277":{"lemma":"נַעַם","xlit":"Naʻam","pron":"nah'-am","derivation":"from H5276 (נָעֵם); pleasure;","strongs_def":"Naam, an Israelite","kjv_def":"Naam."},
"H5278":{"lemma":"נֹעַם","xlit":"nôʻam","pron":"no'-am","derivation":"from H5276 (נָעֵם);","strongs_def":"agreeableness, i.e. delight, suitableness, splendor or grace","kjv_def":"beauty, pleasant(-ness)."},
"H5279":{"lemma":"נַעֲמָה","xlit":"Naʻămâh","pron":"nah-am-aw'","derivation":"feminine of H5277 (נַעַם); pleasantness;","strongs_def":"Naamah, the name of an antediluvian woman, of an Ammonitess, and of a place in Palestine","kjv_def":"Naamah."},
"H5280":{"lemma":"נַעֲמִי","xlit":"Naʻămîy","pron":"nah-am-ee'","derivation":"patronymic from H5283 (נַעֲמָן);","strongs_def":"a Naamanite, or descendant of Naaman (collectively)","kjv_def":"Naamites."},
"H5281":{"lemma":"נׇעֳמִי","xlit":"Noʻŏmîy","pron":"no-om-ee'","derivation":"from H5278 (נֹעַם); pleasant;","strongs_def":"Noomi, an Israelitess","kjv_def":"Naomi."},
"H5282":{"lemma":"נַעֲמָן","xlit":"naʻămân","pron":"nah-am-awn'","derivation":"from H5276 (נָעֵם);","strongs_def":"pleasantness (plural as concrete)","kjv_def":"pleasant."},
"H5283":{"lemma":"נַעֲמָן","xlit":"Naʻămân","pron":"nah-am-awn'","derivation":"the same as H5282 (נַעֲמָן);","strongs_def":"Naaman, the name of an Israelite and of a Damascene","kjv_def":"Naaman."},
"H5284":{"lemma":"נַעֲמָתִי","xlit":"Naʻămâthîy","pron":"nah-am-aw-thee'","derivation":"patrial from a place corresponding in name (but not identical) with H5279 (נַעֲמָה);","strongs_def":"a Naamathite, or inhabitant of Naamah","kjv_def":"Naamathite."},
"H5285":{"lemma":"נַעֲצוּץ","xlit":"naʻătsûwts","pron":"nah-ats-oots'","derivation":"from an unused root meaning to prick;","strongs_def":"probably a brier; by implication, a thicket of thorny bushes","kjv_def":"thorn."},
"H5286":{"lemma":"נָעַר","xlit":"nâʻar","pron":"naw-ar'","derivation":"a primitive root;","strongs_def":"to growl","kjv_def":"yell."},
"H5287":{"lemma":"נָעַר","xlit":"nâʻar","pron":"naw-ar'","derivation":"a primitive root (probably identical with H5286 (נָעַר), through the idea of the rustling of mane, which usually accompanies the lion's roar);","strongs_def":"to tumble about","kjv_def":"shake (off, out, self), overthrow, toss up and down."},
"H5288":{"lemma":"נַעַר","xlit":"naʻar","pron":"nah'-ar","derivation":"from H5287 (נָעַר);","strongs_def":"(concretely) a boy (as active), from the age of infancy to adolescence; by implication, a servant; also (by interch. of sex), a girl (of similar latitude in age)","kjv_def":"babe, boy, child, damsel (from the margin), lad, servant, young (man)."},
"H5289":{"lemma":"נַעַר","xlit":"naʻar","pron":"nah'-ar","derivation":"from H5287 (נָעַר) in its derivative sense of tossing about;","strongs_def":"a wanderer","kjv_def":"young one."},
"H5290":{"lemma":"נֹעַר","xlit":"nôʻar","pron":"no'-ar","derivation":"from H5287 (נָעַר); (compare H5288 (נַעַר))","strongs_def":"(abstractly) boyhood","kjv_def":"child, youth."},
"H5291":{"lemma":"נַעֲרָה","xlit":"naʻărâh","pron":"nah-ar-aw'","derivation":"feminine of H5288 (נַעַר);","strongs_def":"a girl (from infancy to adolescence)","kjv_def":"damsel, maid(-en), young (woman)."},
"H5292":{"lemma":"נַעֲרָה","xlit":"Naʻărâh","pron":"nah-ar-aw'","derivation":"the same as H5291 (נַעֲרָה);","strongs_def":"Naarah, the name of an Israelitess, and of a place in Palestine","kjv_def":"Naarah, Naarath."},
"H5293":{"lemma":"נַעֲרַי","xlit":"Naʻăray","pron":"nah-ar-ah'-ee","derivation":"from H5288 (נַעַר); youthful;","strongs_def":"Naarai, an Israelite","kjv_def":"Naarai."},
"H5294":{"lemma":"נְעַרְיָה","xlit":"Nᵉʻaryâh","pron":"neh-ar-yaw'","derivation":"from H5288 (נַעַר) and H3050 (יָהּ); servant of Jah;","strongs_def":"Nearjah, the name of two Israelites","kjv_def":"Neariah."},
"H5295":{"lemma":"נַעֲרָן","xlit":"Naʻărân","pron":"nah-ar-awn'","derivation":"from H5288 (נַעַר); juvenile;","strongs_def":"Naaran, a place in Palestine","kjv_def":"Naaran."},
"H5296":{"lemma":"נְעֹרֶת","xlit":"nᵉʻôreth","pron":"neh-o'-reth","derivation":"from H5287 (נָעַר);","strongs_def":"something shaken out, i.e. tow (as the refuse of flax)","kjv_def":"tow."},
"H5297":{"lemma":"נֹף","xlit":"Nôph","pron":"nofe","derivation":"a variation of H4644 (מֹף);","strongs_def":"Noph, the capital of Upper Egypt","kjv_def":"Noph."},
"H5298":{"lemma":"נֶפֶג","xlit":"Nepheg","pron":"neh'-feg","derivation":"from an unused root probably meaning to spring forth; a sprout;","strongs_def":"Nepheg, the name of two Israelites","kjv_def":"Nepheg."},
"H5299":{"lemma":"נָפָה","xlit":"nâphâh","pron":"naw-faw'","derivation":"from H5130 (נוּף) in the sense of lifting;","strongs_def":"a height; also a sieve","kjv_def":"border, coast, region, sieve."},
"H5300":{"lemma":"נְפוּשְׁסִים","xlit":"Nᵉphûwshᵉçîym","pron":"nef-oo-shes-eem'","derivation":"for H5304 (נְפִיסִים);","strongs_def":"Nephushesim, a Temple-servant","kjv_def":"Nephisesim (from the margin)."},
"H5301":{"lemma":"נָפַח","xlit":"nâphach","pron":"naw-fakh'","derivation":"a primitive root;","strongs_def":"to puff, in various applications (literally, to inflate, blow hard, scatter, kindle, expire; figuratively, to disesteem)","kjv_def":"blow, breath, give up, cause to lose (life), seething, snuff."},
"H5302":{"lemma":"נֹפַח","xlit":"Nôphach","pron":"no'-fakh","derivation":"from H5301 (נָפַח); a gust;","strongs_def":"Nophach, a place in Moab","kjv_def":"Nophah."},
"H5303":{"lemma":"נְפִיל","xlit":"nᵉphîyl","pron":"nef-eel'","derivation":"or נְפִל; from H5307 (נָפַל);","strongs_def":"properly, a feller, i.e. a bully or tyrant","kjv_def":"giant."},
"H5304":{"lemma":"נְפִיסִים","xlit":"Nᵉphîyçîym","pron":"nef-ee-seem'","derivation":"plural from an unused root meaning to scatter; expansions;","strongs_def":"Nephisim, a Temple-servant","kjv_def":"Nephusim (from the margin)."},
"H5305":{"lemma":"נָפִישׁ","xlit":"Nâphîysh","pron":"naw-feesh'","derivation":"from H5314 (נָפַשׁ); refreshed;","strongs_def":"Naphish, a son of Ishmael, and his posterity","kjv_def":"Naphish."},
"H5306":{"lemma":"נֹפֶךְ","xlit":"nôphek","pron":"no'-fek","derivation":"from an unused root meaning to glisten;","strongs_def":"shining; a gem, probably the garnet","kjv_def":"emerald."},
"H5307":{"lemma":"נָפַל","xlit":"nâphal","pron":"naw-fal'","derivation":"a primitive root;","strongs_def":"to fall, in a great variety of applications (intransitive or causative, literal or figurative)","kjv_def":"be accepted, cast (down, self, (lots), out), cease, die, divide (by lot), (let) fail, (cause to, let, make, ready to) fall (away, down, -en, -ing), fell(-ing), fugitive, have (inheritance), inferior, be judged (by mistake for H6419 (פָּלַל)), lay (along), (cause to) lie down, light (down), be ([idiom] hast) lost, lying, overthrow, overwhelm, perish, present(-ed, -ing), (make to) rot, slay, smite out, [idiom] surely, throw down."},
"H5308":{"lemma":"נְפַל","xlit":"nᵉphal","pron":"nef-al'","derivation":"(Aramaic) corresponding to H5307 (נָפַל)","strongs_def":"{to fall, in a great variety of applications (intransitive or causative, literal or figurative)}","kjv_def":"fall (down), have occasion."},
"H5309":{"lemma":"נֶפֶל","xlit":"nephel","pron":"neh'-fel","derivation":"or נֵפֶל; from H5307 (נָפַל);","strongs_def":"something fallen, i.e. an abortion","kjv_def":"untimely birth."},
"H5310":{"lemma":"נָפַץ","xlit":"nâphats","pron":"naw-fats'","derivation":"a primitive root;","strongs_def":"to dash to pieces, or scatter","kjv_def":"be beaten in sunder, break (in pieces), broken, dash (in pieces), cause to be discharged, dispersed, be overspread, scatter."},
"H5311":{"lemma":"נֶפֶץ","xlit":"nephets","pron":"neh'-fets","derivation":"from H5310 (נָפַץ);","strongs_def":"a storm (as dispersing)","kjv_def":"scattering."},
"H5312":{"lemma":"נְפַק","xlit":"nᵉphaq","pron":"nef-ak'","derivation":"(Aramaic) a primitive root;","strongs_def":"to issue; causatively, to bring out","kjv_def":"come (go, take) forth (out)."},
"H5313":{"lemma":"נִפְקָא","xlit":"niphqâʼ","pron":"nif-kaw'","derivation":"(Aramaic) from H5312 (נְפַק);","strongs_def":"an outgo, i.e. expense","kjv_def":"expense."},
"H5314":{"lemma":"נָפַשׁ","xlit":"nâphash","pron":"naw-fash'","derivation":"a primitive root;","strongs_def":"to breathe; passively, to be breathed upon , i.e. (figuratively) refreshed (as if by a current of air)","kjv_def":"(be) refresh selves (-ed)."},
"H5315":{"lemma":"נֶפֶשׁ","xlit":"nephesh","pron":"neh'-fesh","derivation":"from H5314 (נָפַשׁ);","strongs_def":"properly, a breathing creature, i.e. animal of (abstractly) vitality; used very widely in a literal, accommodated or figurative sense (bodily or mental)","kjv_def":"any, appetite, beast, body, breath, creature, [idiom] dead(-ly), desire, [idiom] (dis-) contented, [idiom] fish, ghost, [phrase] greedy, he, heart(-y), (hath, [idiom] jeopardy of) life ([idiom] in jeopardy), lust, man, me, mind, mortally, one, own, person, pleasure, (her-, him-, my-, thy-) self, them (your) -selves, [phrase] slay, soul, [phrase] tablet, they, thing, ([idiom] she) will, [idiom] would have it."},
"H5316":{"lemma":"נֶפֶת","xlit":"nepheth","pron":"neh'-feth","derivation":"for H5299 (נָפָה);","strongs_def":"a height","kjv_def":"country."},
"H5317":{"lemma":"נֹפֶת","xlit":"nôpheth","pron":"no'-feth","derivation":"from H5130 (נוּף) in the sense of shaking to pieces;","strongs_def":"a dripping i.e. of honey (from the comb)","kjv_def":"honeycomb."},
"H5318":{"lemma":"נֶפְתּוֹחַ","xlit":"Nephtôwach","pron":"nef-to'-akh","derivation":"from H6605 (פָּתַח); opened, i.e. a spring;","strongs_def":"Nephtoach, a place in Palestine","kjv_def":"Neptoah."},
"H5319":{"lemma":"נַפְתּוּל","xlit":"naphtûwl","pron":"naf-tool'","derivation":"from H6617 (פָּתַל);","strongs_def":"properly, wrestled; but used (in the plural) transitively, a struggle","kjv_def":"wrestling."},
"H5320":{"lemma":"נַפְתֻּחִים","xlit":"Naphtuchîym","pron":"naf-too-kheem","derivation":"plural of foreign origin,","strongs_def":"Naphtuchim, an Egyptian tribe","kjv_def":"Naptuhim."},
"H5321":{"lemma":"נַפְתָּלִי","xlit":"Naphtâlîy","pron":"naf-taw-lee'","derivation":"from H6617 (פָּתַל); my wrestling;","strongs_def":"Naphtali, a son of Jacob, with the tribe descended from him, and its territory","kjv_def":"Naphtali."},
"H5322":{"lemma":"נֵץ","xlit":"nêts","pron":"nayts","derivation":"from H5340 (נָצַץ);","strongs_def":"a flower (from its brilliancy); also a hawk (from it flashing speed)","kjv_def":"blossom, hawk."},
"H5323":{"lemma":"נָצָא","xlit":"nâtsâʼ","pron":"naw-tsaw'","derivation":"a primitive root;","strongs_def":"to go away","kjv_def":"flee."},
"H5324":{"lemma":"נָצַב","xlit":"nâtsab","pron":"naw-tsab'","derivation":"a primitive root;","strongs_def":"to station, in various applications (literally or figuratively)","kjv_def":"appointed, deputy, erect, establish, [idiom] Huzzah (by mistake for a proper name), lay, officer, pillar, present, rear up, set (over, up), settle, sharpen, establish, (make to) stand(-ing, still, up, upright), best state."},
"H5325":{"lemma":"נִצָּב","xlit":"nitstsâb","pron":"nits-twawb'","derivation":"passive participle of H5324 (נָצַב);","strongs_def":"fixed, i.e. a handle","kjv_def":"haft."},
"H5326":{"lemma":"נִצְבָּה","xlit":"nitsbâh","pron":"nits-baw'","derivation":"(Aramaic) from a root corresponding to H5324 (נָצַב);","strongs_def":"fixedness, i.e. firmness","kjv_def":"strength."},
"H5327":{"lemma":"נָצָה","xlit":"nâtsâh","pron":"naw-tsaw'","derivation":"a primitive root;","strongs_def":"properly, to go forth, i.e. (by implication) to be expelled, and (consequently) desolate; causatively, to lay waste; also (specifically), to quarrel","kjv_def":"be laid waste, runinous, strive (together)."},
"H5328":{"lemma":"נִצָּה","xlit":"nitstsâh","pron":"nits-tsaw'","derivation":"feminine of H5322 (נֵץ);","strongs_def":"a blossom","kjv_def":"flower."},
"H5329":{"lemma":"נָצַח","xlit":"nâtsach","pron":"naw-tsakh'","derivation":"a primitive root;  also as denominative from H5331 (נֶצַח)","strongs_def":"properly, to glitter from afar, i.e. to be eminent (as a superintendent, especially of the Temple services and its music); to be permanent","kjv_def":"excel, chief musician (singer), oversee(-r), set forward."},
"H5330":{"lemma":"נְצַח","xlit":"nᵉtsach","pron":"nets-akh'","derivation":"(Aramaic) corresponding to H5329 (נָצַח);","strongs_def":"to become chief","kjv_def":"be preferred."},
"H5331":{"lemma":"נֶצַח","xlit":"netsach","pron":"neh'-tsakh","derivation":"or נֵצַח; from H5329 (נָצַח);","strongs_def":"properly, a goal, i.e. the bright object at adistance travelled towards; hence (figuratively), splendor, or (subjectively) truthfulness, or (objectively) confidence; but usually (adverbially), continually (i.e. to the most distant point of view)","kjv_def":"alway(-s), constantly, end, ([phrase] n-) ever(more), perpetual, strength, victory."},
"H5332":{"lemma":"נֵצַח","xlit":"Nêtsach","pron":"nay'-tsakh","derivation":"probably identical with H5331 (נֶצַח), through the idea of brilliancy of color;","strongs_def":"juice of the grape (as blood red)","kjv_def":"blood, strength."},
"H5333":{"lemma":"נְצִיב","xlit":"nᵉtsîyb","pron":"nets-eeb'","derivation":"or נְצִב; from H5324 (נָצַב);","strongs_def":"something stationary, i.e. a prefect, a military post, a statue","kjv_def":"garrison, officer, pillar."},
"H5334":{"lemma":"נְצִיב","xlit":"Nᵉtsîyb","pron":"nets-eeb'","derivation":"the same as H5333 (נְצִיב); station;","strongs_def":"Netsib, a place in Palestine","kjv_def":"Nezib."},
"H5335":{"lemma":"נְצִיחַ","xlit":"nᵉtsîyach","pron":"nets-ee'-akh","derivation":"from H5329 (נָצַח); conspicuous;","strongs_def":"Netsiach, a Temple-servant","kjv_def":"Neziah."},
"H5336":{"lemma":"נָצִיר","xlit":"nâtsîyr","pron":"naw-tsere'","derivation":"from H5341 (נָצַר);","strongs_def":"properly, conservative; but used passively, delivered","kjv_def":"preserved."},
"H5337":{"lemma":"נָצַל","xlit":"nâtsal","pron":"naw-tsal'","derivation":"a primitive root;","strongs_def":"to snatch away, whether in a good or a bad sense","kjv_def":"[idiom] at all, defend, deliver (self), escape, [idiom] without fail, part, pluck, preserve, recover, rescue, rid, save, spoil, strip, [idiom] surely, take (out)."},
"H5338":{"lemma":"נְצַל","xlit":"nᵉtsal","pron":"nets-al'","derivation":"(Aramaic) corresponding to H5337 (נָצַל);","strongs_def":"to extricate","kjv_def":"deliver, rescue."},
"H5339":{"lemma":"נִצָּן","xlit":"nitstsân","pron":"nits-tsawn'","derivation":"from H5322 (נֵץ);","strongs_def":"a blossom","kjv_def":"flower."},
"H5340":{"lemma":"נָצַץ","xlit":"nâtsats","pron":"naw-tsats'","derivation":"a primitive root;","strongs_def":"to glare, i.e. be bright-colored","kjv_def":"sparkle."},
"H5341":{"lemma":"נָצַר","xlit":"nâtsar","pron":"naw-tsar'","derivation":"a primitive root;","strongs_def":"to guard, in a good sense (to protect, maintain, obey, etc.) or a bad one (to conceal, etc.)","kjv_def":"besieged, hidden thing, keep(-er, -ing), monument, observe, preserve(-r), subtil, watcher(-man)."},
"H5342":{"lemma":"נֵצֶר","xlit":"nêtser","pron":"nay'-tser","derivation":"from H5341 (נָצַר) in the sense of greenness as a striking color;","strongs_def":"a shoot; figuratively, a descendant","kjv_def":"branch."},
"H5343":{"lemma":"נְקֵא","xlit":"nᵉqêʼ","pron":"nek-ay'","derivation":"(Aramaic) from a root corresponding to H5352 (נָקָה);","strongs_def":"clean","kjv_def":"pure."},
"H5344":{"lemma":"נָקַב","xlit":"nâqab","pron":"naw-kab'","derivation":"a primitive root;","strongs_def":"to puncture, literally (to perforate, with more or less violence) or figuratively (to specify, designate, libel)","kjv_def":"appoint, blaspheme, bore, curse, express, with holes, name, pierce, strike through."},
"H5345":{"lemma":"נֶקֶב","xlit":"neqeb","pron":"neh'keb","strongs_def":"a bezel (for a gem)","kjv_def":"pipe."},
"H5346":{"lemma":"נֶקֶב","xlit":"Neqeb","pron":"neh'-keb","derivation":"the same as H5345 (נֶקֶב); dell;","strongs_def":"Nekeb, a place in Palestine","kjv_def":"Nekeb."},
"H5347":{"lemma":"נְקֵבָה","xlit":"nᵉqêbâh","pron":"nek-ay-baw'","derivation":"from H5344 (נָקַב);","strongs_def":"female (from the sexual form)","kjv_def":"female."},
"H5348":{"lemma":"נָקֹד","xlit":"nâqôd","pron":"naw-kode'","derivation":"from an unused root meaning to mark (by puncturing or branding);","strongs_def":"spotted","kjv_def":"speckled."},
"H5349":{"lemma":"נֹקֵד","xlit":"nôqêd","pron":"no-kade'","derivation":"active participle from the same as H5348 (נָקֹד);","strongs_def":"a spotter (of sheep or cattle), i.e. the owner or tender (who thus marks them)","kjv_def":"herdman, sheepmaster."},
"H5350":{"lemma":"נִקֻּד","xlit":"niqqud","pron":"nik-kood'","derivation":"from the same as H5348 (נָקֹד);","strongs_def":"a crumb (as broken to spots); also a biscuit (as pricked)","kjv_def":"cracknel, mouldy."},
"H5351":{"lemma":"נְקֻדָּה","xlit":"nᵉquddâh","pron":"ned-ood-daw'","derivation":"feminine of H5348 (נָקֹד);","strongs_def":"a boss","kjv_def":"stud."},
"H5352":{"lemma":"נָקָה","xlit":"nâqâh","pron":"naw-kaw'","derivation":"a primitive root;","strongs_def":"to be (or make) clean (literally or figuratively); by implication (in an adverse sense) to be bare, i.e. extirpated","kjv_def":"acquit [idiom] at all, [idiom] altogether, be blameless, cleanse, (be) clear(-ing), cut off, be desolate, be free, be (hold) guiltless, be (hold) innocent, [idiom] by no means, be quit, be (leave) unpunished, [idiom] utterly, [idiom] wholly."},
"H5353":{"lemma":"נְקוֹדָא","xlit":"Nᵉqôwdâʼ","pron":"nek-o-daw'","derivation":"feminine of H5348 (נָקֹד) (in the figurative sense of marked); distinction;","strongs_def":"Nekoda, a Temple-servant","kjv_def":"Nekoda."},
"H5354":{"lemma":"נָקַט","xlit":"nâqaṭ","pron":"naw-kat'","derivation":"a primitive root;","strongs_def":"to loathe","kjv_def":"weary."},
"H5355":{"lemma":"נָקִי","xlit":"nâqîy","pron":"naw-kee'","derivation":"or נָקִיא; (Joel 4:19; Jonah [1:14), from H5352 (נָקָה);","strongs_def":"innocent","kjv_def":"blameless, clean, clear, exempted, free, guiltless, innocent, quit."},
"H5356":{"lemma":"נִקָּיוֹן","xlit":"niqqâyôwn","pron":"nik-kaw-yone'","derivation":"or נִקָּיֹן; from H5352 (נָקָה);","strongs_def":"clearness (literally or figuratively)","kjv_def":"cleanness, innocency."},
"H5357":{"lemma":"נָקִיק","xlit":"nâqîyq","pron":"naw-keek'","derivation":"from an unused root meaning to bore;","strongs_def":"a cleft","kjv_def":"hole."},
"H5358":{"lemma":"נָקַם","xlit":"nâqam","pron":"naw-kam'","derivation":"a primitive root;","strongs_def":"to grudge, i.e. avenge or punish","kjv_def":"avenge(-r, self), punish, revenge (self), [idiom] surely, take vengeance."},
"H5359":{"lemma":"נָקָם","xlit":"nâqâm","pron":"naw-kawm'","derivation":"from H5358 (נָקַם);","strongs_def":"revenge","kjv_def":"[phrase] avenged, quarrel, vengeance."},
"H5360":{"lemma":"נְקָמָה","xlit":"nᵉqâmâh","pron":"nek-aw-maw'","derivation":"feminine of H5359 (נָקָם);","strongs_def":"avengement, whether the act of the passion","kjv_def":"[phrase] avenge, revenge(-ing), vengeance."},
"H5361":{"lemma":"נָקַע","xlit":"nâqaʻ","pron":"naw-kah'","derivation":"a primitive root;","strongs_def":"to feel aversion","kjv_def":"be alienated."},
"H5362":{"lemma":"נָקַף","xlit":"nâqaph","pron":"naw-kaf'","derivation":"a primitive root;","strongs_def":"to strike with more or less violence (beat, fell, corrode); by implication (of attack) to knock together, i.e. surround or circulate","kjv_def":"compass (about, -ing), cut down, destroy, go round (about), inclose, round."},
"H5363":{"lemma":"נֹקֶף","xlit":"nôqeph","pron":"no'-kef","derivation":"from H5362 (נָקַף);","strongs_def":"a threshing (of olives)","kjv_def":"shaking."},
"H5364":{"lemma":"נִקְפָּה","xlit":"niqpâh","pron":"nik-paw'","derivation":"from H5362 (נָקַף);","strongs_def":"probably a rope (as encircling)","kjv_def":"rent."},
"H5365":{"lemma":"נָקַר","xlit":"nâqar","pron":"naw-kar'","derivation":"a primitive root;","strongs_def":"to bore (penetrate, quarry)","kjv_def":"dig, pick out, pierce, put (thrust) out."},
"H5366":{"lemma":"נְקָרָה","xlit":"nᵉqârâh","pron":"nek-aw-raw'","derivation":"from H5365 (נָקַר), a fissure","strongs_def":"{to bore (penetrate, quarry)}","kjv_def":"cleft, clift."},
"H5367":{"lemma":"נָקַשׁ","xlit":"nâqash","pron":"naw-kash'","derivation":"a primitive root;","strongs_def":"to entrap (with a noose), literally or figuratively","kjv_def":"catch (lay a) snare."},
"H5368":{"lemma":"נְקַשׁ","xlit":"nᵉqash","pron":"nek-ash'","derivation":"(Aramaic) corresponding to H5367 (נָקַשׁ); used in the sense of H5362 (נָקַף);","strongs_def":"to knock","kjv_def":"smote."},
"H5369":{"lemma":"נֵר","xlit":"Nêr","pron":"nare","derivation":"the same as H5216 (נִיר); lamp;","strongs_def":"Ner, an Israelite","kjv_def":"Ner."},
"H5370":{"lemma":"נֵרְגַּל","xlit":"Nêrᵉgal","pron":"nare-gal'","derivation":"of foreign origin;","strongs_def":"Nergal, a Cuthite deity","kjv_def":"Nergal."},
"H5371":{"lemma":"נֵרְגַּל שַׁרְאֶצֶר","xlit":"Nêrᵉgal Sharʼetser","pron":"nare-gal' shar-eh'-tser","derivation":"from H5370 (נֵרְגַּל) and H8272 (שַׁרְאֶצֶר);","strongs_def":"Nergal-Sharetser, the name of two Babylonians","kjv_def":"Nergal-sharezer."},
"H5372":{"lemma":"נִרְגָּן","xlit":"nirgân","pron":"neer-gawn'","derivation":"from an unused root meaning to roll to pieces;","strongs_def":"a slanderer","kjv_def":"talebearer, whisperer."},
"H5373":{"lemma":"נֵרְדְּ","xlit":"nêrd","pron":"nayrd","derivation":"of foreign origin;","strongs_def":"nard, an aromatic","kjv_def":"spikenard."},
"H5374":{"lemma":"נֵרִיָּה","xlit":"Nêrîyâh","pron":"nay-ree-yaw'","derivation":"or נֵרִיָּהוּ; from H5216 (נִיר) and H3050 (יָהּ); light of Jah;","strongs_def":"Nerijah, an Israelite","kjv_def":"Neriah."},
"H5375":{"lemma":"נָשָׂא","xlit":"nâsâʼ","pron":"naw-saw'","derivation":"or נָסָה; (Psalm 4:6 (H7 (אֲבַד))), a primitive root;","strongs_def":"to lift, in a great variety of applications, literal and figurative, absolute and relative","kjv_def":"accept, advance, arise, (able to, (armor), suffer to) bear(-er, up), bring (forth), burn, carry (away), cast, contain, desire, ease, exact, exalt (self), extol, fetch, forgive, furnish, further, give, go on, help, high, hold up, honorable ([phrase] man), lade, lay, lift (self) up, lofty, marry, magnify, [idiom] needs, obtain, pardon, raise (up), receive, regard, respect, set (up), spare, stir up, [phrase] swear, take (away, up), [idiom] utterly, wear, yield."},
"H5376":{"lemma":"נְשָׂא","xlit":"nᵉsâʼ","pron":"nes-aw'","derivation":"(Aramaic) corresponding to H5375 (נָשָׂא)","strongs_def":"{to lift}","kjv_def":"carry away, make insurrection, take."},
"H5377":{"lemma":"נָשָׁא","xlit":"nâshâʼ","pron":"naw-shaw'","derivation":"a primitive root;","strongs_def":"to lead astray, i.e. (mentally) to delude, or (morally) to seduce","kjv_def":"beguile, deceive, [idiom] greatly, [idiom] utterly."},
"H5378":{"lemma":"נָשָׁא","xlit":"nâshâʼ","pron":"naw-shaw'","derivation":"a primitive root (perhaps identical with H5377 (נָשָׁא), through the idea of imposition);","strongs_def":"to lend on interest; by implication, to dun fordebt","kjv_def":"[idiom] debt, exact, giver of usury."},
"H5379":{"lemma":"נִשֵּׂאת","xlit":"nissêʼth","pron":"nis-sayth'","derivation":"passive participle feminine of H5375 (נָשָׂא);","strongs_def":"something taken, i.e. a present","kjv_def":"gift."},
"H5380":{"lemma":"נָשַׁב","xlit":"nâshab","pron":"naw-shab'","derivation":"a primitive root;","strongs_def":"to blow; by implication, to disperse","kjv_def":"(cause to) blow, drive away."},
"H5381":{"lemma":"נָשַׂג","xlit":"nâsag","pron":"naw-sag'","derivation":"a primitive root;","strongs_def":"to reach (literally or figuratively)","kjv_def":"ability, be able, attain (unto), (be able to, can) get, lay at, put, reach, remove, wax rich, [idiom] surely, (over-) take (hold of, on, upon)."},
"H5382":{"lemma":"נָשָׁה","xlit":"nâshâh","pron":"naw-shaw'","derivation":"a primitive root;","strongs_def":"to forget; figuratively, to neglect; causatively, to remit, remove","kjv_def":"forget, deprive, exact."},
"H5383":{"lemma":"נָשָׁה","xlit":"nâshâh","pron":"naw-shaw'","derivation":"a primitive root (rather identical with H5382 (נָשָׁה), in the sense of H5378 (נָשָׁא));","strongs_def":"to lend or (by reciprocity) borrow on security or interest","kjv_def":"creditor, exact, extortioner, lend, usurer, lend on (taker on) usury."},
"H5384":{"lemma":"נָשֶׁה","xlit":"nâsheh","pron":"naw-sheh'","derivation":"from H5382 (נָשָׁה), in the sense of failure;","strongs_def":"rheumatic or crippled (from the incident to Jacob)","kjv_def":"which shrank."},
"H5385":{"lemma":"נְשׂוּאָה","xlit":"nᵉsûwʼâh","pron":"nes-oo-aw'","derivation":"or rather, נְשֻׂאָה; feminine. passive participle of H5375 (נָשָׂא);","strongs_def":"something borne, i.e. a load","kjv_def":"carriage."},
"H5386":{"lemma":"נְשִׁי","xlit":"nᵉshîy","pron":"nesh-ee'","derivation":"from H5383 (נָשָׁה);","strongs_def":"a debt","kjv_def":"debt."},
"H5387":{"lemma":"נָשִׂיא","xlit":"nâsîyʼ","pron":"naw-see'","derivation":"or נָשִׂא; from H5375 (נָשָׂא);","strongs_def":"properly, an exalted one, i.e. a king or sheik; also a rising mist","kjv_def":"captain, chief, cloud, governor, prince, ruler, vapour."},
"H5388":{"lemma":"נְשִׁיָּה","xlit":"nᵉshîyâh","pron":"nesh-ee-yaw'","derivation":"from H5382 (נָשָׁה);","strongs_def":"oblivion","kjv_def":"forgetfulness."},
"H5389":{"lemma":"נָשִׁין","xlit":"nâshîyn","pron":"naw-sheen'","derivation":"(Aramaic) irregular plural feminine of H606 (אֱנָשׁ)","strongs_def":"{a man}","kjv_def":"women."},
"H5390":{"lemma":"נְשִׁיקָה","xlit":"nᵉshîyqâh","pron":"nesh-ee-kaw'","derivation":"from H5401 (נָשַׁק);","strongs_def":"a kiss","kjv_def":"kiss."},
"H5391":{"lemma":"נָשַׁךְ","xlit":"nâshak","pron":"naw-shak'","derivation":"a primitive root;","strongs_def":"to strike with a sting (as a serpent); figuratively, to oppress with interest on a loan","kjv_def":"bite, lend upon usury."},
"H5392":{"lemma":"נֶשֶׁךְ","xlit":"neshek","pron":"neh'-shek","derivation":"from H5391 (נָשַׁךְ);","strongs_def":"interest on a debt","kjv_def":"usury."},
"H5393":{"lemma":"נִשְׁכָּה","xlit":"nishkâh","pron":"nish-kaw'","derivation":"for H3957 (לִשְׁכָּה);","strongs_def":"a cell","kjv_def":"chamber."},
"H5394":{"lemma":"נָשַׁל","xlit":"nâshal","pron":"naw-shal'","derivation":"a primitive root;","strongs_def":"to pluck off, i.e. divest, eject or drop","kjv_def":"cast (out), drive, loose, put off (out), slip."},
"H5395":{"lemma":"נָשַׁם","xlit":"nâsham","pron":"naw-sham'","derivation":"a primitive root;","strongs_def":"properly, to blow away, i.e. destroy","kjv_def":"destroy."},
"H5396":{"lemma":"נִשְׁמָא","xlit":"nishmâʼ","pron":"nish-maw'","derivation":"(Aramaic) corresponding to H5397 (נְשָׁמָה);","strongs_def":"vital breath","kjv_def":"breath."},
"H5397":{"lemma":"נְשָׁמָה","xlit":"nᵉshâmâh","pron":"nesh-aw-maw'","derivation":"from H5395 (נָשַׁם);","strongs_def":"a puff, i.e. wind, angry or vital breath, divine inspiration, intellect. or (concretely) an animal","kjv_def":"blast, (that) breath(-eth), inspiration, soul, spirit."},
"H5398":{"lemma":"נָשַׁף","xlit":"nâshaph","pron":"naw-shaf'","derivation":"a primitive root;","strongs_def":"to breeze, i.e. blow up fresh (as the wind)","kjv_def":"blow."},
"H5399":{"lemma":"נֶשֶׁף","xlit":"nesheph","pron":"neh'-shef","derivation":"from H5398 (נָשַׁף);","strongs_def":"properly, a breeze, i.e. (by implication) dusk (when the evening breeze prevails)","kjv_def":"dark, dawning of the day (morning), night, twilight."},
"H5400":{"lemma":"נָשַׂק","xlit":"nâsaq","pron":"naw-sak'","derivation":"a primitive root;","strongs_def":"to catch fire","kjv_def":"burn, kindle."},
"H5401":{"lemma":"נָשַׁק","xlit":"nâshaq","pron":"naw-shak'","derivation":"a primitive root (identical with H5400 (נָשַׂק), through the idea of fastening up; compare H2388 (חָזַק), H2836 (חָשַׁק));","strongs_def":"to kiss, literally or figuratively (touch); also (as a mode of attachment), to equip with weapons","kjv_def":"armed (men), rule, kiss, that touched."},
"H5402":{"lemma":"נֶשֶׁק","xlit":"nesheq","pron":"neh'-shek","derivation":"or נֵשֶׁק; from H5401 (נָשַׁק);","strongs_def":"military equipment, i.e. (collectively) arms (offensive or defensive), or (concretely) an arsenal","kjv_def":"armed men, armour(-y), battle, harness, weapon."},
"H5403":{"lemma":"נְשַׁר","xlit":"nᵉshar","pron":"nesh-ar'","derivation":"(Aramaic) corresponding to H5404 (נֶשֶׁר);","strongs_def":"an eagle","kjv_def":"eagle."},
"H5404":{"lemma":"נֶשֶׁר","xlit":"nesher","pron":"neh'-sher","derivation":"from an unused root meaning to lacerate;","strongs_def":"the eagle (or other large bird of prey)","kjv_def":"eagle."},
"H5405":{"lemma":"נָשַׁת","xlit":"nâshath","pron":"naw-shath'","derivation":"a primitive root;","strongs_def":"properly, to eliminate, i.e. (intransitively) to dry up","kjv_def":"fail."},
"H5406":{"lemma":"נִשְׁתְּוָן","xlit":"nishtᵉvân","pron":"nish-tev-awn'","derivation":"probably of Persian origin;","strongs_def":"an epistle","kjv_def":"letter."},
"H5407":{"lemma":"נִשְׁתְּוָן","xlit":"nishtᵉvân","pron":"nish-tev-awn'","derivation":"(Aramaic) corresponding to H5406 (נִשְׁתְּוָן)","strongs_def":"{an epistle}","kjv_def":"letter."},
"H5408":{"lemma":"נָתַח","xlit":"nâthach","pron":"naw-thakh'","derivation":"a primitive root;","strongs_def":"to dismember","kjv_def":"cut (in pieces), divide, hew in pieces."},
"H5409":{"lemma":"נֵתַח","xlit":"nêthach","pron":"nay'-thakh","derivation":"from H5408 (נָתַח);","strongs_def":"a fragment","kjv_def":"part, piece."},
"H5410":{"lemma":"נָתִיב","xlit":"nâthîyb","pron":"naw-theeb'","derivation":"or (feminine) נְתִיבָה; or נְתִבָה; (Jeremiah 6:16), from an unused root meaning to tramp;","strongs_def":"a (beaten) track","kjv_def":"path(-way), [idiom] travel(-ler), way."},
"H5411":{"lemma":"נָתִין","xlit":"Nâthîyn","pron":"naw-theen'","derivation":"or נָתוּן (Ezra 8:17), (the proper form as passive participle), from H5414 (נָתַן); one given, i.e. (in the plural only)","strongs_def":"the Nethinim, or Temple-servants (as given to that duty)","kjv_def":"Nethinims."},
"H5412":{"lemma":"נְתִין","xlit":"Nᵉthîyn","pron":"netheen'","derivation":"(Aramaic) corresponding to H5411 (נָתִין)","strongs_def":"{the Nethinim, or Temple-servants (as given to that duty)}","kjv_def":"Nethinims."},
"H5413":{"lemma":"נָתַךְ","xlit":"nâthak","pron":"naw-thak'","derivation":"a primitive root;","strongs_def":"to flow forth (literally or figuratively); by implication, to liquify","kjv_def":"drop, gather (together), melt, pour (forth, out)."},
"H5414":{"lemma":"נָתַן","xlit":"nâthan","pron":"naw-than'","derivation":"a primitive root;","strongs_def":"to give, used with greatest latitude of application (put, make, etc.)","kjv_def":"add, apply, appoint, ascribe, assign, [idiom] avenge, [idiom] be (healed), bestow, bring (forth, hither), cast, cause, charge, come, commit, consider, count, [phrase] cry, deliver (up), direct, distribute, do, [idiom] doubtless, [idiom] without fail, fasten, frame, [idiom] get, give (forth, over, up), grant, hang (up), [idiom] have, [idiom] indeed, lay (unto charge, up), (give) leave, lend, let (out), [phrase] lie, lift up, make, [phrase] O that, occupy, offer, ordain, pay, perform, place, pour, print, [idiom] pull, put (forth), recompense, render, requite, restore, send (out), set (forth), shew, shoot forth (up), [phrase] sing, [phrase] slander, strike, (sub-) mit, suffer, [idiom] surely, [idiom] take, thrust, trade, turn, utter, [phrase] weep, [phrase] willingly, [phrase] withdraw, [phrase] would (to) God, yield."},
"H5415":{"lemma":"נְתַן","xlit":"nᵉthan","pron":"neth-an'","derivation":"(Aramaic) corresponding to H5414 (נָתַן);","strongs_def":"give","kjv_def":"bestow, give pay."},
"H5416":{"lemma":"נָתָן","xlit":"Nâthân","pron":"naw-thawn'","derivation":"from H5414 (נָתַן); given;","strongs_def":"Nathan, the name of five Israelites","kjv_def":"Nathan."},
"H5417":{"lemma":"נְתַנְאֵל","xlit":"Nᵉthanʼêl","pron":"neth-an-ale'","derivation":"from H5414 (נָתַן) and H410 (אֵל); given of God;","strongs_def":"Nethanel, the name of ten Israelites","kjv_def":"Nethaneel."},
"H5418":{"lemma":"נְתַנְיָה","xlit":"Nᵉthanyâh","pron":"neth-an-yaw'","derivation":"or נְתַנְיָהוּ; from H5414 (נָתַן) and H3050 (יָהּ); given of Jah;","strongs_def":"Nethanjah, the name of four Israelites","kjv_def":"Nethaniah."},
"H5419":{"lemma":"נְתַן־מֶלֶךְ","xlit":"Nᵉthan-Melek","pron":"neth-an' meh'-lek","derivation":"from H5414 (נָתַן) and H4428 (מֶלֶךְ); given of (the) king;","strongs_def":"Nethan-Melek, an Israelite","kjv_def":"Nathanmelech."},
"H5420":{"lemma":"נָתָס","xlit":"nâthâç","pron":"naw-thas'","derivation":"a primitive root;","strongs_def":"to tear up","kjv_def":"mar."},
"H5421":{"lemma":"נָתַע","xlit":"nâthaʻ","pron":"naw-thah'","derivation":"for H5422 (נָתַץ);","strongs_def":"to tear out","kjv_def":"break."},
"H5422":{"lemma":"נָתַץ","xlit":"nâthats","pron":"naw-thats'","derivation":"a primitive root;","strongs_def":"to tear down","kjv_def":"beat down, break down (out), cast down, destroy, overthrow, pull down, throw down."},
"H5423":{"lemma":"נָתַק","xlit":"nâthaq","pron":"naw-thak'","derivation":"a primitive root;","strongs_def":"to tear off","kjv_def":"break (off), burst, draw (away), lift up, pluck (away, off), pull (out), root out."},
"H5424":{"lemma":"נֶתֶק","xlit":"netheq","pron":"neh'-thek","derivation":"from H5423 (נָתַק);","strongs_def":"scurf","kjv_def":"(dry) scall."},
"H5425":{"lemma":"נָתַר","xlit":"nâthar","pron":"naw-thar'","derivation":"a primitive root;","strongs_def":"to jump, i.e. be violently agitated; causatively, to terrify, shake off, untie","kjv_def":"drive asunder, leap, (let) loose, [idiom] make, move, undo."},
"H5426":{"lemma":"נְתַר","xlit":"nᵉthar","pron":"neth-ar'","derivation":"(Aramaic) corresponding to H5425 (נָתַר)","strongs_def":"{to jump, i.e. be violently agitated; causatively, to terrify, shake off, untie}","kjv_def":"shake off."},
"H5427":{"lemma":"נֶתֶר","xlit":"nether","pron":"neh'-ther","derivation":"from H5425 (נָתַר);","strongs_def":"mineral potash (so called from effervescing with acid)","kjv_def":"nitre."},
"H5428":{"lemma":"נָתַשׁ","xlit":"nâthash","pron":"naw-thash'","derivation":"a primitive root;","strongs_def":"to tear away","kjv_def":"destroy, forsake, pluck (out, up, by the roots), pull up, root out (up), [idiom] utterly. s"},
"H5429":{"lemma":"סְאָה","xlit":"çᵉʼâh","pron":"seh-aw'","derivation":"from an unused root meaning to define;","strongs_def":"a seah, or certain measure (as determinative) forgrain","kjv_def":"measure."},
"H5430":{"lemma":"סְאוֹן","xlit":"çᵉʼôwn","pron":"seh-own'","derivation":"from H5431 (סָאַן);","strongs_def":"perhaps a military boot (as aprotection from mud","kjv_def":"battle."},
"H5431":{"lemma":"סָאַן","xlit":"çâʼan","pron":"saw-an'","derivation":"a primitive root; to be miry; used only as denominative from H5430 (סְאוֹן);","strongs_def":"to shoe, i.e. (active participle) a soldier shod","kjv_def":"warrior."},
"H5432":{"lemma":"סַאסְּאָה","xlit":"çaʼçᵉʼâh","pron":"sah-seh-aw'","derivation":"for H5429 (סְאָה);","strongs_def":"measurement, i.e. moderation","kjv_def":"measure."},
"H5433":{"lemma":"סָבָא","xlit":"çâbâʼ","pron":"saw-baw'","derivation":"a primitive root;","strongs_def":"to quaff to satiety, i.e. become tipsy","kjv_def":"drunkard, fill self, Sabean, (wine-) bibber."},
"H5434":{"lemma":"סְבָא","xlit":"Çᵉbâʼ","pron":"seb-aw'","derivation":"of foreign origin;","strongs_def":"Seba, a son of Cush, and the country settled by him","kjv_def":"Seba."},
"H5435":{"lemma":"סֹבֶא","xlit":"çôbeʼ","pron":"so'-beh","derivation":"from H5433 (סָבָא);","strongs_def":"potation, concretely (wine), or abstractly (carousal)","kjv_def":"drink, drunken, wine."},
"H5436":{"lemma":"סְבָאִי","xlit":"Çᵉbâʼîy","pron":"seb-aw-ee'","derivation":"patrial from H5434 (סְבָא);","strongs_def":"a Sebaite, or inhabitant of Seba","kjv_def":"Sabean."},
"H5437":{"lemma":"סָבַב","xlit":"çâbab","pron":"saw-bab'","derivation":"a primitive root;","strongs_def":"to revolve, surround, or border; used in various applications, literally and figuratively","kjv_def":"bring, cast, fetch, lead, make, walk, [idiom] whirl, [idiom] round about, be about on every side, apply, avoid, beset (about), besiege, bring again, carry (about), change, cause to come about, [idiom] circuit, (fetch a) compass (about, round), drive, environ, [idiom] on every side, beset (close, come, compass, go, stand) round about, inclose, remove, return, set, sit down, turn (self) (about, aside, away, back)."},
"H5438":{"lemma":"סִבָּה","xlit":"çibbâh","pron":"sib-baw'","derivation":"from H5437 (סָבַב);","strongs_def":"a (providential) turn (of affairs)","kjv_def":"cause."},
"H5439":{"lemma":"סָבִיב","xlit":"çâbîyb","pron":"saw-beeb'","derivation":"or (feminine) סְבִיבָה; from H5437 (סָבַב);","strongs_def":"(as noun) a circle, neighbour, or environs; but chiefly (as adverb, with or without preposition) around","kjv_def":"(place, round) about, circuit, compass, on every side."},
"H5440":{"lemma":"סָבַךְ","xlit":"çâbak","pron":"saw-bak'","derivation":"a primitive root;","strongs_def":"to entwine","kjv_def":"fold together, wrap."},
"H5441":{"lemma":"סֹבֶךְ","xlit":"çôbek","pron":"so'-bek","derivation":"from H5440 (סָבַךְ);","strongs_def":"a copse","kjv_def":"thicket."},
"H5442":{"lemma":"סְבָךְ","xlit":"çᵉbâk","pron":"seb-awk'","derivation":"from H5440 (סָבַךְ),","strongs_def":"a copse","kjv_def":"thick(-et)."},
"H5443":{"lemma":"סַבְּכָא","xlit":"çabbᵉkâʼ","pron":"sab-bek-aw'","derivation":"(Aramaic) or שַׂבְּכָא; (Aramaic), from a root corresponding to H5440 (סָבַךְ);","strongs_def":"a lyre","kjv_def":"sackbut."},
"H5444":{"lemma":"סִבְּכַי","xlit":"Çibbᵉkay","pron":"sib-bek-ah'-ee","derivation":"from H5440 (סָבַךְ); copse-like;","strongs_def":"Sibbecai, an Israelite","kjv_def":"Sibbecai, Sibbechai."},
"H5445":{"lemma":"סָבַל","xlit":"çâbal","pron":"saw-bal'","derivation":"a primitive root;","strongs_def":"to carry (literally or figuratively), or (reflexively) be burdensome; specifically, to be gravid","kjv_def":"bear, be a burden, carry, strong to labour."},
"H5446":{"lemma":"סְבַל","xlit":"çᵉbal","pron":"seb-al'","derivation":"(Aramaic) corresponding to H5445 (סָבַל);","strongs_def":"to erect","kjv_def":"strongly laid."},
"H5447":{"lemma":"סֵבֶל","xlit":"çêbel","pron":"say'-bel","derivation":"from H5445 (סָבַל);","strongs_def":"a load (literally or figuratively)","kjv_def":"burden, charge."},
"H5448":{"lemma":"סֹבֶל","xlit":"çôbel","pron":"so'-bel","derivation":"(only in the form סֻבָּל; from H5445 (סָבַל);","strongs_def":"a load (figuratively)","kjv_def":"burden."},
"H5449":{"lemma":"סַבָּל","xlit":"çabbâl","pron":"sab-bawl'","derivation":"from H5445 (סָבַל);","strongs_def":"a porter","kjv_def":"(to bear, bearer of) burden(-s)."},
"H5450":{"lemma":"סְבָלָה","xlit":"çᵉbâlâh","pron":"seb-aw-law'","derivation":"from H5447 (סֵבֶל);","strongs_def":"porterage","kjv_def":"burden."},
"H5451":{"lemma":"סִבֹּלֶת","xlit":"çibbôleth","pron":"sib-bo'-leth","derivation":"for H7641 (שִׁבֹּל);","strongs_def":"an ear of grain","kjv_def":"Sibboleth."},
"H5452":{"lemma":"סְבַר","xlit":"çᵉbar","pron":"seb-ar'","derivation":"(Aramaic) a primitive root;","strongs_def":"to bear in mind, i.e. hope","kjv_def":"think."},
"H5453":{"lemma":"סִבְרַיִם","xlit":"Çibrayim","pron":"sib-rah'-yim","derivation":"dual from a root corresponding to H5452 (סְבַר); double hope;","strongs_def":"Sibrajim, a place in Syria","kjv_def":"Sibraim."},
"H5454":{"lemma":"סַבְתָּא","xlit":"Çabtâʼ","pron":"sab-taw'","derivation":"or סַבְתָּה; probably of foreign derivation;","strongs_def":"Sabta or Sabtah, the name of a son of Cush, and the country occupied by his posterity","kjv_def":"Sabta, Sabtah."},
"H5455":{"lemma":"סַבְתְּכָא","xlit":"Çabtᵉkâʼ","pron":"sab-tek-aw'","derivation":"probably of foreign derivation;","strongs_def":"Sabteca, the name of a son of Cush, and the region settled by him","kjv_def":"Sabtecha, Sabtechah."},
"H5456":{"lemma":"סָגַד","xlit":"çâgad","pron":"saw-gad'","derivation":"a primitive root;","strongs_def":"to prostrate oneself (in homage)","kjv_def":"fall down."},
"H5457":{"lemma":"סְגִד","xlit":"çᵉgid","pron":"seg-eed'","derivation":"(Aramaic) corresponding to H5456 (סָגַד)","strongs_def":"{to prostrate oneself (in homage)}","kjv_def":"worship."},
"H5458":{"lemma":"סְגוֹר","xlit":"çᵉgôwr","pron":"seg-ore'","derivation":"from H5462 (סָגַר);","strongs_def":"properly, shut up, i.e. the breast (as enclosing the heart); also gold (as generally shut up safely)","kjv_def":"caul, gold."},
"H5459":{"lemma":"סְגֻלָּה","xlit":"çᵉgullâh","pron":"seg-ool-law'","derivation":"feminine passive participle of an unused root meaning to shut up;","strongs_def":"wealth (as closely shut up)","kjv_def":"jewel, peculiar (treasure), proper good, special."},
"H5460":{"lemma":"סְגַן","xlit":"çᵉgan","pron":"seg-an'","derivation":"(Aramaic) corresponding to H5461 (סָגָן)","strongs_def":"{a prefect of aprovince}","kjv_def":"governor."},
"H5461":{"lemma":"סָגָן","xlit":"çâgân","pron":"saw-gawn'","derivation":"from an unused root meaning to superintend;","strongs_def":"a prefect of a province","kjv_def":"prince, ruler."},
"H5462":{"lemma":"סָגַר","xlit":"çâgar","pron":"saw-gar'","derivation":"a primitive root;","strongs_def":"to shut up; figuratively, to surrender","kjv_def":"close up, deliver (up), give over (up), inclose, [idiom] pure, repair, shut (in, self, out, up, up together), stop, [idiom] straitly."},
"H5463":{"lemma":"סְגַר","xlit":"çᵉgar","pron":"seg-ar'","derivation":"(Aramaic) corresponding to H5462 (סָגַר)","strongs_def":"{to shut up; figuratively, to surrender}","kjv_def":"shut up."},
"H5464":{"lemma":"סַגְרִיד","xlit":"çagrîyd","pron":"sag-reed'","derivation":"probably from H5462 (סָגַר) in the sense of sweeping away;","strongs_def":"a pouring rain","kjv_def":"very rainy."},
"H5465":{"lemma":"סַד","xlit":"çad","pron":"sad","derivation":"from an unused root meaning to estop;","strongs_def":"the stocks","kjv_def":"stocks."},
"H5466":{"lemma":"סָדִין","xlit":"çâdîyn","pron":"saw-deen'","derivation":"from an unused root meaning to envelop;","strongs_def":"a wrapper, i.e. shirt","kjv_def":"fine linen, sheet."},
"H5467":{"lemma":"סְדֹם","xlit":"Çᵉdôm","pron":"sed-ome'","derivation":"from an unused root meaning to scorch; burnt (i.e. volcanic or bituminous) district;","strongs_def":"Sedom, a place near the Dead Sea","kjv_def":"Sodom."},
"H5468":{"lemma":"סֶדֶר","xlit":"çeder","pron":"seh'-der","derivation":"from an unused root meaning to arrange;","strongs_def":"order","kjv_def":"order."},
"H5469":{"lemma":"סַהַר","xlit":"çahar","pron":"cah'-har","derivation":"from an unused root meaning to be round;","strongs_def":"roundness","kjv_def":"round."},
"H5470":{"lemma":"סֹהַר","xlit":"çôhar","pron":"so'-har","derivation":"from the same as H5469 (סַהַר);","strongs_def":"a dungeon (as surrounded by walls)","kjv_def":"prison."},
"H5471":{"lemma":"סוֹא","xlit":"Çôwʼ","pron":"so","derivation":"of foreign derivation;","strongs_def":"So, an Egyptian king","kjv_def":"So."},
"H5472":{"lemma":"סוּג","xlit":"çûwg","pron":"soog","derivation":"a primitive root;","strongs_def":"properly, to flinch, i.e. (by implication) to go back, literally (to retreat) or figuratively (to apostatize)","kjv_def":"backslider, drive, go back, turn (away, back)."},
"H5473":{"lemma":"סוּג","xlit":"çûwg","pron":"soog","derivation":"a primitive root (probably rather identical with H5472 (סוּג) through the idea of shrinking from a hedge; compare H7735 (שׂוּג));","strongs_def":"to hem in, i.e. bind","kjv_def":"set about."},
"H5474":{"lemma":"סוּגַר","xlit":"çûwgar","pron":"soo-gar'","derivation":"from H5462 (סָגַר);","strongs_def":"an inclosure, i.e. cage (for an animal)","kjv_def":"ward."},
"H5475":{"lemma":"סוֹד","xlit":"çôwd","pron":"sode","derivation":"from H3245 (יָסַד);","strongs_def":"a session, i.e. company of persons (in close deliberation); by implication, intimacy, consultation, a secret","kjv_def":"assembly, consel, inward, secret (counsel)."},
"H5476":{"lemma":"סוֹדִי","xlit":"Çôwdîy","pron":"so-dee'","derivation":"from H5475 (סוֹד); a confidant;","strongs_def":"Sodi, an Israelite","kjv_def":"Sodi."},
"H5477":{"lemma":"סוּחַ","xlit":"Çûwach","pron":"soo'-akh","derivation":"from an unused root meaning to wipe away; sweeping;","strongs_def":"Suach, an Israelite","kjv_def":"Suah."},
"H5478":{"lemma":"סוּחָה","xlit":"çûwchâh","pron":"soo-khaw'","derivation":"from the same as H5477 (סוּחַ);","strongs_def":"something swept away, i.e. filth","kjv_def":"torn."},
"H5479":{"lemma":"סוֹטַי","xlit":"Çôwṭay","pron":"so-tah'-ee","derivation":"from H7750 (שׂוּט); roving;","strongs_def":"Sotai, one of the Nethinim","kjv_def":"Sotai."},
"H5480":{"lemma":"סוּךְ","xlit":"çûwk","pron":"sook","derivation":"a primitive root;","strongs_def":"properly, to smear over (with oil), i.e. anoint","kjv_def":"anoint (self), [idiom] at all."},
"H5481":{"lemma":"סוּמְפּוֹנְיָה","xlit":"çûwmᵉpôwnᵉyâh","pron":"soom-po-neh-yaw'","derivation":"(Aramaic) or סוּמְפֹּנְיָה; (Aramaic), or סִיפֹנְיָא; (Daniel 3:10) (Aramaic), of Greek origin;","strongs_def":"a bagpipe (with a double pipe)","kjv_def":"dulcimer."},
"H5482":{"lemma":"סְוֵנֵה","xlit":"Çᵉvênêh","pron":"sev-ay-nay'","derivation":"(rather to be written סְוֵנָה; for סְוֵן; i.e to Seven); of Egyptian derivation;","strongs_def":"Seven, a place in Upper Egypt","kjv_def":"Syene."},
"H5483":{"lemma":"סוּס","xlit":"çûwç","pron":"soos","derivation":"or סֻס; from an unused root meaning to skip (properly, for joy);","strongs_def":"a horse (as leaping); also a swallow (from its rapid flight)","kjv_def":"crane, horse (-back, -hoof). Compare H6571 (פָּרָשׁ)."},
"H5484":{"lemma":"סוּסָה","xlit":"çûwçâh","pron":"soo-saw'","derivation":"feminine of H5483 (סוּס);","strongs_def":"a mare","kjv_def":"company of horses."},
"H5485":{"lemma":"סוּסִי","xlit":"Çûwçîy","pron":"soo-see'","derivation":"from H5483 (סוּס); horse-like;","strongs_def":"Susi, an Israelite","kjv_def":"Susi."},
"H5486":{"lemma":"סוּף","xlit":"çûwph","pron":"soof","derivation":"a primitive root;","strongs_def":"to snatch away, i.e. terminate","kjv_def":"consume, have an end, perish, [idiom] be utterly."},
"H5487":{"lemma":"סוּף","xlit":"çûwph","pron":"soof","derivation":"(Aramaic) corresponding to H5486 (סוּף);","strongs_def":"to come to an end","kjv_def":"consume, fulfill."},
"H5488":{"lemma":"סוּף","xlit":"çûwph","pron":"soof","derivation":"probably of Egyptian origin;","strongs_def":"a reed, especially the papyrus","kjv_def":"flag, Red (sea), weed. Compare H5489 (סוּף)."},
"H5489":{"lemma":"סוּף","xlit":"Çûwph","pron":"soof","derivation":"for H5488 (סוּף) (by ellipsis of H3220 (יָם));","strongs_def":"the Reed (Sea)","kjv_def":"Red Sea."},
"H5490":{"lemma":"סוֹף","xlit":"çôwph","pron":"sofe","derivation":"from H5486 (סוּף);","strongs_def":"a termination","kjv_def":"conclusion, end, hinder participle"},
"H5491":{"lemma":"סוֹף","xlit":"çôwph","pron":"sofe","derivation":"(Aramaic) corresponding to H5490 (סוֹף)","strongs_def":"{a termination}","kjv_def":"end."},
"H5492":{"lemma":"סוּפָה","xlit":"çûwphâh","pron":"soo-faw'","derivation":"from H5486 (סוּף);","strongs_def":"a hurricane","kjv_def":"Red Sea, storm, tempest, whirlwind, Red sea."},
"H5493":{"lemma":"סוּר","xlit":"çûwr","pron":"soor","derivation":"or שׂוּר; (Hosea 9:12), a primitive root;","strongs_def":"to turn off (literal or figurative)","kjv_def":"be(-head), bring, call back, decline, depart, eschew, get (you), go (aside), [idiom] grievous, lay away (by), leave undone, be past, pluck away, put (away, down), rebel, remove (to and fro), revolt, [idiom] be sour, take (away, off), turn (aside, away, in), withdraw, be without."},
"H5494":{"lemma":"סוּר","xlit":"çûwr","pron":"soor","derivation":"probably passive participle of H5493 (סוּר);","strongs_def":"turned off, i.e. deteriorated","kjv_def":"degenerate."},
"H5495":{"lemma":"סוּר","xlit":"Çûwr","pron":"soor","derivation":"the same as H5494 (סוּר);","strongs_def":"Sur, a gate of the temple","kjv_def":"Sur."},
"H5496":{"lemma":"סוּת","xlit":"çûwth","pron":"sooth","derivation":"perhaps denominative from H7898 (שַׁיִת);","strongs_def":"properly, to prick, i.e. (figuratively) stimulate; by implication, to seduce","kjv_def":"entice, move, persuade, provoke, remove, set on, stir up, take away."},
"H5497":{"lemma":"סוּת","xlit":"çûwth","pron":"sooth","derivation":"probably from the same root as H4533 (מַסְוֶה);","strongs_def":"covering, i.e. clothing","kjv_def":"clothes."},
"H5498":{"lemma":"סָחַב","xlit":"çâchab","pron":"saw-khab'","derivation":"a primitive root;","strongs_def":"to trail along","kjv_def":"draw (out), tear."},
"H5499":{"lemma":"סְחָבָה","xlit":"çᵉchâbâh","pron":"seh-khaw-baw'","derivation":"from H5498 (סָחַב);","strongs_def":"a rag","kjv_def":"cast clout."},
"H5500":{"lemma":"סָחָה","xlit":"çâchâh","pron":"saw-khaw'","derivation":"a primitive root;","strongs_def":"to sweep away","kjv_def":"scrape."},
"H5501":{"lemma":"סְחִי","xlit":"çᵉchîy","pron":"seh-khee'","derivation":"from H5500 (סָחָה);","strongs_def":"refuse (as swept off)","kjv_def":"offscouring."},
"H5502":{"lemma":"סָחַף","xlit":"çâchaph","pron":"saw-khaf'","derivation":"a primitive root;","strongs_def":"to scrape off","kjv_def":"sweep (away)."},
"H5503":{"lemma":"סָחַר","xlit":"çâchar","pron":"saw-khar'","derivation":"a primitive root;","strongs_def":"to travel round (specifically as a pedlar); intensively, to palpitate","kjv_def":"go about, merchant(-man), occupy with, pant, trade, traffick."},
"H5504":{"lemma":"סַחַר","xlit":"çachar","pron":"sakh'-ar","derivation":"from H5503 (סָחַר);","strongs_def":"profit (from trade)","kjv_def":"merchandise."},
"H5505":{"lemma":"סָחַר","xlit":"çâchar","pron":"saw-khar'","derivation":"from H5503 (סָחַר);","strongs_def":"an emporium; abstractly, profit (from trade)","kjv_def":"mart, merchandise."},
"H5506":{"lemma":"סְחֹרָה","xlit":"çᵉchôrâh","pron":"sekh-o-raw'","derivation":"from H5503 (סָחַר);","strongs_def":"traffic","kjv_def":"merchandise."},
"H5507":{"lemma":"סֹחֵרָה","xlit":"çôchêrâh","pron":"so-khay-raw'","derivation":"properly, active participle feminine of H5503 (סָחַר);","strongs_def":"something surrounding the person, i.e. a shield","kjv_def":"buckler."},
"H5508":{"lemma":"סֹחֵרֶת","xlit":"çôchêreth","pron":"so-kheh'-reth","derivation":"similar to H5507 (סֹחֵרָה);","strongs_def":"probably a (black) tile (or tessara) forlaying borders with","kjv_def":"black marble."},
"H5509":{"lemma":"סִיג","xlit":"çîyg","pron":"seeg","derivation":"or סוּג; (Ezekiel 22:18), from H5472 (סוּג) in the sense of refuse;","strongs_def":"scoria","kjv_def":"dross."},
"H5510":{"lemma":"סִיוָן","xlit":"Çîyvân","pron":"see-vawn'","derivation":"probably of Persian origin;","strongs_def":"Sivan, the third Hebrew month","kjv_def":"Sivan."},
"H5511":{"lemma":"סִיחוֹן","xlit":"Çîychôwn","pron":"see-khone'","derivation":"or סִיחֹן; from the same as H5477 (סוּחַ); tempestuous;","strongs_def":"Sichon, an Amoritish king","kjv_def":"Sihon."},
"H5512":{"lemma":"סִין","xlit":"Çîyn","pron":"seen","derivation":"of uncertain derivation;","strongs_def":"Sin the name of an Egyptian town and (probably) desert adjoining","kjv_def":"Sin."},
"H5513":{"lemma":"סִינִי","xlit":"Çîynîy","pron":"see-nee'","derivation":"from an otherwise unknown name of a man;","strongs_def":"a Sinite, or descendant of one of the sons of Canaan","kjv_def":"Sinite."},
"H5514":{"lemma":"סִינַי","xlit":"Çîynay","pron":"see-nah'-ee","derivation":"of uncertain derivation;","strongs_def":"Sinai, mountain of Arabia","kjv_def":"Sinai."},
"H5515":{"lemma":"סִינִים","xlit":"Çîynîym","pron":"see-neem'","derivation":"plural of an otherwise unknown name;","strongs_def":"Sinim, a distant Oriental region","kjv_def":"Sinim."},
"H5516":{"lemma":"סִיסְרָא","xlit":"Çîyçᵉrâʼ","pron":"see-ser-aw'","derivation":"of uncertain derivation;","strongs_def":"Sisera, the name of a Canaanitish king and of one of the Nethinim","kjv_def":"Sisera."},
"H5517":{"lemma":"סִיעָא","xlit":"Çîyʻâʼ","pron":"see-ah'","derivation":"or סִיעֲהָא; from an unused root meaning to converse; congregation;","strongs_def":"Sia or Siaha, one of the Nethinim","kjv_def":"Sia, Siaha."},
"H5518":{"lemma":"סִיר","xlit":"çîyr","pron":"seer","derivation":"or (feminine) סִירָה; or סִרָה; (Jeremiah 52:18), from a primitive root meaning to boil up;","strongs_def":"a pot; also a thorn (as springing up rapidly); by implication, a hook","kjv_def":"caldron, fishhook, pan, (wash-)pot, thorn."},
"H5519":{"lemma":"סָךְ","xlit":"çâk","pron":"sawk","derivation":"from H5526 (סָכַךְ);","strongs_def":"properly, a thicket of men, i.e. a crowd","kjv_def":"multitude."},
"H5520":{"lemma":"סֹךְ","xlit":"çôk","pron":"soke","derivation":"from H5526 (סָכַךְ);","strongs_def":"a hut (as of entwined boughs); also a lair","kjv_def":"covert, den, pavilion, tabernacle."},
"H5521":{"lemma":"סֻכָּה","xlit":"çukkâh","pron":"sook-kaw'","derivation":"feminine of H5520 (סֹךְ);","strongs_def":"a hut or lair","kjv_def":"booth, cottage, covert, pavilion, tabernacle, tent."},
"H5522":{"lemma":"סִכּוּת","xlit":"çikkûwth","pron":"sik-kooth'","derivation":"feminine of H5519 (סָךְ);","strongs_def":"an (idolatrous) booth","kjv_def":"tabernacle."},
"H5523":{"lemma":"סֻכּוֹת","xlit":"Çukkôwth","pron":"sook-kohth'","derivation":"or סֻכֹּת; plural of H5521 (סֻכָּה); booths;","strongs_def":"Succoth, the name of a place in Egypt and of three in Palestine","kjv_def":"Succoth."},
"H5524":{"lemma":"סֻכּוֹת בְּנוֹת","xlit":"Çukkôwth bᵉnôwth","pron":"sook-kohth' ben-ohth'","derivation":"from H5523 (סֻכּוֹת) and the (irregular) plural of H1323 (בַּת);","strongs_def":"booths of (the) daughters; brothels, i.e. idoalatrous tents forimpure purpose","kjv_def":"Succothbenoth."},
"H5525":{"lemma":"סֻכִּי","xlit":"Çukkîy","pron":"sook-kee'","derivation":"patrial from an unknown name (perhaps H5520 (סֹךְ));","strongs_def":"a Sukkite, or inhabitant of some place near Egypt (i.e. hut-dwellers)","kjv_def":"Sukkiims."},
"H5526":{"lemma":"סָכַךְ","xlit":"çâkak","pron":"saw-kak'","derivation":"or שָׂכַךְ; (Exodus 33:22), a primitive root;","strongs_def":"properly, to entwine as ascreen; by implication, to fence in, cover over, (figuratively) protect","kjv_def":"cover, defence, defend, hedge in, join together, set, shut up."},
"H5527":{"lemma":"סְכָכָה","xlit":"Çᵉkâkâh","pron":"sek-aw-kaw'","derivation":"from H5526 (סָכַךְ); inclosure;","strongs_def":"Secacah, a place in Palestine","kjv_def":"Secacah."},
"H5528":{"lemma":"סָכַל","xlit":"çâkal","pron":"saw-kal'","derivation":"for H3688 (כָּסַל);","strongs_def":"to be silly","kjv_def":"do (make, play the, turn into) fool(-ish, -ishly, -ishness)."},
"H5529":{"lemma":"סֶכֶל","xlit":"çekel","pron":"seh'-kal","derivation":"from H5528 (סָכַל);","strongs_def":"silliness; concretely and collectively, dolts","kjv_def":"folly."},
"H5530":{"lemma":"סָכָל","xlit":"çâkâl","pron":"saw-kawl'","derivation":"from H5528 (סָכַל);","strongs_def":"silly","kjv_def":"fool(-ish), sottish."},
"H5531":{"lemma":"סִכְלוּת","xlit":"çiklûwth","pron":"sik-looth'","derivation":"or שִׂכְלוּת; (Ecclesiastes 1:17), from H5528 (סָכַל);","strongs_def":"silliness","kjv_def":"folly, foolishness."},
"H5532":{"lemma":"סָכַן","xlit":"çâkan","pron":"saw-kan'","derivation":"a primitive root;","strongs_def":"to be familiar with; by implication, to minister to, be serviceable to, be customary","kjv_def":"acquaint (self), be advantage, [idiom] ever, (be, (un-)) profit(-able), treasurer, be wont."},
"H5533":{"lemma":"סָכַן","xlit":"çâkan","pron":"saw-kan'","derivation":"probably a denominative from H7915 (שַׂכִּין);","strongs_def":"properly, to cut, i.e. damage; also to grow (causatively, make) poor","kjv_def":"endanger, impoverish."},
"H5534":{"lemma":"סָכַר","xlit":"çâkar","pron":"saw-kar'","derivation":"a primitive root;","strongs_def":"to shut up; by implication, to surrender","kjv_def":"stop, give over. See also H5462 (סָגַר), H7936 (שָׂכַר)."},
"H5535":{"lemma":"סָכַת","xlit":"çâkath","pron":"saw-kath'","derivation":"a primitive root to be silent;","strongs_def":"by implication, to observe quietly","kjv_def":"take heed."},
"H5536":{"lemma":"סַל","xlit":"çal","pron":"sal","derivation":"from H5549 (סָלַל);","strongs_def":"properly, a willow twig (as pendulous), i.e. an osier; but only as woven into abasket","kjv_def":"basket."},
"H5537":{"lemma":"סָלָא","xlit":"çâlâʼ","pron":"saw-law'","derivation":"a primitive root;","strongs_def":"to suspend in a balance, i.e. weigh","kjv_def":"compare."},
"H5538":{"lemma":"סִלָּא","xlit":"Çillâʼ","pron":"sil-law'","derivation":"from H5549 (סָלַל); an embankment;","strongs_def":"Silla, a place in Jerusalem","kjv_def":"Silla."},
"H5539":{"lemma":"סָלַד","xlit":"çâlad","pron":"saw-lad'","derivation":"a primitive root;","strongs_def":"probably to leap (with joy), i.e. exult","kjv_def":"harden self."},
"H5540":{"lemma":"סֶלֶד","xlit":"Çeled","pron":"seh'-led","derivation":"from H5539 (סָלַד); exultation;","strongs_def":"Seled, an Israelite","kjv_def":"Seled."},
"H5541":{"lemma":"סָלָה","xlit":"çâlâh","pron":"saw-law'","derivation":"a primitive root;","strongs_def":"to hang up, i.e. weigh, or (figuratively) contemn","kjv_def":"tread down (under foot), value."},
"H5542":{"lemma":"סֶלָה","xlit":"çelâh","pron":"seh'-law","derivation":"from H5541 (סָלָה);","strongs_def":"suspension (of music), i.e. pause","kjv_def":"Selah."},
"H5543":{"lemma":"סַלּוּ","xlit":"Çallûw","pron":"sal-loo'","derivation":"or סַלּוּא; or סָלוּא; or סַלַּי; from H5541 (סָלָה); weighed;","strongs_def":"Sallu or Sallai, the name of two Israelites","kjv_def":"Sallai, Sallu, Salu."},
"H5544":{"lemma":"סִלּוֹן","xlit":"çillôwn","pron":"sil-lone'","derivation":"or סַלּוֹן; from H5541 (סָלָה);","strongs_def":"a prickle (as if pendulous)","kjv_def":"brier, thorn."},
"H5545":{"lemma":"סָלַח","xlit":"çâlach","pron":"saw-lakh'","derivation":"a primitive root;","strongs_def":"to forgive","kjv_def":"forgive, pardon, spare."},
"H5546":{"lemma":"סַלָּח","xlit":"çallâch","pron":"saw-lawkh'","derivation":"from H5545 (סָלַח);","strongs_def":"placable","kjv_def":"ready to forgive."},
"H5547":{"lemma":"סְלִיחָה","xlit":"çᵉlîychâh","pron":"sel-ee-khaw'","derivation":"from H5545 (סָלַח);","strongs_def":"pardon","kjv_def":"forgiveness, pardon."},
"H5548":{"lemma":"סַלְכָה","xlit":"Çalkâh","pron":"sal-kaw'","derivation":"from an unused root. meaning to walk; walking;","strongs_def":"Salcah, a place East of the Jordan","kjv_def":"Salcah, Salchah."},
"H5549":{"lemma":"סָלַל","xlit":"çâlal","pron":"saw-lal'","derivation":"a primitive root;","strongs_def":"to mound up (especially a turnpike); figurative, to exalt; reflexively, to oppose (as by a dam)","kjv_def":"cast up, exalt (self), extol, make plain, raise up."},
"H5550":{"lemma":"סֹלְלָה","xlit":"çôlᵉlâh","pron":"so-lel-aw'","derivation":"or סוֹלְלָה; active participle feminine of H5549 (סָלַל), but used passively;","strongs_def":"a military mound, i.e. rampart of besiegers","kjv_def":"bank, mount."},
"H5551":{"lemma":"סֻלָּם","xlit":"çullâm","pron":"sool-lawm'","derivation":"from H5549 (סָלַל);","strongs_def":"a stair-case","kjv_def":"ladder."},
"H5552":{"lemma":"סַלְסִלָּה","xlit":"çalçillâh","pron":"sal-sil-law'","derivation":"from H5541 (סָלָה);","strongs_def":"a twig (as pendulous)","kjv_def":"basket."},
"H5553":{"lemma":"סֶלַע","xlit":"çelaʻ","pron":"seh'-lah","derivation":"from an unused root meaning to be lofty;","strongs_def":"a craggy rock, literally or figuratively (a fortress)","kjv_def":"(ragged) rock, stone(-ny), strong hold."},
"H5554":{"lemma":"סֶלַע","xlit":"Çelaʻ","pron":"seh'-lah","derivation":"the same as H5553 (סֶלַע);","strongs_def":"Sela, the rock-city of Idumaea","kjv_def":"rock, Sela(-h)."},
"H5555":{"lemma":"סֶלַע הַמַּחְלְקוֹת","xlit":"Çelaʻ ham-machlᵉqôwth","pron":"seh'-lah hammakh-lek-oth'","derivation":"from H5553 (סֶלַע) and the plural of H4256 (מַחֲלֹקֶת) with the article interposed; rock of the divisions;","strongs_def":"Sela-ham-Machlekoth, a place in Palestine","kjv_def":"Sela-hammalekoth."},
"H5556":{"lemma":"סׇלְעָם","xlit":"çolʻâm","pron":"sol-awm'","derivation":"apparently from the same as H5553 (סֶלַע) in the sense of crushing as with a rock, i.e. consuming;","strongs_def":"a kind of locust (from its destructiveness)","kjv_def":"bald locust."},
"H5557":{"lemma":"סָלַף","xlit":"çâlaph","pron":"saw-laf'","derivation":"a primitive root;","strongs_def":"properly, to wrench, i.e. (figuratively) to subvert","kjv_def":"overthrow, pervert."},
"H5558":{"lemma":"סֶלֶף","xlit":"çeleph","pron":"seh'-lef","derivation":"from H5557 (סָלַף);","strongs_def":"distortion, i.e. (figuratively) viciousness","kjv_def":"perverseness."},
"H5559":{"lemma":"סְלִק","xlit":"çᵉliq","pron":"sel-eek'","derivation":"(Aramaic) a primitive root;","strongs_def":"to ascend","kjv_def":"come (up)."},
"H5560":{"lemma":"סֹלֶת","xlit":"çôleth","pron":"so'-leth","derivation":"from an unused root meaning to strip;","strongs_def":"flour (as chipped off)","kjv_def":"(fine) flour, meal."},
"H5561":{"lemma":"סַם","xlit":"çam","pron":"sam","derivation":"from an unused root meaning to smell sweet;","strongs_def":"an aroma","kjv_def":"sweet (spice)."},
"H5562":{"lemma":"סַמְגַּר נְבוֹ","xlit":"Çamgar Nᵉbôw","pron":"sam-gar' neb-o'","derivation":"of foreign origin;","strongs_def":"Samgar-Nebo, a Babylonian general","kjv_def":"Samgar-nebo."},
"H5563":{"lemma":"סְמָדַר","xlit":"çᵉmâdar","pron":"sem-aw-dar'","derivation":"of uncertain derivation;","strongs_def":"a vine blossom; used also adverbially, abloom","kjv_def":"tender grape."},
"H5564":{"lemma":"סָמַךְ","xlit":"çâmak","pron":"saw-mak'","derivation":"a primitive root;","strongs_def":"to prop (literally or figuratively); reflexively, to lean upon or take hold of (in a favorable or unfavorable sense)","kjv_def":"bear up, establish, (up-) hold, lay, lean, lie hard, put, rest self, set self, stand fast, stay (self), sustain."},
"H5565":{"lemma":"סְמַכְיָהוּ","xlit":"Çᵉmakyâhûw","pron":"sem-ak-yaw'-hoo","derivation":"from H5564 (סָמַךְ) and H3050 (יָהּ); supported of Jah;","strongs_def":"Semakjah, an Israelite","kjv_def":"Semachiah."},
"H5566":{"lemma":"סֶמֶל","xlit":"çemel","pron":"seh'-mel","derivation":"or סֵמֶל; from an unused root meaning to resemble;","strongs_def":"a likeness","kjv_def":"figure, idol, image."},
"H5567":{"lemma":"סָמַן","xlit":"çâman","pron":"saw-man'","derivation":"a primitive root;","strongs_def":"to designate","kjv_def":"appointed."},
"H5568":{"lemma":"סָמַר","xlit":"çâmar","pron":"saw-mar'","derivation":"a primitive root;","strongs_def":"to be erect, i.e. bristle as hair","kjv_def":"stand up, tremble."},
"H5569":{"lemma":"סָמָר","xlit":"çâmâr","pron":"saw-mar'","derivation":"from H5568 (סָמַר);","strongs_def":"bristling, i.e. shaggy","kjv_def":"rough."},
"H5570":{"lemma":"סְנָאָה","xlit":"Çᵉnâʼâh","pron":"sen-aw-aw'","derivation":"from an unused root meaning to prick; thorny;","strongs_def":"Senaah, a place in Palestine","kjv_def":"Senaah, Hassenaah (with the article)."},
"H5571":{"lemma":"סַנְבַלַּט","xlit":"Çanballaṭ","pron":"san-bal-lat'","derivation":"of foreign origin;","strongs_def":"Sanballat, a Persian satrap of Samaria","kjv_def":"Sanballat."},
"H5572":{"lemma":"סְנֶה","xlit":"çᵉneh","pron":"sen-eh'","derivation":"from an unused root meaning to prick;","strongs_def":"a bramble","kjv_def":"bush."},
"H5573":{"lemma":"סֶנֶה","xlit":"Çeneh","pron":"seh-neh'","derivation":"the same as H5572 (סְנֶה); thorn;","strongs_def":"Seneh, a crag in Palestine","kjv_def":"Seneh."},
"H5574":{"lemma":"סְנוּאָה","xlit":"Çᵉnûwʼâh","pron":"sen-oo-aw'","derivation":"or סְנֻאָה from the same as H5570 (סְנָאָה);","strongs_def":"pointed; (used with the article as a proper name) Senuah, the name of two Israelites","kjv_def":"Hasenuah (including the art), Senuah."},
"H5575":{"lemma":"סַנְוֵר","xlit":"çanvêr","pron":"san-vare'","derivation":"of uncertain derivation;","strongs_def":"(in plural) blindness","kjv_def":"blindness."},
"H5576":{"lemma":"סַנְחֵרִיב","xlit":"Çanchêrîyb","pron":"san-khay-reeb'","derivation":"of foreign origin;","strongs_def":"Sancherib, an Assyrian king","kjv_def":"Sennacherib."},
"H5577":{"lemma":"סַנְסִן","xlit":"çançin","pron":"san-seen'","derivation":"from an unused root meaning to be pointed;","strongs_def":"a twig (as tapering)","kjv_def":"bough."},
"H5578":{"lemma":"סַנְסַנָּה","xlit":"Çançannâh","pron":"san-san-naw'","derivation":"feminine of a form of H5577 (סַנְסִן); a bough;","strongs_def":"Sansannah, a place in Palestine","kjv_def":"Sansannah."},
"H5579":{"lemma":"סְנַפִּיר","xlit":"çᵉnappîyr","pron":"sen-ap-peer'","derivation":"of uncertain derivation;","strongs_def":"a fin (collectively)","kjv_def":"fins."},
"H5580":{"lemma":"סָס","xlit":"çâç","pron":"sawce","derivation":"from the same as H5483 (סוּס);","strongs_def":"a moth (from the agility of the fly)","kjv_def":"moth."},
"H5581":{"lemma":"סִסְמַי","xlit":"Çiçmay","pron":"sis-mah'-ee","derivation":"of uncertain derivation;","strongs_def":"Sismai, an Israelite","kjv_def":"Sisamai."},
"H5582":{"lemma":"סָעַד","xlit":"çâʻad","pron":"saw-ad'","derivation":"a primitive root;","strongs_def":"to suport (mostly figurative)","kjv_def":"comfort, establish, hold up, refresh self, strengthen, be upholden."},
"H5583":{"lemma":"סְעַד","xlit":"çᵉʻad","pron":"seh-ad'","derivation":"(Aramaic) corresponding to H5582 (סָעַד);","strongs_def":"to aid","kjv_def":"helping."},
"H5584":{"lemma":"סָעָה","xlit":"çâʻâh","pron":"saw-aw'","derivation":"a primitive root;","strongs_def":"to rush","kjv_def":"storm."},
"H5585":{"lemma":"סָעִיף","xlit":"çâʻîyph","pron":"saw-eef'","derivation":"from H5586 (סָעַף);","strongs_def":"a fissure (of rocks); also a bough (as subdivided)","kjv_def":"(outmost) branch, clift, top."},
"H5586":{"lemma":"סָעַף","xlit":"çâʻaph","pron":"saw-af'","derivation":"a primitive root; properly, to divide up; but used only as denominative from H5585 (סָעִיף),","strongs_def":"to disbranch (a tree)","kjv_def":"top."},
"H5587":{"lemma":"סָעִף","xlit":"çâʻiph","pron":"saw-eef'","derivation":"or שָׂעִף; from H5586 (סָעַף);","strongs_def":"divided (in mind), i.e. (abstractly) a sentiment","kjv_def":"opinion."},
"H5588":{"lemma":"סֵעֵף","xlit":"çêʻêph","pron":"say-afe'","derivation":"from H5586 (סָעַף);","strongs_def":"divided (in mind), i.e. (concretely) a skeptic","kjv_def":"thought."},
"H5589":{"lemma":"סְעַפָּה","xlit":"çᵉʻappâh","pron":"seh-ap-paw'","derivation":"feminine of H5585 (סָעִיף);","strongs_def":"a twig","kjv_def":"bough, branch. Compare H5634 (סַרְעַפָּה)."},
"H5590":{"lemma":"סָעַר","xlit":"çâʻar","pron":"saw-ar'","derivation":"a primitive root;","strongs_def":"to rush upon; by implication, to toss (transitive or intransitive, literal or figurative)","kjv_def":"be (toss with) tempest(-uous), be sore, troubled, come out as a (drive with the, scatter with a) whirlwind."},
"H5591":{"lemma":"סַעַר","xlit":"çaʻar","pron":"sah'-ar","derivation":"or (feminine) סְעָרָה; from H5590 (סָעַר);","strongs_def":"a hurricane","kjv_def":"storm(-y), tempest, whirlwind."},
"H5592":{"lemma":"סַף","xlit":"çaph","pron":"saf","derivation":"from H5605 (סָפַף), in its original sense of containing;","strongs_def":"a vestibule (as a limit); also a dish (for holding blood or wine)","kjv_def":"bason, bowl, cup, door (post), gate, post, threshold."},
"H5593":{"lemma":"סַף","xlit":"Çaph","pron":"saf","derivation":"the same as H5592 (סַף);","strongs_def":"Saph, a Philistine","kjv_def":"Saph. Compare H5598 (סִפַּי)."},
"H5594":{"lemma":"סָפַד","xlit":"çâphad","pron":"saw-fad'","derivation":"a primitive root;","strongs_def":"properly, to tear the hair and beat the breasts (as Orientals do in grief); generally to lament; by implication, to wail","kjv_def":"lament, mourn(-er), wail."},
"H5595":{"lemma":"סָפָה","xlit":"çâphâh","pron":"saw-faw'","derivation":"a primitive root;","strongs_def":"properly, to scrape (literally, to shave; but usually figuratively) together (i.e. to accumulate or increase) or away (i.e. to scatter, remove, or ruin; intransitively, to perish)","kjv_def":"add, augment, consume, destroy, heap, join, perish, put."},
"H5596":{"lemma":"סָפַח","xlit":"çâphach","pron":"saw-fakh'","derivation":"or שָׂפַח; (Isaiah 3:17), a primitive root;","strongs_def":"properly, to scrape out, but in certain peculiar senses (of removal or association)","kjv_def":"abiding, gather together, cleave, smite with the scab."},
"H5597":{"lemma":"סַפַּחַת","xlit":"çappachath","pron":"sap-pakh'-ath","derivation":"from H5596 (סָפַח);","strongs_def":"the mange (as making the hair fall off)","kjv_def":"scab."},
"H5598":{"lemma":"סִפַּי","xlit":"Çippay","pron":"sip-pah'-ee","derivation":"from H5592 (סַף); bason-like;","strongs_def":"Sippai, a Philistine","kjv_def":"Sippai. Compare H5593 (סַף)."},
"H5599":{"lemma":"סָפִיחַ","xlit":"çâphîyach","pron":"saw-fee'-akh","derivation":"from H5596 (סָפַח);","strongs_def":"something (spontaneously) falling off, i.e. a self-sown crop; figuratively, a freshet","kjv_def":"(such) things as (which) grow (of themselves), which groweth of its own accord (itself)."},
"H5600":{"lemma":"סְפִינָה","xlit":"çᵉphîynâh","pron":"sef-ee-naw'","derivation":"from H5603 (סָפַן);","strongs_def":"a (sea-going) vessel (as ceiled with a deck)","kjv_def":"ship."},
"H5601":{"lemma":"סַפִּיר","xlit":"çappîyr","pron":"sap-peer'","derivation":"from H5608 (סָפַר);","strongs_def":"a gem (perhaps used for scratching other substances), probably the sapphire","kjv_def":"sapphire."},
"H5602":{"lemma":"סֵפֶל","xlit":"çêphel","pron":"say'-fel","derivation":"from an unused root meaning to depress;","strongs_def":"a basin (as deepened out)","kjv_def":"bowl, dish."},
"H5603":{"lemma":"סָפַן","xlit":"çâphan","pron":"saw-fan'","derivation":"a primitive root;","strongs_def":"to hide by covering; specifically, to roof (passive participle as noun, a roof) or wainscot; figuratively, to reserve","kjv_def":"cieled, cover, seated."},
"H5604":{"lemma":"סִפֻּן","xlit":"çippun","pron":"sip-poon'","derivation":"from H5603 (סָפַן);","strongs_def":"a wainscot","kjv_def":"cieling."},
"H5605":{"lemma":"סָפַף","xlit":"çâphaph","pron":"saw-faf'","derivation":"a primitive root; properly, to snatch away, i.e. terminate; but used only as denominative from H5592 (סַף) (in the sense of a vestibule),","strongs_def":"to wait at the threshold","kjv_def":"be a doorkeeper."},
"H5606":{"lemma":"סָפַק","xlit":"çâphaq","pron":"saw-fak'","derivation":"or שָׂפַק; (1 Kings 20:10; Job 27:23; Isaiah 2:6), a primitive root;","strongs_def":"to clap the hands (in token of compact, derision, grief, indignation, or punishment); by implication of satisfaction, to be enough; by implication of excess, to vomit","kjv_def":"clap, smite, strike, suffice, wallow."},
"H5607":{"lemma":"סֵפֶק","xlit":"çêpheq","pron":"say'-fek","derivation":"or שֶׂפֶק; (Job 20:22; Job 36:18), from H5606 (סָפַק);","strongs_def":"chastisement; also satiety","kjv_def":"stroke, sufficiency."},
"H5608":{"lemma":"סָפַר","xlit":"çâphar","pron":"saw-far'","derivation":"a primitive root;","strongs_def":"properly, to score with a mark as a tally or record, i.e. (by implication) to inscribe, and also to enumerate; intensively, to recount, i.e. celebrate","kjv_def":"commune, (ac-) count; declare, number, [phrase] penknife, reckon, scribe, shew forth, speak, talk, tell (out), writer."},
"H5609":{"lemma":"סְפַר","xlit":"çᵉphar","pron":"sef-ar'","derivation":"(Aramaic) from a root corresponding to H5608 (סָפַר);","strongs_def":"a book","kjv_def":"book, roll."},
"H5610":{"lemma":"סְפָר","xlit":"çᵉphâr","pron":"sef-awr'","derivation":"from H5608 (סָפַר);","strongs_def":"a census","kjv_def":"numbering."},
"H5611":{"lemma":"סְפָר","xlit":"Çᵉphâr","pron":"sef-awr'","derivation":"the same as H5610 (סְפָר);","strongs_def":"Sephar, a place in Arabia","kjv_def":"Sephar."},
"H5612":{"lemma":"סֵפֶר","xlit":"çêpher","pron":"say'-fer","derivation":"or (feminine) סִפְרָה; (Psalm 56:8 (H9 (אֲבֵדָה))), from H5608 (סָפַר);","strongs_def":"properly, writing (the art or a document); by implication, a book","kjv_def":"bill, book, evidence, [idiom] learn(-ed) (-ing), letter, register, scroll."},
"H5613":{"lemma":"סָפֵר","xlit":"çâphêr","pron":"saw-fare'","derivation":"(Aramaic) from the same as H5609 (סְפַר);","strongs_def":"a scribe (secular or sacred)","kjv_def":"scribe."},
"H5614":{"lemma":"סְפָרָד","xlit":"Çᵉphârâd","pron":"sef-aw-rawd'","derivation":"of foreign derivation;","strongs_def":"Sepharad, a region of Assyria","kjv_def":"Sepharad."},
"H5615":{"lemma":"סְפֹרָה","xlit":"çᵉphôrâh","pron":"sef-o-raw'","derivation":"from H5608 (סָפַר);","strongs_def":"a numeration","kjv_def":"number."},
"H5616":{"lemma":"סְפַרְוִי","xlit":"Çᵉpharvîy","pron":"sef-ar-vee'","derivation":"patrial from H5617 (סְפַרְוַיִם);","strongs_def":"a Sepharvite or inhabitant of Sepharvain","kjv_def":"Sepharvite."},
"H5617":{"lemma":"סְפַרְוַיִם","xlit":"Çᵉpharvayim","pron":"sef-ar-vah'-yim","derivation":"(dual)","strongs_def":"Sepharvites"},
"H5618":{"lemma":"סֹפֶרֶת","xlit":"Çôphereth","pron":"so-feh'-reth","derivation":"feminine active participle of H5608 (סָפַר); a scribe (properly, female);","strongs_def":"Sophereth, a temple servant","kjv_def":"Sophereth."},
"H5619":{"lemma":"סָקַל","xlit":"çâqal","pron":"saw-kal'","derivation":"a primitive root;","strongs_def":"properly, to be weighty; but used only in the sense of lapidation or its contrary (as if a delapidation)","kjv_def":"(cast, gather out, throw) stone(-s), [idiom] surely."},
"H5620":{"lemma":"סַר","xlit":"çar","pron":"sar","derivation":"contracted from H5637 (סָרַר);","strongs_def":"peevish","kjv_def":"heavy, sad."},
"H5621":{"lemma":"סָרָב","xlit":"çârâb","pron":"saw-rawb'","derivation":"from an unused root meaning to sting;","strongs_def":"a thistle","kjv_def":"brier."},
"H5622":{"lemma":"סַרְבַּל","xlit":"çarbal","pron":"sar-bal'","derivation":"(Aramaic) of uncertain derivation;","strongs_def":"a cloak","kjv_def":"coat."},
"H5623":{"lemma":"סַרְגּוֹן","xlit":"Çargôwn","pron":"sar-gone'","derivation":"of foreign derivation;","strongs_def":"Sargon, an Assyrian king","kjv_def":"Sargon."},
"H5624":{"lemma":"סֶרֶד","xlit":"Çered","pron":"seh'-red","derivation":"from a primitive root meaning to tremble; trembling;","strongs_def":"Sered, an Israelite","kjv_def":"Sered."},
"H5625":{"lemma":"סַרְדִּי","xlit":"Çardîy","pron":"sar-dee'","derivation":"patron from H5624 (סֶרֶד);","strongs_def":"a Seredite (collectively) or descendants of Sered","kjv_def":"Sardites."},
"H5626":{"lemma":"סִרָה","xlit":"Çirâh","pron":"see-raw'","derivation":"from H5493 (סוּר); departure;","strongs_def":"Sirah, a cistern so-called","kjv_def":"Sirah. See also H5518 (סִיר)."},
"H5627":{"lemma":"סָרָה","xlit":"çârâh","pron":"saw-raw'","derivation":"from H5493 (סוּר);","strongs_def":"apostasy, crime; figuratively, remission","kjv_def":"[idiom] continual, rebellion, revolt(-ed), turn away, wrong."},
"H5628":{"lemma":"סָרַח","xlit":"çârach","pron":"saw-rakh'","derivation":"a primitive root;","strongs_def":"to extend (even to excess)","kjv_def":"exceeding, hand, spread, stretch self, banish."},
"H5629":{"lemma":"סֶרַח","xlit":"çerach","pron":"seh'-rakh","derivation":"from H5628 (סָרַח);","strongs_def":"a redundancy","kjv_def":"remnant."},
"H5630":{"lemma":"סִרְיֹן","xlit":"çiryôn","pron":"sir-yone'","derivation":"for H8302 (שִׁרְיוֹן);","strongs_def":"a coat of mail","kjv_def":"brigandine."},
"H5631":{"lemma":"סָרִיס","xlit":"çârîyç","pron":"saw-reece'","derivation":"or סָרִס; from an unused root meaning to castrate;","strongs_def":"a eunuch; by implication, valet (especially of the female apartments), and thus, a minister of state","kjv_def":"chamberlain, eunuch, officer. Compare H7249 (רַב־סָרִיס)."},
"H5632":{"lemma":"סָרֵךְ","xlit":"çârêk","pron":"saw-rake'","derivation":"(Aramaic) of foreign origin;","strongs_def":"an emir","kjv_def":"president."},
"H5633":{"lemma":"סֶרֶן","xlit":"çeren","pron":"seh'-ren","derivation":"from an unused root of uncertain meaning;","strongs_def":"an axle; figuratively, a peer","kjv_def":"lord, plate."},
"H5634":{"lemma":"סַרְעַפָּה","xlit":"çarʻappâh","pron":"sar-ap-paw'","derivation":"for H5589 (סְעַפָּה);","strongs_def":"a twig","kjv_def":"bough."},
"H5635":{"lemma":"סָרַף","xlit":"çâraph","pron":"saw-raf'","derivation":"a primitive root;","strongs_def":"to cremate, i.e. to be (near) of kin (such being privileged to kindle the pyre)","kjv_def":"burn."},
"H5636":{"lemma":"סַרְפָּד","xlit":"çarpâd","pron":"sar-pawd'","derivation":"from H5635 (סָרַף);","strongs_def":"a nettle (as stinging like a burn)","kjv_def":"brier."},
"H5637":{"lemma":"סָרַר","xlit":"çârar","pron":"saw-rar'","derivation":"a primitive root;","strongs_def":"to turn away, i.e. (morally) be refractory","kjv_def":"[idiom] away, backsliding, rebellious, revolter(-ing), slide back, stubborn, withdrew."},
"H5638":{"lemma":"סְתָו","xlit":"çᵉthâv","pron":"seth-awv'","derivation":"from an unused root meaning to hide;","strongs_def":"winter (as the dark season)","kjv_def":"winter."},
"H5639":{"lemma":"סְתוּר","xlit":"Çᵉthûwr","pron":"seth-oor'","derivation":"from H5641 (סָתַר); hidden;","strongs_def":"Sethur, an Israelite","kjv_def":"Sethur."},
"H5640":{"lemma":"סָתַם","xlit":"çâtham","pron":"saw-tham'","derivation":"or שָׂתַם; (Numbers 24:15), a primitive root;","strongs_def":"to stop up; by implication, to repair; figuratively, to keep secret","kjv_def":"closed up, hidden, secret, shut out (up), stop."},
"H5641":{"lemma":"סָתַר","xlit":"çâthar","pron":"saw-thar'","derivation":"a primitive root;","strongs_def":"to hide (by covering), literally or figuratively","kjv_def":"be absent, keep close, conceal, hide (self), (keep) secret, [idiom] surely."},
"H5642":{"lemma":"סְתַר","xlit":"çᵉthar","pron":"seth-ar'","derivation":"(Aramaic) corresponding to H5641 (סָתַר);","strongs_def":"to conceal; figuratively, to demolish","kjv_def":"destroy, secret thing."},
"H5643":{"lemma":"סֵתֶר","xlit":"çêther","pron":"say'-ther","derivation":"or (feminine) סִתְרָה; (Deuteronomy [32:38), from H5641 (סָתַר);","strongs_def":"a cover (in a good or a bad, a literal or a figurative sense)","kjv_def":"backbiting, covering, covert, [idiom] disguise(-th), hiding place, privily, protection, secret(-ly, place)."},
"H5644":{"lemma":"סִתְרִי","xlit":"Çithrîy","pron":"sith-ree'","derivation":"from H5643 (סֵתֶר); protective;","strongs_def":"Sithri, an Israelite,","kjv_def":"Zithri. ["},
"H5645":{"lemma":"עָב","xlit":"ʻâb","pron":"awb","derivation":"(masculine and feminine); from H5743 (עוּב);","strongs_def":"properly, an envelope, i.e. darkness (or density, 2 Chronicles 4:17); specifically, a (scud) cloud; also a copse","kjv_def":"clay, (thick) cloud, [idiom] thick, thicket. Compare H5672 (עֲבִי)."},
"H5646":{"lemma":"עָב","xlit":"ʻâb","pron":"awb","derivation":"or עֹב; from an unused root meaning to cover; properly, equivalent to H5645 (עָב); but used only as an architectural term,","strongs_def":"an architrave (as shading the pillars)","kjv_def":"thick (beam, plant)."},
"H5647":{"lemma":"עָבַד","xlit":"ʻâbad","pron":"aw-bad'","derivation":"a primitive root;","strongs_def":"to work (in any sense); by implication, to serve, till, (causatively) enslave, etc.","kjv_def":"[idiom] be, keep in bondage, be bondmen, bond-service, compel, do, dress, ear, execute, [phrase] husbandman, keep, labour(-ing man, bring to pass, (cause to, make to) serve(-ing, self), (be, become) servant(-s), do (use) service, till(-er), transgress (from margin), (set a) work, be wrought, worshipper,"},
"H5648":{"lemma":"עֲבַד","xlit":"ʻăbad","pron":"ab-bad'","derivation":"(Aramaic) corresponding to H5647 (עָבַד);","strongs_def":"to do, make, prepare, keep, etc.","kjv_def":"[idiom] cut, do, execute, go on, make, move, work."},
"H5649":{"lemma":"עֲבַד","xlit":"ʻăbad","pron":"ab-bad'","derivation":"(Aramaic) from H5648 (עֲבַד);","strongs_def":"a servant","kjv_def":"servant."},
"H5650":{"lemma":"עֶבֶד","xlit":"ʻebed","pron":"eh'-bed","derivation":"from H5647 (עָבַד);","strongs_def":"a servant","kjv_def":"[idiom] bondage, bondman, (bond-) servant, (man-) servant."},
"H5651":{"lemma":"עֶבֶד","xlit":"ʻEbed","pron":"eh'-bed","derivation":"the same as H5650 (עֶבֶד);","strongs_def":"Ebed, the name of two Israelites","kjv_def":"Ebed."},
"H5652":{"lemma":"עֲבָד","xlit":"ʻăbâd","pron":"ab-awd'","derivation":"from H5647 (עָבַד);","strongs_def":"a deed","kjv_def":"work."},
"H5653":{"lemma":"עַבְדָּא","xlit":"ʻAbdâʼ","pron":"ab-daw'","derivation":"from H5647 (עָבַד); work;","strongs_def":"Abda, the name of two Israelites","kjv_def":"Abda"},
"H5654":{"lemma":"עֹבֵד אֱדוֹם","xlit":"ʻÔbêd ʼĔdôwm","pron":"o-bade' ed-ome'","derivation":"from the active participle of H5647 (עָבַד) and H123 (אֱדֹם); worker of Edom;","strongs_def":"Obed-Edom, the name of five Israelites","kjv_def":"Obed-edom."},
"H5655":{"lemma":"עַבְדְּאֵל","xlit":"ʻAbdᵉʼêl","pron":"ab-deh-ale'","derivation":"from H5647 (עָבַד) and H410 (אֵל); serving God;","strongs_def":"Abdeel, an Israelite","kjv_def":"Abdeel. Compare H5661 (עַבְדִיאֵל)."},
"H5656":{"lemma":"עֲבֹדָה","xlit":"ʻăbôdâh","pron":"ab-o-daw'","derivation":"or עֲבוֹדָה; from H5647 (עָבַד);","strongs_def":"work of any kind","kjv_def":"act, bondage, [phrase] bondservant, effect, labour, ministering(-try), office, service(-ile, -itude), tillage, use, work, [idiom] wrought."},
"H5657":{"lemma":"עֲבֻדָּה","xlit":"ʻăbuddâh","pron":"ab-ood-daw'","derivation":"passive participle of H5647 (עָבַד);","strongs_def":"something wrought, i.e. (concretely) service","kjv_def":"household, store of servants."},
"H5658":{"lemma":"עַבְדוֹן","xlit":"ʻAbdôwn","pron":"ab-dohn'","derivation":"from H5647 (עָבַד); servitude;","strongs_def":"Abdon, the name of a place in Palestine and of four Israelites","kjv_def":"Abdon. Compare H5683 (עֶבְרֹן)."},
"H5659":{"lemma":"עַבְדוּת","xlit":"ʻabdûwth","pron":"ab-dooth'","derivation":"from H5647 (עָבַד);","strongs_def":"servitude","kjv_def":"bondage."},
"H5660":{"lemma":"עַבְדִּי","xlit":"ʻAbdîy","pron":"ab-dee'","derivation":"from H5647 (עָבַד); serviceable;","strongs_def":"Abdi, the name of two Israelites","kjv_def":"Abdi."},
"H5661":{"lemma":"עַבְדִיאֵל","xlit":"ʻAbdîyʼêl","pron":"ab-dee-ale'","derivation":"from H5650 (עֶבֶד) and H410 (אֵל); servant of God;","strongs_def":"Abdiel, an Israelite","kjv_def":"Abdiel. Compare H5655 (עַבְדְּאֵל)."},
"H5662":{"lemma":"עֹבַדְיָה","xlit":"ʻÔbadyâh","pron":"o-bad-yaw'","derivation":"or עֹבַדְיָהוּ; active participle of H5647 (עָבַד) and H3050 (יָהּ); serving Jah;","strongs_def":"Obadjah, the name of thirteen Israelites","kjv_def":"Obadiah."},
"H5663":{"lemma":"עֶבֶד מֶלֶךְ","xlit":"ʻEbed Melek","pron":"eh'-bed meh'-lek","derivation":"from H5650 (עֶבֶד) and H4428 (מֶלֶךְ); servant of a king;","strongs_def":"Ebed-Melek, a eunuch of Zedekeah","kjv_def":"Ebed-melech."},
"H5664":{"lemma":"עֲבֵד נְגוֹ","xlit":"ʻĂbêd Nᵉgôw","pron":"ab-ade' neg-o'","derivation":"the same as H5665 (עֲבֵד נְגוֹא);","strongs_def":"Abed-Nego, the Babylonian name of one of Daniel's companions","kjv_def":"Abed-nego."},
"H5665":{"lemma":"עֲבֵד נְגוֹא","xlit":"ʻĂbêd Nᵉgôwʼ","pron":"ab-ade' neg-o'","derivation":"(Aramaic) of foreign origin;","strongs_def":"Abed-Nego, the name of Azariah","kjv_def":"Abed-nego."},
"H5666":{"lemma":"עָבָה","xlit":"ʻâbâh","pron":"aw-baw'","derivation":"a primitive root;","strongs_def":"to be dense","kjv_def":"be (grow) thick(-er)."},
"H5667":{"lemma":"עֲבוֹט","xlit":"ʻăbôwṭ","pron":"ab-ote'","derivation":"or עֲבֹט; from H5670 (עָבַט);","strongs_def":"a pawn","kjv_def":"pledge."},
"H5668":{"lemma":"עָבוּר","xlit":"ʻâbûwr","pron":"aw-boor'","derivation":"or עָבֻר; passive participle of H5674 (עָבַר);","strongs_def":"properly, crossed, i.e. (abstractly) transit; used only adverbially, on account of, in order that","kjv_def":"because of, for (...'s sake), (intent) that, to."},
"H5669":{"lemma":"עָבוּר","xlit":"ʻâbûwr","pron":"aw-boor'","derivation":"the same as H5668 (עָבוּר);","strongs_def":"passed, i.e. kept over; used only of stored grain","kjv_def":"old corn."},
"H5670":{"lemma":"עָבַט","xlit":"ʻâbaṭ","pron":"aw-bat'","derivation":"a primitive root;","strongs_def":"to pawn; causatively, to lend (on security); figuratively, to entangle","kjv_def":"borrow, break (ranks), fetch (a pledge), lend, [idiom] surely."},
"H5671":{"lemma":"עַבְטִיט","xlit":"ʻabṭîyṭ","pron":"ab-teet'","derivation":"from H5670 (עָבַט);","strongs_def":"something pledged, i.e. (collectively) pawned goods","kjv_def":"thick clay (by a false etym.)."},
"H5672":{"lemma":"עֲבִי","xlit":"ʻăbîy","pron":"ab-ee'","derivation":"or עֳבִי]; from H5666 (עָבָה);","strongs_def":"density, i.e. depth or width","kjv_def":"thick(-ness). Compare H5645 (עָב)."},
"H5673":{"lemma":"עֲבִידָה","xlit":"ʻăbîydâh","pron":"ab-ee-daw'","derivation":"(Aramaic) from H5648 (עֲבַד);","strongs_def":"labor or business","kjv_def":"affairs, service, work."},
"H5674":{"lemma":"עָבַר","xlit":"ʻâbar","pron":"aw-bar'","derivation":"a primitive root;","strongs_def":"to cross over; used very widely of any transition (literal or figurative; transitive, intransitive, intensive, causative); specifically, to cover (in copulation)","kjv_def":"alienate, alter, [idiom] at all, beyond, bring (over, through), carry over, (over-) come (on, over), conduct (over), convey over, current, deliver, do away, enter, escape, fail, gender, get over, (make) go (away, beyond, by, forth, his way, in, on, over, through), have away (more), lay, meddle, overrun, make partition, (cause to, give, make to, over) pass(-age, along, away, beyond, by, -enger, on, out, over, through), (cause to, make) [phrase] proclaim(-amation), perish, provoke to anger, put away, rage, [phrase] raiser of taxes, remove, send over, set apart, [phrase] shave, cause to (make) sound, [idiom] speedily, [idiom] sweet smelling, take (away), (make to) transgress(-or), translate, turn away, (way-) faring man, be wrath."},
"H5675":{"lemma":"עֲבַר","xlit":"ʻăbar","pron":"ab-ar'","derivation":"(Aramaic) corresponding to H5676 (עֵבֶר)","strongs_def":"{properly, a region across; but used only adverbially (with or without a preposition) on the opposite side (especially of the Jordan; ususally meaning the east)}","kjv_def":"beyond, this side."},
"H5676":{"lemma":"עֵבֶר","xlit":"ʻêber","pron":"ay'-ber","derivation":"from H5674 (עָבַר);","strongs_def":"properly, a region across; but used only adverbially (with or without a preposition) on the opposite side (especially of the Jordan; ususally meaning the east)","kjv_def":"[idiom] against, beyond, by, [idiom] from, over, passage, quarter, (other, this) side, straight."},
"H5677":{"lemma":"עֵבֵר","xlit":"ʻÊbêr","pron":"ay'-ber","derivation":"the same as H5676 (עֵבֶר);","strongs_def":"Eber, the name of two patriarchs and four Israelites","kjv_def":"Eber, Heber."},
"H5678":{"lemma":"עֶבְרָה","xlit":"ʻebrâh","pron":"eb-raw'","derivation":"feminine of H5676 (עֵבֶר);","strongs_def":"an outburst of passion","kjv_def":"anger, rage, wrath."},
"H5679":{"lemma":"עֲבָרָה","xlit":"ʻăbârâh","pron":"ab-aw-raw'","derivation":"from H5674 (עָבַר);","strongs_def":"a crossing-place","kjv_def":"ferry, plain (from the margin)."},
"H5680":{"lemma":"עִבְרִי","xlit":"ʻIbrîy","pron":"ib-ree'","derivation":"patronymic from H5677 (עֵבֵר);","strongs_def":"an Eberite (i.e. Hebrew) or descendant of Eber","kjv_def":"Hebrew(-ess, woman)."},
"H5681":{"lemma":"עִבְרִי","xlit":"ʻIbrîy","pron":"ib-ree'","derivation":"the same as H5680 (עִבְרִי);","strongs_def":"Ibri, an Israelite","kjv_def":"Ibri."},
"H5682":{"lemma":"עֲבָרִים","xlit":"ʻĂbârîym","pron":"ab-aw-reem'","derivation":"plural of H5676 (עֵבֶר); regions beyond;","strongs_def":"Abarim, a place in Palestine","kjv_def":"Abarim, passages."},
"H5683":{"lemma":"עֶבְרֹן","xlit":"ʻEbrôn","pron":"eb-rone'","derivation":"from H5676 (עֵבֶר); transitional;","strongs_def":"Ebron, a place in Palestine","kjv_def":"Hebron. Perhaps a clerical error for H5658 (עַבְדוֹן)."},
"H5684":{"lemma":"עֶבְרֹנָה","xlit":"ʻEbrônâh","pron":"eb-raw-naw'","derivation":"feminine of H5683 (עֶבְרֹן);","strongs_def":"Ebronah, place in the Desert","kjv_def":"Ebronah."},
"H5685":{"lemma":"עָבַשׁ","xlit":"ʻâbash","pron":"aw-bash'","derivation":"a primitive root;","strongs_def":"to dry up","kjv_def":"be rotten."},
"H5686":{"lemma":"עָבַת","xlit":"ʻâbath","pron":"aw-bath'","derivation":"a primitive root;","strongs_def":"to interlace, i.e. (figuratively) to pervert","kjv_def":"wrap up."},
"H5687":{"lemma":"עָבֹת","xlit":"ʻâbôth","pron":"aw-both'","derivation":"or עָבוֹת; from H5686 (עָבַת);","strongs_def":"intwined, i.e. dense","kjv_def":"thick."},
"H5688":{"lemma":"עֲבֹת","xlit":"ʻăbôth","pron":"ab-oth'","derivation":"or עֲבוֹת; or (feminine) עֲבֹתָה; the same as H5687 (עָבֹת);","strongs_def":"something intwined, i.e. a string, wreath or foliage","kjv_def":"band, cord, rope, thick bough (branch), wreathen (chain)."},
"H5689":{"lemma":"עֲגַב","xlit":"ʻăgab","pron":"aw-gab'","derivation":"a primitive root;","strongs_def":"to breathe after, i.e. to love (sensually)","kjv_def":"dote, lover."},
"H5690":{"lemma":"עֶגֶב","xlit":"ʻegeb","pron":"eh'-gheb","derivation":"from H5689 (עֲגַב);","strongs_def":"love (concretely), i.e. amative words","kjv_def":"much love, very lovely."},
"H5691":{"lemma":"עֲגָבָה","xlit":"ʻăgâbâh","pron":"ag-aw-baw'","derivation":"from H5689 (עֲגַב);","strongs_def":"love (abstractly), i.e. amorousness","kjv_def":"inordinate love."},
"H5692":{"lemma":"עֻגָּה","xlit":"ʻuggâh","pron":"oog-gaw'","derivation":"from H5746 (עוּג);","strongs_def":"an ash-cake (as round)","kjv_def":"cake (upon the hearth)."},
"H5693":{"lemma":"עָגוּר","xlit":"ʻâgûwr","pron":"aw-goor'","derivation":"passive part (but with active sense) of an unused root meaning to twitter;","strongs_def":"probably the swallow","kjv_def":"swallow."},
"H5694":{"lemma":"עָגִיל","xlit":"ʻâgîyl","pron":"aw-gheel'","derivation":"from the same as H5696 (עָגֹל);","strongs_def":"something round, i.e. a ring (for the ears)","kjv_def":"earring."},
"H5695":{"lemma":"עֵגֶל","xlit":"ʻêgel","pron":"ay-ghel","derivation":"from the same as H5696 (עָגֹל);","strongs_def":"a (male) calf (as frisking round), especially one nearly grown (i.e. a steer)","kjv_def":"bullock, calf."},
"H5696":{"lemma":"עָגֹל","xlit":"ʻâgôl","pron":"aw-gole'","derivation":"or עָגוֹל; from an unused root meaning to revolve,","strongs_def":"circular","kjv_def":"round."},
"H5697":{"lemma":"עֶגְלָה","xlit":"ʻeglâh","pron":"eg-law'","derivation":"feminine of H5695 (עֵגֶל);","strongs_def":"a (female) calf, especially one nearly grown (i.e. a heifer)","kjv_def":"calf, cow, heifer."},
"H5698":{"lemma":"עֶגְלָה","xlit":"ʻEglâh","pron":"eg-law'","derivation":"the same as H5697 (עֶגְלָה);","strongs_def":"Eglah, a wife of David","kjv_def":"Eglah."},
"H5699":{"lemma":"עֲגָלָה","xlit":"ʻăgâlâh","pron":"ag-aw-law'","derivation":"from the same as H5696 (עָגֹל);","strongs_def":"something revolving, i.e. a wheeled vehicle","kjv_def":"cart, chariot, wagon"},
"H5700":{"lemma":"עֶגְלוֹן","xlit":"ʻEglôwn","pron":"eg-lawn'","derivation":"from H5695 (עֵגֶל); vituline;","strongs_def":"Eglon, the name of a place in Palestine and of a Moabitish king","kjv_def":"Eglon."},
"H5701":{"lemma":"עָגַם","xlit":"ʻâgam","pron":"aw-gam'","derivation":"a primitive root;","strongs_def":"to be sad","kjv_def":"grieve."},
"H5702":{"lemma":"עָגַן","xlit":"ʻâgan","pron":"aw-gan'","derivation":"a primitive root;","strongs_def":"to debar, i.e. from marriage","kjv_def":"stay."},
"H5703":{"lemma":"עַד","xlit":"ʻad","pron":"ad","derivation":"from H5710 (עָדָה);","strongs_def":"properly, a (peremptory) terminus, i.e. (by implication) duration, in the sense of advance or perpetuity (substantially as a noun, either with or without a preposition)","kjv_def":"eternity, ever(-lasting, -more), old, perpetually, [phrase] world without end."},
"H5704":{"lemma":"עַד","xlit":"ʻad","pron":"ad","derivation":"properly, the same as H5703 (עַד) (used as a preposition, adverb or conjunction; especially with a preposition);","strongs_def":"as far (or long, or much) as, whether of space (even unto) or time (during, while, until) or degree (equally with)","kjv_def":"against, and, as, at, before, by (that), even (to), for(-asmuch as), (hither-) to, [phrase] how long, into, as long (much) as, (so) that, till, toward, until, when, while, ([phrase] as) yet."},
"H5705":{"lemma":"עַד","xlit":"ʻad","pron":"ad","derivation":"(Aramaic) corresponding to H5704 (עַד)","strongs_def":"{as far (or long, or much) as, whether of space (even unto) or time (during, while, until) or degree (equally with)}","kjv_def":"[idiom] and, at, for, (hither-) to, on till, (un-) to, until, within."},
"H5706":{"lemma":"עַד","xlit":"ʻad","pron":"ad","derivation":"the same as H5703 (עַד) in the sense of the aim of an attack;","strongs_def":"booty","kjv_def":"prey."},
"H5707":{"lemma":"עֵד","xlit":"ʻêd","pron":"ayd","derivation":"contracted from H5749 (עוּד);","strongs_def":"concretely, a witness; abstractly, testimony; specifically, a recorder, i.e. prince","kjv_def":"witness."},
"H5708":{"lemma":"עֵד","xlit":"ʻêd","pron":"ayd","derivation":"from an unused root meaning to set a period (compare H5710 (עָדָה), H5749 (עוּד));","strongs_def":"the menstrual flux (as periodical); by implication (in plural) soiling","kjv_def":"filthy."},
"H5709":{"lemma":"עֲדָא","xlit":"ʻădâʼ","pron":"ad-aw'","derivation":"(Aramaic) or עֲדָה; (Aramaic), corresponding to H5710 (עָדָה)","strongs_def":"{to advance, i.e. pass on or continue; causatively, to remove; specifically, to bedeck (i.e. bring an ornament upon)}","kjv_def":"alter, depart, pass (away), remove, take (away)."},
"H5710":{"lemma":"עָדָה","xlit":"ʻâdâh","pron":"aw-daw'","derivation":"a primitive root;","strongs_def":"to advance, i.e. pass on or continue; causatively, to remove; specifically, to bedeck (i.e. bring an ornament upon)","kjv_def":"adorn, deck (self), pass by, take away."},
"H5711":{"lemma":"עָדָה","xlit":"ʻÂdâh","pron":"aw-daw'","derivation":"from H5710 (עָדָה); ornament;","strongs_def":"Adah, the name of two women","kjv_def":"Adah."},
"H5712":{"lemma":"עֵדָה","xlit":"ʻêdâh","pron":"ay-daw'","derivation":"feminine of H5707 (עֵד) in the original sense of fixture;","strongs_def":"a stated assemblage (specifically, a concourse, or generally, a family or crowd)","kjv_def":"assembly, company, congregation, multitude, people, swarm. Compare H5713 (עֵדָה)."},
"H5713":{"lemma":"עֵדָה","xlit":"ʻêdâh","pron":"ay-daw'","derivation":"feminine of H5707 (עֵד) in its techn. sense;","strongs_def":"testimony","kjv_def":"testimony, witness. Compare H5712 (עֵדָה)."},
"H5714":{"lemma":"עִדּוֹ","xlit":"ʻIddôw","pron":"id-do'","derivation":"or עִדּוֹא; or עִדִּיא; from H5710 (עָדָה); timely;","strongs_def":"Iddo (or Iddi), the name of five Israelites","kjv_def":"Iddo. Compare H3035 (יִדּוֹ), H3260 (יֶעְדִּי)."},
"H5715":{"lemma":"עֵדוּת","xlit":"ʻêdûwth","pron":"ay-dooth'","derivation":"feminine of H5707 (עֵד);","strongs_def":"testimony","kjv_def":"testimony, witness."},
"H5716":{"lemma":"עֲדִי","xlit":"ʻădîy","pron":"ad-ee'","derivation":"from H5710 (עָדָה) in the sense of trappings;","strongs_def":"finery; generally an outfit; specifically, a headstall","kjv_def":"[idiom] excellent, mouth, ornament."},
"H5717":{"lemma":"עֲדִיאֵל","xlit":"ʻĂdîyʼêl","pron":"ad-ee-ale'","derivation":"from H5716 (עֲדִי) and H410 (אֵל); ornament of God;","strongs_def":"Adiel, the name of three Israelites","kjv_def":"Adiel."},
"H5718":{"lemma":"עֲדָיָה","xlit":"ʻĂdâyâh","pron":"ad-aw-yaw'","derivation":"or עֲדָיָהוּ; from H5710 (עָדָה) and H3050 (יָהּ); Jah has adorned;","strongs_def":"Adajah, the name of eight Israelites","kjv_def":"Adaiah."},
"H5719":{"lemma":"עָדִין","xlit":"ʻâdîyn","pron":"aw-deen'","derivation":"from H5727 (עָדַן);","strongs_def":"voluptuous","kjv_def":"given to pleasures."},
"H5720":{"lemma":"עָדִין","xlit":"ʻÂdîyn","pron":"aw-deen'","derivation":"the same as H5719 (עָדִין);","strongs_def":"Adin, the name of two Israelites","kjv_def":"Adin."},
"H5721":{"lemma":"עֲדִינָא","xlit":"ʻĂdîynâʼ","pron":"ad-ee-naw'","derivation":"from H5719 (עָדִין); effeminacy;","strongs_def":"Adina, an Israelite","kjv_def":"Adina."},
"H5722":{"lemma":"עֲדִינוֹ","xlit":"ʻădîynôw","pron":"ad-ee-no'","derivation":"probably from H5719 (עָדִין) in the original sense of slender (i.e. a spear);","strongs_def":"his spear","kjv_def":"Adino."},
"H5723":{"lemma":"עֲדִיתַיִם","xlit":"ʻĂdîythayim","pron":"ad-ee-thah'-yim","derivation":"dual of a feminine of H5706 (עַד); double prey;","strongs_def":"Adithajim, a place in Palestine","kjv_def":"Adithaim."},
"H5724":{"lemma":"עַדְלַי","xlit":"ʻAdlay","pron":"ad-lah'-ee","derivation":"probably from an unused root of uncertain meaning;","strongs_def":"Adlai, an Israelite","kjv_def":"Adlai."},
"H5725":{"lemma":"עֲדֻלָּם","xlit":"ʻĂdullâm","pron":"ad-ool-lawm'","derivation":"probably from the passive participle of the same as H5724 (עַדְלַי);","strongs_def":"Adullam, a place in Palestine","kjv_def":"Adullam."},
"H5726":{"lemma":"עֲדֻלָּמִי","xlit":"ʻĂdullâmîy","pron":"ad-ool-law-mee'","derivation":"patrial from H5725 (עֲדֻלָּם);","strongs_def":"an Adullamite or native of Adullam","kjv_def":"Adullamite."},
"H5727":{"lemma":"עָדַן","xlit":"ʻâdan","pron":"aw-dan'","derivation":"a primitive root;","strongs_def":"to be soft or pleasant; figuratively and reflexively, to live voluptuously","kjv_def":"delight self."},
"H5728":{"lemma":"עֲדֶן","xlit":"ʻăden","pron":"ad-en'","derivation":"or עֲדֶנָּה; from H5704 (עַד) and H2004 (הֵן);","strongs_def":"till now","kjv_def":"yet."},
"H5729":{"lemma":"עֶדֶן","xlit":"ʻEden","pron":"eh'-den","derivation":"from H5727 (עָדַן); pleasure;","strongs_def":"Eden, a place in Mesopotamia","kjv_def":"Eden."},
"H5730":{"lemma":"עֵדֶן","xlit":"ʻêden","pron":"ay'-den","derivation":"or (feminine) עֶדְנָה; from H5727 (עָדַן);","strongs_def":"pleasure","kjv_def":"delicate, delight, pleasure. See also H1040 (בֵּית עֵדֶן)."},
"H5731":{"lemma":"עֵדֶן","xlit":"ʻÊden","pron":"ay'-den","derivation":"the same as H5730 (עֵדֶן) (masculine);","strongs_def":"Eden, the region of Adam's home","kjv_def":"Eden."},
"H5732":{"lemma":"עִדָּן","xlit":"ʻiddân","pron":"id-dawn'","derivation":"(Aramaic) from a root corresponding to that of H5708 (עֵד);","strongs_def":"a set time; technically, a year","kjv_def":"time."},
"H5733":{"lemma":"עַדְנָא","xlit":"ʻAdnâʼ","pron":"ad-naw'","derivation":"from H5727 (עָדַן); pleasure;","strongs_def":"Adna, the name of two Israelites","kjv_def":"Adna."},
"H5734":{"lemma":"עַדְנָה","xlit":"ʻAdnâh","pron":"ad-naw'","derivation":"from H5727 (עָדַן); pleasure;","strongs_def":"Adnah, the name of two Israelites","kjv_def":"Adnah."},
"H5735":{"lemma":"עֲדְעָדָה","xlit":"ʻĂdʻâdâh","pron":"ad-aw-daw'","derivation":"from H5712 (עֵדָה); festival;","strongs_def":"Adadah, a place in Palestine","kjv_def":"Adadah."},
"H5736":{"lemma":"עֲדַף","xlit":"ʻădaph","pron":"aw-daf'","derivation":"a primitive root;","strongs_def":"to be (causatively, have) redundant","kjv_def":"be more, odd number, be (have) over (and above), overplus, remain."},
"H5737":{"lemma":"עֲדַר","xlit":"ʻădar","pron":"aw-dar'","derivation":"a primitive root;","strongs_def":"to arrange as a battle, a vineyard (to hoe); hence, to muster and so to miss (or find wanting)","kjv_def":"dig, fail, keep (rank), lack."},
"H5738":{"lemma":"עֶדֶר","xlit":"ʻEder","pron":"eh'-der","derivation":"from H5737 (עֲדַר); an arrangement (i.e. drove);","strongs_def":"Eder, an Israelite","kjv_def":"Ader."},
"H5739":{"lemma":"עֵדֶר","xlit":"ʻêder","pron":"ay'-der","derivation":"from H5737 (עֲדַר);","strongs_def":"an arrangement, i.e. muster (of animals)","kjv_def":"drove, flock, herd."},
"H5740":{"lemma":"עֵדֶר","xlit":"ʻÊder","pron":"ay'-der","derivation":"the same as H5739 (עֵדֶר);","strongs_def":"Eder, the name of an Israelite and of two places in Palestine","kjv_def":"Edar, Eder."},
"H5741":{"lemma":"עַדְרִיאֵל","xlit":"ʻAdrîyʼêl","pron":"ad-ree-ale'","derivation":"from H5739 (עֵדֶר) and H410 (אֵל); flock of God;","strongs_def":"Adriel, an Israelite","kjv_def":"Adriel."},
"H5742":{"lemma":"עָדָשׁ","xlit":"ʻâdâsh","pron":"aw-dawsh'","derivation":"from an unused root of uncertain meaning;","strongs_def":"a lentil","kjv_def":"lentile."},
"H5743":{"lemma":"עוּב","xlit":"ʻûwb","pron":"oob","derivation":"a primitive root;","strongs_def":"to be dense or dark, i.e. to becloud","kjv_def":"cover with a cloud."},
"H5744":{"lemma":"עוֹבֵד","xlit":"ʻÔwbêd","pron":"o-bade'","derivation":"active participle of H5647 (עָבַד); serving;","strongs_def":"Obed, the name of five Israelites","kjv_def":"Obed."},
"H5745":{"lemma":"עוֹבָל","xlit":"ʻÔwbâl","pron":"o-bawl'","derivation":"of foreign derivation;","strongs_def":"Obal, a son of Joktan","kjv_def":"Obal."},
"H5746":{"lemma":"עוּג","xlit":"ʻûwg","pron":"oog","derivation":"a primitive root; properly, to gyrate; but used only as a denominative from H5692 (עֻגָּה),","strongs_def":"to bake (round cakes on the hearth)","kjv_def":"bake."},
"H5747":{"lemma":"עוֹג","xlit":"ʻÔwg","pron":"ogue","derivation":"probably from H5746 (עוּג); round;","strongs_def":"Og, a king of Bashan","kjv_def":"Og."},
"H5748":{"lemma":"עוּגָב","xlit":"ʻûwgâb","pron":"oo-gawb'","derivation":"or עֻגָּב; from H5689 (עֲגַב) in the original sense of breathing;","strongs_def":"a reed-instrument of music","kjv_def":"organ."},
"H5749":{"lemma":"עוּד","xlit":"ʻûwd","pron":"ood","derivation":"a primitive root;","strongs_def":"to duplicate or repeat; by implication, to protest, testify (as by reiteration); intensively, to encompass, restore (as a sort of reduplication)","kjv_def":"admonish, charge, earnestly, lift up, protest, call (take) to record, relieve, rob, solemnly, stand upright, testify, give warning, (bear, call to, give, take to) witness."},
"H5750":{"lemma":"עוֹד","xlit":"ʻôwd","pron":"ode","derivation":"or עֹד; from H5749 (עוּד);","strongs_def":"properly, iteration or continuance; used only adverbially (with or without preposition), again, repeatedly, still, more","kjv_def":"again, [idiom] all life long, at all, besides, but, else, further(-more), henceforth, (any) longer, (any) more(-over), [idiom] once, since, (be) still, when, (good, the) while (having being), (as, because, whether, while) yet (within)."},
"H5751":{"lemma":"עוֹד","xlit":"ʻôwd","pron":"ode","derivation":"(Aramaic) corresponding to H5750 (עוֹד)","strongs_def":"{properly, iteration or continuance; used only adverbially (with or without preposition), again, repeatedly, still, more}","kjv_def":"while."},
"H5752":{"lemma":"עוֹדֵד","xlit":"ʻÔwdêd","pron":"o-dade'","derivation":"or עֹדֵד; from H5749 (עוּד); reiteration;","strongs_def":"Oded, the name of two Israelites","kjv_def":"Oded."},
"H5753":{"lemma":"עָוָה","xlit":"ʻâvâh","pron":"aw-vaw'","derivation":"a primitive root;","strongs_def":"to crook, literally or figuratively","kjv_def":"do amiss, bow down, make crooked, commit iniquity, pervert, (do) perverse(-ly), trouble, [idiom] turn, do wickedly, do wrong."},
"H5754":{"lemma":"עַוָּה","xlit":"ʻavvâh","pron":"av-vaw'","derivation":"intensive from H5753 (עָוָה) abbreviated;","strongs_def":"overthrow","kjv_def":"[idiom] overturn."},
"H5755":{"lemma":"עִוָּה","xlit":"ʻIvvâh","pron":"iv-vaw'","derivation":"or עַוָּא; (2 Kings 17:24), for H5754 (עַוָּה);","strongs_def":"Ivvah or Avva, a region of Assyria","kjv_def":"Ava, Ivah."},
"H5756":{"lemma":"עוּז","xlit":"ʻûwz","pron":"ooz","derivation":"a primitive root;","strongs_def":"to be strong; causatively, to stregthen, i.e. (figuratively) to save (by flight)","kjv_def":"gather (self, self to flee), retire."},
"H5757":{"lemma":"עַוִּי","xlit":"ʻAvvîy","pron":"av-vee'","derivation":"patrial from H5755 (עִוָּה);","strongs_def":"an Avvite or native of Avvah (only plural)","kjv_def":"Avims, Avites."},
"H5758":{"lemma":"עִוְיָא","xlit":"ʻivyâʼ","pron":"iv-yaw'","derivation":"(Aramaic) from a root corresponding to H5753 (עָוָה);","strongs_def":"perverseness","kjv_def":"iniquity."},
"H5759":{"lemma":"עֲוִיל","xlit":"ʻăvîyl","pron":"av-eel'","derivation":"from H5764 (עוּל);","strongs_def":"a babe","kjv_def":"young child, little one."},
"H5760":{"lemma":"עֲוִיל","xlit":"ʻăvîyl","pron":"av-eel'","derivation":"from H5765 (עֲוַל);","strongs_def":"perverse (morally)","kjv_def":"ungodly."},
"H5761":{"lemma":"עַוִּים","xlit":"ʻAvvîym","pron":"av-veem'","derivation":"plural of H5757 (עַוִּי);","strongs_def":"Avvim (as inhabited by Avvites), a place in Palestine (with the article prefix)","kjv_def":"Avim."},
"H5762":{"lemma":"עֲוִית","xlit":"ʻĂvîyth","pron":"av-veeth'","derivation":"or perhaps (עַיּוֹת (as if plural of H5857 (עַי)) (עַיוּת; from H5753 (עָוָה); ruin;","strongs_def":"Avvith (or Avvoth), a place in Palestine","kjv_def":"Avith."},
"H5763":{"lemma":"עוּל","xlit":"ʻûwl","pron":"ool","derivation":"a primitive root;","strongs_def":"to suckle, i.e. give milk","kjv_def":"milch, (ewe great) with young."},
"H5764":{"lemma":"עוּל","xlit":"ʻûwl","pron":"ool","derivation":"from H5763 (עוּל);","strongs_def":"a babe","kjv_def":"sucking child, infant."},
"H5765":{"lemma":"עֲוַל","xlit":"ʻăval","pron":"aw-val'","derivation":"a primitive root;","strongs_def":"to distort (morally)","kjv_def":"deal unjustly, unrighteous."},
"H5766":{"lemma":"עֶוֶל","xlit":"ʻevel","pron":"eh'-vel","derivation":"or עָוֶל; and (feminine) עַוְלָה; or עוֹלָה; or עֹלָה; from H5765 (עֲוַל);","strongs_def":"(moral) evil","kjv_def":"iniquity, perverseness, unjust(-ly), unrighteousness(-ly); wicked(-ness)."},
"H5767":{"lemma":"עַוָּל","xlit":"ʻavvâl","pron":"av-vawl'","derivation":"intensive from H5765 (עֲוַל);","strongs_def":"evil (morally)","kjv_def":"unjust, unrighteous, wicked."},
"H5768":{"lemma":"עוֹלֵל","xlit":"ʻôwlêl","pron":"o-lale'","derivation":"or עֹלָל; from H5763 (עוּל);","strongs_def":"a suckling","kjv_def":"babe, (young) child, infant, little one."},
"H5769":{"lemma":"עוֹלָם","xlit":"ʻôwlâm","pron":"o-lawm'","derivation":"or עֹלָם; from H5956 (עָלַם);","strongs_def":"properly, concealed, i.e. the vanishing point; generally, time out of mind (past or future), i.e. (practically) eternity; frequentatively, adverbial (especially with prepositional prefix) always","kjv_def":"alway(-s), ancient (time), any more, continuance, eternal, (for, (n-)) ever(-lasting, -more, of old), lasting, long (time), (of) old (time), perpetual, at any time, (beginning of the) world ([phrase] without end). Compare H5331 (נֶצַח), H5703 (עַד)."},
"H5770":{"lemma":"עָוַן","xlit":"ʻâvan","pron":"aw-van'","derivation":"denominative from H5869 (עַיִן);","strongs_def":"to watch (with jealosy)","kjv_def":"eye."},
"H5771":{"lemma":"עָוֺן","xlit":"ʻâvôn","pron":"aw-vone'","derivation":"or עָווֹן; (2 Kings 7:9; Psalm 51:5 (H7 (אֲבַד))), from H5753 (עָוָה);","strongs_def":"perversity, i.e. (moral) evil","kjv_def":"fault, iniquity, mischeif, punishment (of iniquity), sin."},
"H5772":{"lemma":"עוֹנָה","xlit":"ʻôwnâh","pron":"o-naw'","derivation":"from an unused root apparently meaning to dwell together;","strongs_def":"sexual (cohabitation)","kjv_def":"duty of marriage."},
"H5773":{"lemma":"עַוְעֶה","xlit":"ʻavʻeh","pron":"av-eh'","derivation":"from H5753 (עָוָה);","strongs_def":"perversity","kjv_def":"[idiom] perverse."},
"H5774":{"lemma":"עוּף","xlit":"ʻûwph","pron":"oof","derivation":"a primitive root; to cover (with wings or obscurity); hence (as denominative from H5775 (עוֹף))","strongs_def":"to fly; also (by implication of dimness) to faint (from the darkness of swooning)","kjv_def":"brandish, be (wax) faint, flee away, fly (away), [idiom] set, shine forth, weary."},
"H5775":{"lemma":"עוֹף","xlit":"ʻôwph","pron":"ofe","derivation":"from H5774 (עוּף);","strongs_def":"a bird (as covered with feathers, or rather as covering with wings), often collectively","kjv_def":"bird, that flieth, flying, fowl."},
"H5776":{"lemma":"עוֹף","xlit":"ʻôwph","pron":"ofe","derivation":"(Aramaic) corresponding to H5775 (עוֹף)","strongs_def":"{a bird (as covered with feathers, or rather as covering with wings), often collectively}","kjv_def":"fowl."},
"H5777":{"lemma":"עוֹפֶרֶת","xlit":"ʻôwphereth","pron":"o-feh'-reth","derivation":"or עֹפֶרֶת; feminine participle active of H6080 (עָפַר);","strongs_def":"lead (from its dusty color)","kjv_def":"lead."},
"H5778":{"lemma":"עוֹפַי","xlit":"ʻÔwphay","pron":"o-fah'-ee","derivation":"from H5775 (עוֹף); birdlike;","strongs_def":"Ephai, an Israelite","kjv_def":"Ephai (from margin)."},
"H5779":{"lemma":"עוּץ","xlit":"ʻûwts","pron":"oots","derivation":"a primitive root;","strongs_def":"to consult","kjv_def":"take advice ((counsel) together)."},
"H5780":{"lemma":"עוּץ","xlit":"ʻÛwts","pron":"oots","derivation":"apparently from H5779 (עוּץ); consultation;","strongs_def":"Uts, a son of Aram, also a Seirite, and the regions settled by them.","kjv_def":"Uz."},
"H5781":{"lemma":"עוּק","xlit":"ʻûwq","pron":"ook","derivation":"a primitive root;","strongs_def":"to pack","kjv_def":"be pressed."},
"H5782":{"lemma":"עוּר","xlit":"ʻûwr","pron":"oor","derivation":"a primitive root (rather identical with through the idea of opening the eyes);","strongs_def":"to wake (literally or figuratively)","kjv_def":"(a-) wake(-n, up), lift up (self), [idiom] master, raise (up), stir up (self)."},
"H5783":{"lemma":"עוּר","xlit":"ʻûwr","pron":"oor","derivation":"a primitive root;","strongs_def":"to (be) bare","kjv_def":"be made naked."},
"H5784":{"lemma":"עוּר","xlit":"ʻûwr","pron":"oor","derivation":"(Aramaic)","strongs_def":"chaff (as the naked husk)","kjv_def":"chaff."},
"H5785":{"lemma":"עוֹר","xlit":"ʻôwr","pron":"ore","derivation":"from H5783 (עוּר);","strongs_def":"skin (as naked); by implication, hide, leather","kjv_def":"hide, leather, skin."},
"H5786":{"lemma":"עָוַר","xlit":"ʻâvar","pron":"aw-var'","derivation":"a primitive root (rather denominatively from H5785 (עוֹר) through the idea of a film over the eyes);","strongs_def":"to blind","kjv_def":"blind, put out. See also H5895 (עַיִר)."},
"H5787":{"lemma":"עִוֵּר","xlit":"ʻivvêr","pron":"iv-vare'","derivation":"intensive from H5786 (עָוַר);","strongs_def":"blind (literally or figuratively)","kjv_def":"blind (men, people)."},
"H5788":{"lemma":"עִוָּרוֹן","xlit":"ʻivvârôwn","pron":"iv-vaw-rone'","derivation":"and (feminine) עַוֶּרֶת; from H5787 (עִוֵּר);","strongs_def":"blindness","kjv_def":"blind(-ness)."},
"H5789":{"lemma":"עוּשׁ","xlit":"ʻûwsh","pron":"oosh","derivation":"a primitive root;","strongs_def":"to hasten","kjv_def":"assemble self."},
"H5790":{"lemma":"עוּת","xlit":"ʻûwth","pron":"ooth","derivation":"for H5789 (עוּשׁ);","strongs_def":"to hasten, i.e. succor","kjv_def":"speak in season."},
"H5791":{"lemma":"עָוַת","xlit":"ʻâvath","pron":"aw-vath'","derivation":"a primitive root;","strongs_def":"to wrest","kjv_def":"bow self, (make) crooked., falsifying, overthrow, deal perversely, pervert, subvert, turn upside down."},
"H5792":{"lemma":"עַוָּתָה","xlit":"ʻavvâthâh","pron":"av-vaw-thaw'","derivation":"from H5791 (עָוַת);","strongs_def":"oppression","kjv_def":"wrong."},
"H5793":{"lemma":"עוּתַי","xlit":"ʻÛwthay","pron":"oo-thah'-ee","derivation":"from H5790 (עוּת); succoring;","strongs_def":"Uthai, the name of two Israelites","kjv_def":"Uthai."},
"H5794":{"lemma":"עַז","xlit":"ʻaz","pron":"az","derivation":"from H5810 (עָזַז);","strongs_def":"strong, vehement, harsh","kjv_def":"fierce, [phrase] greedy, mighty, power, roughly, strong."},
"H5795":{"lemma":"עֵז","xlit":"ʻêz","pron":"aze","derivation":"from H5810 (עָזַז);","strongs_def":"a she-goat (as strong), but masculine in plural (which also is used elliptically for goat's hair)","kjv_def":"(she) goat, kid."},
"H5796":{"lemma":"עֵז","xlit":"ʻêz","pron":"aze","derivation":"(Aramaic) corresponding to H5795 (עֵז)","strongs_def":"{a she-goat (as strong), but masculine in plural (which also is used elliptically for goat's hair)}","kjv_def":"goat."},
"H5797":{"lemma":"עֹז","xlit":"ʻôz","pron":"oze","derivation":"or (fully) עוֹז; from H5810 (עָזַז);","strongs_def":"strength in various applications (force, security, majesty, praise)","kjv_def":"boldness, loud, might, power, strength, strong."},
"H5798":{"lemma":"עֻזָּא","xlit":"ʻUzzâʼ","pron":"ooz-zaw'","derivation":"or עֻזָּה; feminine of H5797 (עֹז); strength;","strongs_def":"Uzza or Uzzah, the name of five Israelites","kjv_def":"Uzza, Uzzah."},
"H5799":{"lemma":"עֲזָאזֵל","xlit":"ʻăzâʼzêl","pron":"az-aw-zale'","derivation":"from H5795 (עֵז) and H235 (אָזַל);","strongs_def":"goat of departure; the scapegoat","kjv_def":"scapegoat."},
"H5800":{"lemma":"עָזַב","xlit":"ʻâzab","pron":"aw-zab'","derivation":"a primitive root;","strongs_def":"to loosen, i.e. relinquish, permit, etc.","kjv_def":"commit self, fail, forsake, fortify, help, leave (destitute, off), refuse, [idiom] surely."},
"H5801":{"lemma":"עִזָּבוֹן","xlit":"ʻizzâbôwn","pron":"iz-zaw-bone'","derivation":"from H5800 (עָזַב) in the sense of letting go (for a price, i.e. selling);","strongs_def":"trade, i.e. the place (mart) or the payment (revenue)","kjv_def":"fair, ware."},
"H5802":{"lemma":"עַזְבּוּק","xlit":"ʻAzbûwq","pron":"az-book'","derivation":"from H5794 (עַז) and the root of H950 (בּוּקָה); stern depopulator;","strongs_def":"Azbuk, an Israelite","kjv_def":"Azbuk."},
"H5803":{"lemma":"עַזְגָּד","xlit":"ʻAzgâd","pron":"az-gawd'","derivation":"from H5794 (עַז) and H1409 (גָּד); stern troop;","strongs_def":"Azgad, an Israelite","kjv_def":"Azgad."},
"H5804":{"lemma":"עַזָּה","xlit":"ʻAzzâh","pron":"az-zaw'","derivation":"feminine of H5794 (עַז); strong;","strongs_def":"Azzah, a place in Palestine","kjv_def":"Azzah, Gaza."},
"H5805":{"lemma":"עֲזוּבָה","xlit":"ʻăzûwbâh","pron":"az-oo-baw'","derivation":"feminine passive participle of H5800 (עָזַב);","strongs_def":"desertion (of inhabitants)","kjv_def":"forsaking."},
"H5806":{"lemma":"עֲזוּבָה","xlit":"ʻĂzûwbâh","pron":"az-oo-baw'","derivation":"the same as H5805 (עֲזוּבָה);","strongs_def":"Azubah, the name of two Israelitesses","kjv_def":"Azubah."},
"H5807":{"lemma":"עֱזוּז","xlit":"ʻĕzûwz","pron":"ez-ooz'","derivation":"from H5810 (עָזַז);","strongs_def":"forcibleness","kjv_def":"might, strength."},
"H5808":{"lemma":"עִזּוּז","xlit":"ʻizzûwz","pron":"iz-zooz'","derivation":"from H5810 (עָזַז);","strongs_def":"forcible; collectively and concretely, an army","kjv_def":"power, strong."},
"H5809":{"lemma":"עַזּוּר","xlit":"ʻAzzûwr","pron":"az-zoor'","derivation":"or עַזֻּר; from H5826 (עָזַר); helpful;","strongs_def":"Azzur, the name of three Israelites","kjv_def":"Azur, Azzur."},
"H5810":{"lemma":"עָזַז","xlit":"ʻâzaz","pron":"aw-zaz'","derivation":"a primitive root;","strongs_def":"to be stout (literally or figuratively)","kjv_def":"harden, impudent, prevail, strengthen (self), be strong."},
"H5811":{"lemma":"עָזָז","xlit":"ʻÂzâz","pron":"aw-zawz'","derivation":"from H5810 (עָזַז); strong;","strongs_def":"Azaz, an Israelite","kjv_def":"Azaz."},
"H5812":{"lemma":"עֲזַזְיָהוּ","xlit":"ʻĂzazyâhûw","pron":"az-az-yaw'-hoo","derivation":"from H5810 (עָזַז) and H3050 (יָהּ); Jah has strengthened;","strongs_def":"Azazjah, the name of three Israelites","kjv_def":"Azaziah."},
"H5813":{"lemma":"עֻזִּי","xlit":"ʻUzzîy","pron":"ooz-zee'","derivation":"from H5810 (עָזַז); forceful;","strongs_def":"Uzzi, the name of six Israelites","kjv_def":"Uzzi."},
"H5814":{"lemma":"עֻזִּיָּא","xlit":"ʻUzzîyâʼ","pron":"ooz-zee-yaw'","derivation":"perhaps for H5818 (עֻזִּיָּה);","strongs_def":"Uzzija, an Israelite","kjv_def":"Uzzia."},
"H5815":{"lemma":"עֲזִיאֵל","xlit":"ʻĂzîyʼêl","pron":"az-ee-ale'","derivation":"from H5756 (עוּז) and H410 (אֵל); strengthened of God;","strongs_def":"Aziel, an Israelite","kjv_def":"Aziel. Compare H3268 (יַעֲזִיאֵל)."},
"H5816":{"lemma":"עֻזִּיאֵל","xlit":"ʻUzzîyʼêl","pron":"ooz-zee-ale'","derivation":"from H5797 (עֹז) and H410 (אֵל); strength of God;","strongs_def":"Uzziel, the name of six Israelites","kjv_def":"Uzziel."},
"H5817":{"lemma":"עׇזִּיאֵלִי","xlit":"ʻOzzîyʼêlîy","pron":"oz-zee-ay-lee'","derivation":"patronymically from H5816 (עֻזִּיאֵל);","strongs_def":"an Uzzielite (collectively) or descendants of Uzziel","kjv_def":"Uzzielites."},
"H5818":{"lemma":"עֻזִּיָּה","xlit":"ʻUzzîyâh","pron":"ooz-zee-yaw'","derivation":"or עֻזִּיָּהוּ; from H5797 (עֹז) and H3050 (יָהּ); strength of Jah;","strongs_def":"Uzzijah, the name of five Israelites","kjv_def":"Uzziah."},
"H5819":{"lemma":"עֲזִיזָא","xlit":"ʻĂzîyzâʼ","pron":"az-ee-zaw'","derivation":"from H5756 (עוּז); strengthfulness;","strongs_def":"Aziza, an Israelite","kjv_def":"Aziza."},
"H5820":{"lemma":"עַזְמָוֶת","xlit":"ʻAzmâveth","pron":"az-maw'-veth","derivation":"from H5794 (עַז) and H4194 (מָוֶת); strong one of death;","strongs_def":"Azmaveth, the name of three Israelites and of a place in Palestine","kjv_def":"Azmaveth. See also H1041 (בֵּית עַזְמָוֶת)."},
"H5821":{"lemma":"עַזָּן","xlit":"ʻAzzân","pron":"az-zawn'","derivation":"from H5794 (עַז); strong one;","strongs_def":"Azzan, an Israelite","kjv_def":"Azzan."},
"H5822":{"lemma":"עׇזְנִיָּה","xlit":"ʻoznîyâh","pron":"oz-nee-yaw'","derivation":"probably feminine of H5797 (עֹז);","strongs_def":"probably the sea-eagle (from its strength)","kjv_def":"ospray."},
"H5823":{"lemma":"עָזַק","xlit":"ʻâzaq","pron":"aw-zak'","derivation":"a primitive root;","strongs_def":"to grub over","kjv_def":"fence about."},
"H5824":{"lemma":"עִזְקָא","xlit":"ʻizqâʼ","pron":"iz-kaw'","derivation":"(Aramaic) from a root corresponding to H5823 (עָזַק);","strongs_def":"a signet-ring (as engraved)","kjv_def":"signet."},
"H5825":{"lemma":"עֲזֵקָה","xlit":"ʻĂzêqâh","pron":"az-ay-kaw'","derivation":"from H5823 (עָזַק); tilled;","strongs_def":"Azekah, a place in Palestine","kjv_def":"Azekah."},
"H5826":{"lemma":"עָזַר","xlit":"ʻâzar","pron":"aw-zar'","derivation":"a primitive root;","strongs_def":"to surround, i.e. protect or aid","kjv_def":"help, succour."},
"H5827":{"lemma":"עֶזֶר","xlit":"ʻEzer","pron":"eh'-zer","derivation":"from H5826 (עָזַר); help;","strongs_def":"Ezer, the name of two Israelites","kjv_def":"Ezer. Compare H5829 (עֵזֶר)."},
"H5828":{"lemma":"עֵזֶר","xlit":"ʻêzer","pron":"ay'-zer","derivation":"from H5826 (עָזַר);","strongs_def":"aid","kjv_def":"help."},
"H5829":{"lemma":"עֵזֶר","xlit":"ʻÊzer","pron":"ay'-zer","derivation":"the same as H5828 (עֵזֶר);","strongs_def":"Ezer, the name of four Israelites","kjv_def":"Ezer. Compare H5827 (עֶזֶר)."},
"H5830":{"lemma":"עֶזְרָא","xlit":"ʻEzrâʼ","pron":"ez-raw'","derivation":"a variation of H5833 (עֶזְרָה);","strongs_def":"Ezra, an Israelite","kjv_def":"Ezra."},
"H5831":{"lemma":"עֶזְרָא","xlit":"ʻEzrâʼ","pron":"ez-raw'","derivation":"(Aramaic) corresponding to H5830 (עֶזְרָא);","strongs_def":"Ezra, an Israelite","kjv_def":"Ezra."},
"H5832":{"lemma":"עֲזַרְאֵל","xlit":"ʻĂzarʼêl","pron":"az-ar-ale'","derivation":"from H5826 (עָזַר) and H410 (אֵל); God has helped;","strongs_def":"Azarel, the name of five Israelites","kjv_def":"Azarael, Azareel."},
"H5833":{"lemma":"עֶזְרָה","xlit":"ʻezrâh","pron":"ez-raw'","derivation":"or עֶזְרָת; (Psalm 60:11 (13); Psalm 108:12 (13)), feminine of H5828 (עֵזֶר);","strongs_def":"aid","kjv_def":"help(-ed, -er)."},
"H5834":{"lemma":"עֶזְרָה","xlit":"ʻEzrâh","pron":"ez-raw'","derivation":"the same as H5833 (עֶזְרָה);","strongs_def":"Ezrah, an Israelite","kjv_def":"Ezrah."},
"H5835":{"lemma":"עֲזָרָה","xlit":"ʻăzârâh","pron":"az-aw-raw'","derivation":"from H5826 (עָזַר) in its original meaning of surrounding;","strongs_def":"an inclosure; also a border","kjv_def":"court, settle."},
"H5836":{"lemma":"עֶזְרִי","xlit":"ʻEzrîy","pron":"ez-ree'","derivation":"from H5828 (עֵזֶר); helpful;","strongs_def":"Ezri, an Israelite","kjv_def":"Ezri."},
"H5837":{"lemma":"עַזְרִיאֵל","xlit":"ʻAzrîyʼêl","pron":"az-ree-ale'","derivation":"from H5828 (עֵזֶר) and H410 (אֵל); help of God;","strongs_def":"Azriel, the name of three Israelites","kjv_def":"Azriel."},
"H5838":{"lemma":"עֲזַרְיָה","xlit":"ʻĂzaryâh","pron":"az-ar-yaw'","derivation":"or עֲזַרְיָהוּ; from H5826 (עָזַר) and H3050 (יָהּ); Jah has helped;","strongs_def":"Azarjah, the name of nineteen Israelites","kjv_def":"Azariah."},
"H5839":{"lemma":"עֲזַרְיָה","xlit":"ʻĂzaryâh","pron":"az-ar-yaw'","derivation":"(Aramaic) corresponding to H5838 (עֲזַרְיָה);","strongs_def":"Azarjah, one of Daniel's companions","kjv_def":"Azariah."},
"H5840":{"lemma":"עַזְרִיקָם","xlit":"ʻAzrîyqâm","pron":"az-ree-kawm'","derivation":"from H5828 (עֵזֶר) and active participle of H6965 (קוּם); help of an enemy;","strongs_def":"Azrikam, the name of four Israelites","kjv_def":"Azrikam."},
"H5841":{"lemma":"עַזָּתִי","xlit":"ʻAzzâthîy","pron":"az-zaw-thee'","derivation":"patrial from H5804 (עַזָּה);","strongs_def":"an Azzathite or inhabitant of Azzah","kjv_def":"Gazathite, Gazite."},
"H5842":{"lemma":"עֵט","xlit":"ʻêṭ","pron":"ate","derivation":"from H5860 (עִיט) (contracted) in the sense of swooping, i.e. side-long stroke;","strongs_def":"a stylus or marking stick","kjv_def":"pen."},
"H5843":{"lemma":"עֵטָא","xlit":"ʻêṭâʼ","pron":"ay-taw'","derivation":"(Aramaic) from H3272 (יְעַט);","strongs_def":"prudence","kjv_def":"counsel."},
"H5844":{"lemma":"עָטָה","xlit":"ʻâṭâh","pron":"aw-taw'","derivation":"a primitive root;","strongs_def":"to wrap, i.e. cover, veil, cloth, or roll","kjv_def":"array self, be clad, (put a) cover (-ing, self), fill, put on, [idiom] surely, turn aside."},
"H5845":{"lemma":"עֲטִין","xlit":"ʻăṭîyn","pron":"at-een'","derivation":"from an unused root meaning apparently to contain;","strongs_def":"a receptacle (for milk, i.e. pail; figuratively, breast)","kjv_def":"breast."},
"H5846":{"lemma":"עֲטִישָׁה","xlit":"ʻăṭîyshâh","pron":"at-ee-shaw'","derivation":"from an unused root meaning to sneeze;","strongs_def":"sneezing","kjv_def":"sneezing."},
"H5847":{"lemma":"עֲטַלֵּף","xlit":"ʻăṭallêph","pron":"at-al-lafe'","derivation":"of uncertain derivation;","strongs_def":"a bat","kjv_def":"bat."},
"H5848":{"lemma":"עָטַף","xlit":"ʻâṭaph","pron":"aw-taf'","derivation":"a primitive root;","strongs_def":"to shroud, i.e. clothe (whether transitive or reflex.); hence (from the idea of darkness) to languish","kjv_def":"cover (over), fail, faint, feebler, hide self, be overwhelmed, swoon."},
"H5849":{"lemma":"עָטַר","xlit":"ʻâṭar","pron":"aw-tar'","derivation":"a primitive root;","strongs_def":"to encircle (for attack or protection); especially to crown (literally or figuratively)","kjv_def":"compass, crown."},
"H5850":{"lemma":"עֲטָרָה","xlit":"ʻăṭârâh","pron":"at-aw-raw'","derivation":"from H5849 (עָטַר);","strongs_def":"a crown","kjv_def":"crown."},
"H5851":{"lemma":"עֲטָרָה","xlit":"ʻĂṭârâh","pron":"at-aw-raw'","derivation":"the same as H5850 (עֲטָרָה);","strongs_def":"Atarah, an Israelitess","kjv_def":"Atarah."},
"H5852":{"lemma":"עֲטָרוֹת","xlit":"ʻĂṭârôwth","pron":"at-aw-roth'","derivation":"or עֲטָרֹת; plural of H5850 (עֲטָרָה);","strongs_def":"Ataroth, the name (thus simply) of two places in Palestine","kjv_def":"Ataroth."},
"H5853":{"lemma":"עַטְרוֹת אַדָּר","xlit":"ʻAṭrôwth ʼAddâr","pron":"at-roth' ad-dawr'","derivation":"from the same as H5852 (עֲטָרוֹת) and H146 (אַדָּר); crowns of Addar;","strongs_def":"Atroth-Addar, a place in Palestine","kjv_def":"Ataroth-adar(-addar)."},
"H5854":{"lemma":"עַטְרוֹת בֵּית יוֹאָב","xlit":"ʻAṭrôwth bêyth Yôwʼâb","pron":"at-roth' bayth yoawb'","derivation":"from the same as H5852 (עֲטָרוֹת) and H1004 (בַּיִת) and H3097 (יוֹאָב); crowns of the house of Joab;","strongs_def":"Atroth-beth-Joab, a place in Palestine","kjv_def":"Ataroth the house of Joab."},
"H5855":{"lemma":"עַטְרוֹת שׁוֹפָן","xlit":"ʻAṭrôwth Shôwphân","pron":"at-roth' sho-fawn'","derivation":"from the same as H5852 (עֲטָרוֹת) and a name otherwise unused (being from the same as H8226 (שָׂפַן)) meaning hidden; crowns of Shophan;","strongs_def":"Atroth-Shophan, a place in Palestine","kjv_def":"Atroth, Shophan (as if two places)."},
"H5856":{"lemma":"עִי","xlit":"ʻîy","pron":"ee","derivation":"from H5753 (עָוָה);","strongs_def":"a ruin (as if overturned)","kjv_def":"heap."},
"H5857":{"lemma":"עַי","xlit":"ʻAy","pron":"ah'ee","derivation":"or (feminine) עַיָּא; (Nehemiah 11:31), or עַיָּת; (Isaiah 10:28), for H5856 (עִי);","strongs_def":"Ai, Aja or Ajath, a place in Palestine","kjv_def":"Ai, Aija, Aijath, Hai."},
"H5858":{"lemma":"עֵיבָל","xlit":"ʻÊybâl","pron":"ay-bawl'","derivation":"perhaps from an unused root probably meaning to be bald; bare;","strongs_def":"Ebal, a mountain of Palestine","kjv_def":"Ebal."},
"H5859":{"lemma":"עִיּוֹן","xlit":"ʻÎyôwn","pron":"ee-yone'","derivation":"from H5856 (עִי); ruin;","strongs_def":"Ijon, a place in Palestine","kjv_def":"Ijon."},
"H5860":{"lemma":"עִיט","xlit":"ʻîyṭ","pron":"eet","derivation":"a primitive root;","strongs_def":"to swoop down upon (literally or figuratively)","kjv_def":"fly, rail."},
"H5861":{"lemma":"עַיִט","xlit":"ʻayiṭ","pron":"ah'-yit","derivation":"from H5860 (עִיט);","strongs_def":"a hawk or other bird of prey","kjv_def":"bird, fowl, ravenous (bird)."},
"H5862":{"lemma":"עֵיטָם","xlit":"ʻÊyṭâm","pron":"ay-tawm'","derivation":"from H5861 (עַיִט); hawk-ground;","strongs_def":"Etam, a place in Palestine","kjv_def":"Etam."},
"H5863":{"lemma":"עִיֵּי הָעֲבָרִים","xlit":"ʻÎyêy hâ-ʻĂbârîym","pron":"ee-yay' haw-ab-aw-reem'","derivation":"from the plural of H5856 (עִי) and the plural of the active participle of H5674 (עָבַר) with the article interposed; ruins of the passers;","strongs_def":"Ije-ha-Abarim, a place near Palestine","kjv_def":"Ije-abarim."},
"H5864":{"lemma":"עִיִּים","xlit":"ʻÎyîym","pron":"ee-yeem'","derivation":"plural of H5856 (עִי); ruins;","strongs_def":"Ijim, a place in the Desert.","kjv_def":"Iim."},
"H5865":{"lemma":"עֵילוֹם","xlit":"ʻêylôwm","pron":"ay-lome'","derivation":"for H5769 (עוֹלָם)","strongs_def":"{concealed, i.e. the vanishing point; generally, time out of mind (past or future), i.e. (practically) eternity; frequentatively, adverbial (especially with prepositional prefix) always}","kjv_def":"ever."},
"H5866":{"lemma":"עִילַי","xlit":"ʻÎylay","pron":"ee-lah'-ee","derivation":"from H5927 (עָלָה); elevated;","strongs_def":"Ilai, an Israelite","kjv_def":"Ilai."},
"H5867":{"lemma":"עֵילָם","xlit":"ʻÊylâm","pron":"ay-lawm'","derivation":"or עוֹלָם; (Ezra 10:2; Jeremiah 49:36), probably from H5956 (עָלַם); hidden, i.e. distant;","strongs_def":"Elam, a son of Shem and his descendants, with their country; also of six Israelites","kjv_def":"Elam."},
"H5868":{"lemma":"עֲיָם","xlit":"ʻăyâm","pron":"ah-yawm'","derivation":"of doubtful origin and authenticity;","strongs_def":"probably meaning strength","kjv_def":"mighty."},
"H5869":{"lemma":"עַיִן","xlit":"ʻayin","pron":"ah'-yin","derivation":"probably a primitive word;","strongs_def":"an eye (literally or figuratively); by analogy, a fountain (as the eye of the landscape)","kjv_def":"affliction, outward appearance, [phrase] before, [phrase] think best, colour, conceit, [phrase] be content, countenance, [phrase] displease, eye((-brow), (-d), -sight), face, [phrase] favour, fountain, furrow (from the margin), [idiom] him, [phrase] humble, knowledge, look, ([phrase] well), [idiom] me, open(-ly), [phrase] (not) please, presence, [phrase] regard, resemblance, sight, [idiom] thee, [idiom] them, [phrase] think, [idiom] us, well, [idiom] you(-rselves)."},
"H5870":{"lemma":"עַיִן","xlit":"ʻayin","pron":"ah'-yin","derivation":"(Aramaic) corresponding to H5869 (עַיִן);","strongs_def":"an eye","kjv_def":"eye."},
"H5871":{"lemma":"עַיִן","xlit":"ʻAyin","pron":"ah'-yin","derivation":"the same as H5869 (עַיִן); fountain;","strongs_def":"Ajin, the name (thus simply) of two places in Palestine","kjv_def":"Ain."},
"H5872":{"lemma":"עֵין גֶּדִי","xlit":"ʻÊyn Gedîy","pron":"ane geh'-dee","derivation":"from H5869 (עַיִן) and H1423 (גְּדִי); fountain of a kid;","strongs_def":"En-Gedi, a place in Palestine","kjv_def":"En-gedi."},
"H5873":{"lemma":"עֵין גַּנִּים","xlit":"ʻÊyn Gannîym","pron":"ane gan-neem'","derivation":"from H5869 (עַיִן) and the plural of H1588 (גַּן); fountain of gardens;","strongs_def":"En-Gannim, a place in Palestine","kjv_def":"En-gannim."},
"H5874":{"lemma":"עֵין־דֹּאר","xlit":"ʻÊyn-Dôʼr","pron":"ane-dore'","derivation":"or עֵין דּוֹר; or עֵין־דֹּר; from H5869 (עַיִן) and H1755 (דּוֹר); fountain of dwelling;","strongs_def":"En-Dor, a place in Palestine","kjv_def":"En-dor."},
"H5875":{"lemma":"עֵין הַקּוֹרֵא","xlit":"ʻÊyn haq-Qôwrêʼ","pron":"ane-hak-ko-ray'","derivation":"from H5869 (עַיִן) and the active participle of H7121 (קָרָא); fountain of One calling;","strongs_def":"En-hak-Kore, a place near Palestine","kjv_def":"En-hakhore."},
"H5876":{"lemma":"עֵין חַדָּה","xlit":"ʻÊyn Chaddâh","pron":"ane khad-daw'","derivation":"from H5869 (עַיִן) and the feminine of a derivative from H2300 (חָדַד); fountain of sharpness;","strongs_def":"En-Chaddah, a place in Palestine","kjv_def":"En-haddah."},
"H5877":{"lemma":"עֵין חָצוֹר","xlit":"ʻÊyn Châtsôwr","pron":"ane khaw-tsore'","derivation":"from H5869 (עַיִן) and the same as H2674 (חָצוֹר); fountain of a village;","strongs_def":"En-Chatsor, a place in Palestine","kjv_def":"En-hazor."},
"H5878":{"lemma":"עֵין חֲרֹד","xlit":"ʻÊyn Chărôd","pron":"ane khar-ode'","derivation":"from H5869 (עַיִן) and a derivative of H2729 (חָרַד); fountain of trembling;","strongs_def":"En-Charod, a place in Palestine","kjv_def":"well of Harod."},
"H5879":{"lemma":"עֵינַיִם","xlit":"ʻÊynayim","pron":"ay-nah'-yim","derivation":"or עֵינָם; dual of H5869 (עַיִן); double fountain;","strongs_def":"Enajim or Enam, a place in Palestine","kjv_def":"Enaim, openly (Genesis 38:21)."},
"H5880":{"lemma":"עֵין מִשְׁפָּט","xlit":"ʻÊyn Mishpâṭ","pron":"ane mish-pawt'","derivation":"from H5869 (עַיִן) and H4941 (מִשְׁפָּט); fountain of judgment;","strongs_def":"En-Mishpat, a place near Palestine","kjv_def":"En-mishpat."},
"H5881":{"lemma":"עֵינָן","xlit":"ʻÊynân","pron":"ay-nawn'","derivation":"from H5869 (עַיִן); having eyes;","strongs_def":"Enan, an Israelite","kjv_def":"Enan. Compare H2704 (חֲצַר עֵינָן)."},
"H5882":{"lemma":"עֵין עֶגְלַיִם","xlit":"ʻÊyn ʻEglayim","pron":"ane eg-lah'-yim","derivation":"עַיִן and the dual of H5695 (עֵגֶל); fountain of two calves;","strongs_def":"En-Eglajim, a place in Palestine","kjv_def":"En-eglaim."},
"H5883":{"lemma":"עֵין רֹגֵל","xlit":"ʻÊyn Rôgêl","pron":"ane ro-gale'","derivation":"from H5869 (עַיִן) and the active participle of H7270 (רָגַל); fountain of a traveller;","strongs_def":"En-Rogel, a place near Jerusalem","kjv_def":"En-rogel."},
"H5884":{"lemma":"עֵין רִמּוֹן","xlit":"ʻÊyn Rimmôwn","pron":"ane rim-mone'","derivation":"from H5869 (עַיִן) and H7416 (רִמּוֹן); fountain of a pomegranate;","strongs_def":"En-Rimmon, a place in Palestine","kjv_def":"En-rimmon."},
"H5885":{"lemma":"עֵין שֶׁמֶשׁ","xlit":"ʻÊyn Shemesh","pron":"ane sheh'-mesh","derivation":"from H5869 (עַיִן) and H8121 (שֶׁמֶשׁ); fountain of the sun;","strongs_def":"En-Shemesh, a place in Palestine","kjv_def":"Enshemesh."},
"H5886":{"lemma":"עֵין תַּנִּים","xlit":"ʻÊyn Tannîym","pron":"ane tan-neem'","derivation":"from H5869 (עַיִן) and the plural of H8565 (תַּן); fountain of jackals;","strongs_def":"En-Tannim, a pool near Jerusalem","kjv_def":"dragon well."},
"H5887":{"lemma":"עֵין תַּפּוּחַ","xlit":"ʻÊyn Tappûwach","pron":"ane tap-poo'-akh","derivation":"from H5869 (עַיִן) and H8598 (תַּפּוּחַ); fountain of an apple-tree;","strongs_def":"En-Tappuach, a place in Palestine","kjv_def":"En-tappuah."},
"H5888":{"lemma":"עָיֵף","xlit":"ʻâyêph","pron":"aw-yafe'","derivation":"a primitive root;","strongs_def":"to languish","kjv_def":"be wearied."},
"H5889":{"lemma":"עָיֵף","xlit":"ʻâyêph","pron":"aw-yafe'","derivation":"from H5888 (עָיֵף);","strongs_def":"languid","kjv_def":"faint, thirsty, weary."},
"H5890":{"lemma":"עֵיפָה","xlit":"ʻêyphâh","pron":"ay-faw'","derivation":"feminine from H5774 (עוּף);","strongs_def":"obscurity (as if from covering)","kjv_def":"darkness."},
"H5891":{"lemma":"עֵיפָה","xlit":"ʻÊyphâh","pron":"ay-faw'","derivation":"the same as H5890 (עֵיפָה);","strongs_def":"Ephah, the name of a son of Midian, and of the region settled by him; also of an Israelite and of an Israelitess","kjv_def":"Ephah."},
"H5892":{"lemma":"עִיר","xlit":"ʻîyr","pron":"eer","derivation":"or (in the plural) עָר; or עָיַר; (Judges 10:4), from H5782 (עוּר)","strongs_def":"a city (a place guarded by waking or a watch) in the widest sense (even of a mere encampment or post)","kjv_def":"Ai (from margin), city, court (from margin), town."},
"H5893":{"lemma":"עִיר","xlit":"ʻÎyr","pron":"eer","derivation":"the same as H5892 (עִיר);","strongs_def":"Ir, an Israelite","kjv_def":"Ir."},
"H5894":{"lemma":"עִיר","xlit":"ʻîyr","pron":"eer","derivation":"(Aramaic) from a root corresponding to H5782 (עוּר);","strongs_def":"a watcher, i.e. an angel (as guardian)","kjv_def":"watcher."},
"H5895":{"lemma":"עַיִר","xlit":"ʻayir","pron":"ah'-yeer","derivation":"from H5782 (עוּר) in the sense of raising (i.e. bearing a burden);","strongs_def":"properly, a young ass (as just broken to a load); hence an ass-colt","kjv_def":"(ass) colt, foal, young ass."},
"H5896":{"lemma":"עִירָא","xlit":"ʻÎyrâʼ","pron":"ee-raw'","derivation":"from H5782 (עוּר); wakefulness;","strongs_def":"Ira, the name of three Israelites","kjv_def":"Ira."},
"H5897":{"lemma":"עִירָד","xlit":"ʻÎyrâd","pron":"ee-rawd'","derivation":"from the same as H6166 (עֲרָד); fugitive;","strongs_def":"Irad, an antediluvian","kjv_def":"Irad."},
"H5898":{"lemma":"עִיר הַמֶּלַח","xlit":"ʻÎyr ham-Melach","pron":"eer ham-meh'-lakh","derivation":"from H5892 (עִיר) and H4417 (מֶלַח) with the article of substance interp.; city of (the) salt;","strongs_def":"Irham-Melach, a place near Palestine","kjv_def":"the city of salt."},
"H5899":{"lemma":"עִיר הַתְּמָרִים","xlit":"ʻÎyr hat-Tᵉmârîym","pron":"eer hat-tem-aw-reem'","derivation":"from H5892 (עִיר) and the plural of H8558 (תָּמָר) with the article interpolated; city of the palmtrees;","strongs_def":"Ir-hat-Temarim, a place in Palestine","kjv_def":"the city of palmtrees."},
"H5900":{"lemma":"עִירוּ","xlit":"ʻÎyrûw","pron":"ee-roo'","derivation":"from H5892 (עִיר); a citizen;","strongs_def":"Iru, an Israelite","kjv_def":"Iru."},
"H5901":{"lemma":"עִירִי","xlit":"ʻÎyrîy","pron":"ee-ree'","derivation":"from H5892 (עִיר); urbane;","strongs_def":"Iri, an Israelite","kjv_def":"Iri."},
"H5902":{"lemma":"עִירָם","xlit":"ʻÎyrâm","pron":"ee-rawm'","derivation":"from H5892 (עִיר); city-wise;","strongs_def":"Iram, an Idumaean","kjv_def":"Iram."},
"H5903":{"lemma":"עֵירֹם","xlit":"ʻêyrôm","pron":"ay-rome'","derivation":"or עֵרֹם; from H6191 (עָרַם);","strongs_def":"nudity","kjv_def":"naked(-ness)."},
"H5904":{"lemma":"עִיר נָחָשׁ","xlit":"ʻÎyr Nâchâsh","pron":"eer naw-khawsh'","derivation":"from H5892 (עִיר) and H5175 (נָחָשׁ); city of a serpent;","strongs_def":"Ir-Nachash, a place in Palestine","kjv_def":"Irnahash."},
"H5905":{"lemma":"עִיר שֶׁמֶשׁ","xlit":"ʻÎyr Shemesh","pron":"eer sheh'-mesh","derivation":"from H5892 (עִיר) and H8121 (שֶׁמֶשׁ); city of the sun;","strongs_def":"Ir-Shemesh, a place in Palestine","kjv_def":"Irshemesh."},
"H5906":{"lemma":"עַיִשׁ","xlit":"ʻAyish","pron":"ah'-yish","derivation":"or עָשׁ; from H5789 (עוּשׁ);","strongs_def":"the constellation of the Great Bear (perhaps from its migration through the heavens)","kjv_def":"Arcturus."},
"H5907":{"lemma":"עַכְבּוֹר","xlit":"ʻAkbôwr","pron":"ak-bore'","derivation":"probably for H5909 (עַכְבָּר);","strongs_def":"Akbor, the name of an Idumaean and of two Israelites","kjv_def":"Achbor."},
"H5908":{"lemma":"עַכָּבִישׁ","xlit":"ʻakkâbîysh","pron":"ak-kaw-beesh'","derivation":"probably from an unused root in the literal sense of entangling;","strongs_def":"a spider (as weaving a network)","kjv_def":"spider."},
"H5909":{"lemma":"עַכְבָּר","xlit":"ʻakbâr","pron":"ak-bawr'","derivation":"probably from the same as H5908 (עַכָּבִישׁ) in the secondary sense of attacking;","strongs_def":"a mouse (as nibbling)","kjv_def":"mouse."},
"H5910":{"lemma":"עַכּוֹ","xlit":"ʻAkkôw","pron":"ak-ko'","derivation":"apparently from an unused root meaning to hem in;","strongs_def":"Akko (from its situation on a bay)","kjv_def":"Accho."},
"H5911":{"lemma":"עָכוֹר","xlit":"ʻÂkôwr","pron":"aw-kore'","derivation":"from H5916 (עָכַר); troubled;","strongs_def":"Akor, the name of a place in Palestine","kjv_def":"Achor."},
"H5912":{"lemma":"עָכָן","xlit":"ʻÂkân","pron":"aw-kawn'","derivation":"from an unused root meaning to trouble; troublesome;","strongs_def":"Akan, an Israelite","kjv_def":"Achan. Compare H5917 (עָכָר)."},
"H5913":{"lemma":"עָכַס","xlit":"ʻâkaç","pron":"aw-kas'","derivation":"a primitive root; properly, to tie, specifically, with fetters; but used only as denominative from H5914 (עֶכֶס);","strongs_def":"to put on anklets","kjv_def":"make a tinkling ornament."},
"H5914":{"lemma":"עֶכֶס","xlit":"ʻekeç","pron":"eh'-kes","derivation":"from H5913 (עָכַס);","strongs_def":"a fetter; hence, an anklet","kjv_def":"stocks, tinkling ornament."},
"H5915":{"lemma":"עַכְסָה","xlit":"ʻAkçâh","pron":"ak-saw'","derivation":"feminine of H5914 (עֶכֶס); anklet;","strongs_def":"Aksah, an Israelitess","kjv_def":"Achsah."},
"H5916":{"lemma":"עָכַר","xlit":"ʻâkar","pron":"aw-kar'","derivation":"a primitive root;","strongs_def":"properly, to roil water; figuratively, to disturb or affict","kjv_def":"trouble, stir."},
"H5917":{"lemma":"עָכָר","xlit":"ʻÂkâr","pron":"aw-kawr'","derivation":"from H5916 (עָכַר); troublesome;","strongs_def":"Akar, an Israelite","kjv_def":"Achar. Compare H5912 (עָכָן)."},
"H5918":{"lemma":"עׇכְרָן","xlit":"ʻOkrân","pron":"ok-rawn'","derivation":"from H5916 (עָכַר); muddler;","strongs_def":"Okran, an Israelite","kjv_def":"Ocran."},
"H5919":{"lemma":"עַכְשׁוּב","xlit":"ʻakshûwb","pron":"ak-shoob'","derivation":"probably from an unused root meaning to coil;","strongs_def":"an asp (from lurking coiled up)","kjv_def":"adder."},
"H5920":{"lemma":"עַל","xlit":"ʻal","pron":"al","derivation":"from H5927 (עָלָה);","strongs_def":"properly, the top; specifically, the highest (i.e. God); also (adverb) aloft, to Jehovah","kjv_def":"above, high, most High."},
"H5921":{"lemma":"עַל","xlit":"ʻal","pron":"al","derivation":"properly, the same as H5920 (עַל) used as a preposition (in the singular or plural often with prefix, or as conjunction with a particle following);","strongs_def":"above, over, upon, or against (yet always in this last relation with a downward aspect) in a great variety of applications","kjv_def":"above, according to(-ly), after, (as) against, among, and, [idiom] as, at, because of, beside (the rest of), between, beyond the time, [idiom] both and, by (reason of), [idiom] had the charge of, concerning for, in (that), (forth, out) of, (from) (off), (up-) on, over, than, through(-out), to, touching, [idiom] with."},
"H5922":{"lemma":"עַל","xlit":"ʻal","pron":"al","derivation":"(Aramaic) corresponding to H5921 (עַל)","strongs_def":"{above, over, upon, or against (yet always in this last relation with a downward aspect) in a great variety of applications}","kjv_def":"about, against, concerning, for, (there-) fore, from, in, [idiom] more, of, (there-, up-) on, (in-) to, [phrase] why with."},
"H5923":{"lemma":"עֹל","xlit":"ʻôl","pron":"ole","derivation":"or עוֹל; from H5953 (עָלַל);","strongs_def":"a yoke (as imposed on the neck), literally or figuratively","kjv_def":"yoke."},
"H5924":{"lemma":"עֵלָּא","xlit":"ʻêllâʼ","pron":"ale-law'","derivation":"(Aramaic) from H5922 (עַל);","strongs_def":"above","kjv_def":"over."},
"H5925":{"lemma":"עֻלָּא","xlit":"ʻUllâʼ","pron":"ool-law'","derivation":"feminine of H5923 (עֹל); burden;","strongs_def":"Ulla, an Israelite","kjv_def":"Ulla."},
"H5926":{"lemma":"עִלֵּג","xlit":"ʻillêg","pron":"il-layg'","derivation":"from an unused root meaning to stutter;","strongs_def":"stuttering","kjv_def":"stammerer."},
"H5927":{"lemma":"עָלָה","xlit":"ʻâlâh","pron":"aw-law'","derivation":"a primitive root;","strongs_def":"to ascend, intransitively (be high) or actively (mount); used in a great variety of senses, primary and secondary, literal and figurative","kjv_def":"arise (up), (cause to) ascend up, at once, break (the day) (up), bring (up), (cause to) burn, carry up, cast up, [phrase] shew, climb (up), (cause to, make to) come (up), cut off, dawn, depart, exalt, excel, fall, fetch up, get up, (make to) go (away, up); grow (over) increase, lay, leap, levy, lift (self) up, light, (make) up, [idiom] mention, mount up, offer, make to pay, [phrase] perfect, prefer, put (on), raise, recover, restore, (make to) rise (up), scale, set (up), shoot forth (up), (begin to) spring (up), stir up, take away (up), work."},
"H5928":{"lemma":"עֲלָה","xlit":"ʻălâh","pron":"al-law'","derivation":"(Aramaic) corresponding to H5930 (עֹלָה);","strongs_def":"a holocaust","kjv_def":"burnt offering."},
"H5929":{"lemma":"עָלֶה","xlit":"ʻâleh","pron":"aw-leh'","derivation":"from H5927 (עָלָה);","strongs_def":"a leaf (as coming up on a tree); collectively, foliage","kjv_def":"branch, leaf."},
"H5930":{"lemma":"עֹלָה","xlit":"ʻôlâh","pron":"o-law'","derivation":"or עוֹלָה; feminine active participle of H5927 (עָלָה);","strongs_def":"a step or (collectively, stairs, as ascending); usually a holocaust (as going up in smoke)","kjv_def":"ascent, burnt offering (sacrifice), go up to. See also H5766 (עֶוֶל)."},
"H5931":{"lemma":"עִלָּה","xlit":"ʻillâh","pron":"il-law'","derivation":"(Aramaic) feminine from a root corresponding to H5927 (עָלָה);","strongs_def":"a pretext (as arising artificially)","kjv_def":"occasion."},
"H5932":{"lemma":"עַלְוָה","xlit":"ʻalvâh","pron":"al-vaw'","derivation":"for H5766 (עֶוֶל);","strongs_def":"moral perverseness","kjv_def":"iniquity."},
"H5933":{"lemma":"עַלְוָה","xlit":"ʻAlvâh","pron":"al-vaw'","derivation":"or עַליָה; the same as H5932 (עַלְוָה);","strongs_def":"Alvah or Aljah, an Idumaean","kjv_def":"Aliah, Alvah."},
"H5934":{"lemma":"עָלוּם","xlit":"ʻâlûwm","pron":"aw-loom'","derivation":"passive participle of H5956 (עָלַם) in the denominative sense of H5958 (עֶלֶם);","strongs_def":"(only in plural as abstract) adolescence; figuratively, vigor","kjv_def":"youth."},
"H5935":{"lemma":"עַלְוָן","xlit":"ʻAlvân","pron":"al-vawn'","derivation":"or עַלְיָן; from H5927 (עָלָה); lofty;","strongs_def":"Alvan or Aljan, an Idumaean","kjv_def":"Alian, Alvan."},
"H5936":{"lemma":"עֲלוּקָה","xlit":"ʻălûwqâh","pron":"al-oo-kaw'","derivation":"feminine passive participle of an unused root meaning to suck;","strongs_def":"the leech","kjv_def":"horse-leech."},
"H5937":{"lemma":"עָלַז","xlit":"ʻâlaz","pron":"aw-laz'","derivation":"a primitive root;","strongs_def":"to jump for joy, i.e. exult","kjv_def":"be joyful, rejoice, triumph."},
"H5938":{"lemma":"עָלֵז","xlit":"ʻâlêz","pron":"aw-laze'","derivation":"from H5937 (עָלַז);","strongs_def":"exultant","kjv_def":"that rejoiceth."},
"H5939":{"lemma":"עֲלָטָה","xlit":"ʻălâṭâh","pron":"al-aw-taw'","derivation":"feminine from an unused root meaning to cover;","strongs_def":"dusk","kjv_def":"dark, twilight."},
"H5940":{"lemma":"עֱלִי","xlit":"ʻĕlîy","pron":"el-ee'","derivation":"from H5927 (עָלָה);","strongs_def":"a pestle (as lifted)","kjv_def":"pestle."},
"H5941":{"lemma":"עֵלִי","xlit":"ʻÊlîy","pron":"ay-lee'","derivation":"from H5927 (עָלָה); lofty;","strongs_def":"Eli, an Israelite highpriest","kjv_def":"Eli."},
"H5942":{"lemma":"עִלִּי","xlit":"ʻillîy","pron":"il-lee'","derivation":"from H5927 (עָלָה);","strongs_def":"high; i.e. comparative","kjv_def":"upper."},
"H5943":{"lemma":"עִלַּי","xlit":"ʻillay","pron":"il-lah'-ee","derivation":"(Aramaic) corresponding to H5942 (עִלִּי);","strongs_def":"supreme (i.e. God)","kjv_def":"(most) high."},
"H5944":{"lemma":"עֲלִיָּה","xlit":"ʻălîyâh","pron":"al-ee-yaw'","derivation":"feminine from H5927 (עָלָה);","strongs_def":"something lofty, i.e. a stair-way; also a second-story room (or even one on the roof); figuratively, the sky","kjv_def":"ascent, (upper) chamber, going up, loft, parlour."},
"H5945":{"lemma":"עֶלְיוֹן","xlit":"ʻelyôwn","pron":"el-yone'","derivation":"from H5927 (עָלָה);","strongs_def":"an elevation, i.e. (adj.) lofty (compar.); as title, the Supreme","kjv_def":"(Most, on) high(-er, -est), upper(-most)."},
"H5946":{"lemma":"עֶלְיוֹן","xlit":"ʻelyôwn","pron":"el-yone'","derivation":"(Aramaic) corresponding to H5945 (עֶלְיוֹן);","strongs_def":"the Supreme","kjv_def":"Most high."},
"H5947":{"lemma":"עַלִּיז","xlit":"ʻallîyz","pron":"al-leez'","derivation":"from H5937 (עָלַז);","strongs_def":"exultant","kjv_def":"joyous, (that) rejoice(-ing)"},
"H5948":{"lemma":"עֲלִיל","xlit":"ʻălîyl","pron":"al-eel'","derivation":"from H5953 (עָלַל) in the sense of completing;","strongs_def":"probably a crucible (as working over the metal)","kjv_def":"furnace."},
"H5949":{"lemma":"עֲלִילָה","xlit":"ʻălîylâh","pron":"al-ee-law'","derivation":"or עֲלִלָה; from H5953 (עָלַל) in the sense of effecting;","strongs_def":"an exploit (of God), or a performance (of man, often in a bad sense); by implication, an opportunity","kjv_def":"act(-ion), deed, doing, invention, occasion, work."},
"H5950":{"lemma":"עֲלִילִיָּה","xlit":"ʻălîylîyâh","pron":"al-ee-lee-yaw'","derivation":"for H5949 (עֲלִילָה);","strongs_def":"(miraculous) execution","kjv_def":"work."},
"H5951":{"lemma":"עֲלִיצוּת","xlit":"ʻălîytsûwth","pron":"al-ee-tsooth'","derivation":"from H5970 (עָלַץ);","strongs_def":"exultation","kjv_def":"rejoicing."},
"H5952":{"lemma":"עַלִּית","xlit":"ʻallîyth","pron":"al-leeth'","derivation":"from H5927 (עָלָה);","strongs_def":"a second-story room","kjv_def":"chamber. Compare H5944 (עֲלִיָּה)."},
"H5953":{"lemma":"עָלַל","xlit":"ʻâlal","pron":"aw-lal'","derivation":"a primitive root;","strongs_def":"to effect thoroughly; specifically, to glean (also figuratively); by implication (in a bad sense) to overdo, i.e. maltreat, be saucy to, pain, impose (also literal)","kjv_def":"abuse, affect, [idiom] child, defile, do, glean, mock, practise, thoroughly, work (wonderfully)."},
"H5954":{"lemma":"עֲלַל","xlit":"ʻălal","pron":"al-al'","derivation":"(Aramaic) corresponding to H5953 (עָלַל) (in the sense of thrusting oneself in),","strongs_def":"to enter; causatively, to introduce","kjv_def":"bring in, come in, go in."},
"H5955":{"lemma":"עֹלֵלָה","xlit":"ʻôlêlâh","pron":"o-lay-law'","derivation":"feminine active participle of H5953 (עָלַל);","strongs_def":"only in plural gleanings; by extension gleaning-time","kjv_def":"(gleaning) (of the) grapes, grapegleanings."},
"H5956":{"lemma":"עָלַם","xlit":"ʻâlam","pron":"aw-lam'","derivation":"a primitive root;","strongs_def":"to veil from sight, i.e. conceal (literally or figuratively)","kjv_def":"[idiom] any ways, blind, dissembler, hide (self), secret (thing)."},
"H5957":{"lemma":"עָלַם","xlit":"ʻâlam","pron":"aw-lam'","derivation":"(Aramaic) corresponding to H5769 (עוֹלָם);","strongs_def":"remote time, i.e. the future or past indefinitely; often adverb, forever","kjv_def":"for (n-)ever (lasting), old."},
"H5958":{"lemma":"עֶלֶם","xlit":"ʻelem","pron":"eh'-lem","derivation":"from H5956 (עָלַם); (compare H5959 (עַלְמָה))","strongs_def":"properly, something kept out of sight, i.e. a lad","kjv_def":"young man, stripling."},
"H5959":{"lemma":"עַלְמָה","xlit":"ʻalmâh","pron":"al-maw'","derivation":"feminine of H5958 (עֶלֶם);","strongs_def":"a lass (as veiled or private)","kjv_def":"damsel, maid, virgin."},
"H5960":{"lemma":"עַלְמוֹן","xlit":"ʻAlmôwn","pron":"al-mone'","derivation":"from H5956 (עָלַם); hidden; See also H5963 (עַלְמֹן דִּבְלָתָיְמָה).","strongs_def":"Almon, a place in Palestine"},
"H5961":{"lemma":"עֲלָמוֹת","xlit":"ʻĂlâmôwth","pron":"al-aw-moth'","derivation":"plural of H5959 (עַלְמָה);","strongs_def":"properly, girls, i.e. the soprano or female voice, perhaps falsetto","kjv_def":"Alamoth."},
"H5962":{"lemma":"עַלְמִי","xlit":"ʻAlmîy","pron":"al-mee'","derivation":"(Aramaic) patrial from a name corresponding to H5867 (עֵילָם) contracted;","strongs_def":"an Elamite or inhabitant of Elam","kjv_def":"Elamite."},
"H5963":{"lemma":"עַלְמֹן דִּבְלָתָיְמָה","xlit":"ʻAlmôn Diblâthâyᵉmâh","pron":"al-mone' dib-lawthaw'-yem-aw","derivation":"from the same as H5960 (עַלְמוֹן) and the dual of H1960 (הֻיְּדָה) (compare H1015 (בֵּית דִּבְלָתַיִם)) with enclitic of direction; Almon towards Diblathajim;","strongs_def":"Almon-Diblathajemah, a place in Moab","kjv_def":"Almondilathaim."},
"H5964":{"lemma":"עָלֶמֶת","xlit":"ʻÂlemeth","pron":"aw-leh'-meth","derivation":"from H5956 (עָלַם); a covering;","strongs_def":"Alemeth, the name of a place in Palestine and of two Israelites","kjv_def":"Alameth, Alemeth."},
"H5965":{"lemma":"עָלַס","xlit":"ʻâlaç","pron":"aw-las'","derivation":"a primitive root;","strongs_def":"to leap for joy, i.e. exult, wave joyously","kjv_def":"[idiom] peacock, rejoice, solace self."},
"H5966":{"lemma":"עָלַע","xlit":"ʻâlaʻ","pron":"aw-lah'","derivation":"a primitive root;","strongs_def":"to sip up","kjv_def":"suck up."},
"H5967":{"lemma":"עֲלַע","xlit":"ʻălaʻ","pron":"al-ah'","derivation":"(Aramaic) corresponding to H6763 (צֵלָע);","strongs_def":"a rib","kjv_def":"rib."},
"H5968":{"lemma":"עָלַף","xlit":"ʻâlaph","pron":"aw-laf'","derivation":"a primitive root;","strongs_def":"to veil or cover; figuratively, to be languid","kjv_def":"faint, overlaid, wrap self."},
"H5969":{"lemma":"עֻלְפֶּה","xlit":"ʻulpeh","pron":"ool-peh'","derivation":"from H5968 (עָלַף);","strongs_def":"an envelope, i.e. (figuratively) mourning","kjv_def":"fainted."},
"H5970":{"lemma":"עָלַץ","xlit":"ʻâlats","pron":"aw-lats'","derivation":"a primitive root;","strongs_def":"to jump for joy, i.e. exult","kjv_def":"be joyful, rejoice, triumph."},
"H5971":{"lemma":"עַם","xlit":"ʻam","pron":"am","derivation":"from H6004 (עָמַם);","strongs_def":"a people (as a congregated unit); specifically, a tribe (as those of Israel); hence (collectively) troops or attendants; figuratively, a flock","kjv_def":"folk, men, nation, people."},
"H5972":{"lemma":"עַם","xlit":"ʻam","pron":"am","derivation":"(Aramaic) corresponding to H5971 (עַם)","strongs_def":"{a people (as a congregated unit); specifically, a tribe (as those of Israel); hence (collectively) troops or attendants; figuratively, a flock}","kjv_def":"people."},
"H5973":{"lemma":"עִם","xlit":"ʻim","pron":"eem","derivation":"from H6004 (עָמַם);","strongs_def":"adverb or preposition, with (i.e. in conjunction with), in varied applications; specifically, equally with; often with prepositional prefix (and then usually unrepresented in English)","kjv_def":"accompanying, against, and, as ([idiom] long as), before, beside, by (reason of), for all, from (among, between), in, like, more than, of, (un-) to, with(-al)."},
"H5974":{"lemma":"עִם","xlit":"ʻim","pron":"eem","derivation":"(Aramaic) corresponding to H5973 (עִם)","strongs_def":"{adverb or preposition, with (i.e. in conjunction with), in varied applications; specifically, equally with; often with prepositional prefix (and then usually unrepresented in English)}","kjv_def":"by, from, like, to(-ward), with."},
"H5975":{"lemma":"עָמַד","xlit":"ʻâmad","pron":"aw-mad'","derivation":"a primitive root;","strongs_def":"to stand, in various relations (literal and figurative, intransitive and transitive)","kjv_def":"abide (behind), appoint, arise, cease, confirm, continue, dwell, be employed, endure, establish, leave, make, ordain, be (over), place, (be) present (self), raise up, remain, repair, [phrase] serve, set (forth, over, -tle, up), (make to, make to be at a, with-) stand (by, fast, firm, still, up), (be at a) stay (up), tarry."},
"H5976":{"lemma":"עָמַד","xlit":"ʻâmad","pron":"aw-mad'","derivation":"for H4571 (מָעַד);","strongs_def":"to shake","kjv_def":"be at a stand."},
"H5977":{"lemma":"עֹמֶד","xlit":"ʻômed","pron":"o'-med","derivation":"from H5975 (עָמַד);","strongs_def":"a spot (as being fixed)","kjv_def":"place, ([phrase] where) stood, upright."},
"H5978":{"lemma":"עִמָּד","xlit":"ʻimmâd","pron":"im-mawd'","derivation":"prolonged for H5973 (עִם);","strongs_def":"along with","kjv_def":"against, by, from, [phrase] me, [phrase] mine, of, [phrase] that I take, unto, upon, with(-in.)"},
"H5979":{"lemma":"עֶמְדָּה","xlit":"ʻemdâh","pron":"em-daw'","derivation":"from H5975 (עָמַד);","strongs_def":"a station, i.e. domicile","kjv_def":"standing."},
"H5980":{"lemma":"עֻמָּה","xlit":"ʻummâh","pron":"oom-maw'","derivation":"from H6004 (עָמַם);","strongs_def":"conjunction, i.e. society; mostly adverb or preposition (with prepositional prefix), near, beside, along with","kjv_def":"(over) against, at, beside, hard by, in points."},
"H5981":{"lemma":"עֻמָּה","xlit":"ʻUmmâh","pron":"oom-maw'","derivation":"the same as H5980 (עֻמָּה); association;","strongs_def":"Ummah, a place in Palestine","kjv_def":"Ummah."},
"H5982":{"lemma":"עַמּוּד","xlit":"ʻammûwd","pron":"am-mood'","derivation":"or עַמֻּד; from H5975 (עָמַד);","strongs_def":"a column (as standing); also a stand, i.e. platform","kjv_def":"[idiom] apiece, pillar."},
"H5983":{"lemma":"עַמּוֹן","xlit":"ʻAmmôwn","pron":"am-mone'","derivation":"from H5971 (עַם); tribal, i.e. inbred;","strongs_def":"Ammon, a son of Lot; also his posterity and their country","kjv_def":"Ammon, Ammonites."},
"H5984":{"lemma":"עַמּוֹנִי","xlit":"ʻAmmôwnîy","pron":"am-mo-nee'","derivation":"patronymically from H5983 (עַמּוֹן);","strongs_def":"an Ammonite or (the adjective) Ammonitish","kjv_def":"Ammonite(-s)."},
"H5985":{"lemma":"עַמּוֹנִית","xlit":"ʻAmmôwnîyth","pron":"am-mo-neeth'","derivation":"feminine of H5984 (עַמּוֹנִי);","strongs_def":"an Ammonitess","kjv_def":"Ammonite(-ss)."},
"H5986":{"lemma":"עָמוֹס","xlit":"ʻÂmôwç","pron":"aw-moce'","derivation":"from H6006 (עָמַס); burdensome;","strongs_def":"Amos, an Israelite prophet","kjv_def":"Amos."},
"H5987":{"lemma":"עָמוֹק","xlit":"ʻÂmôwq","pron":"aw-moke'","derivation":"from H6009 (עָמַק); deep;","strongs_def":"Amok, an Israelite","kjv_def":"Amok."},
"H5988":{"lemma":"עַמִּיאֵל","xlit":"ʻAmmîyʼêl","pron":"am-mee-ale'","derivation":"from H5971 (עַם) and H410 (אֵל); people of God;","strongs_def":"Ammiel, the name of three or four Israelites","kjv_def":"Ammiel."},
"H5989":{"lemma":"עַמִּיהוּד","xlit":"ʻAmmîyhûwd","pron":"am-mee-hood'","derivation":"from H5971 (עַם) and H1935 (הוֹד); people of splendor;","strongs_def":"Ammihud, the name of three Israelites","kjv_def":"Ammihud."},
"H5990":{"lemma":"עַמִּיזָבָד","xlit":"ʻAmmîyzâbâd","pron":"am-mee-zaw-bawd'","derivation":"from H5971 (עַם) and H2064 (זָבַד); people of endowment;","strongs_def":"Ammizabad, an Israelite","kjv_def":"Ammizabad."},
"H5991":{"lemma":"עַמִּיחוּר","xlit":"ʻAmmîychûwr","pron":"am-mee-khoor'","derivation":"from H5971 (עַם) and H2353 (חוּר); people of nobility;","strongs_def":"Ammichur, a Syrian prince","kjv_def":"Ammihud (from the margin)."},
"H5992":{"lemma":"עַמִּינָדָב","xlit":"ʻAmmîynâdâb","pron":"am-mee-naw-dawb'","derivation":"from H5971 (עַם) and H5068 (נָדַב); people of liberality;","strongs_def":"Amminadab, the name of four Israelites","kjv_def":"Amminadab."},
"H5993":{"lemma":"עַמִּי נָדִיב","xlit":"ʻAmmîy Nâdîyb","pron":"am-mee' naw-deeb'","derivation":"from H5971 (עַם) and H5081 (נָדִיב); my people (is) liberal;","strongs_def":"Ammi-Nadib, probably an Israelite","kjv_def":"Amminadib."},
"H5994":{"lemma":"עֲמִיק","xlit":"ʻămîyq","pron":"am-eek'","derivation":"(Aramaic) corresponding to H6012 (עָמֵק);","strongs_def":"profound, i.e. unsearchable","kjv_def":"deep."},
"H5995":{"lemma":"עָמִיר","xlit":"ʻâmîyr","pron":"aw-meer'","derivation":"from H6014 (עָמַר);","strongs_def":"a bunch of grain","kjv_def":"handful, sheaf."},
"H5996":{"lemma":"עַמִּישַׁדַּי","xlit":"ʻAmmîyshadday","pron":"am-mee-shad-dah'ee","derivation":"from H5971 (עַם) and H7706 (שַׁדַּי); people of (the) Almighty;","strongs_def":"Ammishaddai, an Israelite","kjv_def":"Ammishaddai."},
"H5997":{"lemma":"עָמִית","xlit":"ʻâmîyth","pron":"aw-meeth'","derivation":"from a primitive root meaning to associate;","strongs_def":"companionship; hence (concretely) a comrade or kindred man","kjv_def":"another, fellow, neighbour."},
"H5998":{"lemma":"עָמַל","xlit":"ʻâmal","pron":"aw-mal'","derivation":"a primitive root;","strongs_def":"to toil, i.e. work severely and with irksomeness","kjv_def":"(take) labour (in)."},
"H5999":{"lemma":"עָמָל","xlit":"ʻâmâl","pron":"aw-mawl'","derivation":"from H5998 (עָמַל);","strongs_def":"toil, i.e. wearing effort; hence, worry, wheth. of body or mind","kjv_def":"grievance(-vousness), iniquity, labour, mischief, miserable(-sery), pain(-ful), perverseness, sorrow, toil, travail, trouble, wearisome, wickedness."},
"H6000":{"lemma":"עָמָל","xlit":"ʻÂmâl","pron":"aw-mawl'","derivation":"the same as H5999 (עָמָל);","strongs_def":"Amal, an Israelite","kjv_def":"Amal."},
"H6001":{"lemma":"עָמֵל","xlit":"ʻâmêl","pron":"aw-male'","derivation":"from H5998 (עָמַל);","strongs_def":"toiling; concretely, a laborer; figuratively, sorrowful","kjv_def":"that laboureth, that is a misery, had taken (labour), wicked, workman."},
"H6002":{"lemma":"עֲמָלֵק","xlit":"ʻĂmâlêq","pron":"am-aw-lake'","derivation":"probably of foreign origin;","strongs_def":"Amalek, a descendant of Esau; also his posterity and their country","kjv_def":"Amalek."},
"H6003":{"lemma":"עֲמָלֵקִי","xlit":"ʻĂmâlêqîy","pron":"am-aw-lay-kee'","derivation":"patronymically from H6002 (עֲמָלֵק);","strongs_def":"an Amalekite (or collectively the Amalekites) or descendants of Amalek","kjv_def":"Amalekite(-s)."},
"H6004":{"lemma":"עָמַם","xlit":"ʻâmam","pron":"aw-mam'","derivation":"a primitive root;","strongs_def":"to associate; by implication, to overshadow (by huddling together)","kjv_def":"become dim, hide."},
"H6005":{"lemma":"עִמָּנוּאֵל","xlit":"ʻImmânûwʼêl","pron":"im-maw-noo-ale'","derivation":"from H5973 (עִם) and with a pronominal suffix inserted; with us (is) God;","strongs_def":"Immanuel, a type name of Isaiah's son","kjv_def":"Immanuel."},
"H6006":{"lemma":"עָמַס","xlit":"ʻâmaç","pron":"aw-mas'","derivation":"or עָמַשׂ; a primitive root;","strongs_def":"to load, i.e. impose aburden (or figuratively, infliction)","kjv_def":"be borne, (heavy) burden (self), lade, load, put."},
"H6007":{"lemma":"עֲמַסְיָה","xlit":"ʻĂmaçyâh","pron":"am-as-yaw'","derivation":"from H6006 (עָמַס) and H3050 (יָהּ); Jah has loaded;","strongs_def":"Amasjah, an Israelite","kjv_def":"Amasiah."},
"H6008":{"lemma":"עַמְעָד","xlit":"ʻAmʻâd","pron":"am-awd'","derivation":"from H5971 (עַם) and H5703 (עַד); people of time;","strongs_def":"Amad, a place in Palestine","kjv_def":"Amad."},
"H6009":{"lemma":"עָמַק","xlit":"ʻâmaq","pron":"aw-mak'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) deep (literally or figuratively)","kjv_def":"(be, have, make, seek) deep(-ly), depth, be profound."},
"H6010":{"lemma":"עֵמֶק","xlit":"ʻêmeq","pron":"ay'-mek","derivation":"from H6009 (עָמַק);","strongs_def":"a vale (i.e. broad depression)","kjv_def":"dale, vale, valley (often used as a part of proper names). See also H1025 (בֵּית הָעֵמֶק)."},
"H6011":{"lemma":"עֹמֶק","xlit":"ʻômeq","pron":"o'-mek","derivation":"from H6009 (עָמַק);","strongs_def":"depth","kjv_def":"depth."},
"H6012":{"lemma":"עָמֵק","xlit":"ʻâmêq","pron":"aw-make'","derivation":"from H6009 (עָמַק);","strongs_def":"deep (literally or figuratively)","kjv_def":"deeper, depth, strange."},
"H6013":{"lemma":"עָמֹק","xlit":"ʻâmôq","pron":"aw-moke'","derivation":"from H6009 (עָמַק);","strongs_def":"deep (literally or figuratively)","kjv_def":"([idiom] exceeding) deep (thing)."},
"H6014":{"lemma":"עָמַר","xlit":"ʻâmar","pron":"aw-mar'","derivation":"a primitive root; properly, apparently to heap; figuratively, to chastise (as if piling blows); specifically (as denominative from H6016 (עֹמֶר))","strongs_def":"to gather grain","kjv_def":"bind sheaves, make merchandise of."},
"H6015":{"lemma":"עֲמַר","xlit":"ʻămar","pron":"am-ar'","derivation":"(Aramaic) corresponding to H6785 (צֶמֶר);","strongs_def":"wool","kjv_def":"wool."},
"H6016":{"lemma":"עֹמֶר","xlit":"ʻômer","pron":"o'-mer","derivation":"from H6014 (עָמַר);","strongs_def":"properly, a heap, i.e. a sheaf; also an omer, as a dry measure","kjv_def":"omer, sheaf."},
"H6017":{"lemma":"עֲמֹרָה","xlit":"ʻĂmôrâh","pron":"am-o-raw'","derivation":"from H6014 (עָמַר); a (ruined) heap;","strongs_def":"Amorah, a place in Palestine","kjv_def":"Gomorrah."},
"H6018":{"lemma":"עׇמְרִי","xlit":"ʻOmrîy","pron":"om-ree'","derivation":"from H6014 (עָמַר); heaping;","strongs_def":"Omri, an Israelite","kjv_def":"Omri."},
"H6019":{"lemma":"עַמְרָם","xlit":"ʻAmrâm","pron":"am-rawm'","derivation":"probably from H5971 (עַם) and H7311 (רוּם); high people;","strongs_def":"Amram, the name of two Israelites","kjv_def":"Amram."},
"H6020":{"lemma":"עַמְרָמִי","xlit":"ʻAmrâmîy","pron":"am-raw-mee'","derivation":"from H6019 (עַמְרָם);","strongs_def":"an Amramite or descendant of Amram","kjv_def":"Amramite."},
"H6021":{"lemma":"עֲמָשָׂא","xlit":"ʻĂmâsâʼ","pron":"am-aw-saw'","derivation":"from H6006 (עָמַס); burden;","strongs_def":"Amasa, the name of two Israelites","kjv_def":"Amasa."},
"H6022":{"lemma":"עֲמָשַׂי","xlit":"ʻĂmâsay","pron":"am-aw-sah'-ee","derivation":"from H6006 (עָמַס); burdensome;","strongs_def":"Amasai, the name of three Israelites","kjv_def":"Amasai."},
"H6023":{"lemma":"עֲמַשְׁסַי","xlit":"ʻĂmashçay","pron":"am-ash-sah'-ee","derivation":"probably from H6006 (עָמַס); burdensome;","strongs_def":"Amashsay, an Israelite","kjv_def":"Amashai."},
"H6024":{"lemma":"עֲנָב","xlit":"ʻĂnâb","pron":"an-awb'","derivation":"from the same as H6025 (עֵנָב); fruit;","strongs_def":"Anab, a place in Palestine","kjv_def":"Anab."},
"H6025":{"lemma":"עֵנָב","xlit":"ʻênâb","pron":"ay-nawb'","derivation":"from an unused root probably meaning to bear fruit;","strongs_def":"a grape","kjv_def":"(ripe) grape, wine."},
"H6026":{"lemma":"עָנַג","xlit":"ʻânag","pron":"aw-nag'","derivation":"a primitive root;","strongs_def":"to be soft or pliable, i.e. (figuratively) effeminate or luxurious","kjv_def":"delicate(-ness), (have) delight (self), sport self."},
"H6027":{"lemma":"עֹנֶג","xlit":"ʻôneg","pron":"o'-neg","derivation":"from H6026 (עָנַג);","strongs_def":"luxury","kjv_def":"delight, pleasant."},
"H6028":{"lemma":"עָנֹג","xlit":"ʻânôg","pron":"aw-nogue'","derivation":"from H6026 (עָנַג);","strongs_def":"luxurious","kjv_def":"delicate."},
"H6029":{"lemma":"עָנַד","xlit":"ʻânad","pron":"aw-nad'","derivation":"a primitive root;","strongs_def":"to lace fast","kjv_def":"bind, tie."},
"H6030":{"lemma":"עָנָה","xlit":"ʻânâh","pron":"aw-naw'","derivation":"a primitive root;","strongs_def":"properly, to eye or (generally) to heed, i.e. pay attention; by implication, to respond; by extension to begin to speak; specifically to sing, shout, testify, announce","kjv_def":"give account, afflict (by mistake for H6031 (עָנָה)), (cause to, give) answer, bring low (by mistake for H6031 (עָנָה)), cry, hear, Leannoth, lift up, say, [idiom] scholar, (give a) shout, sing (together by course), speak, testify, utter, (bear) witness. See also H1042 (בֵּית עֲנוֹת), H1043 (בֵּית עֲנָת)."},
"H6031":{"lemma":"עָנָה","xlit":"ʻânâh","pron":"aw-naw'","derivation":"a primitive root (possibly rather identical with H6030 (עָנָה) through the idea of looking down or browbeating);","strongs_def":"to depress literally or figuratively, transitive or intransitive (in various applications, as follows)","kjv_def":"abase self, afflict(-ion, self), answer (by mistake for H6030 (עָנָה)), chasten self, deal hardly with, defile, exercise, force, gentleness, humble (self), hurt, ravish, sing (by mistake for H6030 (עָנָה)), speak (by mistake for H6030 (עָנָה)), submit self, weaken, [idiom] in any wise."},
"H6032":{"lemma":"עֲנָה","xlit":"ʻănâh","pron":"an-aw'","derivation":"(Aramaic) corresponding to H6030 (עָנָה)","strongs_def":"{properly, to eye or (generally) to heed, i.e. pay attention; by implication, to respond; by extension to begin to speak; specifically to sing, shout, testify, announce}","kjv_def":"answer, speak."},
"H6033":{"lemma":"עֲנָה","xlit":"ʻănâh","pron":"an-aw'","derivation":"(Aramaic) corresponding to H6031 (עָנָה)","strongs_def":"{to depress literally or figuratively, transitive or intransitive (in various applications, as follows)}","kjv_def":"poor."},
"H6034":{"lemma":"עֲנָה","xlit":"ʻĂnâh","pron":"an-aw'","derivation":"probably from H6030 (עָנָה); an answer;","strongs_def":"Anah, the name of two Edomites and one Edomitess","kjv_def":"Anah."},
"H6035":{"lemma":"עָנָו","xlit":"ʻânâv","pron":"aw-nawv'","derivation":"or (by intermixture with H6041 (עָנִי)) עָנָיו; from H6031 (עָנָה);","strongs_def":"depressed (figuratively), in mind (gentle) or circumstances (needy, especially saintly)","kjv_def":"humble, lowly, meek, poor. Compare H6041 (עָנִי)."},
"H6036":{"lemma":"עָנוּב","xlit":"ʻÂnûwb","pron":"aw-noob'","derivation":"passive participle from the same as H6025 (עֵנָב); borne (as fruit);","strongs_def":"Anub, an Israelite","kjv_def":"Anub."},
"H6037":{"lemma":"עַנְוָה","xlit":"ʻanvâh","pron":"an-vaw'","derivation":"feminine of H6035 (עָנָו);","strongs_def":"mildness (royal); also (concretely) oppressed","kjv_def":"gentleness, meekness."},
"H6038":{"lemma":"עֲנָוָה","xlit":"ʻănâvâh","pron":"an-aw-vaw'","derivation":"from H6035 (עָנָו);","strongs_def":"condescension, human and subjective (modesty), or divine and objective (clemency)","kjv_def":"gentleness, humility, meekness."},
"H6039":{"lemma":"עֱנוּת","xlit":"ʻĕnûwth","pron":"en-ooth'","derivation":"from H6031 (עָנָה);","strongs_def":"affliction","kjv_def":"affliction."},
"H6040":{"lemma":"עֳנִי","xlit":"ʻŏnîy","pron":"on-ee'","derivation":"from H6031 (עָנָה);","strongs_def":"depression, i.e. misery","kjv_def":"afflicted(-ion), trouble."},
"H6041":{"lemma":"עָנִי","xlit":"ʻânîy","pron":"aw-nee'","derivation":"from H6031 (עָנָה); (practically the same as H6035 (עָנָו), although the margin constantly disputes this, making H6035 (עָנָו) subjective and objective);","strongs_def":"depressed, in mind or circumstances","kjv_def":"afflicted, humble, lowly, needy, poor."},
"H6042":{"lemma":"עֻנִּי","xlit":"ʻUnnîy","pron":"oon-nee'","derivation":"from H6031 (עָנָה); afflicted;","strongs_def":"Unni, the name of two Israelites","kjv_def":"Unni."},
"H6043":{"lemma":"עֲנָיָה","xlit":"ʻĂnâyâh","pron":"an-aw-yaw'","derivation":"from H6030 (עָנָה); Jah has answered;","strongs_def":"Anajah, the name of two Israelites","kjv_def":"Anaiah."},
"H6044":{"lemma":"עָנִים","xlit":"ʻÂnîym","pron":"aw-neem'","derivation":"for plural of H5869 (עַיִן); fountains;","strongs_def":"Anim, a place in Palestine","kjv_def":"Anim."},
"H6045":{"lemma":"עִנְיָן","xlit":"ʻinyân","pron":"in-yawn'","derivation":"from H6031 (עָנָה);","strongs_def":"ado, i.e. (generally) employment or (specifically) an affair","kjv_def":"business, travail."},
"H6046":{"lemma":"עָנֵם","xlit":"ʻÂnêm","pron":"aw-name'","derivation":"from the dual of H5869 (עַיִן); two fountains;","strongs_def":"Anem, a place in Palestine","kjv_def":"Anem."},
"H6047":{"lemma":"עֲנָמִים","xlit":"ʻĂnâmîym","pron":"an-aw-meem'","derivation":"as if plural of some Egyptian word;","strongs_def":"Anamim, a son of Mizraim and his descendants, with their country","kjv_def":"Anamim."},
"H6048":{"lemma":"עֲנַמֶּלֶךְ","xlit":"ʻĂnammelek","pron":"an-am-meh'-lek","derivation":"of foreign origin;","strongs_def":"Anammelek, an Assyrian deity","kjv_def":"Anammelech."},
"H6049":{"lemma":"עָנַן","xlit":"ʻânan","pron":"aw-nan'","derivation":"a primitive root; to cover; used only as a denominative from H6051 (עָנָן), to cloud over;","strongs_def":"figuratively, to act covertly, i.e. practise magic","kjv_def":"[idiom] bring, enchanter, Meonemin, observe(-r of) times, soothsayer, sorcerer."},
"H6050":{"lemma":"עֲנַן","xlit":"ʻănan","pron":"an-an'","derivation":"(Aramaic) corresponding to H6051 (עָנָן)","strongs_def":"{a cloud (as covering the sky), i.e. the nimbus or thunder-cloud}","kjv_def":"cloud."},
"H6051":{"lemma":"עָנָן","xlit":"ʻânân","pron":"aw-nawn'","derivation":"from H6049 (עָנַן);","strongs_def":"a cloud (as covering the sky), i.e. the nimbus or thunder-cloud","kjv_def":"cloud(-y)."},
"H6052":{"lemma":"עָנָן","xlit":"ʻÂnân","pron":"aw-nawn'","derivation":"the same as H6051 (עָנָן); cloud;","strongs_def":"Anan, an Israelite","kjv_def":"Anan."},
"H6053":{"lemma":"עֲנָנָה","xlit":"ʻănânâh","pron":"an-aw-naw'","derivation":"feminine of H6051 (עָנָן);","strongs_def":"cloudiness","kjv_def":"cloud."},
"H6054":{"lemma":"עֲנָנִי","xlit":"ʻĂnânîy","pron":"an-aw-nee'","derivation":"from H6051 (עָנָן); cloudy;","strongs_def":"Anani, an Israelite","kjv_def":"Anani."},
"H6055":{"lemma":"עֲנַנְיָה","xlit":"ʻĂnanyâh","pron":"an-an-yaw'","derivation":"from H6049 (עָנַן) and H3050 (יָהּ); Jah has covered;","strongs_def":"Ananjah, the name of an Israelite and of a place in Palestine","kjv_def":"Ananiah."},
"H6056":{"lemma":"עֲנַף","xlit":"ʻănaph","pron":"an-af'","derivation":"(Aramaic) or עֶנֶף; (Aramaic), corresponding to H6057 (עָנָף)","strongs_def":"{a twig (as covering the limbs)}","kjv_def":"bough, branch."},
"H6057":{"lemma":"עָנָף","xlit":"ʻânâph","pron":"aw-nawf'","derivation":"from an unused root meaning to cover;","strongs_def":"a twig (as covering the limbs)","kjv_def":"bough, branch."},
"H6058":{"lemma":"עָנֵף","xlit":"ʻânêph","pron":"aw-nafe'","derivation":"from the same as H6057 (עָנָף);","strongs_def":"branching","kjv_def":"full of branches."},
"H6059":{"lemma":"עָנַק","xlit":"ʻânaq","pron":"aw-nak'","derivation":"a primitive root; properly, to choke; used only as denominative from H6060 (עָנָק),","strongs_def":"to collar, i.e. adorn with anecklace; figuratively, to fit out with supplies","kjv_def":"compass about as a chain, furnish, liberally."},
"H6060":{"lemma":"עָנָק","xlit":"ʻânâq","pron":"aw-nawk'","derivation":"from H6059 (עָנַק);","strongs_def":"a necklace (as if strangling)","kjv_def":"chain."},
"H6061":{"lemma":"עָנָק","xlit":"ʻÂnâq","pron":"aw-nawk'","derivation":"the same as H6060 (עָנָק);","strongs_def":"Anak, a Canaanite","kjv_def":"Anak."},
"H6062":{"lemma":"עֲנָקִי","xlit":"ʻĂnâqîy","pron":"an-aw-kee'","derivation":"patronymically from H6061 (עָנָק);","strongs_def":"an Anakite or descendant of Anak","kjv_def":"Anakim."},
"H6063":{"lemma":"עָנֵר","xlit":"ʻÂnêr","pron":"aw-nare'","derivation":"probably for H5288 (נַעַר);","strongs_def":"Aner, a Amorite, also a place in Palestine","kjv_def":"Aner."},
"H6064":{"lemma":"עָנַשׁ","xlit":"ʻânash","pron":"aw-nash'","derivation":"a primitive root;","strongs_def":"properly, to urge; by implication, to inflict apenalty, specifically, to fine","kjv_def":"amerce, condemn, punish, [idiom] surely."},
"H6065":{"lemma":"עֲנַשׁ","xlit":"ʻănash","pron":"an-ash'","derivation":"(Aramaic) corresponding to H6066 (עֹנֶשׁ);","strongs_def":"a mulct","kjv_def":"confiscation."},
"H6066":{"lemma":"עֹנֶשׁ","xlit":"ʻônesh","pron":"o'-nesh","derivation":"from H6064 (עָנַשׁ);","strongs_def":"a fine","kjv_def":"punishment, tribute."},
"H6067":{"lemma":"עֲנָת","xlit":"ʻĂnâth","pron":"an-awth'","derivation":"from H6030 (עָנָה); answer;","strongs_def":"Anath, an Israelite","kjv_def":"Anath."},
"H6068":{"lemma":"עֲנָתוֹת","xlit":"ʻĂnâthôwth","pron":"an-aw-thoth'","derivation":"plural of H6067 (עֲנָת);","strongs_def":"Anathoth, the name of two Israelites, also of a place in Pal","kjv_def":"Anathoth."},
"H6069":{"lemma":"עַנְתֹתִי","xlit":"ʻAnthôthîy","pron":"an-tho-thee'","derivation":"or עַנְּתוֹתִי; patrial from H6068 (עֲנָתוֹת);","strongs_def":"a Antothite or inhabitant of Anathoth","kjv_def":"of Anathoth, Anethothite, Anetothite, Antothite."},
"H6070":{"lemma":"עַנְתֹתִיָּה","xlit":"ʻAnthôthîyâh","pron":"an-tho-thee-yaw'","derivation":"from the same as H6068 (עֲנָתוֹת) and H3050 (יָהּ); answers of Jah;","strongs_def":"Anthothijah, an Israelite","kjv_def":"Antothijah."},
"H6071":{"lemma":"עָסִיס","xlit":"ʻâçîyç","pron":"aw-sees'","derivation":"from H6072 (עָסַס);","strongs_def":"must or fresh grape-juice (as just trodden out)","kjv_def":"juice, new (sweet) wine."},
"H6072":{"lemma":"עָסַס","xlit":"ʻâçaç","pron":"aw-sas'","derivation":"a primitive root;","strongs_def":"to squeeze out juice; figuratively, to trample","kjv_def":"tread down."},
"H6073":{"lemma":"עֳפֶא","xlit":"ʻŏpheʼ","pron":"of-eh'","derivation":"from an unused root meaning to cover;","strongs_def":"a bough (as covering the tree)","kjv_def":"branch."},
"H6074":{"lemma":"עֳפִי","xlit":"ʻŏphîy","pron":"of-ee'","derivation":"(Aramaic) corresponding to H6073 (עֳפֶא);","strongs_def":"a twig; bough, i.e. (collectively) foliage","kjv_def":"leaves."},
"H6075":{"lemma":"עָפַל","xlit":"ʻâphal","pron":"aw-fal'","derivation":"a primitive root;","strongs_def":"to swell; figuratively, be elated","kjv_def":"be lifted up, presume."},
"H6076":{"lemma":"עֹפֶל","xlit":"ʻôphel","pron":"o'-fel","derivation":"from H6075 (עָפַל);","strongs_def":"a tumor; also a mound, i.e. fortress","kjv_def":"emerod, fort, strong hold, tower."},
"H6077":{"lemma":"עֹפֶל","xlit":"ʻÔphel","pron":"o'-fel","derivation":"the same as H6076 (עֹפֶל);","strongs_def":"Ophel, a ridge in Jerusalem","kjv_def":"Ophel."},
"H6078":{"lemma":"עׇפְנִי","xlit":"ʻOphnîy","pron":"of-nee'","derivation":"from an unused noun (denoting a place in Palestine; from an unused root of uncertain meaning);","strongs_def":"an Ophnite (collectively) or inhabitants of Ophen","kjv_def":"Ophni."},
"H6079":{"lemma":"עַפְעַף","xlit":"ʻaphʻaph","pron":"af-af'","derivation":"from H5774 (עוּף);","strongs_def":"an eyelash (as fluttering); figuratively, morning ray","kjv_def":"dawning, eye-lid."},
"H6080":{"lemma":"עָפַר","xlit":"ʻâphar","pron":"aw-far'","derivation":"; a primitive root; meaning either to be gray or perhaps rather to pulverize; used only as denominative from H6083 (עָפָר),","strongs_def":"to be dust","kjv_def":"cast (dust)."},
"H6081":{"lemma":"עֵפֶר","xlit":"ʻÊpher","pron":"ay'-fer","derivation":"probably a variation of H6082 (עֹפֶר); gazelle;","strongs_def":"Epher, the name of an Arabian and of two Israelites","kjv_def":"Epher."},
"H6082":{"lemma":"עֹפֶר","xlit":"ʻôpher","pron":"o'-fer","derivation":"from H6080 (עָפַר);","strongs_def":"a fawn (from the dusty color)","kjv_def":"young roe (hart)."},
"H6083":{"lemma":"עָפָר","xlit":"ʻâphâr","pron":"aw-fawr'","derivation":"from H6080 (עָפַר);","strongs_def":"dust (as powdered or gray); hence, clay, earth, mud","kjv_def":"ashes, dust, earth, ground, morter, powder, rubbish."},
"H6084":{"lemma":"עׇפְרָה","xlit":"ʻOphrâh","pron":"of-raw'","derivation":"feminine of H6082 (עֹפֶר); female fawn;","strongs_def":"Ophrah, the name of an Israelite and of two places in Palestine","kjv_def":"Ophrah."},
"H6085":{"lemma":"עֶפְרוֹן","xlit":"ʻEphrôwn","pron":"ef-rone'","derivation":"from the same as H6081 (עֵפֶר); fawn-like;","strongs_def":"Ephron, the name of a Canaanite and of two places in Palestine","kjv_def":"Ephron, Ephrain (from the margin)."},
"H6086":{"lemma":"עֵץ","xlit":"ʻêts","pron":"ates","derivation":"from H6095 (עָצָה);","strongs_def":"a tree (from its firmness); hence, wood (plural sticks)","kjv_def":"[phrase] carpenter, gallows, helve, [phrase] pine, plank, staff, stalk, stick, stock, timber, tree, wood."},
"H6087":{"lemma":"עָצַב","xlit":"ʻâtsab","pron":"aw-tsab'","derivation":"a primitive root;","strongs_def":"properly, to carve, i.e. fabricate or fashion; hence (in a bad sense) to worry, pain or anger","kjv_def":"displease, grieve, hurt, make, be sorry, vex, worship, wrest."},
"H6088":{"lemma":"עֲצַב","xlit":"ʻătsab","pron":"ats-ab'","derivation":"(Aramaic) corresponding to H6087 (עָצַב);","strongs_def":"to afflict","kjv_def":"lamentable."},
"H6089":{"lemma":"עֶצֶב","xlit":"ʻetseb","pron":"eh'-tseb","derivation":"from H6087 (עָצַב);","strongs_def":"an earthen vessel; usually (painful) toil; also a pang (whether of body or mind)","kjv_def":"grievous, idol, labor, sorrow."},
"H6090":{"lemma":"עֹצֶב","xlit":"ʻôtseb","pron":"o'-tseb","derivation":"a variation of H6089 (עֶצֶב);","strongs_def":"an idol (as fashioned); also pain (bodily or mental)","kjv_def":"idol, sorrow, [idiom] wicked."},
"H6091":{"lemma":"עָצָב","xlit":"ʻâtsâb","pron":"aw-tsawb'","derivation":"from H6087 (עָצַב);","strongs_def":"an (idolatrous) image","kjv_def":"idol, image."},
"H6092":{"lemma":"עָצֵב","xlit":"ʻâtsêb","pron":"aw-tsabe'","derivation":"from H6087 (עָצַב);","strongs_def":"a (hired) workman","kjv_def":"labour."},
"H6093":{"lemma":"עִצָּבוֹן","xlit":"ʻitstsâbôwn","pron":"its-tsaw-bone'","derivation":"from H6087 (עָצַב);","strongs_def":"worrisomeness, i.e. labor or pain","kjv_def":"sorrow, toil."},
"H6094":{"lemma":"עַצֶּבֶת","xlit":"ʻatstsebeth","pron":"ats-tseh'-beth","derivation":"from H6087 (עָצַב);","strongs_def":"a idol; also, a pain or wound","kjv_def":"sorrow, wound."},
"H6095":{"lemma":"עָצָה","xlit":"ʻâtsâh","pron":"aw-tsaw'","derivation":"a primitive root;","strongs_def":"properly, to fasten (or make firm), i.e. to close (the eyes)","kjv_def":"shut."},
"H6096":{"lemma":"עָצֶה","xlit":"ʻâtseh","pron":"aw-tseh'","derivation":"from H6095 (עָצָה);","strongs_def":"the spine (as giving firmness to the body)","kjv_def":"backbone."},
"H6097":{"lemma":"עֵצָה","xlit":"ʻêtsâh","pron":"ay-tsaw'","derivation":"feminine of H6086 (עֵץ);","strongs_def":"timber","kjv_def":"trees."},
"H6098":{"lemma":"עֵצָה","xlit":"ʻêtsâh","pron":"ay-tsaw'","derivation":"from H3289 (יָעַץ);","strongs_def":"advice; by implication, plan; also prudence","kjv_def":"advice, advisement, counsel(l-(or)), purpose."},
"H6099":{"lemma":"עָצוּם","xlit":"ʻâtsûwm","pron":"aw-tsoom'","derivation":"or עָצֻם; passive participle of H6105 (עָצַם);","strongs_def":"powerful (specifically, a paw); by implication, numerous","kjv_def":"[phrase] feeble, great, mighty, must, strong."},
"H6100":{"lemma":"עֶצְיוֹן גֶּבֶר","xlit":"ʻEtsyôwn Geber","pron":"ets-yone' gheh'ber","derivation":"from H6096 (עָצֶה) and H1397 (גֶּבֶר); backbone-like of a man;","strongs_def":"Etsjon-Geber, a place on the Red Sea","kjv_def":"Ezion-geber."},
"H6101":{"lemma":"עָצַל","xlit":"ʻâtsal","pron":"aw-tsal'","derivation":"a primitive root;","strongs_def":"to lean idly, i.e. to be indolent or slack","kjv_def":"be slothful."},
"H6102":{"lemma":"עָצֵל","xlit":"ʻâtsêl","pron":"aw-tsale'","derivation":"from H6101 (עָצַל);","strongs_def":"indolent","kjv_def":"slothful, sluggard."},
"H6103":{"lemma":"עַצְלָה","xlit":"ʻatslâh","pron":"ats-law'","derivation":"feminine of H6102 (עָצֵל);","strongs_def":"(as abstractly) indolence","kjv_def":"slothfulness."},
"H6104":{"lemma":"עַצְלוּת","xlit":"ʻatslûwth","pron":"ats-looth'","derivation":"from H6101 (עָצַל);","strongs_def":"indolence","kjv_def":"idleness."},
"H6105":{"lemma":"עָצַם","xlit":"ʻâtsam","pron":"aw-tsam'","derivation":"a primitive root; also denominatively (from H6106 (עֶצֶם))","strongs_def":"to bind fast, i.e. close (the eyes); intransitively, to be (causatively, make) powerful or numerous; to crunch the bones","kjv_def":"break the bones, close, be great, be increased, be (wax) mighty(-ier), be more, shut, be(-come, make) strong(-er)."},
"H6106":{"lemma":"עֶצֶם","xlit":"ʻetsem","pron":"eh'tsem","derivation":"from H6105 (עָצַם);","strongs_def":"a bone (as strong); by extension, the body; figuratively, the substance, i.e. (as pron.) selfsame","kjv_def":"body, bone, [idiom] life, (self-) same, strength, [idiom] very."},
"H6107":{"lemma":"עֶצֶם","xlit":"ʻEtsem","pron":"eh'-tsem","derivation":"the same as H6106 (עֶצֶם); bone;","strongs_def":"Etsem, a place in Palestine","kjv_def":"Azem, Ezem."},
"H6108":{"lemma":"עֹצֶם","xlit":"ʻôtsem","pron":"o'-tsem","derivation":"from H6105 (עָצַם);","strongs_def":"power; hence, body","kjv_def":"might, strong, substance."},
"H6109":{"lemma":"עׇצְמָה","xlit":"ʻotsmâh","pron":"ots-maw'","derivation":"feminine of H6108 (עֹצֶם);","strongs_def":"powerfulness; by extension, numerousness","kjv_def":"abundance, strength."},
"H6110":{"lemma":"עַצֻּמָה","xlit":"ʻatstsumâh","pron":"ats-tsoo-maw'","derivation":"feminine of H6099 (עָצוּם);","strongs_def":"a bulwark, i.e. (figuratively) argument","kjv_def":"strong."},
"H6111":{"lemma":"עַצְמוֹן","xlit":"ʻAtsmôwn","pron":"ats-mone'","derivation":"or עַצְמֹן; from H6107 (עֶצֶם); bone-like;","strongs_def":"Atsmon, a place near Palestine","kjv_def":"Azmon."},
"H6112":{"lemma":"עֵצֶן","xlit":"ʻêtsen","pron":"ay'-tsen","derivation":"from an unused root meaning to be sharp or strong;","strongs_def":"a spear","kjv_def":"Eznite (from the margin)."},
"H6113":{"lemma":"עָצָר","xlit":"ʻâtsâr","pron":"aw-tsar'","derivation":"a primitive root;","strongs_def":"to inclose; by analogy, to hold back; also to maintain, rule, assemble","kjv_def":"[idiom] be able, close up, detain, fast, keep (self close, still), prevail, recover, refrain, [idiom] reign, restrain, retain, shut (up), slack, stay, stop, withhold (self)."},
"H6114":{"lemma":"עֶצֶר","xlit":"ʻetser","pron":"eh'-tser","derivation":"from H6113 (עָצָר);","strongs_def":"restraint","kjv_def":"[phrase] magistrate."},
"H6115":{"lemma":"עֹצֶר","xlit":"ʻôtser","pron":"o'-tser","derivation":"from H6113 (עָצָר);","strongs_def":"closure; also constraint","kjv_def":"[idiom] barren, oppression, [idiom] prison."},
"H6116":{"lemma":"עֲצָרָה","xlit":"ʻătsârâh","pron":"ats-aw-raw'","derivation":"or עֲצֶרֶת; from H6113 (עָצָר);","strongs_def":"an assembly, especially on afestival or holiday","kjv_def":"(solemn) assembly (meeting)."},
"H6117":{"lemma":"עָקַב","xlit":"ʻâqab","pron":"aw-kab'","derivation":"a primitive root; properly, to swell out or up; used only as denominative from H6119 (עָקֵב),","strongs_def":"to seize by the heel; figuratively, to circumvent (as if tripping up the heels); also to restrain (as if holding by the heel)","kjv_def":"take by the heel, stay, supplant, [idiom] utterly."},
"H6118":{"lemma":"עֵקֶב","xlit":"ʻêqeb","pron":"ay'-keb","derivation":"from H6117 (עָקַב) in the sense of H6119 (עָקֵב);","strongs_def":"a heel, i.e. (figuratively) the last of anything (used adverbially, for ever); also result, i.e. compensation; and so (adverb with preposition or relatively) on account of","kjv_def":"[idiom] because, by, end, for, if, reward."},
"H6119":{"lemma":"עָקֵב","xlit":"ʻâqêb","pron":"aw-kabe'","derivation":"or (feminine) עִקְּבָה; from H6117 (עָקַב);","strongs_def":"a heel (as protuberant); hence, a track; figuratively, the rear (of an army)","kjv_def":"heel, (horse-) hoof, last, lier in wait (by mistake for H6120 (עָקֵב)), (foot-) step."},
"H6120":{"lemma":"עָקֵב","xlit":"ʻâqêb","pron":"aw-kabe'","derivation":"from H6117 (עָקַב) in its denominative sense;","strongs_def":"a lier in wait","kjv_def":"heel (by mistake for H6119 (עָקֵב))."},
"H6121":{"lemma":"עָקֹב","xlit":"ʻâqôb","pron":"aw-kobe'","derivation":"from H6117 (עָקַב); in the original sense, a knoll (as swelling up); in the denominative sense","strongs_def":"(transitive) fraudulent or (intransitive) tracked","kjv_def":"crooked, deceitful, polluted."},
"H6122":{"lemma":"עׇקְבָה","xlit":"ʻoqbâh","pron":"ok-baw'","derivation":"feminine of an unused form from H6117 (עָקַב) meaning a trick;","strongs_def":"trickery","kjv_def":"subtilty."},
"H6123":{"lemma":"עָקַד","xlit":"ʻâqad","pron":"aw-kad'","derivation":"a primitive root;","strongs_def":"to tie with thongs","kjv_def":"bind."},
"H6124":{"lemma":"עָקֹד","xlit":"ʻâqôd","pron":"aw-kode'","derivation":"from H6123 (עָקַד);","strongs_def":"striped (with bands)","kjv_def":"ring straked."},
"H6125":{"lemma":"עָקָה","xlit":"ʻâqâh","pron":"aw-kaw'","derivation":"from H5781 (עוּק);","strongs_def":"constraint","kjv_def":"oppression."},
"H6126":{"lemma":"עַקּוּב","xlit":"ʻAqqûwb","pron":"ak-koob'","derivation":"from H6117 (עָקַב); insidious;","strongs_def":"Akkub, the name of five Israelites","kjv_def":"Akkub."},
"H6127":{"lemma":"עָקַל","xlit":"ʻâqal","pron":"aw-kal'","derivation":"a primitive root;","strongs_def":"to wrest","kjv_def":"wrong."},
"H6128":{"lemma":"עֲקַלְקַל","xlit":"ʻăqalqal","pron":"ak-al-kal'","derivation":"from H6127 (עָקַל);","strongs_def":"winding","kjv_def":"by(-way), crooked way."},
"H6129":{"lemma":"עֲקַלָּתוֹן","xlit":"ʻăqallâthôwn","pron":"ak-al-law-thone'","derivation":"from H6127 (עָקַל);","strongs_def":"tortuous","kjv_def":"crooked."},
"H6130":{"lemma":"עָקָן","xlit":"ʻÂqân","pron":"aw-kawn'","derivation":"from an unused root meaning to twist; tortuous;","strongs_def":"Akan, an Idummaean","kjv_def":"Akan. Compare H3292 (יַעֲקָן)."},
"H6131":{"lemma":"עָקַר","xlit":"ʻâqar","pron":"aw-kar'","derivation":"a primitive root;","strongs_def":"to pluck up (especially by the roots); specifically, to hamstring; figuratively, to exterminate","kjv_def":"dig down, hough, pluck up, root up."},
"H6132":{"lemma":"עֲקַר","xlit":"ʻăqar","pron":"ak-ar'","derivation":"(Aramaic) corresponding to H6131 (עָקַר)","strongs_def":"{to pluck up (especially by the roots); specifically, to hamstring; figuratively, to exterminate}","kjv_def":"pluck up by the roots."},
"H6133":{"lemma":"עֵקֶר","xlit":"ʻêqer","pron":"ay'-ker","derivation":"from H6131 (עָקַר);","strongs_def":"figuratively, a transplanted person, i.e. naturalized citizen","kjv_def":"stock."},
"H6134":{"lemma":"עֵקֶר","xlit":"ʻÊqer","pron":"ay'-ker","derivation":"the same as H6133 (עֵקֶר);","strongs_def":"Eker, an Israelite","kjv_def":"Eker."},
"H6135":{"lemma":"עָקָר","xlit":"ʻâqâr","pron":"aw-kawr'","derivation":"from H6131 (עָקַר);","strongs_def":"sterile (as if extirpated in the generative organs)","kjv_def":"([idiom] male or female) barren (woman)."},
"H6136":{"lemma":"עִקַּר","xlit":"ʻiqqar","pron":"ik-kar'","derivation":"(Aramaic) from H6132 (עֲקַר);","strongs_def":"a stock","kjv_def":"stump."},
"H6137":{"lemma":"עַקְרָב","xlit":"ʻaqrâb","pron":"ak-rawb'","derivation":"of uncertain derivation;","strongs_def":"a scorpion; figuratively, a scourge or knotted whip","kjv_def":"scorpion."},
"H6138":{"lemma":"עֶקְרוֹן","xlit":"ʻEqrôwn","pron":"ek-rone'","derivation":"from H6131 (עָקַר); eradication;","strongs_def":"Ekron, a place in Palestine","kjv_def":"Ekron."},
"H6139":{"lemma":"עֶקְרוֹנִי","xlit":"ʻEqrôwnîy","pron":"ek-ro-nee'","derivation":"or עֶקְרֹנִי; patrial from H6138 (עֶקְרוֹן);","strongs_def":"an Ekronite or inhabitant of Ekron","kjv_def":"Ekronite."},
"H6140":{"lemma":"עָקַשׁ","xlit":"ʻâqash","pron":"aw-kash'","derivation":"a primitive root;","strongs_def":"to knot or distort; figuratively, to pervert (act or declare perverse)","kjv_def":"make crooked, (prove, that is) perverse(-rt)."},
"H6141":{"lemma":"עִקֵּשׁ","xlit":"ʻiqqêsh","pron":"ik-kashe'","derivation":"from H6140 (עָקַשׁ);","strongs_def":"distorted; hence, false","kjv_def":"crooked, froward, perverse."},
"H6142":{"lemma":"עִקֵּשׁ","xlit":"ʻIqqêsh","pron":"ik-kashe'","derivation":"the same as H6141 (עִקֵּשׁ); perverse;","strongs_def":"Ikkesh, an Israelite","kjv_def":"Ikkesh."},
"H6143":{"lemma":"עִקְּשׁוּת","xlit":"ʻiqqᵉshûwth","pron":"ik-kesh-ooth'","derivation":"from H6141 (עִקֵּשׁ);","strongs_def":"perversity","kjv_def":"[idiom] froward."},
"H6144":{"lemma":"עָר","xlit":"ʻÂr","pron":"awr","derivation":"the same as H5892 (עִיר); a city;","strongs_def":"Ar, a place in Moab","kjv_def":"Ar."},
"H6145":{"lemma":"עָר","xlit":"ʻâr","pron":"awr","derivation":"from H5782 (עוּר);","strongs_def":"a foe (as watchful for mischief)","kjv_def":"enemy."},
"H6146":{"lemma":"עָר","xlit":"ʻâr","pron":"awr","derivation":"(Aramaic) corresponding to H6145 (עָר)","strongs_def":"{a foe (as watchful for mischief)}","kjv_def":"enemy."},
"H6147":{"lemma":"עֵר","xlit":"ʻÊr","pron":"ayr","derivation":"from H5782 (עוּר); watchful;","strongs_def":"Er, the name of two Israelites","kjv_def":"Er."},
"H6148":{"lemma":"עָרַב","xlit":"ʻârab","pron":"aw-rab'","derivation":"a primitive root;","strongs_def":"to braid, i.e. intermix; technically, to traffic (as if by barter); also or give to be security (as a kind of exchange)","kjv_def":"engage, (inter-) meddle (with), mingle (self), mortgage, occupy, give pledges, be(-come, put in) surety, undertake."},
"H6149":{"lemma":"עָרֵב","xlit":"ʻârêb","pron":"aw-rabe'","derivation":"a primitive root (rather identical with H6148 (עָרַב) through the idea of close association);","strongs_def":"to be agreeable","kjv_def":"be pleasant(-ing), take pleasure in, be sweet."},
"H6150":{"lemma":"עָרַב","xlit":"ʻârab","pron":"aw-rab'","derivation":"a primitive root (rather identical with H6148 (עָרַב) through the idea of covering with a texture);","strongs_def":"to grow dusky at sundown","kjv_def":"be darkened, (toward) evening."},
"H6151":{"lemma":"עֲרַב","xlit":"ʻărab","pron":"ar-ab'","derivation":"(Aramaic) corresponding to H6148 (עָרַב);","strongs_def":"to commingle","kjv_def":"mingle (self), mix."},
"H6152":{"lemma":"עֲרָב","xlit":"ʻĂrâb","pron":"ar-awb'","derivation":"or עֲרַב; from H6150 (עָרַב) in the figurative sense of sterility;","strongs_def":"Arab (i.e. Arabia), a country East of Palestine","kjv_def":"Arabia."},
"H6153":{"lemma":"עֶרֶב","xlit":"ʻereb","pron":"eh'-reb","derivation":"from H6150 (עָרַב);","strongs_def":"dusk","kjv_def":"[phrase] day, even(-ing, tide), night."},
"H6154":{"lemma":"עֵרֶב","xlit":"ʻêreb","pron":"ay'-reb","derivation":"or עֶרֶב; (1 Kings 10:15), (with the article [prefix), from H6148 (עָרַב);","strongs_def":"the web (or transverse threads of cloth); also a mixture, (or mongrel race)","kjv_def":"Arabia, mingled people, mixed (multitude), woof."},
"H6155":{"lemma":"עָרָב","xlit":"ʻârâb","pron":"aw-rawb'","derivation":"from H6148 (עָרַב);","strongs_def":"a willow (from the use of osiers as wattles)","kjv_def":"willow."},
"H6156":{"lemma":"עָרֵב","xlit":"ʻârêb","pron":"aw-rabe'","derivation":"from H6149 (עָרֵב);","strongs_def":"pleasant","kjv_def":"sweet."},
"H6157":{"lemma":"עָרֹב","xlit":"ʻârôb","pron":"aw-robe'","derivation":"from H6148 (עָרַב);","strongs_def":"a mosquito (from its swarming)","kjv_def":"divers sorts of flies, swarm."},
"H6158":{"lemma":"עֹרֵב","xlit":"ʻôrêb","pron":"o-rabe'","derivation":"or עוֹרֵב; from H6150 (עָרַב);","strongs_def":"a raven (from its dusky hue)","kjv_def":"raven."},
"H6159":{"lemma":"עֹרֵב","xlit":"ʻÔrêb","pron":"o-rabe'","derivation":"or עוֹרֵב; the same as H6158 (עֹרֵב);","strongs_def":"Oreb, the name of a Midianite and of the cliff near the Jordan","kjv_def":"Oreb."},
"H6160":{"lemma":"עֲרָבָה","xlit":"ʻărâbâh","pron":"ar-aw-baw'","derivation":"from H6150 (עָרַב) (in the sense of sterility);","strongs_def":"a desert; especially (with the article prefix) the (generally) sterile valley of the Jordan and its continuation to the Red Sea","kjv_def":"Arabah, champaign, desert, evening, heaven, plain, wilderness. See also H1026 (בֵּית הָעֲרָבָה)."},
"H6161":{"lemma":"עֲרֻבָּה","xlit":"ʻărubbâh","pron":"ar-oob-baw'","derivation":"feminine passive participle of H6048 (עֲנַמֶּלֶךְ) in the sense of a bargain or exchange;","strongs_def":"something given as security, i.e. (literally) a token (of safety) or (metaphorically) a bondsman","kjv_def":"pledge, surety."},
"H6162":{"lemma":"עֲרָבוֹן","xlit":"ʻărâbôwn","pron":"ar-aw-bone'","derivation":"from H6148 (עָרַב) (in the sense of exchange);","strongs_def":"a pawn (given as security)","kjv_def":"pledge."},
"H6163":{"lemma":"עֲרָבִי","xlit":"ʻĂrâbîy","pron":"ar-aw-bee'","derivation":"or עַרְבִי; patrial from H6152 (עֲרָב);","strongs_def":"an Arabian or inhabitant of Arab (i.e. Arabia)","kjv_def":"Arabian."},
"H6164":{"lemma":"עַרְבָתִי","xlit":"ʻArbâthîy","pron":"ar-baw-thee'","derivation":"patrial from H1026 (בֵּית הָעֲרָבָה);","strongs_def":"an Arbathite or inhabitant of (Beth-) Arabah","kjv_def":"Arbahite."},
"H6165":{"lemma":"עָרַג","xlit":"ʻârag","pron":"aw-rag'","derivation":"a primitive root;","strongs_def":"to long for","kjv_def":"cry, pant."},
"H6166":{"lemma":"עֲרָד","xlit":"ʻĂrâd","pron":"ar-awd'","derivation":"from an unused root meaning to sequester itself; fugitive;","strongs_def":"Arad, the name of a place near Palestine, also of a Canaanite and an Israelite","kjv_def":"Arad."},
"H6167":{"lemma":"עֲרָד","xlit":"ʻărâd","pron":"ar-awd'","derivation":"(Aramaic) corresponding to H6171 (עָרוֹד);","strongs_def":"an onager","kjv_def":"wild ass."},
"H6168":{"lemma":"עָרָה","xlit":"ʻârâh","pron":"aw-raw'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) bare; hence, to empty, pour out, demolish","kjv_def":"leave destitute, discover, empty, make naked, pour (out), rase, spread self, uncover."},
"H6169":{"lemma":"עָרָה","xlit":"ʻârâh","pron":"aw-raw'","derivation":"feminine from H6168 (עָרָה);","strongs_def":"a naked (i.e. level) plot","kjv_def":"paper reed."},
"H6170":{"lemma":"עֲרוּגָה","xlit":"ʻărûwgâh","pron":"ar-oo-gaw'","derivation":"or עֲרֻגָה; feminine passive participle of H6165 (עָרַג);","strongs_def":"something piled up (as if (figuratively) raised by mental aspiration), i.e. a paterre","kjv_def":"bed, furrow."},
"H6171":{"lemma":"עָרוֹד","xlit":"ʻârôwd","pron":"aw-rode'","derivation":"from the same as H6166 (עֲרָד);","strongs_def":"an onager (from his lonesome habits)","kjv_def":"wild ass."},
"H6172":{"lemma":"עֶרְוָה","xlit":"ʻervâh","pron":"er-vaw'","derivation":"from H6168 (עָרָה);","strongs_def":"nudity, literally (especially the pudenda) or figuratively (disgrace, blemish)","kjv_def":"nakedness, shame, unclean(-ness)."},
"H6173":{"lemma":"עַרְוָה","xlit":"ʻarvâh","pron":"ar-vaw'","derivation":"(Aramaic) corresponding to H6172 (עֶרְוָה);","strongs_def":"nakedness, i.e. (figuratively) impoverishment","kjv_def":"dishonor."},
"H6174":{"lemma":"עָרוֹם","xlit":"ʻârôwm","pron":"aw-rome'","derivation":"or עָרֹם; from H6191 (עָרַם) (in its original sense);","strongs_def":"nude, either partially or totally","kjv_def":"naked."},
"H6175":{"lemma":"עָרוּם","xlit":"ʻârûwm","pron":"aw-room'","derivation":"passive participle of H6191 (עָרַם);","strongs_def":"cunning (usually in a bad sense)","kjv_def":"crafty, prudent, subtil."},
"H6176":{"lemma":"עֲרוֹעֵר","xlit":"ʻărôwʻêr","pron":"ar-o-ayr'","derivation":"or עַרְעָר; from H6209 (עָרַר) reduplicated;","strongs_def":"a juniper (from its nudity of situation)","kjv_def":"health."},
"H6177":{"lemma":"עֲרוֹעֵר","xlit":"ʻĂrôwʻêr","pron":"ar-o-ayr'","derivation":"or עֲרֹעֵר; or עַרְעוֹר; the same as H6176 (עֲרוֹעֵר); nudity of situation;","strongs_def":"Aroer, the name of three places in or near Palestine","kjv_def":"Aroer."},
"H6178":{"lemma":"עָרוּץ","xlit":"ʻârûwts","pron":"aw-roots'","derivation":"passive participle of H6206 (עָרַץ);","strongs_def":"feared, i.e. (concretely) a horrible place or chasm","kjv_def":"cliffs."},
"H6179":{"lemma":"עֵרִי","xlit":"ʻÊrîy","pron":"ay-ree'","derivation":"from H5782 (עוּר); watchful;","strongs_def":"Eri, an Israelite","kjv_def":"Eri."},
"H6180":{"lemma":"עֵרִי","xlit":"ʻÊrîy","pron":"ay-ree'","derivation":"patronymically of H6179 (עֵרִי);","strongs_def":"a Erite (collectively) or descendants of Eri","kjv_def":"Erites."},
"H6181":{"lemma":"עֶרְיָה","xlit":"ʻeryâh","pron":"er-yaw'","derivation":"for H6172 (עֶרְוָה);","strongs_def":"nudity","kjv_def":"bare, naked, [idiom] quite."},
"H6182":{"lemma":"עֲרִיסָה","xlit":"ʻărîyçâh","pron":"ar-ee-saw'","derivation":"from an unused root meaning to comminute;","strongs_def":"meal","kjv_def":"dough."},
"H6183":{"lemma":"עָרִיף","xlit":"ʻârîyph","pron":"aw-reef'","derivation":"from H6201 (עָרַף);","strongs_def":"the sky (as dropping at the horizon)","kjv_def":"heaven."},
"H6184":{"lemma":"עָרִיץ","xlit":"ʻârîyts","pron":"aw-reets'","derivation":"from H6206 (עָרַץ);","strongs_def":"fearful, i.e. powerful or tyrannical","kjv_def":"mighty, oppressor, in great power, strong, terrible, violent."},
"H6185":{"lemma":"עֲרִירִי","xlit":"ʻărîyrîy","pron":"ar-e-ree'","derivation":"from H6209 (עָרַר);","strongs_def":"bare, i.e. destitute (of children)","kjv_def":"childless."},
"H6186":{"lemma":"עָרַךְ","xlit":"ʻârak","pron":"aw-rak'","derivation":"a primitive root;","strongs_def":"to set in a row, i.e. arrange, put in order (in a very wide variety of applications)","kjv_def":"put (set) (the battle, self) in array, compare, direct, equal, esteem, estimate, expert (in war), furnish, handle, join (battle), ordain, (lay, put, reckon up, set) (in) order, prepare, tax, value."},
"H6187":{"lemma":"עֵרֶךְ","xlit":"ʻêrek","pron":"eh'rek","derivation":"from H6186 (עָרַךְ);","strongs_def":"a pile, equipment, estimate","kjv_def":"equal, estimation, (things that are set in) order, price, proportion, [idiom] set at, suit, taxation, [idiom] valuest."},
"H6188":{"lemma":"עָרֵל","xlit":"ʻârêl","pron":"aw-rale'","derivation":"a primitive root; properly, to strip; but used as denominative from H6189 (עָרֵל);","strongs_def":"to expose or remove the prepuce, whether literal (to go naked) or figurative (to refrain from using)","kjv_def":"count uncircumcised, foreskin to be uncovered."},
"H6189":{"lemma":"עָרֵל","xlit":"ʻârêl","pron":"aw-rale'","derivation":"from H6188 (עָרֵל); properly, exposed, i.e. projecting loose (as to the prepuce); used only technically,","strongs_def":"uncircumcised (i.e. still having the prepuce uncurtailed)","kjv_def":"uncircumcised (person)."},
"H6190":{"lemma":"עׇרְלָה","xlit":"ʻorlâh","pron":"or-law'","derivation":"feminine of H6189 (עָרֵל);","strongs_def":"the prepuce","kjv_def":"foreskin, [phrase] uncircumcised."},
"H6191":{"lemma":"עָרַם","xlit":"ʻâram","pron":"aw-ram'","derivation":"a primitive root;","strongs_def":"properly, to be (or make) bare; but used only in the derivative sense (through the idea perhaps of smoothness) to be cunning (usually in a bad sense)","kjv_def":"[idiom] very, beware, take crafty (counsel), be prudent, deal subtilly."},
"H6192":{"lemma":"עָרַם","xlit":"ʻâram","pron":"aw-ram'","derivation":"a primitive root;","strongs_def":"to pile up","kjv_def":"gather together."},
"H6193":{"lemma":"עֹרֶם","xlit":"ʻôrem","pron":"o'-rem","derivation":"from H6191 (עָרַם);","strongs_def":"a stratagem","kjv_def":"craftiness."},
"H6194":{"lemma":"עָרֵם","xlit":"ʻârêm","pron":"aw-rame'","derivation":"(Jeremiah 50:26) or (feminine) עֲרֵמָה; from H6192 (עָרַם);","strongs_def":"a heap; specifically, a sheaf","kjv_def":"heap (of corn), sheaf."},
"H6195":{"lemma":"עׇרְמָה","xlit":"ʻormâh","pron":"or-maw'","derivation":"feminine of H6193 (עֹרֶם);","strongs_def":"trickery; or (in a good sense) discretion","kjv_def":"guile, prudence, subtilty, wilily, wisdom."},
"H6196":{"lemma":"עַרְמוֹן","xlit":"ʻarmôwn","pron":"ar-mone'","derivation":"probably from H6191 (עָרַם);","strongs_def":"the plane tree (from its smooth and shed bark)","kjv_def":"chestnut tree."},
"H6197":{"lemma":"עֵרָן","xlit":"ʻÊrân","pron":"ay-rawn'","derivation":"probably from H5782 (עוּר); watchful;","strongs_def":"Eran, an Israelite","kjv_def":"Eran."},
"H6198":{"lemma":"עֵרָנִי","xlit":"ʻÊrânîy","pron":"ay-raw-nee'","derivation":"patronymically from H6197 (עֵרָן);","strongs_def":"an Eranite or descendant (collectively) of Eran","kjv_def":"Eranites."},
"H6199":{"lemma":"עַרעָר","xlit":"ʻarʻâr","pron":"ar-awr'","derivation":"from H6209 (עָרַר);","strongs_def":"naked, i.e. (figuratively) poor","kjv_def":"destitute. See also H6176 (עֲרוֹעֵר)."},
"H6200":{"lemma":"עֲרֹעֵרִי","xlit":"ʻĂrôʻêrîy","pron":"ar-o-ay-ree'","derivation":"patronymically from H6177 (עֲרוֹעֵר);","strongs_def":"an Aroerite or inhabitant of Aroer","kjv_def":"Aroerite."},
"H6201":{"lemma":"עָרַף","xlit":"ʻâraph","pron":"aw-raf'","derivation":"a primitive root;","strongs_def":"to droop; hence, to drip","kjv_def":"drop (down)."},
"H6202":{"lemma":"עָרַף","xlit":"ʻâraph","pron":"aw-raf'","derivation":"a primitive root (identical with through the idea of sloping); properly, to bend downward; but used only as a denominative from H6203 (עֹרֶף),","strongs_def":"to break the neck; hence (figuratively) to destroy","kjv_def":"that is beheaded, break down, break (cut off, strike off) neck."},
"H6203":{"lemma":"עֹרֶף","xlit":"ʻôreph","pron":"o-ref'","derivation":"from H6202 (עָרַף);","strongs_def":"the nape or back of the neck (as declining); hence, the back generally (whether literal or figurative)","kjv_def":"back ((stiff-) neck((-ed)."},
"H6204":{"lemma":"עׇרְפָּה","xlit":"ʻOrpâh","pron":"or-paw'","derivation":"feminine of H6203 (עֹרֶף); mane;","strongs_def":"Orpah, a Moabites","kjv_def":"Orpah."},
"H6205":{"lemma":"עֲרָפֶל","xlit":"ʻărâphel","pron":"ar-aw-fel'","derivation":"probably from H6201 (עָרַף);","strongs_def":"gloom (as of a lowering sky)","kjv_def":"(gross, thick) dark (cloud, -ness)."},
"H6206":{"lemma":"עָרַץ","xlit":"ʻârats","pron":"aw-rats'","derivation":"a primitive root;","strongs_def":"to awe or (intransitive) to dread; hence, to harass","kjv_def":"be affrighted (afraid, dread, feared, terrified), break, dread, fear, oppress, prevail, shake terribly."},
"H6207":{"lemma":"עָרַק","xlit":"ʻâraq","pron":"aw-rak'","derivation":"a primitive root;","strongs_def":"to gnaw, i.e. (figuratively) eat (by hyberbole); also (participle) a pain","kjv_def":"fleeing, sinew."},
"H6208":{"lemma":"עַרְקִי","xlit":"ʻArqîy","pron":"ar-kee'","derivation":"patrial from an unused name meaning a tush;","strongs_def":"an Arkite or inhabitant of Erek","kjv_def":"Arkite."},
"H6209":{"lemma":"עָרַר","xlit":"ʻârar","pron":"aw-rar'","derivation":"a primitive root;","strongs_def":"to bare; figuratively, to demolish","kjv_def":"make bare, break, raise up (perhaps by clerical error for raze), [idiom] utterly."},
"H6210":{"lemma":"עֶרֶשׂ","xlit":"ʻeres","pron":"eh'res","derivation":"from an unused root meaning perhaps to arch;","strongs_def":"a couch (properly, with a canopy)","kjv_def":"bed(-stead), couch."},
"H6211":{"lemma":"עָשׁ","xlit":"ʻâsh","pron":"awsh","derivation":"from H6244 (עָשֵׁשׁ);","strongs_def":"a moth","kjv_def":"moth. See also H5906 (עַיִשׁ)."},
"H6212":{"lemma":"עֶשֶׂב","xlit":"ʻeseb","pron":"eh'seb","derivation":"from an unused root meaning to glisten (or be green);","strongs_def":"grass (or any tender shoot)","kjv_def":"grass, herb."},
"H6213":{"lemma":"עָשָׂה","xlit":"ʻâsâh","pron":"aw-saw'","derivation":"a primitive root;","strongs_def":"to do or make, in the broadest sense and widest application","kjv_def":"accomplish, advance, appoint, apt, be at, become, bear, bestow, bring forth, bruise, be busy, [idiom] certainly, have the charge of, commit, deal (with), deck, [phrase] displease, do, (ready) dress(-ed), (put in) execute(-ion), exercise, fashion, [phrase] feast, (fight-) ing man, [phrase] finish, fit, fly, follow, fulfill, furnish, gather, get, go about, govern, grant, great, [phrase] hinder, hold (a feast), [idiom] indeed, [phrase] be industrious, [phrase] journey, keep, labour, maintain, make, be meet, observe, be occupied, offer, [phrase] officer, pare, bring (come) to pass, perform, pracise, prepare, procure, provide, put, requite, [idiom] sacrifice, serve, set, shew, [idiom] sin, spend, [idiom] surely, take, [idiom] thoroughly, trim, [idiom] very, [phrase] vex, be (warr-) ior, work(-man), yield, use."},
"H6214":{"lemma":"עֲשָׂהאֵל","xlit":"ʻĂsâhʼêl","pron":"as-aw-ale'","derivation":"from H6213 (עָשָׂה) and H410 (אֵל); God has made;","strongs_def":"Asahel, the name of four Israelites","kjv_def":"Asahel."},
"H6215":{"lemma":"עֵשָׂו","xlit":"ʻÊsâv","pron":"ay-sawv'","derivation":"apparently a form of the passive participle of H6213 (עָשָׂה) in the original sense of handling; rough (i.e. sensibly felt);","strongs_def":"Esav, a son of Isaac, including his posterity","kjv_def":"Esau."},
"H6216":{"lemma":"עָשׁוֹק","xlit":"ʻâshôwq","pron":"aw-shoke'","derivation":"from H6231 (עָשַׁק);","strongs_def":"oppressive (as noun, a tyrant)","kjv_def":"oppressor."},
"H6217":{"lemma":"עָשׁוּק","xlit":"ʻâshûwq","pron":"aw-shook'","derivation":"or עָשֻׁק; passive participle of H6231 (עָשַׁק);","strongs_def":"used in plural masculine as abstractly, tyranny","kjv_def":"oppressed(-ion). (Doubtful.)"},
"H6218":{"lemma":"עָשׂוֹר","xlit":"ʻâsôwr","pron":"aw-sore'","derivation":"or עָשֹׂר; from H6235 (עֶשֶׂר);","strongs_def":"ten; by abbreviated form ten strings, and so a decachord","kjv_def":"(instrument of) ten (strings, -th)."},
"H6219":{"lemma":"עָשׁוֹת","xlit":"ʻâshôwth","pron":"aw-shoth'","derivation":"from H6245 (עָשַׁת);","strongs_def":"shining, i.e. polished","kjv_def":"bright."},
"H6220":{"lemma":"עַשְׁוָת","xlit":"ʻAshvâth","pron":"ash-vawth'","derivation":"for H6219 (עָשׁוֹת); bright;","strongs_def":"Ashvath, an Israelite","kjv_def":"Ashvath."},
"H6221":{"lemma":"עֲשִׂיאֵל","xlit":"ʻĂsîyʼêl","pron":"as-ee-ale'","derivation":"from H6213 (עָשָׂה) and H410 (אֵל); made of God;","strongs_def":"Asiel, an Israelite","kjv_def":"Asiel."},
"H6222":{"lemma":"עֲשָׂיָה","xlit":"ʻĂsâyâh","pron":"aw-saw-yaw'","derivation":"from H6213 (עָשָׂה) and H3050 (יָהּ); Jah has made;","strongs_def":"Asajah, the name of three or four Israelites","kjv_def":"Asaiah."},
"H6223":{"lemma":"עָשִׁיר","xlit":"ʻâshîyr","pron":"aw-sheer'","derivation":"from H6238 (עָשַׁר);","strongs_def":"rich, whether literal or figurative (noble)","kjv_def":"rich (man)."},
"H6224":{"lemma":"עֲשִׂירִי","xlit":"ʻăsîyrîy","pron":"as-ee-ree'","derivation":"from H6235 (עֶשֶׂר);","strongs_def":"tenth; by abbreviation, tenth month or (feminine) part","kjv_def":"tenth (part)."},
"H6225":{"lemma":"עָשַׁן","xlit":"ʻâshan","pron":"aw-shan'","derivation":"a primitive root;","strongs_def":"to smoke, whether literal or figurative","kjv_def":"be angry (be on a) smoke."},
"H6226":{"lemma":"עָשֵׁן","xlit":"ʻâshên","pron":"aw-shane'","derivation":"from H6225 (עָשַׁן);","strongs_def":"smoky","kjv_def":"smoking."},
"H6227":{"lemma":"עָשָׁן","xlit":"ʻâshân","pron":"aw-shawn'","derivation":"from H6225 (עָשַׁן);","strongs_def":"smoke, literally or figuratively (vapor, dust, anger)","kjv_def":"smoke(-ing)."},
"H6228":{"lemma":"עָשָׁן","xlit":"ʻÂshân","pron":"aw-shawn'","derivation":"the same as H6227 (עָשָׁן);","strongs_def":"Ashan, a place in Palestine","kjv_def":"Ashan."},
"H6229":{"lemma":"עָשַׂק","xlit":"ʻâsaq","pron":"aw-sak","derivation":"a primitive root (identical with H6231 (עָשַׁק));","strongs_def":"to press upon, i.e. quarrel","kjv_def":"strive with."},
"H6230":{"lemma":"עֵשֶׂק","xlit":"ʻêseq","pron":"ay'sek","derivation":"from H6229 (עָשַׂק);","strongs_def":"strife","kjv_def":"Esek."},
"H6231":{"lemma":"עָשַׁק","xlit":"ʻâshaq","pron":"aw-shak'","derivation":"a primitive root (compare H6229 (עָשַׂק));","strongs_def":"to press upon, i.e. oppress, defraud, violate, overflow","kjv_def":"get deceitfully, deceive, defraud, drink up, (use) oppress(-ion), -or), do violence (wrong)."},
"H6232":{"lemma":"עֵשֶׁק","xlit":"ʻÊsheq","pron":"ay-shek'","derivation":"from H6231 (עָשַׁק); oppression;","strongs_def":"Eshek, an Israelite","kjv_def":"Eshek."},
"H6233":{"lemma":"עֹשֶׁק","xlit":"ʻôsheq","pron":"o'-shek","derivation":"from H6231 (עָשַׁק);","strongs_def":"injury, fraud, (subjectively) distress, (concretely) unjust gain","kjv_def":"cruelly, extortion, oppression, thing (deceitfully gotten)."},
"H6234":{"lemma":"עׇשְׁקָה","xlit":"ʻoshqâh","pron":"osh-kaw'","derivation":"feminine of H6233 (עֹשֶׁק);","strongs_def":"anguish","kjv_def":"oppressed."},
"H6235":{"lemma":"עֶשֶׂר","xlit":"ʻeser","pron":"eh'ser","derivation":"masculine of term עֲשָׂרָה; from H6237 (עָשַׂר);","strongs_def":"ten (as an accumulation to the extent of the digits)","kjv_def":"ten, (fif-, seven-) teen."},
"H6236":{"lemma":"עֲשַׂר","xlit":"ʻăsar","pron":"as-ar'","derivation":"(Aramaic) masculine עֲשְׂרָה (Aramaic); corresponding to H6235 (עֶשֶׂר);","strongs_def":"ten","kjv_def":"ten, [phrase] twelve."},
"H6237":{"lemma":"עָשַׂר","xlit":"ʻâsar","pron":"aw-sar'","derivation":"a primitive root (identical with H6238 (עָשַׁר)); to accumulate; but used only as denominative from H6235 (עֶשֶׂר);","strongs_def":"to tithe, i.e. to take or give atenth","kjv_def":"[idiom] surely, give (take) the tenth, (have, take) tithe(-ing, -s), [idiom] truly."},
"H6238":{"lemma":"עָשַׁר","xlit":"ʻâshar","pron":"aw-shar'","derivation":"a primitive root;","strongs_def":"properly, to accumulate; chiefly (specifically) to grow (causatively, make) rich","kjv_def":"be(-come, en-, make, make self, wax) rich, make (1 Kings 22:48 marg). See H6240 (עָשָׂר)."},
"H6239":{"lemma":"עֹשֶׁר","xlit":"ʻôsher","pron":"o'-sher","derivation":"from H6238 (עָשַׁר);","strongs_def":"wealth","kjv_def":"[idiom] far (richer), riches."},
"H6240":{"lemma":"עָשָׂר","xlit":"ʻâsâr","pron":"aw-sawr'","derivation":"for H6235 (עֶשֶׂר);","strongs_def":"ten (only in combination), i.e. -teen; also (ordinal) -teenth","kjv_def":"(eigh-, fif-, four-, nine-, seven-, six-, thir-) teen(-th), [phrase] eleven(-th), [phrase] sixscore thousand, [phrase] twelve(-th)."},
"H6241":{"lemma":"עִשָּׂרוֹן","xlit":"ʻissârôwn","pron":"is-saw-rone'","derivation":"or עִשָּׂרֹן; from H6235 (עֶשֶׂר);","strongs_def":"(fractional) a tenth part","kjv_def":"tenth deal."},
"H6242":{"lemma":"עֶשְׂרִים","xlit":"ʻesrîym","pron":"es-reem'","derivation":"from H6235 (עֶשֶׂר);","strongs_def":"twenty; also (ordinal) twentieth","kjv_def":"(six-) score, twenty(-ieth)."},
"H6243":{"lemma":"עֶשְׂרִין","xlit":"ʻesrîyn","pron":"es-reen'","derivation":"(Aramaic) corresponding to H6242 (עֶשְׂרִים)","strongs_def":"{twenty; also (ordinal) twentieth}","kjv_def":"twenty."},
"H6244":{"lemma":"עָשֵׁשׁ","xlit":"ʻâshêsh","pron":"aw-shaysh'","derivation":"a primitive root;","strongs_def":"probably to shrink, i.e. fail","kjv_def":"be consumed."},
"H6245":{"lemma":"עָשַׁת","xlit":"ʻâshath","pron":"aw-shath'","derivation":"a primitive root;","strongs_def":"probably to be sleek, i.e. glossy; hence (through the idea of polishing) to excogitate (as if forming in the mind)","kjv_def":"shine, think."},
"H6246":{"lemma":"עֲשִׁת","xlit":"ʻăshith","pron":"ash-eeth'","derivation":"(Aramaic) corresponding to H6245 (עָשַׁת);","strongs_def":"to purpose","kjv_def":"think."},
"H6247":{"lemma":"עֶשֶׁת","xlit":"ʻesheth","pron":"eh'-sheth","derivation":"from H6245 (עָשַׁת);","strongs_def":"a fabric","kjv_def":"bright."},
"H6248":{"lemma":"עַשְׁתּוּת","xlit":"ʻashtûwth","pron":"ash-tooth'","derivation":"from H6245 (עָשַׁת);","strongs_def":"cogitation","kjv_def":"thought."},
"H6249":{"lemma":"עַשְׁתֵּי","xlit":"ʻashtêy","pron":"ash-tay'","derivation":"apparently masculine plural construction of H6247 (עֶשֶׁת) in the sense of an afterthought (used only in connection with H6240 (עָשָׂר) in lieu of H259 (אֶחָד))","strongs_def":"eleven or (ordinal) eleventh","kjv_def":"[phrase] eleven(-th)."},
"H6250":{"lemma":"עֶשְׁתֹּנָה","xlit":"ʻeshtônâh","pron":"esh-to-naw'","derivation":"from H6245 (עָשַׁת);","strongs_def":"thinking","kjv_def":"thought."},
"H6251":{"lemma":"עַשְׁתְּרָה","xlit":"ʻashtᵉrâh","pron":"ash-ter-aw'","derivation":"probably from H6238 (עָשַׁר);","strongs_def":"increase","kjv_def":"flock."},
"H6252":{"lemma":"עַשְׁתָּרוֹת","xlit":"ʻAshtârôwth","pron":"ash-taw-roth'","derivation":"or עַשְׁתָּרֹת; plural of H6251 (עַשְׁתְּרָה);","strongs_def":"Ashtaroth, the name of a Sidonian deity, and of a place East of the Jordan","kjv_def":"Asharoth, Astaroth. See also H1045 (בֵּית עַשְׁתָּרוֹת), H6253 (עַשְׁתֹּרֶת), H6255 (עַשְׁתְּרֹת קַרְנַיִם)."},
"H6253":{"lemma":"עַשְׁתֹּרֶת","xlit":"ʻAshtôreth","pron":"ash-to'reth","derivation":"probably for H6251 (עַשְׁתְּרָה);","strongs_def":"Ashtoreth, the Phoenician goddess of love (and increase)","kjv_def":"Ashtoreth."},
"H6254":{"lemma":"עַשְׁתְּרָתִי","xlit":"ʻAshtᵉrâthîy","pron":"ash-ter-aw-thee'","derivation":"patrial from H6252 (עַשְׁתָּרוֹת);","strongs_def":"an Ashterathite or inhabitant of Ashtaroth","kjv_def":"Ashterathite."},
"H6255":{"lemma":"עַשְׁתְּרֹת קַרְנַיִם","xlit":"ʻAshtᵉrôth Qarnayim","pron":"ash-ter-oth' kar-nah'-yim","derivation":"from H6252 (עַשְׁתָּרוֹת) and the dual of H7161 (קֶרֶן); Ashtaroth of (the) double horns (a symbol of the deity);","strongs_def":"Ashteroth-Karnaim, a place East of the Jordan","kjv_def":"Ashtoreth Karnaim."},
"H6256":{"lemma":"עֵת","xlit":"ʻêth","pron":"ayth","derivation":"from H5703 (עַד);","strongs_def":"time, especially (adverb with preposition) now, when, etc.","kjv_def":"[phrase] after, (al-) ways, [idiom] certain, [phrase] continually, [phrase] evening, long, (due) season, so (long) as, (even-, evening-, noon-) tide, (meal-), what) time, when."},
"H6257":{"lemma":"עָתַד","xlit":"ʻâthad","pron":"aw-thad'","derivation":"a primitive root;","strongs_def":"to prepare","kjv_def":"make fit, be ready to become."},
"H6258":{"lemma":"עַתָּה","xlit":"ʻattâh","pron":"at-taw'","derivation":"from H6256 (עֵת);","strongs_def":"at this time, whether adverb, conjunction or expletive","kjv_def":"henceforth, now, straightway, this time, whereas."},
"H6259":{"lemma":"עָתוּד","xlit":"ʻâthûwd","pron":"aw-thood'","derivation":"passive participle of H6257 (עָתַד);","strongs_def":"prepared","kjv_def":"ready."},
"H6260":{"lemma":"עַתּוּד","xlit":"ʻattûwd","pron":"at-tood'","derivation":"or עַתֻּד; from H6257 (עָתַד);","strongs_def":"prepared, i.e. full grown; spoken only (in plural) of he-goats, or (figuratively) leaders of the people","kjv_def":"chief one, (he) goat, ram."},
"H6261":{"lemma":"עִתִּי","xlit":"ʻittîy","pron":"it-tee'","derivation":"from H6256 (עֵת);","strongs_def":"timely","kjv_def":"fit."},
"H6262":{"lemma":"עַתַּי","xlit":"ʻAttay","pron":"at-tah'ee","derivation":"for H6261 (עִתִּי);","strongs_def":"Attai, the name of three Israelites","kjv_def":"Attai."},
"H6263":{"lemma":"עֲתִיד","xlit":"ʻăthîyd","pron":"ath-eed'","derivation":"(Aramaic) corresponding to H6264 (עָתִיד);","strongs_def":"prepared","kjv_def":"ready."},
"H6264":{"lemma":"עָתִיד","xlit":"ʻâthîyd","pron":"aw-theed'","derivation":"from H6257 (עָתַד);","strongs_def":"prepared; by implication, skilful; feminine plural the future; also treasure","kjv_def":"things that shall come, ready, treasures."},
"H6265":{"lemma":"עֲתָיָה","xlit":"ʻĂthâyâh","pron":"ath-aw-yaw'","derivation":"from H5790 (עוּת) and H3050 (יָהּ); Jah has helped;","strongs_def":"Athajah, an Israelite","kjv_def":"Athaiah."},
"H6266":{"lemma":"עָתִיק","xlit":"ʻâthîyq","pron":"aw-theek'","derivation":"from H6275 (עָתַק);","strongs_def":"properly, antique, i.e. venerable or splendid","kjv_def":"durable."},
"H6267":{"lemma":"עַתִּיק","xlit":"ʻattîyq","pron":"at-teek'","derivation":"from H6275 (עָתַק);","strongs_def":"removed, i.e. weaned; also antique","kjv_def":"ancient, drawn."},
"H6268":{"lemma":"עַתִּיק","xlit":"ʻattîyq","pron":"at-teek'","derivation":"(Aramaic) corresponding to H6267 (עַתִּיק);","strongs_def":"venerable","kjv_def":"ancient."},
"H6269":{"lemma":"עֲתָךְ","xlit":"ʻĂthâk","pron":"ath-awk'","derivation":"from an unused root meaning to sojourn; lodging;","strongs_def":"Athak, a place in Palestine","kjv_def":"Athach."},
"H6270":{"lemma":"עַתְלַי","xlit":"ʻAthlay","pron":"ath-lah'ee","derivation":"from an unused root meaning to compress; constringent;","strongs_def":"Athlai, an Israelite","kjv_def":"Athlai."},
"H6271":{"lemma":"עֲתַלְיָה","xlit":"ʻĂthalyâh","pron":"ath-al-yaw'","derivation":"or עֲתַלְיָהוּ; from the same as H6270 (עַתְלַי) and H3050 (יָהּ); Jah has constrained;","strongs_def":"Athaljah, the name of an Israelitess and two Israelites","kjv_def":"Athaliah."},
"H6272":{"lemma":"עָתַם","xlit":"ʻâtham","pron":"aw-tham","derivation":"a primitive root;","strongs_def":"probably to glow, i.e. (figuratively) be desolated","kjv_def":"be darkened."},
"H6273":{"lemma":"עׇתְנִי","xlit":"ʻOthnîy","pron":"oth-nee'","derivation":"from an unused root meaning to force; forcible;","strongs_def":"Othni, an Israelite","kjv_def":"Othni."},
"H6274":{"lemma":"עׇתְנִיאֵל","xlit":"ʻOthnîyʼêl","pron":"oth-nee-ale'","derivation":"from the same as H6273 (עׇתְנִי) and H410 (אֵל); force of God;","strongs_def":"Othniel, an Israelite","kjv_def":"Othniel."},
"H6275":{"lemma":"עָתַק","xlit":"ʻâthaq","pron":"aw-thak'","derivation":"a primitive root;","strongs_def":"to remove (intransitive or transitive) figuratively, to grow old; specifically, to transcribe","kjv_def":"copy out, leave off, become (wax) old, remove."},
"H6276":{"lemma":"עָתֵק","xlit":"ʻâthêq","pron":"aw-thake'","derivation":"from H6275 (עָתַק);","strongs_def":"antique, i.e. valued","kjv_def":"durable."},
"H6277":{"lemma":"עָתָק","xlit":"ʻâthâq","pron":"aw-thawk'","derivation":"from H6275 (עָתַק) in the sense of license;","strongs_def":"impudent","kjv_def":"arrogancy, grievous (hard) things, stiff."},
"H6278":{"lemma":"עֵת קָצִין","xlit":"ʻÊth Qâtsîyn","pron":"ayth kaw-tseen'","derivation":"from H6256 (עֵת) and H7011 (קַיָּם); time of a judge;","strongs_def":"Eth-Katsin, a place in Palestine","kjv_def":"Ittahkazin (by including directive enclitic)."},
"H6279":{"lemma":"עָתַר","xlit":"ʻâthar","pron":"aw-thar'","derivation":"a primitive root (rather denominative from H6281 (עֶתֶר));","strongs_def":"to burn incense in worship, i.e. intercede (reciprocally, listen to prayer)","kjv_def":"intreat, (make) pray(-er)."},
"H6280":{"lemma":"עָתַר","xlit":"ʻâthar","pron":"aw-thar'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) abundant","kjv_def":"deceitful, multiply."},
"H6281":{"lemma":"עֶתֶר","xlit":"ʻEther","pron":"eh'ther","derivation":"from H6280 (עָתַר); abundance;","strongs_def":"Ether, a place in Palestine","kjv_def":"Ether."},
"H6282":{"lemma":"עָתָר","xlit":"ʻâthâr","pron":"aw-thawr'","derivation":"from H6280 (עָתַר); also compare H6279 (עָתַר)","strongs_def":"incense (as increasing to a volume of smoke); hence  a worshipper","kjv_def":"suppliant, thick."},
"H6283":{"lemma":"עֲתֶרֶת","xlit":"ʻăthereth","pron":"ath-eh'-reth","derivation":"from H6280 (עָתַר);","strongs_def":"copiousness","kjv_def":"abundance. p"},
"H6284":{"lemma":"פָּאָה","xlit":"pâʼâh","pron":"paw-aw'","derivation":"a primitive root;","strongs_def":"to puff, i.e. blow away","kjv_def":"scatter into corners."},
"H6285":{"lemma":"פֵּאָה","xlit":"pêʼâh","pron":"pay-aw'","derivation":"feminine of H6311 (פֹּה);","strongs_def":"properly, mouth in a figurative sense, i.e. direction, region, extremity","kjv_def":"corner, end, quarter, side."},
"H6286":{"lemma":"פָּאַר","xlit":"pâʼar","pron":"paw-ar'","derivation":"a primitive root; also denominative from H6288 (פְּאֹרָה),","strongs_def":"to gleam, i.e. (causatively) embellish; figuratively, to boast; also to explain (i.e. make clear) oneself; to shake a tree","kjv_def":"beautify, boast self, go over the boughs, glorify (self), glory, vaunt self."},
"H6287":{"lemma":"פְּאֵר","xlit":"pᵉʼêr","pron":"peh-ayr'","derivation":"from H6286 (פָּאַר);","strongs_def":"an embellishment, i.e. fancy head-dress","kjv_def":"beauty, bonnet, goodly, ornament, tire."},
"H6288":{"lemma":"פְּאֹרָה","xlit":"pᵉʼôrâh","pron":"peh-o-raw'","derivation":"or פֹּרָאה; or פֻּארָה; from H6286 (פָּאַר);","strongs_def":"properly, ornamentation, i.e. (plural) foliage (including the limbs) as bright green","kjv_def":"bough, branch, sprig."},
"H6289":{"lemma":"פָּארוּר","xlit":"pâʼrûwr","pron":"paw-roor'","derivation":"from H6286 (פָּאַר);","strongs_def":"properly, illuminated, i.e. a glow; as noun, a flush (of anxiety)","kjv_def":"blackness."},
"H6290":{"lemma":"פָּארָן","xlit":"Pâʼrân","pron":"paw-rawn'","derivation":"from H6286 (פָּאַר); ornamental;","strongs_def":"Paran, a desert of Arabia","kjv_def":"Paran."},
"H6291":{"lemma":"פַּג","xlit":"pag","pron":"pag","derivation":"from an unused root meaning to be torpid, i.e.","strongs_def":"crude; an unripe fig","kjv_def":"green figuratively"},
"H6292":{"lemma":"פִּגּוּל","xlit":"piggûwl","pron":"pig-gool'","derivation":"or פִּגֻּל; from an unused root meaning to stink;","strongs_def":"properly, fetid, i.e. (figuratively) unclean (ceremonially)","kjv_def":"abominable(-tion, thing)."},
"H6293":{"lemma":"פָּגַע","xlit":"pâgaʻ","pron":"paw-gah'","derivation":"a primitive root;","strongs_def":"to impinge, by accident or violence, or (figuratively) by importunity","kjv_def":"come (betwixt), cause to entreat, fall (upon), make intercession, intercessor, intreat, lay, light (upon), meet (together), pray, reach, run."},
"H6294":{"lemma":"פֶּגַע","xlit":"pegaʻ","pron":"peh'-gah","derivation":"from H6293 (פָּגַע);","strongs_def":"impact (casual)","kjv_def":"chance, occurent."},
"H6295":{"lemma":"פַּגְעִיאֵל","xlit":"Pagʻîyʼêl","pron":"pag-ee-ale'","derivation":"from H6294 (פֶּגַע) and H410 (אֵל); accident of God;","strongs_def":"Pagiel, an Israelite","kjv_def":"Pagiel."},
"H6296":{"lemma":"פָּגַר","xlit":"pâgar","pron":"paw-gar'","derivation":"a primitive root;","strongs_def":"to relax, i.e. become exhausted","kjv_def":"be faint."},
"H6297":{"lemma":"פֶּגֶר","xlit":"peger","pron":"peh'gher","derivation":"from H6296 (פָּגַר);","strongs_def":"a carcase (as limp), whether of man or beast; figuratively, an idolatrous image","kjv_def":"carcase, corpse, dead body."},
"H6298":{"lemma":"פָּגַשׁ","xlit":"pâgash","pron":"paw-gash'","derivation":"a primitive root;","strongs_def":"to come in contact with, whether by accident or violence; figuratively, to concur","kjv_def":"meet (with, together)."},
"H6299":{"lemma":"פָּדָה","xlit":"pâdâh","pron":"paw-daw'","derivation":"a primitive root;","strongs_def":"to sever, i.e. ransom; gener. to release, preserve","kjv_def":"[idiom] at all, deliver, [idiom] by any means, ransom, (that are to be, let be) redeem(-ed), rescue, [idiom] surely."},
"H6300":{"lemma":"פְּדַהְאֵל","xlit":"Pᵉdahʼêl","pron":"ped-ah-ale'","derivation":"from H6299 (פָּדָה) and H410 (אֵל); God has ransomed;","strongs_def":"Pedahel, an Israelite","kjv_def":"Pedahel."},
"H6301":{"lemma":"פְּדָהצוּר","xlit":"Pᵉdâhtsûwr","pron":"ped-aw-tsoor'","derivation":"from H6299 (פָּדָה) and H6697 (צוּר); a rock (i.e. God) has ransomed;","strongs_def":"Pedahtsur, an Israelite","kjv_def":"Pedahzur."},
"H6302":{"lemma":"פָּדוּי","xlit":"pâdûwy","pron":"paw-doo'ee","derivation":"passive participle of H6299 (פָּדָה). ransomed (and so occurring under H6299 (פָּדָה));","strongs_def":"as abstractly (in plural masculine) a ransom","kjv_def":"(that are) to be (that were) redeemed."},
"H6303":{"lemma":"פָּדוֹן","xlit":"Pâdôwn","pron":"paw-done'","derivation":"from H6299 (פָּדָה); ransom;","strongs_def":"Padon, one of the Nethinim.","kjv_def":"Padon."},
"H6304":{"lemma":"פְּדוּת","xlit":"pᵉdûwth","pron":"ped-ooth'","derivation":"or פְּדֻת; from H6929 (קֵדְמָה)lemma קדְמָה missing vowel, corrected to קֵדְמָה;","strongs_def":"distinction; also deliverance","kjv_def":"division, redeem, redemption."},
"H6305":{"lemma":"פְּדָיָה","xlit":"Pᵉdâyâh","pron":"ped-aw-yaw'","derivation":"or פְּדָיָהוּ; from H6299 (פָּדָה) and H3050 (יָהּ); Jah has ransomed;","strongs_def":"Pedajah, the name of six Israelites","kjv_def":"Pedaiah."},
"H6306":{"lemma":"פִּדְיוֹם","xlit":"pidyôwm","pron":"pid-yome'","derivation":"or פִּדְיֹם; also פִּדְיוֹן; or פִּדְיֹן; from H6299 (פָּדָה);","strongs_def":"a ransom","kjv_def":"ransom, that were redeemed, redemption."},
"H6307":{"lemma":"פַּדָּן","xlit":"Paddân","pron":"pad-dawn'","derivation":"from an unused root meaning to extend; a plateau; or פַּדַּן אֲרָם; from the same and H758 (אֲרָם); the table-land of Aram;","strongs_def":"Paddan or Paddan-Aram, a region of Syria","kjv_def":"Padan, Padan-aram."},
"H6308":{"lemma":"פָּדַע","xlit":"pâdaʻ","pron":"paw-dah'","derivation":"a primitive root;","strongs_def":"to retrieve","kjv_def":"deliver."},
"H6309":{"lemma":"פֶּדֶר","xlit":"peder","pron":"peh'der","derivation":"from an unused root meaning to be greasy;","strongs_def":"suet","kjv_def":"fat."},
"H6310":{"lemma":"פֶּה","xlit":"peh","pron":"peh","derivation":"from H6284 (פָּאָה);","strongs_def":"the mouth (as the means of blowing), whether literal or figurative (particularly speech); specifically edge, portion or side; adverbially (with preposition) according to","kjv_def":"accord(-ing as, -ing to), after, appointment, assent, collar, command(-ment), [idiom] eat, edge, end, entry, [phrase] file, hole, [idiom] in, mind, mouth, part, portion, [idiom] (should) say(-ing), sentence, skirt, sound, speech, [idiom] spoken, talk, tenor, [idiom] to, [phrase] two-edged, wish, word."},
"H6311":{"lemma":"פֹּה","xlit":"pôh","pron":"po","derivation":"or פֹּא; (Job 38:11), or פּוֹ; probably from a primitive inseparable particle 'p' (of demonstrative force) and H1931 (הוּא);","strongs_def":"this place (French ici), i.e. here or hence","kjv_def":"here, hither, the one (other, this, that) side."},
"H6312":{"lemma":"פּוּאָה","xlit":"Pûwʼâh","pron":"poo-aw'","derivation":"or פֻּוָּה; from H6284 (פָּאָה); a blast;","strongs_def":"Puah or Puvvah, the name of two Israelites","kjv_def":"Phuvah, Pua, Puah."},
"H6313":{"lemma":"פּוּג","xlit":"pûwg","pron":"poog","derivation":"a primitive root;","strongs_def":"to be sluggish","kjv_def":"cease, be feeble, faint, be slacked."},
"H6314":{"lemma":"פּוּגָה","xlit":"pûwgâh","pron":"poo-gaw'","derivation":"from H6313 (פּוּג);","strongs_def":"intermission","kjv_def":"rest."},
"H6315":{"lemma":"פּוּחַ","xlit":"pûwach","pron":"poo'akh","derivation":"a primitive root;","strongs_def":"to puff, i.e. blow with the breath or air; hence, to fan (as a breeze), to utter, to kindle (a fire), to scoff","kjv_def":"blow (upon), break, puff, bring into a snare, speak, utter."},
"H6316":{"lemma":"פּוּט","xlit":"Pûwṭ","pron":"poot","derivation":"of foreign origin;","strongs_def":"Put, a son of Ham, also the name of his descendants or their region, and of a Persian tribe","kjv_def":"Phut, Put."},
"H6317":{"lemma":"פּוּטִיאֵל","xlit":"Pûwṭîyʼêl","pron":"poo-tee-ale'","derivation":"from an unused root (probably meaning to disparage) and H410 (אֵל); contempt of God;","strongs_def":"Putiel, an Israelite","kjv_def":"Putiel."},
"H6318":{"lemma":"פּוֹטִיפַר","xlit":"Pôwṭîyphar","pron":"po-tee-far'","derivation":"of Egyptian derivation;","strongs_def":"Potiphar, an Egyptian","kjv_def":"Potiphar."},
"H6319":{"lemma":"פּוֹטִי פֶרַע","xlit":"Pôwṭîy Pheraʻ","pron":"po-tee feh'-rah","derivation":"of Egyptian derivation;","strongs_def":"Poti-Phera, an Egyptian","kjv_def":"Poti-pherah."},
"H6320":{"lemma":"פּוּךְ","xlit":"pûwk","pron":"pook","derivation":"from an unused root meaning to paint;","strongs_def":"dye (specifically, stibium for the eyes)","kjv_def":"fair colours, glistering, paint(-ed) (-ing)."},
"H6321":{"lemma":"פּוֹל","xlit":"pôwl","pron":"pole","derivation":"from an unused root meaning to be thick;","strongs_def":"a bean (as plump)","kjv_def":"beans."},
"H6322":{"lemma":"פּוּל","xlit":"Pûwl","pron":"pool","derivation":"of foreign origin;","strongs_def":"Pul, the name of an Assyrian king and of an Ethiopian tribe","kjv_def":"Pul."},
"H6323":{"lemma":"פּוּן","xlit":"pûwn","pron":"poon","derivation":"a primitive root meaning to turn, i.e.","strongs_def":"be perplexed","kjv_def":"be distracted."},
"H6324":{"lemma":"פּוּנִי","xlit":"Pûwnîy","pron":"poo-nee'","derivation":"patronymically from an unused name meaning a turn;","strongs_def":"a Punite (collectively) or descendants of an unknown Pun","kjv_def":"Punites."},
"H6325":{"lemma":"פּוּנֹן","xlit":"Pûwnôn","pron":"poo-none'","derivation":"from H6323 (פּוּן); perplexity;","strongs_def":"Punon, a place in the Desert","kjv_def":"Punon."},
"H6326":{"lemma":"פּוּעָה","xlit":"Pûwʻâh","pron":"poo-aw'","derivation":"from an unused root meaning to glitter; brilliancy;","strongs_def":"Puah, an Israelitess","kjv_def":"Puah."},
"H6327":{"lemma":"פּוּץ","xlit":"pûwts","pron":"poots","derivation":"a primitive root;","strongs_def":"to dash in pieces, literally or figuratively (especially to disperse)","kjv_def":"break (dash, shake) in (to) pieces, cast (abroad), disperse (selves), drive, retire, scatter (abroad), spread abroad."},
"H6328":{"lemma":"פּוּק","xlit":"pûwq","pron":"pook","derivation":"a primitive root;","strongs_def":"to waver","kjv_def":"stumble, move."},
"H6329":{"lemma":"פּוּק","xlit":"pûwq","pron":"pook","derivation":"a primitive root (identical with H6328 (פּוּק) through the idea of dropping out; compare H5312 (נְפַק));","strongs_def":"to issue, i.e. furnish; causatively, to secure; figuratively, to succeed","kjv_def":"afford, draw out, further, get, obtain."},
"H6330":{"lemma":"פּוּקָה","xlit":"pûwqâh","pron":"poo-kaw'","derivation":"from H6328 (פּוּק);","strongs_def":"a stumbling-block","kjv_def":"grief."},
"H6331":{"lemma":"פּוּר","xlit":"pûwr","pron":"poor","derivation":"a primitive root;","strongs_def":"to crush","kjv_def":"break, bring to nought, [idiom] utterly take."},
"H6332":{"lemma":"פּוּר","xlit":"Pûwr","pron":"poor","derivation":"also (plural) פּוּרִים; or פֻּרִים; from H6331 (פּוּר);","strongs_def":"a lot (as by means of a broken piece)","kjv_def":"Pur, Purim."},
"H6333":{"lemma":"פּוּרָה","xlit":"pûwrâh","pron":"poo-raw'","derivation":"from H6331 (פּוּר);","strongs_def":"a wine-press (as crushing the grapes)","kjv_def":"winepress."},
"H6334":{"lemma":"פּוֹרָתָא","xlit":"Pôwrâthâʼ","pron":"po-raw-thaw'","derivation":"of Persian origin;","strongs_def":"Poratha, a son of Haman","kjv_def":"Poratha."},
"H6335":{"lemma":"פּוּשׁ","xlit":"pûwsh","pron":"poosh","derivation":"a primitive root;","strongs_def":"to spread; figuratively, act proudly","kjv_def":"grow up, be grown fat, spread selves, be scattered."},
"H6336":{"lemma":"פּוּתִי","xlit":"Pûwthîy","pron":"poo-thee'","derivation":"patronymically from an unused name meaning a hinge;","strongs_def":"a Puthite (collectively) or descendants of an unknown Puth","kjv_def":"Puhites (as if from H6312 (פּוּאָה))."},
"H6337":{"lemma":"פָּז","xlit":"pâz","pron":"pawz","derivation":"from H6338 (פָּזַז);","strongs_def":"pure (gold); hence, gold itself (as refined)","kjv_def":"fine (pure) gold."},
"H6338":{"lemma":"פָּזַז","xlit":"pâzaz","pron":"paw-zaz'","derivation":"a primitive root;","strongs_def":"to refine (gold)","kjv_def":"best (gold)."},
"H6339":{"lemma":"פָּזַז","xlit":"pâzaz","pron":"paw-zaz'","derivation":"a primitive root (identical with H6338 (פָּזַז));","strongs_def":"to solidify (as if by refining); also to spring (as if separating the limbs)","kjv_def":"leap, be made strong."},
"H6340":{"lemma":"פָּזַר","xlit":"pâzar","pron":"paw-zar'","derivation":"a primitive root;","strongs_def":"to scatter, whether in enmity or bounty","kjv_def":"disperse, scatter (abroad)."},
"H6341":{"lemma":"פַּח","xlit":"pach","pron":"pakh","derivation":"from H6351 (פָּחַח);","strongs_def":"a (metallic) sheet (as pounded thin); also a spring net (as spread out like a lamina)","kjv_def":"gin, (thin) plate, snare."},
"H6342":{"lemma":"פָּחַד","xlit":"pâchad","pron":"paw-kkad'","derivation":"a primitive root;","strongs_def":"to be startled (by a sudden alarm); hence, to fear in general","kjv_def":"be afraid, stand in awe, (be in) fear, make to shake."},
"H6343":{"lemma":"פַּחַד","xlit":"pachad","pron":"pakh'-ad","derivation":"from H6342 (פָּחַד);","strongs_def":"a (sudden) alarm (properly, the object feared, by implication, the feeling)","kjv_def":"dread(-ful), fear, (thing) great (fear, -ly feared), terror."},
"H6344":{"lemma":"פַּחַד","xlit":"pachad","pron":"pakh'-ad","derivation":"the same as H6343 (פַּחַד);","strongs_def":"a testicle (as a cause of shame akin to fear)","kjv_def":"stone."},
"H6345":{"lemma":"פַּחְדָּה","xlit":"pachdâh","pron":"pakh-daw'","derivation":"feminine of H6343 (פַּחַד);","strongs_def":"alarm (i.e. awe)","kjv_def":"fear."},
"H6346":{"lemma":"פֶּחָה","xlit":"pechâh","pron":"peh-khaw'","derivation":"of foreign origin;","strongs_def":"a prefect (of a city or small district)","kjv_def":"captain, deputy, governor."},
"H6347":{"lemma":"פֶּחָה","xlit":"pechâh","pron":"peh-khaw'","derivation":"(Aramaic) corresponding to H6346 (פֶּחָה)","strongs_def":"{a prefect (of a city or small district)}","kjv_def":"captain, governor."},
"H6348":{"lemma":"פָּחַז","xlit":"pâchaz","pron":"paw-khaz'","derivation":"a primitive root;","strongs_def":"to bubble up or froth (as boiling water), i.e. (figuratively) to be unimportant","kjv_def":"light."},
"H6349":{"lemma":"פַּחַז","xlit":"pachaz","pron":"pakh'-az","derivation":"from H6348 (פָּחַז);","strongs_def":"ebullition, i.e. froth (figuratively, lust)","kjv_def":"unstable."},
"H6350":{"lemma":"פַּחֲזוּת","xlit":"pachăzûwth","pron":"pakh-az-ooth'","derivation":"from H6348 (פָּחַז);","strongs_def":"frivolity","kjv_def":"lightness."},
"H6351":{"lemma":"פָּחַח","xlit":"pâchach","pron":"paw-khakh'","derivation":"a primitive root; to batter out; but used only as denominative from H6341 (פַּח),","strongs_def":"to spread a net","kjv_def":"be snared."},
"H6352":{"lemma":"פֶּחָם","xlit":"pechâm","pron":"peh-khawm'","derivation":"perhaps from an unused root probably meaning to be black;","strongs_def":"a coal, whether charred or live","kjv_def":"coals."},
"H6353":{"lemma":"פֶּחָר","xlit":"pechâr","pron":"peh-khawr'","derivation":"(Aramaic) from an unused root probably meaning to fashion;","strongs_def":"a potter","kjv_def":"potter."},
"H6354":{"lemma":"פַּחַת","xlit":"pachath","pron":"pakh'-ath","derivation":"probably from an unused root apparently meaning to dig;","strongs_def":"a pit, especially forcatching animals","kjv_def":"hole, pit, snare."},
"H6355":{"lemma":"פַּחַת מוֹאָב","xlit":"Pachath Môwʼâb","pron":"pakh'-ath mo-awb'","derivation":"from H6354 (פַּחַת) and H4124 (מוֹאָב); pit of Moab;","strongs_def":"Pachath-Moab, an Israelite","kjv_def":"Pahathmoab."},
"H6356":{"lemma":"פְּחֶתֶת","xlit":"pᵉchetheth","pron":"pekh-eh'-theth","derivation":"from the same as H6354 (פַּחַת);","strongs_def":"a hole (by mildew in a garment)","kjv_def":"fret inward."},
"H6357":{"lemma":"פִּטְדָה","xlit":"piṭdâh","pron":"pit-daw'","derivation":"of foreign derivation;","strongs_def":"a gem, probably the topaz","kjv_def":"topaz."},
"H6358":{"lemma":"פָּטוּר","xlit":"pâṭûwr","pron":"paw-toor'","derivation":"passive participle of H6362 (פָּטַר);","strongs_def":"opened, i.e. (as noun) a bud","kjv_def":"open."},
"H6359":{"lemma":"פָּטִיר","xlit":"pâṭîyr","pron":"paw-teer'","derivation":"from H6362 (פָּטַר);","strongs_def":"open, i.e. unoccupied","kjv_def":"free."},
"H6360":{"lemma":"פַּטִּישׁ","xlit":"paṭṭîysh","pron":"pat-teesh'","derivation":"intensively from an unused root meaning to pound;","strongs_def":"a hammer","kjv_def":"hammer."},
"H6361":{"lemma":"פַּטִּישׁ","xlit":"paṭṭîysh","pron":"pat-teesh'","derivation":"(Aramaic) from a root corresponding to that of H6360 (פַּטִּישׁ);","strongs_def":"a gown (as if hammered out wide)","kjv_def":"hose."},
"H6362":{"lemma":"פָּטַר","xlit":"pâṭar","pron":"paw-tar'","derivation":"a primitive root;","strongs_def":"to cleave or burst through, i.e. (causatively) to emit, whether literal or figurative (gape)","kjv_def":"dismiss, free, let (shoot) out, slip away."},
"H6363":{"lemma":"פֶּטֶר","xlit":"peṭer","pron":"peh'-ter","derivation":"or פִּטְרָה; from H6362 (פָּטַר);","strongs_def":"a fissure, i.e. (concretely) firstling (as opening the matrix)","kjv_def":"firstling, openeth, such as open."},
"H6364":{"lemma":"פִּי־בֶסֶת","xlit":"Pîy-Beçeth","pron":"pee beh'-seth","derivation":"of Egyptian origin;","strongs_def":"Pi-Beseth, a place in Egypt","kjv_def":"Pi-beseth."},
"H6365":{"lemma":"פִּיד","xlit":"pîyd","pron":"peed","derivation":"from an unused root probably meaning to pierce;","strongs_def":"(figuratively) misfortune","kjv_def":"destruction, ruin."},
"H6366":{"lemma":"פֵּיָה","xlit":"pêyâh","pron":"pay-aw'","derivation":"or פִּיָּה; feminine of H6310 (פֶּה);","strongs_def":"an edge","kjv_def":"(two-) edge(-d)."},
"H6367":{"lemma":"פִּי הַחִירֹת","xlit":"Pîy ha-Chîyrôth","pron":"pee hah-khee-roth'","derivation":"from H6310 (פֶּה) and the feminine plural of a noun (from the same root as H2356 (חוֹר)), with the article interpolated; mouth of the gorges;","strongs_def":"Pi-ha-Chiroth, a place in Egypt","kjv_def":"Pi-hahiroth. (In Numbers 14:19 without Pi-.)"},
"H6368":{"lemma":"פִּיחַ","xlit":"pîyach","pron":"pee'-akh","derivation":"from H6315 (פּוּחַ);","strongs_def":"a powder (as easily puffed away), i.e. ashes or dust","kjv_def":"ashes."},
"H6369":{"lemma":"פִּיכֹל","xlit":"Pîykôl","pron":"pee-kole'","derivation":"apparently from H6310 (פֶּה) and H3605 (כֹּל); mouth of all;","strongs_def":"Picol, a Philistine","kjv_def":"Phichol."},
"H6370":{"lemma":"פִּילֶגֶשׁ","xlit":"pîylegesh","pron":"pee-leh'-ghesh","derivation":"or פִּלֶגֶשׁ; of uncertain derivation;","strongs_def":"a concubine; also (masculine) a paramour","kjv_def":"concubine, paramour."},
"H6371":{"lemma":"פִּימָה","xlit":"pîymâh","pron":"pee-maw'","derivation":"probably from an unused root meaning to be plump;","strongs_def":"obesity","kjv_def":"collops."},
"H6372":{"lemma":"פִּינְחָס","xlit":"Pîynᵉchâç","pron":"pee-nekh-aws'","derivation":"apparently from H6310 (פֶּה) and a variation of H5175 (נָחָשׁ); mouth of a serpent;","strongs_def":"Pinechas, the name of three Israelites","kjv_def":"Phinehas."},
"H6373":{"lemma":"פִּינֹן","xlit":"pîynôn","pron":"pee-none'","derivation":"probably the same as H6325 (פּוּנֹן);","strongs_def":"Pinon, an Idumaean","kjv_def":"Pinon."},
"H6374":{"lemma":"פִּיפִיָּה","xlit":"pîyphîyâh","pron":"pee-fee-yaw'","derivation":"for H6366 (פֵּיָה);","strongs_def":"an edge or tooth","kjv_def":"tooth, [idiom] two-edged."},
"H6375":{"lemma":"פִּיק","xlit":"pîyq","pron":"peek","derivation":"from H6329 (פּוּק);","strongs_def":"a tottering","kjv_def":"smite together."},
"H6376":{"lemma":"פִּישׁוֹן","xlit":"Pîyshôwn","pron":"pee-shone'","derivation":"from H6335 (פּוּשׁ); dispersive;","strongs_def":"Pishon, a river of Eden","kjv_def":"Pison."},
"H6377":{"lemma":"פִּיתוֹן","xlit":"Pîythôwn","pron":"pee-thone'","derivation":"probably from the same as H6596 (פֹּת); expansive;","strongs_def":"Pithon, an Israelite","kjv_def":"Pithon."},
"H6378":{"lemma":"פַּךְ","xlit":"pak","pron":"pak","derivation":"from H6379 (פָּכָה);","strongs_def":"a flask (from which a liquid may flow)","kjv_def":"box, vial."},
"H6379":{"lemma":"פָּכָה","xlit":"pâkâh","pron":"paw-kaw'","derivation":"a primitive root;","strongs_def":"to pour","kjv_def":"run out."},
"H6380":{"lemma":"פֹּכֶרֶת צְבָיִים","xlit":"Pôkereth Tsᵉbâyîym","pron":"po-keh'-reth tseb-awyeem'","derivation":"from the active participle (of the same form as the first word) feminine of an unused root (meaning to entrap) and plural of H6643 (צְבִי); trap of gazelles;","strongs_def":"Pokereth-Tsebajim, one of the 'servants of Solomon'","kjv_def":"Pochereth of Zebaim."},
"H6381":{"lemma":"פָּלָא","xlit":"pâlâʼ","pron":"paw-law'","derivation":"a primitive root;","strongs_def":"properly, perhaps to separate, i.e. distinguish (literally or figuratively); by implication, to be (causatively, make) great, difficult, wonderful","kjv_def":"accomplish, (arise...too, be too) hard, hidden, things too high, (be, do, do a, shew) marvelous(-ly, -els, things, work), miracles, perform, separate, make singular, (be, great, make) wonderful(-ers, -ly, things, works), wondrous (things, works, -ly)."},
"H6382":{"lemma":"פֶּלֶא","xlit":"peleʼ","pron":"peh'-leh","derivation":"from H6381 (פָּלָא);","strongs_def":"a miracle","kjv_def":"marvellous thing, wonder(-ful, -fully)."},
"H6383":{"lemma":"פִּלְאִי","xlit":"pilʼîy","pron":"pil-ee'","derivation":"or פָּלִיא; from H6381 (פָּלָא);","strongs_def":"remarkable","kjv_def":"secret, wonderful."},
"H6384":{"lemma":"פַּלֻּאִי","xlit":"Palluʼîy","pron":"pal-loo-ee'","derivation":"patronymically from H6396 (פַּלּוּא);","strongs_def":"a Palluite (collectively) or descendants of Pallu","kjv_def":"Palluites."},
"H6385":{"lemma":"פָּלַג","xlit":"pâlag","pron":"paw-lag'","derivation":"a primitive root;","strongs_def":"to split (literally or figuratively)","kjv_def":"divide."},
"H6386":{"lemma":"פְּלַג","xlit":"pᵉlag","pron":"pel-ag'","derivation":"(Aramaic) corresponding to H6385 (פָּלַג)","strongs_def":"{to split (literally or figuratively)}","kjv_def":"divided."},
"H6387":{"lemma":"פְּלַג","xlit":"pᵉlag","pron":"pel-ag'","derivation":"(Aramaic) from H6386 (פְּלַג);","strongs_def":"a half","kjv_def":"dividing."},
"H6388":{"lemma":"פֶּלֶג","xlit":"peleg","pron":"peh'-leg","derivation":"from H6385 (פָּלַג);","strongs_def":"a rill (i.e. small channel of water, as in irrigation)","kjv_def":"river, stream."},
"H6389":{"lemma":"פֶּלֶג","xlit":"Peleg","pron":"peh'-leg","derivation":"the same as H6388 (פֶּלֶג); earthquake;","strongs_def":"Peleg, a son of Shem","kjv_def":"Peleg."},
"H6390":{"lemma":"פְּלַגָּה","xlit":"pᵉlaggâh","pron":"pel-ag-gaw'","derivation":"from H6385 (פָּלַג);","strongs_def":"a runlet, i.e. gully","kjv_def":"division, river."},
"H6391":{"lemma":"פְּלֻגָּה","xlit":"pᵉluggâh","pron":"pel-oog-gaw'","derivation":"from H6385 (פָּלַג);","strongs_def":"a section","kjv_def":"division."},
"H6392":{"lemma":"פְּלֻגָּה","xlit":"pᵉluggâh","pron":"pel-oog-gaw'","derivation":"(Aramaic) corresponding to H6391 (פְּלֻגָּה)","strongs_def":"{a section}","kjv_def":"division."},
"H6393":{"lemma":"פְּלָדָה","xlit":"pᵉlâdâh","pron":"pel-aw-daw'","derivation":"from an unused root meaning to divide;","strongs_def":"a cleaver, i.e. iron armature (of a chariot)","kjv_def":"torch."},
"H6394":{"lemma":"פִּלְדָּשׁ","xlit":"Pildâsh","pron":"pil-dawsh'","derivation":"of uncertain derivation;","strongs_def":"Pildash, a relative of Abraham","kjv_def":"Pildash."},
"H6395":{"lemma":"פָּלָה","xlit":"pâlâh","pron":"paw-law'","derivation":"a primitive root;","strongs_def":"to distinguish (literally or figuratively)","kjv_def":"put a difference, show marvellous, separate, set apart, sever, make wonderfully."},
"H6396":{"lemma":"פַּלּוּא","xlit":"Pallûwʼ","pron":"pal-loo'","derivation":"from H6395 (פָּלָה); distinguished;","strongs_def":"Pallu, an Israelite","kjv_def":"Pallu, Phallu."},
"H6397":{"lemma":"פְּלוֹנִי","xlit":"Pᵉlôwnîy","pron":"pel-o-nee'","derivation":"patronymically from an unused name (from H6395 (פָּלָה)) meaning separate;","strongs_def":"a Pelonite or inhabitant of an unknown Palon","kjv_def":"Pelonite."},
"H6398":{"lemma":"פָּלַח","xlit":"pâlach","pron":"paw-lakh'","derivation":"a primitive root;","strongs_def":"to slice, i.e. break open or pierce","kjv_def":"bring forth, cleave, cut, shred, strike through."},
"H6399":{"lemma":"פְּלַח","xlit":"pᵉlach","pron":"pel-akh'","derivation":"(Aramaic) corresponding to H6398 (פָּלַח);","strongs_def":"to serve or worship","kjv_def":"minister, serve."},
"H6400":{"lemma":"פֶּלַח","xlit":"pelach","pron":"peh'-lakh","derivation":"from H6398 (פָּלַח);","strongs_def":"a slice","kjv_def":"piece."},
"H6401":{"lemma":"פִּלְחָא","xlit":"Pilchâʼ","pron":"pil-khaw'","derivation":"from H6400 (פֶּלַח); slicing;","strongs_def":"Pilcha, an Israelite","kjv_def":"Pilcha."},
"H6402":{"lemma":"פׇּלְחָן","xlit":"polchân","pron":"pol-khawn'","derivation":"(Aramaic) from H6399 (פְּלַח);","strongs_def":"worship","kjv_def":"service."},
"H6403":{"lemma":"פָּלַט","xlit":"pâlaṭ","pron":"paw-lat'","derivation":"a primitive root;","strongs_def":"to slip out, i.e. escape; causatively, to deliver","kjv_def":"calve, carry away safe, deliver, (cause to) escape."},
"H6404":{"lemma":"פֶּלֶט","xlit":"Peleṭ","pron":"peh'-let","derivation":"from H6403 (פָּלַט); escape;","strongs_def":"Pelet, the name of two Israelites","kjv_def":"Pelet. See also H1046 (בֵּית פֶּלֶט)."},
"H6405":{"lemma":"פַּלֵּט","xlit":"pallêṭ","pron":"pal-late'","derivation":"from H6403 (פָּלַט);","strongs_def":"escape","kjv_def":"deliverance, escape."},
"H6406":{"lemma":"פַּלְטִי","xlit":"Palṭîy","pron":"pal-tee'","derivation":"from H6403 (פָּלַט); delivered;","strongs_def":"Palti, the name of two Israelites","kjv_def":"Palti, Phalti."},
"H6407":{"lemma":"פַּלְטִי","xlit":"Palṭîy","pron":"pal-tee'","derivation":"patronymically from H6406 (פַּלְטִי);","strongs_def":"a Paltite or descendant of Palti","kjv_def":"Paltite."},
"H6408":{"lemma":"פִּלְטַי","xlit":"Pilṭay","pron":"pil-tah'-ee","derivation":"for H6407 (פַּלְטִי);","strongs_def":"Piltai, an Israelite","kjv_def":"Piltai."},
"H6409":{"lemma":"פַּלְטִיאֵל","xlit":"Palṭîyʼêl","pron":"pal-tee-ale'","derivation":"from the same as H6404 (פֶּלֶט) and H410 (אֵל); deliverance of God;","strongs_def":"Paltiel, the name of two Israelites","kjv_def":"Paltiel, Phaltiel."},
"H6410":{"lemma":"פְּלַטְיָה","xlit":"Pᵉlaṭyâh","pron":"pel-at-yaw'","derivation":"or פְּלַטְיָהוּ; from H6403 (פָּלַט) and H3050 (יָהּ); Jah has delivered;","strongs_def":"Pelatjah, the name of four Israelites","kjv_def":"Pelatiah."},
"H6411":{"lemma":"פְּלָיָה","xlit":"Pᵉlâyâh","pron":"pel-aw-yaw'","derivation":"or פְּלָאיָה; from H6381 (פָּלָא) and H3050 (יָהּ); Jah has distinguished;","strongs_def":"Pelajah, the name of three Israelites","kjv_def":"Pelaiah."},
"H6412":{"lemma":"פָּלִיט","xlit":"pâlîyṭ","pron":"paw-leet'","derivation":"or פָּלֵיט; or פָּלֵט; from H6403 (פָּלַט);","strongs_def":"a refugee","kjv_def":"(that have) escape(-d, -th), fugitive."},
"H6413":{"lemma":"פְּלֵיטָה","xlit":"pᵉlêyṭâh","pron":"pel-ay-taw'","derivation":"or פְּלֵטָה; feminine of H6412 (פָּלִיט);","strongs_def":"deliverance; concretely, an escaped portion","kjv_def":"deliverance, (that is) escape(-d), remnant."},
"H6414":{"lemma":"פָּלִיל","xlit":"pâlîyl","pron":"paw-leel'","derivation":"from H6419 (פָּלַל);","strongs_def":"a magistrate","kjv_def":"judge."},
"H6415":{"lemma":"פְּלִילָה","xlit":"pᵉlîylâh","pron":"pel-ee-law'","derivation":"feminine of H6414 (פָּלִיל);","strongs_def":"justice","kjv_def":"judgment."},
"H6416":{"lemma":"פְּלִילִי","xlit":"pᵉlîylîy","pron":"pel-ee-lee'","derivation":"from H6414 (פָּלִיל);","strongs_def":"judicial","kjv_def":"judge."},
"H6417":{"lemma":"פְּלִילִיָּה","xlit":"pᵉlîylîyâh","pron":"pel-ee-lee-yaw'","derivation":"feminine of H6416 (פְּלִילִי);","strongs_def":"judicature","kjv_def":"judgment."},
"H6418":{"lemma":"פֶּלֶךְ","xlit":"pelek","pron":"peh'-lek","derivation":"from an unused root meaning to be round;","strongs_def":"a circuit (i.e. district); also a spindle (as whirled); hence, a crutch","kjv_def":"(di-) staff, participle"},
"H6419":{"lemma":"פָּלַל","xlit":"pâlal","pron":"paw-lal'","derivation":"a primitive root;","strongs_def":"to judge (officially or mentally); by extension, to intercede, pray","kjv_def":"intreat, judge(-ment), (make) pray(-er, -ing), make supplication."},
"H6420":{"lemma":"פָּלָל","xlit":"Pâlâl","pron":"paw-lawl'","derivation":"from H6419 (פָּלַל); judge;","strongs_def":"Palal, an Israelite","kjv_def":"Palal."},
"H6421":{"lemma":"פְּלַלְיָה","xlit":"Pᵉlalyâh","pron":"pel-al-yaw'","derivation":"from H6419 (פָּלַל) and H3050 (יָהּ); Jah has judged;","strongs_def":"Pelaljah, an Israelite","kjv_def":"Pelaliah."},
"H6422":{"lemma":"פַּלְמוֹנִי","xlit":"palmôwnîy","pron":"pal-mo-nee'","derivation":"probably for H6423 (פְּלֹנִי);","strongs_def":"a certain one, i.e. so-and-so","kjv_def":"certain."},
"H6423":{"lemma":"פְּלֹנִי","xlit":"pᵉlônîy","pron":"pel-o-nee'","derivation":"from H6395 (פָּלָה);","strongs_def":"such a one, i.e. a specified person","kjv_def":"such."},
"H6424":{"lemma":"פָּלַס","xlit":"pâlaç","pron":"paw-las'","derivation":"a primitive root;","strongs_def":"properly, to roll flat, i.e. prepare (a road); also to revolve, i.e. weigh (mentally)","kjv_def":"make, ponder, weigh."},
"H6425":{"lemma":"פֶּלֶס","xlit":"peleç","pron":"peh'-les","derivation":"from H6424 (פָּלַס);","strongs_def":"a balance","kjv_def":"scales, weight."},
"H6426":{"lemma":"פָּלַץ","xlit":"pâlats","pron":"paw-lats'","derivation":"a primitive root;","strongs_def":"properly, perhaps to rend, i.e. (by implication) to quiver","kjv_def":"tremble."},
"H6427":{"lemma":"פַּלָּצוּת","xlit":"pallâtsûwth","pron":"pal-law-tsooth'","derivation":"from H6426 (פָּלַץ);","strongs_def":"affright","kjv_def":"fearfulness, horror, trembling."},
"H6428":{"lemma":"פָּלַשׁ","xlit":"pâlash","pron":"paw-lash'","derivation":"a primitive root;","strongs_def":"to roll (in dust)","kjv_def":"roll (wallow) self."},
"H6429":{"lemma":"פְּלֶשֶׁת","xlit":"Pᵉlesheth","pron":"pel-eh'-sheth","derivation":"from H6428 (פָּלַשׁ); rolling, i.e. migratory;","strongs_def":"Pelesheth, a region of Syria","kjv_def":"Palestina, Palestine, Philistia, Philistines."},
"H6430":{"lemma":"פְּלִשְׁתִּי","xlit":"Pᵉlishtîy","pron":"pel-ish-tee'","derivation":"patrial from H6429 (פְּלֶשֶׁת);","strongs_def":"a Pelishtite or inhabitant of Pelesheth","kjv_def":"Philistine."},
"H6431":{"lemma":"פֶּלֶת","xlit":"Peleth","pron":"peh'-leth","derivation":"from an unused root meaning to flee; swiftness;","strongs_def":"Peleth, the name of two Israelites","kjv_def":"Peleth."},
"H6432":{"lemma":"פְּלֵתִי","xlit":"Pᵉlêthîy","pron":"pel-ay-thee'","derivation":"from the same form as H6431 (פֶּלֶת);","strongs_def":"a courier (collectively) or official messenger","kjv_def":"Pelethites."},
"H6433":{"lemma":"פֻּם","xlit":"pum","pron":"poom","derivation":"(Aramaic) probably for H6310 (פֶּה);","strongs_def":"the mouth (literally or figuratively)","kjv_def":"mouth."},
"H6434":{"lemma":"פֵּן","xlit":"pên","pron":"pane","derivation":"from an unused root meaning to turn;","strongs_def":"an angle (of a street or wall)","kjv_def":"corner."},
"H6435":{"lemma":"פֵּן","xlit":"pên","pron":"pane","derivation":"from H6437 (פָּנָה);","strongs_def":"properly, removal; used only (in the construction) adverb as conjunction, lest","kjv_def":"(lest) (peradventure), that...not."},
"H6436":{"lemma":"פַּנַּג","xlit":"pannag","pron":"pan-nag'","derivation":"of uncertain derivation;","strongs_def":"probably pastry","kjv_def":"Pannag."},
"H6437":{"lemma":"פָּנָה","xlit":"pânâh","pron":"paw-naw'","derivation":"a primitive root;","strongs_def":"to turn; by implication, to face, i.e. appear, look, etc.","kjv_def":"appear, at (even-) tide, behold, cast out, come on, [idiom] corner, dawning, empty, go away, lie, look, mark, pass away, prepare, regard, (have) respect (to), (re-) turn (aside, away, back, face, self), [idiom] right (early)."},
"H6438":{"lemma":"פִּנָּה","xlit":"pinnâh","pron":"pin-naw'","derivation":"feminine of H6434 (פֵּן);","strongs_def":"an angle; by implication, a pinnacle; figuratively, a chieftain","kjv_def":"bulwark, chief, corner, stay, tower."},
"H6439":{"lemma":"פְּנוּאֵל","xlit":"Pᵉnûwʼêl","pron":"pen-oo-ale'","derivation":"or (more properly,) פְּנִיאֵל; from H6437 (פָּנָה) and H410 (אֵל); face of God;","strongs_def":"Penuel or Peniel, a place East of Jordan; also (as Penuel) the name of two Israelites","kjv_def":"Peniel, Penuel."},
"H6440":{"lemma":"פָּנִים","xlit":"pânîym","pron":"paw-neem'","derivation":"plural (but always as singular) of an unused noun פָּנֶה; from H6437 (פָּנָה));","strongs_def":"the face (as the part that turns); used in a great variety of applications (literally and figuratively); also (with prepositional prefix) as a preposition (before, etc.)","kjv_def":"[phrase] accept, a-(be-) fore(-time), against, anger, [idiom] as (long as), at, [phrase] battle, [phrase] because (of), [phrase] beseech, countenance, edge, [phrase] employ, endure, [phrase] enquire, face, favour, fear of, for, forefront(-part), form(-er time, -ward), from, front, heaviness, [idiom] him(-self), [phrase] honourable, [phrase] impudent, [phrase] in, it, look(-eth) (-s), [idiom] me, [phrase] meet, [idiom] more than, mouth, of, off, (of) old (time), [idiom] on, open, [phrase] out of, over against, the partial, person, [phrase] please, presence, propect, was purposed, by reason of, [phrase] regard, right forth, [phrase] serve, [idiom] shewbread, sight, state, straight, [phrase] street, [idiom] thee, [idiom] them(-selves), through ([phrase] -out), till, time(-s) past, (un-) to(-ward), [phrase] upon, upside ([phrase] down), with(-in, [phrase] -stand), [idiom] ye, [idiom] you."},
"H6441":{"lemma":"פְּנִימָה","xlit":"pᵉnîymâh","pron":"pen-ee'-maw","derivation":"from H6440 (פָּנִים) with directive enclitic;","strongs_def":"faceward, i.e. indoors","kjv_def":"(with-) in(-ner part, -ward)."},
"H6442":{"lemma":"פְּנִימִי","xlit":"pᵉnîymîy","pron":"pen-ee-mee'","derivation":"from H6440 (פָּנִים);","strongs_def":"interior","kjv_def":"(with-) in(-ner, -ward)."},
"H6443":{"lemma":"פָּנִין","xlit":"pânîyn","pron":"paw-neen'","derivation":"or פָּנִי; from the same as H6434 (פֵּן);","strongs_def":"probably a pearl (as round)","kjv_def":"ruby."},
"H6444":{"lemma":"פְּנִנָּה","xlit":"Pᵉninnâh","pron":"pen-in-naw'","derivation":"probably feminine from H6443 (פָּנִין) contr.;","strongs_def":"Peninnah, an Israelitess","kjv_def":"Peninnah."},
"H6445":{"lemma":"פָּנַק","xlit":"pânaq","pron":"paw-nak'","derivation":"a primitive root;","strongs_def":"to enervate","kjv_def":"bring up."},
"H6446":{"lemma":"פַּס","xlit":"paç","pron":"pas","derivation":"from H6461 (פָּסַס); properly, the palm (of the hand) or sole (of the foot) (compare H6447 (פַּס)); by implication (plural)","strongs_def":"a long and sleeved tunic (perhaps simply a wide one; from the original sense of the root, i.e. of many breadths)","kjv_def":"(divers) colours."},
"H6447":{"lemma":"פַּס","xlit":"paç","pron":"pas","derivation":"(Aramaic) from a root corresponding to H6461 (פָּסַס);","strongs_def":"the palm (of the hand, as being spread out)","kjv_def":"participle"},
"H6448":{"lemma":"פָּסַג","xlit":"pâçag","pron":"paw-sag'","derivation":"a primitive root;","strongs_def":"to cut up, i.e. (figuratively) contemplate","kjv_def":"consider."},
"H6449":{"lemma":"פִּסְגָּה","xlit":"Piçgâh","pron":"pis-gaw'","derivation":"from H6448 (פָּסַג); a cleft;","strongs_def":"Pisgah, a Mountain East of Jordan","kjv_def":"Pisgah."},
"H6450":{"lemma":"פַּס דַּמִּים","xlit":"Paç Dammîym","pron":"pas dam-meem'","derivation":"from H6446 (פַּס) and the plural of H1818 (דָּם); palm (i.e. dell) of bloodshed;","strongs_def":"Pas-Dammim, a place in Palestine","kjv_def":"Pas-dammim. Compare H658 (אֶפֶס דַּמִּים)."},
"H6451":{"lemma":"פִּסָּה","xlit":"piççâh","pron":"pis-saw'","derivation":"from H6461 (פָּסַס);","strongs_def":"expansion, i.e. abundance","kjv_def":"handful."},
"H6452":{"lemma":"פָּסַח","xlit":"pâçach","pron":"paw-sakh'","derivation":"a primitive root;","strongs_def":"to hop, i.e. (figuratively) skip over (or spare); by implication, to hesitate; also (literally) to limp, to dance","kjv_def":"halt, become lame, leap, pass over."},
"H6453":{"lemma":"פֶּסַח","xlit":"peçach","pron":"peh'-sakh","derivation":"from H6452 (פָּסַח);","strongs_def":"a pretermission, i.e. exemption; used only techically of the Jewish Passover (the festival or the victim)","kjv_def":"passover (offering)."},
"H6454":{"lemma":"פָּסֵחַ","xlit":"Pâçêach","pron":"paw-say'-akh","derivation":"from H6452 (פָּסַח); limping;","strongs_def":"Paseach, the name of two Israelites","kjv_def":"Paseah, Phaseah."},
"H6455":{"lemma":"פִּסֵּחַ","xlit":"piççêach","pron":"pis-say'-akh","derivation":"from H6452 (פָּסַח);","strongs_def":"lame","kjv_def":"lame."},
"H6456":{"lemma":"פְּסִיל","xlit":"pᵉçîyl","pron":"pes-eel'","derivation":"from H6458 (פָּסַל);","strongs_def":"an idol","kjv_def":"carved (graven) image, quarry."},
"H6457":{"lemma":"פָּסַךְ","xlit":"Pâçak","pron":"paw-sak'","derivation":"from an unused root meaning to divide; divider;","strongs_def":"Pasak, an Israelite","kjv_def":"Pasach."},
"H6458":{"lemma":"פָּסַל","xlit":"pâçal","pron":"paw-sal'","derivation":"a primitive root;","strongs_def":"to carve, whether wood or stone","kjv_def":"grave, hew."},
"H6459":{"lemma":"פֶּסֶל","xlit":"peçel","pron":"peh'-sel","derivation":"from H6458 (פָּסַל);","strongs_def":"an idol","kjv_def":"carved (graven) image."},
"H6460":{"lemma":"פְּסַנְטֵרִין","xlit":"pᵉçanṭêrîyn","pron":"pes-an-tay-reen'","derivation":"(Aramaic) or פְּסַנְתֵּרִין; a transliteration of the Greek,","strongs_def":"psalterion; a lyre","kjv_def":"psaltery."},
"H6461":{"lemma":"פָּסַס","xlit":"pâçaç","pron":"paw-sas'","derivation":"a primitive root;","strongs_def":"probably to disperse, i.e. (intransitive) disappear","kjv_def":"cease."},
"H6462":{"lemma":"פִּסְפָּה","xlit":"Piçpâh","pron":"pis-paw'","derivation":"perhaps from H6461 (פָּסַס); dispersion;","strongs_def":"Pispah, an Israelite","kjv_def":"Pispah."},
"H6463":{"lemma":"פָּעָה","xlit":"pâʻâh","pron":"paw-aw'","derivation":"a primitive root;","strongs_def":"to scream","kjv_def":"cry."},
"H6464":{"lemma":"פָּעוּ","xlit":"Pâʻûw","pron":"paw-oo'","derivation":"or פָּעִי; from H6463 (פָּעָה); screaming;","strongs_def":"Pau or Pai, a place in Edom","kjv_def":"Pai, Pau."},
"H6465":{"lemma":"פְּעוֹר","xlit":"Pᵉʻôwr","pron":"peh-ore'","derivation":"from H6473 (פָּעַר); a gap; also H1187 (בַּעַל פְּעוֹר))","strongs_def":"Peor, a mountain East of Jordan; Baal Peor,  (for a deity worshipped there","kjv_def":"Peor. See also H1047 (בֵּית פְּעוֹר)."},
"H6466":{"lemma":"פָּעַל","xlit":"pâʻal","pron":"paw-al'","derivation":"a primitive root;","strongs_def":"to do or make (systematically and habitually), especially to practise","kjv_def":"commit, (evil-) do(-er), make(-r), ordain, work(-er)."},
"H6467":{"lemma":"פֹּעַל","xlit":"pôʻal","pron":"po'-al","derivation":"from H6466 (פָּעַל);","strongs_def":"an act or work (concretely)","kjv_def":"act, deed, do, getting, maker, work."},
"H6468":{"lemma":"פְּעֻלָּה","xlit":"pᵉʻullâh","pron":"peh-ool-law'","derivation":"feminine passive participle of H6466 (פָּעַל);","strongs_def":"(abstractly) work","kjv_def":"labour, reward, wages, work."},
"H6469":{"lemma":"פְּעֻלְּתַי","xlit":"Pᵉʻullᵉthay","pron":"peh-ool-leh-thah'-ee","derivation":"from H6468 (פְּעֻלָּה); laborious;","strongs_def":"Peullethai, an Israelite","kjv_def":"Peulthai."},
"H6470":{"lemma":"פָּעַם","xlit":"pâʻam","pron":"paw-am'","derivation":"a primitive root;","strongs_def":"to tap, i.e. beat regularly; hence (generally) to impel or agitate","kjv_def":"move, trouble."},
"H6471":{"lemma":"פַּעַם","xlit":"paʻam","pron":"pah'-am","derivation":"or (feminine) פַּעֲמָה; from H6470 (פָּעַם);","strongs_def":"a stroke, literally or figuratively (in various applications, as follow)","kjv_def":"anvil, corner, foot(-step), going, (hundred-) fold, [idiom] now, (this) [phrase] once, order, rank, step, [phrase] thrice, (often-), second, this, two) time(-s), twice, wheel."},
"H6472":{"lemma":"פַּעֲמֹן","xlit":"paʻămôn","pron":"pah-am-one'","derivation":"from H6471 (פַּעַם);","strongs_def":"a bell (as struck)","kjv_def":"bell."},
"H6473":{"lemma":"פָּעַר","xlit":"pâʻar","pron":"paw-ar'","derivation":"a primitive root;","strongs_def":"to yawn, i.e. open wide (literally or figuratively)","kjv_def":"gape, open (wide)."},
"H6474":{"lemma":"פַּעֲרַי","xlit":"Paʻăray","pron":"pah-ar-ah'-ee","derivation":"from H6473 (פָּעַר); yawning;","strongs_def":"Paarai, an Israelite","kjv_def":"Paarai."},
"H6475":{"lemma":"פָּצָה","xlit":"pâtsâh","pron":"paw-tsaw'","derivation":"a primitive root;","strongs_def":"to rend, i.e. open (especially the mouth)","kjv_def":"deliver, gape, open, rid, utter."},
"H6476":{"lemma":"פָּצַח","xlit":"pâtsach","pron":"paw-tsakh'","derivation":"a primitive root;","strongs_def":"to break out (in joyful sound)","kjv_def":"break (forth, forth into joy), make a loud noise."},
"H6477":{"lemma":"פְּצִירָה","xlit":"pᵉtsîyrâh","pron":"pets-ee-raw'","derivation":"from H6484 (פָּצַר);","strongs_def":"bluntness","kjv_def":"[phrase] file."},
"H6478":{"lemma":"פָּצַל","xlit":"pâtsal","pron":"paw-tsal'","derivation":"a primitive root;","strongs_def":"to peel","kjv_def":"pill."},
"H6479":{"lemma":"פְּצָלָה","xlit":"pᵉtsâlâh","pron":"pets-aw-law'","derivation":"from H6478 (פָּצַל);","strongs_def":"a peeling","kjv_def":"strake."},
"H6480":{"lemma":"פָּצַם","xlit":"pâtsam","pron":"paw-tsam'","derivation":"a primitive root;","strongs_def":"to rend (by earthquake)","kjv_def":"break."},
"H6481":{"lemma":"פָּצַע","xlit":"pâtsaʻ","pron":"paw-tsah'","derivation":"a primitive root;","strongs_def":"to split, i.e. wound","kjv_def":"wound."},
"H6482":{"lemma":"פֶּצַע","xlit":"petsaʻ","pron":"peh'-tsah","derivation":"from H6481 (פָּצַע);","strongs_def":"a wound","kjv_def":"wound(-ing)."},
"H6483":{"lemma":"פִּצֵּץ","xlit":"Pitstsêts","pron":"pits-tsates'","derivation":"from an unused root meaning to dissever; dispersive;","strongs_def":"Pitstsets, a priest","kjv_def":"Apses (including the article)."},
"H6484":{"lemma":"פָּצַר","xlit":"pâtsar","pron":"paw-tsar'","derivation":"a primitive root;","strongs_def":"to peck at, i.e. (figuratively) stun or dull","kjv_def":"press, urge, stubbornness."},
"H6485":{"lemma":"פָּקַד","xlit":"pâqad","pron":"paw-kad'","derivation":"a primitive root;","strongs_def":"to visit (with friendly or hostile intent); by analogy, to oversee, muster, charge, care for, miss, deposit, etc.","kjv_def":"appoint, [idiom] at all, avenge, bestow, (appoint to have the, give a) charge, commit, count, deliver to keep, be empty, enjoin, go see, hurt, do judgment, lack, lay up, look, make, [idiom] by any means, miss, number, officer, (make) overseer, have (the) oversight, punish, reckon, (call to) remember(-brance), set (over), sum, [idiom] surely, visit, want."},
"H6486":{"lemma":"פְּקֻדָּה","xlit":"pᵉquddâh","pron":"pek-ood-daw'","derivation":"feminine passive participle of H6485 (פָּקַד);","strongs_def":"visitation (in many senses, chiefly official)","kjv_def":"account, (that have the) charge, custody, that which...laid up, numbers, office(-r), ordering, oversight, [phrase] prison, reckoning, visitation."},
"H6487":{"lemma":"פִּקָּדוֹן","xlit":"piqqâdôwn","pron":"pik-kaw-done'","derivation":"from H6485 (פָּקַד);","strongs_def":"a deposit","kjv_def":"that which was delivered (to keep), store."},
"H6488":{"lemma":"פְּקִדֻת","xlit":"pᵉqiduth","pron":"pek-ee-dooth'","derivation":"from H6496 (פָּקִיד);","strongs_def":"supervision","kjv_def":"ward."},
"H6489":{"lemma":"פְּקוֹד","xlit":"Pᵉqôwd","pron":"pek-ode'","derivation":"from H6485 (פָּקַד); punishment;","strongs_def":"Pekod, a symbolic name for Babylon","kjv_def":"Pekod."},
"H6490":{"lemma":"פִּקּוּד","xlit":"piqqûwd","pron":"pik-kood'","derivation":"or פִּקֻּד; from H6485 (פָּקַד);","strongs_def":"properly, appointed, i.e. a mandate (of God; plural only, collectively, for the Law)","kjv_def":"commandment, precept, statute."},
"H6491":{"lemma":"פָּקַח","xlit":"pâqach","pron":"paw-kakh'","derivation":"a primitive root;","strongs_def":"to open (the senses, especially the eyes); figuratively, to be observant","kjv_def":"open."},
"H6492":{"lemma":"פֶּקַח","xlit":"Peqach","pron":"peh'-kakh","derivation":"from H6491 (פָּקַח); watch;","strongs_def":"Pekach, an Israelite king","kjv_def":"Pekah."},
"H6493":{"lemma":"פִּקֵּחַ","xlit":"piqqêach","pron":"pik-kay'-akh","derivation":"from H6491 (פָּקַח);","strongs_def":"clear-sighted; figuratively, intelligent","kjv_def":"seeing, wise."},
"H6494":{"lemma":"פְּקַחְיָה","xlit":"Pᵉqachyâh","pron":"pek-akh-yaw'","derivation":"from H6491 (פָּקַח) and H3050 (יָהּ); Jah has observed;","strongs_def":"Pekachjah, an Israelite king","kjv_def":"Pekahiah."},
"H6495":{"lemma":"פְּקַח־קוֹחַ","xlit":"pᵉqach-qôwach","pron":"pek-akh-ko'-akh","derivation":"from H6491 (פָּקַח) redoubled;","strongs_def":"opening (of a dungeon), i.e. jail-delivery (figuratively, salvation for sin)","kjv_def":"opening of the prison."},
"H6496":{"lemma":"פָּקִיד","xlit":"pâqîyd","pron":"paw-keed'","derivation":"from H6485 (פָּקַד);","strongs_def":"a superintendent (civil, military or religious)","kjv_def":"which had the charge, governor, office, overseer, (that) was set."},
"H6497":{"lemma":"פֶּקַע","xlit":"peqaʻ","pron":"peh'-kah","derivation":"from an unused root meaning to burst; only used as an architectural term of an ornament similar to H6498 (פַּקֻּעָה),","strongs_def":"a semi-globe","kjv_def":"knop."},
"H6498":{"lemma":"פַּקֻּעָה","xlit":"paqquʻâh","pron":"pak-koo-aw'","derivation":"from the same as H6497 (פֶּקַע);","strongs_def":"the wild cucumber (from splitting open to shed its seeds)","kjv_def":"gourd."},
"H6499":{"lemma":"פַּר","xlit":"par","pron":"par","derivation":"or פָּר; from H6565 (פָּרַר);","strongs_def":"a bullock (apparently as breaking forth in wild strength, or perhaps as dividing the hoof)","kjv_def":"([phrase] young) bull(-ock), calf, ox."},
"H6500":{"lemma":"פָּרָא","xlit":"pârâʼ","pron":"paw-raw'","derivation":"a primitive root;","strongs_def":"to bear fruit","kjv_def":"be fruitful."},
"H6501":{"lemma":"פֶּרֶא","xlit":"pereʼ","pron":"peh'-reh","derivation":"or פֶּרֶה; (Jeremiah 2:24), from H6500 (פָּרָא) in the secondary sense of running wild;","strongs_def":"the onager","kjv_def":"wild (ass)."},
"H6502":{"lemma":"פִּרְאָם","xlit":"Pirʼâm","pron":"pir-awm'","derivation":"from H6501 (פֶּרֶא); wildly;","strongs_def":"Piram, a Canaanite","kjv_def":"Piram."},
"H6503":{"lemma":"פַּרְבָּר","xlit":"Parbâr","pron":"par-bawr'","derivation":"or פַּרְוָר; of foreign origin;","strongs_def":"Parbar or Parvar, a quarter of Jerusalem","kjv_def":"Parbar, suburb."},
"H6504":{"lemma":"פָּרַד","xlit":"pârad","pron":"paw-rad'","derivation":"a primitive root;","strongs_def":"to break through, i.e. spread or separate (oneself)","kjv_def":"disperse, divide, be out of joint, part, scatter (abroad), separate (self), sever self, stretch, sunder."},
"H6505":{"lemma":"פֶּרֶד","xlit":"pered","pron":"peh'-red","derivation":"from H6504 (פָּרַד);","strongs_def":"a mule (perhaps from his lonely habits)","kjv_def":"mule."},
"H6506":{"lemma":"פִּרְדָּה","xlit":"pirdâh","pron":"pir-daw'","derivation":"feminine of H6505 (פֶּרֶד);","strongs_def":"a she-mule","kjv_def":"mule."},
"H6507":{"lemma":"פְּרֻדָה","xlit":"pᵉrudâh","pron":"per-oo-daw'","derivation":"feminine passive participle of H6504 (פָּרַד);","strongs_def":"something separated, i.e. a kernel","kjv_def":"seed."},
"H6508":{"lemma":"פַּרְדֵּס","xlit":"pardêç","pron":"par-dace'","derivation":"of foreign origin;","strongs_def":"a park","kjv_def":"forest, orchard."},
"H6509":{"lemma":"פָּרָה","xlit":"pârâh","pron":"paw-raw'","derivation":"a primitive root;","strongs_def":"to bear fruit (literally or figuratively)","kjv_def":"bear, bring forth (fruit), (be, cause to be, make) fruitful, grow, increase."},
"H6510":{"lemma":"פָּרָה","xlit":"pârâh","pron":"paw-raw'","derivation":"feminine of H6499 (פַּר);","strongs_def":"a heifer","kjv_def":"cow, heifer, kine."},
"H6511":{"lemma":"פָּרָה","xlit":"Pârâh","pron":"paw-raw'","derivation":"the same as H6510 (פָּרָה);","strongs_def":"Parah, a place in Palestine","kjv_def":"Parah."},
"H6512":{"lemma":"פֵּרָה","xlit":"pêrâh","pron":"pay-raw'","derivation":"from H6331 (פּוּר);","strongs_def":"a hole (as broken, i.e. dug)","kjv_def":"[phrase] mole. Compare H2661 (חֲפֹר)."},
"H6513":{"lemma":"פֻּרָה","xlit":"Purâh","pron":"poo-raw'","derivation":"for H6288 (פְּאֹרָה); foliage;","strongs_def":"Purah, an Israelite","kjv_def":"Phurah."},
"H6514":{"lemma":"פְּרוּדָא","xlit":"Pᵉrûwdâʼ","pron":"per-oo-daw'","derivation":"or פְּרִידָא; from H6504 (פָּרַד); dispersion;","strongs_def":"Peruda or Perida, one of 'Solomon's servants'","kjv_def":"Perida, Peruda."},
"H6515":{"lemma":"פָּרוּחַ","xlit":"Pârûwach","pron":"paw-roo'-akh","derivation":"passive participle of H6524 (פָּרַח); blossomed;","strongs_def":"Paruach, an Israelite","kjv_def":"Paruah."},
"H6516":{"lemma":"פַּרְוַיִם","xlit":"Parvayim","pron":"par-vah'-yim","derivation":"of foreign origin;","strongs_def":"Parvajim, an Oriental region","kjv_def":"Parvaim."},
"H6517":{"lemma":"פָּרוּר","xlit":"pârûwr","pron":"paw-roor'","derivation":"passive participle of H6565 (פָּרַר) in the sense of spreading out (compare H6524 (פָּרַח));","strongs_def":"a skillet (as flat or deep)","kjv_def":"pan, pot."},
"H6518":{"lemma":"פָּרָז","xlit":"pârâz","pron":"paw-rawz'","derivation":"from an unused root meaning to separate, i.e.","strongs_def":"decide; a chieftain","kjv_def":"village."},
"H6519":{"lemma":"פְּרָזָה","xlit":"pᵉrâzâh","pron":"per-aw-zaw'","derivation":"from the same as H6518 (פָּרָז);","strongs_def":"an open country","kjv_def":"(unwalled) town (without walls), unwalled village."},
"H6520":{"lemma":"פְּרָזוֹן","xlit":"pᵉrâzôwn","pron":"per-aw-zone'","derivation":"from the same as H6518 (פָּרָז);","strongs_def":"magistracy, i.e. leadership (also concretely, chieftains)","kjv_def":"village."},
"H6521":{"lemma":"פְּרָזִי","xlit":"pᵉrâzîy","pron":"per-aw-zee'","derivation":"or פְּרוֹזִי; from H6519 (פְּרָזָה);","strongs_def":"a rustic","kjv_def":"village."},
"H6522":{"lemma":"פְּרִזִּי","xlit":"Pᵉrizzîy","pron":"per-iz-zee'","derivation":"for H6521 (פְּרָזִי); inhabitant of the open country;","strongs_def":"a Perizzite, one of the Canaanitish tribes","kjv_def":"Perizzite."},
"H6523":{"lemma":"פַּרְזֶל","xlit":"parzel","pron":"par-zel'","derivation":"(Aramaic) corresponding to H1270 (בַּרְזֶל);","strongs_def":"iron","kjv_def":"iron."},
"H6524":{"lemma":"פָּרַח","xlit":"pârach","pron":"paw-rakh'","derivation":"a primitive root;","strongs_def":"to break forth as a bud, i.e. bloom; generally, to spread; specifically, to fly (as extending the wings); figuratively, to flourish","kjv_def":"[idiom] abroad, [idiom] abundantly, blossom, break forth (out), bud, flourish, make fly, grow, spread, spring (up)."},
"H6525":{"lemma":"פֶּרַח","xlit":"perach","pron":"peh'-rakh","derivation":"from H6524 (פָּרַח);","strongs_def":"a calyx (natural or artificial); generally, bloom","kjv_def":"blossom, bud, flower."},
"H6526":{"lemma":"פִּרְחַח","xlit":"pirchach","pron":"pir-khakh'","derivation":"from H6524 (פָּרַח);","strongs_def":"progeny, i.e. a brood","kjv_def":"youth."},
"H6527":{"lemma":"פָּרַט","xlit":"pâraṭ","pron":"paw-rat'","derivation":"a primitive root;","strongs_def":"to scatter words, i.e. prate (or hum)","kjv_def":"chant."},
"H6528":{"lemma":"פֶּרֶט","xlit":"pereṭ","pron":"peh'-ret","derivation":"from H6527 (פָּרַט);","strongs_def":"a stray or single berry","kjv_def":"grape."},
"H6529":{"lemma":"פְּרִי","xlit":"pᵉrîy","pron":"per-ee'","derivation":"from H6509 (פָּרָה);","strongs_def":"fruit (literally or figuratively)","kjv_def":"bough, (first-)fruit(-ful), reward."},
"H6530":{"lemma":"פְּרִיץ","xlit":"pᵉrîyts","pron":"per-eets'","derivation":"from H6555 (פָּרַץ)lemma פּרַץ missing vowel, corrected to פָּרַץ;","strongs_def":"violent, i.e. a tyrant","kjv_def":"destroyer, ravenous, robber."},
"H6531":{"lemma":"פֶּרֶךְ","xlit":"perek","pron":"peh'-rek","derivation":"from an unused root meaning to break apart;","strongs_def":"fracture, i.e. severity","kjv_def":"cruelty, rigour."},
"H6532":{"lemma":"פֹּרֶכֶת","xlit":"pôreketh","pron":"po-reh'-keth","derivation":"feminine active participle of the same as H6531 (פֶּרֶךְ);","strongs_def":"a separatrix, i.e. (the sacred) screen","kjv_def":"vail."},
"H6533":{"lemma":"פָּרַם","xlit":"pâram","pron":"paw-ram'","derivation":"a primitive root;","strongs_def":"to tear","kjv_def":"rend."},
"H6534":{"lemma":"פַּרְמַשְׁתָּא","xlit":"Parmashtâʼ","pron":"par-mash-taw'","derivation":"of Persian origin;","strongs_def":"Parmashta, a son of Haman","kjv_def":"Parmasta."},
"H6535":{"lemma":"פַּרְנַךְ","xlit":"Parnak","pron":"par-nak'","derivation":"of uncertain derivation;","strongs_def":"Parnak, an Israelite","kjv_def":"Parnach."},
"H6536":{"lemma":"פָּרַס","xlit":"pâraç","pron":"paw-ras'","derivation":"a primitive root;","strongs_def":"to break in pieces, i.e. (usually without violence) to split, distribute","kjv_def":"deal, divide, have hoofs, part, tear."},
"H6537":{"lemma":"פְּרַס","xlit":"pᵉraç","pron":"per-as'","derivation":"(Aramaic) corresponding to H6536 (פָּרַס);","strongs_def":"to split up","kjv_def":"divide, (U-) pharsin."},
"H6538":{"lemma":"פֶּרֶס","xlit":"pereç","pron":"peh'-res","derivation":"from H6536 (פָּרַס);","strongs_def":"a claw; also a kind of eagle","kjv_def":"claw, ossifrage."},
"H6539":{"lemma":"פָּרַס","xlit":"Pâraç","pron":"paw-ras'","derivation":"of foreign origin;","strongs_def":"Paras (i.e. Persia), an Eastern country, including its inhabitants","kjv_def":"Persia, Persians."},
"H6540":{"lemma":"פָּרַס","xlit":"Pâraç","pron":"paw-ras'","derivation":"(Aramaic) corresponding to H6539 (פָּרַס)","strongs_def":"{Paras (i.e. Persia), an Eastern country, including its inhabitants}","kjv_def":"Persia, Persians."},
"H6541":{"lemma":"פַּרְסָה","xlit":"parçâh","pron":"par-saw'","derivation":"feminine of H6538 (פֶּרֶס);","strongs_def":"a claw or split hoof","kjv_def":"claw, (cloven-) footed, hoof."},
"H6542":{"lemma":"פַּרְסִי","xlit":"Parçîy","pron":"par-see'","derivation":"patrial from H6539 (פָּרַס);","strongs_def":"a Parsite (i.e. Persian), or inhabitant of Peres","kjv_def":"Persian."},
"H6543":{"lemma":"פַּרְסִי","xlit":"Parçîy","pron":"par-see'","derivation":"(Aramaic) corresponding to H6542 (פַּרְסִי)","strongs_def":"{a Parsite (i.e. Persian), or inhabitant of Peres}","kjv_def":"Persian."},
"H6544":{"lemma":"פָּרַע","xlit":"pâraʻ","pron":"paw-rah'","derivation":"a primitive root;","strongs_def":"to loosen; by implication, to expose, dismiss; figuratively, absolve, begin","kjv_def":"avenge, avoid, bare, go back, let, (make) naked, set at nought, perish, refuse, uncover."},
"H6545":{"lemma":"פֶּרַע","xlit":"peraʻ","pron":"peh'-rah","derivation":"from H6544 (פָּרַע);","strongs_def":"the hair (as dishevelled)","kjv_def":"locks."},
"H6546":{"lemma":"פַּרְעָה","xlit":"parʻâh","pron":"par-aw'","derivation":"feminine of H6545 (פֶּרַע) (in the sense of beginning);","strongs_def":"leadership (plural concretely, leaders)","kjv_def":"[phrase] avenging, revenge."},
"H6547":{"lemma":"פַּרְעֹה","xlit":"Parʻôh","pron":"par-o'","derivation":"of Egyptian derivation;","strongs_def":"Paroh, a general title of Egyptian kings","kjv_def":"Pharaoh."},
"H6548":{"lemma":"פַּרְעֹה חׇפְרַע","xlit":"Parʻôh Chophraʻ","pron":"par-o' khof-rah'","derivation":"of Egyptian derivation;","strongs_def":"Paroh-Chophra, an Egyptian king","kjv_def":"Pharaoh-hophra."},
"H6549":{"lemma":"פַּרְעֹה נְכֹה","xlit":"Parʻôh Nᵉkôh","pron":"par-o' nek-o'","derivation":"or פַּרְעֹה נְכוֹ; of Egyptian derivation;","strongs_def":"Paroh-Nekoh (or -Neko), an Egyptian king","kjv_def":"Pharaoh-necho, Pharaohnechoh."},
"H6550":{"lemma":"פַּרְעֹשׁ","xlit":"parʻôsh","pron":"par-oshe'","derivation":"probably from H6544 (פָּרַע) and H6211 (עָשׁ);","strongs_def":"a flea (as the isolated insect)","kjv_def":"flea."},
"H6551":{"lemma":"פַּרְעֹשׁ","xlit":"Parʻôsh","pron":"par-oshe'","derivation":"the same as H6550 (פַּרְעֹשׁ);","strongs_def":"Parosh, the name of our Israelite","kjv_def":"Parosh, Pharosh."},
"H6552":{"lemma":"פִּרְעָתוֹן","xlit":"Pirʻâthôwn","pron":"pir-aw-thone'","derivation":"from H6546 (פַּרְעָה); chieftaincy;","strongs_def":"Pirathon, a place in Palestine","kjv_def":"Pirathon."},
"H6553":{"lemma":"פִּרְעָתוֹנִי","xlit":"Pirʻâthôwnîy","pron":"pir-aw-tho-nee'","derivation":"or פִּרְעָתֹנִי; patrial from H6552 (פִּרְעָתוֹן);","strongs_def":"a Pirathonite or inhabitant of Pirathon","kjv_def":"Pirathonite."},
"H6554":{"lemma":"פַּרְפַּר","xlit":"Parpar","pron":"par-par'","derivation":"probably from H6565 (פָּרַר) in the sense of rushing; rapid;","strongs_def":"Parpar, a river of Syria","kjv_def":"Pharpar."},
"H6555":{"lemma":"פָּרַץ","xlit":"pârats","pron":"paw-rats'","derivation":"a primitive root;","strongs_def":"to break out (in many applications, direct and indirect, literal and figurative)","kjv_def":"[idiom] abroad, (make a) breach, break (away, down, -er, forth, in, up), burst out, come (spread) abroad, compel, disperse, grow, increase, open, press, scatter, urge."},
"H6556":{"lemma":"פֶּרֶץ","xlit":"perets","pron":"peh'-rets","derivation":"from H6555 (פָּרַץ)lemma פּרַץ missing vowel, corrected to פָּרַץ;","strongs_def":"a break (literally or figuratively)","kjv_def":"breach, breaking forth (in), [idiom] forth, gap."},
"H6557":{"lemma":"פֶּרֶץ","xlit":"Perets","pron":"peh'-rets","derivation":"the same as H6556 (פֶּרֶץ);","strongs_def":"Perets, the name of two Israelites","kjv_def":"Perez, Pharez."},
"H6558":{"lemma":"פַּרְצִי","xlit":"Partsîy","pron":"par-tsee'","derivation":"patronymically from H6557 (פֶּרֶץ);","strongs_def":"a Partsite (collectively) or descendants of Perets","kjv_def":"Pharzites."},
"H6559":{"lemma":"פְּרָצִים","xlit":"pᵉrâtsîym","pron":"per-aw-tseem'","derivation":"plural of H6556 (פֶּרֶץ); breaks;","strongs_def":"Peratsim, a mountain in Palestine","kjv_def":"Perazim."},
"H6560":{"lemma":"פֶּרֶץ עֻזָּא","xlit":"Perets ʻUzzâʼ","pron":"peh'-rets ooz-zaw'","derivation":"from H6556 (פֶּרֶץ) and H5798 (עֻזָּא); break of Uzza;","strongs_def":"Perets-Uzza, a place in Palestine","kjv_def":"Perezuzza."},
"H6561":{"lemma":"פָּרַק","xlit":"pâraq","pron":"paw-rak'","derivation":"a primitive root;","strongs_def":"to break off or crunch; figuratively, to deliver","kjv_def":"break (off), deliver, redeem, rend (in pieces), tear in pieces."},
"H6562":{"lemma":"פְּרַק","xlit":"pᵉraq","pron":"per-ak'","derivation":"(Aramaic) corresponding to H6561 (פָּרַק);","strongs_def":"to discontinue","kjv_def":"break off."},
"H6563":{"lemma":"פֶּרֶק","xlit":"pereq","pron":"peh'-rek","derivation":"from H6561 (פָּרַק);","strongs_def":"rapine; also a fork (in roads)","kjv_def":"crossway, robbery."},
"H6564":{"lemma":"פָּרָק","xlit":"pârâq","pron":"paw-rawk'","derivation":"from H6561 (פָּרַק);","strongs_def":"soup (as full of crumbed meat)","kjv_def":"broth. See also H4832 (מַרְפֵּא)."},
"H6565":{"lemma":"פָּרַר","xlit":"pârar","pron":"paw-rar'","derivation":"a primitive root;","strongs_def":"to break up (usually figuratively), i.e. to violate, frustrate","kjv_def":"[idiom] any ways, break (asunder), cast off, cause to cease, [idiom] clean, defeat, disannul, disappoint, dissolve, divide, make of none effect, fail, frustrate, bring (come) to nought, [idiom] utterly, make void."},
"H6566":{"lemma":"פָּרַשׂ","xlit":"pâras","pron":"paw-ras'","derivation":"a primitive root;","strongs_def":"to break apart, disperse, etc.","kjv_def":"break, chop in pieces, lay open, scatter, spread (abroad, forth, selves, out), stretch (forth, out)."},
"H6567":{"lemma":"פָּרָשׁ","xlit":"pârâsh","pron":"paw-rash'","derivation":"a primitive root;","strongs_def":"to separate, literally (to disperse) or figuratively (to specify); also (by implication) to wound","kjv_def":"scatter, declare, distinctly, shew, sting."},
"H6568":{"lemma":"פְּרַשׁ","xlit":"pᵉrash","pron":"per-ash'","derivation":"(Aramaic) corresponding to H6567 (פָּרָשׁ);","strongs_def":"to specify","kjv_def":"distinctly."},
"H6569":{"lemma":"פֶּרֶשׁ","xlit":"peresh","pron":"peh'-resh","derivation":"from H6567 (פָּרָשׁ);","strongs_def":"excrement (as eliminated)","kjv_def":"dung."},
"H6570":{"lemma":"פֶּרֶשׁ","xlit":"Peresh","pron":"peh'-resh","derivation":"the same as H6569 (פֶּרֶשׁ);","strongs_def":"Peresh, an Israelite","kjv_def":"Peresh."},
"H6571":{"lemma":"פָּרָשׁ","xlit":"pârâsh","pron":"paw-rawsh'","derivation":"from H6567 (פָּרָשׁ); (compare H5483 (סוּס))","strongs_def":"a steed (as stretched out to a vehicle, not single nor for mounting ); also (by implication) a driver (in a chariot), i.e. (collectively) cavalry","kjv_def":"horseman."},
"H6572":{"lemma":"פַּרְשֶׁגֶן","xlit":"parshegen","pron":"par-sheh'-ghen","derivation":"or פַּתְשֶׁגֶן; of foreign origin;","strongs_def":"a transcript","kjv_def":"copy."},
"H6573":{"lemma":"פַּרְשֶׁגֶן","xlit":"parshegen","pron":"par-sheh'-ghen","derivation":"(Aramaic) corresponding to H6572 (פַּרְשֶׁגֶן)","strongs_def":"{a transcript}","kjv_def":"copy."},
"H6574":{"lemma":"פַּרְשְׁדֹן","xlit":"parshᵉdôn","pron":"par-shed-one'","derivation":"perhaps by compounding H6567 (פָּרָשׁ) and H6504 (פָּרַד) (in the sense of straddling) (compare H6576 (פַּרְשֵׁז));","strongs_def":"the crotch (or anus)","kjv_def":"dirt."},
"H6575":{"lemma":"פָּרָשָׁה","xlit":"pârâshâh","pron":"paw-raw-shaw'","derivation":"from H6567 (פָּרָשׁ);","strongs_def":"exposition","kjv_def":"declaration, sum."},
"H6576":{"lemma":"פַּרְשֵׁז","xlit":"parshêz","pron":"par-shaze'","derivation":"a root apparently formed by compounding H6567 (פָּרָשׁ) and that of H6518 (פָּרָז) (compare H6574 (פַּרְשְׁדֹן));","strongs_def":"to expand","kjv_def":"spread."},
"H6577":{"lemma":"פַּרְשַׁנְדָּתָא","xlit":"Parshandâthâʼ","pron":"par-shan-daw-thaw'","derivation":"of Persian origin;","strongs_def":"Parshandatha, a son of Haman","kjv_def":"Parshandatha."},
"H6578":{"lemma":"פְּרָת","xlit":"Pᵉrâth","pron":"per-awth'","derivation":"from an unused root meaning to break forth; rushing;","strongs_def":"Perath (i.e. Euphrates), a river of the East","kjv_def":"Euphrates."},
"H6579":{"lemma":"פַּרְתַּם","xlit":"partam","pron":"par-tam'","derivation":"of Persian origin;","strongs_def":"a grandee","kjv_def":"(most) noble, prince."},
"H6580":{"lemma":"פַּשׁ","xlit":"pash","pron":"pash","derivation":"probably from an unused root meaning to disintegrate;","strongs_def":"stupidity (as a result of grossness or of degeneracy)","kjv_def":"extremity."},
"H6581":{"lemma":"פָּשָׂה","xlit":"pâsâh","pron":"paw-saw'","derivation":"a primitive root;","strongs_def":"to spread","kjv_def":"spread."},
"H6582":{"lemma":"פָּשַׁח","xlit":"pâshach","pron":"paw-shakh'","derivation":"a primitive root;","strongs_def":"to tear in pieces","kjv_def":"pull in pieces."},
"H6583":{"lemma":"פַּשְׁחוּר","xlit":"Pashchûwr","pron":"pash-khoor'","derivation":"probably from H6582 (פָּשַׁח); liberation;","strongs_def":"Pashchur, the name of four Israelites","kjv_def":"Pashur."},
"H6584":{"lemma":"פָּשַׁט","xlit":"pâshaṭ","pron":"paw-shat'","derivation":"a primitive root;","strongs_def":"to spread out (i.e. deploy in hostile array); by analogy, to strip (i.e. unclothe, plunder, flay, etc.)","kjv_def":"fall upon, flay, invade, make an invasion, pull off, put off, make a road, run upon, rush, set, spoil, spread selves (abroad), strip (off, self)."},
"H6585":{"lemma":"פָּשַׂע","xlit":"pâsaʻ","pron":"paw-sah'","derivation":"a primitive root;","strongs_def":"to stride (from spreading the legs), i.e. rush upon","kjv_def":"go."},
"H6586":{"lemma":"פָּשַׁע","xlit":"pâshaʻ","pron":"paw-shah'","derivation":"a primitive root (identical with through the idea of expansion);","strongs_def":"to break away (from just authority), i.e. trespass, apostatize, quarrel","kjv_def":"offend, rebel, revolt, transgress(-ion, -or)."},
"H6587":{"lemma":"פֶּשַׂע","xlit":"pesaʻ","pron":"peh'-sah","derivation":"from H6585 (פָּשַׂע);","strongs_def":"a stride","kjv_def":"step."},
"H6588":{"lemma":"פֶּשַׁע","xlit":"peshaʻ","pron":"peh'-shah","derivation":"from H6586 (פָּשַׁע);","strongs_def":"a revolt (national, moral or religious)","kjv_def":"rebellion, sin, transgression, trespass."},
"H6589":{"lemma":"פָּשַׂק","xlit":"pâsaq","pron":"paw-sak'","derivation":"a primitive root;","strongs_def":"to dispart (the feet or lips), i.e. become licentious","kjv_def":"open (wide)."},
"H6590":{"lemma":"פְּשַׁר","xlit":"pᵉshar","pron":"pesh-ar'","derivation":"(Aramaic) corresponding to H6622 (פָּתַר);","strongs_def":"to interpret","kjv_def":"make (interpretations), interpreting."},
"H6591":{"lemma":"פְּשַׁר","xlit":"pᵉshar","pron":"pesh-ar'","derivation":"(Aramaic) from H6590 (פְּשַׁר);","strongs_def":"an interpretation","kjv_def":"interpretation."},
"H6592":{"lemma":"פֵּשֶׁר","xlit":"pêsher","pron":"pay'-sher","derivation":"corresponding to H6591 (פְּשַׁר)","strongs_def":"{an interpretation}","kjv_def":"interpretation."},
"H6593":{"lemma":"פִּשְׁתֶּה","xlit":"pishteh","pron":"pish-teh'","derivation":"from the same as H6580 (פַּשׁ) as in the sense of comminuting;","strongs_def":"linen (i.e. the thread, as carded)","kjv_def":"flax, linen."},
"H6594":{"lemma":"פִּשְׁתָּה","xlit":"pishtâh","pron":"pish-taw'","derivation":"feminine of H6593 (פִּשְׁתֶּה);","strongs_def":"flax; by implication, a wick","kjv_def":"flax, tow."},
"H6595":{"lemma":"פַּת","xlit":"path","pron":"path","derivation":"from H6626 (פָּתַת);","strongs_def":"a bit","kjv_def":"meat, morsel, piece."},
"H6596":{"lemma":"פֹּת","xlit":"pôth","pron":"pohth","derivation":"or פֹּתָה; (Ezekiel 13:19), from an unused root meaning to open;","strongs_def":"a hole, i.e. hinge or the female pudenda","kjv_def":"hinge, secret participle"},
"H6597":{"lemma":"פִּתְאוֹם","xlit":"pithʼôwm","pron":"pith-ome'","derivation":"or פִּתְאֹם; from H6621 (פֶּתַע);","strongs_def":"instantly","kjv_def":"straightway, sudden(-ly)."},
"H6598":{"lemma":"פַּתְבַּג","xlit":"pathbag","pron":"pathbag'","derivation":"of Persian origin;","strongs_def":"a dainty","kjv_def":"portion (provision) of meat."},
"H6599":{"lemma":"פִּתְגָּם","xlit":"pithgâm","pron":"pith-gawm'","derivation":"of Persian origin;","strongs_def":"a (judicial) sentence","kjv_def":"decree, sentence."},
"H6600":{"lemma":"פִּתְגָּם","xlit":"pithgâm","pron":"pith-gawm'","derivation":"(Aramaic) corresponding to H6599 (פִּתְגָּם);","strongs_def":"a word, answer, letter or decree","kjv_def":"answer, letter, matter, word."},
"H6601":{"lemma":"פָּתָה","xlit":"pâthâh","pron":"paw-thaw'","derivation":"a primitive root;","strongs_def":"to open, i.e. be (causatively, make) roomy; usually figuratively (in a mental or moral sense) to be (causatively, make) simple or (in a sinister way) delude","kjv_def":"allure, deceive, enlarge, entice, flatter, persuade, silly (one)."},
"H6602":{"lemma":"פְּתוּאֵל","xlit":"Pᵉthûwʼêl","pron":"peth-oo-ale'","derivation":"from H6601 (פָּתָה) and H410 (אֵל); enlarged of God;","strongs_def":"Pethuel, an Israelite","kjv_def":"Pethuel."},
"H6603":{"lemma":"פִּתּוּחַ","xlit":"pittûwach","pron":"pit-too'-akh","derivation":"or פִּתֻּחַ ; passive participle of H6605 (פָּתַח);","strongs_def":"sculpture (in low or high relief or even intaglio)","kjv_def":"carved (work) (are, en-) grave(-ing, -n)."},
"H6604":{"lemma":"פְּתוֹר","xlit":"Pᵉthôwr","pron":"peth-ore'","derivation":"of foreign origin;","strongs_def":"Pethor, a place in Mesopotamia","kjv_def":"Pethor."},
"H6605":{"lemma":"פָּתַח","xlit":"pâthach","pron":"paw-thakh'","derivation":"a primitive root;","strongs_def":"to open wide (literally or figuratively); specifically, to loosen, begin, plough, carve","kjv_def":"appear, break forth, draw (out), let go free, (en-) grave(-n), loose (self), (be, be set) open(-ing), put off, ungird, unstop, have vent."},
"H6606":{"lemma":"פְּתַח","xlit":"pᵉthach","pron":"peth-akh'","derivation":"(Aramaic) corresponding to H6605 (פָּתַח);","strongs_def":"to open","kjv_def":"open."},
"H6607":{"lemma":"פֶּתַח","xlit":"pethach","pron":"peh'-thakh","derivation":"from H6605 (פָּתַח);","strongs_def":"an opening (literally), i.e. door (gate) or entrance way","kjv_def":"door, entering (in), entrance (-ry), gate, opening, place."},
"H6608":{"lemma":"פֵּתַח","xlit":"pêthach","pron":"pay'-thakh","derivation":"from H6605 (פָּתַח);","strongs_def":"opening (figuratively) i.e. disclosure","kjv_def":"entrance."},
"H6609":{"lemma":"פְּתִחָה","xlit":"pᵉthichâh","pron":"peth-ee-khaw'","derivation":"from H6605 (פָּתַח);","strongs_def":"something opened, i.e. a drawn sword","kjv_def":"drawn sword."},
"H6610":{"lemma":"פִּתְחוֹן","xlit":"pithchôwn","pron":"pith-khone'","derivation":"from H6605 (פָּתַח);","strongs_def":"opening (the act)","kjv_def":"open(-ing)."},
"H6611":{"lemma":"פְּתַחְיָה","xlit":"Pᵉthachyâh","pron":"peth-akh-yaw'","derivation":"from H6605 (פָּתַח) and H3050 (יָהּ); Jah has opened;","strongs_def":"Pethachjah, the name of four Israelites","kjv_def":"Pethakiah."},
"H6612":{"lemma":"פְּתִי","xlit":"pᵉthîy","pron":"peth-ee'","derivation":"or פֶּתִי; or פְּתָאִי; from H6601 (פָּתָה);","strongs_def":"silly (i.e. seducible)","kjv_def":"foolish, simple(-icity, one)."},
"H6613":{"lemma":"פְּתַי","xlit":"pᵉthay","pron":"peth-ah'-ee","derivation":"(Aramaic) from a root corresponding to H6601 (פָּתָה);","strongs_def":"open, i.e. (as noun) width","kjv_def":"breadth."},
"H6614":{"lemma":"פְּתִיגִיל","xlit":"pᵉthîygîyl","pron":"peth-eeg-eel'","derivation":"of uncertain derivation;","strongs_def":"probably a figured mantle forholidays","kjv_def":"stomacher."},
"H6615":{"lemma":"פְּתַיּוּת","xlit":"pᵉthayûwth","pron":"peth-ah-yooth'","derivation":"from H6612 (פְּתִי);","strongs_def":"silliness (i.e. seducibility)","kjv_def":"simple."},
"H6616":{"lemma":"פָּתִיל","xlit":"pâthîyl","pron":"paw-theel'","derivation":"from H6617 (פָּתַל);","strongs_def":"twine","kjv_def":"bound, bracelet, lace, line, ribband, thread, wire."},
"H6617":{"lemma":"פָּתַל","xlit":"pâthal","pron":"paw-thal'","derivation":"a primitive root;","strongs_def":"to twine, i.e. (literally) to struggle or (figuratively) be (morally) tortuous","kjv_def":"(shew self) froward, shew self unsavoury, wrestle."},
"H6618":{"lemma":"פְּתַלְתֹּל","xlit":"pᵉthaltôl","pron":"peth-al-tole'","derivation":"from H6617 (פָּתַל);","strongs_def":"tortuous (i.e. crafty)","kjv_def":"crooked."},
"H6619":{"lemma":"פִּתֹם","xlit":"Pithôm","pron":"pee-thome'","derivation":"of Egyptian derivation;","strongs_def":"Pithom, a place in Egypt","kjv_def":"Pithom."},
"H6620":{"lemma":"פֶּתֶן","xlit":"pethen","pron":"peh'-then","derivation":"from an unused root meaning to twist;","strongs_def":"an asp (from its contortions)","kjv_def":"adder."},
"H6621":{"lemma":"פֶּתַע","xlit":"pethaʻ","pron":"peh'-thah","derivation":"from an unused root meaning to open (the eyes); (compare H6597 (פִּתְאוֹם))","strongs_def":"a wink, i.e. moment(used only (with or without preposition) adverbially, quickly or unexpectedly","kjv_def":"at an instant, suddenly, [idiom] very."},
"H6622":{"lemma":"פָּתַר","xlit":"pâthar","pron":"paw-thar'","derivation":"a primitive root;","strongs_def":"to open up, i.e. (figuratively) interpret (a dream)","kjv_def":"interpret(-ation, -er)."},
"H6623":{"lemma":"פִּתְרוֹן","xlit":"pithrôwn","pron":"pith-rone'","derivation":"or פִּתְרֹן; from H6622 (פָּתַר);","strongs_def":"interpretation (of a dream)","kjv_def":"interpretation."},
"H6624":{"lemma":"פַּתְרוֹס","xlit":"Pathrôwç","pron":"path-roce'","derivation":"of Egyptian derivation;","strongs_def":"Pathros, a part of Egypt","kjv_def":"Pathros."},
"H6625":{"lemma":"פַּתְרֻסִי","xlit":"Pathruçîy","pron":"path-roo-see'","derivation":"patrial from H6624 (פַּתְרוֹס);","strongs_def":"a Pathrusite, or inhabitant of Pathros","kjv_def":"Pathrusim."},
"H6626":{"lemma":"פָּתַת","xlit":"pâthath","pron":"paw-thath'","derivation":"a primitive root;","strongs_def":"to open, i.e. break","kjv_def":"participle x"},
"H6627":{"lemma":"צָאָה","xlit":"tsâʼâh","pron":"tsaw-aw'","derivation":"from H3318 (יָצָא);","strongs_def":"issue, i.e. (human) excrement","kjv_def":"that (which) cometh from (out)."},
"H6628":{"lemma":"צֶאֶל","xlit":"tseʼel","pron":"tseh'-el","derivation":"from an unused root meaning to be slender;","strongs_def":"the lotus tree","kjv_def":"shady tree."},
"H6629":{"lemma":"צֹאן","xlit":"tsôʼn","pron":"tsone","derivation":"or צאוֹן; (Psalm 144:13), from an unused root meaning to migrate;","strongs_def":"a collective name for a flock (of sheep or goats); also figuratively (of men)","kjv_def":"(small) cattle, flock ([phrase] -s), lamb ([phrase] -s), sheep(-cote, -fold, -shearer, -herds)."},
"H6630":{"lemma":"צַאֲנָן","xlit":"Tsaʼănân","pron":"tsah-an-awn'","derivation":"from the same as H6629 (צֹאן) used denominatively; sheep pasture;","strongs_def":"Zaanan, a place in Palestine","kjv_def":"Zaanan."},
"H6631":{"lemma":"צֶאֱצָא","xlit":"tseʼĕtsâʼ","pron":"tseh-ets-aw'","derivation":"from H3318 (יָצָא);","strongs_def":"issue, i.e. produce, children","kjv_def":"that which cometh forth (out), offspring."},
"H6632":{"lemma":"צָב","xlit":"tsâb","pron":"tsawb","derivation":"from an unused root meaning to establish;","strongs_def":"a palanquin or canopy (as a fixture); also a species of lizard (probably as clinging fast)","kjv_def":"covered, litter, tortoise."},
"H6633":{"lemma":"צָבָא","xlit":"tsâbâʼ","pron":"tsaw-baw'","derivation":"a primitive root;","strongs_def":"to mass (an army or servants)","kjv_def":"assemble, fight, perform, muster, wait upon, war."},
"H6634":{"lemma":"צְבָא","xlit":"tsᵉbâʼ","pron":"tseb-aw'","derivation":"(Aramaic) corresponding to H6623 (פִּתְרוֹן) in the figurative sense of summoning one's wishes;","strongs_def":"to please","kjv_def":"will, would."},
"H6635":{"lemma":"צָבָא","xlit":"tsâbâʼ","pron":"tsaw-baw'","derivation":"or (feminine) צְבָאָה; from H6633 (צָבָא);","strongs_def":"a mass of persons (or figuratively, things), especially reg. organized forwar (an army); by implication, a campaign, literally or figuratively (specifically, hardship, worship)","kjv_def":"appointed time, ([phrase]) army, ([phrase]) battle, company, host, service, soldiers, waiting upon, war(-fare)."},
"H6636":{"lemma":"צְבֹאִים","xlit":"Tsᵉbôʼîym","pron":"tseb-o-eem'","derivation":"or (more correctly) צְבִיִּים; or צְבִיִּם; plural of H6643 (צְבִי); gazelles;","strongs_def":"Tseboim or Tsebijim, a place in Palestine","kjv_def":"Zeboiim, Zeboim."},
"H6637":{"lemma":"צֹבֵבָה","xlit":"Tsôbêbâh","pron":"tso-bay-baw'","derivation":"feminine active participle of the same as H6632 (צָב); the canopier (with the article);","strongs_def":"Tsobebah, an Israelitess","kjv_def":"Zobebah."},
"H6638":{"lemma":"צָבָה","xlit":"tsâbâh","pron":"tsaw-baw'","derivation":"a primitive root;","strongs_def":"to amass, i.e. grow turgid; specifically, to array an army against","kjv_def":"fight swell."},
"H6639":{"lemma":"צָבֶה","xlit":"tsâbeh","pron":"tsaw-beh'","derivation":"from H6638 (צָבָה);","strongs_def":"turgid","kjv_def":"swell."},
"H6640":{"lemma":"צְבוּ","xlit":"tsᵉbûw","pron":"tseb-oo'","derivation":"(Aramaic) from H6634 (צְבָא);","strongs_def":"properly, will; concretely, an affair (as a matter of determination)","kjv_def":"purpose."},
"H6641":{"lemma":"צָבוּעַ","xlit":"tsâbûwaʻ","pron":"tsaw-boo'-ah","derivation":"passive participle of the same as H6648 (צֶבַע);","strongs_def":"dyed (in stripes), i.e. the hyena","kjv_def":"speckled."},
"H6642":{"lemma":"צָבַט","xlit":"tsâbaṭ","pron":"tsaw-bat'","derivation":"a primitive root;","strongs_def":"to grasp, i.e. hand out","kjv_def":"reach."},
"H6643":{"lemma":"צְבִי","xlit":"tsᵉbîy","pron":"tseb-ee'","derivation":"from H6638 (צָבָה) in the sense of prominence;","strongs_def":"splendor (as conspicuous); also a gazelle (as beautiful)","kjv_def":"beautiful(-ty), glorious (-ry), goodly, pleasant, roe(-buck)."},
"H6644":{"lemma":"צִבְיָא","xlit":"Tsibyâʼ","pron":"tsib-yaw'","derivation":"for H6645 (צִבְיָּה);","strongs_def":"Tsibja, an Israelite","kjv_def":"Zibia."},
"H6645":{"lemma":"צִבְיָּה","xlit":"Tsibyâh","pron":"tsib-yaw'","derivation":"for H6646 (צְבִיָּה);","strongs_def":"Tsibjah, an Israelite","kjv_def":"Zibiah."},
"H6646":{"lemma":"צְבִיָּה","xlit":"tsᵉbîyâh","pron":"tseb-ee-yaw'","derivation":"feminine of H6643 (צְבִי);","strongs_def":"a female gazelle","kjv_def":"roe."},
"H6647":{"lemma":"צְבַע","xlit":"tsᵉbaʻ","pron":"tseb-ah'","derivation":"(Aramaic) a root corresponding to that of H6648 (צֶבַע);","strongs_def":"to dip","kjv_def":"wet."},
"H6648":{"lemma":"צֶבַע","xlit":"tsebaʻ","pron":"tseh'-bah","derivation":"from an unused root meaning to dip (into coloring fluid);","strongs_def":"a dye","kjv_def":"divers, colours."},
"H6649":{"lemma":"צִבְעוֹן","xlit":"Tsibʻôwn","pron":"tsib-one'","derivation":"from the same as H6648 (צֶבַע); variegated;","strongs_def":"Tsibon, an Idumaean","kjv_def":"Zibeon."},
"H6650":{"lemma":"צְבֹעִים","xlit":"Tsᵉbôʻîym","pron":"tseb-o-eem'","derivation":"plural of H6641 (צָבוּעַ); hyenas;","strongs_def":"Tseboim, a place in Palestine","kjv_def":"Zeboim."},
"H6651":{"lemma":"צָבַר","xlit":"tsâbar","pron":"tsaw-bar'","derivation":"a primitive root;","strongs_def":"to aggregate","kjv_def":"gather (together), heap (up), lay up."},
"H6652":{"lemma":"צִבֻּר","xlit":"tsibbur","pron":"tsib-boor'","derivation":"from H6551 (פַּרְעֹשׁ);","strongs_def":"a pile","kjv_def":"heap."},
"H6653":{"lemma":"צֶבֶת","xlit":"tsebeth","pron":"tseh'-beth","derivation":"from an unused root apparently meaning to grip;","strongs_def":"a lock of stalks","kjv_def":"handful."},
"H6654":{"lemma":"צַד","xlit":"tsad","pron":"tsad","derivation":"contr. from an unused root meaning to sidle off;","strongs_def":"a side; figuratively, an adversary","kjv_def":"(be-) side."},
"H6655":{"lemma":"צַד","xlit":"tsad","pron":"tsad","derivation":"(Aramaic) corresponding to H6654 (צַד);","strongs_def":"used adverbially (with preposition) at or upon the side of","kjv_def":"against, concerning."},
"H6656":{"lemma":"צְדָא","xlit":"tsᵉdâʼ","pron":"tsed-aw'","derivation":"(Aramaic) from an unused root corresponding to H6658 (צָדָה) in the sense of intentness;","strongs_def":"a (sinister) design","kjv_def":"true."},
"H6657":{"lemma":"צְדָד","xlit":"Tsᵉdâd","pron":"tsed-awd'","derivation":"from the same as H6654 (צַד); a siding;","strongs_def":"Tsedad, a place near Palestine","kjv_def":"Zedad."},
"H6658":{"lemma":"צָדָה","xlit":"tsâdâh","pron":"tsaw-daw'","derivation":"a primitive root;","strongs_def":"to chase; by implication, to desolate","kjv_def":"destroy, hunt, lie in wait."},
"H6659":{"lemma":"צָדוֹק","xlit":"Tsâdôwq","pron":"tsaw-doke'","derivation":"from H6663 (צָדַק); just;","strongs_def":"Tsadok, the name of eight or nine Israelites","kjv_def":"Zadok."},
"H6660":{"lemma":"צְדִיָּה","xlit":"tsᵉdîyâh","pron":"tsed-ee-yaw'","derivation":"from H6658 (צָדָה); (compare H6656 (צְדָא))","strongs_def":"design","kjv_def":"lying in wait."},
"H6661":{"lemma":"צִדִּים","xlit":"Tsiddîym","pron":"tsid-deem'","derivation":"plural of H6654 (צַד); sides;","strongs_def":"Tsiddim (with the article), a place in Palestine","kjv_def":"Ziddim."},
"H6662":{"lemma":"צַדִּיק","xlit":"tsaddîyq","pron":"tsad-deek'","derivation":"from H6663 (צָדַק);","strongs_def":"just","kjv_def":"just, lawful, righteous (man)."},
"H6663":{"lemma":"צָדַק","xlit":"tsâdaq","pron":"tsaw-dak'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) right (in a moral or forensic sense)","kjv_def":"cleanse, clear self, (be, do) just(-ice, -ify, -ify self), (be turn to) righteous(-ness)."},
"H6664":{"lemma":"צֶדֶק","xlit":"tsedeq","pron":"tseh'-dek","derivation":"from H6663 (צָדַק);","strongs_def":"the right (natural, moral or legal); also (abstractly) equity or (figuratively) prosperity","kjv_def":"[idiom] even, ([idiom] that which is altogether) just(-ice), (un-)right(-eous) (cause, -ly, -ness)."},
"H6665":{"lemma":"צִדְקָה","xlit":"tsidqâh","pron":"tsid-kaw'","derivation":"(Aramaic) corresponding to H6666 (צְדָקָה);","strongs_def":"beneficence","kjv_def":"righteousness."},
"H6666":{"lemma":"צְדָקָה","xlit":"tsᵉdâqâh","pron":"tsed-aw-kaw'","derivation":"from H6663 (צָדַק);","strongs_def":"rightness (abstractly), subjectively (rectitude), objectively (justice), morally (virtue) or figuratively (prosperity)","kjv_def":"justice, moderately, right(-eous) (act, -ly, -ness)."},
"H6667":{"lemma":"צִדְקִיָּה","xlit":"Tsidqîyâh","pron":"tsid-kee-yaw'","derivation":"or צִדְקִיָּהוּ; from H6664 (צֶדֶק) and H3050 (יָהּ); right of Jah;","strongs_def":"Tsidkijah, the name of six Israelites","kjv_def":"Zedekiah, Zidkijah."},
"H6668":{"lemma":"צָהַב","xlit":"tsâhab","pron":"tsaw-hab'","derivation":"a primitive root;","strongs_def":"to glitter, i.e. be golden in color","kjv_def":"[idiom] fine."},
"H6669":{"lemma":"צָהֹב","xlit":"tsâhôb","pron":"tsaw-obe'","derivation":"from H6668 (צָהַב);","strongs_def":"golden in color","kjv_def":"yellow."},
"H6670":{"lemma":"צָהַל","xlit":"tsâhal","pron":"tsaw-hal'","derivation":"a primitive root;","strongs_def":"to gleam, i.e. (figuratively) be cheerful; by transference to sound clear (of various animal or human expressions)","kjv_def":"bellow, cry aloud (out), lift up, neigh, rejoice, make to shine, shout."},
"H6671":{"lemma":"צָהַר","xlit":"tsâhar","pron":"tsaw-har'","derivation":"a primitive root; to glisten; used only as denominative from H3323 (יִצְהָר),","strongs_def":"to press out oil","kjv_def":"make oil."},
"H6672":{"lemma":"צֹהַר","xlit":"tsôhar","pron":"tso'-har","derivation":"from H6671 (צָהַר);","strongs_def":"a light (i.e. window); dual double light, i.e. noon","kjv_def":"midday, noon(-day, -tide), window."},
"H6673":{"lemma":"צַו","xlit":"tsav","pron":"tsav","derivation":"or צָו; from H6680 (צָוָה);","strongs_def":"an injunction","kjv_def":"commandment, precept."},
"H6674":{"lemma":"צוֹא","xlit":"tsôwʼ","pron":"tso","derivation":"or צֹא; from an unused root meaning to issue;","strongs_def":"soiled (as if excrementitious)","kjv_def":"filthy."},
"H6675":{"lemma":"צוֹאָה","xlit":"tsôwʼâh","pron":"tso-aw'","derivation":"or צֹאָה ; feminine of H6674 (צוֹא);","strongs_def":"excrement; generally, dirt; figuratively, pollution","kjv_def":"dung, filth(-iness). Marg. for H2716 (חֶרֶא)."},
"H6676":{"lemma":"צַוַּאר","xlit":"tsavvaʼr","pron":"tsav-var'","derivation":"(Aramaic) corresponding to H6677 (צַוָּאר)","strongs_def":"{the back of the neck (as that on which burdens are bound)}","kjv_def":"neck."},
"H6677":{"lemma":"צַוָּאר","xlit":"tsavvâʼr","pron":"tsav-vawr'","derivation":"or צַוָּר; (Nehemiah 3:5), or צַוָּרֹןlemma צַוָּרן missing vowel, corrected to צַוָּרֹן; (Song of Solomon 4:9), or (feminine) צַוָּארָה; (Micah 2:3), intensively from H6696 (צוּר) in the sense of binding;","strongs_def":"the back of the neck (as that on which burdens are bound)","kjv_def":"neck."},
"H6678":{"lemma":"צוֹבָא","xlit":"Tsôwbâʼ","pron":"tso-baw'","derivation":"or צוֹבָה; or צֹבָה; from an unused root meaning to station; a station;","strongs_def":"Zoba or Zobah, a region of Syria","kjv_def":"Zoba, Zobah."},
"H6679":{"lemma":"צוּד","xlit":"tsûwd","pron":"tsood","derivation":"a primitive root; also denominative from H6718 (צַיִד)","strongs_def":"to lie alongside (i.e. in wait); by implication, to catch an animal (figuratively, men); to victual (for a journey)","kjv_def":"chase, hunt, sore, take (provision)."},
"H6680":{"lemma":"צָוָה","xlit":"tsâvâh","pron":"tsaw-vaw'","derivation":"a primitive root;","strongs_def":"(intensively) to constitute, enjoin","kjv_def":"appoint, (for-) bid, (give a) charge, (give a, give in, send with) command(-er, -ment), send a messenger, put, (set) in order."},
"H6681":{"lemma":"צָוַח","xlit":"tsâvach","pron":"tsaw-vakh'","derivation":"a primitive root;","strongs_def":"to screech (exultingly)","kjv_def":"shout."},
"H6682":{"lemma":"צְוָחָה","xlit":"tsᵉvâchâh","pron":"tsev-aw-khaw'","derivation":"from H6681 (צָוַח);","strongs_def":"a screech (of anguish)","kjv_def":"cry(-ing)."},
"H6683":{"lemma":"צוּלָה","xlit":"tsûwlâh","pron":"tsoo-law'","derivation":"from an unused root meaning to sink;","strongs_def":"an abyss (of the sea)","kjv_def":"deep."},
"H6684":{"lemma":"צוּם","xlit":"tsûwm","pron":"tsoom","derivation":"a primitive root;","strongs_def":"to cover over (the mouth), i.e. to fast","kjv_def":"[idiom] at all, fast."},
"H6685":{"lemma":"צוֹם","xlit":"tsôwm","pron":"tsome","derivation":"or צֹם; from H6684 (צוּם);","strongs_def":"a fast","kjv_def":"fast(-ing)."},
"H6686":{"lemma":"צוּעָר","xlit":"Tsûwʻâr","pron":"tsoo-awr'","derivation":"from H6819 (צָעַר); small;","strongs_def":"Tsuar, an Israelite","kjv_def":"Zuar."},
"H6687":{"lemma":"צוּף","xlit":"tsûwph","pron":"tsoof","derivation":"a primitive root;","strongs_def":"to overflow","kjv_def":"(make to over-) flow, swim."},
"H6688":{"lemma":"צוּף","xlit":"tsûwph","pron":"tsoof","derivation":"from H6687 (צוּף);","strongs_def":"comb of honey (from dripping)","kjv_def":"honeycomb."},
"H6689":{"lemma":"צוּף","xlit":"Tsûwph","pron":"tsoof","derivation":"or צוֹפַי; or צִיף; from H6688 (צוּף); honey-comb;","strongs_def":"Tsuph or Tsophai or Tsiph, the name of an Israelite and of a place in Palestine","kjv_def":"Zophai, Zuph."},
"H6690":{"lemma":"צוֹפַח","xlit":"Tsôwphach","pron":"tso-fakh'","derivation":"from an unused root meaning to expand, breadth;","strongs_def":"Tsophach, an Israelite","kjv_def":"Zophah."},
"H6691":{"lemma":"צוֹפַר","xlit":"Tsôwphar","pron":"tso-far'","derivation":"from H6852 (צָפַר); departing;","strongs_def":"Tsophar, a friend of Job","kjv_def":"Zophar."},
"H6692":{"lemma":"צוּץ","xlit":"tsûwts","pron":"tsoots","derivation":"a primitive root;","strongs_def":"to twinkle, i.e. glance; by analogy, to blossom (figuratively, flourish)","kjv_def":"bloom, blossom, flourish, shew self."},
"H6693":{"lemma":"צוּק","xlit":"tsûwq","pron":"tsook","derivation":"a primitive root;","strongs_def":"to compress, i.e. (figuratively) oppress, distress","kjv_def":"constrain, distress, lie sore, (op-) press(-or), straiten."},
"H6694":{"lemma":"צוּק","xlit":"tsûwq","pron":"tsook","derivation":"a primitive root (identical with through the idea of narrowness (of orifice));","strongs_def":"to pour out, i.e. (figuratively) smelt, utter","kjv_def":"be molten, pour."},
"H6695":{"lemma":"צוֹק","xlit":"tsôwq","pron":"tsoke","derivation":"or (feminine) צוּקָה; from H6693 (צוּק);","strongs_def":"a strait, i.e. (figuratively) distress","kjv_def":"anguish, [idiom] troublous."},
"H6696":{"lemma":"צוּר","xlit":"tsûwr","pron":"tsoor","derivation":"a primitive root;","strongs_def":"to cramp, i.e. confine (in many applications, literally and figuratively, formative or hostile)","kjv_def":"adversary, assault, beset, besiege, bind (up), cast, distress, fashion, fortify, inclose, lay siege, put up in bags."},
"H6697":{"lemma":"צוּר","xlit":"tsûwr","pron":"tsoor","derivation":"or צֻר; from H6696 (צוּר);","strongs_def":"properly, a cliff (or sharp rock, as compressed); generally, a rock or boulder; figuratively, a refuge; also an edge (as precipitous)","kjv_def":"edge, [idiom] (mighty) God (one), rock, [idiom] sharp, stone, [idiom] strength, [idiom] strong. See also H1049 (בֵּית צוּר)."},
"H6698":{"lemma":"צוּר","xlit":"Tsûwr","pron":"tsoor","derivation":"the same as H6697 (צוּר); rock;","strongs_def":"Tsur, the name of a Midianite and of an Israelite","kjv_def":"Zur."},
"H6699":{"lemma":"צוּרָה","xlit":"tsûwrâh","pron":"tsoo-raw'","derivation":"feminine of H6697 (צוּר);","strongs_def":"a rock (Job 28:10); also a form (as if pressed out)","kjv_def":"form, rock."},
"H6700":{"lemma":"צוּרִיאֵל","xlit":"Tsûwrîyʼêl","pron":"tsoo-ree-ale'","derivation":"from H6697 (צוּר) and H410 (אֵל); rock of God;","strongs_def":"Tsuriel, an Israelite","kjv_def":"Zuriel."},
"H6701":{"lemma":"צוּרִישַׁדַּי","xlit":"Tsûwrîyshadday","pron":"tsoo-ree-shad-dah'-ee","derivation":"from H6697 (צוּר) and H7706 (שַׁדַּי); rock of (the) Almighty;","strongs_def":"Tsurishaddai, an Israelite","kjv_def":"Zurishaddai."},
"H6702":{"lemma":"צוּת","xlit":"tsûwth","pron":"tsooth","derivation":"a primitive root;","strongs_def":"to blaze","kjv_def":"burn."},
"H6703":{"lemma":"צַח","xlit":"tsach","pron":"tsakh","derivation":"from H6705 (צָחַח);","strongs_def":"dazzling, i.e. sunny, bright, (figuratively) evident","kjv_def":"clear, dry, plainly, white."},
"H6704":{"lemma":"צִחֶה","xlit":"tsicheh","pron":"tsee-kheh'","derivation":"from an unused root meaning to glow;","strongs_def":"parched","kjv_def":"dried up."},
"H6705":{"lemma":"צָחַח","xlit":"tsâchach","pron":"tsaw-khakh'","derivation":"a primitive root;","strongs_def":"to glare, i.e. be dazzling white","kjv_def":"be whiter."},
"H6706":{"lemma":"צְחִיחַ","xlit":"tsᵉchîyach","pron":"tsekh-ee'-akh","derivation":"from H6705 (צָחַח);","strongs_def":"glaring, i.e. exposed to the bright sun","kjv_def":"higher place, top."},
"H6707":{"lemma":"צְחִיחָה","xlit":"tsᵉchîychâh","pron":"tsekh-ee-khaw'","derivation":"feminine of H6706 (צְחִיחַ);","strongs_def":"a parched region, i.e. the desert","kjv_def":"dry land."},
"H6708":{"lemma":"צְחִיחִי","xlit":"tsᵉchîychîy","pron":"tsekh-ee-khee'","derivation":"from H6706 (צְחִיחַ);","strongs_def":"bare spot, i.e. in the glaring sun","kjv_def":"higher place."},
"H6709":{"lemma":"צַחֲנָה","xlit":"tsachănâh","pron":"tsakh-an-aw'","derivation":"from an unused root meaning to putrefy;","strongs_def":"stench","kjv_def":"ill savour."},
"H6710":{"lemma":"צַחְצָחָה","xlit":"tsachtsâchâh","pron":"tsakh-tsaw-khaw'","derivation":"from H6705 (צָחַח);","strongs_def":"a dry place, i.e. desert","kjv_def":"drought."},
"H6711":{"lemma":"צָחַק","xlit":"tsâchaq","pron":"tsaw-khak'","derivation":"a primitive root;","strongs_def":"to laugh outright (in merriment or scorn); by implication, to sport","kjv_def":"laugh, mock, play, make sport."},
"H6712":{"lemma":"צְחֹק","xlit":"tsᵉchôq","pron":"tsekh-oke'","derivation":"from H6711 (צָחַק);","strongs_def":"laughter (in pleasure or derision)","kjv_def":"laugh(-ed to scorn)."},
"H6713":{"lemma":"צַחַר","xlit":"tsachar","pron":"tsakh'-ar","derivation":"from an unused root meaning to dazzle;","strongs_def":"sheen, i.e. whiteness","kjv_def":"white."},
"H6714":{"lemma":"צֹחַר","xlit":"Tsôchar","pron":"tso'-khar","derivation":"from the same as H6713 (צַחַר); whiteness;","strongs_def":"Tsochar, the name of a Hittite and of an Israelite","kjv_def":"Zohar. Compare H3328 (יִצְחַר)."},
"H6715":{"lemma":"צָחֹר","xlit":"tsâchôr","pron":"tsaw-khore'","derivation":"from the same as H6713 (צַחַר);","strongs_def":"white","kjv_def":"white."},
"H6716":{"lemma":"צִי","xlit":"tsîy","pron":"tsee","derivation":"from H6680 (צָוָה);","strongs_def":"a ship (as a fixture)","kjv_def":"ship."},
"H6717":{"lemma":"צִיבָא","xlit":"Tsîybâʼ","pron":"tsee-baw'","derivation":"from the same as H6678 (צוֹבָא); station;","strongs_def":"Tsiba, an Israelite","kjv_def":"Ziba."},
"H6718":{"lemma":"צַיִד","xlit":"tsayid","pron":"tsah'-yid","derivation":"from a form of H6679 (צוּד) and meaning the same;","strongs_def":"the chase; also game (thus taken); (generally) lunch (especially for a journey)","kjv_def":"[idiom] catcheth, food, [idiom] hunter, (that which he took in) hunting, venison, victuals."},
"H6719":{"lemma":"צַיָּד","xlit":"tsayâd","pron":"tsah'-yawd","derivation":"from the same as H6718 (צַיִד);","strongs_def":"a huntsman","kjv_def":"hunter."},
"H6720":{"lemma":"צֵידָה","xlit":"tsêydâh","pron":"tsay-daw'","derivation":"or צֵדָה; feminine of H6718 (צַיִד);","strongs_def":"food","kjv_def":"meat, provision, venison, victuals."},
"H6721":{"lemma":"צִידוֹן","xlit":"Tsîydôwn","pron":"tsee-done'","derivation":"or צִידֹן; from H6679 (צוּד) in the sense of catching fish; fishery;","strongs_def":"Tsidon, the name of a son of Canaan, and of a place in Palestine","kjv_def":"Sidon, Zidon."},
"H6722":{"lemma":"צִידֹנִי","xlit":"Tsîydônîy","pron":"tsee-do-nee'","derivation":"patrial from H6721 (צִידוֹן);","strongs_def":"a Tsidonian or inhabitant of Tsidon","kjv_def":"Sidonian, of Sidon, Zidonian."},
"H6723":{"lemma":"צִיָּה","xlit":"tsîyâh","pron":"tsee-yaw'","derivation":"from an unused root meaning to parch;","strongs_def":"aridity; concretely, a desert","kjv_def":"barren, drought, dry (land, place), solitary place, wilderness."},
"H6724":{"lemma":"צִיוֹן","xlit":"tsîyôwn","pron":"tsee-yone'","derivation":"from the same as H6723 (צִיָּה);","strongs_def":"a desert","kjv_def":"dry place."},
"H6725":{"lemma":"צִיּוּן","xlit":"tsîyûwn","pron":"tsee-yoon'","derivation":"from the same as H6723 (צִיָּה) in the sense of conspicuousness (compare H5329 (נָצַח));","strongs_def":"a monumental or guiding pillar","kjv_def":"sign, title, waymark."},
"H6726":{"lemma":"צִיּוֹן","xlit":"Tsîyôwn","pron":"tsee-yone'","derivation":"the same (regularly) as H6725 (צִיּוּן)lemma צִיוּן missing dagesh, corrected to צִיּוּן;","strongs_def":"Tsijon (as a permanent capital), a mountain of Jerusalem","kjv_def":"Zion."},
"H6727":{"lemma":"צִיחָא","xlit":"Tsîychâʼ","pron":"tsee-khaw'","derivation":"or צִחָא; as if feminine of H6704 (צִחֶה); drought;","strongs_def":"Tsicha, the name of two Nethinim","kjv_def":"Ziha."},
"H6728":{"lemma":"צִיִּי","xlit":"tsîyîy","pron":"tsee-ee'","derivation":"from the same as H6723 (צִיָּה);","strongs_def":"a desert-dweller, i.e. nomad or wild beast","kjv_def":"wild beast of the desert, that dwell in (inhabiting) the wilderness."},
"H6729":{"lemma":"צִינֹק","xlit":"tsîynôq","pron":"tsee-noke'","derivation":"from an unused root meaning to confine;","strongs_def":"the pillory","kjv_def":"stocks."},
"H6730":{"lemma":"צִיעֹר","xlit":"Tsîyʻôr","pron":"tsee-ore'","derivation":"from H6819 (צָעַר); small;","strongs_def":"Tsior, a place in Palestine","kjv_def":"Zior."},
"H6731":{"lemma":"צִיץ","xlit":"tsîyts","pron":"tseets","derivation":"or צִץ; from H6692 (צוּץ);","strongs_def":"properly, glistening, i.e. a burnished plate; also a flower (as bright-colored); a wing (as gleaming in the air)","kjv_def":"blossom, flower, plate, wing."},
"H6732":{"lemma":"צִיץ","xlit":"Tsîyts","pron":"tseets","derivation":"the same as H6731 (צִיץ); bloom;","strongs_def":"Tsits, a place in Palestine","kjv_def":"Ziz."},
"H6733":{"lemma":"צִיצָה","xlit":"tsîytsâh","pron":"tsee-tsaw'","derivation":"feminine of H6731 (צִיץ);","strongs_def":"a flower","kjv_def":"flower."},
"H6734":{"lemma":"צִיצִת","xlit":"tsîytsith","pron":"tsee-tseeth'","derivation":"feminine of H6731 (צִיץ);","strongs_def":"a floral or wing-like projection, i.e. a forelock of hair, a tassel","kjv_def":"fringe, lock."},
"H6735":{"lemma":"צִיר","xlit":"tsîyr","pron":"tseer","derivation":"from H6696 (צוּר);","strongs_def":"a hinge (as pressed in turning); also a throe (as a phys. or mental pressure); also a herald or erranddoer (as constrained by the principal)","kjv_def":"ambassador, hinge, messenger, pain, pang, sorrow. Compare H6736 (צִיר)."},
"H6736":{"lemma":"צִיר","xlit":"tsîyr","pron":"tseer","derivation":"the same as H6735 (צִיר);","strongs_def":"a form (of beauty; as if pressed out, i.e. carved); hence, an (idolatrous) image","kjv_def":"beauty, idol."},
"H6737":{"lemma":"צָיַר","xlit":"tsâyar","pron":"tsaw-yar'","derivation":"a denominative from H6735 (צִיר) in the sense of ambassador;","strongs_def":"to make an errand, i.e. betake oneself","kjv_def":"make as if...had been ambassador."},
"H6738":{"lemma":"צֵל","xlit":"tsêl","pron":"tsale","derivation":"from H6751 (צָלַל);","strongs_def":"shade, whether literal or figurative","kjv_def":"defence, shade(-ow)."},
"H6739":{"lemma":"צְלָא","xlit":"tsᵉlâʼ","pron":"tsel-aw'","derivation":"(Aramaic) probably corresponding to H6760 (צָלַע) in the sense of bowing;","strongs_def":"pray","kjv_def":"pray."},
"H6740":{"lemma":"צָלָה","xlit":"tsâlâh","pron":"tsaw-law'","derivation":"a primitive root;","strongs_def":"to roast","kjv_def":"roast."},
"H6741":{"lemma":"צִלָּה","xlit":"Tsillâh","pron":"tsil-law'","derivation":"feminine of H6738 (צֵל);","strongs_def":"Tsillah, an antediluvian woman","kjv_def":"Zillah."},
"H6742":{"lemma":"צְלוּל","xlit":"tsᵉlûwl","pron":"tsel-ool'","derivation":"from H6749 (צָלַל) in the sense of rolling;","strongs_def":"a (round or flattened) cake","kjv_def":"cake."},
"H6743":{"lemma":"צָלַח","xlit":"tsâlach","pron":"tsaw-lakh'","derivation":"or צָלֵחַ; a primitive root;","strongs_def":"to push forward, in various senses (literal or figurative, transitive or intransitive)","kjv_def":"break out, come (mightily), go over, be good, be meet, be profitable, (cause to, effect, make to, send) prosper(-ity, -ous, -ously)."},
"H6744":{"lemma":"צְלַח","xlit":"tsᵉlach","pron":"tsel-akh'","derivation":"(Aramaic) corresponding to H6743 (צָלַח);","strongs_def":"to advance (transitive or intransitive)","kjv_def":"promote, prosper."},
"H6745":{"lemma":"צֵלָחָה","xlit":"tsêlâchâh","pron":"tsay-law-khaw'","derivation":"from H6743 (צָלַח);","strongs_def":"something protracted or flattened out, i.e. a platter","kjv_def":"pan."},
"H6746":{"lemma":"צְלֹחִית","xlit":"tsᵉlôchîyth","pron":"tsel-o-kheeth'","derivation":"from H6743 (צָלַח);","strongs_def":"something prolonged or tall, i.e. a vial or salt-cellar","kjv_def":"cruse."},
"H6747":{"lemma":"צַלַּחַת","xlit":"tsallachath","pron":"tsal-lakh'-ath","derivation":"from H6743 (צָלַח);","strongs_def":"something advanced or deep, i.e. a bowl; figuratively, the bosom","kjv_def":"bosom, dish."},
"H6748":{"lemma":"צָלִי","xlit":"tsâlîy","pron":"tsaw-lee'","derivation":"passive participle of H6740 (צָלָה);","strongs_def":"roasted","kjv_def":"roast."},
"H6749":{"lemma":"צָלַל","xlit":"tsâlal","pron":"tsaw-lal'","derivation":"a primitive root;","strongs_def":"properly, to tumble down, i.e. settle by awaving motion","kjv_def":"sink. Compare H6750 (צָלַל), H6751 (צָלַל)."},
"H6750":{"lemma":"צָלַל","xlit":"tsâlal","pron":"tsaw-lal'","derivation":"a primitive root (identical with through the idea of vibration);","strongs_def":"to tinkle, i.e. rattle together (as the ears in reddening with shame, or the teeth in chattering with fear)","kjv_def":"quiver, tingle."},
"H6751":{"lemma":"צָלַל","xlit":"tsâlal","pron":"tsaw-lal'","derivation":"a primitive root (identical with through the idea of hovering over (compare H6754 (צֶלֶם)));","strongs_def":"to shade, as twilight or an opaque object","kjv_def":"begin to be dark, shadowing."},
"H6752":{"lemma":"צֵלֶל","xlit":"tsêlel","pron":"tsay'-lel","derivation":"from H6751 (צָלַל);","strongs_def":"shade","kjv_def":"shadow."},
"H6753":{"lemma":"צְלֶלְפּוֹנִי","xlit":"Tsᵉlelpôwnîy","pron":"tsel-el-po-nee'","derivation":"from H6752 (צֵלֶל) and the active participle of H6437 (פָּנָה); shade-facing;","strongs_def":"Tselelponi, an Israelitess","kjv_def":"Hazelelponi (including the article)."},
"H6754":{"lemma":"צֶלֶם","xlit":"tselem","pron":"tseh'-lem","derivation":"from an unused root meaning to shade;","strongs_def":"a phantom, i.e. (figuratively) illusion, resemblance; hence, a representative figure, especially an idol","kjv_def":"image, vain shew."},
"H6755":{"lemma":"צֶלֶם","xlit":"tselem","pron":"tseh'-lem","derivation":"(Aramaic) or צְלֶם; (Aramaic), corresponding to H6754 (צֶלֶם);","strongs_def":"an idolatrous figure","kjv_def":"form, image."},
"H6756":{"lemma":"צַלְמוֹן","xlit":"Tsalmôwn","pron":"tsal-mone'","derivation":"from H6754 (צֶלֶם); shady;","strongs_def":"Tsalmon, the name of a place in Palestine and of an Israelite","kjv_def":"Zalmon."},
"H6757":{"lemma":"צַלְמָוֶת","xlit":"tsalmâveth","pron":"tsal-maw'-veth","derivation":"from H6738 (צֵל) and H4194 (מָוֶת);","strongs_def":"shade of death, i.e. the grave (figuratively, calamity)","kjv_def":"shadow of death."},
"H6758":{"lemma":"צַלְמֹנָה","xlit":"Tsalmônâh","pron":"tsal-mo-naw'","derivation":"feminine of H6757 (צַלְמָוֶת); shadiness;","strongs_def":"Tsalmonah, a place in the Desert","kjv_def":"Zalmonah."},
"H6759":{"lemma":"צַלְמֻנָּע","xlit":"Tsalmunnâʻ","pron":"tsal-moon-naw'","derivation":"from H6738 (צֵל) and H4513 (מָנַע); shade has been denied;","strongs_def":"Tsalmunna, a Midianite","kjv_def":"Zalmunna."},
"H6760":{"lemma":"צָלַע","xlit":"tsâlaʻ","pron":"tsaw-lah'","derivation":"a primitive root; probably to curve; used only as denominative from H6763 (צֵלָע),","strongs_def":"to limp (as if one-sided)","kjv_def":"halt."},
"H6761":{"lemma":"צֶלַע","xlit":"tselaʻ","pron":"tseh'-lah","derivation":"from H6760 (צָלַע);","strongs_def":"a limping or full (figuratively)","kjv_def":"adversity, halt(-ing)."},
"H6762":{"lemma":"צֶלַע","xlit":"Tselaʻ","pron":"tseh'-lah","derivation":"the same as H6761 (צֶלַע);","strongs_def":"Tsela, a place in Palestine","kjv_def":"Zelah."},
"H6763":{"lemma":"צֵלָע","xlit":"tsêlâʻ","pron":"tsay-law'","derivation":"or (feminine) צַלְעָה; from H6760 (צָלַע);","strongs_def":"a rib (as curved), literally (of the body) or figuratively (of a door, i.e. leaf); hence, a side, literally (of a person) or figuratively (of an object or the sky, i.e. quarter); architecturally, a (especially floor or ceiling) timber or plank (single or collective, i.e. a flooring)","kjv_def":"beam, board, chamber, corner, leaf, plank, rib, side (chamber)."},
"H6764":{"lemma":"צָלָף","xlit":"Tsâlâph","pron":"tsaw-lawf'","derivation":"from an unused root of unknown meaning;","strongs_def":"Tsalaph, an Israelite","kjv_def":"Zalaph."},
"H6765":{"lemma":"צְלׇפְחָד","xlit":"Tsᵉlophchâd","pron":"tsel-of-chawd'","derivation":"from the same as H6764 (צָלָף) and H259 (אֶחָד);","strongs_def":"Tselophchad, an Israelite","kjv_def":"Zelophehad."},
"H6766":{"lemma":"צֶלְצַח","xlit":"Tseltsach","pron":"tsel-tsakh'","derivation":"from H6738 (צֵל) and H6703 (צַח); clear shade;","strongs_def":"Tseltsach, a place in Palestine","kjv_def":"Zelzah."},
"H6767":{"lemma":"צְלָצַל","xlit":"tsᵉlâtsal","pron":"tsel-aw-tsal'","derivation":"from H6750 (צָלַל) reduplicated;","strongs_def":"a clatter, i.e. (abstractly) whirring (of wings); (concretely) a cricket; also a harpoon (as rattling), a cymbal (as clanging)","kjv_def":"cymbal, locust, shadowing, spear."},
"H6768":{"lemma":"צֶלֶק","xlit":"Tseleq","pron":"tseh'-lek","derivation":"from an unused root meaning to split; fissure;","strongs_def":"Tselek, an Israelite","kjv_def":"Zelek."},
"H6769":{"lemma":"צִלְּתַי","xlit":"Tsillᵉthay","pron":"tsil-leth-ah'-ee","derivation":"from the feminine of H6738 (צֵל); shady;","strongs_def":"Tsillethai, the name of two Israelites","kjv_def":"Zilthai."},
"H6770":{"lemma":"צָמֵא","xlit":"tsâmêʼ","pron":"tsaw-may'","derivation":"a primitive root;","strongs_def":"to thirst (literally or figuratively)","kjv_def":"(be a-, suffer) thirst(-y)."},
"H6771":{"lemma":"צָמֵא","xlit":"tsâmêʼ","pron":"tsaw-may'","derivation":"from H6770 (צָמֵא);","strongs_def":"thirsty (literally or figuratively)","kjv_def":"(that) thirst(-eth, -y)."},
"H6772":{"lemma":"צָמָא","xlit":"tsâmâʼ","pron":"tsaw-maw'","derivation":"from H6770 (צָמֵא);","strongs_def":"thirst (literally or figuratively)","kjv_def":"thirst(-y)."},
"H6773":{"lemma":"צִמְאָה","xlit":"tsimʼâh","pron":"tsim-aw'","derivation":"feminine of H6772 (צָמָא);","strongs_def":"thirst (figuratively, of libidinousnes)","kjv_def":"thirst."},
"H6774":{"lemma":"צִמָּאוֹן","xlit":"tsimmâʼôwn","pron":"tsim-maw-one'","derivation":"from H6771 (צָמֵא);","strongs_def":"a thirsty place, i.e. desert","kjv_def":"drought, dry ground, thirsty land."},
"H6775":{"lemma":"צָמַד","xlit":"tsâmad","pron":"tsaw-mad'","derivation":"a primitive root;","strongs_def":"to link, i.e. gird; figuratively, to serve, (mentally) contrive","kjv_def":"fasten, frame, join (self)."},
"H6776":{"lemma":"צֶמֶד","xlit":"tsemed","pron":"tseh'-med","derivation":"a yoke or team (i.e. pair);","strongs_def":"hence, an acre (i.e. day's task for a yoke of cattle to plough)","kjv_def":"acre, couple, [idiom] together, two (donkeys), yoke (of oxen)."},
"H6777":{"lemma":"צַמָּה","xlit":"tsammâh","pron":"tsam-maw'","derivation":"from an unused root meaning to fasten on;","strongs_def":"a veil","kjv_def":"locks."},
"H6778":{"lemma":"צַמּוּק","xlit":"tsammûwq","pron":"tsam-mook'","derivation":"from H6784 (צָמַק);","strongs_def":"a cake of dried grapes","kjv_def":"bunch (cluster) of raisins."},
"H6779":{"lemma":"צָמַח","xlit":"tsâmach","pron":"tsaw-makh'","derivation":"a primitive root;","strongs_def":"to sprout (transitive or intransitive, literal or figurative)","kjv_def":"bear, bring forth, (cause to, make to) bud (forth), (cause to, make to) grow (again, up), (cause to) spring (forth, up)."},
"H6780":{"lemma":"צֶמַח","xlit":"tsemach","pron":"tseh'-makh","derivation":"from H6779 (צָמַח);","strongs_def":"a sprout (usually concrete), literal or figurative","kjv_def":"branch, bud, that which (where) grew (upon), spring(-ing)."},
"H6781":{"lemma":"צָמִיד","xlit":"tsâmîyd","pron":"tsaw-meed'","derivation":"or צָמִד; from H6775 (צָמַד);","strongs_def":"a bracelet or arm-clasp; generally, a lid","kjv_def":"bracelet, covering."},
"H6782":{"lemma":"צַמִּים","xlit":"tsammîym","pron":"tsam-meem'","derivation":"from the same as H6777 (צַמָּה);","strongs_def":"a noose (as fastening); figuratively, destruction","kjv_def":"robber."},
"H6783":{"lemma":"צְמִיתֻת","xlit":"tsᵉmîythuth","pron":"tsem-ee-thooth'","derivation":"or צְמִתֻת; from H6789 (צָמַת);","strongs_def":"excision, i.e. destruction; used only (adverbially) with prepositional prefix to extinction, i.e. perpetually","kjv_def":"ever."},
"H6784":{"lemma":"צָמַק","xlit":"tsâmaq","pron":"tsaw-mak'","derivation":"a primitive root;","strongs_def":"to dry up","kjv_def":"dry."},
"H6785":{"lemma":"צֶמֶר","xlit":"tsemer","pron":"tseh'-mer","derivation":"from an unused root probably meaning to be shaggy;","strongs_def":"wool","kjv_def":"wool(-len)."},
"H6786":{"lemma":"צְמָרִי","xlit":"Tsᵉmârîy","pron":"tsem-aw-ree'","derivation":"patrial from an unused name of a place in Palestine;","strongs_def":"a Tsemarite or branch of the Canaanites","kjv_def":"Zemarite."},
"H6787":{"lemma":"צְמָרַיִם","xlit":"Tsᵉmârayim","pron":"tsem-aw-rah'-yim","derivation":"dual of H6785 (צֶמֶר); double fleece;","strongs_def":"Tsemarajim, a place in Palestine","kjv_def":"Zemaraim."},
"H6788":{"lemma":"צַמֶּרֶת","xlit":"tsammereth","pron":"tsam-meh'-reth","derivation":"from the same as H6785 (צֶמֶר);","strongs_def":"fleeciness, i.e. foliage","kjv_def":"highest branch, top."},
"H6789":{"lemma":"צָמַת","xlit":"tsâmath","pron":"tsaw-math'","derivation":"a primitive root;","strongs_def":"to extirpate (literally or figuratively)","kjv_def":"consume, cut off, destroy, vanish."},
"H6790":{"lemma":"צִן","xlit":"Tsin","pron":"tseen","derivation":"from an unused root meaning to prick; a crag;","strongs_def":"Tsin, a part of the Desert","kjv_def":"Zin."},
"H6791":{"lemma":"צֵן","xlit":"tsên","pron":"tsane","derivation":"from an unused root meaning to be prickly;","strongs_def":"a thorn; hence, a cactus-hedge","kjv_def":"thorn."},
"H6792":{"lemma":"צֹנֵא","xlit":"tsônêʼ","pron":"tso-nay'","derivation":"or צֹנֶה; for H6629 (צֹאן);","strongs_def":"a flock","kjv_def":"sheep."},
"H6793":{"lemma":"צִנָּה","xlit":"tsinnâh","pron":"tsin-naw'","derivation":"feminine of H6791 (צֵן);","strongs_def":"a hook (as pointed); also a (large) shield (as if guarding by prickliness); also cold (as piercing)","kjv_def":"buckler, cold, hook, shield, target."},
"H6794":{"lemma":"צִנּוּר","xlit":"tsinnûwr","pron":"tsin-noor'","derivation":"from an unused root perhaps meaning to be hollow;","strongs_def":"a culvert","kjv_def":"gutter, water-spout."},
"H6795":{"lemma":"צָנַח","xlit":"tsânach","pron":"tsaw-nakh'","derivation":"a primitive root;","strongs_def":"to alight; (transitive) to cause to descend, i.e. drive down","kjv_def":"fasten, light (from off)."},
"H6796":{"lemma":"צָנִין","xlit":"tsânîyn","pron":"tsaw-neen'","derivation":"or צָנִן; from the same as H6791 (צֵן);","strongs_def":"a thorn","kjv_def":"thorn."},
"H6797":{"lemma":"צָנִיף","xlit":"tsânîyph","pron":"tsaw-neef'","derivation":"or צָנוֹף; or (feminine) צָנִיפָה; from H6801 (צָנַף);","strongs_def":"a head-dress (i.e. piece of cloth wrapped around)","kjv_def":"diadem, hood, mitre."},
"H6798":{"lemma":"צָנַם","xlit":"tsânam","pron":"tsaw-nam'","derivation":"a primitive root;","strongs_def":"to blast or shrink","kjv_def":"withered."},
"H6799":{"lemma":"צְנָן","xlit":"Tsᵉnân","pron":"tsen-awn'","derivation":"probably for H6630 (צַאֲנָן);","strongs_def":"Tsenan, a place near Palestine","kjv_def":"Zenan."},
"H6800":{"lemma":"צָנַע","xlit":"tsânaʻ","pron":"tsaw-nah'","derivation":"a primitive root;","strongs_def":"to humiliate","kjv_def":"humbly, lowly."},
"H6801":{"lemma":"צָנַף","xlit":"tsânaph","pron":"tsaw-naf'","derivation":"a primitive root;","strongs_def":"to wrap, i.e. roll or dress","kjv_def":"be attired, [idiom] surely, violently turn."},
"H6802":{"lemma":"צְנֵפָה","xlit":"tsᵉnêphâh","pron":"tsen-ay-faw'","derivation":"from H6801 (צָנַף);","strongs_def":"a ball","kjv_def":"[idiom] toss."},
"H6803":{"lemma":"צִנְצֶנֶת","xlit":"tsintseneth","pron":"tsin-tseh'-neth","derivation":"from the same as H6791 (צֵן);","strongs_def":"a vase (probably a vial tapering at the top)","kjv_def":"pot."},
"H6804":{"lemma":"צַנְתָּרָה","xlit":"tsantârâh","pron":"tsan-taw-raw'","derivation":"probably from the same as H6794 (צִנּוּר);","strongs_def":"a tube","kjv_def":"pipe."},
"H6805":{"lemma":"צָעַד","xlit":"tsâʻad","pron":"tsaw-ad'","derivation":"a primitive root;","strongs_def":"to pace, i.e. step regularly; (upward) to mount; (along) to march; (down and causatively) to hurl","kjv_def":"bring, go, march (through), run over."},
"H6806":{"lemma":"צַעַד","xlit":"tsaʻad","pron":"tsah'-ad","derivation":"from H6804 (צַנְתָּרָה);","strongs_def":"a pace or regular step","kjv_def":"pace, step."},
"H6807":{"lemma":"צְעָדָה","xlit":"tsᵉʻâdâh","pron":"tseh-aw-daw'","derivation":"feminine of H6806 (צַעַד);","strongs_def":"a march; (concretely) an (ornamental) ankle-chain","kjv_def":"going, ornament of the legs."},
"H6808":{"lemma":"צָעָה","xlit":"tsâʻâh","pron":"tsaw-aw'","derivation":"a primitive root;","strongs_def":"to tip over (for the purpose of spilling or pouring out), i.e. (figuratively) depopulate; by implication, to imprison or conquer; (reflexive) to lie down (for coitus)","kjv_def":"captive exile, travelling, (cause to) wander(-er)."},
"H6809":{"lemma":"צָעִיף","xlit":"tsâʻîyph","pron":"tsaw-eef'","derivation":"from an unused root meaning to wrap over;","strongs_def":"a veil","kjv_def":"vail."},
"H6810":{"lemma":"צָעִיר","xlit":"tsâʻîyr","pron":"tsaw-eer'","derivation":"or צָעוֹר; from H6819 (צָעַר);","strongs_def":"little; (in number) few; (in age) young, (in value) ignoble","kjv_def":"least, little (one), small (one), [phrase] young(-er, -est)."},
"H6811":{"lemma":"צָעִיר","xlit":"Tsâʻîyr","pron":"tsaw-eer'","derivation":"the same as H6810 (צָעִיר);","strongs_def":"Tsair, a place in Idumaea","kjv_def":"Zair."},
"H6812":{"lemma":"צְעִירָה","xlit":"tsᵉʻîyrâh","pron":"tseh-ee-raw'","derivation":"feminine of H6810 (צָעִיר);","strongs_def":"smallness (of age), i.e. juvenility","kjv_def":"youth."},
"H6813":{"lemma":"צָעַן","xlit":"tsâʻan","pron":"tsaw-an'","derivation":"a primitive root;","strongs_def":"to load up (beasts), i.e. to migrate","kjv_def":"be taken down."},
"H6814":{"lemma":"צֹעַן","xlit":"Tsôʻan","pron":"tso'-an","derivation":"of Egyptian derivation;","strongs_def":"Tsoan, a place in Egypt","kjv_def":"Zoan."},
"H6815":{"lemma":"צַעֲנַנִּים","xlit":"Tsaʻănannîym","pron":"tsah-an-an-neem'","derivation":"or (dual) צַעֲנַיִם; plural from H6813 (צָעַן); removals;","strongs_def":"Tsaanannim or Tsaanajim, a place in Palestine","kjv_def":"Zaannannim, Zaanaim."},
"H6816":{"lemma":"צַעְצֻעַ","xlit":"tsaʻtsuaʻ","pron":"tsah-tsoo'-ah","derivation":"from an unused root meaning to bestrew with carvings;","strongs_def":"sculpture","kjv_def":"image (work)."},
"H6817":{"lemma":"צָעַק","xlit":"tsâʻaq","pron":"tsaw-ak'","derivation":"a primitive root;","strongs_def":"to shriek; (by implication) to proclaim (an assembly)","kjv_def":"[idiom] at all, call together, cry (out), gather (selves) (together)."},
"H6818":{"lemma":"צַעֲקָה","xlit":"tsaʻăqâh","pron":"tsah-ak-aw'","derivation":"from H6817 (צָעַק);","strongs_def":"a shriek","kjv_def":"cry(-ing)."},
"H6819":{"lemma":"צָעַר","xlit":"tsâʻar","pron":"tsaw-ar'","derivation":"a primitive root;","strongs_def":"to be small, i.e. (figuratively) ignoble","kjv_def":"be brought low, little one, be small."},
"H6820":{"lemma":"צֹעַר","xlit":"Tsôʻar","pron":"tso'ar","derivation":"from H6819 (צָעַר); little;","strongs_def":"Tsoar, a place East of the Jordan","kjv_def":"Zoar."},
"H6821":{"lemma":"צָפַד","xlit":"tsâphad","pron":"tsaw-fad'","derivation":"a primitive root;","strongs_def":"to adhere","kjv_def":"cleave."},
"H6822":{"lemma":"צָפָה","xlit":"tsâphâh","pron":"tsaw-faw'","derivation":"a primitive root;","strongs_def":"properly, to lean forward, i.e. to peer into the distance; by implication, to observe, await","kjv_def":"behold, espy, look up (well), wait for, (keep the) watch(-man)."},
"H6823":{"lemma":"צָפָה","xlit":"tsâphâh","pron":"tsaw-faw'","derivation":"a primitive root (probably identical with H6822 (צָפָה) through the idea of expansion in outlook, transferring to action);","strongs_def":"to sheet over (especially with metal)","kjv_def":"cover, overlay."},
"H6824":{"lemma":"צָפָה","xlit":"tsâphâh","pron":"tsaw-faw'","derivation":"from H6823 (צָפָה);","strongs_def":"an inundation (as covering)","kjv_def":"[idiom] swimmest."},
"H6825":{"lemma":"צְפוֹ","xlit":"Tsᵉphôw","pron":"tsef-o'","derivation":"or צְפִי; from H6822 (צָפָה); observant;","strongs_def":"Tsepho or Tsephi, an Idumaean","kjv_def":"Zephi, Zepho."},
"H6826":{"lemma":"צִפּוּי","xlit":"tsippûwy","pron":"tsip-poo'-ee","derivation":"from H6823 (צָפָה);","strongs_def":"encasement (with metal)","kjv_def":"covering, overlaying."},
"H6827":{"lemma":"צְפוֹן","xlit":"Tsᵉphôwn","pron":"tsef-one'","derivation":"probably for H6837 (צִפְיוֹן);","strongs_def":"Tsephon, an Israelite","kjv_def":"Zephon."},
"H6828":{"lemma":"צָפוֹן","xlit":"tsâphôwn","pron":"tsaw-fone'","derivation":"or צָפֹן; from H6845 (צָפַן);","strongs_def":"properly, hidden, i.e. dark; used only of the north as aquarter (gloomy and unknown)","kjv_def":"north(-ern, side, -ward, wind)."},
"H6829":{"lemma":"צָפוֹן","xlit":"Tsâphôwn","pron":"tsaw-fone'","derivation":"the same as H6828 (צָפוֹן); boreal;","strongs_def":"Tsaphon, a place in Palestine","kjv_def":"Zaphon."},
"H6830":{"lemma":"צְפוֹנִי","xlit":"tsᵉphôwnîy","pron":"tsef-o-nee'","derivation":"from H6828 (צָפוֹן);","strongs_def":"northern","kjv_def":"northern."},
"H6831":{"lemma":"צְפוֹנִי","xlit":"Tsᵉphôwnîy","pron":"tsef-o-nee'","derivation":"patronymically from H6827 (צְפוֹן);","strongs_def":"a Tsephonite, or (collectively) descendants of Tsephon","kjv_def":"Zephonites."},
"H6832":{"lemma":"צְפוּעַ","xlit":"tsᵉphûwaʻ","pron":"tsef-oo'-ah","derivation":"from the same as H6848 (צֶפַע);","strongs_def":"excrement (as protruded)","kjv_def":"dung."},
"H6833":{"lemma":"צִפּוֹר","xlit":"tsippôwr","pron":"tsip-pore'","derivation":"or צִפֹּר; from H6852 (צָפַר);","strongs_def":"a little bird (as hopping)","kjv_def":"bird, fowl, sparrow."},
"H6834":{"lemma":"צִפּוֹר","xlit":"Tsippôwr","pron":"tsip-pore'","derivation":"the same as H6833 (צִפּוֹר);","strongs_def":"Tsippor, a Moabite","kjv_def":"Zippor."},
"H6835":{"lemma":"צַפַּחַת","xlit":"tsappachath","pron":"tsap-pakh'-ath","derivation":"from an unused root meaning to expand;","strongs_def":"a saucer (as flat)","kjv_def":"cruse."},
"H6836":{"lemma":"צְפִיָּה","xlit":"tsᵉphîyâh","pron":"tsef-ee-yaw'","derivation":"from H6822 (צָפָה);","strongs_def":"watchfulness","kjv_def":"watching."},
"H6837":{"lemma":"צִפְיוֹן","xlit":"Tsiphyôwn","pron":"tsif-yone'","derivation":"from H6822 (צָפָה); watch-tower;","strongs_def":"Tsiphjon, an Israelite","kjv_def":"Ziphion. Compare H6827 (צְפוֹן)."},
"H6838":{"lemma":"צַפִּיחִת","xlit":"tsappîychith","pron":"tsap-pee-kheeth'","derivation":"from the same as H6835 (צַפַּחַת);","strongs_def":"a flat thin cake","kjv_def":"wafer."},
"H6839":{"lemma":"צֹפִים","xlit":"Tsôphîym","pron":"tso-feem'","derivation":"plural of active participle of H6822 (צָפָה); watchers;","strongs_def":"Tsophim, a place East of the Jordan","kjv_def":"Zophim."},
"H6840":{"lemma":"צָפִין","xlit":"tsâphîyn","pron":"tsaw-feen'","derivation":"from H6845 (צָפַן);","strongs_def":"a treasure (as hidden)","kjv_def":"hid."},
"H6841":{"lemma":"צְפִיר","xlit":"tsᵉphîyr","pron":"tsef-eer'","derivation":"(Aramaic) corresponding to H6842 (צָפִיר);","strongs_def":"a he-goat","kjv_def":"he (goat)."},
"H6842":{"lemma":"צָפִיר","xlit":"tsâphîyr","pron":"tsaw-feer'","derivation":"from H6852 (צָפַר);","strongs_def":"a male goat (as prancing)","kjv_def":"(he) goat."},
"H6843":{"lemma":"צְפִירָה","xlit":"tsᵉphîyrâh","pron":"tsef-ee-raw'","derivation":"feminine formed like H6842 (צָפִיר);","strongs_def":"a crown (as encircling the head); also a turn of affairs (i.e. mishap)","kjv_def":"diadem, morning."},
"H6844":{"lemma":"צָפִית","xlit":"tsâphîyth","pron":"tsaw-feeth'","derivation":"from H6822 (צָפָה);","strongs_def":"a sentry","kjv_def":"watchtower."},
"H6845":{"lemma":"צָפַן","xlit":"tsâphan","pron":"tsaw-fan'","derivation":"a primitive root;","strongs_def":"to hide (by covering over); by implication, to hoard or reserve; figuratively to deny; specifically (favorably) to protect, (unfavorably) to lurk","kjv_def":"esteem, hide(-den one, self), lay up, lurk (be set) privily, (keep) secret(-ly, place)."},
"H6846":{"lemma":"צְפַנְיָה","xlit":"Tsᵉphanyâh","pron":"tsef-an-yaw'","derivation":"or צְפַנְיָהוּ; from H6845 (צָפַן) and H3050 (יָהּ); Jah has secreted;","strongs_def":"Tsephanjah, the name of four Israelites","kjv_def":"Zephaniah."},
"H6847":{"lemma":"צׇפְנַת פַּעְנֵחַ","xlit":"Tsophnath Paʻnêach","pron":"tsof-nath' pah-nay'-akh","derivation":"of Egyptian derivation;","strongs_def":"Tsophnath-Paneach, Joseph's Egyptian name","kjv_def":"Zaphnath-paaneah."},
"H6848":{"lemma":"צֶפַע","xlit":"tsephaʻ","pron":"tseh'-fah or tsiphoniy tsif-o-nee'","derivation":"from an unused root meaning to extrude;","strongs_def":"a viper (as thrusting out the tongue, i.e. hissing)","kjv_def":"adder, cockatrice."},
"H6849":{"lemma":"צְפִעָה","xlit":"tsᵉphiʻâh","pron":"tsef-ee-aw'","derivation":"feminine from the same as H6848 (צֶפַע);","strongs_def":"an outcast thing","kjv_def":"issue."},
"H6850":{"lemma":"צָפַף","xlit":"tsâphaph","pron":"tsaw-faf'","derivation":"a primitive root;","strongs_def":"to coo or chirp (as a bird)","kjv_def":"chatter, peep, whisper."},
"H6851":{"lemma":"צַפְצָפָה","xlit":"tsaphtsâphâh","pron":"tsaf-tsaw-faw'","derivation":"from H6687 (צוּף);","strongs_def":"a willow (as growing in overflowed places)","kjv_def":"willow tree."},
"H6852":{"lemma":"צָפַר","xlit":"tsâphar","pron":"tsaw-far'","derivation":"a primitive root;","strongs_def":"to skip about, i.e. return","kjv_def":"depart early."},
"H6853":{"lemma":"צְפַר","xlit":"tsᵉphar","pron":"tsef-ar'","derivation":"(Aramaic) corresponding to H6833 (צִפּוֹר);","strongs_def":"a bird.","kjv_def":"bird."},
"H6854":{"lemma":"צְפַרְדֵּעַ","xlit":"tsᵉphardêaʻ","pron":"tsef-ar-day'-ah","derivation":"from H6852 (צָפַר) and a word elsewhere unused meaning a swamp;","strongs_def":"a marsh-leaper, i.e. frog","kjv_def":"frog."},
"H6855":{"lemma":"צִפֹּרָה","xlit":"Tsippôrâh","pron":"tsip-po-raw'","derivation":"feminine of H6833 (צִפּוֹר); bird;","strongs_def":"Tsipporah, Moses' wife","kjv_def":"Zipporah."},
"H6856":{"lemma":"צִפֹּרֶן","xlit":"tsippôren","pron":"tsip-po'-ren","derivation":"from H6852 (צָפַר) (in the denominative sense (from H6833 (צִפּוֹר)) of scratching);","strongs_def":"properly, a claw, i.e. (human) nail; also the point of astyle (or pen, tipped with adamant)","kjv_def":"nail, point."},
"H6857":{"lemma":"צְפַת","xlit":"Tsᵉphath","pron":"tsef-ath'","derivation":"from H6822 (צָפָה); watch-tower;","strongs_def":"Tsephath, a place in Palestine","kjv_def":"Zephath."},
"H6858":{"lemma":"צֶפֶת","xlit":"tsepheth","pron":"tseh'-feth","derivation":"from an unused root meaning to encircle;","strongs_def":"a capital of a column","kjv_def":"chapiter."},
"H6859":{"lemma":"צְפָתָה","xlit":"Tsᵉphâthâh","pron":"tsef-aw'-thaw","derivation":"the same as H6857 (צְפַת);","strongs_def":"Tsephathah, a place in Palestine","kjv_def":"Zephathah."},
"H6860":{"lemma":"צִקְלַג","xlit":"Tsiqlag","pron":"tsik-lag'","derivation":"or צִיקלַג; (1 Chronicles 12:1,20), of uncertain derivation;","strongs_def":"Tsiklag or Tsikelag, a place in Palestine","kjv_def":"Ziklag."},
"H6861":{"lemma":"צִקְלֹן","xlit":"tsiqlôn","pron":"tsik-lone'","derivation":"from an unused root meaning to wind;","strongs_def":"a sack (as tied at the mouth)","kjv_def":"husk."},
"H6862":{"lemma":"צַר","xlit":"tsar","pron":"tsar","derivation":"or צָר; from H6887 (צָרַר); compare as in H6864 (צֹר)","strongs_def":"narrow; (as a noun) a tight place (usually figuratively, i.e. trouble); also a pebble ; (transitive) an opponent (as crowding)","kjv_def":"adversary, afflicted(-tion), anguish, close, distress, enemy, flint, foe, narrow, small, sorrow, strait, tribulation, trouble."},
"H6863":{"lemma":"צֵר","xlit":"Tsêr","pron":"tsare","derivation":"from H6887 (צָרַר); rock;","strongs_def":"Tser, a place in Palestine","kjv_def":"Zer."},
"H6864":{"lemma":"צֹר","xlit":"tsôr","pron":"tsore","derivation":"from H6696 (צוּר);","strongs_def":"a stone (as if pressed hard or to a point); (by implication, of use) a knife","kjv_def":"flint, sharp stone."},
"H6865":{"lemma":"צֹר","xlit":"Tsôr","pron":"tsore","derivation":"or צוֹר; the same as H6864 (צֹר); a rock;","strongs_def":"Tsor, a place in Palestine","kjv_def":"Tyre, Tyrus."},
"H6866":{"lemma":"צָרַב","xlit":"tsârab","pron":"tsaw-rab'","derivation":"a primitive root;","strongs_def":"to burn","kjv_def":"burn."},
"H6867":{"lemma":"צָרֶבֶת","xlit":"tsârebeth","pron":"tsaw-reh'-beth","derivation":"from H6686 (צוּעָר);","strongs_def":"conflagration (of fire or disease)","kjv_def":"burning, inflammation."},
"H6868":{"lemma":"צְרֵדָה","xlit":"Tsᵉrêdâh","pron":"tser-ay-daw'","derivation":"or צְרֵדָתָה; apparently from an unused root meaning to pierce; puncture;","strongs_def":"Tseredah, a place in Palestine","kjv_def":"Zereda, Zeredathah."},
"H6869":{"lemma":"צָרָה","xlit":"tsârâh","pron":"tsaw-raw'","derivation":"feminine of H6862 (צַר);","strongs_def":"tightness (i.e. figuratively, trouble); transitively, a female rival:","kjv_def":"adversary, adversity, affliction, anguish, distress, tribulation, trouble."},
"H6870":{"lemma":"צְרוּיָה","xlit":"Tsᵉrûwyâh","pron":"tser-oo-yaw'","derivation":"feminine passive participle from the same as H6875 (צְרִי); wounded;","strongs_def":"Tserujah, an Israelitess","kjv_def":"Zeruiah."},
"H6871":{"lemma":"צְרוּעָה","xlit":"Tsᵉrûwʻâh","pron":"tser-oo-aw'","derivation":"feminine passive participle of H6879 (צָרַע); leprous;","strongs_def":"Tseruah, an Israelitess","kjv_def":"Zeruah."},
"H6872":{"lemma":"צְרוֹר","xlit":"tsᵉrôwr","pron":"tser-ore'","derivation":"or (shorter) צְרֹר; from H6887 (צָרַר);","strongs_def":"a parcel (as packed up); also a kernel or particle (as if a package)","kjv_def":"bag, [idiom] bendeth, bundle, least grain, small stone."},
"H6873":{"lemma":"צָרַח","xlit":"tsârach","pron":"tsaw-rakh'","derivation":"a primitive root;","strongs_def":"to be clear (in tone, i.e. shrill), i.e. to whoop","kjv_def":"cry, roar."},
"H6874":{"lemma":"צְרִי","xlit":"Tsᵉrîy","pron":"tser-ee'","derivation":"the same as H6875 (צְרִי);","strongs_def":"Tseri, an Israelite","kjv_def":"Zeri. Compare H3340 (יִצְרִי)."},
"H6875":{"lemma":"צְרִי","xlit":"tsᵉrîy","pron":"tser-ee'","derivation":"or צֳרִי; from an unused root meaning to crack (as by pressure), hence, to leak;","strongs_def":"distillation, i.e. balsam","kjv_def":"balm."},
"H6876":{"lemma":"צֹרִי","xlit":"Tsôrîy","pron":"tso-ree'","derivation":"patrial from H6865 (צֹר);","strongs_def":"a Tsorite or inhabitant of Tsor (i.e. Syrian)","kjv_def":"(man) of Tyre."},
"H6877":{"lemma":"צְרִיחַ","xlit":"tsᵉrîyach","pron":"tser-ee'-akh","derivation":"from H6873 (צָרַח) in the sense of clearness of vision;","strongs_def":"a citadel","kjv_def":"high place, hold."},
"H6878":{"lemma":"צֹרֶךְ","xlit":"tsôrek","pron":"tso'-rek","derivation":"from an unused root meaning to need;","strongs_def":"need","kjv_def":"need."},
"H6879":{"lemma":"צָרַע","xlit":"tsâraʻ","pron":"tsaw-rah'","derivation":"a primitive root;","strongs_def":"to scourge, i.e. (intransitive and figurative) to be stricken with leprosy","kjv_def":"leper, leprous."},
"H6880":{"lemma":"צִרְעָה","xlit":"tsirʻâh","pron":"tsir-aw'","derivation":"from H6879 (צָרַע);","strongs_def":"a wasp (as stinging)","kjv_def":"hornet."},
"H6881":{"lemma":"צׇרְעָה","xlit":"Tsorʻâh","pron":"tsor-aw'","derivation":"apparently another form for H6880 (צִרְעָה);","strongs_def":"Tsorah, a place in Palestine","kjv_def":"Zareah, Zorah, Zoreah."},
"H6882":{"lemma":"צׇרְעִי","xlit":"Tsorʻîy","pron":"tsor-ee'","derivation":"or צׇרְעָתִי; patrial from H6881 (צׇרְעָה);","strongs_def":"a Tsorite or Tsorathite, i.e. inhabitants of Tsorah","kjv_def":"Zorites, Zareathites, Zorathites."},
"H6883":{"lemma":"צָרַעַת","xlit":"tsâraʻath","pron":"tsaw-rah'-ath","derivation":"from H6879 (צָרַע);","strongs_def":"leprosy","kjv_def":"leprosy."},
"H6884":{"lemma":"צָרַף","xlit":"tsâraph","pron":"tsaw-raf'","derivation":"a primitive root;","strongs_def":"to fuse (metal), i.e. refine (literally or figuratively)","kjv_def":"cast, (re-) fine(-er), founder, goldsmith, melt, pure, purge away, try."},
"H6885":{"lemma":"צֹרְפִי","xlit":"Tsôrᵉphîy","pron":"tso-ref-ee'","derivation":"from H6884 (צָרַף); refiner;","strongs_def":"Tsorephi (with the article), an Israelite","kjv_def":"goldsmith's."},
"H6886":{"lemma":"צָרְפַת","xlit":"Tsârᵉphath","pron":"tsaq-ref-ath'","derivation":"from H6884 (צָרַף); refinement;","strongs_def":"Tsarephath, a place in Palestine","kjv_def":"Zarephath."},
"H6887":{"lemma":"צָרַר","xlit":"tsârar","pron":"tsaw-rar'","derivation":"a primitive root;","strongs_def":"to cramp, literally or figuratively, transitive or intransitive","kjv_def":"adversary, (be in) afflict(-ion), beseige, bind (up), (be in, bring) distress, enemy, narrower, oppress, pangs, shut up, be in a strait (trouble), vex."},
"H6888":{"lemma":"צְרֵרָה","xlit":"Tsᵉrêrâh","pron":"tser-ay-raw'","derivation":"apparently by erroneous transcription for H6868 (צְרֵדָה);","strongs_def":"Tsererah for Tseredah","kjv_def":"Zererath."},
"H6889":{"lemma":"צֶרֶת","xlit":"Tsereth","pron":"tseh'-reth","derivation":"perhaps from H6671 (צָהַר); splendor;","strongs_def":"Tsereth, an Israelite","kjv_def":"Zereth."},
"H6890":{"lemma":"צֶרֶת הַשַּׁחַר","xlit":"Tsereth hash-Shachar","pron":"tseh'-reth hash-shakh'-ar","derivation":"from the same as H6889 (צֶרֶת) and H7837 (שַׁחַר) with the article interposed; splendor of the dawn;","strongs_def":"Tsereth-hash-Shachar, a place in Palestine","kjv_def":"Zareth-shahar."},
"H6891":{"lemma":"צָרְתָן","xlit":"Tsârᵉthân","pron":"tsaw-reth-awn'","derivation":"perhaps for H6868 (צְרֵדָה);","strongs_def":"Tsarethan, a place in Palestine","kjv_def":"Zarthan. q"},
"H6892":{"lemma":"קֵא","xlit":"qêʼ","pron":"kay","derivation":"or קִיא; from H6958 (קוֹא);","strongs_def":"vomit","kjv_def":"vomit."},
"H6893":{"lemma":"קָאַת","xlit":"qâʼath","pron":"kaw-ath'","derivation":"from H6958 (קוֹא);","strongs_def":"probably the pelican (from vomiting)","kjv_def":"cormorant."},
"H6894":{"lemma":"קַב","xlit":"qab","pron":"kab","derivation":"from H6895 (קָבַב);","strongs_def":"a hollow, i.e. vessel used as a (dry) measure","kjv_def":"cab."},
"H6895":{"lemma":"קָבַב","xlit":"qâbab","pron":"kaw-bab'","derivation":"a primitive root;","strongs_def":"to scoop out, i.e. (figuratively) to malign or execrate (i.e. stab with words)","kjv_def":"[idiom] at all, curse."},
"H6896":{"lemma":"קֵבָה","xlit":"qêbâh","pron":"kay-baw'","derivation":"from H6895 (קָבַב);","strongs_def":"the paunch (as a cavity) or first stomach of ruminants","kjv_def":"maw."},
"H6897":{"lemma":"קֹבָה","xlit":"qôbâh","pron":"ko'-baw","derivation":"from H6895 (קָבַב);","strongs_def":"the abdomen (as a cavity)","kjv_def":"belly."},
"H6898":{"lemma":"קֻבָּה","xlit":"qubbâh","pron":"koob-baw'","derivation":"from H6895 (קָבַב);","strongs_def":"a pavilion (as a domed cavity)","kjv_def":"tent."},
"H6899":{"lemma":"קִבּוּץ","xlit":"qibbûwts","pron":"kib-boots'","derivation":"from H6908 (קָבַץ);","strongs_def":"a throng","kjv_def":"company."},
"H6900":{"lemma":"קְבוּרָה","xlit":"qᵉbûwrâh","pron":"keb-oo-raw'","derivation":"or קְבֻרָה; feminine passive participle of H6912 (קָבַר);","strongs_def":"sepulture; (concretely) a sepulchre","kjv_def":"burial, burying place, grave, sepulchre."},
"H6901":{"lemma":"קָבַל","xlit":"qâbal","pron":"kaw-bal'","derivation":"a primitive root;","strongs_def":"to admit, i.e. take (literally or figuratively)","kjv_def":"choose, (take) hold, receive, (under-) take."},
"H6902":{"lemma":"קְבַל","xlit":"qᵉbal","pron":"keb-al'","derivation":"(Aramaic) corresponding to H6901 (קָבַל);","strongs_def":"to acquire","kjv_def":"receive, take."},
"H6903":{"lemma":"קְבֵל","xlit":"qᵉbêl","pron":"keb-ale'","derivation":"(Aramaic) or קֳבֵל; (Aramaic), (corresponding to H6905 (קָבָל);","strongs_def":"(adverbially) in front of; usually (with other particles) on account of, so as, since, hence","kjv_def":"[phrase] according to, [phrase] as, [phrase] because, before, [phrase] for this cause, [phrase] forasmuch as, [phrase] by this means, over against, by reason of, [phrase] that, [phrase] therefore, [phrase] though, [phrase] wherefore."},
"H6904":{"lemma":"קֹבֶל","xlit":"qôbel","pron":"ko'-bel","derivation":"from H6901 (קָבַל) in the sense of confronting (as standing opposite in order to receive);","strongs_def":"a battering-ram","kjv_def":"war."},
"H6905":{"lemma":"קָבָל","xlit":"qâbâl","pron":"kaw-bawl'","derivation":"from H6901 (קָבַל) in the sense of opposite (see H6904 (קֹבֶל));","strongs_def":"the presence, i.e. (adverbially) in front of","kjv_def":"before."},
"H6906":{"lemma":"קָבַע","xlit":"qâbaʻ","pron":"kaw-bah'","derivation":"a primitive root;","strongs_def":"to cover, i.e. (figuratively) defraud","kjv_def":"rob, spoil."},
"H6907":{"lemma":"קֻבַּעַת","xlit":"qubbaʻath","pron":"koob-bah'-ath","derivation":"from H6906 (קָבַע);","strongs_def":"a goblet (as deep like a cover)","kjv_def":"dregs."},
"H6908":{"lemma":"קָבַץ","xlit":"qâbats","pron":"kaw-bats'","derivation":"a primitive root;","strongs_def":"to grasp, i.e. collect","kjv_def":"assemble (selves), gather (bring) (together, selves together, up), heap, resort, [idiom] surely, take up."},
"H6909":{"lemma":"קַבְצְאֵל","xlit":"Qabtsᵉʼêl","pron":"kab-tseh-ale'","derivation":"from H6908 (קָבַץ) and H410 (אֵל); God has gathered;","strongs_def":"Kabtseel, a place in Palestine","kjv_def":"Kabzeel. Compare H3343 (יְקַבְצְאֵל)."},
"H6910":{"lemma":"קְבֻצָה","xlit":"qᵉbutsâh","pron":"keb-oo-tsaw'","derivation":"feminine passive participle of H6908 (קָבַץ);","strongs_def":"a hoard","kjv_def":"[idiom] gather."},
"H6911":{"lemma":"קִבְצַיִם","xlit":"Qibtsayim","pron":"kib-tsah'-yim","derivation":"dual from H6908 (קָבַץ); a double heap;","strongs_def":"Kibtsajim, a place in Palestine","kjv_def":"Kibzaim."},
"H6912":{"lemma":"קָבַר","xlit":"qâbar","pron":"kaw-bar'","derivation":"a primitive root;","strongs_def":"to inter","kjv_def":"[idiom] in any wise, bury(-ier)."},
"H6913":{"lemma":"קֶבֶר","xlit":"qeber","pron":"keh'-ber","derivation":"or (feminine) קִבְרָה; from H6912 (קָבַר);","strongs_def":"a sepulchre","kjv_def":"burying place, grave, sepulchre."},
"H6914":{"lemma":"קִבְרוֹת הַתַּאֲוָה","xlit":"Qibrôwth hat-Taʼă-vâh","pron":"kib-roth' hat-tahav-aw'","derivation":"from the feminine plural of H6913 (קֶבֶר) and H8378 (תַּאֲוָה) with the article interposed; graves of the longing;","strongs_def":"Kibroth-hat-Taavh, a place in the Desert","kjv_def":"Kibroth-hattaavah."},
"H6915":{"lemma":"קָדַד","xlit":"qâdad","pron":"kaw-dad'","derivation":"a primitive root;","strongs_def":"to shrivel up, i.e. contract or bend the body (or neck) in deference","kjv_def":"bow (down) (the) head, stoop."},
"H6916":{"lemma":"קִדָּה","xlit":"qiddâh","pron":"kid-daw'","derivation":"from H6915 (קָדַד);","strongs_def":"cassia bark (as in shrivelled rolls)","kjv_def":"cassia."},
"H6917":{"lemma":"קָדוּם","xlit":"qâdûwm","pron":"kaw-doom'","derivation":"passive participle of H6923 (קָדַם);","strongs_def":"a pristine hero","kjv_def":"ancient."},
"H6918":{"lemma":"קָדוֹשׁ","xlit":"qâdôwsh","pron":"kaw-doshe'","derivation":"or קָדֹשׁ; from H6942 (קָדַשׁ);","strongs_def":"sacred (ceremonially or morally); (as noun) God (by eminence), an angel, a saint, a sanctuary","kjv_def":"holy (One), saint."},
"H6919":{"lemma":"קָדַח","xlit":"qâdach","pron":"kaw-dakh'","derivation":"a primitive root to inflame","strongs_def":"to inflame","kjv_def":"burn, kindle."},
"H6920":{"lemma":"קַדַּחַת","xlit":"qaddachath","pron":"kad-dakh'-ath","derivation":"from H6919 (קָדַח);","strongs_def":"inflammation, i.e. febrile disease","kjv_def":"burning ague, fever."},
"H6921":{"lemma":"קָדִים","xlit":"qâdîym","pron":"kaw-deem'","derivation":"or קָדִם; from H6923 (קָדַם);","strongs_def":"the fore or front part; hence (by orientation) the East (often adverbially, eastward, for brevity the east wind)","kjv_def":"east(-ward, wind)."},
"H6922":{"lemma":"קַדִּישׁ","xlit":"qaddîysh","pron":"kad-deesh'","derivation":"(Aramaic) corresponding to H6918 (קָדוֹשׁ).","strongs_def":"{sacred (ceremonially or morally); (as noun) God (by eminence), an angel, a saint, a sanctuary}","kjv_def":"holy (One), saint."},
"H6923":{"lemma":"קָדַם","xlit":"qâdam","pron":"kaw-dam'","derivation":"a primitive root;","strongs_def":"to project (one self), i.e. precede; hence, to anticipate, hasten, meet (usually for help)","kjv_def":"come (go, (flee)) before, [phrase] disappoint, meet, prevent."},
"H6924":{"lemma":"קֶדֶם","xlit":"qedem","pron":"keh'-dem","derivation":"or קֵדְמָה; from H6923 (קָדַם);","strongs_def":"the front, of place (absolutely, the fore part, relatively the East) or time (antiquity); often used adverbially (before, anciently, eastward)","kjv_def":"aforetime, ancient (time), before, east (end, part, side, -ward), eternal, [idiom] ever(-lasting), forward, old, past. Compare H6926 (קִדְמָה)."},
"H6925":{"lemma":"קֳדָם","xlit":"qŏdâm","pron":"kod-awm'","derivation":"(Aramaic) or קְדָם; (Aramaic) (Daniel 7:13), corresponding to H6924 (קֶדֶם);","strongs_def":"before","kjv_def":"before, [idiom] from, [idiom] I (thought), [idiom] me, [phrase] of, [idiom] it pleased, presence."},
"H6926":{"lemma":"קִדְמָה","xlit":"qidmâh","pron":"kid-maw'","derivation":"feminine of H6924 (קֶדֶם);","strongs_def":"the forward part (or relatively) East (often adverbially, on the east or in front)","kjv_def":"east(-ward)."},
"H6927":{"lemma":"קַדְמָה","xlit":"qadmâh","pron":"kad-maw'","derivation":"from H6923 (קָדַם);","strongs_def":"priority (in time); also used adverbially (before)","kjv_def":"afore, antiquity, former (old) estate."},
"H6928":{"lemma":"קַדְמָה","xlit":"qadmâh","pron":"kad-maw'","derivation":"(Aramaic) corresponding to H6927 (קַדְמָה);","strongs_def":"former time","kjv_def":"afore(-time), ago."},
"H6929":{"lemma":"קֵדְמָה","xlit":"Qêdᵉmâh","pron":"kayd'-maw","derivation":"from H6923 (קָדַם); precedence;","strongs_def":"Kedemah, a son of Ishmael","kjv_def":"Kedemah."},
"H6930":{"lemma":"קַדְמוֹן","xlit":"qadmôwn","pron":"kad-mone'","derivation":"from H6923 (קָדַם);","strongs_def":"eastern","kjv_def":"east."},
"H6931":{"lemma":"קַדְמוֹנִי","xlit":"qadmôwnîy","pron":"kad-mo-nee'","derivation":"or קַדְמֹנִי; from H6930 (קַדְמוֹן);","strongs_def":"(of time) anterior or (of place) oriental","kjv_def":"ancient, they that went before, east, (thing of) old."},
"H6932":{"lemma":"קְדֵמוֹת","xlit":"Qᵉdêmôwth","pron":"ked-ay-mothe'","derivation":"from H6923 (קָדַם); beginnings;","strongs_def":"Kedemoth, a place in eastern Palestine","kjv_def":"Kedemoth."},
"H6933":{"lemma":"קַדְמַי","xlit":"qadmay","pron":"kad-mah'-ee","derivation":"(Aramaic) from a root corresponding to H6923 (קָדַם);","strongs_def":"first","kjv_def":"first."},
"H6934":{"lemma":"קַדְמִיאֵל","xlit":"Qadmîyʼêl","pron":"kad-mee-ale'","derivation":"from H6924 (קֶדֶם) and H410 (אֵל); presence of God;","strongs_def":"Kadmiel, the name of three Israelites","kjv_def":"Kadmiel."},
"H6935":{"lemma":"קַדְמֹנִי","xlit":"Qadmônîy","pron":"kad-mo-nee'","derivation":"the same as H6931 (קַדְמוֹנִי); ancient, i.e. aboriginal;","strongs_def":"Kadmonite (collectively), the name of a tribe in Palestine","kjv_def":"Kadmonites."},
"H6936":{"lemma":"קׇדְקֹד","xlit":"qodqôd","pron":"kod-kode'","derivation":"from H6915 (קָדַד);","strongs_def":"the crown of the head (as the part most bowed)","kjv_def":"crown (of the head), pate, scalp, top of the head."},
"H6937":{"lemma":"קָדַר","xlit":"qâdar","pron":"kaw-dar'","derivation":"a primitive root;","strongs_def":"to be ashy, i.e. darkcolored; by implication, to mourn (in sackcloth or sordid garments)","kjv_def":"be black(-ish), be (make) dark(-en), [idiom] heavily, (cause to) mourn."},
"H6938":{"lemma":"קֵדָר","xlit":"Qêdâr","pron":"kay-dawr'","derivation":"from H6937 (קָדַר); dusky (of the skin or the tent);","strongs_def":"Kedar, a son of Ishmael; also (collectively) Bedouin (as his descendants or representatives)","kjv_def":"Kedar."},
"H6939":{"lemma":"קִדְרוֹן","xlit":"Qidrôwn","pron":"kid-rone'","derivation":"from H6937 (קָדַר); dusky place;","strongs_def":"Kidron, a brook near Jerusalem","kjv_def":"Kidron."},
"H6940":{"lemma":"קַדְרוּת","xlit":"qadrûwth","pron":"kad-rooth'","derivation":"from H6937 (קָדַר);","strongs_def":"duskiness","kjv_def":"blackness."},
"H6941":{"lemma":"קְדֹרַנִּית","xlit":"qᵉdôrannîyth","pron":"ked-o-ran-neeth'","derivation":"adverb from H6937 (קָדַר);","strongs_def":"blackish ones (i.e. in sackcloth); used adverbially, in mourning weeds","kjv_def":"mournfully."},
"H6942":{"lemma":"קָדַשׁ","xlit":"qâdash","pron":"kaw-dash'","derivation":"a primitive root;","strongs_def":"to be (causatively, make, pronounce or observe as) clean (ceremonially or morally)","kjv_def":"appoint, bid, consecrate, dedicate, defile, hallow, (be, keep) holy(-er, place), keep, prepare, proclaim, purify, sanctify(-ied one, self), [idiom] wholly."},
"H6943":{"lemma":"קֶדֶשׁ","xlit":"Qedesh","pron":"keh'-desh","derivation":"from H6942 (קָדַשׁ); a sanctum;","strongs_def":"Kedesh, the name of four places in Palestine","kjv_def":"Kedesh."},
"H6944":{"lemma":"קֹדֶשׁ","xlit":"qôdesh","pron":"ko'-desh","derivation":"from H6942 (קָדַשׁ);","strongs_def":"a sacred place or thing; rarely abstract, sanctity","kjv_def":"consecrated (thing), dedicated (thing), hallowed (thing), holiness, ([idiom] most) holy ([idiom] day, portion, thing), saint, sanctuary."},
"H6945":{"lemma":"קָדֵשׁ","xlit":"qâdêsh","pron":"kaw-dashe'","derivation":"from H6942 (קָדַשׁ);","strongs_def":"a (quasi) sacred person, i.e. (technically) a (male) devotee (by prostitution) to licentious idolatry","kjv_def":"sodomite, unclean."},
"H6946":{"lemma":"קָדֵשׁ","xlit":"Qâdêsh","pron":"kaw-dashe'","derivation":"the same as H6945 (קָדֵשׁ); sanctuary;","strongs_def":"Kadesh, a place in the Desert","kjv_def":"Kadesh. Compare H6947 (קָדֵשׁ בַּרְנֵעַ)."},
"H6947":{"lemma":"קָדֵשׁ בַּרְנֵעַ","xlit":"Qâdêsh Barnêaʻ","pron":"kaw-dashe' bar-nay'-ah","derivation":"from the same as H6946 (קָדֵשׁ) and an otherwise unused word (apparently compounded of a correspondent to H1251 (בַּר) and a derivative of H5128 (נוּעַ)) meaning desert of a fugitive; Kadesh of (the) Wilderness of Wandering;","strongs_def":"Kadesh-Barnea, a place in the Desert","kjv_def":"Kadeshbarnea."},
"H6948":{"lemma":"קְדֵשָׁה","xlit":"qᵉdêshâh","pron":"ked-ay-shaw'","derivation":"feminine of H6945 (קָדֵשׁ);","strongs_def":"a female devotee (i.e. prostitute)","kjv_def":"harlot, whore."},
"H6949":{"lemma":"קָהָה","xlit":"qâhâh","pron":"kaw-haw'","derivation":"a primitive root;","strongs_def":"to be dull","kjv_def":"be set on edge, be blunt."},
"H6950":{"lemma":"קָהַל","xlit":"qâhal","pron":"'kaw-hal'","derivation":"a primitive root;","strongs_def":"to convoke","kjv_def":"assemble (selves) (together), gather (selves) (together)."},
"H6951":{"lemma":"קָהָל","xlit":"qâhâl","pron":"kaw-hawl'","derivation":"from H6950 (קָהַל);","strongs_def":"assemblage (usually concretely)","kjv_def":"assembly, company, congregation, multitude."},
"H6952":{"lemma":"קְהִלָּה","xlit":"qᵉhillâh","pron":"keh-hil-law'","derivation":"from H6950 (קָהַל);","strongs_def":"an assemblage","kjv_def":"assembly, congregation."},
"H6953":{"lemma":"קֹהֶלֶת","xlit":"qôheleth","pron":"ko-heh'-leth","derivation":"feminine of active participle from H6950 (קָהַל);","strongs_def":"a (female) assembler (i.e. lecturer); abstractly, preaching (used as a 'nom de plume', Koheleth)","kjv_def":"preacher."},
"H6954":{"lemma":"קְהֵלָתָה","xlit":"Qᵉhêlâthâh","pron":"keh-hay-law'-thaw","derivation":"from H6950 (קָהַל); convocation;","strongs_def":"Kehelathah, a place in the Desert","kjv_def":"Kehelathah."},
"H6955":{"lemma":"קְהָת","xlit":"Qᵉhâth","pron":"keh-hawth'","derivation":"from an unused root meaning to ally oneself; allied;","strongs_def":"Kehath, an Israelite","kjv_def":"Kohath."},
"H6956":{"lemma":"קֳהָתִי","xlit":"Qŏhâthîy","pron":"ko-haw-thee'","derivation":"patronymically from H6955 (קְהָת);","strongs_def":"a Kohathite (collectively) or descendants of Kehath","kjv_def":"Kohathites."},
"H6957":{"lemma":"קַו","xlit":"qav","pron":"kav","derivation":"or קָו; from H6960 (קָוָה) (compare H6961 (קָוֶה));","strongs_def":"a cord (as connecting), especially formeasuring; figuratively, a rule; also a rim, a musical string or accord","kjv_def":"line. Compare H6978 (קַו־קַו)lemma קַו־קַי yod, corrected to קַו־קַו."},
"H6958":{"lemma":"קוֹא","xlit":"qôwʼ","pron":"ko","derivation":"or קָיָה; (Jeremiah 25:27), a primitive root;","strongs_def":"to vomit","kjv_def":"spue (out), vomit (out, up, up again)."},
"H6959":{"lemma":"קוֹבַע","xlit":"qôwbaʻ","pron":"ko'-bah or ko-bah'","derivation":"a form collateral to H3553 (כּוֹבַע);","strongs_def":"a helmet","kjv_def":"helmet."},
"H6960":{"lemma":"קָוָה","xlit":"qâvâh","pron":"kaw-vaw'","derivation":"a primitive root;","strongs_def":"to bind together (perhaps by twisting), i.e. collect; (figuratively) to expect","kjv_def":"gather (together), look, patiently, tarry, wait (for, on, upon)."},
"H6961":{"lemma":"קָוֶה","xlit":"qâveh","pron":"kaw-veh'","derivation":"from H6960 (קָוָה);","strongs_def":"a (measuring) cord (as if for binding)","kjv_def":"line."},
"H6962":{"lemma":"קוּט","xlit":"qûwṭ","pron":"koot","derivation":"a primitive root;","strongs_def":"properly, to cut off, i.e. (figuratively) detest","kjv_def":"begrieved, loathe self."},
"H6963":{"lemma":"קוֹל","xlit":"qôwl","pron":"kole","derivation":"or קֹל; from an unused root meaning to call aloud;","strongs_def":"a voice or sound","kjv_def":"[phrase] aloud, bleating, crackling, cry ([phrase] out), fame, lightness, lowing, noise, [phrase] hold peace, (pro-) claim, proclamation, [phrase] sing, sound, [phrase] spark, thunder(-ing), voice, [phrase] yell."},
"H6964":{"lemma":"קוֹלָיָה","xlit":"Qôwlâyâh","pron":"ko-law-yaw'","derivation":"from H6963 (קוֹל) and H3050 (יָהּ); voice of Jah;","strongs_def":"Kolajah, the name of two Israelites","kjv_def":"Kolaiah."},
"H6965":{"lemma":"קוּם","xlit":"qûwm","pron":"koom","derivation":"a primitive root;","strongs_def":"to rise (in various applications, literal, figurative, intensive and causative)","kjv_def":"abide, accomplish, [idiom] be clearer, confirm, continue, decree, [idiom] be dim, endure, [idiom] enemy, enjoin, get up, make good, help, hold, (help to) lift up (again), make, [idiom] but newly, ordain, perform, pitch, raise (up), rear (up), remain, (a-) rise (up) (again, against), rouse up, set (up), (e-) stablish, (make to) stand (up), stir up, strengthen, succeed, (as-, make) sure(-ly), (be) up(-hold, -rising)."},
"H6966":{"lemma":"קוּם","xlit":"qûwm","pron":"koom","derivation":"(Aramaic) corresponding to H6965 (קוּם)","strongs_def":"{to rise (in various applications, literal, figurative, intensive and causative)}","kjv_def":"appoint, establish, make, raise up self, (a-) rise (up), (make to) stand, set (up)."},
"H6967":{"lemma":"קוֹמָה","xlit":"qôwmâh","pron":"ko-maw'","derivation":"from H6965 (קוּם);","strongs_def":"height","kjv_def":"[idiom] along, height, high, stature, tall."},
"H6968":{"lemma":"קוֹמְמִיּוּת","xlit":"qôwmᵉmîyûwth","pron":"ko-mem-ee-yooth'","derivation":"from H6965 (קוּם);","strongs_def":"elevation, i.e. (adverbially) erectly (figuratively)","kjv_def":"upright."},
"H6969":{"lemma":"קוּן","xlit":"qûwn","pron":"koon","derivation":"a primitive root;","strongs_def":"to strike a musical note, i.e. chant or wail (at a funeral)","kjv_def":"lament, mourning woman."},
"H6970":{"lemma":"קוֹעַ","xlit":"Qôwaʻ","pron":"ko'-ah","derivation":"probably from H6972 (קוּץ) in the original sense of cutting off; curtailment;","strongs_def":"Koa, a region of Babylon","kjv_def":"Koa."},
"H6971":{"lemma":"קוֹף","xlit":"qôwph","pron":"kofe","derivation":"or קֹף; of foreign origin,","strongs_def":"a monkey","kjv_def":"ape."},
"H6972":{"lemma":"קוּץ","xlit":"qûwts","pron":"koots","derivation":"a primitive root; to clip off; used only as denominative from H7019 (קַיִץ);","strongs_def":"to spend the harvest season","kjv_def":"summer."},
"H6973":{"lemma":"קוּץ","xlit":"qûwts","pron":"koots","derivation":"a primitive root (identical with through the idea of severing oneself from (compare H6962 (קוּט)));","strongs_def":"to be (causatively, make) disgusted or anxious","kjv_def":"abhor, be distressed, be grieved, loathe, vex, be weary."},
"H6974":{"lemma":"קוּץ","xlit":"qûwts","pron":"koots","derivation":"a primitive root (identical with through the idea of abruptness in starting up from sleep (compare H3364 (יָקַץ)));","strongs_def":"to awake (literally or figuratively)","kjv_def":"arise, (be) (a-) wake, watch."},
"H6975":{"lemma":"קוֹץ","xlit":"qôwts","pron":"kotse","derivation":"or קֹץ; from H6972 (קוּץ) (in the sense of pricking);","strongs_def":"a thorn","kjv_def":"thorn."},
"H6976":{"lemma":"קוֹץ","xlit":"Qôwts","pron":"kotse","derivation":"the same as H6975 (קוֹץ);","strongs_def":"Kots, the name of two Israelites","kjv_def":"Koz, Hakkoz (including the article)."},
"H6977":{"lemma":"קְוֻצָּה","xlit":"qᵉvutstsâh","pron":"kev-oots-tsaw'","derivation":"feminine passive participle of H6972 (קוּץ) in its original sense;","strongs_def":"a forelock (as shorn)","kjv_def":"lock."},
"H6978":{"lemma":"קַו־קַו","xlit":"qav-qav","pron":"kav-kav'","derivation":"from H6957 (קַו) (in the sense of a fastening);","strongs_def":"stalwart","kjv_def":"[idiom] meted out."},
"H6979":{"lemma":"קוּר","xlit":"qûwr","pron":"koor","derivation":"a primitive root; also denominative from H7023 (קִיר)","strongs_def":"to trench; by implication, to throw forth;  to wall up, whether literal (to build a wall) or figurative (to estop)","kjv_def":"break down, cast out, destroy, dig."},
"H6980":{"lemma":"קוּר","xlit":"qûwr","pron":"koor","derivation":"from H6979 (קוּר);","strongs_def":"(only plural) trenches, i.e. a web (as if so formed)","kjv_def":"web."},
"H6981":{"lemma":"קוֹרֵא","xlit":"Qôwrêʼ","pron":"ko-ray'","derivation":"or קֹרֵא; (1 Chronicles 26:1), active participle of H7121 (קָרָא); crier;","strongs_def":"Kore, the name of two Israelites","kjv_def":"Kore."},
"H6982":{"lemma":"קוֹרָה","xlit":"qôwrâh","pron":"ko-raw'","derivation":"or קֹרָה; from H6979 (קוּר);","strongs_def":"a rafter (forming trenches as it were); by implication, a roof","kjv_def":"beam, roof."},
"H6983":{"lemma":"קוֹשׁ","xlit":"qôwsh","pron":"koshe","derivation":"a primitive root; to bend; used only as denominative for H3369 (יָקֹשׁ),","strongs_def":"to set a trap","kjv_def":"lay a snare."},
"H6984":{"lemma":"קוּשָׁיָהוּ","xlit":"qûwshâyâhûw","pron":"koo-shaw-yaw'-hoo","derivation":"from the passive participle of H6983 (קוֹשׁ) and H3050 (יָהּ); entrapped of Jah;","strongs_def":"Kushajah, an Israelite","kjv_def":"Kushaiah."},
"H6985":{"lemma":"קַט","xlit":"qaṭ","pron":"kat","derivation":"from H6990 (קָטַט) in the sense of abbreviation;","strongs_def":"a little, i.e. (adverbially) merely","kjv_def":"very."},
"H6986":{"lemma":"קֶטֶב","xlit":"qeṭeb","pron":"keh'-teb","derivation":"from an unused root meaning to cut off;","strongs_def":"ruin","kjv_def":"destroying, destruction."},
"H6987":{"lemma":"קֹטֶב","xlit":"qôṭeb","pron":"ko'-teb","derivation":"from the same as H6986 (קֶטֶב);","strongs_def":"extermination","kjv_def":"destruction."},
"H6988":{"lemma":"קְטוֹרָה","xlit":"qᵉṭôwrâh","pron":"ket-o-raw'","derivation":"from H6999 (קָטַר);","strongs_def":"perfume","kjv_def":"incense."},
"H6989":{"lemma":"קְטוּרָה","xlit":"Qᵉṭûwrâh","pron":"ket-oo-raw'","derivation":"feminine passive participle of H6999 (קָטַר); perfumed;","strongs_def":"Keturah, a wife of Abraham","kjv_def":"Keturah."},
"H6990":{"lemma":"קָטַט","xlit":"qâṭaṭ","pron":"kaw-tat'","derivation":"a primitive root;","strongs_def":"to clip off, i.e. (figuratively) destroy","kjv_def":"be cut off."},
"H6991":{"lemma":"קָטַל","xlit":"qâṭal","pron":"kaw-tal'","derivation":"a primitive root;","strongs_def":"properly, to cut off, i.e. (figuratively) put to death","kjv_def":"kill, slay."},
"H6992":{"lemma":"קְטַל","xlit":"qᵉṭal","pron":"ket-al'","derivation":"(Aramaic) corresponding to H6991 (קָטַל);","strongs_def":"to kill","kjv_def":"slay."},
"H6993":{"lemma":"קֶטֶל","xlit":"qeṭel","pron":"keh'-tel","derivation":"from H6991 (קָטַל);","strongs_def":"a violent death","kjv_def":"slaughter."},
"H6994":{"lemma":"קָטֹן","xlit":"qâṭôn","pron":"kaw-tone'","derivation":"a primitive root (rather denominative from H6996 (קָטָן));","strongs_def":"to diminish, i.e. be (causatively, make) diminutive or (figuratively) of no account","kjv_def":"be a (make) small (thing), be not worthy."},
"H6995":{"lemma":"קֹטֶן","xlit":"qôṭen","pron":"ko'-ten","derivation":"from H6994 (קָטֹן);","strongs_def":"a pettiness, i.e. the little finger","kjv_def":"little finger."},
"H6996":{"lemma":"קָטָן","xlit":"qâṭân","pron":"kaw-tawn'","derivation":"or קָטֹן; from H6962 (קוּט);","strongs_def":"abbreviated, i.e. diminutive, literally (in quantity, size or number) or figuratively (in age or importance)","kjv_def":"least, less(-er), little (one), small(-est, one, quantity, thing), young(-er, -est)."},
"H6997":{"lemma":"קָטָן","xlit":"Qâṭân","pron":"kaw-tawn'","derivation":"the same as H6996 (קָטָן); small;","strongs_def":"Katan, an Israelite","kjv_def":"Hakkatan (including the article)."},
"H6998":{"lemma":"קָטַף","xlit":"qâṭaph","pron":"kaw-taf'","derivation":"a primitive root;","strongs_def":"to strip off","kjv_def":"crop off, cut down (up), pluck."},
"H6999":{"lemma":"קָטַר","xlit":"qâṭar","pron":"kaw-tar'","derivation":"a primitive root (identical with through the idea of fumigation in a close place and perhaps thus driving out the occupants);","strongs_def":"to smoke, i.e. turn into fragrance by fire (especially as an act of worship)","kjv_def":"burn (incense, sacrifice) (upon), (altar for) incense, kindle, offer (incense, a sacrifice)."},
"H7000":{"lemma":"קָטַר","xlit":"qâṭar","pron":"kaw-tar'","derivation":"a primitive root;","strongs_def":"to inclose","kjv_def":"join."},
"H7001":{"lemma":"קְטַר","xlit":"qᵉṭar","pron":"ket-ar'","derivation":"(Aramaic) from a root corresponding to H7000 (קָטַר);","strongs_def":"a knot (as tied up), i.e. (figuratively) a riddle; also a vertebra (as if a knot)","kjv_def":"doubt, joint."},
"H7002":{"lemma":"קִטֵּר","xlit":"qiṭṭêr","pron":"kit-tare'","derivation":"from H6999 (קָטַר);","strongs_def":"perfume","kjv_def":"incense."},
"H7003":{"lemma":"קִטְרוֹן","xlit":"Qiṭrôwn","pron":"kit-rone'","derivation":"from H6999 (קָטַר); fumigative;","strongs_def":"Kitron, a place in Palestine","kjv_def":"Kitron."},
"H7004":{"lemma":"קְטֹרֶת","xlit":"qᵉṭôreth","pron":"ket-o'-reth","derivation":"from H6999 (קָטַר);","strongs_def":"a fumigation","kjv_def":"(sweet) incense, perfume."},
"H7005":{"lemma":"קַטָּת","xlit":"Qaṭṭâth","pron":"kat-tawth'","derivation":"from H6996 (קָטָן); littleness;","strongs_def":"Kattath, a place in Palestine","kjv_def":"Kattath."},
"H7006":{"lemma":"קָיָה","xlit":"qâyâh","pron":"kaw-yaw'","derivation":"a primitive root;","strongs_def":"to vomit","kjv_def":"spue."},
"H7007":{"lemma":"קַיִט","xlit":"qayiṭ","pron":"kah'-yit","derivation":"(Aramaic) corresponding to H7019 (קַיִץ);","strongs_def":"harvest","kjv_def":"summer."},
"H7008":{"lemma":"קִיטוֹר","xlit":"qîyṭôwr","pron":"kee-tore'","derivation":"or קִיטֹר; from H6999 (קָטַר);","strongs_def":"a fume, i.e. cloud","kjv_def":"smoke, vapour."},
"H7009":{"lemma":"קִים","xlit":"qîym","pron":"keem","derivation":"from H6965 (קוּם);","strongs_def":"an opponent (as rising against one), i.e. (collectively) enemies","kjv_def":"substance."},
"H7010":{"lemma":"קְיָם","xlit":"qᵉyâm","pron":"keh-yawm'","derivation":"(Aramaic) from H6966 (קוּם);","strongs_def":"an edict (as arising in law)","kjv_def":"decree, statute."},
"H7011":{"lemma":"קַיָּם","xlit":"qayâm","pron":"kah-yawm'","derivation":"(Aramaic) from H6966 (קוּם);","strongs_def":"permanent (as rising firmly)","kjv_def":"stedfast, sure."},
"H7012":{"lemma":"קִימָה","xlit":"qîymâh","pron":"kee-maw'","derivation":"from H6965 (קוּם);","strongs_def":"an arising","kjv_def":"rising up."},
"H7013":{"lemma":"קַיִן","xlit":"qayin","pron":"kah'-yin","derivation":"from H6969 (קוּן) in the original sense of fixity;","strongs_def":"a lance (as striking fast)","kjv_def":"spear."},
"H7014":{"lemma":"קַיִן","xlit":"Qayin","pron":"kah'-yin","derivation":"the same as H7013 (קַיִן) (with a play upon the affinity to H7069 (קָנָה));","strongs_def":"Kajin, the name of the first child, also of a place in Palestine, and of an Oriental tribe","kjv_def":"Cain, Kenite(-s)."},
"H7015":{"lemma":"קִינָה","xlit":"qîynâh","pron":"kee-naw'","derivation":"from H6969 (קוּן);","strongs_def":"a dirge (as accompanied by beating the breasts or on instruments)","kjv_def":"lamentation."},
"H7016":{"lemma":"קִינָה","xlit":"Qîynâh","pron":"kee-naw'","derivation":"the same as H7015 (קִינָה);","strongs_def":"Kinah, a place in Palestine","kjv_def":"Kinah."},
"H7017":{"lemma":"קֵינִי","xlit":"Qêynîy","pron":"kay-nee'","derivation":"or קִינִי; (1 Chronicles 2:55), patronymic from H7014 (קַיִן);","strongs_def":"a Kenite or member of the tribe of Kajin","kjv_def":"Kenite."},
"H7018":{"lemma":"קֵינָן","xlit":"Qêynân","pron":"kay-nawn'","derivation":"from the same as H7064 (קֵן); fixed;","strongs_def":"Kenan, an antediluvian","kjv_def":"Cainan, Kenan."},
"H7019":{"lemma":"קַיִץ","xlit":"qayits","pron":"kah'-yits","derivation":"from H6972 (קוּץ);","strongs_def":"harvest (as the crop), whether the product (grain or fruit) or the (dry) season","kjv_def":"summer (fruit, house)."},
"H7020":{"lemma":"קִיצוֹן","xlit":"qîytsôwn","pron":"kee-tsone'","derivation":"from H6972 (קוּץ);","strongs_def":"terminal","kjv_def":"out-(utter-) most."},
"H7021":{"lemma":"קִיקָיוֹן","xlit":"qîyqâyôwn","pron":"kee-kaw-yone'","derivation":"perhaps from H7006 (קָיָה);","strongs_def":"the gourd (as nauseous)","kjv_def":"gourd."},
"H7022":{"lemma":"קִיקָלוֹן","xlit":"qîyqâlôwn","pron":"kee-kaw-lone'","derivation":"from H7036 (קָלוֹן);","strongs_def":"intense disgrace","kjv_def":"shameful spewing."},
"H7023":{"lemma":"קִיר","xlit":"qîyr","pron":"keer","derivation":"or קִר; (Isaiah 22:5), or (feminine) קִירָה; from H6979 (קוּר);","strongs_def":"a wall (as built in a trench)","kjv_def":"[phrase] mason, side, town, [idiom] very, wall."},
"H7024":{"lemma":"קִיר","xlit":"Qîyr","pron":"keer","derivation":"the same as H7023 (קִיר); fortress;","strongs_def":"Kir, a place in Assyrian; also one in Moab","kjv_def":"Kir. Compare H7025 (קִיר חֶרֶשׂ)."},
"H7025":{"lemma":"קִיר חֶרֶשׂ","xlit":"Qîyr Cheres","pron":"keer kheh'-res","derivation":"or (feminine of the latter word) קִיר חֲרֶשֶׂת; from H7023 (קִיר) and H2789 (חֶרֶשׂ); fortress of earthenware;","strongs_def":"Kir-Cheres or Kir-Chareseth, a place in Moab","kjv_def":"Kir-haraseth, Kir-hareseth, Kirharesh, Kir-heres."},
"H7026":{"lemma":"קֵירֹס","xlit":"Qêyrôç","pron":"kay-roce'","derivation":"or קֵרֹס; from the same as H7166 (קַרְסֹל); ankled;","strongs_def":"Keros, one of the Nethinim","kjv_def":"Keros."},
"H7027":{"lemma":"קִישׁ","xlit":"Qîysh","pron":"keesh","derivation":"from H6983 (קוֹשׁ); a bow;","strongs_def":"Kish, the name of five Israelites","kjv_def":"Kish."},
"H7028":{"lemma":"קִישׁוֹן","xlit":"Qîyshôwn","pron":"kee-shone'","derivation":"from H6983 (קוֹשׁ); winding;","strongs_def":"Kishon, a river of Palestine","kjv_def":"Kishon, Kison."},
"H7029":{"lemma":"קִישִׁי","xlit":"Qîyshîy","pron":"kee-shee'","derivation":"from H6983 (קוֹשׁ); bowed;","strongs_def":"Kishi, an Israelite","kjv_def":"Kishi."},
"H7030":{"lemma":"קִיתָרֹס","xlit":"qîythârôç","pron":"kee-thaw-roce'","derivation":"(Aramaic) of Greek origin;","strongs_def":"a lyre","kjv_def":"harp."},
"H7031":{"lemma":"קַל","xlit":"qal","pron":"kal","derivation":"contracted from H7043 (קָלַל);","strongs_def":"light; (by implication) rapid (also adverbial)","kjv_def":"light, swift(-ly)."},
"H7032":{"lemma":"קָל","xlit":"qâl","pron":"kawl","derivation":"(Aramaic) corresponding to H6963 (קוֹל)","strongs_def":"{a voice or sound}","kjv_def":"sound, voice."},
"H7033":{"lemma":"קָלָה","xlit":"qâlâh","pron":"kaw-law'","derivation":"a primitive root (rather identical with H7034 (קָלָה) through the idea of shrinkage by heat);","strongs_def":"to toast, i.e. scorch partially or slowly","kjv_def":"dried, loathsome, parch, roast."},
"H7034":{"lemma":"קָלָה","xlit":"qâlâh","pron":"kaw-law'","derivation":"a primitive root;","strongs_def":"to be light (as implied in rapid motion), but figuratively, only (be (causatively, hold) in contempt)","kjv_def":"base, contemn, despise, lightly esteem, set light, seem vile."},
"H7035":{"lemma":"קָלַהּ","xlit":"qâlahh","pron":"kaw-lah'","derivation":"for H6950 (קָהַל);","strongs_def":"to assemble","kjv_def":"gather together."},
"H7036":{"lemma":"קָלוֹן","xlit":"qâlôwn","pron":"kaw-lone'","derivation":"from H7034 (קָלָה);","strongs_def":"disgrace; (by implication) the pudenda","kjv_def":"confusion, dishonour, ignominy, reproach, shame."},
"H7037":{"lemma":"קַלַּחַת","xlit":"qallachath","pron":"kal-lakh'-ath","derivation":"apparently but a form for H6747 (צַלַּחַת);","strongs_def":"a kettle","kjv_def":"caldron."},
"H7038":{"lemma":"קָלַט","xlit":"qâlaṭ","pron":"kaw-lat'","derivation":"a primitive root;","strongs_def":"to maim","kjv_def":"lacking in his parts."},
"H7039":{"lemma":"קָלִי","xlit":"qâlîy","pron":"kaw-lee'","derivation":"or קָלִיא; from H7033 (קָלָה);","strongs_def":"roasted ears of grain","kjv_def":"parched corn."},
"H7040":{"lemma":"קַלַּי","xlit":"Qallay","pron":"kal-lah'-ee","derivation":"from H7043 (קָלַל); frivolous;","strongs_def":"Kallai, an Israelite","kjv_def":"Kallai."},
"H7041":{"lemma":"קֵלָיָה","xlit":"Qêlâyâh","pron":"kay-law-yaw'","derivation":"from H7034 (קָלָה); insignificance;","strongs_def":"Kelajah, an Israelite","kjv_def":"Kelaiah."},
"H7042":{"lemma":"קְלִיטָא","xlit":"Qᵉlîyṭâʼ","pron":"kel-ee-taw'","derivation":"from H7038 (קָלַט); maiming;","strongs_def":"Kelita, the name of three Israelites","kjv_def":"Kelita."},
"H7043":{"lemma":"קָלַל","xlit":"qâlal","pron":"kaw-lal'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) light, literally (swift, small, sharp, etc.) or figuratively (easy, trifling, vile, etc.)","kjv_def":"abate, make bright, bring into contempt, (ac-) curse, despise, (be) ease(-y, -ier), (be a, make, make somewhat, move, seem a, set) light(-en, -er, -ly, -ly afflict, -ly esteem, thing), [idiom] slight(-ly), be swift(-er), (be, be more, make, re-) vile, whet."},
"H7044":{"lemma":"קָלָל","xlit":"qâlâl","pron":"kaw-lawl'","derivation":"from H7043 (קָלַל);","strongs_def":"brightened (as if sharpened)","kjv_def":"burnished, polished."},
"H7045":{"lemma":"קְלָלָה","xlit":"qᵉlâlâh","pron":"kel-aw-law'","derivation":"from H7043 (קָלַל);","strongs_def":"vilification","kjv_def":"(ac-) curse(-d, -ing)."},
"H7046":{"lemma":"קָלַס","xlit":"qâlaç","pron":"kaw-las'","derivation":"a primitive root;","strongs_def":"to disparage, i.e. ridicule","kjv_def":"mock, scoff, scorn."},
"H7047":{"lemma":"קֶלֶס","xlit":"qeleç","pron":"keh'-les","derivation":"from H7046 (קָלַס);","strongs_def":"a laughing-stock","kjv_def":"derision."},
"H7048":{"lemma":"קַלָּסָה","xlit":"qallâçâh","pron":"kal-law-saw'","derivation":"intensive from H7046 (קָלַס);","strongs_def":"ridicule","kjv_def":"mocking."},
"H7049":{"lemma":"קָלַע","xlit":"qâlaʻ","pron":"kaw-lah'","derivation":"a primitive root;","strongs_def":"to sling; also to carve (as if a circular motion, or into light forms)","kjv_def":"carve, sling (out)."},
"H7050":{"lemma":"קֶלַע","xlit":"qelaʻ","pron":"keh'-lah","derivation":"from H7049 (קָלַע);","strongs_def":"a sling; also a (door) screen (as if slung across), or the valve (of the door) itself","kjv_def":"hanging, leaf, sling."},
"H7051":{"lemma":"קַלָּע","xlit":"qallâʻ","pron":"kal-law'","derivation":"intensive from H7049 (קָלַע);","strongs_def":"a slinger","kjv_def":"slinger."},
"H7052":{"lemma":"קְלֹקֵל","xlit":"qᵉlôqêl","pron":"kel-o-kale'","derivation":"from H7043 (קָלַל);","strongs_def":"insubstantial","kjv_def":"light."},
"H7053":{"lemma":"קִלְּשׁוֹן","xlit":"qillᵉshôwn","pron":"kil-lesh-one'","derivation":"from an unused root meaning to prick;","strongs_def":"a prong, i.e. hay-fork","kjv_def":"fork."},
"H7054":{"lemma":"קָמָה","xlit":"qâmâh","pron":"kuw-maw'","derivation":"feminine of active participle of H6965 (קוּם);","strongs_def":"something that rises, i.e. a stalk of grain","kjv_def":"(standing) corn, grown up, stalk."},
"H7055":{"lemma":"קְמוּאֵל","xlit":"Qᵉmûwʼêl","pron":"kem-oo-ale'","derivation":"from H6965 (קוּם) and H410 (אֵל); raised of God;","strongs_def":"Kemuel, the name of a relative of Abraham, and of two Israelites","kjv_def":"Kemuel."},
"H7056":{"lemma":"קָמוֹן","xlit":"Qâmôwn","pron":"kaw-mone'","derivation":"from H6965 (קוּם); an elevation;","strongs_def":"Kamon, a place East of the Jordan","kjv_def":"Camon."},
"H7057":{"lemma":"קִמּוֹשׁ","xlit":"qimmôwsh","pron":"kim-moshe'","derivation":"or קִימוֹשׁ; from an unused root meaning to sting;","strongs_def":"a prickly plant","kjv_def":"nettle. Compare H7063 (קִמָּשׁוֹן)."},
"H7058":{"lemma":"קֶמַח","xlit":"qemach","pron":"keh'-makh","derivation":"from an unused root probably meaning to grind;","strongs_def":"flour","kjv_def":"flour, meal."},
"H7059":{"lemma":"קָמַט","xlit":"qâmaṭ","pron":"kaw-mat'","derivation":"a primitive root;","strongs_def":"to pluck, i.e. destroy","kjv_def":"cut down, fill with wrinkles."},
"H7060":{"lemma":"קָמַל","xlit":"qâmal","pron":"kaw-mal'","derivation":"a primitive root;","strongs_def":"to wither","kjv_def":"hew down, wither."},
"H7061":{"lemma":"קָמַץ","xlit":"qâmats","pron":"kaw-mats'","derivation":"a primitive root;","strongs_def":"to grasp with the hand","kjv_def":"take an handful."},
"H7062":{"lemma":"קֹמֶץ","xlit":"qômets","pron":"ko'mets","derivation":"from H7061 (קָמַץ);","strongs_def":"a grasp, i.e. handful","kjv_def":"handful."},
"H7063":{"lemma":"קִמָּשׁוֹן","xlit":"qimmâshôwn","pron":"kim-maw-shone'","derivation":"from the same as H7057 (קִמּוֹשׁ);","strongs_def":"a prickly plant","kjv_def":"thorn."},
"H7064":{"lemma":"קֵן","xlit":"qên","pron":"kane","derivation":"contracted from H7077 (קָנַן);","strongs_def":"a nest (as fixed), sometimes including the nestlings; figuratively, a chamber or dwelling","kjv_def":"nest, room."},
"H7065":{"lemma":"קָנָא","xlit":"qânâʼ","pron":"kaw-naw'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) zealous, i.e. (in a bad sense) jealous or envious","kjv_def":"(be) envy(-ious), be (move to, provoke to) jealous(-y), [idiom] very, (be) zeal(-ous)."},
"H7066":{"lemma":"קְנָא","xlit":"qᵉnâʼ","pron":"ken-aw'","derivation":"(Aramaic) corresponding to H7069 (קָנָה);","strongs_def":"to purchase","kjv_def":"buy."},
"H7067":{"lemma":"קַנָּא","xlit":"qannâʼ","pron":"kan-naw'","derivation":"from H7065 (קָנָא);","strongs_def":"jealous","kjv_def":"jealous. Compare H7072 (קַנּוֹא)."},
"H7068":{"lemma":"קִנְאָה","xlit":"qinʼâh","pron":"kin-aw'","derivation":"from H7065 (קָנָא);","strongs_def":"jealousy or envy","kjv_def":"envy(-ied), jealousy, [idiom] sake, zeal."},
"H7069":{"lemma":"קָנָה","xlit":"qânâh","pron":"kaw-naw'","derivation":"a primitive root;","strongs_def":"to erect, i.e. create; by extension, to procure, especially by purchase (causatively, sell); by implication to own","kjv_def":"attain, buy(-er), teach to keep cattle, get, provoke to jealousy, possess(-or), purchase, recover, redeem, [idiom] surely, [idiom] verily."},
"H7070":{"lemma":"קָנֶה","xlit":"qâneh","pron":"kaw-neh'","derivation":"from H7069 (קָנָה);","strongs_def":"a reed (as erect); by resemblance a rod (especially for measuring), shaft, tube, stem, the radius (of the arm), beam (of a steelyard)","kjv_def":"balance, bone, branch, calamus, cane, reed, [idiom] spearman, stalk."},
"H7071":{"lemma":"קָנָה","xlit":"Qânâh","pron":"kaw-naw'","derivation":"feminine of H7070 (קָנֶה); reediness;","strongs_def":"Kanah, the name of a stream and of a place in Palestine","kjv_def":"Kanah."},
"H7072":{"lemma":"קַנּוֹא","xlit":"qannôwʼ","pron":"kan-no'","derivation":"for H7067 (קַנָּא);","strongs_def":"jealous or angry","kjv_def":"jealous."},
"H7073":{"lemma":"קְנַז","xlit":"Qᵉnaz","pron":"ken-az'","derivation":"probably from an unused root meaning to hunt; hunter;","strongs_def":"Kenaz, the name of an Edomite and of two Israelites","kjv_def":"Kenaz."},
"H7074":{"lemma":"קְנִזִּי","xlit":"Qᵉnizzîy","pron":"ken-iz-zee'","derivation":"patronymic from H7073 (קְנַז),","strongs_def":"a Kenizzite or descendant of Kenaz","kjv_def":"Kenezite, Kenizzites."},
"H7075":{"lemma":"קִנְיָן","xlit":"qinyân","pron":"kin-yawn'","derivation":"from H7069 (קָנָה);","strongs_def":"creation, i.e. (concretely) creatures; also acquisition, purchase, wealth","kjv_def":"getting, goods, [idiom] with money, riches, substance."},
"H7076":{"lemma":"קִנָּמוֹן","xlit":"qinnâmôwn","pron":"kin-naw-mone'","derivation":"from an unused root (meaning to erect);","strongs_def":"cinnamon bark (as in upright rolls)","kjv_def":"cinnamon."},
"H7077":{"lemma":"קָנַן","xlit":"qânan","pron":"kaw-nan'","derivation":"a primitive root; to erect; but used only as denominative from H7064 (קֵן);","strongs_def":"to nestle, i.e. build or occupy as anest","kjv_def":"make...nest."},
"H7078":{"lemma":"קֶנֶץ","xlit":"qenets","pron":"keh'-nets","derivation":"from an unused root probably meaning to wrench;","strongs_def":"perversion","kjv_def":"end."},
"H7079":{"lemma":"קְנָת","xlit":"Qᵉnâth","pron":"ken-awth'","derivation":"from H7069 (קָנָה); possession;","strongs_def":"Kenath, a place East of the Jordan","kjv_def":"Kenath."},
"H7080":{"lemma":"קָסַם","xlit":"qâçam","pron":"kaw-sam'","derivation":"a primitive root;","strongs_def":"properly, to distribute, i.e. determine by lot or magical scroll; by implication, to divine","kjv_def":"divine(-r, -ation), prudent, soothsayer, use (divination)."},
"H7081":{"lemma":"קֶסֶם","xlit":"qeçem","pron":"keh'-sem","derivation":"from H7080 (קָסַם);","strongs_def":"a lot; also divination (including its fee), oracle","kjv_def":"(reward of) divination, divine sentence, witchcraft."},
"H7082":{"lemma":"קָסַס","xlit":"qâçaç","pron":"kaw-sas'","derivation":"a primitive root;","strongs_def":"to lop off","kjv_def":"cut off."},
"H7083":{"lemma":"קֶסֶת","xlit":"qeçeth","pron":"keh'-seth","derivation":"from the same as H3563 (כּוֹס) (or as H7185 (קָשָׁה));","strongs_def":"properly, a cup, i.e. an ink-stand","kjv_def":"inkhorn."},
"H7084":{"lemma":"קְעִילָה","xlit":"Qᵉʻîylâh","pron":"keh-ee-law'","derivation":"perhaps from H7049 (קָלַע) in the sense of enclosing; citadel;","strongs_def":"Keilah, a place in Palestine","kjv_def":"Keilah."},
"H7085":{"lemma":"קַעֲקַע","xlit":"qaʻăqaʻ","pron":"kah-ak-ah'","derivation":"from the same as H6970 (קוֹעַ);","strongs_def":"an incision or gash","kjv_def":"[phrase] mark."},
"H7086":{"lemma":"קְעָרָה","xlit":"qᵉʻârâh","pron":"keh-aw-raw'","derivation":"probably from H7167 (קָרַע);","strongs_def":"a bowl (as cut out hollow)","kjv_def":"charger, dish."},
"H7087":{"lemma":"קָפָא","xlit":"qâphâʼ","pron":"kaw-faw'","derivation":"a primitive root;","strongs_def":"to shrink, i.e. thicken (as unracked wine, curdled milk, clouded sky, frozen water)","kjv_def":"congeal, curdle, dark, settle."},
"H7088":{"lemma":"קָפַד","xlit":"qâphad","pron":"kaw-fad'","derivation":"a primitive root;","strongs_def":"to contract, i.e. roll together","kjv_def":"cut off."},
"H7089":{"lemma":"קְפָדָה","xlit":"qᵉphâdâh","pron":"kef-aw-daw'","derivation":"from H7088 (קָפַד);","strongs_def":"shrinking, i.e., terror","kjv_def":"destruction."},
"H7090":{"lemma":"קִפּוֹד","xlit":"qippôwd","pron":"kip-pode'","derivation":"or קִפֹּדlemma קִכֹּד kaf, corrected to קִפֹּד; from H7088 (קָפַד);","strongs_def":"a species of bird, perhaps the bittern (from its contracted form)","kjv_def":"bittern."},
"H7091":{"lemma":"קִפּוֹז","xlit":"qippôwz","pron":"kip-poze'","derivation":"from an unused root meaning to contract, i.e.","strongs_def":"spring forward; an arrow-snake (as darting on its prey)","kjv_def":"great owl."},
"H7092":{"lemma":"קָפַץ","xlit":"qâphats","pron":"kaw-fats'","derivation":"a primitive root;","strongs_def":"to draw together, i.e. close; by implication, to leap (by contracting the limbs); specifically, to die (from gathering up the feet)","kjv_def":"shut (up), skip, stop, take out of the way."},
"H7093":{"lemma":"קֵץ","xlit":"qêts","pron":"kates","derivation":"contracted from H7112 (קָצַץ);","strongs_def":"an extremity; adverbially (with prepositional prefix) after","kjv_def":"[phrase] after, (utmost) border, end, (in-) finite, [idiom] process."},
"H7094":{"lemma":"קָצַב","xlit":"qâtsab","pron":"kaw-tsab'","derivation":"a primitive root;","strongs_def":"to clip, or (generally) chop","kjv_def":"cut down, shorn."},
"H7095":{"lemma":"קֶצֶב","xlit":"qetseb","pron":"keh'-tseb","derivation":"from H7094 (קָצַב);","strongs_def":"shape (as if cut out); base (as if there cut off)","kjv_def":"bottom, size."},
"H7096":{"lemma":"קָצָה","xlit":"qâtsâh","pron":"kaw-tsaw'","derivation":"a primitive root;","strongs_def":"to cut off; (figuratively) to destroy; (partially) to scrape off","kjv_def":"cut off, cut short, scrape (off)."},
"H7097":{"lemma":"קָצֶה","xlit":"qâtseh","pron":"kaw-tseh'","derivation":"or (negative only) קֵצֶה; from H7096 (קָצָה); (used in a great variety of applications and idioms; compare H7093 (קֵץ))","strongs_def":"an extremity","kjv_def":"[idiom] after, border, brim, brink, edge, end, (in-) finite, frontier, outmost coast, quarter, shore, (out-) side, [idiom] some, ut(-ter-) most (part)."},
"H7098":{"lemma":"קָצָה","xlit":"qâtsâh","pron":"kaw-tsaw'","derivation":"feminine of H7097 (קָצֶה); (used like H7097 (קָצֶה))","strongs_def":"a termination","kjv_def":"coast, corner, (selv-) edge, lowest, (uttermost) participle"},
"H7099":{"lemma":"קֶצֶו","xlit":"qetsev","pron":"keh'-tsev","derivation":"and (feminine) קִצְוָה; from H7096 (קָצָה); (used like H7097 (קָצֶה), but with less variety)","strongs_def":"a limit","kjv_def":"end, edge, uttermost participle"},
"H7100":{"lemma":"קֶצַח","xlit":"qetsach","pron":"keh'-tsakh","derivation":"from an unused root apparently meaning to incise;","strongs_def":"fennelflower (from its pungency)","kjv_def":"fitches."},
"H7101":{"lemma":"קָצִין","xlit":"qâtsîyn","pron":"kaw-tseen'","derivation":"from H7096 (קָצָה) in the sense of determining;","strongs_def":"a magistrate (as deciding) or other leader","kjv_def":"captain, guide, prince, ruler. Compare H6278 (עֵת קָצִין)."},
"H7102":{"lemma":"קְצִיעָה","xlit":"qᵉtsîyʻâh","pron":"kets-ee-aw'","derivation":"from H7106 (קָצַע);","strongs_def":"cassia (as peeled; plural the bark)","kjv_def":"cassia."},
"H7103":{"lemma":"קְצִיעָה","xlit":"Qᵉtsîyʻâh","pron":"kets-ee-aw'","derivation":"the same as H7102 (קְצִיעָה);","strongs_def":"Ketsiah, a daughter of Job","kjv_def":"Kezia."},
"H7104":{"lemma":"קְצִיץ","xlit":"Qᵉtsîyts","pron":"kets-eets'","derivation":"from H7112 (קָצַץ); abrupt;","strongs_def":"Keziz, a valley in Palestine","kjv_def":"Keziz."},
"H7105":{"lemma":"קָצִיר","xlit":"qâtsîyr","pron":"kaw-tseer'","derivation":"from H7114 (קָצַר);","strongs_def":"severed, i.e. harvest (as reaped), the crop, the time, the reaper, or figuratively; also a limb (of a tree, or simply foliage)","kjv_def":"bough, branch, harvest (man)."},
"H7106":{"lemma":"קָצַע","xlit":"qâtsaʻ","pron":"kaw-tsah'","derivation":"a primitive root;","strongs_def":"to strip off, i.e. (partially) scrape; by implication, to segregate (as an angle)","kjv_def":"cause to scrape, corner."},
"H7107":{"lemma":"קָצַף","xlit":"qâtsaph","pron":"kaw-tsaf'","derivation":"a primitive root;","strongs_def":"to crack off, i.e. (figuratively) burst out in rage","kjv_def":"(be) anger(-ry), displease, fret self, (provoke to) wrath (come), be wroth."},
"H7108":{"lemma":"קְצַף","xlit":"qᵉtsaph","pron":"kets-af'","derivation":"(Aramaic) corresponding to H7107 (קָצַף);","strongs_def":"to become enraged","kjv_def":"be furious."},
"H7109":{"lemma":"קְצַף","xlit":"qᵉtsaph","pron":"kets-af'","derivation":"(Aramaic) from H7108 (קְצַף);","strongs_def":"rage","kjv_def":"wrath."},
"H7110":{"lemma":"קֶצֶף","xlit":"qetseph","pron":"keh'-tsef","derivation":"from H7107 (קָצַף);","strongs_def":"a splinter (as chipped off); figuratively, rage or strife","kjv_def":"foam, indignation, [idiom] sore, wrath."},
"H7111":{"lemma":"קְצָפָה","xlit":"qᵉtsâphâh","pron":"kets-aw-faw'","derivation":"from H7107 (קָצַף);","strongs_def":"a fragment","kjv_def":"bark(-ed)."},
"H7112":{"lemma":"קָצַץ","xlit":"qâtsats","pron":"kaw-tsats'","derivation":"a primitive root;","strongs_def":"to chop off (literally or figuratively)","kjv_def":"cut (asunder, in pieces, in sunder, off), [idiom] utmost."},
"H7113":{"lemma":"קְצַץ","xlit":"qᵉtsats","pron":"kets-ats'","derivation":"(Aramaic) corresponding to H7112 (קָצַץ)","strongs_def":"{to chop off (literally or figuratively)}","kjv_def":"cut off."},
"H7114":{"lemma":"קָצַר","xlit":"qâtsar","pron":"kaw-tsar'","derivation":"a primitive root;","strongs_def":"to dock off, i.e. curtail (transitive or intransitive, literal or figurative); especially to harvest (grass or grain)","kjv_def":"[idiom] at all, cut down, much discouraged, grieve, harvestman, lothe, mourn, reap(-er), (be, wax) short(-en, -er), straiten, trouble, vex."},
"H7115":{"lemma":"קֹצֶר","xlit":"qôtser","pron":"ko'-tser","derivation":"from H7114 (קָצַר);","strongs_def":"shortness (of spirit), i.e. impatience","kjv_def":"anguish."},
"H7116":{"lemma":"קָצֵר","xlit":"qâtsêr","pron":"kaw-tsare'","derivation":"from H7114 (קָצַר);","strongs_def":"short (whether in size, number, life, strength or temper)","kjv_def":"few, hasty, small, soon."},
"H7117":{"lemma":"קְצָת","xlit":"qᵉtsâth","pron":"kets-awth'","derivation":"from H7096 (קָצָה);","strongs_def":"a termination (literally or figuratively); also (by implication) a portion; adverbially (with prepositional prefix) after","kjv_def":"end, part, [idiom] some."},
"H7118":{"lemma":"קְצָת","xlit":"qᵉtsâth","pron":"kets-awth'","derivation":"(Aramaic) corresponding to H7117 (קְצָת)","strongs_def":"{a termination (literally or figuratively); also (by implication) a portion; adverbially (with prepositional prefix) after}","kjv_def":"end, partly."},
"H7119":{"lemma":"קַר","xlit":"qar","pron":"kar","derivation":"contracted from an unused root meaning to chill;","strongs_def":"cool; figuratively, quiet","kjv_def":"cold, excellent (from the margin)."},
"H7120":{"lemma":"קֹר","xlit":"qôr","pron":"kore","derivation":"from the same as H7119 (קַר);","strongs_def":"cold","kjv_def":"cold."},
"H7121":{"lemma":"קָרָא","xlit":"qârâʼ","pron":"kaw-raw'","derivation":"a primitive root (rather identical with H7122 (קָרָא) through the idea of accosting a person met);","strongs_def":"to call out to (i.e. properly, address by name, but used in a wide variety of applications)","kjv_def":"bewray (self), that are bidden, call (for, forth, self, upon), cry (unto), (be) famous, guest, invite, mention, (give) name, preach, (make) proclaim(-ation), pronounce, publish, read, renowned, say."},
"H7122":{"lemma":"קָרָא","xlit":"qârâʼ","pron":"kaw-raw'","derivation":"a primitive root;","strongs_def":"to encounter, whether accidentally or in a hostile manner","kjv_def":"befall, (by) chance, (cause to) come (upon), fall out, happen, meet."},
"H7123":{"lemma":"קְרָא","xlit":"qᵉrâʼ","pron":"ker-aw'","derivation":"(Aramaic) corresponding to H7121 (קָרָא)","strongs_def":"{to call out to (i.e. properly, address by name, but used in a wide variety of applications)}","kjv_def":"call, cry, read."},
"H7124":{"lemma":"קֹרֵא","xlit":"qôrêʼ","pron":"ko-ray'","derivation":"properly, active participle of H7121 (קָרָא);","strongs_def":"a caller, i.e. partridge (from its cry)","kjv_def":"partridge. See also H6981 (קוֹרֵא)."},
"H7125":{"lemma":"קִרְאָה","xlit":"qirʼâh","pron":"keer-aw'","derivation":"from H7122 (קָרָא);","strongs_def":"an encountering, accidental, friendly or hostile (also adverbially, opposite)","kjv_def":"[idiom] against (he come), help, meet, seek, [idiom] to, [idiom] in the way."},
"H7126":{"lemma":"קָרַב","xlit":"qârab","pron":"kaw-rab'","derivation":"a primitive root;","strongs_def":"to approach (causatively, bring near) for whatever purpose","kjv_def":"(cause to) approach, (cause to) bring (forth, near), (cause to) come (near, nigh), (cause to) draw near (nigh), go (near), be at hand, join, be near, offer, present, produce, make ready, stand, take."},
"H7127":{"lemma":"קְרֵב","xlit":"qᵉrêb","pron":"ker-abe'","derivation":"(Aramaic) corresponding to H7126 (קָרַב)","strongs_def":"{to approach (causatively, bring near) for whatever purpose}","kjv_def":"approach, come (near, nigh), draw near."},
"H7128":{"lemma":"קְרָב","xlit":"qᵉrâb","pron":"ker-awb'","derivation":"from H7126 (קָרַב);","strongs_def":"hostile encounter","kjv_def":"battle, war."},
"H7129":{"lemma":"קְרָב","xlit":"qᵉrâb","pron":"ker-awb'","derivation":"(Aramaic) corresponding to H7128 (קְרָב)","strongs_def":"{hostile encounter}","kjv_def":"war."},
"H7130":{"lemma":"קֶרֶב","xlit":"qereb","pron":"keh'-reb","derivation":"from H7126 (קָרַב);","strongs_def":"properly, the nearest part, i.e. the center, whether literal, figurative or adverbial (especially with preposition)","kjv_def":"[idiom] among, [idiom] before, bowels, [idiom] unto charge, [phrase] eat (up), [idiom] heart, [idiom] him, [idiom] in, inward ([idiom] -ly, part, -s, thought), midst, [phrase] out of, purtenance, [idiom] therein, [idiom] through, [idiom] within self."},
"H7131":{"lemma":"קָרֵב","xlit":"qârêb","pron":"kaw-rabe'","derivation":"from H7126 (קָרַב);","strongs_def":"near","kjv_def":"approach, come (near, nigh), draw near."},
"H7132":{"lemma":"קְרָבָה","xlit":"qᵉrâbâh","pron":"ker-aw-baw'","derivation":"from H7126 (קָרַב);","strongs_def":"approach","kjv_def":"approaching, draw near."},
"H7133":{"lemma":"קׇרְבָּן","xlit":"qorbân","pron":"kor-bawn'","derivation":"or קֻרְבָּן; from H7126 (קָרַב);","strongs_def":"something brought near the altar, i.e. a sacrificial present","kjv_def":"oblation, that is offered, offering."},
"H7134":{"lemma":"קַרְדֹּם","xlit":"qardôm","pron":"kar-dome'","derivation":"perhaps from H6923 (קָדַם) in the sense of striking upon;","strongs_def":"an axe","kjv_def":"ax."},
"H7135":{"lemma":"קָרָה","xlit":"qârâh","pron":"kaw-raw'","derivation":"feminine of H7119 (קַר);","strongs_def":"coolness","kjv_def":"cold."},
"H7136":{"lemma":"קָרָה","xlit":"qârâh","pron":"kaw-raw'","derivation":"a primitive root;","strongs_def":"to light upon (chiefly by accident); causatively, to bring about; specifically, to impose timbers (for roof or floor)","kjv_def":"appoint, lay (make) beams, befall, bring, come (to pass unto), floor, (hap) was, happen (unto), meet, send good speed."},
"H7137":{"lemma":"קָרֶה","xlit":"qâreh","pron":"kaw-reh'","derivation":"from H7136 (קָרָה);","strongs_def":"an (unfortunate) occurrence, i.e. some accidental (ceremonial) disqualification","kjv_def":"uncleanness that chanceth."},
"H7138":{"lemma":"קָרוֹב","xlit":"qârôwb","pron":"kaw-robe'","derivation":"or קָרֹב; from H7126 (קָרַב);","strongs_def":"near (in place, kindred or time)","kjv_def":"allied, approach, at hand, [phrase] any of kin, kinsfold(-sman), (that is) near (of kin), neighbour, (that is) next, (them that come) nigh (at hand), more ready, short(-ly)."},
"H7139":{"lemma":"קָרַח","xlit":"qârach","pron":"kaw-rakh'","derivation":"a primitive root;","strongs_def":"to depilate","kjv_def":"make (self) bald."},
"H7140":{"lemma":"קֶרַח","xlit":"qerach","pron":"keh'-rakh","derivation":"or קֹרַח; from H7139 (קָרַח);","strongs_def":"ice (as if bald, i.e. smooth); hence, hail; by resemblance, rock crystal","kjv_def":"crystal, frost, ice."},
"H7141":{"lemma":"קֹרַח","xlit":"Qôrach","pron":"ko'rakh","derivation":"from H7139 (קָרַח); ice;","strongs_def":"Korach, the name of two Edomites and three Israelites","kjv_def":"Korah."},
"H7142":{"lemma":"קֵרֵחַ","xlit":"qêrêach","pron":"kay-ray'-akh","derivation":"from H7139 (קָרַח);","strongs_def":"bald (on the back of the head)","kjv_def":"bald (head)."},
"H7143":{"lemma":"קָרֵחַ","xlit":"Qârêach","pron":"kaw-ray'-akh","derivation":"from H7139 (קָרַח); bald;","strongs_def":"Kareach, an Israelite","kjv_def":"Careah, Kareah."},
"H7144":{"lemma":"קׇרְחָה","xlit":"qorchâh","pron":"kor-khaw'","derivation":"or קׇרְחָא; (Ezekiel 27:31), from H7139 (קָרַח);","strongs_def":"baldness","kjv_def":"bald(-ness), [idiom] utterly."},
"H7145":{"lemma":"קׇרְחִי","xlit":"Qorchîy","pron":"kor-khee'","derivation":"patronymic from H7141 (קֹרַח);","strongs_def":"a Korchite (collectively) or descendants of Korach","kjv_def":"Korahite, Korathite, sons of Kore, Korhite."},
"H7146":{"lemma":"קָרַחַת","xlit":"qârachath","pron":"kaw-rakh'-ath","derivation":"from H7139 (קָרַח);","strongs_def":"a bald spot (on the back of the head); figuratively, a threadbare spot (on the back side of the cloth)","kjv_def":"bald head, bare within."},
"H7147":{"lemma":"קְרִי","xlit":"qᵉrîy","pron":"ker-ee'","derivation":"from H7136 (קָרָה);","strongs_def":"hostile encounter","kjv_def":"contrary."},
"H7148":{"lemma":"קָרִיא","xlit":"qârîyʼ","pron":"kaw-ree'","derivation":"from H7121 (קָרָא);","strongs_def":"called, i.e. select","kjv_def":"famous, renowned."},
"H7149":{"lemma":"קִרְיָא","xlit":"qiryâʼ","pron":"keer-yaw'","derivation":"(Aramaic) or קִרְיָה; (Aramaic), corresponding to H7151 (קִרְיָה)","strongs_def":"{building; a city}","kjv_def":"city."},
"H7150":{"lemma":"קְרִיאָה","xlit":"qᵉrîyʼâh","pron":"ker-ee-aw'","derivation":"from H7121 (קָרָא);","strongs_def":"a proclamation","kjv_def":"preaching."},
"H7151":{"lemma":"קִרְיָה","xlit":"qiryâh","pron":"kir-yaw'","derivation":"from H7136 (קָרָה) in the sense of flooring, i.e.","strongs_def":"building; a city","kjv_def":"city."},
"H7152":{"lemma":"קְרִיּוֹת","xlit":"Qᵉrîyôwth","pron":"ker-ee-yoth'","derivation":"plural of H7151 (קִרְיָה); buildings;","strongs_def":"Kerioth, the name of two places in Palestine","kjv_def":"Kerioth, Kirioth."},
"H7153":{"lemma":"קִרְיַת אַרְבַּע","xlit":"Qiryath ʼArbaʻ","pron":"keer-yath' ar-bah'","derivation":"or (with the article interposed) קִרְיַת הָאַרְבַּע; (Nehemiah 11:25), from H7151 (קִרְיָה) and H704 (אַרְבַּע) or H702 (אַרְבַּע); city of Arba, or city of the four (giants);","strongs_def":"Kirjath-Arba or Kirjath-ha-Arba, a place in Palestine","kjv_def":"Kirjath-arba."},
"H7154":{"lemma":"קִרְיַת בַּעַל","xlit":"Qiryath Baʻal","pron":"keer-yath' bah'-al","derivation":"from H7151 (קִרְיָה) and H1168 (בַּעַל); city of Baal;","strongs_def":"Kirjath-Baal, a place in Palestine","kjv_def":"Kirjathbaal."},
"H7155":{"lemma":"קִרְיַת חֻצוֹת","xlit":"Qiryath Chutsôwth","pron":"keer-yath' khoo-tsoth'","derivation":"from H7151 (קִרְיָה) and the feminine plural of H2351 (חוּץ); city of streets;","strongs_def":"Kirjath-Chutsoth, a place in Moab","kjv_def":"Kirjath-huzoth."},
"H7156":{"lemma":"קִרְיָתַיִם","xlit":"Qiryâthayim","pron":"keer-yaw-thah'-yim","derivation":"dual of H7151 (קִרְיָה); double city;","strongs_def":"Kirjathaim, the name of two placed in Palestine","kjv_def":"Kiriathaim, Kirjathaim."},
"H7157":{"lemma":"קִרְיַת יְעָרִים","xlit":"Qiryath Yᵉʻârîym","pron":"keer-yath' yeh-aw-reem'","derivation":"or (Jeremiah 26:20) with the article interposed; or (Joshua 18:28) simply the former part of the word; or קִרְיַת עָרִים; from H7151 (קִרְיָה) and the plural of H3293 (יַעַר) or H5892 (עִיר); city of forests, or city of towns;","strongs_def":"Kirjath-Jearim or Kirjath-Arim, a place in Palestine","kjv_def":"Kirjath, Kirjath-jearim, Kirjath-arim."},
"H7158":{"lemma":"קִרְיַת סַנָּה","xlit":"Qiryath Çannâh","pron":"keer-yath' san-naw'","derivation":"or קִרְיַת סֵפֶר; from H7151 (קִרְיָה) and a simpler feminine from the same as H5577 (סַנְסִן), or (for the latter name) H5612 (סֵפֶר); city of branches, or of a book;","strongs_def":"Kirjath-Sannah or Kirjath-Sepher, a place in Palestine","kjv_def":"Kirjath-sannah, Kirjath-sepher."},
"H7159":{"lemma":"קָרַם","xlit":"qâram","pron":"kaw-ram'","derivation":"a primitive root;","strongs_def":"to cover","kjv_def":"cover."},
"H7160":{"lemma":"קָרַן","xlit":"qâran","pron":"kaw-ran'","derivation":"a primitive root; to push or gore; used only as denominative from H7161 (קֶרֶן),","strongs_def":"to shoot out horns; figuratively, rays","kjv_def":"have horns, shine."},
"H7161":{"lemma":"קֶרֶן","xlit":"qeren","pron":"keh'-ren","derivation":"from H7160 (קָרַן);","strongs_def":"a horn (as projecting); by implication, a flask, cornet; by resemblance. an elephant's tooth (i.e. ivory), a corner (of the altar), a peak (of a mountain), a ray (of light); figuratively, power","kjv_def":"[idiom] hill, horn."},
"H7162":{"lemma":"קֶרֶן","xlit":"qeren","pron":"keh'-ren","derivation":"(Aramaic) corresponding to H7161 (קֶרֶן);","strongs_def":"a horn (literally or for sound)","kjv_def":"horn, cornet."},
"H7163":{"lemma":"קֶרֶן הַפּוּךְ","xlit":"qeren hap-pûwk","pron":"keh'-ren hap-pook'","derivation":"from H7161 (קֶרֶן) and H6320 (פּוּךְ); horn of cosmetic;","strongs_def":"Keren-hap-Puk, one of Job's daughters","kjv_def":"Keren-happuch."},
"H7164":{"lemma":"קָרַס","xlit":"qâraç","pron":"kaw-ras'","derivation":"a primitive root; properly, to protrude; used only as denominative from H7165 (קֶרֶס) (for alliteration with H7167 (קָרַע)),","strongs_def":"to hunch, i.e. be hump-backed","kjv_def":"stoop."},
"H7165":{"lemma":"קֶרֶס","xlit":"qereç","pron":"keh'-res","derivation":"from H7164 (קָרַס);","strongs_def":"a knob or belaying-pin (from its swelling form)","kjv_def":"tache."},
"H7166":{"lemma":"קַרְסֹל","xlit":"qarçôl","pron":"kar-sole'","derivation":"from H7164 (קָרַס);","strongs_def":"an ankle (as a protuberance or joint)","kjv_def":"foot."},
"H7167":{"lemma":"קָרַע","xlit":"qâraʻ","pron":"kaw-rah'","derivation":"a primitive root;","strongs_def":"to rend, literally or figuratively (revile, paint the eyes, as if enlarging them)","kjv_def":"cut out, rend, [idiom] surely, tear."},
"H7168":{"lemma":"קֶרַע","xlit":"qeraʻ","pron":"keh'-rah","derivation":"from H7167 (קָרַע);","strongs_def":"a rag","kjv_def":"piece, rag."},
"H7169":{"lemma":"קָרַץ","xlit":"qârats","pron":"kaw-rats'","derivation":"a primitive root;","strongs_def":"to pinch, i.e. (partially) to bite the lips, blink the eyes (as a gesture of malice), or (fully) to squeeze off (a piece of clay in order to mould a vessel from it)","kjv_def":"form, move, wink."},
"H7170":{"lemma":"קְרַץ","xlit":"qᵉrats","pron":"ker-ats'","derivation":"(Aramaic) corresponding to H7171 (קֶרֶץ) in the sense of a bit (to 'eat the morsels of' any one, i.e.","strongs_def":"chew him up (figuratively) by slander","kjv_def":"[phrase] accuse."},
"H7171":{"lemma":"קֶרֶץ","xlit":"qerets","pron":"keh'-rets","derivation":"from H7169 (קָרַץ);","strongs_def":"extirpation (as if by constriction)","kjv_def":"destruction."},
"H7172":{"lemma":"קַרְקַע","xlit":"qarqaʻ","pron":"kar-kah'","derivation":"from H7167 (קָרַע);","strongs_def":"floor (as if a pavement of pieces or tesseroe), of abuilding or the sea","kjv_def":"bottom, ([idiom] one side of the) floor."},
"H7173":{"lemma":"קַרְקַע","xlit":"Qarqaʻ","pron":"kar-kah'","derivation":"the same as H7172 (קַרְקַע); ground-floor;","strongs_def":"Karka (with the article prefix), a place in Palestine","kjv_def":"Karkaa."},
"H7174":{"lemma":"קַרְקֹר","xlit":"Qarqôr","pron":"kar-kore'","derivation":"from H6979 (קוּר); foundation;","strongs_def":"Karkor, a place East of the Jordan","kjv_def":"Karkor."},
"H7175":{"lemma":"קֶרֶשׁ","xlit":"qeresh","pron":"keh'-resh","derivation":"from an unused root meaning to split off;","strongs_def":"a slab or plank; by implication, a deck of aship","kjv_def":"bench, board."},
"H7176":{"lemma":"קֶרֶת","xlit":"qereth","pron":"keh'-reth","derivation":"from H7136 (קָרָה) in the sense of building;","strongs_def":"a city","kjv_def":"city."},
"H7177":{"lemma":"קַרְתָּה","xlit":"Qartâh","pron":"kar-taw'","derivation":"from H7176 (קֶרֶת); city;","strongs_def":"Kartah, a place in Palestine","kjv_def":"Kartah."},
"H7178":{"lemma":"קַרְתָּן","xlit":"Qartân","pron":"kar-tawn'","derivation":"from H7176 (קֶרֶת); city-plot;","strongs_def":"Kartan, a place in Palestine","kjv_def":"Kartan."},
"H7179":{"lemma":"קַשׁ","xlit":"qash","pron":"kash","derivation":"from H7197 (קָשַׁשׁ);","strongs_def":"straw (as dry)","kjv_def":"stubble."},
"H7180":{"lemma":"קִשֻּׁא","xlit":"qishshuʼ","pron":"kish-shoo'","derivation":"from an unused root (meaning to be hard);","strongs_def":"a cucumber (from the difficulty of digestion)","kjv_def":"cucumber."},
"H7181":{"lemma":"קָשַׁב","xlit":"qâshab","pron":"kaw-shab'","derivation":"a primitive root;","strongs_def":"to prick up the ears, i.e. hearken","kjv_def":"attend, (cause to) hear(-ken), give heed, incline, mark (well), regard."},
"H7182":{"lemma":"קֶשֶׁב","xlit":"qesheb","pron":"keh'-sheb","derivation":"from H7181 (קָשַׁב);","strongs_def":"a hearkening","kjv_def":"[idiom] diligently, hearing, much heed, that regarded."},
"H7183":{"lemma":"קַשָּׁב","xlit":"qashshâb","pron":"kash-shawb'","derivation":"or קַשֻּׁב; from H7181 (קָשַׁב);","strongs_def":"hearkening","kjv_def":"attent(-ive)."},
"H7184":{"lemma":"קָשָׂה","xlit":"qâsâh","pron":"kaw-saw'","derivation":"or קַשְׂוָה; from an unused root meaning to be round;","strongs_def":"a jug (from its shape)","kjv_def":"cover, cup."},
"H7185":{"lemma":"קָשָׁה","xlit":"qâshâh","pron":"kaw-shaw'","derivation":"a primitive root;","strongs_def":"properly, to be dense, i.e. tough or severe (in various applications)","kjv_def":"be cruel, be fiercer, make grievous, be ((ask a), be in, have, seem, would) hard(-en, (labour), -ly, thing), be sore, (be, make) stiff(-en, (-necked))."},
"H7186":{"lemma":"קָשֶׁה","xlit":"qâsheh","pron":"kaw-sheh'","derivation":"from H7185 (קָשָׁה);","strongs_def":"severe (in various applications)","kjv_def":"churlish, cruel, grievous, hard((-hearted), thing), heavy, [phrase] impudent, obstinate, prevailed, rough(-ly), sore, sorrowful, stiff(necked), stubborn, [phrase] in trouble."},
"H7187":{"lemma":"קְשׁוֹט","xlit":"qᵉshôwṭ","pron":"kesh-ote'","derivation":"(Aramaic) or קשֹׁט; (Aramaic), corresponding to H7189 (קֹשֶׁט);","strongs_def":"fidelity","kjv_def":"truth."},
"H7188":{"lemma":"קָשַׁח","xlit":"qâshach","pron":"kaw-shakh'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) unfeeling","kjv_def":"harden."},
"H7189":{"lemma":"קֹשֶׁט","xlit":"qôsheṭ","pron":"ko'-shet","derivation":"or קֹשְׁטְ; from an unused root meaning to balance;","strongs_def":"equity (as evenly weighed), i.e. reality","kjv_def":"truth."},
"H7190":{"lemma":"קְשִׁי","xlit":"qᵉshîy","pron":"kesh-ee'","derivation":"from H7185 (קָשָׁה);","strongs_def":"obstinacy","kjv_def":"stubbornness."},
"H7191":{"lemma":"קִשְׁיוֹן","xlit":"Qishyôwn","pron":"kish-yone'","derivation":"from H7190 (קְשִׁי); hard ground;","strongs_def":"Kishjon, a place in Palestine","kjv_def":"Kishion, Keshon."},
"H7192":{"lemma":"קְשִׂיטָה","xlit":"qᵉsîyṭâh","pron":"kes-ee-taw'","derivation":"from an unused root (probably meaning to weigh out);","strongs_def":"an ingot (as definitely estimated and stamped for a coin)","kjv_def":"piece of money (silver)."},
"H7193":{"lemma":"קַשְׂקֶשֶׂת","xlit":"qasqeseth","pron":"kas-keh'-seth","derivation":"by reduplication from an unused root meaning to shale off as bark;","strongs_def":"a scale (of a fish); hence a coat of mail (as composed of or covered with jointed plates of metal)","kjv_def":"mail, scale."},
"H7194":{"lemma":"קָשַׁר","xlit":"qâshar","pron":"kaw-shar'","derivation":"a primitive root;","strongs_def":"to tie, physically (gird, confine, compact) or mentally (in love, league)","kjv_def":"bind (up), (make a) conspire(-acy, -ator), join together, knit, stronger, work (treason)."},
"H7195":{"lemma":"קֶשֶׁר","xlit":"qesher","pron":"keh'-sher","derivation":"from H7194 (קָשַׁר);","strongs_def":"an (unlawful) alliance","kjv_def":"confederacy, conspiracy, treason."},
"H7196":{"lemma":"קִשֻּׁר","xlit":"qishshur","pron":"kish-shoor'","derivation":"from H7194 (קָשַׁר);","strongs_def":"an (ornamental) girdle (for women)","kjv_def":"attire, headband."},
"H7197":{"lemma":"קָשַׁשׁ","xlit":"qâshash","pron":"kaw-shash'","derivation":"a primitive root; to become sapless through drought; used only as denominative from H7179 (קַשׁ);","strongs_def":"to forage for straw, stubble or wood; figuratively, to assemble","kjv_def":"gather (selves) (together)."},
"H7198":{"lemma":"קֶשֶׁת","xlit":"qesheth","pron":"keh'-sheth","derivation":"from H7185 (קָשָׁה) in the original sense (of H6983 (קוֹשׁ)) of bending;","strongs_def":"a bow, forshooting (hence, figuratively, strength) or the iris","kjv_def":"[idiom] arch(-er), [phrase] arrow, bow(-man, -shot)."},
"H7199":{"lemma":"קַשָּׁת","xlit":"qashshâth","pron":"kash-shawth'","derivation":"intensive (as denominative) from H7198 (קֶשֶׁת);","strongs_def":"a bowman","kjv_def":"[idiom] archer. r"},
"H7200":{"lemma":"רָאָה","xlit":"râʼâh","pron":"raw-aw'","derivation":"a primitive root;","strongs_def":"to see, literally or figuratively (in numerous applications, direct and implied, transitive, intransitive and causative)","kjv_def":"advise self, appear, approve, behold, [idiom] certainly, consider, discern, (make to) enjoy, have experience, gaze, take heed, [idiom] indeed, [idiom] joyfully, lo, look (on, one another, one on another, one upon another, out, up, upon), mark, meet, [idiom] be near, perceive, present, provide, regard, (have) respect, (fore-, cause to, let) see(-r, -m, one another), shew (self), [idiom] sight of others, (e-) spy, stare, [idiom] surely, [idiom] think, view, visions."},
"H7201":{"lemma":"רָאָה","xlit":"râʼâh","pron":"raw-aw'","derivation":"from H7200 (רָאָה);","strongs_def":"a bird of prey (probably the vulture, from its sharp sight)","kjv_def":"glede. Compare H1676 (דָּאָה)."},
"H7202":{"lemma":"רָאֶה","xlit":"râʼeh","pron":"raw-eh'","derivation":"from H7200 (רָאָה);","strongs_def":"seeing, i.e. experiencing","kjv_def":"see."},
"H7203":{"lemma":"רֹאֶה","xlit":"rôʼeh","pron":"ro-eh'","derivation":"active participle of H7200 (רָאָה);","strongs_def":"a seer (as often rendered); but also (abstractly) a vision","kjv_def":"vision."},
"H7204":{"lemma":"רֹאֵה","xlit":"Rôʼêh","pron":"ro-ay'","derivation":"for H7203 (רֹאֶה); prophet;","strongs_def":"Roeh, an Israelite","kjv_def":"Haroeh (including the article)."},
"H7205":{"lemma":"רְאוּבֵן","xlit":"Rᵉʼûwbên","pron":"reh-oo-bane'","derivation":"from the imperative of H7200 (רָאָה) and H1121 (בֵּן); see ye a son;","strongs_def":"Reuben, a son of Jacob","kjv_def":"Reuben."},
"H7206":{"lemma":"רְאוּבֵנִי","xlit":"Rᵉʼûwbênîy","pron":"reh-oob-ay-nee'","derivation":"patronymic from H7205 (רְאוּבֵן);","strongs_def":"a Reubenite or descendant of Reuben","kjv_def":"children of Reuben, Reubenites."},
"H7207":{"lemma":"רַאֲוָה","xlit":"raʼăvâh","pron":"rah-av-aw'","derivation":"from H7200 (רָאָה);","strongs_def":"sight, i.e. satisfaction","kjv_def":"behold."},
"H7208":{"lemma":"רְאוּמָה","xlit":"Rᵉʼûwmâh","pron":"reh-oo-maw'","derivation":"feminine passive participle of H7213 (רָאַם); raised;","strongs_def":"Reumah, a Syrian woman","kjv_def":"Reumah."},
"H7209":{"lemma":"רְאִי","xlit":"rᵉʼîy","pron":"reh-ee'","derivation":"from H7200 (רָאָה);","strongs_def":"a mirror (as seen)","kjv_def":"looking glass."},
"H7210":{"lemma":"רֳאִי","xlit":"rŏʼîy","pron":"ro-ee'","derivation":"from H7200 (רָאָה);","strongs_def":"sight, whether abstractly (vision) or concretely (a spectacle)","kjv_def":"gazingstock, look to, (that) see(-th)."},
"H7211":{"lemma":"רְאָיָה","xlit":"Rᵉʼâyâh","pron":"reh-aw-yaw'","derivation":"from H7200 (רָאָה) and H3050 (יָהּ); Jah has seen;","strongs_def":"Reajah, the name of three Israelites","kjv_def":"Reaia, Reaiah."},
"H7212":{"lemma":"רְאִית","xlit":"rᵉʼîyth","pron":"reh-eeth'","derivation":"from H7200 (רָאָה);","strongs_def":"sight","kjv_def":"beholding."},
"H7213":{"lemma":"רָאַם","xlit":"râʼam","pron":"raw-am'","derivation":"a primitive root;","strongs_def":"to rise","kjv_def":"be lifted up."},
"H7214":{"lemma":"רְאֵם","xlit":"rᵉʼêm","pron":"reh-ame'","derivation":"or רְאֵיםlemma רְאיֵם second vowel, corrected to רְאֵים; or רֵים; or רֵם; from H7213 (רָאַם);","strongs_def":"a wild bull (from its conspicuousness)","kjv_def":"unicorn."},
"H7215":{"lemma":"רָאמָה","xlit":"râʼmâh","pron":"raw-maw'","derivation":"from H7213 (רָאַם);","strongs_def":"something high in value, i.e. perhaps coral","kjv_def":"coral."},
"H7216":{"lemma":"רָאמוֹת","xlit":"Râʼmôwth","pron":"raw-moth'","derivation":"or רָאמֹתxlit Râmôth corrected to Râʼmôth; plural of H7215 (רָאמָה); heights;","strongs_def":"Ramoth, the name of two places in Palestine","kjv_def":"Ramoth."},
"H7217":{"lemma":"רֵאשׁ","xlit":"rêʼsh","pron":"raysh","derivation":"(Aramaic) corresponding to H7218 (רֹאשׁ);","strongs_def":"the head; figuratively, the sum","kjv_def":"chief, head, sum."},
"H7218":{"lemma":"רֹאשׁ","xlit":"rôʼsh","pron":"roshe","derivation":"from an unused root apparently meaning to shake;","strongs_def":"the head (as most easily shaken), whether literal or figurative (in many applications, of place, time, rank, itc.)","kjv_def":"band, beginning, captain, chapiter, chief(-est place, man, things), company, end, [idiom] every (man), excellent, first, forefront, (be-)head, height, (on) high(-est part, (priest)), [idiom] lead, [idiom] poor, principal, ruler, sum, top."},
"H7219":{"lemma":"רֹאשׁ","xlit":"rôʼsh","pron":"roshe","derivation":"or רוֹשׁ; (Deuteronomy 32:32), apparently the same as H7218 (רֹאשׁ);","strongs_def":"a poisonous plant, probably the poppy (from its conspicuous head); generally poison (even of serpents)","kjv_def":"gall, hemlock, poison, venom."},
"H7220":{"lemma":"רֹאשׁ","xlit":"Rôʼsh","pron":"roshe","derivation":"probably the same as H7218 (רֹאשׁ);","strongs_def":"Rosh, the name of an Israelite and of a foreign nation","kjv_def":"Rosh."},
"H7221":{"lemma":"רִאשָׁה","xlit":"riʼshâh","pron":"ree-shaw'","derivation":"from the same as H7218 (רֹאשׁ);","strongs_def":"a beginning","kjv_def":"beginning."},
"H7222":{"lemma":"רֹאשָׁה","xlit":"rôʼshâh","pron":"ro-shaw'","derivation":"feminine of H7218 (רֹאשׁ);","strongs_def":"the head","kjv_def":"head(-stone)."},
"H7223":{"lemma":"רִאשׁוֹן","xlit":"riʼshôwn","pron":"ree-shone'","derivation":"or רִאשֹׁן; from H7221 (רִאשָׁה);","strongs_def":"first, in place, time or rank (as adjective or noun)","kjv_def":"ancestor, (that were) before(-time), beginning, eldest, first, fore(-father) (-most), former (thing), of old time, past."},
"H7224":{"lemma":"רִאשֹׁנִי","xlit":"riʼshônîy","pron":"ree-sho-nee'","derivation":"from H7223 (רִאשׁוֹן);","strongs_def":"first","kjv_def":"first."},
"H7225":{"lemma":"רֵאשִׁית","xlit":"rêʼshîyth","pron":"ray-sheeth'","derivation":"from the same as H7218 (רֹאשׁ);","strongs_def":"the first, in place, time, order or rank (specifically, a firstfruit)","kjv_def":"beginning, chief(-est), first(-fruits, part, time), principal thing."},
"H7226":{"lemma":"רַאֲשֹׁת","xlit":"raʼăshôth","pron":"rah-ash-oth'","derivation":"from H7218 (רֹאשׁ);","strongs_def":"a pillow (being for the head)","kjv_def":"bolster."},
"H7227":{"lemma":"רַב","xlit":"rab","pron":"rab","derivation":"by contracted from H7231 (רָבַב);","strongs_def":"abundant (in quantity, size, age, number, rank, quality)","kjv_def":"(in) abound(-undance, -ant, -antly), captain, elder, enough, exceedingly, full, great(-ly, man, one), increase, long (enough, (time)), (do, have) many(-ifold, things, a time), (ship-)master, mighty, more, (too, very) much, multiply(-tude), officer, often(-times), plenteous, populous, prince, process (of time), suffice(-lent)."},
"H7228":{"lemma":"רַב","xlit":"rab","pron":"rab","derivation":"by contraction from H7232 (רָבַב); (or perhaps the same as H7227 (רַב))","strongs_def":"an archer","kjv_def":"archer."},
"H7229":{"lemma":"רַב","xlit":"rab","pron":"rab","derivation":"(Aramaic) corresponding to H7227 (רַב)","strongs_def":"{abundant}","kjv_def":"captain, chief, great, lord, master, stout."},
"H7230":{"lemma":"רֹב","xlit":"rôb","pron":"robe","derivation":"from H7231 (רָבַב);","strongs_def":"abundance (in any respect)","kjv_def":"abundance(-antly), all, [idiom] common (sort), excellent, great(-ly, -ness, number), huge, be increased, long, many, more in number, most, much, multitude, plenty(-ifully), [idiom] very (age)."},
"H7231":{"lemma":"רָבַב","xlit":"râbab","pron":"raw-bab'","derivation":"a primitive root; (compare H7241 (רָבִיב)); also as denominative from H7233 (רְבָבָה);","strongs_def":"properly, to cast together , i.e. increase, especially in number; to multiply by the myriad","kjv_def":"increase, be many(-ifold), be more, multiply, ten thousands."},
"H7232":{"lemma":"רָבַב","xlit":"râbab","pron":"raw-bab'","derivation":"a primitive root (rather identical with H7231 (רָבַב) through the idea of projection);","strongs_def":"to shoot an arrow","kjv_def":"shoot."},
"H7233":{"lemma":"רְבָבָה","xlit":"rᵉbâbâh","pron":"reb-aw-baw'","derivation":"from H7231 (רָבַב);","strongs_def":"abundance (in number), i.e. (specifically) a myriad (whether definite or indefinite)","kjv_def":"many, million, [idiom] multiply, ten thousand."},
"H7234":{"lemma":"רָבַד","xlit":"râbad","pron":"raw-bad'","derivation":"a primitive root;","strongs_def":"to spread","kjv_def":"deck."},
"H7235":{"lemma":"רָבָה","xlit":"râbâh","pron":"raw-baw'","derivation":"a primitive root;","strongs_def":"to increase (in whatever respect)","kjv_def":"(bring in) abundance ([idiom] -antly), [phrase] archer (by mistake for H7232 (רָבַב)), be in authority, bring up, [idiom] continue, enlarge, excel, exceeding(-ly), be full of, (be, make) great(-er, -ly, [idiom] -ness), grow up, heap, increase, be long, (be, give, have, make, use) many (a time), (any, be, give, give the, have) more (in number), (ask, be, be so, gather, over, take, yield) much (greater, more), (make to) multiply, nourish, plenty(-eous), [idiom] process (of time), sore, store, thoroughly, very."},
"H7236":{"lemma":"רְבָה","xlit":"rᵉbâh","pron":"reb-aw'","derivation":"(Aramaic) corresponding to H7235 (רָבָה)","strongs_def":"{to increase (in whatever respect)}","kjv_def":"make a great man, grow."},
"H7237":{"lemma":"רַבָּה","xlit":"Rabbâh","pron":"rab-baw'","derivation":"feminine of H7227 (רַב); great;","strongs_def":"Rabbah, the name of two places in Palestine, East and West","kjv_def":"Rabbah, Rabbath."},
"H7238":{"lemma":"רְבוּ","xlit":"rᵉbûw","pron":"reb-oo'","derivation":"(Aramaic) from a root corresponding to H7235 (רָבָה);","strongs_def":"increase (of dignity)","kjv_def":"greatness, majesty."},
"H7239":{"lemma":"רִבּוֹ","xlit":"ribbôw","pron":"rib-bo'","derivation":"from H7231 (רָבַב); or רִבּוֹא from H7231 (רָבַב);","strongs_def":"a myriad, i.e. indefinitely, large number","kjv_def":"great things, ten ((eight) -een, (for) -ty, [phrase] sixscore, [phrase] threescore, [idiom] twenty, (twen) -ty) thousand."},
"H7240":{"lemma":"רִבּוֹ","xlit":"ribbôw","pron":"rib-bo'","derivation":"(Aramaic) corresponding to H7239 (רִבּוֹ)","strongs_def":"{a myriad, i.e. indefinitely, large number}","kjv_def":"[idiom] ten thousand times ten thousand."},
"H7241":{"lemma":"רָבִיב","xlit":"râbîyb","pron":"raw-beeb'","derivation":"from H7231 (רָבַב);","strongs_def":"a rain (as an accumulation of drops)","kjv_def":"shower."},
"H7242":{"lemma":"רָבִיד","xlit":"râbîyd","pron":"raw-beed'","derivation":"from H7234 (רָבַד);","strongs_def":"a collar (as spread around the neck)","kjv_def":"chain."},
"H7243":{"lemma":"רְבִיעִי","xlit":"rᵉbîyʻîy","pron":"reb-ee-ee'","derivation":"or רְבִעִי; from H7251 (רָבַע);","strongs_def":"fourth; also (fractionally) a fourth","kjv_def":"foursquare, fourth (part)."},
"H7244":{"lemma":"רְבִיעַי","xlit":"rᵉbîyʻay","pron":"reb-ee-ah'-ee","derivation":"(Aramaic) corresponding to H7243 (רְבִיעִי)","strongs_def":"{fourth; also (fractionally) a fourth}","kjv_def":"fourth."},
"H7245":{"lemma":"רַבִּית","xlit":"Rabbîyth","pron":"rab-beeth'","derivation":"from H7231 (רָבַב); multitude;","strongs_def":"Rabbith, a place in Palestine","kjv_def":"Rabbith."},
"H7246":{"lemma":"רָבַךְ","xlit":"râbak","pron":"raw-bak'","derivation":"a primitive root;","strongs_def":"to soak (bread in oil)","kjv_def":"baken, (that which is) fried."},
"H7247":{"lemma":"רִבְלָה","xlit":"Riblâh","pron":"rib-law'","derivation":"from an unused root meaning to be fruitful; fertile;","strongs_def":"Riblah, a place in Syria","kjv_def":"Riblah."},
"H7248":{"lemma":"רַב־מָג","xlit":"Rab-Mâg","pron":"rab-mawg'","derivation":"from H7227 (רַב) and a foreign word for a Magian; chief Magian;","strongs_def":"Rab-Mag, a Babylonian official","kjv_def":"Rab-mag."},
"H7249":{"lemma":"רַב־סָרִיס","xlit":"Rab-Çârîyç","pron":"rab-saw-reece'","derivation":"from H7227 (רַב) and a foreign word for a eunuch; chief chamberlain;","strongs_def":"Rab-Saris, a Babylonian official","kjv_def":"Rab-saris."},
"H7250":{"lemma":"רָבַע","xlit":"râbaʻ","pron":"raw-bah'","derivation":"a primitive root;","strongs_def":"to squat or lie out flat, i.e. (specifically) in copulation","kjv_def":"let gender, lie down."},
"H7251":{"lemma":"רָבַע","xlit":"râbaʻ","pron":"raw-bah'","derivation":"a primitive root (rather identical with H7250 (רָבַע) through the idea of sprawling 'at all fours' (or possibly the reverse is the order of deriv.); compare H702 (אַרְבַּע)); properly, to be four (sided); used only as denominative of H7253 (רֶבַע);","strongs_def":"to be quadrate","kjv_def":"(four-) square(-d)."},
"H7252":{"lemma":"רֶבַע","xlit":"rebaʻ","pron":"reh'-bah","derivation":"from H7250 (רָבַע);","strongs_def":"prostration (for sleep)","kjv_def":"lying down."},
"H7253":{"lemma":"רֶבַע","xlit":"rebaʻ","pron":"reh'-bah","derivation":"from H7251 (רָבַע);","strongs_def":"a fourth (part or side)","kjv_def":"fourth part, side, square."},
"H7254":{"lemma":"רֶבַע","xlit":"Rebaʻ","pron":"reh'-bah","derivation":"the same as H7253 (רֶבַע);","strongs_def":"Reba, a Midianite","kjv_def":"Reba."},
"H7255":{"lemma":"רֹבַע","xlit":"rôbaʻ","pron":"ro'-bah","derivation":"from H7251 (רָבַע);","strongs_def":"a quarter","kjv_def":"fourth participle"},
"H7256":{"lemma":"רִבֵּעַ","xlit":"ribbêaʻ","pron":"rib-bay'-ah","derivation":"from H7251 (רָבַע);","strongs_def":"a descendant of the fourth generation, i.e. great great grandchild","kjv_def":"fourth."},
"H7257":{"lemma":"רָבַץ","xlit":"râbats","pron":"raw-bats'","derivation":"a primitive root;","strongs_def":"to crouch (on all four legs folded, like a recumbent animal); be implication, to recline, repose, brood, lurk, imbed","kjv_def":"crouch (down), fall down, make a fold, lay, (cause to, make to) lie (down), make to rest, sit."},
"H7258":{"lemma":"רֵבֶץ","xlit":"rêbets","pron":"reh'-bets","derivation":"from H7257 (רָבַץ);","strongs_def":"a couch or place of repose","kjv_def":"where each lay, lie down in, resting place."},
"H7259":{"lemma":"רִבְקָה","xlit":"Ribqâh","pron":"rib-kaw'","derivation":"from an unused root probably meaning to clog by tying up the fetlock; fettering (by beauty);","strongs_def":"Ribkah, the wife of Isaac","kjv_def":"Rebekah."},
"H7260":{"lemma":"רַבְרַב","xlit":"rabrab","pron":"rab-rab'","derivation":"(Aramaic) from H7229 (רַב);","strongs_def":"huge (in size); domineering (in character)","kjv_def":"(very) great (things)."},
"H7261":{"lemma":"רַבְרְבָן","xlit":"rabrᵉbân","pron":"rab-reb-awn'","derivation":"(Aramaic) from H7260 (רַבְרַב);","strongs_def":"a magnate","kjv_def":"lord, prince."},
"H7262":{"lemma":"רַבְשָׁקֵה","xlit":"Rabshâqêh","pron":"rab-shaw-kay'","derivation":"from H7227 (רַב) and H8248 (שָׁקָה); chief butler;","strongs_def":"Rabshakeh, a Babylonian official","kjv_def":"Rabshakeh."},
"H7263":{"lemma":"רֶגֶב","xlit":"regeb","pron":"reh'-gheb","derivation":"from an unused root meaning to pile together;","strongs_def":"a lump of clay","kjv_def":"clod."},
"H7264":{"lemma":"רָגַז","xlit":"râgaz","pron":"raw-gaz'","derivation":"a primitive root;","strongs_def":"to quiver (with any violent emotion, especially anger or fear)","kjv_def":"be afraid, stand in awe, disquiet, fall out, fret, move, provoke, quake, rage, shake, tremble, trouble, be wroth."},
"H7265":{"lemma":"רְגַז","xlit":"rᵉgaz","pron":"reg-az'","derivation":"(Aramaic) corresponding to H7264 (רָגַז)","strongs_def":"{to quiver (with any violent emotion, especially anger or fear)}","kjv_def":"provoke unto wrath."},
"H7266":{"lemma":"רְגַז","xlit":"rᵉgaz","pron":"reg-az'","derivation":"(Aramaic) from H7265 (רְגַז);","strongs_def":"violent anger","kjv_def":"rage."},
"H7267":{"lemma":"רֹגֶז","xlit":"rôgez","pron":"ro'-ghez","derivation":"from H7264 (רָגַז);","strongs_def":"commotion, restlessness (of a horse), crash (of thunder), disquiet, anger","kjv_def":"fear, noise, rage, trouble(-ing), wrath."},
"H7268":{"lemma":"רַגָּז","xlit":"raggâz","pron":"rag-gawz'","derivation":"intensive from H7264 (רָגַז);","strongs_def":"timid","kjv_def":"trembling."},
"H7269":{"lemma":"רׇגְזָה","xlit":"rogzâh","pron":"rog-zaw'","derivation":"feminine of H7267 (רֹגֶז);","strongs_def":"trepidation","kjv_def":"trembling."},
"H7270":{"lemma":"רָגַל","xlit":"râgal","pron":"raw-gal'","derivation":"a primitive root;  also as denominative from H7272 (רֶגֶל)","strongs_def":"to walk along; but only in specifically, applications, to reconnoiter, to be atale-bearer (i.e. slander); to lead about","kjv_def":"backbite, search, slander, (e-) spy (out), teach to go, view."},
"H7271":{"lemma":"רְגַל","xlit":"rᵉgal","pron":"reg-al'","derivation":"(Aramaic) corresponding to H7272 (רֶגֶל)","strongs_def":"{a foot, a step; by euphemistically the pudenda}","kjv_def":"foot."},
"H7272":{"lemma":"רֶגֶל","xlit":"regel","pron":"reh'-gel","derivation":"from H7270 (רָגַל);","strongs_def":"a foot (as used in walking); by implication, a step; by euphemistically the pudenda","kjv_def":"[idiom] be able to endure, [idiom] according as, [idiom] after, [idiom] coming, [idiom] follow, (broken-)foot(-ed, -stool), [idiom] great toe, [idiom] haunt, [idiom] journey, leg, [phrase] piss, [phrase] possession, time."},
"H7273":{"lemma":"רַגְלִי","xlit":"raglîy","pron":"rag-lee'","derivation":"from H7272 (רֶגֶל);","strongs_def":"a footman (soldier)","kjv_def":"(on) foot(-man)."},
"H7274":{"lemma":"רֹגְלִים","xlit":"Rôgᵉlîym","pron":"ro-gel-eem'","derivation":"plural of active participle of H7270 (רָגַל); fullers (as tramping the cloth in washing);","strongs_def":"Rogelim, a place East of the Jordan","kjv_def":"Rogelim."},
"H7275":{"lemma":"רָגַם","xlit":"râgam","pron":"raw-gam'","derivation":"a primitive root (compare H7263 (רֶגֶב), H7321 (רוּעַ), H7551 (רָקַם));","strongs_def":"to cast together (stones), i.e. to lapidate","kjv_def":"[idiom] certainly, stone."},
"H7276":{"lemma":"רֶגֶם","xlit":"Regem","pron":"reh'-gem","derivation":"from H7275 (רָגַם); stoneheap;","strongs_def":"Regem, an Israelite","kjv_def":"Regem."},
"H7277":{"lemma":"רִגְמָה","xlit":"rigmâh","pron":"rig-maw'","derivation":"feminine of the same as H7276 (רֶגֶם);","strongs_def":"a pile (of stones), i.e. (figuratively) a throng","kjv_def":"council."},
"H7278":{"lemma":"רֶגֶם מֶלֶךְ","xlit":"Regem Melek","pron":"reh'-gem meh'-lek","derivation":"from H7276 (רֶגֶם) and H4428 (מֶלֶךְ); king's heap;","strongs_def":"Regem-Melek, an Israelite","kjv_def":"Regem-melech."},
"H7279":{"lemma":"רָגַן","xlit":"râgan","pron":"raw-gan'","derivation":"a primitive root;","strongs_def":"to grumble, i.e. rebel","kjv_def":"murmur."},
"H7280":{"lemma":"רָגַע","xlit":"râgaʻ","pron":"raw-gah'","derivation":"a primitive root;","strongs_def":"properly, to toss violently and suddenly (the sea with waves, the skin with boils); figuratively (in a favorable manner) to settle, i.e. quiet; specifically, to wink (from the motion of the eye-lids)","kjv_def":"break, divide, find ease, be a moment, (cause, give, make to) rest, make suddenly."},
"H7281":{"lemma":"רֶגַע","xlit":"regaʻ","pron":"reh'-gah","derivation":"from H7280 (רָגַע).","strongs_def":"a wink (of the eyes), i.e. a very short space of time","kjv_def":"instant, moment, space, suddenly."},
"H7282":{"lemma":"רָגֵעַ","xlit":"râgêaʻ","pron":"raw-gay'-ah","derivation":"from H7280 (רָגַע);","strongs_def":"restful, i.e. peaceable","kjv_def":"that are quiet."},
"H7283":{"lemma":"רָגַשׁ","xlit":"râgash","pron":"raw-gash'","derivation":"a primitive root;","strongs_def":"to be tumultuous","kjv_def":"rage."},
"H7284":{"lemma":"רְגַשׁ","xlit":"rᵉgash","pron":"reg-ash'","derivation":"(Aramaic) corresponding to H7283 (רָגַשׁ);","strongs_def":"to gather tumultuously","kjv_def":"assemble (together)."},
"H7285":{"lemma":"רֶגֶשׁ","xlit":"regesh","pron":"reh'-ghesh","derivation":"or (feminine) רִגְשָׁה; from H7283 (רָגַשׁ);","strongs_def":"a tumultuous crowd","kjv_def":"company, insurrection."},
"H7286":{"lemma":"רָדַד","xlit":"râdad","pron":"raw-dad'","derivation":"a primitive root;","strongs_def":"to tread in pieces, i.e. (figuratively) to conquer, or (specifically) to overlay","kjv_def":"spend, spread, subdue."},
"H7287":{"lemma":"רָדָה","xlit":"râdâh","pron":"raw-daw'","derivation":"a primitive root;","strongs_def":"to tread down, i.e. subjugate; specifically, to crumble off","kjv_def":"(come to, make to) have dominion, prevail against, reign, (bear, make to) rule,(-r, over), take."},
"H7288":{"lemma":"רַדַּי","xlit":"Radday","pron":"rad-dah'-ee","derivation":"intensive from H7287 (רָדָה); domineering;","strongs_def":"Raddai, an Israelite","kjv_def":"Raddai."},
"H7289":{"lemma":"רָדִיד","xlit":"râdîyd","pron":"raw-deed'","derivation":"from H7286 (רָדַד) in the sense of spreading;","strongs_def":"a veil (as expanded)","kjv_def":"vail, veil."},
"H7290":{"lemma":"רָדַם","xlit":"râdam","pron":"raw-dam'","derivation":"a primitive root;","strongs_def":"to stun, i.e. stupefy (with sleep or death)","kjv_def":"(be fast a-, be in a deep, cast into a dead, that) sleep(-er, -eth)."},
"H7291":{"lemma":"רָדַף","xlit":"râdaph","pron":"raw-daf'","derivation":"a primitive root;","strongs_def":"to run after (usually with hostile intent; figuratively (of time) gone by)","kjv_def":"chase, put to flight, follow (after, on), hunt, (be under) persecute(-ion, -or), pursue(-r)."},
"H7292":{"lemma":"רָהַב","xlit":"râhab","pron":"raw-hab'","derivation":"a primitive root;","strongs_def":"to urge severely, i.e. (figuratively) importune, embolden, capture, act insolently","kjv_def":"overcome, behave self proudly, make sure, strengthen."},
"H7293":{"lemma":"רַהַב","xlit":"rahab","pron":"rah'-hab","derivation":"from H7292 (רָהַב),","strongs_def":"bluster(-er)","kjv_def":"proud, strength."},
"H7294":{"lemma":"רַהַב","xlit":"Rahab","pron":"rah'-hab","derivation":"the same as H7293 (רַהַב);","strongs_def":"Rahab (i.e. boaster), an epithet of Egypt","kjv_def":"Rahab."},
"H7295":{"lemma":"רָהָב","xlit":"râhâb","pron":"raw-hawb'","derivation":"from H7292 (רָהַב);","strongs_def":"insolent","kjv_def":"proud."},
"H7296":{"lemma":"רֹהָב","xlit":"rôhâb","pron":"ro'-hab","derivation":"from H7292 (רָהַב);","strongs_def":"pride","kjv_def":"strength."},
"H7297":{"lemma":"רָהָה","xlit":"râhâh","pron":"raw-haw'","derivation":"a primitive root;","strongs_def":"to fear","kjv_def":"be afraid."},
"H7298":{"lemma":"רַהַט","xlit":"rahaṭ","pron":"rah'-hat","derivation":"from an unused root apparently meaning to hollow out;","strongs_def":"a channel or watering-box; by resemblance a ringlet of hair (as forming parallel lines)","kjv_def":"gallery, gutter, trough."},
"H7299":{"lemma":"רֵו","xlit":"rêv","pron":"rave","derivation":"(Aramaic) from a root corresponding to H7200 (רָאָה);","strongs_def":"aspect","kjv_def":"form."},
"H7300":{"lemma":"רוּד","xlit":"rûwd","pron":"rood","derivation":"a primitive root;","strongs_def":"to tramp about, i.e. ramble (free or disconsolate)","kjv_def":"have the dominion, be lord, mourn, rule."},
"H7301":{"lemma":"רָוָה","xlit":"râvâh","pron":"raw-vaw'","derivation":"a primitive root;","strongs_def":"to slake the thirst (occasionally of other appetites)","kjv_def":"bathe, make drunk, (take the) fill, satiate, (abundantly) satisfy, soak, water (abundantly)."},
"H7302":{"lemma":"רָוֶה","xlit":"râveh","pron":"raw-veh'","derivation":"from H7301 (רָוָה);","strongs_def":"sated (with drink)","kjv_def":"drunkenness, watered."},
"H7303":{"lemma":"רוֹהֲגָה","xlit":"Rôwhăgâh","pron":"ro-hag-aw'","derivation":"from an unused root probably meaning to cry out; outcry;","strongs_def":"Rohagah, an Israelite","kjv_def":"Rohgah."},
"H7304":{"lemma":"רָוַח","xlit":"râvach","pron":"raw-vakh'","derivation":"a primitive root (identical with H7306 (רוּחַ));","strongs_def":"properly, to breathe freely, i.e. revive; by implication, to have ample room","kjv_def":"be refreshed, large."},
"H7305":{"lemma":"רֶוַח","xlit":"revach","pron":"reh'-vakh","derivation":"from H7304 (רָוַח);","strongs_def":"room, literally (an interval) or figuratively (deliverance)","kjv_def":"enlargement, space."},
"H7306":{"lemma":"רוּחַ","xlit":"rûwach","pron":"roo'-akh","derivation":"a primitive root;","strongs_def":"properly, to blow, i.e. breathe; only (literally) to smell or (by implication, perceive (figuratively, to anticipate, enjoy)","kjv_def":"accept, smell, [idiom] touch, make of quick understanding."},
"H7307":{"lemma":"רוּחַ","xlit":"rûwach","pron":"roo'-akh","derivation":"from H7306 (רוּחַ);","strongs_def":"wind; by resemblance breath, i.e. a sensible (or even violent) exhalation; figuratively, life, anger, unsubstantiality; by extension, a region of the sky; by resemblance spirit, but only of a rational being (including its expression and functions)","kjv_def":"air, anger, blast, breath, [idiom] cool, courage, mind, [idiom] quarter, [idiom] side, spirit(-ual), tempest, [idiom] vain, (whirl-) wind(-y)."},
"H7308":{"lemma":"רוּחַ","xlit":"rûwach","pron":"roo'-akh","derivation":"(Aramaic) corresponding to H7307 (רוּחַ)","strongs_def":"{wind; by resemblance breath, i.e. a sensible (or even violent) exhalation; figuratively, life, anger, unsubstantiality; by extension, a region of the sky; by resemblance spirit, but only of a rational being (including its expression and functions)}","kjv_def":"mind, spirit, wind."},
"H7309":{"lemma":"רְוָחָה","xlit":"rᵉvâchâh","pron":"rev-aw-khaw'","derivation":"feminine of H7305 (רֶוַח);","strongs_def":"relief","kjv_def":"breathing, respite."},
"H7310":{"lemma":"רְוָיָה","xlit":"rᵉvâyâh","pron":"rev-aw-yaw'","derivation":"from H7301 (רָוָה);","strongs_def":"satisfaction","kjv_def":"runneth over, wealthy."},
"H7311":{"lemma":"רוּם","xlit":"rûwm","pron":"room","derivation":"a primitive root;","strongs_def":"to be high actively, to rise or raise (in various applications, literally or figuratively)","kjv_def":"bring up, exalt (self), extol, give, go up, haughty, heave (up), (be, lift up on, make on, set up on, too) high(-er, one), hold up, levy, lift(-er) up, (be) lofty, ([idiom] a-) loud, mount up, offer (up), [phrase] presumptuously, (be) promote(-ion), proud, set up, tall(-er), take (away, off, up), breed worms."},
"H7312":{"lemma":"רוּם","xlit":"rûwm","pron":"room","derivation":"or רֻם; from H7311 (רוּם);","strongs_def":"(literally) elevation or (figuratively) elation","kjv_def":"haughtiness, height, [idiom] high."},
"H7313":{"lemma":"רוּם","xlit":"rûwm","pron":"room","derivation":"(Aramaic) corresponding to H7311 (רוּם);","strongs_def":"{to be high actively, to rise or raise (in various applications, literally or figuratively)}","kjv_def":"extol, lift up (self), set up."},
"H7314":{"lemma":"רוּם","xlit":"rûwm","pron":"room","derivation":"(Aramaic) from H7313 (רוּם);","strongs_def":"(literally) altitude","kjv_def":"height."},
"H7315":{"lemma":"רוֹם","xlit":"rôwm","pron":"rome","derivation":"from H7311 (רוּם);","strongs_def":"elevation, i.e. (adverbially) aloft","kjv_def":"on high."},
"H7316":{"lemma":"רוּמָה","xlit":"Rûwmâh","pron":"roo-maw'","derivation":"from H7311 (רוּם); height;","strongs_def":"Rumah, a place in Palestine","kjv_def":"Rumah."},
"H7317":{"lemma":"רוֹמָה","xlit":"rôwmâh","pron":"ro-maw'","derivation":"feminine of H7315 (רוֹם);","strongs_def":"elation, i.e. (adverbially) proudly","kjv_def":"haughtily."},
"H7318":{"lemma":"רוֹמָם","xlit":"rôwmâm","pron":"ro-mawm'","derivation":"from H7426 (רָמַם);","strongs_def":"exaltation, i.e. (figuratively and specifically) praise","kjv_def":"be extolled."},
"H7319":{"lemma":"רוֹמְמָה","xlit":"rôwmᵉmâh","pron":"ro-mem-aw'","derivation":"feminine active participle of H7426 (רָמַם);","strongs_def":"exaltation, i.e. praise","kjv_def":"high."},
"H7320":{"lemma":"רוֹמַמְתִּי עֶזֶר","xlit":"Rôwmamtîy ʻEzer","pron":"ro-mam'-tee eh'-zer","derivation":"(or Romam from H7311 (רוּם) and H5828 (עֵזֶר); I have raised up a help;","strongs_def":"Romamti-Ezer, an Israelite","kjv_def":"Romamti-ezer."},
"H7321":{"lemma":"רוּעַ","xlit":"rûwaʻ","pron":"roo-ah'","derivation":"a primitive root;","strongs_def":"to mar (especially by breaking); figuratively, to split the ears (with sound), i.e. shout (for alarm or joy)","kjv_def":"blow an alarm, cry (alarm, aloud, out), destroy, make a joyful noise, smart, shout (for joy), sound an alarm, triumph."},
"H7322":{"lemma":"רוּף","xlit":"rûwph","pron":"roof","derivation":"a primitive root;","strongs_def":"properly, to triturate (in a mortar), i.e. (figuratively) to agitate (by concussion)","kjv_def":"tremble."},
"H7323":{"lemma":"רוּץ","xlit":"rûwts","pron":"roots","derivation":"a primitive root;","strongs_def":"to run (for whatever reason, especially to rush)","kjv_def":"break down, divide speedily, footman, guard, bring hastily, (make) run (away, through), post."},
"H7324":{"lemma":"רוּק","xlit":"rûwq","pron":"rook","derivation":"a primitive root;","strongs_def":"to pour out (literally or figuratively), i.e. empty","kjv_def":"[idiom] arm, cast out, draw (out), (make) empty, pour forth (out)."},
"H7325":{"lemma":"רוּר","xlit":"rûwr","pron":"roor","derivation":"a primitive root;","strongs_def":"to slaver (with spittle), i.e. (by analogy) to emit afluid (ulcerous or natural)","kjv_def":"run."},
"H7326":{"lemma":"רוּשׁ","xlit":"rûwsh","pron":"roosh","derivation":"a primitive root;","strongs_def":"to be destitute","kjv_def":"lack, needy, (make self) poor (man)."},
"H7327":{"lemma":"רוּת","xlit":"Rûwth","pron":"rooth","derivation":"probably for H7468 (רְעוּת); friend;","strongs_def":"Ruth, a Moabitess","kjv_def":"Ruth."},
"H7328":{"lemma":"רָז","xlit":"râz","pron":"rawz","derivation":"(Aramaic) from an unused root probably meaning","strongs_def":"to attenuate, i.e. (figuratively) hide; a mystery","kjv_def":"secret."},
"H7329":{"lemma":"רָזָה","xlit":"râzâh","pron":"raw-zaw'","derivation":"a primitive root;","strongs_def":"to emaciate, i.e. make (become) thin (literally or figuratively)","kjv_def":"famish, wax lean."},
"H7330":{"lemma":"רָזֶה","xlit":"râzeh","pron":"raw-zeh'","derivation":"from H7329 (רָזָה);","strongs_def":"thin","kjv_def":"lean."},
"H7331":{"lemma":"רְזוֹן","xlit":"Rᵉzôwn","pron":"rez-one'","derivation":"from H7336 (רָזַן); prince;","strongs_def":"Rezon, a Syrian","kjv_def":"Rezon."},
"H7332":{"lemma":"רָזוֹן","xlit":"râzôwn","pron":"raw-zone'","derivation":"from H7329 (רָזָה);","strongs_def":"thinness","kjv_def":"leanness, [idiom] scant."},
"H7333":{"lemma":"רָזוֹן","xlit":"râzôwn","pron":"raw-zone'","derivation":"from H7336 (רָזַן);","strongs_def":"a dignitary","kjv_def":"prince."},
"H7334":{"lemma":"רָזִי","xlit":"râzîy","pron":"raw-zee'","derivation":"from H7329 (רָזָה);","strongs_def":"thinness","kjv_def":"leanness."},
"H7335":{"lemma":"רָזַם","xlit":"râzam","pron":"raw-zam'","derivation":"a primitive root;","strongs_def":"to twinkle the eye (in mockery)","kjv_def":"wink."},
"H7336":{"lemma":"רָזַן","xlit":"râzan","pron":"raw-zan'","derivation":"a primitive root;","strongs_def":"probably to be heavy, i.e. (figuratively) honorable","kjv_def":"prince, ruler."},
"H7337":{"lemma":"רָחַב","xlit":"râchab","pron":"raw-khab'","derivation":"a primitive root;","strongs_def":"to broaden (intransitive or transitive, literal or figurative)","kjv_def":"be an en-(make) large(-ing), make room, make (open) wide."},
"H7338":{"lemma":"רַחַב","xlit":"rachab","pron":"rakh'-ab","derivation":"from H7337 (רָחַב);","strongs_def":"a width","kjv_def":"breadth, broad place."},
"H7339":{"lemma":"רְחֹב","xlit":"rᵉchôb","pron":"rekh-obe'","derivation":"or רְחוֹב; from H7337 (רָחַב);","strongs_def":"a width, i.e. (concretely) avenue or area","kjv_def":"broad place (way), street. See also H1050 (בֵּית רְחוֹב)."},
"H7340":{"lemma":"רְחֹב","xlit":"Rᵉchôb","pron":"rekh-obe'","derivation":"or רְחוֹב; the same as H7339 (רְחֹב);","strongs_def":"Rechob, the name of a place in Syria, also of a Syrian and an Israelite","kjv_def":"Rehob."},
"H7341":{"lemma":"רֹחַב","xlit":"rôchab","pron":"ro'-khab","derivation":"from H7337 (רָחַב);","strongs_def":"width (literally or figuratively)","kjv_def":"breadth, broad, largeness, thickness, wideness."},
"H7342":{"lemma":"רָחָב","xlit":"râchâb","pron":"raw-khawb'","derivation":"from H7337 (רָחַב);","strongs_def":"roomy, in any (or every) direction, literally or figuratively","kjv_def":"broad, large, at liberty, proud, wide."},
"H7343":{"lemma":"רָחָב","xlit":"Râchâb","pron":"raw-khawb'","derivation":"the same as H7342 (רָחָב); proud;","strongs_def":"Rachab, a Canaanitess","kjv_def":"Rahab."},
"H7344":{"lemma":"רְחֹבוֹת","xlit":"Rᵉchôbôwth","pron":"rekh-o-both'","derivation":"or רְחֹבֹת; plural of H7339 (רְחֹב); streets;","strongs_def":"Rechoboth, a place in Assyria and one in Palestine","kjv_def":"Rehoboth."},
"H7345":{"lemma":"רְחַבְיָה","xlit":"Rᵉchabyâh","pron":"rekh-ab-yaw'","derivation":"or רְחַבְיָהוּ; from H7337 (רָחַב) and H3050 (יָהּ); Jah has enlarged;","strongs_def":"Rechabjah, an Israelite","kjv_def":"Rehabiah."},
"H7346":{"lemma":"רְחַבְעָם","xlit":"Rᵉchabʻâm","pron":"rekh-ab-awm'","derivation":"from H7337 (רָחַב) and H5971 (עַם); a people has enlarged;","strongs_def":"Rechabam, an Israelite king","kjv_def":"Rehoboam."},
"H7347":{"lemma":"רֵחֶה","xlit":"rêcheh","pron":"ray-kheh'","derivation":"from an unused root meaning to pulverize;","strongs_def":"a mill-stone","kjv_def":"mill (stone)."},
"H7348":{"lemma":"רְחוּם","xlit":"Rᵉchûwm","pron":"rekh-oom'","derivation":"a form of H7349 (רַחוּם);","strongs_def":"Rechum, the name of a Persian and of three Israelites","kjv_def":"Rehum."},
"H7349":{"lemma":"רַחוּם","xlit":"rachûwm","pron":"rakh-oom'","derivation":"from H7355 (רָחַם);","strongs_def":"compassionate","kjv_def":"full of compassion, merciful."},
"H7350":{"lemma":"רָחוֹק","xlit":"râchôwq","pron":"raw-khoke'","derivation":"or רָחֹק; from H7368 (רָחַק);","strongs_def":"remote, literally or figuratively, of place or time; specifically, precious; often used adverbially (with preposition)","kjv_def":"(a-) far (abroad, off), long ago, of old, space, great while to come."},
"H7351":{"lemma":"רְחִיט","xlit":"rᵉchîyṭ","pron":"rekh-eet'","derivation":"from the same as H7298 (רַהַט);","strongs_def":"a panel (as resembling a trough)","kjv_def":"rafter."},
"H7352":{"lemma":"רַחִיק","xlit":"rachîyq","pron":"rakh-eek'","derivation":"(Aramaic) corresponding to H7350 (רָחוֹק)","strongs_def":"{remote, literally or figuratively, of place or time; specifically, precious; often used adverbially (with preposition)}","kjv_def":"far."},
"H7353":{"lemma":"רָחֵל","xlit":"râchêl","pron":"raw-kale'","derivation":"from an unused root meaning to journey;","strongs_def":"a ewe (the females being the predominant element of a flock) (as a good traveller)","kjv_def":"ewe, sheep."},
"H7354":{"lemma":"רָחֵל","xlit":"Râchêl","pron":"raw-khale'","derivation":"the same as H7353 (רָחֵל);","strongs_def":"Rachel, a wife of Jacob","kjv_def":"Rachel."},
"H7355":{"lemma":"רָחַם","xlit":"râcham","pron":"raw-kham'","derivation":"a primitive root;","strongs_def":"to fondle; by implication, to love, especially to compassionate","kjv_def":"have compassion (on, upon), love, (find, have, obtain, shew) mercy(-iful, on, upon), (have) pity, Ruhamah, [idiom] surely."},
"H7356":{"lemma":"רַחַם","xlit":"racham","pron":"rakh'-am","derivation":"from H7355 (רָחַם);","strongs_def":"compassion (in the plural); by extension, the womb (as cherishing the fetus); by implication, a maiden","kjv_def":"bowels, compassion, damsel, tender love, (great, tender) mercy, pity, womb."},
"H7357":{"lemma":"רַחַם","xlit":"Racham","pron":"rakh'-am","derivation":"the same as H7356 (רַחַם); pity;","strongs_def":"Racham, an Israelite","kjv_def":"Raham."},
"H7358":{"lemma":"רֶחֶם","xlit":"rechem","pron":"rekh'-em","derivation":"from H7355 (רָחַם); (compare H7356 (רַחַם))","strongs_def":"the womb","kjv_def":"matrix, womb."},
"H7359":{"lemma":"רְחֵם","xlit":"rᵉchêm","pron":"rekh-ame'","derivation":"(Aramaic) corresponding to H7356 (רַחַם);","strongs_def":"(plural) pity","kjv_def":"mercy."},
"H7360":{"lemma":"רָחָם","xlit":"râchâm","pron":"raw-khawm'","derivation":"or (feminine) רָחָמָה; from H7355 (רָחַם);","strongs_def":"a kind of vulture (supposed to be tender towards its young)","kjv_def":"gier-eagle."},
"H7361":{"lemma":"רַחֲמָה","xlit":"rachămâh","pron":"rakh-am-aw'","derivation":"feminine of H7356 (רַחַם);","strongs_def":"a maiden","kjv_def":"damsel."},
"H7362":{"lemma":"רַחְמָנִי","xlit":"rachmânîy","pron":"rakh-maw-nee'","derivation":"from H7355 (רָחַם);","strongs_def":"compassionate","kjv_def":"pitiful."},
"H7363":{"lemma":"רָחַף","xlit":"râchaph","pron":"raw-khaf'","derivation":"a primitive root;","strongs_def":"to brood; by implication, to be relaxed","kjv_def":"flutter, move, shake."},
"H7364":{"lemma":"רָחַץ","xlit":"râchats","pron":"raw-khats'","derivation":"a primitive root;","strongs_def":"to lave (the whole or a part of a thing)","kjv_def":"bathe (self), wash (self)."},
"H7365":{"lemma":"רְחַץ","xlit":"rᵉchats","pron":"rekh-ats'","derivation":"(Aramaic) corresponding to H7364 (רָחַץ) (probably through the accessory idea of ministering as a servant at the bath);","strongs_def":"to attend upon","kjv_def":"trust."},
"H7366":{"lemma":"רַחַץ","xlit":"rachats","pron":"rakh'-ats","derivation":"from H7364 (רָחַץ);","strongs_def":"a bath","kjv_def":"wash(-pot)."},
"H7367":{"lemma":"רַחְצָה","xlit":"rachtsâh","pron":"rakh-tsaw'","derivation":"feminine of H7366 (רַחַץ);","strongs_def":"a bathing place","kjv_def":"washing."},
"H7368":{"lemma":"רָחַק","xlit":"râchaq","pron":"raw-khak'","derivation":"a primitive root;","strongs_def":"to widen (in any direction), i.e. (intransitively) recede or (transitively) remove (literally or figuratively, of place or relation)","kjv_def":"(a-, be, cast, drive, get, go, keep (self), put, remove, be too, (wander), withdraw) far (away, off), loose, [idiom] refrain, very, (be) a good way (off)."},
"H7369":{"lemma":"רָחֵק","xlit":"râchêq","pron":"raw-khake'","derivation":"from H7368 (רָחַק);","strongs_def":"remote","kjv_def":"that are far."},
"H7370":{"lemma":"רָחַשׁ","xlit":"râchash","pron":"raw-khash'","derivation":"a primitive root;","strongs_def":"to gush","kjv_def":"indite."},
"H7371":{"lemma":"רַחַת","xlit":"rachath","pron":"rakh'-ath","derivation":"from H7306 (רוּחַ);","strongs_def":"a winnowing-fork (as blowing the chaff away)","kjv_def":"shovel."},
"H7372":{"lemma":"רָטַב","xlit":"râṭab","pron":"raw-tab'","derivation":"a primitive root;","strongs_def":"to be moist","kjv_def":"be wet."},
"H7373":{"lemma":"רָטֹב","xlit":"râṭôb","pron":"raw-tobe'","derivation":"from H7372 (רָטַב);","strongs_def":"moist (with sap)","kjv_def":"green."},
"H7374":{"lemma":"רֶטֶט","xlit":"reṭeṭ","pron":"reh'-tet","derivation":"from an unused root meaning to tremble;","strongs_def":"terror","kjv_def":"fear."},
"H7375":{"lemma":"רֻטֲפַשׁ","xlit":"ruṭăphash","pron":"roo-taf-ash'","derivation":"a root compounded from H7373 (רָטֹב) and H2954 (טָפַשׁ);","strongs_def":"to be rejuvenated","kjv_def":"be fresh."},
"H7376":{"lemma":"רָטָשׁ","xlit":"râṭâsh","pron":"raw-tash'","derivation":"a primitive root;","strongs_def":"to dash down","kjv_def":"dash (in pieces)."},
"H7377":{"lemma":"רִי","xlit":"rîy","pron":"ree","derivation":"from H7301 (רָוָה);","strongs_def":"irrigation, i.e. a shower","kjv_def":"watering."},
"H7378":{"lemma":"רִיב","xlit":"rîyb","pron":"reeb","derivation":"or רוּב; a primitive root;","strongs_def":"properly, to toss, i.e. grapple; mostly figuratively, to wrangle, i.e. hold a controversy; (by implication) to defend","kjv_def":"adversary, chide, complain, contend, debate, [idiom] ever, [idiom] lay wait, plead, rebuke, strive, [idiom] thoroughly."},
"H7379":{"lemma":"רִיב","xlit":"rîyb","pron":"reeb","derivation":"or רִב; from H7378 (רִיב);","strongs_def":"a contest (personal or legal)","kjv_def":"[phrase] adversary, cause, chiding, contend(-tion), controversy, multitude (from the margin), pleading, strife, strive(-ing), suit."},
"H7380":{"lemma":"רִיבַי","xlit":"Rîybay","pron":"ree-bah'-ee","derivation":"from H7378 (רִיב); contentious;","strongs_def":"Ribai, an Israelite","kjv_def":"Ribai."},
"H7381":{"lemma":"רֵיחַ","xlit":"rêyach","pron":"ray'-akh","derivation":"from H7306 (רוּחַ);","strongs_def":"odor (as if blown)","kjv_def":"savour, scent, smell."},
"H7382":{"lemma":"רֵיחַ","xlit":"rêyach","pron":"ray'-akh","derivation":"(Aramaic) corresponding to H7381 (רֵיחַ)","strongs_def":"{odor (as if blown)}","kjv_def":"smell."},
"H7383":{"lemma":"רִיפָה","xlit":"rîyphâh","pron":"ree-faw'","derivation":"or רִפָה; from H7322 (רוּף);","strongs_def":"(only plural), grits (as pounded)","kjv_def":"ground corn, wheat."},
"H7384":{"lemma":"רִיפַת","xlit":"Rîyphath","pron":"ree-fath'","derivation":"or (probably by orthographical error) דִּיפַת; of foreign origin;","strongs_def":"Riphath, a grandson of Japheth and his descendants","kjv_def":"Riphath."},
"H7385":{"lemma":"רִיק","xlit":"rîyq","pron":"reek","derivation":"from H7324 (רוּק);","strongs_def":"emptiness; figuratively, a worthless thing; adverbially, in vain","kjv_def":"empty, to no purpose, (in) vain (thing), vanity."},
"H7386":{"lemma":"רֵיק","xlit":"rêyq","pron":"rake","derivation":"or (shorter) רֵק; from H7324 (רוּק);","strongs_def":"empty; figuratively, worthless","kjv_def":"emptied(-ty), vain (fellow, man)."},
"H7387":{"lemma":"רֵיקָם","xlit":"rêyqâm","pron":"ray-kawm'","derivation":"from H7386 (רֵיק);","strongs_def":"emptily; figuratively (objective) ineffectually, (subjective) undeservedly","kjv_def":"without cause, empty, in vain, void."},
"H7388":{"lemma":"רִיר","xlit":"rîyr","pron":"reer","derivation":"from H7325 (רוּר);","strongs_def":"saliva; by resemblance, broth","kjv_def":"spittle, white (of an egg)."},
"H7389":{"lemma":"רֵישׁ","xlit":"rêysh","pron":"raysh","derivation":"or רֵאשׁ; or רִישׁ; from H7326 (רוּשׁ);","strongs_def":"poverty","kjv_def":"poverty."},
"H7390":{"lemma":"רַךְ","xlit":"rak","pron":"rak","derivation":"from H7401 (רָכַךְ);","strongs_def":"tender (literally or figuratively); by implication, weak","kjv_def":"faint((-hearted), soft, tender ((-hearted), one), weak."},
"H7391":{"lemma":"רֹךְ","xlit":"rôk","pron":"roke","derivation":"from H7401 (רָכַךְ);","strongs_def":"softness (figuratively)","kjv_def":"tenderness."},
"H7392":{"lemma":"רָכַב","xlit":"râkab","pron":"raw-kab'","derivation":"a primitive root;","strongs_def":"to ride (on an animal or in a vehicle); causatively, to place upon (for riding or generally), to despatch","kjv_def":"bring (on (horse-) back), carry, get (oneself) up, on (horse-) back, put, (cause to, make to) ride (in a chariot, on, -r), set."},
"H7393":{"lemma":"רֶכֶב","xlit":"rekeb","pron":"reh'-keb","derivation":"from H7392 (רָכַב);","strongs_def":"a vehicle; by implication, a team; by extension, cavalry; by analogy a rider, i.e. the upper millstone","kjv_def":"chariot, (upper) millstone, multitude (from the margin), wagon."},
"H7394":{"lemma":"רֵכָב","xlit":"Rêkâb","pron":"ray-kawb'","derivation":"from H7392 (רָכַב); rider;","strongs_def":"Rekab, the name of two Arabs and of two Israelites","kjv_def":"Rechab."},
"H7395":{"lemma":"רַכָּב","xlit":"rakkâb","pron":"rak-kawb'","derivation":"from H7392 (רָכַב);","strongs_def":"a charioteer","kjv_def":"chariot man, driver of a chariot, horseman."},
"H7396":{"lemma":"רִכְבָּה","xlit":"rikbâh","pron":"rik-baw'","derivation":"feminine of H7393 (רֶכֶב);","strongs_def":"a chariot (collectively)","kjv_def":"chariots."},
"H7397":{"lemma":"רֵכָה","xlit":"Rêkâh","pron":"ray-kaw'","derivation":"probably feminine from H7401 (רָכַךְ); softness;","strongs_def":"Rekah, a place in Palestine","kjv_def":"Rechah."},
"H7398":{"lemma":"רְכוּב","xlit":"rᵉkûwb","pron":"rek-oob'","derivation":"from passive participle of H7392 (רָכַב);","strongs_def":"a vehicle (as ridden on)","kjv_def":"chariot."},
"H7399":{"lemma":"רְכוּשׁ","xlit":"rᵉkûwsh","pron":"rek-oosh'","derivation":"or רְכֻשׁ; from passive participle of H7408 (רָכַשׁ);","strongs_def":"property (as gathered)","kjv_def":"good, riches, substance."},
"H7400":{"lemma":"רָכִיל","xlit":"râkîyl","pron":"raw-keel'","derivation":"from H7402 (רָכַל)","strongs_def":"a scandal-monger (as travelling about)","kjv_def":"slander, carry tales, talebearer."},
"H7401":{"lemma":"רָכַךְ","xlit":"râkak","pron":"raw-kak'","derivation":"a primitive root;","strongs_def":"to soften (intransitively or transitively), used figuratively","kjv_def":"(be) faint(-hearted), mollify, (be, make) soft(-er), be tender."},
"H7402":{"lemma":"רָכַל","xlit":"râkal","pron":"raw-kal'","derivation":"a primitive root;","strongs_def":"to travel for trading","kjv_def":"(spice) merchant."},
"H7403":{"lemma":"רָכָל","xlit":"Râkâl","pron":"raw-kawl'","derivation":"from H7402 (רָכַל); merchant;","strongs_def":"Rakal, a place in Palestine","kjv_def":"Rachal."},
"H7404":{"lemma":"רְכֻלָּה","xlit":"rᵉkullâh","pron":"rek-ool-law'","derivation":"feminine passive participle of H7402 (רָכַל);","strongs_def":"trade (as peddled)","kjv_def":"merchandise, traffic."},
"H7405":{"lemma":"רָכַס","xlit":"râkaç","pron":"raw-kas'","derivation":"a primitive root;","strongs_def":"to tie","kjv_def":"bind."},
"H7406":{"lemma":"רֶכֶס","xlit":"rekeç","pron":"reh'-kes","derivation":"from H7405 (רָכַס);","strongs_def":"a mountain ridge (as of tied summits)","kjv_def":"rough place."},
"H7407":{"lemma":"רֹכֶס","xlit":"rôkeç","pron":"ro'-kes","derivation":"from H7405 (רָכַס);","strongs_def":"a snare (as of tied meshes)","kjv_def":"pride."},
"H7408":{"lemma":"רָכַשׁ","xlit":"râkash","pron":"raw-kash'","derivation":"a primitive root;","strongs_def":"to lay up, i.e. collect","kjv_def":"gather, get."},
"H7409":{"lemma":"רֶכֶשׁ","xlit":"rekesh","pron":"reh'-kesh","derivation":"from H7408 (רָכַשׁ);","strongs_def":"a relay of animals on a post-route (as stored up for that purpose); by implication, a courser","kjv_def":"dromedary, mule, swift beast."},
"H7410":{"lemma":"רָם","xlit":"Râm","pron":"rawm","derivation":"active participle of H7311 (רוּם); high;","strongs_def":"Ram, the name of an Arabian and of an Israelite","kjv_def":"Ram. See also H1027 (בֵּית הָרָם)."},
"H7411":{"lemma":"רָמָה","xlit":"râmâh","pron":"raw-maw'","derivation":"a primitive root;","strongs_def":"to hurl; specifically, to shoot; figuratively, to delude or betray (as if causing to fall)","kjv_def":"beguile, betray, (bow-) man, carry, deceive, throw."},
"H7412":{"lemma":"רְמָה","xlit":"rᵉmâh","pron":"rem-aw'","derivation":"(Aramaic) corresponding to H7411 (רָמָה);","strongs_def":"to throw, set, (figuratively) assess","kjv_def":"cast (down), impose."},
"H7413":{"lemma":"רָמָה","xlit":"râmâh","pron":"raw-maw'","derivation":"feminine active participle of H7311 (רוּם);","strongs_def":"a height (as a seat of idolatry)","kjv_def":"high place."},
"H7414":{"lemma":"רָמָה","xlit":"Râmâh","pron":"raw-maw'","derivation":"the same as H7413 (רָמָה);","strongs_def":"Ramah, the name of four places in Palestine","kjv_def":"Ramah."},
"H7415":{"lemma":"רִמָּה","xlit":"rimmâh","pron":"rim-maw'","derivation":"from H7426 (רָמַם) in the sense of breading (compare H7311 (רוּם));","strongs_def":"a maggot (as rapidly bred), literally or figuratively","kjv_def":"worm."},
"H7416":{"lemma":"רִמּוֹן","xlit":"rimmôwn","pron":"rim-mone'","derivation":"or רִמֹּן; from H7426 (רָמַם);","strongs_def":"a pomegranate, the tree (from its upright growth) or the fruit (also an artificial ornament)","kjv_def":"pomegranate."},
"H7417":{"lemma":"רִמּוֹן","xlit":"Rimmôwn","pron":"rim-mone'","derivation":"or (shorter) (רִמֹּן; or רִמּוֹנוֹ; (1 Chronicles 6:62 (H77 (אֶבֶץ))), the same as H7416 (רִמּוֹן); in Joshua 19:13: הַמְּתֹאָר is mistaken for part of a name; article with the passive participle of H8388 (תָּאַר); the (one) marked off, i.e. which pertains).","strongs_def":"Rimmon, the name of a Syrian deity, also of five places in Palestine","kjv_def":"Remmon, Rimmon."},
"H7418":{"lemma":"רָמוֹת","xlit":"Râmôwth","pron":"raw-moth","derivation":"or רָמַת נֶגֶב; from the plural or construct form of H7413 (רָמָה) and H5045 (נֶגֶב); heights (or height) of the South;","strongs_def":"Ramoth-Negeb or Ramath-Negeb, a place in Palestine","kjv_def":"south Ramoth, Ramath of the south."},
"H7419":{"lemma":"רָמוּת","xlit":"râmûwth","pron":"raw-mooth'","derivation":"from H7311 (רוּם);","strongs_def":"a heap (of carcases)","kjv_def":"height."},
"H7420":{"lemma":"רֹמַח","xlit":"rômach","pron":"ro'-makh","derivation":"from an unused root meaning to hurl;","strongs_def":"a lance (as thrown); especially the iron point","kjv_def":"buckler, javelin, lancet, spear."},
"H7421":{"lemma":"רַמִּי","xlit":"rammîy","pron":"ram-mee'","derivation":"for H761 (אֲרַמִּי);","strongs_def":"a Ramite, i.e. Aramaean","kjv_def":"Syrian."},
"H7422":{"lemma":"רַמְיָה","xlit":"Ramyâh","pron":"ram-yaw'","derivation":"from H7311 (רוּם) and H3050 (יָהּ); Jah has raised;","strongs_def":"Ramjah, an Israelite","kjv_def":"Ramiah."},
"H7423":{"lemma":"רְמִיָּה","xlit":"rᵉmîyâh","pron":"rem-ee-yaw'","derivation":"from H7411 (רָמָה);","strongs_def":"remissness, treachery","kjv_def":"deceit(-ful, -fully), false, guile, idle, slack, slothful."},
"H7424":{"lemma":"רַמָּךְ","xlit":"rammâk","pron":"ram-mawk'","derivation":"of foreign origin;","strongs_def":"a brood mare","kjv_def":"dromedary."},
"H7425":{"lemma":"רְמַלְיָהוּ","xlit":"Rᵉmalyâhûw","pron":"rem-al-yaw'-hoo","derivation":"from an unused root and H3050 (יָהּ) (perhaps meaning to deck); Jah has bedecked;","strongs_def":"Remaljah, an Israelite","kjv_def":"Remaliah."},
"H7426":{"lemma":"רָמַם","xlit":"râmam","pron":"raw-mam'","derivation":"a primitive root;","strongs_def":"to rise (literally or figuratively)","kjv_def":"exalt, get (oneself) up, lift up (self), mount up."},
"H7427":{"lemma":"רֹמֵמֻת","xlit":"rômêmuth","pron":"ro-may-mooth'","derivation":"from the active participle of H7426 (רָמַם);","strongs_def":"exaltation","kjv_def":"lifting up of self."},
"H7428":{"lemma":"רִמֹּן פֶּרֶץ","xlit":"Rimmôn Perets","pron":"rim-mone' peh'-rets","derivation":"from H7416 (רִמּוֹן) and H6556 (פֶּרֶץ); pomegranate of the breach;","strongs_def":"Rimmon-Perets, a place in the Desert","kjv_def":"Rimmon-parez."},
"H7429":{"lemma":"רָמַס","xlit":"râmaç","pron":"raw-mas'","derivation":"a primitive root;","strongs_def":"to tread upon (as a potter, in walking or abusively)","kjv_def":"oppressor, stamp upon, trample (under feet), tread (down, upon)."},
"H7430":{"lemma":"רָמַשׂ","xlit":"râmas","pron":"raw-mas'","derivation":"a primitive root;","strongs_def":"properly, to glide swiftly, i.e. to crawl or move with short steps; by analogy to swarm","kjv_def":"creep, move."},
"H7431":{"lemma":"רֶמֶשׂ","xlit":"remes","pron":"reh'-mes","derivation":"from H7430 (רָמַשׂ);","strongs_def":"a reptile or any other rapidly moving animal","kjv_def":"that creepeth, creeping (moving) thing."},
"H7432":{"lemma":"רֶמֶת","xlit":"Remeth","pron":"reh'-meth","derivation":"from H7411 (רָמָה); height;","strongs_def":"Remeth, a place in Palestine","kjv_def":"Remeth."},
"H7433":{"lemma":"רָמֹת גִּלעָד","xlit":"Râmôth Gilʻâd","pron":"raw-moth' gil-awd'","derivation":"or רָמוֹת גִּלעָד; (2 Chronicles 22:5), from the plural of H7413 (רָמָה) and H1568 (גִּלְעָד); heights of Gilad;","strongs_def":"Ramoth-Gilad, a place East of the Jordan","kjv_def":"Ramoth-gilead, Ramoth in Gilead. See also H7216 (רָאמוֹת)."},
"H7434":{"lemma":"רָמַת הַמִּצְפֶּה","xlit":"Râmath ham-Mit-speh","pron":"raw-math' ham-mitspeh'","derivation":"from H7413 (רָמָה) and H4707 (מִצְפֶּה) with the article interpolated; height of the watch-tower;","strongs_def":"Ramath-ham-Mitspeh, a place in Palestine","kjv_def":"Ramath-mizpeh."},
"H7435":{"lemma":"רָמָתִי","xlit":"Râmâthîy","pron":"raw-maw-thee'","derivation":"patronymic of H7414 (רָמָה);","strongs_def":"a Ramathite or inhabitant of Ramah","kjv_def":"Ramathite."},
"H7436":{"lemma":"רָמָתַיִם צוֹפִים","xlit":"Râmâthayim Tsôwphîym","pron":"raw-maw-thah'-yim tso-feem'","derivation":"from the dual of H7413 (רָמָה) and the plural of the active participle of H6822 (צָפָה); double height of watchers;","strongs_def":"Ramathajim-Tsophim, a place in Palestine","kjv_def":"Ramathaimzophim."},
"H7437":{"lemma":"רָמַת לֶחִי","xlit":"Râmath Lechîy","pron":"raw'-math lekh'-ee","derivation":"from H7413 (רָמָה) and H3895 (לְחִי); height of a jaw-bone;","strongs_def":"Ramath-Lechi, a place in Palestine","kjv_def":"Ramath-lehi."},
"H7438":{"lemma":"רֹן","xlit":"rôn","pron":"rone","derivation":"from H7442 (רָנַן);","strongs_def":"a shout (of deliverance)","kjv_def":"song."},
"H7439":{"lemma":"רָנָה","xlit":"rânâh","pron":"raw-naw'","derivation":"a primitive root;","strongs_def":"to whiz","kjv_def":"rattle."},
"H7440":{"lemma":"רִנָּה","xlit":"rinnâh","pron":"rin-naw'","derivation":"from H7442 (רָנַן);","strongs_def":"properly, a creaking (or shrill sound), i.e. shout (of joy or grief)","kjv_def":"cry, gladness, joy, proclamation, rejoicing, shouting, sing(-ing), triumph."},
"H7441":{"lemma":"רִנָּה","xlit":"Rinnâh","pron":"rin-naw'","derivation":"the same as H7440 (רִנָּה);","strongs_def":"Rinnah, an Israelite","kjv_def":"Rinnah."},
"H7442":{"lemma":"רָנַן","xlit":"rânan","pron":"raw-nan'","derivation":"a primitive root;","strongs_def":"properly, to creak (or emit a stridulous sound), i.e. to shout (usually for joy)","kjv_def":"aloud for joy, cry out, be joyful (greatly, make to) rejoice, (cause to) shout (for joy), (cause to) sing (aloud, for joy, out), triumph."},
"H7443":{"lemma":"רֶנֶן","xlit":"renen","pron":"reh'-nen","derivation":"from H7442 (רָנַן);","strongs_def":"an ostrich (from its wail)","kjv_def":"[idiom] goodly."},
"H7444":{"lemma":"רַנֵּן","xlit":"rannên","pron":"ran-nane'","derivation":"intensive from H7442 (רָנַן);","strongs_def":"shouting (for joy)","kjv_def":"singing."},
"H7445":{"lemma":"רְנָנָה","xlit":"rᵉnânâh","pron":"ren-aw-naw'","derivation":"from H7442 (רָנַן);","strongs_def":"a shout (for joy)","kjv_def":"joyful (voice), singing, triumphing."},
"H7446":{"lemma":"רִסָּה","xlit":"Riççâh","pron":"ris-saw'","derivation":"from H7450 (רָסַס); a ruin (as dripping to pieces);","strongs_def":"Rissah, a place in the Desert","kjv_def":"Rissah."},
"H7447":{"lemma":"רָסִיס","xlit":"râçîyç","pron":"raw-sees'","derivation":"from H7450 (רָסַס);","strongs_def":"properly, dripping to pieces, i.e. a ruin; also a dew-drop","kjv_def":"breach, drop."},
"H7448":{"lemma":"רֶסֶן","xlit":"reçen","pron":"reh'-sen","derivation":"from an unused root meaning to curb;","strongs_def":"a halter (as restraining); by implication, the jaw","kjv_def":"bridle."},
"H7449":{"lemma":"רֶסֶן","xlit":"Reçen","pron":"reh'-sen","derivation":"the same as H7448 (רֶסֶן);","strongs_def":"Resen, a place in Assyrian","kjv_def":"Resen."},
"H7450":{"lemma":"רָסַס","xlit":"râçaç","pron":"raw-sas'","derivation":"a primitive root; to comminute; used only as denominative from H7447 (רָסִיס),","strongs_def":"to moisten (with drops)","kjv_def":"temper."},
"H7451":{"lemma":"רַע","xlit":"raʻ","pron":"rah","derivation":"from H7489 (רָעַע);","strongs_def":"bad or (as noun) evil (natural or moral)","kjv_def":"adversity, affliction, bad, calamity, [phrase] displease(-ure), distress, evil((-favouredness), man, thing), [phrase] exceedingly, [idiom] great, grief(-vous), harm, heavy, hurt(-ful), ill (favoured), [phrase] mark, mischief(-vous), misery, naught(-ty), noisome, [phrase] not please, sad(-ly), sore, sorrow, trouble, vex, wicked(-ly, -ness, one), worse(-st), wretchedness, wrong. (Incl. feminine raaah; as adjective or noun.)."},
"H7452":{"lemma":"רֵעַ","xlit":"rêaʻ","pron":"ray'-ah","derivation":"from H7321 (רוּעַ);","strongs_def":"a crash (of thunder), noise (of war), shout (of joy)","kjv_def":"[idiom] aloud, noise, shouted."},
"H7453":{"lemma":"רֵעַ","xlit":"rêaʻ","pron":"ray'-ah","derivation":"or רֵיעַ; from H7462 (רָעָה);","strongs_def":"an associate (more or less close)","kjv_def":"brother, companion, fellow, friend, husband, lover, neighbour, [idiom] (an-) other."},
"H7454":{"lemma":"רֵעַ","xlit":"rêaʻ","pron":"ray'-ah","derivation":"from H7462 (רָעָה);","strongs_def":"a thought (as association of ideas)","kjv_def":"thought."},
"H7455":{"lemma":"רֹעַ","xlit":"rôaʻ","pron":"ro'-ah","derivation":"from H7489 (רָעַע);","strongs_def":"badness (as marring), physically or morally","kjv_def":"[idiom] be so bad, badness, ([idiom] be so) evil, naughtiness, sadness, sorrow, wickedness."},
"H7456":{"lemma":"רָעֵב","xlit":"râʻêb","pron":"raw-abe'","derivation":"a primitive root;","strongs_def":"to hunger","kjv_def":"(suffer to) famish, (be, have, suffer, suffer to) hunger(-ry)."},
"H7457":{"lemma":"רָעֵב","xlit":"râʻêb","pron":"raw-abe'","derivation":"from H7456 (רָעֵב);","strongs_def":"hungry (more or less intensely)","kjv_def":"hunger bitten, hungry."},
"H7458":{"lemma":"רָעָב","xlit":"râʻâb","pron":"raw-awb'","derivation":"from H7456 (רָעֵב);","strongs_def":"hunger (more or less extensive)","kjv_def":"dearth, famine, [phrase] famished, hunger."},
"H7459":{"lemma":"רְעָבוֹן","xlit":"rᵉʻâbôwn","pron":"reh-aw-bone'","derivation":"from H7456 (רָעֵב);","strongs_def":"famine","kjv_def":"famine."},
"H7460":{"lemma":"רָעַד","xlit":"râʻad","pron":"raw-ad'","derivation":"a primitive root;","strongs_def":"to shudder (more or less violently)","kjv_def":"tremble."},
"H7461":{"lemma":"רַעַד","xlit":"raʻad","pron":"rah'-ad","derivation":"or (feminine) רְעָדָהxlit rᵉâdâh corrected to rᵉʻâdâh; from H7460 (רָעַד);","strongs_def":"a shudder","kjv_def":"trembling."},
"H7462":{"lemma":"רָעָה","xlit":"râʻâh","pron":"raw-aw'","derivation":"a primitive root;","strongs_def":"to tend a flock; i.e. pasture it; intransitively, to graze (literally or figuratively); generally to rule; by extension, to associate with (as a friend)","kjv_def":"[idiom] break, companion, keep company with, devour, eat up, evil entreat, feed, use as a friend, make friendship with, herdman, keep (sheep) (-er), pastor, [phrase] shearing house, shepherd, wander, waste."},
"H7463":{"lemma":"רֵעֶה","xlit":"rêʻeh","pron":"ray-eh'","derivation":"from H7462 (רָעָה);","strongs_def":"a (male) companion","kjv_def":"friend."},
"H7464":{"lemma":"רֵעָה","xlit":"rêʻâh","pron":"ray'-aw","derivation":"feminine of H7453 (רֵעַ);","strongs_def":"a female associate","kjv_def":"companion, fellow."},
"H7465":{"lemma":"רֹעָה","xlit":"rôʻâh","pron":"ro-aw'","derivation":"for H7455 (רֹעַ);","strongs_def":"breakage","kjv_def":"broken, utterly."},
"H7466":{"lemma":"רְעוּ","xlit":"Rᵉʻûw","pron":"reh-oo'","derivation":"for H7471 (רְעִי) in the sense of H7453 (רֵעַ); friend;","strongs_def":"Reu, a postdiluvian patriarch","kjv_def":"Reu."},
"H7467":{"lemma":"רְעוּאֵל","xlit":"Rᵉʻûwʼêl","pron":"reh-oo-ale'","derivation":"from the same as H7466 (רְעוּ) and H410 (אֵל); friend of God;","strongs_def":"Reuel, the name of Moses' father-in-law, also of an Edomite and an Israelite","kjv_def":"Raguel, Reuel."},
"H7468":{"lemma":"רְעוּת","xlit":"rᵉʻûwth","pron":"reh-ooth'","derivation":"from H7462 (רָעָה) in the sense of H7453 (רֵעַ);","strongs_def":"a female associate; generally an additional one","kjv_def":"[phrase] another, mate, neighbour."},
"H7469":{"lemma":"רְעוּת","xlit":"rᵉʻûwth","pron":"reh-ooth'","derivation":"probably from H7462 (רָעָה);","strongs_def":"a feeding upon, i.e. grasping after","kjv_def":"vexation."},
"H7470":{"lemma":"רְעוּת","xlit":"rᵉʻûwth","pron":"reh-ooth'","derivation":"(Aramaic) corresponding to H7469 (רְעוּת);","strongs_def":"desire","kjv_def":"pleasure, will."},
"H7471":{"lemma":"רְעִי","xlit":"rᵉʻîy","pron":"reh-ee'","derivation":"from H7462 (רָעָה);","strongs_def":"pasture","kjv_def":"pasture."},
"H7472":{"lemma":"רֵעִי","xlit":"Rêʻîy","pron":"ray-ee'","derivation":"from H7453 (רֵעַ); social;","strongs_def":"Rei, an Israelite","kjv_def":"Rei."},
"H7473":{"lemma":"רֹעִי","xlit":"rôʻîy","pron":"ro-ee'","derivation":"from active participle of H7462 (רָעָה);","strongs_def":"pastoral; as noun, a shepherd","kjv_def":"shipherd."},
"H7474":{"lemma":"רַעְיָה","xlit":"raʻyâh","pron":"rah-yaw'","derivation":"feminine of H7453 (רֵעַ);","strongs_def":"a female associate","kjv_def":"fellow, love."},
"H7475":{"lemma":"רַעְיוֹן","xlit":"raʻyôwn","pron":"rah-yone'","derivation":"from H7462 (רָעָה) in the sense of H7469 (רְעוּת);","strongs_def":"desire","kjv_def":"vexation."},
"H7476":{"lemma":"רַעְיוֹן","xlit":"raʻyôwn","pron":"rah-yone'","derivation":"(Aramaic) corresponding to H7475 (רַעְיוֹן);","strongs_def":"a grasp. i.e. (figuratively) mental conception","kjv_def":"cogitation, thought."},
"H7477":{"lemma":"רָעַל","xlit":"râʻal","pron":"raw-al'","derivation":"a primitive root;","strongs_def":"to reel, i.e. (figuratively) to brandish","kjv_def":"terribly shake."},
"H7478":{"lemma":"רַעַל","xlit":"raʻal","pron":"rah'-al","derivation":"from H7477 (רָעַל);","strongs_def":"a reeling (from intoxication)","kjv_def":"trembling."},
"H7479":{"lemma":"רַעֲלָה","xlit":"raʻălâh","pron":"rah-al-aw'","derivation":"feminine of H7478 (רַעַל);","strongs_def":"a long veil (as fluttering)","kjv_def":"muffler."},
"H7480":{"lemma":"רְעֵלָיָה","xlit":"Rᵉʻêlâyâh","pron":"reh-ay-law-yaw'","derivation":"from H7477 (רָעַל) and H3050 (יָהּ); made to tremble (i.e. fearful) of Jah;","strongs_def":"Reelajah, an Israelite","kjv_def":"Reeliah."},
"H7481":{"lemma":"רָעַם","xlit":"râʻam","pron":"raw-am'","derivation":"a primitive root;","strongs_def":"to tumble, i.e. be violently agitated; specifically, to crash (of thunder); figuratively, to irritate (with anger)","kjv_def":"make to fret, roar, thunder, trouble."},
"H7482":{"lemma":"רַעַם","xlit":"raʻam","pron":"rah'am","derivation":"from H7481 (רָעַם);","strongs_def":"a peal of thunder","kjv_def":"thunder."},
"H7483":{"lemma":"רַעְמָה","xlit":"raʻmâh","pron":"rah-maw'","derivation":"feminine of H7482 (רַעַם);","strongs_def":"the mane of a horse (as quivering in the wind)","kjv_def":"thunder."},
"H7484":{"lemma":"רַעְמָה","xlit":"Raʻmâh","pron":"rah-maw'","derivation":"the same as H7483 (רַעְמָה);","strongs_def":"Ramah, the name of a grandson of Ham, and of a place (perhaps founded by him)","kjv_def":"Raamah."},
"H7485":{"lemma":"רַעַמְיָה","xlit":"Raʻamyâh","pron":"rah-am-yaw'","derivation":"from H7481 (רָעַם) and H3050 (יָהּ); Jah has shaken;","strongs_def":"Raamjah, an Israelite","kjv_def":"Raamiah."},
"H7486":{"lemma":"רַעְמְסֵס","xlit":"Raʻmᵉçêç","pron":"rah-mes-ace'","derivation":"or רַעַמְסֵס; of Egyptian origin;","strongs_def":"Rameses or Raamses, a place in Egypt","kjv_def":"Raamses, Rameses."},
"H7487":{"lemma":"רַעֲנַן","xlit":"raʻănan","pron":"rah-aw-nan'","derivation":"(Aramaic) corresponding to H7488 (רַעֲנָן);","strongs_def":"green, i.e. (figuratively) prosperous","kjv_def":"flourishing."},
"H7488":{"lemma":"רַעֲנָן","xlit":"raʻănân","pron":"rah-an-awn'","derivation":"from an unused root meaning to be green;","strongs_def":"verdant; by analogy, new; figuratively, prosperous","kjv_def":"green, flourishing."},
"H7489":{"lemma":"רָעַע","xlit":"râʻaʻ","pron":"raw-ah'","derivation":"a primitive root;","strongs_def":"properly, to spoil (literally, by breaking to pieces); figuratively, to make (or be) good fornothing, i.e. bad (physically, socially or morally)","kjv_def":"afflict, associate selves (by mistake for H7462 (רָעָה)), break (down, in pieces), [phrase] displease, (be, bring, do) evil (doer, entreat, man), show self friendly (by mistake for H7462 (רָעָה)), do harm, (do) hurt, (behave self, deal) ill, [idiom] indeed, do mischief, punish, still, vex, (do) wicked (doer, -ly), be (deal, do) worse."},
"H7490":{"lemma":"רְעַע","xlit":"rᵉʻaʻ","pron":"reh-ah'","derivation":"(Aramaic) corresponding to H7489 (רָעַע)","strongs_def":"{properly, to spoil (literally, by breaking to pieces); figuratively, to make (or be) good fornothing, i.e. bad (physically, socially or morally)}","kjv_def":"break, bruise."},
"H7491":{"lemma":"רָעַף","xlit":"râʻaph","pron":"raw-af'","derivation":"a primitive root;","strongs_def":"to drip","kjv_def":"distil, drop (down)."},
"H7492":{"lemma":"רָעַץ","xlit":"râʻats","pron":"raw-ats'","derivation":"a primitive root;","strongs_def":"to break in pieces; figuratively, harass","kjv_def":"dash in pieces, vex."},
"H7493":{"lemma":"רָעַשׁ","xlit":"râʻash","pron":"raw-ash","derivation":"a primitive root;","strongs_def":"to undulate (as the earth, the sky, etc.; also a field of grain), particularly through fear; specifically, to spring (as a locust)","kjv_def":"make afraid, (re-) move, quake, (make to) shake, (make to) tremble."},
"H7494":{"lemma":"רַעַשׁ","xlit":"raʻash","pron":"rah'-ash","derivation":"from H7493 (רָעַשׁ);","strongs_def":"vibration, bounding, uproar","kjv_def":"commotion, confused noise, earthquake, fierceness, quaking, rattling, rushing, shaking."},
"H7495":{"lemma":"רָפָא","xlit":"râphâʼ","pron":"raw-faw'","derivation":"or רָפָה; a primitive root;","strongs_def":"properly, to mend (by stitching), i.e. (figuratively) to cure","kjv_def":"cure, (cause to) heal, physician, repair, [idiom] thoroughly, make whole. See H7503 (רָפָה)."},
"H7496":{"lemma":"רָפָא","xlit":"râphâʼ","pron":"raw-faw'","derivation":"from H7495 (רָפָא) in the sense of H7503 (רָפָה);","strongs_def":"properly, lax, i.e. (figuratively) a ghost (as dead; in plural only)","kjv_def":"dead, deceased."},
"H7497":{"lemma":"רָפָא","xlit":"râphâʼ","pron":"raw-faw'","derivation":"or רָפָה; from H7495 (רָפָא) in the sense of invigorating;","strongs_def":"a giant","kjv_def":"giant, Rapha, Rephaim(-s). See also H1051 (בֵּית רָפָא)."},
"H7498":{"lemma":"רָפָא","xlit":"Râphâʼ","pron":"raw-faw'","derivation":"or רָפָה; probably the same as H7497 (רָפָא); giant;","strongs_def":"Rapha or Raphah, the name of two Israelites","kjv_def":"Rapha."},
"H7499":{"lemma":"רְפֻאָה","xlit":"rᵉphuʼâh","pron":"ref-oo-aw'","derivation":"feminine passive participle of H7495 (רָפָא);","strongs_def":"a medicament","kjv_def":"heal(-ed), medicine."},
"H7500":{"lemma":"רִפְאוּת","xlit":"riphʼûwth","pron":"rif-ooth'","derivation":"from H7495 (רָפָא);","strongs_def":"a cure","kjv_def":"health."},
"H7501":{"lemma":"רְפָאֵל","xlit":"Rᵉphâʼêl","pron":"ref-aw-ale'","derivation":"from H7495 (רָפָא) and H410 (אֵל); God has cured;","strongs_def":"Rephael, an Israelite","kjv_def":"Rephael."},
"H7502":{"lemma":"רָפַד","xlit":"râphad","pron":"raw-fad'","derivation":"a primitive root;","strongs_def":"to spread (a bed); by implication, to refresh","kjv_def":"comfort, make (a bed), spread."},
"H7503":{"lemma":"רָפָה","xlit":"râphâh","pron":"raw-faw'","derivation":"a primitive root;","strongs_def":"to slacken (in many applications, literal or figurative)","kjv_def":"abate, cease, consume, draw (toward evening), fail, (be) faint, be (wax) feeble, forsake, idle, leave, let alone (go, down), (be) slack, stay, be still, be slothful, (be) weak(-en). See H7495 (רָפָא)."},
"H7504":{"lemma":"רָפֶה","xlit":"râpheh","pron":"raw-feh'","derivation":"from H7503 (רָפָה);","strongs_def":"slack (in body or mind)","kjv_def":"weak. raphah, Raphah. See H7497 (רָפָא), H7498 (רָפָא). riphah. See H7383 (רִיפָה)."},
"H7505":{"lemma":"רָפוּא","xlit":"Râphûwʼ","pron":"raw-foo'","derivation":"passive participle of H7495 (רָפָא); cured;","strongs_def":"Raphu, an Israelite","kjv_def":"Raphu."},
"H7506":{"lemma":"רֶפַח","xlit":"Rephach","pron":"reh'-fakh","derivation":"from an unused root apparently meaning to sustain; support;","strongs_def":"Rephach, an Israelite","kjv_def":"Rephah."},
"H7507":{"lemma":"רְפִידָה","xlit":"rᵉphîydâh","pron":"ref-ee-daw'","derivation":"from H7502 (רָפַד);","strongs_def":"a railing (as spread along)","kjv_def":"bottom."},
"H7508":{"lemma":"רְפִידִים","xlit":"Rᵉphîydîym","pron":"ref-ee-deem'","derivation":"plural of the masculine of the same as H7507 (רְפִידָה); ballusters;","strongs_def":"Rephidim, a place in the Desert","kjv_def":"Rephidim."},
"H7509":{"lemma":"רְפָיָה","xlit":"Rᵉphâyâh","pron":"ref-aw-yaw'","derivation":"from H7495 (רָפָא) and H3050 (יָהּ); Jah has cured;","strongs_def":"Rephajah, the name of five Israelites","kjv_def":"Rephaiah."},
"H7510":{"lemma":"רִפְיוֹן","xlit":"riphyôwn","pron":"rif-yone'","derivation":"from H7503 (רָפָה);","strongs_def":"slackness","kjv_def":"feebleness."},
"H7511":{"lemma":"רָפַס","xlit":"râphaç","pron":"raw-fas'","derivation":"a primitive root;","strongs_def":"to trample, i.e. prostrate","kjv_def":"humble self, submit self."},
"H7512":{"lemma":"רְפַס","xlit":"rᵉphaç","pron":"ref-as'","derivation":"(Aramaic) corresponding to H7511 (רָפַס)","strongs_def":"{to trample, i.e. prostrate}","kjv_def":"stamp."},
"H7513":{"lemma":"רַפְסֹדָה","xlit":"raphçôdâh","pron":"raf-so-daw'","derivation":"from H7511 (רָפַס);","strongs_def":"a raft (as flat on the water)","kjv_def":"flote."},
"H7514":{"lemma":"רָפַק","xlit":"râphaq","pron":"raw-fak'","derivation":"a primitive root;","strongs_def":"to recline","kjv_def":"lean."},
"H7515":{"lemma":"רָפַשׂ","xlit":"râphas","pron":"raw-fas'","derivation":"a primitive root;","strongs_def":"to trample, i.e. roil water","kjv_def":"foul, trouble."},
"H7516":{"lemma":"רֶפֶשׁ","xlit":"rephesh","pron":"reh'-fesh","derivation":"from H7515 (רָפַשׂ);","strongs_def":"mud (as roiled)","kjv_def":"mire."},
"H7517":{"lemma":"רֶפֶת","xlit":"repheth","pron":"reh'-feth","derivation":"probably from H7503 (רָפָה);","strongs_def":"a stall for cattle (from their resting there)","kjv_def":"stall."},
"H7518":{"lemma":"רַץ","xlit":"rats","pron":"rats","derivation":"contracted from H7533 (רָצַץ);","strongs_def":"a fragment","kjv_def":"piece."},
"H7519":{"lemma":"רָצָא","xlit":"râtsâʼ","pron":"raw-tsaw'","derivation":"a primitive root;","strongs_def":"to run; also to delight in","kjv_def":"accept, run."},
"H7520":{"lemma":"רָצַד","xlit":"râtsad","pron":"raw-tsad'","derivation":"a primitive root;","strongs_def":"probably to look askant, i.e. (figuratively) be jealous","kjv_def":"leap."},
"H7521":{"lemma":"רָצָה","xlit":"râtsâh","pron":"raw-tsaw'","derivation":"a primitive root;","strongs_def":"to be pleased with; specifically, to satisfy adebt","kjv_def":"(be) accept(-able), accomplish, set affection, approve, consent with, delight (self), enjoy, (be, have a) favour(-able), like, observe, pardon, (be, have, take) please(-ure), reconcile self."},
"H7522":{"lemma":"רָצוֹן","xlit":"râtsôwn","pron":"raw-tsone'","derivation":"or רָצֹן; from H7521 (רָצָה);","strongs_def":"delight (especially as shown)","kjv_def":"(be) acceptable(-ance, -ed), delight, desire, favour, (good) pleasure, (own, self, voluntary) will, as...(what) would."},
"H7523":{"lemma":"רָצַח","xlit":"râtsach","pron":"raw-tsakh'","derivation":"a primitive root;","strongs_def":"properly, to dash in pieces, i.e. kill (a human being), especially to murder","kjv_def":"put to death, kill, (man-) slay(-er), murder(-er)."},
"H7524":{"lemma":"רֶצַח","xlit":"retsach","pron":"reh-tsakh","derivation":"from H7523 (רָצַח);","strongs_def":"a crushing; specifically, a murder-cry","kjv_def":"slaughter, sword."},
"H7525":{"lemma":"רִצְיָא","xlit":"Ritsyâʼ","pron":"rits-yaw'","derivation":"from H7521 (רָצָה); delight;","strongs_def":"Ritsjah, an Israelite","kjv_def":"Rezia."},
"H7526":{"lemma":"רְצִין","xlit":"Rᵉtsîyn","pron":"rets-een'","derivation":"probably for H7522 (רָצוֹן);","strongs_def":"Retsin, the name of a Syrian and of an Israelite","kjv_def":"Rezin."},
"H7527":{"lemma":"רָצַע","xlit":"râtsaʻ","pron":"raw-tsah'","derivation":"a primitive root;","strongs_def":"to pierce","kjv_def":"bore."},
"H7528":{"lemma":"רָצַף","xlit":"râtsaph","pron":"raw-tsaf'","derivation":"a denominative from H7529 (רֶצֶף);","strongs_def":"to tessellate, i.e. embroider (as if with bright stones)","kjv_def":"pave."},
"H7529":{"lemma":"רֶצֶף","xlit":"retseph","pron":"reh'-tsef","derivation":"for H7565 (רֶשֶׁף);","strongs_def":"a red-hot stone (for baking)","kjv_def":"coal."},
"H7530":{"lemma":"רֶצֶף","xlit":"Retseph","pron":"reh'-tsef","derivation":"the same as H7529 (רֶצֶף);","strongs_def":"Retseph, a place in Assyrian","kjv_def":"Rezeph."},
"H7531":{"lemma":"רִצְפָּה","xlit":"ritspâh","pron":"rits-paw'","derivation":"feminine of H7529 (רֶצֶף);","strongs_def":"a hot stone; also a tessellated pavement","kjv_def":"live coal, pavement."},
"H7532":{"lemma":"רִצְפָּה","xlit":"Ritspâh","pron":"rits-paw'","derivation":"the same as H7531 (רִצְפָּה);","strongs_def":"Ritspah, an Israelitess","kjv_def":"Rizpah."},
"H7533":{"lemma":"רָצַץ","xlit":"râtsats","pron":"raw-tsats'","derivation":"a primitive root;","strongs_def":"to crack in pieces, literally or figuratively","kjv_def":"break, bruise, crush, discourage, oppress, struggle together."},
"H7534":{"lemma":"רַק","xlit":"raq","pron":"rak","derivation":"from H7556 (רָקַק) in its original sense;","strongs_def":"emaciated (as if flattened out)","kjv_def":"lean(-fleshed), thin."},
"H7535":{"lemma":"רַק","xlit":"raq","pron":"rak","derivation":"the same as H7534 (רַק) as a noun;","strongs_def":"properly, leanness, i.e. (figuratively) limitation; only adverbial, merely, or conjunctional, although","kjv_def":"but, even, except, howbeit howsoever, at the least, nevertheless, nothing but, notwithstanding, only, save, so (that), surely, yet (so), in any wise."},
"H7536":{"lemma":"רֹק","xlit":"rôq","pron":"roke","derivation":"from H7556 (רָקַק);","strongs_def":"spittle","kjv_def":"spit(-ting, -tle)."},
"H7537":{"lemma":"רָקַב","xlit":"râqab","pron":"raw-kab'","derivation":"a primitive root;","strongs_def":"to decay (as by worm-eating)","kjv_def":"rot."},
"H7538":{"lemma":"רָקָב","xlit":"râqâb","pron":"raw-kawb'","derivation":"from H7537 (רָקַב);","strongs_def":"decay (by caries)","kjv_def":"rottenness (thing)."},
"H7539":{"lemma":"רִקָּבוֹן","xlit":"riqqâbôwn","pron":"rik-kaw-bone'","derivation":"from H7538 (רָקָב);","strongs_def":"decay (by caries)","kjv_def":"rotten."},
"H7540":{"lemma":"רָקַד","xlit":"râqad","pron":"raw-kad'","derivation":"a primitive root;","strongs_def":"properly, to stamp, i.e. to spring about (wildly or for joy)","kjv_def":"dance, jump, leap, skip."},
"H7541":{"lemma":"רַקָּה","xlit":"raqqâh","pron":"rak-kaw'","derivation":"feminine of H7534 (רַק);","strongs_def":"properly, thinness, i.e. the side of the head","kjv_def":"temple."},
"H7542":{"lemma":"רַקּוֹן","xlit":"Raqqôwn","pron":"rak-kone'","derivation":"from H7534 (רַק); thinness;","strongs_def":"Rakkon, a place in Palestine","kjv_def":"Rakkon."},
"H7543":{"lemma":"רָקַח","xlit":"râqach","pron":"raw-kakh'","derivation":"a primitive root;","strongs_def":"to perfume","kjv_def":"apothecary, compound, make (ointment), prepare, spice."},
"H7544":{"lemma":"רֶקַח","xlit":"reqach","pron":"reh'-kakh","derivation":"from H7543 (רָקַח);","strongs_def":"properly, perfumery, i.e. (by implication) spicery (for flavor)","kjv_def":"spiced."},
"H7545":{"lemma":"רֹקַח","xlit":"rôqach","pron":"ro'-kakh","derivation":"from H7542 (רַקּוֹן);","strongs_def":"an aromatic","kjv_def":"confection, ointment."},
"H7546":{"lemma":"רַקָּח","xlit":"raqqâch","pron":"rak-kawkh'","derivation":"from H7543 (רָקַח);","strongs_def":"a male perfumer","kjv_def":"apothecary."},
"H7547":{"lemma":"רַקֻּחַ","xlit":"raqquach","pron":"rak-koo'-akh","derivation":"from H7543 (רָקַח);","strongs_def":"a scented substance","kjv_def":"perfume."},
"H7548":{"lemma":"רַקָּחָה","xlit":"raqqâchâh","pron":"rak-kaw-khaw'","derivation":"feminine of H7547 (רַקֻּחַ);","strongs_def":"a female perfumer","kjv_def":"confectioner."},
"H7549":{"lemma":"רָקִיעַ","xlit":"râqîyaʻ","pron":"raw-kee'-ah","derivation":"from H7554 (רָקַע);","strongs_def":"properly, an expanse, i.e. the firmament or (apparently) visible arch of the sky","kjv_def":"firmament."},
"H7550":{"lemma":"רָקִיק","xlit":"râqîyq","pron":"raw-keek'","derivation":"from, H7556 (רָקַק) in its original sense;","strongs_def":"a thin cake","kjv_def":"cake, wafer."},
"H7551":{"lemma":"רָקַם","xlit":"râqam","pron":"raw-kam'","derivation":"a primitive root;","strongs_def":"to variegate color, i.e. embroider; by implication, to fabricate","kjv_def":"embroiderer, needlework, curiously work."},
"H7552":{"lemma":"רֶקֶם","xlit":"Reqem","pron":"reh'-kem","derivation":"from H7551 (רָקַם); versi-color;","strongs_def":"Rekem, the name of a place in Palestine, also of a Midianite and an Israelite","kjv_def":"Rekem."},
"H7553":{"lemma":"רִקְמָה","xlit":"riqmâh","pron":"rik-maw'","derivation":"from H7551 (רָקַם);","strongs_def":"variegation of color; specifically, embroidery","kjv_def":"broidered (work), divers colours, (raiment of) needlework (on both sides)."},
"H7554":{"lemma":"רָקַע","xlit":"râqaʻ","pron":"raw-kah'","derivation":"a primitive root;","strongs_def":"to pound the earth (as a sign of passion); by analogy to expand (by hammering); by implication, to overlay (with thin sheets of metal)","kjv_def":"beat, make broad, spread abroad (forth, over, out, into plates), stamp, stretch."},
"H7555":{"lemma":"רִקֻּעַ","xlit":"riqquaʻ","pron":"rik-koo'-ah","derivation":"from H7554 (רָקַע);","strongs_def":"beaten out, i.e. a (metallic) plate","kjv_def":"broad."},
"H7556":{"lemma":"רָקַק","xlit":"râqaq","pron":"raw-kak'","derivation":"a primitive root;","strongs_def":"to spit","kjv_def":"spit."},
"H7557":{"lemma":"רַקַּת","xlit":"Raqqath","pron":"rak-kath'","derivation":"from H7556 (רָקַק) in its original sense of diffusing; a beach (as expanded shingle);","strongs_def":"Rakkath, a place in Palestine","kjv_def":"Rakkath."},
"H7558":{"lemma":"רִשְׁיוֹן","xlit":"rishyôwn","pron":"rish-yone'","derivation":"from an unused root meaning to have leave;","strongs_def":"a permit","kjv_def":"grant."},
"H7559":{"lemma":"רָשַׁם","xlit":"râsham","pron":"raw-sham'","derivation":"a primitive root;","strongs_def":"to record","kjv_def":"note."},
"H7560":{"lemma":"רְשַׁם","xlit":"rᵉsham","pron":"resh-am'","derivation":"(Aramaic) corresponding to H7559 (רָשַׁם)","strongs_def":"{to record}","kjv_def":"sign, write."},
"H7561":{"lemma":"רָשַׁע","xlit":"râshaʻ","pron":"raw-shah'","derivation":"a primitive root;","strongs_def":"to be (causatively, do or declare) wrong; by implication, to disturb, violate","kjv_def":"condemn, make trouble, vex, be (commit, deal, depart, do) wicked(-ly, -ness)."},
"H7562":{"lemma":"רֶשַׁע","xlit":"reshaʻ","pron":"reh'-shah","derivation":"from H7561 (רָשַׁע);","strongs_def":"a wrong (especially moral)","kjv_def":"iniquity, wicked(-ness)."},
"H7563":{"lemma":"רָשָׁע","xlit":"râshâʻ","pron":"raw-shaw'","derivation":"from H7561 (רָשַׁע);","strongs_def":"morally wrong; concretely, an (actively) bad person","kjv_def":"[phrase] condemned, guilty, ungodly, wicked (man), that did wrong."},
"H7564":{"lemma":"רִשְׁעָה","xlit":"rishʻâh","pron":"rish-aw'","derivation":"feminine of H7562 (רֶשַׁע);","strongs_def":"wrong (especially moral)","kjv_def":"fault, wickedly(-ness)."},
"H7565":{"lemma":"רֶשֶׁף","xlit":"resheph","pron":"reh'-shef","derivation":"from H8313 (שָׂרַף);","strongs_def":"a live coal; by analogy lightning; figuratively, an arrow, (as flashing through the air); specifically, fever","kjv_def":"arrow, (burning) coal, burning heat, [phrase] spark, hot thunderbolt."},
"H7566":{"lemma":"רֶשֶׁף","xlit":"Resheph","pron":"reh'-shef","derivation":"the same as H7565 (רֶשֶׁף);","strongs_def":"Resheph, an Israelite","kjv_def":"Resheph."},
"H7567":{"lemma":"רָשַׁשׁ","xlit":"râshash","pron":"raw-shash'","derivation":"a primitive root;","strongs_def":"to demolish","kjv_def":"impoverish."},
"H7568":{"lemma":"רֶשֶׁת","xlit":"resheth","pron":"reh'-sheth","derivation":"from H3423 (יָרַשׁ);","strongs_def":"a net (as catching animals)","kjv_def":"net(-work)."},
"H7569":{"lemma":"רַתּוֹק","xlit":"rattôwq","pron":"rat-toke'","derivation":"from H7576 (רָתַק);","strongs_def":"a chain","kjv_def":"chain."},
"H7570":{"lemma":"רָתַח","xlit":"râthach","pron":"raw-thakh'","derivation":"a primitive root;","strongs_def":"to boil","kjv_def":"boil."},
"H7571":{"lemma":"רֶתַח","xlit":"rethach","pron":"reh'-thakh","derivation":"from H7570 (רָתַח);","strongs_def":"a boiling","kjv_def":"[idiom] (boil) well."},
"H7572":{"lemma":"רַתִּיקָה","xlit":"rattîyqâh","pron":"rat-tee-kaw'","derivation":"from H7576 (רָתַק);","strongs_def":"a chain","kjv_def":"chain."},
"H7573":{"lemma":"רָתַם","xlit":"râtham","pron":"raw-tham'","derivation":"a primitive root;","strongs_def":"to yoke up (to the pole of a vehicle)","kjv_def":"bind."},
"H7574":{"lemma":"רֶתֶם","xlit":"rethem","pron":"reh'-them","derivation":"or רֹתֶם; from H7573 (רָתַם);","strongs_def":"the Spanish broom (from its pole-like stems)","kjv_def":"juniper (tree)."},
"H7575":{"lemma":"רִתְמָה","xlit":"Rithmâh","pron":"rith-maw'","derivation":"feminine of H7574 (רֶתֶם);","strongs_def":"Rithmah, a place in the Desert","kjv_def":"Rithmah."},
"H7576":{"lemma":"רָתַק","xlit":"râthaq","pron":"raw-thak'","derivation":"a primitive root;","strongs_def":"to fasten","kjv_def":"bind."},
"H7577":{"lemma":"רְתֻקָה","xlit":"rᵉthuqâh","pron":"reth-oo-kaw'","derivation":"feminine passive participle of H7576 (רָתַק);","strongs_def":"something fastened, i.e. a chain","kjv_def":"chain."},
"H7578":{"lemma":"רְתֵת","xlit":"rᵉthêth","pron":"reth-ayth'","derivation":"for H7374 (רֶטֶט);","strongs_def":"terror","kjv_def":"trembling. �"},
"H7579":{"lemma":"שָׁאַב","xlit":"shâʼab","pron":"sahw-ab'","derivation":"a primitive root;","strongs_def":"to bale up water","kjv_def":"(woman to) draw(-er, water)."},
"H7580":{"lemma":"שָׁאַג","xlit":"shâʼag","pron":"shaw-ag'","derivation":"a primitive root;","strongs_def":"to rumble or moan","kjv_def":"[idiom] mightily, roar."},
"H7581":{"lemma":"שְׁאָגָה","xlit":"shᵉʼâgâh","pron":"sheh-aw-gaw'","derivation":"from H7580 (שָׁאַג);","strongs_def":"a rumbling or moan","kjv_def":"roaring."},
"H7582":{"lemma":"שָׁאָה","xlit":"shâʼâh","pron":"shaw-aw'","derivation":"a primitive root;","strongs_def":"to rush; by implication, to desolate","kjv_def":"be desolate, (make a) rush(-ing), (lay) waste."},
"H7583":{"lemma":"שָׁאָה","xlit":"shâʼâh","pron":"shaw-aw'","derivation":"a primitive root (identical with through the idea of whirling to giddiness);","strongs_def":"to stun, i.e. (intransitively) be astonished","kjv_def":"wonder."},
"H7584":{"lemma":"שַׁאֲוָה","xlit":"shaʼăvâh","pron":"shah-av-aw'","derivation":"from H7582 (שָׁאָה);","strongs_def":"a tempest (as rushing)","kjv_def":"desolation."},
"H7585":{"lemma":"שְׁאוֹל","xlit":"shᵉʼôwl","pron":"sheh-ole'","derivation":"or שְׁאֹל; from H7592 (שָׁאַל);","strongs_def":"Hades or the world of the dead (as if a subterranean retreat), including its accessories and inmates","kjv_def":"grave, hell, pit."},
"H7586":{"lemma":"שָׁאוּל","xlit":"Shâʼûwl","pron":"shaw-ool'","derivation":"passive participle of H7592 (שָׁאַל); asked;","strongs_def":"Shaul, the name of an Edomite and two Israelites","kjv_def":"Saul, Shaul."},
"H7587":{"lemma":"שָׁאוּלִי","xlit":"Shâʼûwlîy","pron":"shaw-oo-lee'","derivation":"patronymic from H7856 (שִׂטְנָה);","strongs_def":"a Shaulite or descendant of Shaul","kjv_def":"Shaulites."},
"H7588":{"lemma":"שָׁאוֹן","xlit":"shâʼôwn","pron":"shaw-one'","derivation":"from H7582 (שָׁאָה);","strongs_def":"uproar (as of rushing); by implication, destruction","kjv_def":"[idiom] horrible, noise, pomp, rushing, tumult ([idiom] -uous)."},
"H7589":{"lemma":"שְׁאָט","xlit":"shᵉʼâṭ","pron":"sheh-awt'","derivation":"from an unused root meaning to push aside;","strongs_def":"contempt","kjv_def":"despite(-ful)."},
"H7590":{"lemma":"שָׁאט","xlit":"shâʼṭ","pron":"shawt","derivation":"for active part of H7750 (שׂוּט) (compare H7589 (שְׁאָט));","strongs_def":"one contemning","kjv_def":"that (which) despise(-d)."},
"H7591":{"lemma":"שְׁאִיָּה","xlit":"shᵉʼîyâh","pron":"sheh-ee-yaw'","derivation":"from H7582 (שָׁאָה);","strongs_def":"desolation","kjv_def":"destruction."},
"H7592":{"lemma":"שָׁאַל","xlit":"shâʼal","pron":"shaw-al'","derivation":"or שָׁאֵל; a primitive root;","strongs_def":"to inquire; by implication, to request; by extension, to demand","kjv_def":"ask (counsel, on), beg, borrow, lay to charge, consult, demand, desire, [idiom] earnestly, enquire, [phrase] greet, obtain leave, lend, pray, request, require, [phrase] salute, [idiom] straitly, [idiom] surely, wish."},
"H7593":{"lemma":"שְׁאֵל","xlit":"shᵉʼêl","pron":"sheh-ale'","derivation":"(Aramaic) corresponding to H7592 (שָׁאַל)","strongs_def":"{to inquire; by implication, to request; by extension, to demand}","kjv_def":"ask, demand, require."},
"H7594":{"lemma":"שְׁאָל","xlit":"Shᵉʼâl","pron":"sheh-awl'","derivation":"from H7592 (שָׁאַל); request;","strongs_def":"Sheal, an Israelite","kjv_def":"Sheal."},
"H7595":{"lemma":"שְׁאֵלָא","xlit":"shᵉʼêlâʼ","pron":"sheh-ay-law'","derivation":"(Aramaic) from H7593 (שְׁאֵל);","strongs_def":"properly, a question (at law), i.e. judicial decision or mandate","kjv_def":"demand."},
"H7596":{"lemma":"שְׁאֵלָה","xlit":"shᵉʼêlâh","pron":"sheh-ay-law'","derivation":"or שֵׁלָה; (1 Samuel 1:17), from H7592 (שָׁאַל);","strongs_def":"a petition; by implication, a loan","kjv_def":"loan, petition, request."},
"H7597":{"lemma":"שְׁאַלְתִּיאֵל","xlit":"Shᵉʼaltîyʼêl","pron":"sheh-al-tee-ale'","derivation":"or שַׁלְתִּיאֵל; from H7592 (שָׁאַל) and H410 (אֵל); I have asked God;","strongs_def":"Shealtiel, an Israelite","kjv_def":"Shalthiel, Shealtiel."},
"H7598":{"lemma":"שְׁאַלְתִּיאֵל","xlit":"Shᵉʼaltîyʼêl","pron":"sheh-al-tee-ale'","derivation":"(Aramaic) corresponding to H7597 (שְׁאַלְתִּיאֵל)","strongs_def":"{Shealtiel, an Israelite}","kjv_def":"Shealtiel."},
"H7599":{"lemma":"שָׁאַן","xlit":"shâʼan","pron":"shaw-an'","derivation":"a primitive root;","strongs_def":"to loll, i.e. be peaceful","kjv_def":"be at ease, be quiet, rest. See also H1052 (בֵּית שְׁאָן)."},
"H7600":{"lemma":"שַׁאֲנָן","xlit":"shaʼănân","pron":"shah-an-awn'","derivation":"from H7599 (שָׁאַן);","strongs_def":"secure; in a bad sense, haughty","kjv_def":"that is at ease, quiet, tumult. Compare H7946 (שַׁלְאֲנָן)."},
"H7601":{"lemma":"שָׁאַס","xlit":"shâʼaç","pron":"shaw-as'","derivation":"a primitive root;","strongs_def":"to plunder","kjv_def":"spoil."},
"H7602":{"lemma":"שָׁאַף","xlit":"shâʼaph","pron":"shaw-af'","derivation":"a primitive root;","strongs_def":"to inhale eagerly; figuratively, to cover; by implication, to be angry; also to hasten","kjv_def":"desire (earnestly), devour, haste, pant, snuff up, swallow up."},
"H7603":{"lemma":"שְׂאֹר","xlit":"sᵉʼôr","pron":"seh-ore'","derivation":"from H7604 (שָׁאַר);","strongs_def":"barm or yeast-cake (as swelling by fermentation)","kjv_def":"leaven."},
"H7604":{"lemma":"שָׁאַר","xlit":"shâʼar","pron":"shaw-ar'","derivation":"a primitive root;","strongs_def":"properly, to swell up, i.e. be (causatively, make) redundant","kjv_def":"leave, (be) left, let, remain, remnant, reserve, the rest."},
"H7605":{"lemma":"שְׁאָר","xlit":"shᵉʼâr","pron":"sheh-awr'","derivation":"from H7604 (שָׁאַר);","strongs_def":"a remainder","kjv_def":"[idiom] other, remnant, residue, rest."},
"H7606":{"lemma":"שְׁאָר","xlit":"shᵉʼâr","pron":"sheh-awr'","derivation":"(Aramaic) corresponding to H7605 (שְׁאָר)","strongs_def":"{a remainder}","kjv_def":"[idiom] whatsoever more, residue, rest."},
"H7607":{"lemma":"שְׁאֵר","xlit":"shᵉʼêr","pron":"sheh-ayr'","derivation":"from H7604 (שָׁאַר);","strongs_def":"flesh (as swelling out), as living or forfood; generally food of any kind; figuratively, kindred by blood","kjv_def":"body, flesh, food, (near) kin(-sman, -swoman), near (nigh) (of kin)."},
"H7608":{"lemma":"שַׁאֲרָה","xlit":"shaʼărâh","pron":"shah-ar-aw'","derivation":"feminine of H7607 (שְׁאֵר);","strongs_def":"female kindred by blood","kjv_def":"near kinswomen."},
"H7609":{"lemma":"שֶׁאֱרָה","xlit":"Sheʼĕrâh","pron":"sheh-er-aw'","derivation":"the same as H7608 (שַׁאֲרָה);","strongs_def":"Sheerah, an Israelitess","kjv_def":"Sherah."},
"H7610":{"lemma":"שְׁאָר יָשׁוּב","xlit":"Shᵉʼâr Yâshûwb","pron":"sheh-awr'yaw-shoob'","derivation":"from H7605 (שְׁאָר) and H7725 (שׁוּב); a remnant will return;","strongs_def":"Shear-Jashub, the symbolic name of one of Isaiah's sons","kjv_def":"Shear-jashub."},
"H7611":{"lemma":"שְׁאֵרִית","xlit":"shᵉʼêrîyth","pron":"sheh-ay-reeth'","derivation":"from H7604 (שָׁאַר);","strongs_def":"a remainder or residual (surviving, final) portion","kjv_def":"that had escaped, be left, posterity, remain(-der), remnant, residue, rest."},
"H7612":{"lemma":"שֵׁאת","xlit":"shêʼth","pron":"shayth","derivation":"from H7582 (שָׁאָה);","strongs_def":"devastation","kjv_def":"desolation."},
"H7613":{"lemma":"שְׂאֵת","xlit":"sᵉʼêth","pron":"seh-ayth'","derivation":"from H5375 (נָשָׂא);","strongs_def":"an elevation or leprous scab; figuratively, elation or cheerfulness; exaltation in rank or character","kjv_def":"be accepted, dignity, excellency, highness, raise up self, rising."},
"H7614":{"lemma":"שְׁבָא","xlit":"Shᵉbâʼ","pron":"sheb-aw'","derivation":"of foreign origin;","strongs_def":"Sheba, the name of three early progenitors of tribes and of an Ethiopian district","kjv_def":"Sheba, Sabeans."},
"H7615":{"lemma":"שְׁבָאִי","xlit":"Shᵉbâʼîy","pron":"sheb-aw-ee'","derivation":"patronymic from H7614 (שְׁבָא);","strongs_def":"a Shebaite or descendant of Sheba","kjv_def":"Sabean."},
"H7616":{"lemma":"שָׁבָב","xlit":"shâbâb","pron":"shaw-bawb'","derivation":"from an unused root meaning to break up;","strongs_def":"a fragment, i.e. ruin","kjv_def":"broken in pieces."},
"H7617":{"lemma":"שָׁבָה","xlit":"shâbâh","pron":"shaw-baw'","derivation":"a primitive root;","strongs_def":"to transport into captivity","kjv_def":"(bring away, carry, carry away, lead, lead away, take) captive(-s), drive (take) away."},
"H7618":{"lemma":"שְׁבוּ","xlit":"shᵉbûw","pron":"sheb-oo'","derivation":"from an unused root (probably identical with that of H7617 (שָׁבָה) through the idea of subdivision into flashes or streamers (compare H7632 (שָׁבִיב)) meaning to flame;","strongs_def":"a gem (from its sparkle), probably the agate","kjv_def":"agate."},
"H7619":{"lemma":"שְׁבוּאֵל","xlit":"Shᵉbûwʼêl","pron":"sheb-oo-ale'","derivation":"or שׁוּבָאֵל; from H7617 (שָׁבָה) (abbreviated form) or H7725 (שׁוּב) and H410 (אֵל); captive (or returned) of God;","strongs_def":"Shebuel or Shubael, the name of two Israelites","kjv_def":"Shebuel, Shubael."},
"H7620":{"lemma":"שָׁבוּעַ","xlit":"shâbûwaʻ","pron":"shaw-boo'-ah","derivation":"or שָׁבֻעַ; also (feminine) שְׁבֻעָה; properly, passive participle of H7650 (שָׁבַע) as a denominative of H7651 (שֶׁבַע);","strongs_def":"literally, sevened, i.e. a week (specifically, of years)","kjv_def":"seven, week."},
"H7621":{"lemma":"שְׁבוּעָה","xlit":"shᵉbûwʻâh","pron":"sheb-oo-aw'","derivation":"feminine passive participle of H7650 (שָׁבַע);","strongs_def":"properly, something sworn, i.e. an oath","kjv_def":"curse, oath, [idiom] sworn."},
"H7622":{"lemma":"שְׁבוּת","xlit":"shᵉbûwth","pron":"sheb-ooth'","derivation":"or שְׁבִית; from H7617 (שָׁבָה);","strongs_def":"exile, concretely, prisoners; figuratively, a former state of prosperity","kjv_def":"captive(-ity)."},
"H7623":{"lemma":"שָׁבַח","xlit":"shâbach","pron":"shaw-bakh'","derivation":"a primitive root;","strongs_def":"properly, to address in aloud tone, i.e. (specifically) loud; figuratively, to pacify (as if by words)","kjv_def":"commend, glory, keep in, praise, still, triumph."},
"H7624":{"lemma":"שְׁבַח","xlit":"shᵉbach","pron":"sheb-akh'","derivation":"(Aramaic) corresponding to H7623 (שָׁבַח);","strongs_def":"to adulate, i.e. adore","kjv_def":"praise."},
"H7625":{"lemma":"שְׁבַט","xlit":"shᵉbaṭ","pron":"sheb-at'","derivation":"(Aramaic) corresponding to H7626 (שֵׁבֶט);","strongs_def":"a clan","kjv_def":"tribe."},
"H7626":{"lemma":"שֵׁבֶט","xlit":"shêbeṭ","pron":"shay'-bet","derivation":"from an unused root probably meaning to branch off;","strongs_def":"a scion, i.e. (literally) a stick (for punishing, writing, fighting, ruling, walking, etc.) or (figuratively) a clan","kjv_def":"[idiom] correction, dart, rod, sceptre, staff, tribe."},
"H7627":{"lemma":"שְׁבָט","xlit":"Shᵉbâṭ","pron":"sheb-awt'","derivation":"of foreign origin;","strongs_def":"Shebat, a Jewish month","kjv_def":"Sebat."},
"H7628":{"lemma":"שְׁבִי","xlit":"shᵉbîy","pron":"sheb-ee'","derivation":"from H7618 (שְׁבוּ);","strongs_def":"exiled; captured; as noun, exile (abstractly or concretely and collectively); by extension, booty","kjv_def":"captive(-ity), prisoners, [idiom] take away, that was taken."},
"H7629":{"lemma":"שֹׁבִי","xlit":"Shôbîy","pron":"sho-bee'","derivation":"from H7617 (שָׁבָה); captor;","strongs_def":"Shobi, an Ammonite","kjv_def":"Shobi."},
"H7630":{"lemma":"שֹׁבַי","xlit":"Shôbay","pron":"sho-bah'-ee","derivation":"for H7629 (שֹׁבִי);","strongs_def":"Shobai, an Israelite","kjv_def":"Shobai."},
"H7631":{"lemma":"שְׂבִיב","xlit":"sᵉbîyb","pron":"seb-eeb'","derivation":"(Aramaic) corresponding to H7632 (שָׁבִיב)","strongs_def":"{flame (as split into tongues)}","kjv_def":"flame."},
"H7632":{"lemma":"שָׁבִיב","xlit":"shâbîyb","pron":"shaw-beeb'","derivation":"from the same as H7616 (שָׁבָב);","strongs_def":"flame (as split into tongues)","kjv_def":"spark."},
"H7633":{"lemma":"שִׁבְיָה","xlit":"shibyâh","pron":"shib-yaw'","derivation":"feminine of H7628 (שְׁבִי);","strongs_def":"exile (abstractly or concretely and collectively)","kjv_def":"captives(-ity)."},
"H7634":{"lemma":"שׇׁבְיָה","xlit":"Shobyâh","pron":"shob-yaw'","derivation":"feminine of the same as H7629 (שֹׁבִי); captivation;","strongs_def":"Shobjah, an Israelite","kjv_def":"Shachia (from the margin)."},
"H7635":{"lemma":"שָׁבִיל","xlit":"shâbîyl","pron":"shaw-beel'","derivation":"from the same as H7640 (שֹׁבֶל);","strongs_def":"a track or passage-way (as if flowing along)","kjv_def":"path."},
"H7636":{"lemma":"שָׁבִיס","xlit":"shâbîyç","pron":"shaw-beece'","derivation":"from an unused root meaning to interweave;","strongs_def":"a netting for the hair","kjv_def":"caul."},
"H7637":{"lemma":"שְׁבִיעִי","xlit":"shᵉbîyʻîy","pron":"sheb-ee-ee'","derivation":"or שְׁבִעִי; ordinal from H7657 (שִׁבְעִים);","strongs_def":"seventh","kjv_def":"seventh (time)."},
"H7638":{"lemma":"שָׂבָךְ","xlit":"sâbâk","pron":"saw-bawk'","derivation":"from an unused root meaning to intwine;","strongs_def":"a netting (ornament to the capital of a column)","kjv_def":"net."},
"H7639":{"lemma":"שְׂבָכָה","xlit":"sᵉbâkâh","pron":"seb-aw-kaw'","derivation":"feminine of H7638 (שָׂבָךְ);","strongs_def":"a net-work, i.e. (in hunting) a snare, (in architecture) a ballustrade; also a reticulated ornament to apillar","kjv_def":"checker, lattice, network, snare, wreath(-enwork)."},
"H7640":{"lemma":"שֹׁבֶל","xlit":"shôbel","pron":"show'-bel","derivation":"from an unused root meaning to flow;","strongs_def":"a lady's train (as trailing after her)","kjv_def":"leg."},
"H7641":{"lemma":"שִׁבֹּל","xlit":"shibbôl","pron":"shib-bole","derivation":"or (feminine) שִׁבֹּלֶת; from the same as H7640 (שֹׁבֶל);","strongs_def":"a stream (as flowing); also an ear of grain (as growing out); by analogy, a branch","kjv_def":"branch, channel, ear (of corn), (water-)flood, Shibboleth. Compare H5451 (סִבֹּלֶת)."},
"H7642":{"lemma":"שַׁבְלוּל","xlit":"shablûwl","pron":"shab-lool'","derivation":"from the same as H7640 (שֹׁבֶל);","strongs_def":"a snail (as if floating in its own slime)","kjv_def":"snail."},
"H7643":{"lemma":"שְׂבָם","xlit":"Sᵉbâm","pron":"seb-awm'","derivation":"or (feminine) שִׂבְמָה; probably from H1313 (בָּשָׂם); spice;","strongs_def":"Sebam or Sibmah, a place in Moab","kjv_def":"Shebam, Shibmah, Sibmah."},
"H7644":{"lemma":"שֶׁבְנָא","xlit":"Shebnâʼ","pron":"sheb-naw'","derivation":"or שֶׁבְנָה; from an unused root meaning to grow; growth;","strongs_def":"Shebna or Shebnah, an Israelite","kjv_def":"Shebna, Shebnah."},
"H7645":{"lemma":"שְׁבַנְיָה","xlit":"Shᵉbanyâh","pron":"sheb-an-yaw'","derivation":"or שְׁבַנְיָהוּ; from the same as H7644 (שֶׁבְנָא) and H3050 (יָהּ); Jah has grown (i.e. prospered);","strongs_def":"Shebanjah, the name of three or four Israelites","kjv_def":"Shebaniah."},
"H7646":{"lemma":"שָׂבַע","xlit":"sâbaʻ","pron":"saw-bah'","derivation":"or שָׂבֵעַ; a primitive root;","strongs_def":"to sate, i.e. fill to satisfaction (literally or figuratively)","kjv_def":"have enough, fill (full, self, with), be (to the) full (of), have plenty of, be satiate, satisfy (with), suffice, be weary of."},
"H7647":{"lemma":"שָׂבָע","xlit":"sâbâʻ","pron":"saw-baw'","derivation":"from H7646 (שָׂבַע);","strongs_def":"copiousness","kjv_def":"abundance, plenteous(-ness, -ly)."},
"H7648":{"lemma":"שֹׂבַע","xlit":"sôbaʻ","pron":"so'-bah","derivation":"from H7646 (שָׂבַע);","strongs_def":"satisfaction (of food or (figuratively) joy)","kjv_def":"fill, full(-ness), satisfying, be satisfied."},
"H7649":{"lemma":"שָׂבֵעַ","xlit":"sâbêaʻ","pron":"saw-bay'-ah","derivation":"from H7646 (שָׂבַע);","strongs_def":"satiated (in a pleasant or disagreeable sense)","kjv_def":"full (of), satisfied (with)."},
"H7650":{"lemma":"שָׁבַע","xlit":"shâbaʻ","pron":"shaw-bah'","derivation":"a primitive root; properly to be complete, but used only as a denominative from H7651 (שֶׁבַע);","strongs_def":"to seven oneself, i.e. swear (as if by repeating a declaration seven times)","kjv_def":"adjure, charge (by an oath, with an oath), feed to the full (by mistake for H7646 (שָׂבַע)), take an oath, [idiom] straitly, (cause to, make to) swear."},
"H7651":{"lemma":"שֶׁבַע","xlit":"shebaʻ","pron":"sheh'-bah","derivation":"or (masculine) (שִׁבְעָה); from H7650 (שָׁבַע); a primitive cardinal number;","strongs_def":"seven (as the sacred full one); also (adverbially) seven times; by implication, a week; by extension, an indefinite number","kjv_def":"([phrase] by) seven(-fold),-s, (-teen, -teenth), -th, times). Compare H7658 (שִׁבְעָנָה)."},
"H7652":{"lemma":"שֶׁבַע","xlit":"shebaʻ","pron":"sheh'-bah","derivation":"the same as H7651 (שֶׁבַע); seven;","strongs_def":"Sheba, the name of a place in Palestine, and of two Israelites","kjv_def":"Sheba."},
"H7653":{"lemma":"שִׂבְעָה","xlit":"sibʻâh","pron":"sib-aw'","derivation":"feminine of H7647 (שָׂבָע);","strongs_def":"satiety","kjv_def":"fulness."},
"H7654":{"lemma":"שׇׂבְעָה","xlit":"sobʻâh","pron":"sob-aw'","derivation":"feminine of H7648 (שֹׂבַע);","strongs_def":"satiety","kjv_def":"(to have) enough, [idiom] till...be full, (un-) satiable, satisfy, [idiom] sufficiently."},
"H7655":{"lemma":"שִׁבְעָה","xlit":"shibʻâh","pron":"shib-aw'","derivation":"(Aramaic) corresponding to H7651 (שֶׁבַע)","strongs_def":"{seven (as the sacred full one); also (adverbially) seven times; by implication, a week; by extension, an indefinite number}","kjv_def":"seven (times)."},
"H7656":{"lemma":"שִׁבְעָה","xlit":"Shibʻâh","pron":"shib-aw'","derivation":"masculine of H7651 (שֶׁבַע); seven(-th);","strongs_def":"Shebah, a well in Palestine","kjv_def":"Shebah."},
"H7657":{"lemma":"שִׁבְעִים","xlit":"shibʻîym","pron":"shib-eem'","derivation":"multiple of H7651 (שֶׁבַע);","strongs_def":"seventy","kjv_def":"seventy, threescore and ten ([phrase] -teen)."},
"H7658":{"lemma":"שִׁבְעָנָה","xlit":"shibʻânâh","pron":"shib-aw-naw'","derivation":"prolonged for the masculine of H7651 (שֶׁבַע);","strongs_def":"seven","kjv_def":"seven."},
"H7659":{"lemma":"שִׁבְעָתַיִם","xlit":"shibʻâthayim","pron":"shib-aw-thah'-yim","derivation":"dual (adverbially) of H7651 (שֶׁבַע);","strongs_def":"seven-times","kjv_def":"seven(-fold, times)."},
"H7660":{"lemma":"שָׁבַץ","xlit":"shâbats","pron":"shaw-bats'","derivation":"a primitive root;","strongs_def":"to interweave (colored) threads in squares; by implication (of reticulation) to inchase gems in gold","kjv_def":"embroider, set."},
"H7661":{"lemma":"שָׁבָץ","xlit":"shâbâts","pron":"shaw-bawts'","derivation":"from H7660 (שָׁבַץ);","strongs_def":"entanglement, i.e. (figuratively) perplexity","kjv_def":"anguish."},
"H7662":{"lemma":"שְׁבַק","xlit":"shᵉbaq","pron":"sheb-ak'","derivation":"(Aramaic) corresponding to the root of H7733 (שׁוֹבֵק);","strongs_def":"to quit, i.e. allow to remain","kjv_def":"leave, let alone."},
"H7663":{"lemma":"שָׂבַר","xlit":"sâbar","pron":"saw-bar'","derivation":"erroneously שָׁבַר; (Nehemiah 2:13,15), a primitive root;","strongs_def":"to scrutinize; by implication (of watching) to expect (with hope and patience)","kjv_def":"hope, tarry, view, wait."},
"H7664":{"lemma":"שֵׂבֶר","xlit":"sêber","pron":"say'-ber","derivation":"from H7663 (שָׂבַר);","strongs_def":"expectation","kjv_def":"hope."},
"H7665":{"lemma":"שָׁבַר","xlit":"shâbar","pron":"shaw-bar'","derivation":"a primitive root;","strongs_def":"to burst (literally or figuratively)","kjv_def":"break (down, off, in pieces, up), broken (-hearted), bring to the birth, crush, destroy, hurt, quench, [idiom] quite, tear, view (by mistake for H7663 (שָׂבַר))."},
"H7666":{"lemma":"שָׁבַר","xlit":"shâbar","pron":"shaw-bar'","derivation":"denominative from H7668 (שֶׁבֶר);","strongs_def":"to deal in grain","kjv_def":"buy, sell."},
"H7667":{"lemma":"שֶׁבֶר","xlit":"sheber","pron":"sheh'-ber","derivation":"or שֵׁבֶר; from H7665 (שָׁבַר);","strongs_def":"a fracture, figuratively, ruin; specifically, a solution (of a dream)","kjv_def":"affliction, breach, breaking, broken(-footed, -handed), bruise, crashing, destruction, hurt, interpretation, vexation."},
"H7668":{"lemma":"שֶׁבֶר","xlit":"sheber","pron":"sheh'-ber","derivation":"the same as H7667 (שֶׁבֶר);","strongs_def":"grain (as if broken into kernels)","kjv_def":"corn, victuals."},
"H7669":{"lemma":"שֶׁבֶר","xlit":"Sheber","pron":"sheh'-ber","derivation":"the same as H7667 (שֶׁבֶר);","strongs_def":"Sheber, an Israelite","kjv_def":"Sheber."},
"H7670":{"lemma":"שִׁבְרוֹן","xlit":"shibrôwn","pron":"shib-rone'","derivation":"from H7665 (שָׁבַר);","strongs_def":"rupture, i.e. a pang; figuratively, ruin","kjv_def":"breaking, destruction."},
"H7671":{"lemma":"שְׁבָרִים","xlit":"Shᵉbârîym","pron":"sheb-aw-reem'","derivation":"plural of H7667 (שֶׁבֶר); ruins;","strongs_def":"Shebarim, a place in Palestine","kjv_def":"Shebarim."},
"H7672":{"lemma":"שְׁבַשׁ","xlit":"shᵉbash","pron":"sheb-ash'","derivation":"(Aramaic) corresponding to H7660 (שָׁבַץ);","strongs_def":"to entangle, i.e. perplex","kjv_def":"be astonished."},
"H7673":{"lemma":"שָׁבַת","xlit":"shâbath","pron":"shaw-bath'","derivation":"a primitive root;","strongs_def":"to repose, i.e. desist from exertion; used in many implied relations (causative, figurative or specific)","kjv_def":"(cause to, let, make to) cease, celebrate, cause (make) to fail, keep (sabbath), suffer to be lacking, leave, put away (down), (make to) rest, rid, still, take away."},
"H7674":{"lemma":"שֶׁבֶת","xlit":"shebeth","pron":"sheh'-beth","derivation":"from H7673 (שָׁבַת);","strongs_def":"rest, interruption, cessation","kjv_def":"cease, sit still, loss of time."},
"H7675":{"lemma":"שֶׁבֶת","xlit":"shebeth","pron":"sheh'-beth","derivation":"infinitive of H3427 (יָשַׁב);","strongs_def":"properly, session; but used also concretely, an abode or locality","kjv_def":"place, seat. Compare H3429 (יֹשֵׁב בַּשֶּׁבֶת)."},
"H7676":{"lemma":"שַׁבָּת","xlit":"shabbâth","pron":"shab-bawth'","derivation":"intensive from H7673 (שָׁבַת);","strongs_def":"intermission, i.e (specifically) the Sabbath","kjv_def":"([phrase] every) sabbath."},
"H7677":{"lemma":"שַׁבָּתוֹן","xlit":"shabbâthôwn","pron":"shab-baw-thone'","derivation":"from H7676 (שַׁבָּת);","strongs_def":"a sabbatism or special holiday","kjv_def":"rest, sabbath."},
"H7678":{"lemma":"שַׁבְּתַי","xlit":"Shabbᵉthay","pron":"shab-beth-ah'-ee","derivation":"from H7676 (שַׁבָּת); restful;","strongs_def":"Shabbethai, the name of three Israelites","kjv_def":"Shabbethai."},
"H7679":{"lemma":"שָׂגָא","xlit":"sâgâʼ","pron":"saw-gaw'","derivation":"; a primitive root;","strongs_def":"to grow, i.e. (causatively) to enlarge, (figuratively) laud","kjv_def":"increase, magnify."},
"H7680":{"lemma":"שְׂגָא","xlit":"sᵉgâʼ","pron":"seg-aw'","derivation":"(Aramaic) corresponding to H7679 (שָׂגָא);","strongs_def":"to increase","kjv_def":"grow, be multiplied."},
"H7681":{"lemma":"שָׁגֶא","xlit":"Shâgeʼ","pron":"shaw-gay'","derivation":"probably from H7686 (שָׁגָה); erring;","strongs_def":"Shage, an Israelite","kjv_def":"Shage."},
"H7682":{"lemma":"שָׂגַב","xlit":"sâgab","pron":"saw-gab'","derivation":"a primitive root;","strongs_def":"to be (causatively, make) lofty, especially inaccessible; by implication, safe, strong; used literally and figuratively","kjv_def":"defend, exalt, be excellent, (be, set on) high, lofty, be safe, set up (on high), be too strong."},
"H7683":{"lemma":"שָׁגַג","xlit":"shâgag","pron":"shaw-gag'","derivation":"a primitive root;","strongs_def":"to stray, i.e. (figuratively) sin (with more or less apology)","kjv_def":"[idiom] also for that, deceived, err, go astray, sin ignorantly."},
"H7684":{"lemma":"שְׁגָגָה","xlit":"shᵉgâgâh","pron":"sheg-aw-gaw'","derivation":"from H7683 (שָׁגַג);","strongs_def":"a mistake or inadvertent transgression","kjv_def":"error, ignorance, at unawares; unwittingly."},
"H7685":{"lemma":"שָׂגָה","xlit":"sâgâh","pron":"saw-gaw'","derivation":"a primitive root;","strongs_def":"to enlarge (especially upward, also figuratively)","kjv_def":"grow (up), increase."},
"H7686":{"lemma":"שָׁגָה","xlit":"shâgâh","pron":"shaw-gaw'","derivation":"a primitive root;","strongs_def":"to stray (causatively, mislead), usually (figuratively) to mistake, especially (morally) to transgress; by extension (through the idea of intoxication) to reel, (figuratively) be enraptured","kjv_def":"(cause to) go astray, deceive, err, be ravished, sin through ignorance, (let, make to) wander."},
"H7687":{"lemma":"שְׂגוּב","xlit":"Sᵉgûwb","pron":"seg-oob'","derivation":"from H7682 (שָׂגַב); aloft;","strongs_def":"Segub, the name of two Israelites","kjv_def":"Segub."},
"H7688":{"lemma":"שָׁגַח","xlit":"shâgach","pron":"shaw-gakh'","derivation":"a primitive root;","strongs_def":"to peep, i.e. glance sharply at","kjv_def":"look (narrowly)."},
"H7689":{"lemma":"שַׂגִּיא","xlit":"saggîyʼ","pron":"sag-ghee'","derivation":"from H7679 (שָׂגָא);","strongs_def":"(superlatively) mighty","kjv_def":"excellent, great."},
"H7690":{"lemma":"שַׂגִּיא","xlit":"saggîyʼ","pron":"sag-ghee'","derivation":"(Aramaic) corresponding to H7689 (שַׂגִּיא);","strongs_def":"large (in size, quantity or number, also adverbial)","kjv_def":"exceeding, great(-ly); many, much, sore, very."},
"H7691":{"lemma":"שְׁגִיאָה","xlit":"shᵉgîyʼâh","pron":"sheg-ee-aw'","derivation":"from H7686 (שָׁגָה);","strongs_def":"a moral mistake","kjv_def":"error."},
"H7692":{"lemma":"שִׁגָּיוֹן","xlit":"shiggâyôwn","pron":"shig-gaw-yone'","derivation":"or שִׁגָּיֹנָה; from H7686 (שָׁגָה);","strongs_def":"properly, aberration, i.e. (technically) a dithyramb or rambling poem","kjv_def":"Shiggaion, Shigionoth."},
"H7693":{"lemma":"שָׁגַל","xlit":"shâgal","pron":"shaw-gal'","derivation":"a primitive root;","strongs_def":"to copulate with","kjv_def":"lie with, ravish."},
"H7694":{"lemma":"שֵׁגָל","xlit":"shêgâl","pron":"shay-gawl'","derivation":"from H7693 (שָׁגַל);","strongs_def":"a queen (from cohabitation)","kjv_def":"queen."},
"H7695":{"lemma":"שֵׁגָל","xlit":"shêgâl","pron":"shay-gawl'","derivation":"(Aramaic) corresponding to H7694 (שֵׁגָל);","strongs_def":"a (legitimate) queen","kjv_def":"wife."},
"H7696":{"lemma":"שָׁגַע","xlit":"shâgaʻ","pron":"shaw-gah'","derivation":"a primitive root;","strongs_def":"to rave through insanity","kjv_def":"(be, play the) mad (man)."},
"H7697":{"lemma":"שִׁגָּעוֹן","xlit":"shiggâʻôwn","pron":"shig-gaw-yone'","derivation":"from H7696 (שָׁגַע);","strongs_def":"craziness","kjv_def":"furiously, madness."},
"H7698":{"lemma":"שֶׁגֶר","xlit":"sheger","pron":"sheh'-ger","derivation":"from an unused root probably meaning to eject;","strongs_def":"the fetus (as finally expelled)","kjv_def":"that cometh of, increase."},
"H7699":{"lemma":"שַׁד","xlit":"shad","pron":"shad","derivation":"or שֹׁד; probably from H7736 (שׁוּד) (in its original sense) contracted;","strongs_def":"the breast of a woman or animal (as bulging)","kjv_def":"breast, pap, teat."},
"H7700":{"lemma":"שֵׁד","xlit":"shêd","pron":"shade","derivation":"from H7736 (שׁוּד);","strongs_def":"a doemon (as malignant)","kjv_def":"devil."},
"H7701":{"lemma":"שֹׁד","xlit":"shôd","pron":"shode","derivation":"or שׁוֹד; (Job 5:21), from H7736 (שׁוּד);","strongs_def":"violence, ravage","kjv_def":"desolation, destruction, oppression, robbery, spoil(-ed, -er, -ing), wasting."},
"H7702":{"lemma":"שָׂדַד","xlit":"sâdad","pron":"saw-dad'","derivation":"a primitive root;","strongs_def":"to abrade, i.e. harrow afield","kjv_def":"break clods, harrow."},
"H7703":{"lemma":"שָׁדַד","xlit":"shâdad","pron":"shaw-dad'","derivation":"a primitive root;","strongs_def":"properly, to be burly, i.e. (figuratively) powerful (passively, impregnable); by implication, to ravage","kjv_def":"dead, destroy(-er), oppress, robber, spoil(-er), [idiom] utterly, (lay) waste."},
"H7704":{"lemma":"שָׂדֶה","xlit":"sâdeh","pron":"saw-deh'","derivation":"or שָׂדַי; from an unused root meaning to spread out;","strongs_def":"a field (as flat)","kjv_def":"country, field, ground, land, soil, [idiom] wild."},
"H7705":{"lemma":"שִׁדָּה","xlit":"shiddâh","pron":"shid-dah'","derivation":"from H7703 (שָׁדַד);","strongs_def":"a wife (as mistress of the house)","kjv_def":"[idiom] all sorts, musical instrument."},
"H7706":{"lemma":"שַׁדַּי","xlit":"Shadday","pron":"shad-dah'-ee","derivation":"from H7703 (שָׁדַד);","strongs_def":"the Almighty","kjv_def":"Almighty."},
"H7707":{"lemma":"שְׁדֵיאוּר","xlit":"Shᵉdêyʼûwr","pron":"shed-ay-oor'","derivation":"from the same as H7704 (שָׂדֶה) and H217 (אוּר); spreader of light;","strongs_def":"Shedejur, an Israelite","kjv_def":"Shedeur."},
"H7708":{"lemma":"שִׂדִּים","xlit":"Siddîym","pron":"sid-deem'","derivation":"plural from the same as H7704 (שָׂדֶה); flats;","strongs_def":"Siddim, a valley in Palestine","kjv_def":"Siddim."},
"H7709":{"lemma":"שְׁדֵמָה","xlit":"shᵉdêmâh","pron":"shed-ay-maw'","derivation":"apparently from H7704 (שָׂדֶה);","strongs_def":"a cultivated field","kjv_def":"blasted, field."},
"H7710":{"lemma":"שָׁדַף","xlit":"shâdaph","pron":"shaw-daf'","derivation":"a primitive root;","strongs_def":"to scorch","kjv_def":"blast."},
"H7711":{"lemma":"שְׁדֵפָה","xlit":"shᵉdêphâh","pron":"shed-ay-faw'","derivation":"or שִׁדָּפוֹן; from H7710 (שָׁדַף);","strongs_def":"blight","kjv_def":"blasted(-ing)."},
"H7712":{"lemma":"שְׁדַר","xlit":"shᵉdar","pron":"shed-ar'","derivation":"(Aramaic) a primitive root;","strongs_def":"to endeavor","kjv_def":"labour."},
"H7713":{"lemma":"שְׂדֵרָה","xlit":"sᵉdêrâh","pron":"sed-ay-raw'","derivation":"from an unused root meaning to regulate;","strongs_def":"a row, i.e. rank (of soldiers), story (of rooms)","kjv_def":"board, range."},
"H7714":{"lemma":"שַׁדְרַךְ","xlit":"Shadrak","pron":"shad-rak'","derivation":"probably of foreign origin;","strongs_def":"Shadrak, the Babylonian name of one of Daniel's companions","kjv_def":"Shadrach."},
"H7715":{"lemma":"שַׁדְרַךְ","xlit":"Shadrak","pron":"shad-rak'","derivation":"(Aramaic) the same as H7714 (שַׁדְרַךְ)","strongs_def":"{Shadrak, the Babylonian name of one of Daniel's companions}","kjv_def":"Shadrach."},
"H7716":{"lemma":"שֶׂה","xlit":"seh","pron":"seh","derivation":"or שֵׂי; probably from H7582 (שָׁאָה) through the idea of pushing out to graze;","strongs_def":"a member of a flock, i.e. a sheep or goat","kjv_def":"(lesser, small) cattle, ewe, goat, lamb, sheep. Compare H2089 (זֶה)."},
"H7717":{"lemma":"שָׂהֵד","xlit":"sâhêd","pron":"saw-hade'","derivation":"from an unused root meaning to testify;","strongs_def":"a witness","kjv_def":"record."},
"H7718":{"lemma":"שֹׁהַם","xlit":"shôham","pron":"sho'-ham","derivation":"from an unused root probably mean to blanch;","strongs_def":"a gem, probably the beryl (from its pale green color)","kjv_def":"onyx."},
"H7719":{"lemma":"שֹׁהַם","xlit":"Shôham","pron":"sho'-ham","derivation":"the same as H7718 (שֹׁהַם);","strongs_def":"Shoham, an Israelite","kjv_def":"Shoham."},
"H7720":{"lemma":"שַׂהֲרֹן","xlit":"sahărôn","pron":"sah-har-one'","derivation":"from the same as H5469 (סַהַר);","strongs_def":"a round pendant for the neck","kjv_def":"ornament, round tire like the moon."},
"H7721":{"lemma":"שׂוֹא","xlit":"sôwʼ","pron":"so","derivation":"from an unused root (akin to H5375 (נָשָׂא) and H7722 (שׁוֹא)) meaning to rise;","strongs_def":"a rising","kjv_def":"arise."},
"H7722":{"lemma":"שׁוֹא","xlit":"shôwʼ","pron":"sho","derivation":"or (feminine) שׁוֹאָה; or שֹׁאָה; from an unused root meaning to rush over;","strongs_def":"a tempest; by implication, devastation","kjv_def":"desolate(-ion), destroy, destruction, storm, wasteness."},
"H7723":{"lemma":"שָׁוְא","xlit":"shâvᵉʼ","pron":"shawv","derivation":"or שַׁו; from the same as H7722 (שׁוֹא) in the sense of desolating;","strongs_def":"evil (as destructive), literally (ruin) or morally (especially guile); figuratively idolatry (as false, subjective), uselessness (as deceptive, objective; also adverbially, in vain)","kjv_def":"false(-ly), lie, lying, vain, vanity."},
"H7724":{"lemma":"שְׁוָא","xlit":"Shᵉvâʼ","pron":"shev-aw'","derivation":"from the same as H7723 (שָׁוְא); false;","strongs_def":"Sheva, an Israelite","kjv_def":"Sheva."},
"H7725":{"lemma":"שׁוּב","xlit":"shûwb","pron":"shoob","derivation":"a primitive root;","strongs_def":"to turn back (hence, away) transitively or intransitively, literally or figuratively (not necessarily with the idea of return to the starting point); generally to retreat; often adverbial, again","kjv_def":"((break, build, circumcise, dig, do anything, do evil, feed, lay down, lie down, lodge, make, rejoice, send, take, weep)) [idiom] again, (cause to) answer ([phrase] again), [idiom] in any case (wise), [idiom] at all, averse, bring (again, back, home again), call (to mind), carry again (back), cease, [idiom] certainly, come again (back), [idiom] consider, [phrase] continually, convert, deliver (again), [phrase] deny, draw back, fetch home again, [idiom] fro, get (oneself) (back) again, [idiom] give (again), go again (back, home), (go) out, hinder, let, (see) more, [idiom] needs, be past, [idiom] pay, pervert, pull in again, put (again, up again), recall, recompense, recover, refresh, relieve, render (again), requite, rescue, restore, retrieve, (cause to, make to) return, reverse, reward, [phrase] say nay, send back, set again, slide back, still, [idiom] surely, take back (off), (cause to, make to) turn (again, self again, away, back, back again, backward, from, off), withdraw."},
"H7726":{"lemma":"שׁוֹבָב","xlit":"shôwbâb","pron":"sho-bawb'","derivation":"from H7725 (שׁוּב);","strongs_def":"apostate, i.e. idolatrous","kjv_def":"backsliding, frowardly, turn away (from margin)."},
"H7727":{"lemma":"שׁוֹבָב","xlit":"Shôwbâb","pron":"sho-bawb'","derivation":"the same as H7726 (שׁוֹבָב); rebellious;","strongs_def":"Shobab, the name of two Israelites","kjv_def":"Shobab."},
"H7728":{"lemma":"שׁוֹבֵב","xlit":"shôwbêb","pron":"sho-babe'","derivation":"from H7725 (שׁוּב);","strongs_def":"apostate, i.e. heathenish or (actually) heathen","kjv_def":"backsliding."},
"H7729":{"lemma":"שׁוּבָה","xlit":"shûwbâh","pron":"shoo-baw'","derivation":"from H7725 (שׁוּב);","strongs_def":"a return","kjv_def":"returning."},
"H7730":{"lemma":"שׂוֹבֶךְ","xlit":"sôwbek","pron":"so'-bek","derivation":"for H5441 (סֹבֶךְ);","strongs_def":"a thicket, i.e. interlaced branches","kjv_def":"thick boughs."},
"H7731":{"lemma":"שׁוֹבָךְ","xlit":"Shôwbâk","pron":"sho-bawk'","derivation":"perhaps for H7730 (שׂוֹבֶךְ);","strongs_def":"Shobak, a Syrian","kjv_def":"Shobach."},
"H7732":{"lemma":"שׁוֹבָל","xlit":"Shôwbâl","pron":"sho-bawl'","derivation":"from the same as H7640 (שֹׁבֶל); overflowing;","strongs_def":"Shobal, the name of an Edomite and two Israelites","kjv_def":"Shobal."},
"H7733":{"lemma":"שׁוֹבֵק","xlit":"Shôwbêq","pron":"sho-bake'","derivation":"active participle from a primitive root meaning to leave (compare H7662 (שְׁבַק)); forsaking;","strongs_def":"Shobek, an Israelite","kjv_def":"Shobek."},
"H7734":{"lemma":"שׂוּג","xlit":"sûwg","pron":"soog","derivation":"a primitive root;","strongs_def":"to retreat","kjv_def":"turn back."},
"H7735":{"lemma":"שׂוּג","xlit":"sûwg","pron":"soog","derivation":"a primitive root;","strongs_def":"to hedge in","kjv_def":"make to grow."},
"H7736":{"lemma":"שׁוּד","xlit":"shûwd","pron":"shood","derivation":"a primitive root;","strongs_def":"properly, to swell up, i.e. figuratively (by implication of insolence) to devastate","kjv_def":"waste."},
"H7737":{"lemma":"שָׁוָה","xlit":"shâvâh","pron":"shaw-vaw'","derivation":"a primitive root;","strongs_def":"properly, to level, i.e. equalize; figuratively, to resemble; by implication, to adjust (i.e. counterbalance, be suitable, compose, place, yield, etc.)","kjv_def":"avail, behave, bring forth, compare, countervail, (be, make) equal, lay, be (make, a-) like, make plain, profit, reckon."},
"H7738":{"lemma":"שָׁוָה","xlit":"shâvâh","pron":"shaw-vaw'","derivation":"a primitive root;","strongs_def":"to destroy","kjv_def":"[idiom] substance (from the margin)."},
"H7739":{"lemma":"שְׁוָה","xlit":"shᵉvâh","pron":"shev-aw'","derivation":"(Aramaic); corresponding to H7737 (שָׁוָה);","strongs_def":"to resemble","kjv_def":"make like."},
"H7740":{"lemma":"שָׁוֵה","xlit":"Shâvêh","pron":"shaw-vay'","derivation":"from H7737 (שָׁוָה); plain;","strongs_def":"Shaveh, a place in Palestine","kjv_def":"Shaveh."},
"H7741":{"lemma":"שָׁוֵה קִרְיָתַיִם","xlit":"Shâvêh Qiryâthayim","pron":"shaw-vay' kir-yawthah'-yim","derivation":"from the same as H7740 (שָׁוֵה) and the dual of H7151 (קִרְיָה); plain of a double city;","strongs_def":"Shaveh-Kirjathajim, a place East of the Jordan","kjv_def":"Shaveh Kiriathaim."},
"H7742":{"lemma":"שׂוּחַ","xlit":"sûwach","pron":"soo'-akh","derivation":"a primitive root;","strongs_def":"to muse pensively","kjv_def":"meditate."},
"H7743":{"lemma":"שׁוּחַ","xlit":"shûwach","pron":"shoo'-akh","derivation":"a primitive root;","strongs_def":"to sink, literally or figuratively","kjv_def":"bow down, incline, humble."},
"H7744":{"lemma":"שׁוּחַ","xlit":"Shûwach","pron":"shoo'-akh","derivation":"from H7743 (שׁוּחַ); dell;","strongs_def":"Shuach, a son of Abraham","kjv_def":"Shuah."},
"H7745":{"lemma":"שׁוּחָה","xlit":"shûwchâh","pron":"shoo-khaw'","derivation":"from H7743 (שׁוּחַ);","strongs_def":"a chasm","kjv_def":"ditch, pit."},
"H7746":{"lemma":"שׁוּחָה","xlit":"Shûwchâh","pron":"shoo-khaw'","derivation":"the same as H7745 (שׁוּחָה);","strongs_def":"Shuchah, an Israelite","kjv_def":"Shuah."},
"H7747":{"lemma":"שׁוּחִי","xlit":"Shûwchîy","pron":"shoo-khee'","derivation":"patronymic from H7744 (שׁוּחַ);","strongs_def":"a Shuchite or descendant of Shuach","kjv_def":"Shuhite."},
"H7748":{"lemma":"שׁוּחָם","xlit":"Shûwchâm","pron":"shoo-khawm'","derivation":"from H7743 (שׁוּחַ); humbly;","strongs_def":"Shucham, an Israelite","kjv_def":"Shuham."},
"H7749":{"lemma":"שׁוּחָמִי","xlit":"Shûwchâmîy","pron":"shoo-khaw-mee'","derivation":"patronymic from H7748 (שׁוּחָם);","strongs_def":"a Shuchamite (collectively)","kjv_def":"Shuhamites."},
"H7750":{"lemma":"שׂוּט","xlit":"sûwṭ","pron":"soot","derivation":"or (by permutation) סוּט; a primitive root;","strongs_def":"to detrude, i.e. (intransitively and figuratively) become derelict (wrongly practise; namely, idolatry)","kjv_def":"turn aside to."},
"H7751":{"lemma":"שׁוּט","xlit":"shûwṭ","pron":"shoot","derivation":"a primitive root;","strongs_def":"properly, to push forth; (but used only figuratively) to lash, i.e. (the sea with oars) to row; by implication, to travel","kjv_def":"go (about, through, to and fro), mariner, rower, run to and fro."},
"H7752":{"lemma":"שׁוֹט","xlit":"shôwṭ","pron":"shote","derivation":"from H7751 (שׁוּט);","strongs_def":"a lash (literally or figuratively)","kjv_def":"scourge, whip."},
"H7753":{"lemma":"שׂוּךְ","xlit":"sûwk","pron":"sook","derivation":"a primitive root;","strongs_def":"to entwine, i.e. shut in (for formation, protection or restraint)","kjv_def":"fence, (make an) hedge (up)."},
"H7754":{"lemma":"שׂוֹךְ","xlit":"sôwk","pron":"soke","derivation":"or (feminine) שׂוֹכָה; from H7753 (שׂוּךְ);","strongs_def":"a branch (as interleaved)","kjv_def":"bough."},
"H7755":{"lemma":"שׂוֹכֹה","xlit":"Sôwkôh","pron":"so-ko'","derivation":"or שֹׂכֹה; or שׂוֹכוֹ; from H7753 (שׂוּךְ);","strongs_def":"Sokoh or Soko, the name of two places in Palestine","kjv_def":"Shocho, Shochoh, Sochoh, Soco, Socoh."},
"H7756":{"lemma":"שׂוּכָתִי","xlit":"Sûwkâthîy","pron":"soo-kaw-thee'","derivation":"probably patronymic from a name corresponding to H7754 (שׂוֹךְ) (feminine);","strongs_def":"a Sukathite or descendant of an unknown Israelite named Sukah","kjv_def":"Suchathite."},
"H7757":{"lemma":"שׁוּל","xlit":"shûwl","pron":"shool","derivation":"from an unused root meaning to hang down;","strongs_def":"a skirt; by implication, a bottom edge","kjv_def":"hem, skirt, train."},
"H7758":{"lemma":"שׁוֹלָל","xlit":"shôwlâl","pron":"sho-lawl'","derivation":"or שֵׁילָל; (Micah 1:8), from H7997 (שָׁלַל);","strongs_def":"nude (especially bare-foot); by implication, captive","kjv_def":"spoiled, stripped."},
"H7759":{"lemma":"שׁוּלַמִּית","xlit":"Shûwlammîyth","pron":"shoo-lam-meeth'","derivation":"from H7999 (שָׁלַם);","strongs_def":"peaceful (with the article always prefixed, making it a pet name); the Shulammith, an epithet of Solomon's queen","kjv_def":"Shulamite."},
"H7760":{"lemma":"שׂוּם","xlit":"sûwm","pron":"soom","derivation":"or שִׂים; a primitive root;","strongs_def":"to put (used in a great variety of applications, literal, figurative, inferentially, and elliptically)","kjv_def":"[idiom] any wise, appoint, bring, call (a name), care, cast in, change, charge, commit, consider, convey, determine, [phrase] disguise, dispose, do, get, give, heap up, hold, impute, lay (down, up), leave, look, make (out), mark, [phrase] name, [idiom] on, ordain, order, [phrase] paint, place, preserve, purpose, put (on), [phrase] regard, rehearse, reward, (cause to) set (on, up), shew, [phrase] stedfastly, take, [idiom] tell, [phrase] tread down, (over-)turn, [idiom] wholly, work."},
"H7761":{"lemma":"שׂוּם","xlit":"sûwm","pron":"soom","derivation":"(Aramaic) corresponding to H7760 (שׂוּם)","strongs_def":"{to put (used in a great variety of applications, literal, figurative, inferentially, and elliptically)}","kjv_def":"[phrase] command, give, lay, make, [phrase] name, [phrase] regard, set."},
"H7762":{"lemma":"שׁוּם","xlit":"shûwm","pron":"shoom","derivation":"from an unused root meaning to exhale;","strongs_def":"garlic (from its rank odor)","kjv_def":"garlic."},
"H7763":{"lemma":"שׁוֹמֵר","xlit":"Shôwmêr","pron":"sho-mare'","derivation":"or שֹׁמֵר; active participle of H8104 (שָׁמַר); keeper;","strongs_def":"Shomer, the name of two Israelites","kjv_def":"Shomer."},
"H7764":{"lemma":"שׁוּנִי","xlit":"Shûwnîy","pron":"shoo-nee'","derivation":"from an unused root meaning to rest; quiet;","strongs_def":"Shuni, an Israelite","kjv_def":"Shuni."},
"H7765":{"lemma":"שׁוּנִי","xlit":"Shûwnîy","pron":"shoo-nee'","derivation":"patronymic from H7764 (שׁוּנִי);","strongs_def":"a Shunite (collectively) or descendants of Shuni","kjv_def":"Shunites."},
"H7766":{"lemma":"שׁוּנֵם","xlit":"Shûwnêm","pron":"shoo-name'","derivation":"probably from the same as H7764 (שׁוּנִי); quietly;","strongs_def":"Shunem, a place in Pal","kjv_def":"Shunem."},
"H7767":{"lemma":"שׁוּנַמִּית","xlit":"Shûwnammîyth","pron":"shoo-nam-meeth'","derivation":"patrial from H7766 (שׁוּנֵם);","strongs_def":"a Shunammitess, or female inhabitant of Shunem","kjv_def":"Shunamite."},
"H7768":{"lemma":"שָׁוַע","xlit":"shâvaʻ","pron":"shaw-vah'","derivation":"a primitive root;","strongs_def":"properly, to be free; but used only causatively and reflexively, to halloo (for help, i.e. freedom from some trouble)","kjv_def":"cry (aloud, out), shout."},
"H7769":{"lemma":"שׁוּעַ","xlit":"shûwaʻ","pron":"shoo'-ah","derivation":"from H7768 (שָׁוַע);","strongs_def":"a halloo","kjv_def":"cry, riches."},
"H7770":{"lemma":"שׁוּעַ","xlit":"Shûwaʻ","pron":"shoo'-ah","derivation":"the same as H7769 (שׁוּעַ);","strongs_def":"Shua, a Canaanite","kjv_def":"Shua, Shuah."},
"H7771":{"lemma":"שׁוֹעַ","xlit":"shôwaʻ","pron":"sho'-ah","derivation":"from H7768 (שָׁוַע) in the original sense of freedom;","strongs_def":"a noble, i.e. liberal, opulent; also (as noun in the derived sense) a halloo","kjv_def":"bountiful, crying, rich."},
"H7772":{"lemma":"שׁוֹעַ","xlit":"Shôwaʻ","pron":"sho'-ah","derivation":"the same as H7771 (שׁוֹעַ); rich;","strongs_def":"Shoa, an Oriental people","kjv_def":"Shoa."},
"H7773":{"lemma":"שֶׁוַע","xlit":"shevaʻ","pron":"sheh'-vah","derivation":"from H7768 (שָׁוַע);","strongs_def":"a halloo","kjv_def":"cry."},
"H7774":{"lemma":"שׁוּעָא","xlit":"Shûwʻâʼ","pron":"shoo-aw'","derivation":"from H7768 (שָׁוַע); wealth;","strongs_def":"Shua, an Israelitess","kjv_def":"Shua."},
"H7775":{"lemma":"שַׁוְעָה","xlit":"shavʻâh","pron":"shav-aw'","derivation":"feminine of H7773 (שֶׁוַע);","strongs_def":"a hallooing","kjv_def":"crying."},
"H7776":{"lemma":"שׁוּעָל","xlit":"shûwʻâl","pron":"shoo-awl'","derivation":"or שֻׁעָל; from the same as H8168 (שֹׁעַל);","strongs_def":"a jackal (as a burrower)","kjv_def":"fox."},
"H7777":{"lemma":"שׁוּעָל","xlit":"Shûwʻâl","pron":"shoo-awl'","derivation":"the same as H7776 (שׁוּעָל);","strongs_def":"Shual, the name of an Israelite and of a place in Palestine","kjv_def":"Shual."},
"H7778":{"lemma":"שׁוֹעֵר","xlit":"shôwʻêr","pron":"sho-are'","derivation":"or שֹׁעֵר active participle of H8176 (שָׁעַר) (as denominative from H8179 (שַׁעַר));","strongs_def":"a janitor","kjv_def":"doorkeeper, porter."},
"H7779":{"lemma":"שׁוּף","xlit":"shûwph","pron":"shoof","derivation":"a primitive root;","strongs_def":"properly, to gape, i.e. snap at; figuratively, to overwhelm","kjv_def":"break, bruise, cover."},
"H7780":{"lemma":"שׁוֹפָךְ","xlit":"Shôwphâk","pron":"sho-fawk'","derivation":"from H8210 (שָׁפַךְ); poured;","strongs_def":"Shophak, a Syrian","kjv_def":"Shophach."},
"H7781":{"lemma":"שׁוּפָמִי","xlit":"Shûwphâmîy","pron":"shoo-faw-mee'","derivation":"patronymic from H8197 (שְׁפוּפָם);","strongs_def":"a Shuphamite (collectively) or descendants of Shephupham","kjv_def":"Shuphamite."},
"H7782":{"lemma":"שׁוֹפָר","xlit":"shôwphâr","pron":"sho-far'","derivation":"or שֹׁפָר; from H8231 (שָׁפַר) in the original sense of incising;","strongs_def":"a cornet (as giving a clear sound) or curved horn","kjv_def":"cornet, trumpet."},
"H7783":{"lemma":"שׁוּק","xlit":"shûwq","pron":"shook","derivation":"a primitive root;","strongs_def":"to run after or over, i.e. overflow","kjv_def":"overflow, water."},
"H7784":{"lemma":"שׁוּק","xlit":"shûwq","pron":"shook","derivation":"from H7783 (שׁוּק);","strongs_def":"a street (as run over)","kjv_def":"street."},
"H7785":{"lemma":"שׁוֹק","xlit":"shôwq","pron":"shoke","derivation":"from H7783 (שׁוּק);","strongs_def":"the (lower) leg (as a runner)","kjv_def":"hip, leg, shoulder, thigh."},
"H7786":{"lemma":"שׂוּר","xlit":"sûwr","pron":"soor","derivation":"a primitive root;","strongs_def":"properly, to vanquish; by implication, to rule (causatively, crown)","kjv_def":"make princes, have power, reign. See H5493 (סוּר)."},
"H7787":{"lemma":"שׂוּר","xlit":"sûwr","pron":"soor","derivation":"a primitive root (identical with H7786 (שׂוּר) through the idea of reducing to pieces; compare H4883 (מַשּׂוֹר));","strongs_def":"to saw","kjv_def":"cut."},
"H7788":{"lemma":"שׁוּר","xlit":"shûwr","pron":"shoor","derivation":"a primitive root;","strongs_def":"properly, to turn, i.e. travel about (as a harlot or a merchant)","kjv_def":"go, singular See also H7891 (שִׁיר)."},
"H7789":{"lemma":"שׁוּר","xlit":"shûwr","pron":"shoor","derivation":"a primitive root (identical with through the idea of going round for inspection);","strongs_def":"to spy out, i.e. (generally) survey, (for evil) lurk for, (for good) care for","kjv_def":"behold, lay wait, look, observe, perceive, regard, see."},
"H7790":{"lemma":"שׁוּר","xlit":"shûwr","pron":"shoor","derivation":"from H7889 (שִׁימוֹן);","strongs_def":"a foe (as lying in wait)","kjv_def":"enemy."},
"H7791":{"lemma":"שׁוּר","xlit":"shûwr","pron":"shoor","derivation":"from H7788 (שׁוּר);","strongs_def":"a wall (as going about)","kjv_def":"wall."},
"H7792":{"lemma":"שׁוּר","xlit":"shûwr","pron":"shoor","derivation":"(Aramaic) corresponding to H7791 (שׁוּר)","strongs_def":"{a wall (as going about)}","kjv_def":"wall."},
"H7793":{"lemma":"שׁוּר","xlit":"Shûwr","pron":"shoor","derivation":"the same as H7791 (שׁוּר);","strongs_def":"Shur, a region of the Desert","kjv_def":"Shur."},
"H7794":{"lemma":"שׁוֹר","xlit":"shôwr","pron":"shore","derivation":"from H7788 (שׁוּר);","strongs_def":"a bullock (as a traveller)","kjv_def":"bull(-ock), cow, ox, wall (by mistake for H7791 (שׁוּר))."},
"H7795":{"lemma":"שׂוֹרָה","xlit":"sôwrâh","pron":"so-raw'","derivation":"from H7786 (שׂוּר) in the primitive sense of H5493 (סוּר);","strongs_def":"properly, a ring, i.e. (by analogy) a row (adverbially)","kjv_def":"principal."},
"H7796":{"lemma":"שׂוֹרֵק","xlit":"Sôwrêq","pron":"so-rake'","derivation":"the same as H8321 (שֹׂרֵק); a vine;","strongs_def":"Sorek, a valley in Palestine","kjv_def":"Sorek."},
"H7797":{"lemma":"שׂוּשׂ","xlit":"sûws","pron":"soos","derivation":"or שִׂישׂ; a primitive root;","strongs_def":"to be bright, i.e. cheerful","kjv_def":"be glad, [idiom] greatly, joy, make mirth, rejoice."},
"H7798":{"lemma":"שַׁוְשָׁא","xlit":"Shavshâʼ","pron":"shav-shaw'","derivation":"from H7797 (שׂוּשׂ); joyful;","strongs_def":"Shavsha, an Israelite","kjv_def":"Shavsha."},
"H7799":{"lemma":"שׁוּשַׁן","xlit":"shûwshan","pron":"shoo-shan'","derivation":"or שׁוֹשָׁן; or שֹׁשָׁן; and (feminine) שׁוֹשַׁנָּה; from H7797 (שׂוּשׂ);","strongs_def":"a lily (from its whiteness), as aflower of architectural ornament; also a (straight) trumpet (from the tubular shape)","kjv_def":"lily, Shoshannim."},
"H7800":{"lemma":"שׁוּשַׁן","xlit":"Shûwshan","pron":"shoo-shan'","derivation":"the same as H7799 (שׁוּשַׁן);","strongs_def":"Shushan, a place in Persia","kjv_def":"Shushan."},
"H7801":{"lemma":"שׁוּשַׁנְכִי","xlit":"Shûwshankîy","pron":"shoo-shan-kee'","derivation":"(Aramaic) of foreign origin;","strongs_def":"a Shushankite (collectively) or inhabitants of some unknown place in Assyrian","kjv_def":"Susanchites."},
"H7802":{"lemma":"שׁוּשַׁן עֵדוּת","xlit":"Shûwshan ʻÊdûwth","pron":"shoo-shan' ay-dooth'","derivation":"or (plural of former) שׁוֹשַׁנִּים עֵדוּתlemma שׁוֹשַׁנִּיס עֵדוּת samekh, corrected to שׁוֹשַׁנִּים עֵדוּת; from H7799 (שׁוּשַׁן) and H5715 (עֵדוּת); lily (or trumpet) of assemblage;","strongs_def":"Shushan-Eduth or Shoshannim-Eduth, the title of a popular song","kjv_def":"Shoshannim-Eduth, Shushan-eduth."},
"H7803":{"lemma":"שׁוּתֶלַח","xlit":"Shûwthelach","pron":"shoo-theh'-lakh","derivation":"probably from H7582 (שָׁאָה) and the same as H8520 (תֶּלַח); crash of breakage;","strongs_def":"Shuthelach, the name of two Israelites","kjv_def":"Shuthelah."},
"H7804":{"lemma":"שְׁזַב","xlit":"shᵉzab","pron":"shez-ab'","derivation":"(Aramaic) corresponding to H5800 (עָזַב);","strongs_def":"to leave, i.e. (causatively) free","kjv_def":"deliver."},
"H7805":{"lemma":"שָׁזַף","xlit":"shâzaph","pron":"shaw-zaf'","derivation":"a primitive root;","strongs_def":"to tan (by sunburning); figuratively (as if by a piercing ray) to scan","kjv_def":"look up, see."},
"H7806":{"lemma":"שָׁזַר","xlit":"shâzar","pron":"shaw-zar'","derivation":"a primitive root;","strongs_def":"to twist (a thread of straw)","kjv_def":"twine."},
"H7807":{"lemma":"שַׁח","xlit":"shach","pron":"shakh","derivation":"from H7817 (שָׁחַח);","strongs_def":"sunk, i.e. downcast","kjv_def":"[phrase] humble."},
"H7808":{"lemma":"שֵׂחַ","xlit":"sêach","pron":"say'-akh","derivation":"for H7879 (שִׂיחַ);","strongs_def":"communion, i.e. (reflexively) meditation","kjv_def":"thought."},
"H7809":{"lemma":"שָׁחַד","xlit":"shâchad","pron":"shaw-khad'","derivation":"a primitive root;","strongs_def":"to donate, i.e. bribe","kjv_def":"hire, give a reward."},
"H7810":{"lemma":"שַׁחַד","xlit":"shachad","pron":"shakh'-ad","derivation":"from H7809 (שָׁחַד);","strongs_def":"a donation (venal or redemptive)","kjv_def":"bribe(-ry), gift, present, reward."},
"H7811":{"lemma":"שָׂחָה","xlit":"sâchâh","pron":"saw-khaw'","derivation":"a primitive root;","strongs_def":"to swim; causatively, to inundate","kjv_def":"(make to) swim."},
"H7812":{"lemma":"שָׁחָה","xlit":"shâchâh","pron":"shaw-khaw'","derivation":"a primitive root;","strongs_def":"to depress, i.e. prostrate (especially reflexive, in homage to royalty or God)","kjv_def":"bow (self) down, crouch, fall down (flat), humbly beseech, do (make) obeisance, do reverence, make to stoop, worship."},
"H7813":{"lemma":"שָׂחוּ","xlit":"sâchûw","pron":"saw'-khoo","derivation":"from H7811 (שָׂחָה);","strongs_def":"a pond (for swimming)","kjv_def":"to swim in."},
"H7814":{"lemma":"שְׂחוֹק","xlit":"sᵉchôwq","pron":"sekh-oke'","derivation":"or שְׂחֹק; from H7832 (שָׂחַק);","strongs_def":"laughter (in merriment or defiance)","kjv_def":"derision, laughter(-ed to scorn, -ing), mocked, sport."},
"H7815":{"lemma":"שְׁחוֹר","xlit":"shᵉchôwr","pron":"shekh-ore'","derivation":"from H7835 (שָׁחַר);","strongs_def":"dinginess, i.e. perhaps soot","kjv_def":"coal."},
"H7816":{"lemma":"שְׁחוּת","xlit":"shᵉchûwth","pron":"shekh-ooth'","derivation":"from H7812 (שָׁחָה);","strongs_def":"pit","kjv_def":"pit."},
"H7817":{"lemma":"שָׁחַח","xlit":"shâchach","pron":"shaw-khakh'","derivation":"a primitive root;","strongs_def":"to sink or depress (reflexive or causative)","kjv_def":"bend, bow (down), bring (cast) down, couch, humble self, be (bring) low, stoop."},
"H7818":{"lemma":"שָׂחַט","xlit":"sâchaṭ","pron":"saw-khat'","derivation":"a primitive root;","strongs_def":"to tread out, i.e. squeeze (grapes)","kjv_def":"press."},
"H7819":{"lemma":"שָׁחַט","xlit":"shâchaṭ","pron":"shaw-khat'","derivation":"a primitive root;","strongs_def":"to slaughter (in sacrifice or massacre)","kjv_def":"kill, offer, shoot out, slay, slaughter."},
"H7820":{"lemma":"שָׁחַט","xlit":"shâchaṭ","pron":"shaw-khat'","derivation":"a primitive root (identical with H7819 (שָׁחַט) through the idea of striking);","strongs_def":"to hammer out","kjv_def":"beat."},
"H7821":{"lemma":"שְׁחִיטָה","xlit":"shᵉchîyṭâh","pron":"shekh-ee-taw'","derivation":"from H7819 (שָׁחַט);","strongs_def":"slaughter","kjv_def":"killing."},
"H7822":{"lemma":"שְׁחִין","xlit":"shᵉchîyn","pron":"shekh-een'","derivation":"from an unused root probably meaning to burn;","strongs_def":"inflammation, i.e. an ulcer","kjv_def":"boil, botch."},
"H7823":{"lemma":"שָׁחִיס","xlit":"shâchîyç","pron":"shaw-khece'","derivation":"or סָחִישׁ; from an unused root apparently meaning to sprout;","strongs_def":"aftergrowth","kjv_def":"(that) which springeth of the same."},
"H7824":{"lemma":"שָׁחִיף","xlit":"shâchîyph","pron":"shaw-kheef'","derivation":"from the same as H7828 (שַׁחַף);","strongs_def":"a board (as chipped thin)","kjv_def":"cieled with."},
"H7825":{"lemma":"שְׁחִית","xlit":"shᵉchîyth","pron":"shekh-eeth'","derivation":"from H7812 (שָׁחָה);","strongs_def":"a pit-fall (literally or figuratively)","kjv_def":"destruction, pit."},
"H7826":{"lemma":"שַׁחַל","xlit":"shachal","pron":"shakh'-al","derivation":"from an unused root probably meaning to roar;","strongs_def":"a lion (from his characteristic roar)","kjv_def":"(fierce) lion."},
"H7827":{"lemma":"שְׁחֵלֶת","xlit":"shᵉchêleth","pron":"shekh-ay'-leth","derivation":"apparently from the same as H7826 (שַׁחַל) through some obscure idea, perhaps that of peeling off by concussion of sound;","strongs_def":"a scale or shell, i.e. the aromatic mussel.","kjv_def":"onycha."},
"H7828":{"lemma":"שַׁחַף","xlit":"shachaph","pron":"shakh'-af","derivation":"from an unused root meaning to peel, i.e. emaciate;","strongs_def":"the gull (as thin)","kjv_def":"cuckoo."},
"H7829":{"lemma":"שַׁחֶפֶת","xlit":"shachepheth","pron":"shakh-eh'-feth","derivation":"from the same as H7828 (שַׁחַף);","strongs_def":"emaciation","kjv_def":"consumption."},
"H7830":{"lemma":"שַׁחַץ","xlit":"shachats","pron":"shakh'-ats","derivation":"from an unused root apparently meaning to strut;","strongs_def":"haughtiness (as evinced by the attitude)","kjv_def":"[idiom] lion, pride."},
"H7831":{"lemma":"שַׁחֲצוֹם","xlit":"Shachătsôwm","pron":"shakh-ats-ome'","derivation":"from the same as H7830 (שַׁחַץ); proudly;","strongs_def":"Shachatsom, a place in Palestine","kjv_def":"Shahazimah (from the margin)."},
"H7832":{"lemma":"שָׂחַק","xlit":"sâchaq","pron":"saw-khak'","derivation":"a primitive root;","strongs_def":"to laugh (in pleasure or detraction); by implication, to play","kjv_def":"deride, have in derision, laugh, make merry, mock(-er), play, rejoice, (laugh to) scorn, be in (make) sport."},
"H7833":{"lemma":"שָׁחַק","xlit":"shâchaq","pron":"shaw-khak'","derivation":"a primitive root;","strongs_def":"to comminate (by trituration or attrition)","kjv_def":"beat, wear."},
"H7834":{"lemma":"שַׁחַק","xlit":"shachaq","pron":"shakh'-ak","derivation":"from H7833 (שָׁחַק);","strongs_def":"a powder (as beaten small); by analogy, a thin vapor; by extension, the firmament","kjv_def":"cloud, small dust, heaven, sky."},
"H7835":{"lemma":"שָׁחַר","xlit":"shâchar","pron":"shaw-khar'","derivation":"a primitive root (identical with H7836 (שָׁחַר) through the idea of the duskiness of early dawn);","strongs_def":"to be dim or dark (in color)","kjv_def":"be black."},
"H7836":{"lemma":"שָׁחַר","xlit":"shâchar","pron":"shaw-khar'","derivation":"a primitive root;","strongs_def":"properly, to dawn, i.e. (figuratively) be (up) early at any task (with the implication of earnestness); by extension, to search for (with painstaking)","kjv_def":"(do something) betimes, enquire early, rise (seek) betimes, seek diligently) early, in the morning)."},
"H7837":{"lemma":"שַׁחַר","xlit":"shachar","pron":"shakh'-ar","derivation":"from H7836 (שָׁחַר);","strongs_def":"dawn (literal, figurative or adverbial)","kjv_def":"day(-spring), early, light, morning, whence riseth."},
"H7838":{"lemma":"שָׁחֹר","xlit":"shâchôr","pron":"shaw-khore'","derivation":"or שָׁחוֹר; from H7835 (שָׁחַר);","strongs_def":"properly, dusky, but also (absol.) jetty","kjv_def":"black."},
"H7839":{"lemma":"שַׁחֲרוּת","xlit":"shachărûwth","pron":"shakh-ar-ooth'","derivation":"from H7836 (שָׁחַר);","strongs_def":"a dawning, i.e. (figuratively) juvenescence","kjv_def":"youth."},
"H7840":{"lemma":"שְׁחַרְחֹרֶת","xlit":"shᵉcharchôreth","pron":"shekh-ar-kho'-reth","derivation":"from H7835 (שָׁחַר);","strongs_def":"swarthy","kjv_def":"black."},
"H7841":{"lemma":"שְׁחַרְיָה","xlit":"Shᵉcharyâh","pron":"shekh-ar-yaw'","derivation":"from H7836 (שָׁחַר) and H3050 (יָהּ); Jah has sought;","strongs_def":"Shecharjah, an Israelite","kjv_def":"Shehariah."},
"H7842":{"lemma":"שַׁחֲרַיִם","xlit":"Shachărayim","pron":"shakh-ar-ah'-yim","derivation":"dual of H7837 (שַׁחַר); double dawn;","strongs_def":"Shacharajim, an Israelite","kjv_def":"Shaharaim."},
"H7843":{"lemma":"שָׁחַת","xlit":"shâchath","pron":"shaw-khath'","derivation":"a primitive root;","strongs_def":"to decay, i.e. (causatively) ruin (literally or figuratively)","kjv_def":"batter, cast off, corrupt(-er, thing), destroy(-er, -uction), lose, mar, perish, spill, spoiler, [idiom] utterly, waste(-r)."},
"H7844":{"lemma":"שְׁחַת","xlit":"shᵉchath","pron":"shekh-ath'","derivation":"(Aramaic) corresponding to H7843 (שָׁחַת)","strongs_def":"{to decay, i.e. (causatively) ruin (literally or figuratively)}","kjv_def":"corrupt, fault."},
"H7845":{"lemma":"שַׁחַת","xlit":"shachath","pron":"shakh'-ath","derivation":"from H7743 (שׁוּחַ);","strongs_def":"a pit (especially as a trap); figuratively, destruction","kjv_def":"corruption, destruction, ditch, grave, pit."},
"H7846":{"lemma":"שֵׂט","xlit":"sêṭ","pron":"sayte","derivation":"or סֵט; from H7750 (שׂוּט);","strongs_def":"a departure from right, i.e. sin","kjv_def":"revolter, that turn aside."},
"H7847":{"lemma":"שָׂטָה","xlit":"sâṭâh","pron":"saw-taw'","derivation":"a primitive root;","strongs_def":"to deviate from duty","kjv_def":"decline, go aside, turn."},
"H7848":{"lemma":"שִׁטָּה","xlit":"shiṭṭâh","pron":"shit-taw'","derivation":"feminine of a derivative (only in the plural שִׁטִּים; meaning the sticks of wood) from the same as H7850 (שֹׁטֵט);","strongs_def":"the acacia (from its scourging thorns)","kjv_def":"shittah, shittim. See also H1029 (בֵּית הַשִּׁטָּה)."},
"H7849":{"lemma":"שָׁטַח","xlit":"shâṭach","pron":"shaw-takh'","derivation":"a primitive root;","strongs_def":"to expand","kjv_def":"all abroad, enlarge, spread, stretch out."},
"H7850":{"lemma":"שֹׁטֵט","xlit":"shôṭêṭ","pron":"sho-tate'","derivation":"active participle of an otherwise unused root meaning (properly, to pierce; but only as a denominative from H7752 (שׁוֹט))","strongs_def":"to flog; a goad","kjv_def":"scourge."},
"H7851":{"lemma":"שִׁטִּים","xlit":"Shiṭṭîym","pron":"shit-teem'","derivation":"the same as the plural of H7848 (שִׁטָּה); acacia trees;","strongs_def":"Shittim, a place East of the Jordan","kjv_def":"Shittim."},
"H7852":{"lemma":"שָׂטַם","xlit":"sâṭam","pron":"saw-tam'","derivation":"a primitive root;","strongs_def":"properly, to lurk for, i.e. persecute","kjv_def":"hate, oppose self against."},
"H7853":{"lemma":"שָׂטַן","xlit":"sâṭan","pron":"saw-tan'","derivation":"a primitive root;","strongs_def":"to attack, (figuratively) accuse","kjv_def":"(be an) adversary, resist."},
"H7854":{"lemma":"שָׂטָן","xlit":"sâṭân","pron":"saw-tawn'","derivation":"from H7853 (שָׂטַן);","strongs_def":"an opponent; especially (with the article prefixed) Satan, the arch-enemy of good","kjv_def":"adversary, Satan, withstand."},
"H7855":{"lemma":"שִׂטְנָה","xlit":"siṭnâh","pron":"sit-naw'","derivation":"from H7853 (שָׂטַן);","strongs_def":"opposition (by letter)","kjv_def":"accusation."},
"H7856":{"lemma":"שִׂטְנָה","xlit":"Siṭnâh","pron":"sit-naw'","derivation":"the same as H7855 (שִׂטְנָה);","strongs_def":"Sitnah, the name of a well in Pal","kjv_def":"Sitnah."},
"H7857":{"lemma":"שָׁטַף","xlit":"shâṭaph","pron":"shaw-taf'","derivation":"a primitive root;","strongs_def":"to gush; by implication, to inundate, cleanse; by analogy, to gallop, conquer","kjv_def":"drown, (over-) flow(-whelm, rinse, run, rush, (throughly) wash (away)."},
"H7858":{"lemma":"שֶׁטֶף","xlit":"sheṭeph","pron":"sheh'-tef","derivation":"or שֵׁטֶף; from H7857 (שָׁטַף);","strongs_def":"a deluge (literally or figuratively)","kjv_def":"flood, outrageous, overflowing."},
"H7859":{"lemma":"שְׁטַר","xlit":"shᵉṭar","pron":"shet-ar'","derivation":"(Aramaic) of uncertain derivation;","strongs_def":"a side","kjv_def":"side."},
"H7860":{"lemma":"שֹׁטֵר","xlit":"shôṭêr","pron":"sho-tare'","derivation":"active participle of an otherwise unused root probably meaning to write;","strongs_def":"properly, a scribe, i.e. (by analogy or implication) an official superintendent or magistrate","kjv_def":"officer, overseer, ruler."},
"H7861":{"lemma":"שִׁטְרַי","xlit":"Shiṭray","pron":"shit-rah'-ee","derivation":"from the same as H7860 (שֹׁטֵר); magisterial;","strongs_def":"Shitrai, an Israelite","kjv_def":"Shitrai."},
"H7862":{"lemma":"שַׁי","xlit":"shay","pron":"shah'-ee","derivation":"probably from H7737 (שָׁוָה);","strongs_def":"a gift (as available)","kjv_def":"present."},
"H7863":{"lemma":"שִׂיא","xlit":"sîyʼ","pron":"see","derivation":"from the same as H7721 (שׂוֹא) by permutation;","strongs_def":"elevation","kjv_def":"excellency."},
"H7864":{"lemma":"שְׁיָא","xlit":"Shᵉyâʼ","pron":"sheh-yaw'","derivation":"for H7724 (שְׁוָא);","strongs_def":"Sheja, an Israelite","kjv_def":"Sheva (from the margin)."},
"H7865":{"lemma":"שִׂיאֹן","xlit":"Sîyʼôn","pron":"see-ohn'","derivation":"from H7863 (שִׂיא); peak;","strongs_def":"Sion, the summit of Mount Hermon","kjv_def":"Sion."},
"H7866":{"lemma":"שִׁיאוֹן","xlit":"Shîyʼôwn","pron":"shee-ohn'","derivation":"from the same as H7722 (שׁוֹא); ruin;","strongs_def":"Shijon, a place in Palestine","kjv_def":"Shihon."},
"H7867":{"lemma":"שִׂיב","xlit":"sîyb","pron":"seeb","derivation":"a primitive root;","strongs_def":"properly, to become aged, i.e. (by implication) to grow gray","kjv_def":"(be) grayheaded."},
"H7868":{"lemma":"שִׂיב","xlit":"sîyb","pron":"seeb","derivation":"(Aramaic) corresponding to H7867 (שִׂיב)","strongs_def":"{properly, to become aged, i.e. (by implication) to grow gray}","kjv_def":"elder."},
"H7869":{"lemma":"שֵׂיב","xlit":"sêyb","pron":"sabe","derivation":"from H7867 (שִׂיב);","strongs_def":"old age","kjv_def":"age."},
"H7870":{"lemma":"שִׁיבָה","xlit":"shîybâh","pron":"shee-baw'","derivation":"by permutation from H7725 (שׁוּב);","strongs_def":"a return (of property)","kjv_def":"captivity."},
"H7871":{"lemma":"שִׁיבָה","xlit":"shîybâh","pron":"shee-baw'","derivation":"from H3427 (יָשַׁב);","strongs_def":"residence","kjv_def":"while...lay."},
"H7872":{"lemma":"שֵׂיבָה","xlit":"sêybâh","pron":"say-baw'","derivation":"feminine of H7869 (שֵׂיב);","strongs_def":"old age","kjv_def":"(be) gray (grey hoar,-y) hairs (head,-ed), old age."},
"H7873":{"lemma":"שִׂיג","xlit":"sîyg","pron":"seeg","derivation":"from H7734 (שׂוּג);","strongs_def":"a withdrawal (into a private place)","kjv_def":"pursuing."},
"H7874":{"lemma":"שִׂיד","xlit":"sîyd","pron":"seed","derivation":"a primitive root probably meaning to boil up (compare H7736 (שׁוּד)); used only as denominative from H7875 (שִׂיד);","strongs_def":"to plaster","kjv_def":"plaister."},
"H7875":{"lemma":"שִׂיד","xlit":"sîyd","pron":"seed","derivation":"from H7874 (שִׂיד)lemma שׂיד missing vowel, corrected to שִׂיד;","strongs_def":"lime (as boiling when slacked)","kjv_def":"lime, plaister."},
"H7876":{"lemma":"שָׁיָה","xlit":"shâyâh","pron":"shaw-yaw'","derivation":"a primitive root;","strongs_def":"to keep in memory","kjv_def":"be unmindful. (Render Deuteronomy 32:18, 'A Rock bore thee, thou must recollect; and (yet) thou hast forgotten,' etc.)"},
"H7877":{"lemma":"שִׁיזָא","xlit":"Shîyzâʼ","pron":"shee-zaw'","derivation":"of unknown derivation;","strongs_def":"Shiza, an Israelite","kjv_def":"Shiza."},
"H7878":{"lemma":"שִׂיחַ","xlit":"sîyach","pron":"see'-akh","derivation":"a primitive root;","strongs_def":"to ponder, i.e. (by implication) converse (with oneself, and hence, aloud) or (transitively) utter","kjv_def":"commune, complain, declare, meditate, muse, pray, speak, talk (with)."},
"H7879":{"lemma":"שִׂיחַ","xlit":"sîyach","pron":"see'-akh","derivation":"from H7878 (שִׂיחַ);","strongs_def":"a contemplation; by implication, an utterance","kjv_def":"babbling, communication, complaint, meditation, prayer, talk."},
"H7880":{"lemma":"שִׂיחַ","xlit":"sîyach","pron":"see'-akh","derivation":"from H7878 (שִׂיחַ);","strongs_def":"a shoot (as if uttered or put forth), i.e. (generally) shrubbery","kjv_def":"bush, plant, shrub."},
"H7881":{"lemma":"שִׂיחָה","xlit":"sîychâh","pron":"see-khaw'","derivation":"feminine of H7879 (שִׂיחַ);","strongs_def":"reflection; be extension, devotion","kjv_def":"meditation, prayer."},
"H7882":{"lemma":"שִׁיחָה","xlit":"shîychâh","pron":"shee-khaw'","derivation":"for H7745 (שׁוּחָה);","strongs_def":"a pit-fall","kjv_def":"pit."},
"H7883":{"lemma":"שִׁיחוֹר","xlit":"Shîychôwr","pron":"shee-khore'","derivation":"or שִׁחוֹר; or שִׁחֹר; probably from H7835 (שָׁחַר); dark, i.e. turbid;","strongs_def":"Shichor, a stream of Egypt","kjv_def":"Shihor, Sihor."},
"H7884":{"lemma":"שִׁיחוֹר לִבְנָת","xlit":"Shîychôwr Libnâth","pron":"shee-khore' lib-nawth'","derivation":"from the same as H7883 (שִׁיחוֹר) and H3835 (לָבַן); darkish whiteness;","strongs_def":"Shichor-Libnath, a stream of Palestine","kjv_def":"Shihor-libnath."},
"H7885":{"lemma":"שַׁיִט","xlit":"shayiṭ","pron":"shay'-yit","derivation":"from H7751 (שׁוּט); also compare H7752 (שׁוֹט)","strongs_def":"an oar; a scourge (figuratively)","kjv_def":"oar, scourge."},
"H7886":{"lemma":"שִׁילֹה","xlit":"Shîylôh","pron":"shee-lo'","derivation":"from H7951 (שָׁלָה); tranquil;","strongs_def":"Shiloh, an epithet of the Messiah","kjv_def":"Shiloh."},
"H7887":{"lemma":"שִׁילֹה","xlit":"Shîylôh","pron":"shee-lo'","derivation":"or שִׁלֹה; or שִׁילוֹ; or שִׁלוֹ; from the same as H7886 (שִׁילֹה);","strongs_def":"Shiloh, a place in Palestine","kjv_def":"Shiloh."},
"H7888":{"lemma":"שִׁילוֹנִי","xlit":"Shîylôwnîy","pron":"shee-lo-nee'","derivation":"or שִׁילֹנִי; or שִׁלֹנִי; from H7887 (שִׁילֹה);","strongs_def":"a Shilonite or inhabitant of Shiloh","kjv_def":"Shilonite."},
"H7889":{"lemma":"שִׁימוֹן","xlit":"Shîymôwn","pron":"shee-mone'","derivation":"apparently for H3452 (יְשִׁימוֹן); desert;","strongs_def":"Shimon, an Israelite","kjv_def":"Shimon."},
"H7890":{"lemma":"שַׁיִן","xlit":"shayin","pron":"shah'-yin","derivation":"from an unused root meaning to urinate;","strongs_def":"urine","kjv_def":"piss."},
"H7891":{"lemma":"שִׁיר","xlit":"shîyr","pron":"sheer","derivation":"or (the original form) שׁוּר; (1 Samuel 18:6), a primitive root (rather identical with H7788 (שׁוּר) through the idea of strolling minstrelsy);","strongs_def":"to sing","kjv_def":"behold (by mistake for H7789 (שׁוּר)), sing(-er, -ing man, -ing woman)."},
"H7892":{"lemma":"שִׁיר","xlit":"shîyr","pron":"sheer","derivation":"or feminine שִׁירָה; from H7891 (שִׁיר);","strongs_def":"a song; abstractly, singing","kjv_def":"musical(-ick), [idiom] sing(-er, -ing), song."},
"H7893":{"lemma":"שַׁיִשׁ","xlit":"shayish","pron":"shah'-yish","derivation":"from an unused root meaning to bleach, i.e. whiten;  See H8336 (שֵׁשׁ).","strongs_def":"white, i.e. marble."},
"H7894":{"lemma":"שִׁישָׁא","xlit":"Shîyshâʼ","pron":"shee-shaw'","derivation":"from the same as H7893 (שַׁיִשׁ); whiteness;","strongs_def":"Shisha, an Israelite","kjv_def":"Shisha."},
"H7895":{"lemma":"שִׁישַׁק","xlit":"Shîyshaq","pron":"shee-shak'","derivation":"or שׁוּשַׁק; of Egyptian derivation;","strongs_def":"Shishak, an Egyptian king","kjv_def":"Shishak."},
"H7896":{"lemma":"שִׁית","xlit":"shîyth","pron":"sheeth","derivation":"a primitive root;","strongs_def":"to place (in a very wide application)","kjv_def":"apply, appoint, array, bring, consider, lay (up), let alone, [idiom] look, make, mark, put (on), [phrase] regard, set, shew, be stayed, [idiom] take."},
"H7897":{"lemma":"שִׁית","xlit":"shîyth","pron":"sheeth","derivation":"from H7896 (שִׁית);","strongs_def":"a dress (as put on)","kjv_def":"attire."},
"H7898":{"lemma":"שַׁיִת","xlit":"shayith","pron":"shah'-yith","derivation":"from H7896 (שִׁית);","strongs_def":"scrub or trash, i.e. wild growth of weeds or briers (as if put on the field)","kjv_def":"thorns."},
"H7899":{"lemma":"שֵׂךְ","xlit":"sêk","pron":"sake","derivation":"from H5526 (סָכַךְ) in the sense of H7753 (שׂוּךְ);","strongs_def":"a brier (as of a hedge)","kjv_def":"prick."},
"H7900":{"lemma":"שֹׂךְ","xlit":"sôk","pron":"soke","derivation":"from H5526 (סָכַךְ) in the sense of H7753 (שׂוּךְ);","strongs_def":"a booth (as interlaced)","kjv_def":"tabernacle."},
"H7901":{"lemma":"שָׁכַב","xlit":"shâkab","pron":"shaw-kab'","derivation":"a primitive root;","strongs_def":"to lie down (for rest, sexual connection, decease or any other purpose)","kjv_def":"[idiom] at all, cast down, (lover-)lay (self) (down), (make to) lie (down, down to sleep, still with), lodge, ravish, take rest, sleep, stay."},
"H7902":{"lemma":"שְׁכָבָה","xlit":"shᵉkâbâh","pron":"shek-aw-baw'","derivation":"from H7901 (שָׁכַב);","strongs_def":"a lying down (of dew, or for the sexual act)","kjv_def":"[idiom] carnally, copulation, [idiom] lay, seed."},
"H7903":{"lemma":"שְׁכֹבֶת","xlit":"shᵉkôbeth","pron":"shek-o'-beth","derivation":"from H7901 (שָׁכַב);","strongs_def":"a (sexual) lying with","kjv_def":"[idiom] lie."},
"H7904":{"lemma":"שָׁכָה","xlit":"shâkâh","pron":"shaw-kaw'","derivation":"a primitive root;","strongs_def":"to roam (through lust)","kjv_def":"in the morning (by mistake for H7925 (שָׁכַם))."},
"H7905":{"lemma":"שֻׂכָּה","xlit":"sukkâh","pron":"sook-kaw'","derivation":"feminine of H7900 (שֹׂךְ) in the sense of H7899 (שֵׂךְ);","strongs_def":"a dart (as pointed like a thorn)","kjv_def":"barbed iron."},
"H7906":{"lemma":"שֵׂכוּ","xlit":"Sêkûw","pron":"say'-koo","derivation":"from an unused root apparently meaning to surmount; an observatory (with the article);","strongs_def":"Seku, a place in Palestine","kjv_def":"Sechu."},
"H7907":{"lemma":"שֶׂכְוִי","xlit":"sekvîy","pron":"sek-vee'","derivation":"from the same as H7906 (שֵׂכוּ);","strongs_def":"observant, i.e. (concretely) the mind","kjv_def":"heart."},
"H7908":{"lemma":"שְׁכוֹל","xlit":"shᵉkôwl","pron":"shek-ole'","derivation":"infinitive of H7921 (שָׁכֹל);","strongs_def":"bereavement","kjv_def":"loss of children, spoiling."},
"H7909":{"lemma":"שַׁכּוּל","xlit":"shakkûwl","pron":"shak-kool'","derivation":"or שַׁכֻּל; from H7921 (שָׁכֹל);","strongs_def":"bereaved","kjv_def":"barren, bereaved (robbed) of children (whelps)."},
"H7910":{"lemma":"שִׁכּוֹר","xlit":"shikkôwr","pron":"shik-kore'","derivation":"or שִׁכֹּר; from H7937 (שָׁכַר);","strongs_def":"intoxicated, as astate or a habit","kjv_def":"drunk(-ard, -en, -en man)."},
"H7911":{"lemma":"שָׁכַח","xlit":"shâkach","pron":"shaw-kakh'","derivation":"or שָׁכֵחַ; a primitive root;","strongs_def":"to mislay, i.e. to be oblivious of, from want of memory or attention","kjv_def":"[idiom] at all, (cause to) forget."},
"H7912":{"lemma":"שְׁכַח","xlit":"shᵉkach","pron":"shek-akh'","derivation":"(Aramaic) corresponding to H7911 (שָׁכַח) through the idea of disclosure of a covered or forgotten thing;","strongs_def":"to discover (literally or figuratively)","kjv_def":"find."},
"H7913":{"lemma":"שָׁכֵחַ","xlit":"shâkêach","pron":"shaw-kay'-akh","derivation":"from H7911 (שָׁכַח);","strongs_def":"oblivious","kjv_def":"forget."},
"H7914":{"lemma":"שְׂכִיָּה","xlit":"sᵉkîyâh","pron":"sek-ee-yaw'","derivation":"feminine from the same as H7906 (שֵׂכוּ);","strongs_def":"a conspicuous object","kjv_def":"picture."},
"H7915":{"lemma":"שַׂכִּין","xlit":"sakkîyn","pron":"sak-keen'","derivation":"intensive perhaps from the same as H7906 (שֵׂכוּ) in the sense of H7753 (שׂוּךְ);","strongs_def":"a knife (as pointed or edged)","kjv_def":"knife."},
"H7916":{"lemma":"שָׂכִיר","xlit":"sâkîyr","pron":"saw-keer'","derivation":"from H7936 (שָׂכַר);","strongs_def":"a man at wages  by the day or year","kjv_def":"hired (man, servant), hireling."},
"H7917":{"lemma":"שְׂכִירָה","xlit":"sᵉkîyrâh","pron":"sek-ee-raw'","derivation":"feminine of H7916 (שָׂכִיר);","strongs_def":"a hiring","kjv_def":"that is hired."},
"H7918":{"lemma":"שָׁכַךְ","xlit":"shâkak","pron":"shaw-kak'","derivation":"a primitive root;","strongs_def":"to weave (i.e. lay) a trap; figuratively, (through the idea of secreting) to allay (passions; physically, abate a flood)","kjv_def":"appease, assuage, make to cease, pacify, set."},
"H7919":{"lemma":"שָׂכַל","xlit":"sâkal","pron":"saw-kal'","derivation":"a primitive root;","strongs_def":"to be (causatively, make or act) circumspect and hence, intelligent","kjv_def":"consider, expert, instruct, prosper, (deal) prudent(-ly), (give) skill(-ful), have good success, teach, (have, make to) understand(-ing), wisdom, (be, behave self, consider, make) wise(-ly), guide wittingly."},
"H7920":{"lemma":"שְׂכַל","xlit":"sᵉkal","pron":"sek-al'","derivation":"(Aramaic) corresponding to H7919 (שָׂכַל)","strongs_def":"{to be (causatively, make or act) circumspect and hence, intelligent}","kjv_def":"consider."},
"H7921":{"lemma":"שָׁכֹל","xlit":"shâkôl","pron":"shaw-kole'","derivation":"a primitive root;","strongs_def":"properly, to miscarry, i.e. suffer abortion; by analogy, to bereave (literally or figuratively)","kjv_def":"bereave (of children), barren, cast calf (fruit, young), be (make) childless, deprive, destroy, [idiom] expect, lose children, miscarry, rob of children, spoil."},
"H7922":{"lemma":"שֶׂכֶל","xlit":"sekel","pron":"seh'-kel","derivation":"or שֵׂכֶל; from H7919 (שָׂכַל);","strongs_def":"intelligence; by implication, success","kjv_def":"discretion, knowledge, policy, prudence, sense, understanding, wisdom, wise."},
"H7923":{"lemma":"שִׁכֻּלִים","xlit":"shikkulîym","pron":"shik-koo-leem'","derivation":"plural from H7921 (שָׁכֹל);","strongs_def":"childlessness (by continued bereavements)","kjv_def":"to have after loss of others."},
"H7924":{"lemma":"שׇׂכְלְתָנוּ","xlit":"soklᵉthânûw","pron":"sok-leth-aw-noo'","derivation":"(Aramaic) from H7920 (שְׂכַל);","strongs_def":"intelligence","kjv_def":"understanding."},
"H7925":{"lemma":"שָׁכַם","xlit":"shâkam","pron":"shaw-kam'","derivation":"a primitive root; properly, to incline (the shoulder to a burden); but used only as denominative from H7926 (שְׁכֶם);","strongs_def":"literally, to load up (on the back of man or beast), i.e. to start early in the morning","kjv_def":"(arise, be up, get (oneself) up, rise up) early (betimes), morning."},
"H7926":{"lemma":"שְׁכֶם","xlit":"shᵉkem","pron":"shek-em'","derivation":"from H7925 (שָׁכַם);","strongs_def":"the neck (between the shoulders) as the place of burdens; figuratively, the spur of ahill","kjv_def":"back, [idiom] consent, portion, shoulder."},
"H7927":{"lemma":"שְׁכֶם","xlit":"Shᵉkem","pron":"shek-em'","derivation":"the same as H7926 (שְׁכֶם); ridge;","strongs_def":"Shekem, a place in Palestine","kjv_def":"Shechem."},
"H7928":{"lemma":"שֶׁכֶם","xlit":"Shekem","pron":"sheh'-kem","derivation":"for H7926 (שְׁכֶם);","strongs_def":"Shekem, the name of a Hivite and two Israelites","kjv_def":"Shechem."},
"H7929":{"lemma":"שִׁכְמָה","xlit":"shikmâh","pron":"shik-maw'","derivation":"feminine of H7926 (שְׁכֶם);","strongs_def":"the shoulderbone","kjv_def":"shoulder blade."},
"H7930":{"lemma":"שִׁכְמִי","xlit":"Shikmîy","pron":"shik-mee'","derivation":"patronymic from H7928 (שֶׁכֶם);","strongs_def":"a Shikmite (collectively), or descendants of Shekem","kjv_def":"Shichemites."},
"H7931":{"lemma":"שָׁכַן","xlit":"shâkan","pron":"shaw-kan'","derivation":"a primitive root (apparently akin (by transmission) to H7901 (שָׁכַב) through the idea of lodging; compare H5531 (סִכְלוּת), H7925 (שָׁכַם));","strongs_def":"to reside or permanently stay (literally or figuratively)","kjv_def":"abide, continue, (cause to, make to) dwell(-er), have habitation, inhabit, lay, place, (cause to) remain, rest, set (up)."},
"H7932":{"lemma":"שְׁכַן","xlit":"shᵉkan","pron":"shek-an'","derivation":"(Aramaic) corresponding to H7931 (שָׁכַן)","strongs_def":"{to reside or permanently stay (literally or figuratively)}","kjv_def":"cause to dwell, have habitation."},
"H7933":{"lemma":"שֶׁכֶן","xlit":"sheken","pron":"sheh'-ken","derivation":"from H7931 (שָׁכַן);","strongs_def":"a residence","kjv_def":"habitation."},
"H7934":{"lemma":"שָׁכֵן","xlit":"shâkên","pron":"shaw-kane'","derivation":"from H7931 (שָׁכַן);","strongs_def":"a resident; by extension, a fellow-citizen","kjv_def":"inhabitant, neighbour, nigh."},
"H7935":{"lemma":"שְׁכַנְיָה","xlit":"Shᵉkanyâh","pron":"shek-an-yaw'","derivation":"or (prolonged) שְׁכַנְיָהוּ; from H7931 (שָׁכַן) and H3050 (יָהּ); Jah has dwelt;","strongs_def":"Shekanjah, the name of nine Israelites","kjv_def":"Shecaniah, Shechaniah."},
"H7936":{"lemma":"שָׂכַר","xlit":"sâkar","pron":"saw-kar'","derivation":"or (by permutation) סָכַר; (Ezra 4:5), a primitive root (apparently akin (by prosthesis) to H3739 (כָּרָה) through the idea of temporary purchase; compare H7937 (שָׁכַר));","strongs_def":"to hire","kjv_def":"earn wages, hire (out self), reward, [idiom] surely."},
"H7937":{"lemma":"שָׁכַר","xlit":"shâkar","pron":"shaw-kar'","derivation":"a primitive root;","strongs_def":"to become tipsy; in a qualified sense, to satiate with astimulating drink or (figuratively) influence","kjv_def":"(be filled with) drink (abundantly), (be, make) drunk(-en), be merry. (Superlative of H8248 (שָׁקָה).)"},
"H7938":{"lemma":"שֶׂכֶר","xlit":"seker","pron":"seh'-ker","derivation":"from H7936 (שָׂכַר);","strongs_def":"wages","kjv_def":"reward, sluices."},
"H7939":{"lemma":"שָׂכָר","xlit":"sâkâr","pron":"saw-kawr'","derivation":"from H7936 (שָׂכַר);","strongs_def":"payment of contract; concretely, salary, fare, maintenance; by implication, compensation, benefit","kjv_def":"hire, price, reward(-ed), wages, worth."},
"H7940":{"lemma":"שָׂכָר","xlit":"Sâkâr","pron":"saw-kar'","derivation":"the same as H7939 (שָׂכָר); recompense;","strongs_def":"Sakar, the name of two Israelites","kjv_def":"Sacar."},
"H7941":{"lemma":"שֵׁכָר","xlit":"shêkâr","pron":"shay-kawr'","derivation":"from H7937 (שָׁכַר);","strongs_def":"an intoxicant, i.e. intensely alcoholic liquor","kjv_def":"strong drink, [phrase] drunkard, strong wine."},
"H7942":{"lemma":"שִׁכְּרוֹן","xlit":"Shikkᵉrôwn","pron":"shik-ker-one'","derivation":"for H7943 (שִׁכָּרוֹן);","strongs_def":"drunkenness, Shikkeron, a place in Palestine","kjv_def":"Shicron."},
"H7943":{"lemma":"שִׁכָּרוֹן","xlit":"shikkârôwn","pron":"shik-kaw-rone'","derivation":"from H7937 (שָׁכַר);","strongs_def":"intoxication","kjv_def":"(be) drunken(-ness)."},
"H7944":{"lemma":"שַׁל","xlit":"shal","pron":"shal","derivation":"from H7952 (שָׁלָה) abbreviated form;","strongs_def":"a fault","kjv_def":"error."},
"H7945":{"lemma":"שֶׁל","xlit":"shel","pron":"shel","derivation":"for the rel. H834 (אֲשֶׁר); used with prepositional prefix, and often followed by some pronominal affix;","strongs_def":"on account of, whatsoever, whichsoever","kjv_def":"cause, sake."},
"H7946":{"lemma":"שַׁלְאֲנָן","xlit":"shalʼănân","pron":"shal-an-awn'","derivation":"for H7600 (שַׁאֲנָן);","strongs_def":"tranquil","kjv_def":"being at ease."},
"H7947":{"lemma":"שָׁלַב","xlit":"shâlab","pron":"shaw-lab'","derivation":"a primitive root;","strongs_def":"to space off; intensive (evenly) to make equidistant","kjv_def":"equally distant, set in order."},
"H7948":{"lemma":"שָׁלָב","xlit":"shâlâb","pron":"shaw-lawb'","derivation":"from H7947 (שָׁלַב);","strongs_def":"a spacer or raised interval, i.e. the stile in aframe or panel","kjv_def":"ledge."},
"H7949":{"lemma":"שָׁלַג","xlit":"shâlag","pron":"shaw-lag'","derivation":"a primitive root; properly, meaning to be white; used only as denominative from H7950 (שֶׁלֶג);","strongs_def":"to be snow-white (with the linen clothing of the slain)","kjv_def":"be as snow."},
"H7950":{"lemma":"שֶׁלֶג","xlit":"sheleg","pron":"sheh'-leg","derivation":"from H7949 (שָׁלַג);","strongs_def":"snow (probably from its whiteness)","kjv_def":"snow(-y)."},
"H7951":{"lemma":"שָׁלָה","xlit":"shâlâh","pron":"shaw-law'","derivation":"or שָׁלַו; (Job 3:26), a primitive root;","strongs_def":"to be tranquil, i.e. secure or successful","kjv_def":"be happy, prosper, be in safety."},
"H7952":{"lemma":"שָׁלָה","xlit":"shâlâh","pron":"shaw-law'","derivation":"a primitive root (probably identical with H7953 (שָׁלָה) through the idea of educing);","strongs_def":"to mislead","kjv_def":"deceive, be negligent."},
"H7953":{"lemma":"שָׁלָה","xlit":"shâlâh","pron":"shaw-law'","derivation":"a primitive root (rather cognate (by contraction) to the base of H5394 (נָשַׁל), H7997 (שָׁלַל) and their congeners through the idea of extracting);","strongs_def":"to draw out or off, i.e. remove (the soul by death)","kjv_def":"take away."},
"H7954":{"lemma":"שְׁלָה","xlit":"shᵉlâh","pron":"shel-aw'","derivation":"(Aramaic) corresponding to H7951 (שָׁלָה);","strongs_def":"to be secure","kjv_def":"at rest."},
"H7955":{"lemma":"שָׁלָה","xlit":"shâlâh","pron":"shaw-law'","derivation":"(Aramaic) from a root corresponding to H7952 (שָׁלָה);","strongs_def":"a wrong","kjv_def":"thing amiss."},
"H7956":{"lemma":"שֵׁלָה","xlit":"Shêlâh","pron":"shay-law'","derivation":"the same as H7596 (שְׁאֵלָה) (shortened); request;","strongs_def":"Shelah, the name of a postdiluvian patriarch and of an Israelite","kjv_def":"Shelah."},
"H7957":{"lemma":"שַׁלְהֶבֶת","xlit":"shalhebeth","pron":"shal-heh'-beth","derivation":"from the same as H3851 (לַהַב) with sibilant prefix;","strongs_def":"a flare of fire","kjv_def":"(flaming) flame."},
"H7958":{"lemma":"שְׂלָו","xlit":"sᵉlâv","pron":"sel-awv'","derivation":"or שְׂלָיו; by orthographical variation from H7951 (שָׁלָה) through the idea of sluggishness;","strongs_def":"the quail collectively (as slow in flight from its weight)","kjv_def":"quails."},
"H7959":{"lemma":"שֶׁלֶו","xlit":"shelev","pron":"sheh'-lev","derivation":"from H7951 (שָׁלָה);","strongs_def":"security","kjv_def":"prosperity."},
"H7960":{"lemma":"שָׁלוּ","xlit":"shâlûw","pron":"shaw-loo'","derivation":"(Aramaic) or שָׁלוּת; (Aramaic), from the same as H7955 (שָׁלָה);","strongs_def":"a fault","kjv_def":"error, [idiom] fail, thing amiss."},
"H7961":{"lemma":"שָׁלֵו","xlit":"shâlêv","pron":"shaw-lave'","derivation":"or שָׁלֵיו; feminine שְׁלֵוָה; from H7951 (שָׁלָה);","strongs_def":"tranquil; (in a bad sense) careless; abstractly, security","kjv_def":"(being) at ease, peaceable, (in) prosper(-ity), quiet(-ness), wealthy."},
"H7962":{"lemma":"שַׁלְוָה","xlit":"shalvâh","pron":"shal-vaw'","derivation":"from H7951 (שָׁלָה);","strongs_def":"security (genuine or false)","kjv_def":"abundance, peace(-ably), prosperity, quietness."},
"H7963":{"lemma":"שְׁלֵוָה","xlit":"shᵉlêvâh","pron":"shel-ay-vaw'","derivation":"(Aramaic) corresponding to H7962 (שַׁלְוָה);","strongs_def":"safety","kjv_def":"tranquillity. See also H7961 (שָׁלֵו)."},
"H7964":{"lemma":"שִׁלּוּחַ","xlit":"shillûwach","pron":"shil-loo'-akh","derivation":"or שִׁלֻּחַ; from H7971 (שָׁלַח);","strongs_def":"(only in plural) a dismissal, i.e. (of a wife) divorce (especially the document); also (of a daughter) dower","kjv_def":"presents, have sent back."},
"H7965":{"lemma":"שָׁלוֹם","xlit":"shâlôwm","pron":"shaw-lome'","derivation":"or שָׁלֹם; from H7999 (שָׁלַם);","strongs_def":"safe, i.e. (figuratively) well, happy, friendly; also (abstractly) welfare, i.e. health, prosperity, peace","kjv_def":"[idiom] do, familiar, [idiom] fare, favour, [phrase] friend, [idiom] great, (good) health, ([idiom] perfect, such as be at) peace(-able, -ably), prosper(-ity, -ous), rest, safe(-ty), salute, welfare, ([idiom] all is, be) well, [idiom] wholly."},
"H7966":{"lemma":"שִׁלּוּם","xlit":"shillûwm","pron":"shil-loom'","derivation":"or שִׁלֻּם; from H7999 (שָׁלַם);","strongs_def":"a requital, i.e. (secure) retribution, (venal) a fee","kjv_def":"recompense, reward."},
"H7967":{"lemma":"שַׁלּוּם","xlit":"Shallûwm","pron":"shal-loom'","derivation":"or (shorter) שַׁלֻּם; the same as H7966 (שִׁלּוּם);","strongs_def":"Shallum, the name of fourteen Israelites","kjv_def":"Shallum."},
"H7968":{"lemma":"שַׁלּוּן","xlit":"Shallûwn","pron":"shal-loon'","derivation":"probably for H7967 (שַׁלּוּם);","strongs_def":"Shallun, an Israelite","kjv_def":"Shallum."},
"H7969":{"lemma":"שָׁלוֹשׁ","xlit":"shâlôwsh","pron":"shaw-loshe'","derivation":"or שָׁלֹשׁ; masculine שְׁלוֹשָׁה; or שְׁלֹשָׁה; a primitive number;","strongs_def":"three; occasionally (ordinal) third, or (multiple) thrice","kjv_def":"[phrase] fork, [phrase] often(-times), third, thir(-teen, -teenth), three, [phrase] thrice. Compare H7991 (שָׁלִישׁ)."},
"H7970":{"lemma":"שְׁלוֹשִׁים","xlit":"shᵉlôwshîym","pron":"shel-o-sheem'","derivation":"or שְׁלֹשִׁים; multiple of H7969 (שָׁלוֹשׁ);","strongs_def":"thirty; or (ordinal) thirtieth","kjv_def":"thirty, thirtieth. Compare H7991 (שָׁלִישׁ)."},
"H7971":{"lemma":"שָׁלַח","xlit":"shâlach","pron":"shaw-lakh'","derivation":"a primitive root;","strongs_def":"to send away, for, or out (in a great variety of applications)","kjv_def":"[idiom] any wise, appoint, bring (on the way), cast (away, out), conduct, [idiom] earnestly, forsake, give (up), grow long, lay, leave, let depart (down, go, loose), push away, put (away, forth, in, out), reach forth, send (away, forth, out), set, shoot (forth, out), sow, spread, stretch forth (out)."},
"H7972":{"lemma":"שְׁלַח","xlit":"shᵉlach","pron":"shel-akh'","derivation":"(Aramaic) corresponding to H7971 (שָׁלַח)","strongs_def":"{to send away, for, or out (in a great variety of applications)}","kjv_def":"put, send."},
"H7973":{"lemma":"שֶׁלַח","xlit":"shelach","pron":"sheh'-lakh","derivation":"from H7971 (שָׁלַח);","strongs_def":"a missile of attack, i.e. spear; also (figuratively) a shoot of growth; i.e. branch","kjv_def":"dart, plant, [idiom] put off, sword, weapon."},
"H7974":{"lemma":"שֶׁלַח","xlit":"Shelach","pron":"sheh'-lakh","derivation":"the same as H7973 (שֶׁלַח);","strongs_def":"Shelach, a postdiluvian patriarch","kjv_def":"Salah, Shelah. Compare H7975 (שִׁלֹחַ)lemma שִׁלֹּחַ extra dagesh, corrected to שִׁלֹחַ."},
"H7975":{"lemma":"שִׁלֹחַ","xlit":"Shilôach","pron":"shee-lo'-akh","derivation":"or (in imitation of H7974 (שֶׁלַח)) שֶׁלַח; (Nehemiah 3:15), from H7971 (שָׁלַח); rill;","strongs_def":"Shiloach, a fountain of Jerusalem","kjv_def":"Shiloah, Siloah."},
"H7976":{"lemma":"שִׁלֻּחָה","xlit":"shilluchâh","pron":"shil-loo-khaw'","derivation":"feminine of H7964 (שִׁלּוּחַ);","strongs_def":"a shoot","kjv_def":"branch."},
"H7977":{"lemma":"שִׁלְחִי","xlit":"Shilchîy","pron":"shil-khee'","derivation":"from H7973 (שֶׁלַח); missive, i.e. armed;","strongs_def":"Shilchi, an Israelite","kjv_def":"Shilhi."},
"H7978":{"lemma":"שִׁלְחִים","xlit":"Shilchîym","pron":"shil-kheem'","derivation":"plural of H7973 (שֶׁלַח); javelins or sprouts;","strongs_def":"Shilchim, a place in Palestine","kjv_def":"Shilhim."},
"H7979":{"lemma":"שֻׁלְחָן","xlit":"shulchân","pron":"shool-khawn'","derivation":"from H7971 (שָׁלַח);","strongs_def":"a table (as spread out); by implication, a meal","kjv_def":"table."},
"H7980":{"lemma":"שָׁלַט","xlit":"shâlaṭ","pron":"shaw-lat'","derivation":"a primitive root;","strongs_def":"to dominate, i.e. govern; by implication, to permit","kjv_def":"(bear, have) rule, have dominion, give (have) power."},
"H7981":{"lemma":"שְׁלֵט","xlit":"shᵉlêṭ","pron":"shel-ate'","derivation":"(Aramaic) corresponding to H7980 (שָׁלַט)","strongs_def":"{to dominate, i.e. govern; by implication, to permit}","kjv_def":"have the mastery, have power, bear rule, be (make) ruler."},
"H7982":{"lemma":"שֶׁלֶט","xlit":"sheleṭ","pron":"sheh'-let","derivation":"from H7980 (שָׁלַט);","strongs_def":"probably a shield (as controlling, i.e. protecting the person)","kjv_def":"shield."},
"H7983":{"lemma":"שִׁלְטוֹן","xlit":"shilṭôwn","pron":"shil-tone'","derivation":"(Aramaic) from H7980 (שָׁלַט);","strongs_def":"{power.}","kjv_def":"power."},
"H7984":{"lemma":"שִׁלְטוֹן","xlit":"shilṭôwn","pron":"shil-tone'","derivation":"(Aramaic) or שִׁלְטֹןlemma שִׁלְטן missing vowel, corrected to שִׁלְטֹן; corresponding to H7983 (שִׁלְטוֹן)","strongs_def":"{a potentate}","kjv_def":"ruler."},
"H7985":{"lemma":"שׇׁלְטָן","xlit":"sholṭân","pron":"shol-tawn'","derivation":"(Aramaic) from H7981 (שְׁלֵט);","strongs_def":"empire (abstractly or concretely)","kjv_def":"dominion."},
"H7986":{"lemma":"שַׁלֶּטֶת","xlit":"shalleṭeth","pron":"shal-leh'-teth","derivation":"feminine from H7980 (שָׁלַט);","strongs_def":"a vixen","kjv_def":"imperious."},
"H7987":{"lemma":"שְׁלִי","xlit":"shᵉlîy","pron":"shel-ee'","derivation":"from H7951 (שָׁלָה);","strongs_def":"privacy","kjv_def":"[phrase] quietly."},
"H7988":{"lemma":"שִׁלְיָה","xlit":"shilyâh","pron":"shil-yaw'","derivation":"feminine from H7953 (שָׁלָה);","strongs_def":"a fetus or babe (as extruded in birth)","kjv_def":"young one."},
"H7989":{"lemma":"שַׁלִּיט","xlit":"shallîyṭ","pron":"shal-leet'","derivation":"from H7980 (שָׁלַט);","strongs_def":"potent; concretely, a prince or warrior","kjv_def":"governor, mighty, that hath power, ruler."},
"H7990":{"lemma":"שַׁלִּיט","xlit":"shallîyṭ","pron":"shal-leet'","derivation":"(Aramaic) corresponding to H7989 (שַׁלִּיט);","strongs_def":"mighty; abstractly, permission; concretely, a premier","kjv_def":"captain, be lawful, rule(-r)."},
"H7991":{"lemma":"שָׁלִישׁ","xlit":"shâlîysh","pron":"shaw-leesh'","derivation":"or שָׁלוֹשׁ; (1 Chronicles 11:11; 1 Chronicles 12:18), or שָׁלֹשׁ; (2 Samuel 23:13), from H7969 (שָׁלוֹשׁ);","strongs_def":"a triple, i.e. (as a musical instrument) a triangle (or perhaps rather threestringed lute); also (as an indefinite, great quantity) a three-fold measure (perhaps a treble ephah); also (as an officer) a general of the third rank (upward, i.e. the highest)","kjv_def":"captain, instrument of musick, (great) lord, (great) measure, prince, three (from the margin)."},
"H7992":{"lemma":"שְׁלִישִׁי","xlit":"shᵉlîyshîy","pron":"shel-ee-shee'","derivation":"ordinal from H7969 (שָׁלוֹשׁ);","strongs_def":"third; feminine athird (part); by extension, a third (day, year or time); specifically, a third-story cell)","kjv_def":"third (part, rank, time), three (years old)."},
"H7993":{"lemma":"שָׁלַךְ","xlit":"shâlak","pron":"shaw-lak","derivation":"a primitive root;","strongs_def":"to throw out, down or away (literally or figuratively)","kjv_def":"adventure, cast (away, down, forth, off, out), hurl, pluck, throw."},
"H7994":{"lemma":"שָׁלָךְ","xlit":"shâlâk","pron":"shaw-lawk'","derivation":"from H7993 (שָׁלַךְ);","strongs_def":"bird of prey, usually thought to be the pelican (from casting itself into the sea)","kjv_def":"cormorant."},
"H7995":{"lemma":"שַׁלֶּכֶת","xlit":"shalleketh","pron":"shal-leh'-keth","derivation":"from H7993 (שָׁלַךְ);","strongs_def":"a felling (of trees)","kjv_def":"when cast."},
"H7996":{"lemma":"שַׁלֶּכֶת","xlit":"Shalleketh","pron":"shal-leh'-keth","derivation":"the same as H7995 (שַׁלֶּכֶת);","strongs_def":"Shalleketh, a gate in Jerusalem","kjv_def":"Shalleketh."},
"H7997":{"lemma":"שָׁלַל","xlit":"shâlal","pron":"shaw-lal'","derivation":"a primitive root;","strongs_def":"to drop or strip; by implication, to plunder","kjv_def":"let fall, make self a prey, [idiom] of purpose, (make a, (take)) spoil."},
"H7998":{"lemma":"שָׁלָל","xlit":"shâlâl","pron":"shaw-lawl'","derivation":"from H7997 (שָׁלַל);","strongs_def":"booty","kjv_def":"prey, spoil."},
"H7999":{"lemma":"שָׁלַם","xlit":"shâlam","pron":"shaw-lam'","derivation":"a primitive root;","strongs_def":"to be safe (in mind, body or estate); figuratively, to be (causatively, make) completed; by implication, to be friendly; by extension, to reciprocate (in various applications)","kjv_def":"make amends, (make an) end, finish, full, give again, make good, (re-) pay (again), (make) (to) (be at) peace(-able), that is perfect, perform, (make) prosper(-ous), recompense, render, requite, make restitution, restore, reward, [idiom] surely."},
"H8000":{"lemma":"שְׁלַם","xlit":"shᵉlam","pron":"shel-am'","derivation":"(Aramaic) corresponding to H7999 (שָׁלַם);","strongs_def":"to complete, to restore","kjv_def":"deliver, finish."},
"H8001":{"lemma":"שְׁלָם","xlit":"shᵉlâm","pron":"shel-awm'","derivation":"(Aramaic) corresponding to H7965 (שָׁלוֹם);","strongs_def":"prosperity","kjv_def":"peace."},
"H8002":{"lemma":"שֶׁלֶם","xlit":"shelem","pron":"sheh'-lem","derivation":"from H7999 (שָׁלַם);","strongs_def":"properly, requital, i.e. a (voluntary) sacrifice in thanks","kjv_def":"peace offering."},
"H8003":{"lemma":"שָׁלֵם","xlit":"shâlêm","pron":"shaw-lame'","derivation":"from H7999 (שָׁלַם);","strongs_def":"complete (literally or figuratively); especially friendly","kjv_def":"full, just, made ready, peaceable, perfect(-ed), quiet, Shalem (by mistake for a name), whole."},
"H8004":{"lemma":"שָׁלֵם","xlit":"Shâlêm","pron":"shaw-lame'","derivation":"the same as H8003 (שָׁלֵם); peaceful;","strongs_def":"Shalem, an early name of Jerusalem","kjv_def":"Salem."},
"H8005":{"lemma":"שִׁלֵּם","xlit":"shillêm","pron":"shil-lame'","derivation":"from H7999 (שָׁלַם);","strongs_def":"requital","kjv_def":"recompense."},
"H8006":{"lemma":"שִׁלֵּם","xlit":"Shillêm","pron":"shil-lame'","derivation":"the same as H8005 (שִׁלֵּם);","strongs_def":"Shillem, an Israelite","kjv_def":"Shillem."},
"H8007":{"lemma":"שַׂלְמָא","xlit":"Salmâʼ","pron":"sal-maw'","derivation":"probably for H8008 (שַׂלְמָה); clothing;","strongs_def":"Salma, the name of two Israelites","kjv_def":"Salma."},
"H8008":{"lemma":"שַׂלְמָה","xlit":"salmâh","pron":"sal-maw'","derivation":"transp. for H8071 (שִׂמְלָה);","strongs_def":"a dress","kjv_def":"clothes, garment, raiment."},
"H8009":{"lemma":"שַׂלְמָה","xlit":"Salmâh","pron":"sal-maw'","derivation":"the same as H8008 (שַׂלְמָה); clothing;","strongs_def":"Salmah, an Israelite","kjv_def":"Salmon. Compare H8012 (שַׂלְמוֹן)."},
"H8010":{"lemma":"שְׁלֹמֹה","xlit":"Shᵉlômôh","pron":"shel-o-mo'","derivation":"from H7965 (שָׁלוֹם); peaceful;","strongs_def":"Shelomah, David's successor","kjv_def":"Solomon."},
"H8011":{"lemma":"שִׁלֻּמָה","xlit":"shillumâh","pron":"shil-loo-maw'","derivation":"feminine of H7966 (שִׁלּוּם);","strongs_def":"retribution","kjv_def":"recompense."},
"H8012":{"lemma":"שַׂלְמוֹן","xlit":"Salmôwn","pron":"sal-mone'","derivation":"from H8008 (שַׂלְמָה); investiture;","strongs_def":"Salmon, an Israelite","kjv_def":"Salmon. Compare H8009 (שַׂלְמָה)."},
"H8013":{"lemma":"שְׁלֹמוֹת","xlit":"Shᵉlômôwth","pron":"shel-o-moth'","derivation":"feminine plural of H7965 (שָׁלוֹם); pacifications;","strongs_def":"Shelomoth, the name of two Israelites","kjv_def":"Shelomith (from the margin), Shelomoth. Compare H8019 (שְׁלֹמִית)."},
"H8014":{"lemma":"שַׂלְמַי","xlit":"Salmay","pron":"sal-mah'-ee","derivation":"from H8008 (שַׂלְמָה); clothed;","strongs_def":"Salmai, an Israelite","kjv_def":"Shalmai."},
"H8015":{"lemma":"שְׁלֹמִי","xlit":"Shᵉlômîy","pron":"shel-o-mee'","derivation":"from H7965 (שָׁלוֹם); peaceable;","strongs_def":"Shelomi, an Israelite","kjv_def":"Shelomi."},
"H8016":{"lemma":"שִׁלֵּמִי","xlit":"Shillêmîy","pron":"shil-lay-mee'","derivation":"patronymically from H8006 (שִׁלֵּם);","strongs_def":"a Shilemite (collectively) or descendants of Shillem","kjv_def":"Shillemites."},
"H8017":{"lemma":"שְׁלֻמִיאֵל","xlit":"Shᵉlumîyʼêl","pron":"shel-oo-mee-ale'","derivation":"from H7965 (שָׁלוֹם) and H410 (אֵל); peace of God;","strongs_def":"Shelumiel, an Israelite","kjv_def":"Shelumiel."},
"H8018":{"lemma":"שֶׁלֶמְיָה","xlit":"Shelemyâh","pron":"shel-em-yaw'","derivation":"or שֶׁלֶמְיָהוּ; from H8002 (שֶׁלֶם) and H3050 (יָהּ); thankoffering of Jah;","strongs_def":"Shelemjah, the name of nine Israelites","kjv_def":"Shelemiah."},
"H8019":{"lemma":"שְׁלֹמִית","xlit":"Shᵉlômîyth","pron":"shel-o-meeth'","derivation":"or שׁלוֹמִית; (Ezra 8:10), from H7965 (שָׁלוֹם); peaceableness;","strongs_def":"Shelomith, the name of five Israelites and three Israelitesses","kjv_def":"Shelomith."},
"H8020":{"lemma":"שַׁלְמַן","xlit":"Shalman","pron":"shal-man'","derivation":"of foreign derivation;","strongs_def":"Shalman, a king apparently of Assyria","kjv_def":"Shalman. Compare H8022 (שַׁלְמַנְאֶסֶר)."},
"H8021":{"lemma":"שַׁלְמֹן","xlit":"shalmôn","pron":"shal-mone'","derivation":"from H7999 (שָׁלַם);","strongs_def":"a bribe","kjv_def":"reward."},
"H8022":{"lemma":"שַׁלְמַנְאֶסֶר","xlit":"Shalmanʼeçer","pron":"shal-man-eh'-ser","derivation":"of foreign derivation;","strongs_def":"Shalmaneser, an Assyrian king","kjv_def":"Shalmaneser. Comp H8020 (שַׁלְמַן)."},
"H8023":{"lemma":"שִׁלֹנִי","xlit":"Shilônîy","pron":"shee-lo-nee'","derivation":"the same as H7888 (שִׁילוֹנִי);","strongs_def":"Shiloni, an Israelite","kjv_def":"Shiloni."},
"H8024":{"lemma":"שֵׁלָנִי","xlit":"Shêlânîy","pron":"shay-law-nee'","derivation":"from H7956 (שֵׁלָה);","strongs_def":"a Shelanite (collectively), or descendants of Shelah","kjv_def":"Shelanites."},
"H8025":{"lemma":"שָׁלַף","xlit":"shâlaph","pron":"saw-laf'","derivation":"a primitive root;","strongs_def":"to pull out, up or off","kjv_def":"draw (off), grow up, pluck off."},
"H8026":{"lemma":"שֶׁלֶף","xlit":"sheleph","pron":"sheh'-lef","derivation":"from H8025 (שָׁלַף); extract;","strongs_def":"Sheleph, a son of Jokthan","kjv_def":"Sheleph."},
"H8027":{"lemma":"שָׁלַשׁ","xlit":"shâlash","pron":"shaw-lash'","derivation":"a primitive root perhaps originally to intensify, i.e. treble; but apparently used only as denominative from H7969 (שָׁלוֹשׁ),","strongs_def":"to be (causatively, make) triplicate (by restoration, in portions, strands, days or years)","kjv_def":"do the third time, (divide into, stay) three (days, -fold, parts, years old)."},
"H8028":{"lemma":"שֶׁלֶשׁ","xlit":"Shelesh","pron":"sheh'-lesh","derivation":"from H8027 (שָׁלַשׁ); triplet;","strongs_def":"Shelesh, an Israelite","kjv_def":"Shelesh."},
"H8029":{"lemma":"שִׁלֵּשׁ","xlit":"shillêsh","pron":"shil-laysh'","derivation":"from H8027 (שָׁלַשׁ);","strongs_def":"a descendant of the third degree, i.e. great grandchild","kjv_def":"third (generation)."},
"H8030":{"lemma":"שִׁלְשָׁה","xlit":"Shilshâh","pron":"shil-shaw'","derivation":"feminine from the same as H8028 (שֶׁלֶשׁ); triplication;","strongs_def":"Shilshah, an Israelite","kjv_def":"Shilshah."},
"H8031":{"lemma":"שָׁלִשָׁה","xlit":"Shâlishâh","pron":"shaw-lee-shaw'","derivation":"feminine from H8027 (שָׁלַשׁ); trebled land;","strongs_def":"Shalishah, a place in Palestine","kjv_def":"Shalisha."},
"H8032":{"lemma":"שִׁלְשׁוֹם","xlit":"shilshôwm","pron":"shil-shome'","derivation":"or שִׁלְשֹׁם; from the same as H8028 (שֶׁלֶשׁ);","strongs_def":"trebly, i.e. (in time) day before yesterday","kjv_def":"[phrase] before (that time, -time), excellent things (from the margin), [phrase] heretofore, three days, [phrase] time past."},
"H8033":{"lemma":"שָׁם","xlit":"shâm","pron":"shawm","derivation":"a primitive particle (rather from the relative pronoun, H834 (אֲשֶׁר));","strongs_def":"there (transferring to time) then; often thither, or thence","kjv_def":"in it, [phrase] thence, there (-in, [phrase] of, [phrase] out), [phrase] thither, [phrase] whither."},
"H8034":{"lemma":"שֵׁם","xlit":"shêm","pron":"shame","derivation":"a primitive word (perhaps rather from through the idea of definite and conspicuous position; compare H8064 (שָׁמַיִם));","strongs_def":"an appellation, as amark or memorial of individuality; by implication honor, authority, character","kjv_def":"[phrase] base, (in-) fame(-ous), named(-d), renown, report."},
"H8035":{"lemma":"שֵׁם","xlit":"Shêm","pron":"shame","derivation":"the same as H8034 (שֵׁם); name;","strongs_def":"Shem, a son of Noah (often includ. his posterity)","kjv_def":"Sem, Shem."},
"H8036":{"lemma":"שֻׁם","xlit":"shum","pron":"shoom","derivation":"(Aramaic) corresponding to H8034 (שֵׁם)","strongs_def":"{an appellation, as amark or memorial of individuality; by implication honor, authority, character}","kjv_def":"name."},
"H8037":{"lemma":"שַׁמָּא","xlit":"Shammâʼ","pron":"sham-maw'","derivation":"from H8074 (שָׁמֵם); desolation;","strongs_def":"Shamma, an Israelite","kjv_def":"Shamma."},
"H8038":{"lemma":"שֶׁמְאֵבֶר","xlit":"Shemʼêber","pron":"shem-ay'-ber","derivation":"apparently from H8034 (שֵׁם) and H83 (אֵבֶר); name of pinion, i.e. illustrious;","strongs_def":"Shemeber, a king of Zeboim","kjv_def":"Shemeber."},
"H8039":{"lemma":"שִׁמְאָה","xlit":"Shimʼâh","pron":"shim-aw'","derivation":"perhaps for H8093 (שִׁמְעָה);","strongs_def":"Shimah, an Israelite","kjv_def":"Shimah. Compare H8043 (שִׁמְאָם)."},
"H8040":{"lemma":"שְׂמֹאול","xlit":"sᵉmôʼwl","pron":"sem-ole'","derivation":"or שְׂמֹאל; a primitive word (rather perhaps from the same as H8071 (שִׂמְלָה) (by insertion of the aleph) through the idea of wrapping up);","strongs_def":"properly, dark (as enveloped), i.e. the north; hence (by orientation), the left hand","kjv_def":"left (hand, side)."},
"H8041":{"lemma":"שָׂמַאל","xlit":"sâmaʼl","pron":"saw-mal'","derivation":"a primitive root (denominative from H8040 (שְׂמֹאול));","strongs_def":"to use the left hand or pass in that direction)","kjv_def":"(go, turn) (on the, to the) left."},
"H8042":{"lemma":"שְׂמָאלִי","xlit":"sᵉmâʼlîy","pron":"sem-aw-lee'","derivation":"from H8040 (שְׂמֹאול);","strongs_def":"situated on the left side","kjv_def":"left."},
"H8043":{"lemma":"שִׁמְאָם","xlit":"Shimʼâm","pron":"shim-awm'","derivation":"for H8039 (שִׁמְאָה) (compare H38 (אֲבִיָּם));","strongs_def":"Shimam, an Israelite","kjv_def":"Shimeam."},
"H8044":{"lemma":"שַׁמְגַּר","xlit":"Shamgar","pron":"sham-gar'","derivation":"of uncertain derivation;","strongs_def":"Shamgar, an Israelite judge","kjv_def":"Shamgar."},
"H8045":{"lemma":"שָׁמַד","xlit":"shâmad","pron":"shaw-mad'","derivation":"a primitive root;","strongs_def":"to desolate","kjv_def":"destory(-uction), bring to nought, overthrow, perish, pluck down, [idiom] utterly."},
"H8046":{"lemma":"שְׁמַד","xlit":"shᵉmad","pron":"shem-ad'","derivation":"(Aramaic) corresponding to H8045 (שָׁמַד)","strongs_def":"{to desolate}","kjv_def":"consume."},
"H8047":{"lemma":"שַׁמָּה","xlit":"shammâh","pron":"sham-maw'","derivation":"from H8074 (שָׁמֵם);","strongs_def":"ruin; by implication, consternation","kjv_def":"astonishment, desolate(-ion), waste, wonderful thing."},
"H8048":{"lemma":"שַׁמָּה","xlit":"Shammâh","pron":"sham-maw'","derivation":"the same as H8047 (שַׁמָּה);","strongs_def":"Shammah, the name of an Edomite and four Israelites","kjv_def":"Shammah."},
"H8049":{"lemma":"שַׁמְהוּת","xlit":"Shamhûwth","pron":"sham-hooth'","derivation":"for H8048 (שַׁמָּה); desolation;","strongs_def":"Shamhuth, an Israelite","kjv_def":"Shamhuth."},
"H8050":{"lemma":"שְׁמוּאֵל","xlit":"Shᵉmûwʼêl","pron":"sehm-oo-ale'","derivation":"from the passive participle of H8085 (שָׁמַע) and H410 (אֵל); heard of God;","strongs_def":"Shemuel, the name of three Israelites","kjv_def":"Samuel, Shemuel."},
"H8051":{"lemma":"שַׁמּוּעַ","xlit":"Shammûwaʻ","pron":"sham-moo'-ah","derivation":"from H8074 (שָׁמֵם); renowned;","strongs_def":"Shammua, the name of four Israelites","kjv_def":"Shammua, Shammuah."},
"H8052":{"lemma":"שְׁמוּעָה","xlit":"shᵉmûwʻâh","pron":"sehm-oo-aw'","derivation":"feminine passive participle of H8074 (שָׁמֵם);","strongs_def":"something heard, i.e. an announcement","kjv_def":"bruit, doctrine, fame, mentioned, news, report, rumor, tidings."},
"H8053":{"lemma":"שָׁמוּר","xlit":"Shâmûwr","pron":"shaw-moor'","derivation":"passive participle of H8103 (שִׁמְצָה); observed;","strongs_def":"Shamur, an Israelite","kjv_def":"Shamir (from the margin)."},
"H8054":{"lemma":"שַׁמּוֹת","xlit":"Shammôwth","pron":"sham-moth'","derivation":"plural of H8047 (שַׁמָּה); ruins;","strongs_def":"Shammoth, an Israelite","kjv_def":"Shamoth."},
"H8055":{"lemma":"שָׂמַח","xlit":"sâmach","pron":"saw-makh'","derivation":"a primitive root;","strongs_def":"probably to brighten up, i.e. (figuratively) be (causatively, make) blithe or gleesome","kjv_def":"cheer up, be (make) glad, (have, make) joy(-ful), be (make) merry, (cause to, make to) rejoice, [idiom] very."},
"H8056":{"lemma":"שָׂמֵחַ","xlit":"sâmêach","pron":"saw-may'-akh","derivation":"from H8055 (שָׂמַח);","strongs_def":"blithe or gleeful","kjv_def":"(be) glad, joyful, (making) merry((-hearted), -ily), rejoice(-ing)."},
"H8057":{"lemma":"שִׂמְחָה","xlit":"simchâh","pron":"sim-khaw'","derivation":"from H8056 (שָׂמֵחַ);","strongs_def":"blithesomeness or glee, (religious or festival)","kjv_def":"[idiom] exceeding(-ly), gladness, joy(-fulness), mirth, pleasure, rejoice(-ing)."},
"H8058":{"lemma":"שָׁמַט","xlit":"shâmaṭ","pron":"shaw-mat'","derivation":"a primitive root;","strongs_def":"to fling down; incipiently to jostle; figuratively, to let alone, desist, remit","kjv_def":"discontinue, overthrow, release, let rest, shake, stumble, throw down."},
"H8059":{"lemma":"שְׁמִטָּה","xlit":"shᵉmiṭṭâh","pron":"shem-it-taw'","derivation":"from H8058 (שָׁמַט);","strongs_def":"remission (of debt) or suspension of labor)","kjv_def":"release."},
"H8060":{"lemma":"שַׁמַּי","xlit":"Shammay","pron":"sham-mah'-ee","derivation":"from H8073 (שַׁמְלַי); destructive;","strongs_def":"Shammai, the name of three Israelites","kjv_def":"Shammai."},
"H8061":{"lemma":"שְׁמִידָע","xlit":"Shᵉmîydâʻ","pron":"shem-ee-daw'","derivation":"apparently from H8034 (שֵׁם) and H3045 (יָדַע); name of knowing;","strongs_def":"Shemida, an Israelite","kjv_def":"Shemida, Shemidah."},
"H8062":{"lemma":"שְׁמִידָעִי","xlit":"Shᵉmîydâʻîy","pron":"shem-ee-daw-ee'","derivation":"patronymically from H8061 (שְׁמִידָע);","strongs_def":"a Shemidaite (collectively) or descendants of Shemida","kjv_def":"Shemidaites."},
"H8063":{"lemma":"שְׂמִיכָה","xlit":"sᵉmîykâh","pron":"sem-ee-kaw'","derivation":"from H5564 (סָמַךְ);","strongs_def":"a run (as sustaining the Oriental sitter)","kjv_def":"mantle."},
"H8064":{"lemma":"שָׁמַיִם","xlit":"shâmayim","pron":"shaw-mah'-yim","derivation":"dual of an unused singular שָׁמֶה; from an unused root meaning to be lofty;","strongs_def":"the sky (as aloft; the dual perhaps alluding to the visible arch in which the clouds move, as well as to the higher ether where the celestial bodies revolve)","kjv_def":"air, [idiom] astrologer, heaven(-s)."},
"H8065":{"lemma":"שָׁמַיִן","xlit":"shâmayin","pron":"shaw-mah'-yin","derivation":"(Aramaic) corresponding to H8064 (שָׁמַיִם)","strongs_def":"{the sky (as aloft; the dual perhaps alluding to the visible arch in which the clouds move, as well as to the higher ether where the celestial bodies revolve)}","kjv_def":"heaven."},
"H8066":{"lemma":"שְׁמִינִי","xlit":"shᵉmîynîy","pron":"shem-ee-nee'","derivation":"from H8083 (שְׁמֹנֶה);","strongs_def":"eight","kjv_def":"eight."},
"H8067":{"lemma":"שְׁמִינִית","xlit":"shᵉmîynîyth","pron":"shem-ee-neeth'","derivation":"feminine of H8066 (שְׁמִינִי);","strongs_def":"probably an eight-stringed lyre","kjv_def":"Sheminith."},
"H8068":{"lemma":"שָׁמִיר","xlit":"shâmîyr","pron":"shaw-meer'","derivation":"from H8104 (שָׁמַר) in the original sense of pricking;","strongs_def":"a thorn; also (from its keenness for scratching) a gem, probably the diamond","kjv_def":"adamant (stone), brier, diamond."},
"H8069":{"lemma":"שָׁמִיר","xlit":"Shâmîyr","pron":"shaw-meer'","derivation":"the same as H8068 (שָׁמִיר);","strongs_def":"Shamir, the name of two places in Palestine","kjv_def":"Shamir. Compare H8053 (שָׁמוּר)."},
"H8070":{"lemma":"שְׁמִירָמוֹת","xlit":"Shᵉmîyrâmôwth","pron":"shem-ee-raw-moth'","derivation":"or שְׁמָרִימוֹת; probably from H8034 (שֵׁם) and plural of H7413 (רָמָה); name of heights;","strongs_def":"Shemiramoth, the name of two Israelites","kjv_def":"Shemiramoth."},
"H8071":{"lemma":"שִׂמְלָה","xlit":"simlâh","pron":"sim-law'","derivation":"perhaps by permutation for the feminine of H5566 (סֶמֶל) (through the idea of a cover assuming the shape of the object beneath);","strongs_def":"a dress, especially a mantle","kjv_def":"apparel, cloth(-es, -ing), garment, raiment. Compare H8008 (שַׂלְמָה)."},
"H8072":{"lemma":"שַׂמְלָה","xlit":"Samlâh","pron":"sam-law'","derivation":"probably for the same as H8071 (שִׂמְלָה);","strongs_def":"Samlah, an Edomite","kjv_def":"Samlah."},
"H8073":{"lemma":"שַׁמְלַי","xlit":"Shamlay","pron":"sham-lah'-ee","derivation":"for H8014 (שַׂלְמַי);","strongs_def":"Shamlai, one of the Nethinim","kjv_def":"Shalmai (from the margin)."},
"H8074":{"lemma":"שָׁמֵם","xlit":"shâmêm","pron":"shaw-mame'","derivation":"a primitive root;","strongs_def":"to stun (or intransitively, grow numb), i.e. devastate or (figuratively) stupefy (both usually in a passive sense)","kjv_def":"make amazed, be astonied, (be an) astonish(-ment), (be, bring into, unto, lay, lie, make) desolate(-ion, places), be destitute, destroy (self), (lay, lie, make) waste, wonder."},
"H8075":{"lemma":"שְׁמַם","xlit":"shᵉmam","pron":"shem-am'","derivation":"(Aramaic) corresponding to H8074 (שָׁמֵם)","strongs_def":"{to stun (or intransitively, grow numb), i.e. devastate or (figuratively) stupefy (both usually in a passive sense)}","kjv_def":"be astonied."},
"H8076":{"lemma":"שָׁמֵם","xlit":"shâmêm","pron":"shaw-mame'","derivation":"from H8074 (שָׁמֵם);","strongs_def":"ruined","kjv_def":"desolate."},
"H8077":{"lemma":"שְׁמָמָה","xlit":"shᵉmâmâh","pron":"shem-aw-maw'","derivation":"or שִׁמָמָה; feminine of H8076 (שָׁמֵם);","strongs_def":"devastation; figuratively, astonishment","kjv_def":"(laid, [idiom] most) desolate(-ion), waste."},
"H8078":{"lemma":"שִׁמָּמוֹן","xlit":"shimmâmôwn","pron":"shim-maw-mone'","derivation":"from H8074 (שָׁמֵם);","strongs_def":"stupefaction","kjv_def":"astonishment."},
"H8079":{"lemma":"שְׂמָמִית","xlit":"sᵉmâmîyth","pron":"sem-aw-meeth'","derivation":"probably from H8074 (שָׁמֵם) (in the sense of poisoning);","strongs_def":"a lizard (from the superstition of its noxiousness)","kjv_def":"spider."},
"H8080":{"lemma":"שָׁמַן","xlit":"shâman","pron":"shaw-man'","derivation":"a primitive root;","strongs_def":"to shine, i.e. (by analogy) be (causatively, make) oily or gross","kjv_def":"become (make, wax) fat."},
"H8081":{"lemma":"שֶׁמֶן","xlit":"shemen","pron":"sheh'-men","derivation":"from H8080 (שָׁמַן);","strongs_def":"grease, especially liquid (as from the olive, often perfumed); figuratively, richness","kjv_def":"anointing, [idiom] fat (things), [idiom] fruitful, oil(-ed), ointment, olive, [phrase] pine."},
"H8082":{"lemma":"שָׁמֵן","xlit":"shâmên","pron":"shaw-mane'","derivation":"from H8080 (שָׁמַן);","strongs_def":"greasy, i.e. gross; figuratively, rich","kjv_def":"fat, lusty, plenteous."},
"H8083":{"lemma":"שְׁמֹנֶה","xlit":"shᵉmôneh","pron":"shem-o-neh'","derivation":"or שְׁמוֹנֶה; feminine שְׁמֹנָה; or שְׁמוֹנָה; apparently from H8082 (שָׁמֵן) through the idea of plumpness;","strongs_def":"a cardinal number, eight (as if a surplus above the 'perfect' seven); also (as ordinal) eighth","kjv_def":"eight(-een, -eenth), eighth."},
"H8084":{"lemma":"שְׁמֹנִים","xlit":"shᵉmônîym","pron":"shem-o-neem'","derivation":"or שְׁמוֹנִים; multiple from H8083 (שְׁמֹנֶה);","strongs_def":"eighty, also eightieth","kjv_def":"eighty(-ieth), fourscore."},
"H8085":{"lemma":"שָׁמַע","xlit":"shâmaʻ","pron":"shaw-mah'","derivation":"a primitive root;","strongs_def":"to hear intelligently (often with implication of attention, obedience, etc.; causatively, to tell, etc.)","kjv_def":"[idiom] attentively, call (gather) together, [idiom] carefully, [idiom] certainly, consent, consider, be content, declare, [idiom] diligently, discern, give ear, (cause to, let, make to) hear(-ken, tell), [idiom] indeed, listen, make (a) noise, (be) obedient, obey, perceive, (make a) proclaim(-ation), publish, regard, report, shew (forth), (make a) sound, [idiom] surely, tell, understand, whosoever (heareth), witness."},
"H8086":{"lemma":"שְׁמַע","xlit":"shᵉmaʻ","pron":"shem-ah'","derivation":"(Aramaic) corresponding to H8085 (שָׁמַע)","strongs_def":"{to hear intelligently (often with implication of attention, obedience, etc.; causatively, to tell, etc.)}","kjv_def":"hear, obey."},
"H8087":{"lemma":"שֶׁמַע","xlit":"Shemaʻ","pron":"sheh'-mah","derivation":"for the same as H8088 (שֵׁמַע);","strongs_def":"Shema, the name of a place in Palestine and of four Israelites","kjv_def":"Shema."},
"H8088":{"lemma":"שֵׁמַע","xlit":"shêmaʻ","pron":"shay'-mah","derivation":"from H8085 (שָׁמַע);","strongs_def":"something heard, i.e. a sound, rumor, announcement; abstractly, audience","kjv_def":"bruit, fame, hear(-ing), loud, report, speech, tidings."},
"H8089":{"lemma":"שֹׁמַע","xlit":"shômaʻ","pron":"sho'-mah","derivation":"from H8085 (שָׁמַע);","strongs_def":"a report","kjv_def":"fame."},
"H8090":{"lemma":"שְׁמָע","xlit":"Shᵉmâʻ","pron":"shem-aw'","derivation":"for H8087 (שֶׁמַע);","strongs_def":"Shema, a place in Palestine","kjv_def":"Shema."},
"H8091":{"lemma":"שָׁמָע","xlit":"Shâmâʻ","pron":"shaw-maw'","derivation":"from H8085 (שָׁמַע); obedient;","strongs_def":"Shama, an Israelite","kjv_def":"Shama."},
"H8092":{"lemma":"שִׁמְעָא","xlit":"Shimʻâʼ","pron":"shim-aw'","derivation":"for H8093 (שִׁמְעָה);","strongs_def":"Shima, the name of four Israelites","kjv_def":"Shimea, Shimei, Shamma."},
"H8093":{"lemma":"שִׁמְעָה","xlit":"Shimʻâh","pron":"shim-aw'","derivation":"feminine of H8088 (שֵׁמַע); annunciation;","strongs_def":"Shimah, an Israelite","kjv_def":"Shimeah."},
"H8094":{"lemma":"שְׁמָעָה","xlit":"Shᵉmâʻâh","pron":"shem-aw-aw'","derivation":"for H8093 (שִׁמְעָה);","strongs_def":"Shemaah, an Israelite","kjv_def":"Shemaah."},
"H8095":{"lemma":"שִׁמְעוֹן","xlit":"Shimʻôwn","pron":"shim-one'","derivation":"from H8085 (שָׁמַע); hearing;","strongs_def":"Shimon, one of Jacob's sons, also the tribe descended from him","kjv_def":"Simeon."},
"H8096":{"lemma":"שִׁמְעִי","xlit":"Shimʻîy","pron":"shim-ee'","derivation":"from H8088 (שֵׁמַע); famous;","strongs_def":"Shimi, the name of twenty Israelites","kjv_def":"Shimeah (from the margin), Shimei, Shimhi, Shimi."},
"H8097":{"lemma":"שִׁמְעִי","xlit":"Shimʻîy","pron":"shim-ee'","derivation":"patronymically from H8096 (שִׁמְעִי);","strongs_def":"a Shimite (collectively) or descendants of Shimi","kjv_def":"of Shimi, Shimites."},
"H8098":{"lemma":"שְׁמַעְיָה","xlit":"Shᵉmaʻyâh","pron":"shem-aw-yaw'","derivation":"or שְׁמַעְיָהוּ; from H8085 (שָׁמַע) and H3050 (יָהּ); Jah has heard;","strongs_def":"Shemajah, the name of twenty-five Israelites","kjv_def":"Shemaiah."},
"H8099":{"lemma":"שִׁמְעֹנִי","xlit":"Shimʻônîy","pron":"shim-o-nee'","derivation":"patronymically from H8095 (שִׁמְעוֹן);","strongs_def":"a Shimonite (collectively) or descendants of Shimon","kjv_def":"tribe of Simeon, Simeonites."},
"H8100":{"lemma":"שִׁמְעַת","xlit":"Shimʻath","pron":"shim-awth'","derivation":"feminine of H8088 (שֵׁמַע); annunciation;","strongs_def":"Shimath, an Ammonitess","kjv_def":"Shimath."},
"H8101":{"lemma":"שִׁמְעָתִי","xlit":"Shimʻâthîy","pron":"shim-aw-thee'","derivation":"patronymically from H8093 (שִׁמְעָה);","strongs_def":"a Shimathite (collectively) or descendants of Shimah","kjv_def":"Shimeathites."},
"H8102":{"lemma":"שֶׁמֶץ","xlit":"shemets","pron":"sheh'-mets","derivation":"from an unused root meaning to emit a sound;","strongs_def":"an inkling","kjv_def":"a little."},
"H8103":{"lemma":"שִׁמְצָה","xlit":"shimtsâh","pron":"shim-tsaw'","derivation":"feminine of H8102 (שֶׁמֶץ);","strongs_def":"scornful whispering (of hostile spectators)","kjv_def":"shame."},
"H8104":{"lemma":"שָׁמַר","xlit":"shâmar","pron":"shaw-mar'","derivation":"a primitive root;","strongs_def":"properly, to hedge about (as with thorns), i.e. guard; generally, to protect, attend to, etc.","kjv_def":"beward, be circumspect, take heed (to self), keep(-er, self), mark, look narrowly, observe, preserve, regard, reserve, save (self), sure, (that lay) wait (for), watch(-man)."},
"H8105":{"lemma":"שֶׁמֶר","xlit":"shemer","pron":"sheh'-mer","derivation":"from H8104 (שָׁמַר);","strongs_def":"something preserved, i.e. the settlings (plural only) of wine","kjv_def":"dregs, (wines on the) lees."},
"H8106":{"lemma":"שֶׁמֶר","xlit":"Shemer","pron":"sheh'-mer","derivation":"the same as H8105 (שֶׁמֶר);","strongs_def":"Shemer, the name of three Israelites","kjv_def":"Shamer, Shemer."},
"H8107":{"lemma":"שִׁמֻּר","xlit":"shimmur","pron":"shim-moor'","derivation":"from H8104 (שָׁמַר);","strongs_def":"an observance","kjv_def":"[idiom] be (much) observed."},
"H8108":{"lemma":"שׇׁמְרָה","xlit":"shomrâh","pron":"shom-raw'","derivation":"feminine of an unused noun from H8104 (שָׁמַר) meaning a guard;","strongs_def":"watchfulness","kjv_def":"watch."},
"H8109":{"lemma":"שְׁמֻרָה","xlit":"shᵉmurâh","pron":"shem-oo-raw'","derivation":"feminine of passive participle of H8104 (שָׁמַר);","strongs_def":"something guarded, i.e. an eye-lid","kjv_def":"waking."},
"H8110":{"lemma":"שִׁמְרוֹן","xlit":"Shimrôwn","pron":"shim-rone'","derivation":"from H8105 (שֶׁמֶר) in its original sense; guardianship;","strongs_def":"Shimron, the name of an Israelite and of a place in Palestine","kjv_def":"Shimron."},
"H8111":{"lemma":"שֹׁמְרוֹן","xlit":"Shômᵉrôwn","pron":"sho-mer-one'","derivation":"from the active participle of H8104 (שָׁמַר); watch-station;","strongs_def":"Shomeron, a place in Palestine","kjv_def":"Samaria."},
"H8112":{"lemma":"שִׁמְרוֹן מְראוֹן","xlit":"Shimrôwn Mᵉrʼôwn","pron":"shim-rone' mer-one'","derivation":"from H8110 (שִׁמְרוֹן) and a derivative of H4754 (מָרָא); guard of lashing;","strongs_def":"Shimron-Meron, a place in Palestine","kjv_def":"Shimon-meron."},
"H8113":{"lemma":"שִׁמְרִי","xlit":"Shimrîy","pron":"shim-ree'","derivation":"from H8105 (שֶׁמֶר) in its original sense; watchful;","strongs_def":"Shimri, the name of four Israelites","kjv_def":"Shimri."},
"H8114":{"lemma":"שְׁמַרְיָה","xlit":"Shᵉmaryâh","pron":"shem-ar-yaw'","derivation":"or שְׁמַרְיָהוּ; from H8104 (שָׁמַר) and H3050 (יָהּ); Jah has guarded;","strongs_def":"Shemarjah, the name of four Israelites","kjv_def":"Shamariah, Shemariah."},
"H8115":{"lemma":"שׇׁמְרַיִן","xlit":"Shomrayin","pron":"shom-rah'-yin","derivation":"(Aramaic) corresponding to H8111 (שֹׁמְרוֹן);","strongs_def":"Shomrain, a place in Palestine","kjv_def":"Samaria."},
"H8116":{"lemma":"שִׁמְרִית","xlit":"Shimrîyth","pron":"shim-reeth'","derivation":"feminine of H8113 (שִׁמְרִי); female guard;","strongs_def":"Shimrith, a Moabitess","kjv_def":"Shimrith."},
"H8117":{"lemma":"שִׁמְרֹנִי","xlit":"Shimrônîy","pron":"shim-ro-nee'","derivation":"patronymically from H8110 (שִׁמְרוֹן);","strongs_def":"a Shimronite (collectively) or descendants of Shimron","kjv_def":"Shimronites."},
"H8118":{"lemma":"שֹׁמרֹנִי","xlit":"Shômrônîy","pron":"sho-mer-o-nee'","derivation":"patrial from H8111 (שֹׁמְרוֹן);","strongs_def":"a Shomeronite (collectively) or inhabitants of Shomeron","kjv_def":"Samaritans."},
"H8119":{"lemma":"שִׁמְרָת","xlit":"Shimrâth","pron":"shim-rawth'","derivation":"from H8104 (שָׁמַר); guardship;","strongs_def":"Shimrath, an Israelite","kjv_def":"Shimrath."},
"H8120":{"lemma":"שְׁמַשׁ","xlit":"shᵉmash","pron":"shem-ash'","derivation":"(Aramaic) corresponding to the root of H8121 (שֶׁמֶשׁ) through the idea of activity implied in day-light;","strongs_def":"to serve","kjv_def":"minister."},
"H8121":{"lemma":"שֶׁמֶשׁ","xlit":"shemesh","pron":"sheh'-mesh","derivation":"from an unused root meaning to be brilliant;","strongs_def":"the sun; by implication, the east; figuratively, a ray, i.e. (architectural) a notched battlement","kjv_def":"[phrase] east side(-ward), sun (rising), [phrase] west(-ward), window. See also H1053 (בֵּית שֶׁמֶשׁ)."},
"H8122":{"lemma":"שֶׁמֶשׁ","xlit":"shemesh","pron":"sheh'-mesh","derivation":"(Aramaic) corresponding to H8121 (שֶׁמֶשׁ);","strongs_def":"the sun","kjv_def":"sun."},
"H8123":{"lemma":"שִׁמְשׁוֹן","xlit":"Shimshôwn","pron":"shim-shone'","derivation":"from H8121 (שֶׁמֶשׁ); sunlight;","strongs_def":"Shimshon, an Israelite","kjv_def":"Samson."},
"H8124":{"lemma":"שִׁמְשַׁי","xlit":"Shimshay","pron":"shim-shah'-ee","derivation":"(Aramaic) from H8122 (שֶׁמֶשׁ); sunny;","strongs_def":"Shimshai, a Samaritan","kjv_def":"Shimshai."},
"H8125":{"lemma":"שַׁמְשְׁרַי","xlit":"Shamshᵉray","pron":"sham-sher-ah'-ee","derivation":"apparently from H8121 (שֶׁמֶשׁ); sunlike;","strongs_def":"Shamsherai, an Israelite","kjv_def":"Shamsherai."},
"H8126":{"lemma":"שֻׁמָתִי","xlit":"Shumâthîy","pron":"shoo-maw-thee'","derivation":"patronymically from an unused name from H7762 (שׁוּם) probably meaning garlic-smell;","strongs_def":"a Shumathite (collectively) or descendants of Shumah","kjv_def":"Shumathites."},
"H8127":{"lemma":"שֵׁן","xlit":"shên","pron":"shane","derivation":"from H8150 (שָׁנַן); compare  H8143 (שֶׁנְהַבִּים)","strongs_def":"a tooth (as sharp); specifically  ivory; figuratively, a cliff","kjv_def":"crag, [idiom] forefront, ivory, [idiom] sharp, tooth."},
"H8128":{"lemma":"שֵׁן","xlit":"shên","pron":"shane","derivation":"(Aramaic) corresponding to H8127 (שֵׁן);","strongs_def":"a tooth","kjv_def":"tooth."},
"H8129":{"lemma":"שֵׁן","xlit":"Shên","pron":"shane","derivation":"the same as H8127 (שֵׁן); crag;","strongs_def":"Shen, a place in Palestine","kjv_def":"Shen."},
"H8130":{"lemma":"שָׂנֵא","xlit":"sânêʼ","pron":"saw-nay'","derivation":"a primitive root;","strongs_def":"to hate (personally)","kjv_def":"enemy, foe, (be) hate(-ful, -r), odious, [idiom] utterly."},
"H8131":{"lemma":"שְׂנֵא","xlit":"sᵉnêʼ","pron":"sen-ay'","derivation":"(Aramaic) corresponding to H8130 (שָׂנֵא)","strongs_def":"{to hate (personally)}","kjv_def":"hate."},
"H8132":{"lemma":"שָׁנָא","xlit":"shânâʼ","pron":"shaw-naw'","derivation":"a primitive root;","strongs_def":"to alter","kjv_def":"change."},
"H8133":{"lemma":"שְׁנָא","xlit":"shᵉnâʼ","pron":"shen-aw'","derivation":"(Aramaic) corresponding to H8132 (שָׁנָא)","strongs_def":"{to alter}","kjv_def":"alter, change, (be) diverse."},
"H8134":{"lemma":"שִׁנְאָב","xlit":"Shinʼâb","pron":"shin-awb'","derivation":"probably from H8132 (שָׁנָא) and H1 (אָב); a father has turned;","strongs_def":"Shinab, a Canaanite","kjv_def":"Shinab."},
"H8135":{"lemma":"שִׂנְאָה","xlit":"sinʼâh","pron":"sin-aw'","derivation":"from H8130 (שָׂנֵא);","strongs_def":"hate","kjv_def":"[phrase] exceedingly, hate(-ful, -red)."},
"H8136":{"lemma":"שִׁנְאָן","xlit":"shinʼân","pron":"shin-awn'","derivation":"from H8132 (שָׁנָא);","strongs_def":"change, i.e. repetition","kjv_def":"[idiom] angels."},
"H8137":{"lemma":"שֶׁנְאַצַּר","xlit":"Shenʼatstsar","pron":"shen-ats-tsar'","derivation":"apparently of Babylonian origin;","strongs_def":"Shenatstsar, an Israelite","kjv_def":"Senazar."},
"H8138":{"lemma":"שָׁנָה","xlit":"shânâh","pron":"shaw-naw'","derivation":"a primitive root;","strongs_def":"to fold, i.e. duplicate (literally or figuratively); by implication, to transmute (transitive or intransitive)","kjv_def":"do (speak, strike) again, alter, double, (be given to) change, disguise, (be) diverse, pervert, prefer, repeat, return, do the second time."},
"H8139":{"lemma":"שְׁנָה","xlit":"shᵉnâh","pron":"shen-aw'","derivation":"(Aramaic) corresponding to H8142 (שֵׁנָה)","strongs_def":"{sleep}","kjv_def":"sleep."},
"H8140":{"lemma":"שְׁנָה","xlit":"shᵉnâh","pron":"shen-aw'","derivation":"(Aramaic) corresponding to H8141 (שָׁנֶה)","strongs_def":"{a year (as a revolution of time)}","kjv_def":"year."},
"H8141":{"lemma":"שָׁנֶה","xlit":"shâneh","pron":"shaw-neh'","derivation":"(in plural or (feminine) שָׁנָה; from H8138 (שָׁנָה);","strongs_def":"a year (as a revolution of time)","kjv_def":"[phrase] whole age, [idiom] long, [phrase] old, year([idiom] -ly)."},
"H8142":{"lemma":"שֵׁנָה","xlit":"shênâh","pron":"shay-naw'","derivation":"or שֵׁנָא; (Psalm 127:2), from H3462 (יָשֵׁן);","strongs_def":"sleep","kjv_def":"sleep."},
"H8143":{"lemma":"שֶׁנְהַבִּים","xlit":"shenhabbîym","pron":"shen-hab-beem'","derivation":"from H8127 (שֵׁן) and the plural apparently of a foreign word;","strongs_def":"probably, tooth of elephants, i.e. ivory tusk","kjv_def":"ivory."},
"H8144":{"lemma":"שָׁנִי","xlit":"shânîy","pron":"shaw-nee'","derivation":"of uncertain derivation;","strongs_def":"crimson, properly, the insect or its color, also stuff dyed with it","kjv_def":"crimson, scarlet (thread)."},
"H8145":{"lemma":"שֵׁנִי","xlit":"shênîy","pron":"shay-nee'","derivation":"from H8138 (שָׁנָה);","strongs_def":"properly, double, i.e. second; also adverbially, again","kjv_def":"again, either (of them), (an-) other, second (time)."},
"H8146":{"lemma":"שָׂנִיא","xlit":"sânîyʼ","pron":"saw-nee'","derivation":"from H8130 (שָׂנֵא);","strongs_def":"hated","kjv_def":"hated."},
"H8147":{"lemma":"שְׁנַיִם","xlit":"shᵉnayim","pron":"shen-ah'-yim","derivation":"dual of H8145 (שֵׁנִי); feminine שְׁתַּיִם;","strongs_def":"two; also (as ordinal) twofold","kjv_def":"both, couple, double, second, twain, [phrase] twelfth, [phrase] twelve, [phrase] twenty (sixscore) thousand, twice, two."},
"H8148":{"lemma":"שְׁנִינָה","xlit":"shᵉnîynâh","pron":"shen-ee-naw'","derivation":"from H8150 (שָׁנַן);","strongs_def":"something pointed, i.e. a gibe","kjv_def":"byword, taunt."},
"H8149":{"lemma":"שְׁנִיר","xlit":"Shᵉnîyr","pron":"shen-eer'","derivation":"or שְׂנִיר; from an unused root meaning to be pointed; peak;","strongs_def":"Shenir or Senir, a summit of Lebanon","kjv_def":"Senir, Shenir."},
"H8150":{"lemma":"שָׁנַן","xlit":"shânan","pron":"shaw-nan'","derivation":"a primitive root;","strongs_def":"to point (transitive or intransitive); intensively, to pierce; figuratively, to inculcate","kjv_def":"prick, sharp(-en), teach diligently, whet."},
"H8151":{"lemma":"שָׁנַס","xlit":"shânaç","pron":"shaw-nas'","derivation":"a primitive root;","strongs_def":"to compress (with a belt)","kjv_def":"gird up."},
"H8152":{"lemma":"שִׁנְעָר","xlit":"Shinʻâr","pron":"shin-awr'","derivation":"probably of foreign derivation;","strongs_def":"Shinar, a plain in Babylonia","kjv_def":"Shinar."},
"H8153":{"lemma":"שְׁנָת","xlit":"shᵉnâth","pron":"shen-awth'","derivation":"from H3462 (יָשֵׁן);","strongs_def":"sleep","kjv_def":"sleep."},
"H8154":{"lemma":"שָׁסָה","xlit":"shâçâh","pron":"shaw-saw'","derivation":"or שָׁשָׂה; (Isaiah 10:13), a primitive root;","strongs_def":"to plunder","kjv_def":"destroyer, rob, spoil(-er)."},
"H8155":{"lemma":"שָׁסַס","xlit":"shâçaç","pron":"shaw-sas'","derivation":"a primitive root;","strongs_def":"to plunder","kjv_def":"rifle, spoil."},
"H8156":{"lemma":"שָׁסַע","xlit":"shâçaʻ","pron":"shaw-sah'","derivation":"a primitive root;","strongs_def":"to split or tear; figuratively, to upbraid","kjv_def":"cleave, (be) cloven (footed), rend, stay."},
"H8157":{"lemma":"שֶׁסַע","xlit":"sheçaʻ","pron":"sheh'-sah","derivation":"from H8156 (שָׁסַע);","strongs_def":"a fissure","kjv_def":"cleft, clovenfooted."},
"H8158":{"lemma":"שָׁסַף","xlit":"shâçaph","pron":"shaw-saf'","derivation":"a primitive root;","strongs_def":"to cut in pieces, i.e. slaughter","kjv_def":"hew in pieces."},
"H8159":{"lemma":"שָׁעָה","xlit":"shâʻâh","pron":"shaw-aw'","derivation":"a primitive root;","strongs_def":"to gaze at or about (properly, for help); by implication, to inspect, consider, compassionate, be nonplussed (as looking around in amazement) or bewildered","kjv_def":"depart, be dim, be dismayed, look (away), regard, have respect, spare, turn."},
"H8160":{"lemma":"שָׁעָה","xlit":"shâʻâh","pron":"shaw-aw'","derivation":"(Aramaic) from a root corresponding to H8159 (שָׁעָה);","strongs_def":"properly, a look, i.e. a moment","kjv_def":"hour."},
"H8161":{"lemma":"שַׁעֲטָה","xlit":"shaʻăṭâh","pron":"shah'-at-aw","derivation":"feminine from an unused root meaning to stamp;","strongs_def":"a clatter (of hoofs)","kjv_def":"stamping."},
"H8162":{"lemma":"שַׁעַטְנֵז","xlit":"shaʻaṭnêz","pron":"shah-at-naze'","derivation":"probably of foreign derivation;","strongs_def":"linsey-woolsey, i.e. cloth of linen and wool carded and spun together","kjv_def":"garment of divers sorts, linen and wollen."},
"H8163":{"lemma":"שָׂעִיר","xlit":"sâʻîyr","pron":"saw-eer'","derivation":"or שָׂעִר; from H8175 (שָׂעַר);","strongs_def":"shaggy; as noun, a he-goat; by analogy, a faun","kjv_def":"devil, goat, hairy, kid, rough, satyr."},
"H8164":{"lemma":"שָׂעִיר","xlit":"sâʻîyr","pron":"saw-eer'","derivation":"formed the same as H8163 (שָׂעִיר);","strongs_def":"a shower (as tempestuous)","kjv_def":"small rain."},
"H8165":{"lemma":"שֵׂעִיר","xlit":"Sêʻîyr","pron":"say-eer'","derivation":"formed like H8163 (שָׂעִיר); rough;","strongs_def":"Seir, a mountain of Idumaea and its aboriginal occupants, also one in Palestine","kjv_def":"Seir."},
"H8166":{"lemma":"שְׂעִירָה","xlit":"sᵉʻîyrâh","pron":"seh-ee-raw'","derivation":"feminine of H8163 (שָׂעִיר);","strongs_def":"a she-goat","kjv_def":"kid."},
"H8167":{"lemma":"שְׂעִירָה","xlit":"Sᵉʻîyrâh","pron":"seh-ee-raw'","derivation":"formed as H8166 (שְׂעִירָה); roughness;","strongs_def":"Seirah, a place in Palestine","kjv_def":"Seirath."},
"H8168":{"lemma":"שֹׁעַל","xlit":"shôʻal","pron":"sho'-al","derivation":"from an unused root meaning to hollow out;","strongs_def":"the palm; by extension, a handful","kjv_def":"handful, hollow of the hand."},
"H8169":{"lemma":"שַׁעַלְבִים","xlit":"Shaʻalbîym","pron":"shah-al-beem'","derivation":"or שַׁעֲלַבִּין; plural from H7776 (שׁוּעָל); fox-holes;","strongs_def":"Shaalbim or Shaalabbin, a place in Palestine","kjv_def":"Shaalabbin, Shaalbim."},
"H8170":{"lemma":"שַׁעַלְבֹנִי","xlit":"Shaʻalbônîy","pron":"shah-al-bo-nee'","derivation":"patrial from H8169 (שַׁעַלְבִים);","strongs_def":"a Shaalbonite or inhabitant of Shaalbin","kjv_def":"Shaalbonite."},
"H8171":{"lemma":"שַׁעֲלִים","xlit":"Shaʻălîym","pron":"shah-al-eem'","derivation":"plural of H7776 (שׁוּעָל); foxes;","strongs_def":"Shaalim, a place in Palestine","kjv_def":"Shalim."},
"H8172":{"lemma":"שָׁעַן","xlit":"shâʻan","pron":"shaw-an'","derivation":"a primitive root;","strongs_def":"to support one's self","kjv_def":"lean, lie, rely, rest (on, self), stay."},
"H8173":{"lemma":"שָׁעַע","xlit":"shâʻaʻ","pron":"shaw-ah'","derivation":"a primitive root;","strongs_def":"(in a good acceptation) to look upon (with complacency), i.e. fondle, please or amuse (self); (in a bad one) to look about (in dismay), i.e. stare","kjv_def":"cry (out) (by confusion with H7768 (שָׁוַע)), dandle, delight (self), play, shut."},
"H8174":{"lemma":"שַׁעַף","xlit":"Shaʻaph","pron":"shah'-af","derivation":"from H5586 (סָעַף); fluctuation;","strongs_def":"Shaaph, the name of two Israelites","kjv_def":"Shaaph."},
"H8175":{"lemma":"שָׂעַר","xlit":"sâʻar","pron":"saw-ar'","derivation":"a primitive root;","strongs_def":"to storm; by implication, to shiver, i.e. fear","kjv_def":"be (horribly) afraid, fear, hurl as a storm, be tempestuous, come like (take away as with) a whirlwind."},
"H8176":{"lemma":"שָׁעַר","xlit":"shâʻar","pron":"shaw-ar'","derivation":"a primitive root; to split or open, i.e. (literally, but only as denominative from H8179 (שַׁעַר)); compare see H7778 (שׁוֹעֵר)","strongs_def":"to act as gatekeeper; (figuratively) to estimate","kjv_def":"think."},
"H8177":{"lemma":"שְׂעַר","xlit":"sᵉʻar","pron":"seh-ar'","derivation":"(Aramaic) corresponding to H8181 (שֵׂעָר);","strongs_def":"hair","kjv_def":"hair."},
"H8178":{"lemma":"שַׂעַר","xlit":"saʻar","pron":"sah'-ar","derivation":"from H8175 (שָׂעַר);","strongs_def":"a tempest; also a terror","kjv_def":"affrighted, [idiom] horribly, [idiom] sore, storm. See H8181 (שֵׂעָר)."},
"H8179":{"lemma":"שַׁעַר","xlit":"shaʻar","pron":"shah'-ar","derivation":"from H8176 (שָׁעַר) in its original sense;","strongs_def":"an opening, i.e. door or gate","kjv_def":"city, door, gate, port ([idiom] -er)."},
"H8180":{"lemma":"שַׁעַר","xlit":"shaʻar","pron":"shah'-ar","derivation":"from H8176 (שָׁעַר);","strongs_def":"a measure (as a section)","kjv_def":"(hundred-) fold."},
"H8181":{"lemma":"שֵׂעָר","xlit":"sêʻâr","pron":"say-awr'","derivation":"or שַׂעַר; (Isaiah 7:20), from H8175 (שָׂעַר) in the sense of dishevelling;","strongs_def":"hair (as if tossed or bristling)","kjv_def":"hair(-y), [idiom] rough."},
"H8182":{"lemma":"שֹׁעָר","xlit":"shôʻâr","pron":"sho-awr'","derivation":"from H8176 (שָׁעַר);","strongs_def":"harsh or horrid, i.e. offensive","kjv_def":"vile."},
"H8183":{"lemma":"שְׂעָרָה","xlit":"sᵉʻârâh","pron":"seh-aw-raw'","derivation":"feminine of H8178 (שַׂעַר);","strongs_def":"a hurricane","kjv_def":"storm, tempest."},
"H8184":{"lemma":"שְׂעֹרָה","xlit":"sᵉʻôrâh","pron":"seh-o-raw'","derivation":"or שְׂעוֹרָה (feminine meaning the plant); and (masculine meaning the grain); also שְׂעֹר; or שְׂעוֹר; from H8175 (שָׂעַר) in the sense of roughness;","strongs_def":"barley (as villose)","kjv_def":"barley."},
"H8185":{"lemma":"שַׂעֲרָה","xlit":"saʻărâh","pron":"sah-ar-aw'","derivation":"feminine of H8181 (שֵׂעָר);","strongs_def":"hairiness","kjv_def":"hair."},
"H8186":{"lemma":"שַׁעֲרוּרָה","xlit":"shaʻărûwrâh","pron":"shah-ar-oo-raw'","derivation":"or שַׁעֲרִירִיָּה; or שַׁעֲרֻרִת; feminine from H8176 (שָׁעַר) in the sense of H8175 (שָׂעַר);","strongs_def":"something fearful","kjv_def":"horrible thing."},
"H8187":{"lemma":"שְׁעַרְיָה","xlit":"Shᵉʻaryâh","pron":"sheh-ar-yaw'","derivation":"from H8176 (שָׁעַר) and H3050 (יָהּ); Jah has stormed;","strongs_def":"Shearjah, an Israelite","kjv_def":"Sheariah."},
"H8188":{"lemma":"שְׂעֹרִים","xlit":"Sᵉʻôrîym","pron":"seh-o-reem'","derivation":"masculine plural of H8184 (שְׂעֹרָה); barley grains;","strongs_def":"Seorim, an Israelite","kjv_def":"Seorim."},
"H8189":{"lemma":"שַׁעֲרַיִם","xlit":"Shaʻărayim","pron":"shah-ar-ah'-yim","derivation":"dual of H8179 (שַׁעַר); double gates;","strongs_def":"Shaarajim, a place in Palestine","kjv_def":"Shaaraim."},
"H8190":{"lemma":"שַׁעַשְׁגַּז","xlit":"Shaʻashgaz","pron":"shah-ash-gaz'","derivation":"of Persian derivation;","strongs_def":"Shaashgaz, a eunuch of Xerxes","kjv_def":"Shaashgaz."},
"H8191":{"lemma":"שַׁעְשֻׁעַ","xlit":"shaʻshuaʻ","pron":"shah-shoo'-ah","derivation":"from H8173 (שָׁעַע);","strongs_def":"enjoyment","kjv_def":"delight, pleasure."},
"H8192":{"lemma":"שָׁפָה","xlit":"shâphâh","pron":"shaw-faw'","derivation":"a primitive root;","strongs_def":"to abrade, i.e. bare","kjv_def":"high, stick out."},
"H8193":{"lemma":"שָׂפָה","xlit":"sâphâh","pron":"saw-faw'","derivation":"or (in dual and plural) שֶׂפֶת; probably from H5595 (סָפָה) or H8192 (שָׁפָה) through the idea of termination (compare H5490 (סוֹף));","strongs_def":"the lip (as a natural boundary); by implication, language; by analogy, a margin (of a vessel, water, cloth, etc.)","kjv_def":"band, bank, binding, border, brim, brink, edge, language, lip, prating, (sea-)shore, side, speech, talk, (vain) words."},
"H8194":{"lemma":"שָׁפָה","xlit":"shâphâh","pron":"shaw-faw'","derivation":"from H8192 (שָׁפָה) in the sense of clarifying;","strongs_def":"a cheese (as strained from the whey)","kjv_def":"cheese."},
"H8195":{"lemma":"שְׁפוֹ","xlit":"Shᵉphôw","pron":"shef-o'","derivation":"or שְׁפִי; from H8192 (שָׁפָה); baldness (compare H8205 (שְׁפִי));","strongs_def":"Shepho or Shephi, an Idumaean","kjv_def":"Shephi, Shepho."},
"H8196":{"lemma":"שְׁפוֹט","xlit":"shᵉphôwṭ","pron":"shef-ote'","derivation":"or שְׁפוּט; from H8199 (שָׁפַט);","strongs_def":"a judicial sentence, i.e. punishment","kjv_def":"judgment."},
"H8197":{"lemma":"שְׁפוּפָם","xlit":"Shᵉphûwphâm","pron":"shef-oo-fawm'","derivation":"or שְׁפוּפָן; from the same as H8207 (שְׁפִיפֹן); serpent-like;","strongs_def":"Shephupham or Shephuphan, an Israelite","kjv_def":"Shephuphan, Shupham."},
"H8198":{"lemma":"שִׁפְחָה","xlit":"shiphchâh","pron":"shif-khaw'","derivation":"feminine from an unused root meaning to spread out (as a family; see H4940 (מִשְׁפָּחָה));","strongs_def":"a female slave (as a member of the household)","kjv_def":"(bond-, hand-) maid(-en, -servant), wench, bondwoman, womanservant."},
"H8199":{"lemma":"שָׁפַט","xlit":"shâphaṭ","pron":"shaw-fat'","derivation":"a primitive root;","strongs_def":"to judge, i.e. pronounce sentence (for or against); by implication, to vindicate or punish; by extenssion, to govern; passively, to litigate (literally or figuratively)","kjv_def":"[phrase] avenge, [idiom] that condemn, contend, defend, execute (judgment), (be a) judge(-ment), [idiom] needs, plead, reason, rule."},
"H8200":{"lemma":"שְׁפַט","xlit":"shᵉphaṭ","pron":"shef-at'","derivation":"(Aramaic) corresponding to H8199 (שָׁפַט);","strongs_def":"to judge","kjv_def":"magistrate."},
"H8201":{"lemma":"שֶׁפֶט","xlit":"shepheṭ","pron":"sheh'-fet","derivation":"from H8199 (שָׁפַט);","strongs_def":"a sentence, i.e. infliction","kjv_def":"judgment."},
"H8202":{"lemma":"שָׁפָט","xlit":"Shâphâṭ","pron":"shaw-fawt'","derivation":"from H8199 (שָׁפַט); judge;","strongs_def":"Shaphat, the name of four Israelites","kjv_def":"Shaphat."},
"H8203":{"lemma":"שְׁפַטְיָה","xlit":"Shᵉphaṭyâh","pron":"shef-at-yaw'","derivation":"or שְׁפַטְיָהוּ; from H8199 (שָׁפַט) and H3050 (יָהּ); Jah has judged;","strongs_def":"Shephatjah, the name of ten Israelites","kjv_def":"Shephatiah."},
"H8204":{"lemma":"שִׁפְטָן","xlit":"Shiphṭân","pron":"shif-tawn'","derivation":"from H8199 (שָׁפַט); judge-like;","strongs_def":"Shiphtan, an Israelite","kjv_def":"Shiphtan."},
"H8205":{"lemma":"שְׁפִי","xlit":"shᵉphîy","pron":"shef-ee'","derivation":"from H8192 (שָׁפָה);","strongs_def":"bareness; concretely, a bare hill or plain","kjv_def":"high place, stick out."},
"H8206":{"lemma":"שֻׁפִּים","xlit":"Shuppîym","pron":"shoop-peem'","derivation":"plural of an unused noun from the same as H8207 (שְׁפִיפֹן) and meaning the same; serpents;","strongs_def":"Shuppim, an Israelite","kjv_def":"Shuppim."},
"H8207":{"lemma":"שְׁפִיפֹן","xlit":"shᵉphîyphôn","pron":"shef-ee-fone'","derivation":"from an unused root meaning the same as H7779 (שׁוּף);","strongs_def":"a kind of serpent (as snapping), probably the cerastes or horned adder","kjv_def":"adder."},
"H8208":{"lemma":"שָׁפִיר","xlit":"Shâphîyr","pron":"shaf-eer'","derivation":"from H8231 (שָׁפַר); beautiful;","strongs_def":"Shaphir, a place in Palestine","kjv_def":"Saphir."},
"H8209":{"lemma":"שַׁפִּיר","xlit":"shappîyr","pron":"shap-peer'","derivation":"(Aramaic) intensive of a form corresponding to H8208 (שָׁפִיר);","strongs_def":"beautiful","kjv_def":"fair."},
"H8210":{"lemma":"שָׁפַךְ","xlit":"shâphak","pron":"shaw-fak'","derivation":"a primitive root;","strongs_def":"to spill forth (blood, a libation, liquid metal; or even a solid, i.e. to mound up); also (figuratively) to expend (life, soul, complaint, money, etc.); intensively, to sprawl out","kjv_def":"cast (up), gush out, pour (out), shed(-der, out), slip."},
"H8211":{"lemma":"שֶׁפֶךְ","xlit":"shephek","pron":"sheh'-fek","derivation":"from H8210 (שָׁפַךְ);","strongs_def":"an emptying place, e.g. an ash-heap","kjv_def":"are poured out."},
"H8212":{"lemma":"שׇׁפְכָה","xlit":"shophkâh","pron":"shof-kaw'","derivation":"feminine of a derivative from H8210 (שָׁפַךְ);","strongs_def":"a pipe (for pouring forth, e.g. wine), i.e. the penis","kjv_def":"privy member."},
"H8213":{"lemma":"שָׁפֵל","xlit":"shâphêl","pron":"shaw-fale'","derivation":"a primitive root;","strongs_def":"to depress or sink (expectation figuratively, to humiliate, intransitive or transitive)","kjv_def":"abase, bring (cast, put) down, debase, humble (self), be (bring, lay, make, put) low(-er)."},
"H8214":{"lemma":"שְׁפַל","xlit":"shᵉphal","pron":"shef-al'","derivation":"(Aramaic) corresponding to H8213 (שָׁפֵל)","strongs_def":"{to depress or sink (expectation figuratively, to humiliate, intransitive or transitive)}","kjv_def":"abase, humble, put down, subdue."},
"H8215":{"lemma":"שְׁפַל","xlit":"shᵉphal","pron":"shef-al'","derivation":"(Aramaic) from H8214 (שְׁפַל);","strongs_def":"low","kjv_def":"basest."},
"H8216":{"lemma":"שֶׁפֶל","xlit":"shephel","pron":"shay'-fel","derivation":"from H8213 (שָׁפֵל);","strongs_def":"an humble rank","kjv_def":"low estate (place)."},
"H8217":{"lemma":"שָׁפָל","xlit":"shâphâl","pron":"shaw-fawl'","derivation":"from H8213 (שָׁפֵל);","strongs_def":"depressed, literally or figuratively","kjv_def":"base(-st), humble, low(-er, -ly)."},
"H8218":{"lemma":"שִׁפְלָה","xlit":"shiphlâh","pron":"shif-law'","derivation":"feminine of H8216 (שֶׁפֶל);","strongs_def":"depression","kjv_def":"low place."},
"H8219":{"lemma":"שְׁפֵלָה","xlit":"shᵉphêlâh","pron":"shef-ay-law'","derivation":"from H8213 (שָׁפֵל);","strongs_def":"Lowland, i.e. (with the article) the maritime slope of Palestine","kjv_def":"low country, (low) plain, vale(-ley)."},
"H8220":{"lemma":"שִׁפְלוּת","xlit":"shiphlûwth","pron":"shif-looth'","derivation":"from H8213 (שָׁפֵל);","strongs_def":"remissness","kjv_def":"idleness."},
"H8221":{"lemma":"שְׁפָם","xlit":"Shᵉphâm","pron":"shef-awm'","derivation":"probably from H8192 (שָׁפָה); bare spot;","strongs_def":"Shepham, a place in or near Palestine","kjv_def":"Shepham."},
"H8222":{"lemma":"שָׂפָם","xlit":"sâphâm","pron":"saw-fawm'","derivation":"from H8193 (שָׂפָה);","strongs_def":"the beard (as a lippiece)","kjv_def":"beard, (upper) lip."},
"H8223":{"lemma":"שָׁפָם","xlit":"Shâphâm","pron":"shaw-fawm'","derivation":"formed like H8221 (שְׁפָם); baldly;","strongs_def":"Shapham, an Israelite","kjv_def":"Shapham."},
"H8224":{"lemma":"שִׂפְמוֹת","xlit":"Siphmôwth","pron":"sif-moth'","derivation":"feminine plural of H8221 (שְׁפָם);","strongs_def":"Siphmoth, a place in Palestine","kjv_def":"Siphmoth."},
"H8225":{"lemma":"שִׁפְמִי","xlit":"Shiphmîy","pron":"shif-mee'","derivation":"patrial from H8221 (שְׁפָם);","strongs_def":"a Shiphmite or inhabitant of Shepham","kjv_def":"Shiphmite."},
"H8226":{"lemma":"שָׂפַן","xlit":"sâphan","pron":"saw-fan'","derivation":"a primitive root;","strongs_def":"to conceal (as a valuable)","kjv_def":"treasure."},
"H8227":{"lemma":"שָׁפָן","xlit":"shâphân","pron":"shaw-fawn'","derivation":"from H8226 (שָׂפַן);","strongs_def":"a species of rockrabbit (from its hiding), i.e. probably the hyrax","kjv_def":"coney."},
"H8228":{"lemma":"שֶׁפַע","xlit":"shephaʻ","pron":"sheh'-fah","derivation":"from an unused root meaning to abound;","strongs_def":"resources","kjv_def":"abundance."},
"H8229":{"lemma":"שִׁפְעָה","xlit":"shiphʻâh","pron":"shif-aw'","derivation":"feminine of H8228 (שֶׁפַע);","strongs_def":"copiousness","kjv_def":"abundance, company, multitude."},
"H8230":{"lemma":"שִׁפְעִי","xlit":"Shiphʻîy","pron":"shif-ee'","derivation":"from H8228 (שֶׁפַע); copious;","strongs_def":"Shiphi, an Israelite","kjv_def":"Shiphi."},
"H8231":{"lemma":"שָׁפַר","xlit":"shâphar","pron":"shaw-far'","derivation":"a primitive root;","strongs_def":"to glisten, i.e. (figuratively) be (causatively, make) fair","kjv_def":"[idiom] goodly."},
"H8232":{"lemma":"שְׁפַר","xlit":"shᵉphar","pron":"shef-ar'","derivation":"(Aramaic) corresponding to H8231 (שָׁפַר);","strongs_def":"to be beautiful","kjv_def":"be acceptable, please, [phrase] think good."},
"H8233":{"lemma":"שֶׁפֶר","xlit":"shepher","pron":"sheh'-fer","derivation":"from H8231 (שָׁפַר);","strongs_def":"beauty","kjv_def":"[idiom] goodly."},
"H8234":{"lemma":"שֶׁפֶר","xlit":"Shepher","pron":"sheh'-fer","derivation":"the same as H8233 (שֶׁפֶר);","strongs_def":"Shepher, a place in the Desert","kjv_def":"Shapper."},
"H8235":{"lemma":"שִׁפְרָה","xlit":"shiphrâh","pron":"shif-raw'","derivation":"from H8231 (שָׁפַר);","strongs_def":"brightness","kjv_def":"garnish."},
"H8236":{"lemma":"שִׁפְרָה","xlit":"Shiphrâh","pron":"shif-raw'","derivation":"the same as H8235 (שִׁפְרָה);","strongs_def":"Shiphrah, an Israelitess","kjv_def":"Shiphrah."},
"H8237":{"lemma":"שַׁפְרוּר","xlit":"shaphrûwr","pron":"shaf-roor'","derivation":"from H8231 (שָׁפַר);","strongs_def":"splendid, i.e. a tapestry or canopy","kjv_def":"royal pavilion."},
"H8238":{"lemma":"שְׁפַרְפַר","xlit":"shᵉpharphar","pron":"shef-ar-far'","derivation":"(Aramaic) from H8231 (שָׁפַר);","strongs_def":"the dawn (as brilliant with aurora)","kjv_def":"[idiom] very early in the morning."},
"H8239":{"lemma":"שָׁפַת","xlit":"shâphath","pron":"shaw-fath'","derivation":"a primitive root;","strongs_def":"to locate, i.e. (generally) hang on or (figuratively) establish, reduce","kjv_def":"bring, ordain, set on."},
"H8240":{"lemma":"שָׁפָת","xlit":"shâphâth","pron":"shaw-fawth'","derivation":"from H8239 (שָׁפַת);","strongs_def":"a (double) stall (for cattle); also a (two-pronged) hook (for flaying animals on)","kjv_def":"hook, pot."},
"H8241":{"lemma":"שֶׁצֶף","xlit":"shetseph","pron":"sheh'-tsef","derivation":"from H7857 (שָׁטַף) (for alliteration with H7110 (קֶצֶף));","strongs_def":"an outburst (of anger)","kjv_def":"little."},
"H8242":{"lemma":"שַׂק","xlit":"saq","pron":"sak","derivation":"from H8264 (שָׁקַק);","strongs_def":"properly, a mesh (as allowing a liquid to run through), i.e. coarse loose cloth or sacking (used in mourning and for bagging); hence, a bag (for grain, etc.)","kjv_def":"sack(-cloth, -clothes)."},
"H8243":{"lemma":"שָׁק","xlit":"shâq","pron":"shawk","derivation":"(Aramaic) corresponding to H7785 (שׁוֹק);","strongs_def":"the leg","kjv_def":"leg."},
"H8244":{"lemma":"שָׂקַד","xlit":"sâqad","pron":"saw-kad'","derivation":"a primitive root;","strongs_def":"to fasten","kjv_def":"bind."},
"H8245":{"lemma":"שָׁקַד","xlit":"shâqad","pron":"shaw-kad'","derivation":"a primitive root;","strongs_def":"to be alert, i.e. sleepless; hence to be on the lookout (whether for good or ill)","kjv_def":"hasten, remain, wake, watch (for)."},
"H8246":{"lemma":"שָׁקַד","xlit":"shâqad","pron":"shaw-kad'","derivation":"a denominative from H8247 (שָׁקֵד);","strongs_def":"to be (intensively, make) almond-shaped","kjv_def":"make like (unto, after the fashion of) almonds."},
"H8247":{"lemma":"שָׁקֵד","xlit":"shâqêd","pron":"shaw-kade'","derivation":"from H8245 (שָׁקַד);","strongs_def":"the almond (tree or nut; as being the earliest in bloom)","kjv_def":"almond (tree)."},
"H8248":{"lemma":"שָׁקָה","xlit":"shâqâh","pron":"shaw-kaw'","derivation":"a primitive root;","strongs_def":"to quaff, i.e. (causatively) to irrigate or furnish apotion to","kjv_def":"cause to (give, give to, let, make to) drink, drown, moisten, water. See H7937 (שָׁכַר), H8354 (שָׁתָה)."},
"H8249":{"lemma":"שִׁקֻּו","xlit":"shiqquv","pron":"shif-koov'","derivation":"from H8248 (שָׁקָה);","strongs_def":"(plural collective) a draught","kjv_def":"drink."},
"H8250":{"lemma":"שִׁקּוּי","xlit":"shiqqûwy","pron":"shik-koo'-ee","derivation":"from H8248 (שָׁקָה);","strongs_def":"a beverage; moisture, i.e. (figuratively) refreshment","kjv_def":"drink, marrow."},
"H8251":{"lemma":"שִׁקּוּץ","xlit":"shiqqûwts","pron":"shik-koots'","derivation":"or שִׁקֻּץ; from H8262 (שָׁקַץ);","strongs_def":"disgusting, i.e. filthy; especially idolatrous or (concretely) an idol","kjv_def":"abominable filth (idol, -ation), detestable (thing)."},
"H8252":{"lemma":"שָׁקַט","xlit":"shâqaṭ","pron":"shaw-kat'","derivation":"a primitive root;","strongs_def":"to repose (usually figurative)","kjv_def":"appease, idleness, (at, be at, be in, give) quiet(-ness), (be at, be in, give, have, take) rest, settle, be still."},
"H8253":{"lemma":"שֶׁקֶט","xlit":"sheqeṭ","pron":"sheh'-ket","derivation":"from H8252 (שָׁקַט);","strongs_def":"tranquillity","kjv_def":"quietness."},
"H8254":{"lemma":"שָׁקַל","xlit":"shâqal","pron":"shaw-kal'","derivation":"a primitive root;","strongs_def":"to suspend or poise (especially in trade)","kjv_def":"pay, receive(-r), spend, [idiom] throughly, weigh."},
"H8255":{"lemma":"שֶׁקֶל","xlit":"sheqel","pron":"sheh'-kel","derivation":"from H8254 (שָׁקַל);","strongs_def":"probably a weight; used as a commercial standard","kjv_def":"shekel."},
"H8256":{"lemma":"שִׁקְמָה","xlit":"shiqmâh","pron":"shik-maw'","derivation":"or (feminine) שִׁקְמָה; of uncertain derivation;","strongs_def":"a sycamore (usually the tree)","kjv_def":"sycamore (fruit, tree)."},
"H8257":{"lemma":"שָׁקַע","xlit":"shâqaʻ","pron":"shaw-kah'","derivation":"(abbreviated Amos 8:8); a primitive root;","strongs_def":"to subside; by implication, to be overflowed, cease; causatively, to abate, subdue","kjv_def":"make deep, let down, drown, quench, sink."},
"H8258":{"lemma":"שְׁקַעְרוּרָה","xlit":"shᵉqaʻrûwrâh","pron":"shek-ah-roo-raw'","derivation":"from H8257 (שָׁקַע);","strongs_def":"a depression","kjv_def":"hollow strake."},
"H8259":{"lemma":"שָׁקַף","xlit":"shâqaph","pron":"shaw-kaf'","derivation":"a primitive root;","strongs_def":"properly, to lean out (of a window), i.e. (by implication) peep or gaze (passively, be a spectacle)","kjv_def":"appear, look (down, forth, out)."},
"H8260":{"lemma":"שֶׁקֶף","xlit":"sheqeph","pron":"sheh'-kef","derivation":"from H8259 (שָׁקַף);","strongs_def":"a loophole (for looking out), to admit light and air","kjv_def":"window."},
"H8261":{"lemma":"שָׁקֻף","xlit":"shâquph","pron":"shaw-koof'","derivation":"passive participle of H8259 (שָׁקַף); (compare H8260 (שֶׁקֶף))","strongs_def":"an embrasure or opening  with bevelled jam","kjv_def":"light, window."},
"H8262":{"lemma":"שָׁקַץ","xlit":"shâqats","pron":"shaw-kats'","derivation":"a primitive root;","strongs_def":"to be filthy, i.e. (intensively) to loathe, pollute","kjv_def":"abhor, make abominable, have in abomination, detest, [idiom] utterly."},
"H8263":{"lemma":"שֶׁקֶץ","xlit":"sheqets","pron":"sheh'-kets","derivation":"from H8262 (שָׁקַץ);","strongs_def":"filth, i.e. (figuratively and specifically) an idolatrous object","kjv_def":"abominable(-tion)."},
"H8264":{"lemma":"שָׁקַק","xlit":"shâqaq","pron":"shaw-kak'","derivation":"a primitive root;","strongs_def":"to course (like a beast of prey); by implication, to seek greedily","kjv_def":"have appetite, justle one against another, long, range, run (to and fro)."},
"H8265":{"lemma":"שָׂקַר","xlit":"sâqar","pron":"saw-kar'","derivation":"a primitive root;","strongs_def":"to ogle, i.e. blink coquettishly","kjv_def":"wanton."},
"H8266":{"lemma":"שָׁקַר","xlit":"shâqar","pron":"shaw-kar'","derivation":"a primitive root;","strongs_def":"to cheat, i.e. be untrue (usually in words)","kjv_def":"fail, deal falsely, lie."},
"H8267":{"lemma":"שֶׁקֶר","xlit":"sheqer","pron":"sheh'-ker","derivation":"from H8266 (שָׁקַר);","strongs_def":"an untruth; by implication, a sham (often adverbial)","kjv_def":"without a cause, deceit(-ful), false(-hood, -ly), feignedly, liar, [phrase] lie, lying, vain (thing), wrongfully."},
"H8268":{"lemma":"שֹׁקֶת","xlit":"shôqeth","pron":"sho'-keth","derivation":"from H8248 (שָׁקָה);","strongs_def":"a trough (for watering)","kjv_def":"trough."},
"H8269":{"lemma":"שַׂר","xlit":"sar","pron":"sar","derivation":"from H8323 (שָׂרַר);","strongs_def":"a head person (of any rank or class)","kjv_def":"captain (that had rule), chief (captain), general, governor, keeper, lord,(-task-)master, prince(-ipal), ruler, steward."},
"H8270":{"lemma":"שֹׁר","xlit":"shôr","pron":"shore","derivation":"from H8324 (שָׁרַר); (compare H8306 (שָׁרִיר))","strongs_def":"a string (as twisted ), i.e. (specifically) the umbilical cord (also figuratively, as the centre of strength)","kjv_def":"navel."},
"H8271":{"lemma":"שְׁרֵא","xlit":"shᵉrêʼ","pron":"sher-ay'","derivation":"(Aramaic) a root corresponding to that of H8293 (שֵׁרוּת);","strongs_def":"to free, separate; figuratively, to unravel, commence; by implication (of unloading beasts) to reside","kjv_def":"begin, dissolve, dwell, loose."},
"H8272":{"lemma":"שַׁרְאֶצֶר","xlit":"Sharʼetser","pron":"shar-eh'-tser","derivation":"of foreign derivation;","strongs_def":"Sharetser, the name of an Assyrian and an Israelite","kjv_def":"Sharezer."},
"H8273":{"lemma":"שָׁרָב","xlit":"shârâb","pron":"shaw-rawb'","derivation":"from an unused root meaning to glare;","strongs_def":"quivering glow (of the air), expectation. the mirage","kjv_def":"heat, parched ground."},
"H8274":{"lemma":"שֵׁרֵבְיָה","xlit":"Shêrêbᵉyâh","pron":"shay-rayb-yaw'","derivation":"from H8273 (שָׁרָב) and H3050 (יָהּ); Jah has brought heat;","strongs_def":"Sherebjah, the name of two Israelites","kjv_def":"Sherebiah."},
"H8275":{"lemma":"שַׁרְבִיט","xlit":"sharbîyṭ","pron":"shar-beet'","derivation":"for H7626 (שֵׁבֶט);","strongs_def":"a rod of empire","kjv_def":"sceptre."},
"H8276":{"lemma":"שָׂרַג","xlit":"sârag","pron":"saw-rag'","derivation":"a primitive root;","strongs_def":"to intwine","kjv_def":"wrap together, wreath."},
"H8277":{"lemma":"שָׂרַד","xlit":"sârad","pron":"saw-rad'","derivation":"a primitive root; (compare H8279 (שֶׂרֶד))","strongs_def":"properly, to puncture, i.e. (figuratively through the idea of slipping out) to escape or survive","kjv_def":"remain."},
"H8278":{"lemma":"שְׂרָד","xlit":"sᵉrâd","pron":"ser-awd'","derivation":"from H8277 (שָׂרַד);","strongs_def":"stitching (as pierced with a needle)","kjv_def":"service."},
"H8279":{"lemma":"שֶׂרֶד","xlit":"sered","pron":"seh'-red","derivation":"from H8277 (שָׂרַד);","strongs_def":"a (carpenter's) scribing-awl (for pricking or scratching measurements)","kjv_def":"line."},
"H8280":{"lemma":"שָׂרָה","xlit":"sârâh","pron":"saw-raw'","derivation":"a primitive root;","strongs_def":"to prevail","kjv_def":"have power (as a prince)."},
"H8281":{"lemma":"שָׁרָה","xlit":"shârâh","pron":"shaw-raw'","derivation":"a primitive root;","strongs_def":"to free","kjv_def":"direct."},
"H8282":{"lemma":"שָׂרָה","xlit":"sârâh","pron":"saw-raw'","derivation":"feminine of H8269 (שַׂר);","strongs_def":"a mistress, i.e. female noble","kjv_def":"lady, princess, queen."},
"H8283":{"lemma":"שָׂרָה","xlit":"Sârâh","pron":"saw-raw'","derivation":"the same as H8282 (שָׂרָה);","strongs_def":"Sarah, Abraham's wife","kjv_def":"Sarah."},
"H8284":{"lemma":"שָׁרָה","xlit":"shârâh","pron":"shaw-raw'","derivation":"probably feminine of H7791 (שׁוּר);","strongs_def":"a fortification (literally or figuratively)","kjv_def":"sing (by mistake for H7891 (שִׁיר)), wall."},
"H8285":{"lemma":"שֵׁרָה","xlit":"shêrâh","pron":"shay-raw'","derivation":"from H8324 (שָׁרַר) in its original sense of pressing;","strongs_def":"a wrist-band (as compact or clasping)","kjv_def":"bracelet."},
"H8286":{"lemma":"שְׂרוּג","xlit":"Sᵉrûwg","pron":"ser-oog'","derivation":"from H8276 (שָׂרַג); tendril;","strongs_def":"Serug, a postdiluvian patriarch","kjv_def":"Serug."},
"H8287":{"lemma":"שָׁרוּחֶן","xlit":"Shârûwchen","pron":"shaw-roo-khen'","derivation":"probably from H8281 (שָׁרָה) (in the sense of dwelling (compare H8271 (שְׁרֵא)) and H2580 (חֵן); abode of pleasure;","strongs_def":"Sharuchen, a place in Palestine","kjv_def":"Sharuhen."},
"H8288":{"lemma":"שְׂרוֹךְ","xlit":"sᵉrôwk","pron":"ser-oke'","derivation":"from H8308 (שָׂרַךְ);","strongs_def":"a thong (as laced or tied)","kjv_def":"(shoe-)latchet."},
"H8289":{"lemma":"שָׁרוֹן","xlit":"Shârôwn","pron":"shaw-rone'","derivation":"probably abridged from H3474 (יָשַׁר);","strongs_def":"plain, Sharon, the name of a place in Palestine","kjv_def":"Lasharon, Sharon."},
"H8290":{"lemma":"שָׁרוֹנִי","xlit":"Shârôwnîy","pron":"shaw-ro-nee'","derivation":"patrial from H8289 (שָׁרוֹן);","strongs_def":"a Sharonite or inhabitant of Sharon","kjv_def":"Sharonite."},
"H8291":{"lemma":"שָׂרוּק","xlit":"sârûwq","pron":"sar-ook'","derivation":"passive participle from the same as H8321 (שֹׂרֵק);","strongs_def":"a grapevine","kjv_def":"principal plant. See H8320 (שָׂרֻק), H8321 (שֹׂרֵק)."},
"H8292":{"lemma":"שְׁרוּקָה","xlit":"shᵉrûwqâh","pron":"sher-oo-kaw'","derivation":"or (by permutation) שְׁרִיקָה; feminine passive participle of H8319 (שָׁרַק);","strongs_def":"a whistling (in scorn); by analogy, a piping","kjv_def":"bleating, hissing."},
"H8293":{"lemma":"שֵׁרוּת","xlit":"shêrûwth","pron":"shay-rooth'","derivation":"from H8281 (שָׁרָה) abbreviated;","strongs_def":"freedom","kjv_def":"remnant."},
"H8294":{"lemma":"שֶׂרַח","xlit":"Serach","pron":"seh'-rakh","derivation":"by permutation for H5629 (סֶרַח); superfluity;","strongs_def":"Serach, an Israelitess","kjv_def":"Sarah, Serah."},
"H8295":{"lemma":"שָׂרַט","xlit":"sâraṭ","pron":"saw-rat'","derivation":"a primitive root;","strongs_def":"to gash","kjv_def":"cut in pieces, make (cuttings) pieces."},
"H8296":{"lemma":"שֶׂרֶט","xlit":"sereṭ","pron":"seh'-ret","derivation":"and שָׂרֶטֶת; from H8295 (שָׂרַט);","strongs_def":"an incision","kjv_def":"cutting."},
"H8297":{"lemma":"שָׂרַי","xlit":"Sâray","pron":"saw-rah'-ee","derivation":"from H8269 (שַׂר); dominative;","strongs_def":"Sarai, the wife of Abraham","kjv_def":"Sarai."},
"H8298":{"lemma":"שָׁרַי","xlit":"Shâray","pron":"shaw-rah'-ee","derivation":"probably from H8324 (שָׁרַר); hostile;","strongs_def":"Sharay, an Israelite","kjv_def":"Sharai."},
"H8299":{"lemma":"שָׂרִיג","xlit":"sârîyg","pron":"saw-reeg'","derivation":"from H8276 (שָׂרַג);","strongs_def":"a tendril (as entwining)","kjv_def":"branch."},
"H8300":{"lemma":"שָׂרִיד","xlit":"sârîyd","pron":"saw-reed'","derivation":"from H8277 (שָׂרַד);","strongs_def":"a survivor","kjv_def":"[idiom] alive, left, remain(-ing), remnant, rest."},
"H8301":{"lemma":"שָׂרִיד","xlit":"Sârîyd","pron":"suw-reed'","derivation":"the same as H8300 (שָׂרִיד);","strongs_def":"Sarid, a place in Palestine","kjv_def":"Sarid."},
"H8302":{"lemma":"שִׁרְיוֹן","xlit":"shiryôwn","pron":"shir-yone'","derivation":"or שִׁרְיֹן; and שִׁרְיָן; also (feminine) שִׁרְיָה; and שִׁרְיֹנָה; from H8281 (שָׁרָה) in the original sense of turning;","strongs_def":"a corslet (as if twisted)","kjv_def":"breastplate, coat of mail, habergeon, harness. See H5630 (סִרְיֹן)."},
"H8303":{"lemma":"שִׁרְיוֹן","xlit":"Shiryôwn","pron":"shir-yone'","derivation":"and שִׂרְיֹן; the same as H8304 (שְׂרָיָה) (i.e. sheeted with snow);","strongs_def":"Shirjon or Sirjon, a peak of the Lebanon","kjv_def":"Sirion."},
"H8304":{"lemma":"שְׂרָיָה","xlit":"Sᵉrâyâh","pron":"ser-aw-yaw'","derivation":"or שְׂרָיָהוּ; from H8280 (שָׂרָה) and H3050 (יָהּ); Jah has prevailed;","strongs_def":"Serajah, the name of nine Israelites","kjv_def":"Seraiah."},
"H8305":{"lemma":"שְׂרִיקָה","xlit":"sᵉrîyqâh","pron":"ser-ee-kaw'","derivation":"from the same as H8321 (שֹׂרֵק) in the original sense of piercing;","strongs_def":"hetchelling (or combing flax), i.e. (concretely) tow (by extension, linen cloth)","kjv_def":"fine."},
"H8306":{"lemma":"שָׁרִיר","xlit":"shârîyr","pron":"shaw-reer'","derivation":"from H8324 (שָׁרַר) in the original sense as in H8270 (שֹׁר) (compare H8326 (שֹׁרֶר));","strongs_def":"a cord, i.e. (by analogy) sinew","kjv_def":"navel."},
"H8307":{"lemma":"שְׁרִירוּת","xlit":"shᵉrîyrûwth","pron":"sher-ee-rooth'","derivation":"from H8324 (שָׁרַר) in the sense of twisted, i.e. firm;","strongs_def":"obstinacy","kjv_def":"imagination, lust."},
"H8308":{"lemma":"שָׂרַךְ","xlit":"sârak","pron":"saw-rak'","derivation":"a primitive root;","strongs_def":"to interlace","kjv_def":"traverse."},
"H8309":{"lemma":"שְׁרֵמָה","xlit":"shᵉrêmâh","pron":"sher-ay-maw'","derivation":"probably by an orthographical error for H7709 (שְׁדֵמָה);","strongs_def":"a common","kjv_def":"field."},
"H8310":{"lemma":"שַׂרְסְכִים","xlit":"Sarçᵉkîym","pron":"sar-seh-keem'","derivation":"of foreign derivation;","strongs_def":"Sarsekim, a Babylonian general","kjv_def":"Sarsechim."},
"H8311":{"lemma":"שָׂרַע","xlit":"sâraʻ","pron":"saw-rah'","derivation":"a primitive root;","strongs_def":"to prolong, i.e. (reflex) be deformed by excess of members","kjv_def":"stretch out self, (have any) superfluous thing."},
"H8312":{"lemma":"שַׂרְעַף","xlit":"sarʻaph","pron":"sar-af'","derivation":"for H5587 (סָעִף);","strongs_def":"cogitation","kjv_def":"thought."},
"H8313":{"lemma":"שָׂרַף","xlit":"sâraph","pron":"saw-raf'","derivation":"a primitive root;","strongs_def":"to be (causatively, set) on fire","kjv_def":"(cause to, make a) burn((-ing), up) kindle, [idiom] utterly."},
"H8314":{"lemma":"שָׂרָף","xlit":"sârâph","pron":"saw-rawf'","derivation":"from H8313 (שָׂרַף);","strongs_def":"burning, i.e. (figuratively) poisonous (serpent); specifically, a saraph or symbolical creature (from their copper color)","kjv_def":"fiery (serpent), seraph."},
"H8315":{"lemma":"שָׂרָף","xlit":"Sârâph","pron":"saw-raf'","derivation":"the same as H8314 (שָׂרָף);","strongs_def":"Saraph, an Israelite","kjv_def":"Saraph."},
"H8316":{"lemma":"שְׂרֵפָה","xlit":"sᵉrêphâh","pron":"ser-ay-faw'","derivation":"from H8313 (שָׂרַף);","strongs_def":"cremation","kjv_def":"burning."},
"H8317":{"lemma":"שָׁרַץ","xlit":"shârats","pron":"shaw-rats'","derivation":"a primitive root;","strongs_def":"to wriggle, i.e. (by implication) swarm or abound","kjv_def":"breed (bring forth, increase) abundantly (in abundance), creep, move."},
"H8318":{"lemma":"שֶׁרֶץ","xlit":"sherets","pron":"sheh'-rets","derivation":"from H8317 (שָׁרַץ);","strongs_def":"a swarm, i.e. active mass of minute animals","kjv_def":"creep(-ing thing), move(-ing creature)."},
"H8319":{"lemma":"שָׁרַק","xlit":"shâraq","pron":"shaw-rak'","derivation":"a primitive root;","strongs_def":"properly, to be shrill, i.e. to whistle or hiss (as a call or in scorn)","kjv_def":"hiss."},
"H8320":{"lemma":"שָׂרֻק","xlit":"sâruq","pron":"saw-rook'","derivation":"from H8319 (שָׁרַק);","strongs_def":"bright red (as piercing to the sight), i.e. bay","kjv_def":"speckled. See H8291 (שָׂרוּק)."},
"H8321":{"lemma":"שֹׂרֵק","xlit":"sôrêq","pron":"so-rake'","derivation":"or שׂוֹרֵק; and (feminine) שֹׂרֵקָה; from H8319 (שָׁרַק) in the sense of redness (compare H8320 (שָׂרֻק));","strongs_def":"a vine stock (properly, one yielding purple grapes, the richest variety)","kjv_def":"choice(-st, noble) wine. Compare H8291 (שָׂרוּק)."},
"H8322":{"lemma":"שְׁרֵקָה","xlit":"shᵉrêqâh","pron":"sher-ay-kaw'","derivation":"from H8319 (שָׁרַק);","strongs_def":"a derision","kjv_def":"hissing."},
"H8323":{"lemma":"שָׂרַר","xlit":"sârar","pron":"saw-rar'","derivation":"a primitive root;","strongs_def":"to have (transitively, exercise; reflexively, get) dominion","kjv_def":"[idiom] altogether, make self a prince, (bear) rule."},
"H8324":{"lemma":"שָׁרַר","xlit":"shârar","pron":"shaw-rar'","derivation":"a primitive root;","strongs_def":"to be hostile (only active participle an opponent)","kjv_def":"enemy."},
"H8325":{"lemma":"שָׁרַר","xlit":"Shârar","pron":"shaw-rawr'","derivation":"from H8324 (שָׁרַר); hostile;","strongs_def":"Sharar, an Israelite","kjv_def":"Sharar."},
"H8326":{"lemma":"שֹׁרֶר","xlit":"shôrer","pron":"sho'-rer","derivation":"from H8324 (שָׁרַר) in the sense of twisting (compare H8270 (שֹׁר));","strongs_def":"the umbilical cord, i.e. (by extension) a bodice","kjv_def":"navel."},
"H8327":{"lemma":"שָׁרַשׁ","xlit":"shârash","pron":"shaw-rash'","derivation":"a primitive root;","strongs_def":"to root, i.e. strike into the soil, or (by implication) to pluck from it","kjv_def":"(take, cause to take) root (out)."},
"H8328":{"lemma":"שֶׁרֶשׁ","xlit":"sheresh","pron":"sheh'-resh","derivation":"from H8327 (שָׁרַשׁ);","strongs_def":"a root (literally or figuratively)","kjv_def":"bottom, deep, heel, root."},
"H8329":{"lemma":"שֶׁרֶשׁ","xlit":"Sheresh","pron":"sheh'-resh","derivation":"the same as H8328 (שֶׁרֶשׁ);","strongs_def":"Sheresh, an Israelite","kjv_def":"Sharesh."},
"H8330":{"lemma":"שֹׁרֶשׁ","xlit":"shôresh","pron":"sho'-resh","derivation":"(Aramaic) corresponding to H8328 (שֶׁרֶשׁ)","strongs_def":"{a root (literally or figuratively)}","kjv_def":"root."},
"H8331":{"lemma":"שַׁרְשָׁה","xlit":"sharshâh","pron":"shar-shaw'","derivation":"from H8327 (שָׁרַשׁ);","strongs_def":"a chain (as rooted, i.e. linked)","kjv_def":"chain. Compare H8333 (שַׁרְשְׁרָה)."},
"H8332":{"lemma":"שְׁרֹשׁוּ","xlit":"shᵉrôshûw","pron":"sher-o-shoo'","derivation":"(Aramaic) from a root corresponding to H8327 (שָׁרַשׁ);","strongs_def":"eradication, i.e. (figuratively) exile","kjv_def":"banishment."},
"H8333":{"lemma":"שַׁרְשְׁרָה","xlit":"sharshᵉrâh","pron":"shar-sher-aw'","derivation":"from H8327 (שָׁרַשׁ) (compare H8331 (שַׁרְשָׁה));","strongs_def":"a chain; (architectural) probably a garland","kjv_def":"chain."},
"H8334":{"lemma":"שָׁרַת","xlit":"shârath","pron":"shaw-rath'","derivation":"a primitive root;","strongs_def":"to attend as a menial or worshipper; figuratively, to contribute to","kjv_def":"minister (unto), (do) serve(-ant, -ice, -itor), wait on."},
"H8335":{"lemma":"שָׁרֵת","xlit":"shârêth","pron":"shaw-rayth'","derivation":"infinitive of H8334 (שָׁרַת);","strongs_def":"service (in the Temple)","kjv_def":"minister(-ry)."},
"H8336":{"lemma":"שֵׁשׁ","xlit":"shêsh","pron":"shaysh","derivation":"or (for alliteration with H4897 (מֶשִׁי)) שְׁשִׁי; for H7893 (שַׁיִשׁ);","strongs_def":"bleached stuff, i.e. white linen or (by analogy) marble","kjv_def":"[idiom] blue, fine (twined) linen, marble, silk."},
"H8337":{"lemma":"שֵׁשׁ","xlit":"shêsh","pron":"shaysh","derivation":"masculine שִׁשָּׁה; a primitive number; compare H7797 (שׂוּשׂ)","strongs_def":"six (as an overplus  beyond five or the fingers of the hand); as ord. sixth","kjv_def":"six(-teen, -teenth), sixth."},
"H8338":{"lemma":"שָׁשָׁא","xlit":"shâshâʼ","pron":"shaw-shaw'","derivation":"a primitive root; apparently,","strongs_def":"to annihilate","kjv_def":"leave by the sixth part (by confusion with H8341 (שָׁשָׁה))."},
"H8339":{"lemma":"שֵׁשְׁבַּצַּר","xlit":"Shêshᵉbatstsar","pron":"shaysh-bats-tsar'","derivation":"of foreign derivation;","strongs_def":"Sheshbatstsar, Zerubbabel's Persian name","kjv_def":"Sheshbazzar."},
"H8340":{"lemma":"שֵׁשְׁבַּצַּר","xlit":"Shêshᵉbatstsar","pron":"shaysh-bats-tsar'","derivation":"(Aramaic) corresponding to H8339 (שֵׁשְׁבַּצַּר)","strongs_def":"{Sheshbatstsar, Zerubbabel's Persian name}","kjv_def":"Sheshbazzar."},
"H8341":{"lemma":"שָׁשָׁה","xlit":"shâshâh","pron":"shaw-shaw'","derivation":"a denominative from H8337 (שֵׁשׁ);","strongs_def":"to sixth or divide into sixths","kjv_def":"give the sixth participle"},
"H8342":{"lemma":"שָׂשׂוֹן","xlit":"sâsôwn","pron":"saw-sone'","derivation":"or שָׂשֹׂן; from H7797 (שׂוּשׂ);","strongs_def":"cheerfulness; specifically, welcome","kjv_def":"gladness, joy, mirth, rejoicing."},
"H8343":{"lemma":"שָׁשַׁי","xlit":"Shâshay","pron":"shaw-shah'-ee","derivation":"perhaps from H8336 (שֵׁשׁ); whitish;","strongs_def":"Shashai, an Israelite","kjv_def":"Shashai."},
"H8344":{"lemma":"שֵׁשַׁי","xlit":"Shêshay","pron":"shay-shah'-ee","derivation":"probably for H8343 (שָׁשַׁי);","strongs_def":"Sheshai, a Canaanite","kjv_def":"Sheshai."},
"H8345":{"lemma":"שִׁשִּׁי","xlit":"shishshîy","pron":"shish-shee'","derivation":"from H8337 (שֵׁשׁ);","strongs_def":"sixth, ord. or (feminine) fractional","kjv_def":"sixth (part)."},
"H8346":{"lemma":"שִׁשִּׁים","xlit":"shishshîym","pron":"shish-sheem'","derivation":"multiple of H8337 (שֵׁשׁ);","strongs_def":"sixty","kjv_def":"sixty, three score."},
"H8347":{"lemma":"שֵׁשַׁךְ","xlit":"Shêshak","pron":"shay-shak'","derivation":"of foreign derivation;","strongs_def":"Sheshak, a symbolic name of Babylon","kjv_def":"Sheshach."},
"H8348":{"lemma":"שֵׁשָׁן","xlit":"Shêshân","pron":"shay-shawn'","derivation":"perhaps for H7799 (שׁוּשַׁן); lily;","strongs_def":"Sheshan, an Israelite","kjv_def":"Sheshan."},
"H8349":{"lemma":"שָׁשַׁק","xlit":"Shâshaq","pron":"shaw-shak'","derivation":"probably from the base of H7785 (שׁוֹק); pedestrian;","strongs_def":"Shashak, an Israelite","kjv_def":"Shashak."},
"H8350":{"lemma":"שָׁשַׁר","xlit":"shâshar","pron":"shaw-shar'","derivation":"perhaps from the base of H8324 (שָׁרַר) in the sense of that of H8320 (שָׂרֻק);","strongs_def":"red ochre (from its piercing color)","kjv_def":"vermillion."},
"H8351":{"lemma":"שֵׁת","xlit":"shêth","pron":"shayth","derivation":"(Numbers from H7582 (שָׁאָה);","strongs_def":"tumult","kjv_def":"Sheth."},
"H8352":{"lemma":"שֵׁת","xlit":"Shêth","pron":"shayth","derivation":"from H7896 (שִׁית); put, i.e. substituted;","strongs_def":"Sheth, third son of Adam","kjv_def":"Seth, Sheth."},
"H8353":{"lemma":"שֵׁת","xlit":"shêth","pron":"shayth","derivation":"(Aramaic) or שִׁת; (Aramaic), corresponding to H8337 (שֵׁשׁ);","strongs_def":"{six}","kjv_def":"six(-th)."},
"H8354":{"lemma":"שָׁתָה","xlit":"shâthâh","pron":"shaw-thaw'","derivation":"a primitive root;","strongs_def":"to imbibe (literally or figuratively)","kjv_def":"[idiom] assuredly, banquet, [idiom] certainly, drink(-er, -ing), drunk ([idiom] -ard), surely. (Prop. intensive of H8248 (שָׁקָה).)"},
"H8355":{"lemma":"שְׁתָה","xlit":"shᵉthâh","pron":"sheth-aw'","derivation":"(Aramaic) corresponding to H8354 (שָׁתָה)","strongs_def":"{to imbibe (literally or figuratively)}","kjv_def":"drink."},
"H8356":{"lemma":"שָׁתָה","xlit":"shâthâh","pron":"shaw-thaw'","derivation":"from H7896 (שִׁית);","strongs_def":"a basis, i.e. (figuratively) political or moral support","kjv_def":"foundation, purpose."},
"H8357":{"lemma":"שֵׁתָה","xlit":"shêthâh","pron":"shay-thaw'","derivation":"from H7896 (שִׁית);","strongs_def":"the seat (of the person)","kjv_def":"buttock."},
"H8358":{"lemma":"שְׁתִי","xlit":"shᵉthîy","pron":"sheth-ee'","derivation":"from H8354 (שָׁתָה);","strongs_def":"intoxicaion","kjv_def":"drunkenness."},
"H8359":{"lemma":"שְׁתִי","xlit":"shᵉthîy","pron":"sheth-ee'","derivation":"from H7896 (שִׁית);","strongs_def":"a fixture, i.e. the warp in weaving","kjv_def":"warp."},
"H8360":{"lemma":"שְׁתִיָּה","xlit":"shᵉthîyâh","pron":"sheth-ee-yaw'","derivation":"feminine of H8358 (שְׁתִי);","strongs_def":"potation","kjv_def":"drinking."},
"H8361":{"lemma":"שִׁתִּין","xlit":"shittîyn","pron":"shit-teen'","derivation":"(Aramaic) corresponding to H8346 (שִׁשִּׁים) (compare H8353 (שֵׁת));","strongs_def":"sixty","kjv_def":"threescore."},
"H8362":{"lemma":"שָׁתַל","xlit":"shâthal","pron":"shaw-thal'","derivation":"a primitive root;","strongs_def":"to transplant","kjv_def":"plant."},
"H8363":{"lemma":"שְׁתִל","xlit":"shᵉthil","pron":"sheth-eel'","derivation":"from H8362 (שָׁתַל);","strongs_def":"a sprig (as if transplanted), i.e. sucker","kjv_def":"plant."},
"H8364":{"lemma":"שֻׁתַלְחִי","xlit":"Shuthalchîy","pron":"shoo-thal-kee'","derivation":"patronymically from H7803 (שׁוּתֶלַח);","strongs_def":"a Shuthalchite (collectively) or descendants of Shuthelach","kjv_def":"Shuthalhites."},
"H8365":{"lemma":"שָׁתַם","xlit":"shâtham","pron":"shaw-tham'","derivation":"a primitive root;","strongs_def":"to unveil (figuratively)","kjv_def":"be open."},
"H8366":{"lemma":"שָׁתַן","xlit":"shâthan","pron":"shaw-than'","derivation":"a primitive root;","strongs_def":"(causatively) to make water, i.e. urinate","kjv_def":"piss."},
"H8367":{"lemma":"שָׁתַק","xlit":"shâthaq","pron":"shaw-thak'","derivation":"a primitive root;","strongs_def":"to subside","kjv_def":"be calm, cease, be quiet."},
"H8368":{"lemma":"שָׂתַר","xlit":"sâthar","pron":"saw-thar'","derivation":"a primitive root;","strongs_def":"to break out (as an eruption)","kjv_def":"have in (one's) secret parts."},
"H8369":{"lemma":"שֵׁתָר","xlit":"Shêthâr","pron":"shay-thawr'","derivation":"of foreign derivation;","strongs_def":"Shethar, a Persian satrap","kjv_def":"Shethar."},
"H8370":{"lemma":"שְׁתַר בּוֹזְנַי","xlit":"Shᵉthar Bôwzᵉnay","pron":"sheth-ar' bo-zen-ah'-ee","derivation":"of foreign derivation;","strongs_def":"Shethar-Bozenai, a Persian officer","kjv_def":"Shetharboznai."},
"H8371":{"lemma":"שָׁתַת","xlit":"shâthath","pron":"shaw-thath'","derivation":"a primitive root;","strongs_def":"to place, i.e. array; reflex. to lie","kjv_def":"be laid, set. t"},
"H8372":{"lemma":"תָּא","xlit":"tâʼ","pron":"taw","derivation":"and (feminine) תָּאָה; (Ezekiel 40:12), from (the base of) H8376 (תָּאָה);","strongs_def":"a room (as circumscribed)","kjv_def":"(little) chamber."},
"H8373":{"lemma":"תָּאַב","xlit":"tâʼab","pron":"taw-ab'","derivation":"a primitive root;","strongs_def":"to desire","kjv_def":"long."},
"H8374":{"lemma":"תָּאַב","xlit":"tâʼab","pron":"taw-ab'","derivation":"a primitive root (probably identical with H8373 (תָּאַב) through the idea of puffing disdainfully at; compare H340 (אָיַב));","strongs_def":"to loathe (morally)","kjv_def":"abhor."},
"H8375":{"lemma":"תַּאֲבָה","xlit":"taʼăbâh","pron":"tah-ab-aw'","derivation":"from H8374 (תָּאַב) (compare H15 (אָבֶה));","strongs_def":"desire","kjv_def":"longing."},
"H8376":{"lemma":"תָּאָה","xlit":"tâʼâh","pron":"taw-aw'","derivation":"a primitive root;","strongs_def":"to mark off, i.e. (intensively) designate","kjv_def":"point out."},
"H8377":{"lemma":"תְּאוֹ","xlit":"tᵉʼôw","pron":"teh-o'","derivation":"and תּוֹא; (the original form), from H8376 (תָּאָה);","strongs_def":"a species of antelope (probably from the white stripe on the cheek)","kjv_def":"wild bull (ox)."},
"H8378":{"lemma":"תַּאֲוָה","xlit":"taʼăvâh","pron":"tah-av-aw'","derivation":"from H183 (אָוָה) (abbreviated);","strongs_def":"a longing; by implication, a delight (subjectively, satisfaction, objectively, a charm)","kjv_def":"dainty, desire, [idiom] exceedingly, [idiom] greedily, lust(ing), pleasant. See also H6914 (קִבְרוֹת הַתַּאֲוָה)."},
"H8379":{"lemma":"תַּאֲוָה","xlit":"taʼăvâh","pron":"tah-av-aw'","derivation":"from H8376 (תָּאָה);","strongs_def":"a limit, i.e. full extent","kjv_def":"utmost bound."},
"H8380":{"lemma":"תָּאוֹם","xlit":"tâʼôwm","pron":"taw-ome'","derivation":"or תָּאֹם; from H8382 (תָּאַם);","strongs_def":"a twin (in plural only), literally or figuratively","kjv_def":"twins."},
"H8381":{"lemma":"תַּאֲלָה","xlit":"taʼălâh","pron":"tah-al-aw'","derivation":"from H422 (אָלָה);","strongs_def":"an imprecation","kjv_def":"curse."},
"H8382":{"lemma":"תָּאַם","xlit":"tâʼam","pron":"taw-am'","derivation":"a primitive root; to be complete; but used only as denominative from H8380 (תָּאוֹם),","strongs_def":"to be (causatively, make) twinned, i.e. (figuratively) duplicate or (architectural) jointed","kjv_def":"coupled (together), bear twins."},
"H8383":{"lemma":"תְּאֻן","xlit":"tᵉʼun","pron":"teh-oon'","derivation":"from H205 (אָוֶן);","strongs_def":"naughtiness, i.e. toil","kjv_def":"lie."},
"H8384":{"lemma":"תְּאֵן","xlit":"tᵉʼên","pron":"teh-ane'","derivation":"or (in the singular, feminine) תְּאֵנָה; perhaps of foreign derivation","strongs_def":"the fig (tree or fruit)","kjv_def":"fig (tree)."},
"H8385":{"lemma":"תַּאֲנָה","xlit":"taʼănâh","pron":"tah-an-aw'","derivation":"or תֹּאֲנָה; from H579 (אָנָה);","strongs_def":"an opportunity or (subjectively) purpose","kjv_def":"occasion."},
"H8386":{"lemma":"תַּאֲנִיָּה","xlit":"taʼănîyâh","pron":"tah-an-ee-yaw'","derivation":"from H578 (אָנָה);","strongs_def":"lamentation","kjv_def":"heaviness, mourning."},
"H8387":{"lemma":"תַּאֲנַת שִׁלֹה","xlit":"Taʼănath Shilôh","pron":"tah-an-ath' shee-lo'","derivation":"from H8385 (תַּאֲנָה) and H7887 (שִׁילֹה); approach of Shiloh;","strongs_def":"Taanath-Shiloh, a place in Palestine","kjv_def":"Taanath-shiloh."},
"H8388":{"lemma":"תָּאַר","xlit":"tâʼar","pron":"taw-ar'","derivation":"a primitive root;","strongs_def":"to delineate; reflex. to extend","kjv_def":"be drawn, mark out, (Rimmon-) methoar (by union with H7417 (רִמּוֹן))."},
"H8389":{"lemma":"תֹּאַר","xlit":"tôʼar","pron":"to'-ar","derivation":"from H8388 (תָּאַר);","strongs_def":"outline, i.e. figure or appearance","kjv_def":"[phrase] beautiful, [idiom] comely, countenance, [phrase] fair, [idiom] favoured, form, [idiom] goodly, [idiom] resemble, visage."},
"H8390":{"lemma":"תַּאֲרֵעַ","xlit":"Taʼărêaʻ","pron":"tah-ar-ay'-ah","derivation":"perhaps from H772 (אֲרַע);","strongs_def":"Taarea, an Israelite","kjv_def":"Tarea. See H8475 (תַּחְרֵעַ)."},
"H8391":{"lemma":"תְּאַשּׁוּר","xlit":"tᵉʼashshûwr","pron":"teh-ash-shoor'","derivation":"from H833 (אָשַׁר);","strongs_def":"a species of cedar (from its erectness)","kjv_def":"box (tree)."},
"H8392":{"lemma":"תֵּבָה","xlit":"têbâh","pron":"tay-baw'","derivation":"perhaps of foreign derivation;","strongs_def":"a box","kjv_def":"ark."},
"H8393":{"lemma":"תְּבוּאָה","xlit":"tᵉbûwʼâh","pron":"teb-oo-aw'","derivation":"from H935 (בּוֹא);","strongs_def":"income, i.e. produce (literally or figuratively)","kjv_def":"fruit, gain, increase, revenue."},
"H8394":{"lemma":"תָּבוּן","xlit":"tâbûwn","pron":"taw-boon'","derivation":"and (feminine) תְּבוּנָה; or תּוֹבֻנָה; from H995 (בִּין);","strongs_def":"intelligence; by implication, an argument; by extension, caprice","kjv_def":"discretion, reason, skilfulness, understanding, wisdom."},
"H8395":{"lemma":"תְּבוּסָה","xlit":"tᵉbûwçâh","pron":"teb-oo-saw'","derivation":"from H947 (בּוּס);","strongs_def":"a treading down, i.e. ruin","kjv_def":"destruction."},
"H8396":{"lemma":"תָּבוֹר","xlit":"Tâbôwr","pron":"taw-bore'","derivation":"from a root corresponding to H8406 (תְּבַר); broken region;","strongs_def":"Tabor, a mountain in Palestine, also a city adjacent","kjv_def":"Tabor."},
"H8397":{"lemma":"תֶּבֶל","xlit":"tebel","pron":"teh'-bel","derivation":"apparently from H1101 (בָּלַל);","strongs_def":"mixture, i.e. unnatural bestiality","kjv_def":"confusion."},
"H8398":{"lemma":"תֵּבֵל","xlit":"têbêl","pron":"tay-bale'","derivation":"from H2986 (יָבַל);","strongs_def":"the earth (as moist and therefore inhabited); by extension, the globe; by implication, its inhabitants; specifically, a particular land, as Babylonia, Palestine","kjv_def":"habitable part, world."},
"H8399":{"lemma":"תַּבְלִית","xlit":"tablîyth","pron":"tab-leeth'","derivation":"from H1086 (בָּלָה);","strongs_def":"consumption","kjv_def":"destruction."},
"H8400":{"lemma":"תְּבַּלֻּל","xlit":"tᵉballul","pron":"teb-al-lool'","derivation":"from H1101 (בָּלַל) in the original sense of flowing;","strongs_def":"a cataract (in the eye)","kjv_def":"blemish."},
"H8401":{"lemma":"תֶּבֶן","xlit":"teben","pron":"teh'-ben","derivation":"probably from H1129 (בָּנָה);","strongs_def":"properly, material, i.e. (specifically) refuse haum or stalks of grain (as chopped in threshing and used for fodder)","kjv_def":"chaff, straw, stubble."},
"H8402":{"lemma":"תִּבְנִי","xlit":"Tibnîy","pron":"tib-nee'","derivation":"from H8401 (תֶּבֶן); strawy;","strongs_def":"Tibni, an Israelite","kjv_def":"Tibni."},
"H8403":{"lemma":"תַּבְנִית","xlit":"tabnîyth","pron":"tab-neeth'","derivation":"from H1129 (בָּנָה);","strongs_def":"structure; by implication, a model, resemblance","kjv_def":"figure, form, likeness, pattern, similitude."},
"H8404":{"lemma":"תַּבְעֵרָה","xlit":"Tabʻêrâh","pron":"tab-ay-raw'","derivation":"from H1197 (בָּעַר); burning;","strongs_def":"Taberah, a place in the Desert","kjv_def":"Taberah."},
"H8405":{"lemma":"תֵּבֵץ","xlit":"Têbêts","pron":"tay-bates'","derivation":"from the same as H948 (בּוּץ); whiteness;","strongs_def":"Tebets, a place in Palestine","kjv_def":"Thebez."},
"H8406":{"lemma":"תְּבַר","xlit":"tᵉbar","pron":"teb-ar'","derivation":"(Aramaic) corresponding to H7665 (שָׁבַר);","strongs_def":"to be fragile (figuratively)","kjv_def":"broken."},
"H8407":{"lemma":"תִּגְלַת פִּלְאֶסֶר","xlit":"Tiglath Pilʼeçer","pron":"tig-lath' pil-eh'-ser","derivation":"or תִּגְלַת פְּלֶסֶר; or תִּלְגַּת פִּלְנְאֶסֶר or תִּלְגַּת פִּלְנֶסֶר; of foreign derivation;","strongs_def":"Tiglath-Pileser or Tilgath-pilneser, an Assyryrian king","kjv_def":"Tiglath-pileser, Tilgath-pilneser."},
"H8408":{"lemma":"תַּגְמוּל","xlit":"tagmûwl","pron":"tag-mool'","derivation":"from H1580 (גָּמַל);","strongs_def":"a bestowment","kjv_def":"benefit."},
"H8409":{"lemma":"תִּגְרָה","xlit":"tigrâh","pron":"tig-raw'","derivation":"from H1624 (גָּרָה);","strongs_def":"strife, i.e. infliction","kjv_def":"blow."},
"H8410":{"lemma":"תִּדְהָר","xlit":"tidhâr","pron":"tid-hawr'","derivation":"apparently from H1725 (דָּהַר);","strongs_def":"enduring; a species of hard-wood or lasting tree (perhaps oak)","kjv_def":"pine (tree)."},
"H8411":{"lemma":"תְּדִירָא","xlit":"tᵉdîyrâʼ","pron":"ted-ee-raw'","derivation":"(Aramaic) from H1753 (דּוּר) in the original sense of enduring;","strongs_def":"permanence, i.e. (adverb) constantly","kjv_def":"continually."},
"H8412":{"lemma":"תַּדְמֹר","xlit":"Tadmôr","pron":"tad-more'","derivation":"or תַּמֹּר; (1 Kings 9:18), apparently from H8558 (תָּמָר); palm-city;","strongs_def":"Tadmor, a place near Palestine","kjv_def":"Tadmor."},
"H8413":{"lemma":"תִּדְעָל","xlit":"Tidʻâl","pron":"tid-awl'","derivation":"perhaps from H1763 (דְּחַל); fearfulness;","strongs_def":"Tidal, a Canaanite","kjv_def":"Tidal."},
"H8414":{"lemma":"תֹּהוּ","xlit":"tôhûw","pron":"to'-hoo","derivation":"from an unused root meaning to lie waste;","strongs_def":"a desolation (of surface), i.e. desert; figuratively, a worthless thing; adverbially, in vain","kjv_def":"confusion, empty place, without form, nothing, (thing of) nought, vain, vanity, waste, wilderness."},
"H8415":{"lemma":"תְּהוֹם","xlit":"tᵉhôwm","pron":"teh-home'","derivation":"or תְּהֹם; (usually feminine) from H1949 (הוּם);","strongs_def":"an abyss (as a surging mass of water), especially the deep (the main sea or the subterranean watersupply)","kjv_def":"deep (place), depth."},
"H8416":{"lemma":"תְּהִלָּה","xlit":"tᵉhillâh","pron":"teh-hil-law'","derivation":"from H1984 (הָלַל);","strongs_def":"laudation; specifically (concretely) a hymn","kjv_def":"praise."},
"H8417":{"lemma":"תׇּהֳלָה","xlit":"tohŏlâh","pron":"to-hol-aw'","derivation":"feminine of an unused noun (apparently from H1984 (הָלַל)) meaning bluster;","strongs_def":"braggadocio, i.e. (by implication) fatuity","kjv_def":"folly."},
"H8418":{"lemma":"תַּהֲלֻכָה","xlit":"tahălukâh","pron":"tah-hal-oo-kaw'","derivation":"from H1980 (הָלַךְ);","strongs_def":"a procession","kjv_def":"[idiom] went."},
"H8419":{"lemma":"תַּהְפֻּכָה","xlit":"tahpukâh","pron":"tah-poo-kaw'","derivation":"from H2015 (הָפַךְ);","strongs_def":"a perversity or fraud","kjv_def":"(very) froward(-ness, thing), perverse thing."},
"H8420":{"lemma":"תָּו","xlit":"tâv","pron":"tawv","derivation":"from H8427 (תָּוָה);","strongs_def":"a mark; by implication, a signature","kjv_def":"desire, mark."},
"H8421":{"lemma":"תּוּב","xlit":"tûwb","pron":"toob","derivation":"(Aramaic) corresponding to H7725 (שׁוּב), to come back;","strongs_def":"specifically (transitive and ellip.) to reply","kjv_def":"answer, restore, return (an answer)."},
"H8422":{"lemma":"תּוּבַל","xlit":"Tûwbal","pron":"too-bal'","derivation":"or תֻּבַל; probably of foreign derivation;","strongs_def":"Tubal, a postdiluvian patriarch and his posterity","kjv_def":"Tubal."},
"H8423":{"lemma":"תּוּבַל קַיִן","xlit":"Tûwbal Qayin","pron":"too-bal' kah'-yin","derivation":"apparently from H2986 (יָבַל) (compare H2981 (יְבוּל)) and H7014 (קַיִן); offspring of Cain;","strongs_def":"Tubal-Kajin, an antidiluvian patriarch","kjv_def":"Tubal-cain."},
"H8424":{"lemma":"תּוּגָה","xlit":"tûwgâh","pron":"too-gaw'","derivation":"from H3013 (יָגָה);","strongs_def":"depression (of spirits); concretely a grief","kjv_def":"heaviness, sorrow."},
"H8425":{"lemma":"תּוֹגַרְמָה","xlit":"Tôwgarmâh","pron":"to-gar-maw'","derivation":"or תֹּגַרְמָה; probably of foreign derivation;","strongs_def":"Togarmah, a son of Gomer and his posterity","kjv_def":"Togarmah."},
"H8426":{"lemma":"תּוֹדָה","xlit":"tôwdâh","pron":"to-daw'","derivation":"from H3034 (יָדָה);","strongs_def":"properly, an extension of the hand, i.e. (by implication) avowal, or (usually) adoration; specifically, a choir of worshippers","kjv_def":"confession, (sacrifice of) praise, thanks(-giving, offering)."},
"H8427":{"lemma":"תָּוָה","xlit":"tâvâh","pron":"taw-vaw'","derivation":"a primitive root;","strongs_def":"to mark out, i.e. (primitive) scratch or (definite) imprint","kjv_def":"scrabble, set (a mark)."},
"H8428":{"lemma":"תָּוָה","xlit":"tâvâh","pron":"taw-vaw'","derivation":"a primitive root (or perhaps identical with H8427 (תָּוָה) through a similar idea from scraping to pieces);","strongs_def":"to grieve","kjv_def":"limit (by confusion with H8427 (תָּוָה))."},
"H8429":{"lemma":"תְּוַהּ","xlit":"tᵉvahh","pron":"tev-ah'","derivation":"(Aramaic) corresponding to H8539 (תָּמַהּ) or perhaps to H7582 (שָׁאָה) through the idea of sweeping to ruin (compare H8428 (תָּוָה));","strongs_def":"to amaze, i.e. (reflex. by implication) take alarm","kjv_def":"be astonied."},
"H8430":{"lemma":"תּוֹחַ","xlit":"Tôwach","pron":"to'-akh","derivation":"from an unused root meaning to depress; humble;","strongs_def":"Toach, an Israelite","kjv_def":"Toah."},
"H8431":{"lemma":"תּוֹחֶלֶת","xlit":"tôwcheleth","pron":"to-kheh'-leth","derivation":"from H3176 (יָחַל);","strongs_def":"expectation","kjv_def":"hope."},
"H8432":{"lemma":"תָּוֶךְ","xlit":"tâvek","pron":"taw'-vek","derivation":"from an unused root meaning to sever;","strongs_def":"a bisection, i.e. (by implication) the centre","kjv_def":"among(-st), [idiom] between, half, [idiom] (there-, where-), in(-to), middle, mid(-night), midst (among), [idiom] out (of), [idiom] through, [idiom] with(-in)."},
"H8433":{"lemma":"תּוֹכֵחָה","xlit":"tôwkêchâh","pron":"to-kay-khaw'","derivation":"and תּוֹכַחַת; from H3198 (יָכַח);","strongs_def":"chastisement; figuratively (by words) correction, refutation, proof (even in defence)","kjv_def":"argument, [idiom] chastened, correction, reasoning, rebuke, reproof, [idiom] be (often) reproved."},
"H8434":{"lemma":"תּוֹלָד","xlit":"Tôwlâd","pron":"to-lawd'","derivation":"from H3205 (יָלַד); posterity;","strongs_def":"Tolad, a place in Palestine","kjv_def":"Tolad. Compare H513 (אֶלְתּוֹלַד)."},
"H8435":{"lemma":"תּוֹלְדָה","xlit":"tôwlᵉdâh","pron":"to-led-aw'","derivation":"or תֹּלְדָה; from H3205 (יָלַד);","strongs_def":"(plural only) descent, i.e. family; (figuratively) history","kjv_def":"birth, generations."},
"H8436":{"lemma":"תּוּלוֹן","xlit":"Tûwlôwn","pron":"too-lone'","derivation":"from H8524 (תָּלַל); suspension;","strongs_def":"Tulon, an Israelite","kjv_def":"Tilon (from the margin)."},
"H8437":{"lemma":"תּוֹלָל","xlit":"tôwlâl","pron":"to-lawl'","derivation":"from H3213 (יָלַל);","strongs_def":"causing to howl, i.e. an oppressor","kjv_def":"that wasted."},
"H8438":{"lemma":"תּוֹלָע","xlit":"tôwlâʻ","pron":"to-law'","derivation":"and (feminine) תּוֹלֵעָה; or תּוֹלַעַת; or תֹּלַעַת; from H3216 (יָלַע); a maggot (as voracious); specifically (often with ellipsis of H8144 (שָׁנִי))","strongs_def":"the crimson-grub, but used only (in this connection) of the colorfrom it, and cloths dyed therewith","kjv_def":"crimson, scarlet, worm."},
"H8439":{"lemma":"תּוֹלָע","xlit":"Tôwlâʻ","pron":"to-law'","derivation":"the same as H8438 (תּוֹלָע); worm;","strongs_def":"Tola, the name of two Israelites","kjv_def":"Tola."},
"H8440":{"lemma":"תּוֹלָעִי","xlit":"Tôwlâʻîy","pron":"to-law-ee'","derivation":"patronymically from H8439 (תּוֹלָע);","strongs_def":"a Tolaite (collectively) or descendants of Tola","kjv_def":"Tolaites."},
"H8441":{"lemma":"תּוֹעֵבַה","xlit":"tôwʻêbah","pron":"to-ay-baw'","derivation":"or תֹּעֵבַה; feminine active participle of H8581 (תַּעָב);","strongs_def":"properly, something disgusting (morally), i.e. (as noun) an abhorrence; especially idolatry or (concretely) an idol","kjv_def":"abominable (custom, thing), abomination."},
"H8442":{"lemma":"תּוֹעָה","xlit":"tôwʻâh","pron":"to-aw'","derivation":"feminine active participle of H8582 (תָּעָה);","strongs_def":"mistake, i.e. (morally) impiety, or (political) injury","kjv_def":"error, hinder."},
"H8443":{"lemma":"תּוֹעָפָה","xlit":"tôwʻâphâh","pron":"to-aw-faw'","derivation":"from H3286 (יָעַף);","strongs_def":"(only in plural collective) weariness, i.e. (by implication) toil (treasure so obtained) or speed","kjv_def":"plenty, strength."},
"H8444":{"lemma":"תּוֹצָאָה","xlit":"tôwtsâʼâh","pron":"to-tsaw-aw'","derivation":"or תֹּצָאָה; from H3318 (יָצָא);","strongs_def":"(only in plural collective) exit, i.e. (geographical) boundary, or (figuratively) deliverance, (actively) source","kjv_def":"border(-s), going(-s) forth (out), issues, outgoings."},
"H8445":{"lemma":"תּוֹקַהַת","xlit":"Tôwqahath","pron":"to-kah'-ath","derivation":"from the same as H3349 (יִקָּהָה); obedience;","strongs_def":"Tokahath, an Israelite","kjv_def":"Tikvath (by correction for H8616 (תִּקְוָה))."},
"H8446":{"lemma":"תּוּר","xlit":"tûwr","pron":"toor","derivation":"a primitive root;","strongs_def":"to meander (causatively, guide) about, especially fortrade or reconnoitring","kjv_def":"chap(-man), sent to descry, be excellent, merchant(-man), search (out), seek, (e-) spy (out)."},
"H8447":{"lemma":"תּוֹר","xlit":"tôwr","pron":"tore","derivation":"or תֹּר; from H8446 (תּוּר);","strongs_def":"a succession, i.e. a string or (abstractly) order","kjv_def":"border, row, turn."},
"H8448":{"lemma":"תּוֹר","xlit":"tôwr","pron":"tore","derivation":"probably the same as H8447 (תּוֹר);","strongs_def":"a manner (as a sort of turn)","kjv_def":"estate."},
"H8449":{"lemma":"תּוֹר","xlit":"tôwr","pron":"tore","derivation":"or תֹּר; probably the same as H8447 (תּוֹר);","strongs_def":"a ring-dove, often (figuratively) as a term of endearment","kjv_def":"(turtle) dove."},
"H8450":{"lemma":"תּוֹר","xlit":"tôwr","pron":"tore","derivation":"(Aramaic) corresponding (by permutation) to H7794 (שׁוֹר);","strongs_def":"a bull","kjv_def":"bullock, ox."},
"H8451":{"lemma":"תּוֹרָה","xlit":"tôwrâh","pron":"to-raw'","derivation":"or תֹּרָה; from H3384 (יָרָה);","strongs_def":"a precept or statute, especially the Decalogue or Pentateuch","kjv_def":"law."},
"H8452":{"lemma":"תּוֹרָה","xlit":"tôwrâh","pron":"to-raw'","derivation":"probably feminine of H8448 (תּוֹר);","strongs_def":"a custom","kjv_def":"manner."},
"H8453":{"lemma":"תּוֹשָׁב","xlit":"tôwshâb","pron":"to-shawb'","derivation":"or תֹּשָׁב; (1 Kings 17:1), from H3427 (יָשַׁב); a dweller (but not outlandish (H5237 (נׇכְרִי))); especially (as distinguished from a native citizen (active participle of H3427 (יָשַׁב)) and a temporary inmate (H1616 (גֵּר)) or mere lodger (H3885 (לוּן)))","strongs_def":"resident alien","kjv_def":"foreigner, inhabitant, sojourner, stranger."},
"H8454":{"lemma":"תּוּשִׁיָּה","xlit":"tûwshîyâh","pron":"too-shee-yaw'","derivation":"or תֻּשִׁיָּה; from an unused root probably meaning to substantiate;","strongs_def":"support or (by implication) ability, i.e. (direct) help, (in purpose) an undertaking, (intellectual) understanding","kjv_def":"enterprise, that which (thing as it) is, substance, (sound) wisdom, working."},
"H8455":{"lemma":"תּוֹתָח","xlit":"tôwthâch","pron":"to-thawkh'","derivation":"from an unused root meaning to smite;","strongs_def":"a club","kjv_def":"darts."},
"H8456":{"lemma":"תָּזַז","xlit":"tâzaz","pron":"taw-zaz'","derivation":"a primitive root;","strongs_def":"to lop off","kjv_def":"cut down."},
"H8457":{"lemma":"תַּזְנוּת","xlit":"taznûwth","pron":"taz-nooth'","derivation":"or תַּזְנֻת; from H2181 (זָנָה);","strongs_def":"harlotry, i.e. (figuratively) idolatry","kjv_def":"fornication, whoredom."},
"H8458":{"lemma":"תַּחְבֻּלָה","xlit":"tachbulâh","pron":"takh-boo-law'","derivation":"or תַּחְבּוּלָה; from H2254 (חָבַל) as denominative from H2256 (חֶבֶל);","strongs_def":"(only in plural) properly, steerage (as a management of ropes), i.e. (figuratively) guidance or (by implication) a plan","kjv_def":"good advice, (wise) counsels."},
"H8459":{"lemma":"תֹּחוּ","xlit":"Tôchûw","pron":"to'-khoo","derivation":"from an unused root meaning to depress; abasement;","strongs_def":"Tochu, an Israelite","kjv_def":"Tohu."},
"H8460":{"lemma":"תְּחוֹת","xlit":"tᵉchôwth","pron":"tekh-oth'","derivation":"(Aramaic) or תּחֹת; (Aramaic), corresponding to H8478 (תַּחַת);","strongs_def":"beneath","kjv_def":"under."},
"H8461":{"lemma":"תַּחְכְּמֹנִי","xlit":"Tachkᵉmônîy","pron":"takh-kem-o-nee'","derivation":"probably for H2453 (חַכְמוֹנִי); sagacious;","strongs_def":"Tachkemoni, an Israelite","kjv_def":"Tachmonite."},
"H8462":{"lemma":"תְּחִלָּה","xlit":"tᵉchillâh","pron":"tekh-il-law'","derivation":"from H2490 (חָלַל) in the sense of opening;","strongs_def":"a commencement; rel. original (adverb, -ly)","kjv_def":"begin(-ning), first (time)."},
"H8463":{"lemma":"תַּחֲלוּא","xlit":"tachălûwʼ","pron":"takh-al-oo'","derivation":"or תַּחֲלֻא; from H2456 (חָלָא);","strongs_def":"a malady","kjv_def":"disease, [idiom] grievous, (that are) sick(-ness)."},
"H8464":{"lemma":"תַּחְמָס","xlit":"tachmâç","pron":"takh-mawce'","derivation":"from H2554 (חָמַס);","strongs_def":"a species of unclean bird (from its violence), perhaps an owl","kjv_def":"night hawk."},
"H8465":{"lemma":"תַּחַן","xlit":"Tachan","pron":"takh'-an","derivation":"probably from H2583 (חָנָה); station;","strongs_def":"Tachan, the name of two Israelites","kjv_def":"Tahan."},
"H8466":{"lemma":"תַּחֲנָה","xlit":"tachănâh","pron":"takh-an-aw'","derivation":"from H2583 (חָנָה);","strongs_def":"(only plural collectively) an encampment","kjv_def":"camp."},
"H8467":{"lemma":"תְּחִנָּה","xlit":"tᵉchinnâh","pron":"tekh-in-naw'","derivation":"from H2603 (חָנַן);","strongs_def":"graciousness; causatively, entreaty","kjv_def":"favour, grace, supplication."},
"H8468":{"lemma":"תְּחִנָּה","xlit":"Tᵉchinnâh","pron":"tekh-in-naw'","derivation":"the same as H8467 (תְּחִנָּה);","strongs_def":"Techinnah, an Israelite","kjv_def":"Tehinnah."},
"H8469":{"lemma":"תַּחֲנוּן","xlit":"tachănûwn","pron":"takh-an-oon'","derivation":"or (feminine) תַּחֲנוּנָה; from H2603 (חָנַן);","strongs_def":"earnest prayer","kjv_def":"intreaty, supplication."},
"H8470":{"lemma":"תַּחֲנִי","xlit":"Tachănîy","pron":"takh-an-ee'","derivation":"patronymically from H8465 (תַּחַן);","strongs_def":"a Tachanite (collectively) or descendants of Tachan","kjv_def":"Tahanites."},
"H8471":{"lemma":"תַּחְפַּנְחֵס","xlit":"Tachpanchêç","pron":"takh-pan-khace'","derivation":"or תְּחַפְנְחֵס; (Ezekiel 30:18), or תַּחְפְּנֵס; (Jeremiah 2:16), of Egyptian derivation;","strongs_def":"Tachpanches, Techaphneches or Tachpenes, a place in Egypt","kjv_def":"Tahapanes, Tahpanhes, Tehaphnehes."},
"H8472":{"lemma":"תַּחְפְּנֵיס","xlit":"Tachpᵉnêyç","pron":"takh-pen-ace'","derivation":"of Egyptian derivation;","strongs_def":"Tachpenes, an Egyptian woman","kjv_def":"Tahpenes."},
"H8473":{"lemma":"תַּחֲרָא","xlit":"tachărâʼ","pron":"takh-ar-aw'","derivation":"from H2734 (חָרָה) in the original sense of H2352 (חוּר) or H2353 (חוּר);","strongs_def":"a linen corslet (as white or hollow)","kjv_def":"habergeon."},
"H8474":{"lemma":"תַּחָרָה","xlit":"tachârâh","pron":"takh-aw-raw'","derivation":"a facitious root from H2734 (חָרָה) through the idea of the heat of jealousy;","strongs_def":"to vie with a rival","kjv_def":"close, contend."},
"H8475":{"lemma":"תַּחְרֵעַ","xlit":"Tachrêaʻ","pron":"takh-ray'-ah","derivation":"for H8390 (תַּאֲרֵעַ);","strongs_def":"Tachrea, an Israelite","kjv_def":"Tahrea."},
"H8476":{"lemma":"תַּחַשׁ","xlit":"tachash","pron":"takh'-ash","derivation":"probably of foreign derivation;","strongs_def":"a (clean) animal with fur, probably a species of antelope","kjv_def":"badger."},
"H8477":{"lemma":"תַּחַשׁ","xlit":"Tachash","pron":"takh'-ash","derivation":"the same as H8476 (תַּחַשׁ);","strongs_def":"Tachash, a relative of Abraham","kjv_def":"Thahash."},
"H8478":{"lemma":"תַּחַת","xlit":"tachath","pron":"takh'-ath","derivation":"from the same as H8430 (תּוֹחַ);","strongs_def":"the bottom (as depressed); only adverbially, below (often with prepositional prefix underneath), in lieu of, etc.","kjv_def":"as, beneath, [idiom] flat, in(-stead), (same) place (where...is), room, for...sake, stead of, under, [idiom] unto, [idiom] when...was mine, whereas, (where-) fore, with."},
"H8479":{"lemma":"תַּחַת","xlit":"tachath","pron":"takh'-ath","derivation":"(Aramaic) corresponding to H8478 (תַּחַת)","strongs_def":"{the bottom (as depressed); only adverbially, below (often with prepositional prefix underneath), in lieu of, etc.}","kjv_def":"under."},
"H8480":{"lemma":"תַּחַת","xlit":"Tachath","pron":"takh'-ath","derivation":"the same as H8478 (תַּחַת);","strongs_def":"Tachath, the name of a place in the Desert, also of three Israelites","kjv_def":"Tahath."},
"H8481":{"lemma":"תַּחְתּוֹן","xlit":"tachtôwn","pron":"takh-tone'","derivation":"or תַּחְתֹּן; from H8478 (תַּחַת);","strongs_def":"bottommost","kjv_def":"lower(-est), nether(-most)."},
"H8482":{"lemma":"תַּחְתִּי","xlit":"tachtîy","pron":"takh-tee'","derivation":"from H8478 (תַּחַת);","strongs_def":"lowermost; as noun (feminine plural) the depths (figuratively, a pit, the womb)","kjv_def":"low (parts, -er, -er parts, -est), nether (part)."},
"H8483":{"lemma":"תַּחְתִּים חׇדְשִׁי","xlit":"Tachtîym Chodshîy","pron":"takh-teem' khod-shee'","derivation":"apparently from the plural masculine of H8482 (תַּחְתִּי) or H8478 (תַּחַת) and H2320 (חֹדֶשׁ); lower (ones) monthly;","strongs_def":"Tachtim-Chodshi, a place in Palestine","kjv_def":"Tahtim-hodshi."},
"H8484":{"lemma":"תִּיכוֹן","xlit":"tîykôwn","pron":"tee-kone'","derivation":"or תִּיכֹן; from H8432 (תָּוֶךְ);","strongs_def":"central","kjv_def":"middle(-most), midst."},
"H8485":{"lemma":"תֵּימָא","xlit":"Têymâʼ","pron":"tay-maw'","derivation":"or תֵּמָא; probably of foreign derivation;","strongs_def":"Tema, a son of Ishmael, and the region settled by him","kjv_def":"Tema."},
"H8486":{"lemma":"תֵּימָן","xlit":"têymân","pron":"tay-mawn'","derivation":"or תֵּמָן; denominative from H3225 (יָמִין);","strongs_def":"the south (as being on the right hand of a person facing the east)","kjv_def":"south (side, -ward, wind)."},
"H8487":{"lemma":"תֵּימָן","xlit":"Têymân","pron":"tay-mawn'","derivation":"or תֵּמָן; the same as H8486 (תֵּימָן);","strongs_def":"Teman, the name of two Edomites, and of the region and descendant of one of them","kjv_def":"south, Teman."},
"H8488":{"lemma":"תֵּימְנִי","xlit":"Têymᵉnîy","pron":"tay-men-ee'","derivation":"probably for H8489 (תֵּימָנִי);","strongs_def":"Temeni, an Israelite","kjv_def":"Temeni."},
"H8489":{"lemma":"תֵּימָנִי","xlit":"Têymânîy","pron":"tay-maw-nee'","derivation":"patronymically from H8487 (תֵּימָן);","strongs_def":"a Temanite or descendant of Teman","kjv_def":"Temani, Temanite."},
"H8490":{"lemma":"תִּימָרָה","xlit":"tîymârâh","pron":"tee-maw-raw'","derivation":"or תִּמָרָה; from the same as H8558 (תָּמָר);","strongs_def":"a column, i.e. cloud","kjv_def":"pillar."},
"H8491":{"lemma":"תִּיצִי","xlit":"Tîytsîy","pron":"tee-tsee'","derivation":"patrial or patronymically from an unused noun of uncertain meaning;","strongs_def":"a Titsite or descendant or inhabitant of an unknown Tits","kjv_def":"Tizite."},
"H8492":{"lemma":"תִּירוֹשׁ","xlit":"tîyrôwsh","pron":"tee-roshe'","derivation":"or תִּירֹשׁ; from H3423 (יָרַשׁ) in the sense of expulsion;","strongs_def":"must or fresh grape-juice (as just squeezed out); by implication (rarely) fermented wine","kjv_def":"(new, sweet) wine."},
"H8493":{"lemma":"תִּירְיָא","xlit":"Tîyrᵉyâʼ","pron":"tee-reh-yaw'","derivation":"probably from H3372 (יָרֵא);","strongs_def":"fearful, Tirja, an Israelite","kjv_def":"Tiria."},
"H8494":{"lemma":"תִּירָס","xlit":"Tîyrâç","pron":"tee-rawce'","derivation":"probably of foreign derivation;","strongs_def":"Tiras, a son of Japheth","kjv_def":"Tiras."},
"H8495":{"lemma":"תַּיִשׁ","xlit":"tayish","pron":"tah'-yeesh","derivation":"from an unused root meaning to butt;","strongs_def":"a buck or he-goat (as given to butting)","kjv_def":"he goat."},
"H8496":{"lemma":"תֹּךְ","xlit":"tôk","pron":"toke","derivation":"or תּוֹך; (Psalm 72:14), from the same base as H8432 (תָּוֶךְ) (in the sense of cutting to pieces);","strongs_def":"oppression","kjv_def":"deceit, fraud."},
"H8497":{"lemma":"תָּכָה","xlit":"tâkâh","pron":"taw-kaw'","derivation":"a primitive root;","strongs_def":"to strew, i.e. encamp","kjv_def":"sit down."},
"H8498":{"lemma":"תְּכוּנָה","xlit":"tᵉkûwnâh","pron":"tek-oo-naw'","derivation":"feminine passive participle of H8505 (תָּכַן);","strongs_def":"adjustment, i.e. structure; by implication, equipage","kjv_def":"fashion, store."},
"H8499":{"lemma":"תְּכוּנָה","xlit":"tᵉkûwnâh","pron":"tek-oo-naw'","derivation":"from H3559 (כּוּן); or probably identical with H8498 (תְּכוּנָה);","strongs_def":"something arranged or fixed, i.e. a place","kjv_def":"seat."},
"H8500":{"lemma":"תֻּכִּי","xlit":"tukkîy","pron":"took-kee'","derivation":"or תּוּכִּי;","strongs_def":"probably of foreign derivation; some imported creature, probably a peacock","kjv_def":"peacock."},
"H8501":{"lemma":"תָּכָךְ","xlit":"tâkâk","pron":"taw-kawk'","derivation":"from an unused root meaning","strongs_def":"to dissever, i.e. crush","kjv_def":"deceitful."},
"H8502":{"lemma":"תִּכְלָה","xlit":"tiklâh","pron":"tik-law'","derivation":"from H3615 (כָּלָה);","strongs_def":"completeness","kjv_def":"perfection."},
"H8503":{"lemma":"תַּכְלִית","xlit":"taklîyth","pron":"tak-leeth'","derivation":"from H3615 (כָּלָה);","strongs_def":"completion; by implication, an extremity","kjv_def":"end, perfect(-ion)."},
"H8504":{"lemma":"תְּכֵלֶת","xlit":"tᵉkêleth","pron":"tek-ay'-leth","derivation":"probably for H7827 (שְׁחֶלֶת);","strongs_def":"the cerulean mussel, i.e. the color (violet) obtained therefrom or stuff dyed therewith","kjv_def":"blue."},
"H8505":{"lemma":"תָּכַן","xlit":"tâkan","pron":"taw-kan'","derivation":"a primitive root;","strongs_def":"to balance, i.e. measure out (by weight or dimension); figuratively, arrange, equalize, through the idea of levelling (ment. estimate, test)","kjv_def":"bear up, direct, be (un-)equal, mete, ponder, tell, weigh."},
"H8506":{"lemma":"תֹּכֶן","xlit":"tôken","pron":"to'-ken","derivation":"from H8505 (תָּכַן);","strongs_def":"a fixed quantity","kjv_def":"measure, tale."},
"H8507":{"lemma":"תֹּכֶן","xlit":"Tôken","pron":"to'-ken","derivation":"the same as H8506 (תֹּכֶן);","strongs_def":"Token, a place in Palestine","kjv_def":"Tochen."},
"H8508":{"lemma":"תׇּכְנִית","xlit":"toknîyth","pron":"tok-neeth'","derivation":"from H8506 (תֹּכֶן);","strongs_def":"admeasurement, i.e. consummation","kjv_def":"pattern, sum."},
"H8509":{"lemma":"תַּכְרִיךְ","xlit":"takrîyk","pron":"tak-reek'","derivation":"apparently from an unused root meaning to encompass;","strongs_def":"a wrapper or robe","kjv_def":"garment."},
"H8510":{"lemma":"תֵּל","xlit":"têl","pron":"tale","derivation":"by contraction from H8524 (תָּלַל);","strongs_def":"a mound","kjv_def":"heap, [idiom] strength."},
"H8511":{"lemma":"תָּלָא","xlit":"tâlâʼ","pron":"taw-law'","derivation":"a primitive root;","strongs_def":"to suspend; figuratively (through hesitation) to be uncertain; by implication (of mental dependence) to habituate","kjv_def":"be bent, hang (in doubt)."},
"H8512":{"lemma":"תֵּל אָבִיב","xlit":"Têl ʼÂbîyb","pron":"tale aw-beeb'","derivation":"from H8510 (תֵּל) and H24 (אָבִיב); mound of green growth;","strongs_def":"Tel-Abib, a place in Chaldaea","kjv_def":"Tel-abib."},
"H8513":{"lemma":"תְּלָאָה","xlit":"tᵉlâʼâh","pron":"tel-aw-aw'","derivation":"from H3811 (לָאָה);","strongs_def":"distress","kjv_def":"travail, travel, trouble."},
"H8514":{"lemma":"תַּלְאוּבָה","xlit":"talʼûwbâh","pron":"tal-oo-baw'","derivation":"from H3851 (לַהַב);","strongs_def":"desiccation","kjv_def":"great drought."},
"H8515":{"lemma":"תְּלַאשַּׂר","xlit":"Tᵉlaʼssar","pron":"tel-as-sar'","derivation":"or תְּלַשַּׂר; of foreign derivation;","strongs_def":"Telassar, a region of Assyria","kjv_def":"Telassar."},
"H8516":{"lemma":"תַּלְבֹּשֶׁת","xlit":"talbôsheth","pron":"tal-bo'-sheth","derivation":"from H3847 (לָבַשׁ);","strongs_def":"a garment","kjv_def":"clothing."},
"H8517":{"lemma":"תְּלַג","xlit":"tᵉlag","pron":"tel-ag'","derivation":"(Aramaic) corresponding to H7950 (שֶׁלֶג);","strongs_def":"snow","kjv_def":"snow."},
"H8518":{"lemma":"תָּלָה","xlit":"tâlâh","pron":"taw-law'","derivation":"a primitive root;","strongs_def":"to suspend (especially to gibbet)","kjv_def":"hang (up)."},
"H8519":{"lemma":"תְּלוּנָה","xlit":"tᵉlûwnâh","pron":"tel-oo-naw'","derivation":"or תּלֻנָּה; from H3885 (לוּן) in the sense of obstinacy;","strongs_def":"a grumbling","kjv_def":"murmuring."},
"H8520":{"lemma":"תֶּלַח","xlit":"Telach","pron":"teh'-lakh","derivation":"probably from an unused root meaning to dissever; breach;","strongs_def":"Telach, an Israelite","kjv_def":"Telah."},
"H8521":{"lemma":"תֵּל חַרְשָׁא","xlit":"Têl Charshâʼ","pron":"tale khar-shaw'","derivation":"from H8510 (תֵּל) and the feminine of H2798 (חֲרָשִׁים); mound of workmanship;","strongs_def":"Tel-Charsha, a place in Babylonia","kjv_def":"Tel-haresha, Tel-harsa."},
"H8522":{"lemma":"תְּלִי","xlit":"tᵉlîy","pron":"tel-ee'","derivation":"probably from H8518 (תָּלָה);","strongs_def":"a quiver (as slung)","kjv_def":"quiver."},
"H8523":{"lemma":"תְּלִיתַי","xlit":"tᵉlîythay","pron":"tel-ee-thah'-ee","derivation":"(Aramaic) or תַּלְתִּי; (Aramaic), ordinal from H8532 (תְּלָת);","strongs_def":"third","kjv_def":"third."},
"H8524":{"lemma":"תָּלַל","xlit":"tâlal","pron":"taw-lal'","derivation":"a primitive root;","strongs_def":"to pile up, i.e. elevate","kjv_def":"eminent. Compare H2048 (הָתַל)."},
"H8525":{"lemma":"תֶּלֶם","xlit":"telem","pron":"teh'-lem","derivation":"from an unused root meaning to accumulate;","strongs_def":"a bank or terrace","kjv_def":"furrow, ridge."},
"H8526":{"lemma":"תַּלְמַי","xlit":"Talmay","pron":"tal-mah'-ee","derivation":"from H8525 (תֶּלֶם); ridged;","strongs_def":"Talmai, the name of a Canaanite and a Syrian","kjv_def":"Talmai."},
"H8527":{"lemma":"תַּלְמִיד","xlit":"talmîyd","pron":"tal-meed'","derivation":"from H3925 (לָמַד);","strongs_def":"a pupil","kjv_def":"scholar."},
"H8528":{"lemma":"תֵּל מֶלַח","xlit":"Têl Melach","pron":"tale meh'-lakh","derivation":"from H8510 (תֵּל) and H4417 (מֶלַח); mound of salt;","strongs_def":"Tel-Melach, a place in Babylonia","kjv_def":"Tel-melah."},
"H8529":{"lemma":"תָּלַע","xlit":"tâlaʻ","pron":"taw-law'","derivation":"a denominative from H8438 (תּוֹלָע);","strongs_def":"to crimson, i.e. dye that color","kjv_def":"[idiom] scarlet."},
"H8530":{"lemma":"תַּלְפִּיָּה","xlit":"talpîyâh","pron":"tal-pee-yaw'","derivation":"feminine from an unused root meaning to tower;","strongs_def":"something tall, i.e. (plural collective) slenderness","kjv_def":"armoury."},
"H8531":{"lemma":"תְּלַת","xlit":"tᵉlath","pron":"tel-ath'","derivation":"(Aramaic) from H8532 (תְּלָת);","strongs_def":"a tertiary rank","kjv_def":"third."},
"H8532":{"lemma":"תְּלָת","xlit":"tᵉlâth","pron":"tel-awth'","derivation":"(Aramaic) masculine תְּלָתָה; (Aramaic), or תְּלָתָא; (Aramaic), corresponding to H7969 (שָׁלוֹשׁ);","strongs_def":"three or third","kjv_def":"third, three."},
"H8533":{"lemma":"תְּלָתִין","xlit":"tᵉlâthîyn","pron":"tel-aw-theen'","derivation":"(Aramaic) multiple of H8532 (תְּלָת);","strongs_def":"ten times three","kjv_def":"thirty."},
"H8534":{"lemma":"תַּלְתַּל","xlit":"taltal","pron":"tal-tal'","derivation":"by reduplication, from H8524 (תָּלַל) through the idea of vibration;","strongs_def":"a trailing bough (as pendulous)","kjv_def":"bushy."},
"H8535":{"lemma":"תָּם","xlit":"tâm","pron":"tawm","derivation":"from H8552 (תָּמַם);","strongs_def":"complete; usually (morally) pious; specifically, gentle, dear","kjv_def":"coupled together, perfect, plain, undefiled, upright."},
"H8536":{"lemma":"תָּם","xlit":"tâm","pron":"tawm","derivation":"(Aramaic) corresponding to H8033 (שָׁם);","strongs_def":"there","kjv_def":"[idiom] thence, there, [idiom] where."},
"H8537":{"lemma":"תֹּם","xlit":"tôm","pron":"tome","derivation":"from H8552 (תָּמַם);","strongs_def":"completeness; figuratively, prosperity; usually (morally) innocence","kjv_def":"full, integrity, perfect(-ion), simplicity, upright(-ly, -ness), at a venture. See H8550 (תֻּמִּים)."},
"H8538":{"lemma":"תֻּמָּה","xlit":"tummâh","pron":"toom-maw'","derivation":"feminine of H8537 (תֹּם);","strongs_def":"innocence","kjv_def":"integrity."},
"H8539":{"lemma":"תָּמַהּ","xlit":"tâmahh","pron":"taw-mah'","derivation":"a primitive root;","strongs_def":"to be in consternation","kjv_def":"be amazed, be astonished, marvel(-lously), wonder."},
"H8540":{"lemma":"תְּמַהּ","xlit":"tᵉmahh","pron":"tem-ah'","derivation":"(Aramaic) from a root corresponding to H8539 (תָּמַהּ);","strongs_def":"a miracle","kjv_def":"wonder."},
"H8541":{"lemma":"תִּמָּהוֹן","xlit":"timmâhôwn","pron":"tim-maw-hone'","derivation":"from H8539 (תָּמַהּ);","strongs_def":"consternation","kjv_def":"astonishment."},
"H8542":{"lemma":"תַּמּוּז","xlit":"Tammûwz","pron":"tam-mooz'","derivation":"of uncertain derivation;","strongs_def":"Tammuz, a Phoenician deity","kjv_def":"Tammuz."},
"H8543":{"lemma":"תְּמוֹל","xlit":"tᵉmôwl","pron":"tem-ole'","derivation":"or תְּמֹל; probably for H865 (אֶתְמוֹל); sometimes with  H8032 (שִׁלְשׁוֹם)","strongs_def":"properly, ago, i.e. a (short or long) time since; especially yesterday, or  day before yesterday","kjv_def":"[phrase] before (-time), [phrase] these (three) days, [phrase] heretofore, [phrase] time past, yesterday."},
"H8544":{"lemma":"תְּמוּנָה","xlit":"tᵉmûwnâh","pron":"tem-oo-naw'","derivation":"or תְּמֻנָה; from H4327 (מִין);","strongs_def":"something portioned (i.e. fashioned) out, as ashape, i.e. (indefinitely) phantom, or (specifically) embodiment, or (figuratively) manifestation (of favor)","kjv_def":"image, likeness, similitude."},
"H8545":{"lemma":"תְּמוּרָה","xlit":"tᵉmûwrâh","pron":"tem-oo-raw'","derivation":"from H4171 (מוּר);","strongs_def":"barter, compensation","kjv_def":"(ex-) change(-ing), recompense, restitution."},
"H8546":{"lemma":"תְּמוּתָה","xlit":"tᵉmûwthâh","pron":"tem-oo-thaw'","derivation":"from H4191 (מוּת);","strongs_def":"execution (as a doom)","kjv_def":"death, die."},
"H8547":{"lemma":"תֶּמַח","xlit":"Temach","pron":"teh'-makh","derivation":"of uncertain derivation;","strongs_def":"Temach, one of the Nethinim","kjv_def":"Tamah, Thamah."},
"H8548":{"lemma":"תָּמִיד","xlit":"tâmîyd","pron":"taw-meed'","derivation":"from an unused root meaning to stretch;","strongs_def":"properly, continuance (as indefinite extension); but used only (attributively as adjective) constant (or adverbially, constantly); ellipt. the regular (daily) sacrifice","kjv_def":"alway(-s), continual (employment, -ly), daily, (n-)ever(-more), perpetual."},
"H8549":{"lemma":"תָּמִים","xlit":"tâmîym","pron":"taw-meem'","derivation":"from H8552 (תָּמַם);","strongs_def":"entire (literally, figuratively or morally); also (as noun) integrity, truth","kjv_def":"without blemish, complete, full, perfect, sincerely (-ity), sound, without spot, undefiled, upright(-ly), whole."},
"H8550":{"lemma":"תֻּמִּים","xlit":"Tummîym","pron":"toom-meem'","derivation":"plural of H8537 (תֹּם);","strongs_def":"perfections, i.e. (techn.) one of the epithets of the objects in the high-priest's breastplate as an emblem of complete Truth","kjv_def":"Thummim."},
"H8551":{"lemma":"תָּמַךְ","xlit":"tâmak","pron":"taw-mak'","derivation":"a primitive root;","strongs_def":"to sustain; by implication, to obtain, keep fast; figuratively, to help, follow close","kjv_def":"(take, up-) hold (up), maintain, retain, stay (up)."},
"H8552":{"lemma":"תָּמַם","xlit":"tâmam","pron":"taw-mam'","derivation":"a primitive root;","strongs_def":"to complete, in a good or a bad sense, literal, or figurative, transitive or intransitive","kjv_def":"accomplish, cease, be clean (pass-) ed, consume, have done, (come to an, have an, make an) end, fail, come to the full, be all gone, [idiom] be all here, be (make) perfect, be spent, sum, be (shew self) upright, be wasted, whole."},
"H8553":{"lemma":"תִּמְנָה","xlit":"Timnâh","pron":"tim-naw'","derivation":"from H4487 (מָנָה); a portion assigned;","strongs_def":"Timnah, the name of two places in Palestine","kjv_def":"Timnah, Timnath, Thimnathah."},
"H8554":{"lemma":"תִּמְנִי","xlit":"Timnîy","pron":"tim-nee'","derivation":"patrial from H8553 (תִּמְנָה);","strongs_def":"a Timnite or inhabitant of Timnah","kjv_def":"Timnite."},
"H8555":{"lemma":"תִּמְנָע","xlit":"Timnâʻ","pron":"tim-naw'","derivation":"from H4513 (מָנַע); restraint;","strongs_def":"Timna, the name of two Edomites","kjv_def":"Timna, Timnah."},
"H8556":{"lemma":"תִּמְנַת חֶרֶס","xlit":"Timnath Chereç","pron":"tim-nath kheh'-res","derivation":"or תִּמְנַת סֶרַח; from H8553 (תִּמְנָה) and H2775 (חֶרֶס); portion of (the) sun;","strongs_def":"Timnath-Cheres, a place in Palestine","kjv_def":"Timnath-heres, Timnath-serah."},
"H8557":{"lemma":"תֶּמֶס","xlit":"temeç","pron":"teh'-mes","derivation":"from H4529 (מָסָה);","strongs_def":"liquefaction, i.e. disappearance","kjv_def":"melt."},
"H8558":{"lemma":"תָּמָר","xlit":"tâmâr","pron":"taw-mawr'","derivation":"from an unused root meaning to be erect;","strongs_def":"a palm tree","kjv_def":"palm (tree)."},
"H8559":{"lemma":"תָּמָר","xlit":"Tâmâr","pron":"taw-mawr'","derivation":"the same as H8558 (תָּמָר);","strongs_def":"Tamar, the name of three women and a place","kjv_def":"Tamar."},
"H8560":{"lemma":"תֹּמֶר","xlit":"tômer","pron":"to'-mer","derivation":"from the same root as H8558 (תָּמָר);","strongs_def":"a palm trunk","kjv_def":"palm tree."},
"H8561":{"lemma":"תִּמֹּר","xlit":"timmôr","pron":"tim-more'","derivation":"(plural or (feminine) תִּמֹּרָה; (singular and plural), from the same root as H8558 (תָּמָר);","strongs_def":"(architectural) a palm-like pilaster (i.e. umbellate)","kjv_def":"palm tree."},
"H8562":{"lemma":"תַּמְרוּק","xlit":"tamrûwq","pron":"tam-rook'","derivation":"or תַּמְרֻק; or תַּמְרִיק; from H4838 (מָרַק);","strongs_def":"properly, a scouring, i.e. soap or perfumery forthe bath; figuratively, a detergent","kjv_def":"[idiom] cleanse, (thing for) purification(-fying)."},
"H8563":{"lemma":"תַּמְרוּר","xlit":"tamrûwr","pron":"tam-roor'","derivation":"from H4843 (מָרַר);","strongs_def":"bitterness (plural as collective)","kjv_def":"[idiom] most bitter(-ly)."},
"H8564":{"lemma":"תַּמְרוּר","xlit":"tamrûwr","pron":"tam-roor'","derivation":"from the same root as H8558 (תָּמָר);","strongs_def":"an erection, i.e. pillar (probably for a guide-board)","kjv_def":"high heap."},
"H8565":{"lemma":"תַּן","xlit":"tan","pron":"tan","derivation":"from an unused root probably meaning to elongate;","strongs_def":"a monster (as preternaturally formed), i.e. a sea-serpent (or other huge marine animal); also a jackal (or other hideous land animal)","kjv_def":"dragon, whale. Compare H8577 (תַּנִּין)."},
"H8566":{"lemma":"תָּנָה","xlit":"tânâh","pron":"taw-naw'","derivation":"a primitive root;","strongs_def":"to present (a mercenary inducement), i.e. bargain with (a harlot)","kjv_def":"hire."},
"H8567":{"lemma":"תָּנָה","xlit":"tânâh","pron":"taw-naw'","derivation":"a primitive root (identical with through the idea of attributing honor);","strongs_def":"to ascribe (praise), i.e. celebrate, commemorate","kjv_def":"lament, rehearse."},
"H8568":{"lemma":"תַּנָּה","xlit":"tannâh","pron":"tan-naw'","derivation":"probably feminine of H8565 (תַּן);","strongs_def":"a female jackal","kjv_def":"dragon."},
"H8569":{"lemma":"תְּנוּאָה","xlit":"tᵉnûwʼâh","pron":"ten-oo-aw'","derivation":"from H5106 (נוּא);","strongs_def":"alienation; by implication, enmity","kjv_def":"breach of promise, occasion."},
"H8570":{"lemma":"תְּנוּבָה","xlit":"tᵉnûwbâh","pron":"ten-oo-baw'","derivation":"from H5107 (נוּב);","strongs_def":"produce","kjv_def":"fruit, increase."},
"H8571":{"lemma":"תְּנוּךְ","xlit":"tᵉnûwk","pron":"ten-ook'","derivation":"perhaps from the same as H594 (אֲנָךְ) through the idea of protraction;","strongs_def":"a pinnacle, i.e. extremity","kjv_def":"tip."},
"H8572":{"lemma":"תְּנוּמָה","xlit":"tᵉnûwmâh","pron":"ten-oo-maw'","derivation":"from H5123 (נוּם);","strongs_def":"drowsiness, i.e. sleep","kjv_def":"slumber(-ing)."},
"H8573":{"lemma":"תְּנוּפָה","xlit":"tᵉnûwphâh","pron":"ten-oo-faw'","derivation":"from H5130 (נוּף);","strongs_def":"a brandishing (in threat); by implication, tumult; specifically, the official undulation of sacrificial offerings","kjv_def":"offering, shaking, wave (offering)."},
"H8574":{"lemma":"תַּנּוּר","xlit":"tannûwr","pron":"tan-noor'","derivation":"from H5216 (נִיר);","strongs_def":"a fire-pot","kjv_def":"furnace, oven."},
"H8575":{"lemma":"תַּנְחוּם","xlit":"tanchûwm","pron":"tan-khoom'","derivation":"or תַּנְחֻם; and (feminine) תַּנְחוּמָה; from H5162 (נָחַם);","strongs_def":"compassion, solace","kjv_def":"comfort, consolation."},
"H8576":{"lemma":"תַּנְחֻמֶת","xlit":"Tanchumeth","pron":"tan-khoo'-meth","derivation":"for H8575 (תַּנְחוּם) (feminine);","strongs_def":"Tanchumeth, an Israelite","kjv_def":"Tanhumeth."},
"H8577":{"lemma":"תַּנִּין","xlit":"tannîyn","pron":"tan-neen'","derivation":"or תַּנִּים; (Ezekiel 29:3), intensive from the same as H8565 (תַּן);","strongs_def":"a marine or land monster, i.e. sea-serpent or jackal","kjv_def":"dragon, sea-monster, serpent, whale."},
"H8578":{"lemma":"תִּנְיָן","xlit":"tinyân","pron":"tin-yawn'","derivation":"(Aramaic) corresponding to H8147 (שְׁנַיִם);","strongs_def":"second","kjv_def":"second."},
"H8579":{"lemma":"תִּנְיָנוּת","xlit":"tinyânûwth","pron":"tin-yaw-nooth'","derivation":"(Aramaic) from H8578 (תִּנְיָן);","strongs_def":"a second time","kjv_def":"again."},
"H8580":{"lemma":"תַּנְשֶׁמֶת","xlit":"tanshemeth","pron":"tan-sheh'-meth","derivation":"from H5395 (נָשַׁם);","strongs_def":"properly, a hard breather, i.e. the name of two unclean creatures, a lizard and abird (both perhaps from changing color through their irascibility), probably the tree-toad and the water-hen","kjv_def":"mole, swan."},
"H8581":{"lemma":"תַּעָב","xlit":"taʻâb","pron":"taw-ab'","derivation":"a primitive root;","strongs_def":"to loathe, i.e. (morally) detest","kjv_def":"(make to be) abhor(-red), (be, commit more, do) abominable(-y), [idiom] utterly."},
"H8582":{"lemma":"תָּעָה","xlit":"tâʻâh","pron":"taw-aw'","derivation":"a primitive root;","strongs_def":"to vacillate, i.e. reel or stray (literally or figuratively); also causative of both","kjv_def":"(cause to) go astray, deceive, dissemble, (cause to, make to) err, pant, seduce, (make to) stagger, (cause to) wander, be out of the way."},
"H8583":{"lemma":"תֹּעוּ","xlit":"Tôʻûw","pron":"to'-oo","derivation":"or תֹּעִי; from H8582 (תָּעָה); error","strongs_def":"Tou or Toi, a Syrian king","kjv_def":"Toi, Tou."},
"H8584":{"lemma":"תְּעוּדָה","xlit":"tᵉʻûwdâh","pron":"teh-oo-daw'","derivation":"from H5749 (עוּד);","strongs_def":"attestation, i.e. a precept, usage","kjv_def":"testimony."},
"H8585":{"lemma":"תְּעָלָה","xlit":"tᵉʻâlâh","pron":"teh-aw-law'","derivation":"from H5927 (עָלָה);","strongs_def":"a channel (into which water is raised for irrigation); also a bandage or plaster (as placed upon a wound)","kjv_def":"conduit, cured, healing, little river, trench, watercourse."},
"H8586":{"lemma":"תַּעֲלוּל","xlit":"taʻălûwl","pron":"tah-al-ool'","derivation":"from H5953 (עָלַל);","strongs_def":"caprice (as a fit coming on), i.e. vexation; concretely a tyrant","kjv_def":"babe, delusion."},
"H8587":{"lemma":"תַּעֲלֻמָּה","xlit":"taʻălummâh","pron":"tah-al-oom-maw'","derivation":"from H5956 (עָלַם);","strongs_def":"a secret","kjv_def":"thing that is hid, secret."},
"H8588":{"lemma":"תַּעֲנוּג","xlit":"taʻănûwg","pron":"tah-an-oog'","derivation":"or תַּעֲנֻג; and (feminine) תַּעֲנֻגָה; from H6026 (עָנַג);","strongs_def":"luxury","kjv_def":"delicate, delight, pleasant."},
"H8589":{"lemma":"תַּעֲנִית","xlit":"taʻănîyth","pron":"tah-an-eeth'","derivation":"from H6031 (עָנָה);","strongs_def":"affliction (of self), i.e. fasting","kjv_def":"heaviness."},
"H8590":{"lemma":"תַּעֲנָךְ","xlit":"Taʻănâk","pron":"tah-an-awk'","derivation":"or תַּעְנָךְ; of uncertain derivation;","strongs_def":"Taanak or Tanak, a place in Palestine","kjv_def":"Taanach, Tanach."},
"H8591":{"lemma":"תָּעַע","xlit":"tâʻaʻ","pron":"taw-ah'","derivation":"a primitive root;","strongs_def":"to cheat; by analogy, to maltreat","kjv_def":"deceive, misuse."},
"H8592":{"lemma":"תַּעֲצֻמָה","xlit":"taʻătsumâh","pron":"tah-ats-oo-maw'","derivation":"from H6105 (עָצַם);","strongs_def":"might (plural collective)","kjv_def":"power."},
"H8593":{"lemma":"תַּעַר","xlit":"taʻar","pron":"tah'-ar","derivation":"from H6168 (עָרָה);","strongs_def":"a knife or razor (as making bare); also a scabbard (as being bare, i.e. empty)","kjv_def":"(pen-) knife, razor, scabbard, shave, sheath."},
"H8594":{"lemma":"תַּעֲרֻבָה","xlit":"taʻărubâh","pron":"tah-ar-oo-baw'","derivation":"from H6148 (עָרַב);","strongs_def":"suretyship, i.e. (concretely) a pledge","kjv_def":"[phrase] hostage."},
"H8595":{"lemma":"תַּעְתֻּעַ","xlit":"taʻtuaʻ","pron":"tah-too'-ah","derivation":"from H8591 (תָּעַע);","strongs_def":"a fraud","kjv_def":"error."},
"H8596":{"lemma":"תֹּף","xlit":"tôph","pron":"tofe","derivation":"from H8608 (תָּפַף) contracted;","strongs_def":"a tambourine","kjv_def":"tabret, timbrel."},
"H8597":{"lemma":"תִּפְאָרָה","xlit":"tiphʼârâh","pron":"tif-aw-raw'","derivation":"or תִּפְאֶרֶת; from H6286 (פָּאַר);","strongs_def":"ornament (abstractly or concretely, literally or figuratively)","kjv_def":"beauty(-iful), bravery, comely, fair, glory(-ious), honour, majesty."},
"H8598":{"lemma":"תַּפּוּחַ","xlit":"tappûwach","pron":"tap-poo'-akh","derivation":"from H5301 (נָפַח);","strongs_def":"an apple (from its fragrance), i.e. the fruit or the tree (probably includ. others of the pome order, as the quince, the orange, etc.)","kjv_def":"apple (tree). See also H1054 (בֵּית תַּפּוּחַ)."},
"H8599":{"lemma":"תַּפּוּחַ","xlit":"Tappûwach","pron":"tap-poo'-akh","derivation":"the same as H8598 (תַּפּוּחַ);","strongs_def":"Tappuach, the name of two places in Palestine, also of an Israelite","kjv_def":"Tappuah."},
"H8600":{"lemma":"תְּפוֹצָה","xlit":"tᵉphôwtsâh","pron":"tef-o-tsaw'","derivation":"from H6327 (פּוּץ);","strongs_def":"a dispersal","kjv_def":"dispersion."},
"H8601":{"lemma":"תֻּפִין","xlit":"tuphîyn","pron":"too-feen'","derivation":"from H644 (אָפָה);","strongs_def":"cookery, i.e. (concretely) a cake","kjv_def":"baked piece."},
"H8602":{"lemma":"תָּפֵל","xlit":"tâphêl","pron":"taw-fale'","derivation":"from an unused root meaning to smear;","strongs_def":"plaster (as gummy) or slime; (figuratively) frivolity","kjv_def":"foolish things, unsavoury, untempered."},
"H8603":{"lemma":"תֹּפֶל","xlit":"Tôphel","pron":"to'-fel","derivation":"from the same as H8602 (תָּפֵל); quagmire;","strongs_def":"Tophel, a place near the Desert","kjv_def":"Tophel."},
"H8604":{"lemma":"תִּפְלָה","xlit":"tiphlâh","pron":"tif-law'","derivation":"from the same as H8602 (תָּפֵל);","strongs_def":"frivolity","kjv_def":"folly, foolishly."},
"H8605":{"lemma":"תְּפִלָּה","xlit":"tᵉphillâh","pron":"tef-il-law'","derivation":"from H6419 (פָּלַל);","strongs_def":"intercession, supplication; by implication, a hymn","kjv_def":"prayer."},
"H8606":{"lemma":"תִּפְלֶצֶת","xlit":"tiphletseth","pron":"tif-leh'-tseth","derivation":"from H6426 (פָּלַץ);","strongs_def":"fearfulness","kjv_def":"terrible."},
"H8607":{"lemma":"תִּפְסַח","xlit":"Tiphçach","pron":"tif-sakh'","derivation":"from H6452 (פָּסַח); ford;","strongs_def":"Tiphsach, a place in Mesopotamia","kjv_def":"Tipsah."},
"H8608":{"lemma":"תָּפַף","xlit":"tâphaph","pron":"taw-faf'","derivation":"a primitive root;","strongs_def":"to drum, i.e. play (as) on the tambourine","kjv_def":"taber, play with timbrels."},
"H8609":{"lemma":"תָּפַר","xlit":"tâphar","pron":"taw-far'","derivation":"a primitive root;","strongs_def":"to sew","kjv_def":"(women that) sew (together)."},
"H8610":{"lemma":"תָּפַשׂ","xlit":"tâphas","pron":"taw-fas'","derivation":"a primitive root;","strongs_def":"to manipulate, i.e. seize; chiefly to capture, wield, specifically, to overlay; figuratively, to use unwarrantably","kjv_def":"catch, handle, (lay, take) hold (on, over), stop, [idiom] surely, surprise, take."},
"H8611":{"lemma":"תֹּפֶת","xlit":"tôpheth","pron":"to'-feth","derivation":"from the base of H8608 (תָּפַף);","strongs_def":"a smiting, i.e. (figuratively) contempt","kjv_def":"tabret."},
"H8612":{"lemma":"תֹּפֶת","xlit":"Tôpheth","pron":"to'-feth","derivation":"the same as H8611 (תֹּפֶת);","strongs_def":"Topheth, a place near Jerusalem","kjv_def":"Tophet, Topheth."},
"H8613":{"lemma":"תׇּפְתֶּה","xlit":"Tophteh","pron":"tof-teh'","derivation":"probably a form of H8612 (תֹּפֶת);","strongs_def":"Tophteh, a place of cremation","kjv_def":"Tophet."},
"H8614":{"lemma":"תִּפְתַּי","xlit":"tiphtay","pron":"tif-tah'-ee","derivation":"(Aramaic) perhaps from H8199 (שָׁפַט);","strongs_def":"judicial, i.e. a lawyer","kjv_def":"sheriff."},
"H8615":{"lemma":"תִּקְוָה","xlit":"tiqvâh","pron":"tik-vaw'","derivation":"from H6960 (קָוָה); (compare H6961 (קָוֶה))","strongs_def":"literally a cord (as an attachment); figuratively, expectancy","kjv_def":"expectation(-ted), hope, live, thing that I long for."},
"H8616":{"lemma":"תִּקְוָה","xlit":"Tiqvâh","pron":"tik-vaw'","derivation":"the same as H8615 (תִּקְוָה);","strongs_def":"Tikvah, the name of two Israelites","kjv_def":"Tikvah."},
"H8617":{"lemma":"תְּקוּמָה","xlit":"tᵉqûwmâh","pron":"tek-oo-maw'","derivation":"from H6965 (קוּם);","strongs_def":"resistfulness","kjv_def":"power to stand."},
"H8618":{"lemma":"תְּקוֹמֵם","xlit":"tᵉqôwmêm","pron":"tek-o-mame'","derivation":"from H6965 (קוּם);","strongs_def":"an opponent","kjv_def":"rise up against."},
"H8619":{"lemma":"תָּקוֹעַ","xlit":"tâqôwaʻ","pron":"taw-ko'-ah","derivation":"from H8628 (תָּקַע) (in the musical sense);","strongs_def":"a trumpet","kjv_def":"trumpet."},
"H8620":{"lemma":"תְּקוֹעַ","xlit":"Tᵉqôwaʻ","pron":"tek-o'-ah","derivation":"a form of H8619 (תָּקוֹעַ);","strongs_def":"Tekoa, a place in Palestine","kjv_def":"Tekoa, Tekoah."},
"H8621":{"lemma":"תְּקוֹעִי","xlit":"Tᵉqôwʻîy","pron":"tek-o-ee'","derivation":"or תְּקֹעִי; patronymically from H8620 (תְּקוֹעַ);","strongs_def":"a Tekoite or inhabitant of Tekoah","kjv_def":"Tekoite."},
"H8622":{"lemma":"תְּקוּפָה","xlit":"tᵉqûwphâh","pron":"tek-oo-faw'","derivation":"or תְּקֻפָה; from H5362 (נָקַף);","strongs_def":"a revolution, i.e. (of the sun) course, (of time) lapse","kjv_def":"circuit, come about, end."},
"H8623":{"lemma":"תַּקִּיף","xlit":"taqqîyph","pron":"tak-keef'","derivation":"from H8630 (תָּקַף);","strongs_def":"powerful","kjv_def":"mightier."},
"H8624":{"lemma":"תַּקִּיף","xlit":"taqqîyph","pron":"tak-keef'","derivation":"(Aramaic) corresponding to H8623 (תַּקִּיף)","strongs_def":"{powerful}","kjv_def":"mighty, strong."},
"H8625":{"lemma":"תְּקַל","xlit":"tᵉqal","pron":"tek-al'","derivation":"(Aramaic) corresponding to H8254 (שָׁקַל);","strongs_def":"to balance","kjv_def":"Tekel, be weighed."},
"H8626":{"lemma":"תָּקַן","xlit":"tâqan","pron":"taw-kan'","derivation":"a primitive root;","strongs_def":"to equalize, i.e. straighten (intransitive or transitive); figuratively, to compose","kjv_def":"set in order, make straight."},
"H8627":{"lemma":"תְּקַן","xlit":"tᵉqan","pron":"tek-an'","derivation":"(Aramaic) corresponding to H8626 (תָּקַן);","strongs_def":"to straighten up, i.e. confirm","kjv_def":"establish."},
"H8628":{"lemma":"תָּקַע","xlit":"tâqaʻ","pron":"taw-kah'","derivation":"a primitive root;","strongs_def":"to clatter, i.e. slap (the hands together), clang (an instrument); by analogy, to drive (a nail or tent-pin, a dart, etc.); by implication, to become bondsman by handclasping)","kjv_def":"blow (a trumpet), cast, clap, fasten, pitch (tent), smite, sound, strike, [idiom] suretiship, thrust."},
"H8629":{"lemma":"תֵּקַע","xlit":"têqaʻ","pron":"tay-kah'","derivation":"from H8628 (תָּקַע);","strongs_def":"a blast of a trumpet","kjv_def":"sound."},
"H8630":{"lemma":"תָּקַף","xlit":"tâqaph","pron":"taw-kaf'","derivation":"a primitive root;","strongs_def":"to overpower","kjv_def":"prevail (against)."},
"H8631":{"lemma":"תְּקֵף","xlit":"tᵉqêph","pron":"tek-afe'","derivation":"(Aramaic) corresponding to H8630 (תָּקַף);","strongs_def":"to become (causatively, make) mighty or (figuratively) obstinate","kjv_def":"make firm, harden, be(-come) strong."},
"H8632":{"lemma":"תְּקֹף","xlit":"tᵉqôph","pron":"tek-ofe'","derivation":"(Aramaic) corresponding to H8633 (תֹּקֶף);","strongs_def":"power","kjv_def":"might, strength."},
"H8633":{"lemma":"תֹּקֶף","xlit":"tôqeph","pron":"to'-kef","derivation":"from H8630 (תָּקַף);","strongs_def":"might or (figuratively) positiveness","kjv_def":"authority, power, strength."},
"H8634":{"lemma":"תַּרְאֲלָה","xlit":"Tarʼălâh","pron":"tar-al-aw'","derivation":"probably for H8653 (תַּרְעֵלָה); a reeling;","strongs_def":"Taralah, a place in Palestine","kjv_def":"Taralah."},
"H8635":{"lemma":"תַּרְבּוּת","xlit":"tarbûwth","pron":"tar-booth'","derivation":"from H7235 (רָבָה);","strongs_def":"multiplication, i.e. progeny","kjv_def":"increase."},
"H8636":{"lemma":"תַּרְבִּית","xlit":"tarbîyth","pron":"tar-beeth'","derivation":"from H7235 (רָבָה);","strongs_def":"multiplication, i.e. percentage or bonus in addition to principal","kjv_def":"increase, unjust gain."},
"H8637":{"lemma":"תִּרְגַּל","xlit":"tirgal","pron":"teer-gal'","derivation":"a denominative from H7270 (רָגַל);","strongs_def":"to cause to walk","kjv_def":"teach to go."},
"H8638":{"lemma":"תִּרְגַּם","xlit":"tirgam","pron":"teer-gam'","derivation":"a denominative from H7275 (רָגַם) in the sense of throwing over;","strongs_def":"to transfer, i.e. translate","kjv_def":"interpret."},
"H8639":{"lemma":"תַּרְדֵּמָה","xlit":"tardêmâh","pron":"tar-day-maw'","derivation":"from H7290 (רָדַם);","strongs_def":"a lethargy or (by implication) trance","kjv_def":"deep sleep."},
"H8640":{"lemma":"תִּרְהָקָה","xlit":"Tirhâqâh","pron":"teer-haw'-kaw","derivation":"of foreign derivation;","strongs_def":"Tirhakah, a king of Kush","kjv_def":"Tirhakah."},
"H8641":{"lemma":"תְּרוּמָה","xlit":"tᵉrûwmâh","pron":"ter-oo-maw'","derivation":"or תְּרֻמָה; (Deuteronomy 12:11), from H7311 (רוּם);","strongs_def":"a present (as offered up), especially in sacrifice or as tribute","kjv_def":"gift, heave offering (shoulder), oblation, offered(-ing)."},
"H8642":{"lemma":"תְּרוּמִיָּה","xlit":"tᵉrûwmîyâh","pron":"ter-oo-mee-yaw'","derivation":"formed as H8641 (תְּרוּמָה);","strongs_def":"a sacrificial offering","kjv_def":"oblation."},
"H8643":{"lemma":"תְּרוּעָה","xlit":"tᵉrûwʻâh","pron":"ter-oo-aw'","derivation":"from H7321 (רוּעַ);","strongs_def":"clamor, i.e. acclamation of joy or a battle-cry; especially clangorof trumpets, as an alarum","kjv_def":"alarm, blow(-ing) (of, the) (trumpets), joy, jubile, loud noise, rejoicing, shout(-ing), (high, joyful) sound(-ing)."},
"H8644":{"lemma":"תְּרוּפָה","xlit":"tᵉrûwphâh","pron":"ter-oo-faw'","derivation":"from H7322 (רוּף) in the sense of its congener H7495 (רָפָא);","strongs_def":"a remedy","kjv_def":"medicine."},
"H8645":{"lemma":"תִּרְזָה","xlit":"tirzâh","pron":"teer-zaw'","derivation":"probably from H7329 (רָזָה);","strongs_def":"a species of tree (apparently from its slenderness), perhaps the cypress","kjv_def":"cypress."},
"H8646":{"lemma":"תֶּרַח","xlit":"Terach","pron":"teh'-rakh","derivation":"of uncertain derivation;","strongs_def":"Terach, the father of Abraham; also a place in the Desert","kjv_def":"Tarah, Terah."},
"H8647":{"lemma":"תִּרְחֲנָה","xlit":"Tirchănâh","pron":"teer-khan-aw'","derivation":"of uncertain derivation;","strongs_def":"Tirchanah, an Israelite","kjv_def":"Tirhanah."},
"H8648":{"lemma":"תְּרֵין","xlit":"tᵉrêyn","pron":"ter-ane'","derivation":"(Aramaic) feminine תַּרְתֵּין; corresponding to H8147 (שְׁנַיִם);","strongs_def":"two","kjv_def":"second, [phrase] twelve, two."},
"H8649":{"lemma":"תׇּרְמָה","xlit":"tormâh","pron":"tor-maw'","derivation":"and תַּרְמוּת; or תַּרְמִית; from H7411 (רָמָה);","strongs_def":"fraud","kjv_def":"deceit(-ful), privily."},
"H8650":{"lemma":"תֹּרֶן","xlit":"tôren","pron":"to'-ren","derivation":"probably for H766 (אֹרֶן);","strongs_def":"a pole (as a mast or flagstaff)","kjv_def":"beacon, mast."},
"H8651":{"lemma":"תְּרַע","xlit":"tᵉraʻ","pron":"ter-ah'","derivation":"(Aramaic) corresponding to H8179 (שַׁעַר);","strongs_def":"a door; by implication, a palace","kjv_def":"gate mouth."},
"H8652":{"lemma":"תָּרָע","xlit":"târâʻ","pron":"taw-raw'","derivation":"(Aramaic) from H8651 (תְּרַע);","strongs_def":"a doorkeeper","kjv_def":"porter."},
"H8653":{"lemma":"תַּרְעֵלָה","xlit":"tarʻêlâh","pron":"tar-ay-law'","derivation":"from H7477 (רָעַל);","strongs_def":"reeling","kjv_def":"astonishment, trembling."},
"H8654":{"lemma":"תִּרְעָתִי","xlit":"Tirʻâthîy","pron":"teer-aw-thee'","derivation":"patrial from an unused name meaning gate;","strongs_def":"a Tirathite or inhabitant of an unknown Tirah","kjv_def":"Tirathite."},
"H8655":{"lemma":"תְּרָפִים","xlit":"tᵉrâphîym","pron":"ter-aw-feme'","derivation":"plural from H7495 (רָפָא); a healer;","strongs_def":"Teraphim (singular or plural) a family idol","kjv_def":"idols(-atry), images, teraphim."},
"H8656":{"lemma":"תִּרְצָה","xlit":"Tirtsâh","pron":"teer-tsaw'","derivation":"from H7521 (רָצָה); delightsomeness;","strongs_def":"Tirtsah, a place in Palestine; also an Israelitess","kjv_def":"Tirzah."},
"H8657":{"lemma":"תֶּרֶשׁ","xlit":"Teresh","pron":"teh'-resh","derivation":"of foreign derivation;","strongs_def":"Teresh, a eunuch of Xerxes","kjv_def":"Teresh."},
"H8658":{"lemma":"תַּרְשִׁישׁ","xlit":"tarshîysh","pron":"tar-sheesh'","derivation":"probably of foreign derivation (compare H8659 (תַּרְשִׁישׁ));","strongs_def":"a gem, perhaps the topaz","kjv_def":"beryl."},
"H8659":{"lemma":"תַּרְשִׁישׁ","xlit":"Tarshîysh","pron":"tar-sheesh'","derivation":"probably the same as H8658 (תַּרְשִׁישׁ) (as the region of the stone, or the reverse);","strongs_def":"Tarshish, a place on the Mediterranean, hence, the ephithet of a merchant vessel (as if for or from that port); also the name of a Persian and of an Israelite","kjv_def":"Tarshish, Tharshish."},
"H8660":{"lemma":"תִּרְשָׁתָא","xlit":"Tirshâthâʼ","pron":"teer-shaw-thaw'","derivation":"of foreign derivation;","strongs_def":"Tirshatha, the title of a Persian deputy or governor","kjv_def":"Tirshatha."},
"H8661":{"lemma":"תַּרְתָּן","xlit":"Tartân","pron":"tar-tawn'","derivation":"of foreign derivation;","strongs_def":"Tartan, an Assyrian","kjv_def":"Tartan."},
"H8662":{"lemma":"תַּרְתָּק","xlit":"Tartâq","pron":"tar-tawk'","derivation":"of foreign derivation;","strongs_def":"Tartak, a deity of the Avvites","kjv_def":"Tartak."},
"H8663":{"lemma":"תְּשֻׁאָה","xlit":"tᵉshuʼâh","pron":"tesh-oo-aw'","derivation":"from H7722 (שׁוֹא);","strongs_def":"a crashing or loud clamor","kjv_def":"crying, noise, shouting, stir."},
"H8664":{"lemma":"תִּשְׁבִּי","xlit":"Tishbîy","pron":"tish-bee'","derivation":"patrial from an unused name meaning recourse;","strongs_def":"a Tishbite or inhabitant of Tishbeh (in Gilead)","kjv_def":"Tishbite."},
"H8665":{"lemma":"תַּשְׁבֵּץ","xlit":"tashbêts","pron":"tash-bates'","derivation":"from H7660 (שָׁבַץ);","strongs_def":"checkered stuff (as reticulated)","kjv_def":"broidered."},
"H8666":{"lemma":"תְּשׁוּבָה","xlit":"tᵉshûwbâh","pron":"tesh-oo-baw'","derivation":"or תְּשֻׁבָה; from H7725 (שׁוּב);","strongs_def":"a recurrence (of time or place); a reply (as returned)","kjv_def":"answer, be expired, return."},
"H8667":{"lemma":"תְּשׂוּמֶת","xlit":"tᵉsûwmeth","pron":"tes-oo-meth'","derivation":"from H7760 (שׂוּם);","strongs_def":"a deposit, i.e. pledging","kjv_def":"[phrase] fellowship."},
"H8668":{"lemma":"תְּשׁוּעָה","xlit":"tᵉshûwʻâh","pron":"tesh-oo-aw'","derivation":"or תְּשֻׁעָה; from H7768 (שָׁוַע) in the sense of H3467 (יָשַׁע);","strongs_def":"rescue (literal or figurative, persons, national or spiritual)","kjv_def":"deliverance, help, safety, salvation, victory."},
"H8669":{"lemma":"תְּשׁוּקָה","xlit":"tᵉshûwqâh","pron":"tesh-oo-kaw'","derivation":"from H7783 (שׁוּק) in the original sense of stretching out after;","strongs_def":"a longing","kjv_def":"desire."},
"H8670":{"lemma":"תְּשׁוּרָה","xlit":"tᵉshûwrâh","pron":"tesh-oo-raw'","derivation":"from H7788 (שׁוּר) in the sense of arrival;","strongs_def":"a gift","kjv_def":"present."},
"H8671":{"lemma":"תְּשִׁיעִי","xlit":"tᵉshîyʻîy","pron":"tesh-ee-ee'","derivation":"ordinal from H8672 (תֵּשַׁע);","strongs_def":"ninth","kjv_def":"ninth."},
"H8672":{"lemma":"תֵּשַׁע","xlit":"têshaʻ","pron":"tay'-shah","derivation":"or (masculine) תִּשְׁעָה; perhaps from H8159 (שָׁעָה) through the idea of a turn to the next or full number ten;","strongs_def":"nine or (ordinal) ninth","kjv_def":"nine ([phrase] -teen, [phrase] -teenth, -th)."},
"H8673":{"lemma":"תִּשְׁעִים","xlit":"tishʻîym","pron":"tish-eem'","derivation":"multiple from H8672 (תֵּשַׁע);","strongs_def":"ninety","kjv_def":"ninety."},
"H8674":{"lemma":"תַּתְּנַי","xlit":"Tattᵉnay","pron":"tat-ten-ah'-ee","derivation":"of foreign derivation;","strongs_def":"Tattenai, a Persian","kjv_def":"Tatnai."}
}